/* compile with nauty.o nautil.o naugraph.o tree.o
* gcc testing0.c internal_functions1.o interface1.o graph_tree2.o nauty.o nautil.o naugraph.o -o testing0.out
* work with static memory allocation for the labels etc.
*/

/**
note: number of nodes and number of colours is restricted to MAXN
*/

#include<time.h>
//includes nauty.h
#include "interface.h"


/*
start_graph(6);
add_edge(0,3);
add_edge(1,4);
add_edge(1,2);
....
// alle Knoten haben Farbe 0 als Default

set_colour_of_node(0,1); // Knoten 0 bekommt Farbe 1
set_colour_of_node(2,1);
set_colour_of_node(4,2);

*/

// #define TRACE_VERSION TRUE

void prob_init(); 

void prob_get_number_of_colours(long num_of_col);

void prob_start_graph(long number_of_nodes);

void prob_add_edge(long from, long to);

void prob_set_colour_of_node(long node, long colour);

long prob_exists_graph(void);

void prob_free_storage(void);

graph *global_g;
graph *global_canong; 
struct Node *store_tree; //initialize with NULL pointer (needed for insert()) 
int tree_node_counter;

int global_n;
int global_m;

int global_number_of_colours;

int *global_cell_sizes;
int *global_is_set; //to mark which node has got a colour

int *cell_list;

/* variables to check the program flow */
short is_storage_free = TRUE;
short is_init = FALSE;
short is_set_num_colour = FALSE;
short is_start_graph = FALSE;

/* file to print the function calls */
FILE *pfile;

/* time nauty needs*/
#ifdef TIME_NAUTY
clock_t nauty_time; // time Nauty needs in total
clock_t max_time; // maximal time nauty needs for one graph
int num_graphs_max_time; // number of graphs that nauty takes 'max_time' for
#endif

#ifdef TIME_INTERFACE
clock_t start_time;
clock_t end_time;
clock_t total_time;
#endif

#ifdef GRAPH_COUNT_VERSION
   int state_counter;
   int *num_of_nodes_of_stategraphs;
   int *num_of_edges_of_stategraphs;
#endif

//int state_counter;
//int node_counter;


void prob_init(void){ 
	#ifdef TIME_INTERFACE
		total_time = 0;
		start_time =   clock();
	#endif

	if(DEBUG_VERSION){
		printf("in prob_init: \n");
	}
	/* start with new state space only if old states (graphs) are removed  && is_set_num_colour*/
	if(is_storage_free){

		#ifdef TRACE_VERSION
			printf("prob_init(); \n");
			/* write file new when new problem started */
			pfile = fopen("scheduler0_test1.c", "w");
			if(!pfile){
				printf("Can't open file. Function call is not written in file \n");
			}
			else{
				fprintf(pfile, "prob_init();\n");
				fclose(pfile);
			}
			
		#endif


		/* initialisation of all global variables */
		global_g = NULL;
		global_canong = NULL;
		global_n = 0;
		global_m = 0;
		//global_number_of_colours = 1; set extra with prob_set_number of colours
		cell_list = NULL;
		global_is_set = NULL; 

		store_tree = NULL; //initialize with NULL pointer (needed for insert())
		tree_node_counter = 1; // start counting number of stored graphs with 1
		#ifdef TIME_NAUTY
		nauty_time = 0; //set time that nauty needs to 0 for a new problem
		max_time = 0; 
		num_graphs_max_time = 0; //
		#endif
		
		#ifdef GRAPH_COUNT_VERSION
  			state_counter = 0;
 		  	num_of_nodes_of_stategraphs = malloc(sizeof(int) * MAXN+1);
			/* store number of nodes in [i] and in [MAXN+1+i] the sum of number of edges of graphs that number of nodes */
			num_of_edges_of_stategraphs = malloc(sizeof(int) * (MAXN+1) *2);
			int index;
			for(index = 0; index<=MAXN; index++){
				num_of_nodes_of_stategraphs[index] = 0; //initialize with 0
				num_of_edges_of_stategraphs[index] = 0;
				num_of_edges_of_stategraphs[MAXN+1+index] = 0;
			}
			//number of nodes in a graph can be between 0 and MAXN 
		#endif

		/*size must be at least n, the maximum number of cells
 		the number of cells of each graph is not used here as array size but MAXN,
		since the memory for global_cell_sizes is only once allocated
  		if this is changed in furure versions, also functions that use the size of 
  		the array global_cell_sizes must be changed
		*/
		global_cell_sizes = malloc(sizeof(int) * MAXN);
		if(global_cell_sizes){
			//size of each cell is 0 
			init_cell_sizes(global_cell_sizes, MAXN);
		}
		else{
			printf("Error in prob_init: memory could not be allocated! \n");
			return;
		}

		is_init = TRUE; //successful initialisation
		is_storage_free = FALSE;
	}
	else{
		printf("ERROR in prob_init: old storage space needs to be freed, before initialisation!\n ");
		printf("run function prob_free_storage and prob_get_number_of_colours first \n");
	}

	#ifdef TIME_INTERFACE
		end_time =   clock();
		total_time = total_time + (end_time - start_time);
	#endif
}


/* call from ProB and get the number of number of colours 
 * if parameter <=0 or > MAXN then prints error message and returns
*/
void prob_get_number_of_colours(long num_of_col){
	#ifdef TIME_INTERFACE
		start_time =   clock();
	#endif
	if(DEBUG_VERSION){
		printf("in prob_get_number_of_colours: \n");
	}
	
	#ifdef TRACE_VERSION
		printf("prob_get_number_of_colours(%ld); \n", num_of_col);
		/* append next line in file */
		pfile = fopen("scheduler0_test1.c", "a");
		if(!pfile){
			printf("Can't open file. Function call is not written in file \n");
		}
		else{
			fprintf(pfile, "prob_get_number_of_colours(%ld); \n", num_of_col);
			fclose(pfile);
		}
	#endif
	
	global_number_of_colours  = get_number_of_colours(num_of_col);
	if(global_number_of_colours > MAXN){
		printf("Error: number of colours is too big!");
		printf("Rerun prob_get_number_of_colours with a number <= %d!", MAXN);
		return; // number of colours not set
	}

	is_set_num_colour = TRUE;

	#ifdef TIME_INTERFACE
		end_time =   clock();
		total_time = total_time + (end_time - start_time);
	#endif
}

void prob_start_graph(long number_of_nodes){
	#ifdef TIME_INTERFACE
		start_time =   clock();
	#endif

	if(DEBUG_VERSION){
		printf("in prob_start_graph: \n");
	}

	/* all variables need to be initialized before use*/
	if(is_init && is_set_num_colour){
		#ifdef GRAPH_COUNT_VERSION
  			state_counter++;
 		  	num_of_nodes_of_stategraphs[number_of_nodes]++;
			num_of_edges_of_stategraphs[number_of_nodes]++;
			//number of graphs with "number_of_nodes" nodes is increased by 1
		#endif
		
		#ifdef TRACE_VERSION
			printf(" \nprob_start_graph(%d); \n", number_of_nodes);
			/* append next line in file */
			pfile = fopen("scheduler0_test1.c", "a");
			if(!pfile){
				printf("Can't open file. Function call is not written in file \n");
			}
			else{
				fprintf(pfile, " \nprob_start_graph(%d); \n", number_of_nodes);
				fclose(pfile);
			}
		#endif

		/* if MAXN is not set by nauty then n can be up to INT_MAX*/
		if( (MAXN>0 && number_of_nodes<=MAXN) || (MAXN == 0 && number_of_nodes<=INT_MAX) ) {
			global_n = number_of_nodes;
			global_m = (global_n + WORDSIZE - 1) / WORDSIZE;
	
			//free global_g since it may be used before (NULL if not)
			free(global_g);
			global_g = start_graph(global_n);

			// allocate memory for canong like for g as is will contain an ismorpic graph to g
			free(global_canong);
			global_canong = start_graph(global_n); 

			//size of each cell is 0 
			init_cell_sizes(global_cell_sizes, MAXN);

			// initialize cell_list for each graph new -> free cell_list
			//each cell can contain max. n nodes, which are nat. numbers
			free(cell_list);
			cell_list = malloc( sizeof(int) * global_number_of_colours  * global_n );

			if(cell_list){
				init_cell(global_n, global_number_of_colours, cell_list);
			}
			else{
				printf("Error in prob_start_graph: no memory allocated for cell_list, ");
				printf("run prob_get_number_of_colours and then prob_init \n");
			}

			free(global_is_set);
			global_is_set = malloc( (sizeof(int)) * global_m);
	
			int i=0;

			/* all nodes have no colour yet */
			for(i=0; i<global_m; i++){
				global_is_set[i]=0;
			}
			/* entries that exceed the number of nodes are set to 1 */
			for(i=global_n; i<global_m * WORDSIZE; i++){
				set_node(i, global_is_set);
			}

			is_start_graph = TRUE; //successful memory allocation for graph
		}
		else{
			printf("Error in prob_start_graph: number_of_nodes (%ld) is too big\n", number_of_nodes);
		}
	}
	else{
		printf("ERROR in prob_start_graph: storage space needs to be initialised before use!\n");
		printf("and number of colours has to be set run function prob_init and ");
		printf(" prob_get_number_of_colours first \n");
	}
	#ifdef TIME_INTERFACE
		end_time =   clock();
		total_time = total_time + (end_time - start_time);
	#endif
}

void prob_add_edge(long from, long to){
	#ifdef TIME_INTERFACE
		start_time =   clock();
	#endif
	
	// memory for graph needs be allocated before an edge can be added 
	if(is_start_graph){
		add_edge(from, to, global_g, global_n); //global_n= number_of_nodes
		
		#ifdef GRAPH_COUNT_VERSION
			num_of_edges_of_stategraphs[MAXN+1+global_n]++;
			//number of edges of graphs  with "number_of_nodes" nodes is increased by 1
		#endif

		#ifdef TRACE_VERSION
			printf("prob_add_edge(%d, %d); \n", from, to);
			// append next line in file
			pfile = fopen("scheduler0_test1.c", "a");
			if(!pfile){
				printf("Can't open file. Function call is not written in file \n");
			}
			else{
				fprintf(pfile, "prob_add_edge(%d, %d); \n", from, to);
				fclose(pfile);
			}
		#endif
	}
	else{
		printf("ERROR in prob_add_edge: memory for graph needs be allocated before an edge can be added!\n");
		printf("run function prob_start_graph first\n");
	}
	#ifdef TIME_INTERFACE
		end_time =   clock();
		total_time = total_time + (end_time - start_time);
	#endif
}



void prob_set_colour_of_node(long node, long colour){
	#ifdef TIME_INTERFACE
		start_time =   clock();
	#endif

	if(is_start_graph && is_set_num_colour ){
		if(colour < global_number_of_colours){
			set_colour_of_node(global_n, colour, node, cell_list, global_cell_sizes);
			
			#ifdef TRACE_VERSION
				printf("prob_set_colour_of_node(%d, %d); \n", node, colour);
				/* append next line in file */
				pfile = fopen("scheduler0_test1.c", "a");
				if(!pfile){
					printf("Can't open file. Function call is not written in file \n");
				}
				else{
					fprintf(pfile, "prob_set_colour_of_node(%d, %d); \n", node, colour);
					fclose(pfile);
				}
			#endif
		}
		else{
			printf("Error in prob_set_colour_of_node: number for colour %ld is too big!\n", colour);
			printf("Number for colour must be < %d \n", global_number_of_colours);
		}
	}
	else{
		printf("ERROR in prob_set_colour_of_node: graph does not exist or number of colours is not set!\n");
		printf("run function prob_start_graph and  prob_set_number_of_colours first \n");
	}
	#ifdef TIME_INTERFACE
		end_time =   clock();
		total_time = total_time + (end_time - start_time);
	#endif
}


long prob_exists_graph(void){ 
	#ifdef TIME_INTERFACE
		start_time =   clock();
	#endif
	
	if(DEBUG_VERSION){
		printf("in prob_exists_graph: Anfang\n");
	}

	if(is_start_graph){

		#ifdef TRACE_VERSION
			printf("prob_exists_graph(); \n");
			/* append next line in file */
			pfile = fopen("scheduler0_test1.c", "a");
			if(!pfile){
				printf("Can't open file. Function call is not written in file \n");
			}
			else{
				fprintf(pfile, "prob_exists_graph();\n");
				fclose(pfile);
			}
		#endif
		

		long exists = 0; // set to error value 

 		/* The nauty parameter m is a value such that an array of 
  	    	 * m setwords is sufficient to hold n bits.  The type setword
  	    	 * is defined in nauty.h.  The number of bits in a setword is
   	   	 * WORDSIZE, which is 16, 32 or 64.  Here we calculate
   	   	 * m = ceiling(n/WORDSIZE) */
		int m = (global_n + WORDSIZE - 1) / WORDSIZE;

		/* lab holds the initial order of the nodes and ptn holds where a cell ends */
		int lab[MAXN], ptn[MAXN], orbits[MAXN];

	
		static DEFAULTOPTIONS_GRAPH(options);
		statsblk stats;
    		setword workspace[100*MAXM]; //need more workspace for better speed

		/* Default options are set by the DEFAULTOPTIONS_GRAPH macro above.
 		 *  Here we change those options that we want to be different from the
 		 *  defaults.  writeautoms=TRUE causes automorphisms to be written.    */
		options.getcanon = TRUE;
		options.digraph = TRUE;
		//options.defaultptn = FALSE; //set initial partition manually

		if(DEBUG_VERSION){
			printf("global_number_of_colours  %d\n", global_number_of_colours );
		}
		if(global_number_of_colours  > 1){
			/* variablea lab and ptn need to be set maunually */
			options.defaultptn = FALSE;

			/* check if every node has a colour;  give colour 0 if not */
			
			int i = 0;
			while(i<m){
				if(global_is_set[i] != -1){ //-1 = 111... in binary
					printf("not all nodes have a colour set!\n i=%d, global_is_set[i]=%d \n",i,global_is_set[i]);
					printf("give colour 0 to nodes without colour! \n");
					int k= 0;
					for(k=0; k<global_n; k++){ 
			has_colour(k,cell_list, global_cell_sizes, global_n, global_number_of_colours );
					}
					break; //while
				}
				i++;
			}
			
			int k= 0;
			/*
			for(k=0; k<global_n; k++){ 
			has_colour(k,cell_list, global_cell_sizes, global_n, global_number_of_colours );
			}
			*/		

			/* set nautys variables lab and ptn */
			set_label(lab, ptn, cell_list, global_n, global_number_of_colours );
		}else if(global_number_of_colours  == 1){
			/* partition with unit partition, i.e. one colour don't need to set lab and ptn */
			options.defaultptn = TRUE;
			/* need  to set the variable global_cell_sizes */
			global_cell_sizes[0] = global_n; // all nodes have colour 0
		}else{
			printf("Error: the number of cells in the initial partition is not set \n");
		}

		if(DEBUG_VERSION){ 
			int i;
			printf("lab: ");
			for(i=0; i< global_n; i++){
				printf("%d ",lab[i]);
			}
			printf("\n");
			for(i=0; i< global_n; i++){
				printf("%d ",ptn[i]);
			}
			printf("\n");
		}
		/* The following optional call verifies that we are linking
      		 * to compatible versions of the nauty routines.          */
		 nauty_check(WORDSIZE,m,global_n,NAUTYVERSIONID);

		if(DEBUG_VERSION){ 
			printf("in prob_exists_graph: vor Nauty, %d \n", MAXM);
		}
		#ifdef TIME_NAUTY
			clock_t time1 = clock();
		#endif
		nauty(global_g,lab,ptn,NULL,orbits,&options,&stats, workspace,100*MAXM,m,global_n, global_canong);

		if(DEBUG_VERSION){ 
			printf("in prob_exists_graph: direkt nach Nauty\n");
		}

		#ifdef TIME_NAUTY
		clock_t time2 = clock();
		nauty_time = nauty_time + (time2-time1);

		//calculate maximum of runtime for one graph
		if(max_time < (time2- time1)){
			max_time =  time2-time1; //found new maximum
			num_graphs_max_time = 1; //one graph with that maximum so far
		}
		else{
			max_time = max_time;
			num_graphs_max_time++; //another graph takes 'max_time'
		}
		//max_time = ( max_time < (time2-time1) ? time2-time1 : max_time ); 
		#endif

		if(DEBUG_VERSION){
			printf("in prob_exists_graph: nach Nauty\n");
		}

		/* insert canonical form of graph in tree, if canonical form  is not in the tree yet
   		 *returns 0 if canonical form is already in tree, 1 if it is inserted, -1 otherwise
		 * variable global_cell_sizes has been set with set_colour_of_node  or manually if there is 
		 * only one cell.
		*/	
		exists = insert_canon(global_canong, global_cell_sizes, &store_tree, global_n);

		if(DEBUG_VERSION){
			printf("in prob_exists_graph: Ende\n");
		}
		
		#ifdef TRACE_VERSION
		printf("/* prob_exists_graph() --> %d */\n",exists);
		#endif

		#ifdef TIME_INTERFACE
			end_time =   clock();
			total_time = total_time + (end_time - start_time);
		#endif
		return exists;
	}
	else{
		printf("ERROR in prob_exists_graph: memory for graph needs be allocated before existence can be !\n");
		printf("checked! Run function prob_start_graph first \n");
	}

	return -1; //should return before here
}


/* free all allocated memory and set pointers to NULL */
void prob_free_storage(void){
	#ifdef TIME_INTERFACE
		start_time =   clock();
	#endif

	if(DEBUG_VERSION){
		printf("in prob_free_storage(); \n");
	}
	if(!is_storage_free){
		
		#ifdef TRACE_VERSION
			printf("prob_free_storage(); \n");
			/* append next line in file */
			pfile = fopen("scheduler0_test1.c", "a");
			if(!pfile){
				printf("Can't open file. Function call is not written in file \n");
			}
			else{
				fprintf(pfile, "prob_free_storage();\n");
				fclose(pfile);
			}
		#endif

		//print time nauty uses and max of time for one graph
		#ifdef TIME_NAUTY
			printf("\n");
			printf("max_time in ns: %.2f\n", (double) max_time*1000000/ (double) CLOCKS_PER_SEC);
			printf("number of graphs with max_time: %d \n", num_graphs_max_time);
			printf("nauty_time in ms: %.2f\n", (double) nauty_time*1000/  (double) CLOCKS_PER_SEC);
		#endif
		// print number of graphs with however many nodes before feeing storage 
		#ifdef GRAPH_COUNT_VERSION
			printf("state_counter %d \n", state_counter);
			int index = 0;
			for(index = 0; index<=MAXN; index++){
				/*
				if(num_of_nodes_of_stategraphs[index] != 0){
					printf("%d is number of nodes of %d graphs \n", index, num_of_nodes_of_stategraphs[index]);
				}*/
				if(num_of_edges_of_stategraphs[index] != 0  && num_of_edges_of_stategraphs[MAXN+1+index] == 0) {
					printf("Error occured when counting nodes or edges \n");
					printf("This error should not affect results of graph isomorphism \n");
				}
				if(num_of_edges_of_stategraphs[index] != 0 && num_of_edges_of_stategraphs[MAXN+1+index] != 0){
					printf("%d is number of nodes of %d graphs ", index,num_of_edges_of_stategraphs[index]);
					printf(" with in total %d edges \n", num_of_edges_of_stategraphs[MAXN+1+index]);
printf("or approx. %.2f edges per graph\n \n", (float) num_of_edges_of_stategraphs[MAXN+1+index]/(float)num_of_edges_of_stategraphs[index]);
				}
			}
		#endif

		if(DEBUG_VERSION){
			printf("free: tree \n"); 
		}

		free_tree(&store_tree);

		if(DEBUG_VERSION){
			printf("free: global_g: \n");
		}	

		free(global_g);
		global_g = NULL;	

		free(global_canong);
		global_canong = NULL;
	
		free(global_cell_sizes);
		global_cell_sizes = NULL;

		free(cell_list);
		cell_list = NULL;
		
		free(global_is_set);
		global_is_set = NULL; 

		is_storage_free = TRUE;
		//need to initialise variables again for next problem
		is_init = FALSE; 
		is_set_num_colour = FALSE;
		is_start_graph = FALSE;
	}
	else{
		printf("storage is free, initialize new with prob_init()\n");
	}
	#ifdef TIME_INTERFACE
		end_time =   clock();
		total_time = total_time + (end_time - start_time);
		printf("interface_time in ms: %.2f\n", (double) total_time*1000/ (double) CLOCKS_PER_SEC);
	#endif

}


