/**
bugfix: in function compare the variable k wasn't initialized, all previous versions
may contain this bug!

between two graphs the number of nodes with the same colour were not compared
all previous versions may contain this bug!
*/

//includes nauty.h, graph_tree.h,

#include "interface.h"

int check_number_of_nodes = 0;
int check_number_of_colours = 0;
int check_m = 0;

int *global_is_set; //to mark which node has got a colour

/* takes the number of nodes and allocates the memory for a graph with that number of nodes
  initializes graph with the empty graph 
*/
graph* start_graph(int number_of_nodes){

	/* m-number of unsigned ints to hold number_of_nodes bits*/
	int m = (number_of_nodes + WORDSIZE - 1) / WORDSIZE;

	/* When interface gets a new graph as input, the value of 
	 * the global variables check_number_of_nodes and check_m need to be changed 
	 * as the number of nodes, and therefor m, of the new graph may be different */
	check_number_of_nodes = number_of_nodes;
	check_m = m;

	graph *temp = malloc( (sizeof(graph)) * number_of_nodes * m);

	/* return NULL-pointer and error message if no memory is allocated */
	if(!temp){
		printf("in start_graph: no memory to allocate");
		return NULL;
	}

	/* initialise graph with empty graph with n nodes*/
	initGraph(temp, number_of_nodes);

	return temp; //malloc( (sizeof(graph)) * number_of_nodes * m);
}


/* initialize the adgjacency matrix of the graph with zeros
  used in start_graph()
*/
void initGraph(graph *g, int n){

	int m,v;
    	set *gv;

	m = (n + WORDSIZE - 1) / WORDSIZE;

	for (v = 0; v < n; ++v) 
        {
            gv = GRAPHROW(g,v,m);
            EMPTYSET(gv,m);
        }
}

int add_edge(int origin, int dest, graph *g, int n){

	int m = (n + WORDSIZE - 1) / WORDSIZE;

	/*origin and dest are two nodes of the graph, therefore they 
	  must be a number between 0 and n-1 */
	if(origin<0 || dest <0 || origin>=n || dest>=n){
 		printf("error: origin or dest are not between 0 and %d \n", n-1);
		return -1;
	}
	else{

		set *gv;
		gv = GRAPHROW(g,origin,m);

        	ADDELEMENT(gv, dest%n);
	}
	return 0;
}

/* compares two canonical forms in packed form, (adjacency matrix stored as an array of integers)
 * (see nauty users guide) 
 * need n to finish loop; n must be greater than 0;
 * assumes that both graphs have the same number of nodes, ie. canonical forms have the same length
 * returns 0 if graphs are isomorphic
 * 1 if first canonical label (the label of a new graph) is bigger than the second one (stored graph)
 * -1 if first canonical label (the label of a new graph) is smaller than the stored one (stored graph)
*/
int compare(graph *canon1, graph *canon2, int n){

	//compare status; 0 for equal, 1 for canon1 > canon2, 2 for canon1 < canon2
	int comp = ERROR_VALUE; 

	if(canon1 && canon2){
		int m = (n + WORDSIZE - 1) / WORDSIZE; 
		int k = 0;

		while(k < n*m){ // for every node n g has m sets that need to be considered as one row

			int l=0;
			for(l=0; l<m; l++){ //consider one row
				if(canon1[k+l] == canon2[k+l]){ //consider one setword
					comp = 0;
				}
				else{
					if(canon1[k+l] > canon2[k+l]){
						if(DEBUG_VERSION){
							printf("not isomorphic 1>2\n");
						}
						comp = 1;
					}
					else{
						if(DEBUG_VERSION){
							printf("not isomorphic 2>1\n");
						}
						comp = -1;
					}
					/* return after first difference in canonical labels is found */
					return comp; 
				}
			}	
			k=k+m;
		}
		if(DEBUG_VERSION){
			printf("Isomorphic \n");	
		}
	}
	else{
		printf("error in compare: one parameter is Nullpointer \n");
		return ERROR_VALUE; //-20
	}
	return comp; // 0 if no error occured, like n*m <= 0
}

/* functions to calculate the partition */

/* call from ProB and get the number of colours 
 *  returns -1 if parameter <=0 and error message
*/
int get_number_of_colours(int number_of_colours){
	if(number_of_colours <= 0){
		printf("Error in Function get_number_of_colours: \n") ;
		printf("parameter number_of_colours = %d must be > 0 \n", number_of_colours);
		return -1;
	}
	else{
		/* When interface gets a new graph as input, the value of 
		  the global variable check_number_of_colours needs to be changed 
		  as the number of colours of the new graph may be different */
		check_number_of_colours = number_of_colours; 
		return number_of_colours;
	}
	
}

/* initialize the list wit the cells with invalid -1-entries to distinguish from valid entries 
(node numbers) later */
int init_cell(int number_of_nodes, int number_of_colours, int *cell_list){

	/* return -1 and error message if number_of_nodes or number_of_colours < 1 */
	if(number_of_nodes < 1 || number_of_colours < 1){
		printf("Error in init_cell: number of nodes is %d ", number_of_nodes);
		printf("number of colours is %d ", number_of_colours);
		printf(" both parameters must be must be > 0 \n");
		return -1;
	}
	
	if(cell_list){
		int j;
		for(j=0; j<number_of_nodes*number_of_colours; j++){
			cell_list[j] = -1;
		}
	}
	else{
		printf("third parameter cell_list is Nullpointer \n");
		return -1;
	}

	return 0; //function worked correctly
}


/* inserts a node into the cell list according to the number of the colour 
   cell_list[nodes_with_colour 0 -1 .. -1| nodes_with_colour 2 -1 ...-1 | ...| nodes_with_colour 
	noc-1, -1 ...-1]
*/
int set_colour_of_node(int number_of_nodes, int colour, int node, int *cell_list, int *cell_sizes){

	/* return -1 and error message if node is not in correct interval */
	if(node < 0 || node >= number_of_nodes){
		printf("error in set_colour_of_node: number of node is %d ", node);
		printf(" but must between 0 and %d", number_of_nodes-1);
		return -1;
	}
	/* return -1 and error message if node is not in correct interval */
	if(colour < 0 ){
		printf("error in has_colour: colour is %d but must be >= 0 ", colour);
		return -1;
	}

	int j=0;
	//offset = colour*number_of_nodes
	//search for first empty place in cell: offset + already stored nodes of the same colour
	while( cell_list[colour*number_of_nodes + j] >= 0){
		j++;
	}
	//store node  
	cell_list[colour*number_of_nodes + j] = node;

	/* a node got the the colour  int colour, so the number of nodes with that colour is increased 
	   by one, therefor also the size of the respective cell */
	cell_sizes[colour] = cell_sizes[colour]+1;

	
	//mark that this node has been coloured 
	set_node(node, global_is_set);

	if(DEBUG_VERSION){
		printf("set colour of node: %d  %d \n", node, colour);
	}

	return 0; //function worked correct

}

/* marks a node that has been coloured */
void set_node(int node, int is_set[]){
	
	int array_index = node/WORDSIZE;
	int bit_index = node % WORDSIZE;
	int bit_mask = 1<<bit_index;

	is_set[array_index] = is_set[array_index] | bit_mask ;
}


/* if node has a colour then the number is returned otherwise colour 0 is given to that node 
change: the number of nodes with colour 0 must be incresaed buy one if node has no colour */
int has_colour(int node, int *cell_list, int *cell_sizes, int number_of_nodes, int number_of_colours){
	
	/* return -1 and error message if node is not in correct interval */
	if(node < 0 || node >= number_of_nodes){
		printf("error in has_colour: number of node is %d ", node);
		printf(" but must between 0 and %d \n", number_of_nodes-1);
		return -1;
	}

	int i, j, offset;
	
	/* seach in cell_list cell after cell*/
	for(i=0; i<number_of_colours; i++){ 
		offset = i*number_of_nodes; //next cell
		for(j=0; j<number_of_nodes; j++){ 
			if(cell_list[offset+j] == node){ //node is in cell i, so has colour i
				return i;
			}
		}
	}
	//node is not in cell list
	//printf("Colour of node %d was not set! Set colour to 0 \n", node);
	
	set_colour_of_node(number_of_nodes, 0, node, cell_list, cell_sizes);
	return 0; // node has now colour 0;
}

/* set *lab and *ptn manually in case options.defaultptn = FALSE */
void set_label(int *lab, int *ptn, int *cell_list, int number_of_nodes, int number_of_colours){

	int i,j, lab_index, offset;
	lab_index = 0; // index to fill *lab
	 //default partition
	for(i=0; i< number_of_nodes-1; i++){
		ptn[i] = NAUTY_INFINITY;
	}
	ptn[number_of_nodes-1] = 0; //last element of *lab is always end of a cell

	for(i=0; i<number_of_colours; i++){ 
		offset = i*number_of_nodes; //next cell
		for(j=0; j<number_of_nodes; j++){ 
			if(cell_list[offset+j] >= 0){ //node of cell is put into *lab 
				lab[lab_index] = cell_list[offset+j];
				lab_index++;
			}
		}
		ptn[lab_index-1] = 0; //indicate where end of one cell (nodes with same colour) is
	}

	if(DEBUG_VERSION){
		int i;
		for(i=0; i< MAXN; i++){
			//printf("%d ",lab[i]);
		}
		printf("\n");
	}
	/* lab_index == number_of_nodes otherwise not all or to many nodes have been copied into *lab 
		lab_index starts with 0, but is incemented once more after last node has been inserted
	*/
	if(!(lab_index == number_of_nodes)){
		printf(" Error: In function set_label: ");
		printf("number of nodes in *lab: %d must be the same \n ", lab_index);
		printf(" as the total number of nodes: %d \n", number_of_nodes);
	}

}

/* initialize the array with the cell sizes with 0*/
void init_cell_sizes(int *cell_sizes, int length){

	int i=0;
	for(i=0; i<length; i++){
		cell_sizes[i] = 0;
	}
}


