:- module(pltables_export_html, [
							html_header/2,
							html_footer/2,
							html_table_header/2,
							html_table_footer/2,
							html_table_rows/2
					 	   ]).

:- use_module(pltables_export_tools).

:- use_module(probsrc(module_information)).

:- use_module(library(lists)).

:- module_info(group,pltables).
:- module_info(description,'This module is used to export coverage tables to html.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

% ----
% table headers
% ----
html_header(S, TableName) :-
	pltables:pltable(TableName, Attributes),
	write(S,'<html>\n<head>\n'),
	write(S,'<style type="text/css">\n  a {text-decoration: none; color:#000000;}\ntd {border:1px solid black;}\n</style>'),
	write(S,'</head>\n<body>\n'),
	(
	  member(headline(XH), Attributes) ->
	      format(S, '<b>~w</b><br><br>', XH)
	      ; true
	),
	write(S,'<table>\n').

% ----
% table footers
% ----
html_footer(S, _TableName) :-
  write(S,'</table>\n</body>\n</html>').
  
% ----
% table header / footer
% ----
html_table_header(S, TableName) :-
    pltables:pltable(TableName, TableAttributes),
	pltables:pltable_header(TableName, Columns, Attributes),
	(
	  member(description(XD, DescAttrs), TableAttributes)
	  -> length(Columns,NumColumns),
	     format(S, '<tr><td colspan="~w" style="border:0px;">', [NumColumns]),
	     (member(bold, DescAttrs) -> write(S, '<b>') ; true),
	     format(S, '~w', [XD]),
	     (member(bold, DescAttrs) -> write(S, '</b>') ; true),
	     write(S, '</td></tr>\n')
	  ;  true
	),
	write(S, '<tbody>\n'),
	write(S, '<tr>'),
	write_html_table_header(S, Columns, Attributes),
	write(S, '</tr>').
	
write_html_table_header(_S, [], []).
write_html_table_header(S, [Column | Columns], [Attributes|TAttributes]) :-
    (member(align(left), Attributes) -> format(S, '<td align="left">~w</td>', [Column]) ; true),
    (member(align(right), Attributes) -> format(S, '<td align="right">~w</td>', [Column]) ; true),
    (member(align(center), Attributes) -> format(S, '<td align="center">~w</td>', [Column]) ; true),
    (\+member(align(_), Attributes) ->  format(S, '<td>~w</td>', [Column]) ; true),
	write_html_table_header(S, Columns, TAttributes).
	
html_table_footer(S, _TableName) :-
    write(S, '</tbody>\n').

% ----
% table rows
% ----
html_table_rows(S, TableName) :-
	findall(Entries, pltables:pltable_row(TableName, Entries, Attributes), ListOfEntries),
	findall(Attributes, pltables:pltable_row(TableName, Entries, Attributes), ListOfAttributes),
	pltables:pltable(TableName, TableAttributes),
	pltables:pltable_header(TableName, _, ColumnsAttributes),
	write_html_table_rows(S, ListOfEntries, ListOfAttributes, TableAttributes, ColumnsAttributes).
	
write_html_table_rows(_S, [], [], _TableAttributes, _ColumnsAttributes).
write_html_table_rows(S, [Entries | MoreEntries], [RowAttributes | MoreAttributes], TableAttributes, ColumnsAttributes) :-
	write(S, '<tr>'),
	write_html_table_row(S, Entries, RowAttributes, TableAttributes, ColumnsAttributes),
	write(S, '</tr>\n'),
	write_html_table_rows(S, MoreEntries, MoreAttributes, TableAttributes, ColumnsAttributes).
	
write_html_table_row(_S, [], _RowAttributes, _TableAttributes, []).
write_html_table_row(S, [Entry | Entries], RowAttributes, TableAttributes, [ColumnAttribute | ColumnsAttributes]) :-
	merge_attributes(TableAttributes, ColumnAttribute, TmpAttributes),
	merge_attributes(TmpAttributes, RowAttributes, MergedAttributes),
    (member(align(left), MergedAttributes) -> write(S, '<td align="left">') ; true),
    (member(align(right), MergedAttributes) -> write(S, '<td align="right">') ; true),
    (member(align(center), MergedAttributes) -> write(S, '<td align="center">') ; true),
    (\+member(align(_), MergedAttributes) ->  write(S, '<td>') ; true),
    (member(url(URL), MergedAttributes) -> format(S, '<a href="~w">', [URL]) ; true),
    write(S, '<div style="'),
	(member(color(C), MergedAttributes) -> format(S, 'color:#~w', [C]) ; true),
	write(S, '">'),
	(member(bold, MergedAttributes) -> write(S, '<b>') ; true),
	format(S, '~w', [Entry]),
	(member(bold, MergedAttributes) -> write(S, '</b>') ; true),
	write(S, '</div>'),
    (member(url(_), MergedAttributes) -> write(S, '</a>') ; true),
	write(S, '</td>'),
	write_html_table_row(S, Entries, RowAttributes, TableAttributes, ColumnsAttributes).