:- module(pltables_export_xml, [
							xml_header/2,
							xml_footer/2,
							xml_table_header/2,
							xml_table_footer/2,
							xml_table_rows/2
					 	   ]).

:- use_module(pltables_export_tools).

:- use_module(probsrc(module_information)).

:- module_info(group,pltables).
:- module_info(description,'This module is used to export coverage tables to xml.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

% ----
% table headers
% ----
xml_header(S, TableName) :-
	pltables:pltable(TableName, Attributes),
	write(S,'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'),
	write(S,'<tables>'),
	(
	  member(headline(XHU), Attributes) ->
	      xml_escape(XHU,XH),
	      format(S, '<headline>~w</headline>\n', [XH])
	      ; true
	).

% ----
% table footers
% ----
xml_footer(S, _TableName) :-
  write(S,'</tables>').
  
% ----
% table header / footer
% ----
xml_table_header(S, TableName) :-
	write(S, '<table'),
    pltables:pltable(TableName, TableAttributes),
	pltables:pltable_header(TableName, Columns, _Attributes),
	(
	  member(description(XDU, _DescAttrs), TableAttributes) ->
	      xml_escape(XDU,XD),
	      format(S, ' description="~w"', [XD])
	      ; true
	),
	write(S, '>\n'),
	write_xml_table_header(S, Columns).

write_xml_table_header(S, Columns) :-
    write(S, '    <header>\n'),
    write_xml_table_header2(S, Columns),
    write(S, '    </header>\n').

write_xml_table_header2(_S, []).
write_xml_table_header2(S, [Column | Columns]) :-
    xml_escape(Column, EscapedColumn),
	format(S, '        <column>~w</column>\n', [EscapedColumn]),
	write_xml_table_header2(S, Columns).
	
xml_table_footer(S, _TableName) :-
    write(S,'</table>\n').
	

% ----
% table rows
% ----
xml_table_rows(S, TableName) :-
	findall(Entries, pltables:pltable_row(TableName, Entries, _Attributes), ListOfEntries),
	write_xml_table_rows(S, ListOfEntries).
	
write_xml_table_rows(_S, []).
write_xml_table_rows(S, [Entries | MoreEntries]) :-
    write(S, '    <row>\n'),
	write_xml_table_row(S, Entries),
	write(S, '    </row>\n'),
	write_xml_table_rows(S, MoreEntries).
	
write_xml_table_row(_S, []).
write_xml_table_row(S, [Entry | Entries]) :-
	xml_escape(Entry, EscapedEntry),
	format(S, '        <cell>~w</cell>\n', [EscapedEntry]),
	write_xml_table_row(S, Entries).
	
% ----
% escaping characters
% ----
xml_escape(ToEscape, ToEscape).

%xml_escape(ToEscape, Escaped) :-
%  atom(ToEscape), !,
%  atom_codes(ToEscape, CodeList),
%  xml_escape2(CodeList, EscapedCodeList),
%  atom_codes(Escaped, EscapedCodeList).
%xml_escape(ToEscape, ToEscape).

xml_escape2([],[]).
xml_escape2([95|T], [92,95|EscT]) :-
  xml_escape2(T, EscT).
xml_escape2([35|T], [92,35|EscT]) :-
  xml_escape2(T, EscT).
xml_escape2([36|T], [92,36|EscT]) :-
  xml_escape2(T, EscT).
xml_escape2([37|T], [92,37|EscT]) :-
  xml_escape2(T, EscT).
xml_escape2([H|T], [H|EscT]) :-
  H \= 95,
  xml_escape2(T, EscT).