% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(timer, [timer_init/0, clock_ticks/1, clock_microseconds/1,
	  print_clock_info/1,
	  start_timer/1, stop_timer/1, timer_call/1, timer_call/2,
	  timer_statistics/0,
      test/0]).

% use    timer:timer_statistics.      to get info

:- dynamic timer_started/3.

:- op(300,fx,timer_call). 
:- meta_predicate timer_call(0).
:- meta_predicate timer_call(-,0).

timer_call(Call) :- get_pred(Call,_,Pred,_), timer_call(Pred,Call).

timer_call(PP,Call) :-
   start_bt_timer(PP),
   %statistics(runtime,[Start,_]),
   call(Call),
   %statistics(runtime,[Stop,_]),
   %(Stop-Start>40 -> print(' ** '),T is Stop-Start, print(T), print(' ms '),nl,print(Call),nl,trace ; true),
   stop_bt_timer(PP).

start_bt_timer(X) :- start_timer(X,entry).
start_bt_timer(X) :- stop_timer(X,fail),fail.

stop_bt_timer(X) :- stop_timer(X,success).
stop_bt_timer(X) :- start_timer(X,redo),fail.   

get_pred(Module:Call,Module,Pred,Arity) :- !,get_pred(Call,_,Pred,Arity).
get_pred(Call,user,Pred,Arity) :- functor(Call,Pred,Arity).

start_timer(PP) :- start_timer(PP,entry).
start_timer(ProgramPoint,Info) :-
   (timer_started(ProgramPoint,Info1,_)
   -> print(nested(ProgramPoint,Info1,Info)),nl
   ;  % print(starting_timer(ProgramPoint)),nl,
      clock_microseconds(C), assert(timer_started(ProgramPoint,Info,C))
   ).

:- dynamic total_time/3.

clean_timer(ProgramPoint) :- %print(clean(ProgramPoint)),nl,
    retractall(timer_started(ProgramPoint,_,_)).

stop_timer(PP) :- stop_timer(PP,success).
stop_timer(ProgramPoint,SF) :-  clock_microseconds(Cur),
   (retract(timer_started(ProgramPoint,Info,Old))
     -> Delta is Cur-Old,
        update_total_time(ProgramPoint,Delta),
        update_entry_redo(Info,SF,ProgramPoint)
     ;  print(timer_not_started(ProgramPoint)),nl
   ).

:- dynamic program_point_info/5.
update_entry_redo(entry,fail,PP) :- get_PP_info(PP,EntryFail,EntrySucc,RedoFail,RedoSucc),
  New is EntryFail+1, assert(program_point_info(PP,New,EntrySucc,RedoFail,RedoSucc)).
update_entry_redo(entry,success,PP) :- get_PP_info(PP,EntryFail,EntrySucc,RedoFail,RedoSucc),
  New is EntrySucc+1, assert(program_point_info(PP,EntryFail,New,RedoFail,RedoSucc)).
update_entry_redo(redo,fail,PP) :- get_PP_info(PP,EntryFail,EntrySucc,RedoFail,RedoSucc),
  New is RedoFail+1, assert(program_point_info(PP,EntryFail,EntrySucc,New,RedoSucc)).
update_entry_redo(redo,success,PP) :-  get_PP_info(PP,EntryFail,EntrySucc,RedoFail,RedoSucc),
  New is RedoSucc+1, assert(program_point_info(PP,EntryFail,EntrySucc,RedoFail,New)).

get_PP_info(PP,EntryFail,EntrySucc,RedoFail,RedoSucc) :-
   (retract(program_point_info(PP,EntryFail,EntrySucc,RedoFail,RedoSucc))
    -> true
     ;  EntryFail=0,EntrySucc=0,RedoFail=0,RedoSucc=0
   ).
update_total_time(ProgramPoint,Delta) :-
   	  (retract(total_time(ProgramPoint,OldMax,Total))
        -> true ; Total = 0,OldMax=0),
       NewTotal is Total+Delta,
       % print(stopped_timer(ProgramPoint,NewTotal)),nl,
       (OldMax < Delta -> NewMax = Delta ; NewMax = OldMax),
       assert(total_time(ProgramPoint,NewMax,NewTotal)).


% timer:timer_statistics.
timer_statistics :-  print('Time (milliseconds) [Max] --->   ProgramPoint :: (EntryFail,EntrySucc,RedoFail,RedoSucc)'),nl,
   findall(time(Time,Max,PP),total_time(PP,Max,Time),L),
   sort(L,SL),
   timer_statistics(SL).
timer_statistics([]) :- nl.
timer_statistics([time(Time,Max,PP)|T]) :-
   print_time(Time), print(' ['),print_time(Max),print(']  --->  '), print(PP), print(' :: '),
   program_point_info(PP,EntryFail,EntrySucc,RedoFail,RedoSucc),
   print((EntryFail,EntrySucc,RedoFail,RedoSucc)),nl,
   timer_statistics(T).

print_time(Time) :- TT is Time / 1000.0, format('~3F',[TT]).

/* the interface to a heap global variable stored and maintained in C */

foreign(clock_ticks,clock_ticks([-integer])).
foreign(clock_microseconds,clock_microseconds([-integer])).


% to do: also implement a pop from end to replace the not_all_z/1 relation
% for mixed depth-first, breadth-first traversal

foreign_resource(timer,
  [
   clock_ticks,
   clock_microseconds
  ]).
  
  
  :- dynamic loaded/0.

timer_init :- loadfr.

loadfr :-
  (loaded -> true
   ; (assert(loaded), 
      %assert_dir, 
      assert(user:library_directory('.')),
      load_foreign_resource(library(timer)))
   ).
%assert_dir :- (user:library_directory('.') -> true ; assert(user:library_directory('.'))).

:- loadfr.

test :-  timer_init,
         statistics(runtime,_),
         clock_microseconds(X),
         start_timer(app),
         app([a,b,c,d,e,f,g,h,i,j,k,l,m],[z],_),
         print_clock_info(X),
         app([a,b,c,d,e,f,g,h,i,j,k,l,m],[z],_),
         stop_timer(app),
         print_clock_info(X),
         start_timer(app),
         app([a,b,c,d,e,f,g,h,i,j,k,l,m],[z],_),
         stop_timer(app),
         print_clock_info(X),
         start_timer(do_something),
         do_something,
         stop_timer(do_something),
         print_clock_info(X),
         ( timer_call(app2(A,B,[a,b,c,d,e,f,g,h])), print((A,B)),nl, fail ; true),
         timer_call(app2([a,b,c,d,e,f,g,h],[z],_)),
         statistics(runtime,[A,B]),
         print(runtime(B)),nl,
         timer_statistics.
   
do_something :- do_something(1000000).
do_something(0).
do_something(X) :- X1 is X-1, do_something(X1).

app([],Z,Z).
app([H|X],Y,[H|Z]) :- app(X,Y,Z).
app2([],Z,Z).
app2([H|X],Y,[H|Z]) :- app2(X,Y,Z).


print_clock_info(ReferencePoint) :-
    clock_microseconds(Now),
    Diff is Now-ReferencePoint,
    print(Diff), print(' microseconds'),nl.
