#include "user_signal.h"

/* system independent code */

typedef enum {status_ignore, status_protected, status_default} status_type;

static status_type status = status_ignore;
static int signal_in_queue = 0;
static long int reference_id;

static void raise_prolog_exception(void) {
  SP_atom atom = SP_atom_from_string(message_string);
  SP_term_ref exception_term = SP_new_term_ref();
  SP_put_atom(exception_term, atom);
  SP_raise_exception(exception_term);  
}

void set_user_signal_mode(long int look_for_queued, long int mode) {
  int interrupted = (look_for_queued != 0L) && (mode == status_default) && signal_in_queue;
  signal_in_queue = 0;
  if (mode >= status_ignore && mode <= status_default) {
    status = mode;
  } else {
    fprintf(stderr, "wrong mode: %li\n", mode);
  }
  if (interrupted) {
    raise_prolog_exception();
  }
}

long int get_user_signal_mode(void) {
  return (long) status;
}

long get_user_sig_reference(void) {
  return reference_id;
}

static int user_signal_event(void *handle) {
  switch (status) {
  case status_default:
    raise_prolog_exception();
    return SP_ERROR;
  case status_ignore:
    return SP_SUCCESS;
  case status_protected:
    signal_in_queue = 1;
    return SP_SUCCESS;
  default:
    fprintf(stderr, "user signal handler: unexpected status\n");
    return SP_SUCCESS;
  }
}

static void register_interruption_event(void) {
  int success = SP_event(user_signal_event, (void*)0);
  if (!success) {
    fprintf(stderr,"SP_event failed!\n");
    fflush(stderr);
  }
}

/****************************************************************************/
/* Windows: we use the Windows messages mechanism to receive user interrupts*/
/* To enable this, we must start a dedicated thread with an event loop      */
/* The process that wants to send the message needs the ID of the thread    */
#if _WIN32

static DWORD user_interrupt_event_loop(LPVOID lpdwThreadParam);
static UINT msgid;

void user_signal_init(void)
{
  DWORD thread_id;
  msgid = RegisterWindowMessageA("prob_user_interrupt");
  /* printf("message id: %u\n", msgid); */
  if (CreateThread(NULL,                                 /* Default security level */
               0,                                    /* Default stack size */
               (LPTHREAD_START_ROUTINE) &user_interrupt_event_loop,
               0,                       /* Thread paramater */
               0,                                    /* Run immediately */
               &thread_id
                   ) == NULL) {
    fprintf(stderr, "user interrupt signal: creating thread failed.\n");
  }
  fflush(stdout);
  reference_id = thread_id;
}

static DWORD user_interrupt_event_loop(LPVOID params) {
  MSG msg;
  while(GetMessage(&msg, NULL, 0, 0) != 0) {
    if (msg.message == msgid) {
      /* printf("User interrupt: ProB message received\n"); */
      register_interruption_event();
    } else {
      /* printf("User interrupt: Unknown windows message received\n"); */
    }
    fflush(stdout);
  }
  return 0;
}

/****************************************************************************/
/* Unix-like: we just use the unix signal mechanism (SIGUSR1), the          */
/* process that wants to emit the signal needs the ID of the process        */
#else /* _WIN32 */

static void user_signal_handler(int signal_no, void *arg)
{
  register_interruption_event();
}

void user_signal_init(void)
{
  /* printf("Signal initialized !!!!!"); */
  SP_signal(SIGUSR1, user_signal_handler, (void*)0);
  reference_id = getpid();
}

#endif /* _WIN32 */
