% (c) 2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(user_signal, [ user_interruptable_call_det/2
                       , protect_from_user_interrupt_det/1
                       , get_user_signal_ref/1]).

:- use_module('../../src/module_information.pl').

:- module_info(group,infrastructure).
:- module_info(description,'This is the interface to C code for interrupting Prolog runs with a UNIX signal.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

foreign_resource(user_signal, [set_user_signal_mode,get_user_signal_mode,user_signal_init,get_user_sig_reference]).
foreign(user_signal_init, user_signal_init).
foreign(set_user_signal_mode, set_user_signal_mode(+integer,+integer)).
foreign(get_user_signal_mode, get_user_signal_mode([-integer])).
foreign(get_user_sig_reference, get_user_sig_reference([-integer])).

:- dynamic loaded/0.

init_user_signal :- loadfr.

loadfr :-
    ( loaded -> true
    ; otherwise ->
        load_foreign_resource(library(user_signal)),
        user_signal_init,
        assert(loaded)).

catch_interrupt_exception(Call,Result) :-
    catch( call(Call),
           E,
           ( E==user_interrupt_signal -> Result=interrupted
           ; otherwise -> throw(E))),
    ( Result == interrupted -> true
    ; otherwise             -> Result=ok).

:- meta_predicate user_interruptable_call_det(:,+).
user_interruptable_call_det(Call,Result) :-
    loadfr,
    get_user_signal_mode(Mode),
    call_cleanup( (set_user_signal_mode(1,2),
                      catch_interrupt_exception(Call,Result),!),
                  set_user_signal_mode(1,Mode)).

:- meta_predicate protect_from_user_interrupt_det(:).
protect_from_user_interrupt_det(Call) :-
    loadfr,
    get_user_signal_mode(Mode),
    call_cleanup( (set_user_signal_mode(0,1), call(Call), !),
                  set_user_signal_mode(1,Mode)).

get_user_signal_ref(Ref) :-
    loadfr,
    get_user_sig_reference(Ref).
