#include <stdio.h>
#include <stdlib.h>
#include "external.h"
#include "string.h"
#include "assert.h"

FILE *external_reader = NULL;
FILE *external_writer = NULL;

int buf_size = 0; // number of buffered workpackages

char *external_name;

void init_external(char *id) {
    char *name = (char*) malloc(15); // 15 should be enough, since worker id <= 16
    memset(name, 0, 15);
    sprintf(name, "external%s", id);
    external_name = name;
    external_writer = fopen(name, "w+");
    external_reader = fopen(name, "r");
}

void teardown_external() {
    // close them
    fclose(external_reader);
    fclose(external_writer);
    // delete file
    remove(external_name);
    free(external_name);
}

tCell *build_wp(char *term, char *digest, int term_length, int id) {
    tCell *new = (tCell*) malloc(sizeof(tCell));
    assert(new);
    new->digest = digest;
    new->id = id;
    new->term_length = term_length;
    new->term = term;
    new->next = NULL;

    return new; 
}

void teardown_tcell(tCell *cell) {
    free(cell->digest);
    free(cell->term);
    free(cell);
}


void serialize_package(tCell *wp) {
    assert(external_writer);
    fprintf(external_writer, "%d %d ", wp->id, wp->term_length);
    int i;
    // i < 20: 20 == KEYLENGTH at hashtrie
    for (i = 0; i < 20; i++) {
        fputc(wp->digest[i], external_writer);
    }
    fputc(' ', external_writer);
    for (i = 0; i < wp->term_length; i++) {
        fputc(wp->term[i], external_writer);
    }
    fputc('\n', external_writer);
    buf_size++;
    fflush(external_writer);
    teardown_tcell(wp);
}

tCell *deserialize_package() {
    assert(external_reader);
    if (buf_size == 0) {
        return NULL;
    }

    int id = -1;
    int len = -1;
    int rc = fscanf(external_reader, "%d %d", &id, &len);
    assert(rc == 2);
    assert(id != -1 && len != -1);

    char c = fgetc(external_reader);
    if (c != ' ') {
        printf("%d '%c'\n", c, c);
    }
    assert(c == ' ');

    int i;
    char *digest = (char*) malloc(20);
    assert(digest);

    for (i = 0; i < 20; i++) {
        digest[i] = fgetc(external_reader);
    }

    assert(fgetc(external_reader) == ' ');

    char *term = (char*) malloc(len);
    assert(term);

    for (i = 0; i < len; i++) {
        term[i] = fgetc(external_reader);
    }
    assert(fgetc(external_reader) == '\n');
    buf_size--;

    tCell *new = build_wp(term, digest, len, id);
    return new;
}

tCell *build_internal_wp(char *term, char *digest, int term_length, int id) {
    char *wp_digest = (char*) malloc(20);
    assert(wp_digest);
    memcpy(wp_digest, digest, 20);
    char *wp_term = (char*) malloc(term_length);
    assert(wp_term);
    memcpy(wp_term, term, term_length);
    return build_wp(wp_term, wp_digest, term_length, id);
}

int equals(tCell *cell, char *term, char *digest, int term_length, int id) {
    int digest_ret = memcmp(digest, cell->digest, 20);
    int term_ret = memcmp(term, cell->term, term_length);
    return (digest_ret == 0 && term_ret == 0 && term_length == cell->term_length && id == cell->id);
}

void write_to_file(char *term, char *digest, int len, int id) {
    tCell *tmp = build_internal_wp(term, digest, len, id);
    serialize_package(tmp);
}

int get_external_wp_count() {
    return buf_size;
}

/*
int main(void) {
    init_external("123");
    tCell *wp1 = build_internal_wp("foo", "12345678901234567890", 3, 0);
    serialize_package(wp1); // tears down wp1
    tCell *wp2 = build_internal_wp("bar", "99999999999999999999", 3, 0);
    serialize_package(wp2); // tears down wp2

    tCell *c1 = deserialize_package();
    tCell *c2 = deserialize_package();
    tCell *c3 = deserialize_package();
    assert(equals(c1, "foo", "12345678901234567890", 3, 0));
    assert(equals(c2, "bar", "99999999999999999999", 3, 0));
    assert(c3 == NULL);

    teardown_tcell(c1);
    teardown_tcell(c2);

    teardown_external();

    return 0;
}
*/
