#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "sha.h"
#include "hashtrie.h"
#include "keylist.h"
//#include <sys/time.h>
//#include <sys/resource.h>

#define SIZE 32
#define KEYLENGTH 20

#ifdef VERBOSE 
char *fileid = NULL;

void setId(char id2[1]) {
    fileid = malloc(2);
    fileid[0] = id2[0];
    fileid[1] = 0;
}
FILE *logfile2;
#endif

tNode *root;

void recur_add(tNode *root, char key[21], int depth, int checked);
int contains(char key[21]);
tNode *transform(tNode *node, int newdepth);
tNode *init_leaf(char *digest, int pos, int checked);
tNode *init_node();
int checked_states_count = 0;
int all_states = 0;
void print_key(char *key);

//#ifdef VERBOSE
void print_readable_digest2(unsigned char *key, char *dest) {
	sprintf(dest, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",key[0],key[1],key[2],key[3],key[4],key[5],key[6],key[7],key[8],key[9],key[10],key[11],key[12],key[13],key[14],key[15],key[16],key[17],key[18],key[19]);
}
//#endif

void add(char key[21]) {
    int checked = key[0];
    assert(checked == 0 || checked == 1);
    #ifdef VERBOSE 
    char *dig = malloc(41);
    print_readable_digest2((unsigned char*)key + 1, dig);
    fprintf(logfile2, "adding key %s\n", dig);
    free(dig);
    #endif
    recur_add(root, key + 1, 0, checked);
}


void print_key(char *key) {
    int a;
    printf("key: ");
    for (a = 0; a < KEYLENGTH; a++) {
        printf("%d ", key[a]);
    }
    printf("\n");
}

void recur_add(tNode *parent, char key[20], int depth, int checked) {
    assert(depth < 31);
    assert(parent->type == 'N');
    int pos = calculate_pos(key, depth);

    tNode* *pointers = parent->pointers;
    tNode *next = pointers[pos];
    if (next == NULL) {
        int leaf_pos = calculate_pos(key, depth+1);
        tNode *leaf = init_leaf(key, leaf_pos, checked);
        assert(leaf->type == 'L');
        all_states++;
        if (checked) {
            checked_states_count++;
        }
        pointers[pos] = leaf;
        next = leaf;
        return;
    }
 
    if (next->type == 'L') {
        tLeaf *leaf_data = next->leaf_data;
        keyList *list = leaf_data->list;

        int leaf_pos = calculate_pos(key, depth+1);
        // no key inserted on our new position? great!
        if (free_position(list, leaf_pos)) {
            leaf_data->list = insert(list, depth, leaf_pos, key);
            if (checked == 1) {
                uint32_t b = leaf_data->checked;
                int power = (int) pow(2, leaf_pos); 
                if (!(b & power)) {
                    checked_states_count++;
                }
                b = b | power;
                leaf_data->checked = b;
            } 
            all_states++;
            return;
        // same key inserted on position - maybe set bit flag
        } else if (list_contains(list, key)) {
            // and we want to add the information that we checked it? bit fiddling incoming!
            if (checked == 1) {
                uint32_t b = leaf_data->checked;
                int power = (int) pow(2, leaf_pos); 
                if (!(b & power)) {
                    checked_states_count++;
                }
                b = b | power;
                leaf_data->checked = b;
            }
            return;
        } else {
            // okay, we neither found an empty spot nor our key
            // so we need to transform this leaf into a node.
  
            #ifdef DEBUG
              // TODO: debug stuff, that transformation works correctly
            #endif
            next = transform(next, depth+1);
            assert(next->type == 'N');
            pointers[pos] = next;
            #ifdef DEBUG
              // TODO: more debug stuff
            #endif
            recur_add(next, key, depth+1, checked);
        }
    } else {
        // our next step is a node, so let's go deeper!
        recur_add(next, key, depth+1, checked);
    }
//    assert(0);
    
}


int recur_search(tNode *parent, char key[20], int depth) {
    int pos = calculate_pos(key, depth);
    if (parent->type == 'N') {
        tNode* *pointers = parent->pointers;
        tNode *next = pointers[pos];
        if (next == NULL) {
            return 0;
        }
        return recur_search(next, key, depth+1);
    }
    if (parent->type == 'L') {
        tLeaf *leaf_data = parent->leaf_data;
        keyList *l = leaf_data->list;
        if (list_contains(l, key)) {
            uint32_t b = leaf_data->checked;
            if (b & ((int) pow(2, pos))) {
                return 1;
            }
            return -1;
        }
        return 0;
    }
    assert(0);
}


int contains(char key[21]) {
    return recur_search(root, key, 0);
}

int contains_checked(char key[20]) {
    return recur_search(root, key, 0) == 1;
}



tNode *transform(tNode *leaf, int newdepth) {
    //printf("transform\n");
    // let's initialize some pointers
    tLeaf *leaf_data = leaf->leaf_data;
    uint32_t old_bits = leaf_data->checked;
    keyList *old_keylist = leaf_data->list;

    tNode *new_node = init_node();
    tNode **pointers = new_node->pointers;

    while (old_keylist != NULL) {
        char *current_key = old_keylist->digest;
        int i = old_keylist->pos;
        int pos = calculate_pos(current_key, newdepth+1);

        int checked = 0;
        if (old_bits & ((int) pow(2, i))) {
            checked = 1;
        }
        // TODO: we could reuse the cell
        tNode *new_leaf = init_leaf(old_keylist->digest, pos, checked);
        assert(new_leaf);
        assert(new_leaf->leaf_data);
        assert(new_leaf->leaf_data->list);

        // set the pointer in the new node
        pointers[i] = new_leaf;
        old_keylist = teardown_cell(old_keylist);
    }
    free(leaf_data);
    free(leaf);
    return new_node;
}


int calculate_pos(char *key, int depth) {
    // pattern repeats after 8 times
    int offset = depth / 8;
    int moddepth = depth % 8;

    char *offi = key + 5 * offset;

// for some reason, this does not work :(
    uint64_t a = 0;
    int i;
    for (i = 0; i < 5; i++) {
       a += (uint64_t) (offi[i] * pow(2, 8*i));
    }
    uint64_t c = offi[0] + offi[1] * pow(2, 8) + offi[2] * pow(2, 16)
               + offi[3] * pow(2, 24) + offi[4] * pow(2, 32);
    assert(c == a);

    uint64_t b = a >> (moddepth * 5);
    int x = b & 31;
    return x;
}



tNode *init_node() {
    tNode *n = (tNode*) malloc(sizeof(tNode));
    assert(n != NULL);
    n->type = 'N';
    
    tNode **arr = (tNode**) calloc(SIZE, sizeof(tNode*));
    assert(arr != NULL);
    int i;
    for (i = 0; i < SIZE; i++) {
        arr[i] = NULL;
    }
    n->pointers = arr;
    return n;
}

tNode *init_leaf(char *digest, int pos, int checked) {
    //printf("init leaf\n");
    // init leaf data
    tLeaf *l = (tLeaf*) malloc(sizeof(tLeaf));
    assert(l != NULL);

    if (checked == 0) {
        l->checked = 0;
    } else {
        uint32_t power = (uint32_t) pow(2, pos); 
        l->checked = power;
    }

    keyList *list = create_cell(digest, pos);
    assert(list);
    assert(memcmp(digest, list->digest, 20) == 0);
    l->list = list;

    // init leaf node
    tNode *n = (tNode*) malloc(sizeof(tNode));
    assert(n != NULL);
    n->type = 'L';

    
    n->leaf_data = l;
    return n;
}


int count_elements() {
    return all_states;
}

int get_checked_states_count() {
    return checked_states_count;
}

int h_teardown(tNode *node) {
    assert(node != NULL);
    int count = 0;
    int i;
    if (node->type == 'L') {
        tLeaf *leaf_data = node->leaf_data;
        keyList *list = leaf_data->list;
        while (list != NULL) {
            count++;
            list = teardown_cell(list);
        }
        free(leaf_data);
    } else if (node->type == 'N') {
        tNode **pointers = node->pointers;
        for (i = 0; i < SIZE; i++) {
            if (pointers[i] != NULL) {
                count += h_teardown(pointers[i]);
            }
        }
        free(node->pointers);
    }
    free(node);
    return count;
}

int h_teardown_and_dump(tNode *node, FILE *hashdump) {
    assert(node != NULL);
    int count = 0;
    int i;
    if (node->type == 'L') {
        tLeaf *leaf_data = node->leaf_data;
        keyList *list = leaf_data->list;
        while (list != NULL) {
            count++;
            char readable[41];
            print_readable_digest2((unsigned char *) list->digest, readable);
            fprintf(hashdump, "%s\n", readable);
            list = teardown_cell(list);
        }
        free(leaf_data);
    } else if (node->type == 'N') {
        tNode* *pointers = node->pointers;
        for (i = 0; i < SIZE; i++) {
            if (pointers[i] != NULL) {
                count += h_teardown_and_dump(pointers[i], hashdump);
            }
        }
        free(node->pointers);
    }
    free(node);
    return count;
}

int teardown_and_dump() {
    FILE *hashdump = fopen("hashdump", "w");
    int res = h_teardown_and_dump(root, hashdump);
    fclose(hashdump);
    root = NULL;
    checked_states_count = 0;
    assert(res == all_states);
    all_states = 0;
    return res;
}

int teardown() {
    int res = h_teardown(root);
    root = NULL;
    checked_states_count = 0;
    assert(res == all_states);
    all_states = 0;
    return res;
}


int h_iterate(char *mem_block, tNode *node, int count) {
    int tmp_count = count;
    int i;
    if (node->type == 'L') {
        tLeaf *leaf = node->leaf_data;
        keyList *kl = leaf->list;

        while (kl != NULL) {
            memcpy(mem_block + 1 + ((1+KEYLENGTH) * tmp_count), kl->digest, KEYLENGTH);
            int power = (int) pow(2, kl->pos);
            if (leaf->checked & power) {
                mem_block[((1+KEYLENGTH) * tmp_count)] = 1;
            } else {
                mem_block[((1+KEYLENGTH) * tmp_count)] = 0;
            }
            tmp_count++;
            kl = kl->next;
        }
        return tmp_count;
    } else if (node->type == 'N') {
        for (i = 0; i < 32; i++) {
             tNode *pntr = node->pointers[i];
             if (pntr != NULL) {
                 tmp_count = h_iterate(mem_block, pntr, tmp_count);
             }
        }
        return tmp_count;
    } else {
        assert(0); 
    }
}

char *iterate() {
    char *mem_block = (char*) malloc((1+KEYLENGTH) * all_states);
    int res = h_iterate(mem_block, root, 0);
    assert(res == all_states);
    return mem_block;
}



int contains_processed(char key[21]) {
    return contains_checked(key);
}

void put(char key[21]) {
    add(key);
}

void put_local(char key[20]) {
    recur_add(root, key, 0, 0);
}

void put_checked(char key[20]) {
    recur_add(root, key, 0, 1);
}

int h_count_elements(tNode *node) {
    assert(node != NULL);
    int count = 0;
    int i;
    if (node->type == 'L') {
        tLeaf *leaf_data = node->leaf_data;
        keyList *list = leaf_data->list;
        while (list != NULL) {
            count++;
            list = list->next;
        }
    } else if (node->type == 'N') {
        tNode* *pointers = node->pointers;
        for (i = 0; i < SIZE; i++) {
            if (pointers[i] != NULL) {
                count += h_count_elements(pointers[i]);
            }
        }
    }
    return count;
}

void init_hashmap() {
    #ifdef VERBOSE
    char* name = malloc(20);
    sprintf(name, "hashmap%s.log", fileid);
    logfile2 = fopen(name, "a+");
    free(name);
    #endif
    root = init_node();
}

int main() {
    root = init_node();

    put_checked("012345678901234567890"); 
    assert(contains("012345678901234567890")); 
    assert(contains_checked("012345678901234567890")); 
    put_local("012345678901234567890"); 
    assert(contains("012345678901234567890")); 
    put_local("012345678201234567890"); 
    assert(contains("012345678901234567890")); 
    assert(contains("012345678201234567890")); 
    assert(contains_checked("012345678901234567890")); 
    int a;
    //int debug_count = 2;
    for (a = 1; a <= 5000000; a++) {        
        int len = (int)log10(a) + 2;
        char *res = (char*) malloc(len);
        sprintf(res,"%i",a);
        char *digest = (char*) malloc(20);
        sha1(res, len, digest);
        put_local(digest);
        free(digest);
        free(res);
    }
    printf("# of elements: %d\n", get_checked_states_count());
    //struct rusage r;
    //getrusage(RUSAGE_SELF, &r);
    //printf("Memory allocated: %d\n", r.ru_maxrss);

    char *hashes = iterate();
    for (a = 0; a < all_states; a++) {
        int res = contains(hashes + (a * 21) + 1);
        assert(hashes[a * 21] == 0 || hashes[a * 21] == 1);
        if (hashes[a * 21] == 0)
            assert(res == -1);
        else
            assert(res == 1);
    }

    free(hashes);

    printf("# of elements: %d\n", teardown_and_dump());
    return 0;
}
