// WARNING: Requires C99 compatible compiler
 
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>  
#include "heap.h"
   
#define CMP(a, b) ((a) >= (b))
    
    wQueue *init_queue() {
        struct heap *h = malloc(sizeof(struct heap*));
        heap_init(h);
        return h;
    }

    tCell *dequeue(wQueue *q) {
        tCell *x = heap_front(q);
        heap_pop(q);
        return x;
    }

    void enqueue(wQueue *q, char *term, int len, char digest[20], int thresh, int id) {
        tCell *new = malloc(sizeof(tCell));
    
        new->id = 0;
        new->term = malloc(len);
        memcpy(new->term, term, len);
        new->term_length = len;
        new->digest = malloc(20);
        memcpy(new->digest, digest, 20);
        new->next = NULL;
        enqueue_cell(q, new);
    }

    void enqueue_cell(wQueue *q, tCell *new) {
        heap_push(q, new);
    }

    int is_empty(wQueue *q) {
        return q->count == 0;
    }

    int q_size(wQueue *q) {
        return q->count;
    }

    void teardown_queue(wQueue **q) {
        heap_term(*q);
        free(*q);
    }

     static const unsigned int base_size = 20000;
     
     // Prepares the heap for use
     void heap_init(struct heap *h)
     {
         *h = (struct heap){
             .size = base_size,
             .count = 0,
             .data = malloc(sizeof(type) * base_size)
         };
         if (!h->data) _exit(1); // Exit if the memory allocation fails
     }
      
      // Inserts element to the heap
      void heap_push(struct heap *h, type value)
      {
          int val = rand() % 20000;
          unsigned int index, parent;
          value->val = val; 

           // Resize the heap if it is too small to hold all the data
           if (h->count == h->size)
           {
               h->size <<= 1;
               h->data = realloc(h->data, sizeof(type) * h->size);
               if (!h->data) _exit(1); // Exit if the memory allocation fails
           }
            
            // Find out where to put the element and put it
            for(index = h->count++; index; index = parent)
            {
                parent = (index - 1) >> 1;
		//		printf("%d cmp %d\n",h->data[parent]->val, val);
                if CMP(h->data[parent]->val, val) break;
                h->data[index] = h->data[parent];
            }
            h->data[index] = value;
      }
       
       // Removes the biggest element from the heap
       void heap_pop(struct heap *h)
       {
           unsigned int index, swap, other;
            
            // Remove the biggest element
            type temp = h->data[--h->count];
             
/*             // Resize the heap if it's consuming too much memory
             if ((h->count <= (h->size >> 2)) && (h->size > base_size))
             {
                 h->size >>= 1;
                 h->data = realloc(h->data, sizeof(type) * h->size);
                 if (!h->data) _exit(1); // Exit if the memory allocation fails
             }
*/              
              // Reorder the elements
              for(index = 0; 1; index = swap)
              {
                  // Find the child to swap with
                  swap = (index << 1) + 1;
                  if (swap >= h->count) break; // If there are no children, the heap is reordered
                      other = swap + 1;
                      if ((other < h->count) && CMP(h->data[other]->val, h->data[swap]->val)) swap = other;
                      if CMP(temp, h->data[swap]) break; // If the bigger child is less than or equal to its parent, the heap is reordered
                           
                           h->data[index] = h->data[swap];
              }
              h->data[index] = temp;
       }
        
        // Heapifies a non-empty array
        void heapify(type data[], unsigned int count)
        {
            unsigned int item, index, swap, other;
            type temp;
             
             // Move every non-leaf element to the right position in its subtree
             item = (count >> 1) - 1;
             while (1)
             {
                 // Find the position of the current element in its subtree
                 temp = data[item];
                 for(index = item; 1; index = swap)
                 {
                     // Find the child to swap with
                     swap = (index << 1) + 1;
                     if (swap >= count) break; // If there are no children, the current element is positioned
                         other = swap + 1;
                         if ((other < count) && CMP(data[other]->val, data[swap]->val)) swap = other;
                         if CMP(temp->val, data[swap]->val) break; // If the bigger child is smaller than or equal to the parent, the heap is reordered
                              
                              data[index] = data[swap];
                 }
                 if (index != item) data[index] = temp;
                  
                  if (!item) return;
                  --item;
             }
        }
