#include "include/zmq.h"
#include "include/czmq.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "mqhelper.h"
#include "zhelpers.h"
#include <sicstus/sicstus.h>
#include "sha.h"
#include "hashtrie.h"
#include "prettyprintprolog.h"
#include "queue.h"

#define NOT_INITIALIZED	-1
#define FIRST_WORKER	2
#define N_WORKERS   16
#define TIMEINDEX       8

int hashes = 0, work=0, enq = 0;

static double times[N_WORKERS][TIMEINDEX];

volatile int next_worker_id = FIRST_WORKER;

char  *model = NULL;
int model_len = -1;

char  *fwp = NULL;
int fwp_len = -1;
int minimum_size;
int update_interval;
int state_threshold;

zctx_t *ctx;

/************
 * SWITCHES *
 ************/

volatile int distribute_to_minimum = 1; // toggles distribution strategy
// assumes master_queue_switch is 0
// if not 0, distributes to the worker with the minimum number of packages in queue
// if 0, it uses the digest's first byte modulo number of workers
volatile int terminate_on_first_error = 0;
int central_hashes = 0;

// END SWITCHES

volatile int master_queue_switch = 0; // Set by prolog on startup
// if 1, activate a queue on master
// this overrides any distribute_to_minimum setting


void set_assertion_mode() {
    master_queue_switch = 1;
    distribute_to_minimum = 0;
}


FILE *dest;
#ifdef VERBOSE 
FILE *logfile;
#endif

void *hash_publish, 
*hash_collect, 
*work_publish, 
*work_collect,
*id_response,
*send_ctrl,
*recv_result,
*java_stats;

int queues[N_WORKERS];
int external[N_WORKERS];
int empty_workers[N_WORKERS];

int checkShutdown(void);
int pok (void);

volatile int work_transfer_pending = 0;
int work_sent_to = 2; // root
int work_sent_from = 1;
int last_from_queuesize = -3;

int multiple_calculations = 0;


wQueue *master_queue = NULL;


void print_readable_digest(unsigned char *key, char *dest) {
	sprintf(dest, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",key[0],key[1],key[2],key[3],key[4],key[5],key[6],key[7],key[8],key[9],key[10],key[11],key[12],key[13],key[14],key[15],key[16],key[17],key[18],key[19]);
}

void gen_hash(char *term, int len, char** p) {
	printf("%p %p\n",p,*p);
	char *k = (char*) malloc(20);
    sha1(term, len, k);
	unsigned char *key = (unsigned char*) k;
	char *dig = (char*) malloc(41);
        print_readable_digest(key, dig);
	printf("5\n");
	free(key);
	printf("%p %p %p %s\n",p,*p, dig, dig);
	*p = dig;
}

// puts hashes and transfers them from in to out
int put_hashes(zmsg_t *in, zmsg_t *out) {
    int len = zmsg_size(in);
    assert(len >= TIMEINDEX + 3);
    int i = 0;

    // iterate over hashes
    for (; i < len - TIMEINDEX - 5; i++) {
        zframe_t *frame = zmsg_pop(in);
        assert(zframe_size(frame) == 21);
        char *digest = zframe_strdup(frame);
        #ifdef VERBOSE
          char *readable = (char*) malloc(41); // 41 copied from gen_hash
          print_readable_digest((unsigned char*)digest, readable);
          fprintf(logfile, "received hash: %s\n", readable);
          free(readable);
        #endif
        zmsg_addmem(out, digest, 21);
        zframe_destroy(&frame);
        // first entry is the calculcated hash
        // the following hashcodes are successors
        if (i == 0 && contains_processed(digest + 1)) {
            multiple_calculations++;
        }
        put(digest);
        free(digest);              
    }
    int checked_states = get_checked_states_count();
    if (state_threshold != -1 && checked_states >= state_threshold) {
        printf("Checked %d states.\n", checked_states);
        return -1;        
    }
        
    return i;
}


void dump_statistics(int worker, int qs, zmsg_t *in) {
    // runs
    char *runs = zmsg_popstr(in);
    fprintf(dest, "%d; %s; %d", worker, runs, qs);
    free(runs);
    
    // time statistics
    assert(zmsg_size(in) == TIMEINDEX + 3);

    int i;
    for (i = 0; i < TIMEINDEX; i++) {
        char *pl = zmsg_popstr(in);
        double d;
        sscanf(pl, "%lf", &d);
        fprintf(dest, "; %.3lf", d);
        times[worker][i] = d;
        free(pl);
    }
    char *timestamp = zmsg_popstr(in);
    fprintf(dest, "; %s", timestamp);
    free(timestamp);
    
    char *work_hard_runs = zmsg_popstr(in);
    fprintf(dest, "; %s", work_hard_runs);
    free(work_hard_runs);

    char *digest = zmsg_popstr(in);
    fprintf(dest, "; %s", digest);
    free(digest);
    fprintf(dest, "\n");
}

void send_java_stats(zmsg_t *in) {
    zmsg_t *dup = zmsg_dup(in);
    int len = zmsg_size(dup);
    int i;
    for (i = 0;  i < len - TIMEINDEX - 5; i++) {
        zframe_t *frame = zmsg_pop(dup);
        zframe_destroy(&frame);
    }
    
    zmsg_send(&dup, java_stats);
    zmsg_destroy(&dup);
}

void send_workpackage(int worker) {
    zmsg_t *msg = zmsg_new();
    char *id = int2string(worker);
    zmsg_addstr(msg, id);
    free(id);
    int send = 0;

    while (!is_empty(master_queue)) {
        tCell *item = dequeue(master_queue);
//        char *pretty = pretty_print(item->term, item->term_length);
//        printf("%s --> %d\n", pretty, worker);
        if (!contains_processed(item->digest)) {
            int len = item->term_length;
            char *content = (char*) malloc(20 + len);
            memcpy(content, item->digest, 20);
            memcpy(content+20, item->term, len);
    
            zmsg_addmem(msg, content, 20 + len);
            free(content);
            send = 1;
            break;
        }
        free(item->digest);
        free(item->term);
        free(item);
    }
//    printf("send: %d, size: %d\n", send, (int) zmsg_size(msg));

    if (send) {
        int rc = zmsg_send(&msg, work_publish);
        assert(rc == 0);
        empty_workers[worker] = 0;
    }
    zmsg_destroy(&msg);


}

void check_workers() {
    int i;
    for (i = FIRST_WORKER; i < N_WORKERS; i++) {
        if (is_empty(master_queue) || queues[i] == NOT_INITIALIZED) {
            break;
        }
//        printf("%d is empty: %d\n", i, empty_workers[i]);
        if (empty_workers[i]) {
            send_workpackage(i);
            //empty_workers[i] = 0;
        }
    }
}

int h_hash (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    zmsg_t *in = zmsg_recv(hash_collect);

    send_java_stats(in);

    zmsg_t *out = zmsg_new();

    // this fills the outgoing msg
    int send = put_hashes(in, out);
    // case send == -1: checked states threshold reached, terminate now
    if (send == -1) {
        zmsg_destroy(&in);
        zmsg_destroy(&out);
        return -1;
    }

    // ID + Q
    char *string = zmsg_popstr(in);
    int worker;
    int qs;
    int external_size;
    sscanf(string,"%i %i %i",&worker,&qs, &external_size);
    free(string);
    queues[worker] = qs;
    external[worker] = external_size;
//    printf("w: %d, qs: %d\n", worker, qs);

    dump_statistics(worker, qs, in);
    
    // hashes to send?
    if (send && !central_hashes) {
        zmsg_send(&out, hash_publish);
    } else {
    //    printf("%d is empty\n", worker);
        empty_workers[worker] = 1; // no hashes <=> empty queue
    }
    zmsg_destroy(&out);
    zmsg_destroy(&in);

    if (master_queue_switch) {
        check_workers();
    }

    if (checkShutdown())
        return -1;

    if (queues[FIRST_WORKER] == -2) { pok(); }
    return 0;
}

int pok (void) {
    void *context = zmq_init (1);
    void *back = zmq_socket (context, ZMQ_DEALER);
    
    #ifdef VERBOSE
    fprintf(logfile, "sending pok\n");
    #endif
    zmq_connect (back, "tcp://localhost:5003");
    
    char *digest = (char*) malloc(20);
    
    assert (fwp != NULL);
    sha1(fwp, fwp_len, digest);
    
    char wp[20+fwp_len];
    memcpy(wp, digest, 20);
    put_local(digest);
    memcpy(wp + 20, fwp, fwp_len);

    zmsg_t *message = zmsg_new();  
    zmsg_addmem(message, wp, 20+fwp_len);
    zmsg_pushstr(message, "2");
    zmsg_send(&message, back);

    //sleep(3);
    empty_workers[2] = 0;
    
    zmq_close (back);
    zmq_term (context);
    
    return 0;
}


// if a work transfer is completed, the receiver answers with an "ACK" and his new queuesize
int handle_ACK(char *to, zmsg_t *msg) {
    free(to);
    zframe_t *toframe = zmsg_pop(msg);
    zframe_destroy(&toframe);

    // next work transfer is allowed
    work_transfer_pending = 0;
    // update queuesize
    zframe_t *queue = zmsg_pop(msg);
    char *data = zframe_strdup(queue);
    int worker;
    int queuesize;
    int external_size;
    sscanf(data, "%d %d %d", &worker, &queuesize, &external_size);
    queues[worker] = queuesize;
    external[worker] = external_size;
    free(data);
    zframe_destroy(&queue);
    zmsg_destroy(&msg);

    // check queues, if we are done
    if (checkShutdown())
        return -1;
    return 0;
}



/*****************************
 * work balancing strategies *
 *****************************/


// if workers are not in local_enqueuing mode,
// i.e. every workpackage is sent to master,
// master sends packages to minimum
void minimum_distribution_not_local(char *to, zmsg_t *msg) {
    zframe_t *toframe =  zmsg_first(msg);
    toframe = zmsg_pop(msg);
    zframe_destroy(&toframe);
    // find minimum
    int i = FIRST_WORKER + 1, min = queues[FIRST_WORKER], min_id = FIRST_WORKER;
    for (; i < N_WORKERS; i++) {
        if (queues[i] == -1)
            break;
        if (queues[i] < min)
            min_id = i;
    }
    
    char *newto = int2string(min_id);
    zmsg_pushmem(msg, newto, strlen(newto));
    #ifdef VERBOSE
    fprintf(logfile, "send work to %d\n", min_id);
    #endif
    free(newto);
    
    free(to);
    if (msg != NULL) {
        zmsg_send(&msg, work_publish);       
    }
    zmsg_destroy(&msg);
}

void hash_based_distribution(zmsg_t *msg) {
    zframe_t *toframe = zmsg_pop(msg);
    zframe_destroy(&toframe);
    //int nmbr_workers = next_worker_id - 2;
    int nmbr_workers = 0;
    int i;
    for (i = FIRST_WORKER; i < N_WORKERS; i++) {
        if (queues[i] != NOT_INITIALIZED) {
            nmbr_workers++;
        }
    }
    zmsg_t *msgs[nmbr_workers];
    int *count = (int*) malloc(sizeof(int) * nmbr_workers);
    for (i = 0; i < nmbr_workers; i++) {
        count[i] = 0;
        msgs[i] = zmsg_new();
        char *id = int2string(i+2);
        zmsg_pushmem(msgs[i], id, strlen(id));
        free(id);
    }
    zframe_t *frame;
    while ((frame = zmsg_pop(msg)) != NULL) {
        assert(zframe_size(frame) > 20);
        byte *data = zframe_data(frame);
        int x = data[0] % nmbr_workers;
        if (!contains_processed((char*) data)) { 
            zmsg_add(msgs[x], frame);
        }
        count[x]++;
    }
    for (i = 0; i < nmbr_workers; i++) {
        if (count[i] > 0) {
            if (queues[i+2] == -2) {
                printf("late join\n");
                char *memblock = iterate();
                int states = count_elements();
                zmsg_addmem(msgs[i], "known hashes", 13);
                zmsg_addmem(msgs[i], memblock, states * 21);
                free(memblock);
            }
            zmsg_send(&msgs[i], work_publish);
        }
        zmsg_destroy(&msgs[i]);
    }
    free(count);
}

// work balancing strategie in use
void regular_balancing(char *to, zmsg_t *msg) {
   // printf("Sent: %d\n",zmsg_size(msg));
    int to_int = atoi(to);
    assert (to_int >= FIRST_WORKER);
    assert (to_int <= N_WORKERS);
    free(to);
    if (msg != NULL) {
        if (queues[to_int] == -2) {
            // late joining
            char *memblock = iterate();
            int states = count_elements();
            zmsg_addmem(msg, "known hashes", 13);
            zmsg_addmem(msg, memblock, states * 21);
            free(memblock);
        }
        // iterate over hashtrie
        // add memory block
        zmsg_send(&msg, work_publish);
    }
    zmsg_destroy(&msg);
}

void enqueue_single_wp(char *str, zframe_t *frame) {
    char *digest = (char*) malloc(20);
    int wplen = zframe_size(frame) - 20;
    char *term = (char*) malloc(wplen);
    
    memcpy(term, str + 20, wplen);
    memcpy(digest, str, 20);   
    enqueue(master_queue, term, wplen, digest, 0, 0);
    free(str);
    zframe_destroy(&frame);
}

void enqueue_wps(char *to, zmsg_t *msg) {
    free(to);
    zframe_t *toframe =  zmsg_pop(msg);
    zframe_destroy(&toframe);
    int i;
    int len = zmsg_size(msg);

    for (i = 0; i < len; i++) {
        zframe_t *frame = zmsg_pop (msg);
        char *str = zframe_strdup(frame);
        enqueue_single_wp(str, frame);
    }
}

/*********************************
 * end work balancing strategies *
 *********************************/


// handle incoming work
int h_work (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    work++;
    // Receive and forward
    zmsg_t *msg = zmsg_recv(work_collect);
    
    zframe_t *fromframe = zmsg_pop (msg);
    zframe_destroy(&fromframe);

    
    zframe_t *toframe =  zmsg_first(msg);
    char *to = zframe_strdup(toframe);
    
     if (strcmp(to, "done.") == 0) {
        work_transfer_pending = 0;
        return 0;
     }

    if (strcmp(to, "0") == 0) {
        if (master_queue_switch) {
            enqueue_wps(to, msg);
            check_workers();
        }
        else if (distribute_to_minimum) {
            minimum_distribution_not_local(to, msg);
        }
        else {
            free(to);
            hash_based_distribution(msg);
        }
    } else if (0 == strcmp(to, "ACK")) {
        // need to check for shutdown here
        return handle_ACK(to, msg);
    }
    else {
        regular_balancing(to, msg);
    }
    return 0;
}

int h_id (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    char *string = zstr_recv (id_response);
    if (string != NULL) {
        int len = ((int) log10(next_worker_id+1)) + 1;
        
        if (next_worker_id >= N_WORKERS) {
            printf("We only support %i workers.",N_WORKERS);
            return -1;
        }
        
        char *id = (char*) malloc(len + 1);
        char *mque = (char*) malloc(2);

        sprintf(id, "%d", next_worker_id);
        sprintf(mque, "%d", master_queue_switch);

        queues[next_worker_id] = NOT_INITIALIZED;
        

        zmsg_t *msg =  zmsg_new();
        zmsg_addmem(msg, id, strlen(id) + 1);
        zmsg_addmem(msg, model, model_len);
        zmsg_addmem(msg, mque,strlen(mque));
        zmsg_addstr(msg, "%d", central_hashes);
        zmsg_send(&msg, id_response);
        zmsg_destroy(&msg);
        
        free(id);
        free(mque);
        next_worker_id++;
    }
    free(string);
//    if (master_queue_switch) {
//        s_sleep(200); // trying to avoid package loss
//    }
    return 0;
}

int h_results (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    zmsg_t *msg = zmsg_recv(recv_result);

    zframe_t *cf = zmsg_pop (msg); // remove controlframe
    zframe_destroy(&cf);

    while(0 < zmsg_size(msg)) {
        zframe_t *frame = zmsg_pop (msg);
        int size = zframe_size(frame);
        char *txt = zframe_strdup(frame);
//        printf("Error: %s \n",txt);
        char *pretty = pretty_print(txt, size);
        printf("Error: %s\n", pretty);
        free(pretty);
        zframe_destroy(&frame);
        free(txt);
        if (terminate_on_first_error) {
            // TODO: provide more information
            return -1;
        }
    }
    zmsg_destroy(&msg);
    return 0;
}

void print_queuesizes() {
    int i;
    for(i = 2; i < N_WORKERS; i++) {
        printf("(%i:(%i, %i)) ",i,queues[i], external[i]);
    }
    printf("\n");
}    

void transfer_work(int from, int to, int amount) {
    printf("I demand that %i sends %i workpackages to %i\n", from, amount,to);
    assert(amount >= minimum_size);
    work_sent_to = to;
    work_sent_from = from;
    // only one work transfer at a time is allowed
    work_transfer_pending = 1;
    char *sfrom, *sto, *samount;
    
    sfrom = int2string(from);
    sto = int2string(to);
    samount = int2string(amount);
    
    s_sendmore(work_collect,sfrom);
    s_sendmore(work_collect, sto);
    s_send(work_collect, samount);

    #ifdef VERBOSE 
    fprintf(logfile, "I demand that %i sends %i workpackages to %i\n", from, amount,to);
    #endif
    free(sto);
    free(sfrom);
    free(samount);
}

int checkShutdown(void) {
    int known = count_elements();
    int checked = get_checked_states_count();

    if (checked == 0) {
        return 0;
    }
    if (checked > state_threshold && state_threshold > 0) {
        return 1;
    }
    if (checked == known) {
        return 1;
    }
    return 0;
}

void print_times() {
    int i, j;
    for (i = 0; i < N_WORKERS; i++) {
        for (j = 0; j < TIMEINDEX; j++) {
            printf("%f ", times[i][j]);
            
        }
        printf("\n");
    }
}

void *print_stats(void *arg) {
    int p = 0;
    while(1) {
        if (queues[work_sent_from] != last_from_queuesize) {
            work_sent_from = 1;
        }
        if (p++ % 5 == 0) {
            printf("Checked States: %d/%d/%d Workpackages: %i/%i Min-Size: %i Update-Interval: %i\n",
                multiple_calculations, get_checked_states_count(), count_elements(), work,enq,minimum_size,update_interval);
            print_queuesizes();
            #ifdef DEBUG
            print_times();
            #endif
        }
      s_sleep(update_interval);
        
    if (!master_queue_switch) {
            if (next_worker_id - FIRST_WORKER > 1) {
                int need = 0, has = 0;
                int i;
                for(i = FIRST_WORKER; i<next_worker_id; i++) {
                    if (queues[i] + external[i] > queues[has] + external[has]) has = i;
                    if ((queues[i] == 0 && external[i] == 0) || queues[i] == -2) need = i;
                } 
                
                //printf("Stuff: has=%d queues[has]=%d minimum_size=%d need=%d work_transfer_pending=%d last_from_queuesize=%d work_sent_from=%d queues[work_sent_from]=%d\n", 
                //       has,queues[has],minimum_size, need, work_transfer_pending,
                //       last_from_queuesize, work_sent_from, queues[work_sent_from] );
    //
    //print_queuesizes();
    
                // A very scary if statement 
                if (queues[has] + external[has] > minimum_size*2 && need > 0 && !work_transfer_pending && last_from_queuesize != queues[work_sent_from]) { 
                    last_from_queuesize = queues[has];
                    transfer_work(has,need,(queues[has]+external[has])/2);
                }
            }
        }
    }
    return 0;
}


void set_fwp(char *term, int len) {
    fwp_len = len;
    fwp = (char*) malloc(len);
    memcpy(fwp,term,len);
}


void set_model(char *term, int len) {
    model_len = len;
    model = (char*) malloc(len);
    memcpy(model,term,len);
}

void *hash_rep; // move this up?
int h_hashrequest (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    zmsg_t *msg = zmsg_recv(hash_rep);
    assert(zmsg_size(msg) == 1);

    zframe_t *frame = zmsg_pop(msg);
    assert(zframe_size(frame) == 20);

    char *digest = zframe_strdup(frame);
    int status = contains(digest);

    zmsg_t *out = zmsg_new();
    zmsg_addstr(out, "%d", status);

    zmsg_send(&out, hash_rep);
    zmsg_destroy(&out);    

    free(digest);

    zframe_destroy(&frame);

    zmsg_destroy(&msg);
    return 0;
}

void *start_hash_service(void *arg) {
    assert(ctx != NULL);
    
    hash_rep = zsocket_new (ctx, ZMQ_REP);
    zsocket_bind (hash_rep, "tcp://*:5000");

    zloop_t *reactor = zloop_new ();
    zmq_pollitem_t poller1 = { hash_rep, 0, ZMQ_POLLIN };

    zloop_poller (reactor, &poller1, h_hashrequest, NULL);
    zloop_start  (reactor);
    zloop_destroy(&reactor); // will this be called?
    return 0;
}

int wrapped_main(int delay, int len, int maxstates, char* filename) {
    #ifdef VERBOSE
    setId("1");
    #endif
    if (master_queue_switch) {
        master_queue = init_queue();
        assert(master_queue != NULL);
    }

    init_hashmap();

    dest = fopen(filename, "a+");
    #ifdef VERBOSE
    logfile = fopen("master.log", "a+");
    #endif
    assert(dest != NULL);
    update_interval = delay;
    minimum_size = len;
    state_threshold = maxstates;
    printf("Model %s\n",model);
    fprintf(dest, "\"WorkerID\"; \"Runs\"; \"queuesize\"; \"tProlog\"; \"tWorkhard\"; \"tIdle\"; \"tTotal\"; \"tHashing\"; \"tSending\"; \"tReceiving\"; \"tPutSuccessors\";\"timestamp\";\"WorkerRun\";\"hashcode\" \n");
    printf("First WP %s\n",fwp);
    
    //queues = malloc(5*sizeof(int));
    int i;
    for (i = 0; i < N_WORKERS; i++) {
        queues[i] =  NOT_INITIALIZED;
        external[i] = 0;
        empty_workers[i] = 0;
    }
    
    pthread_t stats;
    pthread_create (&stats, NULL, print_stats, NULL);
    
    ctx = zctx_new ();

    pthread_t hash_service;
    if (central_hashes) {
        pthread_create (&hash_service, NULL, start_hash_service, NULL);
    }
    
    if (!central_hashes) {
        hash_publish = zsocket_new (ctx, ZMQ_PUB);
    }
    hash_collect = zsocket_new (ctx, ZMQ_PULL);
    work_publish = zsocket_new (ctx, ZMQ_ROUTER);
    work_collect = zsocket_new (ctx, ZMQ_ROUTER);
    id_response = zsocket_new (ctx, ZMQ_REP);
    send_ctrl = zsocket_new (ctx, ZMQ_PUB);
    recv_result = zsocket_new (ctx, ZMQ_PULL);
    java_stats = zsocket_new (ctx, ZMQ_PUB);
    
    if (!central_hashes) {
        zsocket_bind (hash_publish, "tcp://*:5000");
    }
    zsocket_bind (hash_collect, "tcp://*:5001");
    zsocket_bind (work_publish, "tcp://*:5002");
    zsocket_bind (work_collect, "tcp://*:5003");
    zsocket_bind (id_response, "tcp://*:5005");
    zsocket_bind (send_ctrl, "tcp://*:5007");
    zsocket_bind (recv_result, "tcp://*:5008");
    zsocket_bind (java_stats, "tcp://*:5010");
    
    
    zloop_t *reactor = zloop_new ();
    zmq_pollitem_t poller2 = { hash_collect, 0, ZMQ_POLLIN };
    zmq_pollitem_t poller4 = { work_collect, 0, ZMQ_POLLIN };
    zmq_pollitem_t poller6 = { id_response, 0, ZMQ_POLLIN };
    zmq_pollitem_t poller20 = { recv_result, 0, ZMQ_POLLIN };
    
    zloop_poller (reactor, &poller2, h_hash, NULL);
    zloop_poller (reactor, &poller4, h_work, NULL);
    zloop_poller (reactor, &poller6, h_id, NULL);
    zloop_poller (reactor, &poller20, h_results, NULL);
    zloop_start  (reactor);
    s_send(send_ctrl, "TERM");
    s_send(java_stats, "TERM");
    pthread_cancel(stats);

    if (central_hashes) {
        pthread_cancel(hash_service);
    }

    sleep(1);
    zloop_destroy (&reactor);
    fprintf(dest, "#%s %d %d\n", model + 2, delay, len);    
    fprintf(dest, "#%d workers\n", next_worker_id - 2);
    fprintf(dest, "#----------------------------\n");
    printf("Unique states on master: %d\n\n", teardown());
    #ifdef VERBOSE
    fclose(logfile);
    #endif
    fclose(dest);
    zctx_destroy (&ctx);
    count_elements();
    return 0;
}
