#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

char *buf;
int pos = 0;

char *pretty_print(char *term, int len);
void add_char_to_buf(char x);

void full_print(char *term, int len) {
    int i;
    for (i = 0; i < len; i++) {
        if (term[i] >= 32) {
            printf("%c", term[i]);
        } else {
            //printf(" ");
            printf("'%d'", term[i]);
        }
    }
    printf("\n");
}


int eat(char *term, int len);
int atom(char *term, int len);
int digit(char *term, int len);
int term(char *term, int len);
int list(char *term, int len);
int tuple(char *term, int len);

int left = 1;

char *pretty_print(char *term, int len) {
    buf = (char*) malloc(2*len);
    assert(term[0] == 'D');
    int res = eat(term + 1, len -1);
    assert(len - res == 1);
    add_char_to_buf(0);
    
    char *pretty = (char*) malloc(pos+1);
    memcpy(pretty, buf, pos+1);
    free(buf);
    pos = 0;
    return pretty;
}

int eat(char *termi, int len) {
    char c = termi[0];
    if (c == 'A') {
        return atom(termi+1, len) + 1;
    } else if (c == 'S') {
        return term(termi+1, len) + 1;
    } else if (c == 'I') {
        return digit(termi+1, len) + 1;
    } else if (c == '[') {
        return list(termi+1, len) + 1;
    } else if (c == ']') {
        add_char_to_buf('[');
        add_char_to_buf(']');
       return 1;
    } else if (c == 0) {
        //printf(".\n");
    }
    printf("unknown char %c %d\n", c, c);
    return -1;
}

int atom(char *term, int len) {
    int i;
    for (i = 0; i < len; i++) {
        if (term[i] == 0) {
            break;
        }
        add_char_to_buf(term[i]);
    }
    return i + 1;
}

int tuple(char *term, int len) {
    int tmpleft = left;
    if (tmpleft)
        add_char_to_buf('(');

    left = 1;
    int first = eat(term + 3, len - 3);
    add_char_to_buf(',');
    left = 0;
    int second = eat(term + 3 + first, len - first - 3);

    if (tmpleft)
        add_char_to_buf(')');
    left = 1;
    return first + second + 3;
}


int term(char *term, int len) {
    int functor = 2;
    if (term[0] == ',' && term[1] == 0 && term[2] == 2) {
        return tuple(term, len);
    } else {
        functor = atom(term, len);
    }
    int arity = term[functor];
    int offset = functor + 1;
    int i;
    add_char_to_buf('(');
    for (i = 0; i < arity; i++) {
        int x = eat(term + offset, len - offset);
        offset += x;
        if (i != arity - 1)
            add_char_to_buf(',');
    }
    add_char_to_buf(')');
    return offset ;
}

int digit(char *term, int len) {
    int i;
    for (i = 0; i < len; i++) {
        if (term[i] == 0) {
            break;
        }
        assert(term[i] >= '0' && term[i] <= '9');
        add_char_to_buf(term[i]);
    }
    return i + 1;
}

int list(char *term, int len) {
    add_char_to_buf('[');
    int offset = 0;
    while (offset <= len) {
        int x = eat(term + offset, len-offset);
        offset += x;
        if (x == -1) {
            return -1;
        }
        //printf("<%c>\n", term[offset]);
        assert(term[offset] == '[' || term[offset] == ']');
        if (term[offset] == ']')
            break;
        offset++;
        add_char_to_buf(',');
    }

    add_char_to_buf(']');
    return offset + 1;
}

void add_char_to_buf(char x) {
    buf[pos] = x;
    pos++;
}


/*
[Sbind'0''2'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I1'0'AUSER'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_response'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I1'0'ARESP'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_card'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I2'0'ACARD'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_state'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I8'0'ASESSION_STATE'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_request'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I2'0'ASESSION_REQUEST'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Auser_hotel_bookings'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AUSER'0'Sfd'0''2'I1'0'AHOTEL'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Auser_rental_bookings'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AUSER'0'Sfd'0''2'I3'0'ACAR_RENT'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Arooms_hotel'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I2'0'AROOM'0'Sfd'0''2'I1'0'AHOTEL'0'Atrue'0'I1'0'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AROOM'0'Sfd'0''2'I1'0'AHOTEL'0'Atrue'0'I0'0'Aempty'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I3'0'AROOM'0'Sfd'0''2'I2'0'AHOTEL'0'Atrue'0'I1'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I4'0'AROOM'0'Sfd'0''2'I2'0'AHOTEL'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Acars_rental'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I2'0'ACAR'0'Sfd'0''2'I1'0'ACAR_RENT'0'Atrue'0'I1'0'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ACAR'0'Sfd'0''2'I1'0'ACAR_RENT'0'Atrue'0'I0'0'Aempty'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I3'0'ACAR'0'Sfd'0''2'I2'0'ACAR_RENT'0'Atrue'0'I1'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I4'0'ACAR'0'Sfd'0''2'I2'0'ACAR_RENT'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Aglobal_room_bookings'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AROOM'0'Sfd'0''2'I1'0'AUSER'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Aglobal_car_bookings'0']]

[bind(avl_set(node((fd(1,SESSION),fd(1,USER)),true,0,empty,empty))),bind(session_response,avl_set(node((fd(1,SESSION),fd(1,RESP)),true,0,empty,empty))),bind(session_card,avl_set(node((fd(1,SESSION),fd(2,CARD)),true,0,empty,empty))),bind(session_state,avl_set(node((fd(1,SESSION),fd(8,SESSION_STATE)),true,0,empty,empty))),bind(session_request,avl_set(node((fd(1,SESSION),fd(2,SESSION_REQUEST)),true,0,empty,empty))),bind(user_hotel_bookings,avl_set(node((fd(1,USER),fd(1,HOTEL)),true,0,empty,empty))),bind(user_rental_bookings,avl_set(node((fd(1,USER),fd(3,CAR_RENT)),true,0,empty,empty))),bind(rooms_hotel,avl_set(node((fd(2,ROOM),fd(1,HOTEL)),true,1,node((fd(1,ROOM),fd(1,HOTEL)),true,0,empty,empty),node((fd(3,ROOM),fd(2,HOTEL)),true,1,empty,node((fd(4,ROOM),fd(2,HOTEL)),true,0,empty,empty))))),bind(cars_rental,avl_set(node((fd(2,CAR),fd(1,CAR_RENT)),true,1,node((fd(1,CAR),fd(1,CAR_RENT)),true,0,empty,empty),node((fd(3,CAR),fd(2,CAR_RENT)),true,1,empty,node((fd(4,CAR),fd(2,CAR_RENT)),true,0,empty,empty))))),bind(global_room_bookings,avl_set(node((fd(2,ROOM),fd(1,USER)),true,0,empty,empty))),bind(global_car_bookings,[])]))

DSinvariant_violated'0''1'Sstate'0''1'[Sbind'0''2'Asession'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I1'0'AUSER'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_response'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I1'0'ARESP'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_card'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I2'0'ACARD'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_state'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I8'0'ASESSION_STATE'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Asession_request'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ASESSION'0'Sfd'0''2'I2'0'ASESSION_REQUEST'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Auser_hotel_bookings'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AUSER'0'Sfd'0''2'I1'0'AHOTEL'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Auser_rental_bookings'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AUSER'0'Sfd'0''2'I3'0'ACAR_RENT'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Arooms_hotel'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I2'0'AROOM'0'Sfd'0''2'I1'0'AHOTEL'0'Atrue'0'I1'0'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AROOM'0'Sfd'0''2'I1'0'AHOTEL'0'Atrue'0'I0'0'Aempty'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I3'0'AROOM'0'Sfd'0''2'I2'0'AHOTEL'0'Atrue'0'I1'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I4'0'AROOM'0'Sfd'0''2'I2'0'AHOTEL'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Acars_rental'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I2'0'ACAR'0'Sfd'0''2'I1'0'ACAR_RENT'0'Atrue'0'I1'0'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'ACAR'0'Sfd'0''2'I1'0'ACAR_RENT'0'Atrue'0'I0'0'Aempty'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I3'0'ACAR'0'Sfd'0''2'I2'0'ACAR_RENT'0'Atrue'0'I1'0'Aempty'0'Snode'0''5'S,'0''2'Sfd'0''2'I4'0'ACAR'0'Sfd'0''2'I2'0'ACAR_RENT'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Aglobal_room_bookings'0'Savl_set'0''1'Snode'0''5'S,'0''2'Sfd'0''2'I1'0'AROOM'0'Sfd'0''2'I1'0'AUSER'0'Atrue'0'I0'0'Aempty'0'Aempty'0'[Sbind'0''2'Aglobal_car_bookings'0']]
    
    
    
*/


