#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "queue.h"


wQueue *init_queue() {
    wQueue *new = malloc(sizeof(wQueue));
    new->first = NULL;
    new->last = NULL;
    new->c = 0;
    return new;
    
}

void print_term(tCell *t) {
	int i;
	for (i=0;i<t->term_length;i++) {
		printf("%c",(t->term)[i]);
	}
	printf("\n");
}

// TODO: use this in enqueue_fifo
void enqueue_cell(wQueue *q, tCell *t) {
    q->c++;
    if (q->first == NULL) {
        q->first = t; 
        q->last = t;
        q->first->next = NULL;
    }
    else { 
        tCell *temp = q->last; // for multithreading
        q->last = t;
        temp->next = t;
    }
}

void enqueue(wQueue *q, char *term, int len, char digest[20], int thresh, int id) {
    int r = rand() % 100;
#ifndef LOCAL
    id = 0;
#endif
    if (r < thresh) {
        enqueue_lifo(q, term, len, digest, id);
    }
    else {
        enqueue_fifo(q, term, len, digest, id);
    }
}

#ifdef DEBUG
void enqueue_debug(wQueue *q, char *term, int len, char digest[20]) {
    q->c++;
    // ignore if in Hashmap 
    // else add to working queue
    tCell *new = malloc(sizeof(tCell));
    
    new->term = malloc(len);
    memcpy(new->term, term, len);
    new->term_length = len;
    new->digest = malloc(20);
    memcpy(new->digest, digest, 20);
    new->next = NULL;
    
    //  printf("queued: %s %s\n",term,digest);
    
    
    
    if (q->first == NULL) {
        q->first = new; 
        q->last = new;
        q->first->next = NULL;
    }
    else { 
        tCell *temp = q->last; // for multithreading
        q->last = new;
        temp->next = new;
    }
        
}

void clear_debug(wQueue *q) {
    while (q->first != NULL) {
        tCell *ele = dequeue(q);
        free(ele->term);
        free(ele->digest);
    }
    free(q);
}
#endif

void enqueue_fifo(wQueue *q, char *term, int len, char digest[20], int id) {
    q->c++;
    // ignore if in Hashmap 
    // else add to working queue
    tCell *new = malloc(sizeof(tCell));
    new->term = term;
    new->term_length = len;
    new->digest = digest;
    new->next = NULL;
    new->id = id;
    
    //	printf("queued: %s %s\n",term,digest);
    
    
    
    if (q->first == NULL) {
        q->first = new; 
        q->last = new;
        q->first->next = NULL;
    }
    else { 
        tCell *temp = q->last; // for multithreading
        q->last = new;
        temp->next = new;
    }
    
}

void enqueue_lifo(wQueue *q, char *term, int len, char digest[20], int id) {
    q->c++;
    // ignore if in Hashmap 
    // else add to working queue
    tCell *new = malloc(sizeof(tCell));
    new->term = term;
    new->term_length = len;
    new->digest = digest;
    new->next = q->first;
    new->id = id;
    q->first = new;
    if (q->last == NULL)
        q->last = new;
}

tCell *dequeue(wQueue *q) {
    q->c--;
    tCell *res = q->first;
    q->first = q->first->next;
    if (q->first == NULL) { q->last = NULL; }
    res->next = NULL;
    return res;
}

int is_empty(wQueue *q) {
    return (q->first == NULL);
}

int q_size(wQueue *q) {
    return q->c;
}


void print_queue(wQueue *q) {
    printf("Queue: [");
    tCell *p = q->first;
    //	if (first != NULL) printf("%s %i | ", first->term, c);
    while (p != NULL) {
        printf("%s ",p->term);
        //		print_key(p->digest);
        //		printf("\n");
        p = p->next;
    }
    printf("]\n");
}

void teardown_queue(wQueue **q) {
    while((*q)->c > 0) {
        tCell *cell = dequeue(*q);
        free(cell->term);
        free(cell->digest);
        free(cell);
    }
    free(*q);
}

/* returns 0 if two different terms share the same digest */
#ifdef DEBUG
int check_debugq(wQueue *q, char digest[20], char *term, int len) {
    tCell *cell = q->first;
    while (cell != NULL) {
        if (memcmp(digest, cell->digest, 20) == 0 && (cell->term_length != len || memcmp(term, cell->term, len)))
            return 0;
        cell = cell->next;
    }
    return 1;
}
#endif
