#include "include/zmq.h"
#include <stdio.h>
#include <string.h>
#include "include/czmq.h"
#include "hashtrie.h"
#include "queue.h"
//#include "heap.h"
#include "sha.h"

#include "zhelpers.h"
#include "mqhelper.h"
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <sicstus/sicstus.h>
#include "external.h"

#define BFS -1
#define DFS 1000

//#define THRESH  BFS
int THRESH = -1;

#ifdef VERBOSE
FILE *logfile;
FILE *toprolog;
FILE *fromprolog;
#endif

FILE *statespace;

enum timer_index {tidx_prolog=0, tidx_work_hard_incl_prolog, tidx_work_hard_incl_empty_queue,
                  tidx_overall, tidx_hashcodes, tidx_workrequests, tidx_incoming, tidx_successors,
                  TIMEINDEX};
/*
 * INDX FUNCTION
 * 0    prolog time (checking states)
 * 1    work_hard (incl. prolog time)
 * 2    work_hard (incl. empty queue)
 * 3    overall
 * 4    hashcodes
 * 5    workrequests
 * 6    incoming work
 * 7    put successors
 */
static double sutimes[TIMEINDEX];
int runs_since_update = 0;
static double sumtimes[TIMEINDEX];
static struct timeval starttimes[TIMEINDEX];

int global_starttime;

void work_hard(void);
void print_stats(zmsg_t *msg, tCell *t);
void add_digest_processed(zmsg_t *msg, char digest[20]);
void init_zmq();
void print_time();
void add_queue_stats(zmsg_t *msg);

char *masteraddr;

void  *recv_hashes, *send_hashes, *recv_work, *send_work, *recv_ctrl, *send_result, *recv_tick, *send_tick, *id_req, *hash_req;
char *id;
zloop_t *reactor;

zmsg_t *successors = NULL;
zmsg_t *results = NULL;
char currentDigest[20];

wQueue *local_queue;
wQueue *debug_queue;

int count = 0;
int hit=0, cache=0, cache_hit = 0;

volatile int running = 1;

/************
 * SWITCHES *
 ************/

//int local_enqueueing = 1; // 1 = keep new tasks , 0 = send new tasks to master

// END SWITCHES

int local_hashes = 0; // set by master
int max_states_reached = 0; // set via master control (?)
int max_queue_size = -1;

volatile int local_enqueueing = 1; // Will be set by master
        //1 = keep new tasks , 0 = send new tasks to master



void disable_localqueing() {
    local_enqueueing = 0;
    printf("Disabled local successors queueing.\n");
}


int countRuns = 0;

int transitions = 0;
int has_results = 0;
int is_processing = 0;
int empty_queue = 1;

void start_timer(enum timer_index index) {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    starttimes[index] = tv;
}

double thrash = 0;

static SP_pred_ref pred_zmq_check, pred_local_zmq_check, pred_load_model;

void print_readable_digest(unsigned char *key, char *dest) {
	sprintf(dest, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",key[0],key[1],key[2],key[3],key[4],key[5],key[6],key[7],key[8],key[9],key[10],key[11],key[12],key[13],key[14],key[15],key[16],key[17],key[18],key[19]);
}



char *get_digest_from_term(char *term, int len) {
    char d[20]; // sha1 returns 20 byte hash
    sha1(term, len, d);
    char *res = malloc(41); // 40 byte readabe digest + null termination
    print_readable_digest((unsigned char*) d, res);
    return res;
}
// should be called after get_digest_from_term to avoid leaking memory
void free_digest(char *readable) {
    free(readable);
}




void stop_timer(enum timer_index index) {
    struct timeval startv = starttimes[index];
    suseconds_t time1 = startv.tv_usec;
    time_t sectime1 = startv.tv_sec;
    struct timeval tv;
    gettimeofday(&tv, NULL);
    suseconds_t time2 = tv.tv_usec;
    time_t sectime2 = tv.tv_sec;
/*    if (index == 0) {
        sutimes[index] += thrash;
        thrash = (double) (time2 - time1)/1000 + (double)(sectime2 - sectime1) * 1000;    
    }
    else*/
        sutimes[index] += (double) (time2 - time1)/1000 + (double)(sectime2 - sectime1) * 1000;    
}

int h_hash(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    start_timer(tidx_hashcodes);
    const int max = 20;
    int count = 0;
    while (zsocket_poll(recv_hashes, 1) == TRUE && count < max) {
        zmsg_t *hashes = zmsg_recv(recv_hashes);
        zframe_t *frame;
        while ((frame = zmsg_pop(hashes)) != NULL) {
            assert(zframe_size(frame) == 21);
            char *digest = zframe_strdup(frame);
            #ifdef VERBOSE
            char *readable = (char*) malloc(41);
            print_readable_digest((unsigned char*) digest, readable);
            fprintf(logfile, "Incoming hash: %s\n", readable);
            free(readable); 
            #endif
            if (local_hashes) {
                put(digest);
            }
            free(digest);
            zframe_destroy(&frame);
        }
        count++;
        zmsg_destroy(&hashes);
    }


    stop_timer(tidx_hashcodes);
    return 0; 
}

int h_tick (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    char *z = s_recv(recv_tick);
    //    printf("R: %s\n",z);
    if (z != NULL) {
        work_hard(); 
    }
    free(z);
    return 0;
}


int req_contains(char *digest) {
    zmsg_t *out = zmsg_new();
    zmsg_addmem(out, digest, 20);
    zmsg_send(&out, hash_req);
    zmsg_destroy(&out);


    zmsg_t *in = zmsg_recv(hash_req);
    zframe_t *frame = zmsg_pop(in);

    char *reply = zframe_strdup(frame);
    assert(reply);
    int i = 4;
    sscanf(reply, "%d", &i);

    free(reply);
    zframe_destroy(&frame);
    zmsg_destroy(&in);
    return i;
}

int req_contains_processed(char *digest) {
    return req_contains(digest) == 1;
}

void append_wp(zmsg_t *wpmsg) {
    tCell *item = NULL;
    if (!is_empty(local_queue)) {
        item = dequeue(local_queue); 
    } else {
        item = deserialize_package();
    }
    assert(item);
    #ifdef VERBOSE
    char *readable = (char*) malloc(41);
    print_readable_digest((unsigned char*) item->digest, readable);
    fprintf(logfile, "Sending wp: %s\n", readable);
    free(readable); 
    #endif
    if (local_hashes) {
        if (!contains_processed(item->digest)) {
            wp_sendmore(wpmsg,item->term,item->term_length,item->digest);
        }
    } else {
        if (!req_contains_processed(item->digest)) {
            wp_sendmore(wpmsg,item->term,item->term_length,item->digest);
        }
    }
    free(item->term);
    free(item->digest);
    free(item);
}

int h_workrequest (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    // get message from send_work
    // first frame = target
    // second frame = amount
    // dequeue amount workitems
    // send them via send_work
    start_timer(tidx_workrequests);
    #ifdef VERBOSE
    fprintf(logfile, "Incoming work request\n");
    #endif
    
    zmsg_t *msg = zmsg_recv(send_work);

    if (is_empty(local_queue)) {
       zmsg_send(send_work,"done.");
       zmsg_destroy(&msg);
       return 0;
    }  

    zframe_t *target_frame = zmsg_pop(msg);
    zframe_t *amount_frame = zmsg_pop(msg);
    char *target, *a;
    target = zframe_strdup(target_frame);

    s_sendmore(send_work,target);
    free(target);
    
    a = zframe_strdup(amount_frame);
    int amount = atoi(a);  
    
    zmsg_t *wpmsg = zmsg_new();


    for(;amount > 0 && (!is_empty(local_queue) || get_external_wp_count() > 0); amount--) {
        append_wp(wpmsg);
    }

    zmsg_send(&wpmsg, send_work);
    zmsg_destroy(&wpmsg);
    zframe_destroy(&target_frame);
    zframe_destroy(&amount_frame);
    zmsg_destroy(&msg);
    
    stop_timer(tidx_workrequests);
    return 0;
}

void tick(void) {
    s_send(send_tick,"TOCK");
}

suseconds_t time1;
time_t sectime1;

void put_late_join_hashes(char *str, zframe_t *frame, zmsg_t *msg) {
    free(str);
    zframe_destroy(&frame);
    frame = zmsg_pop(msg);
    str = zframe_strdup(frame);
    // root state will be hashed by master
    // so if exactly one hash is calculated, the worker is not late
    int sze = zframe_size(frame);
    if (sze != 21) {
        printf("Sorry, I'm late. But I got %d hashes.\n", sze/21);
    }
    assert(sze % 21 == 0);
    int j;
    if (local_hashes) {
        for (j = 0; j < (sze / 21); j++) {
            put(str + j*21);
        }
    }
    free(str);
    zframe_destroy(&frame);
}

void enqueue_wps(char *str, zframe_t *frame) {
    char *digest = (char*) malloc(20);
    int wplen = zframe_size(frame) - 20;
    char *term = (char*) malloc(wplen);
    
    memcpy(term, str + 20, wplen);
    memcpy(digest, str, 20);   
    //put_local(digest);
    #ifdef VERBOSE
    char *readable = (char*) malloc(41);
    print_readable_digest((unsigned char*) digest, readable);
    fprintf(logfile, "Enqueuing wp: %s\n", readable);
    free(readable); 
    #endif
    enqueue(local_queue, term, wplen, digest, THRESH, 0);
   // print_queue(local_queue);
   // printf("Q: %s %d\n",term,wplen);
    free(str);
    zframe_destroy(&frame);
}

void send_ack() {
    zmsg_t *ack = zmsg_new();
    zmsg_addstr(ack, "ACK");  
    add_queue_stats(ack);
    zmsg_send(&ack, send_work);
    zmsg_destroy(&ack);
}

int h_work (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    start_timer(tidx_incoming);
    if (countRuns == 0) {
        printf("Starting to work.\n");
        start_timer(tidx_overall);
    }
    
    // receive wp message  
    // iterate over frames
    // deserialize content and enque
    // send tock
    
    zmsg_t *msg = zmsg_recv(recv_work);
    
    #ifdef VERBOSE
    fprintf(logfile, "Incoming work packages\n");
    #endif
    int len = zmsg_size(msg);
    int i;
    
    for (i = 0; i < len; i++) {
        zframe_t *frame = zmsg_pop (msg);
        char *str = zframe_strdup(frame);
        if (i == len - 2) {
            // late joining - refactor me.
            if (0 == strcmp(str, "known hashes")) {
                put_late_join_hashes(str, frame, msg);
                break;
            }
        }
        enqueue_wps(str, frame);
    }
    zmsg_destroy(&msg);

    send_ack();
    tick();
    
    stop_timer(tidx_incoming);
    return 0;
}

int h_control(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    zmsg_t *msg = zmsg_recv(recv_ctrl);
    char *str = zmsg_popstr (msg);
    zmsg_destroy(&msg);
    if (strcmp(str, "TERM") == 0) {
        #ifdef VERBOSE
        fprintf(logfile, "Received terminate command.\n");
        #endif
        stop_timer(tidx_overall);
        printf("Worked for %f ms.\n", sutimes[tidx_overall]);
        free(str);
        return -1;
    } else if (strcmp(str, "EXTERNAL") == 0) {
        max_states_reached = 1;
        printf("Switching to external mode.\n");
    }
    free(str);
    return 0;
}

static SP_pred_ref init_predicate(const char *name, int arity, const char *module) {
  SP_pred_ref pred = SP_predicate(name,arity,module);
  if (pred == NULL) {
    fprintf(stderr, "Unable to find callback module %s:%s/%i for zmq worker, exiting.\n",
            module, name, arity);
    exit(1);
  }
  return pred;
}

static void init_callback_predicates(void) {
  pred_zmq_check = init_predicate("zmq_check",1,"worker");
  pred_local_zmq_check = init_predicate("local_zmq_check",1,"worker");
  pred_load_model = init_predicate("load",1,"worker");
}

void check_state(char *term, int id) {
    //printf("check len: %d \n",strlen(term));
    assert(id == 0);
    SP_pred_ref pred = pred_zmq_check;
    SP_term_ref val = SP_new_term_ref();
    if (!id) {
        SP_put_address(val, term);
    } else {
        SP_put_integer(val, id);
    }
    start_timer(tidx_prolog);
    int status = SP_query_cut_fail(pred, val);
    stop_timer(tidx_prolog);
    assert(status == SP_SUCCESS);
}

void put_result(char *term, int len) {
    has_results = 1;
    printf("Result: %s \n",term);
    zmsg_addmem(results, term, len);
}

zmsg_t *not_queue;
int not_queue_count = 0;


#ifdef VERBOSE
void put_successor_verbose(char *term, int len, char *d) {
    int y;
    for (y = 0; y < len; y++) {
        fprintf(fromprolog, "%c", term[y]);
    }
    fprintf(fromprolog, "\n");
    char *readable = (char*) malloc(41);
    print_readable_digest((unsigned char*)d, readable);
    fprintf(logfile, "Successor: %s\n", readable);
    free(readable); 
}
#endif

void send_package_to_master(char *d, char *r, int len) {
    not_queue_count++;
    char *content = (char*) malloc(20 + len);
    memcpy(content, d, 20);
    memcpy(content+20, r, len);
    
    zmsg_addmem(not_queue, content, 20 + len);
    free(content);
}

void put_successor(char *term, int len, int id) {
    start_timer(tidx_successors);
    transitions++;
    char *r = (char*) malloc(len);
    memcpy(r,term,len);
    
    char *d = (char*) malloc(20);
    
    sha1(r,len,d);

    if (memcmp(currentDigest, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 20) != 0) {
        char readable_from[41];
        char readable_to[41];
        print_readable_digest((unsigned char*) currentDigest, readable_from);
        print_readable_digest((unsigned char*) d, readable_to);
        fprintf(statespace, "\"%s\", \"%s\"\n", readable_from, readable_to);
    }





      #ifdef VERBOSE
        put_successor_verbose(term, len, d);
      #endif
      #ifdef DEBUG
        assert(check_debugq(debug_queue, d, term, len));
      #endif
    int contained = -2;
    if (local_hashes) {
        contained = contains(d);
    }  else {
        contained = req_contains(d);
    }
    if (!contained) { 
        if (local_enqueueing) {
            #ifdef DEBUG
            enqueue_debug(debug_queue, term, len, d);
            #endif
            if (max_queue_size > 0 && max_queue_size <= q_size(local_queue)) {
                write_to_file(r, d, len, id);
                // FIXME: r, d should be free'd here
            } else {
                enqueue(local_queue, r, len, d, THRESH, id);
            }
            assert(successors != NULL);
            add_queued_digest(successors, d);
            if (local_hashes) {
                put_local(d);
            }

            if (max_states_reached) {
                free(d);
                free(r);
            }

        } else {
            send_package_to_master(d, r, len);
            add_queued_digest(successors, d);
            free(r);
            free(d);
        }
    } else {
        cache++;
        free(d);
        free(r);
    }
    stop_timer(tidx_successors);
}

tCell *get_wp() {
    empty_queue = 0;
    start_timer(tidx_work_hard_incl_prolog);
    is_processing = 1;
    
    countRuns++;
    tCell *t = dequeue(local_queue); 
  
    assert(t->term != NULL);
    assert(t->digest != NULL);
    
    if (!local_enqueueing) {
        not_queue = zmsg_new();
        
        zmsg_addmem(not_queue, "0", 2);
        
    }
    not_queue_count = 0;
    return t;
}

#ifdef VERBOSE
void work_verbose(tCell *t) {
    char *readable = (char*) malloc(41);
    print_readable_digest((unsigned char*) t->digest, readable);
    fprintf(logfile, "Checking: %s\n", readable);
    fprintf(logfile, "Term in decimal: %s\n", readable);
    free(readable); 
    fprintf(toprolog, "Term is: ");
    int k;
    for (k = 0; k < t->term_length; k++) {
        fprintf(toprolog, "%c", t->term[k]);
    }
    fprintf(logfile, "\n");
    for (k = 0; k < t->term_length; k++) {
        fprintf(logfile, "%d", t->term[k]);
    }
    fprintf(toprolog, "\n");
}
#endif

void send_results_and_successors(int has_results, tCell *t) {
    if (has_results) {
        zmsg_pushmem(results, t->digest, 20);
        zmsg_send(&results, send_result);
    }
    // reset result status
    has_results = 0;
    add_digest_processed(successors, t->digest);
    add_queue_stats(successors);
    print_stats(successors, t);
    zmsg_send(&successors, send_hashes);
}

void put_checked_digest(char *digest) {
    char newdigest[21];
    memcpy(newdigest+1, digest, 20);
    newdigest[0] = 1;
    put(newdigest);
}

void clean_up_after_work(tCell *t) {
    zmsg_destroy(&results);
    zmsg_destroy(&successors);
    free(t->term);
    free(t->digest);
    free(t);
    is_processing = 0;

}

void report_empty_queue() {
    tick(); // probably not needed
    empty_queue = 1;
    successors = zmsg_new();
    add_queue_stats(successors);
    print_stats(successors, NULL);
    zmsg_send(&successors, send_hashes);
    zmsg_destroy(&successors); 
}

void work_hard (void) {
    start_timer(tidx_work_hard_incl_empty_queue);
    if (!is_empty(local_queue)) {
        tCell *t = get_wp();
        memcpy(currentDigest, t->digest, 20);

        int already_processed = -2;
        if (local_hashes) {
            already_processed = contains_processed(t->digest);
        } else {
            already_processed = req_contains_processed(t->digest);
        }

        if (!already_processed) {
            runs_since_update++;
            #ifdef VERBOSE
                work_verbose(t);
            #endif
            hit++;
        
            successors = zmsg_new ();
            results = zmsg_new();

            check_state(t->term, t->id);


            if (local_hashes) {
                put_checked_digest(t->digest);
            }
            stop_timer(tidx_work_hard_incl_empty_queue);
            stop_timer(tidx_work_hard_incl_prolog);
            send_results_and_successors(has_results, t);
        } else {
            cache_hit++;
            stop_timer(tidx_work_hard_incl_empty_queue);
            stop_timer(tidx_work_hard_incl_prolog);
        }
        
        if (!local_enqueueing) {
            if (not_queue_count) {
                zmsg_send(&not_queue, send_work);
            }
            zmsg_destroy(&not_queue);
        }
        clean_up_after_work(t);
        tick(); 
    }
    else {
        if (get_external_wp_count() > 0) {
            tCell *wp = deserialize_package();
            enqueue_cell(local_queue, wp);
            tick();
        } else if (!empty_queue) {
            report_empty_queue();
        }
        stop_timer(tidx_work_hard_incl_empty_queue);
    }
}

void load_model(char *model);

char *getId(void) {
    s_send(id_req, "hello");
    zmsg_t *msg = zmsg_recv(id_req);
    assert( ((int) zmsg_size(msg)) == 4);
    zframe_t *idframe = zmsg_pop(msg);
    zframe_t *modelframe = zmsg_pop(msg);
    zframe_t *master_queueing_frame = zmsg_pop(msg);
    zframe_t *central_hashes_frame = zmsg_pop(msg);
    char *id = zframe_strdup(idframe);
    char *model = zframe_strdup(modelframe);
    char *master_queueing = zframe_strdup(master_queueing_frame);
    char *central_hashes = zframe_strdup(central_hashes_frame);

    printf("Model size %i %s %s\n", (int) zframe_size(modelframe), model, master_queueing);
    load_model(model);
    if (0 == strcmp(master_queueing, "1")) {
         disable_localqueing();
    }

    int icentral_hashes;
    sscanf(central_hashes, "%d", &icentral_hashes);
    if (!icentral_hashes) {
        local_hashes = 1; // this will initialize the hashmap anyway, but nothing will be put
    }
    
    free(master_queueing);
    free(central_hashes);
    free(model);
    zframe_destroy(&idframe);
    zframe_destroy(&modelframe);
    zframe_destroy(&master_queueing_frame);
    zframe_destroy(&central_hashes_frame);
    zmsg_destroy(&msg);
    return id;
}


void print_stats(zmsg_t *msg, tCell *t) {
    char runs[2];
    assert(runs_since_update < 10);
    //char *runs = (char*) malloc(((int) log10(runs_since_update+1)) + 1);
    sprintf(runs, "%d", runs_since_update);
    runs_since_update = 0;
    
    zmsg_addstr(msg, runs);
    
    int i;
    for (i = 0; i < TIMEINDEX; i++) {
        zmsg_addstr(msg, "%f", sutimes[i]);
        sumtimes[i] += sutimes[i];
        sutimes[i] = 0;
    }
    struct timeval tv;
    gettimeofday(&tv, NULL);
    int len_sec = (int) log10(tv.tv_sec) + 2;
   // int len_micsec = (int) log10(tv.tv_usec + 1) + 2;
    int len_micsec = 6;
    assert(len_sec > 0);
    assert(len_micsec >= 6);
    char *timestamp = (char*) malloc(len_sec + len_micsec + 1);
    assert(timestamp != NULL);

    struct timeval tmpgt;
    gettimeofday(&tmpgt, NULL);
    int tmpt = ((int)tmpgt.tv_sec * 1000) + (int)(tmpgt.tv_usec/1000);

    sprintf(timestamp, "%d",  (tmpt - global_starttime));
    //printf("%s\n", timestamp);
    zmsg_addstr(msg, timestamp);
    free(timestamp);
    char *work_hard_runs;
    if (countRuns == 0) {
        work_hard_runs = (char*) malloc(2);
        sprintf(work_hard_runs, "0");
    }
    else {
        work_hard_runs = int2string(countRuns);
    }
    zmsg_addstr(msg, work_hard_runs);
    free(work_hard_runs);
    

	if ( t == NULL) {
	    zmsg_addstr(msg,"0000000000000000000000000000000000000000");
	} else {
		unsigned char* key = (unsigned char*) t->digest;
		char *dig = (char*) malloc(43);
		sprintf(dig,"\"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x\"",key[0],key[1],key[2],key[3],key[4],key[5],key[6],key[7],key[8],key[9],key[10],key[11],key[12],key[13],key[14],key[15],key[16],key[17],key[18],key[19]);
        zmsg_addstr(msg,"%s", dig);
        free(dig);   
	}
	
	
	

}

void add_queue_stats(zmsg_t *msg) {
    int s = q_size(local_queue) + is_processing;
    if (countRuns > 0 || s != 0) {
        zmsg_addstr(msg, "%s %d %d", id, s, get_external_wp_count());
    } else {
        zmsg_addstr(msg, "%s %d 0", id, -2);
    }
}

void add_digest_processed(zmsg_t *msg, char *digest) {
    char *pl = (char*) malloc(21);
    assert(pl != NULL);
    memcpy (pl, "\1", 1);
    memcpy (pl + 1, digest, 20);
    zmsg_addmem(msg, pl, 21);
    free(pl);
}

void dump(void) {
    SP_pred_ref pred = SP_predicate("dump",0,"worker");
    int status = SP_query_cut_fail(pred);
    assert(status == SP_SUCCESS);
}


int worker_main (char* addr, char *mode) {

  init_callback_predicates();

  masteraddr = addr;

    if (strcmp(mode, "mixed") == 0) {
        THRESH = 50;
    } else if (strcmp(mode, "breadth_first") == 0) {
        THRESH = 0;
    } else {
        THRESH = 100;
    }


  struct timeval tmpgt;
  gettimeofday(&tmpgt, NULL);
  global_starttime = ((int)tmpgt.tv_sec * 1000) + (int)(tmpgt.tv_usec/1000);

  printf("Trying to connect to master at %s.\n",masteraddr);
  printf("Search strategy: %s.\n",mode);

    bzero(currentDigest, 20);


#ifdef DEBUG
    SP_pred_ref pred = SP_predicate("set_debug",0,"worker");
    int status = SP_query_cut_fail(pred);
    assert(status == SP_SUCCESS);
    debug_queue = init_queue();
#endif
#ifdef LOCAL
    SP_pred_ref pred2 = SP_predicate("set_local",0,"worker");
    int status2 = SP_query_cut_fail(pred2);
    assert(status2 == SP_SUCCESS);
#endif
    
    #ifndef VERBOSE   
    init_hashmap();
    #endif
    local_queue = init_queue();
    init_zmq();

//  --------------------------
//  TERMINATION

    teardown_queue(&local_queue);
    
/*    SP_pred_ref pred3 = SP_predicate("statistics",0,"user");
    int status = SP_query(pred3);
    assert(status == SP_SUCCESS);*/

    teardown_external();
    fclose(statespace);


    free(id);
    printf("# of elements: %d\n", teardown());
 
    #ifdef TIME
    print_time();
    #endif
    printf("Cache hits: %d\n", cache_hit);
    #ifdef DEBUG
    print_collisions();
    printf("----------------------\n");
    clear_debug(debug_queue);
    printf("----------------------\n");
    dump();
    #endif
    #ifdef VERBOSE
    fclose(logfile);
    fclose(toprolog);
    fclose(fromprolog);
    #endif
    printf("Done.\n");
    return 0;
}



void load_model(char *model) {
    SP_term_ref val = SP_new_term_ref(); 
    SP_put_address(val, model);
    int status = SP_query_cut_fail(pred_load_model,val);
    assert(status == SP_SUCCESS);   
}

void print_time() {
    printf("\nTime spent:\n");
    printf("on checking a state (prolog): %f ms\n", sumtimes[tidx_prolog]);
    printf("in work_hard (incl. prolog time): %f ms\n", sumtimes[tidx_work_hard_incl_prolog]);
    printf("on an empty queue: %f ms\n", sumtimes[tidx_work_hard_incl_empty_queue] - sumtimes[tidx_work_hard_incl_prolog]);
    printf("in work_hard (excl. prolog time): %f ms\n", sumtimes[tidx_work_hard_incl_prolog] - sumtimes[tidx_prolog]);
    printf("on incoming hashcodes: %f ms\n", sumtimes[tidx_hashcodes]);
    printf("on handling workrequests: %f ms\n", sumtimes[tidx_workrequests]);
    printf("on handling incoming work: %f ms\n", sumtimes[tidx_incoming]);
    printf("on putting successors: %f ms\n", sumtimes[tidx_successors]);
}

void init_zmq() {

    assert(masteraddr != NULL);
    int iplen = strlen(masteraddr)+strlen("tcp://")+strlen(":00000")+1;

    char p5000[iplen];
    sprintf(p5000,"tcp://%s:%i",masteraddr,5000);
    char p5001[iplen];
    sprintf(p5001,"tcp://%s:%i",masteraddr,5001);
    char p5002[iplen];
    sprintf(p5002,"tcp://%s:%i",masteraddr,5002);
    char p5003[iplen];
    sprintf(p5003,"tcp://%s:%i",masteraddr,5003);
    char p5004[iplen];
    sprintf(p5004,"tcp://%s:%i",masteraddr,5004);
    char p5005[iplen];
    sprintf(p5005,"tcp://%s:%i",masteraddr,5005);
    char p5006[iplen];
    sprintf(p5006,"tcp://%s:%i",masteraddr,5006);
    char p5007[iplen];
    sprintf(p5007,"tcp://%s:%i",masteraddr,5007);
    char p5008[iplen];
    sprintf(p5008,"tcp://%s:%i",masteraddr,5008);
    char p5009[iplen];
    sprintf(p5009,"tcp://%s:%i",masteraddr,5009);

    zctx_t *ctx = zctx_new ();
    
    send_hashes = zsocket_new(ctx, ZMQ_PUSH);
    send_tick = zsocket_new(ctx, ZMQ_PUSH);
    recv_tick = zsocket_new(ctx, ZMQ_PULL);
    recv_ctrl = zsocket_new(ctx, ZMQ_SUB);
    send_result = zsocket_new(ctx, ZMQ_PUSH);

    id_req = zsocket_new(ctx, ZMQ_REQ);

    
 
    zsocket_connect(id_req, p5005);
    id = getId(); // updates local_hashes flag
    init_external(id);
    

    char statespace_name[20];
    snprintf(statespace_name, 20, "statespace%s", id);
    statespace = fopen(statespace_name, "w+");


    if (!local_hashes) {
        hash_req = zsocket_new(ctx, ZMQ_REQ);
        assert(0);
    } else {
        recv_hashes = zsocket_new(ctx, ZMQ_SUB);
    }

    
    #ifdef VERBOSE
    setId(id);
    init_hashmap();
    char *filename = (char*) malloc(15);
    sprintf(filename, "worker%s.log", id);
    logfile = fopen(filename, "a+");
    sprintf(filename, "frompl%s.log", id);
    fromprolog= fopen(filename, "a+");
    sprintf(filename, "topl%s.log", id);
    toprolog= fopen(filename, "a+");
    #endif
    printf("my name is %s\n", id);
    
    recv_work = zsocket_new (ctx, ZMQ_DEALER);
    zsocket_set_identity(recv_work,id);

    
    send_work = zsocket_new (ctx, ZMQ_DEALER);
    zsocket_set_identity(send_work,id);
    

    zsocket_connect(send_hashes, p5001);
    zsocket_connect(send_work, p5003);
    zsocket_connect(recv_work, p5002);
    zsocket_connect(recv_ctrl, p5007);
    zsocket_connect(send_result, p5008);
    if (!local_hashes) {
        zsocket_connect(hash_req, p5000);
        assert(0);
    } else {
        zsocket_connect(recv_hashes, p5000);
    }

    int tickport = zsocket_bind(recv_tick, "tcp://*:*");
    char prot[22]; 
    sprintf(prot,"tcp://localhost:%i",tickport); 
    printf("%s\n",prot);
    
    zsocket_connect(send_tick, prot);
    
    
    //  printf("port: %i\n",port);
    
    
    char *filter = "";
    zmq_setsockopt (recv_ctrl, ZMQ_SUBSCRIBE, filter, strlen (filter));
    if (local_hashes) {
        zmq_setsockopt (recv_hashes, ZMQ_SUBSCRIBE, filter, strlen (filter));
    } else {
        assert(0);
    }
    

    printf("starting\n");
    reactor = zloop_new ();
    zmq_pollitem_t poller2 = { recv_hashes, 0, ZMQ_POLLIN };
    zmq_pollitem_t poller4 = { recv_work, 0, ZMQ_POLLIN };
    zmq_pollitem_t poller6 = { recv_tick, 0, ZMQ_POLLIN };
    zmq_pollitem_t poller8 = { send_work, 0, ZMQ_POLLIN };
    zmq_pollitem_t poller20 = { recv_ctrl, 0, ZMQ_POLLIN };
    
    if (local_hashes) {
        zloop_poller (reactor, &poller2, h_hash, NULL);
    } else {
        assert(0);
    }
    zloop_poller (reactor, &poller4, h_work, NULL);
    zloop_poller (reactor, &poller6, h_tick, NULL);
    zloop_poller (reactor, &poller8, h_workrequest, NULL);
    zloop_poller (reactor, &poller20, h_control, NULL);
    
    zmsg_t *msg = zmsg_new();
    s_sleep(100);
    add_queue_stats(msg);
    print_stats(msg, NULL);
    zmsg_send(&msg, send_hashes);
    zmsg_destroy(&msg);
    zloop_start  (reactor);

    zloop_destroy (&reactor);
    printf("Work runs: %d\nChecked states: %d\nAlready cached successors: %d\nTransitions %d\n", countRuns, hit, cache, transitions);
    
    zctx_destroy(&ctx);
}
