% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(b_compiler,[b_compile/5]).

:- use_module(library(lists)).

:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(error_manager).
:- use_module(btypechecker,[fasttype/2]).
:- use_module(b_interpreter).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module compiles set comprehensions into closures; making them independent of the state of the B machine.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

/* compile boolean expression into a closure where the local state and the global
 state has been incorporated, but any parameter is left in the closure */

:- public test_bexpr/3, test_bexpr2/3.
test_bexpr(Expr,[bind(cc,int(1)),bind(nn,int(2))], [bind(db,[(int(1),int(2))])]) :-
    fasttype( +conjunct( +conjunct( +member( global('Name')<<identifier(nn), +identifier('Name')),
                                    +member( global('Code')<<identifier(cc), +identifier('Code'))),
                         +not_member( global('Name')<<identifier(nn),
                                      +domain(set(couple(global('Name'),global('Code')))<<identifier(db)))), Expr).

test_bexpr2(Expr,[],[]) :-
    fasttype( +implication( +equal( seq(any)<<identifier(ss), +empty_sequence),
                            +equal( set(any)<<identifier(res), +empty_set)), Expr).

:- assert_must_succeed(( b_compiler:test_bexpr(E,L,S),
                         b_compiler:b_compile(E,[nn],L,S,_R)  )).
:- assert_must_succeed(( b_compiler:test_bexpr2(E,L,S),
                         b_compiler:b_compile(E,[ss,res],L,S,_R)  )).
 
%:- assert_pre(b_compiler:b_compile_boolean_expression(E,Parameters,LS,S,_), 
%                  (type_check(E,boolean_expression),type_check(Parameters,list(variable_id)),
%                   type_check(LS,store),type_check(S,store))).
%:- assert_post(b_compiler:b_compile_boolean_expression(_,_,_,_,E), type_check(E,boolean_expression)).

b_compile(TExpr,Parameters,LS,S,NTExpr) :- check_is_id_list(Parameters),
   % print('compile : '),translate:print_bexpr(TExpr),nl,  statistics(runtime,[Start,_]),%%
   b_compile1(TExpr,Parameters,LS,S,NTExpr,eval,_FullyKnown). %, check_infos(TExpr,NTExpr).
   %, print('compiled: '), translate:print_bexpr(NTExpr), print(' '),statistics(runtime,[Stop,_]), T is Stop-Start, print(T), print(' ms') , print(NTExpr),nl.

/*
check_infos(Old,New) :- bsyntaxtree:get_texpr_info(Old,IO),
    bsyntaxtree:get_texpr_info(New,IN),
    member(X,IO), \+ member(X,IN),
    bsyntaxtree:important_info(X),
    print(check_infos(Old,New)),nl,
    print('Not copied: '), print(X),fail.
check_infos(_,_). */

:- use_module(bsyntaxtree,[always_well_defined/1]).
b_compile1(TExpr,Parameters,LS,S,ResultTExpr,Eval,FullyKnown1) :-
    remove_bt(TExpr,Expr,NewUntypedExpr,NTExpr), % remove top-level Type Information
    b_compile2(Expr,Parameters,LS,S,NewUntypedExpr,Eval,FullyKnown),
    % print('compiled : '),translate:print_bexpr(Expr),nl,
    % print(res(FullyKnown,Eval,NTExpr)),nl,
    ((FullyKnown=true, evaluate_this(Eval,NTExpr))
     ->  %  print('eval : '),translate:print_bexpr(NTExpr),nl, %
        ((nonvar(NewUntypedExpr),
          b_interpreter:b_compute_expression(NTExpr,LS,S,ResultValue))
          -> %print(result(NTExpr,ResultValue)),nl,
             %remove_bt(TExpr,Expr,value(ResultValue),ResultTExpr)
             get_texpr_type(TExpr,TT),ResultTExpr = b(value(ResultValue),TT,[]),
             FullyKnown1 = FullyKnown
              %, print(compiled_value(ResultTExpr)),nl
          ;  translate:translate_bexpression_with_limit(NTExpr,TTS),
             add_internal_error('Undefined Value: ',bcompile1(TTS,LS,S)),
             ResultTExpr=NTExpr, FullyKnown1=false
        )
       ;   %(FullyKnown=true -> print('not eval: '), print(NTExpr),nl ; true),
            ResultTExpr=NTExpr, 
            (ResultTExpr = b(value(_),_,_) -> FullyKnown1=FullyKnown ; FullyKnown1=false)
     ).

%is_pred(b(_,pred,_)).

evaluate_this(eval,NTExpr) :- !, worth_evaluating(NTExpr).
evaluate_this(_,NTExpr) :- always_wd_and_efficient(NTExpr).


always_wd_and_efficient(b(X,_,_Info)) :- wd_and_efficient(X).

% things which are very quick to compute
wd_and_efficient(sequence_extension(_)) :- !.
wd_and_efficient(integer_set(_)) :- !. % will be converted into value(global_set(_))
wd_and_efficient(X) :- worth_it_int(X),!.
wd_and_efficient(X) :- worth_it_other(X).

check_wd(NTExpr) :- bsyntaxtree:always_well_defined(NTExpr). % ,print(wd(NTExpr)),nl.
  %(bsyntaxtree:always_well_defined(NTExpr) -> true ; print(not_guaranteed_wd(NTExpr)),nl).

worth_evaluating(b(X,Type,Info)) :- worth_it_type(Type,X,Info).
  % (worth_it_type(Type,X,Info) -> true ;  X=value(_) -> true ; print(not_evaluating(X)),nl,fail).
worth_it_type(integer,X,Info) :- worth_it_int_wd(X),!, check_wd(b(X,integer,Info)).
worth_it_type(T,function(A,B),Info) :- !,
 (check_wd(b(function(A,B),T,Info)) -> true
   ; fail). %print('not_precompiling_function '),translate:print_bexpr(A), print(' @ '), translate:print_bexpr(B),nl,fail).
worth_it_type(integer,X,_) :- worth_it_int(X),!.
worth_it_type(T,X,Info) :- worth_it_wd(X),!, check_wd(b(X,T,Info)).
worth_it_type(set(_),X,_) :- worth_it_set(X),!.
worth_it_type(seq(_),X,_) :- worth_it_set(X),!.
worth_it_type(_,X,_) :- worth_it_other(X).

% TO DO: we should use the predicate has_wd_condition from b_ast_cleanup

% integer operations that could generate WD-errors
worth_it_int_wd(div(_,_)).
worth_it_int_wd(max(_)).
worth_it_int_wd(min(_)).
worth_it_int_wd(mod(_,_)).
worth_it_int_wd(power_of(_,_)).

% other operations that could generate WD-errors
worth_it_wd(first(_)).
worth_it_wd(last(_)).
worth_it_wd(tail(_)).
worth_it_wd(front(_)).
worth_it_wd(restrict_front(_,_)).
worth_it_wd(restrict_tail(_,_)).
worth_it_wd(rel_iterate(_,_)). %% could be expensive
worth_it_wd(general_intersection(_)).
worth_it_wd(general_concat(_)).

% to do: check other operators that could be not well-defined !

worth_it_int(unary_minus(_)).
worth_it_int(add(_,_)).
worth_it_int(minus(_,_)).
worth_it_int(multiplication(_,_)).
worth_it_int(max_int).
worth_it_int(min_int).
worth_it_int(card(_)).
worth_it_int(size(_)).

%worth_it_set(empty_set). % treated below
%worth_it_set(empty_sequence). % treated below
worth_it_set(closure(_)).  %% ?? need to be more careful here; can be expensive --> need to keep track which parts need definitely to be evaluated
worth_it_set(integer_set(_)). % will be converted into value(global_set(_))
worth_it_set(interval(_,_)).
worth_it_set(reflexive_closure(_)). 
worth_it_set(reverse(_)).
worth_it_set(domain(_)). worth_it_set(range(_)).
worth_it_set(union(_,_)). worth_it_set(intersection(_,_)).  worth_it_set(set_subtraction(_,_)).
worth_it_set(image(_,_)). worth_it_set(composition(_,_)). 
worth_it_set(domain_restriction(_,_)). worth_it_set(domain_subtraction(_,_)).
worth_it_set(range_restriction(_,_)). worth_it_set(range_subtraction(_,_)).
worth_it_set(direct_product(_,_)). worth_it_set(parallel_product(_,_)).
worth_it_set(iteration(_,_)).
worth_it_set(set_extension(_)).
worth_it_set(sequence_extension(_)).
worth_it_set(rev(_)).
worth_it_set(concat(_,_)).
worth_it_set(seq(_)). % will be kept symbolic anyway
worth_it_set(seq1(_)). % will be kept symbolic anyway
worth_it_set(iseq(_)). % will be kept symbolic anyway
worth_it_set(iseq1(_)). % will be kept symbolic anyway
worth_it_set(perm(_)). % will be kept symbolic anyway
worth_it_set(pow_subset(_)). % will be kept symbolic anyway
worth_it_set(pow1_subset(_)). % will be kept symbolic anyway
worth_it_set(fin_subset(_)). % will be kept symbolic anyway
worth_it_set(fin1_subset(_)). % will be kept symbolic anyway
worth_it_set(general_union(_)).
worth_it_set(first_of_pair(_)).  % can also produce set
worth_it_set(second_of_pair(_)). % can also produce set


worth_it_other(boolean_true).
worth_it_other(boolean_false).
worth_it_other(string(_)). 
worth_it_other(first_of_pair(_)). 
worth_it_other(second_of_pair(_)).
worth_it_other(couple(_,_)).
%worth_it_other(value(_)). % does not have to be evaluted; already a value

/*
% To do: distinguish which parts definitely need to be evaluated
%dont_eval_subexpressions(member).  % special membership code can be used
%dont_eval_subexpressions(not_member).  % special membership code can be used
dont_eval_subexpressions(disjunct(_,_)). % it may not be necessary to eval RHS
dont_eval_subexpressions(implication(_,_)).  % it may not be necessary to eval RHS
*/

:- use_module(kernel_objects,[equal_object/2]).
:- use_module(kernel_mappings,[binary_boolean_operator/3]).
b_compile2(Exp,Parameters,LS,S,NExpr,Eval,FullyKnown) :- var(Exp), !,
    add_internal_error('Variable Expression: ',b_compile2(Exp,Parameters,LS,S,NExpr,Eval,FullyKnown)),
    NExpr=Exp, FullyKnown=false.
b_compile2(empty_set,_,_LS,_S,NExpr,_Eval,FullyKnown) :- !, NExpr=value([]), FullyKnown=true.
b_compile2(empty_sequence,_,_LS,_S,NExpr,_Eval,FullyKnown) :- !, NExpr=value([]), FullyKnown=true.
b_compile2(value(Val),_Parameters,_LS,_S,NExpr,_Eval,FullyKnown) :- !,
    NExpr=value(Val), (ground(Val) -> FullyKnown=true ; FullyKnown=false).
b_compile2(integer(Val),_Parameters,_LS,_S,NExpr,_Eval,FullyKnown) :- !,
    NExpr=value(int(Val)), (ground(Val) -> FullyKnown=true ; FullyKnown=false).
b_compile2(identifier(Id),Parameters,LS,S,NExpr,_Eval,FullyKnown) :- !,
    ( memberchk(Id,Parameters) ->
	  NExpr = identifier(Id), FullyKnown=false
    ; otherwise -> b_interpreter:lookup_value_in_store_and_global_sets(Id,LS,S,Value),
                   NExpr = value(Value), (known_value(Value) -> FullyKnown=true ; FullyKnown=false)
	 ).
% treat a few common cases, more efficiently than syntaxtransformation
b_compile2(equal(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB),
    ((FullyKnownA=true, FullyKnownB=true,
      NExprA = b(value(VA),_,_), NExprB = b(value(VB),_,_))
      ->  %print(computing_eq(VA,VB)),nl,
          (kernel_objects:equal_object(VA,VB) -> NExpr = truth ; NExpr = falsity)
          %, print(result(NExpr)),nl
       ;  generate_equality(NExprA,NExprB,NExpr)
    ).
b_compile2(not_equal(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB),
    ((FullyKnownA=true, FullyKnownB=true,
      NExprA = b(value(VA),_,_), NExprB = b(value(VB),_,_))
      ->  %print(computing_neq(VA,VB)),nl,
          (kernel_objects:equal_object(VA,VB) -> NExpr = falsity ; NExpr = truth)
          %, print(result(NExpr)),nl
       ;  NExpr = not_equal(NExprA,NExprB)
    ).
b_compile2(member(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    NExpr = member(NExprA,NExprB),FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_).
b_compile2(function(Fun,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    (is_extension_function(Fun) -> Eval1=false % extension function applications treated lazily
     % TO DO: first evaluate B, if known then only evaluate relevant part of extension function
     ; Eval1=Eval),
    b_compile1(Fun,Parameters,LS,S,NExprA,Eval1,FullyKnown1),
    NExpr = function(NExprA,NExprB),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnown2).
b_compile2(conjunct(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    (is_falsity(NExprA) -> NExpr = falsity
     ; NExpr = conjunct(NExprA,NExprB),
    % TO DO: propagate static information from A to B, e.g. if A :: x=10  -> add x=10 to LS/Parameters
    % also: try and detect unsatisfiable predicates beforehand
      b_compile1(B,Parameters,LS,S,NExprB,Eval,_)
    ).
b_compile2(disjunct(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    (is_truth(NExprA) -> NExpr = truth
     ; NExpr = disjunct(NExprA,NExprB),
       b_compile1(B,Parameters,LS,S,NExprB,false,_) % does not have to be executed in a succesful branch: set Eval to false to avoid computing expensive expressions which may not be needed
    ).
b_compile2(implication(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    (is_falsity(NExprA) -> NExpr = truth
     ; NExpr = implication(NExprA,NExprB),
       b_compile1(B,Parameters,LS,S,NExprB,false,_) % does not have to be executed in a succesful branch: set Eval to false to avoid computing expensive expressions which may not be needed
    ).
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown) :- 
    functor(Expr,BOP,2),
    kernel_mappings:binary_boolean_operator(BOP,_,_),!, % slightly more efficient than syntaxtransformation
    functor(NExpr,BOP,2), arg(1,Expr,A), arg(1,NExpr,NExprA),
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    arg(2,Expr,B), arg(2,NExpr,NExprB),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_).
    % TO DO: if fully known: evaluate
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown) :-
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr),
    get_texpr_ids(Names,Ids), append(Parameters,Ids,NParameters),
    b_compile_l(Subs,NParameters,LS,S,NSubs,Eval,FullyKnown).
b_compile_l([],_,_,_,[],_Eval,true).
b_compile_l([Expr|ExprRest],P,LS,S,[NExpr|NExprRest],Eval,FullyKnown) :-
    b_compile1(Expr,P,LS,S,NExpr,Eval,FullyKnown1),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile_l(ExprRest,P,LS,S,NExprRest,Eval,FullyKnown2).

combine_fully_known(true,A,A).
combine_fully_known(false,_,false).

is_extension_function(b(A,_,_Info)) :- is_extension_function_aux(A).
is_extension_function_aux(set_extension(_)).
is_extension_function_aux(sequence_extension(_)).

generate_equality(b(A,_,_),b(B,_,_),Res) :- generate_equality_aux(A,B,Res),!.
generate_equality(A,B,equal(A,B)).

generate_equality_aux(convert_bool(AA),value(PT),Res) :- PT==pred_true,get_texpr_expr(AA,TA),!,
   %print(simplify_bool(AA,PT)),nl,
   Res = TA.
generate_equality_aux(convert_bool(AA),value(PF),Res) :- PF==pred_false,!,
   %print(simplify_bool(AA,PF)),nl,
   Res = negation(AA).
generate_equality_aux(value(PT),convert_bool(AA),Res) :- PT==pred_true,get_texpr_expr(AA,TA),!,
   %print(simplify_bool(AA,PT)),nl,
   Res = TA.
generate_equality_aux(value(PF),convert_bool(AA),Res) :- PF==pred_false,!,
   %print(simplify_bool(AA,PF)),nl,
   Res = negation(AA).


known_value(X) :- nonvar(X), known_value2(X).
known_value2(global_set(GS)) :- nonvar(GS).
known_value2(freetype(GS)) :- nonvar(GS).
known_value2(closure(_,_,_)) :- !,fail. % still may have to be computed
known_value2(avl_set(_)). % already fully normalised
known_value2((X,Y)) :- known_value(X),known_value(Y).
known_value2(fd(A,B)) :- number(A),atomic(B).
known_value2(int(N)) :- number(N).
known_value2([]).
known_value2(pred_true).
known_value2(pred_false).
known_value2(string(S)) :- atomic(S).
known_value2([H|T]) :- known_value(H), known_value(T).

%b_evaluate1(TExpr,Parameters,LS,S,ResultTExpr,FullyKnown) :- 

check_is_id_list([]).
check_is_id_list([H|T]) :-
   (H=b(identifier(_),_,_) -> add_error(b_compile,'Typed id list as argument: ',[H|T]) ; true).
