% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_enumerate, [
  b_type_values_in_store/3,
  b_enumerate_values_in_store/5,
  b_enumerate_values/2,
  b_enumerate_typed_values/3,
  b_tighter_enumerate_values_in_store/4,
  b_tighter_enumerate_values/2,
  b_trace_tighter_enumerate_values/2,
  
  enumerate_integer_with_min_range_wf/4]).
  
  
:- use_module(self_check).
:- use_module(debug).
:- use_module(kernel_objects).
:- use_module(kernel_waitflags).
:- use_module(error_manager,[add_error/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module takes care of enumerating B variables.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

  /* ------------------------------------------ */
    
	
/* can be used to apply the result of 	
	b_getVarTypes_from_boolean_expression to a store */
	
:- use_module(store,[lookup_value_for_existing_id/3]).
	
	
:- assert_pre(b_enumerate:b_type_values_in_store(V,T,S), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_type_values_in_store(_,_,_), true).

b_type_values_in_store([],[],_).
b_type_values_in_store([Variable|VT],[Type|TT],Store) :-
   b_type_value_in_store(Variable,Type,Store),
   b_type_values_in_store(VT,TT,Store).
	
b_type_value_in_store(Var,Type,Store) :- 
   (Var = 'Identifier'(VarID)
      -> print(b_type_value_in_store(Var,Type)),print(' variable not in proper form'),nl
       ; VarID = Var),
   store:lookup_value_for_existing_id(VarID,Store,Val),
   (Val==fail -> add_error(b_type_value_in_store,'Variable has not been given value: ',VarID)
              ; ~~pp_mnf(kernel_objects:basic_type(Val,Type))).




/* ------------------------------------------ */
    

:- assert_pre(b_enumerate:b_enumerate_values_in_store(V,T,_,S,_WF), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_enumerate_values_in_store(_,_,_,_,_), true).
	
b_enumerate_values_in_store(Variables,Types,Values,Store,WF) :-
    generate_typed_var_list(Variables,Types,Store,Values,TypedValues),
    b_enumerate_values(TypedValues,WF).
    
:- block generate_typed_var_list(-,?,?,?,?).
generate_typed_var_list([],[],_,[],[]).
generate_typed_var_list([VarID|VT],[Type|TT],Store,Values,Res) :-
    ((~~pp_cll(store:lookup_value_for_existing_id(VarID,Store,Val)), Val \== fail)
     ->  Res = [typedval(Val,Type,VarID)|Rest], Values=[Val|ValT]
     ;   add_error(b_enumerate_values_in_store,'Variable has not been given value: ',VarID),
          Res=Rest,Values=ValT
   ),
   generate_typed_var_list(VT,TT,Store,ValT,Rest).
 
b_enumerate_sorted_values([],_).
b_enumerate_sorted_values([typedvalc(Val,Type,VarID,Card)|Rest],WF) :-
    get_wait_flag(Card,enum(Val,Type),WF,LWF),
    b_enumerate_sorted_value_and_continue(LWF,Val,Type,VarID,Rest,WF).
 :- use_module(kernel_objects,[enumerate_basic_type/2]).
:- block b_enumerate_sorted_value_and_continue(-,?,?,?, ?,?).
b_enumerate_sorted_value_and_continue(_LWF,Val,Type,_VarID,Rest,WF) :-
    %%  print(enum_sorted_value(_VarID,Val,Type)),nl, %%
    ~~pp_cll( kernel_objects:enumerate_basic_type(Val,Type) ),
    b_enumerate_sorted_values(Rest,WF).


b_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)),
   % print(sort_vals(SortedValues)),nl,
    b_enumerate_sorted_values(SortedValues,WF).
/* ------------------------------------------ */

b_enumerate_typed_values(Values,Types,WF) :-
   b_generate_typed_values(Values,Types,TypedVals),
   b_enumerate_values(TypedVals,WF).
b_generate_typed_values([],[],[]).
b_generate_typed_values([Val|Vrest],[Type|Trest],[typedval(Val,Type,'$UNKNOWN_ID')|TVrest]) :-
   b_generate_typed_values(Vrest,Trest,TVrest).


:- assert_pre(b_enumerate:b_tighter_enumerate_values_in_store(V,T,S,_WF), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_tighter_enumerate_values_in_store(_,_,_,_), true).
	
b_tighter_enumerate_values_in_store(Vars,Types,Store,WF) :-
    % print(tight_enum(Vars)),nl, %
    ~~mnf(b_extract_typedvalc(Vars,Types,Store,TypedVals)),
    % print(typed_vals(TypedVals)),nl, %
    b_tighter_enumerate_values(TypedVals,WF).
   
b_extract_typedvalc([],[],_,[]).
b_extract_typedvalc([Variable|VT],[Type|TT],Store,Res) :-
   (Variable = '_lambda_result_'
    -> b_extract_typedvalc(VT,TT,Store,Res) % lambda result does not need to be enumerated; will be computed from rest
    % TO DO: should we also deal specially with _lambda_result_ below in b_tighter_enumerate_values;
    %  or even infer in ANY statements, ... which variables are simply defined by equalities from the others
    ;  Res=[ExtractedH|ET],
       b_extract_typedvalc_var(Variable,Type,Store,ExtractedH),
       b_extract_typedvalc(VT,TT,Store,ET)).
 
b_extract_typedvalc_var(Var,Type,Store,typedval(Val,Type,VarID)) :- 
   (Var = 'Identifier'(VarID)
      -> print(b_tighter_enumerate_value_in_store(Var,Type)),print(' variable not in proper form'),nl
       ; VarID = Var),
   ((store:lookup_value_for_existing_id(VarID,Store,Val), Val \== fail)
     ->  true
     ;   add_error(b_tighter_enumerate_value_in_store,'Variable has not been given value: ',VarID)
   ).


b_tighter_enumerate_sorted_values([],_WF).
b_tighter_enumerate_sorted_values([typedvalc(Val,Type,VarID,Card)|Rest],WF) :- 
    (ground(Val)
      -> check_reached(Val,Type,VarID,Card,det,WF), %% <-----
         b_tighter_enumerate_sorted_values(Rest,WF)
      ;  get_wait_flag(Card,tighter_enum(VarID,Val,Type),WF,LWF),
         b_tighter_enumerate_sorted_value_and_continue(LWF,Val,Type,VarID,Card,Rest,WF)).
:- block b_tighter_enumerate_sorted_value_and_continue(-,?,?,?,?,?,?).
b_tighter_enumerate_sorted_value_and_continue(_LWF,Val,Type,VarID,Card,Rest,WF) :- 
     % (ground(Val) -> print(ground(Val,VarID,Rest)),nl ;  print(tight_enum_sorted_value(Val,Type,VarID,Rest)),nl), %%
    (( Rest \= [],currently_useless_variable(Val), %print(useless(Val,VarID)),nl,
       get_next_useful_variable(Rest,RVal,RType,RVarID,RCard,RRest)
     ) 
      ->  debug_println(9,skipping(Val,VarID,Card,new(RVal,RType,RVarID,RCard))),
         (RCard =< Card /* we can enumerate straight away */
           ->  check_reached(RVal,RType,RVarID,RCard,WF), %% <------
              ~~pp_cll( kernel_objects:enumerate_tight_type(RVal,RType,RVarID) ),
              b_tighter_enumerate_sorted_values([typedvalc(Val,Type,VarID,Card)|RRest],WF)
            ; b_tighter_enumerate_sorted_values([typedvalc(RVal,RType,RVarID,RCard),
                                                 typedvalc(Val,Type,VarID,Card)|RRest],WF)
         )
     ;   check_reached(Val,Type,VarID,Card,WF), %% <-----
         %% tools:print_bt_message(enumerate(VarID,Val)),
         ~~pp_cll( kernel_objects:enumerate_tight_type(Val,Type,VarID) ),
         %% tools:print_bt_message(after_enumerate(VarID,Val)),
         %(VarID=='VS_NOERRORCOND' -> print(done_enum(VarID,Val)),nl ; true),
         b_tighter_enumerate_sorted_values(Rest,WF)
    ).

get_next_useful_variable([typedvalc(Val,Type,VarID,Card)|Rest],RVal,RType,RVarID,RCard,RRest) :-
  (currently_useless_variable(Val)
    ->  RRest=[typedvalc(Val,Type,VarID,Card)|TRRest],
        get_next_useful_variable(Rest,RVal,RType,RVarID,RCard,TRRest)
    ;   RRest=Rest, RVal=Val, RType=Type, RVarID=VarID, RCard=Card
  ).
  

/* purpose: check if an enumeration is useful or not and give priority to useful variables ? */
%:- use_module(library(clpfd),[fd_dom/2]).
currently_useless_variable(Val) :- var(Val),!,frozen(Val,Goal),  %print(enum(Val,Goal)),nl,
      enumeration_only_goal(Goal,Val).
currently_useless_variable(int(Val)) :- currently_useless_variable(Val).
currently_useless_variable(fd(Val,_)) :- currently_useless_variable(Val).
  %fd_dom(Val,Dom),print(uselessfd(Val,Dom)),nl.
enumeration_only_goal(true,_).
%enumeration_only_goal(fd_utils_clpfd:delayed_enum_fd(V,_,_,_),Val) :- V==Val.
enumeration_only_goal(kernel_objects:enumerate_int_wf(V,_),Val) :- V==Val.
enumeration_only_goal(clpfd:(X in _),Var) :- Var==X. % only goal is the enumerator itself


b_tighter_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)),  % no longer required ??
    b_tighter_enumerate_sorted_values(SortedValues,WF).

:- if(debug:global_debug_flag).
:- dynamic reached/4.
my_spy('P_Env_Output_TargetSpeed_Speed').
check_reached(Val,Type,VarID,Card,WF) :- 
  (my_spy(VarID) -> print(' SPY POINT : '), print(VarID), print(' = '), print(Val),nl,
    portray_waitflags(WF), trace ; true),
  (ground(Val) -> check_reached(Val,Type,VarID,Card,det,WF)
     ; check_reached(Val,Type,VarID,Card,enum,WF)).
check_reached(Val,Type,VarID,Card,Info,WF) :- 
    (retract(reached(VarID,RR,Info)) -> true ; RR=0),
    R1 is RR+1,
    (RR mod 1000 < 1 -> (RR=0 -> print('REACHED : ') ; print(' AGAIN  : ')), print(Info), print(' '),
                 print(VarID), print(' : '), print(Card),
                 print(' : '), print(Type), print(' : '), print(Val), print(' :: '), print(RR), nl,
                 (my_spy(VarID)  %(RR>8000,RR<40001)
                  -> portray_waitflags(WF),trace ; portray_waitflags(WF))
                 %, (RR=1000 -> trace ; true)
           ; true),
    assert(reached(VarID,R1,Info)).
b_trace_tighter_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)), 
     print('ENUMERATING VALUES:'),nl,print_values(SortedValues), %%
     nl, portray_waitflags(WF),nl,
    b_trace_tighter_enumerate_values2(SortedValues,WF),
    print('FINISHED SETTING UP ENUMERATION'),nl, portray_waitflags(WF),nl.
 b_trace_tighter_enumerate_values2(SortedValues,WF) :-   retractall(reached(_,_,_)),
    b_tighter_enumerate_sorted_values(SortedValues,WF).

print_values([]).
print_values([typedvalc(Val,Type,VarID,Card)|Rest]) :-
  print_value(VarID,Card,Type,Val),
  (ground(Val) -> true ; when(ground(Val),(nl,print('GROUND: '),print_value(VarID,Card,Type,Val)))),
  print_values(Rest).
  
print_value(VarID,Card,Type,Val) :-
  print(' '),
  print(VarID), print(' :card='), print(Card), print(' ('), print(Type), print(') = '), print(Val),nl.
:- else.
check_reached(_Val,_Type,_VarID,_Card,_WF).
check_reached(_Val,_Type,_VarID,_Card,_Info,_WF).
b_trace_tighter_enumerate_values(TypedValues,WF) :- b_tighter_enumerate_values(TypedValues,WF).
:- endif.



sort_typed_values([],R,R).
sort_typed_values([typedval(Val,Type,VarID)|Rest],SortedSoFar,Res) :-
   (ground(Val) -> NewSorted=SortedSoFar
    ;  get_max_cardinality_as_priority(Type,CardPrio),
       %% TO DO: maybe also take shape of Val into account; e.g. if cardinality of a set already fixed...
       ~~mnf(insert_val(SortedSoFar,Val,Type,VarID,CardPrio,NewSorted))
   ),
   sort_typed_values(Rest,NewSorted,Res).

get_max_cardinality_as_priority(Type,Prio) :-
  ~~mnf(kernel_objects:max_cardinality(Type,Card)),
  (Card \= inf -> Prio=Card
    ; inf_type_prio(Type,Prio)
  ).

% Try to give priority to simpler infinite types (such as integer over seq(_) ,...)
% TO DO: to a full-blown analysis ...
inf_type_prio(integer,Prio) :- !, integer_priority(Prio).
inf_type_prio(seq(X),Prio) :- !, inf_type_prio(X,P1), Prio is P1+11.
inf_type_prio(set(X),Prio) :- !, inf_type_prio(X,P1), Prio is P1+10.
%inf_type_prio(couple(_,_),Prio) :- !, TO DO check if both args are infinite or not ...
inf_type_prio(_,Prio) :- !, Prio=10000001.

integer_priority(10000000). % priority for X : INTEGER enumerations
%integer_pow_priority(10000020).


%typedvalc = typedval + cardinality information
insert_val([],Val,Type,VarID,Card,[typedvalc(Val,Type,VarID,Card)]).
insert_val([typedvalc(V1,T1,Var1,C1)|T], Val,Type, VarID, Card, Res) :-
   ((cardval_greater_than(C1,Card) ; (Card=C1, var(V1), nonvar(Val)))
     -> Res = [typedvalc(Val,Type,VarID,Card),typedvalc(V1,T1,Var1,C1)|T]
     ;  Res = [typedvalc(V1,T1,Var1,C1)|RT],
        insert_val(T,Val,Type,VarID,Card,RT)
   ).


% call to enumerate an integer ensuring that you go at least from MinFrom to MinTo
%:- block enumerate_integer_with_min_range_wf(-,?,?,?).
enumerate_integer_with_min_range_wf(int(X),MinFrom,MinTo,WF) :-
   integer_priority(Prio), P1 is Prio-1,
   get_wait_flag(P1,enum_int_min_range(X,MinFrom,MinTo),WF,LWF),
    % print(enum_int_min_range(X,MinFrom,MinTo)),nl, portray_waitflags(WF),
    %((MinFrom == -4, MinTo == 17) -> trace ; true),
   enum_int_min_range(X,MinFrom,MinTo,LWF).

:- use_module(clpfd_interface,[clpfd_domain/3]).
:- block enum_int_min_range(?,?,?,-).
enum_int_min_range(X,MinFrom,MinTo,_) :-
   %  print(trigger_enum_int_min_range(X,MinFrom,MinTo)),nl,
  (ground(X) -> true ;
    clpfd_domain(X,FDLow,FDUp),
    (number(FDLow) -> From = FDLow ; From=MinFrom),
    (number(FDUp) -> To = FDUp ; To=MinTo),
    % print(enumerate_int(X,From,To)),nl,
    kernel_objects:enum_warning('INTEGER',FDLow:FDUp,From:To),
    kernel_objects:enumerate_int(X,From,To)).  

