% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_interpreter_components,[ 
  reset_component_info/1,
  b_trace_test_components/3, b_trace_test_components/4,
  unsat_component/1, unsat_boolean_expression/2,
  det_solution_for_constant/2,
  b_tracetest_boolean_expression/5,
  observe_state/3,
  
  construct_optimized_exists/3, construct_optimized_exists/4
  ]).

:- use_module(b_interpreter,[b_test_boolean_expression/6]).
:- use_module(preferences).
:- use_module(kernel_waitflags).
:- use_module(tools).
:- use_module(translate).
:- use_module(error_manager).
:- use_module(library(lists)).
:- use_module(b_enumerate, [b_trace_tighter_enumerate_values/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module can be used to split a predicate into components and solving each componente individually (with tracing feedback).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


/* -------------- */
/*  dynamic bits  */
/* -------------- */

:- dynamic skipped_expression/0, unsat_boolean_expression/2.
reset_unsat_boolean_expression_info :- retractall(skipped_expression), retractall(unsat_boolean_expression(_,_)).
reset_tracetest_for_component :- retractall(skipped_expression). % to be called before trace checking a component
assert_skipped_expression(ComponentNr,BE) :- assert_skipped_expression,
        assert(unsat_boolean_expression(ComponentNr,BE)),
        print(' *** INCONSISTENCY DETECTED, SKIPPING: '), print_bexpr_with_limit(BE,150),nl,
        print(' *** COMPONENT NUMBER: '), print(ComponentNr),nl.
assert_skipped_expression :-
  (skipped_expression -> true ; assert(skipped_expression)).


:- dynamic unsat_component/1.
reset_unsat_component_info :- retractall(unsat_component(_)).
store_unsat_component_nr(X) :- 
  (unsat_component(X) -> true ; assert(unsat_component(X))).
l_ground_wait_flags(WF) :- l_ground_wait_flags(WF,Result),
   (Result=false -> (!,fail) ; true).
unsat_component_exists :- unsat_component(_),!.


:- dynamic nr_of_components/1.
:- dynamic  allow_skipping_over_components/0.
reset_component_info(AllowSkip) :- retractall(allow_skipping_over_components),
 (AllowSkip==true -> assert(allow_skipping_over_components) ; true),
 reset_unsat_component_info, reset_unsat_boolean_expression_info.


/* -------------- */


:- use_module(library(avl)).
% You can use the term one as ComponentNr if you don't know the number or it is not relevant
b_tracetest_boolean_expression(BE,LS,S,WF,ComponentNr) :-
   empty_avl(Ai),
   b_tracetest_boolean_expression1(BE,LS,S,WF,ComponentNr,Ai,_Ao).
   
b_tracetest_boolean_expression1(b(Expr,_,Infos),LS,S,WF,ComponentNr,Ai,Ao) :- !,
	b_tracetest_boolean_expression2(Expr,Infos,LS,S,WF,ComponentNr,Ai,Ao).
b_tracetest_boolean_expression2(conjunct(LHS,RHS),_Info,LocalState,State,WF,ComponentNr,Ai,Ao) :- 
     % \+(Info = [try_smt|_]),
     !,
%%   print('check lhs of conjunct: '), translate:print_bexpr(LHS),nl,
        b_tracetest_boolean_expression1(LHS,LocalState,State,WF,ComponentNr,Ai,Aii),
%%   print('check rhs of conjunct: '), translate:print_bexpr(RHS),nl,
        b_tracetest_boolean_expression1(RHS,LocalState,State,WF,ComponentNr,Aii,Ao).
b_tracetest_boolean_expression2(BE,Infos,LS,S,WF,ComponentNr,Ai,Ao) :- 
    (preferences:get_preference(provide_trace_information,true)
      -> nl,print(' ====> '),print_bexpr_with_limit(BE,150),nl ; true),!,
    if(b_test_boolean_expression(b(BE,pred,Infos),LS,S,WF,Ai,Ao),true,   %% b_det_test_boolean_expression2
       (assert_skipped_expression(ComponentNr,BE),
        Ai=Ao)).
        

:- use_module(bsyntaxtree,[predicate_components/2,predicate_identifiers/2]).
b_trace_test_components(Properties,ConstantsState,TypedVals) :-
   b_trace_test_components(Properties,ConstantsState,TypedVals,all).
b_trace_test_components(Properties,ConstantsState,TypedVals,RelevantVariables) :-
    (preferences:get_preference(partition_predicates,true)
      -> predicate_components(Properties,Components),
         length(Components,Len), print(nr_of_components(Len)),nl
      ;  predicate_identifiers(Properties,PropVars),
         Components = [component(Properties,PropVars)], Len=1
     ),
    retractall(nr_of_components(_)),assert(nr_of_components(Len)),
    reset_observe_state,
     % print(components(Components)),nl,
    b_trace_test_components2(Components,ConstantsState,TypedVals,RelevantVariables,1,WFList),
    (unsat_component_exists
     -> print_message(unsatisfiable_predicate), fail
      ; print_message(grounding_wait_flags), l_ground_wait_flags(WFList)).


b_trace_test_components2([],_ConstantsState,TypedVals,_,Nr,ResWF) :-
  (TypedVals=[] -> print_message(finished_processing_component_predicates), 
                   ResWF=[]
     ; (preference(provide_trace_information,false)
        -> print_message(enumerating_constants_without_constraints)
         ; print_message(enumerating_constants_without_constraints(TypedVals))),
       init_wait_flags(WF),
       b_trace_tighter_enumerate_values(TypedVals,WF), ResWF=[wf_comp(WF,Nr)]).
b_trace_test_components2([component(Pred,ComponentVars)|T],ConstantsState,TypedVals,RelevantVariables,
                         Nr,ResWF) :-
		N1 is Nr+1,
    ((RelevantVariables\=all, ComponentVars\= all,
      lists_are_disjoint(RelevantVariables,ComponentVars))
     -> print_message(skipping_irrelevant_component(Nr,ComponentVars)),
		b_trace_test_components2(T,ConstantsState,TypedVals,RelevantVariables,N1,ResWF)
     ;  debug_println(9,checking_component_properties(Nr,ComponentVars)), %%
        ResWF = [wf_comp(WF,Nr)|TWF],
		init_wait_flags(WF),
		(ComponentVars=all -> InTypedVals=TypedVals, OutTypedVals=[], ProjectedState=ConstantsState
		  ; project_typed_vals(TypedVals,ComponentVars,InTypedVals,OutTypedVals),
			project_state(ConstantsState,ComponentVars,ProjectedState) % relevant, e.g., for exists which may wait on state becoming ground
		 ), 
		observe_state(InTypedVals,WF,Nr), %%
		reset_tracetest_for_component,
		b_tracetest_boolean_expression(Pred,[],ProjectedState,WF,Nr),
		(skipped_expression -> store_unsat_component_nr(Nr), allow_skipping_over_components
		  ; %print_bt_message(start_enumerating_constants(Nr)), %%(TypedVals)), %%
			b_trace_tighter_enumerate_values(InTypedVals,WF)
		),
		% we could ground waitflag 1 (deterministic computations) ?
		b_trace_test_components2(T,ConstantsState,OutTypedVals,RelevantVariables,N1,TWF)
	).

:- use_module(debug,[debug_println/2]).
l_ground_wait_flags([],true).
l_ground_wait_flags([wf_comp(WF,ComponentNr)|TWF],SatResult) :- 
   debug_println(9,grounding_component(ComponentNr)),  
   % print_bt_message(grounding_component(ComponentNr)),   portray_waitflags(WF),nl,%%
   if(ground_wait_flags(WF), l_ground_wait_flags(TWF,SatResult),
       (print(unsatisfiable_component(ComponentNr)),nl,
        store_unsat_component_nr(ComponentNr),
        SatResult=false)).

project_typed_vals([],_,[],[]).
project_typed_vals([typedval(Val,Type,VarID)|Rest],ComponentVars,In,Out) :-
   (member(VarID,ComponentVars) 
     -> In=[typedval(Val,Type,VarID)|In2], Out2=Out
     ;  Out=[typedval(Val,Type,VarID)|Out2], In2=In
   ), project_typed_vals(Rest,ComponentVars,In2,Out2).
project_state([],_,[]).
project_state([bind(VarID,Value)|Rest],ComponentVars,Out) :-
   (member(VarID,ComponentVars) 
     -> Out=[bind(VarID,Value)|Out2]
     ;  Out=Out2
   ), project_state(Rest,ComponentVars,Out2).
   
   
 :- volatile det_solution_for_constant/2.
 :- dynamic det_solution_for_constant/2.

:- use_module(tools_printing,[print_term_summary/1]).
reset_observe_state :- retractall(det_solution_for_constant(_,_)).
observe_state(Env,WF,ComponentNr) :- 
  get_wait_flag0(WF,WF0), get_wait_flag1(WF,WF1),
  (preferences:get_preference(provide_trace_information,true)
    -> when((ground(Env);nonvar(WF0)),
       (var(WF0) -> (nl,print('-->> SUCCESS; all constants valued in component '),
                     print(ComponentNr),nl,nl
                     %%,check_for_multiple_solutions(WF,ComponentNr) %% comment in to check for multiple solutions
                     )
                 ;  true))
      %% ,when(nonvar(WF1),(nl, print('Starting Phase 1'),nl, portray_waitflags(WF)))
    ; true),
  observe_state2(Env,WF0,WF1).
observe_state2([],_,_).
observe_state2([typedval(Val,_Type,VarID)|T],WF0,WF1) :-  
    observe_variable(VarID,Val,WF0,WF1),observe_state2(T,WF0,WF1).
%observe_state2([bind(Var,Val)|T],WF0) :- observe_variable(Var,Val,WF0), observe_state2(T,WF0).

% print a warning message if we find multiple solutions *after* all constants have been valued
:- public check_for_multiple_solutions/2.
check_for_multiple_solutions(WF,ComponentNr) :- get_enumeration_finished_wait_flag(WF,EWF),
    retractall(solution_found(ComponentNr,_)),
    add_solution(EWF,ComponentNr).
:- volatile solution_found/2.
:- dynamic solution_found/2.
:- block add_solution(-,?).
add_solution(_,ComponentNr) :- retract(solution_found(ComponentNr,Nr)),!,
    print('*** Multiple solutions found *after valuation* for component '), print(ComponentNr),nl,
    N1 is Nr+1,
    print('*** Solution: '),print(N1),nl,
    assert(solution_found(ComponentNr,N1)).
add_solution(_,ComponentNr) :- % print(add_sol(ComponentNr)),nl,
    assert(solution_found(ComponentNr,1)).

observe_variable(Var,Val,WF0,WF1) :- 
   when((ground(Val);nonvar(WF0)),
        (var(WF0) -> ((preferences:get_preference(provide_trace_information,true)
                        -> print('-->> '),print(Var), nl,print('     '),print_term_summary(value(Val))
                        ;  true),
                      assert(det_solution_for_constant(Var,Val))
                      ) 
                   ; (preferences:get_preference(provide_trace_information,true)
                       -> print(' :?: '),print(Var),print(' '), print_term_summary(value(Val)),
                          observe_variable1(Var,Val,WF1)
                         % , observe_instantiation(Val,Var,Val)
                        ; true)
                   )).

observe_variable1(Var,Val,WF1) :- when((ground(Val);nonvar(WF1)),
      (ground(Val) -> print('--1-->> '),print(Var), nl,print('     '),print_term_summary(value(Val))
        ; print('  :?1?: '),print(Var),print(' '), print_term_summary(value(Val))
         %% , translate:print_bstate([bind(Var,Val)]),nl % comment in to print full value with CLP(FD) info
      ) ).

/* 
comment back in for finer-grained observation and uncomment call above
:- block observe_instantiation(-,?,?).
observe_instantiation(X,Var,Val) :-
  print(bind(X)),nl,translate:print_bstate([bind(Var,Val)]),nl, (debug:debug_mode(on) -> trace ; true),
   X =.. [_|Args],
   observe2(Args,Var,Val).

observe2([],_,_).
observe2([H|T],Var,Val) :- (var(H) -> observe_instantiation(H,Var,Val)
  ; H=.. [_|Args], observe2(Args,Var,Val)),
  observe2(T,Var,Val).
*/

% --------------------

:- use_module(b_global_sets,[b_global_set/1]).

definitely_not_empty(b(E,T,_)) :- definitely_not_empty2(E,T).

definitely_not_empty2(bool_set,_).
definitely_not_empty2(integer_set(_),_).
definitely_not_empty2(identifier(X),set(global(X))) :- b_global_sets:b_global_set(X).
definitely_not_empty2(string_set,_).
definitely_not_empty2(set_extension([_|_]),_).
definitely_not_empty2(sequence_extension([_|_]),_).
% note: not that important due to the limited way definitely_not_empty is currently being used
%       if it fails a disequality will be generated instead; 1..100 /= {} will also be detected as true quite quickly
% definitely_not_empty2(interval(A,B),_) :- print(interval),nl,print(A),nl,print(B),nl,fail. % interval a..b, 
%definitely_not_empty2(X,T) :- print(empty(X,T)),nl,fail.


construct_optimized_exists(Parameters,Body,Pred) :-   
   construct_optimized_exists(Parameters,Body,Pred,true).

% we could also partition inside b_interpreter ?
construct_optimized_exists(Parameters,Body,Pred,Simplify) :-   
% we can actually do more: first project out all conjuncts which do not have any variables in common with Parameters
%  #x.(P(x) & Q)  <=>  #x.(P(x)) & Q 
% we can also use a stronger version of predicate components  #(x,y).( P(x,v) & Q(v,y)) <=>  #(x).P(x,v) & #(y).Q(v,y)
   get_texpr_ids(Parameters,PIDs),
   %print(constructing_optimized_exists(PIDs)),nl,
   predicate_components_with_restriction(Body,PIDs,Components),
   %length(Components,Len), print(nr_of_components_for_exists(Len,Components)),nl,
   conjunct_components(Components,Parameters,Pred,Simplify).
%   print('* EXISTS: '),translate:print_bexpr(Pred),nl.

:- use_module(bsyntaxtree).
:- use_module(b_ast_cleanup,[clean_up/3]).
conjunct_components(C,Parameters,Res,Simplify) :- conjunct_components2(C,Parameters,List,Simplify),
  conjunct_predicates(List,Res).

conjunct_components2([],_Parameters,[],_). % :- (Parameters=[] -> true ; print(unused_existential_var(Parameters)),nl).
conjunct_components2([component(Pred,ComponentVars)|T],Parameters,Res,Simplify) :-
   augment_ids(ComponentVars,CV), % translate atomic id names into b(identifier(... 
   tools:list_intersection(CV,Parameters,ExistParas),
   %print(ord_inter(CV,Parameters,ExistParas)),nl,
   (Simplify=true -> create_and_simplify_exists(ExistParas,Pred,EP)
    ; create_unsimplified_exists(ExistParas,Pred,EP)),
   (EP=b(truth,pred,[]) -> Res=TRes; Res=[EP|TRes]),
   conjunct_components2(T,Parameters,TRes,Simplify).

% translate a list of atomic ides into ids with the b/3 wrapper
augment_ids([],[]).
augment_ids([ID|T],[b(identifier(ID),_,_)|TT]) :- augment_ids(T,TT).

extract_equalities(Parameters,Pred, ReducedParas,OutPred) :-
   extract_equalities2(Pred,Parameters,ReducedParas,NewPred,[],Subst),
   sort(Subst,SortedSubst),
   apply_subst(SortedSubst,NewPred,OutPred).
   % apply equality inside expressions: note: this may duplicate expressions !
   % However, it also works if the equalities contain parameters

extract_equalities2(b(conjunct(A,B),pred,Info),Ids,RemIds,Res) --> !,
   {Res=b(conjunct(PA,PB),pred,Info)},
   extract_equalities2(A,   Ids,RemIds1,PA),
   extract_equalities2(B,RemIds1,RemIds,PB).
extract_equalities2(b(equal(ID,E),pred,_Info),Ids,RemIds,NewPred,InSubst,OutSubst) :- 
    same_id_member(ID,Ids,SID,RemIds),\+ occurs_in_expr(SID,E),
    \+ in_subst_dom(SID,InSubst),
    !,
    (always_well_defined(E)
      -> NewPred=b(truth,pred,[])
      ; NewPred=b(truth,pred,[]),
        % note: if we keep E=E for deadlock checking we would still not find more deadlock counter examples not(E=E) either fails or if E is not WD it also fails (sometimes generating error message though)
        print('Ignoring potential WD problem: '), translate:print_bexpr(E),nl),
        % NewPred = b(equal(E,E),pred,_Info)),  % this ensures that we evaluate E to detect any WD problems
    OutSubst = [SID/E | InSubst].
extract_equalities2(X,Ids,Ids,X) --> [].

apply_subst([],X,X).
apply_subst([Var/Expr|T],E,NewE) :- 
  (replace_id_by_expr(E,Var,Expr,E2)  % replace identifier Var by expression Expr
    -> apply_to_tail(T,Var,Expr,NT), apply_subst(NT,E2,NewE)
    ;  add_internal_error('Call failed: ',apply_subst([Var/Expr|T],E,NewE)),NewE=E).

% we could have something like [x/f(y), z/f(x)]  or even [x/f(y), y/f(x)]
apply_to_tail([],_,_,[]).
apply_to_tail([TV/TE|T],Var,Expr,Res) :-
   (occurs_in_expr(TV,Expr)
     -> print(removing_cyclic_subst(Var,TV)),nl,
        Res= NT
    ;   Res=[TV/TE2|NT],
        replace_id_by_expr(TE,Var,Expr,TE2)
   ),
   apply_to_tail(T,Var,Expr,NT).

in_subst_dom(X,[I/_|T]) :- (X=I -> true ; in_subst_dom(X,T)).

same_id_member(X,[H|T],SID,Rest) :-
  (same_id(X,H,SID) -> Rest=T ; Rest=[H|TR],same_id_member(X,T,SID,TR)).

create_and_simplify_exists(Paras,Pred,ResultingPred) :-
   % print(exists(Paras)),nl, translate:print_bexpr(Pred),nl,
   extract_equalities(Paras,Pred,NewParas,NewPred),
   % translate:print_bexpr(NewPred),nl,
   %print(create_and_simplify_exists2(NewParas,NewPred)),nl,
   create_and_simplify_exists2(NewParas,NewPred,ResultingPred).

create_and_simplify_exists2([],Pred,R) :- !,R=Pred.
create_and_simplify_exists2([ID],b(member(MID,E),pred,_Info),Res) :- same_id(ID,MID,SID),
  % + check that MID does not occur in E
  \+ occurs_in_expr(SID,E),
  !, 
   % print(simple_member(SID,E)),nl, %
  (definitely_not_empty(E) -> Res= b(truth,pred,[])
   ; get_texpr_type(E,Type), EmptySet=b(empty_set,Type,[]),
     Res = b(not_equal(E,EmptySet),pred,[])).
create_and_simplify_exists2([ID],b(not_member(MID,E),pred,Info),Res) :- same_id(ID,MID,SID),
  % + check that MID does not occur in E
  \+ occurs_in_expr(SID,E),
  get_texpr_type(E,SType), 
  bsyntaxtree:is_set_type(SType,Type),
  translate:set_type_to_maximal_texpr(Type,TypeExpr),!,
  %nl,print(not_member(SID,E,Type,TypeExpr)),nl,nl,
  Res = b(not_equal(E,TypeExpr),pred,Info).
create_and_simplify_exists2([ID],b(Pred,pred,_Info),Res) :- 
  b_interpreter_check:arithmetic_op(Pred,_Op,X,Y), 
  ( (same_id(ID,X,SID), \+ occurs_in_expr(SID,Y), always_well_defined(Y)) ;
    (same_id(ID,Y,SID), \+ occurs_in_expr(SID,X), always_well_defined(X))
  ), 
  !,  % we have a formula of the form #SID.(SID > Expr); provided Expr is well-defined, this is always true
  %print(simple_comp(SID,_Op,X,Y)), nl,
  Res= b(truth,pred,[]).
create_and_simplify_exists2(Parameters,Body,Res) :- create_unsimplified_exists(Parameters,Body,Res).
    
create_unsimplified_exists(Parameters,Body,Res) :-
     %print(create_unsimplified_exists(Parameters,Body)),nl,
     create_exists(Parameters,Body,Res2),
     clean_up(Res2,[],Res).
