:- module(ast_tools,[
		     openClauses/4
		    ,openClause/5
		    ,setDefinitions/3
		    ,getMachineName/2
		    ,getDefinitionName/2
		     ]).
:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'tools for manipulating the B-AST').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).

openClauses(abstract_machine(Pos,Variant,Header,OldClauses),OldClauses,
	    abstract_machine(Pos,Variant,Header,NewClauses),NewClauses) :- !.
openClauses(refinement_machine(Pos,Header,ID,OldClauses),OldClauses,
	    refinement_machine(Pos,Header,ID,NewClauses),NewClauses) :- !.
openClauses(implementation_machine(Pos,Header,ID,OldClauses),OldClauses,
	    implementation_machine(Pos,Header,ID,NewClauses),NewClauses) :- !.
openClauses(parse_errors(E),_,_,_):- !, throw(parse_errors(E)). % can happen when a machine is cached with parse_errors in it
openClauses(M,_,_,_):- throw(error('openClauses cannot find clauses',M)).

openClause(Functor,OldMachine,OldClause,NewMachine,NewClause):-
	openClauses(OldMachine,OldList,NewMachine,NewList),!,
	openClauseL(OldList,OldClause,NewList,NewClause,Functor),!,
	nonvar(OldClause).

openClauseL([],_,[],_,_):-!.
openClauseL([H1|T1],OC,[H2|T2],NC,Funktor):- !,
	(H1=..[Funktor|_]
	-> (OC=H1,H2=NC)
	;  (H1=H2)
	),
	openClauseL(T1,OC,T2,NC,Funktor).


setDefinitions(OldMachine,Definitions,NewMachine):-
	openClause(definitions,OldMachine,definitions(Pos,_)
		  ,NewMachine,definitions(Pos,Definitions)).



getMachineName(abstract_machine(_Pos,_Variant,Header,_Clauses),Res):-!,
	getMachineHeaderName(Header,Res).
getMachineName(refinement_machine(_Pos,Header,_Ref,_Clauses),Res):-!,
	getMachineHeaderName(Header,Res).
getMachineName(implementation_machine(_Pos,Header,_Ref,_Clauses),Res) :-!,
	getMachineHeaderName(Header,Res).
getMachineName(M,_):-throw(getMachineName('cannot determine machine name',M)).

getMachineHeaderName(machine_header(_Pos,IDs,_Args),IDs).


getDefinitionName(predicate_definition(_Pos,Name,_Args,_RHS),Name) :- !.
getDefinitionName(substitution_definition(_Pos,Name,_Args,_RHS),Name) :- !.
getDefinitionName(expression_definition(_Pos,Name,_Args,_RHS),Name) :- !.
getDefinitionName(file_definition(_Pos,Name),Name) :- !.
getDefinitionName(T,_):-throw(ast_diff('findDefName not defined',T)).