:- module(parser_call,[parse_predicate/2
		      ,parse_expression/2
		      ,parse_ltl/2
		      ,parse_csp_assert/2
		      ,loadMachineFile/2
		      ,loadDefinitionFileCaching/1
		      ,loadMachineHierarchy/2
		      ,loadMachineHierarchyPedantic/2]).

:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'An alternative parser for B. Used for REPL (no JVM startup time).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(probsrc(tools)).
:- use_module(probsrc(preferences)).
:- use_module(probsrc(bmachine_structure)).
:- use_module(bparser(sablecc_engine)).
:- use_module(bparser(sablecc_engine_backtracking)).
:- use_module(bparser(pre_parser)).
:- use_module(bparser(ast_tools)).
:- use_module(library(lists)).
:- use_module(library(file_systems)).
	      
/*
some wrappers for the calling the b-parser 
*/

test1:-parse_expression("1+1",_AST).
test2:-parse_predicate("e:M",_AST).
test3:-parse_ltl("G({magnetron=On} => F({magnetron=Off}))",_AST).

parse_predicate(String,Term) :-
	runBParser(String,'token_kw_predicate', predicate(_,Term)).
parse_expression(String,Term) :-
	runBParser(String,'token_kw_expression',expression(_,Term)).

/*
parse_csp_assert(Input,Term) :-
	lexString(Input,LexerTokens),!,
	mapTokenIDs(LexerTokens,ParserTokens),!,
	parseTokens(ParserTokens,Term).


parse_ltl(String,Term) :-
	append("#LTL\n",String,Input),
	lexString(Input,LexerTokens),!,
	mapTokenIDs(LexerTokens,ParserTokens),!,
	parseTokens(ParserTokens,Term).
*/


runBParser(Input,ParserEntry,AST1):-
	lexString(state_BParser_NORMAL_0,Input,LexerTokens),!,
	mapTokenIDs(LexerTokens,LT2),!,
        initDefinitionsFromGlobalMachine,
	patchIdentifierTokens(LT2,Toks),!,
	ParserTokens=[token(dummyPos,ParserEntry,dummyAtom)|Toks],
        (get_preference(bool_expression_as_predicate,true)
	  -> parseTokensAlt(pstate_BParser_0,ParserTokens,AST1)
	  ; parseTokens(pstate_BParser_0,ParserTokens,AST1)),
        resetDefinitions.


printTokens(L):-maplist(printToken,L).
printToken(token(Pos,TokenID,TokenAtom)) :-
	format("tokenID :",[]),
	write(TokenID),
	write(Pos),
	format(" TokenAtom ",[]),
	write(TokenAtom),nl.

:- dynamic machineFileLoaded/2.
loadMachineFileCaching(FileName,AST) :-
	(machineFileLoaded(FileName,_)
	  -> true
	  ;  loadMachineFile(FileName,AST),assert(machineFileLoaded(FileName,AST))
	).

loadMachineFile(FileName,AST) :-
	on_exception(parse_errors(Err),loadMachineFileException(FileName,AST), AST=parse_errors(Err) ).

loadMachineFileException(FileName,MachineAST) :-
        read_string_from_file(FileName,Content),
	lexString(state_BParser_NORMAL_0,Content,LexerTokens),!,
	mapTokenIDs(LexerTokens,LT2),!,        
        initDefinitionsFromTokens(FileName,LT2),!,
	%todo: avoid parsing definitions twice
	patchIdentifierTokens(LT2,ParserTokens),!,
        parseTokens(pstate_BParser_0,ParserTokens,AST),
	setOfDefinitionsTransitive(FileName,AllDefinitions),
	(setDefinitions(AST,AllDefinitions,MachineAST)
	-> true
	; true),
	resetDefinitions,
	true.

test(AST):-
        parseMachineFile('/home/fontaine/t/test.mch',AST).

% performance test: using Prolog parser:
% retractall(parser_call:definition_file_loaded(_)),statistics(walltime,_),parser_call:parseMachineFile('~/svn_root/NewProBPrivate/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_cfg_ipart.mch',AST),statistics(walltime,[_,Time]).
% using Java Parser:
% F='/Users/leuschel/svn_root/NewProBPrivate/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_cfg_ipart.mch',parsercall:prob_filename(F, PB),time(parsercall:call_console_parser(F,PB))
% performance test: 9290 ms runtime, 30640 ms walltime on MBP 3.06GHZ
%retractall(parser_call:definition_file_loaded(_)),statistics(walltime,_),parser_call:parseMachineFile('~/svn_root/NewProB/examples/B/PerformanceTests/SATLIB/sudoku.mch',AST),statistics(walltime,[_,Time]).


loadDefinitionFileCaching(FileName) :-
	(definitionFileLoaded(FileName)
	  -> true
	  ; (loadDefinitionFile(FileName),assert(definitionFileLoaded(FileName)))
	).

loadDefinitionFile(FileName):-
	print('loading File: '),print(FileName),nl,
        read_string_from_file(FileName,Content),
	lexString(state_BParser_NORMAL_0,Content,LexerTokens),!,
	mapTokenIDs(LexerTokens,LT2),!,        
        initDefinitionsFromTokens(FileName,LT2).

setOfDefinitionsTransitive(FileName,Set):-
	(setof(X,transitiveDefinition(FileName,X),Set)
	-> true
	;Set=[]).% if there is no solution return an empty list.

transitiveDefinition(FileName,AST):-
	fileContainsDefinition(FileName,direct_definition(AST)).
transitiveDefinition(FileName,AST):-
	fileContainsDefinition(FileName,include_definition(IncludeName)),
	transitiveDefinition(IncludeName,AST).


loadMachineHierarchy(FileName,Res):-
        print('!Prolog parser in Action !'),nl,
	retractall(machineFileLoaded(_,_)),
	loadMachineHierarchyCaching(FileName),
	findall(XAST,machineFileLoaded(_FileName,XAST),Machines),
	findall(FileName,machineFileLoaded(FileName,_),Files),
        machineFileLoaded(FileName,MainMachine),
	retractall(machineFileLoaded(_,_)),
	getMachineName(MainMachine,Main),
	Pragmas=[],
	Res=complete_machine(Main,Machines,Files,Pragmas).
	
loadMachineHierarchyCaching(FileName):-
	loadMachineFileCaching(FileName,AST),
	openClauses(AST,MachineClauses,_,_),
	maplist(loadReferencedMachine(FileName),MachineClauses),
	loadRefinedMachine(FileName,AST),
	resetDefinitions.

loadMachineHierarchyRelative(BaseName,IncludeName):-
	resolveReferencedMachineName(BaseName,IncludeName,AbsoluteIncludeName),
	loadMachineHierarchyCaching(AbsoluteIncludeName).


loadRefinedMachine(FileName,refinement_machine(_Pos,_Header,RefMachine,_Clauses)):-!,
	loadMachineHierarchyRelative(FileName,RefMachine).
loadRefinedMachine(FileName,implementation_machine(_Pos,_Header,RefMachine,_Clauses)):-!,
	loadMachineHierarchyRelative(FileName,RefMachine).
loadRefinedMachine(_,_).

loadReferencedMachine(FileName,sees(_POS,L)):-!,maplist(loadReferencedMachine_Expr(FileName),L).
loadReferencedMachine(FileName,uses(_Pos,L)):-!,maplist(loadReferencedMachine_Expr(FileName),L).

loadReferencedMachine(FileName,includes(_Pos,L)):-!,maplist(loadReferencedMachine_Ref(FileName),L).
loadReferencedMachine(FileName,extends(_Pos,L)):-!,maplist(loadReferencedMachine_Ref(FileName),L).
loadReferencedMachine(FileName,imports(_Pos,L)):-!,maplist(loadReferencedMachine_Ref(FileName),L).
loadReferencedMachine(_,_).

loadReferencedMachine_Expr(FileName,identifier(_Pos,IncludeName)):-!,
	loadMachineHierarchyRelative(FileName,IncludeName).

loadReferencedMachine_Expr(FileName,X):- throw(parserError('bad machine reference',FileName,X)).

loadReferencedMachine_Ref(FileName,machine_reference(_Pos,[RefName],_Args)):-!,
	loadMachineHierarchyRelative(FileName,RefName).
loadReferencedMachine_Ref(FileName,X):- throw(parserError('bad machine reference',FileName,X)).

resolveReferencedMachineName(FileName,IncludeName,AbsoluteName):-
	absolute_file_name(IncludeName,AbsoluteName,[relative_to(FileName),access(read),extensions(['.mch','.ref'])]).


loadMachineHierarchyPedantic(_FileName,_Res):- fail. % todo