:- module(sablecc_engine_peval,[pe/0]).

:- use_module(bparser(parser_tables_BParser)).
:- use_module(library(lists)).
% SABLE CC ENGINE

%tell('bparser/compiled.pl'), sablecc_engine:pe.

pe :-
	call_residue_vars(performReduce(ID,Stack,AST,RestStack),Vars),
        flatGoals(Vars,Body-[]),
	print(performReduce(ID,Stack,AST,RestStack)), print(' :- '),print_body(Body),nl,fail.

flatGoals([],A-A).
flatGoals([Var|Rest],B1-B2):-
	frozen(Var,Conjunct),
	flatConjunct(Conjunct,B1-T),
	flatGoals(Rest,T-B2).
flatConjunct(true,A-A).
flatConjunct((A,B),L1-L2):-!,
	flatConjunct(A,L1-X),
	flatConjunct(B,X-L2).
flatConjunct(G,L1-L2):- !,L1= [G|L2].

print_body(L):-
	maplist(print_res,L),
	print(true), print('.').

print_res(_:delay_trimNode(A,B)) :- !,print(trimNode(A,B)), print(', ').
print_res(_:delay_append(A,B,C)) :- !,print(append(A,B,C)), print(', ').
print_res(H) :- print(H), print(', ').

% replace append and trimNode in sablecc_engine.pl by the following:
:- block delay_append(-,?,?).
delay_append([],X,X).
delay_append([H|X],Y,[H|Z]) :- delay_append(X,Y,Z).

:- block delay_trimNode(?,?).
delay_trimNode(NormalNode,Res) :- trimNode(NormalNode,Res).


unifyStack([],Stack,Res):-Res=Stack.
unifyStack([H|T],[parserState(_,AstNode)|TStack],Res):-!,
	H=AstNode,
	unifyStack(T,TStack,Res).

performReduce(ID,Stack,AST,RestStack):-
	reductionRule(ID,Nodes,Term),
	unifyStack(Nodes,Stack,RestStack),
        (evalReduction(Term,AST)
	  -> true
	  ;throw(parserBug('error in reduce step',ID,Nodes,AST))).

evalReduction(var(X),X):- !.
evalReduction(emptyList,X-X):- !.
evalReduction(singletonList(L),[Res|X]-X):-!,evalReduction(L,Res).
evalReduction(appendList(L1,L2),R1-R3):- !,
  evalReduction(L1,R1-R2),
  evalReduction(L2,R2-R3).
evalReduction(appendNode(L,N),L1-L3):-!,
  evalReduction(L,L1-L2),
  evalReduction(N,R2),
  L2=[R2|L3].
evalReduction(newNode(Constructor1,_Constructor2,L),Res):-!, % todo refactor generator for reduction rules Constructor2 is no longer needed
  reduceList(L,Nodes),
  NormalNode=..[Constructor1|[pos(0,0,0,0,0,0)|Nodes]],
  delay_trimNode(NormalNode,Res).
evalReduction(head(E),Res):-!,
  evalReduction(E,[Res|_]-_).
evalReduction(resultNothing,nothing):-!.
evalReduction(Reduction,Res):-throw(parserBug('cannot eval reduction',Reduction,Res)).

reduceList([],[]).
reduceList([H|T],[H2|T2]) :-
	evalReduction(H,H2),
	reduceList(T,T2).
