% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% file: btracer_tcl.pl

:- use_module(tools,[read_string_from_file/2]).
:- use_module(value_persistance, [set_storage_directory/1]).

:- dynamic cur_tcl_process/1.

%%:- use_module(cspParserHs,'cia/haskell/java_api/cspParserHs.pl',
%%               [initCspParser/0]).

go :-
    set_search_pathes,
     (host_platform(Host) -> print_message(host_platform(Host))
                          ; print_message('### Could not get host platform !')),
    %%initCspParser,
    initialise_plugins,
    check_types, init_preferences,nl,
    get_cli_options(Options,RemArgs),
    %print(cli_options(Options,RemArgs)),nl,
    go2(Options,RemArgs).
 
 :- dynamic cli_prefs_file/1.
 
tcltk_load_commandline_prefs_file :- cli_prefs_file(File),
    preferences:load_preferences(File),
    print('Loaded preferences file: '),print(File),nl.

go2(Opt,_RemArgs) :-
    member(prefs(File),Opt),
    assert(cli_prefs_file(File)),
    fail.
go2(Opt,_RemArgs) :-
    memberchk(cache_storage(Dir),Opt),
    set_storage_directory(Dir),
    fail.
go2(_,[A1,A2|T]) :- 
    print('### Too many command-line arguments: '),nl,
    print('### '), print([A1,A2|T]),nl,
    print('###'),nl,
    print('### Usage [-pref PrefFile] [B/CSP_File]'),nl,
    fail.
go2(_,_) :-  % TclTk will call tcltk_get_command_line_file_to_open/1
    on_exception(E,tk_new([name('ProB')], X),
                 (%print(E),nl,
                  nl,print('### Unable to start Tcl/Tk !'),nl,
                  print('### Ensure that you have installed the right version of Tcl/Tk'),nl,
                  print('### (consult the ProB website).'),nl,throw(E))),
    assert(cur_tcl_process(X)),
    print('Starting Tcl/Tk'),nl, flush_output(user),
    eval_all_tcl_code(X),    
    nl,
    (tcltk_get_command_line_option(batch_mode) -> true ; tk_main_loop),
    tcl_delete(X),
    retract(cur_tcl_process(_)),
    (\+ cli_error_occured -> prob_finished
       ; nl,print('Abnormal Termination of ProB Tcl/Tk'),nl,
         print_cli_errors,
         halt(1)).

:- if(debug:global_debug_flag).
prob_finished :- print('Finished ProB Tcl/Tk'),nl, timer:timer_statistics.
:- else.
prob_finished :- print('Finished ProB Tcl/Tk'),nl.
:- endif.

eval_all_tcl_code(X) :-
    tcl_code(CODE),
    tcl_eval(X, CODE, _),
    fail.
eval_all_tcl_code(_).
    


gen_txt_files :- print_message('Generating ProBLicence.txt'),
                 tell('ProBLicence.txt'),prob_licence(Lic), write(Lic),fail.
gen_txt_files :- nl,told,
                 print_message('Generating ProBSummary.txt'),
                 tell('ProBSummary.txt'),prob_summary(Lic), write(Lic),fail.
gen_txt_files :- nl,told.

gen :- gen_txt_files.

%
% The .tcl code embedded in Prolog
% ' has to be encoded as \' and \ as \\
%
%
    
tcltk_get_command_line_file_to_open(FileNameToOpen) :-
    get_cli_options(_,AX),
    AX = [AF],
    (absolute_file_name(AF,FileNameToOpen)  % main purpose is to convert Windows \ into / and avoid encoding problems
      -> true
      ; error_manager:add_error(btracer_tcl,'Cannot determine full file name: ',AF),fail).

tcltk_get_command_line_option(Opt) :-  get_cli_options(O,_), member(Opt,O).

tcltk_get_command_line_tcltk_commands(list(Res)) :-
   findall(Cmd,tcltk_get_command_line_option(tcltk_command(Cmd)),Res).

get_cli_options(Options,RemArgs) :-
    prolog_flag(argv,ArgV),
    get_options(ArgV,recognised_option,Options,RemArgs).

recognised_option(X,Opt,Args,true) :- recognised_option(X,Opt,Args).

recognised_option('-prefs',prefs(F),[F]).
recognised_option('-regression',run_regression_tests,[]).
recognised_option('-bench',run_benchmarks,[]).
recognised_option('-laws',run_full_law_tests,[]).
recognised_option('-laws_quick',run_quick_law_tests,[]).
recognised_option('-csptests',run_cspm_tests,[]).
recognised_option('-selfcheck',run_selfcheck,[]).
recognised_option('-batch',batch_mode,[]).
recognised_option('-last',tcltk_open_last_file,[]).
recognised_option('-t',tcltk_command(procCheckTraceFile),[]).
recognised_option('-trace',tcltk_command(procCheckTraceFile),[]).
recognised_option('-c',tcltk_command(procComputeCoverage0),[]).
recognised_option('-mc',tcltk_command(procModelCheck),[]).
recognised_option('-tcl',tcltk_command(Cmd),[Cmd]).
recognised_option('-ref',tcltk_command(default_trace_refinement_search),[]).
recognised_option('-cov',trace_Module(F),[F]). % dealt with by prob_coverage_new_tcltk.pl
recognised_option('-cache',cache_storage(D),[D]).


:- dynamic cli_error_occured/0, cli_error_occured/1.
% should be asserted if the sicstus process should terminate with halt(1)
assert_cli_error_occured(X) :-
   (cli_error_occured -> true ; assert(cli_error_occured)),
   assert(cli_error_occured(X)).
print_cli_errors :- print('Error that occured: '),nl,
   cli_error_occured(X), print(' * '), print(X),nl,fail.
print_cli_errors :- nl.

%*******************************************************************************
% load text from external files

:- use_module(library(lists)).

add_tcl_code(Filename) :-
    absolute_file_name(Filename,AbsFilename),
	read_string_from_file(AbsFilename,String),
	split_string_by_keyword(String,Parts),
	assert_tcl_codes(Parts).
assert_tcl_codes([]).
assert_tcl_codes([String|Rest]) :-
	name(Atom,String),
	assert(tcl_code(Atom)),
	assert_tcl_codes(Rest).

assert_from_file(Predicate,Filename) :-
	read_string_from_file(Filename,String),
	name(Atom,String),
	Fact =.. [Predicate,Atom],
	assert(Fact).

split_string_by_keyword(String,Parts) :-
	split_string_by_keyword(X,X,Parts,String,"").
split_string_by_keyword(Stackstart,"\n",[Stackstart|Parts]) -->
	"\nproc ",!,
	{append("proc ",Stackend,NextStackstart)},
	split_string_by_keyword(NextStackstart,Stackend,Parts).
split_string_by_keyword(Stackstart,[C|Stackend],Parts) -->
	[C],!,split_string_by_keyword(Stackstart,Stackend,Parts).
split_string_by_keyword(Stackstart,[],[Stackstart]) --> "".



:- dynamic tcl_code/1.
:- dynamic prob_licence/1, prob_summary/1, procsp_summary/1, mc_summary/1, proz_summary/1, protla_summary/1.

tcl_source_file(ctext).
tcl_source_file(evaluation_view).
tcl_source_file(btracer).

full_tcl_name(FullName) :-
    tcl_source_file(Name),
    atom_concat(Name,'.tcl',File),
    absolute_file_name(tclsrc(File),FullName).

read_tcl_source_files :-
    findall(File, full_tcl_name(File), Files),
    maplist(add_tcl_code,Files).

:- read_tcl_source_files.

:- assert_from_file(prob_licence,'../tcl/licence.txt').
:- assert_from_file(prob_summary,'../tcl/prob_summary.txt').
:- assert_from_file(procsp_summary,'../tcl/procsp_summary.txt').
:- assert_from_file(mc_summary,'../tcl/mc_summary.txt').
:- assert_from_file(protla_summary,'../tcl/tla_summary.txt').
:- assert_from_file(proz_summary,'../tcl/proz_summary.txt').
:- assert_from_file(symmetry_summary,'../tcl/symmetry_summary.txt').

% for a faster code-test-(rewrite-retest)*- .... -giveup loop ;)
reset_tcl :-
    retractall(tcl_code(_)),
    read_tcl_source_files.

% just for interactive debugging
print_tcl_summary :-
    tcl_code(Atom),
    atom_codes(Atom,Codes),
    length(Codes,N),
    ( N =< 50 -> NCodes = Codes
    ; N > 50 -> prefix_length(Codes,NCodes,50)),
    maplist(is_nl,NCodes,MCodes),
    atom_codes(NAtom,MCodes),
    format('Tcl: ~w~n',[NAtom]),
    fail.
print_tcl_summary.

is_nl(10,32) :- !.
is_nl(X,X).
