% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(bvisual_any_maxsolver, [any_maxsolver/5]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,dot).
:- module_info(description,'Tool for generating dot visualization of B predicates and expressions. Finds a maximally satisfiable subset for predicates.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(ordsets)).
:- use_module(library(timeout)).

:- use_module(bsyntaxtree, [conjunction_to_list/2, conjunct_predicates/2]).
:- use_module(bvisual, [get_any/5]).
:- use_module(maxsolver).

:- use_module(error_manager, [add_error/2]).


any_maxsolver(IDs, Condition, State, LocalState, NewLocalState):-
    conjunction_to_list(Condition, Set),!,
    length(Set, Count),
    % x_debug(Condition, Set, Count),
    Check = bvisual_any_maxsolver:x_check(IDs, LocalState, State),
    Goal = maxsolver(Check, Set, Solver0),
    x_time_out_by_size(Count, TimeOut),
    print(time_out(Goal,TimeOut,Res)),nl,
    time_out(Goal, TimeOut, Res),!,
    ( 
        Res = time_out -> 
            add_error(bvisual, 'Time out at trying to find a solver of maximal length.'),
            longest_satisfiable_prefix(Check, Set, Solver, _)
            ;
        Solver0 = time_out(Solver) ->
            add_error(bvisual, 'Time out at trying to find a solver of maximal length.')
            ;
        Solver = Solver0
    ),
    conjunct_predicates(Solver, SolverExpr),
    get_any(IDs, SolverExpr, State, LocalState, NewLocalState).


:- use_module(preferences).

x_time_out_by_size(Size, TimeOut):-
    preferences:get_preference(time_out, BaseTimeOut),
    TimeOut is max(10,min(msb(Size), 3) * BaseTimeOut). /* Timeout must be at least 1 */


x_debug(Condition, Set, Count):-
    print(bvisual_any_maxsolver:condition(Condition)),
    nl,
    print(bvisual_any_maxsolver:set(Set)),
    nl,
    print(bvisual_any_maxsolver:length_of_set(Count)),
    nl,
    flush_output.

%
% Helpers
%

x_check(_, _, _, []):-!.

x_check(IDs, LocalState, State, Subset):-
    conjunct_predicates(Subset, LHS),!,
    get_any(IDs, LHS, State, LocalState, _).