% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(avlsets,[list_to_avlset/2, avlset_to_list/2,
  avlset_union/3,
  avlset_difference/3, avlset_intersection/3,
  avlset_cardinality/2, avlset_member/2,
  test/0]).

% an interface to the avl library for storing sets


:- use_module(library(avl)).

avlset_to_list(AvlSet,List) :- avl_domain(AvlSet,List).

list_to_avlset(List,AvlSet) :-
   add_true_info(List,S),
   sort(S,SS), ord_list_to_avl(SS,AvlSet).

add_true_info([],[]).
add_true_info([H|T],[H-true|CT]) :- 
    add_true_info(T,CT).

   

avlset_union(A1,A2,ARes) :-  % ARes = A1 \/ A2
     avl_domain(A2,List2),
     add_to_avlset(List2,A1,ARes).
add_to_avlset([],R,R).
add_to_avlset([H|T],AVL,AVLOUT) :-  avl_store(H,AVL,true,AVL1),
   add_to_avlset(T,AVL1,AVLOUT).
     


avlset_difference(A1,A2,ARes) :- % ARes =  A1 - A2
     avl_domain(A2,List2),
     remove_from_avlset(List2,A1,ARes).
remove_from_avlset([],R,R).
remove_from_avlset([H|T],AVL,AVLOUT) :-  
   (avl_delete(H,AVL,_true,AVL1) -> true ; AVL1=AVL),
   remove_from_avlset(T,AVL1,AVLOUT).
 
avlset_intersection(A1,A2,ARes) :- % ARes = A1 /\ A2
     avl_domain(A1,List1), empty_avl(Empty),
     cond_add_to_avlset(List1,A2,Empty,ARes).
cond_add_to_avlset([],_,R,R).
cond_add_to_avlset([H|T],A2,AVL,AVLOUT) :-  
   (avl_fetch(H,A2) -> avl_store(H,AVL,true,AVL1) ; AVL1=AVL),
   cond_add_to_avlset(T,A2,AVL1,AVLOUT).
     
     
avlset_member(X,AVL) :- var(X), !, avl_member(X,AVL).
avlset_member(X,AVL) :- ground(X), !, avl_fetch(X,AVL).
avlset_member(X,AVL) :- avl_member(X,AVL).


avlset_cardinality(S,Card) :- avl_size(S,Card).
     
   
print_avlset(A) :-
        avl_domain(A,List),
        print(List),nl.
     
test :- list_to_avlset([44,2,3,4,5,11,22],A1),
		print_avlset(A1),nl,
        list_to_avlset([5,11,33,44,55,1,0],A2),
		print_avlset(A2),nl,
        avlset_union(A1,A2,A3),
        print_avlset(A3),nl,
        avlset_member(44,A3),
        \+ avlset_member(23,A3),
        avlset_member(X,A3),X=4,
        print(ok_member),nl,
        avlset_cardinality(A3,Card),
        print(card(Card)),nl,
        avlset_difference(A1,A2,A4),
        print_avlset(A4),nl,
        avlset_cardinality(A4,Card4),
        print(card(Card4)),nl,
        avlset_intersection(A1,A2,A5),
        print_avlset(A5),nl.
   