% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(profiler,[reset_profiler/0, profile_call/2, ps/0, test/0]).


:- dynamic redo/2, enter/2, calls/2,  profiled_call/2.
:- dynamic sol_time/2, fail_time/2.

reset_profiler :-
   retractall(enter(_,_)),
   retractall(calls(_,_)),
   retractall(profiled_call(_,_)),
   retractall(sol_time(_,_)),
   retractall(fail_time(_,_)).
   
   
profile_call(_,Call) :- !, call(Call).

profile_call(ID,Call) :-  assert(profiled_call(ID,Call)),
  % (enter(ID,_C) -> (print('### Already profiling id: '), print(ID), print(calls(_C,Call)),nl) ; true),
   (retract(calls(ID,Nr)) -> (N1 is Nr+1, assert(calls(ID,N1))) ; assert(calls(ID,1))),
   %print(enter(ID,Call)),nl,
   statistics(runtime,[EnterTime,_]),
   asserta(enter(ID,EnterTime)),
   call(Call),
   (true ; (statistics(runtime,[RedoTime,_]),asserta(enter(ID,RedoTime))),fail),
   statistics(runtime,[ExitTime,_]),
   add_time(ID,ExitTime).
profile_call(ID,_Call) :- % print(fail(ID)),nl,
   statistics(runtime,[FailTime,_]),
   (retract(enter(ID,EnterTime))
      -> Delta is FailTime-EnterTime
       ; (Delta=0, print(err_fail(ID)),nl)
   ), assert(fail_time(ID,Delta)),
   fail. 
   
 

add_time(ID,ExitTime) :-  %print(add_time(ID,EnterTime,ExitTime)),nl,
   (retract(enter(ID,EnterTime))
      -> (Delta is ExitTime-EnterTime,
          assert(sol_time(ID,Delta))
          )
       ; (print(err_add_time(ID)),nl
         )
   ).
          

ps :- print('-------------------'),nl,
   print('Profiler Statistics'),nl,
   calls(ID,Nr),
   print(ID), print('  calls: '), print(Nr), nl,
   findall(T,sol_time(ID,T),FST), 
   print('  Sol : '),stat_list(FST),
   findall(T,fail_time(ID,T),FT), 
   print('  Fail: '),stat_list(FT),
   fail.
ps :- print('-------------------'),nl.
      
      

stat_list(L) :- sum_list_len(L,Sum,Len,Max),
   print('Nr: '), print(Len), print(' Sum: '), print(Sum),
   (Len>0 -> (print(' Max: '), print(Max),print(' Avg: '), Avg is Sum/Len, print(Avg)) ; true),
   nl.
   
sum_list_len([],0,0,0).
sum_list_len([H|T],Sum,Len,Max) :-
   sum_list_len(T,ST,LT,MT),
   Sum is ST+H, Len is LT+1,
   (MT>H -> Max=MT ; Max=H).
   
 
 /*
app([],X,X).
app([H|X],Y,[H|Z]) :-
  app(X,Y,Z).
  
gen([a|T],X) :- X>0, !,X1 is X-1, gen(T,X1).
gen([],0).


test :-  reset_profiler,
         profile_call(gen,gen(L,5000000)), profile_call(app,app(L,L,R)),
          profile_call(gen,gen(L2,20000)), profile_call(app,app(L2,L2,R)),
         ps.
test :- ps.
*/