% (c) 2009-2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(slicer_csp, [slice_from_program_point/14]).

:- use_module(haskell_csp).
:- use_module(avlsets).

:- use_module(probsrc(error_manager), [extract_line_col/5, add_error/3]).

:- use_module(slicer_csp_CSCFG).
:- use_module(slicer_csp_MEB_CEB).
:- use_module(slicer_csp_executable).
:- use_module(library(file_systems)).


find_process_for_source_location(FromL,FromC,ToL,ToC, RenamedProcess, Body,StartLine,StartCol,EndLine,EndCol) :-
  % symbol(RenamedProcess,Process,SrcSpan,_), is_possible_csp_process(RenamedProcess),
   agent_compiled(RenamedProcess,Body,SrcSpan),
   /*extract_line_col(SrcSpan,StartLine,StartCol,EndLine,EndCol),*/
   SrcSpan = src_span(StartLine,SC,EndLine,EndCol,_,_),
   StartCol is SC-1, 
   within(FromL,FromC,ToL,ToC,  StartLine,StartCol,EndLine,EndCol),!.
find_process_for_source_location(FromL,FromC,ToL,ToC, _, _Body) :- nl,
   print('### NO PROCESS FOUND FOR YOUR SELECTION'),nl,
   print(selection(FromL,FromC,ToL,ToC)),nl,
   fail.
   
within(FromL,FromC,ToL,ToC,  StartLine,StartCol,EndLine,EndCol) :-
   StartLine =< FromL, EndLine>= ToL,
   (StartLine < FromL ; StartCol =< FromC),
   (ToL < EndLine ; ToC =< EndCol).
   
safe_call(X) :- 
   on_exception(E,
      (call(X)
  -> true ; (add_error(slicer,'Internal Error, Call failed:',X),fail)),
      (add_error(slicer,'Exception occured: ',E),fail)).

slice_from_program_point(FromL,FromC,ToL,ToC,Filename,StartLines,StartCols,EndLines,EndCols,StartLines2,StartCols2,EndLines2,EndCols2,Reachable) :-
         safe_call(slice_from_program_point2(FromL,FromC,ToL,ToC,Filename,StartLines,StartCols,EndLines,EndCols,StartLines2,StartCols2,EndLines2,EndCols2,Reachable)).

slice_from_program_point_executableMEB(FromL,FromC,ToL,ToC,Filename,NewFilename):-safe_call(slice_from_program_point_executable2(FromL,FromC,ToL,ToC,Filename,NewFilename)).

slice_from_program_point_executableCEB(FromL,FromC,ToL,ToC,Filename,NewFilename):-safe_call(slice_from_program_point_executable3(FromL,FromC,ToL,ToC,Filename,NewFilename)).
   
slice_from_program_point2(FromL,FromC,ToL,ToC,Filename,StartLines,StartCols,EndLines,EndCols,StartLines2,StartCols2,EndLines2,EndCols2,Reachable) :- 
   %print(slice(FromL,FromC,ToL,ToC)), nl,
   %find_process_for_source_location(FromL,FromC,ToL,ToC,Process,Body,StartLine,StartCol,EndLine,EndCol),
   %print(found_process(Process,Body,StartLine,StartCol,EndLine,EndCol)),nl,  
   /* START INSERTING YOUR SLICING CODE HERE: */
   
   statistics(runtime,_),
   statistics(walltime,_),
   atom_concat(Filename,'_tree.pl',PFile),
   clear_graph,
   (file_exists(PFile) -> 
    (read_graph(PFile),print(loaded_tree_file(PFile)),nl,findall(X,slice_node(X,_,_),Xs),print(xs(Xs)),nl) ; 
    (create_graph,write_all(PFile))
   ),

   NFromC is FromC +1,
   NToC is ToC +1,
   findall(List,slice_compact_node(_,List),Lists),
   look_for_sc(Lists,FromL,NFromC,ToL,NToC),
   findall(Id,slice_node(Id,src_span(FromL,NFromC,ToL,NToC,_,_),_),Ids),
   go_from_node_AVL(0,GoFromMain),
   is_reachable(Ids,GoFromMain,Reachable),
   print(reachable(Reachable)),nl,


   statistics(runtime,[_,RTG]),
   statistics(walltime,[_,WTG]),
   nl,nl,nl,nl,nl,nl,nl,
   print('Build Graph - runtime: '),
   print(RTG),
   print(' ms. / absolute time elapsed: '),
   print(WTG),print(' ms.'),nl,nl,nl,nl,nl,nl,nl,

   
   



   statistics(runtime,_),
   statistics(walltime,_),
   meb(Ids,MEB,MEB_CEB),
   statistics(runtime,[_,RTM]),
   statistics(walltime,[_,WTM]),
   nl,nl,nl,nl,nl,nl,nl,
   print('MEB - runtime: '),
   print(RTM),
   print(' ms. / absolute time elapsed: '),
   print(WTM),print(' ms.'),nl,nl,nl,nl,nl,nl,nl,

   length(MEB,LengthMEB),
   print('Nodes MEB :'),
   print(LengthMEB),nl,


   create_graph_slice(MEB,'green','cia/CSCFG_slice_MEB.dot'),   
   %create_size_file(MEB,'cia/MEB.txt'),

   statistics(runtime,_),
   statistics(walltime,_),
   ceb(FromL,FromC,ToL,ToC,MEB_CEB,CEB),
   statistics(runtime,[_,RTC]),
   statistics(walltime,[_,WTC]),
   nl,nl,nl,nl,nl,nl,nl,
   print('CEB - runtime: '),
   print(RTC),
   print(' ms. / absolute time elapsed: '),
   print(WTC),print(' ms.'),nl,nl,nl,nl,nl,nl,nl,
   
   length(CEB,LengthCEB),
   print('Nodes CEB :'),
   print(LengthCEB),nl,

   create_graph_slice(CEB,'yellow','cia/CSCFG_slice_CEB.dot'),   
   %create_size_file(CEB,'cia/CEB.txt'),


   extract_spans(MEB,SPANS_MEB1),
   look_for_agent_calls(MEB,ACMEB_),
   sort(['MAIN'|ACMEB_],ACMEB),
   %print(acmsb(ACMEB)),
   add_spans_lhs(ACMEB,SPANS_MEB2),
   append(SPANS_MEB1,SPANS_MEB2,SPANS_MEB),
   /*print(spans_meb(SPANS_MEB)),nl,*/
   unzip(SPANS_MEB,StartLines,StartCols,EndLines,EndCols),


   extract_spans(CEB,SPANS_CEB1),
   look_for_agent_calls(CEB,ACCEB_),
   sort(['MAIN'|ACCEB_],ACCEB),
   %print(acmsb(ACCEB)),
   add_spans_lhs(ACCEB,SPANS_CEB2),
   append(SPANS_CEB1,SPANS_CEB2,SPANS_CEB),
   /*print(spans_meb(SPANS_CEB)),nl,*/
   haskell_csp:diff_set(set(SPANS_CEB),set(SPANS_MEB),set(SPANS_CEB_)),
   unzip(SPANS_CEB_,StartLines2,StartCols2,EndLines2,EndCols2),

   /* You can now traverse Body and use within/8 to find the source spans that have been
      selected */
   
   print(finished_slicing),nl.



   
slice_from_program_point_executable2(FromL,FromC,ToL,ToC,Filename,NewFilename):-
   /*print(slice(FromL,FromC,ToL,ToC)), nl,
   find_process_for_source_location(FromL,FromC,ToL,ToC,Process,Body,StartLine,StartCol,EndLine,EndCol),
   print(found_process(Process,Body,StartLine,StartCol,EndLine,EndCol)),nl,*/
   
   /*print(filename_(Filename)),nl,*/


   statistics(runtime,_),
   statistics(walltime,_),
   create_graph,
   statistics(runtime,[_,RTG]),
   statistics(walltime,[_,WTG]),
   nl,nl,nl,nl,nl,nl,nl,
   print('Build Graph - runtime: '),
   print(RTG),
   print(' ms. / absolute time elapsed: '),
   print(WTG),print(' ms.'),nl,nl,

   statistics(runtime,_),
   statistics(walltime,_),
   meb(FromL,FromC,ToL,ToC,MEB,_),
   statistics(runtime,[_,RTM]),
   statistics(walltime,[_,WTM]),
   nl,nl,
   print('MEB - runtime: '),
   print(RTM),
   print(' ms. / absolute time elapsed: '),
   print(WTM),print(' ms.'),nl,nl,nl,nl,nl,nl,nl,


   /*create_graph_slice(MEB,'green','cia/CSCFG_slice_MEB.dot'),   */

   
   atom_concat(Filename,'.sliced.csp',NewFilename),
   print('Writing to: '), print(NewFilename),nl,
   make_executable(MEB,NewFilename),


   
   print(finished_make_executable),nl.

   
slice_from_program_point_executable3(FromL,FromC,ToL,ToC,Filename,NewFilename):-
   /*print(slice(FromL,FromC,ToL,ToC)), nl,
   find_process_for_source_location(FromL,FromC,ToL,ToC,Process,Body,StartLine,StartCol,EndLine,EndCol),
   print(found_process(Process,Body,StartLine,StartCol,EndLine,EndCol)),nl,*/
   
   /*print(filename_(Filename)),nl,*/


   statistics(runtime,_),
   statistics(walltime,_),
   create_graph,
   statistics(runtime,[_,RTG]),
   statistics(walltime,[_,WTG]),
   nl,nl,nl,nl,nl,nl,nl,
   print('Build Graph - runtime: '),
   print(RTG),
   print(' ms. / absolute time elapsed: '),
   print(WTG),print(' ms.'),nl,nl,

   statistics(runtime,_),
   statistics(walltime,_),
   meb(FromL,FromC,ToL,ToC,_,MEB_CEB),
   statistics(runtime,[_,RTM]),
   statistics(walltime,[_,WTM]),
   nl,nl,
   print('MEB - runtime: '),
   print(RTM),
   print(' ms. / absolute time elapsed: '),
   print(WTM),print(' ms.'),nl,nl,nl,nl,nl,nl,nl,

   statistics(runtime,_),
   statistics(walltime,_),
   ceb(FromL,FromC,ToL,ToC,MEB_CEB,CEB),
   statistics(runtime,[_,RTC]),
   statistics(walltime,[_,WTC]),
   nl,nl,nl,nl,nl,nl,nl,
   print('CEB - runtime: '),
   print(RTC),
   print(' ms. / absolute time elapsed: '),
   print(WTC),print(' ms.'),nl,nl,nl,nl,nl,nl,nl,
   


   /*create_graph_slice(MEB,'green','cia/CSCFG_slice_MEB.dot'),   */

   
   atom_concat(Filename,'.sliced.csp',NewFilename),
   print('Writing to: '), print(NewFilename),nl,
   make_executable(CEB,NewFilename),


   
   print(finished_make_executable),nl.

is_reachable([],_,true).
is_reachable(Ids,GoFromMain,Reachable):-
   (Ids=[Id]->
    (avlset_member(Id,GoFromMain) -> Reachable=true;Reachable=false)
    ;
    (
     Ids=[Id|Ids_],
     is_reachable(Ids_,GoFromMain,ReachableIds_),
     ((avlset_member(Id,GoFromMain);ReachableIds_) ->
      Reachable=true;Reachable=false)
    )
   ).

unzip([],[],[],[],[]).
unzip([(slice_span(Fl,Fc,Tl,Tc))|L],Fls,Fcs,Tls,Tcs):-
   unzip(L,Fls1,Fcs1,Tls1,Tcs1),
   (Fc = 0 -> (Fc1 is 0);(Fc1 is Fc - 1)),
   (Tc = 0 -> (Tc1 is 0);(Tc1 is Tc - 1)),
   Fls=[Fl|Fls1],
   Fcs=[Fc1|Fcs1],
   Tls=[Tl|Tls1],
   Tcs=[Tc1|Tcs1].

concatena([],L,L).
concatena([X|L1],L2,[X|L3]):-concatena(L1,L2,L3).

extract_spans([],[]).
extract_spans([Id|Ids],SPANS):-
   /*print(extract_spans([Id|Ids])),nl,*/
   slice_node(Id,src_span(Fl,Fc,Tl,Tc,_,_),Type),
   (Type=9 -> 
    (slice_compact_node(Id,List),extract_spans_aux(List,SPANS1));
    SPANS1 = [slice_span(Fl,Fc,Tl,Tc)]
   ),
   extract_spans(Ids,SPANS2),
   concatena(SPANS1,SPANS2,SPANS).


extract_spans_aux([],[]).
extract_spans_aux([(_,src_span(Fl,Fc,Tl,Tc,_,_),_)|Ns],SPANS):-
   extract_spans_aux(Ns,SPANS1),
   SPANS=[slice_span(Fl,Fc,Tl,Tc)|SPANS1].

create_size_file(Ns,File):-
   tell(File),
   create_size_file_aux(Ns),
   told.

create_size_file_aux([]).
create_size_file_aux([Id|Ids]):-
   slice_node(Id,_,Type),
   print_type(Id,Type),
   create_size_file_aux(Ids).


print_type(Id,Type):-
   (
    Type=1 ->
    (slice_event(Id,Event),
    print(Event));
    (Type=2 ->
     (slice_process_call(Id,Process,_),
     print(Process));
     (
      Type=9 ->
      (
       slice_compact_node(Id,List),
       extract_nodes_compact(List)
      );
      (
       Type=3,slice_parallelism(Id,Events) ->
       (nodetype2string(Type),print_events(Events));
       nodetype2string(Type)
      )
     )
    )
   ).

extract_nodes_compact([]).
extract_nodes_compact([(Id,_,Type)|List]):-
   print_type(Id,Type),
   extract_nodes_compact(List).

print_events([]).
print_events([E|Es]):-
   print(E),
   print_events(Es).
   

nodetype2string(1):-print('event').
nodetype2string(2):-print('process_call').
nodetype2string(3):-print('|||').
nodetype2string(4):-print('[]').
nodetype2string(5):-print('->').
nodetype2string(6):-print(';').
nodetype2string(7):-print('SKIP').
nodetype2string(8):-print('STOP').
nodetype2string(9):-print('CompactNode').


look_for_agent_calls([],[]).
look_for_agent_calls([Id|Ids],ACS):-
    look_for_agent_calls(Ids,ACS_),
    (
     slice_process_call(Id,Process,_)-> 
     ACS =[Process|ACS_]
     ;
     (
      (
       slice_compact_node(Id,List) ->
       (look_for_agent_calls_compact(List,ACS__),
       append(ACS_,ACS__,ACS))
       ;
       ACS=ACS_       
      )
     )
    ).

look_for_agent_calls_compact([],[]).
look_for_agent_calls_compact([(Id,_,_)|Ns],ACS):-
    look_for_agent_calls([Id],ACS1),
    look_for_agent_calls_compact(Ns,ACS2),
    append(ACS1,ACS2,ACS).

add_spans_lhs([],[]).
add_spans_lhs([P|Ps],[slice_span(FL,FC,TL,TC)|SRCSPANS]):-
    haskell_csp:symbol(P,_,src_span(FL,FC,TL,TC,_,_),_),
    add_spans_lhs(Ps,SRCSPANS).

look_for_process(P,Result):-
   agent_compiled(_,Body,SrcSpan),
   extract_line_col(SrcSpan,StartLine,StartCol,EndLine,EndCol),
   look_for_agent_call(Body,set(Ps)),
   is_subset_of(set([P]),set(Ps)),
   Result = slice_span(StartLine,StartCol,EndLine,EndCol).
