% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(slicer_csp_CSCFG, [create_graph/0,clear_graph/0,create_graph_slice/3,
                             write_all/1,read_graph/1,come_to_node/2,come_to_node_AVL/2,
                             slice_node/3,slice_edge/3,slice_event/2, slice_process_call/3,
                             slice_parallelism/4,slice_compact_node/2,slice_sequential/2,
                             go_from_node/2,go_from_node_AVL/2,slice_renamed_event/2,
slice_hiding/2,slice_renaming/3,slice_choice/4,slice_if/2]).

:- use_module(haskell_csp).
:- use_module(avlsets).

:- use_module(probsrc(error_manager),[extract_line_col/5]).



create_graph :-
   clear_graph,
   assertz(slice_node_id(0)),
   assertz(slice_renamings([])),
   agent_compiled('MAIN',Body,_),nl,
   

   statistics(runtime,_),
   statistics(walltime,_),
   build_graph(Body,[],_,_),
   statistics(runtime,[_,RTG]),
   statistics(walltime,[_,WTG]),
   nl,nl,nl,nl,nl,nl,nl,
   print('Compute Graph - runtime: '),
   print(RTG),
   print(' ms. / absolute time elapsed: '),
   print(WTG),print(' ms.'),nl,nl,

   findall(Parallelism,slice_parallelism(Parallelism,_,_,_),Parallelisms),


   statistics(runtime,_),
   statistics(walltime,_),
   build_sync_edges(Parallelisms),
   statistics(runtime,[_,RTS]),
   statistics(walltime,[_,WTS]),
   nl,nl,nl,nl,nl,nl,nl,
   print('Compute Sync - runtime: '),
   print(RTS),
   print(' ms. / absolute time elapsed: '),
   print(WTS),print(' ms.'),nl,nl,

   /*retract(slice_edge(28,40,1)),
   retract(slice_edge(28,47,1)),
   retract(slice_edge(25,33,1)),
   retract(slice_edge(25,47,1)),
   retract(slice_edge(22,33,1)),
   retract(slice_edge(22,40,1)),*/

   /*assert_list([slice_edge(57,37,1),slice_edge(57,6,1),slice_edge(57,44,1),slice_edge(57,12,1),slice_edge(57,51,1),slice_edge(57,18,1),slice_edge(34,29,1),slice_edge(41,26,1),slice_edge(48,23,1),slice_edge(6,37,1),slice_edge(12,44,1),slice_edge(18,51,1)]),*/

   findall(A,slice_node(A,_,_),ListaA),
   findall(B,slice_node(_,B,_),ListaB),
   findall(C,slice_node(_,_,C),ListaC),
   zip3(ListaA,ListaB,ListaC,ListaABC),

   findall(A1,slice_edge(A1,_,_),ListaA1),
   findall(B1,slice_edge(_,B1,_),ListaB1),
   findall(C1,slice_edge(_,_,C1),ListaC1),
   zip3(ListaA1,ListaB1,ListaC1,ListaABC1),

   findall(Node,slice_node(Node,_,_),Nodes),
   length(Nodes,LengthNodes),
   print('Nodes CSCFG :'),
   print(LengthNodes),nl,

   build_dot_graph(ListaABC,ListaABC1,'cia/CSCFG.dot'),
   create_compact_graph.

create_compact_graph:-   
   compact_graph([0],[],[]),
   findall(Node,slice_node(Node,_,_),Nodes),
   length(Nodes,LengthNodes),
   print('Nodes compacted CSCFG :'),
   print(LengthNodes),nl,

   findall(A,slice_node(A,_,_),ListaA),
   findall(B,slice_node(_,B,_),ListaB),
   findall(C,slice_node(_,_,C),ListaC),
   zip3(ListaA,ListaB,ListaC,ListaABC),
   findall(A1,slice_edge(A1,_,_),ListaA1),
   findall(B1,slice_edge(_,B1,_),ListaB1),
   findall(C1,slice_edge(_,_,C1),ListaC1),
   zip3(ListaA1,ListaB1,ListaC1,ListaABC1),
  build_dot_graph(ListaABC,ListaABC1,'cia/CSCFG_compact.dot').
 
create_graph_slice(MEB,Color,File) :-
   findall(A,slice_node(A,_,_),ListaA),
   findall(B,slice_node(_,B,_),ListaB),
   findall(C,slice_node(_,_,C),ListaC),
   zip3(ListaA,ListaB,ListaC,ListaABC),
   findall(A1,slice_edge(A1,_,_),ListaA1),
   findall(B1,slice_edge(_,B1,_),ListaB1),
   findall(C1,slice_edge(_,_,C1),ListaC1),
   zip3(ListaA1,ListaB1,ListaC1,ListaABC1),
   findall_nodes_by_id(MEB,NMEB),
   build_dot_graph_slice(ListaABC,ListaABC1,NMEB,Color,File).



   
assert_list([]).
assert_list([F|Fs]):-
   assert(F),
   assert_list(Fs).

zip3([],[],[],[]).
zip3([A|LA],[B|LB],[C|LC],[E|L]):-
   E=(A,B,C),
   zip3(LA,LB,LC,L).
   
findall_nodes_by_id([],[]).
findall_nodes_by_id([Id|Ids],Ns):-
   slice_node(Id,B,C),
   N=(Id,B,C),
   findall_nodes_by_id(Ids,Ns1),
   Ns=[N|Ns1].
   

clear_graph :-  
   retractall(slice_node_id(_)),
   retractall(slice_node(_,_,_)),
   retractall(slice_edge(_,_,_)),
   retractall(slice_event(_,_)),
   retractall(slice_process_call(_,_,_)),
   retractall(slice_parallelism(_,_,_,_)),
   retractall(slice_compact_node(_,_)),
   retractall(slice_sequential(_,_)),
   retractall(slice_choice(_,_,_,_)),
   retractall(slice_if(_,_)),
   retractall(slice_renamings(_)),
   retractall(slice_renamed_event(_,_)),
   retractall(slice_hiding(_,_)),
   retractall(slice_renaming(_,_,_)).

get_id(Id):-
   slice_node_id(Id),
   retractall(slice_node_id(_)),
   NId is Id+1,
   assertz(slice_node_id(NId)),!.


/*
node types
1- event
2- process call
3- parallelism
4- choice
5- prefixing
6- sequential composition
7- SKIP
8- STOP
9- Compact node
10- Hiding
11- Renaming
*/

/*
edge types
1- synchronization
2- control
3- loop
*/

concatena([],L,L).
concatena([X|L1],L2,[X|L3]):-concatena(L1,L2,L3).


build_graph_sharing(Events,A,B,SPAN,Context,Id,FId):-
   get_id(Id),
   assert(slice_node(Id,SPAN,3)),
   build_graph(A,Context,IdA,FIdA),
   build_graph(B,Context,IdB,FIdB),
   assert(slice_parallelism(Id,Events,IdA,IdB)),
   assert(slice_edge(Id,IdA,2)),
   assert(slice_edge(Id,IdB,2)),
   (
    (haskell_csp:is_member_set(-1,set(FIdA));
    haskell_csp:is_member_set(-1,set(FIdB)))->
    FId = [-1];
    concatena(FIdA,FIdB,FId)
   ).

build_graph_choice(A,B,Src_position,Type,Context,Id,FId):-
   Src_position = src_position(FL,FC,SrcPos,Length),
   get_id(Id),
   TC is FC + Length,
   assert(slice_node(Id,src_span(FL,FC,FL,TC,SrcPos,Length),4)),
   build_graph(A,Context,IdA,FIdA),
   build_graph(B,Context,IdB,FIdB),
   assert(slice_choice(Id,Type,IdA,IdB)),
   assert(slice_edge(Id,IdA,2)),
   assert(slice_edge(Id,IdB,2)),
   concatena(FIdA,FIdB,FId).

build_graph(sharing(closure(Events),A,B,SPAN),Context,Id,FId):-
    build_graph_sharing(Events,A,B,SPAN,Context,Id,FId).

build_graph(sharing(setExp(rangeEnum(Events)),A,B,SPAN),Context,Id,FId):-
    build_graph_sharing(Events,A,B,SPAN,Context,Id,FId).

build_graph(aParallel(closure(EventsA),A,closure(EventsB),B,SPAN),Context,Id,FId):-
    haskell_csp:inter_set(set(EventsA),set(EventsB),set(Events)),
    build_graph_sharing(Events,A,B,SPAN,Context,Id,FId).

build_graph('|||'(A,B,src_position(FL,FC,SrcPos,Length)),Context,Id,FId):-
   TC is FC + Length,
   build_graph_sharing([],A,B,src_span(FL,FC,FL,TC,SrcPos,Length),Context,Id,FId).

/*build_graph('|||'(A,B,src_position(FL,FC,SrcPos,Length)),Context,Id,FId):-
   get_id(Id),
   TC is FC + Length,
   assert(slice_node(Id,src_span(FL,FC,FL,TC,SrcPos,Length),3)),
   build_graph(A,Context,IdA,FIdA),
   build_graph(B,Context,IdB,FIdB),
   assert(slice_edge(Id,IdA,2)),
   assert(slice_edge(Id,IdB,2)),
   (
    (haskell_csp:is_member_set(-1,set(FIdA));
    haskell_csp:is_member_set(-1,set(FIdB)))->
    FId = [-1];
    concatena(FIdA,FIdB,FId)
   ).*/

build_graph('|~|'(A,B,Src_position),Context,Id,FId):-
   build_graph_choice(A,B,Src_position,1,Context,Id,FId).

build_graph('[]'(A,B,Src_position),Context,Id,FId):-
   build_graph_choice(A,B,Src_position,2,Context,Id,FId).

build_graph('ifte'(Con,A,B),Context,Id,FId):-
   build_graph_choice(A,B,src_position(0,0,0,0),3,Context,Id,FId),
   assert(slice_if(Id,Con)).


build_graph(prefix(SPAN1,_,Event,Process,SPAN2),Context,IdEvent,FIdProcess):-
   get_id(IdEvent),
   %TC is FC + Length,
   assert(slice_node(IdEvent,SPAN1,1)),
   (
    slice_renamings(ListRenamings),haskell_csp:is_member_set((Event,Renamer),set(ListRenamings))->
    assert(slice_renamed_event(IdEvent,Event)),
    EventIns = Renamer;
    EventIns = Event
   ),
   assert(slice_event(IdEvent,EventIns)),
   get_id(Id),
   %FC2 is TC + 1,
   %TC2 is FC2 + 2,
   %TC2 is FC2 + Length2,
   assert(slice_node(Id,SPAN2,5)), 
   build_graph(Process,Context,IdProcess,FIdProcess),
   assert(slice_edge(IdEvent,Id,2)),
   assert(slice_edge(Id,IdProcess,2)).

/*build_graph(prefix(src_span(FL,FC,_,_,SrcPos,Length),_,Event,Process),Context,IdEvent,FIdProcess):-
   get_id(IdEvent),
   TC is FC + Length,
   assert(slice_node(IdEvent,src_span(FL,FC,FL,TC,SrcPos,Length),1)),
   assert(slice_event(IdEvent,Event)),
   get_id(Id),
   FC2 is TC + 1,
   TC2 is FC2 + 2,
   assert(slice_node(Id,src_span(FL,FC2,FL,TC2,SrcPos,2),5)), 
   build_graph(Process,Context,IdProcess,FIdProcess),
   assert(slice_edge(IdEvent,Id,2)),
   assert(slice_edge(Id,IdProcess,2)).
*/ 

build_graph(agent_call(SPAN,F,_Paras),Context,Id,FId):-
   get_id(Id),
   assert(slice_node(Id,SPAN,2)),
   assert(slice_process_call(Id,F,_Paras)),
   (
    haskell_csp:is_member_set((F,SPAN,IdCon),set(Context))->
    (assert(slice_edge(Id,IdCon,3)),
    FId = [-1]);
    (
     length(_Paras,Arity), 
     functor(FC,F,Arity),
     agent_compiled(FC,Body,_),
     concatena([(F,SPAN,Id)],Context,NContext),
     print(agent_call_(Body)),nl,
     build_graph(Body,NContext,Id1,FId),
     assert(slice_edge(Id,Id1,2))
    )
   ).
   

/*build_graph(aParallel(_,A,_,B,SPAN),Id,FId):-
   get_id(Id),
   assert(slice_node(Id,SPAN,3)),
   build_graph(A,IdA,FIdA),
   build_graph(B,IdB,FIdB),
   assert(slice_edge(Id,IdA,2)),
   assert(slice_edge(Id,IdB,2)),
   concatena(FIdA,FIdB,FId).*/

build_graph(stop(src_span(FL,FC,_,_,SrcPos,Length)),_,Id,[-1]):-
   get_id(Id),
   TC is FC + Length,
   assert(slice_node(Id,src_span(FL,FC,FL,TC,SrcPos,Length),8)).   

build_graph(skip(src_span(FL,FC,_,_,SrcPos,Length)),_,Id,[Id]):-
   get_id(Id),
   TC is FC + Length,
   assert(slice_node(Id,src_span(FL,FC,FL,TC,SrcPos,Length),7)). 

build_graph('\\'(A,Event,src_position(FL,FC,SrcPos,Length)),Context,Id,FIdA):-
   slice_renamings(ListRenamings),
   (
    haskell_csp:is_member_set((Event,RenamerOld),set(ListRenamings)) ->
    haskell_csp:diff_set(set(ListRenamings),set([(Event,RenamerOld)]),set(ListRenamings_));
    ListRenamings_=ListRenamings
   ),
   retractall(slice_renamings(_)),
   assertz(slice_renamings([(Event,'hiddenEvent')|ListRenamings_])),
   get_id(Id),
   TC is FC + Length,
   assert(slice_node(Id,src_span(FL,FC,FL,TC,SrcPos,Length),10)),
   assert(slice_hiding(Id,Event)),
   build_graph(A,Context,IdA,FIdA),
   assert(slice_edge(Id,IdA,2)),
   retractall(slice_renamings(_)),
   assertz(slice_renamings(ListRenamings)).

build_graph(procRenaming([rename(Renamed,Renamer)],A,src_span(FL,FC,_,_,SrcPos,Length)),Context,Id,FIdA):-
   slice_renamings(ListRenamings),
   (
    haskell_csp:is_member_set((Renamed,RenamerOld),set(ListRenamings)) ->
    haskell_csp:diff_set(set(ListRenamings),set([(Renamed,RenamerOld)]),set(ListRenamings_));
    ListRenamings_=ListRenamings
   ),
   retractall(slice_renamings(_)),
   assertz(slice_renamings([(Renamed,Renamer)|ListRenamings_])),
   get_id(Id),
   TC is FC + Length,
   assert(slice_node(Id,src_span(FL,FC,FL,TC,SrcPos,Length),11)),
   assert(slice_renaming(Id,Renamed,Renamer)),
   build_graph(A,Context,IdA,FIdA),
   assert(slice_edge(Id,IdA,2)),
   retractall(slice_renamings(_)),
   assertz(slice_renamings(ListRenamings)).

build_graph(';'(A,B,src_position(FL,FC,SrcPos,Length)),Context,IdA,FIdB):-
   build_graph(A,Context,IdA,FIdA),   
   get_id(Id),
   TC is FC + Length,
   assert(slice_node(Id,src_span(FL,FC,FL,TC,SrcPos,Length),6)),
   assert(slice_sequential(IdA,Id)),
   build_graph(B,Context,IdB,FIdB),
   build_seq_comp_edges(FIdA,Id),
   assert(slice_edge(Id,IdB,2)).

build_seq_comp_edges([Id|Ids],IdSC):-
   (Id = -1 -> true;
    assert(slice_edge(Id,IdSC,2))),
   ( Ids = [] -> true;
     build_seq_comp_edges(Ids,IdSC) ).
  

build_sync_edges([]).
build_sync_edges([P|Ps]):-
   slice_parallelism(P,Events,_,_),
   findall(S,slice_edge(P,S,2),Ss),
   Ss=[S1,S2],
   /*print(build_sync_edges_(P,Events,S1,S2)),nl,*/
   build_sync_edges_aux(Events,P,Events,S1,S2),
   disconnect_non_sync(P,Events,S1,S2),
   build_sync_edges(Ps).


build_sync_edges_aux([],_,_,_,_).
build_sync_edges_aux([Event|Events],P,WholeEvents,S1,S2):-
   build_sync_edges_aux2(Event,P,WholeEvents,[S1],[S2]),
   build_sync_edges_aux(Events,P,WholeEvents,S1,S2).

build_sync_edges_aux2(_,_,_,[],_).
build_sync_edges_aux2(Event,P,WholeEvents,[S1|S1s],[S2|S2s]):-
   /*print(build_sync_edges_aux2_________(Event,P,WholeEvents,[S1|S1s],[S2|S2s])),nl,*/
   (
    slice_event(S1,Event)->
    /*print(trobat(S1,Event)),nl,*/
    build_sync_edges_aux3(Event,P,S1,WholeEvents,true,Event,[S2|S2s],NS2);
    (slice_event(S1,OtherEvent),haskell_csp:is_member_set(OtherEvent,set(WholeEvents)) -> 
     build_sync_edges_aux3(OtherEvent,P,S1,WholeEvents,false,Event,[S2|S2s],NS2);
     NS2 = [S2|S2s]
    )
   ),
   /*print(ns2(NS2)),nl,*/
   (NS2 = [] -> 
    build_sync_edges_aux2(Event,P,WholeEvents,S1s,[S2|S2s]);
    (
     findall(N,slice_edge(S1,N,2),Ns),
     /*print(ns(Ns)),nl,*/
     (Ns=[] -> 
      (
       slice_edge(S1,Loop,3) ->
       (
        /*print('trobat loop a A'),nl,
        print(go_from_node_to_node_(Loop,S1)),nl,*/
        go_from_node_to_node(Loop,S1,NodesLoop),
        /*print(go_from_node_to_node_(Loop,S1,NodesLoop)),nl,*/
        findall(NEvent,slice_event(NEvent,Event),NEvents),
        haskell_csp:inter_set(set(NodesLoop),set(NEvents),set(Intersection)),
        (Intersection=[] -> build_sync_edges_aux2(Event,P,WholeEvents,S1s,NS2); 
         build_sync_edges_aux2(Event,P,WholeEvents,[Loop|S1s],NS2))
       );
       build_sync_edges_aux2(Event,P,WholeEvents,S1s,NS2)
      );
      (
       Ns=[Next]-> 
       (
        (
         slice_node(Next,_,7) ->
         (
          slice_edge(Next,SeqComp,2)->
          (
           /*print(entra_al_cas_de_un_skip_enganxat_a_un_sc),nl,*/
           findall(Before,slice_edge(Before,SeqComp,2),Befores),
           /*print(befores(Befores)),nl,*/
           come_to_node_common(Befores,CommonPath),
           /*print(befores_commonpath(Befores,CommonPath)),nl,*/
           calculate_last(CommonPath,0,Last),
           /*print(commonpath_last(CommonPath,Last)),nl,*/
           go_from_node_to_node(Last,Next,PathToSkip),
           /*print(p_last_next_pathtoskip(P,Last,Next,PathToSkip)),nl,*/
           (
            ((haskell_csp:is_member_set(P,set(PathToSkip)));(P=Last)) ->
            (/*print(entra_en_if),nl,*/build_sync_edges_aux2(Event,P,WholeEvents,S1s,NS2));
            (/*print(entra_en_else),nl,*/build_sync_edges_aux2(Event,P,WholeEvents,[Next|S1s],NS2))  
           )
          );
          build_sync_edges_aux2(Event,P,WholeEvents,S1s,NS2)  
         );
         build_sync_edges_aux2(Event,P,WholeEvents,[Next|S1s],NS2)
        )
       );
       (
        Ns = [N1,N2],
        /*print(build_sync_edges_aux2_(Event,N1,WholeEvents,N2,NS2)),nl,*/
        build_sync_edges_aux2(Event,P,WholeEvents,[N1],NS2),
        build_sync_edges_aux2(Event,P,WholeEvents,[N2],NS2),
        build_sync_edges_aux2(Event,P,WholeEvents,S1s,[S2|S2s])
       )
      )
     )
    )
   ).


build_sync_edges_aux3(_,_,_,_,_,_,[],[]).   
build_sync_edges_aux3(Event,P,S1,WholeEvents,Draw,EventTreated,[S2|S2s],NSons):-
   /*print(build_sync_edges_aux3__________________(Event,P,S1,WholeEvents,Draw,EventTreated,[S2|S2s])),nl,*/
   (
    slice_event(S2,Event)->
    (
      (
       Draw -> 
       (slice_edge(S1,S2,1)-> Loop = true;(assert(slice_edge(S1,S2,1)),Loop = false));
       Loop = false
      ),
      Found=true,Include = false
    );
    (
     (/*(slice_event(S2,EventTreated),Draw=false)-> 
     (print(entraaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa),nl,Found = true,Loop = false,Include = true);  */
     (( slice_event(S2,OtherEvent),haskell_csp:is_member_set(OtherEvent,set(WholeEvents)) ) ->
     (Found = true,Loop = false,Include = false);
     (Found=false,Loop = false,Include = false)))
    )
   ),
   findall(N,slice_edge(S2,N,2),Ns),
   Include=false,
   /*print(found_include_ns(Found,Include,Ns)),nl,*/
   (Ns=[] -> 
    (
     slice_edge(S2,NodeLoop,3) ->
     (
      go_from_node_to_node(NodeLoop,S2,NodesLoop),
      findall(NEvent,slice_event(NEvent,Event),NEvents),
      haskell_csp:inter_set(set(NodesLoop),set(NEvents),set(Intersection)),
      (Intersection=[] -> NS2 = [-1]; NS2 = [NodeLoop])
     );
     NS2= [-1]
    );
    (
     Ns=[Next]-> 
     (
       slice_node(Next,_,7) ->
       (
        slice_edge(Next,SeqComp,2)->
        (
         findall(Before,slice_edge(Before,SeqComp,2),Befores),
         come_to_node_common(Befores,CommonPath),
         calculate_last(CommonPath,0,Last),
         go_from_node_to_node(Last,Next,PathToSkip),
         (
          (haskell_csp:is_member_set(P,set(PathToSkip));P=Last)->
           NS2=[-1];
           NS2=[Next]
         )
        );
        NS2=[-1]
       );
       NS2 = [Next]
     );
     NS2=Ns
    )
   ),
   /*print(ns2_aux3(NS2)),nl,*/
   (
    Found  -> 
    (
     /*print(found),nl,
     print(loop(Loop)),nl,
     print(build_sync3(Event,P,S1,Draw,S2s)),nl,*/
     Loop -> /*print(loop),nl,*/build_sync_edges_aux3(Event,P,S1,WholeEvents,Draw,EventTreated,S2s,NSons);
     (
      /*print(no_loop(Include)),nl,*/build_sync_edges_aux3(Event,P,S1,WholeEvents,Draw,EventTreated,S2s,NSons1),
      (Include -> NPrefixOpe = S2 ;NS2 = [NPrefixOpe]),
      NSons=[NPrefixOpe|NSons1]
     )
    );
    (
     /*print(no_found),nl,*/
     NS2 = [-1] -> 
     build_sync_edges_aux3(Event,P,S1,WholeEvents,Draw,EventTreated,S2s,NSons);
     (
      concatena(NS2,S2s,NNS2),
      build_sync_edges_aux3(Event,P,S1,WholeEvents,Draw,EventTreated,NNS2,NSons)
     )
    )
   ).

disconnect_non_sync(_,[],_,_).
disconnect_non_sync(P,[Event|Events],S1,S2):-
   disconnect_non_sync_aux(P,Event,[S1],S1,S2),
   disconnect_non_sync_aux(P,Event,[S2],S2,S1),
   disconnect_non_sync(P,Events,S1,S2).

disconnect_non_sync_aux(_,_,[],_,_).
disconnect_non_sync_aux(P,Event,[S1|S1s],S1Ori,S2):-
   slice_node(S1,_,Type),
   findall(NS1,slice_edge(S1,NS1,2),NS1s_),
   haskell_csp:union_set(set(S1s),set(NS1s_),set(NS1s)),
   (
    (slice_event(S1,EventS1),Event=EventS1)->
    (
     go_from_node(S2,GFS2_),
     go_from_loops([S2|GFS2_],GFS2__),
     sort(GFS2__,GFS2),
     findall(SyncS1_1,slice_edge(S1,SyncS1_1,1),SyncS1_1s),print(sync1(SyncS1_1s)),nl,
     findall(SyncS1_2,slice_edge(SyncS1_2,S1,1),SyncS1_2s),print(sync2(SyncS1_2s)),nl,
     (SyncS1_1s=[]-> InterSet1=[];haskell_csp:inter_set(set(SyncS1_1s),set(GFS2),set(InterSet1))),
     (SyncS1_2s=[]-> InterSet2=[];haskell_csp:inter_set(set(SyncS1_2s),set(GFS2),set(InterSet2))),
     (InterSet1=[],InterSet2=[])->
     disconnect_seq_comp(S1Ori,S2),
     retract(slice_edge(_,S1,2)),
     disconnect_non_sync_aux(P,Event,S1s,S1Ori,S2);
     disconnect_non_sync_aux(P,Event,NS1s,S1Ori,S2)
    );
    (
     Type=7,slice_edge(S1,SeqComp,2)->
     (
      findall(Before,slice_edge(Before,SeqComp,2),Befores),
      come_to_node_common(Befores,CommonPath),
      calculate_last(CommonPath,0,Last),
      go_from_node_to_node(Last,S1,PathToSkip),
      (
       ((haskell_csp:is_member_set(P,set(PathToSkip)));(P=Last)) ->
        disconnect_non_sync_aux(P,Event,S1s,S1Ori,S2);
        disconnect_non_sync_aux(P,Event,NS1s,S1Ori,S2)  
      )
     );
     disconnect_non_sync_aux(P,Event,NS1s,S1Ori,S2)
    )
   ).
    
disconnect_seq_comp(S1,S2):-
   disconnect_seq_comp_aux([S1]),
   disconnect_seq_comp_aux([S2]).

disconnect_seq_comp_aux([]).
disconnect_seq_comp_aux([S1|S1s]):-
   slice_node(S1,_,Type),
   findall(NS1,slice_edge(S1,NS1,2),NS1s_),
   haskell_csp:union_set(set(S1s),set(NS1s_),set(NS1s)),
   (
    (Type=7,slice_edge(S1,_,2))-> 
    retract(slice_edge(S1,_,2)),
    disconnect_seq_comp_aux(S1s);
    disconnect_seq_comp_aux(NS1s)
   ). 

go_from_loops([],[]).
go_from_loops([N|Ns],WithLoops):-
   go_from_loops(Ns,WithLoops_),
   (
    slice_edge(N,Loop,3)->
    (
     go_from_node(Loop,WithLoops__),
     haskell_csp:union_set(set(WithLoops_),set([N|WithLoops__]),set(WithLoops))
    );
    haskell_csp:union_set(set(WithLoops_),set([N]),set(WithLoops))
   ).
   

come_to_node_common([N|Ns],CommonPath):-
   come_to_node(N,CTN),
   /*print(acaba_el_come_to_node),nl,*/
   (Ns=[]-> 
    CommonPath = CTN;
    (
     come_to_node_common(Ns,CommonPath1),
     haskell_csp:inter_set(set(CommonPath1),set(CTN),set(CommonPath))
    )
   ).

come_to_node(Id,R):-
   findall(O,slice_edge(O,Id,2),Os),
   (
     (Os = [0];Os = []) -> 
     (R=Os)
     ;
     (
      (
       Os = [O1] ->
       (
        come_to_node(O1,R1),
        haskell_csp:union_set(set(Os),set(R1),set(R))
       )
       ;
       (
        come_to_node_aux(Os,R1),
        haskell_csp:union_set(set(Os),set(R1),set(R))
       )
      )	
     )
    ).


come_to_node_aux([],[]).
come_to_node_aux([O|Os],R):-
   come_to_node(O,R1),
   come_to_node_aux(Os,R2),
   haskell_csp:union_set(set(R1),set(R2),set(R)).

come_to_node_AVL(Id,R):-
   findall(O,slice_edge(O,Id,2),Os_),
   list_to_avlset(Os_,Os),
   (
     (Os = node(0,_,_,empty,empty);Os = empty) -> 
     (R=Os)
     ;
     (
      (
       Os = node(O1,_,_,empty,empty) ->
       (
        come_to_node_AVL(O1,R1),
        avlset_union(Os,R1,R)
       )
       ;
       (
        come_to_node_AVL_aux(Os,R1),
        avlset_union(Os,R1,R)
       )
      )	
     )
    ).

come_to_node_AVL_aux(empty,empty).
come_to_node_AVL_aux(node(O,_,_,T1,T2),R):-
   come_to_node_AVL(O,RO),
   come_to_node_AVL_aux(T1,R1),
   come_to_node_AVL_aux(T2,R2),
   avlset_union(R1,R2,R_),
   avlset_union(RO,R_,R).
/*   haskell_csp:union_set(set(R1),set(R2),set(R)).*/

go_from_node(Id,R):-
   findall(P,slice_edge(Id,P,2),Ps),
   (
     Ps = [] -> 
     (R=[])
     ;
     (
      (
       haskell_csp:my_length(Ps,1) ->
       (
        Ps = [P1],
        go_from_node(P1,R1),
        haskell_csp:union_set(set(Ps),set(R1),set(R))
       )
       ;
       (
        Ps = [P1,P2],
        go_from_node(P1,R1),
        go_from_node(P2,R2),
        haskell_csp:union_set(set(Ps),set(R1),R_),
        haskell_csp:union_set(R_,set(R2),set(R))
       )
      )	
     )
    ).

go_from_node_AVL(Id,R):-
   findall(P,slice_edge(Id,P,2),Ps_),
   list_to_avlset(Ps_,Ps),
   (
     Ps = empty -> 
     R = empty
     ; 
     (
      (
       Ps = node(P1,_,_,empty,empty) ->
       (
        go_from_node_AVL(P1,R1),
        avlset_union(Ps,R1,R)
       )
       ;
       (
        (Ps = node(P1,_,_,empty,(node(P2,_,_,_,_)));Ps = node(P1,_,_,(node(P2,_,_,_,_),empty))),
        go_from_node_AVL(P1,R1),
        go_from_node_AVL(P2,R2),
        avlset_union(Ps,R1,R_),
        avlset_union(R_,R2,R)
       )
      )	
     )
    ).


go_from_node_to_node(Id1,Id2,R):-
   findall(P,slice_edge(Id1,P,2),Ps),
   (
     Ps = [] -> 
     R=[]
     ;
     (
      (
       haskell_csp:my_length(Ps,1) ->
       (
        Ps = [P1],
        (
         P1=Id2 -> R=[Id2];
         (
          go_from_node_to_node(P1,Id2,R1),
          haskell_csp:union_set(set(Ps),set(R1),set(R))
         )
        )
       )
       ;
       (
        Ps = [P1,P2],
        go_from_node_to_node(P1,Id2,R1),
        go_from_node_to_node(P2,Id2,R2),
        (
         haskell_csp:is_member_set(Id2,set(R1))-> R=R1;
         (
          haskell_csp:is_member_set(Id2,set(R2))-> R=R2;
          R=[]
         )
        )
       )
      )	
     )
    ).


 
calculate_last([],Last,Last).
calculate_last([N|Ns],PreviousLast,Last):-
   (
    (N > PreviousLast) ->
    Last1 = N; 
    Last1 = PreviousLast
   ),
   calculate_last(Ns,Last1,Last).


compact_graph([],_,_).
compact_graph([N|Ns],Pending,Treated):-
   /*print(compact_graph_([N|Ns],Pending,Treated)),nl,*/
   /*(haskell_csp:is_member_set(N,set(Treated))->
   compact_graph(Ns,[],Treated);*/
   (N=0->
   Compact=false;
   (
    slice_node(N,_,1)->
    (
     (slice_edge(N,_,1);slice_edge(_,N,1))->
     Compact=false;
     Compact=true
    );
    (
     (slice_node(N,_,3);slice_node(N,_,4)) ->
     Compact=false;
     (
      slice_node(N,_,6)->
      (
       findall(BN,slice_edge(BN,N,2),BNs),
       (
        BNs=[];BNs=[_] -> Compact = true;Compact=false
       )
      );
      (
       (slice_node(N,_,2),(slice_edge(N,_,3);slice_edge(_,N,3)))->
       Compact=false;
       /*(
        slice_node(N,_,7)->
        (
         slice_edge(N,_,2)->
         print(entra_al_then),nl,Compact=true;
         print(entra_al_else),nl,Compact=false
        );*/
        Compact=true
       /*)*/
      )
     )
    )
   )),
   /*print(compact(Compact)),nl,*/
   findall(NN,slice_edge(N,NN,2),NNs1),
   /*print(nns1(NNs1)),nl,*/
   concatena(NNs1,Ns,NNs),
   /*print(nns(NNs)),nl,*/
   (NNs1 = [] -> NCompact=false ; NCompact=Compact),
   /*print(ncompact_nns(NCompact,NNs)),nl,*/
   ( 
    NCompact ->
    ( 
     concatena(Pending,[N],NPending),
     compact_graph(NNs,NPending,Treated)
    );
    (
     (
      Pending=[] ->
      compact_graph(NNs,[],[N|Treated]);
      (
       /*(slice_node(N,_,1);slice_node(N,_,2);slice_node(N,_,3);slice_node(N,_,4);slice_node(N,_,5);slice_node(N,_,6);slice_node(N,_,8))*/
       /*Pending=[_] ->
       (Compact -> (concatena(Pending,[N],NPending),compact_graph(NNs,NPending,Treated));compact_graph(NNs,[],[N|Treated]));*/
       (
        (NNs1=[]->
         /*; slice_node(N,_,7)*/ 
         (Compact -> concatena(Pending,[N],NPending);
         NPending=Pending);
         NPending=Pending
        ),
        (NPending=[_]-> compact_graph(NNs,[],[N|Treated]);
        (
        compact_aux(NPending,[],List),
        get_id(Id),
        /*print(list_compact(Id,List)),nl,*/
        assert(slice_node(Id,src_span(0,0,0,0,0,0),9)),
        assert(slice_compact_node(Id,List)),
        List = [(First,_,_)|_],
        /*print(arriva(First)),nl,*/
        (slice_edge(BFirst,First,2) -> assert(slice_edge(BFirst,Id,2));true),
        /*print(arriva2),nl,*/
        last_of(List,(Last,_,_)),
        findall(NLast,slice_edge(Last,NLast,2),NLasts),
        /*print(last_nlasts(Last,NLasts)),nl,*/
        /*(NNs1=[] -> true;
         (NLasts=[] -> true;
         (NLasts = [NLasts1],assert(slice_edge(Id,NLasts1,2)))
         )
        ),*/
        (NLasts=[] -> true;
        (NLasts = [NLasts1],assert(slice_edge(Id,NLasts1,2)))
        ),
        retract_edges_list(NPending),
        compact_graph(NNs,[],[N|Treated])
       )))
      )  
     )
    )
   ).

last_of([],-1).
last_of([Last],Last).
last_of([_|Ns],Last):-
   last_of(Ns,Last).

compact_aux([],List,List).
compact_aux([P|Pending],List,NList):-
   slice_node(P,PSPAN,PTYPE),
   concatena(List,[(P,PSPAN,PTYPE)],NList1),
   retract(slice_node(P,PSPAN,PTYPE)),
   compact_aux(Pending,NList1,NList).

retract_edges_list([]).
retract_edges_list([N|Ns]):-
   (retract(slice_edge(N,_,2)),retract(slice_edge(_,N,2)) -> true;true),
   retract_edges_list(Ns).

build_dot_graph(N,E,File):-
   tell(File),
   print('digraph CSCFG'),nl,
   print(' {'),nl,
   print('  ordering=out; ranksep=0.5;root=false 0;'),nl,
   /*slice_node(Id,_,Type),
   nodetype2string(Type,SType),*/
   /*write(Stream,id_type_stype(Id,Type,SType)),*/
   build_dot_graph_nodes(N),
   build_dot_graph_edges(E),
   print(' }'),nl,
   told.

build_dot_graph_nodes([]).
build_dot_graph_nodes([N|Ns]):-
   N=(Id,_,Type),
   print('  '),
   print(Id),
   print(' [shape=ellipse, label="('),
   print(Id),
   print(') '),
   print_type(Id,Type),
   print('"];'),nl,
   build_dot_graph_nodes(Ns).


print_type(Id,Type):-
   (Type=1 ->
    (slice_event(Id,Event),
    print(Event));
    (Type=2 ->
     (slice_process_call(Id,Process,_),
     print(Process));
     (
      Type=9 ->
      (
       slice_compact_node(Id,List),
       print('['),
       print_types(List),
       print(']')
      );
      nodetype2string(Type)
     )
    )
   ).

print_types([]).
print_types([(Id,_,T)|List]):-
   print_type(Id,T),
   (List=[] -> true;print(', ')),
   print_types(List).   

build_dot_graph_edges([]).
build_dot_graph_edges([E|Es]):-
   E=(Id1,Id2,Type),
   print('  '),
   print(Id1),
   print(' -> '),
   print(Id2),
   (Type=2 ->
    true;
    (Type=1 ->
     print(' [style=dotted, color=red, arrowhead=none,constraint=false];');
     print('[color=blue];')
    )
   ),nl,
   build_dot_graph_edges(Es).

nodetype2string(1):-print('event').
nodetype2string(2):-print('process_call').
nodetype2string(3):-print('||').
nodetype2string(4):-print('[]').
nodetype2string(5):-print('->').
nodetype2string(6):-print(';').
nodetype2string(7):-print('SKIP').
nodetype2string(8):-print('STOP').
nodetype2string(9):-print('CompactNode').
nodetype2string(10):-print('\\ ').
nodetype2string(11):-print('[[<-]]').


build_dot_graph_slice(N,E,MEB,Color,File):-
   tell(File),
   print('digraph CSCFG'),nl,
   print(' {'),nl,
   print('  ordering=out; ranksep=0.5;root=false 0;'),nl,
   haskell_csp:diff_set(set(N),set(MEB),set(NotSlice)),
   build_dot_graph_nodes(NotSlice),
   build_dot_graph_nodes_slice(MEB,Color),
   build_dot_graph_edges(E),
   print(' }'),nl,
   told.

build_dot_graph_nodes_slice([],_).
build_dot_graph_nodes_slice([N|Ns],Color):-
   N=(Id,_,Type),
   print('  '),
   print(Id),
   print(' [shape=ellipse, label="('),
   print(Id),
   print(') '),
   print_type(Id,Type),
   print('", color='),
   print(Color),
   print('];'),nl,
   build_dot_graph_nodes_slice(Ns,Color).




/*
look_for_agent_call('[]'(A,B),Processes):-
   look_for_agent_call(A,ProcessesA),look_for_agent_call(B,ProcessesB),haskell_csp:union_set(ProcessesA,ProcessesB,Processes).

look_for_agent_call(prefix(_,_,_,CSP),Processes):-
   look_for_agent_call(CSP,Processes).

look_for_agent_call(agent_call(F,_),set([F])).

look_for_agent_call(aParallel(_,A,_,B,_),Processes):-
   look_for_agent_call(A,ProcessesA),look_for_agent_call(B,ProcessesB),haskell_csp:union_set(ProcessesA,ProcessesB,Processes).

look_for_agent_call(skip(_),set([])).

look_for_agent_call(stop,set([])).

look_for_agent_call(';'(A,B),Processes):-
   look_for_agent_call(A,ProcessesA),look_for_agent_call(B,ProcessesB),haskell_csp:union_set(ProcessesA,ProcessesB,Processes).

look_for_agent_call(sharing(_,A,B,_),Processes):-
   look_for_agent_call(A,ProcessesA),look_for_agent_call(B,ProcessesB),haskell_csp:union_set(ProcessesA,ProcessesB,Processes).

look_for_agent_call('|~|'(A,B,_),Processes):-
   look_for_agent_call(A,ProcessesA),look_for_agent_call(B,ProcessesB),haskell_csp:union_set(ProcessesA,ProcessesB,Processes).
*/

write_all(File):-
   tell(File),

   findall(A,slice_node(A,_,_),As),
   findall(B,slice_node(_,B,_),Bs),
   findall(C,slice_node(_,_,C),Cs),
   (As=[] -> (print(slice_node_(-1,src_span(0,0,0,0,0,0),-1)),print('.'),nl);write_all_node(As,Bs,Cs)),

   findall(A1,slice_edge(A1,_,_),A1s),
   findall(B1,slice_edge(_,B1,_),B1s),
   findall(C1,slice_edge(_,_,C1),C1s),
   (A1s=[] -> (print(slice_edge_(-1,-1,-1)),print('.'),nl);write_all_edge(A1s,B1s,C1s)),

   findall(A2,slice_event(A2,_),A2s),
   findall(B2,slice_event(_,B2),B2s),
   (A2s=[] -> (print(slice_event_(-1,a)),print('.'),nl);write_all_event(A2s,B2s)),

   findall(A3,slice_process_call(A3,_,_),A3s),
   findall(B3,slice_process_call(_,B3,_),B3s),
   findall(C3,slice_process_call(_,_,C3),C3s),
   (A3s=[] -> (print(slice_process_call_(-1,'0',[])),print('.'),nl);write_all_process_call(A3s,B3s,C3s)),

   findall(A4,slice_parallelism(A4,_,_,_),A4s),
   findall(B4,slice_parallelism(_,B4,_,_),B4s),
   findall(C4,slice_parallelism(_,_,C4,_),C4s),
   findall(D4,slice_parallelism(_,_,_,D4),D4s),
   (A4s=[] -> (print(slice_parallelism_(-1,[],-1,-1)),print('.'),nl);write_all_parallelism(A4s,B4s,C4s,D4s)),

   findall(A5,slice_compact_node(A5,_),A5s),
   findall(B5,slice_compact_node(_,B5),B5s),
   (A5s=[] -> (print(slice_compact_node_(-1,[])),print('.'),nl);write_all_compact_node(A5s,B5s)),

   findall(A6,slice_sequential(A6,_),A6s),
   findall(B6,slice_sequential(_,B6),B6s),
   (A6s=[] -> (print(slice_sequential_(-1,-1)),print('.'),nl);write_all_sequential(A6s,B6s)),

   findall(A7,slice_choice(A7,_,_,_),A7s),
   findall(B7,slice_choice(_,B7,_,_),B7s),
   findall(C7,slice_choice(_,_,C7,_),C7s),
   findall(D7,slice_choice(_,_,_,D7),D7s),
   (A7s=[] -> (print(slice_choice_(-1,-1,-1,-1)),print('.'),nl);write_all_choice(A7s,B7s,C7s,D7s)),

   findall(A8,slice_if(A8,_),A8s),
   findall(B8,slice_if(_,B8),B8s),
   (A8s=[] -> (print(slice_if_(-1,true)),print('.'),nl);write_all_if(A8s,B8s)),

   findall(A9,slice_renamed_event(A9,_),A9s),
   findall(B9,slice_renamed_event(_,B9),B9s),
   (A9s=[] -> (print(slice_renamed_event_(-1,a)),print('.'),nl);write_all_renamed_event(A9s,B9s)),

   findall(A10,slice_hiding(A10,_),A10s),
   findall(B10,slice_hiding(_,B10),B10s),
   (A10s=[] -> (print(slice_hiding_(-1,a)),print('.'),nl);write_all_hiding(A10s,B10s)),

   findall(A11,slice_renaming(A11,_,_),A11s),
   findall(B11,slice_renaming(_,B11,_),B11s),
   findall(C11,slice_renaming(_,_,C11),C11s),
   (A11s=[] -> (print(slice_renaming_(-1,a,a)),print('.'),nl);write_all_renaming(A11s,B11s,C11s)),

   told.

write_all_node([],[],[]).
write_all_node([A|As],[B|Bs],[C|Cs]):-
   print(slice_node_(A,B,C)),print('.'),nl,
   write_all_node(As,Bs,Cs).

write_all_edge([],[],[]).
write_all_edge([A|As],[B|Bs],[C|Cs]):-
   print(slice_edge_(A,B,C)),print('.'),nl,
   write_all_edge(As,Bs,Cs).

write_all_event([],[]).
write_all_event([A|As],[B|Bs]):-
   print(slice_event_(A,B)),print('.'),nl,
   write_all_event(As,Bs).

write_all_process_call([],[],[]).
write_all_process_call([A|As],[B|Bs],[C|Cs]):-
   print('slice_process_call_('),
   print(A),print(',\''),
   print(B),print('\','),
   print(C),print(').'),nl,
   write_all_process_call(As,Bs,Cs).

write_all_parallelism([],[],[],[]).
write_all_parallelism([A|As],[B|Bs],[C|Cs],[D|Ds]):-
   print(slice_parallelism_(A,B,C,D)),print('.'),nl,
   write_all_parallelism(As,Bs,Cs,Ds).

write_all_compact_node([],[]).
write_all_compact_node([A|As],[B|Bs]):-
   print(slice_compact_node_(A,B)),print('.'),nl,
   write_all_compact_node(As,Bs).

write_all_sequential([],[]).
write_all_sequential([A|As],[B|Bs]):-
   print(slice_sequential_(A,B)),print('.'),nl,
   write_all_sequential(As,Bs).

write_all_choice([],[],[],[]).
write_all_choice([A|As],[B|Bs],[C|Cs],[D|Ds]):-
   print(slice_choice_(A,B,C,D)),print('.'),nl,
   write_all_choice(As,Bs,Cs,Ds).

write_all_if([],[]).
write_all_if([A|As],[B|Bs]):-
   print(slice_if_(A,B)),print('.'),nl,
   write_all_if(As,Bs).

write_all_renamed_event([],[]).
write_all_renamed_event([A|As],[B|Bs]):-
   print(slice_renamed_event_(A,B)),print('.'),nl,
   write_all_renamed_event(As,Bs).

write_all_hiding([],[]).
write_all_hiding([A|As],[B|Bs]):-
   print(slice_hiding_(A,B)),print('.'),nl,
   write_all_hiding(As,Bs).

write_all_renaming([],[],[]).
write_all_renaming([A|As],[B|Bs],[C|Cs]):-
   print(slice_renaming_(A,B,C)),print('.'),nl,
   write_all_renaming(As,Bs,Cs).

read_graph(File):-

   consult(File),

   findall(A,slice_node_(A,_,_),As),
   findall(B,slice_node_(_,B,_),Bs),
   findall(C,slice_node_(_,_,C),Cs),
   translate_node(As,Bs,Cs),

   findall(A1,slice_edge_(A1,_,_),A1s),
   findall(B1,slice_edge_(_,B1,_),B1s),
   findall(C1,slice_edge_(_,_,C1),C1s),
   translate_edge(A1s,B1s,C1s),

   findall(A2,slice_event_(A2,_),A2s),
   findall(B2,slice_event_(_,B2),B2s),
   translate_event(A2s,B2s),

   findall(A3,slice_process_call_(A3,_,_),A3s),
   findall(B3,slice_process_call_(_,B3,_),B3s),
   findall(C3,slice_process_call_(_,_,C3),C3s),
   translate_process_call(A3s,B3s,C3s),

   findall(A4,slice_parallelism_(A4,_,_,_),A4s),
   findall(B4,slice_parallelism_(_,B4,_,_),B4s),
   findall(C4,slice_parallelism_(_,_,C4,_),C4s),
   findall(D4,slice_parallelism_(_,_,_,D4),D4s),
   translate_parallelism(A4s,B4s,C4s,D4s),

   findall(A5,slice_compact_node_(A5,_),A5s),
   findall(B5,slice_compact_node_(_,B5),B5s),
   translate_compact_node(A5s,B5s),

   findall(A6,slice_sequential_(A6,_),A6s),
   findall(B6,slice_sequential_(_,B6),B6s),
   translate_sequential(A6s,B6s),

   findall(A7,slice_choice_(A7,_,_,_),A7s),
   findall(B7,slice_choice_(_,B7,_,_),B7s),
   findall(C7,slice_choice_(_,_,C7,_),C7s),
   findall(D7,slice_choice_(_,_,_,D7),D7s),
   translate_choice(A7s,B7s,C7s,D7s),

   findall(A8,slice_if_(A8,_),A8s),
   findall(B8,slice_if_(_,B8),B8s),
   translate_if(A8s,B8s),

   findall(A9,slice_renamed_event_(A9,_),A9s),
   findall(B9,slice_renamed_event_(_,B9),B9s),
   translate_renamed_event(A9s,B9s),

   findall(A10,slice_hiding_(A10,_),A10s),
   findall(B10,slice_hiding_(_,B10),B10s),
   translate_hiding(A10s,B10s),

   findall(A11,slice_renaming_(A11,_,_),A11s),
   findall(B11,slice_renaming_(_,B11,_),B11s),
   findall(C11,slice_renaming_(_,_,C11),C11s),
   translate_renaming(A11s,B11s,C11s).

translate_node([],[],[]).
translate_node([A|As],[B|Bs],[C|Cs]):-
   (A = -1-> true;assert(slice_node(A,B,C))),
   translate_node(As,Bs,Cs).

translate_edge([],[],[]).
translate_edge([A|As],[B|Bs],[C|Cs]):-
   (A = -1-> true;assert(slice_edge(A,B,C))),
   translate_edge(As,Bs,Cs).

translate_event([],[]).
translate_event([A|As],[B|Bs]):-
   (A = -1-> true;assert(slice_event(A,B))),
   translate_event(As,Bs).

translate_process_call([],[],[]).
translate_process_call([A|As],[B|Bs],[C|Cs]):-
   (A = -1-> true;assert(slice_process_call(A,B,C))),
   translate_process_call(As,Bs,Cs).

translate_parallelism([],[],[],[]).
translate_parallelism([A|As],[B|Bs],[C|Cs],[D|Ds]):-
   (A = -1-> true;assert(slice_parallelism(A,B,C,D))),
   translate_parallelism(As,Bs,Cs,Ds).

translate_compact_node([],[]).
translate_compact_node([A|As],[B|Bs]):-
   (A = -1-> true;assert(slice_compact_node(A,B))),
   translate_compact_node(As,Bs).

translate_sequential([],[]).
translate_sequential([A|As],[B|Bs]):-
   (A = -1-> true;assert(slice_sequential(A,B))),
   translate_sequential(As,Bs).

translate_choice([],[],[],[]).
translate_choice([A|As],[B|Bs],[C|Cs],[D|Ds]):-
   (A = -1-> true;assert(slice_choice(A,B,C,D))),
   translate_choice(As,Bs,Cs,Ds).

translate_if([],[]).
translate_if([A|As],[B|Bs]):-
   (A = -1-> true;assert(slice_if(A,B))),
   translate_if(As,Bs).

translate_renamed_event([],[]).
translate_renamed_event([A|As],[B|Bs]):-
   (A = -1-> true;assert(slice_renamed_event(A,B))),
   translate_renamed_event(As,Bs).

translate_hiding([],[]).
translate_hiding([A|As],[B|Bs]):-
   (A = -1-> true;assert(slice_hiding(A,B))),
   translate_hiding(As,Bs).

translate_renaming([],[],[]).
translate_renaming([A|As],[B|Bs],[C|Cs]):-
   (A = -1-> true;assert(slice_renaming(A,B,C))),
   translate_renaming(As,Bs,Cs).
