% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(slicer_csp_MEB_CEB, [meb/3,ceb/6,look_for_sc/5]).

:- use_module(haskell_csp).

:- use_module(error_manager,'../error_manager', [extract_line_col/5]).

:- use_module(slicer_csp_CSCFG).
:- use_module(avlsets).

look_for_sc([],_,_,_,_).
look_for_sc([List|Lists],FromL,FromC,ToL,ToC):-
   (look_for_sc_aux(List,FromL,FromC,ToL,ToC)-> uncompact(List);true),
   look_for_sc(Lists,FromL,FromC,ToL,ToC).


look_for_sc_aux([(_,src_span(FL,FC,TL,TC,_,_),_)|List],FromL,FromC,ToL,ToC):-
   (FL = FromL,
   FC = FromC,
   TL = ToL,
   TC = ToC);
   look_for_sc_aux(List,FromL,FromC,ToL,ToC).

uncompact(List):-
   /*print(uncompact_(List)),nl,*/
   findall(CN,slice_compact_node(CN,List),[CNList]),
   /*print(cnlist(CNList)),nl,*/
   slice_edge(Start,CNList,2),
   /*print(cnlist(Start)),nl,*/
   uncompact_aux(List,Start,Last),
   (slice_edge(CNList,Next,2)->
    (assert(slice_edge(Last,Next,2)),retract(slice_edge(CNList,Next,2)));
    true
   ),
   retract(slice_edge(_,CNList,2)),
   retract(slice_node(CNList,_,_)).
   

uncompact_aux([],L,L).
uncompact_aux([(Id,SPAN,Type)|Ns],S,F):-
   /*print(uncompact_aux_([(Id,SPAN,Type)|Ns],S,F)),nl,*/
   assert(slice_node(Id,SPAN,Type)),
   assert(slice_edge(S,Id,2)),
   uncompact_aux(Ns,Id,F).

select_top_most([],_,[]).
select_top_most([Id|Ids],FullIds,IdsTM):-
   come_to_node(Id,CTId),
   haskell_csp:inter_set(set(CTId),set(FullIds),set(Common)),
   select_top_most(Ids,FullIds,IdsTM_),
   (Common=[] -> IdsTM = [Id|IdsTM_];IdsTM = IdsTM_).

   
meb(Ids,MEB,MEB_CEB):-
   select_top_most(Ids,Ids,IdsFinal),
   print(select_top_most_(Ids,IdsFinal)),nl,
   meb_aux(IdsFinal,MEB,MEB_CEB).


meb_aux([],[],[]).
meb_aux([N|Ns],MEB,MEB_CEB):-
   build_MEB(N,MEB1),
   (
    Ns = [] -> 
    (MEB = MEB1,MEB_CEB=MEB1) 
    ; 
    (meb_aux(Ns,MEB2,MEB_CEB2),haskell_csp:union_set(set(MEB1),set(MEB_CEB2),set(MEB_CEB)),intersection_meb(MEB1,MEB2,MEB))
   ).

intersection_meb([],Ns,Ns).
intersection_meb(Ns,[],Ns).
intersection_meb(Ns,Ms,Intersection):-
   extract_spans(Ns,SpansNs_),
   extract_spans(Ms,SpansMs_),
   sort(SpansNs_,SpansNs),
   sort(SpansMs_,SpansMs),
   haskell_csp:inter_set(set(SpansNs),set(SpansMs),set(InterSpans)),
   haskell_csp:union_set(set(Ns),set(Ms),set(NsUMs)),
   extract_nodes_from_spans(InterSpans,NsUMs,Intersection).

extract_spans([],[]).
extract_spans([N|Ns],SPANs):-
   slice_node(N,SPAN,_),
   extract_spans(Ns,SPANs_),
   (
    slice_compact_node(N,ListCN)-> 
    (
     extract_spans_cn(ListCN,SPANsListCN),
     haskell_csp:union_set(set(SPANsListCN),set(SPANs_),set(SPANs))
    );
    SPANs = [SPAN|SPANs_]
   ).

extract_spans_cn([],[]).
extract_spans_cn([(_,SPAN,_)|Ns],[SPAN|SPANs]):-extract_spans_cn(Ns,SPANs).  


extract_nodes_from_spans([],_,[]).
extract_nodes_from_spans([SPAN|SPANs],Ns,NSPANs):-
   extract_nodes_from_spans_aux(SPAN,Ns,NSPANs1),
   extract_nodes_from_spans(SPANs,Ns,NSPANs2),
   haskell_csp:union_set(set(NSPANs1),set(NSPANs2),set(NSPANs)).

extract_nodes_from_spans_aux(_,[],[]).
extract_nodes_from_spans_aux(SPAN,[N|Ns],NSPANs):-
   extract_nodes_from_spans_aux(SPAN,Ns,NSPANs_),   
   (slice_node(N,SPAN,_) -> 
    NSPANs = [N|NSPANs_];
    (
     slice_compact_node(N,ListCN)->
     (
      extract_spans_cn(ListCN,SPANSListCN),
      (
       haskell_csp:is_member_set(SPAN,set(SPANSListCN))->
       NSPANs = [N|NSPANs_];
       NSPANs=NSPANs_
      )
     )
     ;
     NSPANs=NSPANs_
    )
   ).

build_MEB(N,NewMEB):-

   findall(O1,slice_edge(O1,N,1),O1s),
   findall(O2,slice_edge(N,O2,1),O2s),   
   list_to_avlset(O1s,O1sAVL),
   list_to_avlset(O2s,O2sAVL),
   list_to_avlset([N],NAVL),
   avlset_union(O1sAVL,O2sAVL,Os1),
   avlset_union(Os1,NAVL,Os),
   build_pending(Os,Init),

   build_MEB_initial(Init,MEB1),
   avlset_union(MEB1,Init,MEB),/*Because come_to_node not include the last node*/


   go_from_node_AVL(N,BlackList_),
   avlset_union(BlackList_,NAVL,BlackList),

   build_sync1(MEB,Pending1),

   avlset_union(MEB,BlackList,NotQ),
   come_to_node_AVL(N,CTN),
   build_sync2(MEB,MEB,N,NotQ,CTN,Pending2),
   
   avlset_union(Pending1,Pending2,Pending),
 
   build_MEB_while(MEB,BlackList,N,Pending,CTN,NewMEB_),
   avlset_to_list(NewMEB_,NewMEB),
   print(build_meb_(N,NewMEB)),nl.

build_MEB_while(MEB,_,_,empty,_,MEB).
build_MEB_while(MEB,BlackList,N,node(M,BT,NT,T1,T2),CTN,NMEB):-


   list_to_avlset([M],MAVL),
   avlset_union(MEB,MAVL,NewMEB1),
   come_to_node_AVL(M,CTR),
   avlset_union(NewMEB1,CTR,NewMEB),

   build_sync1(NewMEB,Sync1),

   statistics(runtime,_),
   statistics(walltime,_),

   avlset_union(NewMEB,BlackList,NotQ_),
   avlset_union(NotQ_,(node(M,BT,NT,T1,T2)),NotQ),

   build_sync2(NewMEB,NewMEB,N,NotQ,CTN,Sync2),  

   avlset_union(Sync1,Sync2,Sync),

   avlset_union((node(M,BT,NT,T1,T2)),Sync,NewPending1),

   (T1=empty,T2=empty -> NewPending = empty;
    (NewMEB = empty -> NewPending = NewPending1;
     avlset_difference(NewPending1,NewMEB,NewPending)
    ) 
   ),
   /*print(sync(Sync)),nl,
   print(new_pending(NewPending)),nl,
   print(newMEB(NewMEB)),nl,*/


   (NewPending= empty -> (NMEB = NewMEB);
    build_MEB_while(NewMEB,BlackList,N,NewPending,CTN,NMEB)
   ).



   
build_pending(empty,empty).
build_pending(node(O,_,_,T1,T2),Pending):-
   findall(N,slice_edge(N,O,2),Ns),
   list_to_avlset(Ns,NsAVL),
   build_pending(T1,Pending1),
   build_pending(T2,Pending2),
   avlset_union(Pending1,Pending2,Pending_),
   avlset_union(Pending_,NsAVL,Pending).


build_MEB_initial(empty,empty).
build_MEB_initial(node(M,_,_,T1,T2),MEB):-
   come_to_node_AVL(M,CTR),
   build_MEB_initial(T1,MEB1),
   build_MEB_initial(T2,MEB2),
   avlset_union(MEB1,MEB2,MEB_),
   avlset_union(MEB_,CTR,MEB).

build_sync1(empty,empty).
build_sync1(node(R,_,_,T1,T2),Sync):-
   findall(Q1,slice_edge(Q1,R,1),Q1s),
   findall(Q2,slice_edge(R,Q2,1),Q2s),
   list_to_avlset(Q1s,Q1sAVL),
   list_to_avlset(Q2s,Q2sAVL),
   avlset_union(Q1sAVL,Q2sAVL,Qs),
   build_sync1(T1,SyncT1),
   build_sync1(T2,SyncT2),
   avlset_union(SyncT1,SyncT2,SyncT),
   avlset_union(Qs,SyncT,Sync).


build_sync2(empty,_,_,_,_,empty).
build_sync2(node(R,_,_,T1,T2),MEB,N,NotQ,CTN,Sync):-
   (
    avlset_member(R,CTN)->
    Qs = empty;
    (
     findall(QL,slice_edge(QL,R,3),QLS1),
     build_sync_aux(QLS1,R,MEB,Qs_),
     list_to_avlset(Qs_,Qs)
     /*print(r_qls(R,QLS)),nl*/
    )
   ),
   (
    Qs=empty-> 
    SyncR = empty;
    avlset_difference(Qs,NotQ,SyncR)/*,print(entra_al_diff(R,Qs,NotQ,SyncR)),nl*/
   ),
   build_sync2(T1,MEB,N,NotQ,CTN,SyncMEB1),
   build_sync2(T2,MEB,N,NotQ,CTN,SyncMEB2),
   avlset_union(SyncMEB1,SyncMEB2,SyncMEBT),
   avlset_union(SyncR,SyncMEBT,Sync).


build_sync_aux([],_,_,[]).
build_sync_aux([QL|QLS1],R,MEB,QLS):-
   go_from_node_to_node_AVL(R,QL,PathRQL_),
   list_to_avlset([QL],QLAVL),
   avlset_union(QLAVL,PathRQL_,PathRQL),
   build_sync_aux2(PathRQL,MEB,SyncMEB1),
   build_sync_aux(QLS1,R,MEB,QLS2),  
   (SyncMEB1=empty -> QLS=QLS2; QLS=[QL|QLS2]).
   
build_sync_aux2(empty,_,empty).
build_sync_aux2(node(A,_,_,T1,T2),MEB,SyncMEB):-
   findall(B,slice_edge(B,A,1),Bs),
   findall(C,slice_edge(A,C,1),Cs),
   list_to_avlset(Bs,BsAVL),
   list_to_avlset(Cs,CsAVL),
   avlset_union(BsAVL,CsAVL,BCs),
   avlset_intersection(BCs,MEB,SyncMEB_),
   build_sync_aux2(T1,MEB,SyncMEBT1),
   build_sync_aux2(T2,MEB,SyncMEBT2),
   avlset_union(SyncMEBT1,SyncMEBT2,SyncMEBT),
   avlset_union(SyncMEB_,SyncMEBT,SyncMEB).



ceb(_,_,_,_,[],[]).
ceb(FromL,FromC,ToL,ToC,MEB,NCEB):-
   /*print(entra_ceb),nl,*/
   retractall(ceb_fp_control(_,_)),
   findall(Choice,slice_node(Choice,_,4),NsChoice),
   haskell_csp:inter_set(set(NsChoice),set(MEB),set(NsChoiceMEB)),
   /*print(build_loopnodes_(NsChoiceMEB,MEB)),nl,*/
   build_loopnodes(NsChoiceMEB,MEB,Loopnodes),
   /*print(loopnodes(Loopnodes)),nl,*/
   haskell_csp:union_set(set(MEB),set(Loopnodes),set(CEB)),
   /*print(ceb(CEB)),nl,*/
   findall(N,slice_node(N,_,_),Ns),
   NFromC is FromC +1,
   NToC is ToC +1,
   findall(NSlice,slice_node(NSlice,src_span(FromL,NFromC,ToL,NToC,_,_),_),NsSlice),
   haskell_csp:union_set(set(CEB),set(NsSlice),NotCEB),
   haskell_csp:diff_set(set(Ns),NotCEB,set(Ms)),
   haskell_csp:inter_set(set(NsChoice),set(CEB),NsChoiceCEB),
   haskell_csp:diff_set(set(CEB),NsChoiceCEB,set(MPs)),
   /*print(build_pending_ceb1_(Ms,MPs)),nl,*/
   build_pending_ceb1(Ms,MPs,Pending1),
   findall(Paralel,slice_node(Paralel,_,3),NsParalel),
   haskell_csp:inter_set(set(NsParalel),set(Loopnodes),set(NsParalelLoopnodes)),
   /*print(build_pending_ceb2_(NsParalelLoopnodes,Loopnodes)),nl,*/
   build_pending_ceb2(NsParalelLoopnodes,Loopnodes,Pending2),
   /*print(pending1_pending2(Pending1,Pending2)),nl,*/
   haskell_csp:union_set(set(Pending1),set(Pending2),set(Pending)),
   /*print(ceb_repeat_(Pending,CEB)),nl,*/
   ceb_repeat(Pending,CEB,NCEB).


build_pending_ceb1([],_,[]).
build_pending_ceb1([N|Ns],Ms,Pending):- 
   find_edge_meb(N,Ms,VN),
   build_pending_ceb1(Ns,Ms,Pending1),
   haskell_csp:union_set(set(VN),set(Pending1),set(Pending)).


find_edge_meb(_,[],[]).
find_edge_meb(N,[M|MPs],VN):-
   (
    slice_edge(M,N,2) ->  
     VN=[N];
     find_edge_meb(N,MPs,VN)
   ).

build_pending_ceb2([],_,[]).
build_pending_ceb2([N|Ns],Loopnodes,Pending):- 
   findall(S,slice_edge(N,S,2),[S1,S2]),
   (haskell_csp:is_member_set(S1,set(Loopnodes))-> P1 = S2 ; P1 = S1),
   build_pending_ceb2(Ns,Loopnodes,Pending1),
   Pending=[P1|Pending1].

build_loopnodes([],_,[]).
build_loopnodes([N|Ns],MEB,Loopnodes):-
   go_from_node(N,RN),
   haskell_csp:diff_set(set(RN),set(MEB),set(RN_not_in_MEB)),
   /*print(build_loopnodes_aux_(RN_not_in_MEB,N,MEB)),nl,*/
   build_loopnodes_aux(RN_not_in_MEB,N,MEB,Loopnodes1),
   /*print(rESULT___________build_loopnodes_aux(Loopnodes1)),nl,*/
   build_loopnodes(Ns,MEB,Loopnodes2),
   haskell_csp:union_set(set(Loopnodes1),set(Loopnodes2),set(Loopnodes)).


build_loopnodes_aux([],_,_,[]).
build_loopnodes_aux([N|Ns],Choice,MEB,Loopnodes):-
    (slice_edge(N,N3,3) -> 
     (
      haskell_csp:is_member_set(N3,set(MEB))-> 
      (
       go_from_node_to_node(Choice,N,Loopnodes2),
       /*print(go_from_node_to_node_(Choice,N,Loopnodes2)),nl,*/
       (find_sync_edge(Loopnodes2,MEB,Loopnodes2)->
        (build_loopnodes_aux(Ns,Choice,MEB,Loopnodes1),
        haskell_csp:union_set(set(Loopnodes1),set(Loopnodes2),set(Loopnodes)));
        Loopnodes = []       
       )
      );
      build_loopnodes_aux(Ns,Choice,MEB,Loopnodes)
     );
     build_loopnodes_aux(Ns,Choice,MEB,Loopnodes)
    ).

find_sync_edge([],_,[]).
find_sync_edge([N|Ns],MEB,NNs):-
   findall(NP1,slice_edge(NP1,N,1),NP1s),
   findall(NP2,slice_edge(N,NP2,1),NP2s),
   haskell_csp:union_set(set(NP1s),set(NP2s),set(NPs)),
   find_sync_edge(Ns,MEB,NNs1),
   (NPs = [] -> 
    NNs = [N|NNs1];
    (haskell_csp:union_set(set(NPs),set(MEB),set(MEB)) -> NNs = [N|NNs1];NNs = NNs1)
   ).

ceb_repeat([],CEB,CEB).
ceb_repeat([M|Pending],CEB,FCEB):-
    /*findall(MP1,slice_edge(NCEB,MP1,1),MP1s),
    findall(MP2,slice_edge(MP2,NCEB,1),MP2s),*/
    findall(MP1,slice_edge(M,MP1,1),MP1s),
    findall(MP2,slice_edge(MP2,M,1),MP2s),
    haskell_csp:union_set(set(MP1s),set(MP2s),MPs1),
    haskell_csp:inter_set(MPs1,set(CEB),set(MPs)),
    /*print(ceb_repeat_(M,MPs1,Pending)),nl,*/
    /*print(mps(MPs)),nl,*/
    (
     MPs1=set([])->
     (
      haskell_csp:diff_set(set(Pending),set([M]),NPending1),
      haskell_csp:union_set(set(CEB),set([M]),set(NCEB)),
      /*print(nceb_npending1(NCEB,NPending1)),nl,*/
      findall(MPP,slice_edge(M,MPP,2),MPPs1),
      haskell_csp:diff_set(set(MPPs1),set(NCEB),MPPs),
      haskell_csp:union_set(NPending1,MPPs,set(NPending))
     );
     (
     MPs=[]->
     (
      haskell_csp:inter_set(MPs1,set(Pending),PendingSync1),
      /*print(pendingsync1(PendingSync1,MPs1)),nl,*/
      (
       (PendingSync1=set([]);MPs1=set([])) -> 
       PendingSync = [-1];
       haskell_csp:diff_set(PendingSync1,MPs1,set(PendingSync))
      ),
      /*print(pendingsync(PendingSync)),nl,*/
      (
       PendingSync = [] ->
       (
        haskell_csp:union_set(set(CEB),set([M]),NCEB1),
        haskell_csp:union_set(NCEB1,MPs1,set(NCEB)),
        haskell_csp:diff_set(set(Pending),set(NCEB),NPending1),
        findall(N,slice_edge(M,N,2),Ns),
        haskell_csp:diff_set(set(Ns),set(NCEB),NPending2),
        MPs1 = set(Candidates),
        next_candidates(Candidates,NextCandidates),
        haskell_csp:diff_set(set(NextCandidates),set(NCEB),NPending3),
        haskell_csp:union_set(NPending1,NPending2,NPending4),
        haskell_csp:union_set(NPending3,NPending4,set(NPending))
       );
       (
        NCEB = CEB,
        concatena(Pending,[M],NPending)
       )
      )
     );
     (
      haskell_csp:diff_set(set(Pending),set([M]),NPending1),
      haskell_csp:union_set(set(CEB),set([M]),set(NCEB)),
      /*print(nceb_npending1(NCEB,NPending1)),nl,*/
      findall(MPP,slice_edge(M,MPP,2),MPPs1),
      haskell_csp:diff_set(set(MPPs1),set(NCEB),MPPs),
      haskell_csp:union_set(NPending1,MPPs,set(NPending))
     )
    )),
    (
     ceb_fp_control(NCEB,NPending)->
     FCEB = CEB;
     (
      assert(ceb_fp_control(NCEB,NPending)),
      ceb_repeat(NPending,NCEB,FCEB)
     )
    ).
    /*ceb_repeat_fp(CEB,NCEB,[M|Pending],NPending,FCEB).
   
ceb_repeat_fp(CEB,NCEB,Pending,NPending,FCEB):-
   (NCEB=CEB,NPending=Pending) ->
   FCEB = CEB;
   ceb_repeat(NPending,NCEB,FCEB).*/

concatena([],L,L).
concatena([X|L1],L2,[X|L3]):-concatena(L1,L2,L3).
   
next_candidates([],[]).
next_candidates([C|Cs],Next):-
   findall(O,slice_edge(C,O,2),Os),
   next_candidates(Cs,Next1),
   haskell_csp:union_set(set(Os),set(Next1),set(Next)).

/*
go_from_node(Id,R):-
   findall(P,slice_edge(Id,P,2),Ps),
   (
     Ps = [] -> 
     (R=[])
     ;
     (
      (
       haskell_csp:my_length(Ps,1) ->
       (
        Ps = [P1],
        go_from_node(P1,R1),
        haskell_csp:union_set(set(Ps),set(R1),set(R))
       )
       ;
       (
        Ps = [P1,P2],
        go_from_node(P1,R1),
        go_from_node(P2,R2),
        haskell_csp:union_set(set(Ps),set(R1),R_),
        haskell_csp:union_set(R_,set(R2),set(R))
       )
      )	
     )
    ).
*/

go_from_node_to_node(Id1,Id2,R):-
   findall(P,slice_edge(Id1,P,2),Ps),
   (
     Ps = [] -> 
     R=[]
     ;
     (
      (
       haskell_csp:my_length(Ps,1) ->
       (
        Ps = [P1],
        (
         P1=Id2 -> R=[Id2];
         (
          go_from_node_to_node(P1,Id2,R1),
          haskell_csp:union_set(set(Ps),set(R1),set(R))
         )
        )
       )
       ;
       (
        Ps = [P1,P2],
        go_from_node_to_node(P1,Id2,R1),
        go_from_node_to_node(P2,Id2,R2),
        (
         haskell_csp:is_member_set(Id2,set(R1))-> haskell_csp:union_set(set([P1]),set(R1),set(R));
         (
          haskell_csp:is_member_set(Id2,set(R2))-> haskell_csp:union_set(set([P2]),set(R2),set(R));
          R=[]
         )
        )
       )
      )	
     )
    ).

go_from_node_to_node_AVL(Id1,Id2,R):-
   findall(P,slice_edge(Id1,P,2),Ps_),
   list_to_avlset(Ps_,Ps),
   (
     Ps = empty -> 
     R=empty
     ;
     (
      (
       Ps = node(P1,_,_,empty,empty)->
       (
        (
         P1=Id2 -> R=node(Id2,_,_,empty,empty);
         (
          go_from_node_to_node_AVL(P1,Id2,R1),
          avlset_union(Ps,R1,R)
         )
        )
       )
       ;
       (
        (Ps = node(P1,_,_,empty,(node(P2,_,_,_,_)));Ps = node(P1,_,_,(node(P2,_,_,_,_),empty))),
        go_from_node_to_node_AVL(P1,Id2,R1),
        go_from_node_to_node_AVL(P2,Id2,R2),
        (
         avlset_member(Id2,R1)-> list_to_avlset([P1],P1AVL),avlset_union(P1AVL,R1,R);
         (
          avlset_member(Id2,R2)-> list_to_avlset([P2],P2AVL),avlset_union(P2AVL,R2,R);
          R=empty
         )
        )
       )
      )	
     )
    ).
