% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(coverage_statistics,[tcltk_compute_coverage/1, tcltk_compute_coverage_and_enabling/1,
                               tcltk_compute_min_max/1,pretty_print_min_max_coverage_to_file/1,
                               compute_the_coverage/4]).

:- use_module(state_space).
:- use_module(tools).
:- use_module(translate).

/* --------------------------------------------------- */
:- dynamic var_min_max/3.

tcltk_compute_min_max(_) :-
   retractall(var_min_max(_,_,_)),
   visited_expression(_,State),
   treat_state(State),fail.
tcltk_compute_min_max(list(Res)) :-
   findall(Info, get_info(Info,constant), CRes),
   findall(Info, get_info(Info,variable), VRes),
   (CRes=[] -> Res = ['VARIABLES'|VRes] ;
     append(['CONSTANTS'|CRes],['VARIABLES'|VRes],Res)).

pretty_print_min_max_coverage_to_file(Filename) :-
   tcltk_compute_min_max(list(Res)),
   tools:write_lines_to_file(Filename,Res).

:- use_module(b_machine_hierarchy,[abstract_constant/2,concrete_constant/2]).
get_info(Info,Type) :- var_min_max(Var,Min,Max),
   (abstract_constant(Var,_) ; concrete_constant(Var,_)
     -> Type = constant ; Type = variable),
   translate_min_max(Var,Min,Max,Info).

translate_min_max(Var,M,M,Info) :- !, translate:translate_bvalue(M,TMin),
   ajoin([Var,'=',TMin],Info).
translate_min_max(Var,pred_false,pred_true,Info) :- !, 
   ajoin([Var,':BOOL'],Info).
translate_min_max(Var,fd(X,T),fd(Y,T),Info) :- !, 
   gen_fd_set(X,Y,T,Set), translate:translate_bvalue(Set,TS),
   ajoin([Var,':',TS],Info).
translate_min_max(Var,Min,Max,Info) :-
    translate:translate_bvalue(Min,TMin),
    translate:translate_bvalue(Max,TMax),
    ajoin([Var,':',TMin,'..',TMax],Info).

gen_fd_set(X,Y,_T,R) :- X>Y,!, R=[].
gen_fd_set(X,Y,T,[fd(X,T)|R]) :- X1 is X+1,
   gen_fd_set(X1,Y,T,R).

treat_state(const_and_vars(_,S)) :- !, treat_state(S).
treat_state(concrete_constants(S)) :- !, treat_state(S).
treat_state(List) :- member(bind(VAR,VAL),List), treat_var(VAR,VAL),fail.

treat_var(Variable,Value) :- retract(var_min_max(Variable,Min,Max)),!,
    (is_smaller(Value,Min) -> NewMin = Value ; NewMin=Min),
    (is_smaller(Max,Value) -> NewMax = Value ; NewMax=Max),
    assert(var_min_max(Variable,NewMin,NewMax)).
treat_var(Variable,Value) :- 
    assert(var_min_max(Variable,Value,Value)).

:- use_module(library(avl),[avl_size/2]).
is_smaller(int(X),int(Y)) :- !, X < Y.
is_smaller(string(X),string(Y)) :- !, X @< Y.
is_smaller(pred_true,_) :- !,fail.
is_smaller(pred_false,_) :- !.
is_smaller(fd(X,T),fd(Y,T)) :- !, X < Y.
is_smaller((X1,X2),(Y1,Y2)) :- !, (is_smaller(X1,Y1) -> true ; X1=Y1, is_smaller(X2,Y2)).
is_smaller([],Y) :- !, Y \== [].
is_smaller(_,[]) :- !, fail.
is_smaller(avl_set(X),avl_set(Y)) :- avl_size(X,SX), avl_size(Y,SY),!,
   (SX<SY -> true ; SX=SY, X@<Y).
is_smaller(X,Y) :- X @< Y.
    
/* --------------------------------------------------- */

tcltk_compute_coverage(Res) :-
   compute_the_coverage(Res,_,_,false).
tcltk_compute_coverage_and_enabling(Res) :-
   compute_the_coverage(Res,_,_,true).
   
compute_the_coverage(Res,TotalNodeNr,TotalTransSum,ShowEnabledInfo) :-
   retractall(operation_hit(_,_,_)),
   reset_node_hit,
   compute_coverage(ShowEnabledInfo),
   findall(S2,  
       (operation_hit(OpS,Nr),
        %mnf(translate:translate_expression(Op,OpS)),
        string_concatenate(':',Nr,S1),
        string_concatenate(OpS,S1,S2)),Res1),
   findall(S2,  
       (query_node_hit(Prop,Nr),
        string_concatenate(':',Nr,S1),
        string_concatenate(Prop,S1,S2)),Res0),
   findall(OpName, uncovered_operation(OpName),Res2),
   total_number_of_transitions(TotalTransSum),
   (ShowEnabledInfo=false
     -> Res00 = []
     ;
	   findall(DE, def_enables(DE),EnRes1),
	   findall(DE, pos_enables(DE),EnRes2),
	   append(['DEFINITELY_ENABLED_AFTER'|EnRes1],['POSSIBLY_ENABLES'|EnRes2],Res00)
   ),
   get_op_string('UNCOVERED_',UNCOVEREDSTR),
   get_op_string('TOTAL_',TOTALSTR), get_op_string('COVERED_',COVEREDSTR),
   append([UNCOVEREDSTR|Res2],Res00,Res01),
   append([TOTALSTR,TotalTransSum,COVEREDSTR|Res1],
          Res01,OpRes),
   append(['STATES'|Res0],OpRes,Res),
   query_node_hit(total,TotalNodeNr).

:- use_module(specfile,[get_specification_description/2]).
get_op_string(PREFIX,Res) :-
   get_specification_description(operations,OP),
   string_concatenate(PREFIX,OP,Res).


def_enables(X) :- operation_hit(OpName1,_,_),
   definitely_enabled_after(OpName1,OpName2),
   string_concatenate('====>',OpName2,R1),
   string_concatenate(OpName1,R1,X).
pos_enables(X) :- operation_hit(OpName1,_,_),
   possibly_enables(OpName1,OpName2),
  % \+ definitely_enabled_after(OpName1,OpName2),
   string_concatenate('=?+=>',OpName2,R1),
   string_concatenate(OpName1,R1,X).
   
/* note: assumes that same operation name cannot appear with multiple arity */

total_number_of_transitions(TotalTransSum) :-
  findall(Nr,operation_hit(_,_,Nr),List),
  sum_list(List,0,TotalTransSum).
  
sum_list([],N,N).
sum_list([H|T],SoFar,Res) :- S1 is SoFar+H, sum_list(T,S1,Res).

operation_hit(Op,Nr) :-
   operation_hit(OpTerm,_ID,Nr),
   ((OpTerm = partition(POp,_,P)) -> mnf(translate:translate_expression(partition(POp,P),Op))
                                  ;  Op = OpTerm).
   
:- dynamic operation_hit/3.
operation_hit(operation(_),node__2,55).

:- use_module(specfile,[get_possible_event/1]).

uncovered_operation(OpName) :- specfile:get_possible_event(OpName),
   \+(operation_hit(OpName,_,_)).
   
add_hit(OpName,ID) :-
  ((OpName = partition(N,_Op,P)) -> LookupName = partition(N,_,P) ; LookupName = OpName),
  (retract(operation_hit(LookupName,NID,Nr))
   -> (N1 is Nr+1, assert(operation_hit(LookupName,NID,N1)))
   ;  (assert(operation_hit(OpName,ID,1)))
  ).

query_node_hit(Prop,Nr) :- node_hit(Prop,Nr).
query_node_hit(total,Nr) :- node_hit(open,N1), node_hit(live,N2), node_hit(deadlocked,N3),
   Nr is N1+N2+N3.
:- dynamic node_hit/2.
node_hit(property,22).
add_node_hit(OpName) :-
  (retract(node_hit(OpName,Nr))
   -> (N1 is Nr+1, assert(node_hit(OpName,N1)))
   ;  (assert(node_hit(OpName,1)))
  ).

reset_node_hit :- retractall(node_hit(_,_)),
    assert(node_hit(deadlocked,0)),
    (specfile:b_or_z_mode -> assert(node_hit(invariant_violated,0)),
                    assert(node_hit(invariant_not_checked,0))
                  ; true),
    assert(node_hit(open,0)), assert(node_hit(live,0)),
    init_enabling_info.

:- use_module(specfile,[get_operation_name/2]).

compute_coverage(_) :-
   visited_expression(ID,_),
   (not_all_transitions_added(ID)
       -> add_node_hit(open)
       ;  (transition(ID,_,_) -> add_node_hit(live)
           ; user:is_not_interesting(ID) -> add_node_hit(ignored)
           ; max_reached_for_node(ID) -> true  % e.g., if MAX_OPERATIONS set to 0
           ; otherwise -> add_node_hit(deadlocked))
    ),
    (max_reached_for_node(ID)
       -> add_node_hit(explored_but_not_all_transitions_computed)
       ; true
    ),
   (invariant_violated(ID) -> add_node_hit(invariant_violated) ; true),
   (time_out_for_invariant(ID) -> add_node_hit(invariant_time_out) ; true),
   (time_out_for_assertions(ID) -> add_node_hit(assertions_time_out) ; true),
   (invariant_not_yet_checked(ID) -> add_node_hit(invariant_not_checked) ; true),
   ((state_error(ID,_,Err),Err\=invariant_violated) -> 
      (Err=abort_error(TYPE,_,_,_)
         -> add_node_hit(TYPE)
         ;  add_node_hit(state_error))
     ; true),
   fail.
compute_coverage(ShowEnabledInfo) :-  % print(transitions),nl,
   transition(ID,Operation,NID),
   nonvar(Operation),
   get_operation_name(Operation,OpName),
   add_hit(OpName,ID),
   (ShowEnabledInfo=true -> update_enabling_info(ID,OpName,NID) ; true),
   fail.
%compute_coverage :- print_message('Covered Properties: '),
%  node_hit(OpName,Nr),
%   print(OpName), print('  nodes:'), print(Nr),nl,
%   fail.
%compute_coverage :- print_message('Covered Operations: '),
%  operation_hit(OpName,_ID,Nr),
%   print(OpName), print('  hits:'), print(Nr),nl,
%   fail.
%compute_coverage :- nl, print_message('Uncovered Operations: '),
%   uncovered_operation(OpName),
%   print(OpName), print(' '), fail.
compute_coverage(_) :- nl.
   



:- dynamic definitely_enabled_after/2, possibly_enables/2.  /* true if after an operation another is always enabled */

:- use_module(specfile,[b_mode/0]).
:- use_module(bmachine,[b_get_machine_operation/4]).
init_enabling_info :-
    retractall(definitely_enabled_after(_,_)),
    retractall(possibly_enables(_,_)),
    b_mode,
    (b_get_machine_operation(OpName1,_,_,_) ; OpName1 = '$initialise_machine'),
    b_get_machine_operation(OpName2,_,_,_),
    assert(definitely_enabled_after(OpName1,OpName2)),
    fail.
init_enabling_info.


update_enabling_info(_ID,OpName,NewID) :- %print(upd(_ID,OpName)),nl,
    definitely_enabled_after(OpName,OpName2),
    (opname_possible(NewID,OpName2)
      -> true
      ; retract(definitely_enabled_after(OpName,OpName2))  
      %(retract(definitely_enabled_after(OpName,OpName2)) -> print(retract(OpName,OpName2)))
    ),fail.
update_enabling_info(ID,OpName,NewID) :- %print(upd(_ID,OpName)),nl,
   opname_possible(NewID,OpName2),  % Note: can succeed multiple times with same OpName2 !
   \+ possibly_enables(OpName,OpName2),
   \+ opname_possible(ID,OpName2),
   assert(possibly_enables(OpName,OpName2)),
   fail.
update_enabling_info(_,_,_).


opname_possible(ID,OpName) :- transition(ID,Op,_),get_operation_name(Op,OpName).
