% (c) 2009-2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 



/* visualize_graph.pl */
:- module(dot_graph_generator,[gen_dot_graph/6,print_graph_header/1,print_graph_footer/0]).

:- use_module(module_information).
:- module_info(group,dot).
:- module_info(description,'This a few tools for generating dot graphs.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

 :- use_module(library(lists)).  
 :- use_module(preferences).    
    
:- use_module(debug).
:- use_module(self_check).

/* --------------------------------------------------- */
/*             MAIN ENTRY POINTS FOR TCL               */
/* --------------------------------------------------- */

gen_dot_graph(F,Module,NodePredicate,TransPredicate,SameRankPred,SubGraphPred) :-
   print('% Generating Dot File: '), print(F),nl,
   tell(F),
   (gen_dot_graph(Module,NodePredicate,TransPredicate,SameRankPred,SubGraphPred) -> true ; true),
   told,
   print('% Done'),nl.


node_id(Module,P,NodeID) :-
   node_predicate_call(Module,P,NodeID,_Sub,_NodeDesc,_Shape,_Style,_Color).
node_predicate_call(Module,P,NodeID,SubGraph,NodeDesc,Shape,Style,Color) :-
   %Call =.. [P,NodeID,SubGraph,NodeDesc,Shape,Style,Color],
   call(Module:P,NodeID,SubGraph,NodeDesc,Shape,Style,Color).
   
trans_predicate_call(Module,P,NodeID,Label,SuccNodeID,Color) :-
   %Call =.. [P,NodeID,Label,SuccNodeID,Color],
   call(Module:P,NodeID,Label,SuccNodeID,Color).
   
same_rank_call(Module,P,Nodes) :- P \= none,
   %Call =.. [P,Nodes],
   call(Module:P,Nodes).
   /* should succeed once for every set of NodeIDs which should be of same rank */
   
subgraph_call(Module,P,SubGraphID,Style,Color) :- P \= none,
   %Call =.. [P,SubGraphID,Style,Color],
   call(Module:P,SubGraphID,Style,Color).
   /* Notes: SubGraphID: should be none if not in a subgraph; Style and Color can be none */
   

/* ---------------------------------------------------------------------- */   
   
gen_dot_graph(Module,NodePredicate,TransPredicate,SameRankPred,SubGraphPred) :-
    print_graph_header(prob_graph),
    (node_id(Module,NodePredicate,_) -> true
       ; (format(user_error,"No nodes in gen_dot_graph: ~w, ~w, ~w ~w ~w.~n~n",
           [Module,NodePredicate,TransPredicate,SameRankPred,SubGraphPred]),fail)),
	print_nodes(Module,NodePredicate,SubGraphPred),
	nl, fail.   
gen_dot_graph(Module,_NodePredicate,_TransPredicate,SameRankPred,_SubGraphPred) :-
    same_rank_call(Module,SameRankPred,Nodes),
    print_same_ranks(Nodes),
    fail.
gen_dot_graph(Module,_NodePredicate,TransPredicate,_SameRankPred,_SubGraphPred) :-
	print_transitions(_NodeID,Module,TransPredicate),
	nl, fail.
gen_dot_graph(_,_,_,_,_) :-
    print_graph_footer.

/* ---------------------------------------------------------------------- */   


print_graph_header(Type) :-
    print('digraph '), print(Type), print(' {'),nl,
   % print('graph [orientation=landscape, page="8.5, 11",ratio=fill,size="7.5,10"];'),nl,
    print(' graph [page="8.5, 11",ratio=fill,size="7.5,10"];'),
    nl.

print_graph_footer :-  print('}'),nl.   
    
    
/* ---------------------------------------------------------------------- */   


print_nodes(Module,NodePredicate,SubGraphPred) :-
   subgraph_call(Module,SubGraphPred,SubgraphID,Style,Color),
   print('  subgraph '), print('cluster_'),print(SubgraphID), print(' {'),nl,
   (Style = none -> true ; (print('      style='), print(Style), print(';'),nl)),
   (Color = none -> true ; (print('      color='), print(Color), print(';'),nl)),
   print('      label='),print(SubgraphID), print(';'),nl,
   print_nodes2(SubgraphID,Module,NodePredicate),
   print('  }'),nl,
   fail.
print_nodes(Module,NodePredicate,_SubGraphPred) :-
   print_nodes2(none,Module,NodePredicate),
   nl.
   
   
print_nodes2(SubGraph,Module,NodePredicate) :-
   node_predicate_call(Module,NodePredicate,NodeID,SubGraph,NodeDesc,Shape,Style,Color),
   print(' '),print(NodeID),
   print(' [shape='), print(Shape),  /* options: triangle, .... */
   (Style = none
    -> true
    ;  (print(', style='), print(Style))  /* options: filled */
   ),
   (Color = none
    -> true
    ;  (print(', color="'), print(Color),print('"'))  /* options: red,green,... */
   ),
   get_preference(dot_node_font_size,FSize),
   print(', fontsize='), print(FSize),
   print(', label="'), 
   ((preference(dot_print_node_ids,true),NodeID \= NodeDesc)
     -> (print(NodeID), print(':'), print('\\'),print('n')) ; true),
   print(NodeDesc),
    print('"];'),nl,
    fail.
print_nodes2(_Subgraph,_,_) :- nl.

print_transitions(NodeID,Module,TransPredicate) :-
	trans_predicate_call(Module,TransPredicate,NodeID,Label,SuccID,Color),
	
    ((NodeID=root) -> preference(dot_print_root,true) ; true),
    (preference(dot_print_leaves,true) -> true
      ; (trans_predicate_call(Module,TransPredicate,SuccID,_,_,_)->true;fail)),

	((NodeID \= SuccID) -> true ; preference(dot_print_self_loops,true)),
	
	print(' '),print(NodeID), print(' -> '),
	print(SuccID), 
	print(' ['),
	(preference(dot_print_arc_colors,true) 
	  -> (print('color="'), print(Color),print('",'))
	  ;  true
	),
	
	print(' label="'),
	print(Label),	    
	print('"'),
	get_preference(dot_edge_font_size,FSize),
	print(', fontsize='), print(FSize),
	print('];'),
	 nl,
	fail.
print_transitions(_NodeID,_,_) :- nl.


print_same_ranks(L) :-
    print(' { rank=same; '),
    print_same_ranks2(L),
    print(' }'),nl.

print_same_ranks2([]).
print_same_ranks2([H|T]) :- 
   print(H), print('; '),
   print_same_ranks2(T).