% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(eclipse_interface,
  [
   get_version/7,
   clear_loaded_machines/0, 
   clear_loaded_machines_wo_errors/0,
  
   get_rand/4, % DEPRECATED, remove when switching to the new core
   set_rand/4, % DEPRECATED, remove when switching to the new core
   
   debug_console/2,
   
   typecheck_b_project/3,
   
   load_b_file/1, load_prob_file/1,
   load_eventb_file/1,
   load_tla_file/1,
   load_xtl_file/1,
   load_event_b_project/4,
   load_b_project/3,
   load_classical_b/1,
   load_classical_b/2, % deprecated
   
   do_modelchecking/3,
   prob2_modelcheck/5,
   find_shortest_trace_to_current_state/1,
   find_shortest_trace_to_current_state2/2,
   compute_coverage/5,

   do_ltl_modelcheck/6,

   get_invariants/1,
   get_name/1,

   evaluate/3,

   es_reset/0,  %evalstore: reset, remove all dynamic stores
   es_copy_from_statespace/2,
   es_eval/4,
   
   load_cspm_spec_from_cspm_file/1, load_cspm_spec_from_pl_file/1, 
   
   init_eclipse_preferences/0,
   set_eclipse_preference/2,
   get_eclipse_preference/2,
   is_normal_eclipse_preference/5,
   list_eclipse_preferences/1,
   list_current_eclipse_preferences/1,
   
   set_eclipse_card/2,
   
   getAllOperations/1, % deprecated, remove in 2.0
   getSetElements/1,
   getInternalRep/1,
   get_machine_objects/5,
   
   get_all_transitions/1,

   invariantKO/1, max_operations_reached/1, timeout_occurred/1, inv_timeout_occurred/1, op_timeout_occurred/2, op_timeout_occurred/1,
   state_property/3,
   
   getErrorMessages/1,
   
   start_animation/0,  update_preferences_from_spec/0,
   
   getCurrentStateID/1, getState/2,
   getVarCstValue/3, applyVarCstValue/4,
   isInitialisedState/1, isRootState/1, isConstantsSetUpState/1,
   getOperationParameterNames/2, % deprecated, remove in 2.0

   get_operation_infos/1,

   
   % getCurrentStateValues/2, no longer exists ??!
   getStateValues/2,
   get_state_errors/2,
   
   test_boolean_expression_in_node/2, 
   computeOperations/1, 
   executeOperationNr/1,
   
   setCurrentState/1,
   computeOperationsForState/2,
   
   evaluate_expression/2,
   
   print_states_for_dot/1,
   
   evaluate_expression_as_tree/4,
   evaluate_properties_as_tree/1,
   evaluate_invariant_as_tree/2,
   evaluate_precondition_as_tree/3,
   evaluate_raw_expression_as_tree/3,
   
   %in contrast to evaluate_expression, it supports predicates and does typing
   evaluate_raw_expression/3,
   evaluate_raw_expressions/3,

   evaluate_formula/3,
   evaluate_formulas/3,
   evaluate_tautology/2,
   evaluate_formula/2,
   
   sap_generate_testcases/7,
   sap_generate_local_testcases/5,
   sap_find_test_path/4,
   execute_custom_operations/6,
   debug_invariants/1,
   
   evaluation_get_top_level/1,
   evaluation_expand_formula/3,
   evaluation_get_values/3,
   evaluation_insert_formula/3,

   deadlock_freedom_check/1,
   deadlock_freedom_check/2,
   invariant_check/2,

   save_as_classical_b/4,
   serialize/2,
   deserialize/2,

   analyse_predicate/6,
   get_classicalb_machine_objects/2,

   activate_plugin/1,
   get_plugin_output/2,

   dot_hierarchy/1,

   get_statespace_edges_newer_than/2,

   get_op_from_id/2
  ]).
  
:- use_module(module_information).
:- module_info(group,cli).
:- module_info(description,'This module provides the Prolog interface to Java and other languages (usually called via socket server).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(system),[environ/2]).
:- use_module(library(lists)).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(version).
:- use_module(eventhandling,[announce_event/1]).

:- use_module(model_checker).

:- use_module(b_global_sets,[enum_global_type/2,add_prob_deferred_set_elements_to_store/2]).
:- use_module(translate).
:- use_module(store).

:- use_module(bvisual2).

:- use_module(library(codesio)).

:- use_module(state_space,[visited_expression/2]).
:- use_module(tcltk_interface,[get_id_of_node_and_add_if_required/4
                              ,tcltk_add_new_transition_transid/6
                              ,check_invariant/2]).

:- use_module(evalstores,[evalstore_reset/0,evalstore_create_store_by_state/2,
                          evalstore_eval_formula/4,evalstore_bindings/2]).

:- use_module(b_state_model_check,[cbc_deadlock_freedom_check/3]).

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).

% analyse_predicate(+Type,-Desc,-Result,-Total,-False,-Unknown)

analyse_predicate(Type,Desc,Result,Total,False,Unknown) :-
 user:tcltk_analyse_option(Type,Desc),
 user:tcltk_analyse(Type,Result,Total,False,Unknown).


serialize(Id,SerializedState) :-
    visited_expression(Id,State),
    write_term_to_codes(State,Codes,".",[quoted(true)]),
    atom_codes(SerializedState,Codes).

deserialize(NewId,SerializedState) :-
    atom_codes(SerializedState,Codes),
    read_from_codes(Codes,State),
    user:get_id_of_node_and_add_if_required(State,NewId,_Res,root).

/* a list of the calls provided by this interface; useful for the whitelist of the socket server */


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       WARNING: This is dangerous in a release. It allows to execute arbitrary code
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


check_key(Key) :-
    environ('PROB_DEBUGGING_KEY',Key).

:- dynamic current_prob_module/1.
current_prob_module(user).

debug_console(Key,Commands) :-
    check_key(Key), debug_console2(Commands).
debug_console2([]).
debug_console2([H|T]) :-
    nonvar(H),
    debug_command(H), debug_console2(T).

debug_command(set_module(X)) :-
    !, current_module(X), retractall(current_prob_module(_)), assert(current_prob_module(X)).
debug_command(Command) :-
    current_prob_module(M),
    write(M:Command),nl,
    call(M:Command).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


get_version(Major,Minor,Service,Qualifier,SvnRevision,LastChangedDate,PrologInfo) :- 
   version(Major,Minor,Service,Qualifier),
   revision(SvnRevision),  lastchangeddate(LastChangedDate), prolog_flag(version,PrologInfo).


/* ------------------------- */
/* Machine Loading Interface */
/* ------------------------- */

:- use_module(pref_definitions,[b_get_preferences_from_machine/0]).
:- use_module(specfile,[b_or_z_mode/0,set_animation_mode/1,set_animation_minor_mode/1,animation_mode/1,
                        process_algebra_mode/0,set_currently_opened_file/1,reset_currently_opened_file/0]).
:- use_module(bmachine_construction,[check_machine/5, get_constructed_machine_name/2]).
:- use_module(ltl,[ltl_model_check_with_ce/4,typecheck_temporal_formula/3,ltl_formula_structure/3]).
%:- use_module(tcltk_interface, [find_shortest_trace_to_node/5]).

:- use_module(state_space).
:- use_module(bmachine).


% DEPRECATED, only here because the old Rodin plugin calls the predicates
get_rand(0,0,0,0).
set_rand(_,_,_,_).

:- use_module(preferences,[reset_all_temporary_preferences/0]).
clear_loaded_machines :- 
   clear_loaded_machines_wo_errors,
   reset_errors, reset_all_temporary_preferences,
   garbage_collect_atoms,!. % reclaim atoms if possible
clear_loaded_machines :- add_error(clear_loaded_machines,'clear_loaded_machines Failed').

clear_loaded_machines_wo_errors :- b_machine_reset, 
   haskell_csp:clear_cspm_spec,
   b_machine_hierarchy:reset_hierarchy,
   b_interpreter:reset_b_interpreter,
   state_space_initialise,
   reset_model_checker,reset_refinement_checker,reduce_graph_reset.

% Errors is a list of error(ErrorDescription,Position)
% Position is the number of the Syntaxtree-Node 
typecheck_b_project(Main,Machines,Errors) :-
    check_machine(Main,Machines,[],_,Errors).

load_b_project(Main,Machines,Errors) :-
    b_set_machine(Main,Machines,Errors).

load_classical_b(Machines) :- 
     set_animation_mode(b),
     load_classical_b(Machines,Errors),
     add_errors(Errors).

load_classical_b(Machines,Errors) :- 
    [First|_] = Machines,
    get_constructed_machine_name(First,Main),
    b_set_machine(Main,Machines,Errors).

add_errors([]).
add_errors([H|T]) :- error_manager:add_error('eclipse_interface','Error loading machine',H),add_errors(T).

load_event_b_project(Machines,Contexts,Proofs,Errors) :-
    set_animation_mode(b), set_animation_minor_mode(eventb),
    b_set_eventb_project_flat(Machines,Contexts,Proofs),
    load_additional_information(Proofs),
    Errors = [].

:- use_module(parsercall,[call_tla2b_parser/1, tla2b_filename/2]).
load_tla_file(Filename) :-
   call_tla2b_parser(Filename),
   tla2b_filename(Filename,GeneratedBFile),
   load_b_file(GeneratedBFile),
   set_animation_minor_mode(tla).
load_b_file(Filename) :-
   set_animation_mode(b),
   reset_currently_opened_file,
   bmachine:b_load_machine(Filename),
   set_currently_opened_file(Filename).
load_prob_file(Filename) :-
   set_animation_mode(b),
   reset_currently_opened_file,
   bmachine:b_load_machine_probfile(Filename),
   set_currently_opened_file(Filename).

save_as_classical_b(File,Machines,Contexts,Errors) :-
    /* first set the number of animated levels to 0*/
    get_preference(number_animated_abstractions,OldNum),
    call_cleanup( ( set_preference(number_animated_abstractions,0),
                    load_event_b_project(Machines,Contexts,[],Errors),!),
                  set_preference(number_animated_abstractions,OldNum)),
    ( Errors==[] ->
        b_show_eventb_as_classicalb(Rep,true),
        tools:write_to_file(File,Rep)
    ; otherwise ->
        /* Do nothing, but return the list of errors*/
        true).

 
load_eventb_file(Filename) :-
   set_animation_mode(b), set_animation_minor_mode(eventb),
   reset_currently_opened_file,
   b_load_eventb_project(Filename),
   set_currently_opened_file(Filename).

:- use_module(xtl_interface,[open_xtl_file/1]).
load_xtl_file(Filename) :- 
   set_animation_mode(xtl),
   reset_currently_opened_file,
   xtl_interface:open_xtl_file(Filename),
   set_currently_opened_file(Filename).
  
get_all_transitions(Transitions) :-
    findall(trans(TransId,Src,Event,PP,Dst,Infos),
            ( transition(Src,Event,TransId,Dst),
                translate_event(Event,PP),
                get_event_infos(TransId,Infos)),
            Transitions).

get_event_infos(TransId,Infos) :-
    findall(Info, transition_info(TransId,Info), Infos1),
    filter_transition_infos(Infos1,Infos).

filter_transition_infos([],[]).
filter_transition_infos([I|Irest],[O|Orest]) :-
    filter_transition_info(I,O),
    filter_transition_infos(Irest,Orest).

filter_transition_info(event(StackI),event(StackO)) :- !,
    filter_event_stack(StackI,StackO).
filter_transition_info(I,I).

filter_event_stack([],[]).
filter_event_stack([event(Name,Params)|Irest],[event(Name,Strings)|Orest]) :-
    translate_bvalue_l(Params,Strings),
    filter_event_stack(Irest,Orest).

translate_bvalue_l([],[]).
translate_bvalue_l([V|Vrest],[S|Srest]) :-
    translate_bvalue(V,S),
    translate_bvalue_l(Vrest,Srest).
    

print_states_for_dot(X) :- visualize_graph:tcltk_print_states_for_dot(X).


% When Time Milliseconds have elapsed the modelchecker should stop after its next step 
do_modelchecking(Time, Options, Result) :- 
     statistics(runtime,[CurTime,_]), /* get current time in ms */
     LimitTime is CurTime+Time,
     option_set(find_deadlocks,Options,Deadlock),
     option_set(find_invariant_violations,Options,Invariant),
     option_set(find_assertion_violations,Options,Assertions),
     option_set(inspect_existing_nodes,Options,InspectExistingNodes),
     Goal = 0,
     StopAtFullCoverage=0,
     WithPOR = 0, % no partial order reduction
     MaxNumberOfStatesToCheck=100000,
     user:do_model_check(MaxNumberOfStatesToCheck,_,LimitTime, Res, Deadlock, Invariant, Goal,Assertions,
                        StopAtFullCoverage, WithPOR, InspectExistingNodes),
     build_modelcheck_return(Res,Result).
   
prob2_modelcheck(Time,Options,Last,Ops,Result) :-
     do_modelchecking(Time,Options,Result),
     get_statespace_edges_newer_than(Last,Ops).

 :- use_module(coverage_statistics).
compute_coverage(TotalNodeNr,TotalTransSum,NodeStat,OpStat,Uncovered)  :- 
   coverage_statistics:compute_the_coverage(_,TotalNodeNr,TotalTransSum,false),
   findall(S2,(coverage_statistics:operation_hit(OpS,Nr),tools:string_concatenate(':',Nr,S1),tools:string_concatenate(OpS,S1,S2)),OpStat),
   findall(S2,(coverage_statistics:query_node_hit(Prop,Nr),tools:string_concatenate(':',Nr,S1),tools:string_concatenate(Prop,S1,S2)),NodeStat),
   findall(OpName, coverage_statistics:uncovered_operation(OpName),Uncovered).
      
   
do_ltl_modelcheck(Formula,Max,Mode,Atomics,Structure,Result) :-
   typecheck_temporal_formula(Formula,TypeCheckedFormula,Status),
   ( Status=ok ->
       ltl_formula_structure(TypeCheckedFormula,Atomics,Structure),
       ltl_model_check_with_ce(TypeCheckedFormula,Max,Mode,Result1),
       ltl_adapt_operations(Result1,Result)
   ; otherwise ->
       Atomics=[], Structure=[],
       Result=typeerror([])).

ltl_adapt_operations(counterexample(CE1,LoopEntry,PathToCE1), counterexample(CE,LoopEntry,PathToCE)) :- !,
    filter_ops_and_ids(PathToCE1,root,PathToCE),
    ltl_adapt_ce(CE1,CE).
ltl_adapt_operations(Result,Result).

ltl_adapt_ce([],[]).
ltl_adapt_ce([atom(State,Evals,OpTuple)|Irest],[atom(State,Evals,Transition)|Orest]) :-
    ltl_adapt_ce2(OpTuple,State,Transition),
    ltl_adapt_ce(Irest,Orest).
ltl_adapt_ce2(none,_State,none).
ltl_adapt_ce2((TransId,Action,DstId),State,Transition) :-
    create_operation_term(TransId,Action,State,DstId,Transition).

find_shortest_trace_to_current_state(TraceStrings) :-
 find_shortest_trace_to_current_state2(TraceStrings, _).

 
find_shortest_trace_to_current_state2(TraceStrings, TraceIDs) :-
 getCurrentStateID(ID),
 user:find_shortest_trace_to_node(root,ID,Trace,TraceIDs),
 translate_events(Trace,TraceStrings).
 
 

option_set(Element, List, Result) :-
  ( member(Element,List)
    -> Result = 1
    ;  Result = 0).


build_modelcheck_return(MCRes,EclipseResult) :-
  build_modelcheck_return2(MCRes,EclipseResult) -> true
  ; nl,print('### Unknown Model Check Error Result:'), print(MCRes),nl,
    getCurrentStateID(State),EclipseResult=general_error(State).

build_modelcheck_return2(no,not_yet_finished(100000)). 
build_modelcheck_return2([timeout,N],not_yet_finished(N1)) :- N1 is 100000 - N. 
build_modelcheck_return2(deadlock,deadlock(State)):- getCurrentStateID(State). 
build_modelcheck_return2(invariant_violation,invariant_violation(State)):- getCurrentStateID(State). 
build_modelcheck_return2(assertion_violation,assertion_violation(State)):- getCurrentStateID(State). 
build_modelcheck_return2(state_error,state_error(State)):- getCurrentStateID(State). 
build_modelcheck_return2(state_error(_),state_error(State)):- getCurrentStateID(State). 
build_modelcheck_return2(well_definedness_error,well_definedness_error(State)) :- getCurrentStateID(State).
build_modelcheck_return2(general_error_occurred,general_error(State)):- getCurrentStateID(State). 
build_modelcheck_return2(all,ok_not_all_nodes_considered) :- max_reached_for_node(_),!. 
build_modelcheck_return2(all,ok).




/* ------------------------- */
/* CSP Loading Interface */
/* ------------------------- */

 :- use_module(probcspsrc(haskell_csp),[parse_and_load_cspm_file/1,load_cspm_pl_file/1]).

load_cspm_spec_from_cspm_file(CSPMFile) :-
   reset_currently_opened_file, specfile:set_animation_mode(cspm),
   parse_and_load_cspm_file(CSPMFile),
   set_currently_opened_file(CSPMFile).
load_cspm_spec_from_pl_file(CSPMFile) :-
   reset_currently_opened_file, specfile:set_animation_mode(cspm),
   load_cspm_pl_file(CSPMFile),
   set_currently_opened_file(CSPMFile).

/* --------------------- */
/* Preferences Interface */
/* --------------------- */

:- use_module(preferences,
  [init_preferences/0,set_preference/2,get_preference/2,
   eclipse_preference/2,
   preference_val_type/2,preference_description/2,preference_default_value/2,
   preference_category/2]).


% will only return normal preferences, not advanced ones
list_eclipse_preferences(L) :- 
   findall(preference(A,B,C,D,E),is_normal_eclipse_preference(A,B,C,D,E),L).

:- use_module(model_checker,[reset_depth_breadth_first_mode/0]).
init_eclipse_preferences :- preferences:init_preferences, model_checker:reset_depth_breadth_first_mode.

% will only return normal preferences, not advanced ones
is_normal_eclipse_preference(PrefString,Type,Description,Category,DefaultValue) :-
   preferences:eclipse_preference(PrefString,PS),
   \+ preferences:advanced_eclipse_preference(PrefString,PS),
   preference_val_type(PS,Type),
   preference_description(PS,Description),
   preference_category(PS,Category),
   preference_default_value(PS,DefaultValue).

set_eclipse_preference(PrefS,PrefVal) :- convert_cli_arg(PrefVal,Value),
   convert_pref_value(Value,CValue),
   (preferences:eclipse_preference(PrefS,P) -> true
    ; preferences:obsolete_eclipse_preference(PrefS)
      -> add_error_fail(eclipse_preference,'Obsolete Eclipse Preference: ',PrefS)
    ; add_error_fail(eclipse_preference,'Unknown Eclipse Preference: ',PrefS)
   ),
   (preferences:set_preference(P,CValue) -> true ; add_error(eclipse_preference,'Could not set preference: ',Value)).

convert_cli_arg(PrefVal,Value) :-
   (atomic(PrefVal)
     ->  (name(PrefVal,ASCIIPREF), name(Value,ASCIIPREF))  /* convert '12' to 12 */
     ;   Value=PrefVal
   ).
  
convert_pref_value('TRUE',V) :- !, V=true.
convert_pref_value('FALSE',V) :- !, V=false.
convert_pref_value(X,X).

% set Cardinality of SET; a bit like scope_SET == Scope DEFINITION
:- use_module(b_global_sets,[set_user_defined_scope/2]).
set_eclipse_card(Set,Scope) :- convert_cli_arg(Scope,Value),
   set_user_defined_scope(Set,Value).
   
get_eclipse_preference(PrefS,PrefVal) :-
   preferences:eclipse_preference(PrefS,PS),
   get_preference(PS,PrefVal).

list_current_eclipse_preferences(L) :-
   findall(preference(A,B),find_eclipse_preference(A,B),L).

find_eclipse_preference(A,B) :-
  list_eclipse_preferences(X),
  member(preference(A,_,_,_,_),X),
  get_eclipse_preference(A,B).


/* ------------------ */
/* getErrorMessages/1 */
/* ------------------ */
/* Returns a list of strings containing all error messages (of the error_manager).
   Clears all the errors. Returns the empty list if no errors occurred since
   the last call to getErrorMessages. */
   
%:- use_module(error_manager,[get_all_errors_and_reset/1]).
getErrorMessages(ListOfErrMsgs) :-
  (get_all_errors_and_reset(ListOfErrMsgs) ->  true  ; ListOfErrMsgs=[]).
  
  
  
/* ---------------------------- */
/* Syntax Information Interface */
/* ---------------------------- */

getAllOperations(L) :-
   (b_or_z_mode
     -> findall(OperationName, b_top_level_operation(OperationName), L)
     ;  L=[]).  % FIXME: provide list for CSP-M and Promela
   

get_operation_infos(L) :- 
    findall(op(Name,Results,Parameters),b_get_machine_operation(Name,Results,Parameters,_,_),L).


  

getSetElements(L) :-  /* get a list of all fd(N,S) elements of all sets and precompute their translation */
   findall(fds(Nr,Set,String),  
           (enum_global_type(fd(Nr,Set),_), translate_bvalue(fd(Nr,Set),String)), L).

           
:- use_module(specfile,[get_internal_representation/1]).
getInternalRep(X) :- get_internal_representation(X).           

get_classicalb_machine_objects(Constants, Variables) :- 
    b_get_machine_constants(C),translate_list(C,Constants),
    b_get_machine_variables(V),translate_list(V,Variables).


translate_list([],[]).
translate_list([H|T],[HT|TT]) :- translate_bexpression(H,HT), translate_list(T,TT).




get_machine_objects(Sections,SetElements,Constants,Variables,Operations) :-
    b_get_animated_sections(Sections),
    findall(TId,
            ( enum_global_type(fd(Nr,Set),_),
                translate_bvalue(fd(Nr,Set),String),
                create_texpr(identifier(String),global(Set),[],TId)),
            SetElements),
    b_get_machine_constants(Constants),
    b_get_machine_variables(Variables),
    b_get_promoted_machine_operations(Operations1),
    maplist(eclipse_interface:remove_op,Operations1,Operations).
remove_op(TId,NTId) :-
    remove_bt(TId,identifier(Id),identifier(NId),NTId),
    remove_op2(Id,NId).
remove_op2(op(Op),Op) :- !.
remove_op2(Op,Op).
   
/* ------------------- */
/* Animation Interface */
/* ------------------- */


get_invariants(List) :- %print(start),nl,
   b_get_invariant_from_machine(B), 
%   print(got(B)),nl,
   conjunction_to_list(B,L), %print(L),nl,
   maplist(mk_inv,L,List).

mk_inv(C,inv(Text,Pos)) :-  translate_bexpression(C,Text),get_texpr_pos(C,P), (P=pos(Pos,_,_,_,_,_) -> true;Pos = P),!.
mk_inv(C,_) :- add_internal_error(mk_inv,'Failed: ',C),fail.

get_name(Name) :- b_machine_name(Name).



:- use_module(self_check,[turn_off_run_time_type_checks/0]).
start_animation :- %turn_off_run_time_type_checks,
                   initialise_animator,!.
start_animation :- add_error(start_animation,'Start Animation Failed').

% update preferences from SET_PREF Definitions in B machines:
update_preferences_from_spec :- (b_or_z_mode -> b_get_preferences_from_machine ; true).

:- use_module(flow,[reset_flow/0]).
:- use_module(model_checker,[reset_model_checker/0]).
:- use_module(refinement_checker,[reset_refinement_checker/0]).
:- use_module(reduce_graph_state_space,[reduce_graph_reset/0]).
initialise_animator :-
   state_space_initialise,
   reset_errors, reset_flow, reset_model_checker,
   reset_refinement_checker,
   reduce_graph_reset,
   initialise_specification. 

:- use_module(symmetry_marker,[precompile_marker_typing_info/0]).
:- use_module(external_functions,[reset_external_functions/0]).
:- use_module(value_persistance,[initialise_operation_caching/0]).
initialise_specification :-  /* call once before starting animation or model checking */
   ( b_or_z_mode -> b_machine_precompile
   ; otherwise -> 
       preferences:get_preference(symmetry_mode,X),
       ( X=off -> true
       ; otherwise ->
           add_message(initialise_specification,'Symmetry can only be used for B & Z specifications'),
           preferences:set_preference(symmetry_mode,off) /* Turn symmetry off */
       ),
       ~~mnf(other_spec_precompile)
   ),
      %TODO(DP,11.8.2008)
   ~~mnf(graph_canon:reset_typing_info_for_symm),
   user:initialise_operation_not_yet_covered,
   reset_external_functions,
   preferences:preference(symmetry_mode,SymMode),
   symmetry_marker:precompile_marker_typing_info,
   ( SymMode=nauty ->
       (graph_canon:initialise_nauty -> true ; preferences:set_preference(symmetry_mode,off))
   ; otherwise -> graph_canon:clear_nauty),
   announce_event(specification_initialised),
   ( initialise_operation_caching -> true
   ; otherwise -> add_error(b_load_machine,'Initialising operation caching failed'),fail).


   /* get the ID of the current state */
getCurrentStateID(CurID) :-
    current_state_id(CurID).
   /* get the ID of the current state */

:- use_module(specfile,[expand_const_and_vars_to_full_store/2]).
getState(es(ID),Store) :-           % use a store from the evalstore module instead of
    !,evalstore_bindings(ID,Store). % a state from the state space
getState(ID,BState) :-
    visited_expression(ID,State),
    get_state2(State,BState).
get_state2(csp_and_b(CSPState,BState),[bind('CSP_CONTROLLER',CSPState)|ExpandedBState]) :- !,
  get_state2(BState,ExpandedBState).
get_state2(concrete_constants(BState),BState) :- !.
get_state2(root,BState) :- !, BState=[].
get_state2(State,Res) :- expand_const_and_vars_to_full_store(State,BState),
  (BState=[_|_] -> Res=BState ; Res=[] /* we have [] or a CSP process or similar */).


isRootState(ID) :- ID=root.

:- use_module(specfile,[state_corresponds_to_set_up_constants/2, state_corresponds_to_initialised_b_machine/1,
                        state_corresponds_to_initialised_b_machine/2]).

isConstantsSetUpState(ID) :-
    % true if no constants exists or if all constants have been valued
    visited_expression(ID,State),
    specfile:state_corresponds_to_set_up_constants(State,_).
   
isInitialisedState(ID) :-
    visited_expression(ID,State), ID \= root,
    specfile:state_corresponds_to_initialised_b_machine(State).

   
   /* compute enabled operations, if required, and return a list of
      strings of all possible operations and arguments */
computeOperations(Operations) :-
   user:tcltk_get_options(list(Operations)). 
 
/* ------------------ */
/*    invariantKO/1   */
/* ------------------ */

:- use_module(b_interpreter,[b_compute_expression/4]).
invariantKO(StateID) :-  isInitialisedState(StateID),
      getState(StateID,State),
      user:check_invariantKO(StateID,State).

:- use_module(state_space,[time_out_for_node/2,time_out_for_invariant/1,max_reached_for_node/1]).

op_timeout_occurred(OpNameList) :-
   getCurrentStateID(CurID), op_timeout_occurred(CurID,OpNameList).
op_timeout_occurred(StateID,OpNameList) :-
    findall(OpName,time_out_for_node(StateID,OpName),OpNameList).
inv_timeout_occurred(StateID) :- time_out_for_invariant(StateID).
timeout_occurred(StateID) :-  time_out_for_node(StateID).
max_operations_reached(StateID) :- 
   max_reached_for_node(StateID).

state_property(Property,StateId,Status) :-
  ( state_property2(Property,StateId) -> Status = true
  ; otherwise -> Status = false).
state_property2(invariantKO,StateId) :- invariantKO(StateId).
state_property2(timeout_occurred,StateId) :- inv_timeout_occurred(StateId).
state_property2(max_operations_reached,StateId) :- max_operations_reached(StateId).
state_property2(initialised,StateId) :- isInitialisedState(StateId).

executeOperationNr(Nr) :-
   user:compute_all_transitions_if_necessary, /* make sure options are computed; TO DO: improve */
   user:tcltk_perform(Nr).
   
build_arguments([],_,_,[],[]).   
build_arguments([H|T],Name,State,[R|RT],[E|ET]) :- 
    b_type_expression(H,[variables,identifier(Name)],subst,RTyped,E),
    emptystate(Empty),
    b_compute_expression(RTyped,Empty,State,R),
    build_arguments(T,Name,State,RT,ET).
   
   
   
setCurrentState(ID) :- /* jumps to the given node; can be backtracked */
   visited_expression(ID,_CurState),
   user:tcltk_goto_state(jump,jump,ID).


test_boolean_expression_in_node(ResID,BoolExpr) :- 
   visited_expression(ResID,CurState),
   specfile:state_corresponds_to_initialised_b_machine(CurState,CurBState),
   ~~pp_cll(b_interpreter:b_test_boolean_expression_wf(BoolExpr,[],CurBState)).
 
computeOperationsForState(StateID, OpsAndIDs) :- /* compute the enabled operations
   (without the backtrack options) for a state ID which is not necessarily the current one */
   visited_expression(StateID,_CurState),
   user:tcltk_compute_options(StateID,OpsSTAndIDs),
   print('OpsSTAndIDs: '),print(OpsSTAndIDs),nl,
   (filter_ops_and_ids(OpsSTAndIDs,StateID,OpsAndIDs) -> true
     ; print(filter_ops_and_ids_failed(OpsSTAndIDs)),nl,fail).

/*
op(140,io,23,23,[[in(r14_1m)],'SetRoute',multi_span(460,35,460,71,16934,36,src_span(460,31,460,32,16930,1))],['{<< in(r14_1m) >>}','<< SetRoute >>','<< multi_span(460,35,460,71,16934,36,src_span(460,31,460,32,16930,1)) >>'],[],')
*/

:- use_module(bmachine,[b_get_machine_operation_parameter_names/2]).
getOperationParameterNames(Operation,Paras) :-
   b_get_machine_operation_parameter_names(Operation,Paras).

filter_ops_and_ids(OpsSTAndIDs,StateID,OpsAndIDs) :- 
   animation_mode(MODE),
   ( MODE == b ->
       filter_ops_and_ids_b(OpsSTAndIDs,StateID,OpsAndIDs)
   ; otherwise->
       filter_ops_and_ids_csp(OpsSTAndIDs,StateID,OpsAndIDs)
   ).

filter_ops_and_ids_b([],_,[]).
filter_ops_and_ids_b([OpTuple|T],StateID,[FOp|FT]) :-
   ((OpTuple = (Id,OpTerm,Dst), create_operation_term(Id,OpTerm,StateID,Dst,FOp))
     -> true
     ; print(filter_op_failed(OpTuple,StateID)),nl, FOp = ('?')),
   filter_ops_and_ids_b(T,StateID,FT).

filter_ops_and_ids_csp([],_,[]).
filter_ops_and_ids_csp([OpTuple|T],StateID,[FOp|FT]) :-
   ((OpTuple = (Id,OpTerm,Dst), create_operation_term(Id,OpTerm,StateID,Dst,FOp))
     -> true
     ; print(filter_op_failed(OpTuple,StateID)),nl, FOp = ('?')),
   filter_ops_and_ids_csp(T,StateID,FT).

filter_op(TransId,OpTerm,Src,Dst,JavaOpTerm) :- 
   (animation_mode(cspm) ->
       filter_op_csp(TransId,OpTerm,Src,Dst,JavaOpTerm)
   ; otherwise ->
       filter_op1(TransId,OpTerm,Src,Dst,JavaOpTerm)).

filter_op1(TransId,OpTerm,Src,Dst,
           op(TransId,Op,Src,Dst,Arguments,ArgumentsPretty,Infos,State)) :-
   extract_args_and_return_values(OpTerm,Op,Arguments),
   getState(Dst,BState), translate:translate_bstate(BState,State),
   prettyprint(Arguments,ArgumentsPretty),
   get_event_infos(TransId,Infos).

filter_op_csp(TransId,OpTerm,Src,Dst,op(TransId,Op,Src,Dst,Arguments,ArgumentsPretty,Infos,CSPState)) :- 
   extract_csp_args_and_return_values(OpTerm,Op,Arguments),
   getState(Dst,CSPState), 
   prettyprint_csp(Arguments,ArgumentsPretty),
   get_event_infos(TransId,Infos).

extract_csp_args_and_return_values(io(Args,ChName,_SPAN),ChName,Args).
extract_csp_args_and_return_values(start_cspm(Name),Name,[]).
extract_csp_args_and_return_values(start_cspm_MAIN,'start_cspm_MAIN',[]).
extract_csp_args_and_return_values(tick(_Span),'tick',[]).
extract_csp_args_and_return_values(tau(_Span),'tau',[]).
extract_csp_args_and_return_values(_OP,'?',[]).

prettyprint_csp([],[]).
prettyprint_csp([H|T],[P1|R]) :- 
   translate_cspm_expression(H,P),
   tools: ajoin(P,P1),
   prettyprint_csp(T,R).

prettyprint([],[]).
prettyprint([H|T],[P|R]) :- translate_bvalue(H,P), prettyprint(T,R).

:- dynamic result_warning_emitted/0.

extract_args_and_return_values(Term,OpName,Arguments) :-
    ( Term='-->'(OpTerm,Results) ->
        ( result_warning_emitted -> true
        ; otherwise ->
            print('Warning: results of operations are currently treated like arguments'),nl,
            assert(result_warning_emitted)),
        append(Results,Args,Arguments)
    ; otherwise -> Term=OpTerm,Args=Arguments
    ),
    OpTerm =.. [OpName|Args].



% returns all variables and their values of the given state as a list
% - constants and variables that are defined for the machine but
%   not present in the given state get the value '?'
% - for each variable a term of the form
%     binding(Id,Value,PPValue,Tag)
%   is in the result list, where
%     Id is the variable's name,
%     Value is its value a Prolog term
%     PPValue is its pretty-printed value
% NOTE: Tag no longer generated: TO DO REVIEW:
%     Tag is one of (cst,var,unknown), depending on whether the
%       variable is a constant, a variable or not defined in the
%       machine description but present in the state
getStateValues(CurID,VariableState) :- /* get variable values */
  getState(CurID,CurBState),
  findall( binding(Id,Value,PPValue),
           (member(bind(Id,Value),CurBState),
               ( translate_bvalue(Value,PPValue) -> true
               ; otherwise                       -> PPValue='**pretty-print failed**')),
           VariableState).

getVarCstValue(ID,VarOrCst,TVals) :-
  /* a function to get the value of a particular Variable or Constant;
     treats relations specially: a list of relation pairs is returned */
  getState(ID,CurBState),
  member(bind(VarOrCst,Val),CurBState),
  (Val = [ (_,_) | _]  /* check if it is a relation */
    ->  findall(TFact,relation_fact(_F,VarOrCst,Val,TFact),TVals)
    ;   (TVals = [TVal], Val = TVal)
   ).
   
 relation_fact(_F,_VarOrCst,Val,TFact) :- 
   member((L,R), Val), Fact = (L,R),
   Fact=TFact.
   
applyVarCstValue(ID,VarOrCstFunction,DomValue,RanResult) :-
  /* a function to apply a particular Variable or Constant Function/Relation to a value;
     treats relations specially: a list of relation pairs is returned */
    getState(ID,CurBState),
    member(bind(VarOrCstFunction,Val),CurBState),
    member((DomValue,R),Val),
    R= RanResult.

:- use_module(tools,[safe_atom_codes/2]).
get_state_errors(State,Errors) :-
  findall(E, (state_error(State,_,E),E \== invariant_violated), Errs),
  convert_errors(Errs,Errors).
convert_errors([],[]).
convert_errors([Error|InRest],[error(Event,Short,Long)|OutRest]) :-
  ( Error = eventerror(Event,EError,Trace) ->
      translate_event_error(EError,Short),
      explain_event_trace(Trace,LongStr),
      safe_atom_codes(Long,LongStr)
  ; otherwise ->
      Event = '*unknown*',
      translate_state_error(Error,Short),
      explain_state_error(Error,_Span,LongStr),
      safe_atom_codes(Long,LongStr)),
  convert_errors(InRest,OutRest).


/* ------------------------------- */

/* evaluates an expression in Prolog format and computes the result in the current state */
evaluate_expression(Expression,Result) :-
  current_expression(_CurID,CurState),
  b_interpreter:b_compute_expression(Expression,[],CurState,Result).
  
  
  
%%
% Code for invariant debug
%

:-use_module(bvisual).

evaluate_properties_as_tree(Tree):-
    bvisual:get_properties(TExpr),
    evaluate_expression_as_tree(root, [], TExpr, Tree).

evaluate_invariant_as_tree(StateID, Tree):-
    bvisual:get_invariant(TExpr),
    evaluate_expression_as_tree(StateID, [], TExpr, Tree).

evaluate_precondition_as_tree(StateID, OpName, Tree):-
    bvisual:get_operation_pre(OpName, TExpr, _),
    evaluate_expression_as_tree(StateID, [], TExpr, Tree).


evaluate_raw_expression_as_tree(StateID, Raw, Tree):-
    bmachine:b_type_expression(Raw, [variables], _, TExpr, Errors),
    ( var(TExpr) ->
        print(error_while_typing_raw_expression(Errors)),nl,
        flush_output,
        !,
        fail
        ;
        evaluate_expression_as_tree(StateID, [], TExpr, Tree)
    ).



evaluate_expression_as_tree1(StateID, LocalState, TExpr, Tree) :-
    getState(StateID, CurState),
    get_tree_from_expr(T0, TExpr, LocalState, CurState),
    get_pretty_print_tree(T0, Tree).

evaluate_expression_as_tree(StateID, LocalState, Expr, Tree) :-
    getState(StateID, State),
    get_tree_from_expr2(State, LocalState, Expr, T),
    get_tree_node_info(T, Children, NewLocalState, Metas),
    translate_bexpression(Expr, Pretty),
    create_tree_node(Tree, Children, NewLocalState, [pretty_long_desc(Pretty) | Metas]).

evaluate_formula(StateId,Element,R) :- 
    user_interruptable_call_det(evaluate_formula2(StateId,Element,RI), InterruptResult),!, 
    (InterruptResult = interrupted -> R = interrupted; R=RI).

evaluate_csp_expression(bindval(_Ident,ParsedExpression,_Span),R) :-
    haskell_csp: evaluate_expression(ParsedExpression,Res),
    translate: translate_cspm_expression(Res, PrettyPrintedResult),
    tools: ajoin(PrettyPrintedResult,R).
evaluate_csp_expression(Other,_R) :- 
    add_error(eclipse_interface,'Unexpected Csp Declaration: ', Other),fail.

evaluate_formula2(_StateId, eval(Raw,csp,Code), result(R,'',Code)):- !,
    evaluate_csp_expression(Raw,R).

evaluate_formula2(StateId, eval(Raw,PredOrExpr,Code), R):- 
    evaluate_formula_typecheck(PredOrExpr,Raw,Typed,Type,Errors),
    ( Errors=[] ->
        getState(StateId,State1),
        add_prob_deferred_set_elements_to_store(State1,State),
        evaluate_formula_eval(Type,State,Typed,Res,Solution,_EnumWarning),
        R=result(Res,Solution,Code)
    ; otherwise ->
        R = [Code|Errors]).

evaluate_tautology(Element,R) :- 
   user_interruptable_call_det(evaluate_tautology2(Element,RI), InterruptResult),!, 
     (InterruptResult = interrupted -> R = interrupted; R=RI).

evaluate_tautology2(Codes, result(R,Solution,Quantor,Vars,Warn)):- 
     eval_strings:eval_codes(Codes,forall,R,Warn,LocalState,TypeInfo),extract_typeinfo(TypeInfo,Quantor,Vars),pp_solution(LocalState,Solution).

evaluate_formula(Element,R) :- 
   user_interruptable_call_det(evaluate_formula2(Element,RI), InterruptResult),!, 
     (InterruptResult = interrupted -> R = interrupted; R=RI).

evaluate_formula2(Codes, result(R,Solution,Quantor,Vars,Warn)):- 
     eval_strings:eval_codes(Codes,exists,R,Warn,LocalState,TypeInfo),extract_typeinfo(TypeInfo,Quantor,Vars),pp_solution(LocalState,Solution).


extract_typeinfo(expression(_),expression,[]).
extract_typeinfo(forall(BL),forall,L) :- translate_list(BL,L).
extract_typeinfo(exists(BL),exists,L) :- translate_list(BL,L).
extract_typeinfo(predicate,predicate,[]).

pp_solution(LocalState, Solution) :-     
   ( LocalState=[] -> Solution=''
    ; otherwise    -> translate:translate_bstate(LocalState,Solution)).

evaluate_formula_typecheck('#PREDICATE',Raw,Typed,Type,Errors) :-
    !,bmachine:type_open_pred(open(exists),Raw,[prob,variables],Typed,Errors),
    Type=pred.
evaluate_formula_typecheck('#EXPRESSION',Raw,Typed,Type,Errors) :-
    !,bmachine:b_type_expression(Raw,[prob,variables],Type,Typed,Errors).

evaluate_formula_typecheck(Else,_Raw,_Typed,_Type,_Errors) :-
    add_error(eclipse_interface,'expected #PREDICATE or #EXPRESSION', Else),fail.

evaluate_formula_eval(pred,State,Typed,Res,Solution,EnumWarning) :-
    !,eval_strings:eval_predicate(State,Typed,Res,EnumWarning,LocalState),
    ( LocalState=[] -> Solution=''
    ; otherwise     -> translate:translate_bstate(LocalState,Solution)).

evaluate_formula_eval(_Type,State,Typed,Res,'',EnumWarning) :-
    eval_strings:eval_expression(State,Typed,Res,EnumWarning).

evaluate_formulas(StateID, Raws, Vals):-
    maplist(evaluate_formula(StateID), Raws, Vals).
    

evaluate_raw_expression(StateID, Raw, Val):-
    bmachine:b_type_expression(Raw, [variables], _, TExpr, Errors),
    ( var(TExpr) ->
        print(error_while_typing_raw_expression(Errors)),nl,
        flush_output,
        !,
        fail
        ;
        getState(StateID, State),
        bvisual:x_get_value_of_expr(State, [], TExpr, Val)
    ).

evaluate_raw_expressions(StateID, Raws, Vals):-
    maplist(evaluate_raw_expression(StateID), Raws, Vals).   

/* ------------------------- */
/* Execute Operation         */
/* ------------------------- */

% MaxNrOfSolutions is the maximum number of solutions ProB returns. Also we should consider existing solutions if the Predicate holds. 
%:- use_module(succeed_max,[succeed_max_call/2]).

execute_custom_operations(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, Errors) :- 
    setCurrentState(CurID),
    
    ( (OpName = '$initialise_machine';OpName =  '$setup_constants' )  -> Scope = [variables];
      otherwise ->    b_top_level_operation(OpName),
                      b_get_machine_operation(OpName,_,V,_),
                       Scope = [identifier(V),variables]
    ), 
    (MaxNrOfSolutions<1
     -> add_message(eclipse_interface,'execute_custom_operations MaxNrOfSolutions < 1: ',MaxNrOfSolutions)
      ; true),
    bmachine:b_type_expression(ParsedPredicate, Scope, _, TypedPred, Errors),
    retractall(bmachine:b_machine_temp_predicate(_)),
    assert(bmachine:b_machine_temp_predicate(TypedPred)),
    %prob_socketserver:writeln_log(parsed_pred(ParsedPredicate)), %%
    %prob_socketserver:writeln_log(succeed_max_call(execute_custom_operation(CurID,OpName,TO),MaxNrOfSolutions)), %%
    findall(TO, execute_custom_operation(CurID,OpName,TO,MaxNrOfSolutions), TOperations),
    %length(TOperations,Len), prob_socketserver:writeln_log(findall_length(Len)),
    retractall(bmachine:b_machine_temp_predicate(_)),
    clear_context_state.

% TO DO: check if we need to recompute the operation effect: if the ParsedPredicate is TRUE 
%  (TRUE = equal(none,integer(none,1),integer(none,1)) ) & MaxNrOfSolutions <=
%        what has already been used previously, we can simply reuse transition from the state space

execute_custom_operation(CurID,OpName,Operation_op,Max) :-
    set_context_state(CurID),
    visited_expression(CurID,InState),
    specfile:compute_operation_effect_max(InState,OpName,Operation,NewState,_TransPathInfo,Max),
    % prob_socketserver:writeln_log(sol(OpName,NewState)), %%
    user:add_trans_id(CurID,Operation,NewState,[],NewID,[],TransId),
    % prob_socketserver:writeln_log(ids(OpName,Operation,NewID,TransId)), %%
    create_operation_term(TransId,Operation,CurID,NewID, Operation_op).

get_op_from_id(Id,Term) :-
    transition(Src,Op,Id,Dest),
    create_operation_term(Id,Op,Src,Dest,Term).

% create a term op(...) with infos for Java interface about an operation-transition
create_operation_term(TransId,Op,CurID,NewID, JavaOpTerm) :-
   (filter_op(TransId,Op,CurID,NewID, JavaOpTerm) -> true
     ; add_error(eclipse_interface,'filter_op failed: ', filter_op(TransId,Op,CurID,NewID, JavaOpTerm)),fail).

/* ------------------------- */
/* SAP Testcase Generation   */
/* ------------------------- */
:- use_module(sap).

sap_generate_testcases(Events,TargetPredicate,MaxDepth,MaxNewNodes,Filename,NumberOfTestcases,UncoveredEvents) :-
    type_predicate(TargetPredicate,TypedPredicate),
    computeOperationsForState_and_generate_testcases(Events,TypedPredicate,MaxDepth,MaxNewNodes,Filename,
                                   NumberOfTestcases,UncoveredEvents).
   
sap_generate_local_testcases(GFile,LFile,MaxNodes,Sat,Unsat) :-    
    sap:tcl_create_local_testsuite(GFile,LFile,MaxNodes,Sat,Unsat).

sap_find_test_path(Events,EndPredicate,Timeout,Operations) :-
    type_predicate(EndPredicate,TypedPredicate),
    ( create_testcase_path(init,Events,TypedPredicate,Timeout,Trace) -> true
    ; otherwise -> Trace = []),
    % Trace can be "timeout" or "interrupt", too
    ( is_list(Trace) -> op_tuples_to_full_operations(Trace,Operations)
    ; otherwise      -> Operations=Trace).

op_tuples_to_full_operations([],[]).
op_tuples_to_full_operations([(Id,OpTerm,Src,Dst)|Trest],[Op|Orest]) :-
    create_operation_term(Id,OpTerm,Src,Dst,Op),
    op_tuples_to_full_operations(Trest,Orest).

type_predicate(Predicate,TypedPredicate) :-
    b_type_expression(Predicate,[variables],pred,TypedPredicate,Errors),
    add_all_perrors(Errors,[],sap_target_predicate_type_error),
    Errors = [].

/* ------------------------- */
/* Formula evaluation        */
/* ------------------------- */
evaluation_get_top_level(Tops) :-
    % Currently, we do not want to see the rodin information positons
    % in the expression viewer
    suppress_rodin_positions,
    bv_get_top_level(TopIds),
    findall( top(Id,Label,Children),
             ( member(Id,TopIds), bv_expand_formula(Id,Label,Children) ),
             Tops).
evaluation_expand_formula(Id,Label,Children) :-
    % Currently, we do not want to see the rodin information positons
    % in the expression viewer
    suppress_rodin_positions,
    bv_expand_formula(Id,Label,Children).
evaluation_get_values(Ids,StateId,Values) :-
    bv_get_values(Ids,StateId,Values).
evaluation_insert_formula(AST,ParentId,Id) :-
    b_type_expression(AST,[variables],_,Typed,Errors),
    ( Errors == [] ->
        bv_insert_formula(Typed,ParentId,Id)
    ; otherwise ->
        add_error_and_fail(eclipse_interface,'Could not type-check AST','')).


/* ------------------------- */
/* Check for deadlocks       */
/* ------------------------- */

:- meta_predicate call_with_smt_mode_enabled(0).
call_with_smt_mode_enabled(Call) :-
    get_preference(use_smt_mode,SMT),
    get_preference(use_clpfd_solver,CLPFD),
    set_preference(use_smt_mode,true),
    set_preference(use_clpfd_solver,true), % we also enable CLPFD
    call_cleanup( (Call,!),
                  (set_preference(use_smt_mode,SMT),set_preference(use_clpfd_solver,CLPFD)
                  )).

deadlock_freedom_check(Result) :-
    create_texpr(truth,pred,[],Predicate),
    deadlock_freedom_check1(Predicate,Result).
deadlock_freedom_check(Predicate,Result) :-
    b_type_expression(Predicate,[variables],pred,TPredicate,Errors),
    ( Errors == [] ->
        deadlock_freedom_check1(TPredicate,Result)
    ; otherwise ->
        Result = errors(Errors)).
deadlock_freedom_check1(Predicate,Result) :-
    % always do a deadlock check with SMT mode enabled
    call_with_smt_mode_enabled(deadlock_freedom_check2(Predicate,Result)).
deadlock_freedom_check2(Predicate,Result) :-
    user_interruptable_call_det(user:catch_clpfd_overflow_call1(cbc_deadlock_freedom_check(State,Predicate,0)),
                                InterruptResult),!,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ; otherwise ->
        Result = deadlock(Transition,StateId),
        create_helper_transition(root,deadlock_check,State,StateId,Transition)).
deadlock_freedom_check2(_Predicate,no_deadlock_found).

create_helper_transition(SrcId,Op,DstState,DstId,Transition) :-
    user:tcltk_add_new_transition_transid(SrcId,Op,DstId,DstState,[],TransId),
    transition(SrcId,Action,TransId,DstId),!,
    create_operation_term(TransId,Action,SrcId,DstId,Transition).

/* ------------------------------- */
/* Find values                     */
/* ------------------------------- */
%find_values(Commands,Predicate,MaxNumberSolutions,Solutions) :-
%    setup_userdef_state_space(Commands,root,State,Ids,ConstantsId).

%setup_userdef_state_space2(copy_state(Id),CState,State,Ids,ConstantsId) :-
%    visited_expression(Id,S),
%    ( S=concrete_constants(Constants) ->
%        ( CState=root -> ConstantsId = S
%        ; otherwise   -> ConstantsId = CState),
        



/* ------------------------------- */
/* Check for invariant violations  */
/* ------------------------------- */
invariant_check(all,Result) :-
    findall(OpName,b_get_machine_operation(OpName,_,_,_),Ops),
    invariant_check2(Ops,Result).
invariant_check(ops(Ops),Result) :-
    invariant_check2(Ops,Result).
invariant_check2(Ops,Result) :-
    call_with_smt_mode_enabled(user_interruptable_call_det(invariant_check3(Ops,Res,[]), Status)),
    ( Status = interrupted -> Result = interrupted
    ; otherwise            -> Result = Res).
invariant_check3([]) --> !.
invariant_check3([Op|Rest]) -->
    invariant_check_for_single_op(Op),
    invariant_check3(Rest).
:- use_module(clpfd_interface,[catch_clpfd_overflow_call1/1]).
invariant_check_for_single_op(OpName,In,Out) :-
    ( clpfd_interface:catch_clpfd_overflow_call1(b_state_model_check:state_model_check(OpName,State1,Operation,State2)) ->
        In = [counterexample(OpName,Trans1,Trans2)|Out],
        atom_concat( invariant_check_ , OpName, RootTrans),
        create_helper_transition(root,    RootTrans,State1,StateId1,Trans1),
        create_helper_transition(StateId1,Operation,State2,_StateId2,Trans2)
    ; otherwise ->
        In = Out).

/* ------------------------- */
/* Generate Visualizations   */
/* ------------------------- */   

evaluate(In,Result,Warnings) :- tcltk_interface:tcltk_eval(In,Result,Warnings,_Solution).

debug_invariants(Result):-
    ~~mnf(bvisual:get_invariant(BExpr)),
    generate_tree_from_boolean_expression(BExpr, Tree,[]),
    new_graph(debug_invariants),
    translate_to_graph(Tree,debug_invariants,_),
    get_graph(debug_invariants,Result).
    
generate_tree_from_boolean_expression(BExpr,Tree,BSVars):-
    ( 
        current_state_corresponds_to_initialised_b_machine,!,
        specfile:current_b_expression(CurState0),!,
        
   ~~mnf(b_interpreter:insert_before_substitution_variables(BSVars,[],CurState0,CurState0,CurState))
   
        ;
        CurState = []
    ),
    ~~mnf(get_tree_from_expr(Tree, BExpr, [], CurState)).
    
   translate_to_graph(bind(Children,Label,Type,Value,_),G,Id) :- add_vertex(G,Id,[Label,Type,Value]), translate_children(Children,G,Id).
   translate_children([],_,_).
   translate_children([H|T],G,Parent) :- translate_to_graph(H,G,Child), add_edge(G,_,Parent,Child,[]), translate_children(T,G,Parent).

/* ------------------------- */
/* Plugin Handling           */
/* ------------------------- */
:- use_module(plugins, [start_plugin_for_current_mode/1,plugin_get_output/2]).
activate_plugin(ID) :-
   start_plugin_for_current_mode(ID).

get_plugin_output(OutputID,Output) :-
   plugin_get_output(OutputID,Output).


/* ------------------------- */
/* Transfer Statespace       */
/* ------------------------- */

% Transfer Statespace structure. Used for model checking. 
get_statespace_edges_newer_than(Last,L) :- 
          state_space:state_counter(X), H is X - 1, 
          (Last < 0 -> findall(edge(ID,root,TO),transition(root,_,ID,TO),A); A = []),
          add_transitions(Last, H, A, L).

add_transitions(X,Y,A,A) :- X >= Y,!.
add_transitions(X,Y,A,R) :- X < Y,!, findall(edge(ID,Y,TO),transition(Y,_,ID,TO),L,A), Y1 is Y - 1, add_transitions(X,Y1,L,R).

/* ------------------------- */
/* Dot Printing              */
/* ------------------------- */

dot_hierarchy(Filename) :- b_machine_hierarchy:print_hierarchy(Filename).

/* ------------------------- */
/* Evalstore                 */
/* ------------------------- */
es_reset :- evalstore_reset.
es_eval(EvalstoreId,UntypedFormula,Timeout,Result) :-
    user_interruptable_call_det(evalstore_eval_formula(EvalstoreId,UntypedFormula,Timeout,Result1),
                                InterruptResult),!,
    ( InterruptResult = interrupted    -> Result = interrupted
    ; Result1 = ok(V,TypedIds,StoreId) ->
        Result = ok(V,PP,Ids,StoreId),
        get_texpr_ids(TypedIds,Ids),
        es_eval_translate(V,PP)
    ; otherwise                        -> Result = Result1).
es_eval_translate(pred(V),V).
es_eval_translate(value(V),PP) :-
    translate_bvalue(V,PP).
es_copy_from_statespace(StateId,EvalstoreId) :-
    evalstore_create_store_by_state(StateId,EvalstoreId).

