% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(enabling_analysis,[tcltk_cbc_enabling_analysis/1, tcltk_cbc_enabling_analysis/2,
                             tcltk_cbc_dependence_analysis/1, print_enable_table/1,
                             tcltk_create_enable_graph/1,
		             tcltk_create_dependence_graph/1,
                             dependent_actions/3]).

% --------------

:- use_module(bmachine,            [b_get_machine_operation/4, b_get_operation_normalized_read_write_info/3]).
:- use_module(b_state_model_check, [get_negated_guard/3]).
:- use_module(sap).
:- use_module(library(ordsets)).
:- use_module(library(lists)).

tcltk_cbc_dependence_analysis(list([list(['Origin'|SOps])|Result])) :- 
  findall(Op, b_get_machine_operation(Op,_,_,_), Ops), sort(Ops,SOps),
  findall(list([OpName1|EnableList]),
     (b_get_sorted_op(OpName1),
      findall(Enable,cbc_dependence_analysis(OpName1,_OpName2,Enable),EnableList)),
     Result),
  print_enable_table([list(['Origin'|Ops])|Result]).

b_get_sorted_op(Op) :- findall(Op, b_get_machine_operation(Op,_,_,_), Ops),
    sort(Ops,SOps), member(Op,SOps).

cbc_dependence_analysis(OpName1,OpName2,Res) :-
    b_get_machine_operation(OpName1,_,_,_),
    b_get_read_write(OpName1,Reads1,Writes1),
    get_negated_guard(OpName1,PosGuard1,NegGuard1),
    b_get_sorted_op(OpName2),
    (OpName1=OpName2 -> Res = '='  % TO DO: check if it is possible that for two different parameter values of the same event there is dependence
      ; OpName2 @< OpName1 -> Res = '-'  % our checking is symmetric; only check one pair
      ; b_get_read_write(OpName2,Reads2,Writes2),
		format("CHECKING DEPENDENCE: ~w r:~w / w:~w  <-->  ~w r:~w / w:~w~n",
				 [OpName1,Reads1,Writes1,OpName2,Reads2,Writes2]),
		(static_independence(Reads1,Writes1,Reads2,Writes2)
		  -> Res = static_independent
		  ;  ord_intersect(Writes1,Writes2) -> Res = race_dependent
		     % TO DO: in this case check if there is indeed a race dependence (e.g. in scheduler new and ready_active are actually independent !)
		     % Set up [OpName1,OpName2], [OpName2,OpName1] and see if for ord_intersect(Writes1,Writes2) we can find different values
		  ;  get_negated_guard(OpName2,PosGuard2,NegGuard2),
			 (sap:testcase_path_timeout(pred(PosGuard1),300,[OpName2],NegGuard1,_Constants,_Ops,_TestS,_TI,R1)
			   ->  format(" ~w may disable ~w (~w)!~n",[OpName2,OpName1,R1]),
				   (R1==timeout -> Res=timeout_dependent; Res = dependent)
			  ; % TO DO: check for enumeration warnings !
			    sap:testcase_path_timeout(pred(PosGuard2),300,[OpName1],NegGuard2,_,_,_,_,R2)
				-> format(" ~w may disable ~w (~w)!~n",[OpName1,OpName2,R2]),
				   (R2==timeout -> Res=timeout_dependent; Res = dependent)
			  ; Res = independent)
		)
    ).

dependent_actions(OpName1,OpName2,Res) :- 
    b_get_read_write(OpName1,Reads1,Writes1),
    get_negated_guard(OpName1,PosGuard1,NegGuard1),
    (OpName1=OpName2 -> Res = '='
      ; OpName2 @< OpName1 -> Res = '-'  % our checking is symmetric; only check one pair
      ; b_get_read_write(OpName2,Reads2,Writes2),
		(static_independence(Reads1,Writes1,Reads2,Writes2)
		  -> Res = static_independent
		  ;  ord_intersect(Writes1,Writes2) 
			-> Res = race_dependent
		  ;  get_negated_guard(OpName2,PosGuard2,NegGuard2),
			  (sap:testcase_path_timeout(pred(PosGuard1),300,[OpName2],NegGuard1,_Constants,_Ops,_TestS,_TI,_R)
			   	-> Res = dependent
			  ; sap:testcase_path_timeout(pred(PosGuard2),300,[OpName1],NegGuard2,_,_,_,_,_R)
				-> Res = dependent
			  ; Res = independent
			  )
		)
    ).

static_independence(Reads1,Writes1,Reads2,Writes2) :-
    \+ ord_intersect(Writes1,Reads2), % op1 does not modify guard/effect of op2
    \+ ord_intersect(Writes2,Reads1), % op2 does not modify guard/effect of op1
    \+ ord_intersect(Writes1,Writes2). % no race condition

% --------------

% check which operations can be enabled after executing another operation
% to do: move maybe to another module; provide proper CSV export using Sebastian's modules
cbc_enable_analysis(OpName1,OpName2,Enable,ExtraTimeout) :- 
   OpName1='INITIALISATION',print('CHECKING ENABLING AFTER INITIALISATION'),nl,
   b_get_machine_operation(OpName2,_,_,_),
    print('INITIALISATION'), print(' ---> '), print(OpName2), print('  :: '),
    Timeout is 200+ExtraTimeout,
    ( sap:testcase_path_timeout(init,Timeout,[OpName2],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,R)
      -> print(R), print(' : '),
         get_negated_guard(OpName2,_,NegGuard),
         (sap:testcase_path_timeout(init,Timeout,[],NegGuard,_Constants2,_Ops2,_TestS2,_TI2,R2)
           -> print(R2), print(' : '),Enable=possible
            ; Enable=guaranteed)
      ;  Enable=impossible),
      print(Enable),nl.
cbc_enable_analysis(OpName1,OpName2,Enable,ExtraTimeout) :-
    Timeout1 is 300+ExtraTimeout,
    Timeout2 is 200+ExtraTimeout,
    b_get_machine_operation(OpName1,_,_,_),
    b_get_read_write(OpName1,Reads1,Writes1),
    %get_negated_guard(OpName1,PosGuard1,_NegGuard1),
    format("CHECKING ENABLING AFTER: ~w r:~w / w:~w~n",[OpName1,Reads1,Writes1]),
    b_get_machine_operation(OpName2,_,_,_),
    b_get_read_write(OpName2,Reads2,Writes2),
    get_negated_guard(OpName2,PosGuard2,NegGuard2),
    print(OpName1), print(' ---> '), print(OpName2), print('  :: '),
    (\+ ord_intersect(Writes1,Reads2) -> WR=independent, print('(wr_indep) ')
      ; WR=dependent, print('(wr_dep) ')),
    ((WR=independent,\+ ord_intersect(Reads1,Reads2))
     -> (\+ ord_intersect(Writes1,Writes2) -> Enable = independent
         ; Enable = activation_indepdendent  /* independent for enabling, but effect of OpName2 can be modified */
        )
     ; % sap:testcase_path_timeout(invariant,Timeout1,[OpName1,OpName2],b(truth,pred,[]),_Csts,_Ops,_TestS,_TI,R)
      sap:testcase_path_timeout(invariant,Timeout1,[OpName1],PosGuard2,_Csts,_Ops,_TestS,_TI,R) % advantage over version with [OpName1,OpName2] : one less state to setup and enumerate; but inner guards may not be checked
      -> print(enabled_after(R)), print(' : '),
         %print('Pos Guard: '),translate:print_bexpr(PosGuard2),nl,
         %print('Neg Guard: '),translate:print_bexpr(NegGuard2),nl,
         (sap:testcase_path_timeout(invariant,Timeout1,[OpName1],NegGuard2,_Csts2,_Ops2,_TestS2,_TI2,R2)
           -> print(disabled_after(R2)), print(' : '),
            % TO DO: test if NegGuard2 holds initially it is possible to execute [OpName1,OpName2]; if not: OpName1 cannot enable OpName2, only keep it -> Enable=keep_possible
            % then we could check if OpName1 can disable OpName2: PosGuard2,[OpName1],NegGuard2
              (WR=dependent 
              -> % b_get_invariant_from_machine(INV), % assume invariant holds ?
              ((OpName1\=OpName2, % otherwise OpName2 must be enabled if [OpName1] can be executed
                %sap:testcase_path_timeout(pred(NegGuard2),Timeout2,[OpName1,OpName2],b(truth,pred,[]),_,_,_,_,R3)
                sap:testcase_path_timeout(pred(NegGuard2),Timeout2,[OpName1],PosGuard2,_,_,_,_,R3))
               -> print(enable(R3)), print(' : '),
                  (sap:testcase_path_timeout(pred(PosGuard2),Timeout2,[OpName1],NegGuard2,_,_,_,_,R4) 
                  -> print(disable(R4)), print(' : '),
                     (member(timeout,[R,R2,R3,R4]) -> Enable=timeout_possible ; Enable=possible)
                  ;  (member(timeout,[R,R2,R3]) -> Enable=timeout_possible_enable ; Enable=possible_enable) /* Opname cannot disable OpName2; only enable it */
                  )
               ; /* OpName1 cannot enable OpName2; only preserve it */
                  (sap:testcase_path_timeout(pred(PosGuard2),Timeout2,[OpName1],NegGuard2,_,_,_,_,R4) 
                  -> print(disable(R4)), print(' : '),
                     (member(timeout,[R,R2,R4]) -> Enable=timeout_possible_disable; Enable=possible_disable)
                  ;  (member(timeout,[R,R2]) -> Enable=timeout_keep; Enable=keep) /* Opname cannot enable or disable */
                  )
              )
              ; % possible after, can also be disabled after but op1 cannot modify guard variables of op2
                ((R==timeout;R2==timeout)
                 -> Enable=timeout_keep
                 ; Enable=keep))  % was possible_act_indep
            ; (R==timeout -> Enable=timeout_guaranteed; Enable=guaranteed))
      ;  Enable=impossible),
     print('Enable='),print(Enable),nl.

tcltk_cbc_enabling_analysis(L) :- tcltk_cbc_enabling_analysis(0,L).
tcltk_cbc_enabling_analysis(ExtraTimeout,list([list(['Origin'|Ops])|Result])) :- 
  findall(Op, b_get_machine_operation(Op,_,_,_), Ops),
  findall(list([OpName1|EnableList]),
     (init_or_op(OpName1),
      findall(Enable,cbc_enable_analysis(OpName1,_OpName2,Enable,ExtraTimeout),EnableList)),
     Result),
  print_enable_table([list(['Origin'|Ops])|Result]).

:- use_module(ample_sets,[dependent_act/2]).
enable_analysis(OpName1,OpName2,Enable) :- 
    b_get_read_write(OpName1,_Reads1,Writes1),
    b_get_read_write(OpName2,Reads2,_Writes2),
    (\+ample_sets:dependent_act(OpName1,OpName2) ->
	Enable=independent
    ; otherwise ->
	get_negated_guard(OpName1,PosGuard1,_NegGuard1),
	(sap:testcase_path_timeout(pred(PosGuard1),300,[OpName1,OpName2],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,R1) -> 
		get_negated_guard(OpName2,_PosGuard,NegGuard),
                (sap:testcase_path_timeout(no_init,300,[OpName1],NegGuard,_Constants2,_Ops2,_TestS2,_TI2,R2) -> 
              		((ord_intersect(Writes1,Reads2),sap:testcase_path_timeout(pred(NegGuard),300,[OpName1,OpName2],b(truth,pred,[]),_,_,_,_,R3)) -> 
					( member(timeout,[R1,R2,R3]) ->
						print('TIMEOUTS [R1,R2,R3]: '),print([R1,R2,R3]),nl,
						Enable=timeout
					;otherwise ->
						Enable=possible
					)
			;otherwise ->
				( member(timeout,[R1,R2]) ->
					print('TIMEOUTS [R1,R2]: '),print([R1,R2]),nl,
					Enable=timeout
				; otherwise ->
					Enable=possible_disable_or_keep
				)
			)
		; otherwise -> 
			(R1==timeout ->
				print('TIMEOUTS R1: '),print(R1),nl,
				Enable=timeout
			; otherwise ->
				Enable=guaranteed
			)
		)
	; otherwise ->
		Enable=impossible
	)
    ).

init_or_op('INITIALISATION').
init_or_op(OpName1) :- b_get_machine_operation(OpName1,_,_,_).

b_get_read_write(O,R,W) :- b_get_operation_normalized_read_write_info(O,R,W),!.
b_get_read_write(_O,[all],[all]). % in Event-B no read information stored at the moment; all calls to b_get_operation_normalized_read_write_info fail

% print table in CSV format
print_enable_table(R) :- print_enable_table_list(R).
print_enable_table_list([]) :- nl,nl.
print_enable_table_list([list(L)|T]) :-
   print_csv_list(L), print_enable_table_list(T).
print_csv_list([]) :- nl.
print_csv_list([H]) :- !,print(H),nl.
print_csv_list([H|T]) :- print(H), print(','), print_csv_list(T).

:- use_module(bmachine,            [b_get_machine_operation/4]).
:- use_module(flow).
tcltk_create_enable_graph(File) :- 
	findall(Action, (bmachine:b_get_machine_operation(Action,_,_,_)),AllActions),
	ample_sets: determine_dependent_actions(AllActions,_DepGraph),
	determine_enabled_actions(AllActions,AllActions,EnableTuples),
	flow:reset_count,
	flow:enumerate_events(AllActions),
	print(writing_to_dot_file(File)),nl,
	tell(File),
	flow: dot_header('EnableGraph'),
	flow: dot_node(AllActions),
	dot_edge(EnableTuples),
	flow: dot_footer,
	told,
	ample_sets: clear_dynamic_predicates_for_POR.

dot_edge([]).
dot_edge([(A1,A2,Enable)|T]) :- 
	flow:eventnumber(A1,A1N), 
	flow:eventnumber(A2,A2N),  
	format_edge(A1N,A2N,Enable), 
	dot_edge(T).

format_edge(FromNumber,ToNumber,Enable) :- 
        ( Enable==dependent -> 
		print(FromNumber), print('->'), print(ToNumber), print(' '), 
		( Enable==dependent         -> print('[color="forestgreen" label=dependent]')
		; Enable==race_dependent    -> print('[color="seablue" label=race_dependent]')
		; otherwise                 -> print('[color="red" label=error]')
		)
	; otherwise ->
		print(FromNumber), print('->'), print(ToNumber), print(' '), 
		( Enable==guaranteed        -> print('[color="forestgreen" label=guaranteed]') 
		; Enable==possible          -> print('[color="blue" label=possible]')
		; Enable==possible_enable   -> print('[color="blue" label=possible_enable]')
		; Enable==timeout           -> print('[color="brown" label=timeout]')
		; otherwise                 -> print('[color="red" label=error]')
		)
	),nl.

determine_enabled_actions([],_AllActions,[]).
determine_enabled_actions([Act|Actions],AllActions,Res) :-
	determine_enabled_ralation(Act,AllActions,ActRes),
	append(ActRes,ActionsRes,Res),
	determine_enabled_actions(Actions,AllActions,ActionsRes).

determine_enabled_ralation(_Act,[],[]).
determine_enabled_ralation(Act,[Act1|Acts],Res) :-
	enable_analysis(Act,Act1,Enable),
	(member(Enable,[guaranteed,possible,timeout,possible_enable]) 
		-> Res = [(Act,Act1,Enable)|R],
		   determine_enabled_ralation(Act,Acts,R)
	; otherwise 
		-> determine_enabled_ralation(Act,Acts,Res)
	).

:- use_module(library(ugraphs)).
tcltk_create_dependence_graph(File) :- 
	findall(Action, (bmachine:b_get_machine_operation(Action,_,_,_)),AllActions),
	ample_sets: determine_dependent_actions(AllActions,DepGraph),
	flow: reset_count,
	flow: enumerate_events(AllActions),
	edges(DepGraph,Edges),
	to_dot_edges(Edges,DotEdges),
	print(writing_to_dot_file(File)),nl,
	tell(File),
	flow: dot_header('DepGraph'),
	flow: dot_node(AllActions),
	dot_edge(DotEdges),
	flow: dot_footer,
	told,
	ample_sets: clear_dynamic_predicates_for_POR.

to_dot_edges([],[]).
to_dot_edges([X-Y|T],[(X,Y,dependent)|TRes]) :- 
	delete(T,Y-X,TR),
	to_dot_edges(TR,TRes).

