% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(eventhandling, [register_event_listener/3,announce_event/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'The eventhandling module is used to register predicates that are called when certain events (like loading a specification) occurr.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).

:- use_module(error_manager,[add_internal_error/2]).
:- use_module(tools,[ajoin/2]).

:- dynamic registered_listener/3.

:- meta_predicate register_event_listener(+,0,+).
register_event_listener(Event,Module:Pred,Description) :-
    ( known_events(Event,_) ->
        register_event_listener2(Event,Module:Pred,Description)
    ; otherwise ->
        add_internal_error('Unknown event',register_for_event(Event,Module:Pred,Description))).
register_event_listener2(Event,Module:Pred,Description) :-
    assert(registered_listener(Event,Module:Pred,Description)).

known_events(specification_initialised,'Directly after the specificiation has been loaded').

announce_event(Event) :-
    ( known_events(Event,_) ->
        announce_event2(Event)
    ; otherwise ->
        add_internal_error('Unknown event',announce_event(Event))).
announce_event2(Event) :-
    findall(Module:Predicate, registered_listener(Event,Module:Predicate,_Desc), Calls),
    maplist(call_for_event(Event), Calls).
call_for_event(Event,Module:Predicate) :-
    ( call(Module:Predicate) -> true
    ; otherwise ->
        ajoin(['Call for event ',Event,' failed.'],Msg),
        add_internal_error(Msg,Module:Predicate)).
