% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

/*
 find trivial whens (directly translateable to block)
 start with: sicstus -l find_whens.pl --goal whens,halt.
 or: sicstus -l find_whens.pl --goal validate_refactoring('orig.pl','new.pl'),halt.
 or: sicstus -l find_whens.pl --goal count_whens,halt.
*/

portray_message(informational, _).

:- use_module(library(file_systems)).
:- use_module(library(lists)).

whens :-
	file_members_of_directory(Dir),
	search_when_l(Dir),
	flush_output.

search_when_l([]).
search_when_l([H|T]) :-
	( is_prolog(H,F,Name) -> search_when(F,Name)
	; otherwise -> true),
	search_when_l(T).

is_prolog(F-_,F,N) :- atom_chars(F,Cs), append(Nc,['.',p,l],Cs), atom_chars(N,Nc).

search_when(F,Name) :-
	search_when2(F,Name).
search_when2(F,Name) :-
	open(F,read,S),
	search(Name,S),
	close(S).

search(F,S) :-
	catch(read(S,Term), _, Term = exception),
	( Term == end_of_file ->
	    true
	; otherwise ->
	    check(Term,F),
	    search(F,S)).

check( (Head :- when(Blocking,_) ),F ) :-
  get_variables(Blocking,Vars),
  Head =.. [Functor|Args],
  check_args(Args,Vars,Functor,BlockList),!,
  to_commas(BlockList,Blocks),
  print(F),print(:),
  print(Functor),print(' -> block '),print(Blocks),print(.),nl.
check(_,_).

get_variables(Blocking, Vars) :-
	get_variables1(Blocking,Vars,[]).
get_variables1((A,B)) --> !,
	get_variables1(A),
	get_variables1(B).
get_variables1(E) -->
	[Vars],
	{get_variables2(E,Vars,[])}.
get_variables2(nonvar(V)) --> [V].
get_variables2((A;B)) -->
	get_variables2(A),
	get_variables2(B).

check_args(_,[],_,[]).
check_args(Args,[Vs|Vss],Functor,[Block|Brest]) :-
	check_args1(Args,Vs,Syms),
	Block =.. [Functor|Syms],
	check_args(Args,Vss,Functor,Brest).
check_args1(Args,Vs,Syms) :-
	sort(Vs,SVs),
	check_args2(Args,SVs,Syms).
check_args2([],[],[]).
check_args2([H|T],Vs,[S|Srest]) :-
	( exact_mem(H,Vs,Vrest) ->
	    S = '-',check_args2(T,Vrest,Srest)
	; otherwise ->
	    S = '?', check_args2(T,Vs,Srest)).
exact_mem(V,[H|T], Rest) :-
	( V==H -> Rest=T
	; otherwise -> Rest=[H|R],exact_mem(V,T,R)).


validate_refactoring(Orig,Ref) :-
	open(Orig,read,O),
	open(Ref,read,R),
	validate_ref(O,R),
	close(O),
	close(R).
validate_ref(O,R) :-
	read(O,OTerm),
	read(R,RTerm),
	( OTerm == end_of_file, OTerm==RTerm ->
	    true
	; val(OTerm,RTerm,R) ->
	    validate_ref(O,R)
	; otherwise ->
	    line_count(R,Line),
	    print('could not validate '),print(Line),nl,
	    fail).

val(Term,Term,_) :- !.
val( (Head :- when(Blocking,Body) ), (:- block Blocks),R ) :-
  get_variables(Blocking,Vars),
  Head =.. [Functor|Args],
  check_args(Args,Vars,Functor,BlockList),
  to_commas(BlockList,Blocks),
  read(R,Term),
  Term = ( Head :- Body ).

to_commas([A],A).
to_commas([A,B|Rest],(A,R)) :- to_commas([B|Rest],R).


count_whens :-
	file_members_of_directory(Dir),
	count_when_l(Dir),
	flush_output.

count_when_l([]).
count_when_l([H|T]) :-
	( is_prolog(H,F,Name) -> count_when(F,Name)
	; otherwise -> true),
	count_when_l(T).

count_when(F,Name) :-
	open(F,read,S),
	count(Name,S,0,Count),
	close(S),
	(Count==0 -> true ; print(Name),print(:),print(Count),nl).

count(F,S,I,O) :-
	catch( read(S,Term), _, Term = exception),
	( Term == end_of_file ->
	    I=O
	; otherwise ->
	    count_in_term(Term,I,X),
	    count(F,S,X,O)).

count_in_term( (_ :- Body), I, O) :- !,count_in_body(Body,I,O).
count_in_term(_,X,X).

count_in_body(when(_,_),I,O) :- !, O is I+1.
count_in_body((A,B),I,O) :- !, count_in_body(A,I,X), count_in_body(B,X,O).
count_in_body((A;B),I,O) :- !, count_in_body(A,I,X), count_in_body(B,X,O).
count_in_body((A->B),I,O) :- !, count_in_body(A,I,X), count_in_body(B,X,O).
count_in_body(_,X,X).
