% (c) 2009-2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


/* ---------------------- */
/* GENERATING NEW SYMBOLS */
/* ---------------------- */
/* File: gensym.pl */


:- module(gensym,
    [gensym/2, reset_gensym/0, gennum/1, gennum_count/1, reset_gennum/0, reset_gennum/1]).

 :- use_module(module_information).
:- module_info(group,tools).
:- module_info(description,'Tools for generating fresh symbols and numbers.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(tools, [string_concatenate/3]).

/* ===================================================== */

/* ------ */
/* GENSYM */
/* ------ */

/* generate a new symbol with the given prefix */
/* code from the "Art of Prolog" */
gensym(Prefix,V) :-
	var(V),
	atom(Prefix),
	oldvalue(Prefix,N),
	N1 is N + 1,
	set_flag(Prefix,N1),
	name(PE_Sep,"__"),
	string_concatenate(Prefix,PE_Sep,PreSep),
	string_concatenate(PreSep,N1,V).

:- dynamic gennum_count/1.
gennum_count(0).

gennum(Nr) :- 
	retract(gennum_count(Nr)), N1 is Nr + 1,
	assert(gennum_count(N1)).

reset_gensym :- retractall(flag(_,_)).
reset_gennum :- reset_gennum(0).

reset_gennum(Nr) :- retractall(gennum_count(_)), assert(gennum_count(Nr)).

oldvalue(Prefix,N) :- flag(Prefix,N),!.
oldvalue(_Prefix,0).
set_flag(Name,X) :-
	nonvar(Name),
	retract(flag(Name,_Val)),!,
	asserta(flag(Name,X)).
set_flag(Name,X) :-
	nonvar(Name),
	asserta(flag(Name,X)).

:- dynamic flag/2.
%flag(foo,foo) :- fail.

/* already defined in ecce_sicstus
:- mode string_concatenate(i,i,o).
string_concatenate(X,Y,XY) :-
   name(X,Xs),name(Y,Ys),append(Xs,Ys,XYs),name(XY,XYs).
*/