:- module(input_syntax_tree, [remove_raw_position_info/2
                             ,extract_raw_identifiers/2
                             ,get_definition_name/2
                             ]).

remove_raw_position_info([],Output) :- !,Output=[].
remove_raw_position_info([I|Irest],[O|Orest]) :-
    !,remove_raw_position_info(I,O),
    remove_raw_position_info(Irest,Orest).
remove_raw_position_info(Input,Output) :-
    compound(Input),functor(Input,Functor,Arity),Arity>0,!,
    Arity2 is Arity-1,
    functor(Output,Functor,Arity2),
    remove_raw_position_info2(1,Arity,Input,Output).
remove_raw_position_info(Input,Input).


remove_raw_position_info2(Pos,Arity,Input,Output) :-
    Pos < Arity,!,Pos2 is Pos+1,
    arg(Pos2,Input,IA), arg(Pos,Output,OA),
    remove_raw_position_info(IA,OA),
    remove_raw_position_info2(Pos2,Arity,Input,Output).
remove_raw_position_info2(_Pos,_Arity,_Input,_Output).


extract_raw_identifiers(Ast,Ids) :-
    extract_raw_identifiers2(Ast,Unsorted,[]),
    sort(Unsorted,Ids).
extract_raw_identifiers2([]) --> !,[].
extract_raw_identifiers2([I|Irest]) -->
    !,extract_raw_identifiers2(I),
    extract_raw_identifiers2(Irest).
extract_raw_identifiers2(identifier(_,I)) --> !,[I].
extract_raw_identifiers2(definition(_,I,Args)) --> !,[I],
    extract_raw_identifiers2(Args).
extract_raw_identifiers2(Input) -->
    {compound(Input),functor(Input,_Functor,Arity),Arity>1,!},
    extract_raw_identifiers3(2,Arity,Input).
extract_raw_identifiers2(_) --> [].

extract_raw_identifiers3(Pos,Arity,Input) -->
    {Pos=<Arity,!,Pos2 is Pos+1, arg(Pos,Input,Arg)},
    extract_raw_identifiers2(Arg),
    extract_raw_identifiers3(Pos2,Arity,Input).
extract_raw_identifiers3(_Pos,_Arity,_Input) --> [].

get_definition_name(expression_definition(_Pos,Name,_Args,_Body),Name).
get_definition_name(predicate_definition(_Pos,Name,_Args,_Body),Name).
get_definition_name(substitution_definition(_Pos,Name,_Args,_Body),Name).
