% (c) 2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(interval_calc,[interval_addition/3
                        ,interval_subtraction/3
                        ,interval_multiplication/3
                        ,interval_negation/2
                        ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,analysis).
:- module_info(description,'This module provides methods to calculate with interval expressions.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(inf_arith).

interval_addition(irange(Am,AM),irange(Bm,BM),irange(Cm,CM)) :-
    Cm infis Am+Bm, CM infis AM+BM.
interval_subtraction(irange(Am,AM),irange(Bm,BM),irange(Cm,CM)) :-
    Cm infis Am-BM, CM infis AM-Bm.
interval_multiplication(irange(Am,AM),irange(Bm,BM),irange(Cm,CM)) :-
    MultExprs = [Am*Bm,Am*BM,AM*Bm,AM*BM],
    get_min(MultExprs,Cm),
    get_max(MultExprs,CM).
interval_negation(irange(Am,AM),irange(Cm,CM)) :-
    CM infis -Am, Cm infis -AM.

get_min([A],Min) :- !, Min infis A.
get_min([A|Rest],Min) :-
    get_min(Rest,Rmin), Min infis min(A,Rmin).
get_max([A],Max) :- !,Max infis A.
get_max([A|Rest],Max) :-
    get_max(Rest,Rmax), Max infis max(A,Rmax).
