% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 
/* ---------------------------------------- */
/* (c) Michael Leuschel, 2001-2010          */
/* All rights reserved.                     */
/* This code may not be distributed without */
/* written permission by Michael Leuschel   */
/* ---------------------------------------- */


:- module(junit_tests,[set_junit_dir/1,print_junit/2,create_junit_result/6,junit_mode/1]).

:- use_module(library(xml)).
:- use_module(library(lists)).
:- use_module(library(codesio)).
:- use_module(library(file_systems)).
:- use_module(tools).
:- use_module(library(process)).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides functionality to output test case results in a format compatible to junit.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').



:- dynamic junit_mode/1.

set_junit_dir(Directory) :- assert(junit_mode(Directory)).


%   <testcase name="Call" classname="Selfcheck.Module" time="2"><error type="error"></testcase>
            
create_junit_result(Name,Section, Module, Time, Verdict, result(Name,Section,Module,TR,Verdict)) :-
    junit_mode(_),!,
    convert_to_junit_time(Time,TR).

create_junit_result(_Call, _Section, _Module, _Time, _Verdict, none).

convert_to_junit_time(T,R) :- number(T),!, R is T / 1000.
            
print_junit(Results,Module) :-
          junit_mode(Dir),!,print_junit2(Dir,Results,Module).
print_junit(_Results,_Module).

:- use_module(error_manager).
print_junit2(Dir,Results,Module) :-
          prepare_xml(Results,Codes),
          open_file(0,Dir,Module,Stream),
          put_codes(Codes,Stream), 
          close(Stream),!.
print_junit2(Dir,_Results,Module) :- add_error(junit_tests,'print_junit failed',Dir:Module).

:- use_module(tools,[safe_on_exception/3]).  
open_file(C,Dir,Module,Stream) :- process_id(PID), number_codes(C,CC), atom_codes(Num,CC), ajoin([Dir,'/',Num,'_',PID,'_',Module,'.xml'], File), open_file2(C,File,Dir,Module,Stream).
open_file2(C,F,D,M,Stream) :- file_exists(F),!, C1 is C+1, open_file(C1,D,M,Stream).
open_file2(_C,File,_D,_M,Stream) :-
   safe_on_exception(E,open(File, write,Stream),
            (print('### Exception while opening Junit File: '), print(File),nl,
             add_error(junit_tests,'Junit File Opening Exception: ',File:E), fail)).

prepare_xml(Results,Codes) :-
 maplist(prepare_testcase_xml, Results,X),
 xml_parse(Codes,xml([],element(testsuite,[],X))).
prepare_xml(Results,_Codes) :- add_error_fail(junit_tests,'prepare_xml failed',Results).


prepare_testcase_xml(result(Call,Section,Module,Time,Verdict),element(testcase,R,Error)) :-
 maplist(nonvar,Result) -> R = [name=Name, classname=Classname, time=T],
                           write_to_codes(Call,Name),
                           atom_codes(Module,MC),
                           atom_codes(Section,SC),
                           ( Verdict=pass     -> Error=[]
                           ; Verdict=error(E) -> (createError(E,Err), Error=[element(error,['='(type,"Error")],Err)])
                           ; Verdict=skip     -> Error=[element(skipped,[],[])]),
                           append([SC,".",MC],Classname),
                           number_codes(Time,T)
 ; otherwise            -> add_error_fail(junit_tests,'prepare_testcase_xml called on non-ground result',Result).


createError([],[]).
createError([E|T],[pcdata(Error)|R]) :- write_to_codes(E,Codes), append(["\n",Codes,"\n"],Error), createError(T,R).
