% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(model_checker,[is_deadlocked/1, open_search/10,
  last_step_breadth_first_in_mixed_mode/0, depth_breadth_first_mode/1,
  set_depth_breadth_first_mode/1, reset_depth_breadth_first_mode/0,
  dbf_modes/3,
  reset_model_checker/0,
  proof_info_model_checking_incomplete/5, model_checking_is_incomplete/6]).

 :- use_module(module_information).
:- module_info(group,model_checker).
:- module_info(description,'The core model checker (does not treat LTL).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(b_interpreter,[state_violates_assertions/2]).
:- use_module(error_manager,[real_error_occurred/0]).

:- use_module(specfile,[animation_mode/1,csp_mode/1]).
:- use_module(tools,[print_message/1]).
:- use_module(state_space,[visited_expression/2,transition/3, operation_not_yet_covered/1]).
:- use_module(debug,[debug_println/2]).

:- dynamic depth_breadth_first_mode/1.
depth_breadth_first_mode(mixed).

dbf_modes(0,mixed,'Mixed DF/BF').
dbf_modes(1,breadth_first,'Breadth First').
dbf_modes(2,depth_first,'Depth First').

reset_depth_breadth_first_mode :- set_depth_breadth_first_mode(mixed).
set_depth_breadth_first_mode(MODE) :-
  retractall(depth_breadth_first_mode(_)),
  debug_println(9,depth_breadth_first_mode(MODE)),
  assert(depth_breadth_first_mode(MODE)).
  
reset_model_checker :- retractall(last_step_breadth_first_in_mixed_mode).

/* open_search: search for errors/goals among open nodes */
/* LimitNr: max number of nodes to check, NodesAnalysed: number of nodes checked */
open_search(LimitNr,NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,StopAtFullCoverage,PartialOrderReduction) :-
    animation_mode(MODE),
    depth_breadth_first_mode(DFMODE),
    statistics(runtime,[ReferenceTime,_]),
    ((PartialOrderReduction == 1, MODE == b, \+ample_sets: dependency_enabling_computed) -> 
	ample_sets:compute_dependendency_relation_of_all_events_in_the_model
    ; true), 
    open_search10(0,LimitNr,ReferenceTime,
                           NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,StopAtFullCoverage,PartialOrderReduction,MODE,DFMODE),
    clear_context_state.


:- use_module(library(random)).
open_search10(Nr,LimitNr,RefTime,NodesAnalysed,LimitTime,Res,
              FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,SFC,PartialOrderReduction,MODE,DF) :-
  %print(os(Nr,LimitNr,RefTime)),nl,
  Nr < LimitNr,!,
  (\+ number(LimitTime)
   -> open_search_do_one_node(Nr,LimitNr,RefTime,NodesAnalysed,LimitTime,Res,
					  FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,SFC,PartialOrderReduction,MODE,DF)
   ;  statistics(runtime,[NewTime,_]),
	  ( (Nr mod 1000 =:= 0,Nr\=0)
		 ->  Delta is NewTime-RefTime,
		     ((Delta>500 ; random(0,500,X), X=<Delta)
		     -> print('* '), print(Nr),print(' states checked (total: '), 
		        get_state_space_stats(States,Transitions,_ProcessedTotal),
		        print(States), print(' states, '),print(Transitions),
		        print(' transitions,'),
			    statistics(memory_used,M), tools:print_mb(M),
			    NodesPerSec is 1000*1000 / (NewTime-RefTime),
			    format('), current ~2F states/sec~n',[NodesPerSec])
			  ; true),
			 RefTime1 = NewTime
		 ;   RefTime1 = RefTime),
	  (NewTime>LimitTime
		-> Res = [timeout,Nr], NodesAnalysed=Nr
		;  open_search_do_one_node(Nr,LimitNr,RefTime1,NodesAnalysed,LimitTime,Res,
					  FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,SFC,PartialOrderReduction,MODE,DF))
	).
open_search10(N,_LimitNr,_RefTime,N,_,no ,_,_,_,_,_,_,_,_).
 
:- use_module(preferences,[preference/2]).
:- use_module(specfile,[prepare_state_for_specfile_trans/2]).
open_search_do_one_node(Nr,LimitNr,RefTime,NodesAnalysed,LimitTime,Res,
                           FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,StopAtFullCoverage,PartialOrderReduction,MODE,DFMODE) :-
  get_next_node_to_check(FindInvViolations,ID,MODE,DFMODE),!,
  set_context_state(ID),
  ((preferences:preference(ignore_hash_collisions,true),
    preferences:preference(forget_state_space,true))
    -> retract_visited_expression(ID,CurState) %, retractall(id_to_marker(ID,_)) not necessary
    ;  visited_expression(ID,CurState)),
  (FindInvViolations=1 % compute invariant status
    -> specfile: prepare_state_for_specfile_trans(CurState,PCurState),
       user:compute_invariant_if_necessary(ID,PCurState,MODE)
     ; CurState=PCurState), 
    %print(processed_invariant(ID)), statistics(memory_used,M), tools:print_mb(M),nl,
  !,
  ((PartialOrderReduction ==1, MODE == b) 
	-> user: compute_ample_set(ID,PCurState) % <----- compute ample set
  ; user:compute_all_transitions(ID,PCurState) -> true
  ),  
  %print_message(mc(ID)),
    %print(computed_all_transitions(ID)), statistics(memory_used,M2), tools:print_mb(M2),nl,
   % !,
  user:do_one_gui_event,
  (open_search_found_error(ID,PCurState,Res,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,MODE)
   ->  print(found_error(Res,ID)),nl, clear_context_state,
       NodesAnalysed=Nr,
       user:tcltk_execute_trace_to_node(ID)
    ;  N1 is Nr + 1, 
	   open_search10(N1,LimitNr,RefTime,NodesAnalysed,LimitTime,Res,
		 FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,PartialOrderReduction,MODE,DFMODE)
  ).
open_search_do_one_node(N,_LimitNr,_RefTime,N,_,all,_,_,_,_,_,_,_,_) :- 
    print_message('All open nodes visited').
                                                      


:- use_module(specfile,[b_or_z_mode/1]).
%%get_next_node_to_check(1,ID,_) :- state_space:not_invariant_checked(ID). % if there is an existing node whose invariant we have not yet checked: check it now
get_next_node_to_check(_,ID,AnimMODE,DFMODE) :- 
  %(nonvar(ID) -> print(nonvar(ID)),nl ; true),
  get_open_node_to_check(DFMODE,ID),
  (b_or_z_mode(AnimMODE) -> assert(not_invariant_checked(ID)) ; true).
  
:- dynamic last_step_breadth_first_in_mixed_mode/0.
get_open_node_to_check(depth_first,NodeID) :- !,
   pop_id_from_front(NodeID).
get_open_node_to_check(mixed,NodeID) :-  %%depth_breadth_first_mode(DFMODE),
   random(1,3,1),
                  /* random(1,4,1) ==> 33 percent chance of going DF */
                  /* random(1,3,1) ==> 50 percent chance of going DF */
                  /* (random(1,3,1);random(1,3,1)) ==> 75 % chance of going DF */
   retractall(last_step_breadth_first_in_mixed_mode),
   !,
   % We are going depth_first
   pop_id_from_front(NodeID). %, print_message(a(NodeID)).
get_open_node_to_check(DFMODE,NodeID) :-
   ((DFMODE=breadth_first ; last_step_breadth_first_in_mixed_mode)
    -> true ; assert(last_step_breadth_first_in_mixed_mode)),
   %print(going_bf(NodeID)),nl,
   pop_id_from_end(NodeID).

:- use_module(state_space).
open_search_found_error(ID,CurState,Error,
                       _FindDeadlocks,1,_FindGoal,_FindAssViolations,_StopAtFullCoverage,MODE) :-
    find_invariant_error(MODE,ID,CurState,Error).
open_search_found_error(ID,_CurState,deadlock,1,_,_,_,_,_) :-
   is_deadlocked(ID). %, print(deadlock(ID)),nl.
open_search_found_error(ID,CurState,assertion_violation,_,_,_,1,_,_MODE) :-
   state_violates_assertions(ID,CurState).
open_search_found_error(ID,_CurState,ERR,_,_,_,_,_,_MODE) :-
   state_space:state_error(ID,_,Error), Error \== invariant_violated,
   ERR = state_error(Error).
open_search_found_error(_ID,_CurState,general_error_occurred,_,_,_,_,_,_MODE) :-
   error_manager:real_error_occurred.  /* do we need to reset_errors_occured before model check ?? */
open_search_found_error(ID,_CurState,goal_found,_,_,1,_,_,MODE) :-
   (user:test_goal_in_node(ID) ; is_csp_goal_state(ID,MODE)).
open_search_found_error(_ID,_CurState,full_coverage,_,_,_,_,1,_MODE) :-
   \+ operation_not_yet_covered(_).

find_invariant_error(b,ID,_CurState,invariant_violation) :- !, % here we know that invariant information has been computed
    (invariant_violated(ID) -> true).
find_invariant_error(cspm,ID,_CurState,csp_error) :- !, % we know that only csp_can be possible
    is_csp_error_state(ID).
find_invariant_error(MODE,ID,CurState,invariant_violation) :- 
    user:check_invariantKO(MODE,ID,CurState).
find_invariant_error(csp_and_b,ID,_CurState,csp_error) :- !, % xtl mode cannot apply
    is_csp_error_state(ID).
find_invariant_error(MODE,ID,CurState,xtl_error) :- 
    is_xtl_error_state(ID,CurState,MODE).
   
is_deadlocked(ID) :-
  preferences:preference(forget_state_space,true),!, /*  transitions not added */
  (retract(state_space:transition(ID,_,_,_)) -> fail ; true).
is_deadlocked(ID) :- % ID must be ground upon call
 %%  visited_expression(ID,_CurState),
  deadlocked_state(ID),
  \+ max_reached_for_node(ID).
/* \+((transition(ID,_,NID),NID \= ID)).   <-- to detect more liberal form of DEADLOCK */

is_csp_error_state(ID) :- 
  transition(ID,io(_,error,_),_).  /* <--- THE CSP has raised the error channel */

:- use_module(xtl_interface,[xtl_property/2, promela_property/2]).
is_xtl_error_state(_ID,CurState,AM) :-  
  ( AM=xtl     -> xtl_property(CurState,unsafe)   /* <--- XTL has found unsafe configuration */
  ; AM=promela -> promela_property(CurState,unsafe)
%  ; AM=smv     -> smv_property(CurState,unsafe)
  ).

is_csp_goal_state(ID,MODE) :- csp_mode(MODE),
  visited_expression(ID,_CurState),
  transition(ID,io(_,goal,_),_).


% ---------------------------------------

:- use_module(tools, [ajoin/2]).
:- use_module(symmetry_marker,[hash_model_checking_imprecise/2]).
%model_checking_is_unsound(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term)
model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,'') :-
    proof_info_model_checking_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg).
model_checking_is_incomplete(_FindInv,_FindDeadlocks,_FindGoal,_FindAss,Msg,Term) :-
   (hash_model_checking_imprecise(ID,BasicType)
    ->  Msg = '*** Hashing may be imprecise for, e.g., for variable ',
        translate:pretty_type(BasicType,S),
        ajoin([ID,' : ',S],Term)).

:- use_module(bmachine, [b_machine_has_proven_invariants/0, b_machine_has_unproven_assertions/0]).
proof_info_model_checking_incomplete(1,_FindDeadlocks,_FindGoal,0,Msg) :-
    b_machine_has_proven_invariants,
    preferences:get_preference(use_po,true),
    b_machine_has_unproven_assertions,
    Msg = '*** Using proof information (-p PROOF_INFO TRUE) is potentially incomplete when not checking theorems (-noass) !'.
