:- module(module_information,[module_info/2,module_info/3,get_module_info/3,get_maximal_module_info/2,number_of_groups/1,number_of_modules/1]).

% a simple mechanism to provide annotations for various modules
% examples:
% :- module_info(description,'....').
% :- module_info(group,kernel).
:- meta_predicate module_info(0,*).
:- dynamic get_module_info/3.
module_info(Module:Attribute,Value) :- !,%print(describe_module(Module,Attribute,Value)),nl,
   assert(get_module_info(Module,Attribute,Value)).
module_info(Attribute,Value) :- nl,print('### Attribute has no module prefix: '), Attribute,nl,
   print('### Value: '), print(Value),nl.
% called for non-module prolog files
module_info(Module,Attribute,Value) :-
   assert(get_module_info(Module,Attribute,Value)).

:- module_info(group,infrastructure).
:- module_info(description,'This module contains functionality to store basic documentation and information about all modules in ProB.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).

get_maximal_module_info(revision,X) :- !,
    get_module_info(module_information,revision,X).
% below does not work properly for revision numbers as 9000 comes after 10000 !
get_maximal_module_info(Info,X) :- findall(Rev,get_module_info(_Module,Info,Rev),RS),
   sort(RS,SortedRevList), last(SortedRevList,X).

% module_information:get_maximal_module_info(revision,X).
%  module_information:get_maximal_module_info(lastchanged,X).

number_of_groups(X) :-
    findall(Group,get_module_info(_Module,group,Group),List),
    remove_dups(List,NList),
    length(NList,X).

number_of_modules(X) :-
    findall(Module,get_module_info(Module,_A,_B),List),
    remove_dups(List,NList),
    length(NList,X).

number_of_modules(Group,X) :-
    findall(Module,get_module_info(Module,group,Group),List),
    remove_dups(List,NList),
    length(NList,X).


