:- module(builtin_tools, 
        [is_builtin/1,
         executable_builtin/1, 
         is_decidable_builtin/1,
         specialize_builtin/2,
         post_process_builtin/2]).



:- dynamic executable_builtin/1.

executable_builtin(number(Expr)) :-
     nonvar(Expr).
executable_builtin(is(_,Expr)) :- 
     ground(Expr).
executable_builtin(arg(Nr,T,_)) :- 
     number(Nr), nonvar(T).
executable_builtin(functor(T,F,N)) :- 
     (nonvar(T) ; atomic(F), number(N)).
executable_builtin('='(_X,_Y)).
%executable_builtin('C'(_X,_Y,_Z)). % old-style DCGs
executable_builtin(true).
executable_builtin(fail).
executable_builtin(false).
executable_builtin(otherwise).
executable_builtin(repeat).
executable_builtin('=..'(X,Y)) :-
	(nonvar(X)) ; (list_ok_for_eqdotdot(Y)).
executable_builtin('<'(X,Y)) :-
	arithmetic_expression(X),arithmetic_expression(Y).
executable_builtin('=<'(X,Y)) :-
	arithmetic_expression(X),arithmetic_expression(Y).
executable_builtin('>'(X,Y)) :-
	arithmetic_expression(X),arithmetic_expression(Y).
executable_builtin('>='(X,Y)) :-
	arithmetic_expression(X),arithmetic_expression(Y).
executable_builtin('@<'(X,Y)) :- '?='(X,Y). % is this sufficient ??
executable_builtin('@=<'(X,Y)) :-'?='(X,Y). % is this sufficient ??
executable_builtin('@>'(X,Y)) :-'?='(X,Y). % is this sufficient ??
executable_builtin('@>='(X,Y)) :-'?='(X,Y). % is this sufficient ??
executable_builtin(nonvar(X)) :-
	nonvar(X).
executable_builtin(ground(X)) :-
	ground(X).
executable_builtin(number(X)) :-
	nonvar(X).
executable_builtin(float(X)) :-
	nonvar(X).
executable_builtin(integer(X)) :-
	nonvar(X).
executable_builtin(atomic(X)) :-
	nonvar(X).
executable_builtin(atom(X)) :-
	nonvar(X).
executable_builtin(call(X)) :-
	nonvar(X),
	executable_builtin(X).
executable_builtin(   dif(X,Y)) :- '?='(X,Y). 
executable_builtin('\\=='(X,Y)) :- '?='(X,Y).
executable_builtin(  '=='(X,Y)) :- '?='(X,Y). 
executable_builtin( '\\='(X,Y)) :- '?='(X,Y). 
executable_builtin('=:='(X,Y)) :-
    arithmetic_expression(X),arithmetic_expression(Y).
executable_builtin('=\\='(X,Y)) :-
    arithmetic_expression(X),arithmetic_expression(Y).
executable_builtin(append(X,_,Y)) :-
    (is_list_skeleton(X) ; is_list_skeleton(Y)).
executable_builtin(sort(X,_)) :-
    ground(X).
executable_builtin(member(_,Y)) :-
    is_list_skeleton(Y).
executable_builtin(length(Y)) :-
    is_list_skeleton(Y).
executable_builtin(atom_concat(X,Y,_)) :-
	atom(X), atom(Y).
executable_builtin(atom_codes(A,_)) :- atom(A).
executable_builtin(number_codes(A,_)) :- number(A).
executable_builtin(atom_chars(A,_)) :- atom(A).
executable_builtin(number_chars(A,_)) :- number(A).
executable_builtin(atom_length(A,_)) :- atom(A).


:- use_module(library(terms),[term_variables/2]).
specialize_builtin(X,Code) :- executable_builtin(X),!,call(X), Code=true.
specialize_builtin(ground(X),Code) :- !,
  term_variables(X,V), Code=ground(V).
specialize_builtin(X,X). % we cannot specialize; just return call as residual code

post_process_builtin(otherwise,C) :- !, C=true.
post_process_builtin(ground(X),Code) :- !,
  term_variables(X,V), Code=ground(V).
post_process_builtin('=..'(X,Y),Code) :- nonvar(X),!,
   X =.. YY, Code = '='(Y,YY).
post_process_builtin('=..'(X,Y),Code) :- list_ok_for_eqdotdot(Y),!,
   XX =.. Y, Code = '='(X,XX).

% ---------------------------


arithmetic_expression(X) :- ground(X).
 /* will cause error to appear at specialization time
    if not really an arithmetic expression !
    TO DO : improve ; or add try-catch around executing builtin's */

list_ok_for_eqdotdot(L) :-
	nonvar(L),L = [H|T],
	atomic(H),
	is_list_skeleton(T).

is_list_skeleton(X) :- nonvar(X),is_list_skeleton_aux(X).
is_list_skeleton_aux([]).
is_list_skeleton_aux([_H|T]) :- is_list_skeleton(T).

% ----------------------------

% update using  builtin_tools:update.
%
update :- clause(executable_builtin(H),_), portray_clause(is_builtin(H)),fail.
update.
is_builtin(number(_)).
is_builtin(_ is _).
is_builtin(arg(_,_,_)).
is_builtin(functor(_,_,_)).
is_builtin(_=_).
is_builtin(true).
is_builtin(fail).
is_builtin(false).
is_builtin(otherwise).
is_builtin(repeat).
is_builtin(_=.._).
is_builtin(_<_).
is_builtin(_=<_).
is_builtin(_>_).
is_builtin(_>=_).
is_builtin(_@<_).
is_builtin(_@=<_).
is_builtin(_@>_).
is_builtin(_@>=_).
is_builtin(nonvar(_)).
is_builtin(ground(_)).
is_builtin(number(_)).
is_builtin(float(_)).
is_builtin(integer(_)).
is_builtin(atomic(_)).
is_builtin(atom(_)).
is_builtin(call(_)).
is_builtin(dif(_,_)).
is_builtin(_\==_).
is_builtin(_==_).
is_builtin(_\=_).
is_builtin(_=:=_).
is_builtin(_=\=_).
is_builtin(append(_,_,_)).
is_builtin(sort(_,_)).
is_builtin(member(_,_)).
is_builtin(length(_)).
is_builtin(atom_concat(_,_,_)).
is_builtin(atom_codes(_,_)).
is_builtin(number_codes(_,_)).
is_builtin(atom_chars(_,_)).
is_builtin(number_chars(_,_)).
is_builtin(atom_length(_,_)).

% return true when a built-in can be decided to be true or false, with on other options
% i.e., when is it safe to execute an if-then-else with it as condition

is_decidable_builtin(true).
is_decidable_builtin(fail).
is_decidable_builtin(false).
is_decidable_builtin(otherwise).
is_decidable_builtin(repeat).
is_decidable_builtin('='(X,Y)) :- '?='(X,Y).
is_decidable_builtin('=='(X,Y)) :- '?='(X,Y).
is_decidable_builtin('\\='(X,Y)) :- '?='(X,Y).
is_decidable_builtin('\\=='(X,Y)) :- '?='(X,Y).
is_decidable_builtin(dif(X,Y)) :- '?='(X,Y).
is_decidable_builtin(member(X,L)) :-
   (ground(X),ground(L)  ; is_list_skeleton(L),\+ member(X,L)).
is_decidable_builtin(var(X)) :- nonvar(X).
is_decidable_builtin(nonvar(X)) :- nonvar(X).
is_decidable_builtin(ground(X)) :- ground(X).
is_decidable_builtin(atom(X)) :- atomic(X).
is_decidable_builtin(atomic(X)) :- atomic(X).
is_decidable_builtin(number(X)) :- atomic(X).
is_decidable_builtin('<'(X,Y)) :- ground(X),ground(Y).
is_decidable_builtin('=<'(X,Y)) :- ground(X),ground(Y).
is_decidable_builtin('>'(X,Y)) :- ground(X),ground(Y).
is_decidable_builtin('>='(X,Y)) :- ground(X),ground(Y).
is_decidable_builtin('@<'(X,Y)) :- '?='(X,Y).
is_decidable_builtin('@=<'(X,Y)) :- '?='(X,Y).
is_decidable_builtin('@>'(X,Y)) :- '?='(X,Y).
is_decidable_builtin('@>='(X,Y)) :- '?='(X,Y).


