:- module(mini_test_pe,[mymem/2, test/2, test2/2, test3/2]).
%:- ensure_loaded(pe).
:- use_module(pe).

:- unfold mymem(_,_). % important to move this before the protect clause !
:- unfold protect(_,_).
protect(X,R) :- (mymem(Y,X) -> print(mem(Y,X)),R=mem(Y) ; print(not_mem(Y,X)),R=not_mem), nl.
mymem(X,[X|_]).
mymem(X,[_|T]) :- mymem(X,T).
test(X,R) :- nl,nl,
   clear_specialized_code,
   partially_evaluate_and_load(mymem(X,R),C), print(calling(C)),nl,call(C), print(C),nl.
test2(X,R) :- nl,nl,
   clear_specialized_code,
   partially_evaluate_and_load(protect(X,R),C), call(C).
   
:- test(X,[a,b,c]), X==a.

:- unfold tom(_,_).
:- use_module(library(timeout),[time_out/3]).
tom(X,R) :- time_out(mymem(X,R),10,TRes),
               (TRes==time_out -> print(time_out),nl ; true).
test3(X,R) :- nl,nl,
   clear_specialized_code,
   partially_evaluate_and_load(tom(X,R),C), call(C).

:- unfold bmem(_,_).
:- unfold btest(_,_).
btest(X,L) :- bmem(X,L).
:- block bmem(?,-), bmem(-,?).
bmem(X,[X|_]).
bmem(X,[_|T]) :- bmem(X,T).

:- use_module(library(plunit)).
:- begin_tests(pe).

test(simple_mymem1) :- R=[a,b,c],
   partially_evaluate_and_load(mymem(X,R),C),
   call(C),X==c.
test(simple_mymem2,[all(X==[a,b,c,a,d,e,f])]) :- R=[a,b,c,a,d,e,f],
   partially_evaluate_and_load(mymem(X,R),C),
   call(C).
test(simple_mymem4,[fail]) :- R=[a,b,c,a,d,e,f], 
   partially_evaluate_and_load(mymem(X,R),C), X=g,
   call(C).
test(simple_mymem5,[fail]) :- R=[a,b,c,a,d,e,f],  X=g,
   partially_evaluate_and_load(mymem(X,R),C),
   call(C).
test(simple_mymem6,[fail]) :- R=[], 
   partially_evaluate_and_load(mymem(_X,R),C),
   call(C).

test(block_member1) :-  R=[a,b,c],
   partially_evaluate_and_load(btest(b,R),C),
   call(C).
test(block_member2) :-  R=[a|T],
   partially_evaluate_and_load(btest(b,R),C),
   T = [b,c],
   call(C).
test(block_member3) :-  R=[a|T],
   partially_evaluate_and_load(btest(b,R),C),
   call(C), var(T), % check that block declaration active
   T = [b,c].
test(block_member4,[true(var(T4))]) :-  R=[a|T],
   partially_evaluate_and_load(btest(b,R),C),
   call(C), var(T), % check that block declaration active
   T = [c,c|T2], var(T2), % check that block declaration active
   T2 = [d,e|T3], var(T3), % check that block declaration active
   T3 = [b|T4].
test(block_member5,[fail]) :-  R=[a|T],
   partially_evaluate_and_load(btest(b,R),C),
   call(C), var(T), % check that block declaration active
   T = [c,d].
:- end_tests(pe).

:- run_tests.
