% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(predicate_debugger,[tcltk_debug_properties/3,
  tcltk_debug_properties_or_op/4,  adapt_tcl_operation_name/2,
  get_operation_enabling_condition/6]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,misc).
:- module_info(description,'The predicate debugger adds conjuncts one by one until the predicate becomes unsatisfiable.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


:- use_module(debug).
:- use_module(self_check).
:- use_module(preferences).
:- use_module(error_manager).

:- use_module(translate).
:- use_module(b_interpreter).
:- use_module(bsyntaxtree).
:- use_module(b_ast_cleanup).

:- use_module(library(timeout)).

/* b_interpreter:b_debug_properties. */

:- volatile checked_property/1, last_checked_solution/2, failed_property/2, time_out_property/2, cur_conjunct_nr/1, projection_onto_component/1.
:- dynamic checked_property/1.
:- dynamic last_checked_solution/2.
:- dynamic failed_property/2.
:- dynamic time_out_property/2.
:- dynamic cur_conjunct_nr/1.
:- dynamic projection_onto_component/1.

reset_debug_properties :-
   retractall(last_debug_info(_,_)),
   retractall(checked_property(_)),
   retractall(last_checked_solution(_,_)),
   retractall(failed_property(_,_)),
   retractall(time_out_property(_,_)),
   retractall(cur_conjunct_nr(_)),
   retractall(projection_onto_component(_)).
   
:- use_module(b_global_sets,[b_fd_type/3]).
tcltk_debug_properties(RealRes,UnsatCore,Satisfiable) :-
   tcltk_debug_properties_or_op('@PROPERTIES',UnsatCore,RealRes,Satisfiable).

% code below because names with $ cause problems in Tcl with eval
adapt_tcl_operation_name(X,R) :- var(X),!,R=X.
adapt_tcl_operation_name('@PROPERTIES','$setup_constants') :- !.
adapt_tcl_operation_name('@INITIALISATION','$initialise_machine') :- !.
adapt_tcl_operation_name(X,X).

tcltk_debug_properties_or_op(TclOpName,ComputeUnsatCore,RealRes,Satisfiable) :-
   adapt_tcl_operation_name(TclOpName,OpName),
   print(tcltk_debug_properties_or_op(OpName,ComputeUnsatCore,RealRes)),nl,
   (ComputeUnsatCore==false ->
      (OpName='$setup_constants' -> b_debug_properties ; b_debug_operation(OpName))
    ;    set_fast_unsat_core(ComputeUnsatCore),
         print('*** Finding Unsat Core : '), print(ComputeUnsatCore),nl,
         (b_find_unsat_core
          -> print('*** Finished '),nl
          ; add_error_fail(predicate_debugger,'Finding Unsat Core failed',OpName))
   ),
   print(debugged(OpName)),nl,
   findall(TP, (checked_property(P),
               translate_bexpression_with_limit(P,TrP),
               ajoin([TrP,' &\n'],TP)),
               Res1),
   ((projection_onto_component(_Nr))
     -> Res1P = ['(only a sub-component was analysed);'|Res1] ; Res1P=Res1),
   (Res1=[] -> R1=Res1P ; 
     R1=['; ********;;SATISFIABLE CONSTRAINTS and PROPERTIES:;'|Res1P]
   ),
   findall(TP2,( ((failed_property(P,_),RES='false') ; (time_out_property(P,_),RES='time_out')),
               translate_status(P,RES,TP2)),
               Res2),
   (Res2=[] -> R2=[';No unsatisfiable PROPERTY or CONSTRAINT'], Satisfiable=true
           ;   R2=['; ********;;First unsatisfiable PROPERTY or CONSTRAINT:;'|Res2],
               Satisfiable=false
   ),
   findall(TP, (last_checked_solution(ConstantsState,_), % TO DO: Project on remaining variables !?
               translate_equivalence(b(identifier('SOLUTION'),boolean,[]),ConstantsState,TP)),
               Res12),
   append(R2,[';; ********;;Solution for satisfiable part:;'|Res12],R2Res12),
   append(R1,R2Res12,Res),
   findall(SetDescr,(b_fd_type(Set,LowBnd,UpBnd), Card is 1+UpBnd-LowBnd,
             create_texpr(identifier(Set),set(global(Set)),[],SetId),
             create_texpr(card(SetId),integer,[],CardExpr),
             create_texpr(value(int(Card)),integer,[],ValExpr),
             create_texpr(equal(CardExpr,ValExpr),prop,[],EqualExpr),
             translate_bexpression(EqualExpr,SetDescr1),
             ajoin([SetDescr1,'\n'],SetDescr)), Sets),
   
   append([';Cardinalities of SETS:;'|Sets],Res,SRes),
   preferences:get_preference(minint,MININT),
   preferences:get_preference(maxint,MAXINT),
   clear_error_context,
   RealRes = ['SETTINGS:;MININT .. MAXINT = ',MININT,'..',MAXINT|SRes].
tcltk_debug_properties_or_op([],_,'Machine has no PROPERTIES',true).

translate_equivalence(A,B,R) :-
   translate_bexpression(A,TA),
   translate_bstate(B,TB),
   ajoin([TA,'\n   == ',TB,'\n'],R).
translate_status(A,Status,R) :-
   translate_bexpression(A,TA),
   ajoin([TA,'\n   == ',Status],R).
   
:- use_module(bmachine).
b_debug_properties :- set_error_context(b_debug_properties),
   reset_debug_properties,
   %%b_extract_constants_and_properties(Constants,CstTypes,Properties),
   %b_machine_has_constants_or_properties,
   b_get_machine_constants(TypedConstants),
   b_get_properties_from_machine(Properties),!,
   adapt_properties(Properties,TypedConstants,AProperties,ATypedConstants),
   create_texpr(truth,pred,[b_debug_properties],Truth),
   nr_of_conjuncts(AProperties,NrConj),
   (b_debug_properties2(AProperties,Truth,ATypedConstants,[],NrConj) -> true ; true).
   
:- use_module(library(lists)).
adapt_properties(Properties,TypedConstants,Pred,CTypedConstants) :-
    %(b_interpreter:nr_of_components(NrC), NrC>1),
    preferences:get_preference(partition_predicates,true),
    b_interpreter:unsat_properties_component(CompNr),!,
    assert(projection_onto_component(CompNr)),
    print(debugging_unsat_or_skipped_component(CompNr)),nl,
    bsyntaxtree:predicate_components(Properties,Components),
    nth1(CompNr,Components,component(Pred,Vars)),
    print(component_vars(Vars)),nl,
    project_texpr_identifiers(TypedConstants,Vars,CTypedConstants).
adapt_properties(Properties,TypedConstants,Properties,TypedConstants).
   
    
% only keep texpr_identifiers whose id is in the provided list
project_texpr_identifiers([],_,[]).
project_texpr_identifiers([TID|T],Vars,Res) :-
   (get_texpr_id(TID,ID),member(ID,Vars)
    -> Res = [TID|R2] ; Res=R2),
   project_texpr_identifiers(T,Vars,R2).

:- use_module(specfile,[state_corresponds_to_initialised_b_machine/2, state_corresponds_to_set_up_constants/2]).
/* b_interpreter:tcltk_debug_properties_or_op(prog1,R)*/
b_debug_operation(OpName) :-
   reset_debug_properties,
   (state_space:current_expression(CurID,CurState) -> true
      ; add_error(b_debug_operation,'Could not get current state',OpName),fail),
   (state_corresponds_to_initialised_b_machine(CurState,CurBState) -> true
     ; (OpName=='$initialise_machine',
        state_corresponds_to_set_up_constants(CurState,CurBState)) -> true
     ; add_error(b_debug_operation,'Machine not yet initialised! ',CurState),fail),
   set_error_context(operation(OpName,CurID)),
   (get_operation_enabling_condition(OpName,Parameters,EnablingCondition,BSVars,true,false) -> true
     ; add_error(b_debug_operation,'Operation does not exist or cannot determine enabling condition: ',OpName),fail),!, 
   print('Enabling: '), print_bexpr(EnablingCondition),nl,
   flush_output(user),
   create_texpr(truth,pred,[op_enabling_condition],Truth),
   nr_of_conjuncts(EnablingCondition,NrConj),
   ~~mnf(insert_before_substitution_variables(BSVars,[],CurBState,CurBState,NewCurBState)),
   extract_exists_parameters(EnablingCondition,Parameters,NewCond,NewParas),
   (b_debug_properties2(NewCond,Truth,NewParas,NewCurBState,NrConj) -> true ; true).
 
 % Try to lift out existentially quantified variables to the outer level; to make the debugging more fine-grained
 extract_exists_parameters(Pred,Parameters, NewPred, NewParas) :-
    get_texpr_expr(Pred,exists(Vars,Body)),
    disjoint_list_union(Parameters,Vars,Paras1), !,  % also checks if no clash exists; otherwise we have to keep exists
    print(augmented_paras(Paras1)),nl,
    extract_exists_parameters(Body,Paras1,NewPred,NewParas).
 extract_exists_parameters(Pred,Parameters, NewPred, NewParas) :-
    get_texpr_expr(Pred,conjunct(LHS,RHS)),!,
    extract_exists_parameters(LHS,Parameters, NewLHS, IntParas),
    extract_exists_parameters(RHS,IntParas,NewRHS,NewParas),
    conjunct_predicates([NewLHS,NewRHS],NewPred).
 extract_exists_parameters(Pred,Par,Pred,Par).
 
 
:- dynamic obtain_action_enabling_conditions/0, simplify_enabling_condition/1.
obtain_action_enabling_conditions. simplify_enabling_condition(true).
set_obtain_action_enabling_conditions(true) :- !,
 (obtain_action_enabling_conditions -> true ; assert(obtain_action_enabling_conditions)).
set_obtain_action_enabling_conditions(_) :- retractall(obtain_action_enabling_conditions).
set_simplify_enabling_condition(X) :- retractall(simplify_enabling_condition(_)),
  assert(simplify_enabling_condition(X)).

:- use_module(b_interpreter_components,[construct_optimized_exists/4]).

get_operation_enabling_condition(OpName,Parameters,EnablingCondition,BecomesSuchVars,GetAct,Simplify) :-
   set_obtain_action_enabling_conditions(GetAct), % specify whether we want also conditions stemming from actions, such as :() or ::
   set_simplify_enabling_condition(Simplify), % specify whether we want to simplify the EnablingCondition if possible
   nonvar(OpName),  %% print(getting_enabling(OpName,GetAct)),nl,%%
   b_get_machine_operation_for_animation(OpName,_Results,Parameters,TBody,_OType,true),
   get_texpr_expr(TBody,Body),
   get_operation_enabling_condition2(Body,EnablingCondition,BecomesSuchVars).
   
 /* TO DO: recursively expand the code below for more complicated SELECT,ANY,... */
get_operation_enabling_condition2(precondition(PreCond,TBody),Res,BecomesSuchVars) :- !, % PRE
     get_texpr_expr(TBody,Body),
     ~~mnf(get_operation_enabling_condition2(Body,RC,BecomesSuchVars)),
     ~~mnf(conjunct_predicates([PreCond,RC],Res)).
get_operation_enabling_condition2(any(Parameters,PreCond,TBody),Res,BecomesSuchVars) :- !, % ANY
     get_texpr_expr(TBody,Body),
     ~~mnf(get_operation_enabling_condition2(Body,RC,BecomesSuchVars)),
     ~~mnf(conjunct_predicates([PreCond,RC],Res1)),
     simplify_enabling_condition(Simplify),
     construct_optimized_exists(Parameters,Res1,Res,Simplify).
get_operation_enabling_condition2(select([b(select_when(PreCond, TBody),_,_)]),Res,BecomesSuchVars) :- !, 
     get_texpr_expr(TBody,Body),
     ~~mnf(get_operation_enabling_condition2(Body,RC,BecomesSuchVars)),
     ~~mnf(conjunct_predicates([PreCond,RC],Res)).
get_operation_enabling_condition2(select([b(select_when(PreCond, TBody),_,_)],b(skip,_,_)),Res,BecomesSuchVars) :- !, 
     get_texpr_expr(TBody,Body),
     ~~mnf(get_operation_enabling_condition2(Body,RC,BecomesSuchVars)),
     ~~mnf(conjunct_predicates([PreCond,RC],Res)).
   %% TO DO: compose all select_when's with an OR 
get_operation_enabling_condition2(parallel([TH|T]),Res,BecomesSuchVars) :- !, 
     get_texpr_expr(TH,H),
     ~~mnf(get_operation_enabling_condition2(H,E1,BecomesSuchVarsH)),
     (T=[] -> Res=E1, BecomesSuchVarsH=BecomesSuchVars
       ; (~~mnf(get_operation_enabling_condition2(parallel(T),E2,BecomesSuchVarsT)),
          append(BecomesSuchVarsH,BecomesSuchVarsT,BecomesSuchVars),
          ~~mnf(conjunct_predicates([E1,E2],Res)))).
get_operation_enabling_condition2(becomes_element_of(_LHS,RHS),Res,[]) :- 
     obtain_action_enabling_conditions,
     get_texpr_expr(RHS,RHSExpr),get_texpr_type(RHS,Type),!,
     ((simplify_enabling_condition(true),definitely_not_empty(RHSExpr,Type))
       -> %print(def_not_empty(RHSExpr)),nl,
          create_texpr(truth,pred,[],Res)
       ; ~~mnf(create_texpr(empty_set,Type,[],EmptySet)),
         ~~mnf(create_texpr(not_equal(RHS,EmptySet),pred,[],Res2)),
         %print(not_equal_empty(RHSExpr,Type)),nl,
         ~~mnf(clean_up(Res2,[],Res))
     ).
get_operation_enabling_condition2(becomes_such(Vars,Condition),Res,Vars) :-
     obtain_action_enabling_conditions,!,
     simplify_enabling_condition(Simplify),
     construct_optimized_exists(Vars,Condition,Res,Simplify).
get_operation_enabling_condition2(rlevent(_Name,_Section,_Status,_Params,Guard,_Theorems,Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents),Res,BecomesSuchVars) :- !,
   %print(actions(Actions)),nl,
   (obtain_action_enabling_conditions
    ->  ~~mnf(get_operation_enabling_for_event_b_actions(Actions,Guard,Res,BecomesSuchVars))
    ;  Res=Guard,BecomesSuchVars=[]).
get_operation_enabling_condition2(assign(_LHS,_RHS),Truth,[]) :- !,
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(assign_single_id(_IDE,_RHS),Truth,[]) :- !,
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(if(_IFLIST),Truth,[]) :- !, % TO DO: Improve; check if guards inside IF
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(skip,Truth,[]) :- !,
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(X,Truth,[]) :- 
     (obtain_action_enabling_conditions -> print(cannot_obtain_enabling(X)),nl ; true),
     create_texpr(truth,pred,[],Truth).
     
definitely_not_empty(bool_set,_).
definitely_not_empty(integer_set(_),_).
definitely_not_empty(identifier(X),set(global(X))).
definitely_not_empty(string_set,_).


:- use_module(btypechecker,[prime_identifiers/2]).
% get operation enabling condition in context of an event-b action list:
get_operation_enabling_condition_event_b(becomes_such(Vars,Condition),Res,Primed) :- !,
      prime_identifiers(Vars,Primed), % this is different from classical B
      simplify_enabling_condition(Simplify),
      construct_optimized_exists(Primed,Condition,Res,Simplify).
get_operation_enabling_condition_event_b(Subst,Condition,BecomesSuchVars) :- 
	get_operation_enabling_condition2(Subst,Condition,BecomesSuchVars). 

get_operation_enabling_for_event_b_actions([],Res,Res,[]).
get_operation_enabling_for_event_b_actions([H|T],InRes,OutRes,BecomesSuchVars) :- 
     get_texpr_expr(H,Body),
     ~~mnf(get_operation_enabling_condition_event_b(Body,ResH,HBecomesSuchVars)),
     ~~mnf(conjunct_predicates([InRes,ResH],IntRes)),
     get_operation_enabling_for_event_b_actions(T,IntRes,OutRes,TBecomesSuchVars),
     append(HBecomesSuchVars,TBecomesSuchVars,BecomesSuchVars).
  
:- use_module(library(timeout)).


nr_of_conjuncts(Pred,Nr) :- 
  is_a_conjunct(Pred,LHS,RHS),!,
  nr_of_conjuncts(LHS,NrL),
  nr_of_conjuncts(RHS,NrR),
  Nr is NrL+NrR.
nr_of_conjuncts(_,1).

:- volatile last_debug_info/2.
:- dynamic last_debug_info/2.
b_debug_properties2(Pred,Rest,Constants,GlobalState,TotalNr) :-
  retractall(last_debug_info(_,_)), assert(last_debug_info(Constants,GlobalState)),
  (b_debug_find_first_unsat_property(Pred,Rest,Constants,GlobalState,TotalNr)
  -> print('ALL PROPERTIES SATISFIABLE'),nl
  ;  true % get_preference(debug_try_minimise,false) -> true ;  b_find_unsat_core(Constants,GlobalState)
  ).

b_find_unsat_core :-     /* assumes that b_debug_properties was run before */
  % b_machine_has_constants_or_properties,
 %  b_get_machine_constants(TypedConstants),
  % b_get_properties_from_machine(Properties),!,
  % adapt_properties(Properties,TypedConstants,_AProperties,ATypedConstants), 
   %b_find_unsat_core(ATypedConstants,[]).
   last_debug_info(Constants,GlobalState), b_find_unsat_core(Constants,GlobalState).
b_find_unsat_core(Constants,GlobalState) :-
     print('TRYING TO MINIMISE UNSATISFIABLE PROPERTIES'),nl,
     (last_checked_solution(_,Time)
      -> print(' Time for last solution in ms: '), print(Time), nl ; true),
     findall(P,checked_property(P),SatProps),
     (failed_property(FailProp,USTime) -> true
      ; time_out_property(FailProp,USTime) -> true
      ; add_error_fail(b_find_unsat_core,'No failed or timed-out property:',SatProps)),
     print(' Time for unsatisfiable conjuncts in ms: '), print(USTime), nl,
     append(SatProps,[FailProp],Props),
     length(SatProps,SL),
     b_debug_try_remove_props(SatProps,Props,[],Constants,GlobalState,1,SL).
     % TO DO: partition into components; try removing compnents without FailProp
     
b_debug_try_remove_props([],_,_,_,_,_,_) :- !.
b_debug_try_remove_props([H|TSat],[H|TUnsat],Required,TypedConstants,GlobalState,Count,Total) :-
  append(Required,TSat,ConjunctsWithoutH),
  conjunct_predicates(ConjunctsWithoutH,SatPred),
  nl,print('CHECKING REMOVAL ('), print(Count), print('/'),print(Total),
  print(') OF: '),translate_bexpression(H,PT), print(PT),nl,
  (timeout_fast_check_property(sat,SatPred,TypedConstants,Sol,GlobalState,TimeOutResSat)
   -> (TimeOutResSat \= time_out 
       -> append(Required,TUnsat,ConjunctsWithoutH2),
          conjunct_predicates(ConjunctsWithoutH2,UnsatPred),
	  ((timeout_fast_check_property(unsat,UnsatPred,TypedConstants,_,GlobalState,TimeOutRes),
	    print(res(TimeOutRes)),nl,
	    (TimeOutRes = time_out ->  \+ time_out_property(_,_) ; true) 
           )
		-> print('Required for inconsistency : '), print(TimeOutRes),nl,
		   append(Required,[H],NewRequired)
		;  /* H is not required for inconsistency/timeout */
		   print('NOT required for INCONSISTENCY !'),nl,
		   % print_bexpr(UnsatPred),nl,
		   Required=NewRequired,
		   % TO DO: update found solution
		   (retract(last_checked_solution(_,LastTime)) -> true
		     ; print('*** Could not remove last_checked_solution'),nl,
		       preferences:get_computed_preference(debug_time_out,LastTime)
		   ),
		   assert(last_checked_solution(Sol,LastTime)),
		   (retract(checked_property(H)) -> true
			; print('*** Could not remove checked_property: '), print(H),nl
	           )
	  )
      ; print('Required for efficiency'),nl,
        append(Required,[H],NewRequired)
      )
  ;  nl,print('*** INTERNAL ERROR: FAILED'),nl,nl,
     append(Required,[H],NewRequired)
  ),
  C1 is Count+1,
  b_debug_try_remove_props(TSat,TUnsat,NewRequired,TypedConstants,GlobalState,C1,Total).
 
b_debug_find_first_unsat_property(Pred,Rest,Constants,GlobalState,TotalNr) :- 
  is_a_conjunct(Pred,LHS,RHS),!,
  (b_debug_find_first_unsat_property(LHS,Rest,Constants,GlobalState,TotalNr)
   -> (conjunct_predicates([Rest,LHS],Rest2),
       b_debug_find_first_unsat_property(RHS,Rest2,Constants,GlobalState,TotalNr))
   ; fail
  ).
b_debug_find_first_unsat_property(Prop,Rest,TypedConstants,GlobalState,TotalNr) :-
  nl,nl,flush_output(user),
  conjunct_predicates([Rest,Prop],Pred),
  (retract(cur_conjunct_nr(CurNr)) -> true ; CurNr=1), CurNrP1 is CurNr+1,
  assert(cur_conjunct_nr(CurNrP1)),
  print('CHECKING '), print(CurNr), print('/'), print(TotalNr), print(': '),
    translate_bexpression(Prop,PT), print(PT),nl,
    statistics(runtime,[Start,_]),
  (timeout_check_property(Pred,TypedConstants,ConstantsState,GlobalState,TimeOutRes)
   -> statistics(runtime,[Stop,_]), Time is Stop - Start,
      (TimeOutRes=time_out
         -> assert(time_out_property(Prop,Time)),
            print('TIMEOUT !! ('),print(Time), print(' ms)'),nl, flush_output(user),fail
         ;  
            print('OK ('), print(Time), print(' ms)'), flush_output(user),
            assert_checked_property(Prop,ConstantsState,Time) 
      )
   ;  statistics(runtime,[Stop,_]), Time is Stop - Start,
      assert(failed_property(Prop,Time)),
      print('No Solution Found !! ('),print(Time), print(' ms)'),nl, flush_output(user),
      %print_bexpr(Pred),nl,
      fail
  ).

last_unsat_time(Time) :- time_out_property(_,Time) -> true ; failed_property(_,Time).

set_fast_unsat_core(T) :- retractall(fast_unsat_core),
  (T==fast -> assert(fast_unsat_core) ; true).
:- dynamic fast_unsat_core/0.
fast_unsat_core.

get_timeout(sat,DebugTimeOut) :- last_checked_solution(_,LTime),!,DebugTimeOut is LTime+301.
get_timeout(unsat,DebugTimeOut) :- last_unsat_time(USTime),last_checked_solution(_,LSTime),
    (fast_unsat_core -> TT is min(USTime,LSTime) ; TT is max(USTime,LSTime)),!,
    DebugTimeOut is TT+301.
get_timeout(S,DebugTimeOut) :- print('*** Could not get TimeOut : '), print(S),nl,
   preferences:get_preference(time_out,CurTO), DebugTimeOut is CurTO * 1.

timeout_fast_check_property(ExpectSat,Pred,TypedConstants,ConstantsState,GlobalState,TimeOutRes) :-
   get_timeout(ExpectSat,DebugTimeOut),
   % print(time_out_value(DebugTimeOut)),nl,
   % print_bexpr(Pred),nl,
  statistics(runtime,[Start,_]),
  time_out(b_check_property_satisfiable(Pred,TypedConstants,ConstantsState,GlobalState),
            DebugTimeOut,TimeOutRes), 
  statistics(runtime,[Stop,_]), Time is Stop - Start, print(time(Time)),nl,
  (TimeOutRes=time_out -> print(time_out_occurred(Time,DebugTimeOut)),nl, print_bexpr(Pred),nl ; true).
timeout_check_property(Pred,TypedConstants,ConstantsState,GlobalState,TimeOutRes) :-
  preferences:get_computed_preference(debug_time_out,DebugTimeOut),
  time_out(b_check_property_satisfiable(Pred,TypedConstants,ConstantsState,GlobalState),
            DebugTimeOut,TimeOutRes).


   
assert_checked_property(P,ConstantsState,Time) :-
   assert(checked_property(P)),
   (retract(last_checked_solution(_,_))
    -> true ; true),
   assert(last_checked_solution(ConstantsState,Time)).
         
:- use_module(kernel_waitflags).
:- use_module(b_enumerate).
b_check_property_satisfiable(Properties,TypedConstants,ConstantsState,GlobalState) :-
    set_up_typed_localstate(TypedConstants,_FreshVars,TypedVals,[],ConstantsState), 
 %   print(checking_properties(TypedConstants,Properties,ConstantsState,GlobalState)),nl,
  %  print('CHECKING: '),
    init_wait_flags(WF),
    b_tighter_enumerate_values(TypedVals,WF),  % user tighter enumerate, rather than standard enumerate
    (GlobalState = []
      ->  b_test_boolean_expression(Properties,[],ConstantsState,WF)
      ;   b_test_boolean_expression(Properties,ConstantsState,GlobalState,WF)
    ),
    %ground_det_wait_flag(WF), print(grounding_wait_flags(ConstantsState,WF)),nl,
    ground_wait_flags(WF).    
   