% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


portray_message(informational, _).
 
%% sp_4_1_3_nested_exception_patch.pl BEGIN
%% Compile this to patch a bug in 4.1.3 affecting nested exception handlers.
%% This source code is a no-op on other versions of sicstus than 4.1.3.
:- if(current_prolog_flag(dialect,sicstus)).
:- if(current_prolog_flag(version_data,sicstus(4,1,3,0,_))).

prolog:(matching_exception(internal, Ball, Templ) :- !,
        Templ = Ball).
prolog:(matching_exception(prolog, '$reserved'(_), _) :- !,
        fail).
prolog:(matching_exception(prolog, time_out, _) :- !,
        fail).
prolog:(matching_exception(_, Ball, Template) :-
        compound(Ball),
        sysball_iso_equiv(Ball,Iso1Exc), !,
        Template = error(Iso1Exc, Ball)).
prolog:(matching_exception(_, Ball, Ball)).

:- endif.                       % sicstus 4.1.3
:- endif.                       % sicstus
%% sp_4_1_3_nested_exception_patch.pl END

%:- include('self_check_off.pl').

% use module_info/3 because we are not in a module here
:- use_module(module_information).
:- module_info(prob_cli,group,infrastructure).
:- module_info(prob_cli,description,'ProB start file in cli mode.').
:- module_info(prob_cli,revision,'$Rev$').
:- module_info(prob_cli,lastchanged,'$LastChangedDate$').


:- use_module(debugging_calls).
:- register_debugging_call pp_mnf(*), pp_cll(*), mnf(*), mnf(-,*), det_call(*).
:- disable_debugging_calls.

:- use_module(tools,[set_search_pathes/0,string_concatenate/3,arg_is_number/2]). % called first to set_compile_time_search_pathes

:- compile(tcltk_interface).
%:- compile(btracer_tcl).

:- use_module(eclipse_interface).


:- use_module(prob_socketserver).

:- use_module(coverage_statistics,[compute_the_coverage/4]).
:- use_module(value_persistance, [set_storage_directory/1,unset_storage_directory/0]).

do_one_gui_event.
tk_do_all_events.

:- dynamic junit_mode/1.

:- use_module(b_trace_checking,[check_default_trace_for_specfile/1]).
%:- use_module(cspParserHs,'cia/haskell/java_api/cspParserHs.pl',
%               [initCspParser/0]).

:- use_module(library(file_systems),[file_exists/1]).
:- use_module(debug).
:- use_module(preferences).
:- use_module(ltl,[ltl_check_assertions/2,ltl_model_check/4,ctl_model_check/4]).
:- use_module(test_typechecker,[run_typecheck_testcases/0]).

:- use_module(self_check,[disable_interaction_on_errors/0]).
:- use_module(error_manager).

:- use_module(extension('zmq/master')).
:- use_module(extension('zmq/worker')).

:- use_module(self_check,[perform_self_check/0,turn_off_run_time_type_checks/0,turn_on_run_time_type_checks/0]).

:- use_module(library(system)).
:- use_module(library(codesio)).

:- use_module(log_analyser).
:- use_module(bmachine,[b_write_machine_representation_to_file/2,full_b_machine/1]).

:- use_module(xtl_interface,[set_cspm_main_process/1]).
:- use_module(visualize_graph,[tcltk_print_states_for_dot/1]).
:- use_module(reduce_graph_state_space,[print_state_merge_for_dot/1,print_dot_for_dfa_from_nfa/1]).
:- use_module('kodkod/kodkod_test').
:- use_module(predicate_analysis,[test_predicate_analysis/0]).
:- use_module(b_show_history,[write_history_to_file/2,write_values_to_file/1]).
:- use_module(bparser(parser_integration_test),[parser_benchmark/0]).
:- use_module(external_functions,[check_if_predicates_were_evaluated/1]).

printsilent(S) :- option(silent) -> true ; print(S).
nls :- option(silent) -> true ; nl.


:- use_module(junit_tests,[set_junit_dir/1, print_junit/2, create_junit_result/6]).
set_junit_mode(X) :- set_junit_dir(X), statistics(runtime,[Start,_]), assert(junit_mode(Start)).


go_proxy :-
    catch( go_cli(['-s','8888']), halt(ExitCode),
           ( nl,write('CLI halt prevented, exit code '),write(ExitCode),nl) ).

go_cli :- 
    prolog_flag(argv,ArgV),
    go_cli_with_junit_check(ArgV).
  
:- use_module(xtl_interface,[reset_cspm_main_process/0]).
:- use_module(ample_sets, [clear_dynamic_predicates_for_POR/0]).
reset_cli :-
  reset_errors, reset_expected_error_occured,
  preferences:reset_to_defaults,
  reset_cspm_main_process,
  unset_storage_directory,
  clear_dynamic_predicates_for_POR,
  (file_loaded(_)
   -> clear_loaded_machines,
      retractall(file_loaded(_,_)),
      retractall(loaded_main_file(_))
   ;  true).

go_cli_with_junit_check(ArgV) :-
    catch( go_cli(ArgV),
           halt(ExitCode),
           ( ExitCode = 0 ->
               true
           ; junit_mode(S) -> 
               statistics(runtime,[E,_]), T is E - S,
	       create_junit_result(ArgV,'Integration Tests','Integration_Test',T,error([ExitCode]),Result),
               print_junit([Result],'Integration_Test')
           ; throw(halt(ExitCode)))).

go_cli(ArgV) :- 
     (go_cli2(ArgV) -> true
     ; print_error('INTERNAL ERROR OCCURED !'),nl,
       print_error(' (go_cli failed)'),nl, halt_exception(1)).



%:- use_module(extension('myheap/myheap')).
go_cli2(ArgV) :-
    set_search_pathes,
    %myheap_init,
    init_eclipse_preferences,
    cli_init_options(ArgV,RemArgV),
    %% cli_init_options(['-ll'|ArgV],RemArgV),  %% comment in to have an automatically logging probcli
    !,
    if_option_set(test_mode,set_random_seed_to_prologs_current_random_state, set_new_random_seed), 
    if_option_set(verbose,
                  tcltk_turn_debugging_on(19),
                  tcltk_turn_debugging_off),
    if_option_set(very_verbose,
                  tcltk_turn_debugging_on(5)),
    debug_print(9,'Command Line Arguments: '),debug_println(9,ArgV),
    if_option_set(very_verbose,
                  print_options),
    debug_print(6,'Command Line File Args: '),debug_println(6,RemArgV),
    if_option_set(cache_storage(StorageDir), set_storage_directory(StorageDir)),
    if_option_set(parser_benchmark,   % temporay option for running parser_benchmarks
                  parser_benchmark),
    if_option_set(csp_main(MAINPROC),
                  set_cspm_main_process(MAINPROC)),
    if_option_set(parsercp(ParserLoc),
                  assert(parsercall:parser_location(ParserLoc))),
    cli_show_help(ArgV,RemArgV),
    generate_time_stamp(Datime,NOW),
    if_option_set(log(F),
                  cli_start_logging(F,NOW,Datime,RemArgV)),
    if_option_set(evaldot(F),
                  set_eval_dot_file(F)),
    if_option_set(runtimechecking,
                  turn_on_run_time_type_checks,
                  turn_off_run_time_type_checks),
    if_option_set(junit(X),
                  set_junit_mode(X)),
    set_prefs,
    if_option_set(selfcheck,
                  cli_start_selfcheck),
    if_option_set(typechecker_tests,
                  (run_typecheck_testcases -> halt_prob(NOW); halt1_prob(NOW))),
    if_option_set(print_version,
                  (print_version, (RemArgV=[] -> halt_prob(NOW) ; true))),
    if_option_set(print_short_version,
                  (print_short_version, (RemArgV=[] -> halt_prob(NOW) ; true))),
    if_option_set(very_verbose,
                  preferences:print_preferences),
    if_option_set(zmq_worker, zmq_start_worker(localhost,NOW)),
    if_option_set(zmq_worker2(MasterIP), zmq_start_worker(MasterIP,NOW)),
    debug_println(6,loading(RemArgV)),
    cli_load_files(RemArgV,NOW), % all CLI arguments which are not understood are supposed to be files to be treated
    debug_println(6,finished_loading(RemArgV)),
    if_options_set(analyse_log(LF,RowID,ForGnu),
                  analyse_log(LF,RowID,ForGnu)),
    if_options_set(check_log(LF),
                  check_log(LF)),
    if_option_set(socket(Port),
                  cli_start_socketserver(Port)),
    % check_all_expected_errors_occured(NOW),  % is now checked for each file; socket_server should not generate errors ?
    debug_println(20,'% ProB Finished'),
    cli_print_junit_results(ArgV),
    writeln_log_time(prob_finished(NOW)).

cli_init_options(ArgV,RemArgV) :-
    retractall(option(_)),
    %%assert(option(log('/tmp/ProBLog.log'))), print('LOGGING'),nl, %% coment in to build a version of probcli that automatically logs
    ( get_options(ArgV,recognised_option,Options,RemArgV) ->
        assert_all_options(Options)
    ; otherwise ->
        print_error(get_options_failed(ArgV)),definite_error_occured).
cli_show_help(ArgV,RemArgV) :-
    ( (option(help) ; ArgV=[]) ->
        print_help, (RemArgV=[] -> halt_exception ; true)
    ; true).
cli_start_logging(F,NOW,Datime,RemArgV) :-
    debug_print(20,'%logging to: '), debug_println(20,F), set_log_file(F),
    writeln_log(start_logging(NOW,F)),
    version(V1,V2,V3,Suffix), revision(Rev),  lastchangeddate(LCD),
    writeln_log(version(NOW,V1,V2,V3,Suffix,Rev,LCD)),
    prolog_flag(version,PV),
    writeln_log(prolog_version(NOW,PV)),
    findall(Opt, option(Opt), Options),
    writeln_log(options(NOW,Options,RemArgV)),
    datime(Datime,DateRecord),
    writeln_log(date(NOW,DateRecord)).

cli_start_selfcheck :-
    ((b_set_initial_machine, start_animation, perform_self_check) -> true ; true),
    (test_error_occured(_) -> error_occured(selfcheck) ; true).

:- dynamic file_loaded/2.
file_loaded(Status) :- file_loaded(Status,_File).

%cli_load_files(RemArgV,NOW) :- 
%    if_options_set(l_cspm_pl_file(PlFile),cli_l_cspm_pl_file(PlFile)),!,
%    assert(file_loaded(true,PlFile)),
%    cli_process_loaded_file(NOW,PlFile),
%(RemArgV=[] -> true ; cli_load_files2(RemArgV,NOW)).
cli_load_files([],NOW) :- !, 
   ( options_allow_start_without_file ->
       set_animation_mode(b),
       bmachine:b_set_empty_machine,
       assert(file_loaded(true,empty_machine))
     ; otherwise ->
       print('No file to process'),nl),
   writeln_log_time(start_processing_empty_machine(NOW)),
   cli_process_loaded_file(NOW,'$EMPTY_MACHINE'),
   check_all_expected_errors_occured(NOW). % check that all expected errors occured; below they will be checked for each file
cli_load_files(RemArgV,NOW) :- cli_load_files2(RemArgV,NOW).

options_allow_start_without_file :- option(eval_repl).
options_allow_start_without_file :- option(eval_string_or_file(_,_,_,_)).

cli_load_files2([],_) :- !.
cli_load_files2(RemArgV,NOW) :-
    (file_loaded(_) -> clear_loaded_machines_wo_errors ; true),
    retractall(file_loaded(_,_)),
    RemArgV = [MainFile0|Rest],!,
    absolute_file_name(MainFile0,MainFile,[access(none)]), % converts Windows slash into Unix slash,...
    if_option_set(file_info,print_file_info(MainFile)),
    ( file_exists(MainFile) ->
        debug_println(6,file_exists(MainFile)),
        ( load_main_file(MainFile,NOW) ->
            assert(file_loaded(true,MainFile)),
            debug_println(20,'% Processing File'),
            writeln_log_time(start_processing(NOW)),
            (cli_process_loaded_file(NOW,MainFile)
              -> debug_println(20,'% Finished Processing File')
              ;  print_error('### Processing file failed: '), print_error(MainFile)
            ),
            writeln_log_time(finished_processing(NOW)),
            get_errors,
            get_state_space_errors
        ; otherwise ->
            assert(file_loaded(error,MainFile)),
            print_error('### Loading Machine Failed'),
            writeln_log_time(loading_failed(NOW,MainFile)),
            error_occured(load_main_file),
            get_errors
        ),nl
    ; otherwise ->
        nl, assert(file_loaded(error,MainFile)),
        print_error('### Specified File does not exist:'),
        print_error(MainFile),definite_error_occured
    ),
    check_all_expected_errors_occured(NOW), 
    reset_expected_error_occured, % reset for next file
    NOW1 is NOW+1, update_time_stamp(NOW1),
    cli_load_files2(Rest,NOW1).

print_file_info(F) :-
   print('Specification_File('), print(F), print(')'),nl.
   
% ---------------------

% process all the commands for a loaded file:
cli_process_loaded_file(NOW,MainFile) :-
    (real_error_occurred -> print_error('% *** Errors occurred while loading ! ***'),nl,nl ; true),
    get_errors,
    if_option_set(kodkod_performance(KPFile,Iterations),
                  compare_kodkod_performance1(KPFile,Iterations,NOW)),
    if_option_set(kodkod_comparision(MaxResiduePreds),
                  test_kodkod_and_exit(MaxResiduePreds,NOW)),
    % if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)), %% moved to later
    
    if_option_set(zmq_master2(N,Time,Max,StatFile), zmq_start_master(invariant,N,Time,Max,StatFile)),
    if_option_set(zmq_master, zmq_start_master(invariant,200,100,-1,zmq_statistics)),
   
    
    
    % STARTING ANIMATION/MODEL CHECKING
    cli_start_animation(NOW),

    ifm_option_set(pretty_print_internal_rep(PPFILE),
                  b_write_machine_representation_to_file(needed,PPFILE)),
    ifm_option_set(pretty_print_all_types(PPFILE),
                  b_write_machine_representation_to_file(all,PPFILE)),
                  
    if_option_set_loaded(trace_check,trace_check,
                         cli_start_trace_check(MainFile)),
    (initialise_required
      -> check_loaded(initialise),
         cli_start_initialisation(NOW),
         writeln_log_time(initialised(NOW))
      ;  true),
   
    if_option_set(zmq_assertion(StatFile), zmq_start_master(assertion,StatFile)),


    if_option_set_loaded(check_properties,check_properties,
                         cli_check_properties(NOW)),
    ifm_option_set_loaded(check_assertions(ALL,ReqInfos),check_assertions,
                         cli_check_assertions(ALL,ReqInfos,NOW)),
    if_option_set(set_goal(GOAL),
                  cli_set_goal(GOAL)),
    ifm_option_set_loaded(mc(Nr),model_check,
                         cli_start_model_check(Nr,NOW)),
    ifm_option_set_loaded(random_animate(Steps,ErrOnDeadlock),animate,
                         cli_random_animate(NOW,Steps,ErrOnDeadlock)),
    if_option_set_loaded(pa_check,predicate_analysis,
                         test_predicate_analysis),

    cbc_check(NOW),
    ifm_option_set_loaded(mcm_tests(ADepth,AMaxS,ATarget,Output),mcm_test_cases,
                         mcm_test_case_generation(ADepth,AMaxS,ATarget,Output)),
    ifm_option_set_loaded(cbc_tests(ADepth,ATarget,Output),cb_test_cases,
                         cbc_test_case_generation(ADepth,ATarget,Output)),
    ifm_option_set_loaded(test_description(TestDescFile),cb_test_cases,
                         test_generation_by_xml_description(TestDescFile)),
    if_options_set(csp_in_situ_refinement_check(Assertion,Negated),
                  cli_csp_in_situ_refinement_check(Assertion,Negated,NOW)),
    if_options_set(csp_checkAssertion(Assertion,Negated),
                   cli_checkAssertion(Assertion,Negated,NOW)),
    if_options_set(refinement_check(RefFile,PerformSingleFailures,RefNrNodes),
                  cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW)),
    if_options_set(ltl_forumla_model_check(Formula),cli_ltl_model_check(Formula)),                  
    if_options_set(ctl_forumla_model_check(Formula),cli_ctl_model_check(Formula)),                  
    if_options_set(csp_get_assertions,cli_csp_get_assertions),
    if_options_set(set_csp_event_as_visible,cli_set_csp_event_as_visible),
    if_options_set(csp_get_enumerated_assertions,cli_csp_get_enumerated_assertions),
    if_options_set(cspm_evaluate_expression(E),cli_cspm_evaluate_expression(E)),
    if_options_set(eval_csp_expression(E),cli_eval_csp_expression(E)),
    if_options_set(csp_add_unicode(File), cli_csp_add_unicode(File)),
    if_options_set(csp_remove_unicode(File), cli_csp_remove_unicode(File)),
    if_options_set(csp_translate_to_file(PlFile),cli_csp_translate_to_file(PlFile)),
    if_options_set(get_min_max_coverage(Filename),cli_get_min_max_coverage(Filename)),
    if_option_set_loaded(check_goal,check_goal,cli_check_goal),
    if_option_set_loaded(animate,animate,
                         (animate_machine -> true ; true)),
    evaluate_from_commandline,
    if_option_set_loaded(ltl_assertions,check_ltl_assertions,
                         (timeout_call(ltl_check_assertions,NOW) -> true; true)),
    ifm_option_set_loaded(ltl_file(LtlFilename),check_ltl_file,
                         (ltl_check_file(LtlFilename) -> true; true)),
    ifm_option_set_loaded(history(HistoryFilename),history,
                         cli_print_history(HistoryFilename)),
    ifm_option_set_loaded(print_values(ValuesFilename),values,
                         cli_print_values(ValuesFilename)),
    if_options_set(save_state_for_refinement(SaveRefF),
                  refinement_checker:tcltk_save_specification_state_for_refinement(SaveRefF)),
    if_options_set(state_space_dot(DotFile),
                  tcltk_print_states_for_dot(DotFile)),
    if_options_set(state_space_merge_dot(DotFile),
                  print_state_merge_for_dot(DotFile)),
    if_options_set(state_space_dfa_dot(DotFile),
                  print_dot_for_dfa_from_nfa(DotFile)),
    if_options_set(state_as_graph_dot(DotFile),
                  graph_canon:print_cstate_graph(DotFile)),
    if_options_set(enabling_analysis_csv(EnablingCsvFile),
                   do_enabling_analysis_csv(EnablingCsvFile)),
    if_options_set(expect_evaluation(EEvals),
                   check_if_predicates_were_evaluated(EEvals)),
    ifm_option_set(coverage(Nodes,Operations,ShowEnabledInfo),
                  cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW)),
    ifm_option_set(variable_coverage,
                  cli_show_variable_coverage(NOW)),
    ifm_option_set(coverage(ShowEnabledInfo),
                  cli_show_coverage(ShowEnabledInfo,NOW)),
    if_option_set(statistics,
                  cli_print_statistics).

:- use_module(profiler).
cli_print_statistics :- profile_statistics,
    nl,garbage_collect,statistics,nl,
    state_space:state_space_initialise_with_stats.
    
% ---------------------

cli_start_animation(NOW) :-
    file_loaded(true),!,
    debug_println(20,'% Starting Animation'),
		
    writeln_log_time(start_animation(NOW)),
    start_animation_without_computing,
    if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)),

    getAllOperations(Ops),
    debug_print(20,'Operations: '), debug_println(20,Ops),
    
    (we_need_only_static_assertions(ALL)
      -> debug_println(20,'% Projecting on static ASSERTIONS'),
         print('% PROJECTING ON STATIC ASSERTIONS'),nl,
         b_interpreter:set_projection_on_static_assertions(ALL) ; true),
    
    (option(load_state(File))
      -> debug_println(20,'% Loading stored state from file'),
         state_space:tcltk_load_state(File)
       ; computeOperations_for_root_required ->
         debug_println(20,'% Searching for valid initial states'),
         cli_computeOperations(EO),
         debug_println(10,EO)
       ; debug_println(20,'% No initialisation required')
    ).
cli_start_animation(_NOW).

start_animation_without_computing :-
    update_preferences_from_spec,
    set_prefs, % override SET_PREF in DEFINITIONS with values from command-line;
    start_animation,
    if_option_set(plugin(Plugin),
            (plugins:initialise_plugins,plugins:start_plugin_for_current_mode(Plugin))),
    get_errors.
    
cli_random_animate(NOW,Steps,ErrorOnDeadlock) :- 
    writeln_log_time(starting_random_animate(NOW,Steps)),
    start_ms_timer(S),
    perform_random_steps(Steps,ErrorOnDeadlock),
    stop_ms_timer(S,Res),
    %tcltk_save_history_as_trace_file(user),
    debug_println(20,finished_random_animate(Steps,Res)),
    print(finished_random_animate(Steps,Res)),nl,
    writeln_log_time(finished_random_animate(NOW,Steps,ErrorOnDeadlock,Res)).

we_need_only_static_assertions(ALL) :- specfile:b_or_z_mode,
  (option(check_assertions(main,_)) -> ALL=main
   ; option(check_assertions(ALL,_))),
   % we do an assertion check
   \+ ((option(A), option_requires_all_properties(A))),
   b_get_assertions(ALL,dynamic,[]). % the assertions do not reference variables 

option_requires_all_properties(mc(_)). 
option_requires_all_properties(check_properties(_)). % we may want to check all properties
option_requires_all_properties(random_animate(_)).
option_requires_all_properties(trace_check).
option_requires_all_properties(mcm_tests(_,_,_,_)).
option_requires_all_properties(cbc_tests(_,_,_)).
option_requires_all_properties(animate).
option_requires_all_properties(eval_repl).
option_requires_all_properties(eval_string_or_file(_,_,_,_)).
option_requires_all_properties(ltl_assertions).
option_requires_all_properties(ltl_file(_)).
option_requires_all_properties(refinement_check(_,_,_)).
option_requires_all_properties(random_animate(_)).

cli_start_trace_check(MainFile) :-
     debug_println(20,'% Starting Trace Check: '),
     ( check_default_trace_for_specfile(MainFile) -> true
     ; otherwise -> error_occured(trace_check)).

% is it necessary to compute enabled operations for root state
computeOperations_for_root_required :- initialise_required.
computeOperations_for_root_required :- option(trace_check).
computeOperations_for_root_required :- option(ltl_assertions).
computeOperations_for_root_required :- option(random_animate(_,_)).
computeOperations_for_root_required :- option(socket(_)).
computeOperations_for_root_required :- option(mc(_)).
computeOperations_for_root_required :- option(ltl_file(_)).
computeOperations_for_root_required :- option(ltl_forumla_model_check(_)).
computeOperations_for_root_required :- option(ctl_forumla_model_check(_)).
computeOperations_for_root_required :- option(refinement_check(_,_,_)).
computeOperations_for_root_required :- option(csp_in_situ_refinement_check(_,_)).
computeOperations_for_root_required :- option(csp_checkAssertion(_,_)).
computeOperations_for_root_required :- option(mcm_tests(_,_,_,_)).
computeOperations_for_root_required :- option(mcm_cover(_)).
   
% is an initialisation mandatory:
initialise_required :- option(initialise).
initialise_required :- \+ option(trace_check), \+ option(load_state(_)),
   init_req2.
init_req2 :- option(check_properties).
init_req2 :- option(zmq_assertion(_)).
init_req2 :- option(check_assertions(_,_)).
init_req2 :- option(eval_string_or_file(_,_,_,_)). % ensure that we initialise/precompile empty machine in case no main file specified

cli_start_initialisation(NOW) :-
    debug_println(20,'% Performing INITIALISATION: '),
    (perform_random_initialisation -> true ; 
      writeln_log_time(cli_start_initialisation_failed(NOW)),
      fail).
cli_check_properties(NOW) :-
    print('% Checking PROPERTIES: '),nl,
    writeln_log_time(starting_check_properties(NOW)),
    ( current_state_corresponds_to_setup_constants_b_machine ->
        set_analyse_hook('_P'),
        predicate_evaluator:tcltk_analyse_properties(_PROPRES,PROPInfos),
        unset_analyse_hook,
        print(PROPInfos),nl,   % ex: [total/33,true/29,false/0,unknown/4,timeout/4,runtime/49950]
        (predicate_evaluator:check_summary_all_true(PROPInfos) -> true
          ; print_error('### Not all PROPERTIES true'), error_occured(check_properties))
    ; otherwise ->
        (tcltk_debug_properties(PROPRES,false,Satisfiable) ->
         print(PROPRES),nl,
         print(Satisfiable),nl
         ; error_occured(debug_properties_failed))
    ),
    writeln_log_time(finished_check_properties(NOW,PROPInfos)),
    print('% Finished checking PROPERTIES'),nl.
cli_check_assertions(ALL,ReqInfos,NOW) :-
    print('% Checking ASSERTIONS: '),nl,
    writeln_log_time(starting_check_assertions(NOW)),
    set_analyse_hook('_A'), % for dot output, in case users wants to generate dot files for assertions
    predicate_evaluator:tcltk_analyse_assertions(ALL,_ASSRES,Infos),
    unset_analyse_hook,
    print(Infos),nl,
    check_required_infos(ReqInfos,Infos,check_assertions),
    writeln_log_time(finished_check_assertions(NOW,Infos)),
    print('% Finished checking ASSERTIONS'),nl,!.
cli_check_assertions(ALL,ReqInfos,NOW) :-
    add_internal_error('Analyse ASSERTIONS unexpectedly failed',cli_check_assertions(ALL,ReqInfos,NOW)),
    error_occured(internal_error).
cli_set_goal(GOAL) :-
    debug_println(20,set_goal(GOAL)), %print(set_goal(GOAL)), nl,
    bmachine:b_set_machine_goal(GOAL).
cli_check_goal :- 
    print('% Checking GOAL predicate: '),nl, 
    tcltk_analyse_goal(List,Summary),
    debug_println(20,List),
    debug_println(20,Summary),
    check_required_infos([false/0,unknown/0],Summary,check_goal).
cli_start_model_check(Nr,NOW) :-
    debug_print(20,'% Starting Model Check, Bound: '), 
    debug_println(20,Nr),
    writeln_log_time(starting_model_check(NOW,Nr)),
    model_check(Nr,Time,MCRes,NOW),
    nl,
    writeln_log_time(model_check(NOW,Nr,Time,MCRes)).

zmq_start_master(invariant,N,Time,Max,StatFile) :-
    start_animation_without_computing,
    zmq_get_initialisation_term(InitTerm),
    start_master(InitTerm,N,Time,Max,StatFile),
    halt.

zmq_start_master(assertion,StatFile) :-
    current_state_corresponds_to_setup_constants_b_machine,
    animation_mode(b),
    full_b_machine(Machine),
    b_get_assertions(_,static,SAss),
    b_get_assertions(_,dynamic,DAss),
    append(SAss,DAss,Ass),
    current_expression(_,State1),
    specfile:state_corresponds_to_set_up_constants(State1,State),
    start_master(assertions(classical_b(Machine),State,Ass),2,200,-1,StatFile),
    halt.

zmq_get_initialisation_term(Term) :-
    animation_mode(b), \+ animation_minor_mode(eventb), !, % Classical B
    full_b_machine(Machine),
    Term = classical_b(Machine).
zmq_get_initialisation_term(Term) :-
    animation_mode(b), animation_minor_mode(eventb), !, % Event-B
    full_b_machine(Machine),
    Term = eventb(Machine).
zmq_get_initialisation_term(_Term) :-
    animation_mode(Mode),
    add_internal_error('Unsupported formalism for ZMQ', Mode),
    fail.
    
zmq_get_initialisation_term(filename(FN)) :-
    loaded_main_file(FN).

zmq_start_worker(Master,NOW) :-
    start_worker(Master,zmq_worker_load_model),halt_prob(NOW).

zmq_worker_load_model(classical_b(Machine)) :-
    bmachine:(b_machine_reset, assert_machine(Machine)),
    set_animation_mode(b), 
    eclipse_interface:start_animation, cli_computeOperations(_)
    .

zmq_worker_load_model(eventb(Machine)) :-
    bmachine:(b_machine_reset, assert_machine(Machine)),
    set_animation_mode(b), set_animation_minor_mode(eventb),
    eclipse_interface:start_animation, cli_computeOperations(_).

zmq_worker_load_model(filename(FN)) :-
    print('loading file by filename\n'),flush_output,
    ( is_eventb_b(FN) ->  
        eclipse_interface:load_eventb_file(FN)
    ; otherwise ->
        bmachine:b_load_machine_probfile(FN)),
    eclipse_interface:start_animation.

zmq_worker_load_model(assertions(Machine,State,Assertions)) :- 
   assert(assertion_counter(0)),
   print(loaded_model_for_assertion_checking),nl,
   zmq_worker_load_model(Machine),
   assert(worker:assertion_state(State)),
   make_assertionwps(Assertions).
   % assert current state

:-dynamic assertion_counter/1.

make_assertionwps([]).
make_assertionwps([H|T]) :- bsyntaxtree:conjunction_to_list(H,HL), 
                                sort_assertions(HL,SL),
                                append_assertion(SL),
                                make_assertionwps(T).

append_assertion([]).
append_assertion([H|T]) :- assertion_counter(N),
                           retractall(assertion_counter(_)),
                           N1 is N + 1,
                           assert(assertion_counter(N1)),
                           assert(worker:assertion_task(N1,H)),
                           append_assertion(T). 

%assertions_order(A,B) :- term_size(A,NA),term_size(B,NB), NA > NB.
sort_assertions(X,X). 
% :- samsort(assertions_order,X,Y).


zmq_worker_load_model(Other) :-
    add_internal_error('ZMQ worker: Unexpected machine description', zmq_worker_load_model(Other)),
    fail.
is_eventb_b(FN) :- append(_,FN,".eventb").
 % load_model(Initialisation)


load_model(classical_b(M)) :- !,eclipse_interface:load_classical_b(M).
load_model(load_event_b_project(A,B,C,D)) :- !, eclipse_interface:load_event_b_project(A,B,C,D), start_animation.
load_model(X) :- print('*** Unknown model: '), print(X),nl.


:- use_module(predicate_evaluator).
:- use_module(bmachine,[b_machine_name/1]).
set_analyse_hook(AddPrefix) :- % set a hook to write false/unknown expressions into a dot file
   reset_dot_file_number,
   if_options_set(dot_analyse_output_prefix(Path),
                  (loaded_main_file(MainFile),
                   get_modulename_filename(MainFile,RootName),
                   % we could also use b_machine_hierarchy:main_machine_name(RootName)
                   string_concatenate(Path,RootName,P1),
                   string_concatenate(P1,AddPrefix,FullPath),
                   set_dot_file_prefix(FullPath),
                   debug_println(9,dot_file_prefix(FullPath)),
                   register_conjunct_error_hook(user:pred_eval_hook))).
unset_analyse_hook :- predicate_evaluator:reset_conjunct_error_hook.

% Status: true, false, unknown
pred_eval_hook(_Conjunct,true,_EnumWarning,_IsExpanded, _CS) :-
  \+ option(dot_generate_for_all_formulas),!. % don't generate .dot for true formulas, unless explicitly requested
pred_eval_hook(Conjunct,Status,_EnumWarning,_IsExpanded, CS) :-
   print('Generating dotfile for: '),print(CS),nl,
   (write_dot_graph_to_new_file(Status,Conjunct) -> true
     ; add_error(dot_output,'Writing dot to file failed: ',CS)).


:- dynamic dot_file_prefix/1.
:- dynamic dot_file_number/1.
dot_file_prefix('~/Desktop/dot').
set_dot_file_prefix(F) :- retractall(dot_file_prefix(_)), assert(dot_file_prefix(F)).
dot_file_number(0).
reset_dot_file_number :- retractall(dot_file_number(_)), assert(dot_file_number(0)).
get_next_nr(GNr) :- retract(dot_file_number(Nr)), N1 is Nr+1,
  assert(dot_file_number(N1)), GNr = Nr.
write_dot_graph_to_new_file(Status,BExpr) :-
    dot_file_prefix(Dir),get_next_nr(Nr),
    string_concatenate('_',Status,Str1),
    string_concatenate(Nr,Str1,NS),
    string_concatenate(Dir,NS,F1),
    atom_concat(F1,'.dot',FileName),
    user:write_dot_file_for_pred(BExpr,FileName).

:- use_module(refinement_checker).
cli_csp_in_situ_refinement_check(Assertion,Negated,NOW) :-
    debug_println(20,'% Starting CSP Refinement Check'),
    writeln_log_time(starting_csp_refinement_check(NOW)),
    ( timeout_call(user:checkAssertion(Assertion,PP,Negated,_,RefTrace),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    writeln_log_time(finished_csp_refinement_check(PP,NOW)).
cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW) :-
    debug_println(20,'% Starting Refinement Check'),
    writeln_log_time(starting_refinement_check(NOW)),
    refinement_checker:tcltk_load_refine_spec_file(RefFile),
    ( timeout_call(tcltk_refinement_search(RefTrace,PerformSingleFailures,RefNrNodes),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    writeln_log_time(finished_refinement_check(NOW)).
check_ref_result(RefTrace) :- 
   ( RefTrace==no_counter_example ->
            print('==> Refinement Check Successful'),nl
        ; otherwise ->
            print('*** Refinement Check Counter-Example: ***'),nl, print(RefTrace),nl,
            print('*** Refinement Check Failed ***'),nl,
            error_occured(refinement_check_fails)).
cli_checkAssertion(Assertion,Negated,NOW) :-
    debug_println(20,'% Starting CSP Deadlock free Check'),
    writeln_log_time(starting_csp_deadlockFree_check(Assertion,NOW)),
    ( timeout_call(user:checkAssertion(Assertion,PP,Negated,_,ResTrace),NOW)
      -> check_model_result(ResTrace)
      ; otherwise -> true),
    writeln_log_time(finished_csp_deadlockFree_check(PP,NOW)).
check_model_result(ResTrace) :- 
    ( ResTrace==no_counter_example -> 
             print('==> Model Check Successful'),nl
         ;otherwise ->

             print('*** Model Check Counter-Example: ***'),nl,print(ResTrace),nl,
             print('*** Model Check Failed ***'),nl,
             error_occured(model_check_fails)).
cli_csp_get_assertions :-
    loaded_main_file(CSPFile),
    get_csp_assertions_as_string(CSPFile,String),
    print('*** Assertions in File (separated by $) ***'),nl,print(String),nl.
cli_set_csp_event_as_visible :- 
    set_preference(csp_event_visible_to_user,true).
cli_csp_get_enumerated_assertions :-
    loaded_main_file(CSPFile),
    get_csp_assertions_enum(CSPFile,String),
    print('*** Enumerated Assertions in File (separated by $) ***'),nl,print(String),nl.
cli_cspm_evaluate_expression(E) :-
    (loaded_main_file(CSPFile) ->
	      haskell_csp: cspm_eval_expression_from_csp_file(CSPFile, '--src=', E, _ResText, _ErrText)
	    ; haskell_csp: cspm_eval_expression(E,_ResText,_ErrText)
    ).
cli_eval_csp_expression(E) :-
    (loaded_main_file(CSPFile) ->
	      haskell_csp: evaluate_csp_expression(E, CSPFile, Res)
	    ; haskell_csp: evaluate_csp_expression(E,Res)
    ), print('Evaluated Expression: '),nl,print(Res),nl.
cli_csp_add_unicode(File) :- 
    loaded_main_file(CSPFile),
    cspm_file_add_remove_unicode(CSPFile, File,'--addUnicode=').
cli_csp_remove_unicode(File) :- 
    loaded_main_file(CSPFile),
    cspm_file_add_remove_unicode(CSPFile, File,'--removeUnicode=').
cli_csp_translate_to_file(PlFile) :-
    loaded_main_file(CSPFile),parse_and_load_cspm_file_into_specific_pl_file(CSPFile,PlFile).
cli_get_min_max_coverage(Filename) :- 
    pretty_print_min_max_coverage_to_file(Filename).
cli_start_socketserver(Port) :-
    printsilent('Starting Socket Server'),nls,
    absolute_file_name(prob_home('.'),AppDir),
    printsilent('Application Path: '),printsilent(AppDir),nls,
    disable_interaction_on_errors,
    ( start(Port) -> true
    ; otherwise ->
        print('Starting socket server failed, Port: '), print(Port),nl),
    printsilent('Finished Socket Server'),nls.
cli_show_coverage(ShowEnabledInfo,NOW) :-
    cli_show_coverage(_Nodes,_Operations,ShowEnabledInfo,NOW).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    print('Coverage:'),nl,
    compute_the_coverage(Res,TotalNodeSum,TotalTransSum,ShowEnabledInfo),
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    print(Res),nl,
    ( Nodes=TotalNodeSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of nodes: ',TotalNodeSum),
        add_error(probcli,'Expected: ',Nodes),error_occured(coverage)),
    ( Operations=TotalTransSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of transitions: ',TotalTransSum),
        add_error(probcli,'Expected: ',Operations),error_occured(coverage)).
cli_show_variable_coverage(NOW) :-
    print('Variable (min/max) coverage: '),nl,
    coverage_statistics:tcltk_compute_min_max(list(R)),
    print_list(R),
    writeln_log(computed_variable_coverage(NOW,R)).
cli_print_junit_results(ArgV) :-
    junit_mode(S),!,
    statistics(runtime,[E,_]),
    T is E - S,
    create_junit_result(ArgV,'Integration Tests','Integration_Test',T,pass,Result),
    print_junit([Result],'Integration_Test').
cli_print_junit_results(_).

cli_print_history(HistFile) :- 
    findall( O, option(history_option(O)), Options),
    debug_println(9,writing_history_to_file(HistFile)),
    (write_history_to_file(HistFile,Options) -> true ; add_error(history,'Writing history to file failed: ',HistFile)).
    
cli_print_values(ValuesFilename) :- 
    (write_values_to_file(ValuesFilename) -> true ; add_error(sptxt,'Writing values to file failed: ',ValuesFilename)).

:- use_module(library(terms),[term_hash/2]).
:- dynamic probcli_time_stamp/1.
generate_time_stamp(NOW,TS) :- retractall(probcli_time_stamp(_)),
  now(NOW),
  prolog_flag(argv,ArgV),term_hash(ArgV,Hash),
  Rnd is Hash mod 1000,
  % random(0,1000,Rnd),  always returns 216 % TO DO: try to get milliseconds from some library function
  TS is (NOW*1000)+Rnd,
  assert(probcli_time_stamp(TS)).
update_time_stamp(NOW1) :- retractall(probcli_time_stamp(_)),
  assert(probcli_time_stamp(NOW1)).

get_errors :- \+ real_error_occurred,!,
              (get_error(_Source,_Msg) -> print('*** Warnings occured'),nl ; true),
              reset_errors.
get_errors :- get_error_sources. 

get_error_sources :- get_error(ErrSource,_), !, error_occured(ErrSource),
      findall(1,get_error(ErrSource,_),L), length(L,Nr), 
      (Nr>0 -> N1 is Nr+1, print_error('*** Occurences of this error: '),print_error(N1) ; true),
      get_error_sources.
get_error_sources.
 
get_state_space_errors :- option(strict_raise_error),!,
              (invariant_violated(ID) -> print(invariant_violation(ID)),nl,
                                         error_occured(invariant_violation) ; true),
              (state_error(_,_,abort_error(TYPE,_,_,_)) -> error_occured(TYPE) ; true),
              get_state_errors.
get_state_space_errors.

get_state_errors :- state_error(_,_,X), X\=invariant_violated, X\=abort_error(_,_,_,_),
  create_state_error_description(X,Msg),error_occured(Msg),fail.
get_state_errors.

create_state_error_description(eventerror(Event,Error,_),Description) :- !,
    functor(Error,Functor,_),
    ajoin(['event_error:',Event,':',Functor],Description).
create_state_error_description(StateError,Description) :-
    functor(StateError,Functor,_),
    atom_concat('state_error:',Functor,Description).

check_loaded(Action) :-
    ( file_loaded(true)  -> true
    ; file_loaded(error) -> fail /* we have already generated error message */
    ; otherwise ->
        add_error(probcli,'No file specified; cannot perform action: ',Action),
        error_occured(loading),fail).

:- dynamic loaded_main_file/1.

load_main_file(MainFile,NOW) :- retractall(loaded_main_file(_)),
               debug_print(20,'% Loading: '), debug_println(20,MainFile),
               writeln_log_time(loading(NOW,MainFile)),
               get_filename_extension(MainFile,Ext),
               debug_println(6,file_extension(Ext)),
               load_spec_file(Ext,MainFile),
               debug_println(20,'% Finished loading'),
               assert(loaded_main_file(MainFile)).
load_spec_file('pl',MainFile) :- !, load_cspm_spec_from_pl_file(MainFile).
load_spec_file('csp',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('cspm',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('P',MainFile) :- !, load_xtl_file(MainFile).
load_spec_file('p',MainFile) :- !, load_xtl_file(MainFile). % sometimes windows is confused about the upper case letter....
load_spec_file('eventb',MainFile) :- !, load_eventb_file(MainFile).
load_spec_file('v',MainFile) :- !,
    print('---------------------------------'),nl,
    print('Loading: '),print(MainFile),nl,
    load_b_file(MainFile). % Siemens Rule File; maybe in future use -eval_rule_file
load_spec_file('prob',MainFile) :- !,load_prob_file(MainFile).
load_spec_file('mch',MainFile) :- !,load_b_file(MainFile).
load_spec_file('sys',MainFile) :- !,load_b_file(MainFile).
load_spec_file('ref',MainFile) :- !,load_b_file(MainFile).
load_spec_file('imp',MainFile) :- !,load_b_file(MainFile).
load_spec_file('fuzz',MainFile) :- !,tcltk_open_z_file(MainFile).
load_spec_file('tla',MainFile) :- !, load_tla_file(MainFile).
load_spec_file('pml',MainFile) :- !,parsercall:call_promela_parser(MainFile),
   parsercall:promela_prolog_filename(MainFile,PrologFile),
   print(consulting(PrologFile)),nl,
   tcltk_open_promela_file(PrologFile).
load_spec_file(_EXT,MainFile) :- print_error('### Unknown file extension, assuming B machine:'),
                                 print_error(_EXT),
                                 load_b_file(MainFile).


test_kodkod_and_exit(MaxResiduePreds,NOW) :-
    start_animation_without_computing,
    test_kodkod(MaxResiduePreds),
    halt_prob(NOW).

compare_kodkod_performance1(KPFile,Iterations,NOW) :-
    start_animation_without_computing,
    compare_kodkod_performance(KPFile,Iterations),
    halt_prob(NOW).

:- use_module(version).

print_short_version :- 
    version(V1,V2,V3,Suffix),revision(Rev),
    format('VERSION ~p.~p.~p-~p (~p)~N',[V1,V2,V3,Suffix,Rev]).
print_version :- 
    print('ProB Command Line Interface'),nl,
    version(V1,V2,V3,Suffix),
    revision(Rev),  lastchangeddate(LCD), prolog_flag(version,PV), 
    ( option(verbose) ->
        absolute_file_name(prob_home('.'),AppDir),
        format('  VERSION ~p.~p.~p-~p (~p)~N  ~p~N  Prolog: ~p~N  Application Path: ~p~N~N',
           [V1,V2,V3,Suffix,Rev,LCD,PV,AppDir])
     ;  
        absolute_file_name(prob_home('.'),AppDir),
        format('  VERSION ~p.~p.~p-~p (~p)~N  ~p~N  Prolog: ~p~N~N',
           [V1,V2,V3,Suffix,Rev,LCD,PV])
    ).
	  
print_help :- print_version,
          print('Usage: probcli FILE [OPTIONS]'),nl,
          print('  OPTIONS are: '),nl,
          print('     -mc Nr       model check; checking at most Nr states'),nl,
          print('     -noXXX       XXX=dead,inv,goal,ass (for model check)'),nl,
          print('     -bf          proceed breadth-first'),nl,
          print('     -df          proceed depth-first'),nl,
          print('     --timeout N  timeout in ms for model checking and refinement checking'),nl,
          print('     -t           trace check (associated .trace file must exist)'),nl,
          print('     -init        initialise specification'),nl,
          print('     -cbc OPNAME  constraint-based invariant checking for an operation'),nl,
          print('                  (also use OPNAME=all)'),nl,
          print('     -cbc_deadlock  constraint-based deadlock checking'),nl,
          print('                  (also use -cbc_deadlock_pred PRED)'),nl,
          print('     -cbc_assertions  constraint-based static assertion checking'),nl,
          print('     -cbc_refinement  constraint-based static refinement checking'),nl,
          print('     -cbc_sequence S  constraint-based search for sequence of operations'),nl,
          print('     -strict      raise error if model-checking finds counter example'),nl,
          print('                  or trace checking fails'),nl,
          print('     -expcterr ERR  expect error to occur (ERR=cbc,mc,ltl,...)'),nl,
          print('     -expcteval Evals  expect that the given predicates have been evaluated or not (see external predicate IS_EVALUATED)'),nl,
          print('     -animate Nr  random animation (max. Nr steps)'),nl,
          print('     -animate_all random animation until a deadlock is reached'),nl,
          print('     -animate_stats  provide feedback which operations are executed'),nl,
          print('     -his File    write history to File'),nl,
          print('     -his_option O  additional option when writing a history (show_init or show_states)'),nl,
          print('     -sptxt File  save constants and variable values to File'),nl,
          print('     -cache Directory  automatically save constants to files and avoid recomputation'),nl,
          print('     -det_check   check if animation steps are deterministic'),nl,
          print('     -det_constants  only check if SETUP_CONSTANTS step is deterministic'),nl,
          print('     -i           interactive animation. Only for interactive sessions,'),nl,
          print('                  the output can arbitrarily change in fututre versions. '),nl,
          print('                  Do not build automatic tools using the interactive mode'),nl,
          print('     -repl        start interactive read-eval-loop'),nl,
          print('     -eval "E"    evaluate expression or predicate'),nl,
          print('     -c           print coverage statistics'),nl,
          print('     -cc Nr Nr    print and check coverage statistics'),nl,
          print('     -statistics  print memory and other statistics at the end'),nl,
          print('     -p PREF Val  set Preference to Value'),nl,
          print('     -prefs FILE  set Preferences from Prolog File'),nl,
          print('     -card GS Val set scope of B deferred set'),nl,
          print('     -goal "PRED" set GOAL predicate for model checker'),nl,
          print('     -check_goal  check GOAL (after -mc, -t, or -animate)'),nl,
          print('     -s Port      start socket server on given port'),nl,
          print('     -ss          start socket server on port 9000'),nl,
          print('     -sf          start socket server on some free port'),nl,
          print('     -l LogFile   log activities in LogFile'),nl,
          print('     -ll          log activities in /tmp/prob_cli_debug.log'),nl,
          print('     -lg LogFile  analyse logfile using gnuplot'),nl,
          print('     -pp FILE     pretty-print internal representation to file'), nl,
          print('     -ppf FILE    like -pp, but force printing of all type infos'),nl,
          print('     -v           verbose'),nl,
 %%         print('     -vv          very verbose'),nl,  % no effect when debugging calls disabled
          print('     -version     print version information (-svers for short info)'),nl,
          print('     -assertions  check ASSERTIONS'),nl,
          print('     -main_assertions  check ASSERTIONS from main file only'),nl,
          print('     -properties  check PROPERTIES'),nl,
          print('     -cache Dir   use directory "Dir" to cache constants and variables'),nl,
          print('     -rc          runtime checking of types/pre-/post-conditions'),nl,
          print('     -ltlfile F   check LTL formulas in file F'),nl,
          print('     -ltlassertions  check LTL assertions (in DEFINITIONS)'),nl,
          print('     -ltllimit L  explore at most L states when model-checking LTL or CTL'),nl,
          print('     -ltlformula \"F\" check the LTL formula F'),nl,
          print('     -ctlformula \"F\" check the CTL formula F'),nl,
          print('     -save File   save state space for later refinement check'),nl,
          print('     -refchk File refinement check against previous saved state space'),nl,
          print('     -mcm_tests Depth MaxStates EndPredicate File'),nl,
          print('                  generate test cases with maximum length Depth, explore'),nl,
          print('                  maximally MaxStates, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -mcm_cover Operation'),nl,
          print('                  when generating MCM test cases, Operation should be covered'),nl,
          print('     -cbc_tests Depth EndPredicate File'),nl,
          print('                  generate test cases by constraint solving with maximum'),nl,
          print('                  length Depth, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -cbc_cover Operation'),nl,
          print('                  when generating CBC test cases, Operation should be covered'),nl,
          print('     -cbc_cover_all   try and cover all operations'),nl,
          print('     -test_description File'),nl,
          print('                  read information for test generation from File'),nl,
          print('     -spdot File  write graph of the state space to a dot file.'),nl,
          print('     -spmdot File write graph signature merge of the state space to a dot file.'),nl,
          print('     -spddot File  write graph of the state space as a DFA to a dot file.'),nl,
          print('     -dot_output Path generate dot files for false assertions/properties'),nl,
          print('     -dot_all         also generate dot files for true assertions/properties'),nl,
          print('     -load_state File load state from a saved state (currently generated by Tcl/Tk version)'),nl,
	  print('     -test_mode  This flag sets the random seed to the prolog\'s current random state'),nl,
       % For Eclipse Version only   
       %%  print('     -parsercp  CP class path of the B Parser, this has to be a valid Java class path'),nl, 
       %%   print('     -cspm         load CSP-M .csp file rather than B Machine .mch/.ref/.imp File'),nl,
       %%   print('     -csp         load CSP-M .pl file rather than B Machine File'),nl,
          print('     -cspref Spec [m= Impl File'),nl, 
          print('                  checks a refinement statement,'),nl,
          print('                  where Spec and Impl are processes from File, and \'m\' the type of the refinement:'),nl,
          print('                  \'T\' for traces, \'F\' for failures, or \'FD\' for failures-divergences.'),nl,
          print('     -cspdeadlock P m File'),nl,
          print('                  checks a process for deadlock,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -cspdeterminism P m File'),nl, 
          print('                  checks a process for determinism,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -csplivelock P File'),nl, 
          print('                  checks a process for divergence,'),nl,
          print('                  where \'P\' is a process from File.'),nl,
	  print('     -csp_eval "E" evaluate CSP-M expression.'),nl,
	  print('     -cspm_eval "E" evaluate CSP-M expression with the CSPM tool.'),nl,
          print('     -csp-guide File   CSP||B: Use the CSP File to control the B machine'),nl,
          print(' FILE extensions are: '),nl,
          print('      .mch         for B abstract machines'),nl,
          print('      .ref         for B refinement machines'),nl,
          print('      .imp         for B implementation machines'),nl,
          print('      .csp, .cspm  for CSP-M files, same format as FDR'),nl,
          print('      .eventb      for Event-B packages exported from Rodin ProB Plugin'),nl,
          ( option(verbose) ->
              print(' Preferences PREF are: '),nl,
              print_eclipse_prefs
          ; otherwise ->
              print(' Use --help -v to print avalaible preferences PREF'),nl
          ),
          print(' More info at: http://www.stups.uni-duesseldorf.de/ProB/'),nl,
          nl.  
      

set_prefs :- 
   if_option_set(socket(_), % then we may need the event(.) transition_info for the Java API
       preferences:set_preference(store_event_transinfo,true)),
   option(set_prefs_from_file(File)),
   debug_println(20,load_preferences(File)),
   preferences:load_preferences(File),
   fail.
set_prefs :-   
   option(set_pref_group(P,V)),
   debug_println(20,set_pref_group(P,V)),
   set_preference_group(P,V),
   fail.
% eclipse preference or 'normal preference'
set_prefs :-   
   option(set_pref(P,V)),
   debug_println(20,set_pref(P,V)),
   ( eclipse_preference(P,_)
     -> set_eclipse_preference(P,V)
     ; % might be a term if its a plugin preference
     atom_codes(P,Codes),
     append(Codes,".",Codes2), % to make term readable by read_from_codes
     read_from_codes(Codes2,Preference),
     set_preference(Preference,V)
   ),
   fail.
set_prefs :- option(set_card(P,V)),
   debug_println(20,set_card(P,V)),
   set_eclipse_card(P,V),
   fail.
set_prefs :- 
    ( option(breadth_first) -> user:set_depth_breadth_first_mode(breadth_first)
    ; option(depth_first)   -> user:set_depth_breadth_first_mode(depth_first)
    ; otherwise -> true
    ).

:- use_module(library(timeout)).
timeout_call(Call,NOW) :- option(timeout(TO)),!,
   statistics(runtime,[T1,_]),
   time_out(Call,TO,Res),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   print('Runtime: '), print(Runtime), print(' ms'),nl,
   (Res=time_out -> print('*** Timeout occurred: '), print(TO),nl,
                    print('*** Call: '), print(Call),nl,
                    nl,
                    writeln_log(timeout_occured(NOW,Call))
                 ;  true).
timeout_call(Call,_NOW) :- 
   statistics(runtime,[T1,_]),
   call(Call),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   print('Runtime: '), print(Runtime), print(' ms'),nl.

model_check(Nr,Time,ErrRes,NOW) :-
   option(timeout(TO)),!,
   time_out(model_check(Nr,Time,ErrRes),TO,Res),
   (Res=time_out -> print('*** Timeout occurred while Model Checking: '), print(TO),nl,nl,
                    writeln_log(timeout_occured(NOW,model_check(Nr,Time,ErrRes))),
                    Time=time_out(TO), ErrRes=time_out
                 ; true).
model_check(Nr,Time,ErrRes,_NOW) :- model_check(Nr,Time,ErrRes).

:- use_module(model_checker,[model_checking_is_incomplete/6]).

add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations) :- 
   %print(check(model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term))),nl,
   model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term),
   nl,add_warning(model_check_incomplete,Msg,Term),
   fail.
add_model_checking_warnings(_,_,_,_).

model_check(Nr,Time,ErrRes) :-
   statistics(runtime,[T1,_]),
   (option(no_deadlocks) -> FindDeadlocks=0 ; FindDeadlocks=1),
   (option(no_invariant_violations) -> FindInvViolations=0 ; FindInvViolations=1),
   (option(no_goal) -> FindGoal=0 ; FindGoal=1),
   (option(no_assertion_violations)
     -> FindAssViolations=0
      ; FindAssViolations=1
    ),
   (option(with_reduction) -> WithPOR=1; WithPOR=0),
   StopAtFullCoverage=0,
   STOPMCAFTER =  86400000, /* 86400000 = 1 day timeout */
   InspectExistingNodes = 1,
   (user:do_model_check(Nr,NodesAnalysed,STOPMCAFTER,ErrRes,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,WithPOR, InspectExistingNodes)  
      -> (statistics(runtime,[T2,_]),
           Time is T2-T1,
           print('Model Checking Time: '), print(Time), print(' ms'),nl,
           print('States analysed: '), print(NodesAnalysed),nl,
          (ErrRes = no
           -> print('No Counter Example Found, not all nodes visited')
           ;  (ErrRes=all
                -> (tcltk_find_max_reached_node(_)
                     -> print('No Counter Example found. However, not all transitions were computed !')
                     ;  print('No Counter Example found. ALL nodes visited.')
                   ),
                   add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations) 
                ;   print('*** COUNTER EXAMPLE FOUND ***'),nl,
                    debug_println(20,ErrRes),nl,
                    user:translate_error_for_tclk(ErrRes,TclTkRes),
                    print(TclTkRes),nl,
                    (option(silent) -> true
                      ; print('*** TRACE: '),nl,
                        user:tcltk_get_history(list(Hist)),
                        reverse(Hist,Trace),
                        print_list(Trace,1)
                    ),
                    error_occured(TclTkRes)
              )
          )
         )
      ;  ( statistics(runtime,[T2,_]),
           Time is T2-T1,
           print('Model Checking Time: '), print(Time), print(' ms'),nl,
           print_error('*** Model Checking FAILED '),nl,
           ErrRes=fail,
           definite_error_occured)
   ).
cbc_check(NOW) :-
     option(cbc_deadlock_check(DeadlockGoalPred)),
     print('% Starting Constraint-Based DEADLOCK check '),nl,
     writeln_log_time(cbc_deadlock_check(NOW)),
     (tcltk_constraint_find_deadlock_state_with_goal(DeadlockGoalPred,false,Res)
       -> (Res=time_out ->
            print('*** TIME_OUT occurred ***'),nl,
            error_occured(cbc_deadlock_check_time_out)
          ; print('*** DEADLOCK state found ***'),nl,
            error_occured(cbc_deadlock_check),
            print('*** STATE = '),nl,
            current_b_expression(DBState), translate:print_bstate(DBState),nl,
            print('*** END DEADLOCKING STATE '),nl
         )
       ;  print('No DEADLOCK state found'),nl
     ),fail.
cbc_check(NOW) :- 
     option(constraint_based_check(OpName)),OpName\=all,
     check_loaded(constraint_based_check),
     print('% Starting Constraint-Based Check for Operation: '), print(OpName),nl,
     writeln_log_time(cbc_operation_check(NOW,OpName)),
	  (tcltk_constraint_based_check_with_timeout(OpName,Result)
		->  print('% Constraint-Based Check Result: '),nl,
			print(Result),nl,
			(Result=time_out
			  ->  print('*** TIMEOUT OCCURRED ***'),nl, error_occured(cbc)
			  ;  (Result=ok -> print('NO ERRORS FOUND'),nl)
				   ; print('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl,error_occured(cbc) )
		;  add_error(probcli,'ConstraintBasedCheck unexpectedly failed'),
		   (b_get_machine_operation(OpName,_,_,_) -> true
		     ; add_error(probcli,'Unknown Operation Name: ',OpName)),
		   definite_error_occured
	  ),nl,
	  fail.
cbc_check(NOW) :-
     option(constraint_based_check(all)),
     check_loaded(constraint_based_check),
     print('% Starting Constraint-Based Check for all Operations: '),nl,
     writeln_log_time(cbc_all_check(NOW)),
          (tcltk_constraint_based_check(list(Result),ErrorsWereFound)
            ->  print('% Constraint-Based Check Result: '),nl,
                print(Result),nl,
                (ErrorsWereFound=true
                  ->  print('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl, error_occured(cbc)
                  ;  (ErrorsWereFound=false -> print('NO ERRORS FOUND'),nl)
                       ; print('*** TIMEOUT OCCURRED ***'),nl,error_occured(cbc)
                )
            ;  add_internal_error('ConstraintBasedCheck unexpectedly failed. ',cbc_check(NOW)),definite_error_occured
          ),nl,fail.
cbc_check(NOW) :-
     option(cbc_assertions),
     check_loaded(cbc_assertions),
     print('% Starting Constraint-Based static ASSERTIONS check '),nl,
     writeln_log_time(cbc_assertions_start(NOW)),
     (tcltk_constraint_find_static_assertion_violation(Res)
       -> (Res=time_out ->
            print('*** TIME_OUT occured ***'),nl,
            error_occured(cbc_assertions_time_out)
          ; print('*** ASSERTION Violation found ***'),nl,
            error_occured(cbc_assertions),
            print('*** STATE = '),nl,
            current_b_expression(DBState), translate:print_bstate(DBState),nl,
            print('*** END ASSERTION Violation STATE '),nl
         )
       ;  print('No static ASSERTION Violation found'),nl, Res=ok
     ),writeln_log_time(cbc_assertions_finished(NOW,Res)),fail.
cbc_check(NOW) :-
     option(cbc_sequence(Sequence)),
     check_loaded(cbc_sequence),
     print('% Starting Constraint-Based Check for all Sequence: '), print(Sequence),nl,
     writeln_log_time(cbc_sequence(NOW)),
     (tcltk_cbc_find_trace(Sequence,Res)
       -> (Res=ok -> print('Sequence found and executed'),nl 
           ; Res=time_out -> error_occured(cbc_sequence_time_out)
           ; Res=no_solution_found -> print('*** NO SOLUTION FOUND '),error_occured(cbc_sequence_no_solution_found)
           ; print('*** Unknown result: '), print(Res),nl,
             error_occured(cbc_sequence)
          )
       ;  print('*** Internal error: Check failed '), error_occured(cbc_sequence), Res=internal_error
     ),writeln_log_time(cbc_sequence(NOW,Res)),fail.
cbc_check(NOW) :-
     option(cbc_refinement),
     check_loaded(cbc_refinement),
     print('% Starting Constraint-Based static refinement check '),nl,
     writeln_log_time(cbc_refinement_start(NOW)),
     tcltk_cbc_refinement_check(list(Result),ErrorsWereFound),
     print('% Constraint-Based Refinement Check Result: '),nl,print(Result),nl,
     (ErrorsWereFound = time_out -> print('*** TIME_OUT occured ***'),nl,error_occured(cbc_refinement_time_out) ;
      ErrorsWereFound = true -> print('*** Refinement Violation found ***'),nl,error_occured(cbc_refinement) ;
      otherwise -> print('No static Refinement Violation found'),nl
     ),writeln_log_time(cbc_refinement_finished(NOW,ErrorsWereFound)),fail.
cbc_check(_).




if_option_set(Option,Call) :-
    if_option_set(Option,Call,true).
if_option_set(Option,Then,Else) :-
    ( option(Option) -> call(Then)
    ; otherwise      -> call(Else)).
ifm_option_set(Option,Call) :-
    ifm_option_set(Option,Call,true).
ifm_option_set(Option,Then,Else) :- % can perform multiple options
   findall(Then,option(Option),As), 
   (As=[] -> call(Else) ; perform(As)).
perform([]).
perform([A|T]) :- call(A), perform(T).
if_option_set_loaded(Option,Action,Call) :-
    ( (option(Option),check_loaded(Action)) ->
        call(Call)
    ; otherwise -> true).
ifm_option_set_loaded(Option,Action,Call) :- % can perform multiple options
   findall(Call,(option(Option),check_loaded(Action)),As), perform(As).

   

if_options_set(Option,Call) :- % allow multiple solutions for Option
   option(Option),Call,fail.
if_options_set(_,_).

print_options :- print('CLI OPTIONS: '),nl,
                 option(Option), print(Option), nl, fail.
print_options :- nl.

mcm_test_case_generation(ADepth,AMaxStates,ATarget,Output) :-
    arg_is_number(ADepth,MaxDepth),
    arg_is_number(AMaxStates,MaxStates),
    bmachine:b_parse_machine_predicate(ATarget,Target),!,
    get_comma_separated_options(mcm_cover,Events),
    print('mcm test case generation, maximum search depth: '),print(MaxDepth),nl,
    print('mcm test case generation, maximum number of states: '),print(MaxStates),nl,
    print('mcm test case generation, target state predicate: '),print_bexpr(Target),nl,
    print('mcm test case generation, output file: '),print(Output),nl,
    print('mcm test case generation, events to cover: '),print_list(Events),nl,
    explore_and_generate_testcases(Events,Target,MaxDepth,MaxStates,Output,NumTests,Uncovered),
    print('mcm test case generation, generated test cases: '),print(NumTests),nl,
    print('mcm test case generation, uncovered events: '),print_list(Uncovered),nl.
mcm_test_case_generation(_ADepth,_AMaxStates,_ATarget,_Output) :-
    print_error('### MCM Test Case Generation failed'),
    error_occured(mcm_tests).

cbc_test_case_generation(ADepth,ATarget,Output) :-
    arg_is_number(ADepth,MaxDepth),
    bmachine:b_parse_machine_predicate(ATarget,Target),!,
    ( option(cbc_cover_all) -> Events=all ; get_comma_separated_options(cbc_cover,Events)),
    print('constraint based test case generation, maximum search depth: '),print(MaxDepth),nl,
    print('constraint based test case generation, target state predicate: '),print_bexpr(Target),nl,
    print('constraint based test case generation, output file: '),print(Output),nl,
    print('constraint based test case generation, events to cover: '),print_list(Events),nl,
    cbc_gen_test_cases(Events,Target,MaxDepth,Output,Uncovered),
    %print('mcm test case generation, generated test cases: '),print(NumTests),nl,
    print('constraint based test case generation, uncovered events: '),print_list(Uncovered),nl.
cbc_test_case_generation(_ADepth,_ATarget,_Output) :-
    print_error('### CBC Test Case Generation failed'),
    error_occured(cbc_tests).

print_list(all) :- print('** all **').
print_list([]) :- print('** none **').
print_list([H|T]) :- print(H),print(' '),print_list2(T).
print_list2([]).
print_list2([H|T]) :- print(H),print(' '),print_list2(T).

get_comma_separated_options(Option,Selection) :-
    functor(O,Option,1),
    findall(E, (option(O),arg(1,O,CommaSep),
                   split_by_comma(CommaSep,Es),
                   member(E,Es)),
            Selection).

split_by_comma(String,Strings) :-
    atom_chars(String,Chars),
    split_by_comma2(Chars,Strings).
split_by_comma2(Chars,[A|Rest]) :-
    append(AChars,[','|B],Chars),!,
    atom_chars(A,AChars),
    split_by_comma2(B,Rest).
split_by_comma2(Chars,[String]) :- atom_chars(String,Chars).


ltl_check_assertions :-
    (option(ltl_limit(Limit)) -> true; Limit=10000),
    print('LTL Model checking assertions'),nl,
    ltl_check_assertions(Limit,Outcome),
    ( Outcome = pass -> print('LTL check passed'),nl
    ; Outcome = fail -> print('*** LTL check failed'),nl,error_occured(ltl)
    ; Outcome = no_tests -> print('*** No LTL assertions found, test failed'),nl,definite_error_occured
    ; otherwise -> print('*** An error occured in the LTL assertion test'),nl,
	definite_error_occured).

ltl_check_file(Filename) :-
    (option(ltl_limit(Limit)) -> true; Limit=10000),
    print('LTL Model checking assertions'),nl,
    parse_ltlfile(Filename, Formulas),
    ( real_error_occurred ->
        print('An error occurred while parsing the LTL file.\n'), fail
    ; otherwise ->
        ltl_check_formulas(Formulas,Limit)).

ltl_check_formulas([],_) :-
    print('All LTL formulas checked.\n').
ltl_check_formulas([formula(Name,F)|Rest],Limit) :-
    print('Checking formula '),print(Name),print(':\n'),
    ltl_model_check2(F,Limit,init,Status),
    ( Status == no ->
        print('Counter-example found for formula \"'),print(Name),
        print('\", saving trace file.\n'),
        ajoin(['ltlce_', Name, '.trace'], Tracefile),
        tcltk_save_history_as_trace_file(Tracefile),
        add_error(ltl_counterexample,'Counter-example was found')
    ; Status == ok ->
        ltl_check_formulas(Rest,Limit)
    ; Status == incomplete ->
        ajoin(['Model was not completly model-checked, aborted after ',Limit,' new states'],
              Msg),
        add_error(ltl,Msg)
    ; otherwise ->
        ajoin(['Model checker returns unexpected result (',Status,')'],Msg),
        add_error(ltl,Msg)).

cli_ltl_model_check(Formula) :-
    (option(ltl_limit(Max)) -> true; Max=1000),
    print('LTL model checking formula'),nl,
    ltl_model_check(Formula,Max,init,Status),
    check_status(Status,Formula,ltl).

cli_ctl_model_check(Formula) :-
    (option(ltl_limit(Max)) -> true; Max=1000),
    print('CTL model checking formula'),nl,
    ctl_model_check(Formula,Max,init,Status),
    check_status(Status,Formula,ctl).

check_status(ok,Formula,ltl) :- !, % TO DO: make uniform ? CTL returns true; LTL returns ok
    print('LTL Formula '),print('TRUE.'),nl,
    print('No counter example found for '),print(Formula),print('.'),nl.
check_status(true,Formula,ctl) :- (OK=ok ; OK=true),!, 
    print('CTL Formula '),print('TRUE.'),nl,
    print('No counter example found for '),print(Formula),print('.'),nl.
check_status(incomplete,Formula,LTLorCTL) :- !, incomplete_warning(LTLorCTL,Warning),
    add_warning(Warning, 'Warning: Model Check incomplete for: ', Formula),nl,
    print('No counter example found so far for '),print(Formula),print('.'),nl.
check_status(NO,Formula,LTLorCTL) :- (NO=no ; NO=false),!, % TO DO: make uniform
    add_error(LTLorCTL, 'Model Check Counter-Example found for: ', Formula),nl,
    print('Formula '), print('FALSE.'),nl, 
        %print('Counter example found for '),print(Formula),print('.'),nl,
    nl.
check_status(Status,Formula,LTLorCTL) :-
    add_internal_error('Unknown status: ', check_status(Status,Formula,LTLorCTL)).

incomplete_warning(ltl,ltl_incomplete) :- !.
incomplete_warning(ctl,ctl_incomplete) :- !.
incomplete_warning(X,X).

evaluate_from_commandline :-
    retractall(eval_result(_,_)),
    option(eval_string_or_file(A,B,Q,E)), % print(eval(A,B,Q,E)),nl,
    % treat eval_string and eval_file together to ensure proper order of evaluation
    %  (only possible side-effect at the moment: formula can adde new machine_string facts)
    evaluate_string_or_file(A,B,Q,E),
    fail.
evaluate_from_commandline :- print_eval_results,
    % treat -repl option
    if_option_set(eval_repl,
                  (evaluate_expressions -> true ; true)).

:- dynamic eval_result/2.
add_eval_result(R) :- retract(eval_result(R,N)),!,
   N1 is N+1, assert(eval_result(R,N1)).
add_eval_result(R) :- assert(eval_result(R,1)).
print_eval_results :- findall(R/N, eval_result(R,N), L), sort(L,SL),
   (SL=[] -> true ; format('Evaluation results: ~w~n',[SL])).

evaluate_string_or_file(string,String,_,Expected) :-
   print('eval: '), print(String),nl,
    eval_string(String,StringResult,EnumWarning),
    add_eval_result(StringResult),
    (Expected\=StringResult
      -> print_error('### Error, expected result to be: '), print_error(Expected),
         error_occured(eval_string)
      ; EnumWarning = false -> true
      ; nonvar(Expected) -> gen_enum_warning(Expected)
    ).
evaluate_string_or_file(file,File,Quantifier,Expected) :-
    turn_show_error_source_off, % reduce clutter in user feedback; eval_file used in ProB Logic Calculator for example
    (eval_file(File,Quantifier,Result,EnumWarning)
      -> add_eval_result(Result),
         (EnumWarning = false -> true ;  gen_enum_warning(EnumWarning)),
         (Result\=Expected -> print_error('### Error, expected result to be: '),
                              print_error(Expected),
                              print_error('### File: '), print_error(File),
                              error_occured(eval_file)
           ; true)
       ;  add_eval_result(eval_file_failed),
          print_error('### Eval from file failed: '), print_error(File),
          error_occured(eval_file)
    ), turn_show_error_source_on.

gen_enum_warning(time_out) :- !,error_occured(eval_string_time_out).
gen_enum_warning(_) :- print_error('### Enumeration warning occured'),
                       error_occured(eval_string_enum_warning,warning).
evaluate_expressions :- nl,print('ProB Interactive Expression and Predicate Evaluator '), nl, 
                        print('Type "help" for more information.'),nl,
                        turn_show_error_source_off, % reduce clutter in user feedback
                        (option(evaldot(File))
                          -> print('Solutions written to dot file: '), print(File),nl
                          ;  true
                        ),
                        eval,
                        turn_show_error_source_on.
eval :- print('>>> '), read_line(Expr), eval_line(Expr).

:- use_module(eval_strings).

eval_line(end_of_file) :- !.
eval_line([]) :- !.
eval_line("+") :- !, add_last_expression_to_unit_tests, eval.
eval_line("$$") :- !, print_last_expression, eval.
eval_line("$") :- !, print_last_info, eval.
eval_line("@") :- !, get_preference(find_abort_values,OldVal),
  print('Try more aggressively to detect ill-defined expressions: '),
  (OldVal=true -> Val=false ; Val=true), print(Val),nl,
  temporary_set_preference(find_abort_values,Val) , eval.
eval_line("!") :- !, toggle_eval_det,eval.
eval_line([104,101,108,112]) :- !, /* help */
       print('ProB Interactive Expression and Predicate Evaluator '), nl,
       print('Type a valid B expressions or predicates, followed by RETURN or ENTER.'),nl,
       print_sets,
       print('Available CONSTANTS: '),
       bmachine:b_get_machine_constants(Constants),
       bsyntaxtree:get_texpr_ids(Constants,CN),print(CN),nl,
       print('Available VARIABLES: '),
       bmachine:b_get_machine_variables(Variables), 
       bsyntaxtree:get_texpr_ids(Variables,VN),print(VN),nl,
       print('Type + to save last expression to ProB unit tests.'),nl,
       print('Type ! to go to deterministic propagation only mode.'),nl,
       print('Type $ to obtain more information about the last evaluation.'),nl,
       print('Type an empty line to exit.'),nl,
       eval.
eval_line(ExpressionOrPredicate) :- eval_codes(ExpressionOrPredicate,exists,_,_,_,_),eval.


print_sets :- print('Avalable SETS: '), b_global_set(GS), print_set(GS),fail.
print_sets :- nl.

print_set(GS) :- print(GS), \+ is_b_global_constant(GS,_,_),!, print(' ').
print_set(GS) :- print(' = {'), is_b_global_constant(GS,_,Cst), print(Cst), print(' '),fail.
print_set(_) :- print(' } ').

% Simple Animator

animate_machine :- nl,print('IMPORTANT: Do not use this mode for automatic tools.'),nl,
                   print('The output format can change arbitrarily in future versions.'),nl,
                   print('Please terminate your input with a dot (.) and then type return.'),nl,nl,
                   animate_machine2.

animate_machine2 :-
  print_current_state,
  cli_computeOperations(Ops),
  length(Ops,Max),
  print('Enabled Operations: '),nl,
  print_options(Ops,1),
  print(' ==> '),!,
  read(Nr),
  ((number(Nr),Nr>0,Nr=<Max)
    -> cli_executeOperationNr(Nr,Ops)
    ;  fail
  ),!,
  animate_machine2.
animate_machine2.

print_current_state :- getCurrentStateID(CurID), print('ID ==> '), print(CurID),nl,
  getStateValues(CurID,State),
  print_bindings(State),
  ((specfile:b_or_z_mode,\+isInitialisedState(CurID))-> print(' Not yet initialised.') ; true),
  nl.

cli_computeOperations(Ops) :- option(animate_stats),!, % provide statistics about the animation
  statistics(runtime,[Start,_]),nl,
  computeOperations(Ops),
  statistics(runtime,[Stop,_]), Time is Stop - Start, print('Time: '), print(Time), print(' ms'),nl.
cli_computeOperations(Ops) :- computeOperations(Ops).

cli_executeOperationNr(Nr,Options) :- 
   (option(animate_stats)
     -> print('performing: '), nth1(Nr,Options,Action), print(Action),nl
     ; true),
   executeOperationNr(Nr).

perform_random_step :- perform_random_step(_Ops,_Len,_RanChoice).
perform_random_step(Ops,Len,RanChoice) :- 
  cli_computeOperations(Ops),
  length(Ops,Len), Len>0,
  debug_println(20,perform_random_step(Len)),
  L1 is Len+1,
  ((do_det_checking, Len>1)
    ->  print_error('### Non-deterministic step in animate or init'),
        print_error('### State:'),
        print_current_state,  print_error('### Enabled Operations: '), print_options(Ops,1),
        error_occured(det_check)
    ; true),
  random(1,L1,RanChoice),
  debug_println(20,random(L1,RanChoice)),
  cli_executeOperationNr(RanChoice,Ops).
  
do_det_checking :- option(det_check),!.
do_det_checking :- option(det_constants_check),getCurrentStateID(root),
   b_mode, b_machine_has_constants_or_properties.

perform_random_steps(Nr,_) :- \+ number(Nr),!,
   print_error('### Argument to animate not a number'), print_error(Nr),error_occured(animate).
perform_random_steps(Nr,_) :- Nr<0, !,
   print_error('### Argument to animate is a negative number'), print_error(Nr),error_occured(animate).
perform_random_steps(Nr,ErrorOnDeadlock) :- perform_random_steps_aux(0,Nr,ErrorOnDeadlock).

perform_random_steps_aux(Nr,Nr,_) :- !, debug_println(9,performed_random_steps(Nr)).
perform_random_steps_aux(Nr,Max,ErrorOnDeadlock) :-
   (perform_random_step -> N1 is Nr+1, perform_random_steps_aux(N1,Max,ErrorOnDeadlock)
     ;  /* deadlock */
      ErrorOnDeadlock=true ->
       print_error('### Deadlock occured during -animate, at step number:'), print_error(Nr),
       error_occured(animate)
     ; print('% Deadlock occured during -animate, at step number:'), print(Nr),nl
    ).

perform_random_initialisation :- getCurrentStateID(State), State \= root, !,
   print_error('### init can only be used in initial state'), print_error(State),error_occured(initialisation).
perform_random_initialisation :- b_mode, b_machine_has_constants_or_properties,!,
   (perform_random_step(Ops,_Len,RanChoice)
    -> nth1(RanChoice,Ops,Choice), %print(Choice),nl,
       (Choice = 'PARTIAL_SETUP_CONSTANTS'
         -> error_occured(setup_constants_fails)
         ; true)
    ; error_occured(setup_constants_fails),fail), % $setup_constants
   perform_random_init_after_setup_constants. % $initialise_machine
perform_random_initialisation :- (perform_random_step -> true ; error_occured(initialisation_fails),fail).


perform_random_init_after_setup_constants :- \+ option(initialise), we_need_only_static_assertions(_),!,
         print('% NOT INITIALISING MACHINE (not required)'),nl.
 %  debug_println(20,'% NOT INITIALISING MACHINE (not required)').
perform_random_init_after_setup_constants :-
   (perform_random_step -> true ; error_occured(initialisation_fails),fail).

:- use_module(enabling_analysis,[tcltk_cbc_enabling_analysis/1, print_enable_table/1]).
do_enabling_analysis_csv(EnablingCsvFile) :-
    debug_println(9,peforming_enabling_analysis),
    tcltk_cbc_enabling_analysis(list(R)),
    debug_println(9,writing_to_file(EnablingCsvFile)),
    tell(EnablingCsvFile),
    print_enable_table(R),
    told.

print_options([],_).
print_options([H|T],N) :-
  print(' '), print(N), print(':'), print(H),nl,
  N1 is N+1,
  print_options(T,N1).
  
print_list([],_) :- !.
print_list([H|T],N) :- !,
  print(' '), print(N), print(': '),print(H), print(': '),
  (visited_expression(H,State) -> print(State) ; true),nl,
  N1 is N+1,
  print_list(T,N1).
print_list(X,_) :- print('### not a list: '), print(X),nl.
  
print_bindings([]) :- !.
print_bindings([binding(Var,_,PPV)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings([binding(Var,_,PPV,_Tag)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings(X) :- print('### Internal Error: illegal binding list: '), print(X),nl.

:- dynamic expected_error_occured/1.
:- dynamic error_did_not_occur/1.
reset_expected_error_occured :- retractall(expected_error_occured(_)).
check_all_expected_errors_occured(_NOW) :-  get_errors, get_state_space_errors,
  % first assert time_out errors; not caught by other mechanisms at the moment
  ((  state_space:time_out_for_node(_ID)
    ; state_space:time_out_for_invariant(_ID)
    ; state_space:time_out_for_assertions(_ID) )
   -> error_occured(time_out),fail ; fail).
check_all_expected_errors_occured(NOW) :- retractall(error_did_not_occur(_)),
  option(expect_error(Type)),
  \+ expected_error_occured(Type),
  print('*** Expected Error of following type to occur: '), print(Type),nl,
  writeln_log_time(expected_error_did_not_occur(NOW,Type)),
  assert(error_did_not_occur(Type)),
  (option(strict_raise_error) -> definite_error_occured ; fail).
check_all_expected_errors_occured(_NOW) :-
   (option(expect_error(_)) 
    -> (error_did_not_occur(_) -> print('*** Some expected errors did NOT occur !')
         ; print('All expected errors occured.')),nl
    ;  true).

error_occured(warning(Type)) :- !, error_occured(Type,warning).
error_occured(Type) :- error_occured(Type,error).

error_occured(Type,_) :- option(expect_error(Type)),!,
    assert(expected_error_occured(Type)).
error_occured(Type,ErrOrWarning) :-
    (probcli_time_stamp(NOW) -> true ; NOW=unknown),
    writeln_log(error_occured(NOW,Type)),
    (option(strict_raise_error)
     -> print_error('*** Unexpected error occured ***'),
        print_error(Type), 
        definite_error_occured
     ; (ErrOrWarning=error,serious_error(Type))
      -> print_error('*** Serious error occured ***'),
         print_error(Type), 
         definite_error_occured
     ;  print_probcli_error_non_strict(Type,ErrOrWarning)).

% a list of serious errors: if these occur; then return code different from 0 even in non-strict mode
serious_error(get_java_command_path).
serious_error(internal_error(_)).

print_probcli_error_non_strict(parse_machine_predicate_error,_) :-
    !. % have already been reported
print_probcli_error_non_strict(Type,ErrOrWarning) :- 
  (ErrOrWarning=warning -> print_error('*** warning occurred ***')
     ; print_error('*** error occured ***')),
  print_error(Type).

definite_error_occured :- print_error('*** Abnormal termination of probcli !'),
  (file_loaded(_,File) -> print_error('*** for_file'(File)) ; true),
  (probcli_time_stamp(NOW) -> halt1_prob(NOW)
   ; writeln_log(halt(1)),
     halt_exception(1)
  ).

halt_prob(NOW) :-
  writeln_log(prob_halted_prematurely(NOW)),
  check_all_expected_errors_occured(NOW),
  halt_exception.
halt1_prob(NOW) :-
  writeln_log(prob_halted_prematurely(NOW)),
  halt_exception(1).


check_required_infos([],_,_).
check_required_infos([H|T],Infos,ErrType) :-
  (member(H,Infos) -> check_required_infos(T,Infos,ErrType)
     ; print('*** Unexpected result while checking: '),
       translate_err_type(ErrType,ES), print(ES),nl,
       print('***    expected : '), print(H),nl,
       print('***          in : '), print(Infos),nl,
       error_occured(ErrType)).
translate_err_type(check_assertions,'ASSERTIONS') :- !.
translate_err_type(check_goal,'GOAL') :- !.
translate_err_type(X,X).


mc_ok_arg(Arg,X) :- Arg==all,!,prolog_flag(max_tagged_integer,X).
mc_ok_arg(Arg,N) :- tools:arg_is_number(Arg,N).

:- dynamic option/1.
assert_all_options([]).
assert_all_options([Opt|T]) :- assert(option(Opt)), assert_all_options(T).

recognised_option(X,Opt,[],true) :- recognised_option(X,Opt).
recognised_option(X,Opt,Args,true) :- recognised_option(X,Opt,Args).

recognised_option('-mc',mc(N),[Arg],user:mc_ok_arg(Arg,N)).
recognised_option('--model_check',mc(N),[Arg],user:mc_ok_arg(Arg,N)).
recognised_option('--timeout',timeout(N),[Arg],tools:arg_is_number(Arg,N)).
recognised_option('-s',socket(S),[Arg],tools:arg_is_number(Arg,S)).
recognised_option('-cc',coverage(N,N2,false),[Arg,Arg2],(tools:arg_is_number(Arg,N),tools:arg_is_number(Arg2,N2))).
recognised_option('-ltllimit',ltl_limit(Nr),[Arg], tools:arg_is_number(Arg,Nr)).
recognised_option('-aa',check_assertions(all,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (tools:arg_is_number(T,TNr),tools:arg_is_number(F,FNr),tools:arg_is_number(U,UNr))).
recognised_option('-ma',check_assertions(main,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (tools:arg_is_number(T,TNr),tools:arg_is_number(F,FNr),tools:arg_is_number(U,UNr))).
recognised_option('-kodkod_comparision',kodkod_comparision(Nr),[Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-kodkod_performance',kodkod_performance(File,Nr),[File,Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-animate',random_animate(N,true),[Steps],tools:arg_is_number(Steps,N)).
recognised_option('-expcteval',expect_evaluation(List),[Arg],user:evalchecklist(Arg,List)).
recognised_option('-zmq_master2',zmq_master2(N,Time,Max,StatFile),[SN,STime,SMax,StatFile],
                  tools:(arg_is_number(SN,N),arg_is_number(STime,Time),arg_is_integer(SMax,Max))).

recognised_option('-zmq_assertions',zmq_assertion(StatFile),[StatFile]).

recognised_option('-plugin',plugin(Plugin), [Plugin]).

recognised_option('-p',set_pref(PREF,PREFVAL),[PREF,PREFVAL]).
recognised_option('-pref',set_pref(PREF,PREFVAL),[PREF,PREFVAL]).
recognised_option('-pref_group',set_pref_group(Group,Val),[Group,Val]).
recognised_option('--pref',set_pref(PREF,PREFVAL),[PREF,PREFVAL]).
recognised_option('-prefs',set_prefs_from_file(PREFFILE),[PREFFILE]).
recognised_option('--prefs',set_prefs_from_file(PREFFILE),[PREFFILE]).
recognised_option('-card',set_card(SET,SCOPE),[SET,SCOPE]).
recognised_option('--card',set_card(SET,SCOPE),[SET,SCOPE]).
recognised_option('-goal',set_goal(GOAL),[GOAL]).
recognised_option('--goal',set_goal(GOAL),[GOAL]).
recognised_option('--eval',eval_string_or_file(string,E,exists,_),[E]).
recognised_option('-eval',eval_string_or_file(string,E,exists,_),[E]).
recognised_option('--evalt',eval_string_or_file(string,E,exists,'TRUE'),[E]).
recognised_option('-evalt',eval_string_or_file(string,E,exists,'TRUE'),[E]).
recognised_option('--evalf',eval_string_or_file(string,E,exists,'FALSE'),[E]).
recognised_option('-evalf',eval_string_or_file(string,E,exists,'FALSE'),[E]).
recognised_option('--eval_file',eval_string_or_file(file,F,exists,_),[F]).
recognised_option('--evalt_file',eval_string_or_file(file,F,exists,'TRUE'),[F]).
recognised_option('-eval_file',eval_string_or_file(file,F,exists,_),[F]).
recognised_option('-evalt_file',eval_string_or_file(file,F,exists,'TRUE'),[F]).
recognised_option('--eval_rule_file',eval_string_or_file(file,F,forall,_),[F]).
recognised_option('-eval_rule_file',eval_string_or_file(file,F,forall,_),[F]).
recognised_option('-parsercp',(L),[L]).
recognised_option('-l',log(Log),[Log]).
recognised_option('-log',log(Log),[Log]).
recognised_option('--log',log(Log),[Log]).
recognised_option('-lg',analyse_log(Log,timestamp,true),[Log]).
recognised_option('-lgr',analyse_log(Log,revision,false),[Log]).
recognised_option('-check_log',check_log(Log),[Log]).
recognised_option('-expcterr',expect_error(Type),[Type]).
recognised_option('-expecterr',expect_error(Type),[Type]).
recognised_option('-expect',expect_error(Type),[Type]).
recognised_option('-pp',pretty_print_internal_rep(File),[File]).
recognised_option('-ppf',pretty_print_all_types(File),[File]).
recognised_option('-his',history(File),[File]).
recognised_option('-his_option',history_option(Option),[Option]).
recognised_option('-sptxt',print_values(Filename),[Filename]).
recognised_option('-cache',cache_storage(D),[D]).

recognised_option('-MAIN',csp_main(ProcessName),[ProcessName]).
recognised_option('--MAIN',csp_main(ProcessName),[ProcessName]).

recognised_option('-ltlfile',ltl_file(Filename),[Filename]).
recognised_option('-ltlformula',ltl_forumla_model_check(Formula),[Formula]).
recognised_option('-ctlformula',ctl_forumla_model_check(Formula),[Formula]).
recognised_option('-load_state',load_state(Filename),[Filename]).
recognised_option('-save',save_state_for_refinement(Filename),[Filename]).
recognised_option('-refchk',refinement_check(Filename,0,100000),[Filename]).

recognised_option('-cspref',csp_in_situ_refinement_check(assertRef('False',val_of(AbsP1,no_loc_info_available),Type,val_of(ImplP2,no_loc_info_available),no_loc_info_available),'False'),[AbsP1,Type,ImplP2]).
%recognised_option('-cspref',csp_in_situ_refinement_check(AbsP1,Type,ImplP2),[AbsP1,Type,ImplP2]).
% -cspref R [F= Q
recognised_option('-cspdeadlock',csp_checkAssertion(assertModelCheckExt('False',val_of(Proc,src_span),'DeadlockFree',Model),'False'),[Proc,Model]).
% -cspdeadlock R F
recognised_option('-cspdeterminism',csp_checkAssertion(assertModelCheckExt('False',val_of(Proc,src_span),'Deterministic',Model),'False'),[Proc,Model]).
% -cspdeterminism R F
recognised_option('-csplivelock',csp_checkAssertion(assertModelCheck('False',val_of(Proc,src_span),'LivelockFree'),'False'),[Proc]).
% -csplivelock R
recognised_option('-cspm_eval', cspm_evaluate_expression(Expr),[Expr]).
recognised_option('-csp_eval', eval_csp_expression(Expr),[Expr]).
recognised_option('-csp_add_unicode', csp_add_unicode(File), [File]).
recognised_option('-csp_remove_unicode', csp_remove_unicode(File), [File]).
recognised_option('-get_csp_assertions_as_string',csp_get_assertions,[]).
recognised_option('-get_csp_enumerated_assertions_as_string',csp_get_enumerated_assertions,[]).
recognised_option('-prologOut',csp_translate_to_file(PlFile),[PlFile]).
recognised_option('-csp-guide',add_csp_guide(Filename),[Filename]).
recognised_option('--csp_event',set_csp_event_as_visible,[]). % used only with the -csp-guide option
recognised_option('-get_min_max_coverage',get_min_max_coverage(FileName),[FileName]).
recognised_option('-cbc',constraint_based_check(OpName),[OpName]).
recognised_option('-cbc_deadlock',cbc_deadlock_check(true),[]).
recognised_option('-cbc_assertions',cbc_assertions,[]).
recognised_option('-cbc_refinement',cbc_refinement,[]).
recognised_option('-cbc_deadlock_pred',cbc_deadlock_check(GoalPred),[GoalPred]).
recognised_option('-cbc_sequence',cbc_sequence(OpSequence),[OpSequence]).
recognised_option('--comment',comment(UserComment),[UserComment]). % not processed by tool, but will be stored in log-file and used by log_analyser
recognised_option('-junit',junit(Dir),[Dir]).
recognised_option('-mcm_tests', mcm_tests(Depth,MaxStates,EndPred,Output),[Depth,MaxStates,EndPred,Output]).
recognised_option('-mcm_cover', mcm_cover(Event),[Event]).
recognised_option('-cbc_tests', cbc_tests(Depth,EndPred,Output),[Depth,EndPred,Output]).
recognised_option('-cbc_cover', cbc_cover(Event),[Event]).
recognised_option('-cbc_cover_all', cbc_cover_all,[]).
recognised_option('-test_description', test_description(File), [File]).
recognised_option('-spdot',state_space_dot(File),[File]).
recognised_option('-spmdot',state_space_merge_dot(File),[File]).
recognised_option('-spddot',state_space_dfa_dot(File),[File]).
recognised_option('-sgdot',state_as_graph_dot(File),[File]).
recognised_option('-dot_output',dot_analyse_output_prefix(Path),[Path]).
recognised_option('-evaldot',evaldot(File),[File]).
recognised_option('-zmq_master',zmq_master,[]).
recognised_option('-zmq_worker2',zmq_worker2(MasterIP),[MasterIP]).
recognised_option('-enabling_analysis_csv',enabling_analysis_csv(EnablingCsvFile),[EnablingCsvFile]).

recognised_option('-dot_all',dot_generate_for_all_formulas). % generate dot also for true formulas
recognised_option('-animate_all',random_animate(2147483647,false)).
recognised_option('-animate_stats',animate_stats).
recognised_option('--check_goal',check_goal).
recognised_option('-check_goal',check_goal).
recognised_option('-ltlassertions',ltl_assertions).
recognised_option('-assertions',check_assertions(all,[false/0,unknown/0])).
recognised_option('-main_assertions',check_assertions(main,[false/0,unknown/0])).
recognised_option('-properties',check_properties).
recognised_option('-typecheckertests',typechecker_tests).
recognised_option('-selfcheck',selfcheck).
recognised_option('-pacheck',pa_check).
recognised_option('-det_check',det_check). % check if animation is deterministic
recognised_option('-det_constants',det_constants_check). % check if animation for setup_constants is deterministic
recognised_option('-bf',breadth_first).
recognised_option('--breadth',breadth_first).
recognised_option('-df',depth_first).
recognised_option('--depth',depth_first).
recognised_option('-strict',strict_raise_error).
recognised_option('--strict',strict_raise_error).
recognised_option('--silent',silent).
recognised_option('-silent',silent).
recognised_option('-statistics',statistics).
recognised_option('--statistics',statistics).
recognised_option('-nodead',no_deadlocks).
recognised_option('-noinv',no_invariant_violations).
recognised_option('-nogoal',no_goal).
recognised_option('-noass',no_assertion_violations).
recognised_option('-POR',with_reduction).
recognised_option('-i',animate).
recognised_option('-repl',eval_repl). % used to be -eval
recognised_option('--repl',eval_repl).
recognised_option('-c',coverage(false)).
recognised_option('--coverage',coverage(false)).
recognised_option('-vc',variable_coverage).
recognised_option('--variable_coverage',variable_coverage).
recognised_option('-cv',coverage(true)).
recognised_option('-v',verbose).
recognised_option('-vv',very_verbose).
recognised_option('-verbose',verbose).
recognised_option('-very_verbose',very_verbose).
recognised_option('--verbose',verbose).
recognised_option('--very_verbose',very_verbose).
recognised_option('--version',print_version).
recognised_option('-version',print_version).
recognised_option('-svers',print_short_version).
recognised_option('--svers',print_short_version).
recognised_option('-short_version',print_short_version).
recognised_option('--short_version',print_short_version).
recognised_option('--file_info',file_info).
recognised_option('-file_info',file_info).
recognised_option('-t',trace_check).
recognised_option('-init',initialise).
recognised_option('--init',initialise).
recognised_option('--initialise',initialise).
recognised_option('-ll',log('/tmp/prob_cli_debug.log')).
recognised_option('-ss',socket(9000)).
recognised_option('-sf',socket(_)).
recognised_option('--help',help).
recognised_option('-help',help).
recognised_option('-h',help).
recognised_option('-rc',runtimechecking).
recognised_option('-parser_benchmark',parser_benchmark). % temporay option for running parser_benchmarks
recognised_option('-test_mode',test_mode).
%recognised_option('-csp',csp_pl_mode).
%recognised_option('-cspm',csp_mode).

recognised_option('-zmq_worker',zmq_worker).

% convert a string of the form "+x,-y,z" to a list [evaluated(x),not_evaluated(y),evaluated(z)]
evalchecklist(Arg,List) :-
    split_atom(Arg,[',',' '],AList),
    evalchecklist2(AList,List).
evalchecklist2([],[]).
evalchecklist2([Atom|Arest],[Term|Trest]) :-
    atom_chars(Atom,Chars),Chars=[C|Crest],
    evalchecklist3(C,Crest,Term),
    evalchecklist2(Arest,Trest).
evalchecklist3(+,Crest,evaluated(Pred)) :- !,atom_chars(Pred,Crest).
evalchecklist3(-,Crest,not_evaluated(Pred)) :- !,atom_chars(Pred,Crest).
evalchecklist3(C,Crest,evaluated(Pred)) :- atom_chars(Pred,[C|Crest]).

set_random_seed_to_prologs_current_random_state :- 
     getrand(CurrState),setrand(CurrState).

set_new_random_seed :-
   now(TimeStamp), % getting the unix time
   setrand(TimeStamp). % setting new random seed by every execution of probcli

halt_exception :- halt_exception(0).
halt_exception(Code) :- throw(halt(Code)).
    

  
runtime_entry(start) :- go_cli.

save :- save_program('probcli.sav').

