/* a file to start prob_cli with coverage/profiling info enabled (for SICStus 4.2 and later) */

:- prolog_flag(compiling,_,debugcode).
:- prolog_flag(source_info,_,on).
:- prolog_flag(profiling,_,on).

:- include(coverage_term_expander).

:- ensure_loaded(prob_cli).

:- use_module(coverage_tools).
:- use_module(test_runner).
:- use_module(tools,[ajoin/2]).
:- use_module(library(system)).
:- use_module(library(codesio)).
:- use_module(library(lists)).

% use module_info/3 because we are not in a module here
:- module_info(prob_cov,group,coverage_analysis).
:- module_info(prob_cov,description,'ProB start file with coverage tests activated. Includes predicates to execute certain tests. Used by Jenkins.').
:- module_info(prob_cov,revision,'$Rev$').
:- module_info(prob_cov,lastchanged,'$LastChangedDate$').

header(units, 'Unit Tests').
header(laws, 'Law Checking').
header(units-and-laws,'Unit and Law Tests').
header(regression,'Regression Tests').
header(all,'All Tests').
header(data_validation,'Data Validation Tests').
header(last,'Last Test Case').
header(X,Joined) :- atom(X), !, ajoin(['Testcases with Group ', X],Joined).
header(X,Joined) :- is_list(X), maplist(atom,X), !, ajoin(['Testcases with Groups ', X],Joined). 
header(X,Joined) :- ajoin(['Testcase(s) with ID(s) ', X],Joined).
atom_header(Keyword, Header) :-
    header(Keyword,Tests),
    datime(datime(Year,Month,Day,Hour,Min,_Sec)),
    format_to_codes('Coverage Report (~w, generated ~w.~w.~w ~w:~w)', [Tests,Day,Month,Year,Hour,Min], Codes),
    atom_codes(Header, Codes).

generate_report(descriptions,Now) :- !,
    ajoin(['./coverage_reports/html/descriptions/descriptions-',Now,'.html'],FilenameHTML),
    ajoin(['./coverage_reports/tex/descriptions/descriptions-',Now,'.tex'],FilenameTEX),
    gen_pltables_descriptions_table(descriptionstable, [headline('Module Descriptions')],
		[kernel, interpreter]),
    tables_to_html([descriptionstable], FilenameHTML),
    tables_to_latex([descriptionstable], FilenameTEX).
generate_report(Keyword,Now) :-
    run(Keyword),
    atom_header(Keyword,Header),
    ajoin(['./coverage_reports/html/',Keyword,'/',Keyword,'-',Now,'.html'],FilenameHTML),
    ajoin(['./coverage_reports/tex/',Keyword,'/',Keyword,'-',Now,'.html'],FilenameTEX),
    ajoin(['./coverage_reports/xml/',Keyword,'/',Keyword,'-',Now,'.xml'],FilenameXML),
    ajoin(['./coverage_reports/color/',Keyword,'/'],ColorizedPath),
    ajoin(['../../color/',Keyword,'/'],ColorizedURL),
    ajoin(['./coverage_reports/emma/',Keyword,'.xml'],FilenameEmma),
    export(Header,FilenameHTML,FilenameTEX,FilenameXML,ColorizedPath,ColorizedURL,FilenameEmma).

export(Header,FilenameHTML,FilenameTEX,FilenameXML,ColorizedPath,ColorizedURL,EmmaPath) :-
    write('collecting coverage information from sicstus'), nl, full_flush,
    get_file_coverage_info(Info),
    write('collecting coverage information from sicstus done'), nl, full_flush,
    write('colorising source files line by line'), nl, full_flush,
    colorize_source_files(Info, ColorizedPath),
    write('colorising source files line by line done'), nl, full_flush,
    write('generation coverage report tables'), nl, full_flush,
    gen_pltables_coverage_rows(Info, [headline(Header)],[kernel,interpreter],ColorizedURL),
    write('generation coverage report tables done'), nl, full_flush,
    write('exporting tables to html, tex and xml'), nl, full_flush,
    export_tables(FilenameHTML,FilenameTEX,FilenameXML),
    write('exporting tables to html, tex and xml done'), nl, full_flush,
    write('exporting coverage data to emma'), nl, full_flush,
    export_emma(EmmaPath),
    write('exporting coverage data to emma done'), nl, full_flush.

run(units) :- run_tests_by_category(unit).
run(laws) :- run_tests_by_category(laws).
run(units-and-laws) :- run_tests_by_category([unit,laws]).
run(regression) :- run_tests_all_except([unit,laws]).
run(all) :- run_tests_all.
run(data_validation) :- run_tests_by_category(data_validation).
run(last) :- run_last_test.
run(X) :- atom(X), !, run_tests_by_category(X).
run(X) :- is_list(X), maplist(atom,X), !, run_tests_by_category(X).
run(ID) :- id(ID).

% from old prob_cov.pl - now ported to new table module
pci :- print_coverage_info.
pci(X) :- print_coverage_info(X).
hci :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_html_stream(user_output).
hci_file :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_html('out.html').
lci :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_latex_stream(user_output).
lci_file :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_latex('out.tex').
