% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(refinement_checker,[
     % tcltk_save_implementation_state_for_refinement/1,
     tcltk_save_specification_state_for_refinement/1,
     tcltk_load_refine_spec_file/1,
     tcltk_refinement_search/3, in_situ_ref_search/5, in_situ_model_check/5,
     reset_refinement_checker/0,impl_trans/3,
     generate_new_tau_div_collapsed_state/3,
     search_for_divergence/5,ignore_infinite_datatypes/0
     ]).

:- use_module(tools).
:- use_module(module_information,[module_info/2]).
:- module_info(group,csp).
:- module_info(description,'A CSP refinement checker, along with determinism, deadlock and livelock checking.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).

:- use_module(self_check).
:- use_module(debug).
:- use_module(error_manager).
:- use_module(preferences).
:- use_module(translate,[translate_event/2, print_state/1]).

:- use_module(state_space).

:- use_module(tools,[safe_on_exception/3]). 
user_consult_without_redefine_warning(File) :-
    prolog_flag(redefine_warnings, Old, off),
    prolog_flag(single_var_warnings, Old2, off),
    (safe_on_exception(Exc,
        load_files(File,[compilation_mode(assert_all)]),
        (nl,print('Exception occured:'),print(Exc),nl,fail))
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old), 
    prolog_flag(single_var_warnings, _, Old2),
    OK=true.

tcltk_save_specification_state_for_refinement(File) :-
   print('saving XSB Specification state to: '), print(File),nl,
   tell(File), print_specification_state_space_for_refinement, told, print(done),nl.
   
print_specification_state_space_for_refinement :-
    ref_transition(From,Label,To),
    write_term(spec_trans(From,Label,To),[quoted(true)]),
    write('.'),nl,
    fail.
print_specification_state_space_for_refinement :-
    not_all_transitions_added(ID), 
    write_term(spec_not_all_transitions_added(ID),[quoted(true)]),
    write('.'),nl,
    fail.
print_specification_state_space_for_refinement :-
    (max_reached_for_node(ID) ; time_out_for_node(ID)),
    write_term(spec_max_reached_for_node(ID),[quoted(true)]),
    write('.'),nl,
    fail.
print_specification_state_space_for_refinement :-
    (not_all_transitions_added(_) -> true ;
        portray_clause((spec_not_all_transitions_added(_) :- fail))
    ),
    ((max_reached_for_node(_);time_out_for_node(_)) -> true ;
        portray_clause((spec_max_reached_for_node(_) :- fail))
    ),
    ((max_reached_for_node(_) ; time_out_for_node(_); not_all_transitions_added(_))
      -> portray_clause((spec_completely_explored :- fail))
      ;  portray_clause((spec_completely_explored :- true))
    ). 
% TO DO: also generate facts for spec_stable (as we eliminate the tau transitions in ref_transition)    

ref_transition(From,Label,To) :- ref_transition(From,Label,To,[From]).

ref_transition(From,Label,To,_TauList) :-
   transition(From,Action1,To1),
   (functor(Action1,'$setup_constants',_)
      -> ref_transition(To1,Label,To)
      ;   ref_generate_label(Action1,Label1), 
           %(Label1 = tau  % we now now longer expand tau transitions in spec_trans
           %    ->  \+ member(To1,TauList), /* no loop */
           %        ref_transition(To1,Label,To,[To1|TauList])
           %     ; 
                Label=Label1, To=To1
          %)
   ).
    
:- use_module(specfile,[animation_mode/1,csp_mode/0,csp_with_bz_mode/0]).
ref_generate_label(Action,Label) :-
  ((functor(Action,'$initialise_machine',_) ; functor(Action,start_cspm_MAIN,_))
    -> Label = '$initialise_machine'
    ;  ((Action=tau(_),\+animation_mode(b))  
         -> Label=tau
         ;  translate_event(Action,Label))
  ).

% these facts can also be found in the _refine_spec.P files
:- volatile spec_trans/3.
:- dynamic spec_trans/3.
:- dynamic spec_max_reached_for_node/1, spec_not_all_transitions_added/1, spec_completely_explored/0.
reset_refine_spec :-
   retractall(spec_trans(_,_,_)),
   retractall(spec_max_reached_for_node(_)),
   retractall(spec_not_all_transitions_added(_)),
   retractall(spec_completely_explored),
   (csp_with_bz_mode -> print(csp_and_b_mode),nl,
      assert((spec_trans(From,Label,To) :- impl_trans_cspb(From,Label,To)))
   ;  assert((spec_trans(From,Label,To) :- impl_trans(From,Label,To)))).

:- public impl_trans_cspb/3. % asserted above in spec_trans
impl_trans_cspb(From,Label,To) :- 
   impl_trans(From,X,To),translate_csp_b_event(X,X1),
   %print(translated_event(X1)),nl,
   (var(Label) -> Label=X
   ; translate_csp_b_event(Label,Label1),X1==Label1).

%spec_trans(From,Label,To) :- impl_trans(From,Label,To).
%spec_trans(_,_,_) :- fail.
spec_max_reached_for_node(_) :- fail.    
spec_not_all_transitions_added(_) :- fail.    
spec_completely_explored :- fail.
% spec_stable(_).

% summaring of the set closures 
reduce_channel_set(_Actions,closure([]),[]).
reduce_channel_set(Actions,closure([tuple([Channel])|Tail]),List) :-
   expand_symbolic_set(closure([tuple([Channel])]),HdList,_C),
   return_csp_closure_value(HdList,AllChannelEvents),
   list_to_ord_set(AllChannelEvents,HdSet),
   ord_intersection(HdSet,Actions,InterSet),
   ( ord_seteq(InterSet,HdSet) -> 
        reduce_channel_set(Actions,closure(Tail),T),
        ord_union([Channel],T,List)
   ; otherwise ->
        reduce_channel_set(Actions,closure(Tail),T),
        ord_union(InterSet,T,List)
   ).
reduce_channel_set(_Actions,Cl,_L) :- 
   add_internal_error('Internal Error: Type error; expected closure: ',Cl),fail.

/* --------------------------------------------------- */
/* REFINEMENT CHECKING */

%:- table not_refines/3.

:- use_module(probcspsrc(haskell_csp), [evaluate_argument/2,is_not_infinite_type/1]).
:- use_module(probcspsrc(csp_sets), [expand_symbolic_set/3]).
:- use_module(translate, [return_csp_closure_value/2]).

:- dynamic not_refines_table/3,not_refusals_table/3,ignore_infinite_datatypes/0.

not_failure_refines(singleton_failures,X,YList,[not_enabled(A)|_T]) :- % Singleton Failures
      spec_trans_all(YList,A),
      \+(impl_trans(X,A,_)),    /* TO DO: handle tau for CSP */
      print_message(cannot_do(A)).
not_failure_refines(failures,X,YList,[cannot_refuse_compl(AllXActions),cannot_refuse(ReducedRefuseSet)|_T]) :- % CSP Failures
      impl_stable(X), % only need to check if X is stable
      impl_all_possible_actions(X,AllXActions),
      % print(check_fail_ref(X,AllXActions,YList)),nl,
      \+ find_failure_abstraction(YList,AllXActions), % if satisfied => an X action can be refused at the current state of Y
      % note: we can use ord_subset since setof returns sorted lists
      get_refused_set(AllXActions,Closure,RefusedSet),
      reduce_channel_set(RefusedSet,Closure,ReducedRefuseSet),
      print_message(cannot_refuse_compl(AllXActions)),
      print_message(cannot_refuse(ReducedRefuseSet)).

not_failure_refines(failure_divergences,X,YList,T) :- 
      (not_failure_refines(failures,X,YList,T)
       -> true /* not we have already checked for divergence in YList in not_refines */
        ; impl_diverges(X), T=[spec_cannot_diverge|_T]).

:- use_module(library(ordsets)).
find_failure_abstraction([AbsState|_T],AllXActions) :- 
     spec_stable(AbsState),
     spec_all_possible_actions(AbsState,AllAbsActions),
     % note: we can use ord_subset as setof returns sorted lists
     ord_subset(AllAbsActions,AllXActions),!.% we have found an abstract state whose failures is a superset of X's failures
find_failure_abstraction([_|T],AllXActions) :- 
     find_failure_abstraction(T,AllXActions).

get_refused_set(AllXActions,R,RefusedSet) :- 
      evaluate_argument('Events',R),
      expand_symbolic_set(R,R1,_C),
      (csp_with_bz_mode -> 
	l_translate_csp_b_event(AllXActions,AllXActionsNew),
	l_translate_csp_b_event(R1,R2)
      ; otherwise -> 
	AllXActionsNew=AllXActions,
	R2=R1
      ),
      return_csp_closure_value(R2,AllPossibleEvents),
      list_to_ord_set(AllPossibleEvents,AllEventsSet),
      list_to_ord_set(AllXActionsNew,AllXActionsSet),
      ord_subtract(AllEventsSet,AllXActionsSet,RefusedSet).

spec_stable(State) :- 
   get_tau_closure(State,_,stable,_).
 % \+ spec_trans(State,tau,_).
impl_stable(State) :- 
   get_tau_closure(State,_,stable,_).
 % \+ impl_trans(State,tau,_).
spec_possible_action(State,Action) :-
   setof(A,NState^spec_trans(State,A,NState),As), member(Action,As).
spec_all_possible_actions(State,ActionList) :- 
   (setof(A,NState^spec_trans(State,A,NState),ActionList) -> true ; ActionList=[]). 
impl_all_possible_actions(State,ActionList) :- 
   (setof(A,NState^impl_trans(State,A,NState),ActionList) -> true ; ActionList=[]).

spec_trans_all([State|MoreStates],Action) :-
    spec_possible_action(State,Action), 
    spec_trans_all2(MoreStates,Action).
    
spec_trans_all2([],_ANY).
spec_trans_all2([State|T],Action) :-
    spec_trans(State,Action,_),!,
    spec_trans_all2(T,Action).

not_refines(X,Y,Tr,_,_YEnabled,_) :- % print_message(nr(X,Y,Tr)),trace, %%
   % the configuration is already in the memo table: do not look for a counter-example here
   not_refines_table(X,Y,Tr),!,fail.
not_refines(X,YList,TraceX,_,_,_) :- 
      assert(not_refines_table(X,YList,TraceX)), % add to memo table
      YList=[X|_],csp_mode,!,fail. % simple check if X appears in the list; if so we will never find a counter example; we could do a full member check (but not sure about performance)
      % in non csp_mode: root state of impl and spec are different
not_refines(_X,YList,_TraceX,_TraceY,_YEnabled,_) :-
      \+(spec_completely_explored),
      spec_list_contains_unexplored_node(YList),!,  %YEnabled=['$unknown'],
      fail.
      /* we don't know what the spec can do */
      /* to do: improve for max_reached_for_node (if we know max>0) */
not_refines(X,YList,TraceX,_TraceY,YEnabledList,FailuresModel) :-       /* use failure refinement */
      not_failure_refines(FailuresModel,X,YList,TraceX),
      findall(AA,spec_par_trans(YList,AA,_),YEnabledList).
not_refines(X,YList,TraceX,TraceY,YEnabledList,FailuresModel) :-   
      get_tau_closure(X,_Closure,Stable,_Diverges),
      Stable=unstable_prio(Dest),!, % only perform tau priority action
      TraceX = [go(tau,X2)|TX], member(X2,Dest),
      not_refines(X2,YList,TX,TraceY,YEnabledList,FailuresModel).
not_refines(X,YList,TraceX,TraceY,YEnabledList,FailuresModel) :-  
      do_one_trace_step_ahead(X,YList,TraceX,TraceY,YEnabledList,FailuresModel).

do_one_trace_step_ahead(X,YList,TraceX,TraceY,YEnabledList,FailuresModel) :- 
      TraceX = [go(A,X2)|TX],
      impl_trans(X,A,X2),
      (A = tau
        -> not_refines(X2,YList,TX,TraceY,YEnabledList,FailuresModel)
        ;  (setof(Y2,spec_par_trans(YList,A,Y2),YS)
              -> TraceY = [go(A,_)|TY],
                 spec_tau_closure(YS,YTS,FailuresModel),
                 not_refines(X2,YTS,TX,TY,YEnabledList,FailuresModel)
              ;  findall(AA,(spec_par_trans(YList,AA,_),AA\=tau),YListChoices), /* no transition: refinement false */
                 remove_duplicates(YListChoices,YEnabledList) 
           )
      ).

% Refusal traces are of the form <X1,a1,X2,a2,...,Xn> 
% where X1..Xn are the refused sets, a1 -> a2 -> ... -> an is the event trace beginning from the initial state.
% We need to keep track of the whole refusal trace beginning from the initial state 
% in order to prove if the Impl process is a refusal refinement of the Spec process.

% Notion: At each point we must check if Impl has the same refusal trace prefix as Spec in 
% order to continue checking of the refusal trace.

% TODO: Comments, which explain the refusal based refinement algorithm, are missing!!!

not_refusals(X,Y,_TrX,_TrY,RefusalTraceX,_RefusalTraceY,_YEnabled) :- 
   not_refusals_table(X,Y,RefusalTraceX),!,fail.
not_refusals(X,YList,_TrX,_TrY,RefusalTrace,_RefusalTraceY,_YEnabled) :- 
      assert(not_refusals_table(X,YList,RefusalTrace)),
      YList=[X|_],csp_mode,!,fail.
not_refusals(_X,YList,_TraceX,_TraceY,_RefusalTraceX,_RefusalTraceY,_YEnabled) :-
      \+(spec_completely_explored),
      spec_list_contains_unexplored_node(YList),!, 
      fail.
not_refusals(X,YList,TraceX,TraceY,RefusalTraceX,RefusalTraceY,YEnabledList) :- 
    impl_stable(X),
    prefix(TraceX,TraceY),
    \+impl_trans(X,_A,_X2), % we have a deadlock here
    spec_does_not_deadlock(YList), 
    %print('Deadlock in SpecY does not occur => FAILURE!'),nl,
    append(TraceX,[refuse('Sigma')],RefusalTraceX),
    findall(AA,(spec_stable_par_trans(YList,AA,_),AA\=tau),YListChoices),
    remove_duplicates(YListChoices,YEnabledList),
    get_refused_set(YEnabledList,_Closure,RefusedSetY),
    append(TraceY,[refuse(RefusedSetY)],RefusalTraceY).
not_refusals(X,YList,TraceX,TraceY,RefusalTraceX,RefusalTraceY,YEnabledList) :- 
    impl_stable(X),
    prefix(TraceX,TraceY),
    impl_trans(X,A,X2),
    impl_all_possible_actions(X,AllXActions),
    get_refused_set(AllXActions,_Closure,RefusedSet),
    append(TraceX,[refuse(RefusedSet),go(A,X2)],NTX),
    check_go_step_refusal(A,RefusedSet,X2,YList,NTX,TraceY,RefusalTraceX,RefusalTraceY,YEnabledList).
not_refusals(X,YList,TraceX,TraceY,RefusalTraceX,RefusalTraceY,YEnabledList) :- 
    unstable_state_with_enabled_visible_actions(X),
    prefix(TraceX,TraceY),
    %print(unstable_state_with_enabled_visible_actions),nl,
    RefusedSet=bullet, % the bullet set
    impl_non_tau_trans(X,A,X2),
    append(TraceX,[refuse(RefusedSet),go(A,X2)],NTX),
    check_go_step_refusal(A,RefusedSet,X2,YList,NTX,TraceY,RefusalTraceX,RefusalTraceY,YEnabledList).
not_refusals(X,YList,TraceX,TraceY,RefusalTraceX,RefusalTraceY,YEnabledList) :- 
    impl_trans(X,tau,X2),
    append(TraceX,[go(tau_direct,X2)],NTX),
    append(TraceY,[go(tau_direct,_)],NTY),
    not_refusals(X2,YList,NTX,NTY,RefusalTraceX,RefusalTraceY,YEnabledList).

check_go_step_refusal(Action,RefusedSet,NextXState,YList,NTX,TraceY,RefusalTraceX,RefusalTraceY,YEnabledList) :- 
    (setof(Y2,spec_par_trans_bullet(YList,RefusedSet,Action,Y2),YS) -> 
           append(TraceY,[refuse(RefusedSet),go(Action,_)],NTY),
           spec_tau_closure(YS,YTS,failures),
           not_refusals(NextXState,YTS,NTX,NTY,RefusalTraceX,RefusalTraceY,YEnabledList)
    ; otherwise -> 
	   RefusalTraceX=NTX,
           findall(AA,(spec_stable_par_trans(YList,AA,_),AA\=tau),YListChoices),
           remove_duplicates(YListChoices,YEnabledList),
	   get_refused_set(YEnabledList,_Closure,YRefusedSet),
	   append(TraceY,[refuse(YRefusedSet)],RefusalTraceY)
    ).

spec_does_not_deadlock([]).
spec_does_not_deadlock([Y|Ys]) :- 
  (impl_trans(Y,_A,_Y1) -> 
	spec_does_not_deadlock(Ys)
  ; otherwise ->
	fail
  ).

spec_list_contains_unexplored_node([ID|T]) :-  % TO DO: we could do this check when generating YList ?
  (spec_not_all_transitions_added(ID) ;
   spec_max_reached_for_node(ID) ;
   spec_list_contains_unexplored_node(T)).

spec_par_trans_bullet([Y|_],RefusedSetX,A,Y2) :- 
    spec_stable(Y),
    spec_all_possible_actions(Y,AllYActions),
    get_refused_set(AllYActions,_Closure,RefusedSetY),
    special_ord_subset(RefusedSetX,RefusedSetY),
    spec_trans(Y,A,Y2).
spec_par_trans_bullet([Y|_],RefusedSetX,A,Y2) :- 
    unstable_spec_state_with_enabled_visible_actions(Y),
    RefusedSetY=bullet,
    special_ord_subset(RefusedSetX,RefusedSetY),
    spec_trans(Y,A,Y2).
spec_par_trans_bullet([_Y|YT],RefusedSetX,A,Y2) :-
    spec_par_trans_bullet(YT,RefusedSetX,A,Y2).   

special_ord_subset(SubSet,Set) :- 
    %print(special_ord_subset(SubSet,Set)),nl,
    ( SubSet == bullet -> 
	true
    ; Set == bullet ->
	fail
    ; otherwise -> 
	ord_subset(SubSet,Set)
    ).

% in case we have a state with enabled visible and non-visible actions
% e.g. a -> STOP [> b -> STOP
unstable_state_with_enabled_visible_actions(State) :- 
    \+impl_stable(State),
    NState^impl_trans(State,A,NState),
    A\=tau.
impl_non_tau_trans(X,A,X2) :- 
    impl_trans(X,A,X2),
    A\=tau.

unstable_spec_state_with_enabled_visible_actions(State) :- 
    \+spec_stable(State),
    NState^spec_trans(State,A,NState),
    A\=tau.

spec_stable_par_trans([Y|_],A,Y2) :- 
    impl_stable(Y),
    spec_trans(Y,A,Y2).
spec_stable_par_trans([_|YT],A,Y2) :- 
    spec_stable_par_trans(YT,A,Y2).

spec_par_trans([Y|_],A,Y2) :- 
    spec_trans(Y,A,Y2).
spec_par_trans([_|YT],A,Y2) :- 
    spec_par_trans(YT,A,Y2).

%:- block translate_csp_b_event(-,?).
translate_csp_b_event(Event,R) :- 
    ( with_csp_label(Event) -> remove_label_from_event(Event,'CSP',R)
    ; otherwise -> convert_to_csp_event(Event,R) % unified event in B mode (e.g. 'link(a,b)' instead of 'link.a.b')
    ).

l_translate_csp_b_event([],[]).
l_translate_csp_b_event([Event|T],[TEvent|R]) :- translate_csp_b_event(Event,TEvent),l_translate_csp_b_event(T,R).

:- assert_must_succeed((remove_label_from_event('CSP:in','CSP',R),R=='in')).
:- assert_must_succeed((remove_label_from_event('B:in','B',R),R=='in')).

remove_label_from_event(Event,Label,R) :- split_atom(Event,[':'],List),remove(List,Label,RList),ajoin(RList,R).

with_csp_label(Event) :- split_atom(Event,[':'],['CSP'|_]).

convert_to_csp_event(Event,CSPEvent) :- split_atom(Event, ['(',',',')'], DotEls),ajoin_with_sep(DotEls,'.',CSPEvent). 

% compute tau closure of list of abstract nodes; fails if there is divergence in FD mode
% (as no refine check counter-example will be found)
spec_tau_closure(SpecList,TauClosure,FailuresModel) :-
   spec_tau_closure_aux(SpecList,[],TauClosure,no_div,DIV),
   %print(spec_tau_closure(SpecList,DIV,TauClosure, FailuresModel)),nl,
   (FailuresModel = failure_divergences
     -> DIV = no_div % if an abstract node diverges: it can do anything -> fail as no counter-example can be found
     % note: fail after computing; to be able to store information for other checks
     ; /* WARNING: if DIV=div then could it be that not all successor nodes will be in TauClosure ??
           Maybe with internal choice & tau_skip we are safe ??
          if so TO DO : fix */
       true
     ).

spec_tau_closure_aux([],Acc,Acc,D,D).
spec_tau_closure_aux([SpecState|T],Acc,Res,DivSoFar,DIVRes) :-
   get_tau_closure(SpecState,SpecClosure,_,DIV),
   comb_div(DIV,DivSoFar,NewDivSoFar),
   ord_union(SpecClosure,Acc,NewAcc),
   spec_tau_closure_aux(T,NewAcc,Res,NewDivSoFar,DIVRes).

%%%%%%%% DEAD CODE %%%%%%%%%
/*
old_spec_tau_closure(SpecList,Res) :-
  tau_closure2(SpecList,SpecList,Res). %, print(tau_closure(SpecList,Res)),nl.


tau_closure2([],X,X).
tau_closure2([H|T],Acc,Res) :- setof(Succ,new_tau_succ(H,Acc,Succ),Succs),!,
  ord_union(Succs,Acc,NewAcc), tau_closure2(Succs,NewAcc,NewAcc2),
  tau_closure2(T,NewAcc2,Res).
tau_closure2([_|T],Acc,Res) :- tau_closure2(T,Acc,Res).

% find a new tau successor which is not yet in the accumulator list
new_tau_succ(H,Acc,Succ) :- spec_trans(H,tau,Succ), \+ member(Succ,Acc).
% TO DO if Succ in Acc --> return info that there is divergence
*/


impl_trans(From,Label,To) :-  /* need to improve efficiency of that : */
   impl_trans_term(From,Action1,To1),
   (functor(Action1,'$setup_constants',_)
      -> impl_trans(To1,Label,To)
      ;  ref_generate_label(Action1,Label), To=To1  % for in-situ refinement we do not need the overhead of this !
   ).

in_situ_model_check(SpecNodeID,ResTrace,Type,ModelStyle,MaxNrOfNewNodes) :-
        set_max_nr_of_new_impl_trans_nodes(MaxNrOfNewNodes),
        interruptable_perform_mc(Type,ModelStyle,SpecNodeID,ResTrace).

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).
interruptable_perform_mc(Type,ModelStyle,SpecNodeID,ResTrace) :- 
   evaluate_argument('Events',R),(is_not_infinite_type(R) -> true; assert(ignore_infinite_datatypes)),
   user_interruptable_call_det(user:catch_interrupt_assertion_call(refinement_checker: perform_mc(Type,ModelStyle,SpecNodeID,ResTrace)),InterruptResult),
   (InterruptResult=interrupted -> ResTrace=[interrupted], print('Assertion check has been interrupted by user!!!'),nl
   ; print('Assertion check has finished normally.')),retractall(ignore_infinite_datatypes).

perform_mc('Deterministic',ModelStyle,X,T) :- !,deterministic_check(X,ModelStyle,T).
perform_mc('DeadlockFree',ModelStyle,X,T) :- !, deadlock_check(X,ModelStyle,T).
perform_mc('LivelockFree',_ModelStyle,X,T) :- !, divergence_check(X,T).
perform_mc(Style,_ModelStyle,_X,T) :- add_internal_error('Internal Error: Unknown checking style: ',Style), T=none_so_far.

% -----------------------

% DETERMINISM CHECKING

%%%% DEAD CODE %%%%%
/*
deterministic_check_interrupted(X,ModelStyle,ResTr) :- 
   user_interruptable_call_det(user:catch_interrupt_assertion_call(refinement_checker: deterministic_check(X,ModelStyle,ResTrace)),InterruptResult),
   (InterruptResult=interrupted -> ResTr=[interrupted|ResTrace], print("DETERMINISM CHECK Interrupted by user!!!"),nl
   ; print('DETERMINISM CHECK has finished normally.')).
*/

deterministic_check(X,ModelStyle,ResTrace) :- 
  det_check(X,TraceX,TraceY,YEnabledList,ModelStyle),
  (TraceX==no_counter_example
       -> ResTrace=no_counter_example %,print_message('No refinement counter example found')  used to be all return value
   ;   inst_list(TraceX,ResTrace0),
       inst_list(TraceY,ResTrace1), /* convert pending free var into [] + go */
       tcltk_execute_string_trace(X,TraceX),
       append(ResTrace0,[' At_last_step_specification_can_do_one_of:'|YEnabledList],ResTraceX),
       append(ResTraceX,[' Trace_of_the_left_specification:'|ResTrace1],ResTrace)
   ).

det_check(X,TraceX,TraceY,YEnabledList,ModelStyle) :- reset_all_dynamic_state_predicates_for_determinism_check,
   cputime(T1),cputime(T_Tau1),
   (ModelStyle = failure_divergences -> generate_new_tau_div_collapsed_state(X,true,false)
   ;  generate_new_tau_div_collapsed_state(X,false,false)),cputime(T_Tau2),retractall(dvisited(_)),retractall(dtvisited(_)),
   print('------------FINISHED TAU COLLAPSE------------'),nl,
   print('% Checking Time for TAU COLLAPSE Algorithm: '),D_Tau is T_Tau2-T_Tau1,print(D_Tau),print(' ms'),nl,
   ( ModelStyle=failure_divergences,tau_collapse_diverges -> 
     cputime(T2),print_state_from_id(X), print(' diverges'),nl,TraceX=[diverges],TraceY=[diverges]
   ; gensym:gensym('det',RootId),  % was gennum
     get_predeterministic_node_id(X,X1),generate_predeterministic_process(X1,RootId),retractall(tau_visited(_)),
     ( spec_tau_closure([RootId],InitialsRootId,failures),
       not_refines(X,InitialsRootId,TraceX,TraceY,YEnabledList,ModelStyle) ->
        cputime(T2),print_state_from_id(X), print(' reaches a '), print('non deterministic choice'),nl,
        print('% Checking Time: '), D is T2-T1, print(D), print(' ms'),nl,
        print(TraceX)
     ; cputime(T2),print_state_from_id(X),print(' is '),print('deterministic'),nl,TraceX=no_counter_example,
       print('% Checking Time: '), D is T2-T1, print(D), print(' ms'),nl
     )
   ).

reset_all_dynamic_state_predicates_for_determinism_check :- retractall(dvisited(_)), retractall(tau_closure(_,_,_,_)),
          retractall(state_pair(_,_)),retractall(tau_collapse_trans(_,_,_)),retractall(tau_visited(_)),
          retractall(eliminate_state(_,_)),retractall(tau_collapse_diverges),retractall(already_in_closure(_)),retractall(dtvisited(_)).

/* Examples for testing the search_after_each_tau_transition_for_divergence/5 and
            search_for_divergence/5 predicates */

test_transition(2,tau,4).
test_transition(4,b,6).
test_transition(2,tau,3).
test_transition(3,tau,5).
test_transition(3,tau,11).
test_transition(11,v,12).
test_transition(11,tau,13).
test_transition(13,tau,11).
test_transition(13,w,15).
test_transition(11,tau,3).
test_transition(5,tau,3).
test_transition(5,c,8).
test_transition(5,c,9).
test_transition(5,k,10).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_after_each_tau_transition_for_divergence(2,[3,4],[],false,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(2,3,[2,3,5],3,true))).
:- assert_must_fail((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(2,4,_Cl,_D,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_after_each_tau_transition_for_divergence(3,[5,11],[[3,5],[3,11,13]],false,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_after_each_tau_transition_for_divergence(5,[3],[[3,5,11,13]],false,true))).

test_transition(22,tau,23).
test_transition(22,tau,24).
test_transition(23,a,28).
test_transition(23,tau,26).
test_transition(26,tau,27).
test_transition(27,tau,26).
test_transition(27,a,28).
test_transition(24,tau,25).
test_transition(25,tau,30).
test_transition(30,tau,25).
test_transition(30,a,27).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(22,23,[22,23,26,27],26,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(22,24,[22,24,25,30],25,true))).

test_transition(32,tau,34).
test_transition(34,a,36).
test_transition(34,tau,35).
test_transition(35,tau,37).
test_transition(37,tau,35).
test_transition(37,b,40).
test_transition(35,tau,38).
test_transition(38,v,39).
test_transition(38,tau,32).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_after_each_tau_transition_for_divergence(32,[34],[[32,34,35,37,38]],false,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(32,34,[32,34,35,38],32,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(35,37,[35,37],35,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_after_each_tau_transition_for_divergence(102,[103,105],[[102,103,104,107],[102,105]],false,true))).

test_transition(101,start,102).
test_transition(102,tau,103).
test_transition(103,tau,102).
test_transition(103,tau,104).
test_transition(104,tau,103).
test_transition(104,tau,107).
test_transition(107,tau,104).
test_transition(102,tau,105).
test_transition(105,tau,102).
test_transition(105,w,106).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_after_each_tau_transition_for_divergence(102,[103,105],[[102,103,104,107],[102,105]],false,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(102,103,[102,103],102,true))).
:- assert_must_fail((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(105,106,_Cl,_D,true))).
:- assert_must_fail((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(101,102,_Cl,_D,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(102,105,[102,105],102,true))).

test_transition(201,start,202).
test_transition(202,tau,203).
test_transition(203,tau,202).
test_transition(203,tau,204).
test_transition(204,tau,203).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
                                refinement_checker: search_after_each_tau_transition_for_divergence(202,[203],[[202,203,204]],false,true))).
:- assert_must_succeed((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check, 
               refinement_checker: search_for_divergence(202,203,[202,203],202,true))).
:- assert_must_fail((refinement_checker: reset_all_dynamic_state_predicates_for_determinism_check,
               refinement_checker: search_for_divergence(201,202,_Cl,_D,true))).

/* Here begins the implementation for collapsing all strong tau connected components in the state space. */

:- use_module(tools,[flatten/2]).
:- dynamic tau_collapse_trans/3, tau_visited/1,eliminate_state/2,tau_collapse_diverges/0,already_in_closure/1.
generate_new_tau_div_collapsed_state(X,DIV,Test) :- 
   (\+tau_visited(X),\+dvisited(X),impl_trans1(X,_,_,Test) -> 
       findall(Y,impl_trans1(X,tau,Y,Test),L),
       (search_after_each_tau_transition_for_divergence(X,L,Closures,DIV,Test) -> 
            /* add all transitions from tau Closure to the new component if there is livelock */
            add_transitions_to_new_tau_collapsed_state(X,Closures,DIV,Test)
       ;    true
       )
   ;   true % deadlock or visited state, do nothing more.
   ).

/* Checks whether divergence occurres after every tau transition and if so then collects the State Closures
   for every divergence that is possible from X.  */
search_after_each_tau_transition_for_divergence(_X,[],[],_DIV,_Test). % _X is stable state, there is no divergence possible 
search_after_each_tau_transition_for_divergence(X,[YState|TY],Closures,DIV,Test) :-
   /* X\=YState means: ignore every 'state fixed' tau loop */
   (X\=YState,search_for_divergence(X,YState,Closure,DivergesBy,Test),DivergesBy=X ->
	/* search_for_further_divergence/3 predicate is needed for getting the whole divergence space which 
	     is accessible from the X state (i.e. we are looking for the closure of all possible tau livelocks from the X node). */
     (DIV=true -> 
	fail
     ; otherwise -> 
	true
     ),
     (tau_collapse_diverges -> 
	true 
   /* tau_collapse_diverges/0 needed for checking if the process diverges (see det_check/5 implementation) */
     ; otherwise -> 
	assert(tau_collapse_diverges)
     ),
     assert_all_as_already_in_closure(Closure),
     retractall(tau_visited(_)),
     search_for_further_divergence(X,Closure,Closure1,DIV,Test),
     list_to_ord_set(Closure,SetClosure),
     list_to_ord_set(Closure1,SetClosure1),
     ord_union(SetClosure,SetClosure1,UnionClosure),
     /* eliminating the states which will be collapsed through asserting eliminate_state/1 predicates,
        by adding transitions to the collapsed state these states will be not added to the new 'tau-collapse' state space. */
     eliminate_states_from_closure(X,UnionClosure),
     retractall(tau_visited(_)), % clean all tau_visited states
     search_after_each_tau_transition_for_divergence(X,TY,TCl,DIV,Test),
     Closures=[UnionClosure|TCl]
   ; otherwise -> 
	search_after_each_tau_transition_for_divergence(X,TY,Closures,DIV,Test)
   ).

assert_all_as_already_in_closure([]).
assert_all_as_already_in_closure([Hd|Tail]) :- 
   assert(already_in_closure(Hd)),
   assert_all_as_already_in_closure(Tail).

search_for_further_divergence(_X,[],[],_DIV,_Test).
search_for_further_divergence(X,Closure,Closure1,DIV,Test) :- 
   list_to_ord_set(Closure,Set),% remove_duplicates(Set,Set1),
   ord_subtract(Set,[X],Diff),
   search_for_further_divergence1(Diff,Closures,DIV,Test),
   flatten(Closures,Closure1).

search_for_further_divergence1([],[],_DIV,_Test).
search_for_further_divergence1([YState|TY],Closures,DIV,Test) :- 
   findall(Z,(impl_trans1(YState,tau,Z,Test),\+already_in_closure(Z)),L),
   search_after_each_tau_transition_for_divergence(YState,L,Cls,DIV,Test),
   flatten(Cls,FlClosure),
   search_for_further_divergence1(TY,Cls1,DIV,Test),
   Closures=[FlClosure|Cls1].

add_transitions_to_new_tau_collapsed_state(X,[],DIV,Test) :- 
   findall((A,Y),impl_trans1(X,A,Y,Test),Tr),
   assert(dvisited(X)),
   add_tau_collapse_transitions(X,Tr,DIV,Test).
add_transitions_to_new_tau_collapsed_state(X,Closures,DIV,Test) :-
   flatten(Closures,Closure),
   list_to_ord_set(Closure,Set),
   ord_subtract(Set,[X],Diff),
   get_all_transitions_from_tau_closure(Diff,Acts,Test),
   add_tau_collapse_transitions(X,Acts,DIV,Test),
   get_predeterministic_node_id(X,X1),
   (\+tau_collapse_trans(X1,tau,X1) -> 
	assert(tau_collapse_trans(X1,tau,X1))
   ;otherwise -> 
	true
   ),add_transitions_to_new_tau_collapsed_state(X,[],DIV,Test).

get_all_transitions_from_tau_closure([],[],_Test).
get_all_transitions_from_tau_closure([From|States],Result,Test) :-
   findall((A,To),impl_trans1(From,A,To,Test),Acts),get_all_transitions_from_tau_closure(States,Res,Test),
   append(Acts,Res,R),remove_dups(R,Result).

add_tau_collapse_transitions(_X,[],_DIV,_Test).
add_tau_collapse_transitions(X,[(A,Y)|TY],DIV,Test) :- 
   get_predeterministic_node_id(X,X1),
   (Y=X,A=tau -> true
   ; (\+eliminate_state(Y,_) ->
        get_predeterministic_node_id(Y,Y1),
        assert(tau_collapse_trans(X1,A,Y1)),
        retractall(tau_visited(_)),generate_new_tau_div_collapsed_state(Y,DIV,Test)
     ;  (eliminate_state(Y,Z),Z\=X -> 
		       get_predeterministic_node_id(Y,Y1),
		       get_predeterministic_node_id(Z,Z1),
                       assert(tau_collapse_trans(X1,A,Z1))
        ; otherwise -> 
		       true
        )
     )
   ),add_tau_collapse_transitions(X,TY,DIV,Test).

/* eliminating closure states by generating dynamic facts for marking them */
eliminate_states_from_closure(_,[]).
eliminate_states_from_closure(X,[X|Closure]) :- eliminate_states_from_closure(X,Closure).
eliminate_states_from_closure(X,[Y|TY]) :- X\=Y, assert(eliminate_state(Y,X)),
      eliminate_states_from_closure(X,TY).

/* Predicate for looking for divergence in the sub-state/state space of the state */
search_for_divergence(X,Y,Closure,DivergesBy,Test) :- 
   ((\+impl_stable1(Y,Test),impl_trans1(X,tau,Y,Test)),
              (\+dvisited(Y),\+eliminate_state(Y,_)) -> assert(tau_visited(X)), 
         (\+tau_visited(Y) -> 
               impl_trans1(Y,tau,Z,Test),Y\=Z, % avoid fixed point tau loops (else endless loop)
                 ((\+already_in_closure(Z),tau_visited(Z),impl_trans1(Y,tau,Z1,Test),Z1\=Z,Z1\=Y) % lookahead to next not visisted state if Z is visited
                                 -> (search_for_divergence(Y,Z1,Cl,DivergesBy,Test) -> true % already tau visited states have to be skipped (complexity)
                                    ; search_for_divergence(Y,Z,Cl,DivergesBy,Test)
                                    )
                    ;  otherwise -> search_for_divergence(Y,Z,Cl,DivergesBy,Test)
                    ),Closure=[X|Cl] % adding state to divergence closure
         ; Closure=[X],DivergesBy=Y
         ) ; fail ). % break the recursion by stable states and backtrack

impl_trans1(From,Label,To,IsTestCase) :- 
    (IsTestCase -> test_transition(From,Label,To)
    ; otherwise -> impl_trans(From,Label,To) ).

impl_stable1(State,Test) :- \+ impl_trans1(State,tau,_,Test).

/* Here ends the implementation for collapsing all strong tau connected components in the state space. */


/* Here are predicates implemented for the generation of the predeterministic state space from the 'tau collapsed' state space,
   see the implementation of the det_check/5 predicate above. */

:- dynamic state_pair/2, dtvisited/1.
generate_predeterministic_process(X,From) :-
   (\+dtvisited(X),tau_collapse_trans(X,_,_) -> 
       (setof(Y,(tau_collapse_trans(X,tau,Y),X\=Y),S) -> 
           S=[Node|_],
	   get_predeterministic_node_id(Node,To),
           assert(transition(From,tau,_Id,To)),
           assert(dtvisited(X)),
           generate_predeterministic_process(Node,To)
       ;otherwise ->
	   setof((A,Set),setof(Y,tau_collapse_trans(X,A,Y),Set),Set1),
           add_all_visible_transitions(From,Set1,ChoosenSuccs,Succs),
           assert(dtvisited(X)),
           generate_predeterministic_process1(ChoosenSuccs,Succs)
       )
   ; true % no more transitions to be added
   ).

get_predeterministic_node_id(Node,To) :- 
   ( \+dtvisited(Node),\+state_pair(Node,_)
     -> gensym:gensym('det',To), % was gennum
        assert(state_pair(Node,To))
   ;   state_pair(Node,To)).

generate_predeterministic_process1([],[]).
generate_predeterministic_process1([Next|Nexts],[N|Ns]) :- 
   generate_predeterministic_process(Next,N),
   generate_predeterministic_process1(Nexts,Ns).

add_all_visible_transitions(_,[],[],[]).
add_all_visible_transitions(From,[(A,[H|_])|T],[H|ChoosenSuccs],[To|Succs]) :- 
   get_predeterministic_node_id(H,To),
   assert(transition(From,A,_Id,To)),
   add_all_visible_transitions(From,T,ChoosenSuccs,Succs).   

%%% print_added_transition/3 is implemented only for debugging
/*print_added_transition(From,Action,To) :- 
   print('From: '),print(From),print(', Label: '),
   print(Action),print(', To: '),print(To),nl.
*/

/* End of the generation of the predeterministic state-implementation. */

:- dynamic dvisited/1. % true if a node has been seen during dfs traversal

% -----------------------

% DEADLOCK CHECKING


deadlock_check(X,ModelStyle,Trace) :- DLK=true,
  (ModelStyle=failure_divergences -> DIV=true ;
   ModelStyle=trace -> add_error(deadlock_check,'Trace model cannot be used for deadlock checking'), DIV=false
   ; DIV=false),
  dfs_check(X,Trace,DLK,DIV).

:- dynamic dincomplete/0.
dfs_check(X,ResTrace,DLK,DIV) :-
    retractall(dvisited(_)), retractall(tau_closure(_,_,_,_)),retractall(dincomplete),
     cputime(T1),
   (dfs_search(X,DLK,DIV,Trace1) ->
      cputime(T2),print_state_from_id(X), print(' reaches a '), printDD(DLK,DIV),nl,
      print(Trace1),nl,inst_list(Trace1,ResTrace),
      tcltk_execute_string_trace(X,Trace1)
    ; dincomplete -> cputime(T2),print('No '),printDD(DLK,DIV), print('found so far for :'),
                     print_state_from_id(X),nl, ResTrace=none_so_far
    ; cputime(T2),print_state_from_id(X), print(' is '),printDD(DLK,DIV), print('free'),nl,ResTrace=no_counter_example
    ),
     print('% Checking Time: '), D is T2-T1, print(D), print(' ms'),nl.

dfs_search(X,_,_DIV,_) :- dvisited(X),!, %print(already_visited(X)),nl,
%  (DIV\=true, TO DO: check that we have not ignored tau or visible transitions in this loop)
                       fail.
%dfs_search(X,_,_DIV,_) :- % comment this clause in if you do not want omega to count as deadlock; FDR does !
%   visited_expression(X,omega),!, assert(dvisited(X)),fail. % this is a termination node; do not look for deadlock or other errors from it
dfs_search(X,DLK,_,[deadlocks]) :- DLK=true,\+ impl_trans_term(X,_,_),!, % will compute transitions if necessary
  (not_all_transitions_added(X) -> (dincomplete -> assert(dincomplete),fail) ; true).
dfs_search(X,DLK,DIV,Result) :-
   get_tau_closure(X,Closure,Stable,Diverges),
   assert(dvisited(X)),
   ((DIV=true, Diverges=div)
     -> Result = [diverges], print(diverges(X,Closure)),nl
     ;  Stable=unstable_prio(Dest) -> /* only perform the first priority tau transitions */
         Result=[go(tau,Y)|R2],
         member(Y,Dest), dfs_search(Y,DLK,DIV,R2)
     ;  % Stable=[X] -> 
        Result=[go(A,Y)|R2], impl_trans(X,A,Y), dfs_search(Y,DLK,DIV,R2)
    % ;  Result=[go(tau,Y)|R2], member(Y,Stable), dfs_search(Y,DLK,DIV,R2)
    ).


printDD(DLK,DIV) :- (DLK=true -> print('deadlock ') ; true),
  (DIV=true -> print('divergence ') ; true).

% -----------------------

% DIVERGENCE CHECKING

divergence_check(X,Trace) :- DLK=false,DIV=true,dfs_check(X,Trace,DLK,DIV).


:- dynamic tau_visited/2,tau_closure/4, tau_loop_back_node/1.

impl_diverges(State) :- get_tau_closure(State,_,_,div).

% we could provide an optimized version of the code below for just checking divergence
% this would be sufficient in FD mode for refinement checking and in deadlock/livelock checking
get_tau_closure(State,Closure,Stable,DIV) :- tau_closure(State,C,S,D),!,Closure=C, Stable=S,DIV=D.
get_tau_closure(State,Closure,Stable,DIV) :- retractall(tau_visited(_,_)),
  compute_tau_closure(State,Closure,Stable,DIV).

get_tau_closure2(State,Closure,Stable,DIV) :- tau_closure(State,C,S,D),!,Closure=C, Stable=S,DIV=D.
get_tau_closure2(State,Closure,Stable,DIV) :- compute_tau_closure(State,Closure,Stable,DIV).

compute_tau_closure(State,Closure,Stable,DIV) :-
   (tau_priority_transition(State,Span,_)
     -> Prio = prio, findall(To,tau_priority_transition(State,Span,To),Dest),
        assert(tau_visited(State,Dest)) % remember that we did not inspect all tau successors, only Dest
     ;  Prio = all,  findall(To,tau_transition(State,_,To),Dest),
        assert(tau_visited(State,[]))
   ),
   %print(visiting(State,Dest,Prio)),nl,
   (Dest=[]
     -> /* stable state */
        DivOccured = no_div, TauClosure=[State],
        DStable = stable
     ; /* unstable state */
       sort(Dest,SDest),
       comp_dest_tau_clos(SDest,[],DestTauClosure,no_div,DivOccured),
       (((DivOccured=no_div ; \+ tau_loop_back_node(State)),
          Prio=prio)
          -> DStable = unstable_prio(SDest),  % it is safe to only treat the priority tau transitions in this node
             TauClosure = DestTauClosure % also: we do not need to add this node to the tau-closure
          ;  DStable = unstable,
             ord_union([State],DestTauClosure,TauClosure) % also add node itself
       )
   ),
   %print(result(State,TauClosure,DStable,DivOccured)),nl, %
   assert(tau_closure(State,TauClosure,DStable,DivOccured)),
   Closure=TauClosure, DIV=DivOccured, Stable=DStable.

comp_dest_tau_clos([],CLOS,CLOS,DIV,DIV).
comp_dest_tau_clos([Dest|T],CLOS_sofar,CLOS_RES,DIV_sofar,DIVRES) :-
    ((tau_visited(Dest,PrioList), /* we have seen the node before */
      \+ tau_closure(Dest,_,_,_)) /* its treatment is not complete; i.e., we have a loop */
     -> /* we have divergence */
        % print(tau_loop(Dest,PrioList)),nl,
        % ord_union([Dest],CLOS_sofar,CLOS_RES)   % Dest already in closure
        DIV_sofar2=div,
        (PrioList=[] -> CLOS_sofar2=CLOS_sofar
          ;  /* we did not add all successors of Dest yet */
             retract(tau_visited(Dest,_)),
             assert(tau_visited(Dest,[])),
             assert(tau_loop_back_node(Dest)),
             findall(To,tau_transition(Dest,_,To),AllDestSuccs),
             sort([To|AllDestSuccs],SADS), % also add To; it was not yet added
             ord_subtract(SADS,PrioList,DestIgnored),
             % print(adding_ignored_succs(Dest,DestIgnored)),nl,
             comp_dest_tau_clos(DestIgnored,CLOS_sofar,CLOS_sofar2,div,_)
        ),
        comp_dest_tau_clos(T,CLOS_sofar2,CLOS_RES,DIV_sofar2,DIVRES)
      ; get_tau_closure2(Dest,DClos,_DStable,DIV),
        comb_div(DIV,DIV_sofar,DIV_sofar2),
        ord_union(CLOS_sofar,DClos,CLOS_sofar2), % combine states reachable via tau
        comp_dest_tau_clos(T,CLOS_sofar2,CLOS_RES,DIV_sofar2,DIVRES)).

comb_div(div,_,div).
comb_div(no_div,D,D).

%%% size_of_tables/0 have only debugging purpose
/*
:- public size_of_tables/0.
size_of_tables :- print_size_of_table(refinement_checker:not_refines_table/3).
*/

reset_refinement_checker :- retractall(tau_visited(_,_)),retractall(tau_loop_back_node(_)),
    retractall(not_refines_table(_,_,_)),
    retractall(dvisited(_)), retractall(tau_closure(_,_,_,_)),retractall(dincomplete),
    reset_refine_spec.
cputime(T) :- statistics(runtime,[T,_]).  

refine_check(X,SpecY,TraceX,TraceY,YEnabledList,FailuresModel) :- 
           retractall(not_refines_table(_,_,_)),
           retractall(not_refusals_table(_,_,_)),
           % ( (not_refines(X,InitialsY,TraceX,YEnabledList,FailuresModel) , TraceY=TraceX)
           (csp_mode -> 
		evaluate_argument('Events',R),
		/* Needed for the Trace Debugger of the Refinement Checker. 
		   It prohibits to expand an infinite data type structures when a counter example has been found.
		   TODO: Consider a more sophisticated solution. This code fragment adds unnecessary overhead to every assertion check.*/
                (is_not_infinite_type(R) -> 
			true
		; otherwise -> 
			assert(ignore_infinite_datatypes))
           ; otherwise -> 
		true),
           ((spec_tau_closure([SpecY],InitialsY,FailuresModel),
             not_refines_or_refusals(X,InitialsY,TraceX,TraceY,YEnabledList,FailuresModel))
               ->
                nl,print_state_from_id(X), print(' is *not* a '), print_ref(FailuresModel),
                print_state_from_id(SpecY),nl,
                (debug_mode(off) -> true
                 ; print('Trace of '), print_state_from_id(X),nl,print(TraceX),nl,
                   print('Trace of '), print_state_from_id(SpecY),nl,print(TraceY),nl,
                   print('Events enabled at last step of '),print_state_from_id(SpecY),nl,
                   print(YEnabledList),nl%,print('Enabled Processes: '),nl
                 )
               ;
                print_state_from_id(X), print(' is a '), print_ref(FailuresModel),
                print_state_from_id(SpecY),nl, TraceX = no_counter_example
             ),retractall(ignore_infinite_datatypes).

not_refines_or_refusals(X,InitialsY,TraceX,TraceY,YEnabledList,FailuresModel) :-
   (FailuresModel == refusals ->
	not_refusals(X,InitialsY,[],[],RefusalTraceX,RefusalTraceY,YEnabledList),
        TraceX=RefusalTraceX,
        TraceY=RefusalTraceY
   ; otherwise ->
	not_refines(X,InitialsY,TraceX,TraceY,YEnabledList,FailuresModel)
   ).
	
print_state_from_id(ID) :- visited_expression(ID,State),
  print_state(State).
print_ref(FailuresModel) :- print(FailuresModel), print(' refinement of ').

tcltk_load_refine_spec_file(SpecFile) :- 
                 print_message(loading_refine_spec(SpecFile)),
                 user_consult_without_redefine_warning(SpecFile).

tcltk_refinement_search(ResTrace,FailuresModel,MaxNrOfNewNodes) :-
     refinement_search(root,root,ResTrace,FailuresModel,MaxNrOfNewNodes).

%  A Procedure to do in_situ_refinement search: impl_trans & spec_trans are represented in the same state_space
in_situ_ref_search(ImplNode,SpecNode,ResTrace,FailuresModel,MaxNrOfNewNodes) :-
  % TO DO: redirect spec_trans,... to impl_trans ....
  % retractall(spec_trans(_,_,_)),
  % assert((spec_trans(From,Label,To) :- impl_trans(From,Label,To))),
   interruptable_refinement_search(ImplNode,SpecNode,ResTrace,FailuresModel,MaxNrOfNewNodes).

interruptable_refinement_search(ImplNode,SpecNode,ResTrace,FailuresModel,MaxNrOfNewNodes) :- 
          user_interruptable_call_det(user:catch_interrupt_assertion_call(refinement_checker: refinement_search(ImplNode,SpecNode,ResTrace,FailuresModel,MaxNrOfNewNodes)),InterruptResult),
   (InterruptResult=interrupted -> ResTrace=[interrupted], print("Refinement check has been interrupted by user!!!"),nl
   ; (real_error_occurred -> print_error('% *** Errors occurred while refinement checking ! ***'),nl,nl,fail
     ; print('Refinement check has finished normally.'),nl
     )
   ).

:- use_module(debug).
%:- use_module(probcspsrc(haskell_csp),[channel_type_list/2]).
refinement_search(ImplNode,SpecNode,ResTrace,FailuresModel,MaxNrOfNewNodes) :-
     debug_println(9,refinement_search(ImplNode,SpecNode,ResTrace,FailuresModel,MaxNrOfNewNodes)),
     set_max_nr_of_new_impl_trans_nodes(MaxNrOfNewNodes),
     cputime(T1),
     refine_check(ImplNode,SpecNode,TraceX,TraceY,YEnabledList,FailuresModel),
     cputime(T2),
     print('% Refinement Check Time: '), D is T2-T1, print(D), print(' ms'),nl,
     (TraceX==no_counter_example
	  -> ResTrace=no_counter_example %,print_message('No refinement counter example found')  used to be all return value
      ;   inst_list(TraceX,ResTrace0),
          inst_list(TraceY,ResTrace1), /* convert pending free var into [] + go */
	      tcltk_execute_string_trace(ImplNode,TraceX),
		  append(ResTrace0,[' At_last_step_specification_can_do_one_of:'|YEnabledList],ResTraceX),
		  append(ResTraceX,[' Trace_of_the_left_specification:'|ResTrace1],ResTrace)
		  %,print('Result trace: '),print(ResTrace),nl
	 ). %size_of_tables.

inst_list([],R) :- !,R=[].
inst_list([go(tau_direct,_ID)|T],['GO:',tau|IT]) :- !, inst_list(T,IT). % NOTE : can be multiple taus now
inst_list([go(Op,_ID)|T],['GO:',Op|IT]) :- csp_mode,!, inst_list(T,IT).
inst_list([go(Op,_ID)|T],[Op|IT]) :- !, inst_list(T,IT).
inst_list([spec_cannot_diverge|T],['DIVERGES'|IT]) :- !, inst_list(T,IT).
inst_list([diverges|T],['DIVERGES'|IT]) :- !, inst_list(T,IT).
inst_list([deadlocks|T],['DEADLOCKS'|IT]) :- !, inst_list(T,IT).
inst_list([not_enabled(Op)|T],['NOT_ENABLED:',Op|IT]) :- !,inst_list(T,IT).
inst_list([determ(Op1,_ID1,Op2,_ID2)|T],['DETERMINISM_BY_EVENT:',Op1,Op2|IT]) :- !,inst_list(T,IT).
inst_list([cannot_refuse_compl(Ops)|T],Res) :- 
    append(['CANNOT_REFUSE_COMPL:'|Ops],IT,Res),
    inst_list(T,IT).
inst_list([cannot_refuse(Ops)|T],Res) :- 
    append(['CANNOT_REFUSE:'|Ops],IT,Res),
    inst_list(T,IT).
inst_list([refuse('Sigma')|T],Res) :- 
    append(['REFUSED_SET:','Sigma'],IT,Res),
    inst_list(T,IT).    
inst_list([refuse(Ops)|T],Res) :- 
    append(['REFUSED_SET:'|Ops],IT,Res),
    inst_list(T,IT).


tcltk_execute_string_trace(StartNode,Trace) :- /* can be useful for TestCase Generation */
   %state_space_reset,
   user:tcltk_execute_trace_to_node(StartNode), % TO DO: check if not too expensive; usually StartNode will be root or an immediate successor of root
   execute_string_trace_to_node(Trace),!.
tcltk_execute_string_trace(StartNode,Trace) :-
   print_error('Could not execute trace'), print_error(StartNode), print_error(Trace).
   
execute_string_trace_to_node([]) :- !.
execute_string_trace_to_node([spec_cannot_diverge|T]) :- !, execute_string_trace_to_node(T).
execute_string_trace_to_node([deadlocks|T]) :- !, execute_string_trace_to_node(T).
execute_string_trace_to_node([diverges|T]) :- 
    current_state_id(CurID),
    find_tau_trace_to(CurID,divergence,NewTrace,T),!,
    execute_string_trace_to_node(NewTrace).
execute_string_trace_to_node([not_enabled(_)|T]) :- !, execute_string_trace_to_node(T).
execute_string_trace_to_node([cannot_refuse_compl(_)|T]) :- !, execute_string_trace_to_node(T).
execute_string_trace_to_node([cannot_refuse(_)|T]) :- !, execute_string_trace_to_node(T).
execute_string_trace_to_node([refuse(_)|T]) :- !, execute_string_trace_to_node(T).
execute_string_trace_to_node([go('$initialise_machine',ID)|T]) :- 
    current_state_id(CurID),
    transition(CurID,Action,ID),
    functor(Action,'$initialise_machine',_),!,
    user:tcltk_perform_action(Action,ID),
    execute_string_trace_to_node(T).
execute_string_trace_to_node([go('$initialise_machine',ID)|T]) :- 
 /* special case for CurID=root, as setup_constants get merged
    into initialise_machine by refinement checker */
    current_state_id(CurID), CurID=root,
    transition(CurID,Action1,ID1),
    functor(Action1,'$setup_constants',_),
    transition(ID1,Action2,ID),
    functor(Action2,'$initialise_machine',_),!,
    user:tcltk_perform_action(Action1,ID1),
    user:tcltk_perform_action(Action2,ID),
    execute_string_trace_to_node(T).
execute_string_trace_to_node([go('$initialise_machine',ID)|T]) :- 
    current_state_id(root), 
    transition(root,start_cspm_MAIN,ID),!,
    user:tcltk_perform_action(start_cspm_MAIN,ID),
    execute_string_trace_to_node(T).
execute_string_trace_to_node([go('$setup_constants',ID)|T]) :- 
    current_state_id(CurID),
    transition(CurID,Action,ID),
    functor(Action,'$setup_constants',_),!,
    user:tcltk_perform_action(Action,ID),
    execute_string_trace_to_node(T).
execute_string_trace_to_node([go(tau,DestID)|T]) :-  % could now be multiple taus ! adapt
    current_state_id(CurID),
    find_tau_trace_to(CurID,DestID,NewTrace,T),!,
    execute_string_trace_to_node(NewTrace).
execute_string_trace_to_node([go(tau_direct,ID)|T]) :- 
    current_state_id(CurID),
    tau_transition(CurID,Action,ID),!,
    user:tcltk_perform_action(Action,ID),
    execute_string_trace_to_node(T).
execute_string_trace_to_node([go(Action,ID)|T]) :-   /* <---- Node ID's have to be added to avoid problems !! */
    current_state_id(CurID),
    transition(CurID,Ev,ID),translate_event(Ev,Action),
    user:tcltk_perform_action(Ev,ID),!,
    execute_string_trace_to_node(T).
execute_string_trace_to_node([go(Action,ID)|T]) :- print('Could not execute: '),
  print(go(Action,ID,T)),nl,
  current_expression(CurID,State),
  print(current_expression(CurID,State)),nl,
  fail.

tau_priority_transition(CurID,Span,ID) :- impl_trans_term(CurID,tau(TAUINFO),ID), priority_tau(TAUINFO,Span).
priority_tau(rep_int_choice(Span),rep_int_choice(Span)). % add wrapper in case Span is unknown
%priority_tau(tick(S),tick(S)). % can also hide choices: tick was visible; resolves external choice
priority_tau(int_choice_left(Span,_),int_choice(Span)). % TO DO: check span
priority_tau(int_choice_right(Span,_),int_choice(Span)).
% TO DO: hide if only taus possible; same with link parallel
% TO DO: replace span by position insisde CSP expression (in case of multiple copies of same operator)

% first execute prioritized tau transitions from a source location if possible; otherwise any tau is ok
%%%%%%%%%%% DEAD CODE %%%%%%%%%%%%
/*
prioritized_tau_trans(CurID,ID,prio) :- tau_priority_transition(CurID,Span,_),!,
  tau_priority_transition(CurID,Span,ID).
prioritized_tau_trans(CurID,ID,all) :- tau_transition(CurID,_Action,ID).
*/

tau_transition(CurID,Action,ID) :- impl_trans_term(CurID,Action,ID), functor(Action,tau,_).

% find a tau trace to a give id; id can also be 'divergence'; last arg is trace as difference list
:- volatile tau_trace_visited/1.
:- dynamic tau_trace_visited/1.
find_tau_trace_to(CurID,DestID,Trace,TraceTail) :- retractall(tau_trace_visited(_)),
   find_tau_trace_to_aux(CurID,DestID,Trace,TraceTail).

find_tau_trace_to_aux(CurID,DestID,T,Tail) :-
   tau_trace_visited(CurID),!,DestID=divergence,T=Tail.
find_tau_trace_to_aux(CurID,ID,T,Tail) :- ID=CurID,!,T=Tail.
find_tau_trace_to_aux(CurID,ID,[go(tau_direct,ID2)|T],Tail) :- assert(tau_trace_visited(CurID)),
   tau_transition(CurID,_,ID2),
   find_tau_trace_to_aux(ID2,ID,T,Tail).

/* ---------------------------------
   Printing gluing invariant is not used anywhere in the source code (DEAD CODE).
   --------------------------------- */
/*
:- public pgt/0.

pgt :- print_gluing_invariant.
print_gluing_invariant :- print('GLUING INVARIANT'),nl,
   not_refines_table(X,Y,_),
   X \= root,
   X \= concrete_constants(_),
   print_state_as_expression(X),
   print(' => '),nl,
   print_states_as_disjunction(Y),nl,
   print(' & '),nl,
   fail.
print_gluing_invariant :- print('  TRUE'),nl.
   
print_state_as_expression(ID) :-
   visited_expression(ID,State), 
   print('   ('),
   print_bindings(State),
   print(') ').
*/
%print_spec_state_as_expression(ID) :-
 /* need to be able to access state of specification machine ! */
% print('not yet implemented, ID:'), print(ID).
/* 
print_states_as_disjunction([]) :- print(' FALSE'),nl.
print_states_as_disjunction([S]) :- !, print_spec_state_as_expression(S).
print_states_as_disjunction([S|T]) :- !, print_spec_state_as_expression(S),
  print(' or '), print_states_as_disjunction(T).

:- use_module(translate).
print_bindings([]) :- print('TRUE').
print_bindings([bind(Var,Val)]) :- !,
  translate_bvalue(Var,TV), print(TV),
  print('='),
  translate_bvalue(Val,TVal), print(TVal).
print_bindings([bind(Var,Val)|T]) :-
  translate_bvalue(Var,TV), print(TV),
  print('='),
  translate_bvalue(Val,TVal), print(TVal),
  print_bindings(T).
*/
