% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(specfile, [
                    specfile_possible_trans_name/2,
                    specfile_trans/6, prepare_state_for_specfile_trans/2,
                    property/2, elementary_prop/2, 
                    partial_trans/4,
                    
                    animation_mode/1, set_animation_mode/1, add_animation_mode/1,
                    animation_minor_mode/1, set_animation_minor_mode/1,
                    b_mode/0, z_mode/0, b_or_z_mode/0, b_or_z_mode/1,
                    csp_mode/0, csp_mode/1,
		            csp_with_bz_mode/0,
					process_algebra_mode/0,
					
					type_check_csp_and_b/0,
					
                    currently_opened_file/1, 
                    reset_currently_opened_file/0,set_currently_opened_file/1,
                    spec_file_has_been_successfully_loaded/0,
                    
					state_corresponds_to_initialised_b_machine/1,
					state_corresponds_to_initialised_b_machine/2,
					state_corresponds_to_set_up_constants/2,
					current_b_expression/1,
					
					get_operation_name/2,
					build_up_b_real_operation/2,
					expand_const_and_vars/3, expand_const_and_vars_to_full_store/2,
                    expand_to_constants_and_variables/3,
					
					compute_operation_effect_max/6, % has replaced compute_operation_effect/5 
					create_setup_constants_operation/2,
					get_internal_representation/1,
					get_possible_event/1, get_specification_description/2
					]).

:- use_module(tools).

:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module computes transitions and properties depending on the current animator mode.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).
:- use_module(library(avl)).
%:- use_module(library(clpfd)).

:- use_module(debug).
:- use_module(b_interpreter).
:- use_module(preferences).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(translate).
:- use_module(succeed_max,[succeed_max_call_id/3]).

:- use_module(store,[empty_state/1]).
:- use_module(bmachine).
:- use_module(bsyntaxtree).
:- use_module(plugins).


:- use_module(delay).

:- use_module(value_persistance, [load_constants/2]).

/* --------------------------------- */




/* --------------------------------- */
/* ANIMATION & TEMPORAL VERIFICATION */
/* --------------------------------- */


:- dynamic animation_mode/1, animation_minor_mode/1.
% not marking as volatile; otherwise we startup with no animation_mode set
% not all parts of ProB can deal with this at the moment
%:- volatile animation_mode/1, animation_minor_mode/1.

animation_mode(b).
/* possible values: b, csp, xtl, z, csp_and_b, promela */

b_mode :- animation_mode(X), (X=b ; X=csp_and_b).
z_mode :- animation_mode(X), (X=b ; X=csp_and_b), animation_minor_mode(z).
csp_with_bz_mode :- animation_mode(csp_and_b).

csp_mode :- animation_mode(X), csp_mode(X).
csp_mode(cspm).
csp_mode(csp_and_b).

b_or_z_mode :- 
  (animation_mode(X) -> (X=b -> true ; X=csp_and_b)).
b_or_z_mode(b). b_or_z_mode(csp_and_b).

process_algebra_mode :- 
 (animation_mode(X) -> \+ X=b, \+X=z).

set_animation_mode(X) :-
   retractall(animation_mode(_)), retractall(animation_minor_mode(_)),
   assert(animation_mode(X)).

set_animation_minor_mode(X) :- retractall(animation_minor_mode(_)),
   assert(animation_minor_mode(X)).

add_animation_mode(X) :-
   (animation_mode(X) -> true ; assert(animation_mode(X))).
   
   
:- dynamic currently_opened_file/1.
:- volatile currently_opened_file/1.
%currently_opened_file(none).

reset_currently_opened_file :-  retractall(currently_opened_file(_)).

set_currently_opened_file(File) :-
  reset_currently_opened_file,
  assert(currently_opened_file(File)).
  
spec_file_has_been_successfully_loaded :- currently_opened_file(_).

% true if state corresponds to an initialised B machine
state_corresponds_to_initialised_b_machine(const_and_vars(_,_)) :- !,
   b_or_z_mode.
state_corresponds_to_initialised_b_machine([]).
state_corresponds_to_initialised_b_machine([_|_]).
state_corresponds_to_initialised_b_machine(csp_and_b(_,BState)) :-
   csp_with_bz_mode,
   state_corresponds_to_initialised_b_machine(BState).

% check if state corresponds to an initialised B machine, and if so return expanded State
state_corresponds_to_initialised_b_machine(const_and_vars(ConstID,Vars),State) :- !,
   b_or_z_mode, expand_const_and_vars(ConstID,Vars,State).
state_corresponds_to_initialised_b_machine(expanded_const_and_vars(_ConstID,_Vars,FullStore),State) :- !,
   % generated by prepare_state_for_specfile_trans
   State=FullStore.
state_corresponds_to_initialised_b_machine([],[]).
state_corresponds_to_initialised_b_machine([H|T],[H|T]).
state_corresponds_to_initialised_b_machine(csp_and_b(_,BState),State) :-
   csp_with_bz_mode,
   state_corresponds_to_initialised_b_machine(BState,State).


state_corresponds_to_set_up_constants(const_and_vars(ConstID,Vars),State) :- !,
   b_or_z_mode, expand_const_and_vars(ConstID,Vars,State).
state_corresponds_to_set_up_constants(concrete_constants(Constants),State) :- !,
   b_or_z_mode, State=Constants.
state_corresponds_to_set_up_constants(csp_and_b(_,BState),State) :- !,
   csp_with_bz_mode,
   state_corresponds_to_set_up_constants(BState,State).
state_corresponds_to_set_up_constants([],State) :- !, State=[],
   animation_mode(b).
state_corresponds_to_set_up_constants(State,RState) :-
   animation_mode(b),
   (State = root -> \+ b_machine_has_constants_or_properties,RState=[] ; RState=State).

:- use_module(state_space,[current_expression/2]).
current_b_expression(CurBState) :-
        current_expression(_CurID,CurState),
        state_corresponds_to_set_up_constants(CurState,CurBState).

/* TRANSITIONS */

:- use_module(xtl_interface).

% ---------------------------------------------------


% Tries to find successor state for Op in State DB. Fails, if no solution is possible
% Input: State DB, Operation Op
% Output: DB2 the sucessor state
% Residue: Unfinished co-routines (should not happen!) 
%specfile_trans(DB,Op,DB2,[]) :- !, specfile:i_transition(DB,Op,DB2).
specfile_trans(DB,OpName,Op,DB2,TransInfo,Residue) :-
%       i_transition(DB,OpName,Op,DB2,TransInfo), Residue = []. % is faster, but no residue checking
         animation_mode(MODE),
         call_residue(i_transition(MODE,DB,OpName,Op,DB2,TransInfo),Residue).

i_transition(b,State,OpName,Op,NewState,TransInfo) :- !,
    b_trans(State,OpName,Op,NewState,TransInfo).
i_transition(xtl,State,_OpName,Op,NewState,[]) :- !,
    xtl_transition(State,Op,NewState).
i_transition(cspm,State,_OpName,Op,NewState,[]) :- !,
    cspm_transition(State,Op,NewState).
i_transition(promela,State,_OpName,Op,NewState,[]) :- !, 
    promela_transition(State,Op,NewState).
 % ; MODE=smv         -> smv_transition(State,Op,NewState),TransInfo=[]
i_transition(csp_and_b,State,_OpName,Op,NewState,TransInfo) :- !,
    csp_and_b_transition(State,Op,NewState,TransInfo).
i_transition(MODE,State,_OpName,Op,NewState,TransInfo) :- 
    plugin_transition(MODE,State,Op,NewState,TransInfo).


% if you call specfile_trans multiple times for the same state, it is
% more efficient to call this predicate first once
prepare_state_for_specfile_trans(const_and_vars(ConstID,Vars),R) :- !, %print(expanding_const(ConstID)),nl,
      R = expanded_const_and_vars(ConstID,Vars,FullStore),
      expand_const_and_vars(ConstID,Vars,FullStore). %, print(full(FullStore)),nl.
% TO DO: also expand for csp_and_b
prepare_state_for_specfile_trans(X,X).


/* can be attempted if trans fails to attempt to use partially computed transitions */
partial_trans(DB,Op,DB2,Residue) :-
  call_residue(partial_i_transition(DB,Op,DB2),Residue).


partial_i_transition(root,'$partial_setup_constants',concrete_constants(ConstantsStore)) :-
     animation_mode(b),
     set_error_context(operation('$partial_setup_constants',root)),
     b_interpreter:b_partial_set_up_concrete_constants(ConstantsStore).
     
/* build up skeletons of possible operations; to be used before time_out call is made */
specfile_possible_trans_name(State,OpName) :- animation_mode(MODE),
  (MODE=b -> b_possible_trans_name(State,OpName)
   ; setup_transition_skeleton(MODE,OpName) /* TO DO: add more possible skeletons */
  ).

:- use_module(library(random)).
b_possible_trans_name(root,OpName) :- !, 
 (b_machine_has_constants_or_properties
   -> OpName = '$setup_constants'
   ;  OpName = '$initialise_machine').
b_possible_trans_name(concrete_constants(_),OpName) :- !, OpName = '$initialise_machine'.
b_possible_trans_name(_State,OpName) :- 
   ((preferences:preference(randomise_operation_order,true),random(1,3,1))
      -> b_machine_operation_names_in_reverse_order(OpName) ; true),
   b_get_machine_operation(OpName,_Res,_Params,_,_).

% --------------

:- use_module(probcspsrc(haskell_csp),[channel_type_list/2]).
:- use_module(debug).
% check whether a currently open CSP is type compatible with a loaded B machine
type_check_csp_and_b :- animation_mode(csp_and_b), debug_println(9,type_check_csp_and_b),
   b_get_machine_operation(ChannelOpName,Res,Params,_), append(Params,Res,ParamsRes),
   channel_type_list(ChannelOpName,CSPTypeList),
   % TO DO: also check B variable names which match channel name
   debug_println(9,checking_synchronised_channel(ChannelOpName,ParamsRes,CSPTypeList)),
   l_check_compatibility(ParamsRes,CSPTypeList,ChannelOpName),
   fail.
type_check_csp_and_b.

%convert_csp_type_to_list(type(T),R) :- convert_csp_type_to_list2(T,R).
%convert_csp_type_to_list2(dotUnitType,[]).
%convert_csp_type_to_list2(dotTupleType(T),T).

l_check_compatibility([],[],_) :- !.
l_check_compatibility([B|BT],[CSP|CSPT],ChannelOpName) :- !, check_compatibility(B,CSP,ChannelOpName),
    l_check_compatibility(BT,CSPT,ChannelOpName).
l_check_compatibility([],CSP,ChannelOpName) :- !,
    add_error(type_check_csp_and_b,'CSP Channel has too many parameters: ', ChannelOpName:CSP).
l_check_compatibility(B,[],ChannelOpName) :- !,
    print('* CSP Channel has too few parameters: '), print(ChannelOpName),nl,
    print('* B Arguments will be ignored: '), print(B),nl.
    %add_error(type_check_csp_and_b,'CSP Channel has too few parameters: ', ChannelOpName:B).
    
check_compatibility(b(identifier(ID),TYPE,INFO),CSP_TYPE,ChannelOpName) :- !,
   (type_ok(TYPE,CSP_TYPE) -> true ;
   % TO DO: pretty print B and CSP Types
     add_error(type_check_csp_and_b,'Incompatible types between B and CSP: ',ChannelOpName:ID:(TYPE:CSP_TYPE),b(identifier(ID),TYPE,INFO))
   ).
check_compatibility(B,CSP,ChannelOpName) :- !,
   add_internal_error('Illegal types: ',check_compatibility(B,CSP,ChannelOpName)).

% TO DO: needs to be refined much more;
type_ok(Type,CSPType) :-
  (type_ok2(Type,CSPType) -> true
    ; add_error(type_check_csp_and_b,'Cannot be converted (B:CSP): ',Type:CSPType),fail).
type_ok2(boolean,X) :- !, is_csp_boolean_type(X).
type_ok2(integer,X) :- !, is_csp_integer_type(X).
type_ok2(string,CSP) :- !,is_csp_global_set_type(CSP,string).
type_ok2(set(T),CSP) :- !, is_csp_set_type(CSP,T).
% TO DO: what if we have set(couple(integer,T)) instead of seq(T) ?
type_ok2(seq(T),CSP) :- !, is_csp_seq_type(CSP,T).
type_ok2(couple(A,B),CSP) :- !, is_csp_couple_type(CSP,A,B).
type_ok2(global(GS),CSP) :- !,is_csp_global_set_type(CSP,GS).
type_ok2(X,Y) :- print(unkown(X,Y)),nl.

is_csp_couple_type(setValue([na_tuple([A|Rest])|_]),TA,TB) :- type_ok(TA,setValue([A])),
     check_rest_couple_els(Rest,TB).
is_csp_couple_type(typeTuple([A|Rest]),TA,TB) :- type_ok(TA,A), check_rest_couple(Rest,TB).

check_rest_couple_els([B],TB) :- !, type_ok(TB,setValue([B])).
check_rest_couple_els([B1|BRest],couple(TB1,TBRest)) :- type_ok(TB1,setValue([B1])),
    check_rest_couple_els(BRest,TBRest).
check_rest_couple([B],TB) :- !, type_ok(TB,B).
check_rest_couple([B1|BRest],couple(TB1,TBRest)) :- type_ok(TB1,B1),
    check_rest_couple(BRest,TBRest).

is_csp_set_type('Set'([H|_]),Type) :- type_ok(Type,H). % just check Head
is_csp_set_type(setValue([H|_]),Type) :- type_ok(Type,H). % just check Head

is_csp_seq_type('Seq'(H),Type) :- type_ok(Type,H).
is_csp_seq_type(setValue([H|_]),Type) :- H=list(L), is_csp_list(L,Type).
is_csp_list([],_).
is_csp_list([H|_],Type) :- type_ok(Type,setValue([H])).

is_csp_boolean_type(boolType).
is_csp_boolean_type(setValue([H|_])) :- is_csp_boolean_value(H).
is_csp_boolean_value(true).
is_csp_boolean_value(false).

is_csp_integer_type(intType).
is_csp_integer_type(setFromTo(_,_)).
is_csp_integer_type(setFrom(_)).
is_csp_integer_type(setValue([int(_)|_])).

is_csp_global_set_type(dataType(_),_GS). % TO DO CHECK MEMBERS
is_csp_global_set_type(setValue([]),_GS). % TO DO CHECK MEMBERS
is_csp_global_set_type(setValue([_|_]),_GS). % TO DO CHECK MEMBERS

% --------------


:- use_module(bmachine,[b_get_machine_operation/4]).
:- use_module(bsets_clp,[tuple_of/3]).
%csp_and_b_transition(root,start_csp_and_b,csp_and_b,false).
%csp_and_b_transition(csp_and_b,Op,NewState,AbortStateReached) :-
csp_and_b_transition(root,Op,NewState,TransInfo) :- !,
  Op = start_cspm_MAIN, NewState = csp_and_b_root,TransInfo = [].
csp_and_b_transition(csp_and_b_root,Trans,NewState,TransInfo) :- !,
  Trans = tau(Op),
  b_trans(root,_,Op,InitialBState,TransInfo),
  (InitialBState = concrete_constants(_)
    -> NewState = InitialBState
    ;  csp_initialisation_for_b(InitialCSPState),
       NewState = csp_and_b(InitialCSPState,InitialBState)
  ).
csp_and_b_transition(concrete_constants(C),Trans,NewState,TransInfo) :-  !,
  Trans = tau(Op),
  b_trans(concrete_constants(C),_,Op,InitialBState,TransInfo),
  csp_initialisation_for_b(InitialCSPState),
  NewState = csp_and_b(InitialCSPState,InitialBState).
csp_and_b_transition(csp_and_b(CSPState,BState),EventVisibleToUser,csp_and_b(NewCSP,NewB),TransInfo) :- !,
  csp_transition_for_b(CSPState,ChOpName,ChArgs,CSPAction,NewCSP),
  % TO DO: split up csp_transition_for_b: into two parts; second part converts datatypes and only called if linking op exists
 %% print(csp_trans(ChOpName,ChArgs)),nl, %%
  length(ChArgs,Len), length(OpArgs,Len),
  (build_up_b_operation_for_csp(ChOpName,OpArgs,Operation)
    ->  %%  print_message(trying_b_trans(Operation,ChArgs,OpArgs)), %%
        generate_b_operationargs_from_csp(ChArgs,OpArgs),
        %% print(trying_b_trans(Operation)),nl, %%
        b_trans(BState,_,Operation,NewB,TransInfo)
        %% ,print_message(csp_btrans(Operation)) %%
    ;   NewB=BState,TransInfo=[],
        %% print(probing(ChOpName)),nl,
        ((expand_const_and_vars_to_full_store(BState,FBState),
          try_lookup_value_in_store_and_global_sets(ChOpName,FBState,IdValue))
         -> /* interpret as probing operation: get value of var,cst,SET */
            generate_b_operationargs_from_csp(ChArgs,OpArgs),
            (OpArgs=[SingleVal]
              ->  SingleVal = IdValue, Operation='-->'(ChOpName,[IdValue])
              ;   OpArgs=[Val1,Val2],
                  bsets_clp:tuple_of(Val1,Val2,IdValue), /* use exact_element_of ??? */
                  OpCall =.. [ChOpName,Val1],
                  Operation='-->'(OpCall,[Val2])
            )
         ; /* no B operation of the name; could be tau or tick or extra comm channel */
          %Op1 =.. [ChOpName|OpArgs], 
           Operation = CSPAction  % removed csp(.) wrapper [for LTL model checker]
        )
  ),
  hide_csp_event(CSPAction,Operation,EventVisibleToUser).
csp_and_b_transition(State,Op,NewState,[]) :-
   % we have a pure CSP state: use CSP-M transition
   cspm_transition(State,Op,NewState).


hide_csp_event(CSPAction,_,EventVisibleToUser) :-
   haskell_csp:symbol('HIDE_CSPB',_,Span,_),!, % if there is a HIDE_CSPB definition then we do MAIN [{| HIDE_CSPB |}] B_MACHINE \ {| HIDE_CSPB |}
   haskell_csp:evaluate_argument(val_of('HIDE_CSPB',Span),EvCList),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCList,ECList,no_loc_info_available)),
  haskell_csp:cspm_hide_action(CSPAction,omega,ECList, Span, EventVisibleToUser,_).
hide_csp_event(CSPAction,Operation,EventVisibleToUser) :-
  ( preferences: get_preference(csp_event_visible_to_user,true) -> EventVisibleToUser = CSPAction
  ; EventVisibleToUser = Operation).

build_up_b_operation_for_csp(OpName,OpArgs,Operation) :-  %print(csp(OpName,OpArgs,Operation)),nl,
  nonvar(OpName),
  b_get_machine_operation(OpName,Res,Params,_), %print(get_machine_op(OpName,Res,Params)),nl,
  specfile:generate_args(OpArgs,Params,Res, BArgs,BResults), %print(gen_args(BArgs,BResults)),nl,
  % TO DO: either also extract types from Params, Res or better write a static checker that tests that the B operation and CSP channel types are compatible
  safe_univ(Op1,[OpName|BArgs]), %print(op1(Op1)),nl,
  (BResults = []  -> Operation = Op1  ;  Operation =.. ['-->',Op1,BResults]).

build_up_b_real_operation(OpName,Operation) :-
  b_get_machine_operation_for_animation(OpName,Res,Params,_,_,true), %true==TopLevel
  length(Params,Len), length(OpArgs,Len),
  length(Res,RLen), length(BResults,RLen),
  safe_univ(Op1,[OpName|OpArgs]),
  (BResults = []  -> Operation = Op1  ;  Operation =.. ['-->',Op1,BResults]).

  
  
generate_args([],[],Res,[],BRes) :- length(Res,L), length(BRes,L).
generate_args([],[_|T],Res,[_|TA],BRes) :-
   generate_args([],T,Res,TA,BRes).
generate_args([CSPArg|TC],[_|T],Res,[CSPArg|TA],BRes) :-
   generate_args(TC,T,Res,TA,BRes).
generate_args([CSPArg|TC],[],[_|TRes],[],[CSPArg|TBRes]) :-
   generate_args(TC,[],TRes,[],TBRes).
generate_args([H|T],[],[],[],[]) :-
   add_error(specfile,'CSP provides extra arguments: ',[H|T]).


% tool to get the basic operation/channel/event name of a transition:
get_operation_name(Op,Name) :- var(Op),!, add_error(get_operation_name,'Variable Transition: ',Op), Name=Op.
get_operation_name('-->'(F,_),N) :- !, functor(F,N,_). 
get_operation_name(io([Action|_],proc(_Nr,_PROC),_SPAN),F) :- 
     animation_mode(promela),!,functor(Action,F,_).
get_operation_name(io(_V,Ch,_SPAN),N) :- animation_mode(cspm),!, functor(Ch,N,_).
get_operation_name(FullOperation,Name) :- functor(FullOperation,Name,_).




:- use_module(state_space,[visited_expression/2]).

b_trans(State,OpName,Operation,NewState,PathInfo) :- 
	 compute_operation_effect_max(State,OpName,Operation,NewState,PathInfo,_Max).

compute_operation_effect_max([],OpName,Operation,NewState,PathInfo,Max) :- 
     compute_operation_on_expanded_store([],OpName,Operation,Updates,PathInfo,Max),
     store:store_updates_and_normalise(Updates,[],NewState).
compute_operation_effect_max([H|T],OpName,Operation,NewState,PathInfo,Max) :- 
     compute_operation_on_expanded_store([H|T],OpName,Operation,Updates,PathInfo,Max),
     store:store_updates_and_normalise(Updates,[H|T],NewState).
compute_operation_effect_max(const_and_vars(ConstID,Vars),OpName,Operation,ResultingStore,PathInfo,Max) :- 
     prepare_state_for_specfile_trans(const_and_vars(ConstID,Vars),R),
     %print(expanded_const_and_vars(ConstID)),nl,
     compute_operation_effect_max(R,OpName,Operation,ResultingStore,PathInfo,Max).
compute_operation_effect_max(concrete_constants(ConstantsStore),'$initialise_machine',OpInit,ResultingStore,PathInfo,Max) :-
     % print('Searching for valid INITIALISATION'),nl,!,
     (var(Max) -> preferences:get_preference(maxNrOfInitialisations,Max) ; true),
     user:do_one_gui_event,user:do_one_gui_event,user:do_one_gui_event,
     succeed_max_call_id(initialisation,b_interpreter:b_initialise_machine(ConstantsStore,InitialVars,InitialStore,PathInfo),Max),
     create_initialisation_operation(InitialVars,OpInit),
     user:do_one_gui_event,
     ((\+ preferences:preference(symmetry_mode,flood), /* TO DO : improve permutation to allow using const_and_vars optimisation */
       state_space:visited_expression(ConstID,concrete_constants(ConstantsStore)))
      -> ResultingStore = const_and_vars(ConstID,InitialVars)  /* avoid storing constant values again */
      ;  ResultingStore = InitialStore).
compute_operation_effect_max(expanded_const_and_vars(ConstID,Vars,FullStore),OpName,Operation,ResultingStore,PathInfo,Max) :- 
     %print(const_and_vars(ConstID,Vars)),nl,
     compute_operation_on_expanded_store(FullStore,OpName,Operation,Updates,PathInfo,Max),
     store:store_updates_and_normalise(Updates,Vars,NewVars),
     ResultingStore = const_and_vars(ConstID,NewVars).
     %ErrorMsg = compute_operation_effect(const_and_vars(ConstID,Vars),Operation,ResultingStore,PathInfo),
     %reconstruct_constants_store(ConstID,ConstantsStore,NewState,ErrorMsg,ResultingStore).
compute_operation_effect_max(root,'$setup_constants',OpSetup,concrete_constants(ConstantsStore),[],Max) :-
     b_machine_has_constants_or_properties,
     (var(Max) -> preferences:get_preference(maxNrOfInitialisations,Max) ; true),
     compute_constants(ConstantsStore,Max),
     ((get_preference(show_initialisation_arguments,true),extract_value_list(ConstantsStore,Vals))
      -> safe_univ(OpSetup,['$setup_constants'|Vals])
       ; OpSetup = '$setup_constants'
     ),
     %%print_message('FOUND_CONSTANTS'(OpName)),
     user:do_one_gui_event.
compute_operation_effect_max(root,'$initialise_machine',OpInit,InitialStore,PathInfo,Max) :-
     \+ b_machine_has_constants_or_properties,
     % print('Searching for valid INITIALISATION'),nl,!,
     (var(Max) -> preferences:get_preference(maxNrOfInitialisations,Max) ; true),
     user:do_one_gui_event, user:do_one_gui_event, user:do_one_gui_event,
     empty_state(EmptyState),
     succeed_max_call_id(initialisation,b_interpreter:b_initialise_machine(EmptyState,InitialVars,InitialStore,PathInfo),Max),
     create_initialisation_operation(InitialVars,OpInit),
     user:do_one_gui_event.

compute_constants(ConstantsStore,_Max) :-
    % check if values have been computed and stored into a file before -- if yes, use them
    load_constants(Stores,MaxReached),!, choose_loaded_solution(Stores,ConstantsStore,MaxReached).
compute_constants(ConstantsStore,Max) :-
    user:do_one_gui_event, user:do_one_gui_event, user:do_one_gui_event,
    % print('Searching for valid CONSTANTS'),nl,!,
    succeed_max_call_id(set_up_constants,b_interpreter:b_set_up_concrete_constants(ConstantsStore),Max).

choose_loaded_solution(Stores,ConstantsStore,_MaxReached) :-
    member(ConstantsStore,Stores).
choose_loaded_solution(_Stores,_ConstantsStore,true) :-
    % provoke the setting of the "maximum reached flag"
    succeed_max_call_id(set_up_constants,member(_,_),1),fail.

expand_const_and_vars_to_full_store(root,R) :- !,R=[].
expand_const_and_vars_to_full_store(concrete_constants(ConstantsStore),FullStore) :- !,
    ConstantsStore = FullStore.
expand_const_and_vars_to_full_store(const_and_vars(ConstID,Vars),FullStore) :- !,
    expand_const_and_vars(ConstID,Vars,FullStore).
expand_const_and_vars_to_full_store(R,R).

expand_const_and_vars(ConstID,Vars,FullStore) :-
    state_space:visited_expression(ConstID,concrete_constants(ConstantsStore)), %% This can take a while if the Constants are big !
    append(ConstantsStore,Vars,FullStore).

expand_to_constants_and_variables(root,[],[]).
expand_to_constants_and_variables(concrete_constants(ConstStore),ConstStore,[]).
expand_to_constants_and_variables(const_and_vars(ConstID,VarStore),ConstStore,VarStore) :-
    state_space:visited_expression(ConstID,concrete_constants(ConstStore)).
expand_to_constants_and_variables([],[],[]).
expand_to_constants_and_variables([H|T],[],[H|T]).
     
% compute operation effect on a store where all constants have been put into the environment
compute_operation_on_expanded_store(InState,OpName,Operation,NewState,PathInfo,Max) :-
     (var(Max) -> preferences:get_preference(maxNrOfEnablingsPerOperation,Max) ; true),
     %randomised_top_level_operation(OpName),
    % (var(OpName) -> b_interpreter:try_get_op_name(Operation,OpName) ; true), % this is also called by execute_top_level_operation
    % print_bt_message(b_execute_top_level_operation(OpName,Operation,InState,NewState,PathInfo)),
     succeed_max_call_id(OpName,b_interpreter:b_execute_top_level_operation_update(OpName,Operation,InState,NewState,PathInfo),Max).
     %b_execute_top_level_operation_update(OpName,Operation,InState,NewState,PathInfo)

create_initialisation_operation(InitialVars,OpInit) :-
    get_preference(show_initialisation_arguments,true),
    extract_value_list(InitialVars,Vals),
    !,safe_univ(OpInit,['$initialise_machine'|Vals]).
create_initialisation_operation(_InitialVars,'$initialise_machine').

create_setup_constants_operation(ConstantVars,Setup) :-
    extract_value_list(ConstantVars,Vals),
    !,safe_univ(Setup,['setup_constants'|Vals]).

%reconstruct_constants_store(ConstID,ConstantsStore,NewState,Msg,ResultingStore) :-
%    ( append(ConstantsStore,NewVars,NewState) ->
%        ResultingStore = const_and_vars(ConstID,NewVars)
%    ; otherwise ->
%        add_error(specfile,'Could not extract constants after operation:', Msg),
%        ResultingStore = NewState
%    ).
                                

extract_value_list([],[]).
extract_value_list([bind(_,X)|T],[X|ET]) :- extract_value_list(T,ET).




/* PROPERTIES */

property(const_and_vars(ConstID,Vars),Property) :- b_or_z_mode,
    expand_const_and_vars(ConstID,Vars,FullStore),!,
    property(FullStore,Property).
property(csp_and_b(CSPState,BState),Property) :- animation_mode(csp_and_b),!,
    %% print(checking_csp_and_b(CSPState)),nl,
    expand_const_and_vars_to_full_store(BState,FBState),
   (xtl_interface:cspm_property(CSPState,Property) ;
    b_invariant_property(b,FBState,Property) ; b_property(FBState,Property)).
property(State,Property) :- animation_mode(AM),
   (AM = b -> (b_invariant_property(AM,State,Property) ; b_property(State,Property))
    ; AM=xtl -> xtl_interface:xtl_property(State,Property)
    ; AM=cspm -> xtl_interface:cspm_property(State,Property)
    ; AM=promela -> xtl_interface:promela_property(State,Property)
 %   ; AM=smv -> xtl_interface:smv_property(State,Property)
    ; AM=csp_and_b -> b_property(State,Property) % State should be root
    ; otherwise -> plugin_state_property(AM,State,Property)
   ).

%b_property(const_and_vars(ConstID,Vars),Prop) :- !,
%   expand_const_and_vars(ConstID,Vars,FullStore),
%   b_property(FullStore,Prop).
b_property(root,Prop) :- !, Prop='='(PARA,VAL), b_preference_par(PARA,VAL).
%b_property([H|T],abort) :- %DB \= concrete_constants(_),
%          debug:debug_mode(on), b_state_is_abort([H|T]).
b_property([H|T],non_ground(Var)) :- debug:debug_mode(on),
         member(bind(Var,Val),[H|T]),\+(ground(Val)).
b_property([H|T],Prop) :- elementary_prop([H|T],Prop).
b_property([],Prop) :- elementary_prop([],Prop).
b_property(concrete_constants(DB),Prop) :- elementary_prop(DB,Prop).


%b_state_is_abort(csp_and_b(_,State)) :- !, b_state_abort_value(State),!.
%b_state_is_abort(State) :- b_state_abort_value(State),!.
%b_state_abort_value(State) :- member(bind(_,abort(_)),State).  % abort(_) isn't really used that much anymore

:- use_module(preferences).
:- use_module(b_global_sets,[b_fd_type/3]).

b_preference_par('MAXINT',MaxInt) :- get_preference(maxint,MaxInt).
b_preference_par('MININT',MinInt) :- get_preference(minint,MinInt).
b_preference_par(card(GlobalSet),Card) :- b_fd_type(GlobalSet,Low,Up),
 Card is Up-Low+1.
 
/* --------------------------------------------------------- */



   
b_invariant_property(b,DB,invariant_false) :-
   DB \= root, DB \= concrete_constants(_),
   preference(do_neginvariant_checking,true),
   (b_interpreter:state_satisfies_negation_of_invariant(DB) -> true ; fail).

/* add stuff for Z */
b_invariant_property(z,DB,Prop) :- DB \= root, DB \= concrete_constants(_),
    b_invariant_property(b,DB,Prop).
%b_invariant_property(_,_,_) :- clear_error_context,fail.

elementary_prop(DB,Prop) :-
    member(bind(Var,Val),DB),
    elementary_prop3(Val,Var,Prop).

:- use_module(custom_explicit_sets,[is_avl_relation/1,get_first_avl_elements/4]).
elementary_prop3(Val,Var,Prop) :- var(Val),!, Prop = '='(Var,'_VAR_').
elementary_prop3(avl_set(Avl),Var,Prop) :- !,
   (custom_explicit_sets:is_avl_relation(Avl) % \+ custom_explicit_sets:is_avl_sequence(Avl))
     -> elementary_rel_prop(avl_set(Avl),Var,Prop) ; elementary_var_prop(avl_set(Avl),Var,Prop)).
elementary_prop3([H|T],Var,Prop) :- H=(_,_), !, elementary_rel_prop([H|T],Var,Prop).
elementary_prop3(Val,Var,Prop) :- elementary_var_prop(Val,Var,Prop).
    
elementary_var_prop(Val,Var,Prop) :-
    translate_bvalue(Val,Text),
    Prop = '='(Var,Text). 

elementary_rel_prop(Val,Var,Prop) :-
    nonvar(Val),
    get_relation_element(Val,X,Y),
    create_texpr(identifier(Var),any,[],Fun),
    create_texpr(value(X),any,[],Arg),
    create_texpr(function(Fun,Arg),any,[],Lhs),
    create_texpr(value(Y),any,[],Rhs),
    create_texpr(equal(Lhs,Rhs),pred,[],Expr),
    translate_bexpression_with_limit(Expr,80,Prop).

get_relation_element(avl_set(A),X,Y) :- custom_explicit_sets:is_avl_relation(A),
    !, %preferences:preference(expand_avl_upto,Limit),
       Limit=30, % reduced limit, as now we can inspect with evaluation view
       custom_explicit_sets:get_first_avl_elements(A,Limit,Els,CutOff),
       ((member((X,Y),Els), X\=term(_))
        ;
       (CutOff=not_all, X='...', Y='...')).
get_relation_element(Value,X,Y) :-
    member((X,Y),Value). /* otherwise we have a record structure */

			  
:- use_module(bmachine,[b_show_machine_representation/2]).
:- use_module(probcspsrc(haskell_csp),[get_internal_csp_representation/1]).
get_internal_representation(X) :- animation_mode(AM), get_internal(AM,X).
get_internal(b,X) :- !, b_show_machine_representation(X,true).
get_internal(cspm,X) :- !, get_internal_csp_representation(X).
get_internal(cspm,X) :- !, get_internal_csp_representation(X).
get_internal(AM,X) :- !, plugin_get_internal(AM,X).

get_internal(_,X) :- X="Internal representation only available in B or CSP mode".


:- use_module(probcspsrc(haskell_csp),[channel/2]).
get_possible_event(OpName) :- b_or_z_mode,
   b_get_machine_operation(OpName,_,_,_B).
get_possible_event(Channel) :- csp_mode,
   channel(Channel,_).
% what about csp_and_b mode ?

% obtain a textual description of specification category names
get_specification_description(Category,Name) :-
  animation_mode(Mode), 
  (animation_minor_mode(Minor) -> true ; Minor = none),
  (get_specific_description(Mode,Minor,Category,N) -> Name=N
   ; get_default_description(Category,N) -> Name = N
   ; add_error(get_specification_description,'Unknown category: ',Category),
     Name=Category).

get_specific_description(cspm,_,C,N) :- get_csp_description(C,N).
get_specific_description(b,eventb,C,N) :- get_eventb_description(C,N).
get_specific_description(b,tla,C,N) :- get_tla_description(C,N).

get_csp_description(operations,'CHANNELS').
get_csp_description(operation,'CHANNEL').
get_csp_description(assertions,'CSP_ASSERTIONS').
get_csp_description(machine,'CSP_SPECIFICATION').

get_tla_description(properties,'ASSUME').
get_tla_description(machine,'MODULE').

get_eventb_description(properties,'AXIOMS').
get_eventb_description(assertions,'THEOREMS').
get_eventb_description(operations,'EVENTS').
get_eventb_description(operation,'EVENT').
get_eventb_description(machine,'MODEL').

get_default_description(properties,'PROPERTIES').
get_default_description(assertions,'ASSERTIONS').
get_default_description(invariant,'INVARIANT').
get_default_description(invariants,'INVARIANTS').
get_default_description(operations,'OPERATIONS').
get_default_description(operation,'OPERATION').
get_default_description(machine,'MACHINE').
get_default_description(constants,'CONSTANTS').
get_default_description(variables,'VARIABLES').
get_default_description(initialisation,'INITIALISATION').
