% (c) 2009-2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(state_custom_dot_graph,[tcltk_generate_state_custom_dot_graph/1]).

:- use_module(module_information).
:- module_info(group,dot).
:- module_info(description,'This module provides a way to generate custom state graphs (using info from DEFINITIONS).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(dot_graph_generator).

:- use_module(translate,[    print_bexpr/1 ]).

:- use_module(self_check).
:- use_module(error_manager).

:- use_module(state_space,[current_expression/2]).
			  
tcltk_generate_state_custom_dot_graph(File) :-
       current_expression(_CurID,State),
       specfile:state_corresponds_to_initialised_b_machine(State,BState),!,
      ( bmachine:b_get_machine_custom_nodes_function(NodesFunction)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_NODES Function in DEFINITIONS')),
       (bmachine:b_get_machine_custom_edges_function(EdgesFunction)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_EDGES Function in DEFINITIONS')),
       eval_definition_fun(NodesFunction, BState, NodesRes),
       print(custom_nodes(NodesRes)),nl,
       retractall(custom_nodes(_)), assert(custom_nodes(NodesRes)),
       eval_definition_fun(EdgesFunction, BState, EdgesRes),
       print(custom_edges(EdgesRes)),nl,
       retractall(custom_edges(_)), assert(custom_edges(EdgesRes)),
       retractall(custom_node_id(_,_)), retractall(nodectr(_)), 
       gen_dot_graph(File,state_custom_dot_graph,node_predicate,trans_predicate,none,none),
       print(finished),nl.
tcltk_generate_state_custom_dot_graph(_) :-
       add_error_and_fail(tcltk_generate_state_custom_dot_graph,'Only possible for initialised B machine.').

:- dynamic custom_nodes/1, custom_edges/1.

node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color) :- SubGraph=none,
     Shape=box, Style=filled,
     custom_nodes(Nodes),
     (var(Nodes) -> add_error(node_predicate,'Variable custom_nodes: ',Nodes),fail ; true),
     member((NodeVal,ColVal),Nodes),
     (translate:translate_bvalue(NodeVal,NodeDesc) -> true ; NodeDesc = '???'),
     translate_colour(ColVal,Color),
     %tools:print_message(node(NodeVal,ColVal,Color,NodeID)),
     gen_id(NodeVal,NodeID). %tools:print_message(nodeid(NodeID)).
 
:- dynamic custom_node_id/2, nodectr/1.
nodectr(0).

get_ctr(Res) :- retract(nodectr(C)),!, C1 is C+1, assert(nodectr(C1)), Res=C.
get_ctr(0) :- assert(nodectr(1)).

gen_id(NodeVal,ID) :-
 (custom_node_id(NodeVal,ID) -> true
   ; get_ctr(C),assert(custom_node_id(NodeVal,C)),ID=C).

translate_colour(int(X),Colour) :- !, (translate_int_col(X,Colour) -> true ; Colour=black).
translate_colour(_,lightgray).
% TO DO : add string conversions, rgb values , ...

translate_int_col(0,white).
translate_int_col(1,blue).
translate_int_col(2,red).
translate_int_col(3,green).
translate_int_col(4,lightgray).
translate_int_col(5,orange).
translate_int_col(6,yellow).
translate_int_col(7,brown).
translate_int_col(8,violet).
translate_int_col(9,tomato).
translate_int_col(10,darkslateblue).
translate_int_col(11,maroon2).
translate_int_col(12,olivedrab2).
translate_int_col(13,chartreuse3).
% red,green,blue,yellow,orange,black,white,gray,brown,violet,darkred,tomato,darkblue,
%             darkgray,darkviolet,darkslateblue,lightblue,lightgray,maroon2,olivedrab2,
%             steelblue,chartreuse3,chartreuse4

trans_predicate(NodeID,Label,SuccID,Color) :- Label = 'custom', Color = blue,
   custom_edges(Edges),
     (var(Edges) -> add_error(trans_predicate,'Variable custom_edges: ',Edges),fail ; true),
   member((From,To),Edges),
   gen_id(From,NodeID),
   gen_id(To,SuccID).

:- use_module(custom_explicit_sets,[try_expand_custom_set/2]).

eval_definition_fun(AnimFunction,BState, FunctionRes) :- %trace,
       b_interpreter:b_compute_expression(AnimFunction,[],BState,FunctionResCl),
       nl, print('FunctionResult'(FunctionResCl)),nl,
       ~~mnf( custom_explicit_sets:try_expand_custom_set(FunctionResCl, FunctionRes) ).
