% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(state_space_open_nodes_c,[not_all_transitions_added/1,
          reset_open_ids/0,
          add_id_at_front/1,
          add_id_at_end/1,
          add_id_random/1,
          add_id_to_process/3,
          add_id_with_weight/2,
          pop_id_from_front_direct/1,
          pop_id_from_end_direct/1,
          retract_open_node_direct/1,
          
          open_ids_empty/0,
          
          retract_open_ids_with_statistics/0,
          print_state_space_open_nodes/0]).

:- use_module(tools).

% Open Node Queue Interface
% The interface for adding new ids and popping them from the list of open nodes:

:- dynamic not_all_transitions_added/1.
%not_all_transitions_added(v_0).


:- use_module(myheap,'../extensions/myheap/myheap.pl',[myheap_init/0,
  myheap_reset/0, myheap_add_node/2, myheap_add_node_at_front/1, myheap_add_node_at_end/1,
  myheap_add_node_random/1,
  myheap_size/1, myheap_isempty/0, 
  myheap_top_id/1, myheap_top_weight/1, 
  myheap_pop/1,
  
  myheap_max_weight/1,
  myheap_pop_max/1]).


open_ids_empty :- myheap_empty(X),X==1.

reset_open_ids :- myheap_init, retractall(ids_still_to_process(_,_,_)),
    retractall(not_all_transitions_added(_)).
retract_open_ids_with_statistics :-
     retract_with_statistics(state_space_open_nodes,
         [not_all_transitions_added(_), ids_still_to_process(_,_,_)]),
     myheap_init. % To do: expand and compute statistics

:- dynamic ids_still_to_process/3. %ids for which we still have to compute the heuristic function
% purpose: delay the computation until all transitions have been added of PredecessorID

add_id_to_process(ID,Hash,PredecessorID) :-
  assert(ids_still_to_process(ID,Hash,PredecessorID)), assert(not_all_transitions_added(ID)).

process_ids :- retract(ids_still_to_process(ID,Hash,PredID)),  % we assume ID cannot be root
   state_space:out_degree(PredID,Degree), % print(add_out_degree_id(ID,PredID,Degree)),nl,
    Weight is Degree*1000 + Hash mod 1000,
  %  Weight = Degree,
   myheap_add_node(ID,Weight),fail.
process_ids.


add_id_at_front(root) :- !, myheap_add_node_at_front(-99), assert(not_all_transitions_added(root)).
add_id_at_front(NodeID) :- myheap_add_node_at_front(NodeID), 
    print_node(add_front,NodeID),
   asserta(not_all_transitions_added(NodeID)).


add_id_at_end(root) :- !, myheap_add_node_at_end(-99), assert(not_all_transitions_added(root)).
add_id_at_end(NodeID) :- myheap_add_node_at_end(NodeID),
    print_node(add_end,NodeID),
   assertz(not_all_transitions_added(NodeID)).
 
 
add_id_with_weight(root,W) :- !, myheap_add_node(-99,W), assert(not_all_transitions_added(root)).
add_id_with_weight(ID,W) :- myheap_add_node(ID,W), assert(not_all_transitions_added(ID)).

add_id_random(root) :- !, myheap_add_node_random(-99), assert(not_all_transitions_added(root)).
add_id_random(ID) :- myheap_add_node_random(ID), assert(not_all_transitions_added(ID)).

pop_id_from_front_direct(NodeID) :- nonvar(NodeID),!, 
   print_message(illegal_call(pop_id_from_front(NodeID))),fail.
pop_id_from_front_direct(RNodeID) :- process_ids,
   myheap_pop(ID), ID \= -1,
   (ID= -99 -> NodeID=root ; NodeID=ID),
    % print_node(pop_front,NodeID),
   (retract(not_all_transitions_added(NodeID)) -> RNodeID=NodeID
     ; print_message(no_node(NodeID)),pop_id_from_front_direct(RNodeID)).
     
pop_id_from_end_direct(NodeID) :- bmachine:b_get_machine_heuristic_function(_),!,
    pop_id_from_front_direct(NodeID). % we are in heuristic mode: always pop from front
pop_id_from_end_direct(NodeID) :- nonvar(NodeID),!, 
   print_message(illegal_call(pop_id_from_end_direct(NodeID))),fail.
pop_id_from_end_direct(RNodeID) :- process_ids, myheap_pop_max(ID),   ID \= -1,
   (ID= -99 -> NodeID=root ; NodeID=ID),
    print_node(pop_end,NodeID),
   (retract(not_all_transitions_added(NodeID)) -> RNodeID=NodeID
    ; print_message(no_node(NodeID)),pop_id_from_end_direct(RNodeID)).
   
print_node(Info,ID) :- state_space:visited_expression(ID,State),
 print(Info), print(' '),print(ID), print(' : '), print(State),nl.

retract_open_node_direct(NodeID) :- retract(not_all_transitions_added(NodeID)),
  print_message('Cannot delete node from C++ multimap yet'),
  print_message(NodeID).
 

print_state_space_open_nodes :- print_dynamic_pred(state_space_open_nodes,not_all_transitions_added,1),
print_message('Cannot print open nodes C++ multimap yet').
  %print_dynamic_pred(state_space_open_nodes,not_all_z,1),