% (c) 2009-2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

/* 
 ensure_loaded('images/state_viewer_images').
*/

/* ------------------------------------------------------------------ */

:- use_module(bmachine,[b_machine_name/1,
                        b_get_machine_animation_function/1,
                        b_get_machine_animation_function_default/1,
                        b_get_definition_string_from_machine/2,
                        get_animation_image/2]).

/* this file is not a module (yet):
:- use_module(module_information).
:- module_info(group,gui).
:- module_info(description,'This module can provide a Tk graphical representation of a B state.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').
*/

:- use_module(specfile,[state_corresponds_to_initialised_b_machine/2]).

/* First argument: image list of pairs: ImageName, ImageFilePath
   Second argument: list of triplets {type framename label}
   */

:- use_module(translate,[    print_bexpr/1 ]).            

tcltk_get_image_list(ImageList,FramesAndItems) :-
       current_expression(_CurID,State),
       specfile:state_corresponds_to_initialised_b_machine(State,BState),
       b_get_machine_animation_function(AnimFunction),
       get_animation_images(ImageList),
       %print(animation_images(ImageList)),
      % print('Animation Function: '),
      % print_bexpr(AnimFunction),
       eval_definition_fun(AnimFunction, BState, FunctionRes1),
       % print('animation_function'(FunctionRes1)),nl,
       ((b_get_machine_animation_function_default(AnimFunctionDefault),
	     eval_definition_fun(AnimFunctionDefault, BState, FunctionRes0))
           ->   %print(overriding),nl, print(FunctionRes0),nl,
               ~~mnf(user:override_anim_fun(FunctionRes0,FunctionRes1,FunctionRes))
            ; FunctionRes = FunctionRes1
       ),
       % print('full_fun'(FunctionRes)),nl,
       (FunctionRes = int(ImageNr)
         ->  /* simple style animation: just a single image */
             convert_nr_into_tk_label(ImageNr,"IMG",Label),
             FramesAndItems = [frame, row1, top, img11, row1, Label]
          ;  ( get_min_max_row(FunctionRes,Min,Max)
               ->  debug_println(9,'Matrix function'),
                   debug_println(9,min_max(Min,Max)),
                   setup_frames(Min,Max,Frames), debug_println(9,frames(Frames)),
                   sort(FunctionRes,SortedFR),
                   get_frame_items(SortedFR,Items),
                 %  print(items(Items)),nl,
                   append(Frames,Items,FramesAndItems)
               ;   FunctionRes=[] -> print('**** Empty ANIMATION_FUNCTION result'),nl,
                   fail
               ;   print('**** Unknown Format for ANIMATION_FUNCTION result'),nl,
                   print('**** must be of type INT or (INT*INT)+->INT'),nl,
                   % print(FunctionRes),nl, %%
                   fail
             )
       ).

:- use_module(custom_explicit_sets,[try_expand_custom_set/2]).

eval_definition_fun(AnimFunction,BState, FunctionRes) :-
       b_interpreter:b_compute_expression(AnimFunction,[],BState,FunctionResCl),
     %%  nl, print('FunctionResult'(FunctionResCl)),nl,
       ~~mnf( custom_explicit_sets:try_expand_custom_set(FunctionResCl, ExpCl) ),
       (translate_to_ints(ExpCl, FunctionRes)
          -> true /* already translate to ints so that override will work properly */
          ; check_anim_function(ExpCl),fail
       )  % ,print('fres'(FunctionRes)),nl
       .

check_anim_function([]) :- !.
check_anim_function(int(_)) :- !. % just a single picture
check_anim_function([H|_]) :- !, 
  (H = ((Row,Col),_Img) ->
         (((get_int(Row,_),get_int(Col,_)) -> true
          ;  add_error(state_viewer_images,'Elements of ANIMATION_FUNCTION should be convertible to (INT*INT)*INT: ',H))
        )
     ;  add_error(state_viewer_images,
            'Elements of the ANIMATION_FUNCTION should be of type (INT*INT)*INT: ',H)
   ).
check_anim_function((A,B)) :- !,
   add_error(state_viewer_images,'ANIMATION_FUNCTION should be a set of pairs, not a pair: ',(A,B)).
check_anim_function(X) :- !,
   add_error(state_viewer_images,'ANIMATION_FUNCTION should be of type (INT*INT) +-> INT (or similar): ',X).

:- use_module(library(avl)).

translate_to_ints(In,Out) :- empty_avl(E),translate_to_ints(In,E,Out).
translate_to_ints([],_,[]) :-!.
translate_to_ints(X,_,int(IX)) :- get_int(X,IX),!.
translate_to_ints([((Row,Col),Img)|T], AVL, Res) :- !,
   get_int(Row,RI), get_int(Col,CI), (Img=II),
   (avl_fetch((RI,CI),AVL,ExistingValue)
    -> add_error(state_viewer_images,
          'Animation function defines two images for (Row,Col):[img1,img2] = ',((RI,CI):[ExistingValue,II])),
       NewAVL = AVL, Res=TT
     ; avl_store((RI,CI),AVL,II,NewAVL), Res=[((RI,CI),II)|TT]
   ),
   translate_to_ints(T,NewAVL,TT).

/* a version of override which does not worry about correct typing */
override_anim_fun([],Fun2,Res) :- !, Res=Fun2.
override_anim_fun(int(_),Fun2,Res) :- !, Res=Fun2.
override_anim_fun([((RowT,Col),Img)|T],Fun2,Res) :- !,
   (member(((RowT,Col),_),Fun2)
     -> Res = ResT
      ; Res = [((RowT,Col),Img)|ResT]
   ), override_anim_fun(T,Fun2,ResT).
override_anim_fun(X,F,R) :-
   add_error(state_viewer_images,'Unknown Format for ANIMATION_FUNCTION_DEFAULT', X),
   R=F.

get_min_max_row([((Row,_Col),_Img)|T],Min,Max) :-
   =(Row,RowInt),
   get_min_max_row2(T,RowInt,RowInt,Min,Max).
get_min_max_row2([],Min,Max,Min,Max).
get_min_max_row2([((Row,_Col),_Img)|T],Min,Max,MinR,MaxR) :-
    (Row<Min -> Min1=Row ; Min1=Min),
    (Row>Max -> Max1=Row ; Max1=Max),
    get_min_max_row2(T,Min1,Max1,MinR,MaxR).
    
get_int(int(R),R).
get_int(fd(R,_Name),R).
get_int(pred_true /* bool_true */,1).
get_int(pred_false /* bool_false */,0).

get_frame_items([],[]).
get_frame_items([((Row,Col),Img)|T],[ItemLbl,FrameLbl,ImageLbl|TR]) :-
    convert_nr_into_tk_label(Row,"frame",FrameLbl),
    ((get_int(Img,ImgNr),images(ImgNr,_,ImageLbl))
      -> convert_2nr_into_tk_label(Row,Col,"item",ItemLbl)
      ;  (convert_2nr_into_tk_label(Row,Col,"text",ItemLbl),
          ((get_int(Img,ImgNr),get_animation_string(ImgNr,ImageLbl)) -> true
            ; translate:translate_bvalue(Img,ImageLbl)
              %convert_nr_into_tk_label(Img,"ERR",ImageLbl),
              %add_error(get_frame_items,'No ANIMATION_IMGxx or ANIMATION_STRxx declared for xx = ',Img)
         ))
    ),
    get_frame_items(T,TR).

setup_frames(Min,Max,[]) :- Min>Max,!.
setup_frames(Min,Max,[frame, FrameLbl, top |T]) :-
    convert_nr_into_tk_label(Min,"frame",FrameLbl), M1 is Min+1,
    setup_frames(M1,Max,T).

:- dynamic images/3.

get_animation_images(Res) :-
    findall( img(Nr,S,Lbl), get_animation_image_and_label(Nr,S,Lbl), List),
    retractall(images(_,_,_)),
    flatten_img_list(List,Res).

flatten_img_list([],[]).
flatten_img_list([img(Nr,S,Lbl)|T],[Lbl,S|FT]) :- assert(images(Nr,S,Lbl)), flatten_img_list(T,FT).

get_animation_string(Nr,S) :-
         number_codes(Nr,TailCodes),
         /* ANIMATION_STR */
         atom_codes(Def_Name,[65,78,73,77,65,84,73,79,78,95,83,84,82|TailCodes]),
         b_get_definition_string_from_machine(Def_Name,S).
         
get_animation_image_and_label(Nr,S,Lbl) :-
    get_animation_image(Nr,S),
    convert_nr_into_tk_label(Nr,"IMG",Lbl).

convert_nr_into_tk_label(Nr,Prefix,Label) :- number_codes(Nr,Codes),
        append(Prefix,Codes,C),
        atom_codes(Label,C).
        
convert_2nr_into_tk_label(Nr1,Nr2,Prefix,Label) :-
        number_codes(Nr1,Codes1), number_codes(Nr2,Codes2),
        append("_",Codes2,C2), append(Codes1,C2,Codes),
        append(Prefix,Codes,C),
        atom_codes(Label,C).
 