% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(store, [ %store_values_for_existing_ids/4,
                  store_value_for_existing_id/4,
                  variable_is_defined/2,
                  store_value/4,
                  store_updates_and_normalise/3,
                  store_intermediate_updates/3,
                  save_updates_in_existing_store/2,
                  lazy_store_updates_and_normalise/3,
                  merge_updates/3,
                  get_id_value/3, 
                  lookup_value/3, lookup_value/4,
                  lookup_value_for_existing_id/3, lookup_value_for_existing_id/4,
                  lookup_and_delete_value/7,
                  empty_state/1,
                  add_var_to_localstate/4,
                  set_up_localstate/4,
                  set_up_undefined_localstate/3,
                  delete_variables_from_state/3,
                  
                  normalise_value_for_var/3, normalise_value_for_var/4,
                  normalise_value/2,
                  l_normalise_values/2,
                  l_expand_and_normalise_values/2, l_expand_and_normalise_values/3,
                  normalise_store/2,
                  copy_variable_store/4,
                  %extract_variable/4,
                  insert_variables/5
                  %,create_substate/3
                  ]).
                  
:- use_module(library(lists)).
:- use_module(self_check).
:- use_module(debug).

:- use_module(typechecker).
:- use_module(preferences).
:- use_module(tools).

:- use_module(bsyntaxtree).

:- use_module(error_manager).

:- use_module(bmachine, [b_is_constant/1]).

:- type variable_id +--> (lambda_res(integer) ; atomic).
:- type binding +--> bind(variable_id,any).
:- type store +--> list(binding).

:- use_module(tools).
:- use_module(module_information).
:- module_info(group,interpreter).
:- module_info(description,'This module takes care of storing B values in an environment and normalising values.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

/* --------------------------------------------------------- */
    
/* store and lookup of values */


% unused at the moment:
%store_values_for_existing_ids([], [], _, _).
%store_values_for_existing_ids([VarsHd|VarsTl], [ValsHd|ValsTl], InState, OutState) :-
%  store_value_for_existing_id(VarsHd, ValsHd, InState, OutState),
%  store_values_for_existing_ids(VarsTl, ValsTl, InState, OutState).


variable_is_defined(Var,Store) :- safe_member(bind(Var,_),Store,variable_is_defined).

store_value(Var,Value,Old,New) :- store_value2(Old,Var,Value,New).
store_value2([],Var,Value,[bind(Var,Value)]).
store_value2([bind(V,VV)|T],Var,Value,[bind(V,NValue)|UT]) :-
  ( Var=V ->
      NValue=Value, UT=T
  ; otherwise ->
      NValue=VV, store_value2(T,Var,Value,UT)).


store_value_for_existing_id(L,ID,V,L2) :- var(ID),!,
    add_internal_error('VarID is var !!',store_value_for_existing_id(ID,V,L,L2)),fail.
store_value_for_existing_id(L,ID,V,L2) :- store_value_for_existing_id_aux(L,ID,V,L2).
store_value_for_existing_id_aux([],Var,Value,[bind(Var,Value)]) :-
    add_error(store_value_for_existing_id,'Variable does not exist in store: ',Var).
store_value_for_existing_id_aux([bind(V,VV)|T],Var,Value,[bind(V,VV2)|UT]) :-
    (Var=V -> VV2=Value, UT=T
     ;  VV2=VV,store_value_for_existing_id_aux(T,Var,Value,UT)).
 




/* store_intermediate_updates(InState,Updates,OutStates) */
/* rewrote it so that it copies state skeleton even if updates are not yet known */

:- assert_must_succeed((store:store_intermediate_updates([bind(y,1),bind(x,2),bind(a,22)],[bind(x,33)],R),R==[bind(y,1),bind(x,33),bind(a,22)])).
:- assert_must_succeed((store:store_intermediate_updates([bind(Y,1),bind(X,2),bind(A,22)],[bind(x,33)],R),X=x,A=a,Y=y,R==[bind(y,1),bind(x,33),bind(a,22)])).

/*
% Note: the InStore is not sorted; nor is Update
% Assumption is that Updates are usually much smaller than Store
store_intermediate_updates_new(InStore,Update,OutStore) :- empty_avl(E),
    store_int_delay(Update,E,InStore,OutStore).

:- block store_int_delay(-,?,?,?).
store_int_delay([],UpdateAVL,InStore,OutStore) :-
   store_avl_update(InStore,UpdateAVL,OutStore).
store_int_delay([bind(Var,Value)|T],UpdateAVL,InStore,OutStore) :-
   store_int_delay2(Var,Value,T,UpdateAVL,InStore,OutStore).
:- block store_int_delay2(-,?,?, ?,?,?).
store_int_delay2(Var,Value,T, UpdateAVL,InStore,OutStore) :-
   (avl_fetch(Var,UpdateAVL,_OtherVal) ->
      add_error(store_intermediate_updates,'Variable has been assigned twice in parallel: ',Var),
      %add_error(store_intermediate_updates,'Values: ',(OtherVal,Value)),
      AVL2 = UpdateAVL
    ;  avl_store(Var,UpdateAVL,Value,AVL2)),
   store_int_delay(T,AVL2,InStore,OutStore).

:- block store_avl_update(-,?,?).
store_avl_update([],_UpdateAVL,[]).
store_avl_update([bind(Var,OldVal)|T],UpdateAVL,[bind(Var,NewValue)|NT]) :-
	store_avl_update_aux2(Var,OldVal,T,UpdateAVL,NewValue,NT).
:- block store_avl_update_aux2(-,?,?,?,?,?).
store_avl_update_aux2(Var,OldVal,T,UpdateAVL,NewValue,NT) :-
   (avl_fetch(Var,UpdateAVL,UpdVal) -> NewValue=UpdVal
     ; NewValue=OldVal),
    store_avl_update(T,UpdateAVL,NT).
*/

store_intermediate_updates(InStore,Update,OutStore) :-
   store_intermediate_updates2(InStore,Update,OutStore,Update).
   
:- block store_intermediate_updates2(-,?,?,?), store_intermediate_updates2(?,-,?,?).
store_intermediate_updates2([],Updates,[],FullUpdate) :-
    store_intermediate_updates3(Updates,FullUpdate).
store_intermediate_updates2([bind(Var,OldVal)|InT],Updates,[bind(Var,NewVal)|OutT],FullUpdate) :-
   search_and_delete_value(Var,OldVal,NewVal,Updates,NUpdates),
   store_intermediate_updates2(InT,NUpdates,OutT,FullUpdate).

:- block store_intermediate_updates3(-,?).
store_intermediate_updates3(Updates,FullUpdate) :-
   ( Updates = [] -> true
   %; Updates = [bind(abort__flag__info,_)] -> true
   ;  add_error(store_intermediate_updates,'Illegal or double assignments: ',(rest(Updates),full(FullUpdate)))).



:- assert_pre(store:store_updates_and_normalise(A,B,_C),(type_check(A,store),type_check(B,store))). 
:- assert_post(store:store_updates_and_normalise(_A,_B,C),(type_check(C,store))).

% TO DO: optimize; probably better to traverse InState once, rather than for each update binding
%:- block store_updates_and_normalise(-,?,?). % block normally not necessary

store_updates_and_normalise(Updates,InState,OutState) :-
   preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,EXPAND),
   store_updates_and_normalise1(Updates,EXPAND,InState,OutState).
  
store_updates_and_normalise1([],_EXPAND,InState,InState).
store_updates_and_normalise1([bind(Var,NewVal)|RestUpdates],EXPAND,InState,OutState) :-
   store_updates_and_normalise2(Var,EXPAND,NewVal,RestUpdates,InState,OutState).
%:- block store_updates_and_normalise2(-,?,?,?,?).
store_updates_and_normalise2(Var,EXPAND,NewVal,RestUpdates,InState,OutState) :-  
  var(Var),!,
  add_internal_error('Prolog variable for Var: ',store_updates_and_normalise2(Var,EXPAND,NewVal,RestUpdates,InState,OutState)),
  lazy_store_updates_and_normalise2(Var,EXPAND,NewVal,RestUpdates,[],InState,OutState).
store_updates_and_normalise2(Var,EXPAND,NewVal,RestUpdates,InState,OutState) :-  
   normalise_value_for_var(Var,EXPAND,NewVal,NormalisedNewVal),
   store_value_for_existing_id(InState,Var,NormalisedNewVal,InState2),
   (safe_member(bind(Var,OtherValue),RestUpdates,store_updates_and_normalise) ->
      add_error(store_updates_and_normalise,'Variable has been assigned twice in parallel: ',Var),
      add_error(store_updates_and_normalise,'Values: ',(NewVal,OtherValue))
    ;  (b_is_constant(Var)
        -> add_error(store_updates_and_normalise,'Assigned value to constant: ',Var) ; true)
   ),
   store_updates_and_normalise1(RestUpdates,EXPAND,InState2,OutState).
   
% a blocking version of store_updates_and_normalise; to be used by CBC

lazy_store_updates_and_normalise(Updates,InState,OutState) :-
   preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,EXPAND),
   lazy_store_updates_and_normalise1(Updates,EXPAND,[],InState,OutState).
   
:- block lazy_store_updates_and_normalise1(-,?,?,?,?). 
lazy_store_updates_and_normalise1([],_,_,InState,InState).
lazy_store_updates_and_normalise1([bind(Var,NewVal)|RestUpdates],EXPAND,UpdSoFar,InState,OutState) :-
   lazy_store_updates_and_normalise2(Var,EXPAND,NewVal,RestUpdates,UpdSoFar,InState,OutState).
:- block lazy_store_updates_and_normalise2(-,?,?,?,?,?,?).
lazy_store_updates_and_normalise2(Var,EXPAND,NewVal,RestUpdates,UpdSoFar,InState,OutState) :-  
   normalise_value_for_var(Var,EXPAND,NewVal,NormalisedNewVal),
   %print(store(Var,NewVal,NormalisedNewVal)),nl,
   store_value_for_existing_id(InState,Var,NormalisedNewVal,InState2),
   (safe_member(bind(Var,OtherValue),UpdSoFar,store_updates_and_normalise) ->
      add_error(store_updates_and_normalise,'Variable has been assigned twice in parallel: ',Var),
      add_error(store_updates_and_normalise,'Values: ',(NewVal,OtherValue))
    ;  (b_is_constant(Var)
        -> add_error(store_updates_and_normalise,'Assigned value to constant: ',Var) ; true)
   ),
   lazy_store_updates_and_normalise1(RestUpdates,EXPAND,[bind(Var,NewVal)|UpdSoFar],InState2,OutState).
 
 
/* use to merge updates, e.g. for sequential composition LHS ; RHS */
:- assert_pre(store:merge_updates(_A,_B,_C),true). % (type_check(_A,store),type_check(_B,store))). 
:- assert_post(store:merge_updates(_A,_B,_C),true). %(type_check(_C,store))).

:- assert_must_succeed((store:merge_updates([bind(x,1)],[bind(x,2)],R),R==[bind(x,2)])).
:- assert_must_succeed((store:merge_updates([bind(x,1)],[bind(X,2),bind(y,3)],R),X=x,R==[bind(x,2),bind(y,3)])).
:- assert_must_succeed((store:merge_updates([bind(x,1)],[bind(X,2),bind(y,3)],R),X=z,R==[bind(z,2),bind(y,3),bind(x,1)])).
:- assert_must_succeed((store:merge_updates([bind(y,1),bind(x,1)],[bind(x,2)],R),R==[bind(x,2),bind(y,1)])).

:- block merge_updates(-,?,?).
merge_updates([],RHSUpdates,RHSUpdates).
merge_updates([Bind|LHT],RHSUpdates,Rest) :-
    merge_updates2( Bind,LHT,RHSUpdates,Rest).
:- block merge_updates2(?,?,-,?).
merge_updates2(bind(Var,NewVal),LHT,RHSUpdates,Rest) :-
    merge_remove(RHSUpdates,Var,NewVal,RHSUpdates2),
    merge_updates(LHT,RHSUpdates2,Rest).

:- block merge_remove(-,?,?,?).
merge_remove([],Var,NewVal,[bind(Var,NewVal)]).
merge_remove([bind(Var2,Val2)|T2],Var1,Val1,[bind(Var2,Val2)|NewT2]) :-
    when(ground(Var2),
      (Var1=Var2
        -> NewT2=T2  /* binding in second updates overrides first update */
         ; merge_remove(T2,Var1,Val1,NewT2)
      )).
/* ----------------------- */

:- block save_updates_in_existing_store(-,?).
save_updates_in_existing_store([],_Store).
save_updates_in_existing_store([bind(Var,Value)|Urest],Store) :-
    save_updates_in_existing_store2(Var,Value,Store),
    save_updates_in_existing_store(Urest,Store).
:- block save_updates_in_existing_store2(-,?,?).
save_updates_in_existing_store2(Var,Value,Store) :-
    lookup_value_for_existing_id(Var,Store,Value).


:- assert_must_succeed(store:l_normalise_values([int(1)],[int(1)])).

l_normalise_values([],[]).
l_normalise_values([H|T],[NH|NT]) :- normalise_value(H,NH), l_normalise_values(T,NT).

:- block l_expand_and_normalise_values(-,?).
l_expand_and_normalise_values([],[]).
l_expand_and_normalise_values([H|T],[NH|NT]) :-
    normalise_value_for_var(unkown,true,H,NH),
    l_expand_and_normalise_values(T,NT).
:- block l_expand_and_normalise_values(-,?,?).
l_expand_and_normalise_values([],[],_).
l_expand_and_normalise_values([H|T],[NH|NT],[VarName|VT]) :-
    %% print(normalising(VarName)),nl, %%
    normalise_value_for_var(VarName,true,H,NH),
    l_expand_and_normalise_values(T,NT,VT).



:- assert_must_succeed(store:normalise_value(int(1),int(1))).
:- assert_must_succeed((store:normalise_value([int(1)],R), R==avl_set(node(int(1),true,0,empty,empty)))).
:- assert_must_succeed((store:normalise_value([int(1),int(2)],R1),
                        store:normalise_value([int(2),int(1)],R2),R1==R2)).
%:- assert_must_succeed((store:normalise_value(X,[int(1),int(2)]),X=[int(2),int(1)])).
:- assert_must_succeed((store:normalise_value(X,_R),X=[Y,Z],
             Y=[int(2),int(1)],Z=[int(1)])).
:- assert_must_succeed(store:normalise_value(_X,[int(1),int(3)])).
:- assert_must_succeed((store:normalise_value(X,[int(3),int(1)]),
                        kernel_objects:equal_object(X,[int(1),int(3)]))).
:- assert_must_fail(store:normalise_value(int(2),int(1))).


normalise_value_for_var(Var,Val,NormVal) :- 
    preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,EXPAND),
    normalise_value_for_var(Var,EXPAND,Val,NormVal).

:- block normalise_value_for_var(?,?,-,-).
normalise_value_for_var(Var,EXPAND,Val,NormVal) :-  % print(norm(Var,EXPAND,Val,NormVal)),nl, trace,
   ~~mnf(expand_closure_value(Val,EXPAND,Var,ENewVal)), % also normalises values inside closure if not expanded
   normalise_value(ENewVal,NormVal). %  , tools_printing:print_term_summary(normalised(ENewVal,NormVal)).

:- block normalise_value(-,-).
normalise_value(V,Res) :-
    ((nonvar(V),var(Res)) -> mnf(normalise2(V,ResR)),Res=ResR ; normalise3(V,Res)).
    % ((nonvar(V),var(Res)) -> normalise2(V,Res) ; normalise3(V,Res)).



:- use_module(b_global_sets,[b_integer_or_string_set/1,all_elements_of_type/2]).
:- use_module(library(avl)).

%:- use_module(debug,[time/1]).
%%normalise2(X,R) :- var(X),!,R=X. /* just in case the when is unfrozen by hypercall */
normalise2([H|T],Res) :- !,normalise_list([H|T],[],Res).
normalise2((A,B),(NA,NB)) :- !, normalise_value(A,NA), normalise_value(B,NB).
normalise2(global_set(GS),N) :- !, 
   (b_integer_or_string_set(GS)
     -> N = global_set(GS) /* can be very large or infinite, so don't expand */
     ; all_elements_of_type(GS,Res),
       normalise2(Res,N)).
normalise2(avl_set(A),R) :- !,
  (empty_avl(A) -> R=[] ; %time(
   (avl_to_list(A,L), ord_list_to_avl(L,B), R=avl_set(B))).
normalise2(string(S),R) :- !, R=string(S).
normalise2(rec(Fields),R) :- !, R=rec(NFields), normalise_fields(Fields,NFields).
normalise2(V,V).

:- block normalise_fields(-,-).
normalise_fields([],[]).
normalise_fields([field(Name,Val)|T], [field(Name,NVal)|NT]) :- normalise_value(Val,NVal),
    normalise_fields(T,NT).
/* version to be called if for some reason second argument already instantiated;
  e.g. when operation arguments and results are already instantiated */
%%normalise3(X,R) :- var(R),var(X),!,R=X. /* just in case the when is unfrozen by hypercall */
:- use_module(kernel_objects,[equal_object/2]).
normalise3(R,[H|T]) :- !,kernel_objects:equal_object(R,[H|T]). /* don't insist on normalising if a set is
  already partially given*/
normalise3((A,B),(NA,NB)) :- !, normalise_value(A,NA), normalise_value(B,NB).
normalise3(V,V).


:- block normalise_list(-,?,?).
normalise_list([],Acc,Res) :- !,
      when(ground(Acc),(sort(Acc,SAcc),ord_list_to_avl(SAcc,AVL),Res=avl_set(AVL))).
normalise_list([H|T],Acc,Res) :- !, normalise_value(H,NH),
  normalise_list(T,[NH-true|Acc],Res).  % - true for ord_list_to_avl
normalise_list(avl_set(AVL),Acc,Res) :- !, avl_to_list(AVL,L), append(Acc,L,Acc2), normalise_list([],Acc2,Res).
normalise_list(global_set(GS),Acc,Res) :- !, all_elements_of_type(GS,G), normalise_list(G,Acc,Res).
normalise_list(closure(P,T,B),Acc,Res) :- !, % in this context we have to expand closure:
  custom_explicit_sets:expand_closure_to_avl_or_list(P,T,B,ExpandedValue,check(normalise_list)),
  normalise_list(ExpandedValue,Acc,Res).
normalise_list(L,Acc,Res) :- add_internal_error('Unexpected argument: ',normalise_list(L,Acc,Res)),
  normalise_list([L-true],Acc,Res).


expand_closure_value(Var,_,_,R) :- var(Var),!,R=Var.
expand_closure_value(closure(P,T,B),EXPAND,VarName,ExpandedValue) :- !, %print(expand_closure_value(P,T,B)),nl,
   expand_closure_value2(P,T,B,EXPAND,VarName,ExpandedValue).
expand_closure_value(closure_x(P,T,B,Exp),EXPAND,VarName,ExpandedValue) :- !,
   (ground(Exp) -> ExpandedValue=Exp ; expand_closure_value(closure(P,T,B),EXPAND,VarName,ExpandedValue)).
expand_closure_value(Val,_Expand,_VarName,Val).

:- use_module(custom_explicit_sets,[dont_expand_this_explicit_set/1]).
:- block expand_closure_value2(-,?,?,?,?,?), expand_closure_value2(?,-,?,?,?,?), expand_closure_value2(?,?,-,?,?,?).
expand_closure_value2(P,T,B,EXPAND,VarName,ExpandedValue) :- !, 
   % print(expand_closure_value(P,T,B)),nl,
   ((EXPAND==false ; dont_expand_this_explicit_set(closure(P,T,B)))
      ->   construct_normalised_closure(P,T,B,ExpandedValue)   %, print(not_expanding(closure(P,T,B))),nl
      ;    %print(expanding(closure(P,T,B))),nl,
      on_exception(error(representation_error('CLPFD integer overflow'),_),
           time_out_with_factor_call(custom_explicit_sets:expand_closure_to_avl_or_list(P,T,B,ExpandedValue,no_check),
                0.3,store:time_out_expand_backup(VarName,P,T,B,ExpandedValue))
             % , print(expanded_closure(ExpandedValue)),nl
             , (print('% CLP(FD) overflow during closure expansion: '), print(VarName),nl,
                construct_normalised_closure(P,T,B,ExpandedValue)))
    ).

/*
:- use_module(custom_explicit_sets,[dont_expand_this_explicit_set/1, is_interval_closure/5]).
% maybe systematically expand small intervals and global_set's for normalization purposes ??
% (alternatively, we could re-convert certain avl-sets to symbolic form here)
dont_expand_closure_value(_,_,_) :- 
   preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,false),!.
%dont_expand_closure_value(P,T,B) :-
% is_interval_closure(P,T,B,Low,Up), number(Low), number(Up),!,
%    % interval closures are quite efficient for certain manipulations
%   Size is Up+1-Low, Size>100. % another magic constant ; which value to choose ??
%dont_expand_closure_value(_,_,_) :- 
%   preferences:get_preference(convert_comprehension_sets_into_closures,false),!.
%   % in the case above: we have already decided not to expand the closure earlier, dont repeat check !? TO DO: check that this is indeed ok; that no closures are constructed otherwise; maybe we should just check for small interval closures ?
dont_expand_closure_value(P,T,B) :- dont_expand_this_explicit_set(closure(P,T,B)).
   */

time_out_expand_backup(VarName,P,T,B,ExpandedValue) :-
   print('% Expansion time-out for: '), print(VarName),nl,
   construct_normalised_closure(P,T,B,ExpandedValue).

construct_normalised_closure(P,T,B,closure(P,T,NB)) :- 
   normalise_closure_body(B,NB).
  % print(norm),nl,print(B),nl,print(NB),nl,nl.

% normalise values that are compiled inside closures
normalise_closure_body(b(OP,T,I),b(NOP,T,FI)) :- 
   (var(I) -> print(b(OP,T,I)),nl , when(nonvar(OP),(print(b(OP,T,I)),nl)) ; true),
   filter_info(I,FI),ncb(OP,NOP).

filter_info(V,R) :- var(V),!,R=V.
filter_info([],[]).
filter_info([H|T],Res) :-
 (important_info(H) -> Res=[H|TR] ; Res=TR), filter_info(T,TR).

important_info(contains_wd_condition).
important_info(prob_annotation(_)).
important_info(used_ids(_)).

:- use_module(kernel_mappings).
% ncb stands for normalise closure body
ncb(V,R) :- var(V),!, R=V.
ncb(truth,R) :- !, R=truth.
ncb(value(V),value(NV)) :- !,normalise_value_for_var(closure_body,V,NV).
ncb(integer(X),integer(NX)) :- !, X=NX.
ncb(identifier(X),identifier(NX)) :- !, X=NX.
ncb(string(X),string(NX)) :- !, X=NX.
ncb(integer_set(X),integer_set(NX)) :- !, X=NX.
ncb(set_extension(X),set_extension(NX)) :- !, X=NX.
ncb(general_sum(Ids,Condition,Expression),general_sum(Ids,NC,NE)) :- !,
   normalise_closure_body(Condition,NC),normalise_closure_body(Expression,NE).
ncb(general_product(Ids,Condition,Expression),general_product(Ids,NC,NE)) :- !,
   normalise_closure_body(Condition,NC),normalise_closure_body(Expression,NE).
ncb(sequence_extension(X),sequence_extension(NX)) :- !, X=NX.
ncb(external_function_call(F,L),external_function_call(F,LL)) :- !, L=LL.
ncb(external_pred_call(F,L),external_pred_call(F,LL)) :- !, L=LL.
ncb(equal(A,B),equal(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(couple(A,B),couple(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(interval(A,B),interval(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(function(A,B),function(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(relations(A,B),relations(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(member(A,B),member(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(not_member(A,B),not_member(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(negation(B),negation(NB)) :- !, normalise_closure_body(B,NB).
ncb(conjunct(A,B),conjunct(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(disjunct(A,B),disjunct(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(equivalence(A,B),equivalence(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(implication(A,B),implication(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(comprehension_set(V,B),comprehension_set(V,NB)) :- !, normalise_closure_body(B,NB).
ncb(convert_bool(B),convert_bool(NB)) :- !, normalise_closure_body(B,NB).
ncb(exists(IDs,B),exists(IDs,NB)) :- !, normalise_closure_body(B,NB).
ncb(forall(IDs,A,B),forall(IDs,NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(assertion_expression(Cond,ErrMsg,Expr),assertion_expression(NA,ErrMsg,NB)) :- !,
  normalise_closure_body(Cond,NA), normalise_closure_body(Expr,NB).
ncb(let_predicate(Ids,AssignmentExprs,Pred),let_predicate(Ids,AssignmentExprs,NPred)) :- !,
  normalise_closure_body(Pred,NPred).
ncb(F,NF) :- functor(F,Op,1), 
  (unary_function(Op,_) ; unary_in_boolean_type(Op,_)), !, 
  arg(1,F,Arg1), functor(NF,Op,1), arg(1,NF,NormArg1),
  normalise_closure_body(Arg1,NormArg1).
ncb(F,NF) :- functor(F,Op,2), 
  (binary_function(Op,_) ; kernel_mappings:binary_boolean_operator(Op,_,_) ; binary_in_boolean_type(Op,_)), !, 
  arg(1,F,Arg1),arg(2,F,Arg2), functor(NF,Op,2), arg(1,NF,NormArg1), arg(2,NF,NormArg2),
  normalise_closure_body(Arg1,NormArg1),
  normalise_closure_body(Arg2,NormArg2).
ncb(X,X) :- print(ncb(X)),nl.

%ncb(set_extension([b(couple(b(add(b(identifier(idx),integer,[nodeid(pos(53,1,11,91,11,93))]),b(integer(1),integer,[nodeid(pos(54,1,11,95,11,95))])),integer,[nodeid(pos(52,1,11,91,11,95))]),b(identifier(bit),integer,[nodeid(pos(55,1,11,101,11,103))])),couple(integer,integer),[nodeid(pos(51,1,11,91,11,103))])]))

normalise_store(T,NT) :-
    preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,EXPAND),
    normalise_store1(T,EXPAND,NT).
normalise_store1([],_Expand,[]).
normalise_store1([bind(Var,Val)|T],EXPAND,[bind(Var,NVal)|NT]) :-
    %print(normalise(Val)),nl,print(Var),nl,
    normalise_value_for_var(Var,EXPAND,Val,NVal),
    normalise_store1(T,EXPAND,NT). 

/* ----------------------- */

:- assert_pre(store:lookup_value_for_existing_id(Id,State,_Val),
                (type_check(Id,variable_id),list_skeleton(State),type_check(State,store))).
:- assert_post(store:lookup_value_for_existing_id(_Id,_State,_Val), true).

lookup_value_for_existing_id(Id,State,Res) :- 
 (safe_member(bind(Id,Val),State,lookup_value_for_existing_id)
   -> check_val(Val,Id,State),Res=Val
    ; print(lookup_failed(Id,State)),nl,Res=fail
  ).
  
:- assert_pre(store:lookup_value_for_existing_id(Id,LS,State,_Val),
                (type_check(Id,variable_id),type_check(LS,store),type_check(State,store),list_skeleton(LS),list_skeleton(State))).
:- assert_post(store:lookup_value_for_existing_id(_Id,_LS,_State,_Val), true).

lookup_value_for_existing_id(Id,LocalState,State,Val) :- 
 (safe_member(bind(Id,Val),LocalState,lookup_value_for_existing_id__4)
   -> check_val(Val,Id,LocalState)
    ; lookup_value_for_existing_id(Id,State,Val)).

lookup_value(Id,State,Val) :- 
 (safe_member(bind(Id,Val),State,lookup_value) -> check_val(Val,Id,State)
    ; fail
  ).
lookup_value(Id,LocalState,State,Val) :- 
 (safe_member(bind(Id,Val),LocalState,lookup_value__4)
   -> check_val(Val,Id,LocalState)
    ; lookup_value(Id,State,Val)).
  
get_id_value(Id,State,Val) :- safe_member(bind(Id,Val),State,get_id_value).  /* can be backtracked over */

%safe_member(X,T) :- safe_member(X,T,safe_member).
safe_member(_,[],_) :- !,fail.
safe_member(X,[H|T],Loc) :- !, (X=H ; safe_member(X,T,Loc)).
safe_member(X,T,Loc) :- add_internal_error('Argument not a list: ',safe_member(X,T,Loc)),fail.

check_val(X,V,State) :- 
   (X==term(undefined) -> 
     add_error(store,'Trying to read undefined variable: ',V:State),fail
     ; true).

:- assert_must_succeed(( store:lookup_and_delete_value(x,X,[bind(y,2),bind(x,3)],R,true,_WF,F),
  X==3, R==[bind(y,2)],F==true )).
:- assert_must_succeed(( store:lookup_and_delete_value(x,X,[bind(y,2)|T],R,true,_WF,F),
  T = [bind(x,3)],
  X==3, R==[bind(y,2)],F==true )).

lookup_and_delete_value(Var,Value,T,UT,ReportError,WF,Finished) :-
   lookup_and_delete_value1(T,Var,Value,UT,ReportError,WF,Finished).
:- use_module(kernel_waitflags,[add_abort_error_span/5]).
:- block lookup_and_delete_value1(-,?,?,?,?,?,?).
lookup_and_delete_value1([],Var,Value,[],ReportError,WF,Finished) :-
   (ReportError=report_error(WF) 
     ->  add_abort_error_span(missing_assignment_error,'Statement did not assign to variable: ',Var,unknown,WF),  /* lookup_and_delete is called only from get_results */
         Value = abort(no_value_for(Var))
     ;   Value = term(no_value_for(Var))
   ), Finished=true.
lookup_and_delete_value1([bind(Var2,Value2)|T],Var,Value,Res,ReportError,WF,Finished) :-
     lookup_and_delete_value2(Var2,Value2,T,Var,Value,Res,ReportError,WF,Finished).
:- block lookup_and_delete_value2(-,?,?,?,?,?,?,?,?).
lookup_and_delete_value2(Var2,Value2,T,Var,Value,Res,ReportError,WF,Finished) :-
   (Var=Var2
      -> Value=Value2, T=Res, Finished=true
      ;  Res=[bind(Var2,Value2)|UT],
          lookup_and_delete_value1(T,Var,Value,UT,ReportError,WF,Finished)
    ).

    
:- assert_must_succeed(( store:search_and_delete_value(x,22,X,[bind(y,2),bind(x,3)],R),
  X==3, R==[bind(y,2)] )).
:- assert_must_succeed(( store:search_and_delete_value(x,22,X,[bind(y,2)],R),
  X==22, R==[bind(y,2)] )).
:- assert_must_succeed(( store:search_and_delete_value(x,22,X,[bind(y,2)|T],R),
  T = [bind(x,3)],
  X==3, R==[bind(y,2)] )).

:- block search_and_delete_value(-,?,?,?,?).
search_and_delete_value(Var,OldVal,Value,InStore,OutStore) :-
  search_and_delete_value2(InStore,Var,OldVal,Value,OutStore).
  
:- block search_and_delete_value2(-,?,?,?,?).
search_and_delete_value2([],_Var,OldVal,Value,[]) :- 
   Value = OldVal.
search_and_delete_value2([bind(Var2,Value2)|T],Var,OldVal,Value,Res) :-
   search_and_delete_value3(Var2,Value2,T,Var,OldVal,Value,Res).

:- block search_and_delete_value3(-,?,?,?,?,?,?).
search_and_delete_value3(Var2,Value2,T,Var,OldVal,Value,Res) :- 
    (Var=Var2
      -> Value=Value2, T=Res
      ;  Res=[bind(Var2,Value2)|UT],
         search_and_delete_value2(T,Var,OldVal,Value,UT)
    ).

empty_state([]).
add_var_to_localstate(Var,Val,InState,[bind(Var,Val)|InState]).
    
:- assert_pre(store:set_up_localstate(Vars,Vals,In,_),
                (type_check(Vars,list(variable_id)),type_check(Vals,vlist(any)),type_check(In,store))).
:- assert_post(store:set_up_localstate(_,_,_,Out), type_check(Out,store)).
    
set_up_localstate([],[],S,S).
set_up_localstate([Var|T],[Val|FT],InS,[bind(VarID,Val)|OutS]) :- 
   /* add Var,Val to Local state at the beginning to properly handle nested
      blocks with same variable names */
   (Var=b(identifier(Id),_,_) -> %print(typed_identifier_for_set_up_localstate(Var)),nl,
                                 VarID = Id
                               ; VarID = Var),
   set_up_localstate(T,FT,InS,OutS).

set_up_undefined_localstate([],S,S).
set_up_undefined_localstate([Var|T],InS,[bind(Id,term(undefined))|OutS]) :-
   def_get_texpr_id(Var,Id),
   /* add Var,Val to Local state at the beginning to properly handle nested
      blocks with same variable names */
   set_up_undefined_localstate(T,InS,OutS).
   

:- assert_pre(store:delete_variables_from_state(Vars,In,_),
                (type_check(Vars,list(variable_id)),type_check(In,store))).
:- assert_post(store:delete_variables_from_state(_,_,Out), type_check(Out,store)).

delete_variables_from_state([],S,S).
delete_variables_from_state([Var|T],InState,OutState) :-
   delete_variable(Var, InState, S2),
   delete_variables_from_state(T,S2,OutState).
   
delete_variable(_,[],[]).
delete_variable(Var,[bind(V,Val)|T],Res) :-
  (Var==V
    ->  Res = T
    ;   Res = [bind(V,Val)|DT], delete_variable(Var,T,DT)
  ).
/* --------------------------------- */

% unused at the moment:
%:- assert_pre(store:copy_store_template(In,_),
%                (type_check(In,store))).
%:- assert_post(store:copy_store_template(_,Out), type_check(Out,store)).
%copy_store_template([],[]).
%copy_store_template([bind(Var,_)|T],[bind(Var,_)|CT]) :- copy_store_template(T,CT).


% copy_store_template copies a store and its values except some changing variables
% Two new stores are created: constants are not copied into the first new update state, but
% to the second new full state
copy_variable_store([],_,[],[]).
copy_variable_store([bind(Var,Old)|T],Changes,Updates,NewFullState) :-
    ( b_is_constant(Var) ->
        Updates = CT, NewFullState = [bind(Var,Old)|FT]
    ; safe_member(Var,Changes,copy_variable_store) ->
        Updates = [bind(Var,New)|CT], NewFullState = [bind(Var,New)|FT]
    ; otherwise ->
        Updates = CT, NewFullState = [bind(Var,Old)|FT]),
    copy_variable_store(T,Changes,CT,FT).


% gets the value of one variable out of the state and returns
% the remaining state without the variable
% - no checking of the variables value is performed
%   (in contrast to lookup_and_delete_value)
% - the variable name should be ground, the predicate cannot
%   be used to iterate over the state
% NOT USED !
%extract_variable([bind(Id,Value)|OutState],Id,Value,OutState) :- !.
%extract_variable([Bind|InState],Id,Value,[Bind|OutState]) :-
%	extract_variable(InState,Id,Value,OutState).

% insert_variables(+Store, -OldVars, ?NewVars, +InStore, -OutStore)
%  Copies all variable values from Store to InStore, the resulting
%  state is OutStore.
%  The list of copied variables is returned as OldVars, NewVars a list
%  of the same size which contains the new names of the variables in OutStore.
%  This predicate is used for copying the values of an after-state to a local
%  state, but adding a prime to all their names.
%  Works only with states that are represented by lists
insert_variables([], [], [], InStore, InStore).
insert_variables([bind(OldVar,Value)|Srest], [OldVar|Orest], [NewVar|Nrest],
		 InStore, OutStore) :-
	insert_variables(Srest,Orest,Nrest,[bind(NewVar,Value)|InStore], OutStore).

% NOT USED !
%create_substate(State,Variables,Substate) :-
%	empty_state(Empty),
%	create_substate2(Variables,State,Empty,Substate).
%create_substate2([],_State,Substate,Substate).
%create_substate2([Var|Vrest],State,In,Out) :-
%	( lookup_value(Var,State,Value) ->
%	    add_var_to_localstate(Var,Value,In,Inter)
%	; otherwise -> In = Inter),
%	create_substate2(Vrest,State,Inter,Out).

