% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(symmetry_marker,[precompile_marker_typing_info/0,
                           compute_marker_for_state/2,
                           type_allows_for_precise_marker/1,
                           hash_model_checking_imprecise/2,
                           non_symmetric_type/1]).
/* File: symmetry_marker.pl */
/* Created: 28/9/06  */


:- use_module(module_information).
:- module_info(group,symmetry).
:- module_info(description,'This module computes symmetry hash markers for B states (used for hash symmetry).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(debug).

:- use_module(library(lists)).

/* marker(Object,  Type, ValueAsTerm) */


:- use_module(b_global_sets,[is_unused_b_global_constant/3,
  b_global_deferred_set_with_card_gt1/1, b_global_set_with_potential_symmetry/1]).

:- use_module(error_manager).

type_allows_for_precise_marker(integer).
type_allows_for_precise_marker(string).
type_allows_for_precise_marker(boolean).
type_allows_for_precise_marker(set(global(_))) :- !. % ok, whether global set deferred or not 
type_allows_for_precise_marker(set(couple(X,Y))) :- !,
  ((Y=global(_),non_symmetric_type(X))  %  type_allows_for_precise_marker(Y) not ok; unless we are guaranteed to have a function
    -> true
    ;  (X=global(_),non_symmetric_type(Y)) % see comment above
     -> true
    ;  (non_symmetric_type(X), non_symmetric_type(Y)) -> true
    ;  print_fail(set(couple(X,Y)))).
type_allows_for_precise_marker(set(X)) :-
  (non_symmetric_type(X) -> true
   ;  print_fail(set(X))).
type_allows_for_precise_marker(seq(X)) :- !, type_allows_for_precise_marker(set(couple(integer,X))).
type_allows_for_precise_marker(global(_)).
type_allows_for_precise_marker(couple(X,Y)) :- !,type_allows_for_precise_marker(X),
  type_allows_for_precise_marker(Y).
type_allows_for_precise_marker(record(F)) :- !, fields_allows_for_precise_marker(F).

fields_allows_for_precise_marker([]).
fields_allows_for_precise_marker([field(_Name,Type)|T]) :-
   type_allows_for_precise_marker(Type),
   fields_allows_for_precise_marker(T).

:- use_module(preferences,[get_preference/2]).
hash_model_checking_imprecise(ID,BasicType) :- %print(check),nl,
   get_preference(symmetry_mode,hash),
   (b_get_machine_constants(TL) ; b_get_machine_variables(TL)),
   member(b(identifier(ID),BasicType,_),TL),
   \+ type_allows_for_precise_marker(BasicType).
   

%:- use_module(translate,[pretty_type/2]).
print_fail(_T) :- 
        %print('*** type may lead to imprecise hash marker:'),
        %translate:pretty_type(T,S),print(S),nl,
        fail.
% a type which does not contain any deferred set elements
non_symmetric_type(integer) :- !.
non_symmetric_type(boolean) :- !.
non_symmetric_type(string) :- !.
non_symmetric_type(set(X)) :- !, non_symmetric_type(X).
non_symmetric_type(couple(X,Y)) :- !,non_symmetric_type(X), non_symmetric_type(Y).
non_symmetric_type(global(X)) :- !, \+ b_global_set_with_potential_symmetry(X).
non_symmetric_type(seq(X)) :- !, non_symmetric_type(X).
non_symmetric_type(record(F)) :- !, non_symmetric_fields(F).
non_symmetric_type(freetype(_)) :- !.
non_symmetric_type(X) :- add_error_fail(non_symmetric_type,'Unknown type: ',X).

non_symmetric_fields([]).
non_symmetric_fields([field(_Name,Type)|T]) :- non_symmetric_type(Type),
  non_symmetric_fields(T).
 
/* first a predicate to count the occurrences of anonymous set elements */

/* possible improvement: count nr of occurences for each variable */

:- use_module(bmachine,[b_get_machine_constants/1, b_get_machine_variables/1]).
:- volatile non_symmetric_var/2.
:- dynamic non_symmetric_var/2.
precompile_marker_typing_info :-
   retractall(non_symmetric_var(_,_)),
   bmachine:b_get_machine_variables(TypedVarList),
   bmachine:b_get_machine_constants(TypedCstList),
   append(TypedCstList,TypedVarList,List),
   analyse_typed_varlist(List).

analyse_typed_varlist([]).
analyse_typed_varlist([b(identifier(ID),Type,_)|T]) :-
   (non_symmetric_type(Type)
     -> assert(non_symmetric_var(ID,Type))
        %% ,print(non_sym(ID,Type)),nl
     ; true),
   analyse_typed_varlist(T).
 
 
%:- dynamic count/4.


count_state(State) :- empty_avl(E),
   %nl,translate:print_bstate(State),nl,
    count_env(State,E,Out),
   %portray_avl(Out),nl,
    compute_deferred_set_el_markers(Out).
 
count_env([],I,O) :- !, O=I.
count_env([bind(Var,Val)|T],In,Out) :- !,
   (non_symmetric_var(Var,_) -> In=I1
     ; %print(counting(Var)),nl,
       count_obj(Val,Var,In,I1)),
   count_env(T,I1,Out).
count_env(E,I,O) :- add_error(symmetry_marker,'Illegal env: ',E), O=I.

:- use_module(library(avl)).
/* IDEA: instead of just storing variable Var, store entire path to 
   occurrence of fd(Nr,T) */
count_obj(fd(Nr,T),PathToObject,In,Out) :-  !, %print(fd(Nr,T)),
  ((b_global_deferred_set_with_card_gt1(T) ;
    b_global_sets:is_unused_b_global_constant(T,Nr,_) /* TO DO: only makes sense if more than one unused global constant */
    ) 
     -> %add_count(PathToObject,Nr,T,In,Out)
        add_avl_count(In,usage_count(Nr,T,PathToObject),Out)
     ;  Out=In
   ).
count_obj((X,Y),PathToObject,In,Out) :- !, count_pair(X,Y,PathToObject,In,Out).
count_obj([X|Y],PathToObject,In,Out) :- !,
   (could_contain_deferred_set_elements(X)
     -> count_obj_list([X|Y],PathToObject,In,Out)  /* only count set if there are deferred set elements inside */
      ; Out=In
   ).
count_obj(avl_set(A),PathToObject,In,Out) :- !,
   (avl_could_contain_deferred_set_element(A)
     -> count_avl(A,el(PathToObject),In,Out)
     ;  Out=In
   ).
%count_obj([],_,O,O).
count_obj(_,_,O,O).
% no need to have base case for empty avl; as empty set represented by []


count_avl(empty,_PathToObject) --> [].
count_avl(node(Key,true,_,L,R),PathToObject) -->
   count_avl(L,PathToObject), count_obj(Key,PathToObject),
   count_avl(R,PathToObject).

count_pair(X,Y,PathToObject,In,Out) :- not_symmetric(Y),!,
   count_obj(X,mapped_to(PathToObject,Y),In,Out).
count_pair(X,Y,PathToObject,In,Out) :- not_symmetric(X),!,
   count_obj(Y,mapped_from(X,PathToObject),In,Out).
count_pair(X,X,PathToObject,In,Out) :- !,  /* is this correct ??
    should we only do it for simple X, what about sets: should
    work if normalised ?? */
   count_obj(X,leftright(PathToObject),In,Out).
count_pair(X,Y,PathToObject,In,Out) :- !,
   count_obj(X,left(PathToObject),In,I1),
   count_obj(Y,right(PathToObject),I1,Out).


could_contain_deferred_set_elements([]).
could_contain_deferred_set_elements([H|_]) :- could_contain_deferred_set_elements(H).
could_contain_deferred_set_elements(fd(_Nr,_T)). 
%b_global_deferred_set_with_card_gt1(T) ;       b_global_sets:is_unused_b_global_constant(T,Nr,_).
could_contain_deferred_set_elements((A,B)) :-
  (could_contain_deferred_set_elements(A) -> true ; could_contain_deferred_set_elements(B)).
could_contain_deferred_set_elements(avl_set(A)) :-  avl_could_contain_deferred_set_element(A).

avl_could_contain_deferred_set_element(node(X,_,_,_L,_R)) :- /* first underscore should be true */
  could_contain_deferred_set_elements(X).
%an_avl_element(node(X,_,_,_L,_R),X).  /* first underscore should be true */

count_obj_list([],_,O,O).
count_obj_list([X|Y],PathToObject,In,Out) :- 
   count_obj(X,el(PathToObject),In,I1), 
   count_obj(Y,PathToObject,I1,Out).
   
/* not_symmetric: check if object has symmetric permutations */
not_symmetric([]).
not_symmetric((X,Y)) :- not_symmetric(X), not_symmetric(Y).
not_symmetric(pred_false /* bool_false */).
not_symmetric(pred_true /* bool_true */).
not_symmetric(string(_)).
not_symmetric(fd(Nr,T)) :-
  \+ b_global_deferred_set_with_card_gt1(T),
  \+ b_global_sets:is_unused_b_global_constant(T,Nr,_).  /* TO DO: only makes sense if more than one unused global constant */
not_symmetric(int(_)).
/* add same for sets ?? */


      %  add_avl_count(usage_count(Nr,T,PathToObject),In,Out)
add_avl_count(AVL,Key,NewAVL) :-  %print(add_count(Key)),nl,
   (avl_fetch(Key,AVL,OldCount)
     -> NewCount is OldCount+1,
        avl_change(Key,AVL,_,NewAVL,NewCount)
     ;  avl_store(Key,AVL,1,NewAVL)
    ). %, print(ok(NewCount)),nl.


/* 
add_count(Var,Nr,T) :-
  (retract(count(Nr,T,Var,C1))
    -> (C is C1+1, assert(count(Nr,T,Var,C)))
    ;  assert(count(Nr,T,Var,1))
  ). */
  
:- volatile deferred_set_el_marker/3.
:- dynamic deferred_set_el_marker/3.

/* compute individual markers for deferred set elements by hashing
   classical path information using sha */

/*
compute_deferred_set_el_markers(CountAVL) :-
   retractall(deferred_set_el_marker(_,_,_)),
   avl_generate_el_markers(CountAVL,'$DUMMY',-1,[],OutType,OutNr,OutCount),
      %%  print(last_marker(OutType,OutNr,OutCount)),nl, %%
   assert_def_marker(OutType,OutNr,OutCount).

avl_generate_el_markers(empty,CurType,CurElNr,CurCount,CurType,CurElNr,CurCount).
avl_generate_el_markers(node(usage_count(Nr,T,Path),Count,_,L,R),CurType,CurElNr,CurCount,
                               OutT,OutNr,OutCount) :-
   avl_generate_el_markers(L,CurType,CurElNr,CurCount,T1,N1,C1),
   ((N1=Nr,T1=T) % we are still treating the same deferred set element fd(T,Nr) 
     -> C2=[count(Path,Count)|C1] % add cur path to list of paths for fd(T,Nr) 
     ;  assert_def_marker(T1,N1,C1), % assert cur paths for fd(T1,N1) 
        % We could use instead: avl_store(fd(T1,N1),AVL,el(T1,C1),NewAVL)
        %print(marker(T1,N1,C1)),nl, %%
        C2 = [count(Path,Count)]  % start fresh; collecting paths for fd(T,Nr) 
   ), avl_generate_el_markers(R,T,Nr,C2,OutT,OutNr,OutCount).
 
:- use_module(extension('probhash/probhash')).
assert_def_marker(T1,N1,C1) :- raw_sha_hash(el(T1,C1),MHash),
 assert(deferred_set_el_marker(T1,N1,MHash)).
 */

/*
 an attempt at not storing paths but just sort the fd-elements according to path
 information and assign them a number: without adding id_couple below in marker/3 we get collisions, e.g, for
 ./probcli examples/B/SymmetryReduction/PhilRing.mch -mc 1000 -p DEFAULT_SETSIZE 3 -p SYMMETRY_MODE hash -cc 13 47 
*/

compute_deferred_set_el_markers(CountAVL) :-
   avl_to_list(CountAVL,AList),
   %print(AList),nl,
   merge_paths(AList,BList),
   sort(BList,SBList),
   %print(merged(SBList)),nl,
   retractall(deferred_set_el_marker(_,_,_)),
   assert_markers(SBList,0),!.

% merge all the information for same deferred set elements (Nr,Type)
merge_paths([],[]).
merge_paths([usage_count(Nr,Type,Path)-Count|Tail],Res) :-
  merge_paths(Tail,Nr,Type,[count(Path,Count)],Res).
  
merge_paths([],Nr,Type,MarkerPaths,[marker(Type,MarkerPaths,Nr)]). % :- print(marker(Type,MarkerPaths,Nr)),nl.
merge_paths([usage_count(Nr,Type,Path)-Count|Tail],Nr,Type,M,Res) :- !,
  merge_paths(Tail,Nr,Type,[count(Path,Count)|M],Res).
merge_paths([usage_count(Nr,Type,Path)-Count|Tail],Nr2,T2,M,[marker(T2,M,Nr2)|Res]) :-
  % print(marker(T2,Nr2,M)),nl,
  merge_paths(Tail,Nr,Type,[count(Path,Count)],Res).

:- use_module(extension('probhash/probhash')).
assert_markers([],_).
assert_markers([marker(Type,Paths,Nr)|T],ID) :-
  get_same_markers(T,Type,Paths,Same,TT,NrSame), % get all other elements with same Paths information
  ID1 is ID+1,
  (NrSame=0 -> assert(deferred_set_el_marker(Type,Nr,fdunique(ID1,Type))) 
   ; assert_same_markers([Nr|Same],Type,ID1,NrSame)),
  assert_markers(TT,ID1).

get_same_markers([marker(Type,Paths,Nr)|T],Type,Paths,[Nr|SM],TT,ResNr) :- !,
  get_same_markers(T,Type,Paths,SM,TT,SNr), ResNr is SNr+1.
get_same_markers(T,_,_,[],T,0).


assert_same_markers([],_,_,_).
assert_same_markers([Nr|T],Type,ID,NrSame) :- 
  assert(deferred_set_el_marker(Type,Nr,fdsame(ID,Type,NrSame))),
  assert_same_markers(T,Type,ID,NrSame).


/* */


% -------------------------------------------------------


/* marker(+ProBOBject, -TypeOfObject, -HashTerm) */  

  
%:- use_module(b_global_sets,[all_elements_of_type/2]).
:- use_module(custom_explicit_sets).

/* missing: closure, records,... */
marker(closure(P,T,B),set(any),closure(P,T,B)).
marker(global_set(GS),set(Type),Local) :-
    marker_explicit_set(global_set(GS),set(Type),Local).
marker(avl_set(S),set(Type),SLocal) :- % marker_explicit_set(avl_set(S),set(Type),Val,Local).
   (avl_marker(S, Type,Local) -> true ; print(failed_avl_marker(S)),nl,fail),
   sort_markers(Local,SLocal).
marker(closure_x(P,T,B,E),set(Type),Local) :- 
    marker_explicit_set(closure_x(P,T,B,E),set(Type),Local).
marker(int(N),  integer,  N). % int(N)
marker(string(S),  string, S). % string(S)).
marker(fd(Nr,T), global(T), Term) :-
   (deferred_set_el_marker(T,Nr,Term)
     -> true
     ;  (b_global_deferred_set_with_card_gt1(T)
          -> print(not_counted(Nr,T)),nl ; true),
         Term=fd(Nr,T)  /*should not happen for deferred sets */
   ).
marker(pred_false /* bool_false */, boolean, pred_false).
marker(pred_true /* bool_true */, boolean, pred_true).
marker((X,Y), couple(T1,T2), Mark) :-
   (X==Y -> Mark = id_couple(L1), marker(X,T1,L1) % added 7/9/2011 by leuschel
    ;  Mark = couple(L1,L2), marker(X,T1,L1),marker(Y,T2,L2)).
marker([],       set(_), []).
marker([H|T], set(Type), Local) :-
   l_marker([H|T],Type,LL),
   sort_markers(LL,Local).


marker_explicit_set(ES,set(Type),Local) :-
    expand_custom_set_to_list_now(ES,Res),
    l_marker(Res,Type,LL),
    sort_markers(LL,Local).


l_marker([], _Type, []).
l_marker([H|T], Type, [LH|LT]) :-
   marker(H,Type,LH),
   l_marker(T,Type,LT).
   
avl_marker(Avl, Type,Local) :- avl_domain(Avl,List),
  l_marker(List,Type,Local).

  
   
   /*  example queries:
      use_module(symmetry_marker),
      symmetry_marker:compute_marker_for_state([bind(tr,[(fd(1,'Session'),fd(1,'Session')),(fd(1,'Session'),fd(2,'Session'))]),
       bind(inv,[fd(1,'Session')]),bind(i,fd(1,'Session')),
       bind(reach,[fd(1,'Session'),fd(2,'Session')]),bind(rac,[fd(1,'Session')])],M).
       
       symmetry_marker:nodes_in_graph([(fd(1,'Session'),fd(1,'Session')),(fd(1,'Session'),fd(2,'Session'))],N).
       
       use_module(symmetry_marker),symmetry_marker:cm.
       
       use_module(symmetry_marker),
       symmetry_marker:count_state([bind(tr,[(fd(1,'Session'),fd(1,'Session')),(fd(1,'Session'),fd(2,'Session'))]),
       bind(inv,[fd(1,'Session')]),bind(i,fd(1,'Session')),
       bind(reach,[fd(1,'Session'),fd(2,'Session')]),bind(rac,[fd(1,'Session')])]),
       symmetry_marker:count(X,Y,Z).
       */
   

:- use_module(library(samsort)).

sort_markers(X,SX) :- samsort(X,SX). % sort without duplicate removal

  

:- use_module(specfile,[expand_const_and_vars_to_full_store/2]).

:- use_module(self_check).
:- assert_must_succeed((symmetry_marker:compute_marker_for_state([bind(x,fd(1,'Name')),bind(y,fd(2,'Name'))],R1),
 symmetry_marker:compute_marker_for_state([bind(x,fd(2,'Name')),bind(y,fd(1,'Name'))],R2),R1==R2)).

compute_marker_for_state(State,Marker) :-
  (State = concrete_constants(BState)
    -> Marker = concrete_constants(BMarker)
    ;  (State = csp_and_b(CSPState,BState)
        -> Marker = csp_and_b(CSPState,BMarker)
        ;  (State=[] ; State = [_|_] ; State = const_and_vars(_,_) ),
            State=BState, Marker=BMarker
       )
  ),
   expand_const_and_vars_to_full_store(BState,FullBState),
   count_state(FullBState), /* first count the number of occurrences
    of every symmetric element in the State; this number will
    be used as the marker for that element */
    marker_for_state(FullBState,BMarker).
 %   ,raw_sha_hash(BMarker0,BMarker). % , print(marker_for_state(FullBState,BMarker)),nl,nl.

/* compute a marker for a given state */
marker_for_state([],[]).
marker_for_state([bind(Var,Val)|T],[M|TM]) :-
  (non_symmetric_var(Var,_) -> TermMarker=Val
     ; marker(Val, _Type,TermMarker)),
  M=mark(Var,TermMarker),
  marker_for_state(T,TM).


% TESTING CODE

/*
:- dynamic marker/2.
% symmetry_marker:cm.
% symmetry_marker:marker(10,M1), symmetry_marker:marker(18,M2), M1=M2.

:- public cm/0.
cm :-
  print('Computing markers for state space'),nl,
  retractall(marker(_,_)),
  cms(ID, M),
  assert(marker(ID,M)),
  fail.
cm :-
  marker(ID,Marker),
  marker(ID2,Marker), ID2 > ID,
  print(same_marker(ID,ID2)),nl,
  fail.
cm :- print('Finished'),nl.

:- use_module(state_space,[visited_expression/2]).
cms(ID, M) :-
  state_space:visited_expression(ID,State),
  compute_marker_for_state(State,M),
  print(marker(ID,State,M)),nl.
  

:- public test_sym/0.
test_sym :- state_space:visited_expression(ID,State),
  print(ID), print(' : '), 
  sym_mark_for_state(State).

sym_mark_for_state(State) :-print(State),nl,
  member(bind(Var,Val),State),
  print(Var), print(' = '),
  marker(Val, T,L),
  print(marker(Val,T,L)),
  print(' '),
  nl,
  fail.
sym_mark_for_state(_) :- nl.
*/
