% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% tcltk_interface.pl (was btracer.pl and btracer2.pl)


prob_use_module(X) :- load_files(X,[if(changed),load_type(source),compilation_mode(compile)]).
prob_use_module(X,I) :-
  load_files(X,[if(changed),load_type(source),compilation_mode(compile),imports(I)]).

/* load all necessary modules */
:- use_module(library(random)).
:- use_module(library(lists)).
:- use_module(library(system)).

:- use_module(library(ordsets)).
:- use_module(library(avl)).
:- use_module(library(process)).

:- prob_use_module(typechecker).
:- prob_use_module(self_check).
:- prob_use_module(kernel_waitflags).
                           
:- prob_use_module(specfile).
:- prob_use_module(translate).
:- prob_use_module(tools).
:- prob_use_module(b_enumerate).
:- prob_use_module(b_state_model_check).
:- prob_use_module(bmachine).


:- prob_use_module(debug).
:- prob_use_module(preferences).
:- prob_use_module(version).
:- prob_use_module(b_interpreter).
:- prob_use_module(predicate_debugger).
:- prob_use_module(b_trace_checking).
:- prob_use_module(gensym).
:- use_module(delay).
:- prob_use_module(bsyntaxtree).

:- prob_use_module(kernel_objects).
:- prob_use_module(bsets_clp).

:- prob_use_module(state_space).


:- prob_use_module(b_compiler).
:- prob_use_module(error_manager).
:- prob_use_module(model_checker).

:- prob_use_module(visualize_graph).
:- prob_use_module(state_custom_dot_graph).
:- prob_use_module(fdr_csp_generator).
:- prob_use_module(coverage_statistics).
:- prob_use_module(succeed_max).

:- prob_use_module(testcase_generator).

:- use_module('proz/proz',[open_proz_file/2]).

:- prob_use_module(reduce_graph_state_space).
:- prob_use_module(flow).

:- prob_use_module(ltl).
:- prob_use_module('kodkod/kodkod').
:- prob_use_module(graph_canon).
:- prob_use_module('promela/promela_ncprinter').

:- prob_use_module(sap).
:- prob_use_module(plugins,[notify_new_transition_to_plugin/5]).

:- prob_use_module(system_call,[generate_temporary_cspm_file/2, get_writable_compiled_filename/3]).

:- prob_use_module(bvisual2).
:- prob_use_module(b_show_history).
:- prob_use_module(enabling_analysis).

:- use_module(probcspsrc(haskell_csp),[parse_and_load_cspm_file/1, cspm_file_add_remove_unicode/3,
                                       get_csp_assertions_as_string/2,get_csp_processes/1,
                                       cspm_eval_expression_from_csp_file/5, cspm_eval_expression/3,
                                       parse_and_load_cspm_file_into_specific_pl_file/2, load_cspm_pl_file/1,
					evaluate_csp_expression/2,evaluate_csp_expression/3,
					parse_single_csp_expression_file/3]).
:- use_module(probcspsrc(slicer_csp),[slice_from_program_point/14]).

:- use_module(value_persistance, [save_constants/1,add_new_transitions_to_cache/1]).

/* main program 
 * ============
 */

 % use module_info/3 because we are not in a module here
:- use_module(module_information).
:- module_info(prob_cli,group,infrastructure).
:- module_info(prob_cli,description,'Interface between the Tcl/Tk GUI, the CLI and the internal modules.').
:- module_info(prob_cli,revision,'$Rev$').
:- module_info(prob_cli,lastchanged,'$LastChangedDate$').


   

/* -------------------------------------------------------------------- */

%tcltk_find_untyped_vars(Vs) :- find_untyped_vars(Vs), Vs \== [].

%tcltk_find_untyped_consts(Vs) :- find_untyped_consts(Vs), Vs \== [].

tcltk_find_max_reached_node :- tcltk_find_max_reached_node(_).
tcltk_find_max_reached_node(ID) :- max_reached_for_node(ID).

tcltk_current_node_has_no_real_transition :- 
  \+ tcltk_current_node_has_real_transition.
tcltk_current_node_has_real_transition :-
  current_state_id(CurID),
  transition(CurID,Trans,_),
  (CurID \= root ; Trans \= '$partial_setup_constants'),!.
tcltk_current_node_has_partial_transition :-
  current_state_id(CurID),
  transition(CurID,'$partial_setup_constants',_),!.

/* --------------------------------------------------------------------- */
tcltk_goto_state(ActionAsTerm,StateID) :-
  ~~mnf(translate:translate_event(ActionAsTerm,ActionAsString)),
  tcltk_goto_state(ActionAsTerm,ActionAsString,StateID).
tcltk_goto_state(ActionAsTerm,ActionAsString,StateID) :-
  (var(StateID) -> print_message(var_state(tcltk_goto_state(ActionAsString,StateID)))  ;true),
   (visited_expression(StateID,_State)
    -> true
    ;  print(state_does_not_exist(tcltk_goto_state(ActionAsString,StateID))),nl,fail
    ),
   (retract(current_state_id(CurID))->true),
   assert(current_state_id(StateID)),
   (transition(CurID,ActionAsTerm,OpID,StateID)
    -> /* ok: proper transition */
       add_to_op_trace_ids(OpID),
       add_id_to_history(CurID)
     ; (CurID=StateID
        -> true
        ;  debug_print(9,jumping(CurID,ActionAsString,StateID)),debug_nl(9),
           add_to_op_trace_ids(jump(StateID)),
           add_id_to_history(CurID)
       )
    ).

add_id_to_history(CurID) :-
   (retract(history(History))->true),
   assert(history([CurID|History])), retractall(forward_history(_)).

initialise_operation_not_yet_covered :- retractall(operation_not_yet_covered(_)),
   b_or_z_mode,
   b_top_level_operation(Name),
  % b_get_machine_operation(Name,_,Par,_),  length(Par,Arity), functor(Op,Name,Arity), 
   % Note: no '-->' added 
   assert(operation_not_yet_covered(Name)), 
   debug_println(operation_not_yet_covered(Name)),
   fail.
   /* Missing: treat operations with return values */
initialise_operation_not_yet_covered.


tcltk_add_new_transition(FromID,Action,ToID,ToTempl,TransInfo) :-
    tcltk_add_new_transition_transid(FromID,Action,ToID,ToTempl,TransInfo,_TransId).

tcltk_add_new_transition_transid(FromID,Action,ToID,ToTempl,TransInfo,TransId) :-
   get_id_of_node_and_add_if_required(ToTempl,ToID,Exists,FromID),
   %% print_message(tcltk_add_new_transition(FromID,Action,ToID,Exists)), %%
   ((Exists=exists,transition(FromID,Action,TransId,ToID)) 
     -> true
     ;  (preferences:preference(forget_state_space,true)
          -> (any_transition(FromID,_,_) -> true
               ; store_transition(FromID,'$NO_DEADLOCK',FromID,TransId)) /* just add dummy transition */
          ;  store_transition(FromID,Action,ToID,TransId),
			 (store_transition_infos(TransInfo,TransId) -> true
					; add_error(tcltk_interface,'storing transition info failed', TransInfo))
		),
	    add_cover(Action,OpName),
	    animation_mode(Mode),
        ((b_or_z_mode(Mode), FromID \= ToID, not_all_transitions_added(ToID))
            -> add_additional_destination_infos(FromID,OpName,ToID)
             ; true
         ),
        notify_new_transition_to_plugin(Mode,FromID,TransId,ToID,Exists)
   ).

add_cover('-->'(Action1,_),OpName) :- !,add_cover_aux(Action1,OpName).
add_cover(Action1,OpName) :- !,add_cover_aux(Action1,OpName).
add_cover_aux(Action1,OpName) :-  functor(Action1,OpName,_),
         (retract(operation_not_yet_covered(OpName))
           -> (preferences:get_preference(provide_trace_information,true)
                -> print(cover(OpName)),nl ; true),
              (operation_not_yet_covered(_) -> true ; nl,print('ALL OPERATIONS COVERED'),nl,nl)
           ; true
         ).


add_additional_destination_infos(FromID,ActionName,ToID) :- 
          preference(try_operation_reuse,true), 
          animation_mode(b),
          preference(symmetry_mode,MODE), (MODE=off ; MODE=flood),
          % TO DO: can we enable this trick in symmetry mode ?
          bmachine:reuse_operation_effect(ActionName,OtherAction),
          \+ reuse_operation(ToID,OtherAction,_,_),
          assert(reuse_operation(ToID,OtherAction,FromID,ActionName)),
      %%    print(reuse_operation(ToID,OtherAction,FromID,ActionName)),nl, %%
          fail.
add_additional_destination_infos(FromID,ActionName,ToID) :- %print(add_dest(FromID,ActionName,ToID)),nl,
          ((b_specialized_invariant_for_op(ActionName,_),
            \+ invariant_not_yet_checked(FromID),
            \+ invariant_violated(FromID)
            )
            -> 
              (retract(specialized_inv(ToID,OldAVL)) -> true ; empty_avl(OldAVL)),
               avl_store(ActionName,OldAVL,nothing,NewAVL),
               % TO DO: provide a special value if an operation preserves the entire invariant ??
               %print(specialized_inv(ToID,NewAVL)),nl,
               assert(specialized_inv(ToID,NewAVL))
             ; true
           ).
/*
Store names of incoming transitions in an AVL tree
Used to calculate invariant in presence of proof information
*/



/* check whether we have visited a node already: if so return
 id of that node; otherwise generate new id and add node */
get_id_of_node_and_add_if_required(State,ID,Res,FromID) :-
  preferences:preference(symmetry_mode,SymMode),
 % print(computing_hash),nl,
  compute_hash(SymMode,State,Hash,Marker),
 % print_message(hash(Hash)),
  ( expression_already_visited(SymMode,State,Hash,Marker,ID)
   -> Res = exists
   ;  Res = new,
      gen_new_state_id(ID), /* gensym('node',ID),*/
      add_new_visited_expression(ID,Hash,State,FromID,SymMode) ,
      assert(hash_to_id(Hash,ID)),
      (SymMode=hash -> assert_id_to_marker(ID,Marker) ; true)
   ).

expression_already_visited(SymMode,State,Hash,Marker,ID) :-
  hash_to_id(Hash,ID),
  ((preferences:preference(ignore_hash_collisions,true) ; SymMode=nauty)
   -> true
   ;  ((SymMode=hash,Marker \= none)
        ->  id_to_marker(ID,Marker) /* just check Markers, not state */
        ;   (visited_expression(ID,State)
             -> true  /* warning: may instantiate State if not ground */
              ; print(hash_collision(Hash,ID)),nl,fail)
      )
  ).

assert_id_to_marker(ID,Marker) :-
    ((Marker = none;preferences:preference(ignore_hash_collisions,true))
        -> true  % don't assert marker info if we ignore hash collisions anyway
        ; assert(id_to_marker(ID,Marker))
    ).

  

:- use_module(model_checker,[set_depth_breadth_first_mode/1]).


tcltk_set_dbf_mode(Nr) :- dbf_modes(Nr,Mode,_),!,
   set_depth_breadth_first_mode(Mode).
tcltk_set_dbf_mode(Nr) :- add_error(tcltk_set_dbf_mode,'Unknown mode: ',Nr).

tcltk_dbf_modes(list(Names)) :-  findall(N,dbf_modes(_,_,N),Names).

/* ---------------------------------- */
/*     add_new_visited_expression/2   */
/* ---------------------------------- */

add_new_visited_expression(ToID,Hash,ToState,FromID,SymMode) :- 
   state_space_add(ToID,ToState),
   add_open_node(ToID,Hash,FromID),
   (SymMode=flood
     -> add_permutations(ToID,ToState) ; true).

:- use_module(model_checker,[last_step_breadth_first_in_mixed_mode/0]).
add_open_node(NodeID,Hash,FromID) :- %print_message(aa(NodeID)),
   (is_not_interesting(NodeID) ->
      asserta(not_interesting(NodeID))
    ; add_open_node2(NodeID,Hash,FromID)
    ).
 :- use_module(library(terms),[term_size/2]).
add_open_node2(NodeID,_Hash,_FromID) :- %print_message(aa(NodeID)),
   %(b_or_z_mode -> assert(not_invariant_checked(NodeID)) ; true), % is now added when we pop an open node
   ((last_step_breadth_first_in_mixed_mode , \+ random(1,8,1) ) 
          -> add_id_at_end(NodeID) /* do not add successor of BF step at top */
                                   /* avoids degeneration into BF when large branching factor */
          ;   add_id_at_front(NodeID) % for default DF/BF behaviour
          
             % add_id_with_weight(NodeID,_Hash) % use hash value
             
             % add_id_random(NodeID)   % use a random number
            
             % visited_expression(NodeID,T), term_size(T,Sz),
             % add_id_with_weight(NodeID,Sz)  % use term_size as heuristic
            
             % add_id_to_process(NodeID,_Hash,_FromID) % for out_degree as heuristic
            
            %add_id_with_heuristic(NodeID)  % for using custom heuristic if possible
   ).
 
 
 add_id_with_heuristic(NodeID) :- %print(adding(NodeID)),nl,
   bmachine:b_get_machine_heuristic_function(HFunction),
   user:visited_expression(NodeID,State),
   % print(cur_expr(State)),nl,
   specfile:state_corresponds_to_initialised_b_machine(State,BState),!,
   (b_interpreter:b_compute_expression(HFunction,[],BState,HRes)
    ->  %print_message(heuristic(NodeID,HRes)),
	   (HRes = int(H)
		-> add_id_with_weight(NodeID,H) % , print(added_id_with_weight(NodeID,H)),nl
		;  print_message('*** Illegal HEURISTIC_FUNCTION result: '), print_message(H),
		   add_id_at_front(NodeID)
	   )
	 ; print_message(heuristic_function_failed(NodeID,BState)),
	   add_id_at_front(NodeID)
    ).
add_id_with_heuristic(NodeID) :- add_id_at_front(NodeID).


is_not_interesting(NodeID) :-
    b_get_machine_searchscope(Scope),
    \+ eclipse_interface:test_boolean_expression_in_node(NodeID,Scope).

tcltk_mark_current_node_to_be_recomputed_wo_timeout :-
  current_state_id(CurID), print(recomputing(CurID)),nl,
  tcltk_mark_node_for_recomputed(CurID),
   (use_no_timeout(CurID) -> true ; assert(use_no_timeout(CurID))).
tcltk_mark_node_for_recomputed(ID) :-
  (not_all_transitions_added(ID) -> true
    ; add_id_at_front(ID)),
  retractall(max_reached_for_node(ID)),
  retractall(time_out_for_node(ID,_)),
  (transition(ID,'$partial_setup_constants',ID2)
     -> delete_node(ID2) ; true),
  retractall(state_space:transition(ID,'$partial_setup_constants',_,_)).

transitions_computed_for_node(ID) :-
  \+(not_all_transitions_added(ID)),  \+ is_not_interesting(ID).


  
  

:- use_module(state_permuter).
add_permutations(ID,Expression) :- %print(add_permutations(ID,Expression)),nl,
  permute_state(Expression,Perm), Perm \= Expression,
  generate_new_flood_id(Perm,PermID),
  state_space_add(PermID,Perm),
  (preferences:preference(forget_state_space,true)
          -> true ; store_transition(PermID,'*permute*',ID,_)),
  fail.
add_permutations(_,_).
  
generate_new_flood_id(State,ID) :-  
  compute_hash(flood,State,Hash,Marker),
  \+ expression_already_visited(flood,State,Hash,Marker,ID),
  gen_new_state_id(ID),
  assert(hash_to_id(Hash,ID)).
  

  
:- use_module(symmetry_marker,[compute_marker_for_state/2]).
:- use_module(hashing).


/* compute_hash(State, HashValue, Marker/SecondHashTerm) */
compute_hash(State,Hash,Marker) :- preferences:preference(symmetry_mode,SymMode),
  compute_hash(SymMode,State,Hash,Marker).

%:- use_module(extension('probhash/probhash')).

compute_hash(hash,State,Hash,Marker) :- 
  (symmetry_marker:compute_marker_for_state(State,HMarker)
    -> HMarker=Marker
      %raw_sha_hash(HMarker,Marker)
     ; print(compute_marker_failed(State)),nl,fail),
  my_term_hash(Marker,Hash),!.
compute_hash(nauty,State,Hash,Marker) :- !,
  compute_hash(off,State,EnvHash,_),
  ((hash_to_nauty_id(EnvHash,NautyID), hash_to_id(NautyID,RealID),
    visited_expression(RealID,State))
    -> %% print(exact_state_exists(EnvHash,NautyID,State)),nl, %%
      % the exact same state already exists; no need to compute normal form
       Hash=NautyID
   ;   (graph_canon:add_state_graph_to_nauty(State,Exists)
        -> true
        ; add_error(compute_hash,'Failed: ',add_state_graph_to_nauty(State,Exists), Exists=1)
        ),
        (Exists>0 -> Hash=Exists
             ; /* new node */ Hash is -Exists, assert(hash_to_nauty_id(EnvHash,Hash)))
   ),
  % print(nauty_add(Exists,Hash,State)),nl,
  Marker=none. 
% my_term_hash now uses if_var(ignore) option of SICStus 4.1; no longer need to check ground
compute_hash(_, Env,Hash,none) :- my_term_hash(Env,Hash),!.


% --------------------------------------------

hash_collisions(N) :-
  findall((ID,ID2), (hash_to_id(Hash,ID),
                     hash_to_id(Hash,ID2), ID2 \= root, (ID=root->true; ID2>ID)),L),
  length(L,N).
hash_collision(ID,ID2,S1,S2) :-
  hash_to_id(Hash,ID),
  hash_to_id(Hash,ID2), ID2 \= root, (ID=root->true; ID2>ID),
  visited_expression(ID,S1),
  %print((ID,S1)),nl,
  visited_expression(ID2,S2),
  %print((ID2,S2)),nl,
  nl.
 
 :- dynamic hash_collision/2. % hash_collision(NrStatesWithSameHash, NrOfSuchClusters)
 analyze_hash_collisions :- retractall(hash_collision(_,_)),
    print('Analyzing Hash Collisions'),nl,
    hash_to_id(Hash,ID),
    \+ (hash_to_id(Hash,ID2), ID2 @< ID),
    findall(X,hash_to_id(Hash,X),L),
    length(L,Nr),
    (retract(hash_collision(Nr,C))
      -> (C1 is C+1, assert(hash_collision(Nr,C1)))
      ;  assert(hash_collision(Nr,1))
    ),fail.
analyze_hash_collisions :-
    print('Hash Collision Distribution (1=no collision): '),
   findall((Nr,X),hash_collision(Nr,X),Coll),
   sort(Coll,SC),
   print(SC),nl,
   (hash_collision(ID,ID2,S1,S2)  ->
      (print('SAMPLE COLLISION: '),nl,
       print(ID),nl,print(S1),nl, print_hash_info(S1),
       print(ID2),nl,print(S2),nl,print_hash_info(S2)
      )
      ; print('NO COLLISIONS')
   ),nl.
   
% use_module(tcltk_interface), print_collisions. 
print_collisions :- hash_collision(ID,ID2,S1,S2),
    print('COLLISION: '), print(ID), print(' <-> '), print(ID2),nl,
    term_hash(S1,TH1), print(sicstus_term_hash(TH1)),nl,
    my_term_hash(S1,TH2), print(my_term_hash(TH2)),nl,
    my_term_hash(S2,TH2),
    check_collision(S1,S2,[range(infinite),depth(infinite)]),
    check_collision(S1,S2,[algorithm(sdbm),range(smallint),depth(infinite)]),
    print('  '),print(S1),nl, print('  '),print(S2),nl,
    fail.
print_collisions.

check_collision(S1,S2,Options) :-
  term_hash(S1,Options,IH1),
  term_hash(S2,Options,IH2),
  (IH1=IH2 -> print(collision_remains(Options,IH1)), print(' **'),nl ; print(collision_disappears(Options,IH1,IH2)),nl).
  
ac :- analyse_collision([range(smallint),depth(infinite)]). % Runtime: 3890 ms
ac0 :- analyse_collision([range(smallint),algorithm(sdbm),depth(infinite)]). % Runtime: 3760 ms
ac1 :- analyse_collision([range(infinite),algorithm(sdbm),depth(infinite)]). % Runtime: 3730 ms
ac2 :- analyse_collision([range(infinite),algorithm(hsieh),depth(infinite)]). % Runtime: 3970 ms
ac3 :- analyse_collision([range(infinite),algorithm(jenkins),depth(infinite)]). % Runtime: 3970 ms
ac4 :- analyse_collision([]). % Runtime: 3720 ms
analyse_collision(Options) :- retractall(hash_collision(_,_)),
  statistics(runtime,[_,_]),
  prolog_flag(max_tagged_integer,Max),
  prolog_flag(min_tagged_integer,Min),
  visited_expression(ID,B),
  term_hash(B,Options,HashB), %
  %super_hash(B,HashB), %
  ((HashB>=Min, HashB =< Max) -> true ; print('*** HASH NOT GOOD FOR INDEXING: '), print(ID),nl),
  (hash_collision(HashB,OtherID)
    -> print('COLLISION : '), print(ID), print(' <-> '), print(OtherID), print(' : '), print(HashB),nl
    ;  assert(hash_collision(HashB,ID))
  ),fail.
analyse_collision(_Options) :-
  statistics(runtime,[_,T]), print('Runtime: '), print(T), print(' ms').

  
:- use_module(library(terms),[term_hash/2, term_hash/3]).
print_hash_info(S1) :- 
       env_hash(S1,EH1), print(env_hash(EH1)),nl,
       my_term_hash(S1,MTH1), print(my_term_hash(MTH1)),nl,
       term_hash(S1,TH1), print(sicstus_term_hash(TH1)),nl.
   
% --------------------------------------------




equivalent_states(S1,true,S2,true) :- \+(\+(S1=S2)).  /* IMPROVE, may even
 allow hash functions */
    
tcltk_set_initial_machine :-
  b_set_initial_machine.

tcltk_clear_machine :- 
   retractall(cached_state_as_strings(_,_,_)),
   set_animation_mode(b),
   bmachine:b_set_empty_machine,
   bmachine:b_machine_reset, 
   haskell_csp:clear_cspm_spec,
   b_machine_hierarchy:reset_hierarchy,
   b_interpreter:reset_b_interpreter,
   state_space_initialise, reset_errors, reset_all_temporary_preferences,
   reset_model_checker,
   reset_refinement_checker,
   reduce_graph_state_space:reduce_graph_reset,
   garbage_collect_atoms, % reclaim atoms if possible
   % dynamic predicates used for POR
   clear_dynamic_predicates_for_POR.

:- use_module(eclipse_interface,[update_preferences_from_spec/0,start_animation/0]).

tcltk_initialise :- eclipse_interface:update_preferences_from_spec,
                    eclipse_interface:start_animation.
 
tcltk_reset :- reset_errors,
               history(H), op_trace_ids(T), 
               current_state_id(CurID),
               (forward_history(FH) -> true ; FH=[]),
               state_space_reset,
               gen_forward_history(T,[CurID|H],FH,FwdHist),
               retractall(forward_history(_)),
               assert(forward_history(FwdHist)).

gen_forward_history([],H,R,R) :- (H=[root] -> true ; print(unexpected_remaining_hist(H)),nl).
gen_forward_history([A|TT],[ID|TH],Acc,Res) :-
  gen_forward_history(TT,TH,[forward(ID,A)|Acc],Res).

tcltk_toggle_switch(Pred) :-
      Call =.. [Pred,X],
      retract(Call),!,
      NewCall =.. [Pred,Y],
      ((X=true) -> (Y=false) ; (Y=true)),
      asserta(NewCall).
      
      
/* ---- getting available options ----- */


tcltk_get_options_dest_info(list(DestinationInfos)) :- debug_println(9,start_tcltk_get_options_dest_info),
  % get a list of all options which are skip events
  current_state_id(CurID),
  compute_all_transitions_if_necessary(CurID),
  findall(Res, (transition(CurID,ActionAsTerm,ID),
               option_dest_info(ID,ActionAsTerm,CurID,Res)),DestinationInfos).

option_dest_info(DestID,_,CurID,Res) :- DestID=CurID,!,Res=skip.
option_dest_info(DestID,_,_,Res) :- not_all_transitions_added(DestID),!,Res=open.
option_dest_info(DestID,_,_,Res) :- invariant_violated(DestID),!,Res=invariant_violated.
option_dest_info(DestID,_,_,Res) :- is_deadlocked(DestID),!,Res=deadlock.
option_dest_info(_DestID,_,_,unknown).

tcltk_get_options(list(Options)) :- debug_println(9,start_tcltk_get_options),
  debug_println(9,tcltk_get_options),
  current_state_id(CurID),
  tcltk_compute_options(CurID,TransSpecs),
  extract_actions(TransSpecs,ActionsAndIDs,Actions),
  set_current_options(ActionsAndIDs),
  %print(Actions),nl,
  debug_println(9,finished_tcltk_get_options),
  Actions = Options.
  
extract_actions([],[],[]).
extract_actions([(Id,Term,Dst)|T],[(Id,Str,Term,Dst)|Srest],[Str|ET]) :-
	time_out_call(translate_event(Term,Str),
				  Str='**TIMEOUT**'),
	extract_actions(T,Srest,ET).


compute_state_information_if_necessary(CurID) :-
  compute_all_transitions_if_necessary(CurID,true).
%  compute_all_transitions_if_necessary(CurID), % first compute transitions, will assert not_invariant_checked if needed
%  compute_invariant_if_necessary_and_requested(CurID).
  
tcltk_compute_options(CurID,ActionsAndIDs) :-
  compute_state_information_if_necessary(CurID),
  %print(found_transitions(CurID)),nl,
  findall((ActId,ActionAsTerm,DestID),
	 (   transition(CurID,ActionAsTerm,ActId,DestID),
	     time_out_call(translate_event(ActionAsTerm,ActString),
	                   ActString='**TIMEOUT**')),
	                   % TO DO: sort or group, especially if we randomize transition/3
	 ActionsAndIDs).
  

tcltk_get_status(INVVIOLATED,MAXREACHED,TIMEOUT) :- debug_println(9,start_tcltk_get_status),
  current_state_id(CurID),
  (time_out_for_invariant(CurID) -> INVVIOLATED = 2
     ;  (invariant_violated(CurID)
           -> INVVIOLATED = 1 ; INVVIOLATED = 0)
   ),
  (max_reached_for_node(CurID) -> MAXREACHED = 1 ; MAXREACHED = 0),
  (time_out_for_node(CurID) -> TIMEOUT = 1 ; TIMEOUT = 0),
  debug_println(9,got_status(INVVIOLATED,MAXREACHED,TIMEOUT)).


%%fd_copy_term(X,Y,true) :- copy_term(X,Y).

    

/* --------------------------------------------------------------------- */

tcltk_get_history(list(StringHistory)) :- debug_println(9,start_tcltk_get_trace),
   get_action_trace(History),
   convert_trace_to_list_of_strings(History,StringHistory).

convert_trace_to_list_of_strings([],[]).
convert_trace_to_list_of_strings([action(AS,_)|T],[AS|CT]) :- !, 
      convert_trace_to_list_of_strings(T,CT).
convert_trace_to_list_of_strings([H|T],[H|CT]) :- convert_trace_to_list_of_strings(T,CT).


/* --------------------------------------------------------------------- */

:- dynamic cached_state_as_strings/3.

tcltk_get_state(list([S1|StateAsStrings])) :- debug_println(9,start_tcltk_get_state),
  current_expression(CurID,CurState),
  compute_state_information_if_necessary(CurID),
  (time_out_for_invariant(CurID) -> S1 = invariant_time_out
    ; invariant_violated(CurID) -> S1=invariant_violated
    ; invariant_not_yet_checked(CurID) -> S1=invariant_not_checked
    ; S1=invariant_ok),
  (cached_state_as_strings(CurID,CurState,StateAsStrings)
   -> true
   ;  retractall(cached_state_as_strings(_,_,_)),
      time_out_call(tcltk_get_state(StateAsStrings,CurState),
                     StateAsStrings = ['***TIMEOUT-PROPERTY***']),
      assert(cached_state_as_strings(CurID,CurState,StateAsStrings))
  ),
  debug_println(9,finished_tcltk_get_state).
  

tcltk_get_state(StateAsStrings,State) :-
  %print(get_state),nl,
  findall(Prop,property(State,Prop),StateProps),
  %print(translating(StateProps)),nl,
  ~~mnf(tcltk_get_state,translate:translate_properties_with_limit(StateProps,StateAsStrings)).
  %print(translated(StateAsStrings)),nl.

:- dynamic current_state_errors/1.

tcltk_get_state_errors(list(ErrorsAsStrings)) :-
  current_state_id(CurID),
  findall(Error,( state_error(CurID,_,Error), Error \== invariant_violated ), Errors),
  ~~mnf(tcltk_get_state,translate:translate_state_errors(Errors,ErrorsAsStrings)),
  retractall(current_state_errors(_)),
  assert(current_state_errors(Errors)).

:- dynamic last_detailed_error/2.
tcltk_get_detailed_state_error(Num,ErrorMsg,Srow,Scol,Erow,Ecol) :-
  retractall(last_detailed_error(_,_)),
  current_state_errors(Errors),
  nth0(Num,Errors,Error),
  !,  %print(explaining(Error)),nl,
  explain_state_error(Error,Span,ErrorMsg), %print(span(Span)),nl,
  assert(last_detailed_error(Span,ErrorMsg)), % save info, so that we can debug if Span contains a span_predicate
  %(generate_dot_from_last_span_predicate('~/Desktop/err.dot') -> true ; true),
  (extract_line_col(Span,Srow,Scol,Erow,Ecol)
    -> true
    ;  Srow= -1, Scol= -1, Erow= -2, Ecol= -2).

can_generate_dot_from_last_state_error :- last_detailed_error(span_predicate(_,_,_),_).
generate_dot_from_last_span_predicate(FileName) :- 
  last_detailed_error(span_predicate(P,LS,State),_ErrorMsg), % also works with expressions
  write_dot_file_for_pred_expr_and_state(P, LS, State,FileName).
  
tcltk_current_state_invariant_violated :-
  current_state_id(CurID),
  invariant_violated(CurID).
  
  

/* --------------------------------------------------------------------- */
 
   
tcltk_save_history_as_trace_file(File) :-
   print_message('saving history to: '), print_message(File),
   (get_action_term_trace(TrA),reverse(TrA,Tr)
   ;
   find_trace_to_current_node(Tr)),!,
   tcltk_save_history_as_trace_file(Tr,File).
tcltk_save_history_as_trace_file(_File) :-
   print_error('### Cannot produce trace to current state !'),
   fail.
   
tcltk_save_history_as_trace_file(Tr,File) :-
   tell(File),
   (specfile:b_or_z_mode
    -> b_machine_name(OurName)
     ; specfile:currently_opened_file(OurName)
   ),
   print_quoted(machine(OurName)), print('.'),nl,
   member(O,Tr),
   print_quoted(O), print('.'),nl,
   fail.
tcltk_save_history_as_trace_file(_,_) :-
   nl, told, 
   print_message(done).   

      
/* --------------------------------------------------------------------- */


tcltk_random_perform :- 
  current_options(Options),
  length(Options,Len),
  (Len>0
  ->  L1 is Len+1,
      random(1,L1,RanChoice),
      tcltk_perform2(RanChoice,Options)
  ;   true
  ).
   
 
tcltk_get_line_col(Nr,list(StartLine),list(StartCol),list(EndLine),list(EndCol)) :-  /* just get the location in the source code */
   current_options(Options),
   nth1(Nr,Options,(_Id,_Action,ActionAsTerm,_NewID)),
   %print(get_source_text(ActionAsTerm)),nl,
   time_out_call(get_source_text_positions(ActionAsTerm,StartLine,StartCol,EndLine,EndCol),
     add_error_and_fail(tcltk_get_line_col,'Timeout determining source code position: ',ActionAsTerm)),
   StartLine \= []. % otherwise fail
%   print(startlines(StartLine)),nl,print(startcols(StartCol)),nl,
%   print(endlines(EndLine)),nl, print(endcol(EndCol)),nl.
   
tcltk_is_sync_event(Nr) :-
   current_options(Options),
   nth1(Nr,Options,(_Id,_Action,ActionAsTerm,_NewID)),
   extract_span_from_event(ActionAsTerm,SPAN,_,_), 
   extract_span_info(SPAN,Info),
  % print(extracted_span_info_for_sync(Info,SPAN)),nl,
   member(sharing,Info).
   
tcltk_perform(Nr) :-
   current_options(Options),
   tcltk_perform2(Nr,Options).
  

tcltk_perform2(Nr,Options) :-
   (nth1(Nr,Options,(_Id,Action,ActionAsTerm,NewID))
     ->  tcltk_perform3(Action,ActionAsTerm,NewID),
         retractall(forward_history(_))
	 ;   tcltk_backtrack /* interpret as backtrack */
   ).
   

tcltk_can_backtrack :- history([_|_]).
tcltk_backtrack :- \+ tcltk_can_backtrack,!,
   print_message('Cannot backtrack'),fail.
tcltk_backtrack :-
               remove_from_op_trace_ids(LastTraceH),
               retract(history(History)),
               retract(current_state_id(CurID)),!,
	           History= [LastID|EHist],
               assert(history(EHist)),
               %visited_expression(LastID,LastExpr,LastBody),
               assert(current_state_id(LastID)),
               (retract(forward_history(FwdHist)) -> true ; FwdHist=[]),
               assert(forward_history([forward(CurID,LastTraceH)|FwdHist])).
   
tcltk_can_forward :- forward_history([_|_]).
tcltk_forward :- \+ tcltk_can_forward,!,
   print_message('Cannot go forward'),fail.
tcltk_forward :-
               retract(forward_history([forward(FwdID,LastTraceH)|FwdHist])),
               retract(history(EHist)),
               retract(current_state_id(CurID)),!,
               assert(forward_history(FwdHist)),
               assert(history([CurID|EHist])),
               add_to_op_trace_ids(LastTraceH),
               %visited_expression(FwdID,FwdExpr,FwdBody),
               assert(current_state_id(FwdID)).
   

tcltk_perform3(Action,ActionAsTerm,NewID) :-  %print(tcltk_perform3(Action,ActionAsTerm,NewID)),nl,
  ((var(Action),var(ActionAsTerm))
    -> add_error(tcltk_perform3,'Illegal call: ',tcltk_perform3(Action,ActionAsTerm,NewID)) ; true),
	( animating_cspm_process_with_arguments(ActionAsTerm,ActionTerm) % in case we want to animate processes with arguments, like P(-,-)
	    -> cur_tcl_process(X), eval_callback_proc(X,ActionTerm)
	  ; ~~mnf(user:tcltk_goto_state(ActionAsTerm,Action,NewID))
	).

animating_cspm_process_with_arguments(ActionAsTerm,ActionTerm) :-
	tcltk_cspm_mode, animation_mode(cspm),
	ActionAsTerm=start_cspm(ActionTerm),
	haskell_csp: string_term_with_args(ActionTerm).

eval_callback_proc(TclProc,ActionTerm) :- 
	string_concatenate('procShowCSPProcessToAnimate ', ActionTerm, Command),
	tcl_eval(TclProc,Command,[]).		     
	
	/* allows one to execute an action in term form, from the current state: */	  
%tcltk_perform_action(T,Cur) :- print_message(perf(T,Cur)),fail.

tcltk_perform_action(ActionAsTerm,NewID) :-
  current_state_id(CurID),
  transition(CurID,ActionAsTerm,NewID),
  ~~mnf(user:tcltk_goto_state(ActionAsTerm,NewID)).
 
 tcltk_no_constants_or_no_inititalisation_found :-
    (current_expression(root,_) ; current_expression(_,concrete_constants(_))),
    current_options([]).

/*  ------------ */
/*  tcltk_open/1 */
/*  ------------ */
  

%prolog_check_load_errors :-
%  (tcltk_find_untyped_consts(ErrRes)
%   -> add_error(prolog_open,'These constants were not given a proper type:', ErrRes)
%   ;  true),
%  (tcltk_find_untyped_vars(ErrRes)
%   -> add_error(prolog_open,'These variables were not given a proper type:', ErrRes)
%   ;  true).
   
tcltk_open_b_file_for_minor_mode(File,Minor) :- tcltk_open_b_file(File),
  set_animation_minor_mode(Minor).

tcltk_open_b_file(File) :- print(tcltk_open_b_file(File)),nl,
  tcltk_clear_machine,
  reset_currently_opened_file,
  set_animation_mode(b),
  print_message(loading(File)),
  bmachine:b_load_machine(File),
  print_message(loaded),
  set_currently_opened_file(File).

:- use_module(xtl_interface,[open_xtl_file/1]).
tcltk_open_xtl_file(File) :- 
  tcltk_clear_machine,
  reset_currently_opened_file,
  set_animation_mode(xtl),
  xtl_interface:open_xtl_file(File), set_currently_opened_file(File).

  
tcltk_open_cspm_file(File) :- 
  tcltk_clear_machine,
  reset_currently_opened_file,
  set_animation_mode(cspm),
  xtl_interface:open_cspm_file(File), set_currently_opened_file(File).
  
:- prob_use_module(xtl_interface).
tcltk_open_promela_file(File) :- 
  tcltk_clear_machine,
  reset_currently_opened_file,
  set_animation_mode(promela), 
  xtl_interface:open_promela_file(File), set_currently_opened_file(File).

% disabled because Java parser seems broken  
%tcltk_open_smv_file(File) :- 
%  tcltk_clear_machine,
%  reset_currently_opened_file,
%  set_animation_mode(smv), 
%  xtl_interface:open_smv_file(File), set_currently_opened_file(File).

:- dynamic currently_opened_cspm_file_in_csp_and_b_mode/1.
:- use_module(specfile,[type_check_csp_and_b/0]).
tcltk_add_csp_file(File) :- 
  set_animation_mode(csp_and_b), 
  retractall(currently_opened_cspm_file_in_csp_and_b_mode(_)),
  xtl_interface:open_cspm_file(File),
  assert(currently_opened_cspm_file_in_csp_and_b_mode(File)),
  notify_change_of_animation_mode,
  specfile:type_check_csp_and_b.

:- use_module(value_persistance,[clear_applicable_flag/0]).
notify_change_of_animation_mode :-
   clear_applicable_flag.
  

tcltk_add_xtl_file(File) :- 
  xtl_interface:open_xtl_file(File),
  add_animation_mode(xtl).
  
tcltk_open_z_file(File) :- 
  open_proz_file(File,BMachine),
  tcltk_clear_machine,
  set_animation_mode(b), set_animation_minor_mode(z),
  b_set_typed_machine(BMachine),
  set_currently_opened_file(File).

:- use_module(eclipse_interface,[load_eventb_file/1]).
tcltk_load_packaged_eventb_file(File) :-
  tcltk_clear_machine,
  load_eventb_file(File).
   
    
/* --------------------------------------------------------------------- */

tcltk_exists_an_open_node :-
  not_all_transitions_added(_) ; not_interesting(_).
tcltk_goto_an_open_node :-
  (not_all_transitions_added(ID) ; not_interesting(ID)),!,
  tcltk_goto_state(find_open_node,find_open_node,ID).
tcltk_goto_max_reached_node :-
  (max_reached_for_node(ID) ; time_out_for_node(ID)),!,
  tcltk_goto_state(find_max_reached_node,find_max_reached_node,ID).

tcltk_goto_an_invariant_violation :-
  invariant_violated(ID),!,
  tcltk_goto_state(find_inv_violation,find_inv_violation,ID).  
  
/* --------------------------------------------------------------------- */

% TO DO: clean up and move into model_checker module

tcltk_state_space_only_has_root_node :- transition(root,_,ID),
    \+ not_all_transitions_added(ID),!,fail.
tcltk_state_space_only_has_root_node.

tcltk_search_among_existing_nodes(TclTkRes,FindDeadlocks,FindInvViolations,
                                                      FindGoal,FindAssViolations) :-         
   search_among_existing_nodes(Res,FindDeadlocks,FindInvViolations,
                                                      FindGoal,FindAssViolations),
   translate_error_for_tclk(Res,TclTkRes).

% if tcltk_state_space_only_has_root_node is true: then we should call search_among_existing_nodes before doing open_search to ensure that we uncover state_errors in the root node
search_among_existing_nodes(invariant_violation,_FindDeadlocks,FindInvViolations,
                                                      _FindGoal,_FindAssViolations) :-
   ((get_state_space_stats(_,_,PrN), PrN>100)
     -> print_message('Search Among Processed Nodes'), print_message(number_of_nodes(PrN)) ; true),
   FindInvViolations=1,
   invariant_violated(ResID),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(DEADLOCK,FindDeadlocks,_,_,_) :-
   FindDeadlocks=1,
   visited_expression(ResID,_CurState),
   transitions_computed_for_node(ResID),
   is_deadlocked(ResID),
   (time_out_for_node(ResID)
    -> DEADLOCK = possible_deadlock_timeout ; DEADLOCK = deadlock),
   print_message(deadlock(ResID)),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(goal_found,_,_,FindGoal,_) :-
   FindGoal=1,
   %visited_expression(ResID,_CurState),
   %\+(not_all_transitions_added(ResID)),  % if commented out: will also look at open nodes
   test_goal_in_node(ResID),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(assertion_violation,_,_,_,FindAssViolations) :-
   FindAssViolations=1,
   visited_expression(ResID,CurState),
   transitions_computed_for_node(ResID),
   state_violates_assertions(ResID,CurState),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(ERR,_,_,_,_) :-
   state_space:state_error(ResID,_,Error), Error \== invariant_violated,
   (Error = abort_error(ERR,_,_,_) -> true ; ERR = state_error),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(all,_,_,_,_).

test_goal_in_node(ResID) :- b_get_machine_goal(Goal), 
  eclipse_interface:test_boolean_expression_in_node(ResID,Goal).

   
compute_all_transitions_if_necessary :-
  current_state_id(CurID),
  compute_all_transitions_if_necessary(CurID,false).
compute_all_transitions_if_necessary(CurID) :-
  compute_all_transitions_if_necessary(CurID,false).
  
compute_all_transitions_if_necessary(CurID,CheckInvariant) :-
  (retract_open_node(CurID) % will assert not_invariant_checked
    -> set_context_state(CurID),
       % first check invariant, relevant if we want to use specialized_inv proof information
       (CheckInvariant==true -> compute_invariant_if_necessary_and_requested(CurID) ; true),
       compute_all_transitions(CurID),
       clear_context_state
   ; true).
%compute_all_transitions_if_necessary(CurID,CurState) :-
%  ((not_all_transitions_added(CurID);not_interesting(CurID))
%    -> compute_all_transitions(CurID,CurState) ; true).
    
compute_all_transitions(ID) :-
  visited_expression(ID,CurState),
 %% print(computing_transitions(ID)),nl, %%%
  compute_all_transitions(ID,CurState).
 %% print(all_transitions_computed(ID)),nl. %%%

compute_invariant_if_necessary_and_requested(ID) :-
  visited_expression(ID,State),
  compute_invariant_if_necessary(ID,State).
  % previously we checked preference(do_invariant_checking,true)
%compute_invariant_if_necessary_and_requested(ID,State) :-
%   (preference(do_invariant_checking,true) -> compute_invariant_if_necessary(ID,State) ; true).
compute_invariant_if_necessary(ID,State) :-
    animation_mode(MODE),compute_invariant_if_necessary(ID,State,MODE).
compute_invariant_if_necessary(ID,State,MODE) :- 
   ( check_invariantKO(MODE,ID,State) -> true ; true).

check_invariantKO(ID,State) :-
    animation_mode(MODE),check_invariantKO(MODE,ID,State).
check_invariantKO(b,ID,State) :- !, check_invariantKO2(ID,State).
check_invariantKO(csp_and_b,ID,State) :- !, check_invariantKO2(ID,State).
check_invariantKO(MODE,ID,State) :-  % print(check_invariant(MODE,ID,State)),nl,
    MODE \= cspm, MODE \= xtl, MODE \= promela,
    %MODE \= smv, 
    plugins:plugin_check_invariantKO(MODE,ID,State).
    
check_invariantKO2(ID,_) :- invariant_violated(ID),!.
check_invariantKO2(ID,CurState) :- 
    set_invariant_checked(ID), % only succeeds if we haven't checked invariant for ID yet
    % print(check_inv(ID)),nl,
    state_corresponds_to_initialised_b_machine(CurState,CurBState),
    (catch_clpfd_overflow_call_for_state(ID,b_state_violates_invariant(ID,CurBState),true) ->
    %(b_state_violates_invariant(ID,CurBState) ->
        set_invariant_violated(ID)).

:- use_module(probsrc(ample_sets)).

compute_ample_set(CurID,CurState) :-
  catch_clpfd_overflow_call_for_state(CurID,compute_ample_set2(CurID,CurState),clear_context_state).

% preferences:preference(use_clpfd_solver,false)
compute_all_transitions(CurID,CurState) :-
  catch_clpfd_overflow_call_for_state(CurID,compute_all_transitions2(CurID,CurState),clear_context_state).


%catch_clpfd_overflow_call_for_state(_CurID, Call,_ExtraCleanUpCode) :-
%    preferences:preference(use_clpfd_solver,false),!, % Overflow should not occur
%    call(Call).
catch_clpfd_overflow_call_for_state(CurID, Call,ExtraCleanUpCode) :-
  on_exception(error(representation_error('CLPFD integer overflow'),_),
               Call,
               ( clpfd_interface:clpfd_overflow_error_message,
               store_state_error(CurID,'CLPFD_integer_overflow',_),
               call(ExtraCleanUpCode))).

:- use_module(clpfd_interface,[catch_clpfd_overflow_call1/1]).

:- use_module(extension('user_signal/user_signal'), [get_user_signal_ref/1]).
catch_interrupt_assertion_call(Call) :- 
   on_exception(error(forced_interrupt_error('User has interrupted the current execution'),_),
                Call,
                (process_interrupted_error_message,fail)).

process_interrupted_error_message :- 
    add_error(probcli,'User has interrupted current execution.').

compute_all_transitions2(CurID,CurState) :- /* compute all outgoing transitions for a given node */
   % set_context_state(CurID), % error_context is set for each operation below anyway.
   reset_max_reached,
   prepare_state_for_specfile_trans(CurState,PreparedCurState), % will perform unpacking of constants just once, for example
   (use_no_timeout(CurID)
     -> add_transitions_fail_loop(CurID,PreparedCurState)
     ;   preferences:preference(time_out,CurTO),
         add_transitions__with_timeout_fail_loop(CurID,PreparedCurState,CurTO),
	     add_partial_transitions(CurID,PreparedCurState)
   ),
   (max_reached -> assert_max_reached_for_node(CurID),MaxReached=true ; MaxReached=false),
   (CurID==root -> save_constants(MaxReached) ; add_new_transitions_to_cache(CurID))
   . %,clear_context_state.  
   
add_transitions_fail_loop(CurID,CurState) :- 
    specfile_possible_trans_name(CurState,ActionName),
   %print(adding_transition_no_timeout(CurID,ActionName)),nl,
   set_error_context(operation(ActionName,CurID)),
   add_transition(CurID,CurState,ActionName),fail.
add_transitions_fail_loop(_,_) :- clear_error_context.

:- use_module(library(timeout),[time_out/3]).
my_timeout_test(1000000) :- print('.'), !, my_timeout_test(0).	
my_timeout_test(X) :- X1 is X+1, my_timeout_test(X1).
my_timeout_test :- time_out(my_timeout_test(0),500,TO), nl,
	    print(time_out_result(TO)),nl, TO==time_out.
:- assert_must_succeed(user:my_timeout_test).	    

add_transitions__with_timeout_fail_loop(CurID,CurState,CurTO) :-    % statistics(runtime,_),
   % TO DO: prepare computing the state: expanding const and vars, unpacking, ...
    specfile_possible_trans_name(CurState,ActionName),
    % Note: we enumerate possible transition names so that we can catch the time-out per name !
       %  statistics(runtime,[T1,DeltaT]), print(delta_time(ActionName,DeltaT)),nl, %%
         % first enumerate possible Operations/... to obtain timeout per operation/...
   % TO DO: divide CurTO by number of transitions ???
       %% statistics(runtime,[T1,_]), print(try(ActionName,T1)),nl,%%
      set_error_context(operation(ActionName,CurID)),
      time_out(add_transition(CurID,CurState,ActionName),CurTO,TimeOutRes),
       %% statistics(runtime,[T2,_]), TDiff is T2-T1, print(added_transition(CurID,ActionName,TDiff,TimeOutRes)),nl, %%
   %%  add_transition(CurID,CurState,ActionName) ,
	  (TimeOutRes==time_out 
	   ->  print_message('TIME OUT: '), print_message(CurID), print_message(CurTO),
	       print_message(ActionName),
		   (var(ActionName)
		     -> assert_time_out_for_node(CurID,'unspecified_operation')
		     ; assert_time_out_for_node(CurID,ActionName))  % TO DO: assert info per ActionSkeleton ?
		; true
	  ),fail.
add_transitions__with_timeout_fail_loop(_,_,_) :- clear_error_context.
    % statistics(runtime,[_,DeltaT]), print(delta_time(end,DeltaT)),nl.
   
add_transition(CurID,CurState,ActionName) :- 
    %(var(ActionName) -> add_error_fail(add_transition,'ActionName is a variable: ',CurID) ; true),
    nonvar(ActionName), % can happen in CSP, CSP||B, XTL, ... modes
    candidate_operation_for_reuse(ActionName),
    retract(reuse_operation(CurID,ActionName,PreviousID,_FromOp)),!,
    % we don't have to compute the operation again; simply reuse pre-computed effect
    b_get_operation_normalized_read_write_info(ActionName,_Read,WrittenVariables),
    build_up_b_real_operation(ActionName,ActionTemplate), % missing add any parameters
   %  print(reusing_operation(ActionName,ActionTemplate, from(PreviousID), for(CurID),via(FromOp))),nl,
    transition(PreviousID,ActionTemplate,TransID,PreviousSuccessorID),
 %   print(trans(PreviousID,ActionTemplate,TransID,PreviousSuccessorID)),nl,
    findall(Info,transition_info(TransID,Info),TransInfos),
  %  print(transinfo(TransInfos)),nl,
     visited_expression(PreviousSuccessorID,PreviousExpression,true),
     ~~mnf(patch_state(CurState,WrittenVariables,PreviousExpression,NewExpression)),
  %  print_bstate(NewExpression),nl,
%     (specfile_trans(CurState,ActionName,ActionTemplate,NewExpression,_,_)
%      -> true ; add_error(add_transition,'Mismatch in operation reuse',ActionName),
%                print(CurID),nl,
%                print(ActionName), print(' via '),  print(via(PrevOp)),nl,
%                print_bstate(CurState), print( ' ---> '),nl,
%                print_bstate(NewExpression),nl, print(written(WrittenVariables)),nl,nl,
%                print(prev(PreviousID,PreviousSuccessorID)),nl,  print_bstate(PreviousExpression),nl
%     ),
    ~~mnf(add_trans_id(CurID,ActionTemplate,NewExpression,[],_NewID,TransInfos,_)).  
:- if(debug:global_debug_flag).
add_transition(CurID,CurState,ActionName) :-
   %debug:watch(30,specfile:specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue)),
   %debug:watch(30,user:add_trans_id(CurID,Act,NewExpression,Residue,_NewId,TransInfo,_)).
   debug:timer_call(ActionName,specfile:specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue)),
   debug:timer_call(user:add_trans_id(CurID,Act,NewExpression,Residue,_NewId,TransInfo,_)).
:- else.
add_transition(CurID,CurState,ActionName) :-
  /* ActionName maybe pre-instantiated */
   specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue),
 %% print_bt_message(specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue)), %%
   add_trans_id(CurID,Act,NewExpression,Residue,_NewId,TransInfo,_).
:- endif.

patch_state(const_and_vars(ConstID,Vars),WrittenVariables,PrevState,
            const_and_vars(ConstID,PatchedVars)) :-
   patch_state(Vars,WrittenVariables,PrevState,PatchedVars).
patch_state([],_Write,_Prev,[]).
patch_state([bind(Variable,Value)|T],WrittenVariables,PrevState,[bind(Variable,NewValue)|PT]) :-
    (ord_member(Variable,WrittenVariables)
     ->  lookup_variable(PrevState,Variable,NewValue)
     ; NewValue=Value
    ), patch_state(T,WrittenVariables,PrevState,PT).
    
lookup_variable(const_and_vars(_ConstID,Vars),Var,Val) :- !,
    lookup_value_for_existing_id(Var,Vars,Val).
lookup_variable(Store,Var,Val) :- lookup_value_for_existing_id(Var,Store,Val). 
   

add_trans_id(CurID,Act,NewExpression,Residue,NewID,TransInfo,TransId) :-  
  (Residue=[] ->  true ; add_error(add_trans_id,'Residue: ',CurID:Act:Residue)),
  tcltk_add_new_transition_transid(CurID,Act,NewID,NewExpression,TransInfo,TransId).
    

add_partial_transitions(root,CurState) :-
   /* for the moment the only state where partial transitions are computed */
   \+ any_transition(root,_,_), /* adding normal transitions was not successful */
   partial_trans(CurState,Act,NewExpression,_Residue),
   add_trans_id(root,Act,NewExpression,_Residue,_NewID,[],_),fail.
add_partial_transitions(_CurID,_).
   
  
/* --------------------------------------------------------------------- */

/* A bit like AO* algorithm: expands open nodes (i.e., states whose transitions
 have not yet been computed */
 
%:- use_module(linda_slave).
:- use_module(model_checker).

tcltk_model_check(Nr,TclTkRes,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopFullCoverage,PartialOrderReduction,TIMELIMIT,TotalTime) :-
   statistics(runtime,[CurTime,_]), /* get current time in ms */
   LimitTime is CurTime+TIMELIMIT,
   InspectExistingNodes = 0,
   do_model_check(Nr,_,LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,StopFullCoverage,PartialOrderReduction,InspectExistingNodes),
   statistics(runtime,[CurTime2,_]),
   TotalTime is CurTime2-CurTime,
   translate_error_for_tclk(Res,TclTkRes).

translate_error_for_tclk([timeout,Nr],Res) :- !, Res=[timeout,Nr].
translate_error_for_tclk(state_error(S),String) :- !, translate_state_err_for_tcltk(S,String).
translate_error_for_tclk(S,R) :- atomic(S),!,R=S.
translate_error_for_tclk(S,R) :- nonvar(S), add_error(translate_error_for_tclk,'Unknown Model Check result: ',S), R=unknown.

translate_state_err_for_tcltk(abort_error(TYPE,_,_,_),ERR) :- !, ERR = TYPE.
translate_state_err_for_tcltk(eventerror(Event,Error,_),ERR) :- !,
    functor(Error,F,_),
    ajoin(['event_error:',Event,':',F],ERR).
translate_state_err_for_tcltk(Error,ERR) :-
   functor(Error, StateError, _), atom_concat('state_error:',StateError,ERR).
   


do_model_check(_Nr,NodesAnalysed,_LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,_StopFullCoverage,_PartialOrderReduction,ForceInspectExistingNodes) :-
    (tcltk_state_space_only_has_root_node -> true % Then also look at the root node
      ; ForceInspectExistingNodes == 1),
    search_among_existing_nodes(Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations), Res \=all, NodesAnalysed=0.
do_model_check(LimitNr,NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,StopFullCoverage,PartialOrderReduction,_) :- 
   open_search(LimitNr,NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,StopFullCoverage,PartialOrderReduction).


/* --------------------------------------------------- */

/* constraint-based checking of refinements */
:- use_module(cbc_refinement_checks).
tcltk_cbc_refinement_check(list(List),ErrorOccured) :-
    cbc_refinement_check(List,ErrorOccured).

/* constraint-based/model search for invariant violations */

:- dynamic cbc_inv_error_found/2.
tcltk_constraint_based_check(L) :- tcltk_constraint_based_check(L,_).
tcltk_constraint_based_check(list(List),ErrorsOrTimeoutWereFound) :- 
   retractall(cbc_inv_error_found(_,_)),
   findall(Res, (tcltk_constraint_based_check_with_timeout(OpName,ErrorDesc),
                   (ErrorDesc=ok -> true ; assert(cbc_inv_error_found(OpName,ErrorDesc))),
                   string_concatenate(' == ',ErrorDesc,Res1),
                   string_concatenate(OpName,Res1,Res)), List),
   %%print(list(List)),nl, %%
   (cbc_inv_error_found(_,_) -> 
      ((cbc_inv_error_found(_,X),X\=time_out)
        -> ErrorsOrTimeoutWereFound=true ; ErrorsOrTimeoutWereFound=time_out
      )
      ; ErrorsOrTimeoutWereFound = false).

% can also be called with uninstantiated OpName
tcltk_constraint_based_check_with_timeout(OpName,ErrorDesc) :-
   preferences:get_computed_preference(debug_time_out,DebugTimeOut),
   b_get_machine_operation(OpName,_,_,_),
   time_out(tcltk_constraint_based_check_op(OpName,ErrorDesc),DebugTimeOut,TimeOutRes),
    (TimeOutRes=time_out 
		   -> print_message('TIME OUT: '), print_message(OpName), print_message(DebugTimeOut),
			   ErrorDesc = timeout
			; true
	).
tcltk_constraint_based_check_op(_OpName,ErrorDesc) :-
  catch_clpfd_overflow_call1(state_model_check(_OpName,State,Operation,NewState)),!,
  %fd_copy_term(NewState,ResTempl,ResBody), print(ResTempl),print(ResBody),nl,
  %fd_copy_term(State,PriorTempl,PriorBody),
  ~~mnf(tcltk_add_new_transition_transid(root,'$JUMP'(constraint_based_check),PriorID,State,[],TransId1)),
  ~~mnf(translate_event(Operation,Action)),
  ~~mnf(tcltk_add_new_transition_transid(PriorID,Operation,NewID,NewState,[],TransId2)),
  retractall(current_state_id(_)),
  assert(current_state_id(NewID)),
  reset_op_trace_ids,
  add_to_op_trace_ids(TransId1), add_to_op_trace_ids(TransId2),
  retractall(history(_)),
  retractall(forward_history(_)),
  assert(history([PriorID,root])), print(State),nl,
  ajoin(['invariant_violated_after(',Action,')'],ErrorDesc).
tcltk_constraint_based_check_op(_Depth,'ok').


% try to find constants, parameters that make a given trace feasible
tcltk_cbc_find_trace(TraceToBeFound,Res) :-
    preferences:get_computed_preference(debug_time_out,DebugTimeOut),
    print(looking_for_solution_for_trace(TraceToBeFound)),nl,
    split_atom(TraceToBeFound,[',',';',' '],TraceAsList),
    print(list(TraceAsList)),nl,
    time_out(sap:create_testcase_path(TraceAsList,0,Transitions),DebugTimeOut,TimeOutRes),
    !,
    (TimeOutRes=time_out
      -> print_message('*** TIME OUT: '), print_message(DebugTimeOut), Res = time_out
      ;  Res = ok,
         print_message('*** FOUND PATH: '), print_message(Transitions),
         maplist(extract_trans_id,Transitions,TransIds),
         set_trace_by_transition_ids(TransIds)
    ).
tcltk_cbc_find_trace(_TraceToBeFound,no_solution_found) :- print_message('*** NO SOLUTION FOUND'),nl.
extract_trans_id( (TransId,_,_,_), TransId).


% ------------
tcltk_constraint_find_valid_state :-
   catch_clpfd_overflow_call1(b_set_up_valid_state(State)), % TO DO: does not yet separate constants and variables !
   tcltk_add_cbc_state(State,find_valid_state).
 
tcltk_constraint_find_deadlock_state(Res) :-
   check_we_are_in_b_mode(cbc_deadlock_freedom_check),
   create_texpr(truth,pred,[],True),
   catch_clpfd_overflow_call1(cbc_deadlock_freedom_check(State,True,0)), % TO DO: does not yet separate constants and variables !
   (State == time_out -> Res = time_out
    ; tcltk_add_cbc_state(State,cbc_deadlock), Res=deadlock).
tcltk_constraint_find_deadlock_state_with_goal(Filter,Res) :-
   check_we_are_in_b_mode(cbc_deadlock_freedom_check),
   ( b_get_machine_goal(Goal) -> true
   ; otherwise                -> create_texpr(truth,pred,[],Goal)),
   catch_clpfd_overflow_call1(cbc_deadlock_freedom_check(State,Goal,Filter)), % TO DO: does not yet separate constants and variables !
   (State == time_out -> Res = time_out
    ; tcltk_add_cbc_state(State,cbc_deadlock), Res=deadlock).

check_we_are_in_b_mode(_) :- b_or_z_mode,!.
check_we_are_in_b_mode(Cmd) :- add_error(check_we_are_in_b_mode,'This command can only be performed in B,Z or TLA+ mode: ',Cmd),fail.
 
tcltk_constraint_find_deadlock_state_with_goal(true,_Filter,Res) :- 
  !,tcltk_constraint_find_deadlock_state(Res).
tcltk_constraint_find_deadlock_state_with_goal(GoalPred,Filter,Res) :-
  bmachine:b_set_machine_goal(GoalPred),
  tcltk_constraint_find_deadlock_state_with_goal(Filter,Res).

tcltk_add_cbc_state(State,OpName) :- 
  ~~mnf(tcltk_add_new_transition_transid(root,'$JUMP'(OpName),NewID,State,[],TransId)),
  retractall(current_state_id(_)),
  assert(current_state_id(NewID)),
  reset_op_trace_ids,
  add_to_op_trace_ids(TransId), 
  retractall(history(_)),
  retractall(forward_history(_)),
  assert(history([root])). %, print(State),nl.

tcltk_constraint_find_static_assertion_violation(Result) :-
    catch_clpfd_overflow_call1(cbc_static_assertions_check(State)),
    ( State = counterexample_found(CE) ->
        Result = counterexample_found,
        tcltk_add_cbc_state(CE,static_assertion_check)
    ; State = no_counterexample_found ->
        fail
    ; otherwise ->
        Result = State).
/* --------------------------------------------------- */
  

:- use_module(error_manager,[extract_all_line_col/5]).
:- use_module(probcspsrc(haskell_csp),[extract_span_from_event/4,extract_span_info/2]).

get_source_text_positions(Event,Line,Col,EndLine,EndCol) :- 
      extract_span_from_event(Event,SPAN,_,_),  
      extract_all_line_col(SPAN,Line,Col,EndLine,EndCol).


/* -------------------------------------------------------------------- */

get_current_option_eventtrace(OptionNum,Trace) :-
    % the option number starts with 0 (coming from a TK listbox)
    current_options(ActionsAndIDs),
    nth0(OptionNum,ActionsAndIDs,(Id,_,_,_)),
    transition_info(Id,eventtrace(Trace)),!.

tcltk_has_eventtrace(OptionNum) :-
    get_current_option_eventtrace(OptionNum,_Trace).

tcltk_show_eventtrace(OptionNum,Desc) :-
    get_current_option_eventtrace(OptionNum,Trace),
    explain_event_trace(Trace,Desc).

    
    
    

/* --------------------------------------------------- */

tcltk_slowly_execute_trace_to_current_node :-
   print_message(informational,'Executing History'),
   current_state_id(CurID),
   tcltk_slowly_execute_trace_to_node(CurID).
   
tcltk_slowly_execute_trace_to_node(ID) :- /* can be useful for TestCase Generation */
   find_shortest_path_from_to(root,ID,OpL,IDList),
   state_space_reset,
   slowly_execute_trace_to_node(OpL,IDList).
   
:- use_module(library(system),[sleep/1]).

slowly_execute_trace_to_node([],[]).
slowly_execute_trace_to_node([OpID|T],[ToID|RestIDs]) :-
    print(exec(ActionAsTerm)),nl,
    user:force_do_one_gui_event,
    transition(_,ActionAsTerm,OpID,_),
    tcltk_perform_action(ActionAsTerm,ToID),
    user:force_do_one_gui_event,
    (cur_tcl_process(X)
     -> tcl_eval(X,procInsertHistoryOptionsState,_) /* a bit of a hack: updated window */
     ;  true
    ),
    user:tk_do_all_events,
    user:(tk_do_one_event([tk_window_events,tk_idle_events ]) -> true ; true),
    sleep(0.2),
    slowly_execute_trace_to_node(T,RestIDs).
    
    
/* --------------------------------------------------- */

   
tcltk_find_shortest_trace_to_current_node(list(StringList)) :-
   find_trace_to_current_node(OpL),
   translate_events(OpL,StringList).
   
tcltk_execute_trace_to_current_node :-
   current_state_id(CurID),
   tcltk_execute_trace_to_node(CurID).
   
tcltk_execute_trace_to_node(ToID) :- /* can be useful for TestCase Generation */
  set_target_id(ToID),
  tcltk_execute_trace_from_to_found_predicate(root,_,_).
 
tcltk_execute_trace_from_to_found_predicate(From,OpL,IDList) :-
     print(finding_trace_from_to(From)),nl, flush_output(user),
   find_shortest_path_from_to_found_predicate(From,OpL,IDList,ToID), !,
       /* mal: replaced by shortest 5/12/2006 */
   (current_state_id(From) -> true ; state_space_reset),
     debug_println(9,executing_trace(From,OpL,ToID)), flush_output(user),
   execute_id_trace_from_current(ToID,OpL,IDList).
 %  print(executing_trace_to(ID,OpL,IDList)),nl,
   %compute_history(OpL,IDList,History), print(new_history(History)),nl,

   
execute_trace_to_node([],[]).
execute_trace_to_node([ActionAsTerm|T],[ToID|RestIDs]) :-
   %print(exec(ActionAsTerm)),nl,
    tcltk_perform_action(ActionAsTerm,ToID),
    execute_trace_to_node(T,RestIDs).

find_trace_to_current_node(Trace) :-
   current_state_id(CurID),
   find_shortest_path_from_to(root,CurID,L,_),
   extract_term_trace_from_transition_ids(L,Trace).
   
:- use_module(state_space_dijkstra).
 
find_shortest_trace_to_node(FromID,ToID,ActionTermList,IDList) :-
   find_shortest_path_from_to(FromID,ToID,ActionTermList,IDList).

    
 /* --------------------------------------------------- */ 

    

tcltk_goto_a_non_deterministic_node :- /* same transition leads to different resulting state */
    transition(ID,X,ID2),
    transition(ID,X,ID3), dif_ID(ID2,ID3),
    print_message(found_non_det_op(X)),
    %tcltk_goto_state(find_non_det_node,find_non_det_node,ID).
    tcltk_execute_trace_to_node(ID).

dif_ID(ID1,ID2) :- ((number(ID1),number(ID2)) -> ID2>ID1 ; ID1 \= ID2).

tcltk_goto_a_branching_node :- /* two different events/operations enabled */
    transition(ID,X,ID2), get_operation_name(X,NX),
    transition(ID,Y,ID3), dif_ID(ID2,ID3), get_operation_name(Y,NY), NX\=NY,
    print_message(found_branching_op(ID,NX,NY)),
    %tcltk_goto_state(find_branching_node,find_branching_node,ID),
    tcltk_execute_trace_to_node(ID).

tcltk_goto_a_non_resetable_node :-
    find_resetable_nodes,
    visited_expression(ID,_),
    \+(resetable_node(ID)),
    \+(not_all_transitions_added(ID)), !,
    %tcltk_goto_state(find_non_resetable_node,find_non_resetable_node,ID).
    tcltk_execute_trace_to_node(ID).
    
operation_name_covered(OpName) :- operation_name_covered(OpName,_).
operation_name_covered(OpName,Template) :- b_get_machine_operation(OpName,Results,Par,_),
   \+ operation_not_yet_covered(OpName),
   length(Par,Arity),
   functor(Op,OpName,Arity),
   (Results\=[] -> Template = '-->'(Op,_) ; Template = Op).

tcltk_goto_state_enabling_operation(TclOpName,FromCurrent) :-
   predicate_debugger:adapt_tcl_operation_name(TclOpName,OpName),
   operation_name_covered(OpName,Template),
   set_found_predicate(ID, transition(ID,Template,_)),
   print(template(Template,FromCurrent)),nl,
    print_message('Computing trace to node'),
    (FromCurrent=1
      -> current_state_id(FromID) % inefficient: better to do one dijkstra from FromID !
      ; FromID = root
    ),
    tcltk_execute_trace_from_to_found_predicate(FromID,_,_).
%    tcltk_goto_state(find_enabled(OpName),find_enabled,ID).
    
:- dynamic resetable_node/1.  

resetable_node(a).

:- dynamic find_flag/2.
find_resetable_nodes :-
    retractall(find_flag(_,_)),
    retractall(resetable_node(_)),
    any_transition(root,_,ID), /* add nodes reachable from root */
    assert(find_flag(ID,0)),
    fail.
find_resetable_nodes :-
    visited_expression(CID,concrete_constants(_C)),
    any_transition(CID,_,ID), /* add nodes reachable after a setup_constants */
    assert(find_flag(ID,0)),
    fail.
find_resetable_nodes :- find_resetable_nodes_loop,
    print_message('Not resetable:'),
    visited_expression(ID,_),
    \+(resetable_node(ID)),
    print_message(ID),
    fail.
find_resetable_nodes :-
    print_message('Done').
   
    
find_resetable_nodes_loop :-
    retract(find_flag(ID,N)), !, find_resetable_nodes_loop(ID,N).
find_resetable_nodes_loop.    

find_resetable_nodes_loop(ID,N) :- N1 is N+1,
    assert(resetable_node(ID)),
    any_transition(NewID,_,ID),
    \+(resetable_node(NewID)),
    \+(find_flag(NewID,_)),
    assert(find_flag(NewID,N1)),
    fail.
find_resetable_nodes_loop(_,_) :- find_resetable_nodes_loop.


/* ------------------------------------------------------------------ */

:- ensure_loaded(state_viewer_images).
 
/* ------------------------------------------------------------------ */



print_property_partitions :- print('PARTITIONS OF PROPERTIES'),nl,
   b_interpreter:b_get_properties_from_machine(Properties),
   bsyntaxtree:predicate_components(Properties,Comp),
   length(Comp,Len), print(Len), print(' partitions found'),nl,
   member(component(P,Vars),Comp), nl,print(Vars),nl,translate:print_bexpr(P),nl,fail.
print_property_partitions :- nl, print(' ============== '),nl.


tcltk_add_user_executed_operation(TclOpName,CustomPredicate) :- 
   predicate_debugger:adapt_tcl_operation_name(TclOpName,OpName),
    print(tcltk_add_user_executed_operation(OpName,CustomPredicate)),nl,
    current_expression(CurID,InState),
    print(cur_id(CurID)),nl,
    bmachine:set_temp_predicate(CustomPredicate,OpName),
    print(exec(OpName)),nl,
    set_context_state(CurID),
     print(Operation),nl,
     preferences:get_preference(maxNrOfEnablingsPerOperation,Max), 
     preferences:get_preference(maxNrOfInitialisations,MaxI), preferences:set_preference(maxNrOfEnablingsPerOperation,1),
     preferences:set_preference(maxNrOfInitialisations,1),
     (specfile:b_trans(InState,OpName,Operation,NewState,TransInfo)
      -> preferences:set_preference(maxNrOfEnablingsPerOperation,Max),
         preferences:set_preference(maxNrOfInitialisations,MaxI),
         print(Operation),nl,
         %print(NewState),nl,
         bmachine:reset_temp_predicate,
         % now add a transition to the state space:
         Residue=[],
         add_trans_id(CurID,Operation,NewState,Residue,NewID,TransInfo,_),
         clear_context_state,
         tcltk_goto_state(Operation,NewID)
      ;  preferences:set_preference(maxNrOfEnablingsPerOperation,Max),
         preferences:set_preference(maxNrOfInitialisations,MaxI),
         bmachine:reset_temp_predicate,clear_context_state,
         fail
    ).
    

% --------------------------------------------------------------------------------------

% :- ensure_loaded(btracer2).
 

:- use_module(bmachine).
:- use_module(specfile).
:- use_module(tools).
:- use_module(library(lists)).
:- use_module(store).
:- use_module(self_check).
:- use_module(b_global_sets).
:- use_module(translate).
:- use_module(predicate_debugger).


/* --------------------------------------------------------- */
/* --------------------------------------------------------- */

        
/* ----------------------------------*/
/*       b_analyse_conjunction       */
/* ----------------------------------*/

/* provide list of possible options for analysis */
tcltk_analyse_option(invariant,'INVARIANT') :- b_or_z_mode.
tcltk_analyse_option(properties,S) :- b_or_z_mode,
    get_specification_description(properties,S).
tcltk_analyse_option(assertions,S) :- (b_or_z_mode ; csp_mode),
    get_specification_description(assertions,S).
tcltk_analyse_option(deadlock,'DEADLOCKED') :- b_or_z_mode.

:- use_module(predicate_evaluator).
/* perform the various analysis */
tcltk_analyse(Type,ListWithResults,Total,False,Unknown) :-
   tcltk_analyse2(Type,ListWithResults,Summary),
   (member(total/Total,Summary) -> true ; Total=0),
   (member(false/False,Summary) -> true ; False=0),
   (member(unknown/Unknown,Summary) -> true ; Unknown=0).
tcltk_analyse2(invariant,L,Summary) :- !,tcltk_analyse_invariant(L,Summary).
tcltk_analyse2(properties,L,Summary) :- !,tcltk_analyse_properties(L,Summary).
tcltk_analyse2(assertions,L,Summary) :- !,tcltk_analyse_assertions(L,Summary).
tcltk_analyse2(deadlock,L,Summary) :- !,tcltk_analyse_deadlock(L,Summary).
tcltk_analyse2(X,L,Summary) :- add_error(tcltk_analyse, 'Unknown type of formula: ', X),
 L=[], Summary=[].

   
/* --------------------------------- */

tcltk_show_current_state(TransBState) :-
   current_b_expression(B),
   mnf(translate_bstate(B,TransBState)).
        
tcltk_show_typing_of_variables_and_constants(List) :-
      b_get_machine_variables(TypedVarList),
      convert_typed_varlist(TypedVarList,VarList,Types),
      generate_typing_list(VarList,Types,VList),
      
      b_get_machine_all_constants(TypedCstList),
      convert_typed_varlist(TypedCstList,CVarList,CTypes),
      find_non_det_constants,
      generate_typing_list(CVarList,CTypes,CList),
      ( CList = []
        -> List = [' VARIABLES;'|VList]
        ;  append([' CONSTANTS;'|CList],['VARIABLES;'|VList],List)
      ),!.
tcltk_show_typing_of_variables_and_constants('TYPE ERROR in MACHINE').

convert_typed_varlist([],[],[]).
convert_typed_varlist([b(identifier(ID),Type,_)|T],[ID|IT],[Type|TT]) :-
  %print(ID), print(' : '), print(Type),nl,
  % db : set(couple(global(Name),global(Code)))   a : set(integer)  boolean
  convert_typed_varlist(T,IT,TT).

:- use_module(symmetry_marker,[type_allows_for_precise_marker/1]).
generate_typing_list([],[],[]).
generate_typing_list([Var|VT],[BasicType|TT],[Translation|Rest]) :-
   pretty_type(BasicType,PrettyType),
   (kernel_objects:max_cardinality(BasicType,Card) -> true ; Card='??'),
   (b_get_constant_represented_inside_global_set(Var,BasicType)
     -> PP='] (treated as enumerated);;'
    ; (preference(symmetry_mode,hash),
         \+type_allows_for_precise_marker(BasicType))
       -> PP = '] (IMPRECISE for HASH);;'
       ;  PP = '];;'
   ),
   (non_det_constant(Var) -> string_concatenate('][multiple solutions in SETUP_CONSTANTS',PP,PPN) ; PPN=PP),
   string_concatenate(Card,PPN,P0),
   string_concatenate('  [card=',P0,P00),
   string_concatenate(PrettyType,P00,P1),
   string_concatenate(':',P1,P),
   string_concatenate(Var,P,Translation),
%   mnf(translate:translate_bexpression(b(member(identifier(Var),BasicType),pred,[]),Translation)),
   generate_typing_list(VT,TT,Rest).
  
:- use_module(symmetry_marker,[hash_model_checking_imprecise/2]).
:- use_module(translate,[pretty_type/2]).
tcltk_hash_model_checking_imprecise :-
   hash_model_checking_imprecise(ID,BasicType),!,
   print(' * Note: Hashing may be imprecise for: '),
   print(ID), print(' : '),translate:pretty_type(BasicType,S),print(S), print(' !'),nl.


/* find constants which are assigned multiple values */
:- dynamic non_det_constant/1.
find_non_det_constants :-
   retractall(non_det_constant(_)),
   findall(Csts,visited_expression(_,concrete_constants(Csts),_),CL),
   ((CL=[Valuation1|O],O=[_|_]) -> find_non_det_constants1(Valuation1,O)
    ;  true /* 0 or 1 valuation only */
    ).

find_non_det_constants1([],_).
find_non_det_constants1([bind(C,Value)|T],Other) :- check_cst(Other, C, Value, PeeledOther),
  find_non_det_constants1(T,PeeledOther).

check_cst([],_,_,[]).
check_cst([ [bind(C,OtherVal)|T1] | TT], C, Value, [ T1 | PTT]) :-
   (Value=OtherVal -> check_cst(TT,C,Value,PTT)
    ; assert(non_det_constant(C)), peel2(TT,PTT)).
peel2([],[]).
peel2([ [_|T1] | TT], [ T1 | PTT]) :- peel2(TT,PTT).


/* --------------------------------- */

	
   /* for tcltk: */
:- use_module(pref_definitions,[b_get_definition_string_list_from_machine/2]).
/* --------------------------------- */

%%
% Visualize Invariant...
%
% Siehe bvisual.pl

:- use_module(bvisual).

generate_dot_from_invariant(FileName):-
    bmachine:b_get_invariant_from_machine(BExpr),
    write_dot_file_for_pred(BExpr,FileName).


generate_dot_from_operation(TclOpName,FileName) :-  
    predicate_debugger:adapt_tcl_operation_name(TclOpName,OpName),
    bvisual:get_operation_pre(OpName, BExpr, BecomesSuchVars),
    print_bexpr(BExpr),nl,
    generate_tree_from_boolean_expression(BExpr, Tree,BecomesSuchVars),
    write_dot_graph_to_file(Tree,FileName).
   	

generate_dot_from_properties(FileName) :- 
    (current_state_corresponds_to_initialised_b_machine -> CreateExists=false; CreateExists=true),
    bvisual:get_properties(BExpr,CreateExists), 
    write_dot_file_for_pred(BExpr,FileName).
    
:- use_module(bsyntaxtree,[conjunct_predicates/2]).

generate_dot_from_assertions(FileName) :- 
    ((current_state_corresponds_to_setup_constants_b_machine,
     bmachine:b_get_dynamic_assertions_from_machine(BExpr), BExpr \=[])
      -> true
       ; bmachine:b_get_static_assertions_from_machine(BExpr)
    ),
    conjunct_predicates(BExpr,BE),
    write_dot_file_for_pred(BE,FileName).
    
generate_dot_from_goal(FileName) :-  
    b_get_machine_goal(BExpr),
    print('% CHECKING: '),translate_bexpression(BExpr,PT), print(PT),nl,
    write_dot_file_for_pred(BExpr,FileName).
 
generate_dot_from_deadlock_po(FileName) :-  
    b_state_model_check:get_unsorted_all_guards_false_pred(BExpr),
    print('% CHECKING: '),translate_bexpression(BExpr,PT), print(PT),nl,
    write_dot_file_for_pred(BExpr,FileName).


write_dot_file_for_pred(BExpr,FileName) :- 
    generate_tree_from_boolean_expression(BExpr, Tree,[]),
    print('% Writing to: '), print(FileName),nl,
    mnf(write_dot_graph_to_file(Tree,FileName)).
write_dot_file_for_pred_expr_and_state(BExpr, LocalState, State,FileName) :- 
    mnf(get_tree_from_expr(Tree, BExpr, LocalState, State)),
    print('% Writing to: '), print(FileName),nl,
    mnf(write_dot_graph_to_file(Tree,FileName)).
 
    
    
generate_tree_from_boolean_expression(BExpr,Tree,BSVars):-
    ( 
        current_state_corresponds_to_setup_constants_b_machine,
        current_b_expression(CurState0), 
        % print(cur(CurState0)),nl,
   ~~mnf(insert_before_substitution_variables(BSVars,[],CurState0,CurState0,CurState))
   
        ;
        CurState = []
    ),!,
    mnf(get_tree_from_expr(Tree, BExpr, [], CurState)).

tcltk_bv_get_tops(Tops) :-
    bv_get_top_level(Tops).    
tcltk_bv_get_structure(Id,Label,list(Subs)) :-
    bv_expand_formula(Id,Label,Subs).
tcltk_bv_is_leaf(Id,Label) :-
    bv_expand_formula(Id,Label,[]).
tcltk_bv_get_value(Id,Text,Tag) :-
    current_state_id(StateId),
    bv_get_values([Id],StateId,[Value]),
    retrieve_tag_text(Value,Tag,Text).
retrieve_tag_text(i,       inac,   '').
retrieve_tag_text(e(Text), error,  Text).
retrieve_tag_text(v(Text), value,  Text).
retrieve_tag_text(p(true), ptrue,  true).
retrieve_tag_text(p(false),pfalse, false).

tcltk_bv_get_unlimited_value(Id,Text,Tag) :-
    current_state_id(StateId),
    bv_get_value_unlimited(Id,StateId,Value),
    retrieve_tag_text(Value,Tag,Text1),
    atom_codes(Text1,Text).   % TODO: In the end, the result was converted from codes to atom and back

% -----------------------------------------

% STILL TO DO: refactor eval from here and probcli into separate module

:- dynamic eval_elapsed_time/1.

cputime(T) :- statistics(runtime,[T,_]).

tcltk_time_call(Call) :-
  cputime(T1),Call,cputime(T2),ElapsedTime is T2-T1,
  retractall(eval_elapsed_time(_)),assert(eval_elapsed_time(ElapsedTime)).

:- use_module(eval_strings).  

tcltk_eval(String,StringResult,EnumWarning,Solution) :-
  get_computed_preference(debug_time_out,DTO),
  %print(debug_time_out(DTO)),nl,
  time_out(eval_string(String,StringResult,EnumWarning,LocalState),DTO,TimeOutRes),
  (LocalState=[] -> Solution='' ; translate_bstate(LocalState,Solution)),
  (TimeOutRes=time_out ->
    StringResult = '**** TIME-OUT ****', print(StringResult), print(' ('),print(DTO), print('ms)'),nl,
    EnumWarning = no
    ; print_last_info).

% code to translate an expression into a list of lists that can be displayed as a table in Tk
% TO DO: make more robust and add more fancy translations
tcltk_eval_as_table(String,Res) :-
    atom_codes(String,Codes),
    bmachine:b_parse_machine_expression_from_codes(Codes,Typed,_Type,false,Error), Error=none,
    current_state_id(CurID), eclipse_interface:getState(CurID,EState),
    b_interpreter:b_compute_expression(Typed,[],EState,Value),
    expand(Value, ExpVal),
    Res = list([list(Header)|Table]),
    translate_to_table(ExpVal,Table,Header). %, print(table(Res)),nl.

expand((A,B),[(A,B)]) :- !. % wrap into a set
expand(Value,ExpVal) :- custom_explicit_sets:try_expand_custom_set(Value, ExpVal).

translate_to_table([],[],['Elements']) :- !.
translate_to_table([A|T],[list(LA)|TT],Header) :- !,
    get_table_entry(A,LA,Header),
    translate_to_table(T,TT,_).
translate_to_table(A,[list([AT])],['Value']) :- !,
    translate:translate:translate_bvalue(A,AT).

get_table_entry(A,[AT],['Elements']) :- A \= (_,_),!,
    translate:translate:translate_bvalue(A,AT).
get_table_entry(A,AT,HA) :- %A=(A1,A2), (A1 = (_,_) ; A2=(_,_)),
    !,
    get_table_entry2(A,[],AT,HA).
%get_table_entry((A,B),[AT,BT],['prj1','prj2']) :- !,
%    translate:translate:translate_bvalue(A,AT),
%    translate:translate:translate_bvalue(B,BT).
    
get_table_entry2((A,B),Path,ABT,HAB) :- !,
    get_table_entry2(A,[1|Path],AT,HA),
    get_table_entry2(B,[2|Path],BT,HB),
    append(AT,BT,ABT), append(HA,HB,HAB).
get_table_entry2(A,Path,[AT],[Header]) :- !,
    reverse(Path,RP), ajoin(['prj'|RP],Header), % to do: do all this more efficiently
    translate:translate:translate_bvalue(A,AT).

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).
interruptable_tcltk_eval(String,StringResult,EnumWarning) :- 
   user_interruptable_call_det(user: catch_interrupt_assertion_call(tcltk_eval(String,StringResult,EnumWarning)),InterruptResult),
   (InterruptResult=interrupted -> nl,print('^C'),nl; otherwise -> true).

% -----------------------------------------

% CSP Refinement Search

:- dynamic tcltk_cspm_mode/0.

set_tcltk_cspm_mode :- assert(tcltk_cspm_mode).
unset_tcltk_cspm_mode :- retractall(tcltk_cspm_mode).

get_new_csp_process_id(E,File,ID) :- 
	( parse_single_csp_expression_file(E,File,Proc) -> 
		haskell_csp: check_compiled_term(Proc),	add_csp_process_id1(Proc,ID,animatable_process)
	; add_error(parse_single_csp_expression_file,'Invalid process expression: ',E),fail
	).

parse_and_analyze_process_expression(ProcExp,File,Proc) :- 
	( parse_single_csp_expression_file(ProcExp,File,Proc) -> true
	; add_error(parse_single_csp_expression_file,'Invalid process expression: ',ProcExp),fail
	).

:- use_module(refinement_checker).
get_csp_process_id(Process,ID) :- get_start_transition_term(Process,Trans),
   transition(root,Trans,ID).

:- use_module(probcspsrc(haskell_csp),[animatable_process_without_arguments/1,animatable_process/1,animatable_process_cli/1]).
add_csp_process_id(Process,NewID) :-
    add_csp_process_id1(Process,NewID,animatable_process_cli).

add_csp_process_id1(Process,NewID,Pred) :-
    functor(Call,Pred,1),arg(1,Call,Process),
    (Call
     -> get_start_transition_term(Process,Trans),
        (transition(root,Trans,_,DestID) -> NewID=DestID
         ;  xtl_interface:get_start_expr(Process,NewState),
            add_trans_id(root,Trans,NewState,[],NewID,[],_)
        )
      ; add_error_fail(add_csp_process_id,'Not an animatable CSP Process: ',Process)).

get_start_transition_term('MAIN',R) :- !, R= start_cspm_MAIN.
get_start_transition_term(P,start_cspm(P)).

add_csp_general_process_id(val_of(Process,_),ID) :- !,
   ( animatable_process_without_arguments(Process) % in case the called processes is one without arguments, then we don't have to parse the expression
	-> add_csp_process_id(Process,ID)
   ; animation_mode(cspm),loaded_main_file(CSPFile),parse_and_analyze_process_expression(Process,CSPFile,Proc),
	       add_trans_id(root,start_cspm(Proc),Proc,[],ID,[],_)
   ).

add_csp_general_process_id(GenProcExpr,ID) :-
   % TO DO: maybe strip out source-loc info?
   add_trans_id(root,start_cspm(GenProcExpr),GenProcExpr,[],ID,[],_).

tcltk_csp_in_situ_refinement(ImplProcess,AbsProcess,ResTrace,Style) :-
   format('Checking ~w ~w ~w~n',[AbsProcess,Style,ImplProcess]),
   tcltk_csp_in_situ_refinement(ImplProcess,AbsProcess,ResTrace,Style,10000000).
csp_ref_style('[T=',trace). % Trace Refinement
csp_ref_style('T',trace). % Trace Refinement
csp_ref_style('[SF=',singleton_failures). % Singleton Failures Refinement
csp_ref_style('[F=',failures). % Failures Refinement
csp_ref_style('[R=',refusals). % Refusals Refinement
csp_ref_style('[RD=',refusals_div). % Refusals-Divergence Refinement
csp_ref_style('F',failures). % Failures Refinement
csp_ref_style('[FD=',X) :- csp_ref_style('FailureDivergence',X).
csp_ref_style('FD',X) :- csp_ref_style('FailureDivergence',X).
csp_ref_style('Trace',trace).
csp_ref_style('Failure',failures).
csp_ref_style('FailureDivergence',failure_divergences).
csp_ref_style('RefusalTesting',refusals).
csp_ref_style('RefusalTestingDiv',refusals_div) :- add_error(unsupported_refinement_operator,'[RD= refinement not yet supported; checking [R=').
csp_ref_style('RevivalTesting',revival) :- add_error(unsupported_refinement_operator,'[V= refinement not yet supported.').
csp_ref_style('RevivalTestingDiv',revival_div) :- add_error(unsupported_refinement_operator,'[VD= refinement not yet supported.').

translate_csp_ref_style(Style,RefStyle) :- (csp_ref_style(Style,RefStyle) -> true ;
     add_error_fail(csp,'Unknown refinement style: ',Style)).

% tcltk_csp_in_situ_refinement('P','Q',R,0,1000). tcltk_csp_in_situ_refinement('Q','P',R,1000).
tcltk_csp_in_situ_refinement(ImplProcess,AbsProcess,ResTrace,Style,MaxNrOfNewNodes) :-
    add_csp_process_id(ImplProcess,ImplNodeID),
    add_csp_process_id(AbsProcess,SpecNodeID),
    translate_csp_ref_style(Style,RefStyle),
    refinement_checker:in_situ_ref_search(ImplNodeID,SpecNodeID,ResTrace,RefStyle,MaxNrOfNewNodes).

:- use_module(probcspsrc(haskell_csp),[get_csp_assertions/1,translate_csp_assertion/2, get_csp_assertions_enum/2]).
tcltk_check_csp_assertions([],Summary) :- get_csp_assertions(Assertions),
   treat_csp_ass(Assertions,_List,0,0,0,Summary). % TO DO: display _List

treat_csp_ass([],[],T,F,U,[total/Tot, true/T, false/F, unknown/U]) :- Tot is T+F+U.
treat_csp_ass([Assertion|T],['='(PP,Ok)|TPP],True,False,U,Summary) :- !,
   (checkAssertion(Assertion,PP,_,Ok,_) 
    -> U1=U, (Ok=true -> T1 is True+1, F1=False ; T1=True, F1 is False+1) 
    ;  U1 is U+1, T1=True, F1=False,Ok=unknown),
   treat_csp_ass(T,TPP,T1,F1,U1,Summary).
treat_csp_ass([H|T],TP,T1,F1,U,Summary) :- add_error(treat_csp_ass,'Unknown CSP assertion: ',H),
   U1 is U+1, treat_csp_ass(T,TP,T1,F1,U1,Summary).

% assertModelCheckExt(False,val_of(P3,src_span(25,8,25,10,562,2)),DeadlockFree,F)

run_assertion_check(assertModelCheckExt(Negated,Spec,Type,ModelStyle),Negated,ResTrace) :-
   debug_println(9,assertModelCheckExt(Negated,Spec,Type,ModelStyle)),
   add_csp_general_process_id(Spec,SpecNodeID),
   translate_csp_ref_style(ModelStyle,RefStyle),
   refinement_checker:in_situ_model_check(SpecNodeID,ResTrace,Type,RefStyle,10000000),
   print(result_trace),nl,print(ResTrace),nl.
run_assertion_check(assertModelCheck(Negated,Spec,Style),Negated,ResTrace) :-
   run_assertion_check(assertModelCheckExt(Negated,Spec,Style,'FailureDivergence'),Negated,ResTrace).
run_assertion_check(assertRef(Negated,Spec,Style,Impl,Span),Negated,ResTrace) :-
   debug_println(9,assertRef(Negated,Spec,Style,Impl,Span)),
   add_csp_general_process_id(Spec,SpecNodeID),
   add_csp_general_process_id(Impl,ImplNodeID),
   translate_csp_ref_style(Style,RefStyle),
   refinement_checker:in_situ_ref_search(ImplNodeID,SpecNodeID,ResTrace,RefStyle,10000000).

checkAssertion(Assertion,PP,Negated,Ok,ResTrace) :-
   print('CHECKING '),
   translate_csp_assertion(Assertion,PP), 
   print(PP),nl,
   debug_println(9,checkAssertion(Assertion)),
   reset_refinement_checker,
   (run_assertion_check(Assertion,Negated,ResTrace)
     -> debug_println(9,result(ResTrace)),
       (ResTrace=no_counter_example
           -> (Negated = 'False' -> Ok=true
                ; %add_error(check_assertRef,'Assertion FAILS (expected refinement not to hold): ',PP),
                  Ok=false)
           ; (ResTrace=[interrupted|_]
                 -> Ok=interrupted
                 ;  (Negated = 'False' -> %add_error(check_assertRef,'Assertion FAILS: ',PP),
                                    Ok=false ; Ok=true)
             )
        ),
        (Ok=true -> print('OK'),nl ; (Ok=interrupted -> print('*** Execution interrupted by user ***'),nl
                                                  ;  print('*** FALSE ***')),nl
                                  )
     ;  add_error_fail(checkAssertion,'Internal Error, Assertion Checking Failed for: ',PP)
    ).

/* For starting assertion check we have to call the checkAssertion/5 predicate with the assertion predicate (assertRef/5, assertModelCheck/3 or assertModelCheckExt/4) 
    from the Csp prolog file. Tcl_Tk does not know anything about the assertion syntax in prolog, that's why each assertion is refered to a number, which
    is refered to the same number in the listbox of the 'Check Csp Assertion' window in Tcl_Tk. Therefore we need the predicate run_assertion_check_by_number/4 here. */ 
run_assertion_check_by_number(CSPFile,Number, Negated, Result) :-
      get_csp_assertions_enum(CSPFile,CSP_Assertions),
      %print('Assertions: '),print(CSP_Assertions),nl,
      run_assertion_check_by_num(Number, Negated, Result, CSP_Assertions).
run_assertion_check_by_num(Number,_,_Result,[]) :-
   add_error_and_fail(run_assertion_check_by_num,'No Assertion could be found under this number: ', Number).
run_assertion_check_by_num(Number, Negated, Result,[(H,Number)|_]) :- !,
      % print('checking_assertion: '),print(H),nl, %%
      checkAssertion(H,_,Negated,_Ok,Result).
run_assertion_check_by_num(Number, Negated, Result,[(_,N)|T]):- Number \= N,
      run_assertion_check_by_num(Number,Negated,Result,T).

tcltk_machine_has_assertions :-
    ( b_get_static_assertions_from_machine([_|_]); b_get_dynamic_assertions_from_machine([_|_]) ), !.
    
tcltk_unique_top_level_operation(Name) :- b_top_level_operation(Name),
  \+ (b_top_level_operation(N2), N2 \= Name).
