% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(testcase_generator,[generate_testcase_graph/1, compute_testcases/0]).
/* File: testcase_generator.pl */
/* Created: 1/6/06 by Michael Leuschel */

:-use_module(dot_graph_generator).
:-use_module(library(lists)).
% use_module(testcase_generator),testcase_generator:cs.

:- dynamic testcase_nr/1.
:- dynamic testcase_config/2.
:- dynamic testcase_operation/3.

reset_testcase :-
   retractall(testcase_nr(_)),
   assert(testcase_nr(0)),
   retractall(testcase_config(_,_)), retractall(testcase_operation(_,_,_)),
   retractall(max_nr_of_new_test_nodes(_)),
   assert(max_nr_of_new_test_nodes(0)).

add_testcase_config(Config) :-
  retract(testcase_nr(N)),
  N1 is N+1,
  assert(testcase_nr(N1)),
  assert(testcase_config(Config,N1)).
  
/* ------------------------------------------------ */
  
% A Testcase configuration is a set of states
% Intuitively: the implementation can choose which one

compute_successors([X|TConfig],FullLabel,NewConfig) :-
  /* find a possible enabled operation */
  setof(Lbl,R^X2^impl_test_trans(X,Lbl,R,X2),Labels),
  %print(labels(X,Labels)),nl,
  member(Label,Labels),
  possible_in_all_states([X|TConfig],Label,NewConfigs),
  %print(label_successful(Label,NewConfigs)),nl,
  member(config(ReturnValue,NewConfig),NewConfigs),
  (ReturnValue = [] -> FullLabel = Label ; FullLabel='-->'(Label,ReturnValue)).


possible_in_all_states([],_,[]).
possible_in_all_states([X|T],Label,AllSuccs) :- 
 possible_in_all_states(T,Label,TSuccs),
 %print(chck(X,Label)),
% setof(X2,R^impl_test_trans(X,Label,R,X2),XSuccs),
 findall(edge(R,X2),impl_test_trans(X,Label,R,X2),XSuccs),
 %print(succs(XSuccs)),nl,
 XSuccs \= [],
 insert_configs(XSuccs,TSuccs,AllSuccs).
 
 
merge([],T,T).
merge([X|T],L,Res) :- insert(L,X,L2), merge(T,L2,Res).
 
insert([],State,[State]).
insert([N|T],State,Res) :-
  (State=N -> Res = [N|T]
   ;(State @<N
     -> Res = [State,N|T]
     ; (Res = [N|TT],
       insert(T,State,TT))
     )
  ).
 
insert_config_with_return_value([],R,State,[config(R,[State])]).
insert_config_with_return_value([config(R1,States)|T], R,State, OutList) :-
  (R=R1 -> (OutList = [config(R1,NewStates)|T], insert(States,State,NewStates))
   ;(R @<R1
     -> OutList = [config(R,[State]),config(R1,States)|T]
     ; (OutList = [config(R1,States)|TT],
       insert_config_with_return_value(T,R,State,TT))
     )
  ).
  
insert_configs([],R,R).
insert_configs([edge(R,State)|T],In,Out) :- 
   remove_non_observable_values(R,NewResult),
  insert_config_with_return_value(In,NewResult,State,O2),
  insert_configs(T,O2,Out).

remove_non_observable_values([],[]).
remove_non_observable_values([X|T],Res) :-
   (observable(X) -> Res = [X|RT] ; Res = RT),
   remove_non_observable_values(T,RT).

observable(pred_false /* bool_false */).
observable(pred_true /* bool_true */).

/* ------------------------------------------------ */

cs :- generate_testcase_graph('~/Desktop/testcasegraph.dot').

generate_testcase_graph(File) :-
      reset_testcase,
      treat_config([root]),
      print(finished),nl,
      print_covered_ops,
      gen_dot_graph(File,testcase_generator,dot_node,dot_trans,dot_same_rank,dot_subgraph).

dot_node(ID,none,Desc,box,none,blue) :-
  testcase_config(Config,ID), Desc=Config.
  
dot_trans(NodeID,Label,SuccNodeID,Color) :-
  testcase_config(Config,NodeID),
  testcase_operation(Config,Label,SuccNodeConfig),
  testcase_config(SuccNodeConfig,SuccNodeID),
  (Label = '-->'(_,_) -> Color=green ; Color=black).
  
dot_same_rank(_) :- fail.
dot_subgraph(_,_,_) :- fail.

 :- use_module(bmachine,[b_get_machine_operation/4]).

print_covered_ops :- 
   print('NUMBER OF CONFIGURATIONS: '),
   testcase_nr(X),
   print(X),nl,
   print('COVERED OPERATIONS: '),nl, print('  {'),
   compute_covered_operations,
   covered_op(OpName,_Arity),
   print(' '), print(OpName),
   fail.
print_covered_ops :- print(' }'),nl,
   print('UNCOVERED OPERATIONS: '),nl, print('  {'),
   b_get_machine_operation(OpName,_R,_P,_B),
   \+ covered_op(OpName,_),
   print(' '), print(OpName),
   fail.
print_covered_ops :- print(' }'),nl,nl.


:- dynamic covered_op/2.

compute_covered_operations :- 
   retractall(covered_op(_,_)),
   testcase_operation(_,Label,_),
   label_functor(Label,OpName,Arity),
   \+ covered_op(OpName,_),
   assert(covered_op(OpName,Arity)),
   fail.
compute_covered_operations.
 
       
label_functor('-->'(Op,_Return),OpName,Arity) :- !,
   functor(Op,OpName,Arity).
label_functor(Op,OpName,Arity) :-
   functor(Op,OpName,Arity).
   
 
compute_testcases :-
      reset_testcase,
      print('Normalising state space'),nl,
      treat_config([root]),
      print(finished),nl,
      print_covered_ops,
      compute_coverage.
      
      

:- dynamic change/0.
:- dynamic can_cover/3.


init_can_cover :- print(init_can_cover),nl,
   retractall(can_cover(_,_,_)),
   testcase_operation(ID,Label,_),
   label_functor(Label,OpName,_),
   findall(PP,get_postlude(ID,PP),Postlude),
   add_can_cover(ID,OpName,[Label,'POSTLUDE'|Postlude]),
   fail.
init_can_cover.


add_can_cover(ID,OpName,Trace) :-
  \+ can_cover(ID,OpName,_),
  assert(can_cover(ID,OpName,Trace)), assert(change),
  %print(can_cover(ID,OpName)),nl,
  testcase_operation(PreID,Label,ID), /* add can_cover to all predecessors */
  (Label \= '-->'(Op,Return)
    -> \+ (testcase_operation(PreID,'-->'(Op,R2),_ID2),R2\=Return)
    ; true
  ),
  add_can_cover(PreID,OpName,[Label|Trace]).
   
propagation_loop :-
   print(propagating),nl,
   propagate_split_can_cover,
   fail.
propagation_loop :-
   (change -> propagation_loop ; (print('fixpoint reached'),nl)).

propagate_split_can_cover :-
   retractall(change),
   can_cover(ID,OpName,T1),
   testcase_operation(PreID,'-->'(_Label,R1),ID),
   \+ can_cover(PreID,OpName,_),
   propagate_split_can_cover2(PreID,OpName,R1,ID,T1).

propagate_split_can_cover2(PreID,OpName,R1,_ID,T1) :-   
   testcase_operation(PreID,'-->'(Label,R2),ID2),
   R2 \= R1,
   can_cover(ID2,OpName,T2),
   add_can_cover(PreID,OpName,[split(Label,R1,T1,R2,T2)]).
   /* Missing: add splits for more than one return value ! */
propagate_split_can_cover2(PreID,OpName,R1,_ID,T1) :- 
   \+ (testcase_operation(PreID,'-->'(Label,R2),_), R2 \=R1),
    /* there is only one possible return value */
   add_can_cover(PreID,OpName,[check_return_value(Label,R1,T1)]).
   

/*
   testcase_generator:testcase_operation(ID,Label,_),
   testcase_generator:label_functor(Label,OpName,_),
 testcase_generator:testcase_operation(ID,'-->'(Label,R1),ID) */
 
get_postlude(ID,'-->'(Label,R1)) :- 
  testcase_operation(ID,'-->'(Label,R1),ID),
   \+ (testcase_operation(ID,'-->'(Label,R2),_), R2 \=R1).
   /* we have a probing operation with only a single return value R1
      which is always applicable */ 


compute_coverage :-  /* compute which operations can actually be covered
 by test case traces or test case trees */
   init_can_cover,
   propagation_loop,
   print_root_coverage.
   
:- use_module(translate,[translate_events/2]).

print_root_coverage :- print('REALLY COVERED OPERATIONS WITH TEST CASES: '),nl,
   can_cover([root],OpName,Trace),
   print('* '),print(OpName),print(' :'),nl,
   print('   '),translate_events(Trace,PPTrace),print(PPTrace),nl,
   last(Trace,config(ID)),
   print_postlude(ID),
   fail.
print_root_coverage :-  nl,
   print('UNCOVERED OPERATIONS: '),nl, print('  {'),
   b_get_machine_operation(OpName,_R,_P,_B),
   \+ can_cover([root],OpName,_),
   print(' '), print(OpName),
   fail.
print_root_coverage :- print(' }'),nl,nl.

/* ------------------------------------------------ */

treat_config(X) :-
  \+ testcase_config(X,_),
  add_testcase_config(X),
  print(treating(X)),nl,
  compute_successors(X,Label,Succ), 
  % X \= Succ,
  assert(testcase_operation(X,Label,Succ)),
  treat_config(Succ),
  fail.
treat_config(_).  

:- dynamic max_nr_of_new_test_nodes/1.
max_nr_of_new_test_nodes(0).

impl_test_trans(X,Label,Return,X2) :-  /* need to improve efficiency of that : */
   (state_space_open_nodes:retract_open_node(X)
     -> ((retract(max_nr_of_new_test_nodes(Max)),Max>0)
          -> (M1 is Max-1, assert(max_nr_of_new_test_nodes(M1)),
              user:do_one_gui_event,
              print(computing(X,Max)),nl,
              user:compute_all_transitions(X))
          ;  fail
        )
     ; true),
   user:transition(X,Action,X2),
   (functor(Action,'$initialise_machine',_)
      -> (Label = '$initialise_machine',Return=[])
       ; (functor(Action,'$setup_constants',_)
          -> (Label = '$setup_constants',Return=[])
            ; (Action='-->'(Op,Return)
                -> (Label = Op)
                ;   (Label = Action,Return=[])))).