% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(translate,
          [print_bexpr/1, debug_print_bexpr/1, nested_print_bexpr/1, print_bexpr_stream/2,
           print_components/1,
           print_bexpr_with_limit/2,
           print_unwrapped_bexpr_with_limit/1,print_bvalue/1, l_print_bvalue/1, print_bvalue_stream/2,
           translate_params_for_dot/2,
           print_machine/1,
           translate_machine/3,
           translate_bexpression/2,
           translate_bexpression_with_limit/3, translate_bexpression_with_limit/2,
           translate_bexpression_to_codes/2,
           translate_substitution/2, print_subst/1,
           translate_bvalue/2, translate_bvalue_with_limit/3,
           translate_bvalue_with_type/3,
           print_state/1,
           translate_bstate/2, print_bstate/1, translate_context/2, print_context/1,
           translate_any_state/2,
           print_value_variable/1,
           print_cspm_state/1, translate_cspm_state/2, print_csp_value/1, pp_csp_value/3,
	   translate_cspm_expression/2,
           translate_properties_with_limit/2,
           translate_event/2,translate_events/2,
           translate_state_errors/2,translate_state_error/2,
           translate_event_error/2,
           translate_prolog_constructor/2,
           pretty_type/2,
           explain_state_error/3,
           explain_event_trace/2,
           
           print_span/1,
       
           set_type_to_maximal_texpr/2, type_set/2, % should probably be moved to other module

           translate_error_term/2,

           set_translation_constants/1, set_translation_context/1,
           clear_translation_constants/0,

           set_print_type_infos/1,
           suppress_rodin_positions/0,

           install_b_portray_hook/0,remove_b_portray_hook/0,

           translate_eventb_to_classicalb/3,
           return_csp_closure_value/2
           ]).

:- use_module(tools).

:- use_module(module_information).
:- module_info(group,tools).
:- module_info(description,'This module is responsible for pretty-printing B and CSP, source spans, ...').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).
:- use_module(library(codesio)).
:- use_module(library(terms)).
:- use_module(library(avl)).

:- use_module(debug).
:- use_module(error_manager).
:- use_module(self_check).
:- use_module(b_global_sets).
:- use_module(specfile,[csp_with_bz_mode/0,process_algebra_mode/0,
                        animation_minor_mode/1,set_animation_minor_mode/1,
                        animation_mode/1,set_animation_mode/1, csp_mode/0]).
:- use_module(bsyntaxtree).
%:- use_module('smv/smv_trans',[smv_print_initialisation/2]).
:- use_module(preferences,[get_preference/2]).
:- use_module(bmachine_structure).

:- use_module(plugins,[plugin_prettyprint_transition/3,active_plugin/1,
                       plugin_prettyprint_property/3,
		       plugin_prettyprint_value/3]).



print_unwrapped_bexpr_with_limit(Expr) :- print_unwrapped_bexpr_with_limit(Expr,200).
print_unwrapped_bexpr_with_limit(Expr,Limit) :-
   translate:print_bexpr_with_limit(b(Expr,pred,[]),Limit),nl.
debug_print_bexpr(E) :- debug:debug_mode(off) -> true ; print_bexpr(E).
print_bexpr(Expr) :- translate_bexpression(Expr,R), print(R).
print_bexpr_with_limit(Expr,Limit) :- translate_bexpression_with_limit(Expr,Limit,R), print(R).
print_bvalue(Val) :- translate_bvalue(Val,TV), print(TV).
print_bexpr_stream(S,Expr) :- translate_bexpression(Expr,R), write(S,R).
print_bvalue_stream(S,Val) :- translate_bvalue(Val,R), write(S,R).

print_components(C) :- print_components(C,0).
print_components([],Nr) :- print('Nr of components: '),print(Nr),nl.
print_components([component(Pred,Vars)|T],Nr) :- N1 is Nr+1,
   print('Component: '), print(N1), print(' over '), print(Vars),nl,
   print_bexpr(Pred),nl, 
   print_components(T,N1).
   
l_print_bvalue([]).
l_print_bvalue([H|T]) :- print_bvalue(H), print(' : '),l_print_bvalue(T).


nested_print_bexpr(Expr) :- nbp(Expr,conjunct,0).

nbp(b(E,_,_),Type,Indent) :- nbp2(E,Type,Indent).
nbp2(conjunct(LHS,RHS),Type,Indent) :- !,inc_indent(conjunct,Type,Indent,NIndent),
  nbp(LHS,conjunct,NIndent), print_indent(NIndent), print('&'),nl,
  nbp(RHS,conjunct,NIndent).
nbp2(disjunct(LHS,RHS),Type,Indent) :- !,inc_indent(disjunct,Type,Indent,NIndent),
  nbp(LHS,disjunct,NIndent), print_indent(NIndent), print('or'),nl,
  nbp(RHS,disjunct,NIndent).
nbp2(negation(LHS),_Type,Indent) :- !,inc_indent(negation,false,Indent,NIndent),
  print_indent(Indent), print('not('),nl,
  nbp(LHS,negation,NIndent),
  print_indent(Indent), print(')'),nl.
nbp2(E,_,Indent) :- print_indent(Indent), print_bexpr(b(E,pred,[])),nl.

inc_indent(Type,CurType,I,NewI) :- (Type=CurType -> NewI=I ; NewI=s(I)).

print_indent(s(X)):- !,
    print('  '),
    print_indent(X).
print_indent(_).

/* =============================================== */
/* Translating expressions and values into strings */
/* =============================================== */



  
  
translate_params_for_dot(List,TransList) :- translate_params_for_dot(List,TransList,-3).
translate_params_for_dot([],'',_).
translate_params_for_dot([H|T],Res,Nr) :-
  translate_property_with_limit(H,100,TH),
  (Nr>2 -> N1=1 ; N1 is Nr+1),
  translate_params_for_dot(T,TT,N1),
  string_concatenate(TH,TT,Res1),
  (N1=1 -> string_concatenate(',\n',Res1,Res)
    ; (Nr>(-3) -> string_concatenate(',',Res1,Res) ; Res=Res1)).
  
translate_channel_values(X,['_'|T],T) :- var(X),!.
translate_channel_values([],S,S) :- !.
translate_channel_values([tuple([])|T],S,R) :- !,
  translate_channel_values(T,S,R).
translate_channel_values([in(tuple([]))|T],S,R) :- !,
  translate_channel_values(T,S,R).
translate_channel_values([H|T],['.'|S],R) :-  !,
  pp_csp_value(H,S,S2),
  translate_channel_values(T,S2,R).
translate_channel_values(tail_in(X),S,T) :-
      (X=[] ; X=[_|_]), !, translate_channel_values(X,S,T).
translate_channel_values(_X,['??'|S],S).



pp_single_csp_value(V,'_') :- var(V),!.
pp_single_csp_value(X,'_cyclic_') :- cyclic_term(X),!.
pp_single_csp_value(int(X),A) :- atomic(X),!,number_chars(X,Chars),atom_chars(A,Chars).

:- assert_must_succeed((translate_cspm_expression(listExp(rangeOpen(2)),R), ajoin(R,R1), R1 == '<2..>')).
:- assert_must_succeed((translate_cspm_expression(listFrom(2),R), ajoin(R,R1), R1 == '<2..>')).
:- assert_must_succeed((translate_cspm_expression(listFromTo(2,6),R), ajoin(R,R1), R1 == '<2..6>')).
:- assert_must_succeed((translate_cspm_expression(setFromTo(2,6),R), ajoin(R,R1), R1 == '{2..6}')).
:- assert_must_succeed((translate_cspm_expression('#'(listFromTo(2,6)),R), ajoin(R,R1), R1 == '#<2..6>')).
:- assert_must_succeed((translate_cspm_expression(inGuard(x,setFromTo(1,5)),R), ajoin(R,R1), R1 == 'x:{1..5}')).
:- assert_must_succeed((translate_cspm_expression(builtin_call(int(3)),R), ajoin(R,R1), R1 == '3')).
:- assert_must_succeed((translate_cspm_expression(set_to_seq(setValue([int(1),int(2)])),R), ajoin(R,R1), R1 == 'seq({1,2})')).
:- assert_must_succeed((translate_cspm_expression(diff(setValue([int(1)]),setValue([])),R), ajoin(R,R1), R1 == 'diff({1},{})')).
:- assert_must_succeed((translate_cspm_expression(inter(setValue([int(1)]),setValue([])),R), ajoin(R,R1), R1 == 'inter({1},{})')).
:- assert_must_succeed((translate_cspm_expression(lambda([x,y],'*'(x,y)),R), ajoin(R,R1), R1 == '\\ x,y @ (x*y)')).
:- assert_must_succeed((translate_cspm_expression(lambda([x,y],'/'(x,y)),R), ajoin(R,R1), R1 == '\\ x,y @ (x/y)')).
:- assert_must_succeed((translate_cspm_expression(lambda([x,y],'%'(x,y)),R), ajoin(R,R1), R1 == '\\ x,y @ (x%y)')).
:- assert_must_succeed((translate_cspm_expression(rename(x,y),R), ajoin(R,R1), R1 == 'x <- y')).
:- assert_must_succeed((translate_cspm_expression(link(x,y),R), ajoin(R,R1), R1 == 'x <-> y')).
:- assert_must_succeed((translate_cspm_expression(agent_call_curry(f,[[a,b],[c]]),R), ajoin(R,R1), R1 == 'f(a,b)(c)')).

translate_cspm_expression(Expr, Text) :-
	(pp_csp_value(Expr,Text,[]) -> true
	; print('Pretty printing expression failed: '),print(Expr),nl).

pp_csp_value(X,[A|S],S) :- pp_single_csp_value(X,A),!.
pp_csp_value(setValue(L),['{'|S],T) :- !,pp_csp_value_l(L,',',S,['}'|T],inf).
pp_csp_value(setExp(rangeEnum(L)),['{'|S],T) :- !,pp_csp_value_l(L,',',S,['}'|T],inf).
pp_csp_value(avl_set(A),['{'|S],T) :- !,avl_domain(A,L),pp_csp_value_l(L,',',S,['}'|T],inf).
pp_csp_value(setExp(rangeClosed(L,U)),['{'|S],T) :- !, pp_csp_value(L,S,['..'|S2]),pp_csp_value(U,S2,['}'|T]).
pp_csp_value(setExp(rangeOpen(L)),['{'|S],T) :- !, pp_csp_value(L,S,['..}'|T]).
pp_csp_value(listExp(rangeEnum(L)),['<'|S],T) :- !,pp_csp_value_l(L,',',S,['>'|T],inf).
pp_csp_value(listExp(rangeClosed(L,U)),['<'|S],T) :- !, pp_csp_value(L,S,['..'|S2]),pp_csp_value(U,S2,['>'|T]).
pp_csp_value(listExp(rangeOpen(L)),['<'|S],T) :- !, pp_csp_value(L,S,['..>'|T]).
pp_csp_value(setFromTo(L,U),['{'|S],T) :- !,
   pp_csp_value(L,S,['..'|S2]),pp_csp_value(U,S2,['}'|T]).
pp_csp_value(setFrom(L),['{'|S],T) :- !,
   pp_csp_value(L,S,['..}'|T]).
pp_csp_value(closure(L), ['{|'|S],T) :- !,pp_csp_value_l(L,',',S,['|}'|T],inf).
pp_csp_value(list(L),['<'|S],T) :- !,pp_csp_value_l(L,',',S,['>'|T],inf).
pp_csp_value(listFromTo(L,U),['<'|S],T) :- !,
   pp_csp_value(L,S,['..'|S2]),pp_csp_value(U,S2,['>'|T]).
pp_csp_value(listFrom(L),['<'|S],T) :- !,
   pp_csp_value(L,S,['..>'|T]).
pp_csp_value('#'(L),['#'|S],T) :- !,pp_csp_value(L,S,T).
pp_csp_value('^'(X,Y),S,T) :- !,pp_csp_value(X,S,['^'|S1]), pp_csp_value(Y,S1,T).
pp_csp_value(linkList(L),S,T) :- !,pp_csp_value_l(L,',',S,T,inf).
pp_csp_value(in(X),S,T) :- !,pp_csp_value(X,S,T).
pp_csp_value(inGuard(X,Set),S,T) :- !,pp_csp_value(X,S,[':'|S1]),
  pp_csp_value(Set,S1,T).
pp_csp_value(out(X),['!'|S],T) :- !,pp_csp_value(X,S,T).
pp_csp_value(alsoPat(X,_Y),S,T) :- !,pp_csp_value(X,S,T).
pp_csp_value(appendPat(X,_Fun),S,T) :- !,pp_csp_value(X,S,T).
pp_csp_value(int(X),S,T) :- !,pp_csp_value(X,S,T).
pp_csp_value(tuple(vclosure),S,T) :- !, S=T.
pp_csp_value(tuple([X]),S,T) :- !,pp_csp_value(X,S,T).
pp_csp_value(tuple([X|vclosure]),S,T) :- !,pp_csp_value(X,S,T).
pp_csp_value(tuple([H|TT]),S,T) :- !,pp_csp_value(H,S,['.'|S1]),pp_csp_value(tuple(TT),S1,T).
pp_csp_value(dotTuple([H]),S,T) :- !, pp_csp_value(H,S,T).
pp_csp_value(dotTuple([H|TT]),S,T) :- !, pp_csp_value(H,S,['.'|S1]),
  pp_csp_value(dotTuple(TT),S1,T).
pp_csp_value(tupleExp(Args),S,T) :- !,pp_csp_args(Args,S,T,'(',')').
pp_csp_value(na_tuple(Args),S,T) :- !,pp_csp_args(Args,S,T,'(',')').
pp_csp_value(record(Name,Args),['('|S],T) :- !,pp_csp_value(tuple([Name|Args]),S,[')'|T]).
pp_csp_value(val_of(Name,_Span),S,T) :- !, pp_csp_value(Name,S,T).
pp_csp_value(builtin_call(X),S,T) :- !,pp_csp_value(X,S,T).
pp_csp_value(seq_to_set(X),['set('|S],T) :- !,pp_csp_value(X,S,[')'|T]).
pp_csp_value(set_to_seq(X),['seq('|S],T) :- !,pp_csp_value(X,S,[')'|T]).
%pp_csp_value('\\'(B,C,S),S1,T) :- !, pp_csp_process(ehide(B,C,S),S1,T).
pp_csp_value(agent_call(_Span,Agent,Parameters),['('|S],T) :- !,
  pp_csp_value(Agent,S,S1),
  pp_csp_args(Parameters,S1,[')'|T],'(',')').
pp_csp_value(agent_call_curry(Agent,Parameters),S,T) :- !,
  pp_csp_value(Agent,S,S1),
  pp_csp_curry_args(Parameters,S1,T).
pp_csp_value(lambda(Parameters,Body),['\\ '|S],T) :- !,
  pp_csp_args(Parameters,S,[' @ '|S1],'',''),
  pp_csp_value(Body,S1,T).
pp_csp_value(rename(X,Y),S,T) :- !,pp_csp_value(X,S,[' <- '|S1]),
  pp_csp_value(Y,S1,T).
pp_csp_value(link(X,Y),S,T) :- !,pp_csp_value(X,S,[' <-> '|S1]),
  pp_csp_value(Y,S1,T).
% binary operators:
pp_csp_value(diff(A1,A2),[diff,'('|S],T) :- !,pp_csp_args([A1,A2],S,[')'|T],'','').
pp_csp_value(inter(A1,A2),[inter,'('|S],T) :- !,pp_csp_args([A1,A2],S,[')'|T],'','').
pp_csp_value(Expr,['('|S],T) :- bynary_numeric_operation(Expr,E1,E2,OP),!,
    pp_csp_value(E1,S,[OP|S2]),
    pp_csp_value(E2,S2,[')'|T]).
% unary operators:
pp_csp_value(card(A1),[card,'('|S],T) :- !,pp_csp_value(A1,S,[')'|T]).
pp_csp_value('Union'(A1),['Union','('|S],T) :- !,pp_csp_value(A1,S,[')'|T]).
pp_csp_value('Inter'(A1),['Inter','('|S],T) :- !,pp_csp_value(A1,S,[')'|T]).
% vclosure
pp_csp_value(Expr,S,T) :- is_list(Expr),!,pp_csp_value(closure(Expr),S,T).
% Argument of function can be process
pp_csp_value(Expr,S,T) :- pp_csp_process(Expr,S,T),!.
pp_csp_value(Expr,S,T) :- csp_with_bz_mode,!,pp_value(Expr,S,T).
pp_csp_value(X, [A|S], S) :- % ['<< ',A,' >>'|S],S) :-  % the << >> pose problems when checking against FDR
    write_to_codes(X,Codes),atom_codes_with_limit(A,Codes).

print_csp_value(Val) :- pp_csp_value(Val,Atoms,[]), ajoin(Atoms,Text),
  print(Text).

return_csp_closure_value(closure(S),List) :- pp_csp_value_l1(S,List).
return_csp_closure_value(setValue(S),List) :- pp_csp_value_l1(S,List).

pp_csp_value_l1([Expr|Rest],List) :- 
    ( nonvar(Rest),Rest=[] ->
       pp_csp_value(Expr,T,[]),ajoin(T,Value),List=[Value]
    ;  pp_csp_value_l1(Rest,R),pp_csp_value(Expr,T,[]),ajoin(T,Value),List=[Value|R]
    ).

pp_csp_args([],T,T,_LPar,_RPar).
pp_csp_args([H|TT],[LPar|S],T,LPar,RPar) :- pp_csp_value(H,S,S1), pp_csp_args2(TT,S1,T,RPar).
pp_csp_args2([],[RPar|T],T,RPar).
pp_csp_args2([H|TT],[','|S],T,RPar) :- pp_csp_value(H,S,S1), pp_csp_args2(TT,S1,T,RPar).

pp_csp_curry_args([],T,T).
pp_csp_curry_args([H|TT],S,T) :- is_list(H), pp_csp_args(H,S,S1,'(',')'), pp_csp_curry_args(TT,S1,T).

pp_csp_value_l(V,_Sep,['...'|S],S,N) :- (var(V) ; (N \= inf -> N<1;fail)), !.
pp_csp_value_l([],_Sep,S,S,_).
pp_csp_value_l([Expr|Rest],Sep,S,T,Nr) :-
    ( nonvar(Rest),Rest=[] ->
        pp_csp_value(Expr,S,T)
    ; otherwise ->
        (Nr=inf -> N1 = Nr ; N1 is Nr-1),
        pp_csp_value(Expr,S,[Sep|S1]),pp_csp_value_l(Rest,Sep,S1,T,N1)).

:- assert_must_succeed((translate:convert_set_into_sequence([(int(1),int(5))],Seq),
                        Seq == [int(5)])).
:- assert_must_succeed((translate:convert_set_into_sequence([(int(2),X),(int(1),int(5))],Seq),
                        Seq == [int(5),X])).

convert_set_into_sequence(Set,Seq) :-
    nonvar(Set), \+ animation_minor_mode(eventb),
	convert_set_into_sequence1(Set,Seq).
convert_set_into_sequence1(avl_set(A),Seq) :- !,
     avl_size(A,Sz),size_is_in_set_limit(Sz),convert_avlset_into_sequence(A,Seq).
convert_set_into_sequence1([],Seq) :- !, Seq=[].
convert_set_into_sequence1(Set,Seq) :-
    convert_set_into_sequence2(Set,0,_,SetElems,Seq),ground(SetElems).
convert_set_into_sequence2([],_Max,([],[]),_,_Seq).
convert_set_into_sequence2([Pair|T],Max,Last,SetElems,Seq) :-
    nonvar(Pair),nonvar(T),Pair=(Index,H),ground(Index),
    Index=int(Nr),
    insert_el_into_seq(Nr,H,Seq,SetElems,L),
    (Nr>Max -> NMax=Nr,NLast=L ; NMax=Max,NLast=Last),
    convert_set_into_sequence2(T,NMax,NLast,SetElems,Seq).
insert_el_into_seq(1,H,[H|L],[set|L2],(L,L2)) :- !.
insert_el_into_seq(N,H,[_|T],[_|T2],Last) :- N>1, N1 is N-1, insert_el_into_seq(N1,H,T,T2,Last).

convert_avlset_into_sequence(Avl,Sequence) :-
    \+ animation_minor_mode(eventb),
    convert_avlset_into_sequence2(Avl,1,Sequence).
convert_avlset_into_sequence2(Avl,_Index,[]) :-
    empty_avl(Avl),!.
convert_avlset_into_sequence2(Avl,Index,[Head|Tail]) :-
    avl_del_min(Avl, Min, _ ,NewAvl),
    nonvar(Min), Min=(L,Head),
    ground(L), L=int(Index),
    Index2 is Index + 1,
    convert_avlset_into_sequence2(NewAvl,Index2,Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translate new syntax tree -- work in progress
:- assert_must_succeed((translate_cspm_state(lambda([x,y],'|~|'(prefix(_,[],x,skip(_),_),prefix(_,[],y,skip(_),_),_)),R), R == 'CSP: \\ x,y @ x->SKIP |~| y->SKIP')).
:- assert_must_succeed((translate_cspm_state(agent_call_curry('F',[[a,b],[c]]),R), R == 'CSP: F(a,b)(c)')).
:- assert_must_succeed((translate_cspm_state(ifte(bool_not('<'(x,3)),';'(esharing([a],'/\\'('P1','P2',span),procRenaming([rename(r,s)],'Q',span),span),lParallel([link(b,c)],'R','S',span),span),'[>'(elinkParallel([link(h1,h2)],'G1','G2',span),exception([a],'H1','H2',span),span),span1,span2,span3),R), R == 'CSP:  if not((x<3)) then P1 /\\ P2 [|{|a|}|] Q[[r <- s]] ; R [{|b <-> c|}] S else G1 [{|h1 <-> h2|}] G2 [> H1 [|{|a|}|> H2')).
:- assert_must_succeed((translate_cspm_state(aParallel([a,b],'P',[b,c],'Q',span),R), R == 'CSP: P [{|a,b|} || {|b,c|}] Q')).
:- assert_must_succeed((translate_cspm_state(eaParallel([a,b],'P',[b,c],'Q',span),R), R == 'CSP: P [{|a,b|} || {|b,c|}] Q')).
:- assert_must_succeed((translate_cspm_state(eexception([a,b],'P','Q',span),R), R == 'CSP: P [|{|a,b|}|> Q')).


print_cspm_state(State) :- translate_cspm_state(State,T), print(T).


translate_cspm_state(State,Text) :- 
   ( pp_csp_process(State,Atoms,[]) -> true
     ; print(pp_csp_process_failed(State)),nl,Atoms=State),
   ajoin(['CSP: '|Atoms],Text).

pp_csp_process(skip(_Span),S,T) :- !, S=['SKIP'|T]. 
pp_csp_process(stop(_Span),S,T) :- !, S=['STOP'|T]. 
pp_csp_process('CHAOS'(_Span,Set),['CHAOS('|S],T) :- !,
  pp_csp_value(Set,S,[')'|T]).
pp_csp_process(val_of(Agent,_Span),S,T) :- !,
  pp_csp_value(Agent,S,T).
pp_csp_process(builtin_call(X),S,T) :- !,pp_csp_process(X,S,T).
pp_csp_process(agent_call(_Span,Agent,Parameters),S,T) :- !,
  pp_csp_value(Agent,S,S1),
  pp_csp_args(Parameters,S1,T,'(',')').
pp_csp_process(agent_call_curry(Agent,Parameters),S,T) :- !,
  pp_csp_value(Agent,S,S1),
  pp_csp_curry_args(Parameters,S1,T).
pp_csp_process(lambda(Parameters,Body),['\\ '|S],T) :- !,
  pp_csp_args(Parameters,S,[' @ '|S1],'',''),
  pp_csp_value(Body,S1,T).
pp_csp_process('\\'(B,C,S),S1,T) :- !, pp_csp_process(ehide(B,C,S),S1,T).
pp_csp_process(ehide(Body,ChList,_Span),['('|S],T) :- !,
  pp_csp_process(Body,S,[')\\('|S1]),
  pp_csp_value(ChList,S1,[')'|T]).
pp_csp_process(Expr,S,T) :- binary_csp_op(Expr,X,Y,Op),!,
  pp_csp_process(X,S,[' ',Op,' '|S1]),
  pp_csp_process(Y,S1,T).
pp_csp_process(Expr,S,T) :- sharing_csp_op(Expr,X,Middle,Y,Op1,Op2),!,
  pp_csp_process(X,S,[Op1|S1]),
  pp_csp_value(Middle,S1,[Op2|S2]),
  pp_csp_process(Y,S2,T).
pp_csp_process(Expr,S,T) :- asharing_csp_op(Expr,X,MiddleX,MiddleY,Y,Op1,MOp,Op2),!,
  pp_csp_process(X,S,[Op1|S1]),
  pp_csp_value(MiddleX,S1,[MOp|S2]),
  pp_csp_value(MiddleY,S2,[Op2|S3]),
  pp_csp_process(Y,S3,T).
pp_csp_process(Expr,S,T) :- renaming_csp_op(Expr,X,RList,Op1,Op2),!,
  pp_csp_process(X,S,[Op1|S1]),
  pp_csp_value_l(RList,',',S1,[Op2|T],10).
pp_csp_process(prefix(_SPAN1,Values,ChannelExpr,CSP,_SPAN2),S,T) :- !,
  pp_csp_value_l([ChannelExpr|Values],'',S,['->'|S2],20),
  pp_csp_process(CSP,S2,T).
pp_csp_process('&'(Test,Then),S,T) :- !,
  pp_csp_bool_expr(Test,S,['&'|S2]),
  pp_csp_process(Then,S2,T).
pp_csp_process(ifte(Test,Then,Else,_SPAN1,_SPAN2,_SPAN3),[' if '|S],T) :- !,
  pp_csp_bool_expr(Test,S,[' then '|S2]),
  pp_csp_process(Then,S2,[' else '|S3]),
  pp_csp_process(Else,S3,T).
pp_csp_process(X,[X|T],T).

pp_csp_bool_expr(bool_not(BE),['not('|S],T) :- !, pp_csp_bool_expr(BE,S,[')'|T]).
pp_csp_bool_expr(BE,['('|S],T) :- binary_bool_op(BE,BE1,BE2,OP), !,
    pp_csp_bool_expr(BE1,S,[OP|S2]),
    pp_csp_bool_expr(BE2,S2,[')'|T]).
pp_csp_bool_expr(BE,[OP,'('|S],T) :- binary_pred(BE,BE1,BE2,OP), !,
    pp_csp_value(BE1,S,[','|S2]),
    pp_csp_value(BE2,S2,[')'|T]).
pp_csp_bool_expr(BE,S,T) :- pp_csp_value(BE,S,T).

bynary_numeric_operation('+'(X,Y),X,Y,'+').
bynary_numeric_operation('-'(X,Y),X,Y,'-').
bynary_numeric_operation('*'(X,Y),X,Y,'*').
bynary_numeric_operation('/'(X,Y),X,Y,'/').
bynary_numeric_operation('%'(X,Y),X,Y,'%').

binary_pred('member'(X,Y),X,Y,member).
binary_pred('<'(X,Y),X,Y,'<').
binary_pred('>'(X,Y),X,Y,'>').
binary_pred('>='(X,Y),X,Y,'>=').
binary_pred('<='(X,Y),X,Y,'=<').
binary_pred('elem'(X,Y),X,Y,is_elem_list).
binary_pred('=='(X,Y),X,Y,equal_element).
binary_pred('!='(X,Y),X,Y,not_equal_element).


binary_bool_op('<'(X,Y),X,Y,'<').
binary_bool_op('>'(X,Y),X,Y,'>').
binary_bool_op('>='(X,Y),X,Y,'>=').
binary_bool_op('<='(X,Y),X,Y,'=<').
binary_bool_op('=='(X,Y),X,Y,'==').
binary_bool_op('!='(X,Y),X,Y,'!=').
binary_bool_op(bool_and(X,Y),X,Y,'&&').
binary_bool_op(bool_or(X,Y),X,Y,'||').

binary_csp_op('|||'(X,Y,_Span),X,Y,'|||').
binary_csp_op('[]'(X,Y,_Span),X,Y,'[]').
binary_csp_op('|~|'(X,Y,_Span),X,Y,'|~|').
binary_csp_op(';'(X,Y,_Span),X,Y,';').
binary_csp_op('[>'(P,Q,_SrcSpan),P,Q,'[>').
binary_csp_op('/\\'(P,Q,_SrcSpan),P,Q,'/\\').

sharing_csp_op(esharing(CList,X,Y,_SrcSpan),X,CList,Y,' [|','|] ').
sharing_csp_op(sharing(CList,X,Y,_SrcSpan),X,CList,Y,' [|','|] ').
sharing_csp_op(lParallel(LinkList,X,Y,_Span),X,LinkList,Y,' [','] ').
sharing_csp_op(elinkParallel(LinkList,X,Y,_Span),X,LinkList,Y,' [','] ').
sharing_csp_op(exception(CList,X,Y,_SrcSpan),X,CList,Y,' [|','|> ').
sharing_csp_op(eexception(CList,X,Y,_SrcSpan),X,CList,Y,' [|','|> ').

asharing_csp_op(aParallel(CListX,X,CListY,Y,_SrcSpan),X,CListX,CListY,Y,' [',' || ','] ').
asharing_csp_op(eaParallel(CListX,X,CListY,Y,_SrcSpan),X,CListX,CListY,Y,' [',' || ','] ').

renaming_csp_op(procRenaming(RenameList,X,_SrcSpan),X,RenameList,'[[',']]').
renaming_csp_op(eprocRenaming(RenameList,X,_SrcSpan),X,RenameList,'[[',']]').

translate_events([],[]).
translate_events([E|Erest],[Out|Orest]) :-
    translate_event(E,Out),
    translate_events(Erest,Orest).
% translate_event(smv(Module,ChangedVars),Out) :-
%   !,smv_print_transition(Module,ChangedVars,Out).
translate_event(Transition,Out) :-
    active_plugin(Id),!,
    plugin_prettyprint_transition(Id,Transition,Out).
translate_event(Transition,Out) :-
    translate_event_without_plugin(Transition,Out).
%translate_event_without_plugin(smv_init(InitState),Out) :-
%    !,smv_print_initialisation(smv_init(InitState),Out).
translate_event_without_plugin(Event,Out) :-
    translate_event2(Event,Atoms,[]),!,
    ajoin(Atoms,Out).
translate_event_without_plugin(Event,Out) :- add_error(translate_event,'Could not translate event: ', Event),
    Out = '???'.
/* BEGIN CSP */
translate_event2(start_cspm(Process),['start_cspm('|S],T) :- process_algebra_mode,!,pp_csp_value(Process,S,[')'|T]). 
%% translate_event2(i(_Span),['i'|T],T) :- process_algebra_mode,!.   /* CSP */ %% deprecated
translate_event2(tick(_Span),['tick'|T],T) :- process_algebra_mode,!.   /* CSP */
translate_event2(tau(hide(Action)),['tau(hide('|S],T) :- process_algebra_mode,nonvar(Action), !,
     translate_event2(Action,S,['))'|T]).   /* CSP */
translate_event2(tau(link(Action1,Action2)),['tau(link('|S],T) :- /* CSP */
     nonvar(Action1), nonvar(Action2), process_algebra_mode, !,
     translate_event2(Action1,S,['<->'|S1]),
     translate_event2(Action2,S1,['))'|T]).
translate_event2(tau(Info),['tau(',Fun,')'|T],T) :-
     nonvar(Info), process_algebra_mode,!,   /* CSP */
     functor(Info,Fun,_). %(translate_event(Info,Fun) -> true ; functor(Info,Fun,_)),
translate_event2(io(Ch,V,_Span),S,T) :- process_algebra_mode,!, /* CSP */
    (specfile:csp_with_bz_mode ->
        S=['CSP:'|S1],
        translate_event2(io(Ch,V),S1,T)
    ; otherwise ->
        translate_event2(io(Ch,V),S,T)
    ).
translate_event2(io(X,Channel),S,T) :- process_algebra_mode,!,   /* CSP */
  (X=[] -> translate_event2(Channel,S,T)
   ; (translate_event2(Channel,S,S1),
      translate_channel_values(X,S1,T))
   ).
/* END CSP */
translate_event2('$JUMP'(Name),[A|T],T) :- write_to_codes(Name,Codes),
  atom_codes_with_limit(A,Codes).
translate_event2('-->'(Operation,ResultValues),S,T) :- nonvar(ResultValues),
  ResultValues=[_|_],!,
  translate_event2(Operation,S,['-->',ValuesStr|T]),
  translate_bvalues(ResultValues,ValuesStr).
translate_event2(Op,[TOpName|S],T) :-
  nonvar(Op), Op =.. [OpName|Args],
  translate_operation(OpName,TOpName),
  ( Args=[] -> S=T
  ; otherwise ->
      S=['(',ValuesStr,')'|T],
      translate_bvalues(Args,ValuesStr)),!.
translate_event2(Op,[A|T],T) :-  
   %['<< ',A,' >>'|T],T) :-   % the << >> pose problems when checking against FDR
  write_to_codes(Op,Codes),
  atom_codes_with_limit(A,Codes).

translate_operation('$initialise_machine',T) :- !,T='INITIALISATION'.
translate_operation('$setup_constants',T) :- !,T='SETUP_CONSTANTS'.
translate_operation('$partial_setup_constants',T) :- !,T='PARTIAL_SETUP_CONSTANTS'.
translate_operation(X,X).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% pretty-print properties
translate_properties_with_limit([],[]).
translate_properties_with_limit([P|Prest],[Out|Orest]) :-
    translate_property_with_limit(P,320,Out), % reduced limit as we now have evaluation view + possibility to inspect all of value
    translate_properties_with_limit(Prest,Orest).

translate_property_with_limit(Prop,Limit,Output) :-
    (pp_property(Prop,Limit,Output) -> true ; (add_error(translate_property,'Could not translate property: ',Prop),Output='???')).
pp_property(Prop,_Limit,Output) :-
    active_plugin(Id),!,
    plugin_prettyprint_property(Id,Prop,Output).
pp_property(Prop,Limit,Output) :-
    pp_property_without_plugin(Prop,Limit,Output).
pp_property_without_plugin(=(Key,Value),_,A) :-
    !,ajoin([Key,' = ',Value],A).
pp_property_without_plugin(info(I),_,I) :- !.
pp_property_without_plugin(Prop,Limit,A) :-
    write_to_codes(Prop,Codes),
    atom_codes_with_limit(A,Limit,Codes).

     

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pretty-print errors belonging to a certain state
translate_error_term(Var,S) :- var(Var),!,
   translate_term_into_atom_with_max_depth(Var,S).
translate_error_term('@'(X,F),S) :-
   translate:translate_bvalue(X,TX), translate:translate_bvalue(F,TF),!,
   % translate_term_into_atom_with_max_depth('@'(TX,TF),S).
   ajoin(['Function argument: ',TX, ', function value: ',TF],S).
translate_error_term('@rel'(Arg,Res1,Res2),S) :-
   translate:translate_bvalue(Arg,TA), translate:translate_bvalue(Res1,R1),
   translate:translate_bvalue(Res2,R2),!,
   ajoin(['Function argument: ',TA, ', two possible values: ',R1,', ',R2],S).
translate_error_term([operation(Name,Env)],S) :-
    translate:translate_any_state(Env,TEnv), !,
    translate_term_into_atom_with_max_depth(operation(Name,TEnv),S).
translate_error_term(error(E1,E2),S) :- !, translate_prolog_error(E1,E2,S).
translate_error_term(T,S) :- translate_term_into_atom_with_max_depth(T,S).

translate_prolog_error(existence_error(procedure,Pred),_,S) :- !,
    translate_term_into_atom_with_max_depth('Unknown Prolog predicate:'(Pred),S).
translate_prolog_error(existence_error(source_sink,File),_,S) :- !,
    translate_term_into_atom_with_max_depth('Could not access file:'(File),S).
translate_prolog_error(E1,_,S) :- translate_term_into_atom_with_max_depth(E1,S).

translate_state_errors([],[]).
translate_state_errors([E|ERest],[Out|ORest]) :-
    ( E = eventerror(Event,EError,_) ->
        translate_event_error(EError,Msg),
        ajoin([Event,': ',Msg],Out)
    ; translate_state_error(E,Out) -> true
    ; otherwise -> functor(E,Out,_) ),
    translate_state_errors(ERest,ORest).

translate_error_context(E,TE) :- translate_error_context2(E,Codes,[]),
    atom_codes_with_limit(TE,Codes).
translate_error_context2(span_context(Span,Context)) --> !,
   translate_error_context2(Context),
   translate_span(Span).
translate_error_context2([H]) --> !,translate_error_context2(H).
translate_error_context2(checking_invariant) --> !,
    "INVARIANT".
translate_error_context2(checking_assertions) --> !,
    "ASSERTIONS".
translate_error_context2(checking_negation_of_invariant(_State)) --> !,
    "not(INVARIANT)".
translate_error_context2(operation(OpName,_State)) --> !,
    {translate_operation(OpName,TOp)},
    ppterm(TOp).
translate_error_context2(operation(OpName,_State)) --> !,
    {translate_operation(OpName,TOp)},
    ppterm(TOp).
translate_error_context2(X) --> "???:", ppterm(X).

print_span(Span) :- translate_span(Span,Codes,[]),
    atom_codes_with_limit(TE,Codes), !, print(TE).
print_span(S) :- print(span(S)).
    
translate_span(Span) --> {extract_line_col(Span,Srow,Scol,_Erow,_Ecol)},!,
   " (",ppterm(Srow),":",ppterm(Scol),
     %"-",ppterm(Erow),":",ppterm(Ecol),
    ")".
translate_span(_) --> "".

explain_span(span_predicate(Pred,LS,S)) --> !, explain_span2(pan_predicate(Pred,LS,S)),
    explain_local_state(LS).
explain_span(Span) --> explain_span2(Span).
explain_span2(Span) --> {extract_line_col(Span,Srow,Scol,Erow,Ecol)},!,
   "\n### Line: ",ppterm(Srow),", Column: ",ppterm(Scol)," -- Line: ",ppterm(Erow),", Column: ",ppterm(Ecol).
explain_span2(_) --> "".

explain_local_state([]) --> !, "".
explain_local_state(LS) --> "\n  Local State: ", pp_b_state(LS).

translate_event_error(Error,Out) :-
    ( translate_event_error2(Error,Out) -> true
    ; otherwise ->
        functor(Error,F,_),
        ajoin(['** Unable to translate event error: ',F,' **'],Out)).
translate_event_error2(no_witness_found(Type,Var,_Predicate),Out) :-
    def_get_texpr_id(Var,Id),
    ajoin(['no witness was found for ',Type,' ',Id],Out).
translate_event_error2(simulation_error(_Events),Out) :-
    Out = 'no matching abstract event was found'.
translate_event_error2(action_not_executable(_Action),Out) :-
    Out = 'action was not executable'.
translate_event_error2(invalid_modification(Var,_Pre,_Post),Out) :-
    def_get_texpr_id(Var,Id),
    ajoin(['modification of variable ', Id, ' not allowed'],Out).
translate_event_error2(variant_negative(_CType,_Variant,_Value),Out) :-
    Out = 'enabled for negative variant'.
translate_event_error2(invalid_variant(anticipated,_Expr,_Pre,_Post),Out) :-
    Out = 'variant increased'.
translate_event_error2(invalid_variant(convergent,_Expr,_Pre,_Post),Out) :-
    Out = 'variant not decreased'.
translate_event_error2(invalid_theorem_in_guard(_Theorem),Out) :-
    Out = 'theorem in guard evaluates to false'.

translate_state_error(abort_error(_TYPE,Msg,ErrTerm,ErrorContext),Out) :- !,
    translate_error_term(ErrTerm,ES),
    translate_error_context(ErrorContext,EC),
    ajoin([EC,': ',Msg,': ',ES],Out).
translate_state_error('CLPFD_integer_overflow',Out) :- !,
    Out = 'CLPFD integer overflow'.
translate_state_error(max_state_errors_reached(Nr),Out) :- !,
     ajoin(['Max. number of state errors reached: ', Nr],Out).
translate_state_error(Unknown,Out) :-
    add_error(translate_state_error,'Unknown state error: ',Unknown),
    Out = '*** Unknown State Error ***'.


explain_error_context1([H],Span) --> !,explain_error_context1(H,Span).
explain_error_context1(span_context(Span,Context),Span) --> !,
   explain_span(Span),"\n",
   explain_error_context2(Context).
explain_error_context1(Ctxt,unknown) --> explain_error_context2(Ctxt).

explain_error_context2([H]) --> !,explain_error_context2(H).
explain_error_context2(span_context(Span,Context)) --> !,
   explain_span(Span),"\n",
   explain_error_context2(Context).
explain_error_context2(checking_invariant) --> !,
    "INVARIANT:\n ",
    pp_current_state. % assumes explain is called in the right state ! ; otherwise we need to store the state id
explain_error_context2(checking_assertions) --> !,
    "ASSERTIONS:\n ",
    pp_current_state. % assumes explain is called in the right state ! ; otherwise we need to store the state id
explain_error_context2(checking_negation_of_invariant(State)) --> !,
    "not(INVARIANT):\n  State: ",
    pp_b_state(State).
explain_error_context2(operation('$setup_constants',StateID)) --> !,
    "PROPERTIES: ", "\n  State: ",
    pp_context_state(StateID).
explain_error_context2(operation(OpName,StateID)) --> !,
    "EVENT/OPERATION: ",
    {translate_operation(OpName,TOp)},
    ppterm(TOp), "\n ",
    pp_context_state(StateID).
explain_error_context2(X) --> "UNKNOWN ERROR CONTEXT:\n  ", ppterm(X).

explain_state_error(Error,Span,Out) :-
    explain_state_error2(Error,Span,Out,[]),!.
explain_state_error(_Error,unknown,"Sorry, the detailed output failed.\n").

explain_abort_error_type(well_definedness_error) --> !, "An expression was not well-defined.\n".
explain_abort_error_type(while_variant_error) --> !, "A while-loop VARIANT error occured.\n".
explain_abort_error_type(while_invariant_violation) --> !, "A while-loop INVARIANT error occured.\n".
explain_abort_error_type(precondition_error) --> !, "A precondition (PRE) error occured.\n".
explain_abort_error_type(assert_error) --> !, "An ASSERT error occured.\n".
explain_abort_error_type(Type) --> "Error occurred: ", ppterm(Type), "\n".

explain_state_error2(abort_error(TYPE,Msg,ErrTerm,ErrContext),Span) -->
    explain_abort_error_type(TYPE),
    "Reason:  ", ppterm(Msg), "\n",
    "Details: ", {translate_error_term(ErrTerm,ErrS)},ppterm(ErrS), "\n",
    "Context: ", explain_error_context1(ErrContext,Span).
explain_state_error2(max_state_errors_reached(Nr),unknown) -->
    "Too many error occured for this state.\n",
    "Not all errors are shown.\n",
    "Number of errors is at least: ",   ppterm(Nr).
explain_state_error2(eventerror(_Event,Error,Trace),unknown) -->  % TO DO: also extract loc info ?
    {translate_event_error(Error,Msg)},
    ppatom(Msg),
    "\nA detailed trace containing the error:\n",
    "--------------------------------------\n",
    explain_event_trace(Trace).

show_parameter_values([],[]) --> !.
show_parameter_values([P|Prest],[V|Vrest]) -->
    show_parameter_value(P,V),
    show_parameter_values(Prest,Vrest).
show_parameter_value(P,V) -->
    "    ",pp_expr(P,_)," = ",pp_value(V),"\n".

explain_event_trace(Trace,Codes) :-
    explain_event_trace(Trace,Codes,[]).

explain_event_trace([]) --> !.
explain_event_trace([Step|Trest]) -->
    "\n",
    ( explain_event_step(Step) -> ""
    ; {functor(Step,F,_)} ->
        "  (no rule to explain event step ",ppatom(F),")\n"),
    explain_event_trace(Trest).
explain_event_step(event(Name,Section)) --> !,
    "Event ",ppterm(Name)," in model ",ppterm(Section),
    ":\n".
explain_event_step(true_guard(Parameters,Values,Guard)) --> !,
    ( {Parameters==[]} -> ""
    ; "  for the parameters:\n",
        show_parameter_values(Parameters,Values)),
    "  the guard is true:",
    explain_predicate(Guard,4),"\n".
explain_event_step(variant_checked_pre(CType,Variant,Value)) -->
    "  ",ppatom(CType)," event: checking if the variant if non-negative:\n",
    "    variant: ",pp_expr(Variant,_),"\n",
    "    its value: ",pp_value(Value),"\n".
explain_event_step(variant_negative(CType,Variant,Value)) -->
    explain_event_step(variant_checked_pre(CType,Variant,Value)),
    "  ERROR: variant is negative\n".
explain_event_step(variant_checked_post(CType,Variant,EntryValue,ExitValue)) -->
    "  ",ppatom(CType)," event: checking if the variant is ",
    ( {CType==convergent} -> "decreased:\n" ; "not increased:\n"),
    "    variant: ", pp_expr(Variant,_), "\n",
    "    its value before: ", pp_value(EntryValue),"\n",
    "    its value after:  ", pp_value(ExitValue),"\n".
explain_event_step(invalid_variant(CType,Variant,EntryValue,ExitValue)) -->
    explain_event_step(variant_checked_post(CType,Variant,EntryValue,ExitValue)),
    "  ERROR: variant has ",
    ({CType==convergent} -> "not been decreased\n"; "has been increased\n").
explain_event_step(eval_witness(Type,Id,Value,Predicate)) -->
    witness_intro(Id,Predicate,Type),
    "    found witness:\n",
    "    ", pp_expr(Id,_), " = ", pp_value(Value), "\n".
explain_event_step(no_witness_found(Type,Id,Predicate)) -->
    witness_intro(Id,Predicate,Type),
    "  ERROR: no solution for witness predicate found!\n".
explain_event_step(action(Lhs,_Rhs,Values)) -->
    "  executing an action:\n",
    show_assignments(Lhs,Values).
explain_event_step(action_set(Lhs,_Rhs,ValueSet,Values)) -->
    "  executing an action:\n    ",
    pp_expr_l(Lhs)," :: ",pp_value(ValueSet),"\n    choosing\n",
    show_assignments(Lhs,Values).
explain_event_step(action_pred(Ids,Pred,Values)) -->
    "  executing an action:\n    ",
    pp_expr_l(Ids)," :| ",pp_expr(Pred,_),"\n    choosing\n",
    show_assignments(Ids,Values).
explain_event_step(action_not_executable(TAction)) -->
    "  ERROR: the following action was not executable:",
    translate_subst_with_indention(TAction,4).
explain_event_step(error(Error,_Id)) -->
    % the error marker serves to link to a stored state-error by its ID
    explain_event_step(Error).
explain_event_step(simulation_error(Errors)) -->
    "  no guard of a refined event was satisfiable:\n",
    explain_simulation_errors(Errors).
explain_event_step(invalid_theorem_in_guard(Theorem)) -->
    "  the following theorem evaluates to false:",
    explain_predicate(Theorem,4),"\n".
explain_event_step(invalid_modification(Var,Pre,Post)) -->
    "  the variable ", pp_expr(Var,_), " has been modified.\n",
    "    The event is not allowed to modify the variable because its abstract event does not modify it.\n",
    "    Old value: ", pp_value(Pre), "\n",
    "    New value: ", pp_value(Post), "\n".



witness_intro(Id,Predicate,Type) -->
    "  evaluating witness for abstract ", ppatom(Type), " ", pp_expr(Id,_), "\n",
    "    witness predicate: ", pp_expr(Predicate,_), "\n".

show_assignments([],[]) --> !.
show_assignments([Lhs|Lrest],[Val|Vrest]) -->
    "    ",pp_expr(Lhs,_), " := ", pp_value(Val), "\n",
    show_assignments(Lrest,Vrest).

explain_state([]) --> !.
explain_state([bind(Varname,Value)|Rest]) --> !,
    " ",ppterm(Varname)," = ",pp_value(Value),"\n",
    explain_state(Rest).

explain_guards([]) --> "".
explain_guards([Event|Rest]) -->
    {get_texpr_expr(Event,rlevent(Name,_Section,_Status,_Params,Guard,_Theorems,_Act,_VWit,_PWit,_Unmod,_Evt))},
    "\n",ppatom(Name),":",
    explain_predicate(Guard),
    explain_guards(Rest).
explain_predicate(Guard,Indention,I,O) :-
    pred_over_lines(0,Guard,(Indention,I),(_,O)).
explain_predicate(Guard,I,O) :-
    explain_predicate(Guard,2,I,O).

explain_simulation_errors([]) --> !.
explain_simulation_errors([Error|Rest]) -->
    explain_simulation_error(Error),
    explain_simulation_errors(Rest).
explain_simulation_error(event(Name,Section,Guard)) -->
    "    guard for event ", ppatom(Name),
    " in ", ppatom(Section), ":",
    explain_predicate(Guard,6),"\n".

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pretty-print a state

print_state(State) :- b_state(State), !,print_bstate(State).
print_state(csp_and_b_root) :- csp_with_bz_mode, !,
   print('(MAIN || B)').
print_state(csp_and_b(CSPState,BState)) :- csp_with_bz_mode, !,
   print_bstate(BState), translate_cspm_state(CSPState,Text), print(Text).
print_state(CSPState) :- csp_mode,!,translate_cspm_state(CSPState,Text), print(Text).
print_state(State) :- print('*** Unknown state: '),print(State).

b_state(root).
b_state(concrete_constants(_)).
b_state(const_and_vars(_,_)).
b_state([bind(_,_)|_]).
b_state([]).

print_bstate(State) :- translate_bstate(State,Output), print('     '),print(Output).

translate_any_state(State,Output) :-
    pp_any_state(State,Codes,[]),
    atom_codes_with_limit(Output,Codes).
translate_bstate(State,Output) :-
    pp_b_state(State,Codes,[]),
    atom_codes_with_limit(Output,Codes).
pp_b_state(X) --> try_pp_b_state(X),!.
pp_b_state(X) --> {add_error(pp_b_state,'Could not translate state: ',X)}.

try_pp_b_state(VAR) --> {var(VAR)},!, "_?VAR?_", {add_error(pp_b_state,'Variable state: ',VAR)}.
try_pp_b_state(root) --> !, "root".
try_pp_b_state(concrete_constants(Constants)) --> !,"Constants: ",
   pp_b_state(Constants).
try_pp_b_state(const_and_vars(ID,Vars)) --> !,
   "Constants:",ppterm(ID),", Vars:",
   {set_translation_constants(ID)}, /* extract constants which stand for deferred set elements */
   pp_b_state(Vars),
   {clear_translation_constants}.
try_pp_b_state([]) --> !, "/* empty state */".
try_pp_b_state([bind(Varname,Value)|Rest]) --> !,
    "( ",ppterm(Varname),"=",
    pp_value(Value),
    ({Rest = []} -> []; " &\n       "),
    pp_b_state_list(Rest).


pp_b_state_list([]) --> !, " )".
pp_b_state_list([bind(Varname,Value)|Rest]) --> !,
    ppterm(Varname),"=",
    pp_value(Value),
    ({Rest = []} -> []; " &\n       "),
    pp_b_state_list(Rest).
pp_b_state_list(X) --> {add_error(pp_b_state_list,'Could not translate: ',X)}.

% ----------------

% printing and translating error contexts
print_context(State) :- translate_context(State,Output), print(Output).

translate_context(Context,Output) :-
    pp_b_context(Context,Codes,[]),
    atom_codes_with_limit(Output,Codes).

pp_b_context([]) --> !.
pp_b_context([C|Rest]) --> !,
    pp_b_context(C),
    pp_b_context(Rest).
pp_b_context(translate_context) --> !, " ERROR CONTEXT: translate_context". % error occured within translate_context
pp_b_context(operation(Name,StateID)) --> !,
    " ERROR CONTEXT: OPERATION:",ppterm(Name), ",",pp_context_state(StateID).
pp_b_context(checking_invariant) --> !,
    " ERROR CONTEXT: INVARIANT CHECKING,", pp_cur_context_state.
pp_b_context(checking_negation_of_invariant(State)) --> !,
    " ERROR CONTEXT: NEGATION_OF_INVARIANT CHECKING, State:", pp_b_state(State).
pp_b_context(checking_assertions) --> !,
    " ERROR CONTEXT: ASSERTION CHECKING,", pp_cur_context_state.
pp_b_context(C) --> ppterm(C),pp_cur_context_state.

pp_cur_context_state --> {state_space:get_current_context_state(ID)}, !,pp_context_state(ID).
pp_cur_context_state --> ", unknown context state.".

% assumes we are in the right state:
pp_current_state --> {state_space:current_expression(ID,_)}, !,pp_context_state(ID).
pp_current_state --> ", unknown context state.".

pp_context_state(ID) --> {state_space:visited_expression(ID,State)},!, % we have a state ID
                          " State ID:", ppterm(ID), ", State: ",pp_any_state(State).
pp_context_state(State) --> " State: ",pp_any_state(State).


pp_any_state(X) --> try_pp_b_state(X),!.
pp_any_state(P) -->  "CSP: ",{pp_csp_process(P,Atoms,[])},!,atoms_to_codelist(Atoms).
pp_any_state(X) -->   "Other formalism: ",ppterm(X). % CSP state

atoms_to_codelist([]) --> [].
atoms_to_codelist([Atom|T]) --> ppterm(Atom), atoms_to_codelist(T).

% ----------------

:- dynamic deferred_set_constant/3.

set_translation_context(const_and_vars(ConstID,_)) :- !, 
 %% print_message(setting_translation_constants(ConstID)),
  set_translation_constants(ConstID).
set_translation_context(_).

set_translation_constants(_) :- clear_translation_constants,
    preferences:preference(dot_print_use_constants,false),!.
set_translation_constants(ConstID) :- var(ConstID),!,
   add_error(set_translation_constants,'Variable used as ConstID: ',ConstID).
set_translation_constants(ConstID) :- 
    state_space:visited_expression(ConstID,concrete_constants(ConstantsStore)),!,
   %% print_message(setting_constants(ConstID)),%%
    (treat_constants(ConstantsStore) -> true ; print_message(fail)).
set_translation_constants(ConstID) :- 
   add_error(set_translation_constants,'Unknown ConstID: ',ConstID).
    
clear_translation_constants :- %print_message(clearing),%%
    retractall(deferred_set_constant(_,_,_)).

treat_constants([]).
treat_constants([bind(CstName,Val)|T]) :-
   ((Val=fd(X,GSet),b_global_deferred_set(GSet))
     -> (deferred_set_constant(GSet,X,_)
          -> true /* duplicate def of value */
           ; assert(deferred_set_constant(GSet,X,CstName))
        )
      ; true
   ),
   treat_constants(T).
   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pretty-print a value
translate_bvalue(V,Output) :- 
    ( pp_value(V,Codes,[]) ->
        atom_codes_with_limit(Output,Codes)
    ; add_error(translate_bvalue,'Could not translate bvalue: ',V),
        Output='???').
:- use_module(preferences).
translate_bvalue_with_limit(V,Limit,Output) :- 
    preferences:get_preference(expand_avl_upto,Max),
    (Max>Limit ->
        preferences:set_preference(expand_avl_upto,Limit) % no sense in printing larger AVL trees,
        ; true),
    call_cleanup(translate_bvalue_with_limit_aux(V,Limit,Output),
                 preferences:set_preference(expand_avl_upto,Max)).
translate_bvalue_with_limit_aux(V,Limit,Output) :-
    ( pp_value(V,Codes,[]) ->
        atom_codes_with_limit(Output,Limit,Codes)
    ; add_error(translate_bvalue_with_limit,'Could not translate bvalue: ',V),
        Output='???').

translate_bvalues(Values,Output) :-
    pp_value_l(Values,',',Codes,[]),!,
    atom_codes_with_limit(Output,Codes).
translate_bvalues(Values,_) :- add_error_fail(translate_bvalues,'Could not translate B values: ',Values).


% a more refined pretty printing: takes Type information into account; useful for detecting sequences
translate_bvalue_with_type(Value,Type,Output) :-
    adapt_value_according_to_type(Type,Value,NewValue),
    translate_bvalue(NewValue,Output).
adapt_value_according_to_type(integer,V,R) :- !,R=V.
adapt_value_according_to_type(string,V,R) :- !,R=V.
adapt_value_according_to_type(boolean,V,R) :- !,R=V.
adapt_value_according_to_type(global(_),V,R) :- !,R=V.
adapt_value_according_to_type(couple(TA,TB),(VA,VB),(RA,RB)) :-
   adapt_value_according_to_type(TA,VA,RA),
   adapt_value_according_to_type(TB,VB,RB).
adapt_value_according_to_type(set(_Type),V,R) :- !,R=V. % to do: convert inside set if small enough ?
adapt_value_according_to_type(seq(Type),V,R) :- !,
    (convert_set_into_sequence(V,VS)
        -> l_adapt_value_according_to_type(VS,Type,AVS),
           R=sequence(AVS)
        ;  R=V).
adapt_value_according_to_type(record(_Fields),Value,R) :- !, % to do: convert inside record
    R=Value.
adapt_value_according_to_type(Type,Value,R) :- print(unknown(Type,Value)),nl,
     R=Value.

l_adapt_value_according_to_type([],_Type,R) :- !,R=[].
l_adapt_value_according_to_type([H|T],Type,[AH|AT]) :-
    adapt_value_according_to_type(Type,H,AH),
    l_adapt_value_according_to_type(T,Type,AT).

pp_value(V,I,[]) :- active_plugin(Id), !, plugin_prettyprint_value(Id,V,I).
pp_value(V) --> {var(V)},!,"_".
pp_value(X) --> {cyclic_term(X)},!,"_cyclic_".
pp_value(fd(X)) --> !,ppatom(X).  %% ???
pp_value(fd(X,GSet)) --> {var(X)},!,
   ppatom(GSet),":??".
pp_value(fd(X,GSet)) -->
    {b_global_sets:is_b_global_constant(GSet,X,Res)},!,
    ppatom(Res).
pp_value(fd(X,GSet)) --> {deferred_set_constant(GSet,X,Cst)},!, 
         ppatom(Cst).
pp_value(fd(X,M)) --> !,ppatom(M),ppnumber(X).
pp_value(int(X)) --> !,ppnumber(X).
pp_value(string(X)) --> !,"""",ppatom(X),"""".
pp_value(global_set(X)) --> {atomic(X),integer_set_mapping(X,Y)},!,ppatom(Y).
pp_value(term(X)) --> {var(X)},!,"term(_)".
pp_value(freetype(X)) --> {pretty_freetype(X,P)},!,ppatom(P).
pp_value(pred_true /* bool_true */) --> !,"TRUE".
pp_value(pred_false /* bool_false */) --> !,"FALSE".
%pp_value(bool_true) --> !,"TRUE". % old version; still in some test traces which are printed
%pp_value(bool_false) --> !,"FALSE".
pp_value([]) --> !,"{}".
pp_value(freeval(_,Case,Value)) -->
      {ground(Case),ground(Value),Value=term(Case)},!,ppatom(Case).
pp_value(sequence(List)) --> !,
    ({List=[]} -> "[]" ; ppatom('['),pp_value_l(List,','),ppatom(']')).
pp_value([Head|Tail]) --> !,
    ppatom('{'),pp_value_l([Head|Tail],','),ppatom('}').
%pp_value([Head|Tail]) --> !,
%    {( convert_set_into_sequence([Head|Tail],Elements) ->
%       (Start,End) = ('[',']')
%     ; otherwise ->
%       Elements = [Head|Tail],
%       (Start,End) = ('{','}'))},
%    ppatom(Start),pp_value_l(Elements,','),ppatom(End).
pp_value( (A,B) ) --> !,
    "(",pp_value(A),"|->",pp_value(B),")".
pp_value(avl_set(A)) --> !,
    {avl_size(A,Sz)},
    ( {size_is_in_set_limit(Sz),Sz>2,avl_max(A,(int(Sz),_)),convert_avlset_into_sequence(A,Seq)} ->
        "[",pp_value_l(Seq,','),"]"
    ; {otherwise} ->
        ( {size_is_in_set_limit(Sz)} ->
            {avl_domain(A,List)},
            "{",pp_value_l(List,','),"}"
        ; {otherwise} ->
            {avl_min(A,Min),avl_next(Min,A,Nxt),avl_max(A,Max),avl_prev(Max,A,Prev)},
            "#",ppnumber(Sz),":{",
            pp_value(Min),",",pp_value(Nxt),",...,",
            pp_value(Prev),",",pp_value(Max),"}" )).
pp_value(field(Name,Value)) --> !,
    ppatom(Name),":",pp_value(Value). % : for fields has priority 120 in French manual
pp_value(rec(Rec)) --> !,
    "rec(",pp_value_l(Rec,','),")".
pp_value(struct(Rec)) --> !,
    "struct(",pp_value_l(Rec,','),")".
pp_value(term(no_value_for(Id))) --> !,
    "undefined ",ppatom(Id).
pp_value(closure(Ids,Type,B)) --> 
   {var(Ids) ; var(Type) ; var(B)},!,
   add_internal_error('Illegal value: ',pp_value_illegal_closure(Ids,Type,B)),
   "<< ILLEGAL ",write_to_codes(closure(Ids,Type,B))," >>".
pp_value(closure([Id],[Type],Membership)) --> 
    { get_texpr_expr(Membership,member(Elem,Set)),
      get_texpr_id(Elem,Id),get_texpr_type(Elem,Type),!},
    pp_expr_m(Set,299).
pp_value(closure(Variables,Types,Predicate)) --> !,
    {make_closure_ids(Variables,Types,Ids)},
    pp_expr2(comprehension_set(Ids,Predicate),_).
pp_value(freeval(_,Case,Value)) --> {ground(Case)},!,
    ppatom(Case),"(",pp_value(Value),")".
pp_value(X) -->  % the << >> pose problems when checking against FDR
    "<< ",write_to_codes(X)," >>".

size_is_in_set_limit(Size) :- preferences:preference(expand_avl_upto,Max),
    (Max<0 -> true /* no limit */
     ; Size =< Max).
        
      
pp_value_l(V,_Sep) --> {var(V)},!,"...".
pp_value_l([],_Sep) --> !.
pp_value_l([Expr|Rest],Sep) -->
    ( {nonvar(Rest),Rest=[]} ->
        pp_value(Expr)
    ; {otherwise} ->
        pp_value(Expr),ppatom(Sep),pp_value_l(Rest,Sep)).
pp_value_l(avl_set(A),_Sep) --> "}\\/", pp_value(avl_set(A)).
pp_value_l(closure(A,B,C),_Sep) --> "}\\/", pp_value(closure(A,B,C)).

make_closure_ids([],[],[]).
make_closure_ids([V|Vrest],[T|Trest],[TExpr|TErest]) :-
    create_texpr(identifier(V),T,[],TExpr),
    make_closure_ids(Vrest,Trest,TErest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% machines
:- volatile  print_typing_infos/0, force_all_typing_infos/0, suppress_rodin_positions_flag/0.
:- dynamic print_typing_infos/0.
:- dynamic force_all_typing_infos/0.
:- dynamic suppress_rodin_positions_flag/0.

suppress_rodin_positions :-
    assert_if_not_set( suppress_rodin_positions_flag ).

set_print_type_infos(none) :- !,
    retractall( force_all_typing_infos ),
    retractall( print_typing_infos ).
set_print_type_infos(needed) :- !,
    retractall( force_all_typing_infos ),
    assert_if_not_set( print_typing_infos ).
set_print_type_infos(all) :- !,
    assert_if_not_set( print_typing_infos ),
    assert_if_not_set( force_all_typing_infos ).
set_print_type_infos(Err) :-
    add_internal_error('Illegal type: ',set_print_type_infos(Err)).

assert_if_not_set(Fact) :-
    ( call(Fact) -> true
    ; otherwise  -> assert(Fact)).
    

print_machine(M) :-
    nl, translate_machine(M,Msg,true), put_codes(Msg), nl,
    flush_output(user_output).


translate_machine(M,Codes,AdditionalInfo) :-
    retractall(print_additional_machine_info),
    (AdditionalInfo=true -> assert(print_additional_machine_info) ; true),
    translate_machine1(M,(0,Codes),(_,[])).

translate_machine1(machine(Name,Sections)) -->
    indent('MACHINE '), {adapt_machine_name(Name,AName)}, insertstr(AName),
    {init_machine_translation},
    translate_machine2(Sections,Sections).
translate_machine2([],_) --> !, insertstr('\nEND\n').
translate_machine2([P|Rest],All) -->
    translate_mpart(P,All),
    translate_machine2(Rest,All).

adapt_machine_name('dummy(uses)',R) :- !,R='MAIN'.
adapt_machine_name(X,X).

:- dynamic section_header_generated/1.
:- dynamic print_additional_machine_info/0.
print_additional_machine_info.

init_machine_translation :- retractall(section_header_generated(_)).

% start a part of a section
mpstart(Title,I) --> 
    insertstr('\n'),insertstr(Title),
    indention_level(I,I2), {I2 is I+2}.
% end a part of a section
mpend(I) -->
    indention_level(_,I).

mpstart_section(Section,Title,AltTitle,I,In,Out) :-
   (\+ section_header_generated(Section)
      -> mpstart(Title,I,In,Out), assert(section_header_generated(Section))
      ;  mpstart(AltTitle,I,In,Out) /* use alternative title; section header already generated */
   ).

translate_mpart(Section/I,All) --> %{print(Section),nl},
         ( {I=[]}  -> {true}
         ; translate_mpart2(Section,I,All) -> {true}
         ; {otherwise} ->
             insertstr('\nSection '),insertstr(Section),insertstr(': '),
             insertstr('<< pretty-print failed >>')
         ).
translate_mpart2(deferred_sets,I,_) -->
      mpstart_section(sets,'SETS /* deferred */',' ; /* deferred */',P),
      indent_expr_l_sep(I,';'),mpend(P).
translate_mpart2(enumerated_sets,_I,_) --> [].  % these are now pretty printed below
     %mpstart('ENUMERATED SETS',P),indent_expr_l_sep(I,';'),mpend(P).
translate_mpart2(enumerated_elements,I,_) -->  %{print(enum_els(I)),nl},
      {translate_enums(I,[],Res)},
      mpstart_section(sets,'SETS /* enumerated */',' ; /* enumerated */',P),
      indent_expr_l_sep(Res,';'),mpend(P).
translate_mpart2(parameters,I,_) --> mpstart('PARAMETERS',P),indent_expr_l_sep(I,','),mpend(P).
translate_mpart2(internal_parameters,I,_) --> {print_additional_machine_info},!,mpstart('/* INTERNAL_PARAMETERS',P),indent_expr_l_sep(I,','),insertstr(' */'),mpend(P).
translate_mpart2(internal_parameters,_I,_) --> [].
translate_mpart2(abstract_variables,I,_) --> mpstart('ABSTRACT_VARIABLES',P),indent_expr_l_sep(I,','),mpend(P).
translate_mpart2(concrete_variables,I,_) --> mpstart('CONCRETE_VARIABLES',P),indent_expr_l_sep(I,','),mpend(P).
translate_mpart2(abstract_constants,I,_) --> mpstart('ABSTRACT_CONSTANTS',P),indent_expr_l_sep(I,','),mpend(P).
translate_mpart2(concrete_constants,I,_) --> mpstart('CONCRETE_CONSTANTS',P),indent_expr_l_sep(I,','),mpend(P).
translate_mpart2(promoted,I,_) --> {print_additional_machine_info},!,
       mpstart('/* PROMOTED OPERATIONS',P),indent_expr_l_sep(I,','),insertstr(' */'),mpend(P).
translate_mpart2(promoted,_I,_) --> [].
translate_mpart2(unpromoted,I,_) --> {print_additional_machine_info},!,
       mpstart('/* NOT PROMOTED OPERATIONS',P),indent_expr_l_sep(I,','),insertstr(' */'),mpend(P).
translate_mpart2(unpromoted,_I,_) --> [].
translate_mpart2(constraints,I,All) --> mpart_typing('CONSTRAINTS',[parameters],All,I).
translate_mpart2(invariant,I,All) -->  mpart_typing('INVARIANT', [abstract_variables,concrete_variables],All,I).
translate_mpart2(linking_invariant,_I,_) --> [].
translate_mpart2(properties,I,All) --> mpart_typing('PROPERTIES',[abstract_constants,concrete_constants],All,I).
translate_mpart2(assertions,I,_) --> mpstart('ASSERTIONS',P),indent_expr_l_sep(I,';'),mpend(P).
translate_mpart2(initialisation,S,_) --> mpstart('INITIALISATION',P),translate_inits(S),mpend(P).
translate_mpart2(definitions,Defs,_) --> {print_additional_machine_info},!,mpstart('/* DEFINITIONS',P),translate_defs(Defs),insertstr(' */'),mpend(P).
translate_mpart2(definitions,_Defs,_) --> [].
translate_mpart2(operation_bodies,Ops,_) --> mpstart('OPERATIONS',P),translate_ops(Ops),mpend(P).
translate_mpart2(used,Used,_) --> {print_additional_machine_info},!,mpstart('/* USED',P),translate_used(Used),insertstr(' */'),mpend(P).
translate_mpart2(used,_Used,_) --> [].
translate_mpart2(freetypes,Freetypes,_) --> mpstart('FREETYPES',P),translate_freetypes(Freetypes),mpend(P).
translate_mpart2(meta,_Infos,_) --> [].

% Add typing predicates to a predicate
mpart_typing(Title,Section,Sections,PredI) -->
    {mpart_typing2(Section,Sections,PredI,PredO)},
    ( {is_truth(PredO)} -> []
    ; {otherwise} ->
        mpstart(Title,P),
        pred_over_lines(0,PredO),
        mpend(P)).
    
mpart_typing2(Sections,AllSections,PredI,PredO) :-
    print_typing_infos,!,
    get_all_ids(Sections,AllSections,Ids),
    add_typing_predicates(Ids,PredI,PredO).
mpart_typing2(_Section,_Sections,Pred,Pred).

get_all_ids([],_Sections,[]).
get_all_ids([Section|Srest],Sections,Ids) :-
    memberchk(Section/Ids1,Sections),
    append(Ids1,Ids2,Ids),
    get_all_ids(Srest,Sections,Ids2).

add_optional_typing_predicates(Ids,In,Out) :-
    ( print_typing_infos -> add_typing_predicates(Ids,In,Out)
    ; otherwise          -> In=Out).
    
add_typing_predicates([],P,P) :- !.
add_typing_predicates(Ids,Pin,Pout) :-
    remove_already_typed_ids(Pin,Ids,UntypedIds),
    generate_typing_predicates(UntypedIds,Typing),
    conjunction_to_list(Pin,Pins),
    remove_duplicate_predicates(Typing,Pins,Typing2),
    append(Typing2,[Pin],Preds),
    conjunct_predicates(Preds,Pout).

remove_already_typed_ids(_TExpr,Ids,Ids) :-
    force_all_typing_infos,!.
remove_already_typed_ids(TExpr,Ids,UntypedIds) :-
    get_texpr_expr(TExpr,Expr),
    remove_already_typed_ids2(Expr,Ids,UntypedIds).
remove_already_typed_ids2(conjunct(A,B),Ids,UntypedIds) :- !,
    remove_already_typed_ids(A,Ids,I1),
    remove_already_typed_ids(B,I1,UntypedIds).
remove_already_typed_ids2(Expr,Ids,UntypedIds) :-
    is_typing_predicate(Expr,Id),
    create_texpr(identifier(Id),_,_,TId),
    select(TId,Ids,UntypedIds),!.
remove_already_typed_ids2(_,Ids,Ids).
is_typing_predicate(member(A,_),Id) :- get_texpr_id(A,Id).
is_typing_predicate(subset(A,_),Id) :- get_texpr_id(A,Id).
is_typing_predicate(subset_strict(A,_),Id) :- get_texpr_id(A,Id).

remove_duplicate_predicates([],_Old,[]).
remove_duplicate_predicates([Pred|Prest],Old,Result) :-
    ( is_duplicate_predicate(Pred,Old) -> Result = Rest
    ; otherwise -> Result = [Pred|Rest]),
    remove_duplicate_predicates(Prest,Old,Rest).
is_duplicate_predicate(Pred,List) :-
    remove_all_infos(Pred,Pattern),
    memberchk(Pattern,List).
    
generate_typing_predicates(TIds,Preds) :-
    maplist(generate_typing_predicate, TIds, Preds).
generate_typing_predicate(TId,Pred) :-
    get_texpr_type(TId,Type),
    remove_all_infos(TId,TId2), get_texpr_info(TId2,[]), % clear all infos
    (type_set(Type,TSet) -> true ; add_error_fail(translate,'Illegal type: ',Type)),
    create_texpr(member(TId2,TSet),pred,[],Pred).

% translate a ProB type to a typed expression representing the maximal type
set_type_to_maximal_texpr(ProBType,TExpr) :- replace_seq(ProBType,MType),type_set(MType,TExpr).

% replace all occurences of seq/1 by couple(....) 
replace_seq(set(T),Res) :- !, Res=set(RT),replace_seq(T,RT).
replace_seq(seq(T),Res) :- !, Res=set(couple(integer,RT)),replace_seq(T,RT).
replace_seq(couple(A,B),Res) :- !, Res=couple(RA,RB),replace_seq(A,RA), replace_seq(B,RB).
replace_seq(X,X).

% translate a ProB type into a B expression so that we can write x:TSet
type_set(Type,TSet) :-
    ( var(Type) -> add_error_and_fail(translate,'type_set Type is variable: ',Type)
    ; otherwise ->
        type_set2(Type,Set,Infos),
        create_texpr(Set,set(Type),Infos,TSet)).
type_set2(set(T),pow_subset(Set),[]) :-
    type_set(T,Set).
type_set2(seq(T),seq(Set),[]) :- type_set(T,Set).
type_set2(couple(TA,TB),cartesian_product(SetA,SetB),[]) :-
    type_set(TA,SetA), type_set(TB,SetB).
type_set2(global(G),identifier(G),[given_set]).
type_set2(integer,integer_set('INTEGER'),[]).
type_set2(string,string_set,[]).
type_set2(boolean,bool_set,[]).
type_set2(record(TFields), struct(Rec),[]) :-
    create_texpr(rec(SFields),record(TFields),[],Rec),
    type_set_fields(TFields,SFields).
%type_set2(freetype(Id),identifier(Id),[given_set]).
type_set2(freetype(Id),freetype_set(Id),[]).
type_set_fields([],[]).
type_set_fields([field(Id,Type)|TFrest],[field(Id,TSet)|EFrest]) :-
    type_set(Type,TSet),
    type_set_fields(TFrest,EFrest).


% translate enumerated constant list into enumerate set definition
translate_enums([],Acc,Acc).
translate_enums([EnumCst|T],Acc,Res) :- %get_texpr_id(EnumCst,Id),
    get_texpr_type(EnumCst,global(GlobalSet)),
    insert_enum_cst(Acc,EnumCst,GlobalSet,Acc2),
    translate_enums(T,Acc2,Res).

insert_enum_cst([],ID,Type,[enumerated_set_def(Type,[ID])]).
insert_enum_cst([enumerated_set_def(Type,Lst)|T],ID,Type2,[enumerated_set_def(Type,Lst2)|TT]) :-
   (Type=Type2
     -> Lst2 = [ID|Lst], TT=T
     ;  Lst2 = Lst, insert_enum_cst(T,ID,Type2,TT)
   ).

% pretty-print the initialisation section of a machine
translate_inits(Inits) -->
    ( {is_list_simple(Inits)} ->
        translate_inits2(Inits)
    ; {otherwise} ->
        indention_level(I,I2),{I2 is I+2},
        translate_subst_begin_end(Inits),
        indention_level(_,I)).
translate_inits2([]) --> !.
translate_inits2([init(Name,Subst)|Rest]) -->
    indent('/* '),insertstr(Name),insertstr(': */ '),
    translate_subst_begin_end(Subst),
    translate_inits2(Rest).

% pretty-print the definitions of a machine
translate_defs([]) --> !.
translate_defs([Def|Rest]) --> translate_def(Def),translate_defs(Rest).
translate_def(definition_decl(Name,DefType,_Pos,_Args,_Expr,_Deps)) -->
    {def_description(DefType,Desc)}, indent(Desc),insertstr(Name).
def_description(substitution,'SUBSTITUTION ').
def_description(expression,'EXPRESSION ').
def_description(predicate,'PREDICATE ').
                
% pretty-print the operations of a machine
translate_ops([]) --> !.
translate_ops([Op|Rest]) -->
    translate_op(Op),
    ({Rest=[]} -> {true}; insertstr(';'),indent),
    translate_ops(Rest).
translate_op(Op) -->
    { get_texpr_expr(Op,operation(Id,Res,Params,Body)) },
    indent,translate_op_results(Res),
    pp_expr_indent(Id),
    translate_op_params(Params),
    insertstr(' = '),
    indention_level(I1,I2),{I2 is I1+2,type_infos_in_subst(Params,Body,Body2)},
    translate_subst_begin_end(Body2),
    indention_level(_,I1).
translate_op_results([]) --> !.
translate_op_results(Ids) --> pp_expr_indent_l(Ids), insertstr(' <-- ').
translate_op_params([]) --> !.
translate_op_params(Ids) --> insertstr('('),pp_expr_indent_l(Ids), insertstr(')').

translate_subst_begin_end(TSubst) -->
    {get_texpr_expr(TSubst,Subst),subst_needs_begin_end(Subst),
     create_texpr(block(TSubst),subst,[],Block)},!,
    translate_subst(Block).
translate_subst_begin_end(Subst) -->
    translate_subst(Subst).

subst_needs_begin_end(assign(_,_)).
subst_needs_begin_end(assign_single_id(_,_)).
subst_needs_begin_end(parallel(_)).
subst_needs_begin_end(sequence(_)).
subst_needs_begin_end(operation_call(_,_,_)).

type_infos_in_subst([],Subst,Subst) :- !.
type_infos_in_subst(Ids,SubstIn,SubstOut) :-
    print_typing_infos,!,
    type_infos_in_subst2(Ids,SubstIn,SubstOut).
type_infos_in_subst(_Ids,Subst,Subst).
type_infos_in_subst2(Ids,SubstIn,SubstOut) :-
    get_texpr_expr(SubstIn,precondition(P1,S)),!,
    get_texpr_info(SubstIn,Info),
    create_texpr(precondition(P2,S),pred,Info,SubstOut),
    add_typing_predicates(Ids,P1,P2).
type_infos_in_subst2(Ids,SubstIn,SubstOut) :-
    create_texpr(precondition(P,SubstIn),pred,[],SubstOut),
    generate_typing_predicates(Ids,Typing),
    conjunct_predicates(Typing,P).
    


    

% pretty-print the internal section about included and used machines
translate_used([]) --> !.
translate_used([Used|Rest]) -->
    translate_used2(Used),
    translate_used(Rest).
translate_used2(includeduse(Name,Id,NewTExpr)) -->
    indent,pp_expr_indent(NewTExpr),
    insertstr(' --> '), insertstr(Name), insertstr(':'), insertstr(Id).

% pretty-print the internal information about freetypes
translate_freetypes([]) --> !.
translate_freetypes([Freetype|Frest]) -->
    translate_freetype(Freetype),
    translate_freetypes(Frest).
translate_freetype(freetype(Name,Cases)) -->
    {pretty_freetype(Name,PName)},
    indent(PName),insertstr(': '),
    indention_level(I1,I2),{I2 is I1+2},
    translate_freetype_cases(Cases),
    indention_level(_,I2).
translate_freetype_cases([]) --> !.
translate_freetype_cases([case(Name,constant(_))|Rest]) -->
    !,indent(Name),translate_freetype_cases(Rest).
translate_freetype_cases([case(Name,Type)|Rest]) -->
    {pretty_type(Type,PT)},
    indent(Name),
    insertstr('('),insertstr(PT),insertstr(')'),
    translate_freetype_cases(Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% substitutions

print_subst(Stmt) :- translate_substitution(Stmt,T), print(T).
translate_substitution(Stmt,String) :- translate_subst_with_indention(Stmt,0,Codes,[]),
   atom_codes_with_limit(String, Codes),!.
translate_substitution(Stmt,String) :-
   add_error(translate_substitution,'Could not translate substitution: ',Stmt),String='???'.

translate_subst_with_indention(TS,Indention,I,O) :-
    translate_subst(TS,(Indention,I),(_,O)).

translate_subst(TS) -->
    ( {get_texpr_expr(TS,S)} ->
        %indent_rodin_label(TS), % comment in to pretty print substitution labels
        translate_subst2(S)
    ; {otherwise} -> translate_subst2(TS)).

% will print (first) rodin or pragma label indendent
indent_rodin_label(_TExpr) --> {suppress_rodin_positions_flag,!}.
indent_rodin_label(TExpr) --> {bsyntaxtree:get_texpr_label(TExpr,Names)},!, % note: this will only get the first label
   indent('/* @'),pp_ids_indent(Names),insertstr('*/ ').
indent_rodin_label(_TExpr) --> [].

pp_ids_indent([]) --> !, [].
pp_ids_indent([ID|T]) --> !,pp_expr_indent(identifier(ID)), insertstr(' '),pp_ids_indent(T).
pp_ids_indent(X) --> {add_error(pp_ids_indent,'Not a list of atoms: ',pp_ids_indent(X))}.

translate_subst2(skip) -->
    indent(skip).
translate_subst2(precondition(P,S)) -->
    indent('PRE '), pred_over_lines(2,P), indent('THEN'), insert_subst(S), indent('END').
translate_subst2(assertion(P,S)) -->
    indent('ASSERT '), pp_expr_indent(P), indent('THEN'), insert_subst(S), indent('END').
translate_subst2(block(S)) -->
    indent('BEGIN'), insert_subst(S), indent('END').
translate_subst2(assign(L,R)) -->
    indent,pp_expr_indent_l(L),insertstr(' := '),pp_expr_indent_l(R).
translate_subst2(assign_single_id(L,R)) -->
    translate_subst2(assign([L],[R])).
translate_subst2(becomes_element_of(L,R)) -->
    indent,pp_expr_indent_l(L),insertstr(' :: '),pp_expr_indent(R).
translate_subst2(becomes_such(L,R)) -->
    indent,pp_expr_indent_l(L),insertstr(' : '), insertstr('('),
    { add_optional_typing_predicates(L,R,R1) },
    pp_expr_indent(R1), insertstr(')').
translate_subst2(if([Elsif|Rest])) -->
    { optional_type(Elsif,if_elsif(P,S)) },
    indent('IF '), pp_expr_indent(P), insertstr(' THEN'),
    insert_subst(S),
    translate_ifs(Rest).
translate_subst2(parallel(Ss)) -->
    split_over_lines(Ss,'||').
translate_subst2(init_parallel(Ss)) -->
    split_over_lines(Ss,'||').
translate_subst2(sequence(Ss)) -->
    split_over_lines(Ss,';').
translate_subst2(operation_call(Id,Rs,As)) -->
    indent,translate_op_results(Rs),
    pp_expr_indent(Id),
    translate_op_params(As).
translate_subst2(any(Ids,Pred,Subst)) -->
    indent('ANY '), pp_expr_indent_l(Ids),
    indent('WHERE '),
    {add_optional_typing_predicates(Ids,Pred,Pred2)},
    pred_over_lines(2,Pred2), indent('THEN'),
    insert_subst(Subst),
    indent('END').
translate_subst2(select(Whens)) -->
    translate_whens(Whens,'SELECT '),
    indent('END').
translate_subst2(select(Whens,Else)) -->
    translate_whens(Whens,'SELECT '),
    indent('ELSE'), insert_subst(Else),
    indent('END').
translate_subst2(var(Ids,S)) -->
    indent('VAR '),
    pp_expr_indent_l(Ids),
    indent('IN'),insert_subst(S),
    indent('END').
translate_subst2(let(Ids,P,S)) -->
    indent('LET '),
    pp_expr_indent_l(Ids),
    insertstr(' BE '), pp_expr_indent(P),
    indent('IN'), insert_subst(S),
    indent('END').
translate_subst2(while(Pred,Subst,Inv,Var)) -->
    indent('WHILE '), pp_expr_indent(Pred),
    indent('DO'),insert_subst(Subst),
    indent('INVARIANT '),pp_expr_indent(Inv),
    indent('VARIANT '),pp_expr_indent(Var),
    indent('END').
translate_subst2(while1(Pred,Subst,Inv,Var)) -->
    indent('WHILE /* 1 */ '), pp_expr_indent(Pred),
    indent('DO'),insert_subst(Subst),
    indent('INVARIANT '),pp_expr_indent(Inv),
    indent('VARIANT '),pp_expr_indent(Var),
    indent('END').
translate_subst2(rlevent(Id,Section,Status,Parameters,Guard,Theorems,Actions,VWitnesses,PWitnesses,_Unmod,Refines)) -->
    indent,
    insert_status(Status),
    insertstr('EVENT '), insertstr(Id), insertstr(' = /'), insertstr('* of machine '),
    insertstr(Section),insertstr(' */'),
    insert_variant(Status),
    ( {Parameters=[], get_texpr_expr(Guard,truth)} ->
        {true}                  % indent('BEGIN ')
    ; {Parameters=[]} ->
        indent('WHEN '),
        pred_over_lines(2,Guard)
    ; {otherwise} ->
        indent('ANY '),pp_expr_indent_l(Parameters),
        indent('WHERE '),
        pred_over_lines(2,Guard) ),
    ( {VWitnesses=[],PWitnesses=[]} ->
        []
    ; {otherwise} ->
        {append(VWitnesses,PWitnesses,Witnesses)},
        indent('WITH '),pp_witness_l(Witnesses) ),
    {( Actions=[] ->
       create_texpr(skip,subst,[],Subst)
     ; otherwise ->
       create_texpr(parallel(Actions),subst,[],Subst))},
    ( {Theorems=[]} -> {true}
    ; {otherwise} ->
        indent('THEOREMS '),
        preds_over_lines(2,Theorems)),
    indent('THEN '), insert_subst(Subst),
    pp_refines_l(Refines,Id),
    indent('END').

insert_status(TStatus) -->
    {get_texpr_expr(TStatus,Status),
     status_string(Status,String)},
    insertstr(String).
status_string(ordinary,'').
status_string(anticipated(_),'ANTICIPATED ').
status_string(convergent(_),'CONVERGENT ').

insert_variant(TStatus) -->
    {get_texpr_expr(TStatus,Status)},
    insert_variant2(Status).
insert_variant2(ordinary) --> !.
insert_variant2(anticipated(Variant)) --> insert_variant3(Variant).
insert_variant2(convergent(Variant)) --> insert_variant3(Variant).
insert_variant3(Variant) -->
    indent('USING VARIANT '),pp_expr_indent(Variant).

pp_refines_l([],_) --> [].
pp_refines_l([Ref|Rest],Id) -->
    pp_refines(Ref,Id),pp_refines_l(Rest,Id).
pp_refines(Refined,_Id) -->
    % indent(Id), insertstr(' REFINES '),
    indent('REFINES '),
    insert_subst(Refined).

pp_witness_l([]) --> [].
pp_witness_l([Witness|WRest]) -->
        pp_witness(Witness),pp_witness_l(WRest).
pp_witness(Expr) -->
    indention_level(I1,I2),
    {get_texpr_expr(Expr,witness(Id,Pred)),
     I2 is I1+2},
    indent, pp_expr_indent(Id), insertstr(': '),
    pp_expr_indent(Pred),
    indention_level(_,I1).


translate_whens([],_) --> !.
translate_whens([When|Rest],T) -->
    {get_texpr_expr(When,select_when(P,S))},
    indent(T), pred_over_lines(2,P),
    indent('THEN '),
    insert_subst(S),
    translate_whens(Rest,'WHEN ').

split_over_lines([],_) --> !.
split_over_lines([S|Rest],Symbol) --> !,
    indention_level(I1,I2),{name(Symbol,X),length(X,N),I2 is I1+N+1},
    translate_subst(S),
    split_over_lines2(Rest,Symbol,I1,I2).
split_over_lines(S,Symbol) --> {add_error(split_over_lines,'Illegal argument: ',Symbol:S)}.

split_over_lines2([],_,_,_) --> !.
split_over_lines2([S|Rest],Symbol,I1,I2) -->
    indention_level(_,I1), indent(Symbol),
    indention_level(_,I2), translate_subst(S),
    split_over_lines2(Rest,Symbol,I1,I2).

pred_over_lines(N,Pred) -->
    {conjunction_to_list(Pred,List)},
    preds_over_lines(N,List).
preds_over_lines(N,Preds) -->
    indention_level(I1,I2),{I2 is I1+N},
    preds_over_lines1(Preds),
    indention_level(_,I1).
preds_over_lines1([]) --> !, preds_over_lines1([b(truth,pred,[])]).
preds_over_lines1([H|T]) -->
    indent('  '),pp_expr_indent(H),
    preds_over_lines2(T).
preds_over_lines2([]) --> !.
preds_over_lines2([E|Rest]) -->
    indent('& '),
    pp_expr_m_indent(E,40),
    preds_over_lines2(Rest).

translate_ifs([]) --> !,
    indent('END').
translate_ifs([Elsif]) -->
    {optional_type(Elsif,if_elsif(P,S)),
     optional_type(P,truth)},!,
    indent('ELSE'), insert_subst(S), indent('END').
translate_ifs([Elsif|Rest]) -->
    {optional_type(Elsif,if_elsif(P,S))},
    indent('ELSIF '), pp_expr_indent(P), insertstr(' THEN'),
    insert_subst(S),
    translate_ifs(Rest).

insert_subst(S) -->
    indention_level(I,I2),{I2 is I+2},
    translate_subst(S),
    indention_level(_,I).

indent_expr(Expr) -->
    indent, pp_expr_indent(Expr).
%indent_expr_l([]) --> !.
%indent_expr_l([Expr|Rest]) -->
%    indent_expr(Expr), indent_expr_l(Rest).
indent_expr_l_sep([],_) --> !.
indent_expr_l_sep([Expr|Rest],Sep) --> 
    indent_expr(Expr),
    {(Rest=[] -> RealSep='' ;  RealSep=Sep)},
    insert_atom(RealSep), % the threaded argument is a pair, not directly a string !
    indent_expr_l_sep(Rest,Sep).
%indention_level(L) --> indention_level(L,L).
indention_level(Old,New,(Old,S),(New,S)).
indent --> indent('').
indent(M,(I,S),(I,T)) :- indent2(I,M,S,T).
indent2(Level,Msg) -->
    "\n",do_indention(Level),ppatom(Msg).

insert_atom(Sep,(I,S),(I,T)) :- ppatom(Sep,S,T).

insertstr(M,(I,S),(I,T)) :- ppterm(M,S,T).

do_indention(0,T,R) :- !, R=T.
do_indention(N,[32|I],O) :-
    N>0,N2 is N-1, do_indention(N2,I,O).

optional_type(Typed,Expr) :- get_texpr_expr(Typed,E),!,Expr=E.
optional_type(Expr,Expr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% expressions and predicates

% pretty-type an expression in an indent-environment
% currently, the indent level is just thrown away
pp_expr_indent(Expr,(I,S),(I,T)) :-
    pp_expr(Expr,_,S,T).
pp_expr_indent_l(Exprs,(I,S),(I,T)) :-
    pp_expr_l(Exprs,S,T).
pp_expr_m_indent(Expr,MinPrio,(I,S),(I,T)) :-
    pp_expr_m(Expr,MinPrio,S,T).

constants(boolean_true,'TRUE').
constants(boolean_false,'FALSE').
constants(max_int,'MAXINT').
constants(min_int,'MININT').
constants(empty_set,'{}').
constants(bool_set,'BOOL').
constants(string_set,'STRING').
constants(empty_sequence,'[]').
constants(event_b_identity,'id').

constants(truth,Res) :- animation_minor_mode(eventb),!,Res=true.
constants(truth,'(TRUE:BOOL)'). % __truth;  we could also do TRUE=TRUE
constants(falsity,Res) :- animation_minor_mode(eventb),!,Res=false.
constants(falsity,'(TRUE=FALSE)'). % __falsity

function_like(convert_bool,bool).
function_like(successor,succ).
function_like(predecessor,pred).
function_like(max,max).
function_like(min,min).
function_like(card,card).
function_like(pow_subset,'POW').
function_like(pow1_subset,'POW1').
function_like(fin_subset,'FIN').
function_like(fin1_subset,'FIN1').
function_like(identity,id).
function_like(first_projection,prj1).
function_like(first_of_pair,'__first_of_pair').  % will be dealt with separately to generate parsable representation
function_like(second_projection,prj2).
function_like(second_of_pair,'__second_of_pair'). % will be dealt with separately to generate  parsable representation
function_like(iteration,iterate).
function_like(reflexive_closure,closure).
function_like(closure,closure1).
function_like(domain,dom).
function_like(range,ran).
function_like(seq,seq).
function_like(seq1,seq1).
function_like(iseq,iseq).
function_like(iseq1,iseq1).
function_like(perm,perm).
function_like(size,size).
function_like(first,first).
function_like(last,last).
function_like(front,front).
function_like(tail,tail).
function_like(rev,rev).
function_like(general_concat,conc).
function_like(general_union,union).
function_like(general_intersection,inter).

function_like(negation,not).
function_like(bag_items,items).

function_like(finite,finite). % from Event-B, TO DO: if \+ animation_minor_mode(eventb) then print as S:FIN(S)

unary_prefix(unary_minus,-,210).
unary_prefix(mu,'MU',210).

unary_prefix_parentheses(compaction,'compaction').
unary_prefix_parentheses(bag_items,'bag_items').

unary_postfix(reverse,'~',230).


always_surround_by_parentheses(parallel_product).
always_surround_by_parentheses(composition).

% EXPR * EXPR --> EXPR
binary_infix(composition,';',20,left).
binary_infix(overwrite,'<+',160,left).
binary_infix(direct_product,'><',160,left).
binary_infix(parallel_product,'||',20,left).
binary_infix(concat,'^',160,left).
binary_infix(relations,'<->',125,left).
binary_infix(partial_function,'+->',125,left).
binary_infix(total_function,'-->',125,left).
binary_infix(partial_injection,'>+>',125,left).
binary_infix(total_injection,'>->',125,left).
binary_infix(partial_surjection,'+->>',125,left).
binary_infix(total_surjection,Symbol,125,left) :-
  (animation_minor_mode(eventb) -> Symbol = '->>'; Symbol = '-->>').
binary_infix(total_bijection,'>->>',125,left).
binary_infix(partial_bijection,'>+>>',125,left).
binary_infix(total_relation,'<<->',125,left).             % only in Event-B
binary_infix(surjection_relation,'<->>',125,left).        % only in Event-B
binary_infix(total_surjection_relation,'<<->>',125,left). % only in Event-B
binary_infix(insert_front,'->',160,left).
binary_infix(insert_tail,'<-',160,left).
binary_infix(domain_restriction,'<|',160,left).
binary_infix(domain_subtraction,'<<|',160,left).
binary_infix(range_restriction,'|>',160,left).
binary_infix(range_subtraction,'|>>',160,left).
binary_infix(intersection,'/\\',160,left).
binary_infix(union,'\\/',160,left).
binary_infix(restrict_front,'/|\\',160,left).
binary_infix(restrict_tail,'\\|/',160,left).
binary_infix(couple,'|->',160,left).
binary_infix(interval,'..',170,left).
binary_infix(add,+,180,left).
binary_infix(minus,-,180,left).
binary_infix(set_subtraction,'\\',180,left) :- animation_minor_mode(eventb),!.
binary_infix(set_subtraction,-,180,left).
binary_infix(minus_or_set_subtract,-,180,left).
binary_infix(multiplication,*,190,left).
binary_infix(cartesian_product,**,190,left) :- animation_minor_mode(eventb),!.
binary_infix(cartesian_product,*,190,left).
binary_infix(mult_or_cart,*,190,left). % in case type checker not yet run
binary_infix(div,/,190,left).
binary_infix(modulo,mod,190,left).
binary_infix(power_of,**,200,right).


% PRED * PRED --> PRED
binary_infix(implication,'=>',30,left).
binary_infix(conjunct,'&',40,left).
binary_infix(disjunct,or,40,left).
binary_infix(equivalence,'<=>',60,left).

% EXPR * EXPR --> PRED
binary_infix(equal,=,60,left).
binary_infix(not_equal,'/=',160,left).
binary_infix(less_equal,'<=',160,left).
binary_infix(less,'<',160,left).
binary_infix(greater_equal,'>=',160,left).
binary_infix(greater,'>',160,left).
binary_infix(member,':',60,left).
binary_infix(not_member,'/:',160,left).
binary_infix(subset,'<:',110,left).
binary_infix(subset_strict,'<<:',110,left).
binary_infix(not_subset,'/<:',110,left).
binary_infix(not_subset_strict,'/<<:',110,left).

quantified(general_sum,'SIGMA').
quantified(general_product,'PI').
quantified(quantified_union,'UNION').
quantified(quantified_intersection,'INTER').
quantified(lambda,X) :- name(X,[37]).


translate_prolog_constructor(C,R) :- unary_prefix(C,R,_),!.
translate_prolog_constructor(C,R) :- unary_postfix(C,R,_),!.
translate_prolog_constructor(C,R) :- binary_infix(C,R,_,_),!.
translate_prolog_constructor(C,R) :- function_like(C,R),!.
translate_prolog_constructor(C,R) :- constants(C,R),!.
translate_prolog_constructor(C,R) :- quantified(C,R),!.


   
translate_bexpression(TExpr,String) :-
    (pp_expr(TExpr,String) -> true
     ; (add_error(translate_bexpression,'Could not translate bexpression: ',TExpr),String='???')).

translate_bexpression_to_codes(TExpr,Codes) :-
    pp_expr(TExpr,_,Codes,[]).

pp_expr(TExpr,String) :-
    translate_bexpression_to_codes(TExpr,Codes),
    atom_codes_with_limit(String, Codes).

translate_bexpression_with_limit(T,S) :- translate_bexpression_with_limit(T,200,S).
translate_bexpression_with_limit(TExpr,Limit,String) :-
    (catch_call(pp_expr_with_limit(TExpr,Limit,String)) -> true
     ; (add_error(translate_bexpression,'Could not translate bexpression: ',TExpr),String='???')).

pp_expr_with_limit(TExpr,Limit,String) :-
    pp_expr(TExpr,_,Codes,[]),
    atom_codes_with_limit(String, Limit, Codes).


% pretty-type an expression, if the expression has a priority >MinPrio, parenthesis
% can be ommitted, if not the expression has to be put into parenthesis
pp_expr_m(TExpr,MinPrio,S,Srest) :-
    pp_expr(TExpr,Prio,X,Xrest),
    ( Prio > MinPrio -> % was >=, but problem with & / or with same priority  or with non associative operators !
        S=X, Srest=Xrest
    ; otherwise ->
        [Open] = "(", [Close] = ")",
        S = [Open|X], Xrest = [Close|Srest]).

translated_identifier('_zzzz_binary',R) :- !, R='z__'.
translated_identifier('_zzzz_unary',R) :- !, R='z_'.
translated_identifier(NameWithDots,R) :-
    atom_codes(NameWithDots,Codes),[Dot]=".",
    memberchk(Dot,Codes),!,
    replace_codes(Codes,Dot,"__",NewCodes,[]),
    atom_codes(R,NewCodes).
translated_identifier(X,X).

replace_codes([],_,_) --> !.
replace_codes([C|Rest],S,N) --> replace_codes2(C,S,N),replace_codes(Rest,S,N).
replace_codes2(S,S,N,I,O) :- !,append(N,O,I).
replace_codes2(S,_,_) --> [S].

is_lambda_result_id(b(identifier(ID),_,_INFO),Suffix) :-  % _INFO=[lambda_result], sometiems _INFO=[]
    is_lambda_result_name(ID,Suffix).
is_lambda_result_name(LAMBDA_RESULT,Suffix) :- atomic(LAMBDA_RESULT),
   name(LAMBDA_RESULT,[95,108,97,109,98,100,97,95,114,101,115,117,108,116,95|Suffix]).

pp_expr(TExpr,Prio) --> {var(TExpr)},!,"_",{Prio=500}.
pp_expr(LambdaID,Prio) --> {is_lambda_result_id(LambdaID,Suffix)},!, {Prio=500},
    {append("LAMBDA_RESULT___",Suffix,ASCII), name(R,ASCII)}, ppatom(R).
pp_expr(TExpr,Prio) -->
    ( {is_texpr(TExpr)} ->
        { %get_texpr_type(TExpr,Type),  % we could dispatch depending on type
          get_texpr_expr(TExpr,Expr),get_texpr_info(TExpr,Info)},
        pp_rodin_label(TExpr),
        pp_output_pragma(TExpr),
        pp_expr1(Expr,Info,Prio)
    ; pp_expr2(TExpr,Prio)).

% pretty print output pragmas for internal representation
pp_output_pragma(TExpr) --> {bsyntaxtree:get_texpr_output_pragma(TExpr,Pragma,Content)},!,
                            "/*@ ",ppatoms(Pragma),ppatoms(Content),"*/ ".
pp_output_pragma(_TExpr) --> [].

% will pretty print (first) rodin or pragma label
pp_rodin_label(_TExpr) --> {suppress_rodin_positions_flag,!}.
pp_rodin_label(TExpr) --> {bsyntaxtree:get_texpr_label(TExpr,Names)},!, "/* @",ppatoms(Names),"*/ ".
pp_rodin_label(_TExpr) --> [].

ppatoms([]) --> !, [].
ppatoms([ID|T]) --> !,ppatom(ID), " ", ppatoms(T).
ppatoms(X) --> {add_error(ppatoms,'Not a list of atoms: ',ppatoms(X))}.

pp_expr1(Expr,_,Prio) --> {var(Expr)},!,"_",Prio=500.
pp_expr1(comprehension_set([I1,I2,_I3],P),Info,500) -->
    /* This comprehension set was a projection function (prj1/prj2) */
    {memberchk(was(Prj),Info),(Prj=prj1;Prj=prj2),!,
     lookup_membership(I1,P,Set1), lookup_membership(I2,P,Set2)},
    ppatom(Prj),"(",pp_expr(Set1,_),",",pp_expr(Set2,_),")".
pp_expr1(comprehension_set([_],_),Info,500) -->
    {memberchk(freetype(P),Info),!},ppatom(P).
pp_expr1(Expr,_Info,Prio) --> pp_expr2(Expr,Prio).

lookup_membership(TId,P,Set) :-
    get_texpr_id(TId,Id),get_texpr_id(Temp,Id),
    create_texpr(member(Temp,Set),pred,_,Member),
    member_in_conjunction(Member,P),!.
lookup_membership(TId,_P,atom_string(TS)) :- % atom_string used as wrapper for pp_expr2
    get_texpr_type(TId,Type),pretty_type(Type,TS).
%    add_error(translate,'Failed: ',lookup_membership(TId,P,Set)), Set=unknown.
    

pp_expr2(Expr,_) --> {var(Expr)},!,"_".
pp_expr2(value(V),Prio) --> !,
 {(nonvar(V),V=closure(_,_,_) -> Prio=300 ; Prio=500)}, pp_value(V).
pp_expr2(atom_string(V),500) --> !,ppatom(V). % hardwired_atom
pp_expr2(global_set(V),500) --> !,ppatom(V).
pp_expr2(freetype_set(V),500) --> !,{pretty_freetype(V,P)},ppatom(P).
pp_expr2(identifier(I),500) --> !,
    {( I=op(Id) -> true; I=Id)},
    ( {atomic(Id)} -> ({translated_identifier(Id,TId)},ppatom(TId))
    ; {otherwise} ->
        "'",ppterm(Id), "'").
pp_expr2(integer(N),500) --> !, ppnumber(N).
pp_expr2(integer_set(S),500) --> !,
    {integer_set_mapping(S,T)},ppatom(T).
pp_expr2(string(S),500) --> !, """",ppatom(S),"""".
pp_expr2(set_extension(Ext),500) --> !,
    pp_expr_wrap_l('{',Ext,'}').
pp_expr2(sequence_extension(Ext),500) --> !,
    pp_expr_wrap_l('[',Ext,']').
pp_expr2(assign(LHS,RHS),10) --> !,
    pp_expr_wrap_l(',',LHS,''), ":=", pp_expr_wrap_l(',',RHS,'').
pp_expr2(assign_single_id(LHS,RHS),10) --> !, pp_expr2(assign([LHS],[RHS]),10).
pp_expr2(parallel(RHS),10) --> !,
    pp_expr_wrap_l('||',RHS,'').
pp_expr2(sequence(RHS),10) --> !,
    pp_expr_wrap_l(';',RHS,'').
pp_expr2(comprehension_set(Ids,P1),500) --> !,
    pp_comprehension_set(Ids,P1).
pp_expr2(image(A,B),300) --> !,
    pp_expr_m(A,249),"[", % was 0; but we may have to bracket A; e.g., f <| {2} [{2}] is not ok; 250 is priority of lambda
    pp_expr_m(B,0),"]". % was 500, now set to 0: we never need an outer pair of () !?
pp_expr2(function(A,B),300) --> !,
    pp_expr_m(A,249),"(", % was 0; but we may have to bracket A; e.g., f <| {2} (2) is not ok; 250 is priority of lambda
    pp_expr_m(B,0),")". % was 500, now set to 0: we never need an outer pair of () !?
pp_expr2(enumerated_set_def(GS,ListEls),200) --> !,  % for pretty printing enumerate set defs
    {reverse(ListEls,RLE)}, /* they have been inserted in inverse order */
    ppatom(GS), "=", pp_expr_wrap_l('{',RLE,'}').
pp_expr2(forall(Ids,D1,P),250) --> !,
    "!",pp_expr_ids(Ids),
    {add_typing_predicates(Ids,D1,D)},
    ".",pp_expr_m(implication(D,P),221).
pp_expr2(exists(Ids,P1),250) --> !,
    "#",pp_expr_ids(Ids),
    {add_typing_predicates(Ids,P1,P)},
    ".",pp_expr_m(P,221).
pp_expr2(record_field(R,I),250) --> !,
    pp_expr_m(R,251),"'",ppatom(I).
pp_expr2(rec(Fields),500) --> !,
    "rec(",pp_expr_fields(Fields),")".
pp_expr2(struct(Rec),500) --> 
    {get_texpr_expr(Rec,rec(Fields)) ; get_texpr_expr(Rec,value(rec(Fields)))},!,
    "struct(",pp_expr_fields(Fields),")".
pp_expr2(freetype_set(FT),500) --> !,{pretty_freetype(FT,P)},ppatom(P).
pp_expr2(freetype_case(FT,L,Expr),500) --> !,
    pp_freetype_term('__is_ft_case',FT,L,Expr).
pp_expr2(freetype_constructor(_FT,L,Expr),500) --> !,
    ppatom(L),ppatom('('),pp_expr(Expr,_),ppatom(')').
pp_expr2(freetype_destructor(FT,L,Expr),500) --> !,
    pp_freetype_term('__ft~',FT,L,Expr).
pp_expr2(let_predicate(Ids,Exprs,P),1) --> !,
    pp_expr_let_exists(Ids,Exprs,P). % instead of pp_expr_let
pp_expr2(let_expression(Ids,Exprs,P),1) --> !,
    pp_expr_let(Ids,Exprs,P).
pp_expr2(assertion_expression(Cond,_Msg,Expr),1) --> !,
    "__ASSERT ",pp_expr_m(Cond,30),
    " IN ", pp_expr_m(Expr,30).
pp_expr2(partition(S,Elems),500) --> {animation_minor_mode(eventb)},!,
    "partition(",pp_expr(S,_),
    ({Elems=[]} -> ")" ; pp_expr_wrap_l(',',Elems,')')).
pp_expr2(partition(S,Elems),500) --> !,
    "(",pp_expr(S,_), " = ", 
     ({Elems=[]} -> "{})"
      ; pp_expr_l_sep(Elems,"\\/"), pp_all_disjoint(Elems),")").
pp_expr2(if_then_else(If,Then,Else),1) --> !,
    "\\IF ",pp_expr_m(If,30),
    " \\THEN ",pp_expr_m(Then,30),
    " \\ELSE ",pp_expr_m(Else,30).
pp_expr2(kodkod(Id,Identifiers),300) -->
    "KODKOD_CALL(",ppnumber(Id),": ",pp_expr_ids(Identifiers),")".
pp_expr2(Expr,500) -->
    {constants(Expr,Symbol)},!,ppatom(Symbol).
pp_expr2(Expr,Prio) -->
    {functor(Expr,F,1),
    unary_prefix(F,Symbol,Prio),!,
    arg(1,Expr,Arg),APrio is Prio+1},
    ppatom(Symbol), pp_expr_m(Arg,APrio).
pp_expr2(Expr,500) -->
    {functor(Expr,F,1),
    unary_prefix_parentheses(F,Symbol),!,
    arg(1,Expr,Arg)},
    ppatom(Symbol), "(", pp_expr(Arg,_), ")".
pp_expr2(Expr,Prio) -->
    {functor(Expr,F,1),
    unary_postfix(F,Symbol,Prio),!,
    arg(1,Expr,Arg),APrio is Prio+1},
    pp_expr_m(Arg,APrio),ppatom(Symbol).
pp_expr2(Expr,OPrio) -->
    {functor(Expr,F,2),
    binary_infix(F,Symbol,Prio,Ass),!,
    arg(1,Expr,Left),
    arg(2,Expr,Right),
    ( Ass = left  -> LPrio = Prio, RPrio is Prio+1
    ; Ass = right -> LPrio is Prio+1, RPrio = Prio
    ; otherwise   -> LPrio is Prio+1, RPrio is Prio+1)},
    ({always_surround_by_parentheses(F)} -> "(" ; []),
    pp_expr_m(Left,LPrio),
    " ",ppatom(Symbol)," ",
    pp_expr_m(Right,RPrio),
    ({always_surround_by_parentheses(F)} -> ")",{OPrio=1000} ; {OPrio=Prio}).

pp_expr2(first_of_pair(X),500) -->  {get_texpr_type(X,couple(From,To))},!,
    {pretty_type(From,FromT), pretty_type(To,ToT)},
    "prj1(", ppatom(FromT), ",", ppatom(ToT), ")(",
    pp_expr(X,_),")".
pp_expr2(second_of_pair(X),500) -->  {get_texpr_type(X,couple(From,To))},!,
    {pretty_type(From,FromT), pretty_type(To,ToT)},
    "prj2(", ppatom(FromT), ",", ppatom(ToT), ")(",
    pp_expr(X,_),")".
pp_expr2(Call,500) --> {external_call(Call,Symbol,Args)},!,
    pp_expr_m(atom_string(Symbol),20),pp_expr_wrap_l('/*EXT:*/(',Args,')').
pp_expr2(Expr,500) -->
    {functor(Expr,F,_),
    function_like(F,Symbol),!,
    Expr =.. [F|Args]},
    ppatom(Symbol),pp_expr_wrap_l('(',Args,')').
pp_expr2(Expr,250) -->
    {functor(Expr,F,3),
    quantified(F,Symbol),!,
    Expr =.. [F,Ids,P1,E],
    add_typing_predicates(Ids,P1,P)},
    ppatom(Symbol),pp_expr_ids(Ids),".(",
    pp_expr_m(P,11),pp_such_that_bar(E),
    pp_expr_m(E,11),")".
pp_expr2(Expr,Prio) -->
    {functor(Expr,F,N),print(pp_expr2_unknown_expr(F/N,Prio,Expr)),nl, Prio=20},
        ppterm(Expr).

external_call(external_function_call(Symbol,Args),Symbol,Args).
external_call(external_pred_call(Symbol,Args),Symbol,Args).
external_call(external_subst_call(Symbol,Args),Symbol,Args).

pp_all_disjoint([H1,H2]) --> !, " & ", pp_disjoint(H1,H2).
pp_all_disjoint([H1|T]) --> pp_all_disjoint_aux(T,H1), pp_all_disjoint(T).
pp_all_disjoint([]) --> "".

pp_all_disjoint_aux([],_) --> "".
pp_all_disjoint_aux([H2|T],H1) --> " & ", pp_disjoint(H1,H2), pp_all_disjoint_aux(T,H1).

pp_disjoint(H1,H2) --> pp_expr(H1,_), "/\\", pp_expr(H2,_), " = {}".

:- use_module(bsyntaxtree,[is_a_disjunct/3]).
   
pp_comprehension_set([ID1|T],Body) --> {is_a_disjunct(Body,B1,B2),
    get_last(T,ID1,_FrontIDs,LastID),
    is_lambda_result_id(LastID,_Suffix)},!, % we seem to have the union of two lambda expressions
    "(", pp_comprehension_set([ID1|T],B1),
    " \\/ ", pp_comprehension_set([ID1|T],B2), ")".
pp_comprehension_set([b(identifier('_pred_'),integer,_),
                      b(identifier('_lambda_result_'),integer,_)],Body) -->
   {Body = b(equal(LR,T),pred,_),
    LR = b(identifier('_lambda_result_'),integer,_),
    T = b(minus(ARG,One),integer,_),
    One = b(integer(1),integer,_),
    ARG = b(identifier('_pred_'),integer,_)},
   !,
   "pred".
pp_comprehension_set([b(identifier('_succ_'),integer,_),
                      b(identifier('_lambda_result_'),integer,_)],Body) -->
   {Body = b(equal(LR,T),pred,_),
    LR = b(identifier('_lambda_result_'),integer,_),
    T = b(add(ARG,One),integer,_),
    One = b(integer(1),integer,_),
    ARG = b(identifier('_succ_'),integer,_)},
   !,
   "succ".
pp_comprehension_set([ID1|T],Body) --> {get_last(T,ID1,FrontIDs,LastID),
    is_lambda_result_id(LastID,Suffix),
     % nl, print(lambda(Body,T,ID1)),nl,
    (is_an_equality(Body,From,ToExpr) -> LambdaBody = b(truth,pred,[])
      ; is_a_conjunct(Body,LambdaBody,Equality),
        is_an_equality(Equality,From,ToExpr)),
    is_lambda_result_id(From,Suffix)},!,
    {add_typing_predicates(FrontIDs,LambdaBody,TLambdaBody)},
    "%",pp_lambda_identifiers(FrontIDs),
    ".(",pp_expr_m(TLambdaBody,11), % Check 11 against prio of . and |
    pp_such_that_bar(ToExpr),pp_expr_m(ToExpr,11),")".
pp_comprehension_set(Ids,P1) -->  
    "{",pp_expr_l(Ids),
    {add_typing_predicates(Ids,P1,P)},
    pp_such_that_bar(P),pp_expr_m(P,11),"}".

pp_lambda_identifiers([H1,H2|T]) --> !, "(",pp_expr_l([H1,H2|T]),")".
pp_lambda_identifiers(L) --> pp_expr_l(L).

pp_such_that_bar(b(unary_minus(_),_,_)) --> !, " | ".  % otherwise AtelierB complains about illegal token |- 
pp_such_that_bar(_Next) --> "|".

is_an_equality(b(equal(A,B),_,_),A,B).

integer_set_mapping(A,B) :-
    animation_minor_mode(eventb),eventb_integer_mapping(A,B),!.
integer_set_mapping(ISet,Res) :- atomic(ISet),!,Res=ISet.
integer_set_mapping(_ISet,'integer_set(??)').

eventb_integer_mapping('INTEGER','INT').
eventb_integer_mapping('NATURAL','NAT').
eventb_integer_mapping('NATURAL1','NAT1').

%get_last([b(identifier(_lambda_result_10),set(couple(integer,set(couple(integer,integer)))),[])],b(identifier(i),integer,[]),[b(identifier(i),integer,[])],b(identifier(_lambda_result_10),set(couple(integer,set(couple(integer,integer)))),[]))

get_last([],Last,[],Last).
get_last([H2|T],H1,[H1|LT],Last) :- get_last(T,H2,LT,Last).

pp_expr_wrap_l(Pre,Expr,Post) -->
    ppatom(Pre),pp_expr_l(Expr),ppatom(Post).
pp_freetype_term(Term,FT,L,Expr) -->
    {pretty_freetype(FT,P)},
    ppatom(Term),"(",ppatom(P),",",
    ppatom(L),",",pp_expr_m(Expr,500),")".

% print a list of expressions, seperated by commas
pp_expr_l(List) --> pp_expr_l_sep(List,",").
pp_expr_l_sep([Expr],_) --> !,
    pp_expr_m(Expr,0).
pp_expr_l_sep(List,Sep) --> pp_expr_l2(List,Sep).
pp_expr_l2([],_Sep) --> !.
pp_expr_l2([Expr|Rest],Sep) -->
    pp_expr_m(Expr,116),pp_expr_l3(Rest,Sep).
pp_expr_l3([],_Sep) --> !.
pp_expr_l3(Rest,Sep) -->
    Sep,pp_expr_l2(Rest,Sep).

% print the fields of a record
pp_expr_fields([field(Name,Expr)]) --> !,
    ppatom(Name),":",pp_expr_m(Expr,120).
pp_expr_fields(Fields) --> 
    pp_expr_fields2(Fields).
pp_expr_fields2([]) --> !.
pp_expr_fields2([field(Name,Expr)|Rest]) -->
    ppatom(Name),":",
    pp_expr_m(Expr,116),
    pp_expr_fields3(Rest).
pp_expr_fields3([]) --> !.
pp_expr_fields3(Rest) -->
    ",",pp_expr_fields2(Rest).

% TO DO: test more fully; identifiers seem to be wrapped in brackets
pp_expr_let_exists(Ids,Exprs,P) -->
    "#(",pp_expr_let_ids(Ids),").(",
     pp_expr_let_pred_exprs(Ids,Exprs),
    " & ",pp_expr_m(P,5), ")".
pp_expr_let_ids([]) --> [].
pp_expr_let_ids([ID|T]) --> pp_expr_let_id(ID), pp_expr_let_ids(T).
pp_expr_let_pred_exprs([],[]) --> !.
pp_expr_let_pred_exprs([Id|Irest],[Expr|Erest]) -->
    " ",pp_expr_let_id(Id),
    "=",pp_expr_m(Expr,400),
    ( {Irest=[]} -> []; " &"),
    pp_expr_let_pred_exprs(Irest,Erest).

% print a LET expression (no B syntax, used in translated Z
% models)
pp_expr_let(Ids,Exprs,P) -->
    "LET",pp_expr_let_expressions(Ids,Exprs),
    " IN ",pp_expr_m(P,5).
pp_expr_let_expressions([],[]) --> !.
pp_expr_let_expressions([Id|Irest],[Expr|Erest]) -->
    " ",pp_expr_let_id(Id),
    "==",pp_expr_m(Expr,400),
    ( {Irest=[]} -> []; ";"),
    pp_expr_let_expressions(Irest,Erest).

pp_expr_let_id(ID) --> {atomic(ID),!, print(unwrapped_let_id(ID)),nl}, pp_expr_m(identifier(ID),500).
pp_expr_let_id(ID) --> pp_expr_m(ID,500).

% print a list of identifiers
pp_expr_ids([]) --> !.
pp_expr_ids(Ids) -->
%    ( {Ids=[Id]} -> pp_expr_m(Id,221)
%    ; {otherwise} ->
    "(",pp_expr_l(Ids),")".


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pretty print types for error messages

pretty_type(Type,String) :-
    pretty_type_l([Type],[String]).

pretty_type_l(Types,Strings) :-
    extract_vartype_names(Types,N),
    pretty_type2_l(Types,N,Strings).
pretty_type2_l([],_,[]).
pretty_type2_l([T|TRest],Names,[S|SRest]) :-
    pretty_type2(T,Names,noparen,S),
    pretty_type2_l(TRest,Names,SRest).
extract_vartype_names(Types,names(Variables,Names)) :-
    term_variables(Types,Variables),
    name_variables(Variables,1,Names).

pretty_type2(X,names(Vars,Names),_,Name) :- var(X),!,exact_member_lookup(X,Name,Vars,Names).
pretty_type2(any,_,_,'?').
pretty_type2(set(T),N,_,Text) :-
    pretty_type2(T,N,noparen,TT), ajoin(['POW(',TT,')'],Text).
pretty_type2(seq(T),N,_,Text) :-
    pretty_type2(T,N,noparen,TT), ajoin(['seq(',TT,')'],Text).
pretty_type2(couple(A,B),N,Paren,Text) :-
    pretty_type2(A,N,paren,AT),pretty_type2(B,N,paren,BT),
    ajoin([AT,'*',BT],Prod),
    ( Paren == noparen ->
      Text = Prod
    ; otherwise ->
      ajoin(['(',Prod,')'],Text)).
pretty_type2(string,_,_,'STRING').
pretty_type2(integer,_,_,'INTEGER').
pretty_type2(boolean,_,_,'BOOL').
pretty_type2(global(G),_,_,A) :- ajoin([G],A).
pretty_type2(freetype(Id),N,_,A) :- pretty_freetype2(Id,N,A).
pretty_type2(pred,_,_,predicate).
pretty_type2(subst,_,_,substitution).
pretty_type2(record(Fields),N,_,Text) :-
    pretty_type_fields(Fields,N,FText),
    ajoin(['struct(',FText,')'],Text).
pretty_type2(op(Params,Results),N,_,Text) :-
    pretty_type_l(Params,N,PText),
    ( nonvar(Results),Results=[] ->
        ajoin(['op(',PText,')'],Text)
    ; otherwise ->
        pretty_type_l(Results,N,RText),
        ajoin([RText,'<--op(',PText,')'],Text) ).
pretty_type2(definition(DefType,_,_),_,_,DefType).
pretty_type2(witness,_,_,witness).
pretty_type_l(L,_,'...') :- var(L),!.
pretty_type_l([],_,'') :- !.
pretty_type_l([E|Rest],N,Text) :-
    pretty_type2(E,N,noparen,EText),
    ( nonvar(Rest),Rest=[] ->
        EText=Text
    ; otherwise ->
        pretty_type_l(Rest,N,RText),
        ajoin([EText,',',RText],Text)).

pretty_type_fields(L,_,'...') :- var(L),!.
pretty_type_fields([],_,'').
pretty_type_fields([field(Name,Type)|FRest],N,Text) :-
    pretty_type2(Type,N,noparen,TText),
    ptf_seperator(FRest,Sep),
    pretty_type_fields(FRest,N,RestText),
    ajoin([Name,':',TText,Sep,RestText],Text).
ptf_seperator(L,', ') :- var(L),!.
ptf_seperator([],'') :- !.
ptf_seperator(_,', ').

pretty_freetype(Id,A) :-
    extract_vartype_names(Id,N),
    pretty_freetype2(Id,N,A).
pretty_freetype2(Id,_,A) :- var(Id),!,A='_'.
pretty_freetype2(Id,_,A) :- atomic(Id),!,Id=A.
pretty_freetype2(Id,N,A) :-
    Id=..[Name|TypeArgs],
        pretty_type2_l(TypeArgs,N,PArgs),
        ajoin_with_sep(PArgs,',',P),
        ajoin([Name,'(',P,')'],A).

name_variables([],_,[]).
name_variables([_|VRest],Index,[Name|NRest]) :-
    ( nth1(Index,"ABCDEFGHIJKLMNOPQRSTUVWXYZ",C) -> SName = [C]
    ; otherwise ->  number_codes(Index,SName)),
    append("_",SName,CName),name(Name,CName),
    Next is Index+1,
    name_variables(VRest,Next,NRest).

ppatom(Var) --> {var(Var)},!, ppatom('$VARIABLE').
ppatom(Atom) --> {safe_atom_codes(Atom,Codes)}, ppcodes(Codes).

ppnumber(Number) --> {var(Number)},!,pp_clpfd_variable(Number).
ppnumber(inf) --> !,"inf".
ppnumber(minus_inf) --> !,"minus_inf".
ppnumber(Number) --> {number_codes(Number,Codes)}, ppcodes(Codes).

pp_clpfd_variable(X) --> "?:",{fd_dom(X,Dom)},write_to_codes(Dom),
 ":(",{frozen(X,Goal)},write_to_codes(Goal),")".

ppterm(Term) --> write_to_codes(Term).
ppcodes([],S,S).
ppcodes([C|Rest],[C|In],Out) :- ppcodes(Rest,In,Out).

% for debugging:
b_portray_hook(X) :-
    ground(X),is_texpr(X),print('{# '),print_bexpr(X),print(' #}').
install_b_portray_hook :-
    assert(( user:portray(X) :- translate:b_portray_hook(X) )).
remove_b_portray_hook :-
    retractall( user:portray(_) ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pretty-print of Event-B models as classical B

translate_eventb_to_classicalb(EBMachine,AddInfo,Rep) :-
    ( conversion_check(EBMachine) ->
        convert_eventb_classicalb(EBMachine,CBMachine),
        call_cleanup(( set_animation_mode(b), % clear minor mode "eventb"
                       translate_machine(CBMachine,Rep,AddInfo),!),
                     set_animation_minor_mode(eventb))
    ; otherwise ->
        add_error_and_fail(translate,'Conversion not applicable, check if you limited the number of abstract level to 0')
    ).

convert_eventb_classicalb(EBMachine,CBMachine) :-
    select_section(operation_bodies,In,Out,EBMachine,CBMachine1),
    maplist(convert_eventop,In,Out),
    select_section(initialisation,IIn,IOut,CBMachine1,CBMachine),
    convert_event(IIn,[],IOut).
convert_eventop(EBOp,CBOp) :-
    get_texpr_expr(EBOp,operation(Id,[],Args,EBBody)),
    get_texpr_info(EBOp,Info),
    convert_event(EBBody,Args,CBBody),
    % Remove the arguments
    create_texpr(operation(Id,[],[],CBBody),op([],[]),Info,CBOp).
convert_event(TEvent,Parameters,TSubstitution) :-
    get_texpr_expr(TEvent,rlevent(_Id,_Section,_Status,_Parameters,Guard,_Theorems,Actions,_VariableWitnesses,_ParameterWitnesses,_Ums,_Refined)),    
    in_parallel(Actions,PAction),
    convert_event2(Parameters,Guard,PAction,TSubstitution).
convert_event2([],Guard,Action,Action) :-
    is_truth(Guard),!.
convert_event2([],Guard,Action,Select) :-
    !,create_texpr(select([When]),subst,[],Select),
    create_texpr(select_when(Guard,Action),subst,[],When).
convert_event2(Parameters,Guard,Action,Any) :-
    create_texpr(any(Parameters,Guard,Action),subst,[],Any).
in_parallel([],Skip) :- !,create_texpr(skip,subst,[],Skip).
in_parallel([A],A) :- !.
in_parallel(Actions,Parallel) :- create_texpr(parallel(Actions),subst,[],Parallel).

conversion_check(Machine) :-
    animation_mode(b),
    animation_minor_mode(eventb),
    get_section(initialisation,Machine,Init),
    get_texpr_expr(Init,rlevent(_Id,_Sec,_St,_Par,_Grd,_Thms,_Act,_VW,_PW,_Ums,[])).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(clpfd)).

% print also partially instantiated variables with CLP(FD) Info
print_value_variable(X) :- var(X), !, print(X).
print_value_variable(int(X)) :- var(X),!, print('int('), print_clpfd_variable(X), print(')').
print_value_variable(fd(X,T)) :- var(X),!, print('fd('), print_clpfd_variable(X), print(','), print(T),print(')').
print_value_variable(X) :- print(X).

print_clpfd_variable(X) :- print(X), print(':'), fd_dom(X,Dom), print(Dom), print_frozen_info(X).

:- public l_print_frozen_info/1. 
l_print_frozen_info([]).
l_print_frozen_info([H|T]) :- print(H), print(' '),
  (var(H) -> print_frozen_info(H) ; 
    H=fd_var(V,_) -> print_frozen_info(V) ; true), l_print_frozen_info(T).

print_frozen_info(X) :- frozen(X,Goal), print_frozen_goal(Goal).
print_frozen_goal((A,B)) :- !, print_frozen_goal(A), print(','), print_frozen_goal(B).
print_frozen_goal(prolog:trig_nondif(_A,_B,R,_S)) :- !, frozen(R,G2), print_frozen_goal2(G2).
print_frozen_goal(G) :- print_frozen_goal2(G).
print_frozen_goal2(V) :- var(V),!, print(V).
print_frozen_goal2(true) :- !.
print_frozen_goal2((A,B)) :- !, print_frozen_goal2(A), print(','), print_frozen_goal2(B).
print_frozen_goal2(G) :- print(' :: '), tools_printing:print_term_summary(G).
