% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(b_interpreter,
   [b_test_boolean_expression/4, b_test_boolean_expression/6, 
    b_test_boolean_expression_wf/3, b_det_test_boolean_expression/6,
    b_not_test_boolean_expression/4, b_not_test_boolean_expression/6,
    b_not_test_boolean_expression_wf/3,
    
   b_compute_expression/5,
  %  b_execute_statement/6,
    b_execute_top_level_statement/7,
  
    b_compute_expression/4,
    
    b_set_up_concrete_constants/1, b_partial_set_up_concrete_constants/1,
    b_initialise_machine/4,
    
    b_execute_top_level_operation_update/5, b_execute_top_level_operation_wf/6,
    
    convert_list_of_expressions_into_set/2,
    state_satisfies_negation_of_invariant/1,
    
    b_state_violates_invariant/2,
    state_violates_assertions/2,
    
    insert_before_substitution_variables/5,  % add $0 values to a store
    
    try_lookup_value_in_store_and_global_sets/3, lookup_value_in_store_and_global_sets/4,
    set_up_typed_localstate/5, set_up_localstate_for_let/6,
    
    b_get_machine_operation_for_animation/6,
    set_projection_on_static_assertions/1, reset_b_interpreter/0, split_names_and_types/3,
    
    tcltk_unsatisfiable_components_exist/0, tcltk_quick_describe_unsat_properties/1
    ]).
    
    
    
:- use_module(kernel_objects).
:- use_module(bsets_clp).
:- use_module(delay).


:- use_module(kernel_mappings).
:- use_module(b_global_sets).
:- use_module(store).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).

:- use_module(b_global_sets).
:- use_module(bmachine).
:- use_module(b_enumerate, [b_enumerate_values_in_store/5,
      b_enumerate_values/2, 
      b_tighter_enumerate_values/2]).
:- use_module(debug).
:- use_module(self_check).
:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module provides the basic interpreter for expressions, predicates and substitutions.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(preferences).
:- use_module(error_manager).
:- use_module(typechecker).
:- use_module(specfile,[state_corresponds_to_initialised_b_machine/2
                       ,state_corresponds_to_set_up_constants/2]).

:- use_module(btypechecker,[couplise_list/2]).
:- use_module(bsyntaxtree).

:- use_module(translate).

:- use_module(kernel_objects).
:- use_module(kernel_waitflags).
:- use_module(kernel_records).
:- use_module(kernel_z).

:- use_module('kodkod/kodkod').

:- use_module(custom_explicit_sets).

:- use_module(state_space).
:- use_module(b_interpreter_eventb).
:- use_module(tools_printing,[print_term_summary/1,print_functor/1]).
:- use_module(value_persistance,[load_partial_constants/3,lookup_cached_transitions/3]).


/* ----------------------------------*/

/* For partial evaluation */

%:- use_module('~/cvs_root/cogen2/logen_source/runtime/logen_dispatcher.pl').

b_state_violates_invariant(ID,CurBState) :-
   retract(specialized_inv(ID,AVL)),  %print(specialized_inv(ID,AVL)),nl,
   intersect(AVL,SPO), !,
   \+ check_invariant_predicate_with_time_out(ID,SPO,CurBState).
b_state_violates_invariant(ID,CurBState) :-
  %%  print(checking_NORMAL_inv(_ID)),nl, %,print_bexpr(CurBState),nl,
   state_violates_invariant(ID,CurBState).

intersect(A,Inv) :- 
 % trace,
  avl_domain(A,List),
  get_predicates(List,PList),
  ord_intersection(PList,InvList), 
  conjunct_predicates(InvList,Inv).
  
get_predicates([],[]).
get_predicates([H|T],[OHL|TR]) :- 
   b_specialized_invariant_for_op(H,HR),
   conjunction_to_list(HR,HL),
   list_to_ord_set(HL,OHL), 
   get_predicates(T,TR).  
/* ----------------------------------*/

 :- load_files(library(system), [when(compile_time), imports([environ/2])]).
 :- if(\+ environ(partially_evaluate, true)).
 
reset_partial_evaluator. % nothing to be done

state_satisfies_invariant_aux(ID,State) :- 
   check_invariant_with_time_out(ID,State).  % version without partial evaluation 
   
:- else.
/* partial evaluation enabled */
:- use_module(partial_evaluator(pe)).
:- unfold(b_test_boolean_expression_for_ground_state(_,_,_)). 
:- unfold(b_test_boolean_expression_for_ground_state2(_,_,_,_)).
:- unfold(b_test_boolean_expression_wf(_,_,_)).
:- unfold(b_test_boolean_expression(_,_,_,_)).
:- unfold(b_test_boolean_expression(_,_,_,_,_,_)).
:- unfold(b_test_boolean_expression2(_Expr,_Infos,_LS,_S,_WF,_Ai,_Ao)).
:- unfold(b_test_atomic_boolean_expression2(_,_,_,_,_)).
:- unfold(b_test_implication(_CLHS,_RHS,_LocalState,_State,_WF,_Ai,_Ao)).
:- unfold(block_test_implication(_LHS,_RHS,_LocalState,_State,_WF,_WF2)).
:- unfold(b_test_equivalence(_LHS,_RHS,_LocalState,_State,_WF,_Ai,_Ao)).
:- unfold(b_check_test_equivalence(_PredRes,_LHS,_RHS,_LS,_S,_CWF,_Aii,_Ao)).
:- unfold(b_enum_test_equivalence(_LHS,_RHS,_LS,_S,_WF,_WF2)).
:- unfold(b_test_disjunction(_LHS,_RHS,_LS,_S,_WF,_WF2,_Ai)).
:- unfold(b_not_test_conjunction(_LHS,_RHS,_LS,_S,_WF,_WF2,_Ai)).
:- unfold(b_compute_expression(_, _, _, _, _)).
:- unfold(b_compute_expression2(_,_,_,_, _, _, _)).
:- execute(special_binary_predicate(_,_,_,_,_,_)).
:- execute(preferences:preference(_,_)).
:- execute(preferences:get_computed_preference(_,_)).
:- execute(get_texpr_type(B, _T), nonvar(B)).
:- execute(get_texpr_expr(B, T), (nonvar(B) ; nonvar(T))).
:- execute(always_well_defined(A),ground(A)).
:- execute(create_texpr(_,_,_,_)).
:- execute(construct_member_closure(ID,_Type,_ClosureSetExpression,_Result),ground(ID)). % could in principle be always executed
:- execute(is_set_type(T, _), nonvar(T)).
:- execute(b_interpreter:is_set_type(T, _), nonvar(T)).
:- execute(binary_function(X,M),(nonvar(X);ground(M))).
:- execute(lookup_value_in_store_and_global_sets(Id,LS,S,_Val,_Span),    (list_skeleton(LS),list_skeleton(S),ground(Id))).
%:- execute(lookup_value(Id,LS,S,R),(list_skeleton(LS),list_skeleton(S),ground(Id))).
:- execute(kernel_mappings:binary_boolean_operator(B,_,_), nonvar(B)). % does not seem to work
:- execute(kernel_mappings:binary_boolean_operator(_B,_,_)).
:- execute(empty_avl(_)).
:- execute(b_interpreter_check:register_predicate(Pred,_Infos,_PredTruthVar,_Reused,Ai,_Ao), (ground(Pred),nonvar(Ai))).
:- execute(b_interpreter:register_predicate(Pred,_Infos,_PredTruthVar,_Reused,Ai,_Ao), (ground(Pred),nonvar(Ai))).
:- execute(register_predicate(Pred,_Infos,_PredTruthVar,_Reused,Ai,_Ao), (ground(Pred),nonvar(Ai))).
:- execute(b_interpreter_check:register_predicate(_Pred,_Infos,_PredTruthVar,_Reused,_Ai,_Ao)).
:- execute(b_interpreter:register_predicate(_Pred,_Infos,_PredTruthVar,_Reused,_Ai,_Ao)).
:- execute(b_interpreter:init_wait_flags(_)).
:- execute(b_interpreter:b_get_invariant_from_machine(_)).
:- execute(same_texpr(A,B),(ground(A),ground(B))).
%% :- execute(b_interpreter:ground_wait_flags(_)). %% not safe; just for experimenting
:- unfold(b_test_member_expression(_El,_Set,_LocalState,_State,_WF)).
:- unfold(b_test_member_expression2(_S,_T,_I,_El,_LocalState,_State,_WF)).
:- execute(get_priority_of_boolean_expression2(A,_),ground(A)).
:- execute(cst_in_boolean_type(S,_), nonvar(S)).
:- execute(binary_in_boolean_type(S,_), nonvar(S)).
:- execute(unary_in_boolean_type(S,_), nonvar(S)).
:- execute(unary_function(S,_), nonvar(S)).
:- execute(symbolic_closure_unary_operator(S), nonvar(S)).
:- unfold(b_test_atomic_equal_boolean_expression(_Arg1,_Arg2,Info,LocalState,_State,_WF), 
  (ground(Info),ground(LocalState))). % TO DO: partial evaluator should itself determine when it is safe to unfold with the cuts
:- unfold(b_compute_card(_Arg1,_Type,_Info,_LocalState,_State,_Card,_WF)).
:- unfold(b_compute_card2(Arg1,_Type,_Info,_LocalState,_State,_Card,_WF),nonvar(Arg1)).
:- execute(kernel_objects:max_cardinality(ST,_MaxCard), ground(ST)).
:- unfold(b_compute_max(Ex,_,_Info,_LocalState,_State,_Max,_WF),nonvar(Ex)).
:- unfold(b_compute_min(Ex,_,_Info,_LocalState,_State,_Max,_WF),nonvar(Ex)).
:- unfold(b_compute_expressions(L,_,_,_,_),nonvar(L)).
:- unfold(b_not_test_boolean_expression(P,_LocalState,_State,_WF),nonvar(P)).
:- unfold(b_not_test_boolean_expression(P,_LocalState,_State,_WF,_,_),nonvar(P)).
:- unfold(b_not_test_boolean_expression2(P,_,_LocalState,_State,_WF,_,_),nonvar(P)).
:- unfold(b_not_test_atomic_boolean_expression2(P,_Info,_LS,_S,_WF,_Ai,_Ao),nonvar(P)).
:- unfold(b_test_notmember_expression(Set,_TElem,_TSet,_LocalState,_State,_WF),ground(Set)).
:- unfold(check_invariant_with_time_out(_,_)).
:- unfold(check_invariant_predicate_with_time_out(_,_,_)).
:- unfold(b_compute_arith_expression(Arg2,_LS,_State,_CLPFD_Expr2,_WF),nonvar(Arg2)).
:- unfold(b_compute_arith_expression2(Arg2,_I,_LS,_State,_CLPFD_Expr2,_WF),nonvar(Arg2)).
% annotations for statements:
:- execute(try_get_op_name(_FullOperation,_Name)).
:- execute(b_top_level_operation(_Name)).
:- unfold(b_execute_top_level_operation_update_wf(_Name,_FullOp,_InState,_Updates,_Path,_WF)).
:- unfold(b_execute_operation2(_Name,_FullOperation,_InState,_Updates,_,_,_,_Path,_WF,_)).
:- execute(lookup_cached_transitions(_Name,_InState,_Solutions)). % WARNING: will only work if cache disabled !!!! TO DO : fix
:- execute(b_get_machine_operation_for_animation(_,_,_,_,_,_)).
:- unfold(b_execute_operation3(_,_,_,_,_,_,_,_,_,_,_,_,_,_)).
:- execute(empty_state(_)).
:- execute(create_full_operation(_Name,_Results,ParaValues,_ResultValues,_FullOperation), builtin_tools:is_list_skeleton(ParaValues)).
%:- unfold(check_additional_guard(_LocalState,_State,_WF)).
:- execute(check_additional_guard(_LocalState,_State,_WF), \+ b_machine_temp_predicate(_)).
:- unfold(b_execute_top_level_statement(_Body,_LocalState,_NewInState,_OutState,_WF,_Path,_OR)).
:- unfold(b_execute_top_level_statement2(_Stmt,_Body,_LocalState,_NewInState,_OutState,_WF,_Path,_OR)).
:- unfold(b_execute_statement(_Body,_LocalState,_NewInState,_OutState,_WF,_Path)).
:- unfold(b_execute_statement(_Body,_LocalState,_NewInState,_OutState,_WF,_Path,_OR)).
:- unfold(b_execute_statement2(_Stmt,_Infos,_LocalState,_NewInState,_OutState,_WF,_Path,_OR)).
:- execute(get_texpr_id(IDE,_),nonvar(IDE)).
:- execute(def_get_texpr_id(IDE,_),nonvar(IDE)).
:- unfold(get_results(Results,_,_,_,_ReportError,_WF),nonvar(Results)).
%:- execute(bmachine:b_machine_temp_predicate(_)).
:- execute(b_machine_temp_predicate(_)).
:- unfold(set_up_typed_localstate(Ids,_Vals,_TVs,_In,_Out),nonvar(Ids)).
:- execute(l_expand_and_normalise_values(V,_,_),V==[]). % Replace by unfold
:- execute(set_up_undefined_localstate(V,_,_),ground(V)).
:- execute(b_tighter_enumerate_values(Vals,_WF),Vals==[]).
:- execute(nth1(A,L,_),(nonvar(A) ; builtin_tools:is_list_skeleton(L))).
:- unfold(b_execute_inner_statement(_,_,_,_,_,_,_)).
:- unfold(b_execute_inner_statement(_Body,_LocalState,_InState,_OutState,_WF,_Path,_,_OR)). % TO DO: ensure block decl. generated !
:- unfold(b_execute_statements_in_parallel(S,_,_,_,_,_,_),nonvar(S)).
:- unfold(b_execute_else_list(L,_LocalState,_InState,_OutState,_WF,_LWF,_OR),nonvar(L)).
:- unfold(b_compute_assign_expressions_when(_EXPRsTl,_LocalState,_InState,_ValTl,_WF,_OR)).
:- unfold(b_compute_expressions_when2(_EXPRsTl,_LocalState,_InState,_ValTl,_WF,_LWF)).

:- unfold(b_compute_expression_when2(_EXPRsHd,_LocalState,_InState,_ValHd,_WF,_)). % TO DO : ensure block decl. respected !
:- execute(pe_print(_)).
pe_print(X) :- print('*** PE INFO: '), print(X),nl.

reset_partial_evaluator :- clear_specialized_code.
state_satisfies_invariant_aux(ID,State) :- 
   (fail %specialized_predicate(check_invariant_with_time_out(ID,State),ResCall) % seems to slow down slightly !!
     -> true
     ;  %abstract_state(State,StateSkel),
        generalize_call(list(term(bind,[atomic,dynamic])),State,StateSkel),
        partially_evaluate_and_load(check_invariant_with_time_out(PID,StateSkel ),ResCall), % version with
        StateSkel=State, PID=ID
    ),
    set_error_context(checking_invariant),
    call(ResCall), % still missing : time_out
    clear_error_context.

%abstract_state([],[]).
%abstract_state([bind(ID,_)|T],[bind(ID,_)|AT]) :- abstract_state(T,AT).
:- endif.

:- use_module(library(timeout)).


check_invariant_with_time_out(ID,State) :- 
   b_get_invariant_from_machine(Invariant),
   check_invariant_predicate_with_time_out(ID,Invariant,State).

check_invariant_predicate_with_time_out(_ID,b(truth,_,_),_State) :- !. % only typing in invariant
check_invariant_predicate_with_time_out(ID,Invariant,State) :-
    set_error_context(checking_invariant),
    %debug_println(10,checking_invariant), % translate:print_bexpr(Invariant),nl, print(State),nl,
    preferences:get_computed_preference(debug_time_out,DebugTimeOut),
    % print(checking_invariant(ID,State)),nl,
    % print_bexpr(Invariant),nl,
   % statistics(runtime,_),
    time_out(b_interpreter:b_test_boolean_expression_for_ground_state(Invariant,[],State ),
            DebugTimeOut,TimeOutRes),
    clear_error_context,
    check_timeout(TimeOutRes,ID).
    %% , statistics(runtime,[_,DeltaT]), print(delta_time(inv,DeltaT)),nl.

check_timeout(time_out,ID) :- !,
         nl,print_message('**** TIMEOUT during invariant evaluation !'),nl,
         print_message(node_id(ID)),nl,
         assert_time_out_for_invariant(ID),
         fail.
check_timeout(_,_).

% b_test_boolean_expression_for_ground_state:
%     idea: check each conjunct individually; avoids multiplication of choice points between conjuncts
b_test_boolean_expression_for_ground_state(b(Expr,_,Infos),LS,S) :- !,
	b_test_boolean_expression_for_ground_state2(Expr,Infos,LS,S).
b_test_boolean_expression_for_ground_state2(conjunct(LHS,RHS),_,LocalState,State) :- !,
        b_test_boolean_expression_for_ground_state(LHS,LocalState,State),!,
        b_test_boolean_expression_for_ground_state(RHS,LocalState,State).
b_test_boolean_expression_for_ground_state2(BE,Infos,LS,S) :- 
 %   (preferences:get_preference(provide_trace_information,true)
 %     -> nl,print(' ====> '),print_bexpr_with_limit(BE,150),nl ; true),!,
% print_bexpr_with_limit(BE,150),nl, statistics(runtime,_),
    b_test_boolean_expression_wf(b(BE,pred,Infos),LS,S)
    % , statistics(runtime,[_,DeltaT]), print(delta_time(inv,DeltaT)),nl
     .

state_violates_invariant(I,const_and_vars(ID,S)) :- !,
  add_internal_error('State not expanded: ',state_violates_invariant(I,const_and_vars(ID,S))).
state_violates_invariant(ID,State) :- State \== root, State \= concrete_constants(_),
   \+(state_satisfies_invariant_aux(ID,State)).




state_satisfies_negation_of_invariant(const_and_vars(ID,S)) :- !,
  add_error(state_satisfies_negation_of_invariant,'State not expanded: ',const_and_vars(ID,S)).
state_satisfies_negation_of_invariant(State) :- State \== root, State \= concrete_constants(_),
   print_message(informational,checking_negation_of_invariant),
   b_get_invariant_from_machine(Invariant),
   set_error_context(checking_negation_of_invariant(State)),
   b_not_test_boolean_expression_wf(Invariant,[],State ),
   clear_error_context,
   print_message(informational,invariant_violated).

state_violates_assertions(ID,State) :-
    state_corresponds_to_initialised_b_machine(State,BState), !,
    \+(state_satisfies_assertions(ID,dynamic,BState)).
state_violates_assertions(ID,State) :-
    state_corresponds_to_set_up_constants(State,BState),
    \+(state_satisfies_assertions(ID,static,BState)).
state_satisfies_assertions(ID,Type,State) :-
   ( get_assertions_from_machine(Type,Ass) -> 
       set_error_context(checking_assertions),
       preferences:get_computed_preference(debug_time_out,DebugTimeOut),
       time_out(b_test_list_of_boolean_expression(Ass,[],State),
            DebugTimeOut,TimeOutRes),
       clear_error_context,
       (TimeOutRes = time_out 
       -> (nl,print_message('**** TIMEOUT during assertion checking !'),nl,
           print_message(node_id(ID)),nl,
           assert_time_out_for_assertions(ID)
          )
       ; true %debug_println(10,assertions_ok)
      )
   ; true
   ).

get_assertions_from_machine(Type,Assertions) :-
    ( Type=(dynamic) -> 
        (preferences:get_preference(use_po,true)
         -> b_get_unproven_dynamic_assertions_from_machine(Assertions)
         ;  b_get_dynamic_assertions_from_machine(Assertions)
        )
    ; Type=(static)  -> 
        (preferences:get_preference(use_po,true)
          -> b_get_unproven_static_assertions_from_machine(Assertions)
          ;  b_get_static_assertions_from_machine(Assertions)
        )
    ),
    Assertions = [_|_].


/* ----------------------------------*/
/*     b_det_test_boolean_expression     */
/* ----------------------------------*/
   
   % a tool to check for non-deterministic predicates
   
b_det_test_boolean_expression(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
	b_det_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao).

b_det_test_boolean_expression2(conjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,  
%%   print('check lhs of conjunct: '), translate:print_bexpr(LHS),nl,
        b_det_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
%%   print('check rhs of conjunct: '), translate:print_bexpr(RHS),nl,
        b_det_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
b_det_test_boolean_expression2(Pred,Infos,LS,S,WF,Ai,Ao) :-    
   debug:watch_det(500,b_interpreter:b_test_boolean_expression(b(Pred,pred,Infos),LS,S,WF,Ai,Ao)). 
   
   
/* ----------------------------------*/
/*     b_test_boolean_expression     */
/* ----------------------------------*/


:- type variable_id_list +--> list(type(variable_id)).
                                 
:- type boolean_expression +--> call(bsyntaxtree:check_if_typed_predicate).


b_test_list_of_boolean_expression(List,LS,S) :-
      init_wait_flags(WF),
      b_test_list_of_boolean_expression_wf(List,LS,S,WF),
      ground_wait_flags(WF).
  
b_test_list_of_boolean_expression_wf([],_,_,_WF).
b_test_list_of_boolean_expression_wf([B1|T],LS,S,WF) :-
   ~~pp_cll(b_test_boolean_expression(B1,LS,S,WF)),
   b_test_list_of_boolean_expression_wf(T,LS,S,WF).


:- use_module(b_interpreter_components).

    
:- assert_pre(b_interpreter:b_test_boolean_expression_wf(E,LS,S),
	      (bsyntaxtree:check_if_typed_predicate(E),type_check(LS,store),type_check(S,store) )).
:- assert_post(b_interpreter:b_test_boolean_expression_wf(_,_,_), true ).

%:- dynamic concount/1.
%inc_cc(N) :- ( concount(OldCount) -> retractall(concount(OldCount)), NewCount is OldCount + N; NewCount = N), assert(concount(NewCount)).




% --------------------- BOOLEAN EXPRESSIONS ----------------------------

:- use_module(b_interpreter_check).

b_test_boolean_expression_wf(E,LS,S) :-
      init_wait_flags(WF),
      b_test_boolean_expression(E,LS,S,WF),
      %(debug_mode(on) -> portray_waitflags(WF) ; true),
      %print('GROUNDING DET'),nl,trace,
      %ground_det_wait_flag(WF), nl,print('DET'),nl,
      %portray_waitflags(WF),nl,
      ground_wait_flags(WF).
b_test_boolean_expression(E,LS,S,WF) :-
      empty_avl(Ai), b_test_boolean_expression(E,LS,S,WF,Ai,_).
      
:- assert_pre(b_interpreter:b_test_boolean_expression(E,LS,S,WF),
	      (nonvar(E),bsyntaxtree:check_if_typed_predicate(E),type_check(LS,store),type_check(S,store),
		  type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_test_boolean_expression(_,_,_,WF), type_check(WF,wait_flag)).
      

%b_test_boolean_expression(Expr,_,_,_) :- print('test: '),translate:print_bexpr(Expr),nl,print(Expr),nl,fail. %%

:- if(debug:global_debug_flag).
b_test_boolean_expressiond(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
	b_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao).
b_test_boolean_expression(Expr,LS,S,WF,Ai,Ao) :- !,
	(waitflag0_is_set(WF)
	 -> debug:watch(300,b_interpreter:b_test_boolean_expressiond(Expr,LS,S,WF,Ai,Ao))
	 ;  debug:watch_det(300,b_interpreter:b_test_boolean_expressiond(Expr,LS,S,WF,Ai,Ao)) ).
:- else.
b_test_boolean_expression(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
	b_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao).
:- endif.
b_test_boolean_expression(E,LS,S,WF,Ai,Ao) :- 
  add_internal_error('Boolean expression not properly wrapped: ',E),
  b_test_boolean_expression2(E,[],LS,S,WF,Ai,Ao).


b_test_boolean_expression2(truth,_,_,_,_WF,Ai,Ao) :- !,Ai=Ao.
b_test_boolean_expression2(falsity,_,_,_,_WF,_Ai,_Ao) :- !,fail.
b_test_boolean_expression2(negation(BExpr),_,LocalState,State,WF,Ai,Ao) :- !,
        b_not_test_boolean_expression(BExpr,LocalState,State,WF,Ai,Ao).
%b_test_boolean_expression2(conjunct(LHS,RHS),[try_smt|Info],LocalState,State,WF,Ai,Ao) :- 
%       print('TRY SMT for Conjunct'),nl,
%       b_check_boolean_expression(b(conjunct(LHS,RHS),pred,[try_smt|Info]),LocalState,State,WF,PredRes,Ai,Ao),!,
%       print('SMT Check of Conjunct: '),nl, translate:print_bexpr(LHS),nl, translate:print_bexpr(RHS),nl,
%       PredRes=pred_true.  
b_test_boolean_expression2(conjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,  
%%   print('check lhs of conjunct: '), translate:print_bexpr(LHS),nl,
        b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
%%   print('check rhs of conjunct: '), translate:print_bexpr(RHS),nl,
        b_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
b_test_boolean_expression2(implication(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :-
    CLHS=LHS, 
    %b_compiler:b_compile(LHS,[],LocalState,State,CLHS), % this slows down Cruise benchmark from 3.75 s to 4.3
    !,
    b_test_implication(CLHS,RHS,LocalState,State,WF,Ai,Ao).
b_test_boolean_expression2(equivalence(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
    b_test_equivalence(LHS,RHS,LocalState,State,WF,Ai,Ao).
b_test_boolean_expression2(disjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
    % print(disjunct),nl, print_bexpr(LHS),nl, print_bexpr(RHS),nl,
   % TO DO: normalise disjunct first ?
   copy_wf_start(WF,CWF),
   if(b_check_boolean_expression(LHS,LocalState,State,CWF,PredRes,Ai,Aii),
      % we can check lhs
       (b_interpreter_check:b_check_boolean_expression0(pred_false,PredRes,RHS,LocalState,State,CWF,RR,Aii,Ao)  % TO DO <--------- REPLACE CUT
          -> /* we can also check rhs */
             b_interpreter_check:disjoin(PredRes,RR,pred_true,LHS,RHS,WF),
             copy_wf_finish(WF,CWF)
          ; Ao=Aii, % print(disjunct_rhs_cannot_be_checked),nl, print_bexpr(LHS),nl, print_bexpr(RHS),nl,
            copy_wf_finish(WF,CWF),
            imply_test_boolean_expression(PredRes,pred_false,RHS,LocalState,State,WF),
            %,print(predres(PredRes)),nl,nl,
            enumerate_bool(PredRes,WF)
         ),
      (Ai=Ao,  
       get_priority_of_boolean_expression2(disjunct(LHS,RHS),Prio),
       get_wait_flag(Prio,disjunct,WF,WF2), 
      % be sure that e.g. for (x=2 or x=3) both branches will be tested once deterministic propagation is over
      %print(prio(Prio)), print(' : '),print_bexpr(LHS), print(' or '), print_bexpr(RHS),nl,
	  b_test_disjunction(LHS,RHS,LocalState,State,WF,WF2,Ai)
	  )
    ).
b_test_boolean_expression2(exists(Parameters,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,Ai=Ao,
        b_test_exists(Parameters,RHS,Infos,LocalState,State,WF).
%b_test_boolean_expression2(forall(PARS,LHS,RHS),_Inf,LocalState,State,WF) :-
%   PARS = [b(identifier(R1),integer,_),b(identifier(R2),integer,_)],
%   LHS = b(conjunct(b(conjunct(MEM1,MEM2),pred,_I1),NEQ),pred,_I2),
%   NEQ = b(not_equal(b(identifier(RR1),integer,_I3),b(identifier(RR2),integer,_I4)),pred,_I5),
%   (RR1=R1, RR2=R2 ; RR1=R2, RR2=R1),
%   MEM1 = b(member(b(identifier(RRR1),integer,_),b(SET,set(integer),_)),pred,_),
%   MEM2 = b(member(b(identifier(RRR2),integer,_),b(SET,set(integer),_)),pred,_),
%   (RRR1=R1, RRR2=R2 ; RRR1=R2, RRR2=R1),
%   /* TO DO: check that RHS is symmetric ! */  /* TO DO : generalize symmetry breaking */
%   !, %   print(symmetry_breaking_forall(R1,R2)),nl,
%   GT = b(less(b(identifier(R1),integer,_I3),b(identifier(R2),integer,_I4)),pred,_I5),
%   NewLHS = b(conjunct(b(conjunct(MEM1,MEM2),pred,_I1),GT),pred,_I2),
%   b_test_boolean_expression2(forall(PARS,NewLHS,RHS),_Inf,LocalState,State,WF).
b_test_boolean_expression2(forall(Parameters,LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,Ai=Ao,
                             /* !ID.(LHS => RHS) */
       % print_message(forall(Parameters,LHS,RHS)),   
      %  get_wait_flag0(WF,WF0),  when(nonvar(WF0),
         b_for_all(Parameters,LHS,RHS,LocalState,State,WF). %).
b_test_boolean_expression2(let_predicate(Ids,AssignmentExprs,Pred),_,LocalState,State,WF,Ai,Ao) :- !,Ai=Ao,
	!,set_up_localstate_for_let(Ids,AssignmentExprs,LocalState,State,LetState,WF),
	b_test_boolean_expression(Pred,LetState,State,WF). % we could pass Ai in if Ids fresh
b_test_boolean_expression2(Pred,Info,LocalState,State,WF,Ai,Ao) :-
  % TO DO: we could put some of the operators into the Ai store (comparison operators)
   b_interpreter_check:register_predicate(Pred,Info,pred_true,Reused,Ai,Ao),
   %portray_avl(Ai),nl,
   (Reused==true -> true %print('REUSED: '), translate:print_bexpr(Pred),nl
    ; b_test_atomic_boolean_expression2(Pred,Info,LocalState,State,WF)).

b_test_implication(LHS,RHS,LocalState,State,WF,Ai,Ao) :- 
   copy_wf_start(WF,CWF),
   if(b_check_boolean_expression(LHS,LocalState,State,CWF,PredRes,Ai,Aii), 
  %% print('checking => '), print_bexpr(LHS), print(' '), print(PredRes),nl,
  %% The following improves constraint propagation; but can be tricky with undefinedness
	   if((var(PredRes),always_well_defined(RHS),
		   b_check_boolean_expression(RHS,LocalState,State,CWF,RHSPredRes,Aii,Ao)),
		  (/* also propagate from right to left */
		  % print('checking inverse => '), print_bexpr(RHS), print(' '), print(RHSPredRes),nl,
		   b_interpreter_check:imply_true(PredRes,RHSPredRes),
		   copy_wf_finish(WF,CWF)),
		 (Aii=Ao, 
		  copy_wf_finish(WF,CWF),
		  imply_test_boolean_expression(PredRes,pred_true,RHS,LocalState,State,WF))
	   ),
	   (Ai=Ao,
	    %nl,print('*** CANNOT CHECK BOOLEAN EXPRESSION: '), translate:print_bexpr(LHS),nl,
	    %add_error(cannot_check,'LHS cannot be checked',LHS),
		get_binary_choice_wait_flag(implication,WF,WF2),
		block_test_implication(LHS,RHS,LocalState,State,WF,WF2))
	).

% ------------
:- discontiguous b_test_atomic_boolean_expression2/5.
b_test_atomic_boolean_expression2(partition(LHS,RHS),_,LocalState,State,WF) :- !,
    b_compute_expression(LHS,LocalState,State,SV1,WF), %print(lhs(SV1)),nl,
    l_compute_expression(RHS,LocalState,State,SV2,WF), % RHS is Prolog list of expressions
   %%  print(partition_wf(SV1,SV2,WF)),nl, %%
    kernel_call_predicate(partition_wf(SV1,SV2,WF),kernel_objects,WF,partition(LHS,RHS)).
b_test_atomic_boolean_expression2(member(TElem,TSet),_,LocalState,State,WF) :- !,
	b_test_member_expression(TElem,TSet,LocalState,State,WF).
b_test_atomic_boolean_expression2(not_member(TElem,TSet),_,LocalState,State,WF) :- !,
	get_texpr_expr(TSet,Set),
	b_test_notmember_expression(Set,TElem,TSet,LocalState,State,WF).
b_test_atomic_boolean_expression2(freetype_case(Freetype,Case,Expr),_,LocalState,State,WF) :-
	!,equal_object(Value,freeval(Freetype,Case,_)),
	b_compute_expression(Expr,LocalState,State,Value,WF). % NOT COVERED (16)
b_test_atomic_boolean_expression2(finite(SetExpr),_,LocalState,State,WF) :-
	!, b_compute_expression(SetExpr,LocalState,State,Value,WF),
	kernel_call_predicate(is_finite_set_wf(Value,WF),kernel_objects,WF,finite(SetExpr)).
/* Extension for Kodkod */
b_test_atomic_boolean_expression2(kodkod(ProblemId,_),_,LocalState,State,WF) :-
	!,get_wait_flag(1,kodkod_call,WF,WF0),
    kodkod_request(ProblemId,pos,LocalState,State,WF0). % NOT COVERED (17)
b_test_atomic_boolean_expression2(Expression,_,LocalState,State,WF) :-
    preferences:preference(use_clpfd_solver,false),
	special_binary_predicate(Expression,Module:Kernel_predicate,Arg1,T1,Arg2,T2),!,
	 % print('Special call: '), print_bexpr(Expression),nl, %
	 %print(Kernel_predicate),print(' : '),print(Arg1), print(' '),print(Arg2),nl, %
	 (Arg1==list_of_expressions /* indicates Arg2 is list of expressions */
	  -> b_compute_expressions(Arg2,LocalState,State,ListSV2,WF),
	     KernelCall =.. [Kernel_predicate,ListSV2,WF]
      ;  b_compute_expression(Arg1,LocalState,State,SV1,WF),
         b_compute_expression(Arg2,LocalState,State,SV2,WF),
	     KernelCall =.. [Kernel_predicate,SV1,T1,SV2,T2,WF]
	), %print_term_summary(KernelCall),
    kernel_call_predicate(KernelCall,Module,WF,Expression).
%b_test_atomic_boolean_expression2(equal(Arg1,Arg2),_,LocalState,State,WF) :- get_texpr_type(Arg1,integer),!,
%  b_compute_expression(Arg1,LocalState,State,ExprRes,WF),
%  b_compute_expression(Arg2,LocalState,State,ExprRes,WF). /* avoids one extra variable compared to code below + could fail earlier ; but problem as b_compute_expression still uses = in some places */   

:- include(b_arithmetic_expressions).
%The following clause improves Nqueens performance by about 50 %; by avoiding intermediate variables
% TO DO: do the same for >, <, >= ...
b_test_atomic_boolean_expression2(equal(Arg1,Arg2),Info,LocalState,State,WF) :- !,
   b_test_atomic_equal_boolean_expression(Arg1,Arg2,Info,LocalState,State,WF).
b_test_atomic_boolean_expression2(not_equal(Arg1,Arg2),_,LocalState,State,WF) :-
   get_texpr_type(Arg1,boolean),!,  % special treatment : not_equal_object has no type information
   b_compute_expression(Arg1,LocalState,State,B1,WF),
   b_compute_expression(Arg2,LocalState,State,B2,WF),
   %print(not_eq(B1,B2)),nl,
   bool_pred:negate(B1,B2).
b_test_atomic_boolean_expression2(not_equal(Arg1,Arg2),_,LocalState,State,WF) :-
   get_texpr_type(Arg1,integer),preferences:preference(use_clpfd_solver,true),!,
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   %print(not_eq(CLPFD_Expr1,CLPFD_Expr2)),nl,
   clpfd_interface:clpfd_neq_expr(CLPFD_Expr1,CLPFD_Expr2).
b_test_atomic_boolean_expression2(less_equal(Arg1,Arg2),_,LocalState,State,WF) :-
   get_texpr_type(Arg1,integer),preferences:preference(use_clpfd_solver,true),!,
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   %print(less_eq(CLPFD_Expr1,CLPFD_Expr2)),nl,
   clpfd_interface:clpfd_leq_expr(CLPFD_Expr1,CLPFD_Expr2).
b_test_atomic_boolean_expression2(external_pred_call(FunName,Args),Info,LocalState,State,WF) :- 
      !,  
      b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF),
      % print('EXTERNAL PRED CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_predicate(FunName,EvaluatedArgs,pred_true,Info,WF).
b_test_atomic_boolean_expression2(Expression,_,LocalState,State,WF) :-
        functor(Expression,BOP,2),
        kernel_mappings:binary_boolean_operator(BOP,Module:Kernel_predicate,WFEnabled),!,
     %%  print(bop(BOP)),nl, %%
        arg(1,Expression,Arg1), arg(2,Expression,Arg2),
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
      %% print_term_summary(bop_arg1(BOP,Arg1,State,SV1)),nl, %%
        b_compute_expression(Arg2,LocalState,State,SV2,WF),
      %% print_term_summary(bop_arg2(BOP,Arg2,State,SV2)),nl, %%
	    ( WFEnabled==yes ->
	       KernelCall =.. [Kernel_predicate,SV1,SV2,WF]
	     ; KernelCall =.. [Kernel_predicate,SV1,SV2]),
       %% print(binary_bool_kernel_call(KernelCall)),nl,  %%
        kernel_call_predicate(KernelCall,Module,WF,Expression).
b_test_atomic_boolean_expression2(E,Info,_LS,_S,_WF) :-
    add_internal_error('Uncovered boolean expression: ',b(E,pred,Info)),
    print_functor(E),nl, 
    %print(local_state(LS)),nl, print(state(S)),nl,
    fail.
    
:- use_module(specfile,[animation_minor_mode/1]).
:- use_module(b_machine_hierarchy,[abstract_constant/2,concrete_constant/2]).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
   get_texpr_type(Arg1,integer),preferences:preference(use_clpfd_solver,true),!,
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   %print(eq(CLPFD_Expr1,CLPFD_Expr2)),nl,
   clpfd_interface:clpfd_eq_expr(CLPFD_Expr1,CLPFD_Expr2).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,Info,LocalState,State,WF) :-
    member(lambda_result,Info),
    \+ simple_expression(Arg2),
    !, % this is an equality for the result of a lambda abstraction; only evaluate at the end
    b_compute_expression(Arg1,LocalState,State,SV1,WF), % this is the lambda result identifier _lambda_result_
    %get_enumeration_finished_wait_flag(WF,EWF),
    %get_wait_flag(10000,lambda_result,WF,EWF),
    % compute Arg2 in wf0 phase: avoid computing this stuff over again
    kernel_waitflags:copy_wf0e_wait_flags(WF,WF2), b_compute_expression(Arg2,LocalState,State,SV2,WF2),
    get_last_wait_flag(lambda_result,WF,EWF),
    b_lambda_result_equal_boolean_expression(SV1,SV2,WF,EWF,WF2).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,Info,LocalState,State,WF) :-
     LocalState = [], % only check if we are at the outer level (not inside a quantifier,...)
     get_texpr_type(Arg1,Type1), is_set_type(Type1,SetType1), % check that we have a relation; for pure sets there is no easy way to construct a well-founded recursion; 
     % previously we required SetType1 = couple(_,_)
     get_texpr_id(Arg1,ID), % check if ABSTRACT_CONSTANT
     %print(try_symbolic1(ID,SetType1)),nl,
     (abstract_constant(ID,_) -> ABSTRACT=true
      ; animation_minor_mode(eventb),concrete_constant(ID,_),ABSTRACT=false ),  % maybe z or tla minor_mode: disable this check ??
     kernel_objects:max_cardinality(Type1,MaxCard1),
     (number(MaxCard1) -> MaxCard1>100 ; true), % check that type is large enough to warrant recursive treatment
     % Should we maybe even restrict it to infinite types ??
     % in future: we may do size change analysis, or something like that to check that we have properly defined
     % recursive function
     %print(try_symbolic(ID,ABSTRACT)),nl,
     (member(prob_annotation(recursive(ID)),Info)
       -> SetType1=couple(_,_), % disallow recursive set definitions; there is no way they can be well-founded ??
          print('RECURSIVE equation defining constant: '),  print(ID), print(' = '),
          StringID = b(string(ID),string,[]),
          RecExpr = external_function_call('RECx',[StringID]),
          TRec = b(RecExpr,Type1,[]),
          replace_id_by_expr(Arg2,ID,TRec,RHS),  % replace ID by RECx("ID"): allows compiling closure before ID fully known
          mark_bexpr_as_symbolic(RHS,SymRHS),
          SymArg2 = b(external_function_call('REC_LET',[StringID,SymRHS]),Type1,[]),
          ANN=['RECURSIVE'],
          translate:print_bexpr(SymArg2),nl
       ; ABSTRACT=true -> /* in animation_minor_mode(eventb) we only detect recursive functions and keep them symbolic */
          print('Equation defining abstract constant: '), print(ID),nl,
          mark_bexpr_as_symbolic(Arg2,SymArg2), ANN=[]),
    !,
     b_compute_expression(Arg1,LocalState,State,SV1,WF),
     b_compute_expression(SymArg2,LocalState,State,SV2,WF),
     custom_explicit_sets:mark_closure(SV2,['SYMBOLIC'|ANN],SV3), % necessary ?
     kernel_call_predicate_equal_object_optimized(SV1,SV3,WF).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
    b_compute_expression(Arg1,LocalState,State,SV1,WF),
    b_compute_expression(Arg2,LocalState,State,SV2,WF),
    kernel_call_predicate_equal_object_optimized(SV1,SV2,WF).

simple_expression(b(A,_,_)) :- simple_expression_aux(A).
simple_expression_aux(boolean_true).
simple_expression_aux(boolean_false).
simple_expression_aux(empty_set).
simple_expression_aux(empty_sequence).
simple_expression_aux(integer(_)).
simple_expression_aux(max_int).
simple_expression_aux(min_int).
simple_expression_aux(string(_)).
simple_expression_aux(value(_)).

:- block b_lambda_result_equal_boolean_expression(-,?,?, ?,-).
% only compute lambda result when finished finding domain element
b_lambda_result_equal_boolean_expression(SV1,SV2,WF,_EWF,WF2) :-
    %print(' LAMBDA : '), print((SV1,EWF)),nl, translate:print_bexpr(Arg2),nl,
    /* (var(EWF) -> b_compute_expression(Arg2,LocalState,State,SV2,WF)
      ; b_compute_expression(Arg2,LocalState,State,SV2)), */
    kernel_waitflags:ground_inner_wait_flags(WF2),        
    kernel_call_predicate_equal_object_optimized(SV1,SV2,WF).


% TO DO: add avl_set+closure expansion
%b_test_member_expression(Expr,El,_Set,_LocalState,_State,_WF) :-
%        print('test member: '),print_bexpr(Expr),nl,fail.
b_test_member_expression(TElem,b(Set,Type,Info),LocalState,State,WF) :-
        b_test_member_expression2(Set,Type,Info,TElem,LocalState,State,WF).

%:- use_module(profiler).
b_test_member_expression2(SymbolicSetConstant,_Type,_Info,El,LocalState,State,WF) :-
        %profiler:add_profile_hit(SymbolicSetConstant),
        cst_in_boolean_type(SymbolicSetConstant,Module:Kernel_predicate),!, 
        b_compute_expression(El,LocalState,State,ElValue,WF),
	    functor(KernelCall,Kernel_predicate,2),
	    arg(1,KernelCall,ElValue),
	    arg(2,KernelCall,WF),
     %  print_message(cst_in_boolean_type_call(KernelCall)),
        kernel_call_predicate(KernelCall,Module,WF,SymbolicSetConstant).  % should be member(... SymbolicSetConstant)
b_test_member_expression2(Expression,_Type,_Info,El,LocalState,State,WF) :-
        functor(Expression,BOP,2),
        binary_in_boolean_type(BOP,Module:Kernel_predicate),!,
        arg(1,Expression,Arg1),arg(2,Expression,Arg2),
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
        b_compute_expression(Arg2,LocalState,State,SV2,WF),
        b_compute_expression(El,LocalState,State,ElValue,WF),
        binary_in_kernel_call(Module,Kernel_predicate,ElValue,SV1,SV2,WF,Expression).
b_test_member_expression2(Expression,_Type,_Info,El,LocalState,State,WF) :-
        functor(Expression,UnOP,1),
        unary_in_boolean_type(UnOP,Module:Kernel_predicate),!, 
        arg(1,Expression,Arg1),
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
        b_compute_expression(El,LocalState,State,ElValue,WF),
        unary_in_kernel_call(Module,Kernel_predicate,ElValue,SV1,WF,Expression).
b_test_member_expression2(Expression,Type,Info,El,LocalState,State,WF) :- 
        (Expression=comprehension_set(Par,Cond) % TO DO: --> replace by functor for partial evaluation
          -> b_generate_closure_if_necessary(Par,Cond,LocalState,State,SetValue) /* will not expand closure */
          ; b_compute_expression2(Expression,Type,Info,LocalState,State,SetValue,WF),
            SetValue \== [] % if we have the empty set: check_element will always fail
            % we could disable this check if find_abort_values=true
        ),
        b_compute_expression(El,LocalState,State,Element,WF),
        kernel_call_predicate_check_element_of_wf(Element,SetValue,WF).
	     
b_test_notmember_expression(SymbolicSetCst,El,_Set,LocalState,State,WF) :- 
        kernel_mappings:cst_not_in_boolean_type(SymbolicSetCst,Module:Kernel_predicate),!,
        b_compute_expression(El,LocalState,State,ElValue,WF),
	    functor(KernelCall,Kernel_predicate,1),
	    arg(1,KernelCall,ElValue),
     %  debug_print(cst_not_in_boolean_type(9,KernelCall)),
        kernel_call_predicate(KernelCall,Module,WF,SymbolicSetCst).
b_test_notmember_expression(Expression,El,_Set,LocalState,State,WF) :-
        functor(Expression,BOP,2),
        kernel_mappings:binary_not_in_boolean_type(BOP,Module:Kernel_predicate,HasWF),!,
        arg(1,Expression,Arg1),arg(2,Expression,Arg2),
       ~~pp_cll(b_compute_expression(Arg1,LocalState,State,SV1,WF)),
       ~~pp_cll(b_compute_expression(Arg2,LocalState,State,SV2,WF)),
       ~~pp_cll(b_compute_expression(El,LocalState,State,ElValue,WF)),
        (HasWF=true
          -> KernelCall =.. [Kernel_predicate,ElValue,SV1,SV2,WF]
          ;  KernelCall =.. [Kernel_predicate,ElValue,SV1,SV2]
        ),
     %  debug_println(9,binary_not_in_bool_kernel_call_predicate(KernelCall)),
        kernel_call_predicate(KernelCall,Module,WF,Expression). 
b_test_notmember_expression(Expression,El,_Set,LocalState,State,WF) :- 
        functor(Expression,UnOP,1),
        kernel_mappings:unary_not_in_boolean_type(UnOP,Module:Kernel_predicate,WFEnabled),!,
        arg(1,Expression,Arg1),
        ~~pp_cll(b_compute_expression(Arg1,LocalState,State,SV1,WF)),
        ~~pp_cll(b_compute_expression(El,LocalState,State,ElValue,WF)),
        (WFEnabled==yes
          -> KernelCall =.. [Kernel_predicate,ElValue,SV1,WF]
          ;  KernelCall =.. [Kernel_predicate,ElValue,SV1]
        ),
      % debug_println(9,unary_not_in_bool_kernel_call(KernelCall)),
        kernel_call_predicate(KernelCall,Module,WF,Expression).
b_test_notmember_expression(Expr,Elem,Set,LocalState,State,WF) :-
	(Expr=comprehension_set(Par,Cond)
          -> b_generate_closure_if_necessary(Par,Cond,LocalState,State,SetValue) /* will not expand closure */
          ; b_compute_expression(Set,LocalState,State,SetValue,WF)
    ),
    (SetValue == [] 
      -> true /* nothing can be member of empty set */
         % we could disable this check if find_abort_values=true
      ;  b_compute_expression(Elem,LocalState,State,ElemValue,WF),
	     kernel_call_predicate(not_element_of_wf(ElemValue,SetValue,WF),kernel_objects,WF,Expr)
	).
		      
:- block block_test_implication(?,?,?,?,?,-).
block_test_implication(LHS,RHS,LocalState,State,WF,_WF2) :-
    %% print(impl(LHS)),nl, %%
	b_test_boolean_expression(LHS,LocalState,State,WF),
	b_test_boolean_expression(RHS,LocalState,State,WF).
block_test_implication(LHS,_RHS,LocalState,State,WF,_WF2) :- %% print(impl_not(LHS)),nl, %%
        b_not_test_boolean_expression(LHS,LocalState,State,WF).


b_test_equivalence(LHS,RHS,LocalState,State,WF,Ai,Ao) :-
  copy_wf_start(WF,CWF),
  if(b_check_boolean_expression(LHS,LocalState,State,CWF,PredRes,Ai,Aii),
      (b_check_test_equivalence(PredRes,LHS,RHS,LocalState,State,CWF,Aii,Ao),
       copy_wf_finish(WF,CWF)),
      (% check LHS fails, We could check if RHS can be checked with b_check_boolean_expression
       Ai=Ao, get_binary_choice_wait_flag(equivalence,WF,WF2),
       b_enum_test_equivalence(LHS,RHS,LocalState,State,WF,WF2))).
b_check_test_equivalence(PredRes,LHS,RHS,LocalState,State,WF,Aii,Ao) :- 
  % print('checking <=> '), print_bexpr(LHS), print(' '), print(PredRes),nl,
  ((var(PredRes),b_check_boolean_expression(RHS,LocalState,State,WF,PredRes2,Aii,Ao))  % TO DO <---- REPLACE CUT
    -> /* also propagate from right to left */
      %print('checking inverse <=> '), print_bexpr(RHS), print(' '), print(PredRes2),nl,
     equiv_bidrectional_test_boolean_expression(PredRes,PredRes2,LHS,RHS,LocalState,State,WF)
   ;  Aii=Ao,
      equiv_test_boolean_expression(PredRes,pred_true,RHS,LocalState,State,WF)
  ). % should we also add a binary choice flag ?? what about a=2 <=> b/=3 ??

:- block b_enum_test_equivalence(?,?,?,?,?,-).
b_enum_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2) :-
	b_test_boolean_expression(LHS,LocalState,State,WF),
	b_test_boolean_expression(RHS,LocalState,State,WF).
b_enum_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2) :-
    b_not_test_boolean_expression(LHS,LocalState,State,WF),
	b_not_test_boolean_expression(RHS,LocalState,State,WF).

%enumerate_bool(X,WF) :- !.
enumerate_bool(X,WF) :- (nonvar(X) -> true /* already set */
       ; get_wait_flag0(WF,WF0),
         enumerate_bool0(X,WF,WF0)).
         
:- block enumerate_bool0(?,?,-).
enumerate_bool0(X,WF,_) :- (nonvar(X) -> true /* already set */
      ;  %print(enum_bool(X)), translate:print_bexpr(LHS),nl,
         % get_binary_choice_wait_flag(enumerate_bool,WF,WF2),
         get_wait_flag(15,enumerate_bool,WF,WF2), % what priority should we use ?? give time for other enumerators to decide the lhs of the disjunct; actions_cbtc is slower to setup with low priority
         enumerate_bool_aux(X,WF2)).
:- block enumerate_bool_aux(-,-).
%enumerate_bool_aux(P,WF) :- var(P),print(forcing_bool(P,WF)),nl,fail. %
enumerate_bool_aux(pred_true,_).
enumerate_bool_aux(pred_false,_).

:- block b_test_disjunction(?,?,?,?,?,-,?).
b_test_disjunction(LHS,_RHS,LocalState,State,WF,_WF2,Ai) :- 
    %% print(disjunct_left),nl, print_bexpr(LHS), print(' or '), print_bexpr(_RHS),nl, %%
	b_test_boolean_expression(LHS,LocalState,State,WF,Ai,_).
b_test_disjunction(LHS,RHS,LocalState,State,WF,_WF2,Ai) :- 
    %% print(disjunct_right),nl, print_bexpr(RHS),nl, %%
	b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).

:- block b_not_test_conjunction(?,?,?,?,?,-,?).
b_not_test_conjunction(LHS,_RHS,LocalState,State,WF,_WF2,Ai) :-
    %%  print(' making_lhs_conjunct_false '), print_bexpr(LHS),nl, %% portray_waitflags(WF),
	b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,_).
b_not_test_conjunction(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
    %%  print(' making_lhs_conjunct_true '), print_bexpr(LHS), print(' ::: '), print_bexpr(RHS),nl, %%
	b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).

:- block b_not_test_equivalence(?,?,?,?,?,-,?).
b_not_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
	b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).
b_not_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
	b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).
				   


/* --------------------------------------*/
/*     b_not_test_boolean_expression     */
/* --------------------------------------*/

b_not_test_list_of_boolean_expression([],_,_,_WF). % :- nl, print(done_not),nl,nl,flush_output.
b_not_test_list_of_boolean_expression([B1|T],LS,S,WF) :-
   %print('not '),translate:print_bexpr(B1),nl,flush_output,
   ~~pp_cll(b_not_test_boolean_expression(B1,LS,S,WF)),
   b_not_test_list_of_boolean_expression(T,LS,S,WF).
   
:- assert_pre(b_interpreter:b_not_test_boolean_expression_wf(E,LS,S), 
                  (bsyntaxtree:check_if_typed_predicate(E),type_check(LS,store),type_check(S,store))).
:- assert_post(b_interpreter:b_not_test_boolean_expression_wf(_,_,_), true).


b_not_test_boolean_expression_wf(E,LS,S) :-
      init_wait_flags(WF),b_not_test_boolean_expression(E,LS,S,WF),
      ground_wait_flags(WF).
      
b_not_test_boolean_expression(E,LS,S,WF) :-
      empty_avl(Ai), b_not_test_boolean_expression(E,LS,S,WF,Ai,_).

:- assert_pre(b_interpreter:b_not_test_boolean_expression(E,LS,S,WF), 
                  (type_check(E,boolean_expression),type_check(LS,store),type_check(S,store), type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_not_test_boolean_expression(_,_,_,_), true).

:- if(debug:global_debug_flag).
b_not_test_boolean_expression(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
	(waitflag0_is_set(WF)
	 -> debug:watch(400,b_interpreter:b_not_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao))
	 ;  debug:watch_det(400,b_interpreter:b_not_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao)) ).
:- else.
b_not_test_boolean_expression(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
	b_not_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao).
:- endif.
b_not_test_boolean_expression(E,LS,S,WF,Ai,Ao) :- 
  add_internal_error('Boolean (not) expression not properly wrapped: ',E),
  b_not_test_boolean_expression2(E,[],LS,S,WF,Ai,Ao).

b_not_test_boolean_expression2(truth,_,_,_,_WF,_Ai,_Ao) :- !,fail.
b_not_test_boolean_expression2(falsity,_,_,_,_WF,Ai,Ao) :- !,Ai=Ao. % NOT COVERED (2)
b_not_test_boolean_expression2(negation(BExpr),_,LocalState,State,WF,Ai,Ao) :- !,
        b_test_boolean_expression(BExpr,LocalState,State,WF,Ai,Ao).
/*  TO DO comment in after norm_conjunct issue addressed
b_not_test_boolean_expression2(conjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
   % print(not_conjunct),nl, print_bexpr(LHS),nl, print_bexpr(RHS),nl,
   % TO DO: construct construct_norm_conjunct(RHS,NC)
   (b_check_boolean_expression(LHS,LocalState,State,WF,PredRes,Ai,Aii)
      % we can check lhs
     ->  (b_interpreter_check:b_check_boolean_expression0(pred_true,PredRes,RHS,LocalState,State,WF,RR,Aii,Ao)
          -> % we can also check rhs 
             b_interpreter_check:conjoin(PredRes,RR,pred_false,LHS,RHS,WF)
          ; Ao=Aii, 
            imply_test_boolean_expression(PredRes,pred_true,b(negation(RHS),pred,[]),LocalState,State,WF)
         )
     ; % TO DO: check if RHS can be checked and then use imply_test in other order ??? WD issue ??
       Ai=Ao,
       %get_priority_of_boolean_expression2(disjunct(LHS,RHS),Prio),
       %get_wait_flag(Prio,not_conjunct,WF,WF2),
       get_binary_choice_wait_flag(not_conjunct,WF,WF2),
      % be sure that e.g. for (x=2 or x=3) both branches will be tested once deterministic propagation is over
      %print(prio(Prio)), print(' : '),print_bexpr(LHS), print(' or '), print_bexpr(RHS),nl,
	  b_not_test_conjunction(LHS,RHS,LocalState,State,WF,WF2,Ai)
    ).
    */
b_not_test_boolean_expression2(conjunct(LHS,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,
    b(conjunct(LHS,RHS),pred,Infos) = Conj,
    copy_wf_start(WF,CWF),
    if(b_check_boolean_expression(Conj,LocalState,State,CWF,PredRes,Ai,Ao),
        (PredRes=pred_false,
         copy_wf_finish(WF,CWF) %,  print(' check not conjunct: '), print_bexpr(Conj),nl,
         ),
        (Ai=Ao,
	     get_binary_choice_wait_flag(not_conjunct,WF,WF2),
	     %print(not_conjunct),nl, translate:print_bexpr(LHS),nl,nl,
	     % TO DO: we could check if LHS or RHS are registered in Ai as pred_true/pred_false ??
	     b_not_test_conjunction(LHS,RHS,LocalState,State,WF,WF2,Ai))
      ).
b_not_test_boolean_expression2(implication(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
    b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
b_not_test_boolean_expression2(equivalence(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,Ai=Ao,
    get_binary_choice_wait_flag(not_equivalence,WF,WF2),
	b_not_test_equivalence(LHS,RHS,LocalState,State,WF,WF2,Ai).
b_not_test_boolean_expression2(disjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
        b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
        b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
b_not_test_boolean_expression2(Expr,Info,LocalState,State,WF,Ai,Ao) :-
  % TO DO: we could put some of the operators into the Ai store (comparison operators)
   b_not_test_atomic_boolean_expression2(Expr,Info,LocalState,State,WF,Ai,Ao).

% Other boolean expressions, not involving boolean connectives
b_not_test_atomic_boolean_expression2(exists(Parameters,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,
                             /* #ID.(RHS) */
        Ai=Ao,
        b_not_test_exists(Parameters,RHS,Infos,LocalState,State,WF). 
b_not_test_atomic_boolean_expression2(forall(Parameters,LHS,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,
                             /* !ID.(LHS => RHS) */
	Ai=Ao,create_texpr(negation(RHS),pred,[],Negation),
	conjunct_predicates([LHS,Negation],Condition),
	%print('neg: '),print_bexpr(Condition),nl,
        b_test_exists(Parameters,Condition,Infos,LocalState,State,WF).
/* TODO(DP, 2.8.2008): Extension for Z: a let-statement as predicate */
b_not_test_atomic_boolean_expression2(let_predicate(Ids,AssignmentExprs,Pred),_,LocalState,State,WF,Ai,Ao) :-
	!,Ai=Ao,set_up_localstate_for_let(Ids,AssignmentExprs,LocalState,State,LetState,WF),
	b_not_test_boolean_expression(Pred,LetState,State,WF).  % NOT COVERED (10)
/* TODO(DP, 2.8.2008): */
b_not_test_atomic_boolean_expression2(freetype_case(Freetype,Case,Expr),_,LocalState,State,WF,Ai,Ao) :-
	!,Ai=Ao,b_compute_expression(Expr,LocalState,State,freeval(Freetype,RealCase,_),WF),
	dif(RealCase,Case).  % NOT COVERED (11)
b_not_test_atomic_boolean_expression2(finite(SetExpr),_,LocalState,State,WF,Ai,Ao) :-
	!, Ai=Ao,b_compute_expression(SetExpr,LocalState,State,Value,WF),
	kernel_call_predicate(is_infinite_set_wf(Value,WF),kernel_objects,WF,finite(SetExpr)).
/* Extension for Kodkod */
b_not_test_atomic_boolean_expression2(kodkod(ProblemId,_),_,LocalState,State,WF,Ai,Ao) :-
	!,Ai=Ao,get_wait_flag(1,kodkod_negative_call,WF,WF0),
    kodkod_request(ProblemId,neg,LocalState,State,WF0). % NOT COVERED (12)
b_not_test_atomic_boolean_expression2(partition(LHS,RHS),Info,LocalState,State,WF,Ai,Ao) :- !,
   b_interpreter_check:register_predicate(partition(LHS,RHS),Info,pred_false,Reused,Ai,Ao),
   (Reused==true -> true
    ;  %print(not_partition(LHS,RHS)),nl,
       b_compute_expression(LHS,LocalState,State,SV1,WF), %print(lhs(SV1)),nl,
       l_compute_expression(RHS,LocalState,State,SV2,WF), % RHS is Prolog list of expressions
       %print(not_partition_wf(SV1,SV2,WF)),nl,
       kernel_call_predicate(not_partition_wf(SV1,SV2,WF),kernel_objects,WF,partition(LHS,RHS))
    ).
b_not_test_atomic_boolean_expression2(external_pred_call(FunName,Args),Info,LocalState,State,WF,Ai,Ao) :- 
      !,  Ai=Ao,
      b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF),
      % print('EXTERNAL PRED CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_predicate(FunName,EvaluatedArgs,pred_false,Info,WF).
b_not_test_atomic_boolean_expression2(Expression,Info,LocalState,State,WF,Ai,Ao) :-
        functor(Expression,BOP,2),
        kernel_mappings:negate_binary_boolean_operator(BOP,NBOP),
        arg(1,Expression,Arg1),arg(2,Expression,Arg2),!,
        NExpression =.. [NBOP,Arg1,Arg2],
      %% print_message(negating(BOP,NBOP)), %%
	   %%create_texpr(NExpression,pred,[generated(negation)],TNExpression),
        b_test_boolean_expression2(NExpression,Info,LocalState,State,WF,Ai,Ao).
b_not_test_atomic_boolean_expression2(E,_,_,_,_WF,_Ai,_Ao) :-
    add_internal_error('Uncovered boolean expression (not): ',E),
    print_functor(E),nl,fail.




/* -----------------------------*/
/*     b_compute_expression     */
/* -----------------------------*/

        

:- assert_must_succeed((b_interpreter:b_compute_expression(b(pow_subset(
                      b(identifier(nm),set(integer),[])),set(set(integer)),[]),
                      [], [bind(nm,[])],R,_WF),
                             nonvar(R), custom_explicit_sets:expand_custom_set_to_list(R,[[]]))).
:- assert_must_succeed((b_interpreter:b_compute_expression(b(pow_subset(
                      b(identifier(nm),set(boolean),[])),set(set(boolean)),[]),[],
                                        [bind(nm,[pred_false /* bool_false */])],R,_WF),
                             nonvar(R), custom_explicit_sets:expand_custom_set_to_list(R,ER),
                             kernel_objects:equal_object(ER,[[pred_false /* bool_false */],[]]))).

:- assert_pre(b_interpreter:b_compute_expression(Exp,LS,State,_Val,WF),
                (nonvar(Exp),
		  bsyntaxtree:check_if_typed_expression(Exp),
                  type_check(LS,store),type_check(State,store), type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_compute_expression(_E,_LS,_State,Val,_WF), 
                 (b_interpreter:value_type(Val), type_check(_WF,wait_flag))). %, nonvar(Val)

b_compute_expression(E,LS,S,R) :-
      %add_message(b_compute_expression,'Called Version wo WFLAGS: ', b_compute_expression(E,LS,S,R)),
      init_wait_flags(WF),
      b_compute_expression(E,LS,S,R,WF),
      ground_wait_flags(WF).

l_compute_expression([],_LS,_S,[],_WF).
l_compute_expression([H|T],LS,S,[CH|CT],WF) :-
   b_compute_expression(H,LS,S,CH,WF),
   l_compute_expression(T,LS,S,CT,WF).

%b_compute_expression(Expr,_,_,_,_) :-
%      print('expr: '),translate:print_bexpr(X),nl,fail.
:- if(debug:global_debug_flag).
b_compute_expressiond(b(Expr,Type,Info),LS,S,R,WF) :-  !,
    ((ground(Type), Type  \== pred, Type \== subst) -> true
     ; add_error(b_compute_expression,'Expression has illegal type: ',b(Expr,Type,Info))),
     check_value(R,b(Expr,Type,Info)),
     b_compute_expression2(Expr,Type,Info,LS,S,R,WF).
:- block check_value(-,?).
check_value(pred_true,_) :- !. check_value(pred_false,_) :- !.
check_value(string(_),_) :- !.
check_value([],_) :- !.
check_value([H|T],E) :- !, check_value(H,E), check_value(T,E).
check_value((A,B),E) :- !, check_value(A,E), check_value(B,E).
check_value(fd(N,T),E) :- !, (nonvar(T) -> check_fd(N,T,E) ; check_err(fd(N,T),E)).
check_value(global_set(GS),E) :- !,(nonvar(GS) -> true ; check_err(global_set(GS),E)).
check_value(avl_set(AS),E) :- !,((nonvar(AS),AS=node(_,_,_,_,_)) -> true ; check_err(global_set(AS),E)).
check_value(int(X),E) :- !,check_int(X,E).
check_value(closure(P,T,B),E) :- !,((nonvar(P),nonvar(T),nonvar(B)) -> true ; check_err(closure(P,T,B),E)).
check_value(X,E) :- check_err(uncovered(X),E).
% TO DO: add records + more thorough checking of global_set and closures and avl_sets
:- block check_int(-,?).
check_int(X,E) :- (number(X) -> true ; check_err(int(X),E)).
:- block check_fd(-,?,?).
check_fd(X,T,E) :- ((number(X),b_fd_type(T,Low,Up),X>=Low,X=<Up) -> true ; check_err(fd(X,T),E)).

check_err(V,E) :- add_internal_error('Illegal value: ',V:E).
b_compute_expression(E,LS,S,R,WF) :- !,
    (waitflag0_is_set(WF)
	 -> debug:watch(100,b_interpreter:b_compute_expressiond(E,LS,S,R,WF))
	 ;  debug:watch_det(100,b_interpreter:b_compute_expressiond(E,LS,S,R,WF))).
:- else.
b_compute_expression(b(Expr,Type,Info),LS,S,R,WF) :-  !,
    ((ground(Type), Type  \== pred, Type \== subst) -> true
     ; add_internal_error('Expression has illegal type: ',b(Expr,Type,Info))),
     %% check_value(R,b(Expr,Type,Info)), %% comment in to check values at runtime <---
     b_compute_expression2(Expr,Type,Info,LS,S,R,WF).
:- endif.
b_compute_expression(Expr,LS,S,R,WF) :-
  add_internal_error('Expression not properly wrapped: ',Expr),
  b_compute_expression2(Expr,unknown,[],LS,S,R,WF).



:- use_module(b_global_sets,[b_type2global_set/2]).
b_compute_expression2(value(Val),_T,_I,_LS,_S,V,_WF) :- !,Val=V.
b_compute_expression2(boolean_true,_T,_I,_LState,_State,Res,_WF) :- !, Res = pred_true /* bool_true */.
b_compute_expression2(boolean_false,_T,_I,_LState,_State,Res,_WF) :- !, Res = pred_false /* bool_false */.
b_compute_expression2(integer_set(NSET),_T,_I,_LState,_State,Res,_WF) :- !, Res = global_set(NSET).
b_compute_expression2(bool_set,_T,_I,_LState,_State,Res,_WF) :- !, % b_type2global_set(bool,Res).
 	Res = [pred_true /* bool_true */,pred_false /* bool_false */]. % changing the order breaks Bosch Deadlock v9

b_compute_expression2(string_set,_T,_I,_LState,_State,Res,_WF) :- !, Res = global_set('STRING').
	% was all_strings(Res).  %% NOT COVERED (6)
b_compute_expression2(string(S),_T,_I,_LState,_State,Res,_WF) :- !,
     % TODO(ML,26.8.2008): Check if we should use a different functor for syntax tree vs data 
     Res = string(S).
b_compute_expression2(empty_set,_T,_I,_LState,_State,Res,_WF) :- !, empty_set(Res).
b_compute_expression2(empty_sequence,_T,_I,_LState,_State,Res,_WF) :- !, empty_sequence(Res).
b_compute_expression2(event_b_identity,EType,_I,_LState,_State,Res,WF) :- 
 is_set_type(EType,couple(Type,Type)),!,
 (true %preferences:get_preference(convert_types_into_closures,true)
   ->
	    create_texpr(identifier('_zzzz_unary'),Type,[],TIdentifier1), % was [generated]
	    create_texpr(identifier('_zzzz_binary'),Type,[],TIdentifier2), % was [generated]
        create_texpr(equal(TIdentifier1,TIdentifier2),pred,[],TPred),
         construct_closure(['_zzzz_unary','_zzzz_binary'],[Type,Type],TPred,Res)
    ;    all_objects_of_type(Type,All),
         identity_relation_over_wf(All,Res,WF)
  )   %, print(constructed_eventb_identity(Res)),nl
  .
b_compute_expression2(integer(Val),_T,_I,_LocalState,_State,int(CVal),_WF) :- !, CVal = Val.
b_compute_expression2(max_int,_T,_I,_LS,_S,int(CVal),_WF) :-
	preferences:get_preference(maxint,CVal),!.
b_compute_expression2(min_int,_T,_I,_LS,_S,int(CVal),_WF) :-
	preferences:get_preference(minint,CVal),!.   %% NOT COVERED (13)
b_compute_expression2(identifier(Id),_Type,Info,LocalState,State,Value,_WF) :- !,
        lookup_value_in_store_and_global_sets(Id,LocalState,State,Value,Info).
b_compute_expression2(rec(Fields),_T,_I,LocalState,State,Record,WF) :- !,
        ~~mnf(l_compute_field(Fields,LocalState,State,FieldValues,WF)),
        construct_record(FieldValues,Record).
b_compute_expression2(record_field(RecEx,Name),_T,_I,LocalState,State,Value,WF) :- !,
        b_compute_expression(RecEx,LocalState,State,RecValue,WF), 
        access_record(RecValue,Name,Value).
b_compute_expression2(freetype_set(Id),_T,_I,_LState,_State,Val,_WF) :- !, Val=freetype(Id). %% NOT COVERED (17)
b_compute_expression2(function(Function,FunArg),Type,Info,LocalState,State,Res,WF) :-
       preferences:preference(find_abort_values,false), % the treatment below will not check, e.g., if the set extension is really a function
       is_extension_function(Function,FInfo,ExFun), !,
      /* a set extension applied to an argument: in some cases we can can first evaluate the argument,
        and then avoid having to compute the set_extension for all non-matching args */
         %  print(is_set_extension_function),nl,  print(set_extension(ExFun)),nl, %print(funarg(FunArg)),nl,
        b_compute_expression(FunArg,LocalState,State,ArgValue,WF),
         %  print(funval(ArgValue)),nl,
        % TO DO : should we only do this if ArgValue is known ?? otherwise the standard treatment is better ??
        %(ground(ArgValue) -> true ; true),
        ((member(contains_wd_condition,FInfo) ;
          preferences:preference(use_clpfd_solver,false) ; ground(ArgValue))
         -> %print(apply_function_set_extension(FInfo,ArgValue)),nl, 
            b_apply_function_set_extension(ExFun,ArgValue,LocalState,State,Res,WF,Function)
         ;  /* the standard treatment can infer information about the result,... in clpfd_mode */
            %print(non_ground_fun_arg(ArgValue,FInfo,LocalState,State,Res)),nl, translate:print_bexpr(b(function(Function,FunArg),Type,Info)),nl,
            b_compute_expression(Function,LocalState,State,FValue,WF),
            Span = span_predicate(b(function(Function,FunArg),Type,Info),LocalState,State),
            %kernel_call(bsets_clp:apply_to(FValue,ArgValue,Res,Span,WF),(FValue,ArgValue),WF,function(Function,FunArg))).
            kernel_call_apply_to(FValue,ArgValue,Res,Span,WF)).
b_compute_expression2(function(Function,Arg),Type,Info,LocalState,State,Value,WF) :- !,
     %  print_message(call_function(Function,Arg)), 
        b_compute_expression(Function,LocalState,State,FValue,WF),
     %  print_message(function(FValue)),
        b_compute_expression(Arg,LocalState,State,ArgValue,WF),
        Span = span_predicate(b(function(Function,Arg),Type,Info),LocalState,State),
        %kernel_call(bsets_clp:apply_to(FValue,ArgValue,Value,Span,WF),(FValue,ArgValue),WF,function(Function,Arg)).
        kernel_call_apply_to(FValue,ArgValue,Value,Span,WF).
    %   print_message(kernel_res(Value)).
b_compute_expression2(card(b(Arg1,Type,Info)),integer,_I,LocalState,State,Card,WF) :- !,
       b_compute_card(Arg1,Type,Info,LocalState,State,Card,WF).
b_compute_expression2(max(b(Arg1,Type,Info)),integer,_I,LocalState,State,Max,WF) :- !,
       b_compute_max(Arg1,Type,Info,LocalState,State,Max,WF).
b_compute_expression2(min(b(Arg1,Type,Info)),integer,_I,LocalState,State,Min,WF) :- !,
       b_compute_min(Arg1,Type,Info,LocalState,State,Min,WF).
b_compute_expression2(set_extension(Ex),Type,_I,LocalState,State,ValueSet,WF) :- !,
       % print_bexpr(b(set_extension(Ex),none,[])),nl,
        b_compute_expressions(Ex,LocalState,State,ExValue,WF),
        % print_term_summary(b_compute_expressions(Ex,LocalState,State,ExValue,WF)),
        kernel_call(b_interpreter:convert_list_of_expressions_into_set(ExValue,ValueSet,Type),ExValue,WF, set_extension(Ex)).
        %kernel_call_convert_list_of_expressions_into_set(ExValue,ValueSet,Type, WF).
%b_compute_expression('ListTermes'(ListOfEx),LocalState,State,ValueSet,WF) :- !,
%        /* convert list of expressions of CASE statement into  set  of elements */
%        b_compute_expression(ListOfEx,LocalState,State,Value,WF),
%        convert_list_of_expressions_into_set(Value,ValueSet).
b_compute_expression2(sequence_extension(Ex),_T,_I,LocalState,State,ValueSeq,WF) :- !,
        b_compute_expressions(Ex,LocalState,State,Value,WF),
        kernel_call(b_interpreter:convert_list_of_expressions_into_sequence(Value,ValueSeq),Value,WF,sequence_extension(Ex)).
b_compute_expression2(convert_bool(PRED),_T,_I,LocalState,State,Val,WF) :- !, /* B bool(.) operator */
        b_convert_bool(PRED,LocalState,State,WF,Val).
b_compute_expression2(couple(El1,El2),_T,_I,LocalState,State,(Val1,Val2),WF) :- !,
        b_compute_expression(El1,LocalState,State,Val1,WF),
        b_compute_expression(El2,LocalState,State,Val2,WF).
b_compute_expression2(comprehension_set(Parameters,Condition),_T,Info,LocalState,State,Result,WF) :- !,
        b_compute_comprehension_set(Parameters,Condition,Info,LocalState,State,Result,WF).
%b_compute_expression2(explicit_comprehension_set(Parameters,Condition),_T,_Info,LocalState,State,Result,WF) :- !,
%        b_compute_explicit_comprehension_set(Parameters,Condition,LocalState,State,Result,WF).
b_compute_expression2(general_sum(Ids,Condition,Expression),_T,_I,LocalState,State,SumResult,WF) :- !,
       b_general_sum_or_mul(Ids,Condition,Expression,LocalState,State,SumResult,WF,sum).
b_compute_expression2(general_product(Ids,Condition,Expression),_T,_I,LocalState,State,MulResult,WF) :- !,
       b_general_sum_or_mul(Ids,Condition,Expression,LocalState,State,MulResult,WF,mul).
/* Begin: Extensions for Z */
b_compute_expression2(if_then_else(If,Then,Else),_T,_I,LocalState,State,Value,WF) :- !,
    b_try_check_boolean_expression(If,LocalState,State,WF,PredRes),
    b_compute_if_then_else(PredRes,Then,Else,LocalState,State,Value,WF).
b_compute_expression2(let_expression(Ids,AssignmentExprs,Expr),_T,_I,LocalState,State,Value,WF) :-
	!, set_up_localstate_for_let(Ids,AssignmentExprs,LocalState,State,LetState,WF),
	b_compute_expression(Expr,LetState,State,Value,WF). %% NOT COVERED (30)
b_compute_expression2(freetype_constructor(Id,Case,Expr),_T,_I,LocalState,State,FreeValue,WF) :-
	!, FreeValue = freeval(Id,Case,Value),
	b_compute_expression(Expr,LocalState,State,Value,WF). %% NOT COVERED (32)
b_compute_expression2(freetype_destructor(Id,Case,Expr),_T,Info,LocalState,State,Value,WF) :-
	!, b_compute_expression(Expr,LocalState,State,freeval(Id,VCase,VValue),WF),
    check_freetype_case(Id,Case,VCase,VValue,Value,Info,WF).
/* End: Extensions for Z */
b_compute_expression2(assertion_expression(Cond,ErrMsg,Expr),_T,_Info,LocalState,State,Value,WF) :- !,
    b_try_check_boolean_expression_lwf(Cond,LocalState,State,WF,PredRes,'$GENERATE_ON_DEMAND' ),
     %nl,translate:print_bexpr(Cond),nl,  translate:print_bexpr(Expr),nl, print(predRes(PredRes)),nl,
    ( PredRes==pred_false ->
      add_wd_error_span(ErrMsg,pred_false,span_predicate(Expr,LocalState,State),WF)
     ; b_compute_expression(Expr,LocalState,State,Value,WF),
       when(nonvar(PredRes),(PredRes=pred_true -> true ; 
             add_wd_error_span(ErrMsg,pred_false,span_predicate(Expr,LocalState,State),WF)))
     ). %% NOT COVERED (34)
b_compute_expression2(multiplication(Arg1,Arg2),integer,Info,LocalState,State,Value,WF) :-  
        same_texpr(Arg1,Arg2), !,  % is less useful when CLPFD is already turned on
        b_compute_expression(Arg1,LocalState,State,SV1,WF), 
        unary_kernel_call(kernel_objects,square,SV1,Value,WF,multiplication(Arg1,Arg1),Info).      
b_compute_expression2(image(Relation,Set),_EType,_Info,LocalState,State,Value,WF) :-
        % print(image(Relation,Set)),nl,
       Relation = b(closure(Rel),_,_CInfo),!, % we have closure1(Rel)[Set] -> avoid computing full closure
        b_compute_expression(Rel,LocalState,State,RV,WF), %print(rel(RV)),nl,
        b_compute_expression(Set,LocalState,State,SV,WF), %print(set(SV)),nl,
        bsets_clp:image_for_closure1_wf(RV,SV,Value,WF).
:- use_module(external_functions).
b_compute_expression2(external_function_call(FunName,Args),_EType,Info,LocalState,State,Value,WF) :- 
      !,
      b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF),
      /*  (external_function_argument_list(FunName, ArgList)
       -> b_compute_expressions_symbolic(Args, ArgList, LocalState,State, EvaluatedArgs, WF)
       ; b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF)  ), */
      % print('EXTERNAL FUNCTION CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_function(FunName,EvaluatedArgs,Value,Info,WF).
b_compute_expression2(Expression,EType,Info,LocalState,State,Value,WF) :- 
       is_set_type(EType,Type),
      /* avoid evaluating certain expensive expression:
         convert them into symbolic closures and expand only if really required */
        functor(Expression,UnOp,1),
        symbolic_closure_unary_operator(UnOp), 
            /* TO DO: UNFOLD THIS OR WRAP in CONSTRUCTOR */
        arg(1,Expression,BType),
        !, %preferences:get_preference(convert_types_into_closures,true),!,
     %  print_message(delayed_un_op(UnOp,BType)),
        b_compute_expression(BType,LocalState,State,ArgValue,WF),
	( (ArgValue ==[], unary_function(UnOp,Module:KernelFunction)) ->
	    /* special treatment for empty set: no need to keep symbolic */
	    unary_kernel_call(Module,KernelFunction,ArgValue,Value,WF,Expression,Info)
	; otherwise ->
	    get_texpr_type(BType,ArgType),
	    create_texpr(value(ArgValue),ArgType,[],TValue),
	    ClosureSetExpression =.. [UnOp,TValue],
	    construct_member_closure('_zzzz_unary',Type,ClosureSetExpression,Value)
	).
	
b_compute_expression2(Expression,EType,_Info,LocalState,State,Value,WF) :- 
       is_set_type(EType,Type),
      /* avoid evaluating certain expensive expressions:
         convert them into symbolic closures and expand only if really required */
        functor(Expression,BinOp,2),
        arg(1,Expression,BType1),
        arg(2,Expression,BType2),
        kernel_mappings:symbolic_closure_binary_operator(BinOp), 
        !,
        %preferences:get_preference(convert_types_into_closures,true),!,
        b_compute_expression(BType1,LocalState,State,ArgValue1,WF),
        b_compute_expression(BType2,LocalState,State,ArgValue2,WF),
        % print_message(delayed_bin_op(BinOp,ArgValue1,ArgValue2)),
        (is_definitely_empty(BinOp,ArgValue1,ArgValue2)
          -> %print(empty_symbolic(BinOp,ArgValue1,ArgValue2)),nl,
             Value = []
          ; do_not_keep_symbolic(BinOp,ArgValue1,ArgValue2,ERes)
           ->  %print('Not keeping symbolic: '),print(BinOp), nl,print(ArgValue1),nl, print(ArgValue2),nl,
               equal_object(Value,ERes)
          ;  get_texpr_type(BType1,TypeArg1), create_texpr(value(ArgValue1),TypeArg1,[],TArg1),
	         get_texpr_type(BType2,TypeArg2), create_texpr(value(ArgValue2),TypeArg2,[],TArg2),
             ClosureSetExpression =.. [BinOp,TArg1,TArg2],
             construct_member_closure('_zzzz_unary',Type,ClosureSetExpression,Value)
         ).
b_compute_expression2(Expression,_T,Info,LocalState,State,Value,WF) :- 
        functor(Expression,Op,1), 
        unary_function(Op,Module:KernelFunction),
        arg(1,Expression,Arg1),!,
        b_compute_expression(Arg1,LocalState,State,SV1,WF), 
     %  print(unary_fun(Op,SV1)),nl,
        unary_kernel_call(Module,KernelFunction,SV1,Value,WF,Expression,Info).
b_compute_expression2(Expression,_T,Info,LocalState,State,Value,WF) :- 
        functor(Expression,Op,2),
        binary_function(Op,Module:KernelFunction),
        arg(1,Expression,Arg1),
        arg(2,Expression,Arg2),!,
        b_compute_expression(Arg1,LocalState,State,SV1,WF), 
        (binary_arg1_determines_value(Op,SV1,Result)  % we could disable this if find_abort_values=true or if SV1 not guaranteed to be wd
          -> %print(det(Op,SV1,Result)),nl,
             equal_object(Value,Result)
          ;  b_compute_expression(Arg2,LocalState,State,SV2,WF),
             binary_kernel_call(Module,KernelFunction,SV1,SV2,Value,WF,Expression,Info)
        ).
b_compute_expression2(E,T,Info,L,S,R,WF) :-
   (var(R) -> add_internal_error('Uncovered expression: ',b(E,T,Info)),
              print_functor(E), print(' : '), print(T),nl
            ; print_error(b_compute_expression_failed(E,T,L,S,R,WF))),fail.
                 
:- use_module(bsyntaxtree,[is_set_type/2]).
b_compute_card(Arg1,Type,Info,LocalState,State,Card,WF) :-
      b_compute_card2(Arg1,Type,Info,LocalState,State,Card,WF),
      % maybe we should move this code below inside kernel_objects ??
      (ground(Card) -> true
        ; is_set_type(Type,SetElType),
          kernel_objects:max_cardinality(SetElType,MaxCard), % this could be precomputed statically
          %print(check_card(Card,MaxCard)),nl,
          (number(MaxCard) -> kernel_objects:in_nat_range(Card,int(0),int(MaxCard))
           ; kernel_objects:greater_than_equal(Card,int(0)) % not sure this is required
          )
      ).
b_compute_card2(set_extension(Ex),_,_,LocalState,State,Card,WF) :- !, 
       b_compute_expressions(Ex,LocalState,State,ExValue,WF),
       kernel_objects:cardinality_of_set_extension_list(ExValue,Card).
b_compute_card2(Arg1,Type,Info,LocalState,State,Card,WF) :-
       b_compute_expression2(Arg1,Type,Info,LocalState,State,SV1,WF),
       kernel_mappings:must_succ_kernel_call(kernel_objects:finite_cardinality_as_int(SV1,Card,WF),SV1,WF,card).
b_compute_max(set_extension(Ex),_,Info,LocalState,State,Max,WF) :- !, 
       b_compute_expressions(Ex,LocalState,State,ExValue,WF),
       %print(max_set_extension(ExValue)),nl,
       kernel_objects:maximum_of_set_extension_list(ExValue,Max,Info,WF).
b_compute_max(Arg1,Type,Info,LocalState,State,Max,WF) :-
       b_compute_expression2(Arg1,Type,Info,LocalState,State,SV1,WF),
       kernel_mappings:must_succ_kernel_call(kernel_objects:maximum_of_set(SV1,Max,Info,WF),SV1,WF,max).
b_compute_min(set_extension(Ex),_,Info,LocalState,State,Min,WF) :- !, 
       b_compute_expressions(Ex,LocalState,State,ExValue,WF),
       kernel_objects:minimum_of_set_extension_list(ExValue,Min,Info,WF).
/* doesnot really buy a lot: */
/*b_compute_min(union(Arg1,Arg2),_,Info,LocalState,State,Min,WF) :- %print(union(Arg1,Arg2)),nl,
       Arg1 = b(set_extension(Ex),Type1,Info1),
       !, %print(start_min_union),nl,
       b_compute_min(set_extension(Ex),Type1,Info1,LocalState,State,Min1,WF), %print(arg1(Min1)),nl,
       b_compute_expression(Arg2,LocalState,State,SV2,WF),
         print(min_union(Min1,SV2)),nl,
       b_compute_min2(SV2,Min1,Min,Info,WF).
:- block b_compute_min2(-,?,?,?,?).
b_compute_min2([],Min1,Res,_,_WF) :- !, Res=Min1.
b_compute_min2(SV1,Min1,Res,Info,WF) :-
       kernel_mappings:must_succ_kernel_call(kernel_objects:minimum_of_set(SV1,Min2,Info,WF),SV1,WF,min),
       min(Min1,Min2,Res).
min(int(A),int(B),int(Res)) :- kernel_objects:minimum(A,B,Res). */
b_compute_min(Arg1,Type,Info,LocalState,State,Min,WF) :-
       b_compute_expression2(Arg1,Type,Info,LocalState,State,SV1,WF),
       kernel_mappings:must_succ_kernel_call(kernel_objects:minimum_of_set(SV1,Min,Info,WF),SV1,WF,min).

% check if a freetype value has the expected case, if not
% generate a WD-error
check_freetype_case(Id,Case,VCase,Inner,Value,Info,WF) :-
    get_enumeration_finished_wait_flag(WF,F),
    check_freetype_case2(Id,Case,VCase,Inner,Value,Info,F,WF).
:- block check_freetype_case2(?,?,-,?,?,?,-,?).
check_freetype_case2(Id,Case,VCase,Inner,Value,Info,_F,WF) :-
    ( nonvar(VCase) ->
        ( Case=VCase -> equal_object(Inner,Value)
        ; otherwise ->
            ajoin(['unexpected case for user defined type ',Id,
                   ' (expected ',Case,')'],Msg),
            add_wd_error_set_result(Msg,VCase,Inner,Value,Info,WF))
    ; otherwise ->       % the computation has finished without value,
        true).           % just drop the check

% check if we have a set_extension argument representing an explicit function
% {a |-> f1, b |-> f2, ...}  or  [f1,f2]
is_extension_function(b(A,_,Info),Info,C) :- is_extension_function_aux(A,C).
is_extension_function_aux(set_extension(Ex),CaseList) :- is_set_extension_function(Ex,CaseList).
is_extension_function_aux(sequence_extension(Ex),CaseList) :- is_sequence_extension_function(Ex,1,CaseList).
is_set_extension_function([],[]).
is_set_extension_function([b(couple(A,B),_,_)|T],[case(A,B)|ST]) :- is_set_extension_function(T,ST).
is_sequence_extension_function([],_,[]).
is_sequence_extension_function([B|T],Nr,[case(b(value(int(Nr)),integer,[]),B)|ST]) :- N1 is Nr+1,
    is_sequence_extension_function(T,N1,ST).

b_apply_function_set_extension([],FunArgVal,LocalState,State,_Res,WF,Function) :-
    add_wd_error_span('Function applied outside of domain: ',FunArgVal,span_predicate(Function,LocalState,State),WF).
b_apply_function_set_extension([case(A,B)|T],FunArgVal,LocalState,State,Res,WF,Function) :-
    b_compute_expression(A,LocalState,State,AVal,WF),
    %print(' --> '),translate:print_bexpr(B),nl,
    equality_objects(AVal,FunArgVal,EqRes),  % equality_objects_lwf ??
    % TO DO: if Res is known, we can call equality_objects(BVal,Res,ResEqRes)
    b_apply_function_set_extension_aux(EqRes,B,T,FunArgVal,LocalState,State,Res,WF,Function).

:- block b_apply_function_set_extension_aux(-,?,?,?, ?,?,?,?,?).
b_apply_function_set_extension_aux(pred_true,B,_T,_FunArgVal,LocalState,State,Res,WF,_Function) :-
    b_compute_expression(B,LocalState,State,BVal,WF), %print(value(BVal)),nl,
    equal_object(BVal,Res).
b_apply_function_set_extension_aux(pred_false,_B,T,FunArgVal,LocalState,State,Res,WF,Function) :-
    b_apply_function_set_extension(T,FunArgVal,LocalState,State,Res,WF,Function).


% --------------------- SIGMA & PI ------------------------

/* first clause: try to do a specific treatment for formulas of the form SIGMA(ID).(ID:SET|EXPR) */
/* Advantage: will not delay as long before expanding SET  */
b_general_sum_or_mul([b(identifier(ID),TYPE,_)],b(member(b(identifier(ID),TYPE,_),SET),pred,_),
              Expression,LocalState,State,SumResult,WF,SUMorMUL) :- !,
       % nl,print(b_general_sum_or_mul_set(ID,SET,SUMorMUL)),nl, %%
       b_compute_expression(SET,LocalState,State,SETResult,WF),
       b_general_sum_or_mul_over_set(ID,TYPE,SETResult,Expression,LocalState,State,SumResult,WF,SUMorMUL).
b_general_sum_or_mul(Ids,Condition,Expression,LocalState,State,SumMulResult,WF,SUMorMUL) :-
       %print(general_sum_or_mul(Ids,Condition,Expression)),nl,
	    create_map_to_int(Ids,Condition,Expression,general_sum_or_mul,Set),
        b_compute_expression(Set,LocalState,State,SETResult,WF),
        % TO DO: maybe call directly expand_normal_closure2(no_check,Ids+...,ParameterTypes,ClosureBody,Result,Done)
        %prints(sum_or_mul_over_range(EResult,LocalState,State,SumMulResult,SUMorMUL)),nl,
        (SUMorMUL=sum
        -> kernel_call(bsets_clp:sum_over_range(SETResult,SumMulResult),(SETResult,SumMulResult),WF,
                    general_sum(Ids,Condition,Expression))
         ; %expand_custom_set_to_list(SETResult,EResult),
           kernel_call(bsets_clp:mul_over_range(SETResult,SumMulResult),(SETResult,SumMulResult),WF,
                    general_product(Ids,Condition,Expression))).

:- block b_general_sum_or_mul_over_set(?,?,-, ?,?,?, ?,?,?).
b_general_sum_or_mul_over_set(ID,_TYPE,SETResult,Expression,_LS,_S,SumResult,_WF,SUMorMUL) :- 
       get_texpr_id(Expression,ID), % check if we just sum the elements of the list
       sum_or_mul_of_explicit_set(SETResult,SUMorMUL,R),!,
       % print_term_summary(sum_or_mul(R,SETResult)),nl, %%
       equal_object(SumResult,R).
b_general_sum_or_mul_over_set(ID,TYPE,SETResult,Expression,LocalState,State,SumResult,WF,SUMorMUL) :- 
       % print_term_summary(sum_or_mul(SETResult)),nl,%%
       expand_custom_set_to_list(SETResult,ESETResult,_Done,b_general_sum_or_mul_over_set),
       %print(sum(SETResult,ESETResult)),nl,
       b_sum_or_mul_over_list(ESETResult,ID,TYPE,Expression,LocalState,State,SumResult,WF,SUMorMUL).

b_sum_or_mul_over_list(L,ID,TYPE,Expr,LS,State,Result,WF,SUMorMUL) :-
   get_acc_base(SUMorMUL,Acc),
   b_sum_or_mul_over_list_acc(L,ID,TYPE,Expr,LS,State,Acc,Result,WF,SUMorMUL).

:- block b_sum_or_mul_over_list_acc(-, ?,?, ?,?,?,?, ?,?,?).
% better if list is fully known: no constraint propagation performed during computation !
% for example: SIGMA(x).(x:{-1} \/ 1..30000|x+1) goes down from 2050 ms to 1210 ms
b_sum_or_mul_over_list_acc([],_,_,_,_,_,Acc,Result,_WF,_SUMorMUL) :- Result=int(Acc).
b_sum_or_mul_over_list_acc([H|T],ID,TYPE,Expr,LocalState,State,Acc,Result,WF,SUMorMUL) :-
   % print(sum_over_acc(H,ID,Acc,Result)),nl,
   add_typed_var_to_localstate(ID,H,TYPE,LocalState,NewLocalState),
   b_compute_expression(Expr,NewLocalState,State,HExprVal,WF),
   ((nonvar(T),nonvar(Acc),ground(HExprVal))
     -> HExprVal=int(HVal), % we can continue computing the system without constraint propagation; is faster
        compose_acc(SUMorMUL,Acc,HVal,NewAcc), 
        b_sum_or_mul_over_list_acc(T,ID,TYPE,Expr,LocalState,State,NewAcc,Result,WF,SUMorMUL)
    ;   compose_clp(SUMorMUL,HExprVal,Res_wo_H,Result),
        compose_clp(SUMorMUL,int(Acc),Res_wo_HACC,Res_wo_H),
        b_sum_or_mul_over_list_clp(T,ID,TYPE,Expr,LocalState,State,Res_wo_HACC,WF,SUMorMUL)
   ).
:- block b_sum_or_mul_over_list_clp(-, ?,?, ?,?,?, ?,?,?).
b_sum_or_mul_over_list_clp([],_,_, _,_,_, Result,_WF,SUMorMUL) :-
   get_acc_base(SUMorMUL,Base), Result = int(Base).
b_sum_or_mul_over_list_clp([H|T],ID,TYPE,Expr,LocalState,State,Result,WF,SUMorMUL) :-
   % print(sum_over(H,ID,Result)),nl,
   add_typed_var_to_localstate(ID,H,TYPE,LocalState,NewLocalState),
   b_compute_expression(Expr,NewLocalState,State,HExprVal,WF),
   (T==[] -> Result=HExprVal
    ;   compose_clp(SUMorMUL,HExprVal,TExprSum,Result),
        b_sum_or_mul_over_list_clp(T,ID,TYPE,Expr,LocalState,State,TExprSum,WF,SUMorMUL)
   ).

get_acc_base(sum,0).
get_acc_base(mul,1).
compose_acc(sum,A,B,C) :- C is A+B.
compose_acc(mul,A,B,C) :- C is A*B.
compose_clp(sum,A,B,C) :- int_plus(A,B,C).
compose_clp(mul,A,B,C) :- times(A,B,C).

create_map_to_int(Ids,Condition,Expression,Loc,Set) :-
	Info = [generated(Loc)],
	create_texpr(identifier(maptoint___tmpvar),integer,Info,TmpVar),
	append(Ids,[TmpVar],Vars),
	create_texpr(equal(TmpVar,Expression),pred,Info,Equal),
	conjunct_predicates([Condition,Equal],Pred),
	get_texpr_types(Ids,Types),couplise_list(Types,ArgType),
	create_texpr(comprehension_set(Vars,Pred),set(couple(ArgType,integer)),Info,Set).

/* for if-then-else in Z: compute Then or Else,
   depending on the outcome of the predicate */
:- block b_compute_if_then_else(-,?,?,?,?,?,?).
b_compute_if_then_else(pred_true,Then,_Else,LocalState,State,Value,WF) :-
    b_compute_expression(Then,LocalState,State,Value,WF).
b_compute_if_then_else(pred_false,_Then,Else,LocalState,State,Value,WF) :-
    b_compute_expression(Else,LocalState,State,Value,WF).

/* for let-statements in Z: expand localstate by new variables and assign
   values to them */
set_up_localstate_for_let(Ids,Exprs,LocalState,State,LetState,WF) :-
	set_up_localstate(Ids,Vars,LocalState,LetState),
	compute_let_expressions(Exprs,Vars,LocalState,State,WF).
compute_let_expressions([],[],_,_,_).
compute_let_expressions([Expr|RestExprs],[Var|RestVars],LocalState,State,WF) :-
	b_compute_expression(Expr,LocalState,State,Value,WF),
	equal_object_optimized(Var,Value),
	compute_let_expressions(RestExprs,RestVars,LocalState,State,WF).

l_compute_field([],_LS,_S,[],_WF).
l_compute_field([Field|T],LS,S,[FVal|CT],WF) :-
   b_compute_field(Field,LS,S,FVal,WF),
   l_compute_field(T,LS,S,CT,WF).

b_compute_field(field(Name,TypeOrVal),LocalState,State,field(Name,TypeOrFieldValue),WF) :- 
        b_compute_expression(TypeOrVal,LocalState,State,TypeOrFieldValue,WF).
        
        

:- use_module(b_compiler).

/* :- type bsets_closure +--> closure(list(type(variable_id)),
                                   list(type(basic_type_descriptor)),type(boolean_expression)). */

:- assert_pre(b_interpreter:b_compute_comprehension_set(Parameters,Cond,_Info,LS,State,_R,_WF),
     (ground_check(Parameters),
      bsyntaxtree:check_if_typed_predicate(Cond),type_check(LS,store),type_check(State,store)) ).

:- assert_post(b_interpreter:b_compute_comprehension_set(_P,_C,_Info,_LS,_State,Val,_WF), 
                 (b_interpreter:value_type(Val))). %, nonvar(Val)
                 
mark_bexpr_as_symbolic(b(E,T,I),b(E2,T,[prob_annotation('SYMBOLIC')|I])) 
  :- mark_aux(E,E2).
mark_aux(union(A,B),R) :- !, R=union(SA,SB),
   mark_bexpr_as_symbolic(A,SA), mark_bexpr_as_symbolic(B,SB).
% union is currently the only operator that is kept symbolically
mark_aux(A,A).  

%b_compute_explicit_comprehension_set(Parameters,Condition,LocalState,State,Result,WF) :-
%    b_generate_closure(Parameters,Condition,LocalState,State,CResult),
%    expand_only_custom_closure_global(CResult,Result,no_check).

b_compute_comprehension_set(Parameters,Condition,Info,LocalState,State,Result,WF) :-
    b_generate_closure_if_necessary(Parameters,Condition,LocalState,State,CResult),
    %print(gen(Parameters,Condition,CResult)),nl,
    b_compute_comprehension_set_aux(CResult,Info,Result,WF).

b_compute_comprehension_set_aux(CResult,_,Result,_WF) :- 
  (var(CResult) ; CResult \= closure(_,_,_)),!,
  %% print(not_generated_closure(CResult)),nl, %%
  Result=CResult.
b_compute_comprehension_set_aux(CResult,Info,Result,_WF) :-
   member(prob_annotation('SYMBOLIC'),Info),!,
   custom_explicit_sets:mark_closure(CResult,['SYMBOLIC'],CRS),Result=CRS.
b_compute_comprehension_set_aux(CResult,_Info,Result,_WF) :-
   (preferences:get_preference(convert_comprehension_sets_into_closures,true) ;
    %% print(check(CResult)),nl,trace, %%
    dont_expand_this_explicit_set(CResult)),!,
   Result=CResult.
b_compute_comprehension_set_aux(CResult,_Info,Result,WF) :- % print(expand(CResult)),nl,
     get_wait_flag(1,b_compute_comprehension_set,WF,WF0),
	 when( (ground(CResult);nonvar(WF0)),
		   %(nonvar(WF0);ground(CResult)), % sometimes explicit values defined by comprehension_sets: will generate AVL
	     % (print(expand_comprehension_set_closure(Parameters)),nl, %%
		expand_only_custom_closure_global(CResult,Result,check_no_inf) /* check_no_inf: we already know it is not infinite */
		% ,print(done_expand_custom_set(Result)),nl )%%
		).

split_names_and_types([],[],[]).
split_names_and_types([Identifier|IdRest],[Name|NRest],[Type|TRest]) :-
    get_texpr_expr(Identifier,identifier(Name)),
    get_texpr_type(Identifier,Type),
    split_names_and_types(IdRest,NRest,TRest).

:- use_module(closures,[construct_closure_if_necessary/4]).
b_generate_closure(Parameters,Condition,LocalState,State,Result) :-
    split_names_and_types(Parameters,Names,Types),
    construct_closure(Names,Types,ClosurePred,Result),
    ~~mnf( b_compiler:b_compile(Condition,Names,LocalState,State,ClosurePred) ).

b_generate_closure_if_necessary(Parameters,Condition,LocalState,State,Result) :-
    split_names_and_types(Parameters,Names,Types),
    ~~mnf( b_compiler:b_compile(Condition,Names,LocalState,State,ClosurePred) ),
    construct_closure_if_necessary(Names,Types,ClosurePred,Result).


:- public convert_list_of_expressions_into_sequence/2.
convert_list_of_expressions_into_sequence(List,SeqValue) :-
   convert_list_of_expressions_into_sequence(List,1,Seq),
   equal_object_optimized(Seq,SeqValue).

convert_list_of_expressions_into_sequence([],_,[]).
convert_list_of_expressions_into_sequence([H|T],Cur,[(int(Cur),H)|ConvT]) :-
  C1 is Cur+1,
  convert_list_of_expressions_into_sequence(T,C1,ConvT).

convert_list_of_expressions_into_set(List,Set) :-
   convert_list_of_expressions_into_set(List,Set,set(any)).


convert_list_of_expressions_into_set(ListOfElements,Set,Type) :-
   list_of_expressions_is_ground(ListOfElements,Ground,OrderedListOfElements),
   convert_list_of_expressions_into_set1(Ground,OrderedListOfElements,Set,Type).
convert_list_of_expressions_into_set1(true,ListOfElements,Set,_Type) :- %var(Set),
   !,  construct_avl(ListOfElements,AVLSet),
   equal_object(AVLSet,Set).
convert_list_of_expressions_into_set1(_,Seq,Set,Type) :- 
   kernel_objects:max_cardinality(Type,TypeMaxCard),
   convert_list_of_expressions_into_set2(Seq,1,[],Set,TypeMaxCard). % , print(converted(Seq,Set)),nl.

% TO DO: sort elements : ground elements first
convert_list_of_expressions_into_set2([],_,Set,Res,_) :- %print(equal_object(Set,Res)),nl,
   equal_object(Set,Res).
%  debug:timer_call(kernel_objects:equal_object(Set,Res)).
convert_list_of_expressions_into_set2([H|T],MaxSze,SetSoFar,OutSet,TypeMaxCard) :-
   %kernel_objects:basic_type_set(Type,IntSet,MaxSze), 
   check_set_lists_bounded(IntSet,TypeMaxCard,MaxSze),
   % TO DO: optimize; avoid traversing or typing elements over and over again
   % maybe add a max. sze argument to add_element
   add_element(H,SetSoFar,IntSet), % print(add_el(H,SetSoFar,IntSet)),nl,
   % TO DO: check if H in OutSet and keep track of which elements have already been covered in OutSet
   M1 is MaxSze+1,
   convert_list_of_expressions_into_set2(T,M1,IntSet,OutSet,TypeMaxCard).


:- use_module(inf_arith).
% check that a set does not exceed a certain bound (to prevent infinite
%  or overeager enumeration of sets); Warning: will not check custom_explicit_set sizes
check_set_lists_bounded(Set,TypeMaxCard,Bound) :-
   inf_arith:infmin(TypeMaxCard,Bound,Card),
   (Card=inf -> true
     ; Card<1 -> empty_set(Set)
     ; check_set_bounded_aux(Set,Card)).

:- block check_set_bounded_aux(-,?).
check_set_bounded_aux(CustomSet,_) :- is_custom_explicit_set(CustomSet,check_set_bounded_aux),!.
check_set_bounded_aux([],_).
check_set_bounded_aux([_|T],Card) :-
   C1 is Card-1,
   (C1=0 -> empty_set(T) ; check_set_bounded_aux(T,C1)).



% check if the list can be converted into AVL form; if not: re-order list so that ground elements are present first
list_of_expressions_is_ground([],true,[]).
list_of_expressions_is_ground([E|T],Ground,ResList) :- 
   (should_be_converted_to_avl(E) -> ResList=[E|RT], list_of_expressions_is_ground(T,Ground,RT)
    ; Ground=false,
      bring_ground_elements_forward(T,[E],ResList)
    ).

bring_ground_elements_forward([],Acc,Acc).
bring_ground_elements_forward([H|T],Acc,Res) :-
  (should_be_converted_to_avl(H) -> Res=[H|RT], bring_ground_elements_forward(T,Acc,RT)
    ; bring_ground_elements_forward(T,[H|Acc],Res)).

% compute cardinality of a list of parameters, such as for forall
parameter_list_cardinality([],1).
parameter_list_cardinality([b(identifier(_),BasicType,_)|T],Res) :-
   parameter_list_cardinality(T,TCard),
   (TCard=inf -> Res=inf
    ; kernel_objects:max_cardinality(BasicType,Card),
      kernel_objects:safe_mul(Card,TCard,Res)
   ).
   
b_for_all(Parameters,LHS,RHS,LocalState,State,WF) :- 
     get_wait_flag(1,for_all(Parameters),WF,LWF), % hack: give total_function,... priority to set up domain/range
     %debug_print(9,'b_for_all: '),debug_print_bexpr(LHS), 
     b_generate_for_all_domain(LWF,Parameters,LHS,RHS,LocalState,State,AllSols,NewRHS,WF),
     %debug_println(9,'generated_for_all_domain'),
     % TO DO: move checking of NewRHS into the generation of the domain !?
     % (so that we check directly when we generate a solution; not generating the domain first ?)
     expand_forall(AllSols,NewRHS,Parameters,LocalState,State,WF,0).

% generate the domain of a Universally Quantified Formula
:- block b_generate_for_all_domain(-,?,?, ?,?,?,?,?,?).
b_generate_for_all_domain(_LWF,Parameters,LHS,RHS,LocalState,State,AllSols,NewRHS,WF) :-
	b_generate_for_all_domain_nolwf(Parameters,LHS,RHS,LocalState,State,AllSols,NewRHS,WF).

b_generate_for_all_domain_nolwf(Parameters,LHS,RHS,LocalState,State,ForAllDomain,NewRHS,WF) :-
  %print(forall),nl, print(Parameters),nl,print(LHS),nl,
  % this patterns occurs often in Event-B translations to classical B
   LHS=b(truth,_,_), RHS=b(implication(LHS2,RHS2),_,_),
   large_parameter_list_cardinality(Parameters,50), % larger bound than below: maybe the user explicitly wants to expand it
   !,
   b_generate_for_all_domain_nolwf(Parameters,LHS2,RHS2,LocalState,State,ForAllDomain,NewRHS,WF).
b_generate_for_all_domain_nolwf(Parameters,LHS,RHS,LocalState,State,ForAllDomain,NewRHS,WF) :-
   % optimized treatment for formulas of the form !x.(x:S => RHS) or !(x,y).(x|->y:S => RHS)
   % to do: expand to more formulas, such as !(x,y).(x:S & y:T => RHS)
    is_for_all_set_membership_predicate(LHS,Parameters,Set,Pattern,ParameterValues),!, NewRHS=RHS,
    b_generate_set_for_all_domain(Set,LocalState,State,Pattern,ParameterValues,ForAllDomain,WF).
b_generate_for_all_domain_nolwf(Parameters,LHS,RHS,LocalState,State,ForAllDomain,NewRHS,WF) :-
     ~~mnf( b_generate_closure(Parameters,LHS,LocalState,State,Closure) ),
     b_generate_for_all_domain2(Closure,Parameters,LHS,RHS,LocalState,State,ForAllDomain,NewRHS,WF).

b_generate_for_all_domain2(Closure,_Parameters,LHS,RHS,_LocalState,_State,ForAllDomain,NewRHS,_WF) :-
      (LHS=b(truth,pred,[]) ; ground(Closure)),
      !, NewRHS=RHS,
       %(ground(Closure) -> print(ground(_Parameters)),nl, print_bexpr(LHS),nl ; true), %%
      ~~pp_mnf( b_expand_closure_forall(Closure,ForAllDomain)).
b_generate_for_all_domain2(_,Parameters,LHS,RHS,LocalState,State,ForAllDomain,NewRHS,WF) :-
     is_a_conjunct(LHS,_,_),
     LHS_Membership = b(member(_,_),_,_), % TO DO: also allow conjuncts
     member_conjunct(LHS_Membership,LHS,RestLHS),
     is_for_all_set_membership_predicate(LHS_Membership,Parameters,Set,Pattern,ParameterValues),
     max_cardinality_ok_for_expansion(Set,100,LocalState,State),!, %% maximum cardinality of Set (could be infinite !!) !x.(x:NATURAL1 & x<N => P)
     % print('Splitting for all conjunct: '), print(Parameters),nl, print_bexpr(LHS_Membership),nl,print_bexpr(RestLHS),nl, %%
     NewRHS = b(implication(RestLHS,RHS),pred,[]), % TO DO: extract & join info
     b_generate_set_for_all_domain(Set,LocalState,State,Pattern,ParameterValues,ForAllDomain,WF).
b_generate_for_all_domain2(Closure,Parameters,_LHS,RHS,_LocalState,_State,ForAllDomain,NewRHS,_WF) :-
      %print_bt_message(compiled_forall(Closure,LocalState,State)),%%
     large_parameter_list_cardinality(Parameters,8),
     !,
      NewRHS=RHS, ~~pp_mnf( b_expand_closure_forall(Closure,ForAllDomain)).
b_generate_for_all_domain2(_Closure,Parameters,LHS,RHS,LocalState,State,ForAllDomain,NewRHS,_WF) :-
      % print('Inlining Forall '),parameter_list_cardinality(Parameters,PC), print(PC),nl,print(Parameters),nl,
      NewLHS=b(truth,pred,[]), NewRHS=b(implication(LHS,RHS),pred,[]),
      ~~mnf( b_generate_closure(Parameters,NewLHS,LocalState,State,Closure) ),
      ~~pp_mnf( b_expand_closure_forall(Closure,ForAllDomain))
      %% ,print(domain(ForAllDomain)),nl %%
      .

% code to decide whether a for all should be expanded if the
% predicate contains a set membership predicate : !x.( (P & x:S & P') => Q)
max_cardinality_ok_for_expansion(b(Expr,Type,_Info),Limit,LS,S) :- % which value to choose for Limit ??
  %print('Checking Max Card: '), print_bexpr(b(Expr,Type,_Info)),nl,
  kernel_objects:max_cardinality(Type,TypeCard), %print(type_card(TypeCard)),nl,
  ( (number(TypeCard),TypeCard =< Limit) -> true ; 
   max_cardinality_ok_for_expansion2(Expr,Limit,TypeCard,LS,S)). 

max_cardinality_ok_for_expansion1(b(E,_,_),Limit,TC,LS,S) :- max_cardinality_ok_for_expansion2(E,Limit,TC,LS,S).

:- use_module(custom_explicit_sets,[efficient_card_for_set/3]).

max_cardinality_ok_for_expansion2(set_extension(E),Limit,_TC,_,_) :- !, length(E,Len), Len =< Limit.
max_cardinality_ok_for_expansion2(interval(Up,Low),Limit,_TC,LS,S) :- !,
  %print(interval(Up,Low,Limit)),nl,
  get_integer(UpI,Up,LS,S), get_integer(LowI,Low,LS,S),
  UpI-LowI<Limit.
max_cardinality_ok_for_expansion2(integer_set(GS),Limit,_TC,_LS,_S) :- !, b_global_set_cardinality(GS,Card),
  number(Card), Card =< Limit.
max_cardinality_ok_for_expansion2(set_subtraction(A,_),Limit,TC,LS,S) :- !,
    max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(image(RelA,_),Limit,TC,LS,S) :- !,
    max_cardinality_ok_for_expansion1(RelA,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(intersection(A,B),Limit,TC,LS,S) :- !,
 (max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S) -> true ; max_cardinality_ok_for_expansion1(B,Limit,TC,LS,S)).
max_cardinality_ok_for_expansion2(union(A,B),Limit,TC,LS,S) :- !,
  max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S), max_cardinality_ok_for_expansion1(B,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(range(A),Limit,TC,LS,S) :- !,
  max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(domain(A),Limit,TC,LS,S) :- !,
  max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(direct_product(A,B),Limit,TC,LS,S) :- !,
   max_cardinality_ok_for_expansion2(cartesian_product(A,B),Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(parallel_product(A,B),Limit,TC,LS,S) :- !,
   max_cardinality_ok_for_expansion2(cartesian_product(A,B),Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(cartesian_product(A,B),Limit,TC,LS,S) :- !,
  ILimit is Limit//2, ILimit>0, %use sqrt ?
  % print(cartesian(Limit,ILimit,A,B)),nl,
  max_cardinality_ok_for_expansion1(A,ILimit,TC,LS,S),
  max_cardinality_ok_for_expansion1(B,ILimit,TC,LS,S).
max_cardinality_ok_for_expansion2(value(S),Limit,_TC,_LS,_S) :- !, 
   efficient_card_for_set(S,Card,C),call(C),
   number(Card), Card =< Limit.
max_cardinality_ok_for_expansion2(identifier(_ID),_Limit,TC,_LS,_S) :- !, number(TC).
  % TO DO: check if ID is fully defined; if so use card else accept as it will
  % probably be represented by a list and the forall constraints could be useful in finding ID's value
  % we have to store/compute ID anyway (could be infinite closure; hence check that TC is a number);  a bit of a hack; is there a more principled way?
max_cardinality_ok_for_expansion2(E,Limit,TC,_LS,_S) :- print(unknown_max_card(E,Limit,TC)),nl,fail.

get_integer(X, b(Expr,integer,_),LS,S) :- get_integer2(Expr,X,LS,S).
get_integer2(integer(X),X,_,_) :- number(X).
get_integer2(identifier(Id),Val,LS,S) :- lookup_value(Id,LS,S,int(Val)), number(Val).

member_conjunct(Predicate,Conjunction,Rest) :-
  conjunction_to_list(Conjunction,List),
  select(Predicate,List,RestList),
  conjunct_predicates(RestList,Rest).
     
member_conjunct(Conj,Conj,b(truth,pred,[])).

b_generate_set_for_all_domain(Set,LocalState,State,Pattern,ParameterValues,ForAllDomain,WF) :-
     ~~pp_cll(b_compute_expression(Set,LocalState,State,EvSet,WF)),
    expand_custom_set_to_list(EvSet,ESet,_Done,b_generate_set_for_all_domain),    %% print(b_set_forall_domain(ESet)),nl, %%%
    convert_list_to_parameter_tuples(ESet,Pattern,ParameterValues,ForAllDomain).


% check if it is a predicate of the form x:S or x:Sx & y:Sy ... which we can optimize
%  (i.e. compute SET before all values inside it are fully known)
is_for_all_set_membership_predicate(LHS,Parameters,Set,Pattern,ParameterValues) :-
    get_texpr_ids(Parameters,ParIDs), same_length(ParIDs,ParameterValues),
    is_for_all_set_membership_predicate2(LHS,ParIDs,ParIDs,UnmatchedParIDs,Set,Pattern,ParameterValues,_UPV),
    (UnmatchedParIDs=[] -> debug_println(9,set_membership_forall(ParIDs))
     ; debug_println(9,unmatched(UnmatchedParIDs)),fail).

is_for_all_set_membership_predicate2(LHS,All,ParIDs,UnmatchedParIDs,Set,Pattern,ParameterValues,UPV) :-
    LHS = b(member(Element,Set),_,_),!, 
    identifiers_not_used_in_bexp(All,Set),
    non_failing_pattern_match(Element,ParIDs,UnmatchedParIDs,Pattern,ParameterValues,UPV). % TO DO: also allow failing pattern matches ?!; at least when splitting
is_for_all_set_membership_predicate2(LHS,All,ParIDs,ParIDs2,Set,
                                    (Pat1,Pat2),ParameterValues,UPV) :-
    is_a_conjunct(LHS,LHS1,LHS2), % we could have something of the form x:Sx & y:Sy ...
    Set = b(cartesian_product(Set1,Set2),couple(T1,T2),[]),
    is_for_all_set_membership_predicate2(LHS1,All,ParIDs, ParIDs1,Set1,Pat1,ParameterValues,PV1),
    is_for_all_set_membership_predicate2(LHS2,All,ParIDs1,ParIDs2,Set2,Pat2,PV1,UPV),
    get_texpr_type(Set1,T1),get_texpr_type(Set2,T2).

identifiers_not_used_in_bexp(IDList,Expr) :-
   find_identifier_uses(Expr,[],Used),
   id_check(IDList,Used).
id_check([],_).
id_check([ID|T],Used) :- %get_texpr_id(IDE,ID),
  \+ member(ID,Used), id_check(T,Used).
  
non_failing_pattern_match(IDE,IDsIn,IDsOut,PatVariable,VarsIn,VarsOut) :- get_texpr_id(IDE,ID),!,
   nth1(Nr,IDsIn,ID,IDsOut), nth1(Nr,VarsIn,PatVariable,VarsOut).
%   select(ID,IDsIn,IDsOut).
non_failing_pattern_match(b(couple(ID1,ID2),_,_),IDS,OutIDS,Pattern,VarsIn,VarsOut) :-
   Pattern = (Pat1,Pat2),
   non_failing_pattern_match(ID1,IDS,RemIDS,Pat1,VarsIn,VarsInt),
   non_failing_pattern_match(ID2,RemIDS,OutIDS,Pat2,VarsInt,VarsOut).

large_parameter_list_cardinality(Parameters,Limit) :-
  parameter_list_cardinality(Parameters,ParCard),  (ParCard=inf -> true; ParCard > Limit).

:- block convert_list_to_parameter_tuples(-,?,?,?).    
convert_list_to_parameter_tuples([],_,_,[]).
convert_list_to_parameter_tuples([H|T],Pattern,ParameterValues,Res) :-
   copy_term((Pattern,ParameterValues),(CPat,CPV)),
   (CPat=H -> Res = [CPV|RT]
     ; print(error_could_not_unify_pattern(H,Pattern)),nl,
       Res=RT),
    convert_list_to_parameter_tuples(T,Pattern,ParameterValues,RT).


:- block expand_forall(-,?,?, ?,?,?,?).
expand_forall([],_RHS,_Parameters,_LocalState,_State,_WF,_Nr) :- !.
      %% :- print('finished_expanding forall bodies: '),print_bexpr(_RHS),nl , print(number_of_instances(_Nr)),nl,kernel_waitflags:portray_waitflags(_WF). %%
expand_forall([ParResult|T],RHS,Parameters,LocalState,State,WF,Nr) :- !,
     %% print(test_forall_instance(Nr,ParResult,LocalState)),nl, %%
   test_forall_instance(ParResult,RHS,Parameters,LocalState,State,WF),
   N1 is Nr+1,
   expand_forall(T,RHS,Parameters,LocalState,State,WF,N1).
expand_forall(Set,RHS,Parameters,LocalState,State,WF,Nr) :- 
    add_internal_error('Illegal call: ', expand_forall(Set,RHS,Parameters,LocalState,State,WF,Nr)),
    fail.

test_forall_instance(ParResult,RHS,Parameters,LocalState,State,WF) :-
   %(Parameters == [b(identifier(ss),global('SESSION'),[nodeid(pos(307,1,79,6,79,7))])] -> trace ; true),
   if(set_up_typed_localstate(Parameters,ParResult,_,LocalState,NewLocalState),true,
      (nl,print(set_up_typed_localstate_failed(Parameters,ParResult,_,LocalState,NewLocalState)),nl,
     % it can happen that ParResult has constraints on it and that setting up the typing info detects the error
     % add_internal_error('Call failed: ', set_up_typed_localstate(Parameters,ParResult,_,LocalState,NewLocalState)),
     fail)),
   %init_wait_flags(InnerWF),
   %  print('test: '),print_bexpr(RHS), print(' : '), print(ParResult),nl,
   b_test_boolean_expression(RHS,NewLocalState,State,WF). %InnerWF),
 %  WF=InnerWF. /* give the boolean expression a chance to do ground propagations first */

   /* now the same as above; but without converting list of values into pairs */
:- assert_pre(b_interpreter:b_expand_closure_forall(Closure,_Result),
                (nonvar(Closure), custom_explicit_sets:is_closure(Closure,_Parameters,_ParameterTypes,ClosureBody),
                 nonvar(ClosureBody),
                 bsyntaxtree:check_if_typed_predicate(ClosureBody))).
:- assert_post(b_interpreter:b_expand_closure_forall(_Closure,Result), 
                 b_interpreter:value_type(Result)).
b_expand_closure_forall(Closure,Result) :- nonvar(Closure),
    is_closure_x(Closure, Parameters, ParameterTypes, ClosureBody, XtraVar),!,
    delay_setof_list( ParValues,
                     (%% print_bt_message(expanding_forall_closure(Parameters,ClosureBody,ParValues)), %%
                      custom_explicit_sets:b_test_closure(Parameters,ParameterTypes,ClosureBody,ParValues)
                      %% ,print_bt_message(expanded(Parameters,ParValues)) %%
                      ),
                     Result,
                     [ParValues,XtraVar]).
 b_expand_closure_forall(Closure,Result) :-
       add_internal_error('Illegal Call: ', b_expand_closure_forall(Closure,Result)),
       fail.
/*  replace last clause by following if you want to use  b_generate_closure_if_necessary
b_expand_closure_forall(Set,LESet) :- % not generated a closure; could be global_set, avl_set or list
   expand_custom_set_to_list(Set,ESet,_,b_expand_closure_forall),
   wrap_setelements_into_list(ESet,LESet). % the expand_forall expects a list of values; one for each parameter
:- block wrap_setelements_into_list(-,?).
wrap_setelements_into_list([],[]).
wrap_setelements_into_list([H|T],[[H]|WT]) :- wrap_setelements_into_list(T,WT).
*/

:- assert_pre(b_interpreter:set_up_typed_localstate(V,S,_), (list_skeleton(V),list_skeleton(S))).
:- assert_post(b_interpreter:set_up_typed_localstate(_,_,R),list_skeleton(R)).

set_up_typed_localstate(Identifiers,InState,OutState) :-
	set_up_typed_localstate(Identifiers,_FreshVars,_TypedVals,InState,OutState).

set_up_typed_localstate([],[],[],State,State).
set_up_typed_localstate([Identifier|IdRest],[Val|ValRest],[typedval(Val,Type,Var)|TRest],InState,OutState) :-
    % The other way round (first add Identfier, then IdRest) would be more efficient,
    % but we do it this way to keep compatibility with a previous version.
    % In specfile:compute_operation_effect, the arguments of an operation are in the same order than in the store
    set_up_typed_localstate(IdRest,ValRest,TRest,InState,InterState),
    def_get_texpr_id(Identifier,Var), get_texpr_type(Identifier,Type),
    add_typed_var_to_localstate(Var,Val,Type,InterState,OutState).


add_typed_var_to_localstate(Var,Val,Type,InState,OutState) :-
   %% print(adding(Var,Val,Type)),nl, %%
    add_var_to_localstate(Var,Val,InState,OutState),
    %%when(ground(Val), (print(' ++ '),print(Var),print(' ---> '), print(Val),nl)),
    %%print(basic_type(Val,Type,InState,OutState)),nl, %translate:print_bvalue(Val),nl,
    ~~pp_cll( kernel_objects:basic_type(Val,Type) ). %% ,print(ok(Val)),nl.

/*
leftmost_conjunct(b(conjunct(A,B),pred,Info),Left,Right) :- !,
    (leftmost_conjunct(A,AL,AR) 
      -> Left = AL, Right = b(conjunct(AR,B),pred,Info)
      ;  Left=A, Right = B ).

xxxx_b_test_exists([OneParaID|RestIDs],PRED,Infos,LS,State,WF) :- 
    % try and detect exists used as replacement for a let; 
    % it is also done in type-checker cleanup, do we still need this ??? ...
    % some redundant code with create_and_simplify_exists in  b_interpreter_components
    leftmost_conjunct(PRED,EQUALITY,RestPred),
    %print(try_let(OneParaID,EQUALITY,RestPred)),nl,
    EQUALITY = b(equal(ID,Expr),_,_), %print(equal(ID)),nl,
    same_id(OneParaID,ID,SID), %print(same_id(SID)),nl,
    find_identifier_uses(Expr,[],Used),
    get_texpr_id_names([OneParaID|RestIDs],UP), sort(UP,Parameters),
    %print(check_occur(Used,Parameters)),nl,
    \+ ord_intersect(Used,Parameters), % we can only extract the Expr if it does not use any parameter
    !, debug_println(9,let(SID,RestIDs)),
    set_up_typed_localstate([OneParaID],_FreshOutputVars,_TypedVals,LS,NewLocalState),
    b_test_boolean_expression(EQUALITY,NewLocalState,State,WF),
    (RestIDs=[] -> b_test_boolean_expression(RestPred,NewLocalState,State,WF)
      ; b_test_exists(RestIDs,RestPred,Infos,NewLocalState,State,WF)).
xxb_test_exists(Parameters,Condition,Infos,LocalState,State,WF) :-
     b_check_boolean_expression(b(exists(Parameters,Condition),pred,Infos),LocalState,State,WF,NR),
     get_exists_used_ids(Condition,Infos,Used),lookup_values_if_vars(Used,LocalState,State,WaitVars),
     \+ ground(WaitVars),
     !,
     debug_println(9,unfolding_exists(Parameters)), print(unfolding(Parameters)),nl,
     %translate:print_bexpr(Condition),nl, print(State),nl, print(LocalState),nl,nl,
     NR=pred_true. %, print(set_res(pred_true)),nl.
     % if enabled it slows down: ./probcli examples/EventBPrologPackages/SET_Game/SET_GAM_Sym_NoSet20_mch.eventb -mc 10000000 -p SYMMETRY_MODE hash -p TIME_OUT 7000 -p CLPFD TRUE -df -goal "n=18" -p MAX_OPERATIONS 20 -strict -expcterr goal_found 
*/

b_test_exists(Parameters,Condition,Infos,LocalState,State,WF) :- 
        get_wait_flag0(WF,WF0), 
	 	b_test_exists(WF0,Parameters,Condition,Infos,LocalState,State,WF).
:- block b_test_exists(-,?,?, ?,?,?,?).
b_test_exists(_LWF,Parameters,Condition,Infos,LocalState,State,WF) :-
    get_exists_used_ids(Condition,Infos,Used),
     %  print(b_test_exists(Parameters)),nl, print(state(State)),nl,
    set_up_typed_localstate(Parameters,_FreshOutputVars,TypedVals,LocalState,NewLocalState),
    lookup_values_if_vars(Used,LocalState,State,WaitVars), % TO DO: we could use b_compiler instead
    create_inner_wait_flags(WF,LocalWF), % we are creating inner waitflags here as the exist will wait anyway; LocalWF shares just WFE flag with WF
    b_test_boolean_expression(Condition,NewLocalState,State,LocalWF), /* check Condition */
    %% TO DO: check if Condition is deterministic; if so we do not need to use delay_once_wait_vars
      %   print('tested_exists: '), print_bexpr(Condition),nl, % print(local(NewLocalState)),nl, %%
     %  print(waitvars(WaitVars)),nl,kernel_waitflags:portray_waitflags(LocalWF),nl,
    ground_det_wait_flag(LocalWF),  % print('det no contradiction'),nl,flush_output,
    b_enumerate:b_tighter_enumerate_values(TypedVals,LocalWF), % will not yet enumerate
    get_enumeration_starting_wait_flag(WF,ESWF),  %% TO DO: get flag when enumeration of infinite type starts
    when((nonvar(ESWF);ground(WaitVars)),b_enumerate_exists(WaitVars,Parameters,LocalWF,WF)). 

b_enumerate_exists(WaitVars,Parameters,LocalWF,WF) :-
    (ground(WaitVars)
      -> kernel_waitflags:get_idle_wait_flag(b_enumerate_exists,WF,LWF), 
         /* allow all other co-routines waiting on WaitVars to complete */
         b_enumerate_exists_aux_ground(LWF,WaitVars,Parameters,LocalWF,WF)
       ; b_enumerate_exists_aux_non_ground(WaitVars,Parameters,LocalWF,WF)
    ).

:- block b_enumerate_exists_aux_ground(-,?,?,?,?).
b_enumerate_exists_aux_ground(_,_WaitVars,_Parameters,LocalWF,_WF) :-
   /* Note: this does a local cut: so it is important
       that all conditions are fully evaluated before performing the cut ; otherwise
       solutions will be lost */
        %print_bt_message('ENUMERATE EXISTS'(_Parameters)),nl, % portray_waitflags(_WF),
         (kernel_waitflags:ground_constraintprop_wait_flags(LocalWF)
           -> true). % local cut: one solution is sufficient.

b_enumerate_exists_aux_non_ground(WaitVars,_Parameters,LocalWF,WF) :- 
         %%  print(b_test_exists_enumerating(_WaitVars,_Parameters,LocalWF)),nl,
         kernel_waitflags:ground_wait_flag_to(LocalWF,999999),
         % print(waiting_for_ground(WaitVars)),nl, portray_waitflags(LocalWF),nl,
         when(ground(WaitVars),b_enum_exists_with_cut(LocalWF,WF)).

b_enum_exists_with_cut(LocalWF,WF) :-
     kernel_waitflags:get_idle_wait_flag(b_enumerate_exists,WF,LWF),
     % allow all other co-routines waiting on WaitVars to complete
     b_enum_exists_with_cut_aux(LWF,LocalWF).
:- block b_enum_exists_with_cut_aux(-,?).
b_enum_exists_with_cut_aux(_LWF,LocalWF) :-
     % print(b_enum_exists_with_cut(LocalWF)),nl,
     (kernel_waitflags:ground_constraintprop_wait_flags(LocalWF)
           -> true).

get_exists_used_ids(Condition,Infos,Used) :-
    member(used_ids(Used),Infos) -> true
    ; otherwise ->
	     add_error(b_interpreter,
		  'Expected information of used identifiers in exists operation information',Infos),
		  bsyntaxtree:find_identifier_uses(Condition, [], Used).
% what if we call test_exists for a negated universal quantifier ??

is_an_equality(b(equal(A,B),_,_),A,B).

b_not_test_exists([OnePara],Condition,_Infos,LocalState,State,WF) :-
 % This case arises often when performing CBC deadlock checking with parameters  c = Expression
  % TO DO: generalise ?? 
   is_an_equality(Condition,LHS,RHS), get_texpr_id(OnePara,ID),
   (get_texpr_id(LHS,ID) -> Other=RHS ; get_texpr_id(RHS,ID) -> Other=LHS),
    print(equality_in_not_exists(ID,Other)),nl, 
    \+ occurs_in_expr(ID,Other),  % OCCURS CHECK
    !,   %  not(#ID.(ID=Other)) is equivalent to false
    (always_well_defined(Other) -> 
       fail  % equality will always be true 
       ; % compute Other, in case it is not well-defined
         % print(computing_other(Other)),nl,
         b_compute_expression(Other,LocalState,State,_IgnoredResult,WF)
    ).
/* the following causes performance problems with examples/B/Mathematical/GraphIso/CheckGraphIsomorphism 
   and with lausanne.mch in SMT mode */
b_not_test_exists(Parameters,Condition,_Infos,LocalState,State,WF) :- 
  preferences:preference(use_smt_mode,true),
  is_a_conjunct(Condition,LHS,RHS),
  (\+ ground(LocalState) ; \+ ground(State)), % TO DO: check if any of the variables in Condition are not ground
  !, 
  %print(translating_not_exists_into_forall),nl,print('  '),print_bexpr(LHS),nl, print('=> '),print_bexpr(RHS),nl,
  ((is_for_all_set_membership_predicate(RHS,Parameters,_Set,_Pattern,_ParameterValues),
    \+ is_for_all_set_membership_predicate(LHS,Parameters,_,_,_))
    -> LLHS=RHS, RRHS=LHS % swap LHS and RHS; TO DO: search for membership predicate inside LHS,RHS !!
       %, print(swap_lhs_rhs),nl
     ; LLHS=LHS, RRHS=RHS
  ),
  %print(state(State)),nl, print(local(LocalState)),nl,
  NegRHS = b(negation(RRHS),pred,[]),
  b_for_all(Parameters,LLHS,NegRHS,LocalState,State,WF).
b_not_test_exists(Parameters,Condition,_Infos,LocalState,State,_WF) :- 
   % print(not_exists),nl,
    % compile to remove dependency on unused parts of State+LocalState; will make it delay less below
    get_texpr_id_names(Parameters,ParaIDs),
     %%  print(b_not_test_exists(ParaIDs)),nl, print(state(LocalState,State)),nl, %%
    ~~mnf( b_compiler:b_compile(Condition,ParaIDs,LocalState,State,CompiledCond) ),
   %% print('compiled: '),print_bexpr(CompiledCond),nl, %%
    delay_not( ( 
                 b_interpreter:set_up_typed_localstate(Parameters,FreshOutputVars,TypedVals,[],NewLocalState),
                 kernel_waitflags:init_wait_flags(LocalWF), % we are not creating inner waitflags here as the exist will wait anyway
                 %   print(testing_not_exists_condition(Parameters,FreshOutputVars)),nl, %%
                 b_interpreter:b_test_boolean_expression(CompiledCond,NewLocalState,[],LocalWF),  % check Condition 
                 % print(enumerating(TypedVals)),nl,
                 b_enumerate:b_enumerate_values(TypedVals,LocalWF),
                 kernel_waitflags:ground_wait_flags(LocalWF)
                 %  , print(finished_enum_exists(TypedVals)),nl %%
              ), [LocalWF,TypedVals,NewLocalState|FreshOutputVars] ). 
               


   
/* --------------------------------- */

value_type(_X). % only used in post_conditions (assert_post)
/*
value_type(X) :- nonvar(X),
 (X = value(_) ; X = global_set(_) ; 
  X = global_constant(_) ; X = term(_)).*/


% look up identifiers in store, if the identifier
% is not in the store, assume it's a constant and ignore it
lookup_values_if_vars([],_,_,[]).
lookup_values_if_vars([Id|IRest],LocalStore,Store,Values) :-
	( lookup_value(Id,LocalStore,Store,Val) -> Values = [Val|VRest]
	; Values = VRest),
	lookup_values_if_vars(IRest,LocalStore,Store,VRest).


  

 /* will not generate error messages; used for CSP||B when we don't know if something is a B variable or not */
   /* TO DO: add peel_prefix for global_sets and constants ?? */
try_lookup_value_in_store_and_global_sets(Id,State,Val) :- 
 (lookup_value(Id,State,RVal) -> (RVal=Val) 
    ; (b_global_set(Id) -> (Val=global_set(Id))  /* global set evaluates to itself */
        ; (b_global_sets:lookup_global_constant(Id,VGC)
              -> (Val = VGC)
              ;  fail
          ))
  ).
  
  
:- assert_pre(b_interpreter:lookup_value_in_store_and_global_sets(Id,LS,State,_Val),
                (nonvar(Id),type_check(LS,store),type_check(State,store))).
:- assert_post(b_interpreter:lookup_value_in_store_and_global_sets(_Id,_LS,_State,_Val), true).

lookup_value_in_store_and_global_sets(Id,LocalState,State,Val) :-
  lookup_value_in_store_and_global_sets(Id,LocalState,State,Val,unknown).
lookup_value_in_store_and_global_sets(Id,LocalState,State,Val,Span) :- 
 (lookup_value(Id,LocalState,State,RVal) -> RVal=Val
    ; lookup_value_in_global_sets(Id,Val,LocalState,State,Span)
  ).  
  
lookup_value_in_global_sets(Id,Val,LocalState,State,Span) :- 
   (b_global_set(Id)
        -> Val=global_set(Id)  /* global set evaluates to itself */
        ; (b_global_sets:lookup_global_constant(Id,VGC)
              ->  Val = VGC
              ;   %translate_environment(LocalState,LS),
                  %translate_environment(State,S),
                  %nl,print(local(LS)),nl,  print(global(S)),nl,nl,
                  translate:translate_bstate(LocalState,LS), %context(
                  translate:translate_bstate(State,S),
                  ((member(bind(Id,_),LocalState) ; member(bind(Id,_),State))
                    -> Msg = 'Cannot determine value for identifier: '
                    ;  Msg = 'Cannot find identifier: '
                  ),
                  add_error(b_interpreter,Msg,Id:local(LS):global(S),Span),
                  fail % Val = term(Id)  /* Val=fail */
          )
   ).

%translate_environment([],[]) :- !.
%translate_environment([bind(V,_)|T],[V|TT]) :- !, translate_environment(T,TT). 
%translate_environment(A,A).

/* -----------------------------*/
/*    b_compute_expressions     */
/* -----------------------------*/

b_compute_expressions([], _, _, [],_WF).
b_compute_expressions([EXPRsHd|EXPRsTl],LocalState,InState,[ValHd|ValTl],WF) :-
   ~~pp_cll(b_compute_expression(EXPRsHd,LocalState,InState,ValHd,WF)),
   b_compute_expressions(EXPRsTl,LocalState,InState,ValTl,WF). 

/*
b_compute_expressions_symbolic([],[], _, _, [],_WF).
b_compute_expressions_symbolic([EXPRsHd|EXPRsTl],[SYMBOLIC|TS],LocalState,InState,[ValHd|ValTl],WF) :-
   (SYMBOLIC==symbolic -> EXPRsHd=b(E,T,I), EXPR = b(E,T,[prob_annotation('SYMBOLIC')|I])
     ; EXPR = EXPRsHd),
   ~~pp_cll(b_compute_expression(EXPR,LocalState,InState,ValHd,WF)),
   b_compute_expressions_symbolic(EXPRsTl,TS,LocalState,InState,ValTl,WF). 
*/

% a variation of above that delays computing the expressions until Waitflag 1/2 is set
b_compute_assign_expressions_when(EXPRsTl,LocalState,InState,ValTl,WF,OR) :-
   get_assign_expr_priority_wf(OR,WF,LWF),
   b_compute_expressions_when2(EXPRsTl,LocalState,InState,ValTl,WF,LWF).
b_compute_expressions_when2([], _, _, [],_WF,_).
b_compute_expressions_when2([EXPRsHd|EXPRsTl],LocalState,InState,[ValHd|ValTl],WF,LWF) :-
  b_compute_expression_when2(EXPRsHd,LocalState,InState,ValHd,WF,LWF),
  b_compute_expressions_when2(EXPRsTl,LocalState,InState,ValTl,WF,LWF). 

get_assign_expr_priority_wf(output_required,WF,LWF) :- !,
  % the Output of the statement is required; do not delay computing it
  get_wait_flag0(WF,LWF).
  %get_wait_flag(1,get_assign_expr_priority_wf,WF,LWF).
get_assign_expr_priority_wf(_,WF,LWF) :-
 % the Output is not directly required; compute it later (when we are sure that the statement actually succeeds)
  (preferences:preference(use_smt_mode,true)
   -> Prio = 2.0 % SMT Mode: we are probably doing CBC checking, we may have set up constraints on ValTl
   ; Prio=4000), % unless the expression is inside an sequential composition: nothing can depend on ValTl; TO DO: check if we are inside a sequential composition; if not use inf priority; otherwise use 2.0 or 4
   get_wait_flag(Prio,get_assign_expr_priority_wf,WF,LWF).
   % LHS/RHS of assignments do not affect constraint propagation (at least in normal model checking mode; in CBC or bounded MC this would be different) ??!!

:- block b_compute_expression_when2(?,?,?,?,?,-).
b_compute_expression_when2(EXPRsHd,LocalState,InState,ValHd,WF,_) :-
    %% translate:print_bexpr(EXPRsHd),nl, print(value(ValHd)),nl, frozen(ValHd,G), print('Frozen: '),print(G),nl, kernel_waitflags:portray_waitflags(WF), %%
    b_compute_expression(EXPRsHd,LocalState,InState,ValHd,WF).
/* -----------------------------*/
/*     b_execute_statement      */
/* -----------------------------*/
/* b_execute_statement(XMLStatements:input, StateofLocalVariables:input,
                      StateOfGlobalVariables:input,
                      ListOfUpdateBindingsForGlobalVariablesAndResultVariables:output) */

      

:- assert_pre(b_interpreter:b_execute_statement(Stmt,LS,IS,_OS,_WF,_Path,_OR),
	      (bsyntaxtree:check_if_typed_substitution(Stmt),type_check(LS,store),type_check(IS,store) )).
:- assert_post(b_interpreter:b_execute_statement(_Stmt,_LS,_IS,OS,_WF,_Path,_OR),
	      (type_check(OS,store))).

%b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path) :- !,
%  b_execute_statement(Body,LocalState,InState,OutState,WF,Path). % code below has still problems in INITIALISATION
b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,OR) :-
  get_wait_flag(2.0,inner_stmt,WF,LWF),
  clone_wait_flags_from1(WF,InnerWF),
  b_execute_inner_statement(Body,LocalState,InState,OutState,InnerWF,Path,LWF,OR).
:- block b_execute_inner_statement(?,?,?,?,?,?,-,?).
b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,_,OR) :-
   b_execute_statement(Body,LocalState,InState,OutState,WF,Path,OR),
   ground_wait_flag0(WF).



b_execute_statement(Stmt, LS, S, OS, WF,Path) :- 
   b_execute_statement(Stmt, LS, S, OS, WF,Path,output_not_required).
b_execute_statement(b(Stmt,_,Infos), LS, S, OS, WF,Path,OR) :- !,
    %%  print_term_summary(Stmt), %%
    b_execute_statement2(Stmt, Infos, LS, S, OS, WF,Path,OR).
b_execute_statement(Stmt, LS, S, OS, WF,Path,OR) :- 
  add_internal_error('Statement not properly wrapped: ',b_execute_statement(Stmt, LS, S, OS, WF,Path,OR)),
  b_execute_statement2(Stmt, [], LS, S, OS, WF,Path,OR).

% blocks should be removed by type-checker
b_execute_statement2(block(Stmt),_,LS,IS,OS,WF,Path,OR) :- !,
    b_execute_statement(Stmt,LS,IS,OS,WF,Path,OR).  %% NOT COVERED 
/*  Var1, ..., VarN := Expr1, ..., ExprN */
/*  also: Var1, fun1(arg), ... := Expr1, Expr2, ... */
b_execute_statement2(assign(LHS,Exprs),_,LocalState,InState,OutState,WF,assign,OR) :- !,
  %% print(assign(LHS,Exprs)),nl,
   b_compute_assign_expressions_when(Exprs,LocalState,InState,VALs,WF,OR),% VALs - a list of computed values 
   ~~mnf(b_assign_values_or_functions(LHS,VALs,LocalState,InState,IDs,NewValues,WF)),
   ~~mnf(store_values(IDs, NewValues, OutState)).
b_execute_statement2(assign_single_id(IDE,Expr),_,LocalState,InState,OutState,WF,assign,_OR) :- 
   get_texpr_id(IDE,ID),!,
   b_compute_expression(Expr,LocalState,InState,Value,WF), % b_compute_assign_expressions_when
   %b_compute_assign_expressions_when([Expr],LocalState,InState,[Value],WF),
   OutState=[bind(ID,Value)].
 b_execute_statement2(choice(ChoiceList),_,LocalState,InState,OutState,WF,Path,OR) :- !,
   /*  Choice LHS Or RHS Or ... End */
   get_wait_flag(2.0,choice,WF,WFC), % maybe use length of ChoiceList as Priority ??
   b_execute_choice(ChoiceList,LocalState,InState,OutState,WF,Path,WFC,OR).

b_execute_statement2(becomes_element_of(LHS,RHS),  /* LHS :: RHS */
   _,LocalState,InState,OutState,WF,becomes_element_of,_OR) :- !,
%% print_message('becomes_element_of'(RHS)),
   b_compute_expression(RHS,LocalState,InState,ValueSet,WF),
   %print(check_element_of_wf(Val,ValueSet)),nl,
   check_element_of_wf(Val,ValueSet,WF), 
      ground_det_wait_flag(WF),
   convert_pairs_into_list(LHS,Val,ValList),
   b_assign_values_or_functions(LHS,ValList,LocalState,InState,IDs,NewValues,WF),   
   store_values(IDs,NewValues,OutState).
  
b_execute_statement2(becomes_such(Vars,Condition),  /* Vars : Condition */
   _,LocalState,InState,OutState,WF,becomes_such,_OR) :- !,
   % create the LHS variables in the local state
   set_up_typed_localstate(Vars,Values,TypedVals,LocalState,TempLocalState1),
   get_texpr_ids(Vars,Ids),
   % store those values also the outgoing state
   store_values(Ids,Values,OutState),
   % there might be some variables of the form x$0 to access the
   % values before the substitution. We put them into the local state
   insert_before_substitution_variables(Vars,LocalState,InState,TempLocalState1,TempLocalState),
   LWF=WF,%% create_inner_wait_flags(WF,LWF), %%
   b_tighter_enumerate_values(TypedVals,LWF),
   ~~pp_cll(b_test_boolean_expression(Condition,TempLocalState,InState,LWF)),
 % print_message(calling_b_tighter_enumerate_values(VarList,VarTypes,OutState)),
   %ground_constraintprop_wait_flags(LWF). %%%
   true.
 % print_message(sol_found(VarList,OutState)). 
  
b_execute_statement2(sequence(Seq),   /* LHS ; RHS */
   _,LocalState,InState,OutUpdates,WF,Path,OR) :- !,
   ( Seq = [First|Rest] ->
       (Rest == [] -> b_execute_statement(First,LocalState,InState,OutUpdates,WF,Path,OR)
        ;  b_execute_sequence(First,sequence(Rest),LocalState,InState,OutUpdates,WF,Path,OR)
       )
   ; Seq = [] -> OutUpdates = [], Path=[]
   ; add_internal_error('Illegal argument to sequence: ',Seq),fail).
b_execute_statement2(precondition(PreCond,Body),_,LocalState,InState,OutState,WF,Path,OR) :- !,
      ground_det_wait_flag(WF),
  %print('*** Non-outermost PRE construct'),nl,
  (   ~~pp_cll(b_test_boolean_expression(PreCond,LocalState,InState,WF)),
         /* PRECONDITION SATISFIED */
      ground_det_wait_flag(WF),
      Path=pre(IPath),
      b_execute_inner_statement(Body,LocalState,InState,OutState,WF,IPath,OR)
   ;
      %print_quoted(b_not_test_boolean_expression(PreCond,LocalState,InState,WF)),
      b_not_test_boolean_expression(PreCond,LocalState,InState,WF),
          /* PRE-CONDITION VIOLATED */
      ground_det_wait_flag(WF),
      translate_bexpression(PreCond,PreString),
      add_abort_error_span(precondition_error,'Precondition violated: ',PreString,span_predicate(PreCond,LocalState,InState),WF),
      Path = pre_violated
  ).
b_execute_statement2(assertion(PreCond,Body),_,LocalState,InState,OutState,WF,Path,OR) :- !,
   get_last_wait_flag(assertion,WF,WFC),
  b_execute_assertion(PreCond,Body,LocalState,InState,OutState,WF,Path,WFC,OR).

b_execute_statement2(select(Whens),_, LocalState, InState, OutState,WF,select(Nr,Path),OR) :- !,
  % print(select(Whens)),nl,
   get_texpr_expr(TExpr,select_when(PreCond, Body)),
   nth1(Nr, Whens, TExpr),
   ~~pp_cll(b_test_boolean_expression(PreCond,LocalState,InState,WF)),
   ground_det_wait_flag(WF), % print(try_select_inner_statement(Nr)), translate:print_bexpr(PreCond),nl,
   b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,OR).

b_execute_statement2(select(Whens,Else), Infos, LocalState, InState, OutState,WF,Path,OR) :- !, 
  (%print(select_with_else),nl,
   b_execute_statement2(select(Whens), Infos, LocalState, InState, OutState,WF,Path,OR)
  ;
   findall(PC, 
      (member(SW,Whens),get_texpr_expr(SW,select_when(PC,_))), AllPreconds),
    % nl, print(select_else),nl, %%
    b_not_test_list_of_boolean_expression(AllPreconds,LocalState,InState,WF),
    Path = select(else,IPath),
    % print(exec_else),nl, portray_waitflags(WF),nl, %%
    b_execute_inner_statement(Else,LocalState,InState,OutState,WF,IPath,OR)
   % ,print(finished_inner_statement_else_SELECT),nl %, portray_waitflags(WF)
    ).

b_execute_statement2(case(Expression,Cases,Else), _, LocalState, InState, OutState,WF,case(Nr,Path),OR) :- !, 
 % According to the AtelierB handbook the cases must be distinct one-by-one; Schneider's book does not seem to require that
 % ProB will at the moment allow multiple entries in which case CASE is non-deterministic
 % Atelier-B will actually only allow literal integers, booleans, enumerated set values as cases !
   b_compute_expression(Expression,LocalState,InState,EValue,WF),
   b_compute_cases(Cases,LocalState,InState,CompCases,AllEitherValues,WF),
   ground_det_wait_flag(WF),
   ( nth1(Nr, CompCases, comp_case_or(ListOfValues, Body)),
	 member(EitherValue,ListOfValues),
	 equal_object(EValue,EitherValue),
     ground_det_wait_flag(WF),
	 b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,OR)
   ; b_not_member_of_value_lists(AllEitherValues,EValue),
     ground_det_wait_flag(WF),
     % print_message(else_case(Body2)),
     Nr=else,
	 b_execute_inner_statement(Else,LocalState,InState,OutState,WF,Path,OR)
	).

b_execute_statement2(var(Parameters,Body),   /* VAR _ IN _ */
   _,LocalState,InState,NewOutState,WF,var(Path),OR) :- !,
   /* introduces a new variable that can be assigned to */
   ~~mnf( set_up_undefined_localstate(Parameters,InState,NewInState) ),
   ~~mnf( split_names_and_types(Parameters,Names,_) ),
   ~~pp_mnf(store:delete_variables_from_state(Names,LocalState,NewLocalState)),
   b_execute_statement(Body,NewLocalState,NewInState,OutState,WF,Path,OR),
 %%print_message(var_res(OutState)),
   ~~mnf(get_results(Parameters,OutState,_ResultValues,NewOutState,false,WF)).
   /* it can be ok not to assign to a  local variable */
   
%TODO(DP, 5.8.2008):
b_execute_statement2(let(Parameters,Defs,Body),   /* LET _ BE _ IN _ */
   _,LocalState,InState,OutState,WF,let(Path),OR) :- !,
   set_up_typed_localstate(Parameters,LocalState,NewLocalState),
   ~~mnf(b_execute_let_definitions(Defs,NewLocalState,InState,WF)),
   b_execute_statement(Body,NewLocalState,InState,OutState,WF,Path,OR).
:- if(true=false).
b_execute_statement2(any(Parameters,PreCond,Body),   /* DEBUGGING VERSION OF ANY _ WHERE */
   _,LocalState,InState,OutState,WF,any(Bindings,Path),OR) :- !, %statistics(runtime,_),
    print('ANY: '), print_bexpr(PreCond),nl, %%
   set_up_typed_localstate(Parameters,FreshVars,TypedVals,LocalState,NewLocalState),
   create_any_bindings(Parameters,FreshVars,Bindings),
  % create_inner_wait_flags(WF,LWF),
  b_tighter_enumerate_values(TypedVals,WF),
   %~~pp_cll(b_test_boolean_expression(PreCond,NewLocalState,InState,WF)), /* check WHERE */
   % TO DO: maybe use component decomposition here !!?
   observe_state(TypedVals,WF,0), b_tracetest_boolean_expression(PreCond,NewLocalState,InState,WF,one), %% comment in for debugging
   print('ANY - BODY: '), print_subst(Body),nl, %%
   b_execute_statement(Body,NewLocalState,InState,OutState,WF,Path,OR),
   kernel_waitflags:portray_waitflags(WF).
:- else.
b_execute_statement2(any(Parameters,PreCond,Body),   /* ANY _ WHERE */
   _,LocalState,InState,OutState,WF,any(Bindings,Path),OR) :- !, %statistics(runtime,_),
   %print('ANY: '), print_bexpr(PreCond),nl, %%
   set_up_typed_localstate(Parameters,FreshVars,TypedVals,LocalState,NewLocalState),
   create_any_bindings(Parameters,FreshVars,Bindings),
  % create_inner_wait_flags(WF,LWF),
  b_tighter_enumerate_values(TypedVals,WF),
   ~~pp_cll(b_test_boolean_expression(PreCond,NewLocalState,InState,WF)), /* check WHERE */
  %% print(any_body(Body)),nl,
   b_execute_statement(Body,NewLocalState,InState,OutState,WF,Path,OR). /* should we use b_execute_inner_statement/LWF here ?? */
:- endif.
b_execute_statement2(skip,    /* SKIP */
   _,_LocalState,_InState,OutState,_WF,skip,_OR) :- !, OutState = [].
b_execute_statement2(parallel(Statements),  /*   LHS  || RHS  */ %@@@
   _,LocalState,InState,OutState,WF,parallel(Paths),OR) :-  !, 
   b_execute_statements_in_parallel(Statements,LocalState,InState,OutState,WF,Paths,OR).
b_execute_statement2(init_parallel(Statements),  /*   Like Parallel: but 
   writes error messages if a branch fails */ %@@@
   _,LocalState,InState,OutState,WF,parallel,OR) :-  !,  
   b_execute_initialisation_statements_in_parallel(Statements,LocalState,InState,OutState,WF,OR).   
b_execute_statement2(if(IfList),_,LocalState,InState,OutState,WF,if,OR) :- !, 
  %LWF=1,
  (preferences:preference(use_smt_mode,true) -> Prio = 2.0
  ; Prio=2000), % TO DO: check if there is a guard inside; if not we can wait much longer; unless we are in CBC mode
  get_wait_flag(Prio,if_else,WF,LWF),  
   b_execute_else_list(IfList,LocalState,InState,OutState,WF,LWF,OR).
b_execute_statement2(operation_call(Operation,Results,Parameters),_,LocalState,InState,OutState,WF,
                     operation_call(OperationName,ParamValues,ResultValues,InnerPath),OR) :- !,
   def_get_texpr_id(Operation,op(OperationName)),
   b_compute_expressions(Parameters,LocalState,InState,ParamValues,WF),
 %  print_message(b_execute_operation_with_parameters(OperationName,ParamValues)), %%
   /* NOT CLEAR: is it an error of this fails ?? */
 %  print(op_call_paras(ParamValues)),nl,
   ~~pp_mnf(b_execute_operation_with_parameters(OperationName,InState,ParamValues,OpOutState,ResultValues,InnerPath,WF,OR)), 
 % print_message(result_values(OperationName,ParamValues,ResultValues,InnerPath)), %%
   split_names_and_types(Results,ResultingIDs,_),
   store_values(ResultingIDs, ResultValues, ResultsOutState),
  %  print(store_values(ResultingIDs, ResultValues, ResultsOutState)),nl,
   combine_updates(ResultsOutState,OpOutState,OutState).
           
b_execute_statement2(while(COND,STMT,INV,VARIANT),
                    _,LocalState,InState,OutState,WF,while,_OR) :- !,
  %print_message(while(LocalState,InState)),
  %profiler:add_profile_hit(while(COND,WF),2),
   b_compute_expression(VARIANT,LocalState,InState,VarVal,WF),
   %print_message(execute_while(LocalState,InState)),
   b_execute_while(COND,STMT,INV,VARIANT,none,VarVal,LocalState,InState,OutState,WF).
b_execute_statement2(while1(COND,STMT,INV,VARIANT,PrevVariantValue),
                    _,LocalState,InState,OutState,WF,while,_OR) :- !,  /* second time in loop; we now have the
                    previous value of the variant againts which we can compare */
 % print_message(while1(PrevVariantValue,LocalState,InState)),
  %profiler:add_profile_hit(while1(COND,PrevVariantValue,WF),2),
   b_compute_expression(VARIANT,LocalState,InState,VarVal,WF),
   b_execute_while1(COND,STMT,INV,VARIANT,PrevVariantValue,VarVal,LocalState,InState,OutState,WF).
b_execute_statement2(external_subst_call(FunName,Args),Info,LocalState,State,OutState,WF,external(FunName),_OR) :- 
      !, b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF),
      % print('EXTERNAL SUBST CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_substitution(FunName,EvaluatedArgs,State,OutState,Info,WF).
b_execute_statement2(E,I,LS,S,O,WF,Path,OR) :-
   add_internal_error('Uncovered statement: ',b_execute_statement2(E,I,LS,S,O,WF,Path,OR)),
   print_functor(E),nl,fail.

% b_execute_sequence(FirstStmt, UnwrappedRest,..)
b_execute_sequence(First,UnwrappedRest,LocalState,InState,OutUpdates,WF,Path,OR) :-
          % print('===> LHS: '), translate:print_bexpr(First),nl,
          Path = sequence(Path1,RestPath),
          b_execute_statement(First,LocalState,InState,LHSUpdates,WF,Path1,output_required), 
          ground_det_wait_flag(WF), % maybe better to delay next execute_statement ?
          % print('===> RHS: '), translate:print_bexpr(b(UnwrappedRest,subst,[])),nl,
          store_intermediate_updates(InState,LHSUpdates,IntermediateState),
          % print(store_intermediate_updates(InState,LHSUpdates,IntermediateState)),nl,
          b_execute_statement_when_state_set(UnwrappedRest,LocalState,IntermediateState,
                                              LHSUpdates,OutUpdates,WF,RestPath,OR).

create_any_bindings([],[],[]).
create_any_bindings([Param|Prest],[Value|Vrest],[bind(Id,Value)|Brest]) :-
    get_texpr_id(Param,Id),
    create_any_bindings(Prest,Vrest,Brest).

% insert the values of variables before the substitution under an
% alternative name. The alternative name is given as an optional extra
% information in the variable (see type-checking of becomes_such for
% details). Usually (i.e. always) the alternative name is x$0 for the variable x.
insert_before_substitution_variables([],_LocalState,_InState,State,State).
insert_before_substitution_variables([Var|VRest],LocalState,InState,StateA,StateB) :-
	get_texpr_info(Var,Infos),
	( member(before_substitution(_,Alternative),Infos) ->
	    def_get_texpr_id(Var,Id),
	    lookup_value(Id,LocalState,InState,Value),
	    store_value(Alternative,Value,StateA,StateC)
	; otherwise ->
	    StateA = StateC),
	insert_before_substitution_variables(VRest,LocalState,InState,StateC,StateB).


:- block b_execute_while(?,?,?,?,?,-, ?,?,?, ?).
b_execute_while(COND,STMT,INV,VARIANT,PrevVariantValue,int(VarVal),LocalState,InState,OutState,WF) :- 
    %get_last_wait_flag(while(VarVal),WF,LWF),
   ground_constraintprop_wait_flags(WF),
    when(ground(InState),
        b_execute_while_loop(COND,STMT,INV,VARIANT,PrevVariantValue,VarVal,LocalState,InState,OutState,WF)).

:- block b_execute_while1(?,?,?,?,?,-,?,?,?,?).
b_execute_while1(COND,STMT,INV,VARIANT,PrevVariantValue,int(VarVal),LocalState,InState,OutState,WF) :- 
   %  get_last_wait_flag(while(VarVal),WF,LWF),
   ground_constraintprop_wait_flags(WF),
   b_execute_while_loop(COND,STMT,INV,VARIANT,PrevVariantValue,VarVal,LocalState,InState,OutState,WF).

% treat the bool(.) operator
b_convert_bool(Pred,LocalState,State,WF,PredRes) :-
    get_wait_flag0(WF,WF0), % first give a chance for PredRes to be instantiated
    b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,WF0).

:- block b_convert_bool_wf0(?,?,?,?,-,-).
b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,_) :- PredRes==pred_true,!,
    %% print('bool(.)=TRUE '), translate:print_bexpr(Pred),nl, %%
    b_test_boolean_expression(Pred,LocalState,State,WF).
b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,_) :- PredRes==pred_false,!,
    %% print('bool(.)=FALSE '), translate:print_bexpr(Pred),nl, %%
    b_not_test_boolean_expression(Pred,LocalState,State,WF).
b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,_) :-
    %% print('bool(.) check : '), translate:print_bexpr(Pred),nl, obsv(LocalState),
    b_try_check_boolean_expression_lwf(Pred,LocalState,State,WF,PredRes,'$GENERATE_ON_DEMAND' ).
    %% print('finished bool(.) check : '), translate:print_bexpr(Pred),nl,translate:print_bstate(LocalState),nl.
    
%%obsv([]).
%%obsv([bind(Var,Val) | T] ) :- when(ground(Val), format("~w = ~w~n",[Var,Val])), obsv(T).

% try to do check_boolean expression; if not possible create choice point if LWF is grounded
b_try_check_boolean_expression_lwf(Pred,LocalState,InState,WF,PredRes,LWF) :-
    %print('Compile: '), translate:print_bexpr(Pred),nl,
    b_compiler:b_compile(Pred,[],LocalState,InState,CPRED), % this increases chance that check can be applied+avoids multiple computations in two branches
    b_try_check_boolean_expression_lwf_c(CPRED,LocalState,InState,WF,PredRes,LWF).
b_try_check_boolean_expression_lwf_c(Pred,LocalState,InState,WF,PredRes,LWF) :-
    copy_wf_start(WF,CWF),
    if(b_check_boolean_expression(Pred,LocalState,InState,CWF,Res),
     % Note: we do not use a cut here ! usually b_check will not enumerate; but on rare occasions it can (apply function in inverse,...)
         %print(check(Res)), print(' : '), translate:print_bexpr(Pred),nl,
         (PredRes=Res, copy_wf_finish(WF,CWF)),
         b_try_check_failed(Pred,LocalState,InState,WF,PredRes,LWF)
      ).
b_try_check_failed(Pred,LocalState,InState,WF,PredRes,LWF) :-
    %print(choice_point(LWF)),nl, translate:print_bexpr(Pred),nl,
    (LWF=='$GENERATE_ON_DEMAND' -> get_binary_choice_wait_flag('$GENERATE_ON_DEMAND',WF,WF1) ; WF1=LWF),
    b_try_check_boolean_expression_lwf_block(Pred,LocalState,InState,WF,PredRes,WF1).

:- block b_try_check_boolean_expression_lwf_block(?,?,?, ?,-,-).
b_try_check_boolean_expression_lwf_block(Pred,LocalState,InState,WF,pred_true,_) :-
    % print(' TRUE +++++> '), translate:print_bexpr(Pred),nl,
    b_test_boolean_expression(Pred,LocalState,InState,WF).
b_try_check_boolean_expression_lwf_block(Pred,LocalState,InState,WF,pred_false,_) :-
    % print(' FALSE +++++> '), translate:print_bexpr(Pred),nl,
    b_not_test_boolean_expression(Pred,LocalState,InState,WF).

% try to do check_boolean expression; if not possible create choice point
b_try_check_boolean_expression(Pred,LocalState,InState,WF,PredRes) :-
    b_compiler:b_compile(Pred,[],LocalState,InState,CPRED), % this increases chance that check can be applied+avoids multiple computations in two branches
    b_try_check_boolean_expression_c(CPRED,LocalState,InState,WF,PredRes).
b_try_check_boolean_expression_c(Pred,LocalState,InState,WF,PredRes) :-
    copy_wf_start(WF,CWF),
    if(b_check_boolean_expression(Pred,LocalState,InState,CWF,Res),
        (PredRes=Res,copy_wf_finish(WF,CWF)),
        b_try_check_boolean_expression_c2(Pred,LocalState,InState,WF,PredRes)).
b_try_check_boolean_expression_c2(Pred,LocalState,InState,WF,pred_true) :-
    %profiler:add_profile_hit(b_try_check_failed(Pred),3),
    b_test_boolean_expression(Pred,LocalState,InState,WF).
b_try_check_boolean_expression_c2(Pred,LocalState,InState,WF,pred_false) :-
    b_not_test_boolean_expression(Pred,LocalState,InState,WF).

:- block b_execute_while_loop(?,?,?,?,-,?,?,?,?,?), b_execute_while_loop(?,?,?,?,?,-,?,?,?,?).
b_execute_while_loop(COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF) :- 
        b_try_check_boolean_expression(INV,LocalState,InState,WF,INVRes),
        %ground_constraintprop_wait_flags(WF), % seems to slow down interpretation
        b_check_execute_while_loop1(INVRes,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF).

:- block b_check_execute_while_loop1(-,?,?,?,?,?,?,?,?,?,?).
b_check_execute_while_loop1(pred_false,_COND,_STMT,INV,_VAR,_PrevVarVal,VarVal,LocalState,InState,_OutState,WF) :-
        add_abort_error_span(while_invariant_violation,
         'While INVARIANT VIOLATED, current VARIANT value: ',VarVal,span_predicate(INV,LocalState,InState),WF).
b_check_execute_while_loop1(pred_true,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF) :-
        b_try_check_boolean_expression(COND,LocalState,InState,WF,CondRes),
        %ground_constraintprop_wait_flags(WF), % seems to slow down interpretation
        b_check_execute_while_loop2(CondRes,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF).

:- block b_check_execute_while_loop2(-,?,?,?,?,?,?,?,?,?,?).
b_check_execute_while_loop2(pred_false,_COND,_STMT,_INV,_VARIANT,_PrevVarVal,_VarVal,_LS,_InState,OutState,_WF) :-
        OutState=[].
b_check_execute_while_loop2(pred_true,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF) :-
        
        %profiler:add_profile_hit(b_execute_while_body(STMT,VarVal,PrevVarVal,LocalState),2),
        b_execute_while_body(COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF).

:- block b_execute_while_body(?,?,?,?,-,?,?,?,?,?), b_execute_while_body(?,?,?,?,?,-,?,?,?,?).
% VarVal and PrevVaraintValue known
b_execute_while_body(_COND,_STMT,_INVARIANT,VAR,_PV,VarVal,LocalState,InState,_OutState,WF) :- 
      VarVal<0, !, % more efficient than is_not_natural(int(VarVal)),
      ground_constraintprop_wait_flags(WF),
      add_abort_error_span(while_variant_error,'While VARIANT not NATURAL:',VarVal,span_predicate(VAR,LocalState,InState),WF).
b_execute_while_body(_COND,_STMT,_INVARIANT,VAR,PrevVariantValue,VarVal,LocalState,InState,_OutState,WF) :- 
      % check no longer required VarVal>=0,%is_natural(int(VarVal),WF),
      PrevVariantValue \= none,
      PrevVariantValue =< VarVal, !, %less_than_equal(int(PrevVariantValue),int(VarVal)),
      ground_constraintprop_wait_flags(WF),
      add_abort_error_span(while_variant_error,'While VARIANT not decreasing:',(VarVal,PrevVariantValue),span_predicate(VAR,LocalState,InState),WF).
b_execute_while_body(COND,STMT,INV,VARIANT,_PrevVariantValue,VarVal,LocalState,InState,OutState,WF) :- 
      % check no longer required due to cuts above %VarVal>=0, %is_natural(int(VarVal),WF),
      %(PrevVariantValue==none -> true  ; less_than(int(VarVal),int(PrevVariantValue))),
      %ground_constraintprop_wait_flags(WF), % seems to slow down interpretation
 %      create_texpr(while1(COND,STMT,INV,VARIANT,VarVal),subst,[],TWhile),
      S2 = while1(COND,STMT,INV,VARIANT,VarVal), % was [TWhile]
      b_execute_sequence(STMT,S2,LocalState,InState,OutState,WF,_Path,output_required).

b_execute_else_list([],_,_State,[],_WF,_LWF,_OR). /* skip */  
b_execute_else_list([TExpr|Rest],LocalState,InState,OutState,WF,LWF,OR) :-
  get_texpr_expr(TExpr,if_elsif(Test,Body)),
  % get_enumeration_finished_wait_flag(WF,LWF),
  InnerWF=WF, %clone_wait_flags_from1(WF,InnerWF),
  b_try_check_boolean_expression_lwf(Test,LocalState,InState,InnerWF,PredRes,LWF),
  %ground_wait_flag0(InnerWF),
  b_execute_else_list2(PredRes,Body,Rest,LocalState,InState,OutState,InnerWF,LWF,OR).
 
:- block b_execute_else_list2(-,?,?,?, ?,?,?,-,?). % should we also unblock if LWF set ? needed for seq. composition to set up state ??
b_execute_else_list2(pred_true,Body,_,LocalState,InState,OutState,WF,_,OR) :- 
    % print('IF TEST: '),print_subst(Body), nl,
  b_execute_statement(Body,LocalState,InState,OutState,WF,_Path,OR).
b_execute_else_list2(pred_false,_Body,Rest,LocalState,InState,OutState,WF,LWF,OR) :-
    % print('IF - ELSE NEG TEST: '), print(LWF),nl, print(LocalState),nl, print(InState),nl,
  b_execute_else_list(Rest,LocalState,InState,OutState,WF,LWF,OR).


:- block b_execute_choice(?,?,?,?,?,?,-,?).
b_execute_choice(ChoiceList,LocalState,InState,OutState,WF,choice(Nr,Path),_WFC,OR) :-
   nth1(Nr,ChoiceList,Choice),
   b_execute_statement(Choice,LocalState,InState,OutState,WF,Path,OR).
   
:- block b_execute_assertion(?,?,?,?,?,?,?,-,?).
b_execute_assertion(PreCond,Body,LocalState,InState,OutState,WF,Path,_,OR) :- 
   (  ~~pp_cll(b_test_boolean_expression(PreCond,LocalState,InState,WF)),
         /* ASSERT SATISFIED */
      ground_det_wait_flag(WF),
      Path = assertion(IPath),
      b_execute_statement(Body,LocalState,InState,OutState,WF,IPath,OR)
   ;
      b_not_test_boolean_expression(PreCond,LocalState,InState,WF),
          /* ASSERT VIOLATED */
      ground_det_wait_flag(WF),
      Path = assertion_violated,
      translate_bexpression(PreCond,PreString),
      add_abort_error_span(assert_error,'ASSERT violated: ',PreString,span_predicate(PreCond,LocalState,InState),WF)
      %,store_value(abort__flag__info,abort(assertion_violated(PreCond)),[],OutState)
  ).


:- block check_state_skeleton_bound(-,?).
check_state_skeleton_bound([],true).
check_state_skeleton_bound([bind(Var,_)|T],Res) :-
  check_state_skeleton_bound_aux(Var,T,Res).
  
:- block check_state_skeleton_bound_aux(-,?,?).
check_state_skeleton_bound_aux(Var,T,Res) :-
   (atomic(Var) -> true ; print('*** variable name not atomic: '), print(Var),nl),
   check_state_skeleton_bound(T,Res).


b_execute_statement_when_state_set(Stmt,LS,State,USoFar,Res,WF,Path,OR) :-
   check_state_skeleton_bound(State,Bound),
   b_execute_statement_when_state_set2(Stmt,LS,State,USoFar,Res,WF,Path,Bound,OR).
:- block b_execute_statement_when_state_set2(?,?,?,?,?,?,?,-,?).
b_execute_statement_when_state_set2(Stmt,LS,State,UpdatesSofar,Res,WF,Path,_,OR) :-
   b_execute_statement2(Stmt,[],LS,State,StmtUpdates,WF,Path,OR),
   ~~pp_mnf(store:merge_updates(UpdatesSofar,StmtUpdates,Res)).
 
store_values([], [], []).
store_values([IDsHd|IDsTl], [VALsHd|VALsTl], [OutStateHd|OutStateTl]) :-
  store_value(IDsHd,VALsHd,[],[OutStateHd]),
  %store_values_check_abort(VALsHd,IDsHd),
  store_values(IDsTl, VALsTl, OutStateTl).

%:- block store_values_check_abort(-,?).
%store_values_check_abort(abort(AbortMsg),IDsHd) :-
%    !, add_message(b_interpreter,'Assignment to variable generated abort:',abort(IDsHd,AbortMsg)).
%store_values_check_abort(_,_).
  

b_compute_cases(Cases,LS,S,CompCases,Values,WF) :-
	b_compute_cases2(Cases,LS,S,CompCases,Values,[],WF).
b_compute_cases2([],_,_,[],AllValues,AllValues,_).
b_compute_cases2([CaseOr|Crest],LS,S,[comp_case_or(Values,Subst)|CCrest],AllValues,VR,WF) :-
	get_texpr_expr(CaseOr,case_or(Exprs,Subst)),
	b_compute_cases3(Exprs,LS,S,Values,AllValues,VI,WF),
	b_compute_cases2(Crest,LS,S,CCrest,VI,VR,WF).
b_compute_cases3([],_,_,[],Values,Values,_).
b_compute_cases3([Expr|Erest],LS,S,[Value|Vrest],[Value|ValuesI],Values,WF) :-
	b_compute_expression(Expr,LS,S,Value,WF),
	b_compute_cases3(Erest,LS,S,Vrest,ValuesI,Values,WF).

b_not_member_of_value_lists([],_).
b_not_member_of_value_lists([EitherValue|Rest],Value) :-
    not_equal_object(Value,EitherValue),
    b_not_member_of_value_lists(Rest,Value).
    
    /*  the B syntax  does allow multiple definitions in a LET: */

b_execute_let_definitions(Pred,LocalState,State,WF) :- 
   is_a_conjunct(Pred,LHS,RHS),!,
   b_execute_let_definitions(LHS,LocalState,State,WF),
   b_execute_let_definitions(RHS,LocalState,State,WF).
b_execute_let_definitions(b(equal(LHS,RHS),_,_),LocalState,State,WF) :- !,
   b_compute_expression(LHS,LocalState,State,LHSVal,WF),
   b_compute_expression(RHS,LocalState,State,RHSVal,WF),
   % TO DO: ideally we should check if LHS is a declared identifier and is only defined once
   kernel_objects:equal_object_optimized(LHSVal,RHSVal).
b_execute_let_definitions(P,_L,_S,_WF) :-
   translate:translate_bexpression_with_limit(P,70,TP),
   add_error(illegal_let,'Let contains a predicate which is not an equality: ',TP,P),fail.


/* -----------------------------------------*/
/*             b_assign_function            */ 
/* -----------------------------------------*/

b_assign_values_or_functions([],[],_LS,_S,[],[],_WF).
b_assign_values_or_functions([TExpr|ERest],[Value|VRest],LS,S,[Id|IRest],[New|NRest],WF) :-
    get_texpr_expr(TExpr,Expr),
    b_assign_value_or_function(Expr,Value,LS,S,Id,New,WF),
    b_assign_values_or_functions(ERest,VRest,LS,S,IRest,NRest,WF).

b_assign_value_or_function(identifier(Id),Value,_LS,_S,Id,Value,_WF).
b_assign_value_or_function(function(Fun,Arg),Value,LocalState,InState,Id,FullValue,WF) :-
    %b_compute_expression(Arg,LocalState,InState,ArgVal,WF),
    %expand_global_sets_but_not_closures(ArgVal,EArgVal),
    b_compute_assign_expressions_when([Arg,Fun],LocalState,InState,[ArgVal,FunVal],WF,output_required), % or should we pass the OR info around ?
    /* we do not examine local state: these vars cannot be modified by operations !? */
    override(FunVal,ArgVal,Value,NewFun,WF), /* <---- */
    get_texpr_expr(Fun,Lhs),
    b_assign_value_or_function(Lhs,NewFun,LocalState,InState,Id,FullValue,WF).
   	
    

/* -----------------------------------------*/
/*     b_execute_statements_in_parallel     */ 
/* -----------------------------------------*/
/* execute a parallel composition of statements */
:- assert_pre(b_interpreter:b_execute_statements_in_parallel(Ss,LS,State,_Os,WF,_Paths,_OR),
                (ground_check(Ss),type_check(LS,store),
                 type_check(State,store),type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_execute_statements_in_parallel(_Ss,_LS,_State,Os,WF,_Paths,_OR),
                (type_check(Os,store),type_check(WF,wait_flag))).
                 
b_execute_statements_in_parallel([],_,_,[],_WF,[],_OR).
b_execute_statements_in_parallel([Stmt|TS],LocalState,InState,OutState,WF,[Path1|TPath],OR) :-
  % print('PAR: '),translate:print_subst(Stmt),nl,
  b_execute_statement(Stmt,LocalState,InState,OutState1,WF,Path1,OR),
  combine_updates(OutState1,OutState2,OutState), /* merge updates */
  b_execute_statements_in_parallel(TS,LocalState,InState,OutState2,WF,TPath,OR).
   
:- block combine_updates(-,?,?). % can also be used to merge states
combine_updates([],X,R) :- !,R=X.
combine_updates([H|T],Y,Res) :- !, Res=[H|Z], combine_updates(T,Y,Z).
combine_updates(X,Y,Z) :- add_internal_error('Illegal call: ',combine_updates(X,Y,Z)),fail.
  
  
:- assert_pre(b_interpreter:b_execute_initialisation_statements_in_parallel(Ss,LS,State,_Os,_WF,_OR),
                (ground_check(Ss),type_check(LS,store),type_check(State,store))).
:- assert_post(b_interpreter:b_execute_initialisation_statements_in_parallel(_Ss,_LS,_State,Os,_WF,_OR),
                (type_check(Os,store))).

b_execute_initialisation_statements_in_parallel([],_,_,[],_WF,_OR).
b_execute_initialisation_statements_in_parallel([Stmt|TS],LocalState,InState,OutState,WF,OR) :-
  (preferences:get_preference(provide_trace_information,true)
    -> nl,print(' =INIT=> '),translate:print_subst(Stmt),nl ; true),
  if(b_execute_initialisation_statement(Stmt,LocalState,InState,OutState1,WF,OR),
     (
      (preferences:get_preference(provide_trace_information,true) -> print(' [=OK=]'),nl ; true),
      b_execute_initialisation_statements_in_parallel(TS,LocalState,InState,OutState2,WF,OR),
      combine_updates(OutState1,OutState2,OutState)
     ),
     (translate:translate_substitution(Stmt,TStmt),
      add_error(initialisation_fails,'Initialisation statement fails: ',TStmt), fail,
      b_execute_initialisation_statements_in_parallel(TS,LocalState,InState,OutState,WF,OR)
      )
    ). 

b_execute_initialisation_statement(b(parallel(Statements),_,_),
				    LocalState,InState,OutState,WF,OR) :-  !,
    print('.'), flush_output(user_output),
    ~~pp_mnf(b_execute_initialisation_statements_in_parallel(Statements,LocalState,InState,OutState,WF,OR)).
b_execute_initialisation_statement(Stmt,LS,In,Out,_OUTERWF,OR) :-
    init_wait_flags(WF), % each initialisation can be executed in isolation
    b_execute_statement(Stmt,LS,In,Out,WF,_Path,OR),ground_wait_flags(WF).

% replace certain parallels by init_parallel for better error reporting in INITIALISATION
translate_init_parallel(b(Stmt,T,I),b(TStmt,T,I)) :- %print_term_summary(Stmt),
  tp2(Stmt,TStmt).
tp2(parallel(S),init_parallel(S)) :- !, debug_println(9,converting_init_parallel).
tp2(assign(LHS,RHS),init_parallel(S)) :- LHS=[_,_|_],
   parallelize_assign(LHS,RHS,S),!, debug_println(9,converting_assign_init_parallel).
tp2(sequence(A),sequence(LA)) :- !,l_translate_parallel(A,LA).
tp2(X,X).
l_translate_parallel([],[]).
l_translate_parallel([H|T],[LH|LT]) :- translate_init_parallel(H,LH), l_translate_parallel(T,LT).

parallelize_assign([],[],[]).
parallelize_assign([V|VT],[RHS|RT],[b(assign([V],[RHS]),subst,[])|PT]) :- parallelize_assign(VT,RT,PT).
/* ---------------------------------------*/
/*     b_execute_top_level_statement      */
/* ---------------------------------------*/

/* this is just like b_execute_statement with the
 only difference that a PRE-condition is treated like
 a select, i.e., it does not generate an abort if used
 outside of its condition */
b_execute_top_level_statement(TExpr,LocalState,InState,OutState,WF,Path,OR) :-
   get_texpr_expr(TExpr,Stmt),
   b_execute_top_level_statement2(Stmt,TExpr,LocalState,InState,OutState,WF,Path,OR).
b_execute_top_level_statement2(precondition(PreCond,Body),_,
   LocalState,InState,OutState,WF,Path,OR) :- 
   preferences:preference(treat_outermost_pre_as_select,true),!,
   %debug_println(9,'Top level PRE treated as SELECT'),
   ~~pp_cll(b_test_boolean_expression(PreCond,LocalState,InState,WF)),
  % b_tracetest_boolean_expression(PreCond,LocalState,InState,WF,one),  % comment in for debugging
   b_execute_statement(Body,LocalState,InState,OutState,WF,Path,OR).
b_execute_top_level_statement2(_,TExpr,LocalState,InState,OutState,WF,Path,OR) :- 
  b_execute_statement(TExpr,LocalState,InState,OutState,WF,Path,OR).

  
  
/* -----------------------------*/
/*      b_execute_operation     */
/* -----------------------------*/
/* execute a single operation */

try_get_op_name(FullOperation,_Name) :- var(FullOperation),!.
try_get_op_name('-->'(FO,_),Name) :- !,try_get_op_name(FO,Name).
try_get_op_name(FullOperation,Name) :- functor(FullOperation,Name,_).

     
b_get_machine_operation_for_animation('$initialise_machine',[],Parameters,Body,OType,_TopLevel) :- 
     b_get_machine_variables(DeclaredVars), Parameters=DeclaredVars,
     b_get_initialisation_from_machine(Body,OType).
b_get_machine_operation_for_animation(Name,Results,Parameters,Body,OType,TopLevel) :- 
    preference(show_eventb_any_arguments,EVENTB),
    b_get_machine_operation(Name,Results,RealParameters,RealBody,OType),
    ((Results=[],TopLevel==true,EVENTB=true, RealParameters=[],
      RealBody=b(any(Parameters,PreCond,ABody),Type,Loc))
     % Then pretend the any parameters are parameters of the event
      -> Body = b(select([b(select_when(PreCond,ABody),Type,Loc)]),Type,Loc)
        % , print(faked(Name,Parameters,Body)),nl
      ;  Parameters=RealParameters,Body=RealBody
    ).



:- assert_pre(b_interpreter:b_execute_top_level_operation_update(_Name,_FullOperation,InState,_NewState,_P),
                (ground_check(InState),type_check(InState,store))).
:- assert_post(b_interpreter:b_execute_top_level_operation_update(Name,_FullOperation,_InState,NewState,_Path),
                (type_check(NewState,store),ground_check(Name))).

 :- if(\+ environ(partially_evaluate, true)).
 
 b_execute_top_level_operation_update(Name,FullOperation,InState,Updates,Path) :-
    b_execute_top_level_operation_update_wf(Name,FullOperation,InState,Updates,Path,WF),
    ground_wait_flags(WF).
    %store:store_updates_and_normalise(Updates,InState,NewState).
 
 :- else.
/* partial evaluation enabled */

 b_execute_top_level_operation_update(Name,FullOp,InState,Updates,Path) :-
    %print(compile_op(Name,FullOp)),nl,
    % THIS SHOULD BE DONE BY AN ANNOTATION: start_specialization(p(static,dynamic,...)) rather than by hand
    generalize_call(list(term(bind,[atomic,dynamic])),InState,StateSkel),
    partially_evaluate_and_load(b_execute_top_level_operation_update_wf(Name,FullOp,StateSkel,Updates,Path,WF),ResCall),
    StateSkel=InState, call(ResCall),
    ground_wait_flags(WF).
:- endif.

b_execute_top_level_operation_update_wf(Name,FullOperation,InState,Updates,Path,WF) :-
    init_wait_flags(WF),
    try_get_op_name(FullOperation,Name), b_top_level_operation(Name),
    b_execute_operation2(Name,FullOperation,InState,Updates,_,_,true,Path,WF,output_not_required).

% for use by CBC:
b_execute_top_level_operation_wf(Name,FullOperation,InState,NewState,Path,WF) :-
    try_get_op_name(FullOperation,Name), b_top_level_operation(Name),
    b_execute_operation2(Name,FullOperation,InState,NewOutState,_,_,true,Path,WF,output_required),
     % we should ensure that Operation part is executed to instantiate NewOutState skeleton
    store:lazy_store_updates_and_normalise(NewOutState,InState,NewState).
   % ~~pp_mnf(store:store_updates_and_normalise(NewOutState,InState,NewState)).
    /* actually execute the updates
     to the global store computed by b_execute_statement */





:- assert_pre(b_interpreter:b_execute_operation_with_parameters(Name,InState,Para,_NewState,_RV,_P,_WF,_OR),
                (ground_check(InState),type_check(InState,store),list_skeleton(Para),
                  atomic(Name))).
:- assert_post(b_interpreter:b_execute_operation_with_parameters(Name,_,_,NewState,RV,_Path,_WF,_OR),
                (type_check(NewState,store),ground_check(Name),list_skeleton(RV))).
b_execute_operation_with_parameters(Name,InState,Parameters,OutState,ResultValues,Path,WF,OR) :-
     (b_get_machine_operation(Name,_,_,_)
      -> b_execute_operation2(Name,_,InState,OutState,Parameters,ResultValues,false,Path,WF,OR)
      ;  add_internal_error('Unknown B operation, cannot call: ',Name),fail
     ).

b_execute_operation2(Name,Operation,InState,Updates,ParaValues,ResultValues,_TopLevel,TransInfo,_WF,_OR) :-
    lookup_cached_transitions(Name,InState,Solutions),!,
    member(transition(ParaValues,ResultValues,Updates,TransInfo),Solutions),
    create_full_operation(Name,ResultValues,ParaValues,ResultValues,Operation).

b_execute_operation2(Name,Operation,InState,NewOutState,ParaValues,ResultValues,TopLevel,TransInfo,WF,OR) :-
    b_get_machine_operation_for_animation(Name,Results,Parameters,Body,OType,TopLevel),
    b_execute_operation3(OType,Name,Operation,
			 InState,NewOutState,
			 Body,Parameters,ParaValues,Results,ResultValues,TopLevel,TransInfo,WF,OR).	

create_full_operation(Name,Results,ParaValues,ResultValues,FullOperation) :-
    Operation =.. [Name|ParaValues],
    ( Results == [] -> /* we have an ordinary operation */
        FullOperation = Operation
    ; FullOperation = '-->'(Operation,ResultValues)
    ).

b_execute_operation3(classic,Name,Operation,InState,NewOutState,
		             Body,Parameters,ParaValues,Results,ResultValues,TopLevel,[path(Path)],WF,OR) :-
    %print('Attempting Operation ----> '),print(Name),nl,
    empty_state(EmptyState),
    set_up_typed_localstate(Parameters,ParaValues,ParamTypedVals,EmptyState,LocalState), 
    l_expand_and_normalise_values(ParaValues,NormalisedParaValues,Parameters),
    create_full_operation(Name,Results,NormalisedParaValues,
                          NormalisedResultValues,Operation),
    %%  statistics(runtime,[Time1,_]), %%%
    set_up_undefined_localstate(Results,InState,NewInState), 
     /* to avoid error messages with sequential composition (store_intermediate_updates) */
    check_additional_guard(LocalState,NewInState,WF),
    b_tighter_enumerate_values(ParamTypedVals,WF), 
   (TopLevel==true
    -> b_execute_top_level_statement(Body,LocalState,NewInState,OutState,WF,Path,OR)
    ;  b_execute_statement(Body,LocalState,NewInState,OutState,WF,Path,OR)
   ),
       /* first look in OutState then in InState */
  %% statistics(runtime,[Time2,_]), Time is Time2-Time1, %%%
  %  print_message(operation(Time,Name)), %%%
     % print_bt_message('Operation Successful ------> '),
     ((preferences:preference(treat_outermost_pre_as_select,true),
       preferences:preference(require_operations_to_assign_all_outputs,true))
       -> ReportError = true ; ReportError = false),
     get_results(Results,OutState,ResultValues,NewOutState,ReportError,WF),
     % print_bt_message(l_expand_and_normalise_values(ResultValues)),
     l_expand_and_normalise_values(ResultValues,NormalisedResultValues,Results)
  %%   ,print_message(paras_results(ParaValues,NormalisedResultValues)) %%%
     .
b_execute_operation3(eventb(ChangeSet,ParaValues,Operation),
		     _Name,Operation,  InState,Updates,
		     TBody,Parameters,ParaValues,[],[],true,TransInfo,_WF,OR) :-
    %debug_println(9,attempting(Name)),
    % set up environment
    copy_variable_store(InState,ChangeSet,Updates,FullOutState),
    b_event(TBody,Parameters,ParaValues,InState,FullOutState,TransInfo).  % note: WF not passed

% additional guards can be set by the user to constrain/direct the execution of operations
check_additional_guard(LocalState,State,WF) :-
   (bmachine:b_machine_temp_predicate(TempPred)
      /* an additional precondition typed in by the user */
     -> print(checking),nl,
        print_bexpr(TempPred),nl,
       ~~pp_cll(b_test_boolean_expression(TempPred,LocalState,State,WF))
        ,print(checked),nl
     ;  true
   ).

add_additional_guard(OldGuard,NewGuard) :-
   (bmachine:b_machine_temp_predicate(TempPred)
     -> conjunct_predicates([TempPred,OldGuard],NewGuard)
     ;  NewGuard = OldGuard
   ).

get_results(Results,OutState,ResultValues,NewOutState,ReportError,WF) :-
  (Results==[] -> (ResultValues,NewOutState) = ([],OutState)
    ; get_results1(Results,OutState,ResultValues,NewOutState,ReportError,WF)).
:- block get_results1(?,-,?,?,?,?).
get_results1([],OutState,[],OutState,_ReportError,_WF).
get_results1([R|Results],OutState,[NRV|ResultValues],NewOutState,ReportError,WF) :-
    def_get_texpr_id(R,ResultId),
    ~~mnf(lookup_and_delete_value(ResultId,RV,OutState,OutState2,ReportError,WF,Finished)),
    get_results2(Finished,Results,RV,NRV,OutState2,ResultValues,NewOutState,ReportError,WF).
:- block get_results2(-,?,?,?,?,?,?,?,?).    
get_results2(_,Results,RV,NRV,OutState2,ResultValues,NewOutState,ReportError,WF) :-    
%     print_message(get_result(ResultId,RV,OutState)), %%
    normalise_value(RV,NRV),
    get_results1(Results,OutState2,ResultValues,NewOutState,ReportError,WF).


  
  /* ------------------------------------------------------------------ */
  /*                       SET UP CONSTANTS                             */
  /* ------------------------------------------------------------------ */
  
:- use_module(static_ordering).

:- volatile no_solution_found_for_constants/0, unsat_properties_component/1,
 unsat_properties_boolean_expression/2.
:- dynamic no_solution_found_for_constants/0, unsat_properties_component/1,
 unsat_properties_boolean_expression/2.

reset_unsat_component_infos :- 
  retractall(unsat_properties_component(_)),
  retractall(unsat_properties_boolean_expression(_,_)).
% store information about unsatisfiable components for later usage
assert_unsat_component_infos :-
  unsat_component(X),  % from b_interpreter_components
  assert(unsat_properties_component(X)),fail.
assert_unsat_component_infos :-
  unsat_boolean_expression(X,BE), % from b_interpreter_components
  assert(unsat_properties_boolean_expression(X,BE)),fail.
assert_unsat_component_infos.

tcltk_unsatisfiable_components_exist :- unsat_component(_),!.

%tcltk_unsatisfiable_components(list(UnsatComponents)) :-
%  findall(C,unsat_properties_component(C),UnsatComponents), UnsatComponents\=[].

get_property_components(Components) :-
    b_machine_has_constants_or_properties,
    b_get_properties_from_machine(Properties),
    bsyntaxtree:predicate_components(Properties,Components),!.
get_property_components([]).

tcltk_quick_describe_unsat_properties(Descr) :- 
   findall(TP, (unsat_properties_boolean_expression(CompNr,P),
               translate_bexpression_with_limit(P,TrP),
               ajoin([TrP,'; == false;'],TP)),
               %ajoin(['Component:',CompNr,'; ',TrP,'; == false;'],TP)),
               D1),
   get_property_components(Components),
   findall(['; * Component ',CompNr,' : ',PS|DescrVars], (unsat_component(CompNr),
                   \+ unsat_properties_boolean_expression(CompNr,_),
                    nth1(CompNr,Components,component(Pred,Vars)),
                    describe_constants(Vars,DescrVars),
                    translate_bexpression_with_limit(Pred,70,PS) ),
               UC),
   append(UC,UCL),
   (D1=[]
    -> Descr = D2,
       (UCL=[] -> D2=['The properties were satisfiable'] ; D2 = ['The following components were unsatisfiable: '|UCL])
    ; Descr = ['The following properties directly led to an inconsistency:;'|D12],
      (UCL=[] -> D2=[] ; D2 = [';In addition the following components were unsatisfiable: '|UCL]),
      append(D1,D2,D12)).

describe_constants([],R) :- !,R=[].
describe_constants(Vars,[';   over identifiers : '|Vars]).

b_set_up_concrete_constants(NormalisedConstantsState) :- 
    reset_unsat_component_infos,
    retractall(no_solution_found_for_constants),assert(no_solution_found_for_constants),
    % preference(allow_incomplete_partial_setup_constants,AllowSkipping),
    reset_component_info(true),
    statistics(runtime,_), statistics(walltime,_),%%
    b_machine_has_constants_or_properties,
    b_get_machine_constants(UnsortedConstants),!,
    b_get_properties_from_machine(MProperties),
    add_additional_guard(MProperties,Properties),
    
    project_onto_static_assertions(Properties,UnsortedConstants,PProperties,PUnsortedConstants),
    
    (preferences:preference(use_static_ordering,true)
     -> static_ordering:sort_ids_by_usage(PUnsortedConstants,PProperties,Constants)
        %,reverse(Constants1,Constants)
        %,print(sorted_by_usage(Constants)),nl
     ;  Constants =  PUnsortedConstants),
    
    set_up_typed_localstate(Constants,_FreshVars,TypedVals,[],ConstantsState),
    load_partial_constants(ConstantsState,PProperties,LPProperties),  % check if some values have been computed and stored in a file
      
   %%  print_bt_message(checking_properties), %%
    b_trace_test_components(LPProperties,ConstantsState,TypedVals),
    
    (unsat_component(_) -> (!,fail) ; true), % otherwise some expressions were skipped; use partial_set_up
 
    statistics(runtime,[_,Time]),
    statistics(walltime,[_,WTime]),
    print_bt_message(found_enumeration_of_constants(Time,WTime)), %%
    % print(ConstantsState),nl,
    normalise_store(ConstantsState,NormalisedConstantsState),
     % print_bt_message(finished_normalising),
    retractall(no_solution_found_for_constants). /* ensure that b_partial_set_up_concrete_constants fails */
b_set_up_concrete_constants([]).

:- volatile project_properties_on_identifiers/1.
:- dynamic project_properties_on_identifiers/1.
set_projection_on_static_assertions(ALL) :-
  % set interpreter in mode which projects out all constants not needed
  % for static assertion checking; all variables are also projected out
   retractall(project_properties_on_identifiers(_)),
   (b_get_assertions(ALL,dynamic,[]) -> true
     ; nl,print('*** WARNING: DYNAMIC ASSERTIONS NOT EMPTY'),nl,nl),
   b_get_assertions(ALL,static,L),
   (L=[] -> nl,print('*** WARNING: NO STATIC ASSERTIONS ***'),nl,nl ; true),
   conjunct_predicates(L,Conj),
   (debug_mode(on) -> print('Static Assertion: '),translate:print_bexpr(Conj),nl ; true),
   bsyntaxtree:predicate_identifiers(Conj,IDs),
   (debug_mode(on) -> print('IDS: '), print(IDs),nl ; true),
   assert(project_properties_on_identifiers(IDs)).

project_onto_static_assertions(Properties,Constants,NewProperties,NewConstants) :-
    %% set_projection_on_static_assertions(main), %% main or all
    project_properties_on_identifiers(ProjIDs),
    bsyntaxtree:project_predicate_on_identifiers(Properties,ProjIDs,NewProperties,PIDs),
    debug_println(9,projecting(PIDs)),
    remove_constants(Constants,PIDs,NewConstants),!. % constants is a typed identifier list.
project_onto_static_assertions(Properties,Constants,Properties,Constants).

remove_constants([],_,[]).
remove_constants([TypedID|T],ProjIDS,Res) :-
   get_texpr_id(TypedID,ID),
   (member(ID,ProjIDS) -> Res = [TypedID|PT] ; Res = PT),
   remove_constants(T,ProjIDS,PT).
   
	
/* can be called if b_set_up_concrete_constants failed; will partially set up the state using values found in first ground value propagation phase */
b_partial_set_up_concrete_constants(NormalisedConstantsState) :- 
    assert_unsat_component_infos,
    no_solution_found_for_constants,
    b_get_machine_constants(Constants),
    Constants \= [],
    (det_solution_for_constant(_,_) -> true ; fail),
    set_up_typed_localstate(Constants,[],ConstantsState),
    print('Solutions found for the following constants:'),nl,
    fill_in_det_solutions(ConstantsState,FC),
    (preference(allow_incomplete_partial_setup_constants,true) -> true
      ; print('All constants were valued.'),nl),
    normalise_store(FC,NormalisedConstantsState).
   
   %extend_state(Constants,NormalisedConstantsState,UndefinedVals),
   %append(NormalisedConstantsState,UndefinedVals, FullNormalisedConstantsState).

fill_in_det_solutions([],[]).
fill_in_det_solutions([bind(Var,Val)|T],[bind(Var,FVal)|FT]) :- %print(bind(Var,Val)),nl,
  (det_solution_for_constant(Var,Val) ->   print_term_summary(bind(Var,Val)),FVal=Val
     ; print(no_sol_for(Var)),nl,
       preference(allow_incomplete_partial_setup_constants,true),
       FVal = abort(could_not_determine_value(Var))
  ),
  fill_in_det_solutions(T,FT).

%extend_state([],_,[]).
%extend_state([b(identifier(Val),_Type,_I)|TRest],C,R) :- 
%   (member(bind(Val,_),C) -> R=TR ; R=[bind(Val,abort(could_not_determine_value))|TR]),
%   extend_state(TRest,C,TR).
   
 

/* --------------------------------- */
  
:- use_module(b_state_model_check, [b_check_valid_state/1]).

b_initialise_machine(ConstantsState,NormalisedInitialVarsState,FullInitialState,Path) :-
   if(b_initialise_machine2(ConstantsState,NormalisedInitialVarsState,FullInitialState,Path),
      true,
      %(add_message(initialisation_fails,'INITIALISATION FAILS'),fail)).
      (add_error(initialisation_fails,'INITIALISATION FAILS'),fail)).
    
b_initialise_machine2(ConstantsState,NormalisedInitialVarsState,FullInitialState,Path) :- 
/* InitialVarsState is an ouput */
   b_get_initialisation_from_machine(InitStatement,OType),
   ( InitStatement=[] ->
	   Path = [],
       add_message(b_initialise_machine,'Machine has no INITIALISATION statement(s)!'),
       NormalisedInitialVarsState = [],
       FullInitialState = ConstantsState
   ; OType == classic ->
	   Path = [path(PathInfo)],
       b_get_machine_variables(DeclaredVars),
       set_up_undefined_localstate(DeclaredVars,ConstantsState,NewInState),
       /* this is if the INITIALISATION has a sequential composition; to avoid error messages */
       init_wait_flags(WF),
       translate_init_parallel(InitStatement,InitStatement2),
       b_execute_statement(InitStatement2,[],NewInState,InitialVarsState,WF,PathInfo),
    %  print_message(init(InitialVarsState)), %%
       b_check_that_only_declared_variables(InitialVarsState,CorrectedState),
       normalise_store(CorrectedState,NormalisedInitialVarsState),
       combine_updates(ConstantsState,NormalisedInitialVarsState,FullInitialState),
       split_names_and_types(DeclaredVars,DeclaredNames,DeclaredTypes),
       check_additional_guard([],InitialVarsState,WF),
    %  print_message(grounding_constraint_waitflags(InitialVarsState)), %%
       ground_wait_flags(WF),
       
       b_enumerate_values_in_store(DeclaredNames,DeclaredTypes,_,InitialVarsState,WF)  %% MOVED HERE TO AVOID that target values are enumerated before initialisation effect computed
   ; OType = eventb(_ChangeSet,ParaValues,_Operation) ->
       prepare_eventb_initialisation(ConstantsState,InitialVarsUnNormState,FullInitialUnNormState),
       get_texpr_expr(InitStatement,
                      rlevent(_Name,_Sec,_Stat,Parameters,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_AbsEv)),
       b_event(InitStatement,Parameters,ParaValues,ConstantsState,FullInitialUnNormState,Path),
       normalise_store(InitialVarsUnNormState,NormalisedInitialVarsState),
       normalise_store(FullInitialUnNormState,FullInitialState) % a bit of redundant work here; we do the work twice and try to normalise constants again !
   ),
   b_check_valid_state(FullInitialState).

prepare_eventb_initialisation(ConstantsState,InitialVarsState,FullInitialState) :-
    b_get_machine_variables(Variables),
    empty_state(Empty),
    set_up_typed_localstate(Variables,_,_,Empty,InitialVarsState),
    % NormalisedInitialVarsState = InitialVarsState, % TO DO : check that b_event normalises this
    combine_updates(ConstantsState,InitialVarsState,FullInitialState).



b_check_that_only_declared_variables(InitialVarsState,CorrectedState) :-
  b_get_machine_variables(DeclaredVars),
  split_names_and_types(DeclaredVars,Names,_Types),
  b_check_only_vars(InitialVarsState,Names,[],CorrectedState).

:- block b_check_only_vars(-,?,?,?).
b_check_only_vars([],_,_,[]).
b_check_only_vars([bind(VarID,Val)|T],Decl,PrevVars,Res) :-
	b_check_only_vars2(VarID,Val,T,Decl,PrevVars,Res).
:- block b_check_only_vars2(-,?,?,?,?,?).
b_check_only_vars2(VarID,Val,T,Decl,PrevVars,Res) :-
   ( member(VarID,Decl)
     -> (member(VarID,PrevVars) 
         -> add_error(b_initialise_machine,'INITIALISATION assigns twice to a VARIABLE: ',VarID),
             Res=RT
          ; Res = [bind(VarID,Val)|RT]
        )
      ; add_error(b_initialise_machine,
        'The INITIALISATION assigns to a variable which is not declared as a VARIABLE of the machine: ',VarID),
         Res = RT
   ),
   b_check_only_vars(T,Decl,[VarID|PrevVars],RT).


reset_b_interpreter :-
  reset_partial_evaluator,
  retractall(project_properties_on_identifiers(_)),
  reset_unsat_component_infos,
   custom_explicit_sets:reset_memo_table,
  retractall(unsat_properties_component(_)),
  retractall(no_solution_found_for_constants).
