% (c) 2004-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(bsets_clp,
   [empty_sequence/1, 
    is_sequence/2,  is_sequence_wf/3, not_is_sequence/2, not_is_sequence_wf/3,
    not_is_non_empty_sequence/2,
    injective_sequence_wf/3,
    not_injective_sequence/3,
    not_non_empty_injective_sequence/3,
    injective_non_empty_sequence/3,
    finite_non_empty_sequence/3,
    permutation_sequence_wf/3,
    not_permutation_sequence/3,
    not_surjective_seq/3,
    size_of_sequence/3,
    prepend_sequence/4, append_sequence/4, prefix_sequence/3,
    suffix_sequence/4, concat_sequence/4,
    disjoint_union_wf/4,
    concatentation_of_sequences/3, 
    tail_sequence/4, first_sequence/4, front_sequence/4, last_sequence/4,
    rev_sequence/3, 
    
    
   % maplet/3,
   % relation/1,
    relation_over/3, relation_over_wf/4,
    not_relation_over/4, 
    domain_wf/3, 
    
    range_wf/3,
    identity_relation_over_wf/3, in_identity/3, not_in_identity/3,
    invert_relation_wf/3,
    tuple_of/3, 
    rel_composition/3, rel_composition_wf/4,
    direct_product_wf/4,
    parallel_product/3,
    rel_iterate/4,
    
    
    not_partial_function/4,
    partial_function/3, partial_function_wf/4,
    
    total_function/3, total_function_wf/4,
    
    enumerate_total_bijection/3,
    total_bijection/3, total_bijection_wf/4,
    
    not_total_function/4,
    not_total_bijection/4,
    
    
    range_restriction/3, range_subtraction/3,
    domain_restriction/3, domain_subtraction/3,
    override_relation/4,  
    image_wf/4, image_for_closure1_wf/4,
    element_of_closure1_wf/3, not_element_of_closure1_wf/3,
    in_domain_wf/3, not_in_domain/2,
    apply_to/4, apply_to/5,
    override/5,
    
    sum_over_range/2, mul_over_range/2,
    
    disjoint_union_generalized_wf/3,
    
    partial_surjection/3, not_partial_surjection_wf/4,
    total_relation_wf/4,
    not_total_relation_wf/4,

    surjection_relation_wf/4, total_surjection_relation_wf/4,
    not_surjection_relation_wf/4, not_total_surjection_relation_wf/4,
    
    total_surjection/3, total_surjection_wf/4,
    not_total_surjection_wf/4,
    
    partial_injection/3, partial_injection_wf/4,
    not_partial_injection/4,
    
    total_injection/3, total_injection_wf/4,
    not_total_injection/4,
    
    partial_bijection/3, partial_bijection_wf/4,
    not_partial_bijection/4,
    
    relational_trans_closure/2, relational_reflexive_closure/2
   ]).


%:- print(loading_bsets_clp),nl.
%:- use_module(library(clpfd)).
%portray_message(informational, _).
:- use_module(library(terms)).
:- use_module(self_check).

:- use_module(debug).
:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides more advanced operations for the basic datatypes of ProB (mainly for relations, functions, sequences).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').
:- use_module(tools_printing,[print_term_summary/1]).

:- use_module(delay).

%:- use_module(labeling).

:- use_module(typechecker).
:- use_module(error_manager).

:- use_module(kernel_objects).
:- use_module(kernel_records).
      
:- use_module(kernel_waitflags).

%:- use_module(b_enumerate,[b_enumerate_values/1]).
:- use_module(custom_explicit_sets).

/* --------- */
/* SEQUENCES */
/* ------- - */

:- assert_must_succeed((bsets_clp:empty_sequence([]))).
:- assert_must_fail((bsets_clp:empty_sequence([int(1)]))).
empty_sequence(X) :- empty_set(X).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_empty_sequence([(int(2),int(33)),(int(1),int(22))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_empty_sequence([(int(2),int(33))]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_empty_sequence([]))).
not_empty_sequence(X) :-
   nonvar(X),is_custom_explicit_set(X,not_empty_sequence),!, is_non_empty_explicit_set(X).
not_empty_sequence([(int(_),_)|_]).


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:is_sequence([(int(1),int(22))],[int(22)]))).
:- assert_must_succeed(bsets_clp:is_sequence(closure(['_zzzz_unit_tests'],[couple(integer,integer)],b(member(b(identifier('_zzzz_unit_tests'),couple(integer,integer),[generated]),b(value([(int(1),int(22))]),set(couple(integer,integer)),[])),pred,[])),[int(22)])).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:is_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:is_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:is_sequence([(int(1),int(33)),(int(0),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:is_sequence([(int(3),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed((is_sequence(R,global_set('Name')),R = [])).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed((is_sequence([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name')) )).
:- assert_must_succeed((is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_fail((is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_fail((is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(3))] )).
:- assert_must_fail((is_sequence(R,global_set('Name')),
                        R = [(int(0),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(X,global_set('Name')),
                        kernel_objects:enumerate_basic_type(X,seq(global('Name'))),
                        X = [(int(1),fd(1,'Name'))])).

is_sequence(X,Type) :- init_wait_flags(WF),
  is_sequence_wf(X,Type,WF),
  ground_wait_flags(WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_domain([int(1),int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_domain([int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_domain([],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:is_sequence_domain([int(0)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:is_sequence_domain([int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_is_sequence_domain([int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_is_sequence_domain([int(0)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_is_sequence_domain([int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_is_sequence_domain([],WF),WF)).

% check if a set is the domain of a sequence, i.e., an interval 1..n with n>=0
:- use_module(custom_explicit_sets,[construct_interval_closure/3]).
:- block is_sequence_domain(-,?).
is_sequence_domain(Domain,WF) :-
   kernel_objects:finite_cardinality_as_int(Domain,int(Max),WF),
   %print(size(Max)),nl,
   construct_interval_closure(1,Max,Interval), equal_object(Domain,Interval).
:- block not_is_sequence_domain(-,?).
not_is_sequence_domain(Domain,WF) :-
   kernel_objects:finite_cardinality_as_int(Domain,int(Max),WF),
   %print(size(Max)),nl,
   construct_interval_closure(1,Max,Interval), not_equal_object(Domain,Interval).

:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_wf([(int(1),pred_true)],
[pred_true,pred_false],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_wf([(int(1),pred_true),(int(2),pred_false),(int(3),pred_true)],
[pred_true,pred_false],WF),WF)).
:- assert_must_succeed((bsets_clp:is_sequence_wf([(int(X),R)],[pred_true],_WF),X==1, R==pred_true)).
:- assert_must_succeed((bsets_clp:is_sequence_wf([(int(X),R),(int(Y),R)],[pred_true],_WF),X=2,
  (preferences:preference(use_clpfd_solver,true) -> Y==1 ; Y=1), R==pred_true)).

is_sequence_wf(Seq,Range,WF) :- is_sequence_wf_ex(Seq,Range,WF,_).
% is_sequence_wf_ex also returns expansion; if it was done
:- block is_sequence_wf_ex(-,?,?,?).
is_sequence_wf_ex(FF,Range,WF,FF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    % print_term_summary(checking_sequence(FFDomain)),nl, %%
    is_sequence_domain(FFDomain,WF).
is_sequence_wf_ex(Seq,Range,WF,avl_set(AER)) :-
   expand_and_convert_to_avl_set(Seq,AER),!,
   is_avl_sequence(AER),
   is_avl_relation_over_range(AER,Range,WF).
is_sequence_wf_ex(X,Type,WF,EX) :- %writeq(is_sequence(X,Type)),nl,
     expand_custom_set_to_list(X,EX,_,is_sequence_wf_ex),
     is_sequence2(EX,[],Type,0,WF).
:- block is_sequence2(-,?,?,?,?).
is_sequence2([],IndexesSoFar,_Type,Size,_WF) :- contiguous_set_of_indexes(IndexesSoFar,Size).
   /* not very good to do the checking at the end; can we move part of the checking earlier ? */
is_sequence2([(int(Idx),X)|Tail],IndexesSoFar,Type,Size,WF) :- 
    less_than_direct(0,Idx), %is_index_greater_zero(Idx),
    not_element_of_wf(int(Idx),IndexesSoFar,WF),
    check_element_of_wf(X,Type,WF), S1 is Size+1,
    is_sequence2(Tail,[int(Idx)|IndexesSoFar],Type,S1,WF).

:- block contiguous_set_of_indexes(-,?).
contiguous_set_of_indexes([],_).
contiguous_set_of_indexes([H|T],Size) :- contiguous_set_of_indexes1(T,H,Size).

:- block contiguous_set_of_indexes1(-,?,?).
contiguous_set_of_indexes1([],int(1),_).
contiguous_set_of_indexes1([int(H2)|T],int(H1),Size) :- less_than_equal_direct(H1,Size),
   less_than_equal_direct(H2,Size), less_than_equal_indexes(T,[H1,H2],Size).
/*  
   Indexes = [H1,H2|T],
   when(ground(Indexes),(sort(Indexes,Sorted),contiguous_set_of_indexes2(Sorted))).   
contiguous_set_of_indexes2([]).
contiguous_set_of_indexes2([int(1)|T]) :- contiguous_set_of_indexes3(T,1).
contiguous_set_of_indexes3([],_).
contiguous_set_of_indexes3([int(N)|T],N1) :- N is N1+1, contiguous_set_of_indexes3(T,N).
*/

less_than_equal_indexes([],All,_) :- clpfd_interface:clpfd_alldifferent(All).
less_than_equal_indexes([int(H)|T],All,Size) :- less_than_equal_direct(H,Size),less_than_equal_indexes(T,[H|All],Size).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(2),int(7)),(int(4),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(2),int(7)),(int(3),int(8))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(2),int(6)),(int(3),int(7)),(int(4),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(0),int(7)),(int(2),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_is_sequence([(int(1),int(22))],[int(22)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_is_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_is_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_is_sequence([(int(3),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),R = [])).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(2,'Name'))] )).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_fail((not_is_sequence([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name')) )).
:- assert_must_fail((not_is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_succeed((not_is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_succeed((not_is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(3))] )).
                        
                        
not_is_sequence(X,Type) :- init_wait_flags(WF),
  not_is_sequence_wf(X,Type,WF),
  ground_wait_flags(WF).
  
:- block not_is_sequence_wf(-,?,?).
not_is_sequence_wf(FF,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
     % print_term_summary(not_checking_sequence(FFDomain)),nl, %
    not_is_sequence_domain(FFDomain,WF).
not_is_sequence_wf(Seq,Range,_WF) :-
   expand_and_convert_to_avl_set(Seq,AER),!, %print_term_summary(not_is_sequence_wf(Seq,Range,WF)),
   (is_avl_sequence(AER) -> is_not_avl_relation_over_range(AER,Range)
      ; true).
not_is_sequence_wf(X,Type,WF) :- expand_custom_set_to_list(X,EX,_Done,not_is_sequence_wf),
    not_is_sequence2(EX,[],Type,WF).

:- block not_is_sequence2(-,?,?,?).
not_is_sequence2([],IndexesSoFar,_,_WF) :- not_contiguous_set_of_indexes(IndexesSoFar).
not_is_sequence2([(int(Idx),X)|Tail],IndexesSoFar,Type,WF) :- 
    membership_test(IndexesSoFar,int(Idx),MemRes),
    not_is_sequence3(MemRes,Idx,X,Tail,IndexesSoFar,Type,WF).

:- block not_is_sequence3(-,?,?,?,?,?,?).
not_is_sequence3(pred_true,_Idx,_X,_Tail,_IndexesSoFar,_Type,_WF).
not_is_sequence3(pred_false,Idx,_X,_Tail,_IndexesSoFar,_Type,_WF) :- nonvar(Idx),Idx<1,!.
not_is_sequence3(pred_false,Idx,X,Tail,IndexesSoFar,Type,WF) :- 
      membership_test(Type,X,MemRes),
      not_is_sequence4(MemRes,Idx,Tail,IndexesSoFar,Type,WF).
      
:- block not_is_sequence4(-,?,?,?,?,?).
not_is_sequence4(pred_false,_Idx,_Tail,_IndexesSoFar,_Type,_WF).
not_is_sequence4(pred_true,Idx,Tail,IndexesSoFar,Type,WF) :- 
     not_is_sequence2(Tail,[int(Idx)|IndexesSoFar],Type,WF).

not_contiguous_set_of_indexes(Indexes) :-
   when(ground(Indexes),(sort(Indexes,Sorted),not_contiguous_set_of_indexes2(Sorted,1))).
not_contiguous_set_of_indexes2([int(N)|T],N1) :- 
       when(?=(N,N1),
          ((N \= N1) ;  (N=N1, N2 is N1+1, not_contiguous_set_of_indexes2(T,N2)))).



    
    
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_is_non_empty_sequence([(int(1),int(22))],[int(22)]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_is_non_empty_sequence([(int(1),int(2))],[int(22)]))).
:- assert_must_succeed((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),R = [])).
:- assert_must_fail((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(4),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(1,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_is_non_empty_sequence(R,[int(1),int(2)]),
                        R = [(int(1),int(2)),(int(2),int(3))] )).
:- block not_is_non_empty_sequence(-,?).
not_is_non_empty_sequence(X,Type) :-
      empty_sequence(X) ; not_is_sequence(X,Type).

    
    
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_sequence_wf([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_sequence_wf([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:injective_sequence_wf([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:injective_sequence_wf([(int(2),int(22)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_sequence_wf([],global_set('Name'),WF),WF)).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
            kernel_waitflags:ground_det_wait_flag(WF), R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
            ground_det_wait_flag(WF),  R = [(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
           ground_det_wait_flag(WF), R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
             ground_det_wait_flag(WF), R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:injective_sequence_wf([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name'),WF),WF) ).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_fail((bsets_clp:injective_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_fail((bsets_clp:injective_sequence_wf(R,[int(1),int(2)],WF),
                       ground_det_wait_flag(WF), R = [(int(2),int(2)),(int(1),int(3))] )).
                        
  

:- block injective_sequence_wf(-,?,?).
injective_sequence_wf(Seq,Type,WF) :- /* corresponds to iseq */
   expand_and_convert_to_avl_set(Seq,AER),!,
   %print_term_summary(inj_avl_seq(AER,Type)),
   is_avl_sequence(AER),
   is_injective_avl_relation(AER,_ExactRange), % Should we check _ExactRange <: Type ??
   is_avl_relation_over_range(AER,Type,WF).
injective_sequence_wf(Seq,Type,WF) :- 
  expand_custom_set_to_list(Seq,ESeq,_,injective_sequence_wf),
  is_sequence_wf(ESeq,Type,WF),injective_sequence2(ESeq,[],Type,WF).
  
:- block injective_sequence2(-,?,?,?).
injective_sequence2([],_,_Type,_WF).
injective_sequence2([(int(_),X)|Tail],SoFar,Type,WF) :- 
    check_element_of_wf(X,Type,WF),
    not_element_of_wf(X,SoFar,WF),
    add_new_element(X,SoFar,SoFar2),
    injective_sequence2(Tail,SoFar2,Type,WF).
 


:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_injective_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_injective_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_injective_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_injective_sequence([(int(2),int(22)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),_WF),R = [])).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(1),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),int(2)),(int(1),int(1))] )).                       
:- assert_must_succeed((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_injective_sequence([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name'),WF),
                        ground_det_wait_flag(WF) )).
:- assert_must_succeed((bsets_clp:not_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_succeed((bsets_clp:not_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),int(2)),(int(1),int(3))] )).
:- block not_injective_sequence(-,?,?), not_injective_sequence(?,-,?).
not_injective_sequence(Seq,Type,WF) :- nonvar(Seq),expand_and_convert_to_avl_set(Seq,AER),!,
    %print_term_summary(not_inj_seq(Seq,Type,WF)),
    (\+ is_avl_sequence(AER) -> true
     ; is_injective_avl_relation(AER,ExactRange) -> not_subset_of_wf(ExactRange,Type,WF)
     ; true).    
not_injective_sequence(Seq,Type,WF) :- /* corresponds to Iseq */
    %get_middle_wait_flag(not_injective_sequence,WF,LWF),
    when(ground(Seq),not_injective_sequence1(Seq,Type,WF)).
%:- block not_injective_sequence1(?,?,?,-).
not_injective_sequence1(Seq,Type,WF) :-    
   expand_custom_set_to_list(Seq,ESeq,_,not_injective_sequence1),
   (not_is_sequence_wf(ESeq,Type,WF) ;
    (is_sequence_wf(ESeq,Type,WF),not_injective_sequence2(ESeq,[],Type,WF))).
:- block not_injective_sequence2(-,?,?,?).
not_injective_sequence2([(int(_),X)|Tail],SoFar,Type,WF) :- 
    membership_test(SoFar,X,MemRes),
    not_injective_sequence3(MemRes,X,Tail,SoFar,Type,WF).

:- block not_injective_sequence3(-,?,?,?,?,?).
not_injective_sequence3(pred_true,_X,_Tail,_SoFar,_Type,_WF). 
not_injective_sequence3(pred_false,X,Tail,SoFar,Type,WF) :-
         add_new_element(X,SoFar,SoFar2),
         not_injective_sequence2(Tail,SoFar2,Type,WF).

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(2),int(33)),(int(1),int(33))],[int(44),int(33),int(22)],WF),WF)).
:- assert_must_succeed((bsets_clp:not_non_empty_injective_sequence(R,global_set('Name'),WF),
  ground_det_wait_flag(WF), R = [])).
:- assert_must_fail((bsets_clp:not_non_empty_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF), R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_non_empty_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF), R = [(int(2),int(2)),(int(1),int(3))] )).

:- block not_non_empty_injective_sequence(-,?,?).
not_non_empty_injective_sequence(X,Type,WF) :-
      empty_sequence(X) ; not_injective_sequence(X,Type,WF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_non_empty_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_non_empty_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:injective_non_empty_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:injective_non_empty_sequence([(int(2),int(44)),(int(1),int(44))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_fail((bsets_clp:injective_non_empty_sequence(R,global_set('Name'),WF),
                       ground_det_wait_flag(WF),R = [])).
:- assert_must_succeed((bsets_clp:injective_non_empty_sequence(R,global_set('Name'),WF),
             ground_det_wait_flag(WF),R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- block injective_non_empty_sequence(-,?,?). /* corresponds to iseq1 */
injective_non_empty_sequence(avl_set(AS),Type,WF) :- !,
   injective_sequence_wf(avl_set(AS),Type,WF),is_non_empty_explicit_set(avl_set(AS)).
injective_non_empty_sequence(Seq,Type,WF) :- 
   try_expand_custom_set(Seq,ESeq),
   injective_sequence_wf(ESeq,Type,WF),not_empty_sequence(ESeq).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:finite_non_empty_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:finite_non_empty_sequence([(int(1),int(33)),(int(2),int(33))],[int(22),int(33)],WF),WF)).
:- assert_must_fail((bsets_clp:finite_non_empty_sequence(R,global_set('Name'),WF),ground_det_wait_flag(WF),ground_det_wait_flag(WF),R = [])).
:- assert_must_succeed((bsets_clp:finite_non_empty_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- block finite_non_empty_sequence(-,?,?).
finite_non_empty_sequence(Seq,Type,WF) :- /* corresponds to Seq1 */
   is_sequence_wf_ex(Seq,Type,WF,ESeq),(var(ESeq) -> not_empty_sequence(Seq) ; not_empty_sequence(ESeq)).


  

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:permutation_sequence_wf([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:permutation_sequence_wf([(int(2),int(33)),(int(1),int(22))],[int(22),int(33)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:permutation_sequence_wf([(int(2),int(33)),(int(1),int(23))],[int(23),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:permutation_sequence_wf([(int(2),int(44)),(int(1),int(44))],[int(44)],WF),WF)).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(1)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(2))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(2)),(int(2),int(1))] )).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:permutation_sequence_wf(R,[pred_true /* bool_true */,pred_false /* bool_false */],WF), kernel_waitflags:ground_wait_flags(WF), nonvar(R),
                        R = [(int(1),pred_false /* bool_false */),(int(2),pred_true /* bool_true */)] )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,[int(1)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(1))] )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),ground_det_wait_flag(WF),R = [])).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),
                 ground_det_wait_flag(WF),R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),
   ground_det_wait_flag(WF),
  kernel_objects:equal_object(R,[(int(1),fd(1,'Name')),(int(3),fd(2,'Name')),(int(2),fd(3,'Name'))]) )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(3))] )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(4),int(3),int(2),int(1)],WF),
                       ground_det_wait_flag(WF), R=[(int(1),int(1)),(int(2),int(2)),(int(3),int(3)),(int(4),int(4))])).
                        
:- block permutation_sequence_wf(-,-,?).
permutation_sequence_wf(SeqFF,Type,WF) :- nonvar(SeqFF),
    custom_explicit_sets:dom_range_for_specific_closure(SeqFF,FFDomain,FFRange,function(bijection)),!,
    equal_object(FFRange,Type), is_sequence_domain(FFDomain,WF).
permutation_sequence_wf(Seq,Type,_WF) :-
   expand_and_convert_to_avl_set(Seq,AER),!,
  %% print_term_summary(checking_avl_perm_seq(AER,Type)),nl, %%
   is_avl_sequence(AER),
   %% print(is_sequence),nl, %%
   is_injective_avl_relation(AER,Range),
%%   print_term_summary(is_injective(Range)), %%
   kernel_objects:equal_object(Range,Type).
permutation_sequence_wf(Seq,Type,WF) :-
   try_expand_custom_set(Seq,ESeq),
   cardinality(Type,Card),
   when(ground(Card), (setup_sequence(Card,SkelSeq,1), %print(setup(SkelSeq,ESeq)),nl,
                       kernel_objects:equal_object(SkelSeq,ESeq))),
   %injective_sequence_wf(ESeq,Type,WF,LWF),
   surjective_iseq(ESeq,Type,WF,Card).

setup_sequence(0,[],_).
setup_sequence(s(C),[(int(Nr),_)|T], Nr ) :-
   N1 is Nr+1,
   setup_sequence(C,T,N1).

:- block surjective_iseq(?,?,?,-),surjective_iseq(?,-,?,?), surjective_iseq(-,?,?,?).
surjective_iseq(avl_set(S),Type,WF,Done) :- expand_custom_set(avl_set(S),ES),
  surjective_iseq(ES,Type,WF,Done).
surjective_iseq(closure(P,T,B),Type,WF,Done) :- expand_custom_set(closure(P,T,B),ES),
  surjective_iseq(ES,Type,WF,Done).
% no case for global_set: cannot be a relation
surjective_iseq([],T,_WF,_) :- empty_set(T).
surjective_iseq([(int(_Nr),El)|Tail],Type,WF,_PreviousRemoveDone) :-
  % print(surjective_iseq([(int(_Nr),El)|Tail],Type,WF)),nl,
   remove_element_wf(El,Type,NType,WF,Done),
  % when(nonvar(Done), (print(removed(_Nr,El)),nl)),
   surjective_iseq(Tail,NType,WF,Done).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_permutation_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_permutation_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_permutation_sequence([(int(2),int(33)),(int(1),int(23))],[int(23),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_permutation_sequence([(int(2),int(44)),(int(1),int(44))],[int(44)],WF),WF)).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,[int(1)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1))] )).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(2)),(int(2),int(1))] )).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,global_set('Name'),WF),
   ground_det_wait_flag(WF), R = [(int(1),fd(1,'Name')),(int(3),fd(2,'Name')),(int(2),fd(3,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,[int(1)],WF),
      ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(1))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,global_set('Name'),_WF),R = [])).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,global_set('Name'),WF),
        ground_det_wait_flag(WF),R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,[int(1),int(2)],WF),
        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(3))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,global_set('Name'),WF),
        ground_det_wait_flag(WF),R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- block not_permutation_sequence(-,?,?).
not_permutation_sequence(SeqFF,Type,WF) :- nonvar(SeqFF),
    custom_explicit_sets:dom_range_for_specific_closure(SeqFF,FFDomain,FFRange,function(bijection)),!,
    kernel_equality:equality_objects(FFRange,Type,Result), 
    when(nonvar(Result),(Result=pred_false -> true ; not_is_sequence_domain(FFDomain,WF))).
not_permutation_sequence(Seq,Type,WF) :-  
    when(ground(Seq),not_permutation_sequence1(Seq,Type,WF)).
%:- block not_permutation_sequence1(?,?,?,-).
not_permutation_sequence1(Seq,Type,WF) :-
   try_expand_custom_set(Seq,ESeq),not_permutation_sequence2(ESeq,Type,WF).
   
not_permutation_sequence2(Seq,Type,WF) :- not_injective_sequence(Seq,Type,WF).
not_permutation_sequence2(Seq,Type,WF) :-
    injective_sequence_wf(Seq,Type,WF), not_surjective_seq(Seq,Type,WF).

:- block not_surjective_seq(-,?,?).
not_surjective_seq([],T,WF) :- not_empty_set_wf(T,WF).
not_surjective_seq([(int(_),El)|Tail],Type,WF) :-
   delete_element_wf(El,Type,NType,WF),
   not_surjective_seq(Tail,NType,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:size_of_sequence([(int(1),int(22))],int(1),_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:size_of_sequence([(int(2),int(22)),(int(1),int(22))],int(2),_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:size_of_sequence([(int(2),int(22)),(int(1),int(22))],int(3),_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:size_of_sequence([(int(2),int(22)),(int(1),int(22)),(int(3),int(33))],int(3),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence(X,R,_WF),
                        X = [(int(1),int(2)),(int(2),int(1))],
                        R = int(2))).
:- assert_must_succeed((preferences:preference(use_clpfd_solver,false) -> true
                        ; bsets_clp:size_of_sequence(X,R,_WF), R=int(RI), 
                          clpfd_interface:clpfd_geq2(RI,2), nonvar(X), X = [(I1,_),(I2,_)|T],
                          I1==int(1), I2==int(2), T=[], RI==2 )).
:- assert_must_succeed((bsets_clp:size_of_sequence(X,R,_WF),X = [(int(1),_),(int(2),_)],R = int(2))).
:- assert_must_succeed((bsets_clp:size_of_sequence(X,_R,_WF),X =[(int(1),_),(int(2),_)] )).
:- assert_must_succeed_any((bsets_clp:size_of_sequence(X,int(2),_WF),nonvar(X),X=[_|Y],nonvar(Y),Y=[_|Z],Z==[])).
:- assert_must_succeed((bsets_clp:size_of_sequence([],int(0),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence([],int(0),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence([(int(1),int(4))],int(1),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence([],_,_WF))).
:- assert_must_fail((bsets_clp:size_of_sequence(X,int(1),_WF),
                        X = [(int(1),_),(int(2),_)|_])).
:- block size_of_sequence(-,-,?).
size_of_sequence(Seq,int(Res),WF) :- size_of_sequence1(Seq,Res,WF),
   set_up_sequence_skel(Seq,Res).

% setup sequence skeleton if we have some CLPFD bounds information about the size
% currently still quite limited: only sets up if sequence is a variable; + does the setup only once
:- use_module(clpfd_interface,[clpfd_domain/3]).
set_up_sequence_skel(Seq,Res) :- %print(size(Seq,Res)),nl,
   var(Seq), % to do: also deal with cases when Seq partially instantiated
   var(Res),
   preferences:preference(use_clpfd_solver,true),
   !,
   clpfd_interface:clpfd_geq2(Res,0), % assert that size must not be negative
   clpfd_interface:post_constraint( (Res#>0) #<=> Trigger), % generate reified trigger for when we can instantiate Seq
   set_up_sequence_skel_aux(Seq,Res,Trigger).
set_up_sequence_skel(_,_). % TO DO: check if Size interval shrinks
:- block set_up_sequence_skel_aux(-,?,-).
set_up_sequence_skel_aux(Seq,_Res,_Trigger) :-
   nonvar(Seq),!.  % to do: also deal with cases when Seq partially instantiated
set_up_sequence_skel_aux(Seq,Res,_Trigger) :-
   clpfd_domain(Res,FDLow,FDUp), % FDLow could also be 0
   %print(setup_sequence(Seq,Res,FDLow)),nl, % <=>
   gen_sequence_skeleton(1,FDLow,FDUp,S), %print(S),nl,
   Seq=S.
gen_sequence_skeleton(N,M,FDUp,S) :- N>M,!,(FDUp==M -> S=[] ; true).
gen_sequence_skeleton(N,Max,FDUp,[(int(N),_)|T]) :- 
    N1 is N+1,
    gen_sequence_skeleton(N1,Max,FDUp,T).

:- block size_of_sequence1(-,-,?).
size_of_sequence1(Seq,ResInt,WF) :- 
  nonvar(Seq),is_custom_explicit_set(Seq,size_of_sequence1),
  size_of_custom_explicit_set(Seq,Size,WF),!,
  equal_object(Size,int(ResInt)).
  /* TO DO: CHECK BELOW: would it not  be better to use cardinality ?? */
/* 
size_of_sequence1(Seq,Size,WF) :- !,finite_cardinality_as_int(Seq,int(Size),WF), check_indexes(Seq,Size).

   construct_interval_closure(1,Size,Domain),
   total_function_wf(FF,Domain,Range,_WF)
% we could also call total_function 1..Size --> _RangeType; would setup domain ?
:- block check_indexes(-,?).
check_indexes([],_) :- !.
check_indexes([(int(X),_)|T],Size) :- !,
   less_than_equal_direct(X,Size), check_indexes(T,Size).
check_indexes(_,_).
*/
size_of_sequence1(Seq,Size,_WF) :- Size==0,!, empty_sequence(Seq).
size_of_sequence1(Seq,Size,_WF) :- 
   expand_custom_set_to_list(Seq,ESeq,_,size_of_sequence1),  % print_term_summary(size1(Seq,ESeq,Size)),
   ((var(ESeq),nonvar(Size)) -> size_of_var_seq(ESeqR,0,Size),
                                ESeqR=ESeq % unify after to do propagation in one go, without triggering coroutines inbetween
                              ; size_of_seq2(ESeq,0,Size)).
           /* small danger of expanding closure while still var !*/
:- block size_of_seq2(-,?,-).
size_of_seq2([],Size,Size).
size_of_seq2([I|Tail],SizeSoFar,Res) :-
    S2 is SizeSoFar + 1,
    check_index(I,Res), % don't instantiate I yet; allow other kernel_predicates to freely instantiate it 
    less_than_equal_direct(S2,Res),
    %(ground(Res) -> safe_less_than_equal(size_of_seq2,S2,Res) ; true),
    size_of_seq2(Tail,S2,Res).
size_of_var_seq([],Size,Size).
size_of_var_seq([(int(S2),_)|Tail],SizeSoFar,Res) :-
    S2 is SizeSoFar + 1,safe_less_than_equal(size_of_var_seq,S2,Res),
    (var(Tail) -> size_of_var_seq(Tail,S2,Res) ; size_of_seq2(Tail,S2,Res)).
        
        
:- block check_index(-,?).
check_index((I,_),Res) :- check_index1(I,Res).
:- block check_index1(-,?).
check_index1(int(Idx),Res) :- less_than_equal_direct(Idx,Res).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:prepend_sequence(int(33),[(int(1),int(22))],[(int(2),int(22)),(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:prepend_sequence(int(33),[],[(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:prepend_sequence(int(33),[(int(2),int(44)),(int(1),int(22))],[(int(1),int(33)),(int(3),int(44)),(int(2),int(22))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:prepend_sequence(int(33),[(int(1),int(22))],[(int(1),int(22)),(int(2),int(33))],WF),WF)).
:- assert_must_succeed((bsets_clp:prepend_sequence(int(7),[],[(int(1),int(7))],_WF))).
:- assert_must_succeed((bsets_clp:prepend_sequence(int(7),X,R,_WF),
                        X = [(int(2),int(4)),(int(1),int(3))],
                        kernel_objects:equal_object(R,[(int(1),int(7)),(int(2),int(3)),(int(3),int(4))]))).
:- block prepend_sequence(?,-,-,?).
prepend_sequence(El,Seq,Res,_WF) :- Seq==[],!,
   equal_object_optimized([(int(1),El)],Res).
prepend_sequence(El,Seq,Res,_WF) :- nonvar(Seq),is_custom_explicit_set(Seq,prepend_sequence),
   prepend_custom_explicit_set(Seq,El,ERes),!, %print(prepend(Seq,El)),nl,
   equal_object(Res,ERes).
prepend_sequence(El,Seq,Res,WF) :- %print_message(prepend_sequence(El,Seq,Res)),
   equal_cons_wf(Res,(int(1),El),ShiftSeq,WF),
   shift_seq_indexes(Seq,1,ShiftSeq,WF).
%   print_message(prepend_result(El,Res)).

:- block shift_seq_indexes(-,-,?,?),shift_seq_indexes(-,?,-,?).
shift_seq_indexes(Seq,Offset,ShiftedSeq,WF) :- Offset == 0,!,  equal_sequence(Seq,ShiftedSeq,WF).
shift_seq_indexes(Seq,Offset,ShiftedSeq,WF) :- nonvar(Seq),!,
   expand_custom_set_to_list(Seq,ESeq,_,shift_seq_indexes),
   shift_seq_indexes2(ESeq,Offset,ShiftedSeq,WF).
shift_seq_indexes(Seq,Offset,ShiftedSeq,WF) :- NegOffset is -Offset,
   % compute in the other direction; TO DO: make a more fine-grained two-ways propagation
   expand_custom_set_to_list(ShiftedSeq,ESeq,_,shift_seq_indexes),
   shift_seq_indexes2(ESeq,NegOffset,Seq,WF).

:- block shift_seq_indexes2(-,?,?,?).
shift_seq_indexes2([],_,R,_WF) :- empty_set(R).
shift_seq_indexes2([(int(N),El)|Tail],Offset,Res,WF) :-
   equal_cons_wf(Res,(int(NewN),El),ShiftTail,WF),
   int_plus(int(N),int(Offset),int(NewN)),
   shift_seq_indexes2(Tail,Offset,ShiftTail,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:append_sequence([(int(1),int(22))],int(33),[(int(2),int(33)),(int(1),int(22))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:append_sequence([],int(33),[(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:append_sequence([(int(2),int(44)),(int(1),int(22))],int(33),[(int(1),int(22)),(int(3),int(33)),(int(2),int(44))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:append_sequence([(int(1),int(22))],int(33),[(int(1),int(33)),(int(2),int(22))],WF),WF)).
:- assert_must_succeed((bsets_clp:append_sequence([],int(7),[(int(1),int(7))],_WF))).
:- assert_must_succeed((bsets_clp:append_sequence(X,int(7),R,_WF),
                        X = [(int(2),int(4)),(int(1),int(3))],
                        kernel_objects:equal_object(R,[(int(1),int(3)),(int(2),int(4)),(int(3),int(7))]))).

:- block append_sequence(-,?,-,?).
append_sequence(Seq,El,Res,_WF) :- Seq==[],!,
   equal_object_optimized([(int(1),El)],Res).
append_sequence(Seq,El,Res,WF) :- %print_term_summary(append_sequence(Seq,El,Res)),
   nonvar(Seq),is_custom_explicit_set(Seq,append_sequence),
   append_custom_explicit_set(Seq,El,ERes),!,
   equal_sequence(Res,ERes,WF).
append_sequence(Seq,El,Res,WF) :- equal_cons_wf(Res,(int(NewSize),El),ResT,WF),
     append_sequence2(Seq,ResT,NewSize,WF).

:- block append_sequence2(-,?,?,?).
append_sequence2(Seq,ResT,NewSize,WF) :-
   try_expand_custom_set(Seq,ESeq), equal_sequence(ESeq,ResT,WF),
   size_of_sequence(ESeq,Size,WF),
   int_plus(Size,int(1),int(NewSize)).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:prefix_sequence([(int(1),int(22))],int(1),[(int(1),int(22))]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(bsets_clp:prefix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(2),[(int(1),int(11)),(int(2),int(22))]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:prefix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(3),[(int(1),int(11)),(int(2),int(22))]))).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(1),X),X = [(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(0),[]),X = [(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(-1),R),X = [(int(1),int(1))],R==[])).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(2),Y), 
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(1)),(int(2),int(3))]) )).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(1),Y), 
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(1))]) )).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(3),Y), 
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,X) )).

% Prefix of a sequence (s /|\ n)
:- block prefix_sequence(-,?,?).
prefix_sequence(Seq,int(Num),Res) :- prefix_sequence1(Seq,Num,Res).
:- block prefix_sequence1(?,-,?).
prefix_sequence1(_Seq,Num,Res) :- Num=<0,!, empty_set(Res).
prefix_sequence1(Seq,Num,Res) :- is_custom_explicit_set(Seq,prefix),
   prefix_of_custom_explicit_set(Seq,Num,ERes),!,
   equal_object(Res,ERes).
prefix_sequence1(Seq,Num,Res) :-
  expand_custom_set_to_list(Seq,ESeq,_,prefix_sequence1), 
  prefix_seq(ESeq,Num,0,Res). %,print(prefix_seq_exit(ESeq,Num,0,Res)),nl
:- block  prefix_seq(-,?,?,?).
prefix_seq([],_Max,_Sze,Res) :-  empty_set(Res).
          %(less_than(int(_Sze),int(_Max))
         %  ->  (print_message('Index bigger than sequence size in prefix_sequence (/|\\) !'),
         %       print_message(Max))
         %     /* in the AtelierB book this is allowed, in Wordsworth + AMN on web it is not ?? */
         %  ;  true).
prefix_seq([(int(N),El)|Tail],Max,SizeSoFar,Res) :-
    prefix_seq2(N,El,Tail,Max,SizeSoFar,Res).
:- block prefix_seq2(-,?,?,?,?,?).
prefix_seq2(N,El,Tail,Max,SizeSoFar,Res) :- % SizeSoFar is always ground
    (less_than_equal_direct(N,Max), equal_cons(Res,(int(N),El),PTail)  % to do: _wf version
      ;
     less_than_direct(Max,N), equal_object(Res,PTail)
    ),
    ( SizeSoFar<N -> NewSizeSoFar=N ; NewSizeSoFar = SizeSoFar ),
    prefix_seq(Tail,Max,NewSizeSoFar,PTail).
    
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:suffix_sequence([(int(1),int(22))],int(0),[(int(1),int(22))],WF),ground_det_wait_flag(WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(bsets_clp:suffix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(1),[(int(1),int(22)),(int(2),int(33))],_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:suffix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(2),[(int(1),int(22)),(int(2),int(33))],_WF))).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(0),X,_WF),X = [(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(1),[],_WF),X = [(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(2),Y,_WF), 
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(4))]) )).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(1),Y,_WF), 
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(3)),(int(2),int(4))]) )).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(2),Y,_WF), 
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(4))]) )).
:- assert_must_abort_wf(bsets_clp:suffix_sequence([(int(1),int(11)),(int(2),int(22))],int(-1),_R,WF),WF).
:- assert_must_abort_wf(bsets_clp:suffix_sequence([(int(1),int(11)),(int(2),int(22))],int(3),_R,WF),WF).

% kernel_waitflags:assert_must_abort2_wf(bsets_clp:suffix_sequence([int(11),int(22)],int(-1),_R,WF),WF)

% suffix of a sequence (s \|/ n); also called restrict at tail (Atelier B) or Drop
:- block suffix_sequence(-,?,?,?).
suffix_sequence(Seq,int(Num),Res,WF) :- % print(suffix_sequence(Seq)),nl,
   suffix_sequence1(Seq,Num,Res,WF).
:- block suffix_sequence1(?,-,?,?).
suffix_sequence1(Seq,Num,Res,_WF) :- Num==0, !, equal_object(Res,Seq).
suffix_sequence1(_Seq,Num,_Res,WF) :- Num<0, !, add_wd_error('negative index in suffix_sequence (\\|/)! ', Num,WF).
suffix_sequence1(Seq,Num,Res,WF) :- is_custom_explicit_set(Seq,suffix),
   suffix_of_custom_explicit_set(Seq,Num,ERes,WF),!,
   equal_object(Res,ERes).
suffix_sequence1(Seq,Num,Res,WF) :-
  expand_custom_set_to_list(Seq,ESeq,_,suffix_sequence), suffix_seq(ESeq,Num,0,Res,WF).
:- block suffix_seq(-,?,?,?,?).
suffix_seq([],Max,Sze,Res,WF) :-
   (less_than_direct(Sze,Max)
          -> add_wd_error('index larger than size of sequence in suffix_sequence (\\|/)! ', '>'(Max,Sze),WF)
          ; true), empty_set(Res).
suffix_seq([(int(N),El)|Tail],Max,SizeSoFar,Res,WF) :-
    suffix_seq2(N,El,Tail,Max,SizeSoFar,Res,WF).
:- block suffix_seq2(-,?,?,?,?,?,?).
suffix_seq2(N,El,Tail,Max,SizeSoFar,Res,WF) :-
    (less_than_equal_direct(N,Max), equal_object(Res,PTail)
      ;
     less_than_direct(Max,N),int_minus(int(N),int(Max),int(NN)),
     equal_cons_wf(Res,(int(NN),El),PTail,WF)
    ),
    (N>SizeSoFar -> (NewSizeSoFar=N)
                  ; (NewSizeSoFar = SizeSoFar)),
    suffix_seq(Tail,Max,NewSizeSoFar,PTail,WF).
  
 

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:concat_sequence([],[(int(1),int(33))],[(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:concat_sequence([(int(1),int(22)),(int(2),int(33))],[(int(1),int(33)),(int(2),int(44))],[(int(2),int(33)),(int(3),int(33)),(int(1),int(22)),(int(4),int(44))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:concat_sequence([(int(1),int(22))],[(int(1),int(33))],[(int(2),int(33)),(int(1),int(22))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:concat_sequence([(int(1),int(22))],[(int(1),int(33))],[(int(2),int(22)),(int(1),int(33))],WF),WF)).
:- assert_must_succeed((bsets_clp:concat_sequence([],X,Y,_WF),
                       X = [(int(1),int(1))], Y==X)).
:- assert_must_succeed((bsets_clp:concat_sequence(X,[],Y,_WF), X = [(int(1),int(1))], Y==X)).
:- assert_must_succeed((bsets_clp:concat_sequence([(int(1),int(1))],[],Y,_WF), Y==[(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:concat_sequence(X,X,Y,_WF),
                X = [(int(1),int(1))], kernel_objects:equal_object(Y,[(int(1),int(1)),(int(2),int(1))]))).
:- assert_must_succeed((bsets_clp:concat_sequence(X,X,Y,_WF),
      X = [(int(2),int(88)),(int(1),int(77))],
      kernel_objects:equal_object(Y,[(int(1),int(77)),(int(2),int(88)),(int(3),int(77)),(int(4),int(88))]))).

:- block /* concat_sequence(-,-,?,?), */
         concat_sequence(?,-,-,?), concat_sequence(-,?,-,?).
concat_sequence(S1,S2,Res,_WF) :- Res==[],!, empty_set(S1), empty_set(S2).
concat_sequence(S1,S2,Res,WF) :-
   (var(S1),var(S2) -> get_wait_flag(2,concat,WF,LWF)  % we have at least two solutions; TODO maybe use cardinality as wait_flag?
    ; LWF=1),
    concat_sequence2(LWF,S1,S2,Res,WF).

:- block concat_sequence2(-,?,-,?,?), concat_sequence2(-,-,?,?,?).
concat_sequence2(_,S1,S2,Res,WF) :- S1==[],!,equal_sequence(S2,Res,WF).
concat_sequence2(_,S1,S2,Res,WF) :- S2==[],!,equal_sequence(S1,Res,WF).
concat_sequence2(LWF,S1,S2,Res,WF) :-
   try_expand_and_convert_to_avl_with_check(S1,AS1,concat1),
   try_expand_and_convert_to_avl_with_check(S2,AS2,concat2),
   concat_sequence3(LWF,AS1,AS2,Res,WF).

concat_sequence3(_,S1,S2,Res,WF) :- nonvar(S1),is_custom_explicit_set(S1,concat_sequence),
   concat_custom_explicit_set(S1,S2,ERes),!,
   %print_term_summary(concat(S1,S2,ERes)),nl,
   equal_sequence(Res,ERes,WF).
concat_sequence3(_LWF,S1,S2,Res,WF) :- 
  % print_term_summary(ordinary_concat_sequence(_LWF,S1,S2,Res)),
  %try_expand_custom_set(S1,ES1), 
  size_of_sequence(S1,int(Size1),WF),
  (number(Size1) -> true
    ; size_of_sequence(S2,Size2,WF), size_of_sequence(Res,SizeRes,WF),
      int_minus(SizeRes,Size2,int(Size1)),
      in_nat_range_wf(int(Size1),int(0),SizeRes,WF)
  ),
   %print(call_concat_sequence_aux(Size1,S1,S2,Res,WF)),nl,
   concat_sequence_aux(Size1,S1,S2,Res,WF).

:- assert_must_succeed( (bsets_clp:equal_sequence([(int(1),A)|T1],[(int(1),int(22))|T2],_WF), 
A==int(22),T2=[],T1==[] )) .
:- assert_must_succeed( (bsets_clp:equal_sequence([(int(1),A)|T],avl_set(node((int(2),string(a)),true,0,node((int(1),string(c)),true,0,empty,empty),node((int(3),string(b)),true,0,empty,empty))),_WF),
A==string(c), T = [(int(2),B)|T2],  B==string(a), T2=[(int(3),C)], C==string(b)) ).
% equal_object optimized for sequences; can infer that pairs with same index have same value
% TO DO: complete and make more efficient
%equal_sequence(X,Y,_WF) :- nonvar(X),nonvar(Y),
%   is_custom_explicit_set(X,eval_sequence), is_custom_explicit_set(Y,eval_sequence),!,
%   equal_explicit_sets(X,Y).
equal_sequence(X,Y,WF) :- nonvar(X),nonvar(Y),
   get_seq_head(X,XI,XEl,XT), get_seq_head(Y,YI,YEl,YT), XI==YI,!,
   % THIS CURRENTLY ONLY CHECKS FRONTMOST indexes
   %print(eq_seq(XI,XEl,YI,YEl,XT,YT)),nl,
   equal_object(XEl,YEl), %print(done(XEl,YEl)),nl,
   equal_sequence(XT,YT,WF).
equal_sequence(X,Y,_WF) :- 
   %print(equal_sequence(X,Y)),nl,
   /* (is_custom_explicit_set(Y) -> monitor_equal_sequence_againts_custom_set(X,Y,WF)
    ; is_custom_explicit_set(X) -> monitor_equal_sequence_againts_custom_set(Y,X,WF)
    ; true),  does not seem to buy anything; equal_object already powerful enough */
   equal_object(X,Y).

% enforces the constraint that there is only one possible elemenent per index:
%:- block monitor_equal_sequence_againts_custom_set(-,?,?).
%monitor_equal_sequence_againts_custom_set([],_,_) :- !.
%monitor_equal_sequence_againts_custom_set([El|T],CS,WF) :- !,
%   element_of_custom_set_wf(El,CS,WF),
%   monitor_equal_sequence_againts_custom_set(T,CS,WF).
%monitor_equal_sequence_againts_custom_set(_,_,_).

get_seq_head([(Idx,El)|Tail],Idx,El,Tail).
%get_seq_head(avl_set(AVL),Idx,El,Tail) :-  does not seem to buy anything; equal_object already powerful enough
%   custom_explicit_sets:avl_min_pair(AVL,Idx,El),
%   custom_explicit_sets:direct_remove_element_from_avl(AVL,(Idx,El),Tail). % TO DO: only compute if indexes ==


:- block concat_sequence_aux(-,?,?,?,?).
concat_sequence_aux(Size1,S1,S2,Res,WF) :- nonvar(Res),Res=avl_set(_),
   prefix_of_custom_explicit_set(Res,Size1,Prefix),
   suffix_of_custom_explicit_set(Res,Size1,Postfix,WF),
   %print(split_avl_set(Size1,S1,S2,Res,Prefix,Postfix,WF)),nl,
   !,
   equal_sequence(S1,Prefix,WF), equal_sequence(S2,Postfix,WF).
concat_sequence_aux(Size1,S1,S2,Res,WF) :- 
   % print(concat_size1(S1,Size1)),nl,
  shift_seq_indexes(S2,Size1,NewS2,WF),   %    print(shifted(S2,Size1,NewS2)),nl,
  disjoint_union_wf(S1,NewS2,Res,WF). % , print(disj_res(S1,NewS2,Res)),nl.
  
% TO DO: add waitflags + use within partition_wf
% computes union of two sets which are guaranteed to be disjoint: means that if two of three sets known the other one can be determined

:- assert_must_succeed(exhaustive_kernel_check_wf([commutative],bsets_clp:disjoint_union_wf([int(3)],[int(2),int(1)],[int(1),int(3),int(2)],WF),WF)).	
:- assert_must_succeed(exhaustive_kernel_check_wf([commutative],bsets_clp:disjoint_union_wf([],[int(2),int(1)],[int(1),int(2)],WF),WF)).	
:- assert_must_succeed(exhaustive_kernel_check_wf([commutative],bsets_clp:disjoint_union_wf([int(1),int(2)],[],[int(2),int(1)],WF),WF)).	
:- assert_must_succeed((bsets_clp:disjoint_union_wf([int(1)],[int(2)],Res,_WF),kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((bsets_clp:disjoint_union_wf(A,B,[int(1)],_WF),B=[H],H==int(1),A==[])).

disjoint_union_wf(Set1,Set2,Res,WF) :-
  disjoint_union_wf0(Set1,Set2,Res,Res,WF).

% disjoint_union_wf0(Set1,Set2,UnionOfSet1Set2, SuperSet, WF)
:- block disjoint_union_wf0(-,-,-,?,?).
disjoint_union_wf0(Set1,Set2,Res,_,_WF) :- Set1==[],!,equal_object(Set2,Res).
disjoint_union_wf0(Set1,Set2,Res,_,_WF) :- Set2==[],!,equal_object(Set1,Res).
disjoint_union_wf0(Set1,Set2,Res,_,_WF) :- Res==[],!,empty_set(Set1), empty_set(Set2).
disjoint_union_wf0(Set1,Set2,Res,FullRes,WF) :- disjoint_union0(Set1,Set2,Res,FullRes,WF).

:- block disjoint_union0(-,-,?,?,?), disjoint_union0(-,?,-,-,?), disjoint_union0(?,-,-,-,?).
disjoint_union0(Set1,Set2,Res,_,_WF) :- Set1==[],!,equal_object(Set2,Res).
disjoint_union0(Set1,Set2,Res,_,_WF) :- Set2==[],!,equal_object(Set1,Res).
disjoint_union0(S1,S2,Res,_F,WF) :-  % print(disjoint_union0(S1,S2,Res,_F)),nl,
    ground(Res), 
    ( ground(S1) -> !, %print(diff(Res,S1,S2)),nl,
                       check_subset_of_wf(S1,Res,WF), % TO DO: check if we can merge the check_subset and difference set in one predicate
                       difference_set_wf(Res,S1,S2,WF)
     ; ground(S2) -> !, %print(diff(Res,S2,S1)),nl,
                       check_subset_of_wf(S2,Res,WF),
                       difference_set_wf(Res,S2,S1,WF)).
disjoint_union0(Set1,Set2,Res,_,_WF) :- nonvar(Set1),
    is_custom_explicit_set(Set1,disjoint_union0), 
    union_of_explicit_set(Set1,Set2,Union), !,
    equal_object(Union,Res).
disjoint_union0(Set1,Set2,Res,Full,WF) :- expand_custom_set_to_list(Set1,ESet1,_,disjoint_union0_1),
    expand_custom_set_to_list(Set2,ESet2,_,disjoint_union0_2),
    %print(disjoint_union0(ESet1,ESet2,Res,WF)),nl,
    disj_union1(ESet1,ESet2,Res,Full,WF). 

:- block disj_union1(-,-,?,?,?).
disj_union1(X,Y,Res,FullRes,WF) :-
   var(X) -> disj_union2(Y,X,Res,FullRes,WF) ; disj_union2(X,Y,Res,FullRes,WF).

disj_union2([],Y,Res,_,_WF) :- equal_object(Y,Res).
disj_union2([X|TX],Y,Res,FullRes,WF) :-
   remove_element_wf(X,Res,TR,WF), % was: equal_cons_wf(Res,X,TR,WF)  but error was that it could force X to be a certain value
   (ground(X) -> equal_cons_wf(Res,X,TR,WF)
    ; check_element_of_wf(X,FullRes,WF), % ensure that we set up proper constraints for X; e.g., for x \/ y = 1..10 & x /\ y = {}
      when(ground(X), equal_cons_wf(Res,X,TR,WF))
    ), % ensure that we instantiate Res if TR known; otherwise we may get pending co-routines, e.g. test 506, SyracuseGrammar
   disj_union3(TX,Y,TR,FullRes,WF).

:- block disj_union3(-,-,-,?,?).
disj_union3(X,Y,Res,_,_WF) :- Res==[],!,empty_set(X),empty_set(Y).
disj_union3(X,Y,Res,FullRes,WF) :- disj_union1(X,Y,Res,FullRes,WF).


:- block disjoint_union_generalized_wf(-,?,?).
disjoint_union_generalized_wf(SetsOfSets,Res,WF) :- 
        expand_custom_set_to_list(SetsOfSets,ESetsOfSets,_,disjoint_union_generalized_wf), 
        disjoint_union_generalized2(ESetsOfSets,[],Res,WF). %, print(res(Res)),nl.
:- block disjoint_union_generalized2(-,?,?,?).
disjoint_union_generalized2([],S,Res,_WF) :-  equal_object(S,Res).
disjoint_union_generalized2([H|T],UnionSoFar,Res,WF) :-
     disjoint_union_wf0(H,UnionSoFar,UnionSoFar2,Res,WF),
     %  print_message(called_disjoint_union(H,UnionSoFar,UnionSoFar2)), %%
     disjoint_union_generalized2(T,UnionSoFar2,Res,WF).

% TO DO: if there are more than two sets: it may be interesting to set up constraint that
%        each set is a subset of the full set;
%  (would avoid enumeration warning in, e.g.,  x \/ y \/ z = 1..10 & x /\ y = {} & x /\ z = {} & y /\ z = {} & card(x)=card(y)+2 )

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:concatentation_of_sequences([(int(1),[]),(int(3),[(int(1),int(22)),(int(2),int(33))]),(int(2),[(int(1),int(11))])],
   [(int(1),int(11)),(int(2),int(22)),(int(3),int(33))],_WF))).  
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:concatentation_of_sequences([(int(1),[]),(int(2),[(int(1),int(33))])],[(int(1),int(33))],_WF))).   
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:concatentation_of_sequences([(int(1),[]),(int(2),[(int(1),int(55))])],Res,WF),
                        kernel_waitflags:ground_wait_flags(WF),
                        kernel_objects:equal_object(Res,[(int(1),int(55))]) )).        
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:concatentation_of_sequences([(int(1),[(int(1),int(22))]),(int(2),[(int(1),int(55))])],Res,WF),
                        kernel_waitflags:ground_wait_flags(WF),
                        kernel_objects:equal_object(Res,[(int(1),int(22)),(int(2),int(55))]) )).
:- block concatentation_of_sequences(-,?,?).
concatentation_of_sequences(SeqOfSeq,Res,WF) :- %print_message(conc(SeqOfSeq)),
    try_expand_and_convert_to_avl_with_check(SeqOfSeq,ES,conc),
    concs2(ES,Res,WF).

concs2(SeqOfSeq,Res,_) :- is_custom_explicit_set(SeqOfSeq,conc),
   conc_custom_explicit_set(SeqOfSeq,CRes),!,
   equal_object(CRes,Res).
concs2(SeqOfSeq,Res,_) :- empty_set(SeqOfSeq),empty_set(Res).
concs2(SeqOfSeq,Res,WF) :- not_empty_set_wf(SeqOfSeq,WF),
    front_sequence(SeqOfSeq,Front,WF),
    concatentation_of_sequences(Front,FrontRes,WF),
    last_sequence(SeqOfSeq,Last,WF),
    concat_sequence(FrontRes,Last,Res,WF).

:- assert_must_abort_wf(bsets_clp:tail_sequence([],_R,unknown,WF),WF).
:- assert_must_abort_wf(bsets_clp:tail_sequence([],[],unknown,WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check( 
 bsets_clp:tail_sequence([(int(1),int(4)),(int(2),int(5))],[(int(1),int(5))],unknown,_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:tail_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],
[(int(1),int(5)),(int(2),int(6))],unknown,_WF)) ).
:- assert_must_succeed((bsets_clp:tail_sequence(X,R,unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5))],
                        kernel_objects:equal_object(R,[(int(1),int(5))]) )).
:- assert_must_succeed((bsets_clp:tail_sequence(X,[(int(1),int(5))],unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5))] )).
:- assert_must_succeed((bsets_clp:tail_sequence(X,[(int(1),int(5)),(int(2),int(7))],unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5)),(int(3),int(7))] )).
:- assert_must_succeed((bsets_clp:tail_sequence(X,[(int(2),int(7)),(int(1),int(5))],unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5)),(int(3),int(7))] )).
:- block tail_sequence(-,?,?,?).
tail_sequence(S1,Res,Span,WF) :- is_custom_explicit_set(S1,tail_sequence),
  tail_sequence_custom_explicit_set(S1,TRes,Span,WF),!,
  equal_object(TRes,Res).
tail_sequence(S1,Res,Span,WF) :- expand_custom_set_to_list(S1,ES1,_,tail_sequence),
  tail2(ES1,Res,Span,WF).

tail2([],_,Span,WF) :-
    add_wd_error_span('tail applied to empty sequence!',[],Span,WF).
tail2([H|T],Res,_Span,WF) :- domain_subtraction([int(1)],[H|T],IntRes),
   shift_seq_indexes(IntRes,-1,Res,WF).


:- assert_must_abort_wf(bsets_clp:first_sequence([],_R,unknown,WF),WF).
:- assert_must_abort_wf(bsets_clp:first_sequence([],int(1),unknown,WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:first_sequence([(int(1),int(4)),(int(2),int(5))],int(4),unknown,_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:first_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],int(4),unknown,_WF)) ).
:- assert_must_succeed((bsets_clp:first_sequence(X,R,unknown,_WF),
                        X = [(int(1),int(2)),(int(2),int(1))],
                        R = int(2))).

:- block first_sequence(-,?,?,?).
%first_sequence(Seq,Res,Span,_) :- % print(first_seq(Seq,Res,Span)),nl,fail.
%  is_custom_explicit_set(Seq,first_sequence),
%  first_sequence_explicit_set(Seq,First), !,
%  equal_object(First,Res).
first_sequence([],_,Span,WF) :- !,add_wd_error_span('first applied to empty sequence!',[],Span,WF).
first_sequence(Seq,Res,Span,WF) :- apply_to(Seq,int(1),Res,Span,WF).
   

   
:- assert_must_abort_wf(bsets_clp:front_sequence([],_R,WF),WF).
:- assert_must_abort_wf(bsets_clp:front_sequence([],[],WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:front_sequence([(int(1),int(4)),(int(2),int(5))],[(int(1),int(4))],_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:front_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],[(int(1),int(4)),(int(2),int(5))],_WF)) ).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:front_sequence(X,R,WF),
                        X = [(int(1),int(2)),(int(2),int(55))],kernel_waitflags:ground_wait_flags(WF),
                        kernel_objects:equal_object(R,[(int(1),int(2))]))).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:front_sequence(X,R,WF),
                        X = [(int(3),int(33))|R], kernel_waitflags:ground_wait_flags(WF),
                        print(R),nl,
                        kernel_objects:equal_object(R,[(int(1),int(2)),(int(2),int(55))]) )).
                        
front_sequence(Seq,Res,WF) :- front_sequence(Seq,Res,unknown,WF).
:- block front_sequence(-,?,?,?).
front_sequence(S1,Res,_Span,_) :- 
  is_custom_explicit_set(S1,front_sequence),
  front_sequence_custom_explicit_set(S1,FRes),!,
  equal_object(FRes,Res).
front_sequence(Seq,Res,Span,WF) :-  expand_custom_set_to_list(Seq,ESeq,_,front_sequence),
  front2(ESeq,Res,Span,WF).
front2([],_,Span,WF) :- add_wd_error_span('front applied to empty sequence!',[],Span,WF).
front2([H|T],Res,_Span,WF) :- size_of_sequence([H|T],int(Size),WF),
  (number(Size) -> true ; size_of_sequence(Res,SizeRes,WF), int_plus(SizeRes,int(1),int(Size))),
  when(ground(Size), domain_subtraction([int(Size)],[H|T],Res)).
   

:- assert_must_abort_wf(bsets_clp:last_sequence([],_R,WF),WF).
:- assert_must_abort_wf(bsets_clp:last_sequence([],int(1),WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:last_sequence([(int(1),int(4)),(int(2),int(5))],int(5),_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:last_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],int(6),_WF)) ).
:- assert_must_succeed((bsets_clp:last_sequence(X,R,_WF),
                        X = [(int(1),int(2)),(int(2),int(55))],R = int(55))).
:- assert_must_succeed((bsets_clp:last_sequence(X,R,_WF), X = [(int(1),int(55))], R = int(55))).
:- assert_must_succeed((bsets_clp:last_sequence([(int(1),[(int(1),int(22))]),(int(2),[(int(1),int(55))])],R,_WF), R == [(int(1),int(55))])).

last_sequence(Seq,Res,WF) :- last_sequence(Seq,Res,unknown,WF).
:- block last_sequence(-,?,?,?).
last_sequence(Seq,Res,_Span,_) :- %print(last_sequence(Seq,Res,_)),nl,
  is_custom_explicit_set(Seq,last_sequence),
  last_sequence_explicit_set(Seq,Last), !, %print(last(Last)),nl,
  equal_object(Last,Res).
last_sequence([],_,Span,WF) :- !,add_wd_error_span('last applied to empty sequence!',[],Span,WF).
%last_sequence(Seq,Res,Span,WF) :- !,  /* I am not sure whether this brings any advantage */
%   expand_custom_set_to_list(Seq,ESeq),
%   last_seq(ESeq,0,abort(undefined_last),0,Res,Span,WF), print(last(ESeq,Res)),nl.
last_sequence(Seq,Res,Span,WF) :-
    size_of_sequence(Seq,int(Size),WF),
    last_sequence_aux(Size,Seq,Res,Span,WF).
:- block last_sequence_aux(-,?,?,?,?).
last_sequence_aux(Size,Seq,Res,Span,WF) :-
    apply_to(Seq,int(Size),Res,Span,WF).
/*
:- block last_seq(-, ?,?,?,?, ?,?).
last_seq([],LastIdx,LastVal,SizeSoFar,LastResult, Span,WF) :-
   (LastIdx=SizeSoFar -> LastResult = LastVal ;
     add_wd_error_span('last applied to invalid sequence (biggest index different from size): ',LastIdx/SizeSoFar,Span,WF),
     LastResult=LastVal).
last_seq([(ICurIdx,CurVal)|T],LastIdx,LastVal,SizeSoFar,LastResult,Span,WF) :-
  last_seq2(ICurIdx,CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF).
:- block last_seq2(-,?,?, ?,?,?,?, ?,?).
last_seq2(int(CurIdx),CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF) :-
   last_seq3(CurIdx,CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF).
:- block last_seq3(-,?,?, ?,?,?,?, ?,?).
last_seq3(CurIdx,CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF) :-
   S1 is SizeSoFar+1,
   (CurIdx>LastIdx -> last_seq(T,CurIdx,CurVal,S1,LastResult,Span,WF)
     ; last_seq(T,LastIdx,LastVal,S1,LastResult,Span,WF)).
*/
   

:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([(int(1),int(4)),(int(2),int(5))],[(int(1),int(5)),(int(2),int(4))],WF),WF )).
:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([(int(1),int(4))],[(int(1),int(4))],WF),WF )).
:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([],[],WF),WF )).
:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],[(int(1),int(6)),(int(3),int(4)),(int(2),int(5))],WF),WF )).
:- assert_must_succeed((bsets_clp:rev_sequence([],[],_WF))).
:- assert_must_succeed((bsets_clp:rev_sequence(X,R,_WF),
                        X = [(int(1),int(2)),(int(2),int(1))],
                        kernel_objects:equal_object(R,[(int(2),int(2)),(int(1),int(1))]) )).
:- assert_must_succeed((bsets_clp:rev_sequence(X,R,_WF),
                        X = [(int(1),int(23)),(int(2),int(1)),(int(3),int(55))],
              kernel_objects:equal_object(R,[(int(3),int(23)),(int(2),int(1)),(int(1),int(55))]) )).
:- assert_must_succeed((bsets_clp:rev_sequence(R,X,_WF),
                        X = [(int(1),int(23)),(int(2),int(1)),(int(3),int(55))],
              kernel_objects:equal_object(R,[(int(3),int(23)),(int(2),int(1)),(int(1),int(55))]) )).
:- assert_must_succeed((bsets_clp:rev_sequence(X,_R,_WF),
                        X = [(int(2),int(1)),(int(1),int(23)),(int(3),int(55))] )).
:- assert_must_succeed((bsets_clp:rev_sequence(_R,X,_WF),
                        X = [(int(3),int(55)),(int(1),int(23)),(int(2),int(1))] )).

:- block rev_sequence(-,-,?).
rev_sequence(S1,Res,WF) :-
  (nonvar(S1) -> rev_sequence2(S1,Res,WF)
               ; rev_sequence2(Res,S1,WF)).

rev_sequence2(S1,Res,WF) :-       
  expand_custom_set_to_list(S1,ES1,_,rev_sequence2),
 % print(rev(ES1)),nl,
  size_of_sequence(ES1,int(Size1),WF),
  rev_sequence3(ES1,Size1,Res,WF).

:- block rev_sequence3(?,-,?,?).
rev_sequence3(E,S,R,WF) :- rev_sequence4(E,S,R,WF).

:- block rev_sequence4(-,?,?,?).
rev_sequence4([],_,Res,_WF) :- empty_set(Res).
rev_sequence4([(int(N),El)|Tail],Size1,Res,WF) :-
   equal_cons_wf(Res,(int(NewN),El),RTail,WF),
   NewN is Size1 - N + 1,
   rev_sequence4(Tail,Size1,RTail,WF).

   
/* --------- */
/* RELATIONS */
/* --------- */

%maplet(X,Y,(X,Y)).

% relation([]).
% relation([(_X,_Y)|T]) :- relation(T).
   
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:relation_over_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check( bsets_clp:relation_over([],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([(int(1),int(2))],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([([int(1)],[int(2)])],[[int(1)],[],[int(2)]],[[int(2)]]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([(pred_true /* bool_true */,pred_false /* bool_false */)],[pred_false /* bool_false */,pred_true /* bool_true */],[pred_false /* bool_false */]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((pred_true /* bool_true */,int(2)),fd(1,'Name'))],[(pred_false /* bool_false */,int(1)),(pred_true /* bool_true */,int(2))],[fd(2,'Name'),fd(1,'Name')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((rec([field(a,fd(1,'Name'))]),int(2)),fd(1,'Name'))],[(rec([field(a,fd(1,'Name'))]),int(1)),(rec([field(a,fd(1,'Name'))]),int(2))],[fd(2,'Name'),fd(1,'Name')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((rec([field(a,fd(2,'Name')),field(b,fd(1,'Name'))]),int(2)),fd(1,'Name'))],[(rec([field(a,fd(1,'Name')),field(b,fd(1,'Name'))]),int(1)),(rec([field(a,fd(1,'Name')),field(b,fd(2,'Name'))]),int(2)),(rec([field(a,fd(2,'Name')),field(b,fd(1,'Name'))]),int(2))],[fd(2,'Name'),fd(1,'Name')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((pred_true /* bool_true */,int(2)),string('"STR1"'))],[(pred_false /* bool_false */,int(1)),(pred_true /* bool_true */,int(2))],[string('"STR2"'),string('"STR1"')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( /* multiple solutions !!*/ bsets_clp:relation_over([(int(1),int(2)),(int(2),int(2))],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([(int(1),int(2)),(int(1),int(3))],[int(1),int(2)],[int(3),int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_fail_check( bsets_clp:relation_over([(int(1),int(2)),(int(2),int(1))],[int(1),int(2)],[int(2)]) )).
:- assert_must_fail(( bsets_clp:relation_over([(int(1),int(1))],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(( bsets_clp:relation_over(X,[int(1),int(2)],[int(3)]), 
                         X==[(int(1),int(3))] )).
:- assert_must_succeed(( bsets_clp:relation_over(X,[int(1),int(2)],[int(3)]), 
                         X==[(int(1),int(3)),(int(2),int(3))] )).
:- assert_must_succeed(( bsets_clp:relation_over(X,[int(1),int(2)],[int(4),int(5)]), 
                         X==[(int(2),int(4)),(int(2),int(5))] )).

relation_over(R,Dom,Ran) :- init_wait_flags(WF),
  relation_over_wf(R,Dom,Ran,WF),
  ground_wait_flags(WF).
  
:- block relation_over_wf(-,-,-,?).
relation_over_wf(R,Dom,Ran,WF) :- 
  kernel_equality:get_cardinality_relation_over_wait_flag(Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels),
  relation_over1(R,Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels).

:- block relation_over1(-,?,?,?,-,?,?).
relation_over1(FF,Domain,Range,WF,_WFR,_MaxCard,_MaxNrOfRels) :- nonvar(FF),
    custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,_),!,
    check_subset_of_wf(FFDomain,Domain,WF).
relation_over1(FF,Domain,Range,WF,_WFR,_MaxCard,_MaxNrOfRels) :- nonvar(FF),
    % print_term_summary(relation_over1(FF,Domain,Range,WF)), %
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,_),!,
    check_subset_of_wf(FFDomain,Domain,WF),
    check_subset_of_wf(FFRange,Range,WF).
relation_over1(R,Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels) :- var(R),!,
    relation_over2(R,[],Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels,none).
relation_over1(R,Domain,Range,WF,_WFR,_MaxCard,_) :- 
   expand_and_convert_to_avl_set(R,AER),!,
   is_avl_relation_over_domain(AER,Domain,WF), %print(dom_ok),nl,
   is_avl_relation_over_range(AER,Range,WF).
relation_over1(R,Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels) :-
    expand_custom_set_to_list(R,ER,_,relation_over1),
    relation_over2(ER,[],Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels,none).

:- block relation_over2(-,?,?,?,?,-,?,?,?).
relation_over2([],_,_,_,_WF,_WFR,_MaxCard,_MaxNrOfRels,_LastPair).
relation_over2(REL,SoFar,Domain,Range,WF,WFR,MaxCard,MaxNrOfRels,LastPair) :- 
   (var(REL) -> NewLastPair=(X,Y) ; NewLastPair=none), %remember whether we freely chose X,Y
   REL = [(X,Y)|T],
    % print(rel_over2(X,Y,T,sofar(SoFar),WFR,MaxCard,LastPair)),nl,
      (number(MaxCard)
       -> MaxCard>0,C1 is MaxCard-1  ,(C1=0 -> T=[] ; true)
        ; C1=MaxCard),
      % TO DO: try to enumerate elements in order
      ordered_pair(LastPair,X,Y,not_equal),
      check_element_of_wf(X,Domain,WF), 
      check_element_of_wf(Y,Range,WF),
      not_element_of_wf((X,Y),SoFar,WF),
      %print_bt_message(done_enum_relation_over2(X,Y,T,sofar(SoFar),MaxCard)),nl,
      update_waitflag(MaxNrOfRels,WFR,NewWFR,WF),
      relation_over2(T,[(X,Y)|SoFar],Domain,Range,WF,NewWFR,C1,MaxNrOfRels,NewLastPair).

% check that new pair is greater than previous pair, if that pair was freely chosen
ordered_pair(none,_,_,_).
ordered_pair((LastX,LastY),NewX,NewY,Eq) :- ordered_value(LastX,NewX,EqualX),
  check_second_component(EqualX,LastY,NewY,Eq).

:- block check_second_component(-,?,?,?).
check_second_component(equal,X,Y,EqRes) :- ordered_value(X,Y,EqRes).
check_second_component(not_equal,_X,_Y,not_equal). % no need to check 2nd component

:- block ordered_value(-,?,?), ordered_value(?,-,?).
ordered_value(pred_true /* bool_true */,B,Eq) :- !, (B=pred_true /* bool_true */ -> Eq=equal ; Eq=not_equal).
ordered_value(pred_false /* bool_false */,B,Eq) :- !, B=pred_false /* bool_false */, Eq=equal.
ordered_value(int(X),int(Y),Eq) :- !,
   kernel_objects:less_than_equal_direct(X,Y), equal_atomic_term(X,Y,Eq).
ordered_value(fd(NrX,T),fd(NrY,T),Eq) :- !,
   kernel_objects:less_than_equal_direct(NrX,NrY), equal_atomic_term(NrX,NrY,Eq).
ordered_value((X1,X2),(Y1,Y2),Eq) :- !, ordered_pair((X1,X2),Y1,Y2,Eq).
ordered_value(string(X),string(Y),Eq) :- !, less_equal_atomic_term(X,Y,Eq).
ordered_value(rec(FX),rec(FY),Eq) :- !,  
   ordered_fields(FX,FY,Eq).
ordered_value([],Y,Eq) :- !, (Y==[] -> Eq=equal ; Eq=not_equal).
ordered_value(avl_set(A),Y,Eq) :- !,
  (Y==[] -> fail
    ; Y=avl_set(B) -> (A @< B -> Eq=not_equal ; A@>B -> fail ; Eq=equal)
    ; print(assuming_strictly_ordered(avl_set(A),Y)),nl,
      Eq=not_equal). % TO DO: treat sets better
ordered_value([H|T],Y,Eq) :- !, 
  (Y==[] -> fail ; (Y==[H|T] -> Eq=equal
                    ; print(assuming_strictly_ordered([H|T],Y)),nl,
                      Eq=not_equal)).
ordered_value(A,B,not_equal) :- print(assuming_strictly_ordered(A,B)),nl. 

:- block ordered_fields(-,?,?).
ordered_fields([],RHS,Eq) :- !,RHS=[], Eq=equal.
ordered_fields([field(Name,ValX)|TX],RHS,Eq) :- !,RHS=[field(Name,ValY)|TY],
  ordered_value(ValX,ValY,Equal1), check_next_field(Equal1,TX,TY,Eq).
ordered_fields(FX,FY,Eq) :- add_internal_error('Unknown fields: ',ordered_fields(FX,FY,Eq)), Eq=not_equal.

:- block check_next_field(-,?,?,?).
check_next_field(equal,TX,TY,EqRes) :- ordered_fields(TX,TY,EqRes).
check_next_field(not_equal,_X,_Y,not_equal). % no need to check next field

:- block less_equal_atomic_term(-,?,?), less_equal_atomic_term(?,-,?).
less_equal_atomic_term(A,B,Res) :- (A==B -> Res=equal ; A @<B, Res=not_equal).

:- block equal_atomic_term(-,?,?), equal_atomic_term(?,-,?).
equal_atomic_term(A,B,Res) :- (A==B -> Res=equal ; Res=not_equal).


:- assert_must_succeed(exhaustive_kernel_check( bsets_clp:not_relation_over([(int(1),int(2)),(int(2),int(1))],[int(1),int(2)],[int(2)],_WF) )).
:- assert_must_succeed(exhaustive_kernel_check( bsets_clp:not_relation_over([(int(1),int(2))],[],[int(2)],_WF) )).
:- assert_must_succeed(exhaustive_kernel_fail_check( bsets_clp:not_relation_over([(int(1),pred_true)],[int(1)],[pred_true],_WF) )).
:- assert_must_succeed(exhaustive_kernel_fail_check( bsets_clp:not_relation_over([],[int(1)],[pred_true],_WF) )).
:- assert_must_succeed( bsets_clp:not_relation_over([(int(1),int(2))],[int(3)],[int(1),int(2)],_) ).
:- assert_must_succeed( bsets_clp:not_relation_over([(int(1),int(2))],[int(1)],[int(3)],_) ).
:- assert_must_succeed( bsets_clp:not_relation_over([(int(1),int(3)),(int(1),int(2))],[int(1)],[int(3)],_) ).
:- assert_must_fail( bsets_clp:not_relation_over([(int(1),int(3))],[int(1)],[int(3)],_) ).
:- assert_must_fail( bsets_clp:not_relation_over([],[int(1)],[int(3)],_) ).
:- assert_must_fail( bsets_clp:not_relation_over([],[],[],_) ).
:- block not_relation_over(-,?,?,?).

not_relation_over(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,_),!,
    % print_term_summary(not_relation_over(FF,FFDomain,Range)),
    not_subset_of_wf(FFDomain,Domain,WF).
not_relation_over(FF,Domain,Range,WF) :- nonvar(FF), % print_term_summary(not_relation_over(FF,Domain,Range,WF)), %
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,_),!,
    %print_term_summary(check_special_not_pf(FF,FFDomain,FFRange)),nl,
    not_both_subset_of(FFDomain,FFRange,Domain,Range,WF).
/* could be slightly more efficient: but not clear if warrants additional complexity in code:
not_relation_over(FF,Domain,Range,WF) :- nonvar(FF),
   check_element_can_be_decided(Domain), % ensures that check_element_of_wf will not block below
   check_element_can_be_decided(Range), % ensures that check_element_of_wf will not block below
   expand_and_convert_to_avl_set(FF,AER),!,
   (is_avl_relation_over_domain(AER,Domain,WF)
     -> \+ is_avl_relation_over_range(AER,Range,WF)
     ;  true).
check_element_can_be_decided(X) :- var(X),!,fail.
check_element_can_be_decided(avl_set(_)).
check_element_can_be_decided([]).
check_element_can_be_decided(closure(P,T,B)) :-
   custom_explicit_sets:is_interval_closure_or_integerset(closure(P,T,B),Low,Up),
   ground(Low), ground(Up).
*/
not_relation_over(R,Dom,Ran,WF) :- 
   expand_custom_set_to_list(R,ER,_,not_relation_over),
 %%  print(not_rel(ER,Dom,Ran)),nl,
   not_relation_over2(ER,Dom,Ran,WF).
   

%not_relation_over2(R,_,_) :- when(nonvar(R), (R\=[], R\=[_|_])) . % TYPE ERROR !
:- block not_relation_over2(-,?,?,?).
not_relation_over2([(X,Y)|T],Domain,Range,WF) :- %print(not_rel2(X,Y,T)),nl,
      membership_test(Domain,X,MemRes),
      not_relation_over3(MemRes,Y,T,Domain,Range,WF).
      
:- block not_relation_over3(-,?,?,?,?,?).
not_relation_over3(pred_false,_Y,_T,_Domain,_Range,_WF).
not_relation_over3(pred_true,Y,T,Domain,Range,WF) :-
      membership_test(Range,Y,MemRes),
      not_relation_over4(MemRes,T,Domain,Range,WF).

:- block not_relation_over4(-,?,?,?,?).
not_relation_over4(pred_false,_T,_Domain,_Range,_WF).
not_relation_over4(pred_true,T,Domain,Range,WF) :-
      not_relation_over2(T,Domain,Range,WF).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf([],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf([(int(1),int(3))],[int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf(
    [(int(0),int(55)),(int(2),int(3)),(int(1),int(3))],[int(1),int(2),int(0)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf(
    [(int(99),int(55)),(int(2),int(3)),(int(99),int(4))],[int(2),int(99)],WF),WF)).
:- assert_must_succeed((bsets_clp:domain_wf([],Res,_WF),Res=[])).
:- assert_must_succeed((bsets_clp:domain_wf([(int(1),int(2))],Res,_WF),
        kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((bsets_clp:domain_wf([(int(1),int(2)),(int(1),int(1))],Res,_WF),
            kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((bsets_clp:domain_wf([(int(2),int(2)),(int(1),int(2))],Res,_WF),
                        kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((bsets_clp:domain_wf(X,Res,_WF),kernel_objects:equal_object(Res,[int(1),int(3),int(2)]),
                        X = [(int(2),int(2)),(int(1),int(1)),(int(3),int(2))])).
:- assert_must_succeed((bsets_clp:domain_wf(X,Res,_WF),kernel_objects:equal_object(Res,[int(1),int(2)]),
                        X = [(int(2),int(2)),(int(1),int(1)),(int(1),int(2))])).
:- assert_must_fail((bsets_clp:domain_wf(X,Res,_WF),kernel_objects:equal_object(Res,[int(1),int(2)]),
                        X = [(int(2),int(2)),(int(1),int(1)),(int(3),int(2))])).

:- block domain_wf(-,-,?).
domain_wf(Rel,Res,_WF) :- Res == [],!,
   % print(empty_domain_res(Rel,Res)),nl,
   empty_set(Rel).
domain_wf(Rel,Res,WF) :- domain_wf1(Rel,Res,WF).

:- block domain_wf1(-,?,?).
domain_wf1(Rel,Res,_WF) :- is_custom_explicit_set(Rel,domain_wf),
 % print_term_summary(try_dom_explicit(Rel,Res)),
  domain_of_explicit_set(Rel,Dom), !,
  % print_term_summary(dom_res(Rel,Dom)),%%
  equal_object(Dom,Res).
domain_wf1(Rel,Res,WF) :-   %% print_term_summary(domain_normal(Rel,Res)), %%
   expand_custom_set_to_list(Rel,Relation,_,domain_wf),
   newdomain1(Relation,[],Res,WF).


%:- block newdomain1(-,?,-,?).
:- block newdomain1(-,?,?,?).
newdomain1(Rel,SoFar,Res,WF) :-
    var(Rel),
    !, % print(prop(Res)),nl,
    domain_propagate_result(Res,Rel,SoFar,WF).
newdomain1(Dom,SoFar,Res,WF) :- newdomain2(Dom,SoFar,Res,WF).

% propagate information from domain to relation
domain_propagate_result([],Rel,SoFar,WF) :- !, check_domain_subset(Rel,SoFar,WF).
domain_propagate_result([H|TR],Rel,SoFar,WF) :-  !,
   % propagate known domain element into Relation Rel
    Rel = [(H,_)|TD], % equal_cons_wf(Rel,(H,_),TD,WF) ?
    not_element_of_wf(H,SoFar,WF),
    add_element(H,SoFar,SoFar2),
    newdomain1(TD,SoFar2,TR,WF).
domain_propagate_result(Res,Rel,SoFar,WF) :- % not a list
    expand_custom_set_to_list(Res,ResultList,_,domain_propagate_result),
    newdomain1(Rel,SoFar,ResultList,WF).

:- block check_domain_subset(-,?,?).
check_domain_subset([],_,_).
check_domain_subset([(X,_Y)|T],SoFar,WF) :- % print(check_dom(X,_Y,SoFar)),nl,
  check_element_of_wf(X,SoFar,WF),
  check_domain_subset(T,SoFar,WF).
  
%:- block newdomain2(-,?,?,?).
newdomain2([],_SoFar,Res,_WF) :- empty_set(Res). 
newdomain2([(X,_Y)|T],SoFar,Res,WF) :-  
  (Res==[]
    -> MemRes=pred_true, check_element_of_wf(X,SoFar,WF)
    ;  membership_test(SoFar,X,MemRes)
   ),
   newdomain3(MemRes,X,T,SoFar,Res,WF).

:- block newdomain3(-,?,?,?,?,?).
newdomain3(pred_true,_,T,SoFar,Res,WF) :- newdomain1(T,SoFar,Res,WF).
newdomain3(pred_false,X,T,SoFar,Res,WF) :- add_element(X,SoFar,SoFar2),
    equal_cons_wf(Res,X,Res2,WF), %print(added(X,SoFar,SoFar2,Res,Res2)),nl,
    newdomain1(T,SoFar2,Res2,WF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_domain_wf(int(2),[(int(2),int(7))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_domain_wf(int(2),[(int(1),int(6)),(int(2),int(7))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_domain_wf(int(22),[(int(1),int(6)),(int(22),int(7)),(int(33),int(7))],WF),WF)).
:- assert_must_succeed((bsets_clp:in_domain_wf(int(1),[(int(1),int(2))],_))).
:- assert_must_succeed((bsets_clp:in_domain_wf(int(3),[(int(1),int(2)),(int(3),int(4))],_))).
:- assert_must_fail((bsets_clp:in_domain_wf(int(3),[],_))).
:- assert_must_fail((bsets_clp:in_domain_wf(int(3),[(int(1),int(2))],_))).
 /* a more efficient version than using element_of and computing domain */
 
:- block in_domain_wf(?,-,?).
in_domain_wf(_DomainElement,[],_WF) :- !,fail. 
in_domain_wf(DomainElement,avl_set(A),_WF) :-  
     ground(DomainElement), !, %print(enter_in_dom_avl(DomainElement)),nl,
     check_in_domain_of_avlset(DomainElement,A).
 % TO DO: check for infinite closures
in_domain_wf(DomainElement,ES,WF) :-
   is_custom_explicit_set(ES,in_domain_wf), %print(enter_in_dom(ES)),nl,
    domain_of_explicit_set(ES,Dom),!,
   % print(in_domain_of_explicit_set(DomainElement,Dom,ES)),nl,
    check_element_of_wf(DomainElement,Dom,WF).
in_domain_wf(El,Rel,WF) :- %print(in_dom(El)),nl,
   expand_custom_set_to_list(Rel,Relation,_,in_domain_wf), 
   get_binary_choice_wait_flag(in_domain_wf(El),WF,LWF),
   in_domain2(El,Relation,WF,LWF).
 
:- block in_domain2(?,-,?,?).
in_domain2(El,[(X,_Y)|T],WF,LWF) :- 
   kernel_objects:equality_objects_lwf(El,X,EqRes,LWF),
   in_domain3(EqRes,El,T,WF,LWF).

:- block in_domain3(-,?,?,?,?).
in_domain3(pred_true,_El,_T,_WF,_LWF).
in_domain3(pred_false,El,T,WF,LWF) :- 
   get_new_subsidiary_wait_flag(LWF,in_domain2(El,T),WF,NewLWF), in_domain2(El,T,WF,NewLWF).


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_in_domain(int(3),[]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_in_domain(int(3),[(int(2),int(7))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_in_domain(int(3),[(int(2),int(7)),(int(4),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_in_domain(int(4),[(int(2),int(7)),(int(4),int(3))]))).
:- assert_must_fail((bsets_clp:not_in_domain(int(1),[(int(1),int(2))]))).
:- assert_must_fail((bsets_clp:not_in_domain(int(3),[(int(1),int(2)),(int(3),int(4))]))).
:- assert_must_succeed((bsets_clp:not_in_domain(int(3),[]))).
:- assert_must_succeed((bsets_clp:not_in_domain(int(3),[(int(1),int(2))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_in_domain(int(3),[(int(1),int(2)),(int(2),int(3))]))).
 /* a more efficient version than using not_element_of and computing domain */

:- block not_in_domain(?,-).
not_in_domain(DomainElement,ES) :- is_custom_explicit_set(ES,not_in_domain), 
    domain_of_explicit_set(ES,Dom),!,
    %print_term_summary(not_in_domain2(DomainElement,Dom,ES)),
    not_element_of(DomainElement,Dom).   
%:- use_module(btypechecker,[couplise_list/2]).
%not_in_domain(DomainElement,closure(Params,Types,Pred)) :- !,
%    couplise_list(Types,Type),
%    ~~mnf( Type = couple(_DomainType,RangeType) ),
%    Couple = (DomainElement,RangeElement),
%    delay_not( (
%                 kernel_waitflags:init_wait_flags(LocalWF), % we are not creating inner waitflags here as the exist will wait anyway
%		        kernel_objects:check_element_of_wf(Couple,closure(Params,Types,Pred),LocalWF),
%                 kernel_waitflags:ground_constraintprop_wait_flags(LocalWF),
%                 b_enumerate:b_enumerate_values([typedval(RangeElement,RangeType,_Var)]),
%                 kernel_waitflags:ground_wait_flags(LocalWF)
%	       ), [LocalWF,RangeElement]).
not_in_domain(El,Rel) :- %%print(not_in_dom(El,Rel)),nl,
   expand_custom_set_to_list(Rel,Relation,_,not_in_domain), 
   not_in_domain2(Relation,El).
:- block not_in_domain2(-,?).
not_in_domain2([],_).
not_in_domain2([(X,_Y)|T],E) :- not_equal_object(E,X), not_in_domain2(T,E).




:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:range_wf([],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:range_wf([(int(1),int(3))],[int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:range_wf(
    [(int(0),int(55)),(int(2),int(3)),(int(1),int(3))],[int(3),int(55)],WF),WF)).
:- assert_must_succeed((bsets_clp:range_wf([],Res,_WF),Res=[])).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(2))],Res,_WF),
           kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(1)),(int(2),int(1))],Res,_WF),
           kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(2)),(int(1),int(1))],Res,_WF),
                        kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(2)),(int(1),int(1)),(int(2),int(3))],Res,_WF),
                        kernel_objects:equal_object(Res,[int(1),int(3),int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,_WF),
                        X = [(int(1),int(2)),(int(1),int(1)),(int(2),int(3))],
                        kernel_objects:equal_object(Res,[int(1),int(3),int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,WF), bsets_clp:domain_wf(X,Res2,WF), kernel_objects:equal_object(Res,Res2),
                        X = [(int(1),int(2)),(int(1),int(1)),(int(2),int(2))])).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,WF), bsets_clp:domain_wf(X,Res2,WF), kernel_objects:equal_object(Res,Res2),
                        X = [(int(2),int(1)),(int(1),int(2)),(int(2),int(2))])).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,WF), bsets_clp:domain_wf(X,Res2,WF), kernel_objects:equal_object(Res,Res2),
                        X = [])).
:- assert_must_succeed((bsets_clp:range_wf([([],[]),([int(0)],[int(0)]),
 ([int(0),int(1)],[int(0),int(1)]),([int(0),int(2)],[int(0),int(2)]),
 ([int(0),int(3)],[int(0),int(3)]),([int(0),int(4)],[int(0),int(4)]),([int(1)],[int(1)]),
 ([int(1),int(2)],[int(1),int(2)]),([int(1),int(3)],[int(1),int(3)]),
 ([int(1),int(4)],[int(1),int(4)]),([int(2)],[int(2)]),([int(2),int(3)],[int(2),int(3)]),
 ([int(2),int(4)],[int(2),int(4)]),([int(3)],[int(3)]),([int(3),int(4)],
                                   [int(3),int(4)]),([int(4)],[int(4)])],_Res,_WF))).
:- assert_must_succeed((bsets_clp:range_wf([([],[]),([int(0)],[int(0)]),
 ([int(0),int(1)],[int(0),int(1)]),
 ([int(0),int(3)],[int(0),int(3)]),([int(0),int(4)],[int(0),int(4)]),([int(1)],[int(1)]),
 ([int(1),int(2)],[int(1),int(2)])],_Res,_WF))).
                        
:- block range_wf(-,?,?).
range_wf(Rel,Res,_WF) :-  % print_term_summary(range_wf(Rel,Res)),
  is_custom_explicit_set(Rel,range_wf),
  range_of_explicit_set(Rel,Range), !,
   % print_term_summary(range_of_explicit_set(Rel,Range)),
  equal_object(Range,Res).
range_wf(Rel,Res,WF) :-  % print_term_summary(range_normal(Rel,Res)),
 % TO DO : propagate information that card of Res <= card of Rel; similar thing for domain
   expand_custom_set_to_list(Rel,Relation,_,range_wf), %print(exp(Relation)),nl,
   % print(newrange2(Relation,[],Res)),nl,
   newrange2(Relation,[],Res,WF).  
   
:- block newrange2(-,?,?,?).
newrange2([],_SoFar,Res,_WF) :-  %print(range2_base(Acc,Res)),nl,
   empty_set(Res). % ,print(eq),nl.
newrange2([(_X,Y)|T],SoFar,Res,WF) :-  %print(range2(Y,T,SoFar,Res)),nl,
  (Res==[]
    -> MemRes=pred_true, check_element_of_wf(Y,SoFar,WF)
    ;  membership_test(SoFar,Y,MemRes)
   ),
   newrange3(MemRes,Y,T,SoFar,Res,WF).

:- block newrange3(-,?,?,?,?,?).
newrange3(pred_true,_Y,T,SoFar,Res,WF) :- newrange2(T,SoFar,Res,WF).
newrange3(pred_false,Y,T,SoFar,Res,WF) :- add_element(Y,SoFar,SoFar2),
    equal_cons_wf(Res,Y,Res2,WF),
    % print_bt_message(added(Y,T,SoFar,sofar2(SoFar2),Res,res2(Res2))),
    newrange2(T,SoFar2,Res2,WF).
   
   
:- assert_must_succeed((bsets_clp:identity_relation_over_wf([],Res,_WF),Res=[])).
:- assert_must_succeed((bsets_clp:identity_relation_over_wf([int(1),int(2)],Res,_WF),
    Res=[(int(1),int(1)),(int(2),int(2))])).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:identity_relation_over_wf([int(2),int(4)],[(int(4),int(4)),(int(2),int(2))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:identity_relation_over_wf([int(1),int(2),int(4)],[(int(4),int(4)),(int(2),int(2)),(int(1),int(1))],WF),WF)).
:- assert_must_fail((bsets_clp:identity_relation_over_wf([int(1)|_],_,_WF),fail)). /* check: no loop */

:- block identity_relation_over_wf(-,?,?).
identity_relation_over_wf(Set1,IDRel,WF) :-
    expand_custom_set_to_list(Set1,ESet1,_,identity_relation_over_wf),
    identity_relation_over2(ESet1,IDRel,WF).

:- block identity_relation_over2(-,?,?).
identity_relation_over2([],Res,_) :- empty_set(Res).
identity_relation_over2([X|T1],Res,WF) :- equal_cons_wf(Res,(X,X),T2,WF), % equal_object([(X,X)|T2],Res),
    identity_relation_over2(T1,T2,WF).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_identity((int(1),int(1)),[int(1),int(2)],WF),WF)).
:- assert_must_fail((bsets_clp:in_identity((int(1),int(2)),[int(1),int(2)],_WF))).
:- assert_must_fail((bsets_clp:in_identity((int(3),int(3)),[int(1),int(2)],_WF))).
:- assert_must_fail((bsets_clp:in_identity((int(1),int(2)),[],_WF))).
in_identity((X,Y),Domain,WF) :- %print(in_identity((X,Y),Domain)),nl,
    equal_object(X,Y), check_element_of_wf(X,Domain,WF). 
:- assert_must_fail((bsets_clp:not_in_identity((int(1),int(1)),[int(1),int(2)],_WF))).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_identity((int(1),int(2)),[int(1),int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_identity((int(3),int(3)),[int(1),int(2)],WF),WF)).
:- assert_must_succeed((bsets_clp:not_in_identity((int(1),int(2)),[],_WF))).
not_in_identity((X,Y),Domain,WF) :- %print(not_in_identity((X,Y),Domain)),nl,
    equality_objects(X,Y,Eq),not_in_id2(Eq,X,Domain,WF).

:- block not_in_id2(-,?,?,?).
not_in_id2(pred_true,X,Domain,WF) :- not_element_of_wf(X,Domain,WF).
not_in_id2(pred_false,_,_,_).
 

:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:invert_relation_wf([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], [(int(6),int(5)),(int(2),int(1)),(int(4),int(3))],WF),WF)).  
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:invert_relation_wf([(int(1),int(2))], [(int(2),int(1))],WF),WF)).    
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:invert_relation_wf([], [],WF),WF)).  
:- assert_must_succeed((bsets_clp:invert_relation_wf(X,X,_),X = [])).
:- assert_must_succeed((bsets_clp:invert_relation_wf(X,X,_),X = [(int(2),int(2))])).
:- assert_must_succeed((bsets_clp:invert_relation_wf(X,[(int(1),int(2)),(int(7),int(6))],_WF),
                     X = [(int(2),int(1)),(int(6),int(7))])).
:- assert_must_succeed((bsets_clp:invert_relation_wf([(int(1),int(2)),(int(7),int(6))],X,_WF),
                     X = [(int(2),int(1)),(int(6),int(7))])).
:- assert_must_succeed((bsets_clp:invert_relation_wf([(int(1),int(2)),(int(7),int(6))],
                           [(int(6),int(7)),(int(2),int(1))],_WF))).
:- assert_must_succeed((bsets_clp:invert_relation_wf(closure([a,b],[string,boolean],b(truth,pred,[])),
					     closure([b,a],[boolean,string],b(truth,pred,[])),_WF))).

:- block invert_relation_wf(-,-,?).
invert_relation_wf(R,IR,WF) :- % print_message(invert_relation(R,IR)),
  %  (nonvar(R) -> invert_relation2(R,IR) ; invert_relation2(IR,R)).
     invert_relation2(R,IR,WF). % , print_term_summary(invert_relation(R,IR)).
/* Optimization for some types of closures: Instead of expanding the closures, we just
   swap the parameters. This does not work with closures wich have only one parameter
   wich is a pair */
invert_relation2(CS,R,_WF) :- nonvar(CS),is_custom_explicit_set(CS,invert_relation),!,
     invert_explicit_set(CS,ICS), equal_object(R,ICS).
invert_relation2(R,CS,_WF) :- nonvar(CS),is_custom_explicit_set(CS,invert_relation),!,
     invert_explicit_set(CS,ICS), equal_object(R,ICS).
%invert_relation2(closure([P1,P2],[T1,T2],Clo),closure([P2,P1],[T2,T1],Clo)) :- !.
invert_relation2(R,IR,WF) :- %try_expand_custom_set(R,ER),
  %  (nonvar(R) -> invert_relation3(R,IR)
  %              ; invert_relation3(IR,R),(ground(IR)-> true ; invert_relation3(R,IR))).
  % print(inverting(R,IR,WF)),nl,
    invert_relation3(R,IR,WF,1), invert_relation3(IR,R,WF,1).
  %  ,print(inverted(R,IR)),nl. % propagates both ways; but multiple solutions ?!

:- block invert_relation3(-,?,?,?).
invert_relation3(closure(P,T,B),Res,_WF,_) :- invert_explicit_set(closure(P,T,B),ICS), equal_object(Res,ICS).
invert_relation3(avl_set(S),Res,_WF,_) :- invert_explicit_set(avl_set(S),ICS), equal_object(Res,ICS).
invert_relation3([],Res,_LWF,_) :- empty_set(Res).
invert_relation3([(X,Y)|T],Res,WF,Depth) :- %prints(invert([(X,Y)|T],Res,Depth)),
       D1 is Depth+1, get_wait_flag(D1,invert_relation3,WF,LWF),
       equal_cons_lwf(Res,(Y,X),IT,LWF),
       invert_relation3(T,IT,WF,D1).

      
 

tuple_of(X,Y,R) :- check_element_of((X,Y),R).
%tuple_of_wf(X,Y,R,WF) :- check_element_of_wf((X,Y),R,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], [(int(6),int(7)),(int(2),int(1)),(int(22),int(22)),(int(4),int(33))],
[(int(1),int(1)),(int(5),int(7)),(int(3),int(33))]))). 
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([], [(int(6),int(7)),(int(2),int(1)),(int(22),int(22)),(int(4),int(33))],[]))). 
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([(int(6),int(7)),(int(2),int(1)),(int(22),int(22)),(int(4),int(33))],[],[]))).   
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([],[],[]))).  
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(6))],
                                             [(int(1),int(11))],X),X = [])).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(6))],[],X),X = [])).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(6))],
                                                  [(int(2),int(11))],X),
                     kernel_objects:equal_object(X,[(int(1),int(11))]))).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(2))],[(int(2),int(11))],X),
                    ground(X), bsets_clp:equal_object(X,[(int(1),int(11)),(int(7),int(11))]))).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(5))],
                       [(int(2),int(11)),(int(2),int(4))],X),
                     kernel_objects:equal_object(X,[(int(1),int(11)),(int(1),int(4))]))).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(1),int(5))],
                       [(int(2),int(11)),(int(5),int(11))],X),
                    kernel_objects:equal_object(X,[(int(1),int(11))]))).

rel_composition(Rel1,Rel2,Comp) :-
      init_wait_flags(WF), % TO DO: remove and add WF arg
      rel_composition_wf(Rel1,Rel2,Comp,WF),
      ground_wait_flags(WF).

:- block rel_composition_wf(-,-,?,?).
rel_composition_wf(Rel1,Rel2,Comp,_WF) :- (Rel1==[] ; Rel2==[]),!, empty_set(Comp).
rel_composition_wf(Rel1,Rel2,Comp,WF) :- rel_composition1(Rel1,Rel2,Comp,WF).

:- block rel_composition1(-,?,?,?),rel_composition1(?,-,?,?).
rel_composition1(Rel1,Rel2,Comp,_WF) :- (Rel1==[] ; Rel2==[]),!, empty_set(Comp).
rel_composition1(Rel1,Rel2,Comp,_WF) :- rel_composition_for_explicit_set(Rel1,Rel2,Res),!,
  %print_term_summary(rel_composition_for_explicit_set(Rel1,Rel2,Res)),
  equal_object(Res,Comp).
rel_composition1(Rel1,Rel2,Comp,WF) :- Rel2=closure(_,_,_),
  keep_symbolic(Rel2),
  dom_for_specific_closure(Rel2,Domain,function(_)),!,
  %print(rel_compose_inf_function(Rel1,Rel2,Domain)),nl,
  expand_custom_set_to_list(Rel1,Relation1,_,rel_composition1),
  rel_compose_inf(Relation1,Domain,Rel2,Comp,WF).
rel_composition1(Rel1,Rel2,Comp,_WF) :- 
  expand_custom_set_to_list(Rel1,Relation1,_,rel_composition1_2),
  expand_custom_set_to_list(Rel2,Relation2,_,rel_composition1_3),
  rel_compose2(Relation1,Relation2,Comp).

:- block rel_compose2(-,?,?).
rel_compose2([],_,Out) :- empty_set(Out).
rel_compose2([(X,Y)|T],Rel2,Out) :-
   rel_extract(Rel2,X,Y,OutXY,[]),
  % rel_extract(Rel2,X,Y,Out,OutRem),
   rel_compose2(T,Rel2,OutRem),
   union(OutRem,OutXY,Out).

:- block rel_extract(-,?,?,?,?).
rel_extract([],_,_,Rem,Rem).
rel_extract([(Y1,Z)|T],X,Y,Res,Rem) :- 
   rel_extract(T,X,Y,CT,Rem),
   kernel_objects:equality_objects(Y1,Y,EqRes),
   rel_extract2(EqRes,Z,X,CT,Res).

:- block rel_extract2(-,?,?,?,?).
rel_extract2(pred_true,  Z, X,CT,Res) :- add_element((X,Z),CT,Res).
rel_extract2(pred_false,_Z,_X,CT,Res) :- Res = CT.

   
:- block rel_compose_inf(-,?,?,?,?).
rel_compose_inf([],_Dom,_Rel2,Comp,_WF) :- empty_set(Comp).
rel_compose_inf([(X,Y)|T],Dom,Fun,CompRes,WF) :-
   membership_test(Dom,Y,MemRes),
   rel_compose_inf_aux(MemRes,X,Y,T,Dom,Fun,CompRes,WF).

:- block rel_compose_inf_aux(-,?,?,?, ?,?,?,?).
rel_compose_inf_aux(pred_true,X,Y,T,Dom,Fun,CompRes,WF) :-
   % print(rel_compose_inf_aux(X,Y)),nl,
   apply_to(Fun,Y,FY,WF), % TO DO: generalize to image so that we can apply it also to infinite relations ?
   add_element((X,FY),CT,CompRes),
   rel_compose_inf(T,Dom,Fun,CT,WF).
rel_compose_inf_aux(pred_false,_X,_Y,T,Dom,Fun,Comp,WF) :-
    % print(rel_compose_not_in_domain(_X,_Y)),nl,
    rel_compose_inf(T,Dom,Fun,Comp,WF).

:- assert_must_abort_wf(bsets_clp:rel_iterate([],int(-1),_R,WF),WF).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([], int(2),[],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], int(1),[(int(1),int(2)),(int(3),int(4)),(int(5),int(6))],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], int(0),[(int(1),int(1)),(int(2),int(2)),(int(3),int(3)),(int(4),int(4)),(int(5),int(5)),(int(6),int(6))],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([(int(1),int(2)),(int(2),int(4)),(int(4),int(6))], int(2),[(int(1),int(4)),(int(2),int(6))],_WF))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(1),X,_WF), R=[],
                        bsets_clp:equal_object(X,R))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(1),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3))],
                        bsets_clp:equal_object(X,R))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(2),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3))],
         bsets_clp:equal_object(X,[(int(1),int(3))]))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(3),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3))],
         bsets_clp:equal_object(X,[]))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(3),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3)),(int(1),int(1))],
         bsets_clp:equal_object(X,[(int(1),int(1)),(int(1),int(2)),(int(1),int(3))]))).

rel_iterate(Rel,int(Nr),Res,WF) :- rel_iterate2(Nr,Rel,Res,WF).
  
:- block rel_iterate2(-,?,?,?).
rel_iterate2(X,Rel,Res,WF) :- 
    ( X=1 -> equal_object(Res,Rel)
    ; X>1 -> X1 is X-1,
        rel_iterate2(X1,Rel,R1,WF),
        rel_composition(Rel,R1,Res) % TO DO: call WF version
    ; X=0 -> rel_iterate0(Rel,Res)
    ; otherwise -> add_wd_error('negative index in iterate',X,WF)
    ).

:- block rel_iterate0(-,?).     
rel_iterate0(Rel,Res) :-
    expand_custom_set_to_list(Rel,Relation,_,rel_iterate0),
   % nl,print(start_rel_iterate0(Relation)),nl,
   rel_iterate0(Relation,[],Res).

:- block rel_iterate0(-,?,?).
rel_iterate0([],D,Res) :-  %print(result_rel_iterate0(D,Res)),nl,nl,
    equal_object(D,Res).
rel_iterate0([(X,Y)|T],ResSoFar,Res) :-
 % TO DO: add (X,X) and (Y,Y) earlier to Res  to improve constraint propagation
   add_element((X,X),ResSoFar,Res2),
   add_element((Y,Y),Res2,Res3),
   rel_iterate0(T,Res3,Res).


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:direct_product_wf([],[(int(1),int(11))],[],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
 [(int(1),int(11))],[(int(1),(int(2),int(11)))],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
 [(int(2),int(11))],[],_WF))).
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
                                             [(int(2),int(11))],X,_WF),
                                             X = [])).
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
                                             [(int(1),int(11))],X,_WF),
                                             kernel_objects:equal_object(X,[(int(1),(int(2),int(11)))]))).   
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(1),int(6))],
                                             [(int(1),int(11))],X,_WF),
          kernel_objects:equal_object(X,[(int(1),(int(2),int(11))),(int(1),(int(6),int(11)))]))). 
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(2),int(6))],
                                             [(int(1),int(11)),(int(1),int(12))],X,_WF),
          kernel_objects:equal_object(X,[(int(1),(int(2),int(11))),(int(1),(int(2),int(12)))]))). 
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(2),int(6))],
                                             [(int(1),int(11)),(int(1),int(12))],
                          [(int(1),(int(2),int(11))),(int(1),(int(2),int(12)))],_WF))). 
:- assert_must_succeed((bsets_clp:direct_product_wf(avl_set(node((fd(1,'Name'),fd(2,'Name')),true,1,empty,node((fd(2,'Name'),fd(3,'Name')),true,0,empty,empty))),
avl_set(node((fd(1,'Name'),fd(2,'Name')),true,1,empty,node((fd(2,'Name'),fd(3,'Name')),true,0,empty,empty))),
avl_set(node((fd(1,'Name'),fd(2,'Name'),fd(2,'Name')),true,1,empty,node((fd(2,'Name'),fd(3,'Name'),fd(3,'Name')),true,0,empty,empty)))
,_WF))). 

:- block direct_product_wf(-,?,?,?),direct_product_wf(?,-,?,?).
direct_product_wf(Rel1,Rel2,Prod,WF) :-
   try_expand_and_convert_to_avl_with_check(Rel1,E1,direct_product),
   try_expand_and_convert_to_avl_with_check(Rel2,E2,direct_product),
   direct_product_wf1(E1,E2,Prod,WF).

direct_product_wf1(Rel1,Rel2,Prod,_WF) :-  % print_term_summary(direct_product(Rel1,Rel2,Prod)), %
  direct_product_explicit_set(Rel1,Rel2,Res),!,
 % print_term_summary(direct_product_result(Res)),
  equal_object(Prod,Res).
direct_product_wf1(Rel1,Rel2,Prod,WF) :-  %print_term_summary(direct_product_list(Rel1,Rel2,Prod)),
  expand_custom_set_to_list(Rel1,Relation1,_,direct_product_wf1_1),
  expand_custom_set_to_list(Rel2,Relation2,_,direct_product_wf1_2),
  direct_product2(Relation1,Relation2,Prod,WF),
  direct_product_backwards(Relation1,Relation2,Prod,WF).

%direct_product2(R1,R2,Res) :- print_message(direct_product2(R1,R2,Res)),fail.  
:- block direct_product2(-,?,?,?).
direct_product2([],_,Out,_WF) :- equal_object(Out,[]).
direct_product2([(X,Y)|T],Rel2,Out,WF) :-
   direct_product_tuple(Rel2,X,Y,Out,OutRem,WF),
   direct_product2(T,Rel2,OutRem,WF).

:- block direct_product_tuple(-,?,?,?,?,?).
direct_product_tuple([],_,_,Res,Rem,_WF) :- equal_object(Res,Rem).
direct_product_tuple([(X2,Z)|T],X,Y,Res,Rem,WF) :- 
   direct_product_tuple(T,X,Y,CT,Rem,WF),
   kernel_objects:equality_objects(X2,X,EqRes),
   direct_product_tuple3(EqRes,X,Y,Z,CT,Res,WF).
   
:- block direct_product_tuple3(-,?,?,?,?,?,?).
direct_product_tuple3(pred_true,X,Y,Z,CT,Res,WF) :- 
    equal_cons_wf(Res,(X,(Y,Z)),CT,WF).  /* no need for add_element as output uniquely determines X,Y,Z !?*/
direct_product_tuple3(pred_false,_X,_Y,_Z,CT,Res,_WF) :-  equal_object(Res,CT).

:- block direct_product_backwards(?,?,-,?).
% Propagate information backwards from result to arguments
direct_product_backwards(R1,R2,Prod,WF) :-
   ((ground(R1);ground(R2)) -> true
    ; expand_custom_set_to_list(Prod,ProdList,_,direct_product_backwards),
      direct_product_propagate_back(ProdList,R1,R2,WF)
   ).

:- block direct_product_propagate_back(-,?,?,?).
direct_product_propagate_back([],_,_,_WF).
direct_product_propagate_back([(X,(Y,Z))|T],R1,R2,WF) :- %% print(prop_back(X,Y,Z,R1,R2)),nl, %
   check_element_of_wf((X,Y),R1,WF), check_element_of_wf((X,Z),R2,WF),
   direct_product_propagate_back(T,R1,R2,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:parallel_product([],[(int(3),int(4))],[]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:parallel_product([(int(1),int(2))],
                        [(int(3),int(4))],[((int(1),int(3)),(int(2),int(4)))]))).
:- assert_must_succeed((bsets_clp:parallel_product([(int(1),int(2))],
                                             [(int(3),int(4))],X),
                                             X == [((int(1),int(3)),(int(2),int(4)))])).
:- assert_must_succeed((bsets_clp:parallel_product([(int(1),int(2))],
                                             [(int(3),int(4))],[((int(1),int(3)),(int(2),int(4)))]))).
:- assert_must_succeed((bsets_clp:parallel_product([(int(1),int(2))],
                                             [],X),X == [])).
:- assert_must_succeed((bsets_clp:parallel_product([],
                                             [(int(3),int(4))],X),X == [])).

:- block parallel_product(-,?,?),parallel_product(?,-,?).
parallel_product(Rel1,Rel2,Prod) :- 
  expand_custom_set_to_list(Rel1,Relation1,_,parallel_product_1),
  expand_custom_set_to_list(Rel2,Relation2,_,parallel_product_2),
  parallel_product2(Relation1,Relation2,ProdRes),
  equal_object(ProdRes,Prod).
 
:- block parallel_product2(-,?,?).
parallel_product2([],_,[]).
parallel_product2([(X,Y)|T],Rel2,Out) :-
   parallel_product_tuple(Rel2,X,Y,Out,Tail),
   parallel_product2(T,Rel2,Tail).

:- block parallel_product_tuple(-,?,?,?,?).
parallel_product_tuple([],_,_,Tail,Tail).
parallel_product_tuple([(X2,Y2)|T],X,Y,[((X,X2),(Y,Y2))|RT],Tail) :- 
    parallel_product_tuple(T,X,Y,RT,Tail).


% -------------------------------------------------
   
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7))],[int(1)],[int(7),int(6)],WF),WF)). %% with wf_det leads to residue custom_explicit_sets:b_not_test_closure_enum
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(8),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(1),int(7))],[int(1)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_partial_function([],[int(1)],[int(7)],_WF))).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1)],[int(7)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1)],[[int(7),int(6)]],_WF),
                     X = [(int(1),[int(6),int(7)])])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(2))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(3),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(5))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(1),int(7)),(int(2),int(6)),(int(2),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(5),int(75))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(-1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('Name'),global_set('Code'),_WF),
                     X = [(fd(1,'Name'),fd(1,'Code'))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(88),fd(2,'Code'))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(2),fd(2,'Code'))])).
:- assert_must_succeed((bsets_clp:not_partial_function([(fd(1,'Code'),int(1)),(fd(1,'Code'),int(2))],
                       global_set('Code'),global_set('NAT1'),_WF) )).

:- block not_partial_function(-,?,?,?).
not_partial_function([],_Domain,_Range,_WF) :- !,fail.
not_partial_function(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    %print_term_summary(not_partial_fun_dom(FF,FFDomain,Range)),
    not_subset_of_wf(FFDomain,Domain,WF).
not_partial_function(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    %print_term_summary(check_special_not_pf(FF,FFDomain,FFRange)),nl,
    not_both_subset_of(FFDomain,FFRange,Domain,Range,WF).
not_partial_function(R,Domain,Range,_WF) :-
   expand_and_convert_to_avl_set(R,AER),!,
   % print_term_summary(check_is_not_avl_partial_function(avl_set(AER),Domain,Range)),
   (is_avl_partial_function(AER)
     ->  % print(is_pf),nl,
        is_not_avl_relation_over_domain_range(AER,Domain,Range)
        %, print(not_in_domain_range),nl
     ;  true %,print(not_pf),nl
   ).
not_partial_function(R,Domain,Range,WF) :- 
   expand_custom_set_to_list(R,ER,_,not_partial_function),
   not_pf(ER,[],Domain,Range,WF).

:- block not_pf(-,?,?,?,?).
not_pf([],_,_,_,_) :- fail.
not_pf([(X,Y)|T],SoFar,Dom,Ran,WF) :- not_pf2(X,Y,T,SoFar,Dom,Ran,WF).

:- block not_pf2(-,?,?,?,?,?,?).
not_pf2(X,_Y,_T,SoFar,_Dom,_Ran,WF) :- 
     % TO DO: use memberhip_test
    check_element_of_wf(X,SoFar,WF). /* then not a function; was check_element_of_lazy */
not_pf2(X,_Y,_T,SoFar,Dom,_Ran,WF) :- 
    not_element_of_wf(X,SoFar,WF), not_element_of_wf(X,Dom,WF). /* function, but domain wrong */
not_pf2(X,Y,T,SoFar,Dom,Ran,WF) :- 
    not_element_of_wf(X,SoFar,WF),
    remove_element_wf_if_not_infinite_or_closure(X,Dom,Dom2,WF,_LWF,Done),  %% provide _LWF ??
    not_pf2b(Done,X,Y,T,SoFar,Dom2,Ran,WF).
:- block not_pf2b(-, ?,?,?, ?,?,?, ?).
not_pf2b(_Done, X,Y,T, SoFar,Dom2,Ran, WF) :-   
    add_element(X,SoFar,SoFar2),
    membership_test(Ran,Y,MemRes),
    not_pf3(MemRes,T,SoFar2,Dom2,Ran,WF).

:- block not_pf3(-,?,?,?,?,?).
not_pf3(pred_false,_T,_SoFar,_Dom2,_Ran,_WF).   /* illegal range */
not_pf3(pred_true,T,SoFar,Dom2,Ran,WF)  :- 
    not_pf(T,SoFar,Dom2,Ran,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_function_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_function_wf([(int(1),int(1)),(int(2),int(1))],global_set('NATURAL'),global_set('NATURAL'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_function_wf([(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).  
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:partial_function_wf([(int(2),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed((bsets_clp:partial_function([],[int(1)],[int(7)]))).
:- assert_must_succeed((bsets_clp:partial_function(X,[int(1)],[int(7)]),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:partial_function(X,[int(1),int(2)],[int(7)]),
                     equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_succeed((findall(X,bsets_clp:partial_function(X,[int(1),int(2)],[int(7)]),L),
                          length(L,4) )).
:- assert_must_succeed((bsets_clp:partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     equal_object(X,[([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))]))).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,[int(1)],[[int(7),int(6)]],_WF),
                     X = [(int(1),[int(6),int(7)])])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(5),int(75))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_fail((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(-1),int(7))])).
:- assert_must_fail((bsets_clp:partial_function_wf(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_fail((bsets_clp:partial_function_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('Name'),global_set('Code'),_WF),
                     X = [(fd(1,'Name'),fd(1,'Code'))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(88),fd(2,'Code'))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(2),fd(2,'Code'))])).
                        
partial_function(R,Domain,Range) :- init_wait_flags(WF),
  partial_function_wf(R,Domain,Range,WF),
 % print(grounding_pf_wait_flags(WF)),nl,
  ground_wait_flags(WF).
  
:- use_module(kernel_equality,[get_cardinality_wait_flag/5]).
:- block partial_function_wf(-,-,?,?).
partial_function_wf(R,_Domain,_Range,_WF) :- R==[], !.
partial_function_wf(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    % print_term_summary(partial_fun_dom(FF,FFDomain,Range)),
    check_subset_of_wf(FFDomain,Domain,WF).
partial_function_wf(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    % same as for total_function_wf check
    %print(check_special(FF,FFDomain,FFRange)),nl,
    check_subset_of_wf(FFDomain,Domain,WF),
    check_subset_of_wf(FFRange,Range,WF). %, print(special_partial_fun(FF)),nl. 
partial_function_wf(R,Domain,Range,WF) :- %print(pf(R)),nl,
   expand_and_convert_to_avl_set(R,AER),!,
   % print_term_summary(check_is_avl_partial_function(avl_set(AER),Domain,Range)),
   is_avl_partial_function(AER),
   % print(is_pf),nl,
   is_avl_relation_over_domain(AER,Domain,WF), %print(dom_ok),nl,
   is_avl_relation_over_range(AER,Range,WF).
partial_function_wf(R,Domain,Range,WF) :-  % print(pf(R,Domain,Range,WF)),nl, %%
   %get_real_cardinality(Domain,Card),  /* Card is free variable if Domain infinite */
   get_cardinality_wait_flag(Domain,partial_function_wf,WF,Card,CWF),
   %% print_bt_message(pf_dom_card(Card)),nl, %%%
   when((nonvar(R);nonvar(CWF)),
       (expand_custom_set_to_list(R,ER,_,partial_function_wf),
        %print(expanded(R,ER,CWF)),nl,
        %get_last_wait_flag(partial_fun(Domain),WF,LWF),
        pf_w(ER,[],Domain,Range,Card,_Large,WF,CWF))).


pf_w(T,SoFar,Dom,Ran,Card,Large,WF,LWF) :-
    (Card==0 -> T=[]
     ; pf(T,SoFar,Dom,Ran,Card,Large,WF,LWF)).

:- block pf(-,?,?,?,?,?,?,-).
pf([],_,_,_,_,_WF,_,_LWF).
pf([(X,Y)|T],SoFar,Dom,Ran,Card,Large,WF,LWF) :- 
    dec_card(Card,NC),/* Card ensures we do not build too big lists */
    Dom \== [],
    remove_domain_element(X,Y,Dom,Dom2,Large,WF,LWF,Done),
       %%  prints(rem_dom(X,Dom,Dom2,Card,Done)),nl, %%
    pf1(Done, X,Y,T,SoFar,Dom2,Ran,NC,Large,WF,LWF).
pf(AVL,SoFar,Dom,Ran,Card,Large,WF,LWF) :- nonvar(AVL),AVL=avl_set(_A),
   add_internal_error('AVL arg: ',pf(AVL,SoFar,Dom,Ran,Card,Large,WF,LWF)),fail.

:- block dec_card(-,?).
dec_card(inf,NewC) :- !, NewC=inf.
dec_card(C,NewC) :- C>0, NewC is C-1.

:- block pf1(-, ?,?,?,?,?,?,?,?,?,?).
pf1(_Done, X,Y,T,SoFar,Dom2,Ran,Card,Large,WF,LWF) :-
    not_element_of_wf(X,SoFar,WF), /* check that it is a function */
    % prints(check(Y,Ran)), %%
    check_element_of_wf(Y,Ran,WF),
    add_new_element(X,SoFar,SoFar2), %print(added(SoFar2)),nl,
    pf_w(T,SoFar2,Dom2,Ran,Card,Large,WF,LWF).
    
remove_domain_element(X,Y,Dom,Dom2,Large,WF,LWF,Done) :- compute_large(Dom,Large),
   ((var(X),var(Y),Large==false,ground(Dom))  
      -> %% (X, Y are free: we can influence which element is taken from Dom
          % print_term_summary(removing_minimal_element(X,Y,Dom)),
          remove_a_minimal_element(X,Dom,Dom2,WF,Done)  %%%%%%%%%% added Jul 15 2008
          %,print_term_summary(done(X,Dom2,Done)) ) 
      ;  remove_element_wf_if_not_infinite_or_closure(X,Dom,Dom2,WF,LWF,Done)
    ).
compute_large(Dom,Large) :- % check if the domain is large; ensure that we compute this only once
 (nonvar(Large) -> true
   ; var(Dom) -> true
   ; dont_expand_this_explicit_set(Dom) -> Large=large
   ; Large=false).
    
:- assert_must_succeed(( bsets_clp:remove_a_minimal_element(X,[int(1)],R,_WF,Done),
                        X==int(1), Done==true, R=[] )).
:- assert_must_succeed(( init_wait_flags(WF), bsets_clp:remove_a_minimal_element(X,[int(1),int(2),int(3)],R,WF,Done), ground_wait_flags(WF),
                        X==int(2), Done==true, R=[int(3)] )).
:- assert_must_succeed(( init_wait_flags(WF), bsets_clp:remove_a_minimal_element(X,[int(1),int(2),int(3)],R,WF,Done), ground_wait_flags(WF),
                        X==int(1), R=[int(2),int(3)], Done==true )).
:- assert_must_succeed(( init_wait_flags(WF), bsets_clp:remove_a_minimal_element(X,[int(1),int(2),int(3)],R,WF,Done), ground_wait_flags(WF),
                        X==int(3), R=[], Done==true )).
:- assert_must_succeed((  init_wait_flags(WF), CL=closure(['_zzzz_binary'],[integer],b(member( b(identifier('_zzzz_binary'),integer,[]),
 b(interval(b(value(int(1)),integer,[]),b(value(int(10)),integer,[])),set(integer),[])),pred,[])),
               bsets_clp:remove_a_minimal_element(X,CL,R,WF,Done), ground_wait_flags(WF),
                        X=int(9), Done==true, kernel_objects:equal_object(R,[int(10)]) )).
          
/* usage: restrict number of possible choices if element to remove is free */
/* select one element; and disallow all elements appearing before it in the list */
remove_a_minimal_element(X,Set,Res,WF,Done) :- 
  expand_custom_set_to_list(Set,ESet,EDone,remove_a_minimal_element),
  remove_a_minimal_element2(X,ESet,EDone,Res,WF,Done).

:- use_module(kernel_equality,[get_cardinality_wait_flag/4]).
:- block remove_a_minimal_element2(?,?,-,?,?,?).
remove_a_minimal_element2(X,ESet,_EDone,Res,WF,Done) :-
 % print(removing_minimal_element2(X,ESet,Res,WF)),nl,%
  ESet \= [],
  (ESet = [El] -> (X=El, equal_object(Res,[]), Done=true ) % only one choice
    ; get_cardinality_wait_flag(ESet,remove_a_minimal_element2,WF,CWF),
      when(nonvar(CWF),(append(_,[X|TRes],ESet),equal_object(Res,TRes),Done=true))
  ).
 



    
 
    
 
   
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_function_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed((bsets_clp:total_function(X,[int(1)],[int(7)]),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3))]],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(3))],int(7)),([(int(1),int(2))],int(7))]))).
:- assert_must_succeed((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))]))).
:- assert_must_succeed((bsets_clp:total_function(X,[int(1)],[[int(7),int(6)]]),
                     kernel_objects:equal_object(X,[(int(1),[int(6),int(7)])]))).
:- assert_must_succeed((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [[int(7),int(6)]]),
                    kernel_objects:equal_object(X,[([(int(1),int(2))],[int(6),int(7)]),
                          ([(int(2),int(3)),(int(1),int(3))],[int(6),int(7)])]))).
:- assert_must_succeed((bsets_clp:total_function(X,[ [(int(1),int(3)),(int(2),int(3))]],
                                                [int(6)]),
                     kernel_objects:equal_object(X,[ ([(int(2),int(3)),(int(1),int(3))], int(6)) ]))).
:- assert_must_succeed((bsets_clp:total_function(X,global_set('Name'),
                 [[],[fd(1,'Code'),fd(2,'Code')],[fd(1,'Code')],[fd(2,'Code')]]),
   kernel_objects:enumerate_basic_type(X,set(couple(global('Name'),set(global('Code'))))),
    kernel_objects:equal_object(X,[(fd(3,'Name'),[fd(2,'Code')]),(fd(1,'Name'),[fd(2,'Code')]),(fd(2,'Name'),[])]))).

%:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),bsets_clp:total_function_wf(TF,global_set('Code'),
% closure([zzzz],[set(set(couple(integer,boolean)))],
%  member(identifier(zzzz),
%  pow_subset(value(closure([zzzz],[set(couple(integer,boolean))],
%  member('ListExpression'(['Identifier'(zzzz)]),
%  'Seq'(value([pred_true /* bool_true */,pred_false /* bool_false */])))))))),WF),
%  % print(tf(TF)),nl,
%   kernel_objects:equal_object(TF,[ (fd(1,'Code'), [[],[(int(1),pred_true /* bool_true */)],[(int(1),pred_true /* bool_true */),(int(2),pred_true /* bool_true */)]]),
%   (fd(2,'Code'), [[],[(int(1),pred_true /* bool_true */)],[(int(1),pred_true /* bool_true */),(int(2),pred_true /* bool_true */)]]) ]),
  % print(gr(WF)),nl,
%   kernel_waitflags:ground_wait_flags(WF) )).

:- assert_must_succeed((bsets_clp:total_function([],[],[int(7)]))).

:- assert_must_fail((bsets_clp:total_function([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).
:- assert_must_fail((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(5))]))).
:- assert_must_fail((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7))]))).
:- assert_must_fail((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(2))],int(7)),
                          ([(int(1),int(3)),(int(1),int(3))],int(6))]))).
:- assert_must_fail((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(3)),(int(1),int(3))],int(6))]))).

total_function(R,Domain,Range) :- init_wait_flags(WF),
  total_function_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).
  
  
:- assert_must_succeed((bsets_clp:total_function_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
 nonvar(X),X=[(A,B),(C,D)],A==int(1),C==int(2),\+ ground(B),\+ ground(D),  B=int(7),D=int(7)     )).

:- block total_function_wf(?,-,?,?). 
total_function_wf(FF,Domain,Range,_WF) :- 
    nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    % print_term_summary(total_fun_dom(FF,FFDomain,Range)),
    equal_object(FFDomain,Domain).
total_function_wf(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    equal_object(FFDomain,Domain), check_subset_of_wf(FFRange,Range,WF)
    .%, print(special_total_fun(FF)),nl. 
total_function_wf(R,Domain,Range,WF) :- nonvar(R), R=avl_set(AEF), !,
   total_function_avl_set(AEF,Domain,Range,WF).
total_function_wf(R,Domain,Range,WF) :-   % print(total_function(R,Domain,Range,WF)),nl,
     try_expand_and_convert_to_avl_with_check(Domain,EDomain,total_function),  /* avoid multiple expansions */
     %print(EDomain),nl,
     % TO DO: maybe avoid converting intervals which are not fully instantiated ?
     try_expand_and_convert_to_avl_unless_large(R,ER), 
     total_function_wf2(ER,EDomain,Range,WF).

:- block total_function_wf2(?,-,?,?). 
total_function_wf2(R,Domain,Range,WF) :- nonvar(R), R=avl_set(AEF), !,
   total_function_avl_set(AEF,Domain,Range,WF).
total_function_wf2(R,Domain,Range,WF) :- %print_term_summary(total_function_wf2(R,Domain,Range,WF)),
    cardinality(Domain,Card),
    %print(tf_card(R,Domain,Card)),nl,
    ((nonvar(R);ground(Card))-> true
      ; get_last_wait_flag(total_fun(Domain),WF,WF1)),
    when((nonvar(R);ground(Card);(nonvar(Card),nonvar(WF1))), /* mal 12/5/04: changed , into ; 17/3/2008: added WF1 */
          /* reason for delaying nonvar(Card): Card grounded bit by bit by cardinality; avoid
             triggering too early and missing tf_var */
       total_function1(R,Card,Domain,Range,WF
     )).

total_function_avl_set(AEF,Domain,Range,WF) :-
    %print(total_function_avl(avl_set(AEF),Dom,Range)),nl,
   (Domain = avl_set(Dom) -> is_avl_total_function_over_domain(AEF,Dom)
     ; is_avl_partial_function(AEF),
       domain_of_explicit_set(avl_set(AEF),AEF_Domain),
       equal_object(AEF_Domain,Domain)
    ),
   is_avl_relation_over_range(AEF,Range,WF).
%total_function1(FF,Card,Domain,Range,WF) :- nonvar(FF),
%   custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),
%   print(dom_range_for_specific_closure(FF,FFDomain,Domain,FFRange,Range)),nl,fail.
total_function1(FF,_,Domain,Range,WF) :- expand_and_convert_to_avl_set(FF,AEF),!,
    total_function_avl_set(AEF,Domain,Range,WF).
total_function1(R,Card,Domain,Range,WF) :- try_expand_custom_set(R,ER),
      total_function2(ER,Card,Domain,Range,WF).
    
total_function2(ER,Card,Domain,Range,WF) :-
   var(ER),ground(Card),!, % print(tf_var(Card)),nl,
   tf_var(TotalFunction,[],Card,Domain,Range,WF),  %print(TotalFunction),nl,
   ER=TotalFunction.
total_function2(ER,Card,Domain,Range,WF) :- tf(ER,[],Card,Domain,Range,WF).

%tf(X,SoFar,C,Dom,Ran,WF) :- print_message(tf(X,SoFar,C,Dom,Ran,WF)),fail.
:- block tf(-,?,-,?,?,?).
tf([],_,0,Dom,_,_WF) :- empty_set(Dom).
tf(FUN,SoFar,s(Card),Dom,Ran,WF) :- var(FUN),nonvar(Dom), % try setting up skeleton for total fun
   remove_exact_first_element(X,Dom,Dom2),not_element_of_wf(X,SoFar,WF),var(FUN),!,
   FUN = [(X,Y)|T], tf1(X,Y,T,SoFar,Card,Dom2,Ran,WF).
tf([(X,Y)|T],SoFar,s(Card),Dom,Ran,WF) :- 
    not_element_of_wf(X,SoFar,WF), %print(not_el(X)),nl,
    remove_element_wf(X,Dom,Dom2,WF), %mal: 17/3/08 changed to _wf version
    tf1(X,Y,T,SoFar,Card,Dom2,Ran,WF).
tf1(X,Y,T,SoFar,Card,Dom2,Ran,WF) :-
    check_element_of_wf(Y,Ran,WF), % print_message(chk(Y,Ran)),
    %when((nonvar(T);nonvar(Card)), /* mal 12/5/04: changed , into ; */
    add_new_element(X,SoFar,SoFar2),  %%% try_expand_and_convert_to_avl
    tf(T,SoFar2,Card,Dom2,Ran,WF).
       
:- block tf_var(-,?,-,?,?,?).
tf_var([],_,0,Dom,_,_WF) :- empty_set(Dom). %, print_bt_message(finished_tf_var).
tf_var([(X,Y)|T],SoFar,s(Card),Dom,Ran,WF) :-  %print_bt_message(tf_var(X,Y,SoFar,Card)),
    /* supposes that X + Y are unbound */
    /* TO DO: rewrite like enumerate <-------------------------- */
    ((var(X),var(Y)) -> true ; (print_message(warning,'Nonvar in tf_var: '),
                                print_message(warning,((X,Y))))),
    remove_exact_first_element(X,Dom,Dom2),
    not_element_of_wf(X,SoFar,WF),
 %  print_message(check_element_of_wf(Y,Ran,WF)),
    check_element_of_wf(Y,Ran,WF),
   % print_message(checked(Y,Ran)),
    add_new_element(X,SoFar,SoFar2),
    tf_var(T,SoFar2,Card,Dom2,Ran,WF).
    
   
:- assert_must_succeed((bsets_clp:enumerate_total_bijection([int(1)],[int(2),int(3)],R),
                        kernel_objects:equal_object(R, []) )).
:- assert_must_succeed((bsets_clp:enumerate_total_bijection([int(1)],[int(3)],R),
                        kernel_objects:equal_object(R, [[(int(1),int(3))]]) )).
                         
enumerate_total_bijection(Dom,Range,AllTotalFunctions) :-
   debug_print(9,enum_total_bijection(Dom,Range)),debug_nl(9),
   OutputVars = TF, cardinality(Dom,Card),
   delay_findall(TF,bsets_clp:enum_tbij(TF,Card,Dom,Range),AllTotalFunctions,OutputVars).  
enum_tbij([],0,Dom,Ran) :- empty_set(Dom),empty_set(Ran).
enum_tbij([(X,Y)|T],s(Card),Dom,Ran) :- 
    /* supposes that X + Y are unbound */
    ((var(X),var(Y)) -> true ; (print_message(warning,'Nonvar in enum_tf: '),
                                print_message(warning,((X,Y))))),
    remove_exact_first_element(X,Dom,Dom2),
    remove_exact_element(Y,Ran,Ran2),
    enum_tbij(T,Card,Dom2,Ran2).
    
:- assert_must_succeed((bsets_clp:total_bijection(X,[int(1)],[int(7)]),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:total_bijection(X,[int(1),int(2)],[int(7),int(8)]),
                     kernel_objects:equal_object(X,[(int(2),int(8)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1)],[int(7),int(3)]),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1),int(2)],[int(3)]),
                     X = [(int(1),int(3)),(int(2),int(3))])).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1),int(2)],[int(7),int(8)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1),int(2)],[int(7),int(8)]),
                     X = [(int(1),int(7)),(int(1),int(8))])).


                     
total_bijection(R,Domain,Range) :- init_wait_flags(WF),
  total_bijection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

:- block total_bijection_wf(?,-,?,?).
total_bijection_wf(FF,Domain,Range,_WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),!,
    equal_object(FFDomain,Domain), equal_object(FFRange,Range).
total_bijection_wf(R,Domain,Range,WF) :-   % print_term_summary(total_bijection_wf(R,Domain,Range)),
    cardinality(Domain,Card),
    %print(card_domain(Domain,Card)),nl,
    cardinality(Range,Card),
    %print(card_range(Range,Card)),nl, 
    %when(ground(Range),(print(ground_range(Range,Card,Domain)),nl)),
    %when(nonvar(Card), (print(nonvar_card(Card,Domain,Range)),nl)),
    %when(ground(Card), (print(ground_card(Card,Domain,Range)),nl)),
    total_bijection_wf2(R,Domain,Card,Range,WF).

:- block total_bijection_wf2(?,?,-,?,?).
total_bijection_wf2(FF,Domain,_,Range,_WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),!,
    equal_object(FFDomain,Domain), equal_object(FFRange,Range).
total_bijection_wf2(R,Domain,Card,Range,WF) :-
    expand_custom_set_to_list(R,ER,_,total_bijection_wf2),
    total_bij(ER,Card,Domain,Range,WF).

total_bij(BIJ,Card,Domain,Range,WF) :-   % print(total_bij(BIJ,Card,Domain,Range,WF)),nl, %
   ((ground(Card),var(BIJ)) -> setup_domain(BIJ,Card,Domain) /* set up domain to avoid repeated
     versions of the same relation */ ; true),
   tbij(BIJ,[],Card,Domain,Range,WF).

setup_domain([],0,Dom) :- empty_set(Dom).
setup_domain([(X,_)|T],s(Card),Dom) :-
    remove_exact_first_element(X,Dom,Dom2),
    setup_domain(T,Card,Dom2).
    
:- block tbij(-,?,?,?,?,?), tbij(?,?,-,?,?,?).
tbij([],_,0,Dom,Ran,_WF) :- empty_set(Dom),empty_set(Ran).
tbij([(X,Y)|T],SoFar,s(Card),Dom,Ran,WF) :-   % prints(tbij(X,Y,T,s(Card))),
    %not_element_of(X,SoFar),  necessary ??
    remove_element_wf(X,Dom,Dom2,WF),   % instead of setup_domain, we could remove first element if X,Y are free
    remove_element_wf(Y,Ran,Ran2,WF),
    add_new_element(Y,SoFar,SoFar2), 
    not_element_of_wf(Y,SoFar,WF),
    %prints(chooseY(Y,forx(X),Ran,Ran2,sofar(SoFar))),  portray_waitflags(WF),nl,
    tbij(T,SoFar2,Card,Dom2,Ran2,WF).

 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(8),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(1),int(7))],[int(1)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_total_function(X,[int(1)],[int(7)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_total_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_function([],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function([],[global_set('NAT1')],[global_set('Name')],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function([(int(7),int(7))],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).

:- block not_total_function(-,?,?,?), not_total_function(?,-,?,?).
not_total_function(FF,Domain,Range,_WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    not_equal_object(FFDomain,Domain).
not_total_function(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    %print_term_summary(check_special_not_tf(FF,FFDomain,FFRange)),nl,
    kernel_equality:equality_objects(FFDomain,Domain,Result), % not yet implemented ! % TODO ! -> sub_set,equal,super_set
    when(nonvar(Result),(Result=pred_false -> true ; not_subset_of_wf(FFRange,Range,WF))). 
not_total_function(R,Domain,Range,WF) :-
   % print_term_summary(not_total_function(R,Domain,Range,WF)),
     try_expand_and_convert_to_avl_with_check(R,ER,not_total_function_range),
     try_expand_and_convert_to_avl_with_check(Domain,EDomain,not_total_function_domain),  /* avoid multiple expansions */
     % print_term_summary(exp_domain(EDomain)),
     try_expand_and_convert_to_avl_unless_large(Range,ERange), 
     % print_term_summary(exp_range(ERange)),
    not_total_function2(ER,EDomain,ERange,WF).

% repeat block, in case Domain or R is a closure
:- block not_total_function2(-,?,?,?), not_total_function2(?,-,?,?).
not_total_function2(R,Domain,Range,_WF) :-    %print_term_summary(not_tot2(R,Domain,Range)),
   expand_and_convert_to_avl_set(R,AER), 
   expand_and_convert_to_avl_set(Domain,ADom),!,
    % print_term_summary(check_is_not_avl_total_function(avl_set(AER),Domain,Range)),
   (is_avl_total_function_over_domain(AER,ADom)
     ->  % print(is_tf),nl,
        is_not_avl_relation_over_range(AER,Range)  % TO DO: check only Range !
     ;  true %,print(not_pf),nl
   ).
not_total_function2(R,EDomain,ERange,WF) :-    
    expand_custom_set_to_list(R,ER,_,not_total_function2),
    not_tf(ER,[],EDomain,ERange,WF).
     
:- block not_tf(-,?,?,?,?).
not_tf([],_,Domain,_,WF) :- not_empty_set_wf(Domain,WF).
not_tf([(X,Y)|T],SoFar,Dom,Ran,WF) :- membership_test(SoFar,X,MemRes),
   not_tf2(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- block not_tf2(-,?,?,?, ?,?,?,?). %, not_tf2(?,?,?,?, -,?,?), not_tf2(?,?,?,?, ?,-,?).
not_tf2(pred_true,_X,_,_T,_SoFar,_Dom,_Ran,_WF).% :- check_element_of_lazy(X,SoFar,WF).
not_tf2(pred_false,X,Y,T,SoFar,Dom,Ran,WF) :-
    %not_element_of_wf(X,SoFar,WF),
    membership_test(Dom,X,MemRes),
    not_tf3(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- block not_tf3(-, ?,?,?,?, ?,?,?).
not_tf3(pred_false,_X,_Y,_T,_SoFar,_Dom,_Ran,_WF).
not_tf3(pred_true,X,Y,T,SoFar,Dom,Ran,WF) :-
    remove_element_wf(X,Dom,Dom2,WF),
     membership_test(Ran,Y,MemRes),
     not_tf4(MemRes,X,Y,T,SoFar,Dom2,Ran,WF).
 
:- block not_tf4(-, ?,?,?,?, ?,?,?).
not_tf4(pred_false,_X,_Y,_T,_SoFar,_Dom2,_Ran,_WF).
not_tf4(pred_true,X,Y,T,SoFar,Dom2,Ran,WF) :-
       check_element_of_wf(Y,Ran,WF), %was check_element_of_lazy(Y,Ran,WF)
       add_new_element(X,SoFar,SoFar2),
       not_tf(T,SoFar2,Dom2,Ran,WF).
    
    
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_bijection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_bijection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(8),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_bijection([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_total_bijection(X,[int(1)],[int(7)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_fail((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(1),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(3)],_WF),
                     X = [(int(1),int(3)),(int(2),int(3))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1)],[int(7),int(8)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7)],_WF),
                     X = [(int(2),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_bijection([],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_bijection([(int(7),int(7))],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_bijection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).

:- block not_total_bijection(-,?,?,?).
not_total_bijection(FF,Domain,Range,_WF) :-
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),!,
    %% print(not_total_bijection_wf_closure(FFDomain,FFRange)),nl, %%
    not_equal_object((FFDomain,FFRange),(Domain,Range)).
not_total_bijection(R,Domain,Range,WF) :-
    try_expand_custom_set(R,ER),not_tot_bij(ER,[],Domain,Range,WF).
     
:- block not_tot_bij(-,?,?,?,?).
not_tot_bij([],_,Domain,Range,_WF) :- empty_not_tot_bij(Domain,Range).
not_tot_bij([(X,Y)|T],SoFar,Dom,Ran,WF) :- membership_test(SoFar,X,MemRes),
    not_tot_bij2(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- use_module(kernel_equality,[empty_set_test/2]).
:- block empty_not_tot_bij(-,?).
empty_not_tot_bij(Domain,Range) :- 
      empty_set_test(Domain,EqRes), empty_not_tot_bij2(EqRes,Range).
:- block empty_not_tot_bij2(-,?).
empty_not_tot_bij2(pred_false,_).
empty_not_tot_bij2(pred_true,Range) :-  not_empty_set(Range).

:- block not_tot_bij2(-,?,?,?,?,?,?,?). 
not_tot_bij2(pred_true,_X,_,_T,_SoFar,_Dom,_Ran,_WF).
not_tot_bij2(pred_false,X,Y,T,SoFar,Dom,Ran,WF) :-
    membership_test(Dom,X,MemRes),
    not_tot_bij3(MemRes,X,Y,T,SoFar,Dom,Ran,WF).
 
:- block not_tot_bij3(-,?,?,?,?,?,?,?). 
not_tot_bij3(pred_false,_X,_,_T,_SoFar,_Dom,_Ran,_WF). %  X not a member of domain
not_tot_bij3(pred_true,X,Y,T,SoFar,Dom,Ran,WF) :-
    remove_element_wf(X,Dom,Dom2,WF),
    membership_test(Ran,Y,MemRes),
    not_tot_bij4(MemRes,X,Y,T,SoFar,Dom2,Ran,WF).
 
:- block not_tot_bij4(-,?,?,?,?,?,?,?). 
not_tot_bij4(pred_false,_X,_,_T,_SoFar,_Dom2,_Ran,_WF). %  Y not a member of range
not_tot_bij4(pred_true,X,Y,T,SoFar,Dom2,Ran,WF) :-
        remove_element_wf(Y,Ran,Ran2,WF),
        add_element(X,SoFar,SoFar2),
        not_tot_bij(T,SoFar2,Dom2,Ran2,WF).
    
  
  
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:range_restriction([(int(1),int(2)),(int(2),int(3))],[int(3)],[(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:range_restriction([(int(1),int(2)),(int(2),int(3))],[int(2),int(3)],[(int(1),int(2)),(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:range_restriction([],[int(2),int(3)],[]))).
:- assert_must_succeed((bsets_clp:range_restriction([],[int(1)],[]))).
:- assert_must_succeed((bsets_clp:range_restriction([],[],[]))).
:- assert_must_succeed((bsets_clp:range_restriction([(int(1),int(2))],[int(1)],[]))).
:- assert_must_succeed((bsets_clp:range_restriction([(int(1),int(2))],[int(2)],[(int(1),int(2))]))).
:- assert_must_succeed((bsets_clp:range_restriction(X,[fd(3,'Name')],R),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],
                        kernel_objects:equal_object(X,R))).
:- assert_must_succeed((bsets_clp:range_restriction(X,Y,R),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],Y=global_set('Name'),
                        kernel_objects:equal_object(X,R))).
:- assert_must_fail((bsets_clp:range_restriction(X,[fd(3,'Name')],R),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(1,'Name'))],
                        kernel_objects:equal_object(X,R))).

:- block range_restriction(-,?,?),range_restriction(?,-,-).

range_restriction(R,S,Res) :-  /* R <| S */ 
    ok_to_try_restriction_explicit_set(S,R,Res),
    range_restriction_explicit_set(R,S,SR),!,
    equal_object(SR,Res).
range_restriction(R,S,Res) :-  /* R |> S */
    expand_custom_set_to_list(R,ER,_,range_restriction),
    relation_restriction(ER,S,Res,pred_true,range).

% heuristic: should we try restriction_explicit_set or
% is relation_restriction with its stronger constraint propagation better
ok_to_try_restriction_explicit_set(S,R,Res) :-
   nonvar(S),
   (var(Res) -> true
    ; S=avl_set(_),
      nonvar(R), R=avl_set(_)  % otherwise constraint propagation from normal relation_restriction better
   ).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:range_subtraction([],[int(2)],[]))).  
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:range_subtraction([(int(1),int(2)),(int(2),int(3))],[int(2)],[(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:range_subtraction([(int(1),int(2)),(int(2),int(3))],[],[(int(1),int(2)),(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:range_subtraction([(int(1),int(2)),(int(2),int(3))],[int(1)],[(int(1),int(2)),(int(2),int(3))]))).

:- block range_subtraction(-,?,?),range_subtraction(?,-,-).
range_subtraction(R,S,Res) :-  /* R |>> S */
    S==[],!,
    equal_object(R,Res).
range_subtraction(R,S,Res) :-  /* R <| S */ 
    %print(range_subtraction(R,S,Res)),nl,trace,
    ok_to_try_restriction_explicit_set(S,R,Res),
    range_subtraction_explicit_set(R,S,SR),!,
    equal_object(SR,Res).
range_subtraction(R,S,Res) :-  /* R |>> S */
    expand_custom_set_to_list(R,ER,_,range_subtraction),
    relation_restriction(ER,S,Res,pred_false,range).


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:domain_restriction([int(2),int(4)],[(int(1),int(4)),(int(2),int(3))],[(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:domain_restriction([int(1),int(2)],[(int(1),int(2)),(int(2),int(3))],[(int(1),int(2)),(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:domain_restriction([int(2),int(3)],[],[]))).
:- assert_must_succeed((bsets_clp:domain_restriction([int(1)],[],[]))).
:- assert_must_succeed((bsets_clp:domain_restriction([int(1)],[],R), R==[])).
:- assert_must_fail((bsets_clp:domain_restriction(_,[],R), R=[int(_)|_])).
:- assert_must_succeed((bsets_clp:domain_restriction([int(2)],[(int(1),int(2))],[]))).
:- assert_must_succeed((bsets_clp:domain_restriction([],[(int(1),int(2))],[]))).
:- assert_must_succeed((bsets_clp:domain_restriction([int(1)],[(int(1),int(2))],[(int(1),int(2))]))).
:- assert_must_succeed((bsets_clp:domain_restriction([int(1)],[(int(1),int(2)),(int(2),_)],_))).
:- assert_must_succeed((bsets_clp:domain_restriction([int(2),int(1)],X,R),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],
                        kernel_objects:equal_object(X,R))).
                        
:- block domain_restriction(?,-,?),domain_restriction(-,?,-).
domain_restriction(S,R,Res) :-  /* S <| R */ 
    %print_term_summary(domain_restriction(S,R,Res)),
    ok_to_try_restriction_explicit_set(S,R,Res),
    domain_restriction_explicit_set(S,R,SR),!,
     %  print_term_summary(domain_restriction_explicit_set(S,R,SR)), %%
    equal_object(SR,Res).
domain_restriction(S,R,Res) :-  /* S <| R */ 
    % print_term_summary(call_domain_restriction(S,R,Res)), translate:print_bvalue(S),nl,
    expand_custom_set_to_list(R,ER,_,domain_restriction),
    relation_restriction(ER,S,Res,pred_true,domain).

% a predicate to compute domain/range restriction/subtraction
:- block relation_restriction(?,-,- ,?,?),
         relation_restriction(-,?,? ,?,?).
relation_restriction([],_S,Res,_AddWhen,_DomOrRange) :-
    %print_term_summary(domain_restriction_finished([],_S,Res)),
    empty_set(Res).
relation_restriction([(X,Y)|T],S,Res,AddWhen,DomOrRange) :- 
    (DomOrRange=domain
     -> membership_test(S,X,MemRes)
      ; membership_test(S,Y,MemRes)),
    (nonvar(MemRes) 
     %MemRes==AddWhen  % MemRes already set; we will ensure that (X,Y) in Res below; this slows down Alstom Compilation Regle !
      % doing the membership_test on the result Res if MemRes\==AddWhen only makes sense if we cannot fully compute the restriction ?? i.e. if T is not a closed list ?
     -> true  %,(MemRes==AddWhen -> true ; print_term_summary(relation_restriction([(X,Y)|T],S,Res,AddWhen,DomOrRange)),nl)
     ;   (AddWhen=pred_true -> InResult=MemRes
           ; bool_pred:negate(InResult,MemRes)),
          membership_test(Res,(X,Y),InResult)
                    % TO DO: same for explicit version; gets called e.g. if S = 1..n  (1..n <| [1,2,3] = [1,2])
                    % can now solve e.g. {x|x <| [1,2,3] = [1,2] & card(x)=2} = {{1,2}}
                    % or  x <| s = [1,2,3] \/ {29|->29}  & x <: 1..100 & s = %i.(i:1..50|i)
     ),
    % print(restrict(X,Y,MemRes)),nl,
    relation_restriction_aux(MemRes,X,Y,T,S,Res,AddWhen,DomOrRange).
:- block relation_restriction_aux(-,?,?,?,?,?, ?,?).
relation_restriction_aux(MemRes,X,Y,T,S,Res,AddWhen,DomOrRange) :- 
     MemRes==AddWhen,!, % (X,Y) should be added to result  %print(add(X,Y)),nl,
     equal_cons(Res,(X,Y),RT), % was : equal_object([(X,Y)|RT],Res),
     relation_restriction(T,S,RT,AddWhen,DomOrRange).  
relation_restriction_aux(_MemRes,_X,_,T,S,RT,AddWhen,DomOrRange) :-
     % the couple is filtered out   %print(filter(_X)),nl,
     relation_restriction(T,S,RT,AddWhen,DomOrRange).   


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:domain_subtraction([int(1),int(3)],[(int(1),int(4)),(int(2),int(3))],[(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:domain_subtraction([int(3),int(4)],[(int(1),int(2)),(int(2),int(3))],[(int(1),int(2)),(int(2),int(3))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:domain_subtraction([int(1)],[],[]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:domain_subtraction([],[(int(11),int(21))],[(int(11),int(21))]))).
:- assert_must_succeed((bsets_clp:domain_subtraction([int(1)],[(int(1),int(2))],[]))).
:- assert_must_succeed((bsets_clp:domain_subtraction([int(3)],[(int(1),int(2))],[(int(1),int(2))]))).
:- assert_must_succeed((bsets_clp:domain_subtraction([int(1)],[(int(1),int(2)),(int(2),int(X))],R),
                         R=[(int(2),int(YY))], YY==X)).
:- assert_must_succeed((bsets_clp:domain_subtraction([int(5),int(3)],X,R),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],
                        kernel_objects:equal_object(X,R))).
:- block domain_subtraction(?,-,?),domain_subtraction(-,?,-).
%domain_subtraction(S,R,Res) :-  /* S <<| R */
%    is_custom_explicit_set(R), nonvar(S),
%    fail.
domain_subtraction(S,R,Res) :- S==[],!,
    equal_object(R,Res).
domain_subtraction(S,R,Res) :-  /* S <<| R */ 
    ok_to_try_restriction_explicit_set(S,R,Res),
    domain_subtraction_explicit_set(S,R,SR),!,
    equal_object(SR,Res).
domain_subtraction(S,R,Res) :-  /* S <<| R */
    %print_message(dom_sub(S,R,Res)),
    expand_custom_set_to_list(R,ER,_,domain_subtraction),
    try_expand_and_convert_to_avl_with_check(S,AS,domain_subtraction),
   % print_term_summary(dom_sub2(ER,AS,Res)),
   % (ground(ER) -> domain_subtraction_acc(ER,AS,[],Res) ; 
                 relation_restriction(ER,AS,Res,pred_false,domain)
  %               )
                 .
 
% version which uses accumulator; does not instantiate directly but can use AVL-tree    
/*
domain_subtraction_acc([],_,Acc,Res) :- %print_term_summary(domain_subtraction_acc_res(Acc)), 
     equal_object(Acc,Res).
domain_subtraction_acc([(X,Y)|T],S,Acc,Res) :- membership_test(S,X,MemRes),
     domain_subtraction_acc2(MemRes,X,Y,T,S,Acc,Res).

:- block domain_subtraction_acc2(-,?,?,?,?,?,?).  
domain_subtraction_acc2(pred_false,X,Y,T,S,Acc,Res) :- %not_element_of(X,S), 
      add_element((X,Y),Acc,NewAcc),
      domain_subtraction_acc(T,S,NewAcc,Res).
domain_subtraction_acc2(pred_true,_X,_Y,T,S,Acc,Res) :- %check_element_of(X,S), 
     domain_subtraction_acc(T,S,Acc,Res).
     */
 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override_relation([(int(1),int(2))],[(int(1),int(3))],[(int(1),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override_relation([(int(1),int(2))],[(int(2),int(3))],[(int(1),int(2)),(int(2),int(3))],WF),WF)).
:- assert_must_succeed((bsets_clp:override_relation([(int(1),int(2)),(int(2),int(4))],[(int(1),int(3))],X,_WF),
                        kernel_objects:equal_object(X,[(int(2),int(4)),(int(1),int(3))]))). 
:- assert_must_succeed((bsets_clp:override_relation([(int(1),int(2)),(int(2),int(4))],[(int(3),int(6))],X,_WF),
                        kernel_objects:equal_object(X,[(int(2),int(4)),(int(1),int(2)),(int(3),int(6))]))). 

:- block override_relation(-,-,?,?).
override_relation(R,S,Res,_WF) :- R==[],!, equal_object(S,Res).
override_relation(R,S,Res,_WF) :- S==[],!, equal_object(R,Res).
override_relation(R,S,Res,_WF) :- Res==[],!, empty_set(S), empty_set(R).
override_relation(R,S,Res,_WF) :-  /* R <+ S */
   % print_term_summary(override_relation(R,S,Res,_WF)),nl,
   override_custom_explicit_set(R,S,ORes),!,
   % print_term_summary(override_custom_explicit_set(R,S,ORes)),
   equal_object(ORes,Res).
override_relation(R,S,Res,WF) :-  /* R <+ S */
  % print('override_relation: '), translate:print_bvalue(S),nl,
  /* TO DO: first check if S is empty */
   % print(override__enter_domain(S,DS)),nl,
    domain_wf(S,DS,WF),
 %   when(ground(DS), (print(override_domain(DS)),nl)),
  % print_term_summary(override__exit__domain(S,DS)),nl,
 %   print_term_summary(override__enter_domain_subtraction(DS,R,DSR)),
    domain_subtraction(DS,R,DSR),
   % print(override__exit_domain_subtraction(DS,R,DSR)),nl,
  %  when(ground(DSR), (print(override_domain_subtraction(DSR)),nl)),
   % print(calling_union(DSR,S,Res)),nl,
    union_wf(DSR,S,Res,WF). % ,print(done_union(DSR,S,Res)),nl,trace.
  %  , print_term_summary(override__exit_union(DSR,S,Res)),nl.
 %  when(ground(Res), (print(override_UNION_RESULT(R,S,Res)),nl)).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override([],int(1),int(3),[(int(1),int(3))],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override([(int(1),int(2)),(int(2),int(6))],int(1),int(3),[(int(1),int(3)),(int(2),int(6))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override([(int(1),int(2)),(int(2),int(6))],int(2),int(3),[(int(1),int(2)),(int(2),int(3))],WF),WF)). 
:- block override(-,?,?,?,?), override(?,-,?,?,?),
         override(?,?,-,?,?). % also wait on Y; try to generate avl if possible; can only be used in substitution anyway
/*  R <+ {X |-> Y} as used by substitution R(X) := Y */
override(R,X,Y,Res,_WF) :- % print_term_summary(override(R,X,Y,Res)),
   override_explicit_set(R,X,Y,ORes),!,
   %% print_term_summary(override_explicit_set(R,X,Y,ORes)), %%
   equal_object(ORes,Res).
override(R,X,Y,Res,WF) :-  
     expand_custom_set_to_list(R,ER,_,override), %print_term_summary(override2(ER,X,Y)),
     override2(ER,X,Y,[(X,Y)],ORes,WF),
     equal_object(ORes,Res).

:- block override2(-,?,?,?,?,?).
override2([],_X,_Y,Remainder,Res,_WF) :- Remainder=Res. %equal_object(Remainder,Res).
override2([(V,W)|T],X,Y,Remainder,Res,WF) :- equality_objects(V,X,EqRes),
   override2c(EqRes,V,W,T,X,Y,Remainder,Res,WF).

:- block override2c(-, ?,?,?, ?,?,?,?,?).
override2c(pred_true,_V,_W,T,X,Y,_Remainder,Res,WF) :- 
    equal_cons_wf(Res,(X,Y),T2,WF), %equal_object([(X,Y)|T2],Res),
    override2(T,X,Y,[],T2,WF).  /* set remainder to [], we have already added (X,Y) */
override2c(pred_false,V,W,T,X,Y,Remainder,Res,WF) :- 
    equal_cons_wf(Res,(V,W),T2,WF), %equal_object([(V,W)|T2],Res),
    override2(T,X,Y,Remainder,T2,WF).

    
    
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2)),(int(3),int(3))],[int(1),int(2)],[int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2)),(int(1),int(3)),(int(4),int(4))],[int(1),int(2)],[int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2)),(int(1),int(3)),(int(4),int(4))],[int(2)],[int(2),int(3)],WF),WF)).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[int(2)],_WF)).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2))],[int(2)],[],_WF)).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2))],[int(3)],[],_WF)).
:- assert_must_succeed((bsets_clp:image_wf([(int(1),int(2)),(int(1),int(3))],
                                  [int(X)],R,_WF), X=1, kernel_objects:equal_object(R,[int(2),int(3)]))).
:- assert_must_succeed((bsets_clp:image_wf([([int(1),int(2)],int(6)),
                                   ([int(1),int(2),int(3)],int(7)),
                                   ([int(2),int(1)],int(8))],
                                  [[int(X),int(1)]],R,_WF), X=2, 
				  kernel_objects:equal_object(R,[int(6),int(8)]))).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2))],[int(1),int(2)],[int(2)],_WF)).
:- assert_must_fail(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[int(1)],_WF)).
:- assert_must_fail(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[],_WF)).


:- block image_wf(-,?,?,?).
image_wf(Rel,_,Res,_WF) :- Rel==[],!,empty_set(Res).
image_wf(Rel,S,Res,_WF) :- image_for_id_closure(Rel,S,Img),!, % we don't require S to be known here
  % print(image_for_id(Rel,S,Img)),nl,
  equal_object(Img,Res). %, print_term_summary(image_for_id(Rel,S,Res)).
image_wf(Rel,S,Res,WF) :-   %print_term_summary(image_wf(Rel,S,Res)),
  image_wf0(Rel,S,Res,WF).

:- block image_wf0(?,-,?,?).
image_wf0(Rel,S,Res,WF) :-  /*  Res = Rel[S] */
     %  nl,print_term_summary(image_wf0(Rel,S,Res)),nl,
    % print(image(Rel,S,Res,WF)),nl,
    (S==[] -> empty_set(Res)
            ; image1(Rel,S,Res,WF)  ) . %,print(done_image(Rel,S,Res)),nl. %,watch(Rel,arg1_of_image,0).

keep_symbolic(R) :- var(R),!,fail.
keep_symbolic(closure(_,_,_)) :- preferences:get_preference(convert_comprehension_sets_into_closures,true),!.
keep_symbolic(R) :- dont_expand_this_explicit_set(R).

:- block image1(-,?,?,?).
image1(Rel,S,Res,WF) :-   %print_term_summary(image1(Rel,S,Res)),
  image_for_explicit_set(Rel,S,Img,WF),!,  % print(image_explicit(Img)),nl,
  equal_object(Img,Res).
%image1(Rel,S,Res,WF) :- expand_custom_set_to_list(S,ES),!, image_of_set(ES,Rel,Res,WF).
image1(Rel,Set,Res,WF) :- %print(rel(Rel,Set)),nl,
   keep_symbolic(Rel), %print(don_expand_this),nl,
   ( dom_for_specific_closure(Rel,Domain,function(_))
     % print(image_for_inf_fun(Rel,Set)),nl,
     -> !, expand_custom_set_to_list(Set,ESet,_,image1),
        image_for_inf_fun(ESet,Domain,Rel,Res,WF)
     ;  get_closure_types(Rel,DomType,RangeType),!,
        % Rel should not be expanded, compute closure by calculating {yy|#(xx).(xx:Set & xx|->yy:Rel)}
        %print(image_for_large_relation(DomType,RangeType,Set,Rel)),nl,
        image_closure(Set,Rel,DomType,RangeType,Closure ),
        expand_custom_set(Closure,CRes), equal_object(CRes,Res)
   ).
image1(Rel,S,Res,WF) :- expand_custom_set_to_list(Rel,Relation,_,image1_2), % bad if Rel is a big closure !
          % print(image2(Relation,S,Res)),nl,
 %        image_for_list_relation(Relation,S,Res).
         image_for_list_relation(Relation,S,[],Res,WF). %,print_term_summary(image2_res(Relation,S,Res)).

:- use_module(btypechecker,[couplise_list/2]).
get_closure_types(closure(_,Types,_),Domain,Range) :-
   couplise_list(Types,couple(Domain,Range)).
%get_closure_types(closure([_,_,_,_],[A,B,C,D],_), couple(couple(A,B),C),D).
 
% construct a closure for {yy|#(xx).(xx:Set & xx|->yy:Rel)}
image_closure(Set,Rel,DomType,RangeType,Closure ) :-
  Closure = closure([yy],[RangeType],Body),
  Body =  b(exists([b(identifier(xx),DomType,[])],
   b(conjunct(
    b(member(b(identifier(xx),DomType,[]),b(value(Set),set(DomType),[])),pred,[]),
  b(member(b(couple(b(identifier(xx),DomType,[]),
  b(identifier(yy),RangeType,[])),couple(DomType,RangeType),[]),
  b(value(Rel),set(couple(DomType,RangeType)),[])),pred,[])),pred,[])),pred,[used_ids([yy])]).



% very similar to rel_compose_inf, indeed f[S] = ran((id(S);f)) 
   :- block image_for_inf_fun(-,?,?,?,?). 
image_for_inf_fun([],_Dom,_Rel2,Comp,_WF) :- empty_set(Comp).
image_for_inf_fun([X|T],Dom,Fun,CompRes,WF) :-
   membership_test(Dom,X,MemRes),
   image_for_inf_fun_aux(MemRes,X,T,Dom,Fun,CompRes,WF).

:- block image_for_inf_fun_aux(-,?,?, ?,?,?,?).
image_for_inf_fun_aux(pred_true,X,T,Dom,Fun,CompRes,WF) :-
   % print(image_for_inf_fun_aux(X)),nl,
   apply_to(Fun,X,FX,WF), % TO DO: generalize to image so that we can apply it also to infinite relations ?
   add_element(FX,CT,CompRes),
   image_for_inf_fun(T,Dom,Fun,CT,WF).
image_for_inf_fun_aux(pred_false,_X,T,Dom,Fun,Comp,WF) :-
    % print(image_for_inf_fun_aux_not_in_domain(_X)),nl,
    image_for_inf_fun(T,Dom,Fun,Comp,WF).


/*
:- block image_of_set(-,?,?,?,?), image_of_set(?,?,-,?,?). 
image_of_set([],Rel,ImageSoFar,Res,WF) :- equal_object(ImageSoFar,Res).
image_of_set([H|T],Rel,ImageSoFar,Res,WF) :-
   image_of_element(Rel,H,ImageSoFar,SF2,WF),
   image_of_set(T,Rel,SF2,Res,WF).
   
image_of_element([],_,Acc,Res,WF) :- equal_object(Acc,Res).
image_of_element([(A,B)|T],H,Acc,Res,WF) :- equality....
image_of_element(avl_set(),H,Acc,Res,WF) :- ....
image_of_element(closure(),....
*/
   
% Computing the image of a relation which is stored as a list: traverse the relation
:- block image_for_list_relation(-,?,?,?,?). 
image_for_list_relation([],_,_,Res,_WF) :- empty_set(Res).
image_for_list_relation([(X,Y)|T],S,ImageSoFar,Res,WF) :-  % prints(image(X,Y,T,ImageSoFar)),trace,
   ((T==[], definitely_not_empty(Res))
    -> MemRes=pred_true, % we need at least one more element for Res
       check_element_of_wf(X,S,WF)
    ;  (Res==[],ImageSoFar==[]) -> MemRes=pred_false, not_element_of_wf(X,S,WF) % Result empty: X cannot be in S
    ;  membership_test(S,X,MemRes)
   ),
   image4(MemRes,Y,T,S,ImageSoFar,Res,WF).

definitely_not_empty(Set) :- nonvar(Set), Set \== [], Set \= closure(_,_,_).

:- block image4(-, ?,?,?, ?,?,?).
image4(pred_true, Y,T,S, ImageSoFar,Res,WF) :- 
  (Res==[]
    -> MemRes=pred_true, check_element_of_wf(Y,ImageSoFar,WF)
    ;  membership_test(ImageSoFar,Y,MemRes)
  ),
  image5(MemRes,Y,T,S,ImageSoFar,Res,WF).
image4(pred_false, _Y,T,S, ImageSoFar,Res,WF) :- 
  image_for_list_relation(T,S,ImageSoFar,Res,WF).
  
:- block image5(-, ?,?,? ,?,?,?).
image5(pred_true,_Y,T,S,ImageSoFar,Res,WF) :- /* we have already added Y to the image */
  image_for_list_relation(T,S,ImageSoFar,Res,WF).
image5(pred_false,Y,T,S,ImageSoFar,Res,WF) :- %print(add_to_image_sofar(Y,ImageSoFar)),nl,
  add_element(Y,ImageSoFar,ImageSoFar2), %print(adding_y_to_image_result(Y,Res)),nl,
  equal_cons_wf(Res,Y,Res2,WF), %print(added(Y,Res)),nl,
  image_for_list_relation(T,S,ImageSoFar2,Res2,WF).

% --------------------
% a new experimental version of image; unfortunately it does not yet seem to pay off :-(
% it is not yet called from above
% use_module(bsets_clp), bsets_clp:image_for_list_relation([(int(1),int(2))],[int(1)],Result)
% use_module(bsets_clp), bsets_clp:image_for_list_relation([(int(1),int(2))|T],[int(1),int(3)],Result)
% bsets_clp:image_for_list_relation([(int(1),int(2))|T],[int(1),int(3)],Result), T=[(int(2),A)], ground(Result), Result==[int(2)]
% bsets_clp:image_for_list_relation([(int(1),int(2))|T],[int(1),int(3)],Result), T=[(int(3),int(55))|T2], nonvar(Result),...
/*
image_for_list_relation(List,S,Result) :-
  image_for_list_relation_propagate(List,S,PQ,PQ,true,Count),
  terminate_propagator_queue(Count,PQ),
  add_image_to_result(PQ,[],Result).

:- block add_image_to_result(-,?,?). 
add_image_to_result(empty,_,Result) :- empty_set(Result).
add_image_to_result(solution(YMEMRES,Y,PQ),ImageSoFar,Result) :- % print(sol(YMEMRES,Y,PQ)),nl, %%
   (YMEMRES=pred_true
      -> membership_test(ImageSoFar,Y,MemRes),add_image_to_result_mem(MemRes,Y,PQ,ImageSoFar,Result)
      ;  add_image_to_result(PQ,ImageSoFar,Result)
   ).
:- block add_image_to_result_mem(-,?,?,?,?). 
add_image_to_result_mem(pred_false,Y,PQ,ImageSoFar,Result) :- % print(add_to_result(Y,ImageSoFar,Result)),nl,
   add_element(Y,ImageSoFar,ImageSoFar2), 
   equal_cons(Result,Y,Res2), % was: equal_object([Y|Res2],Result), TO DO: use _wf version
   add_image_to_result(PQ,ImageSoFar2,Res2).
add_image_to_result_mem(pred_true,_Y,PQ,ImageSoFar,Result) :- % print(skip(_Y)),nl,
  add_image_to_result(PQ,ImageSoFar,Result).

:- block image_for_list_relation_propagate(-,?, ?,?,?,?). 
image_for_list_relation_propagate([],_S, _PQ1,_PQ2,_Done,0).
image_for_list_relation_propagate([(X,Y)|T],S,PQ1,PQ2,PrevDone,s(Count)) :-  
   membership_test(S,X,MemRes),
   (PrevDone==true -> PQ=PQ2 ; PQ=PQ1),
   add_to_propagator_queue(MemRes,Y,PQ,NewPQ,Done),
   image_for_list_relation_propagate(T,S,PQ,NewPQ,Done,Count).

:- block terminate_propagator_queue(-,?).
terminate_propagator_queue(0,PQ) :-
  (PQ=empty -> true ; 
    var(PQ) -> fail ; add_error_fail(terminate_propagator_queue,'Illegal tail (should be var): ',PQ)).
terminate_propagator_queue(s(X),PQ) :- terminate_propagator_queue2(PQ,X).
:- block terminate_propagator_queue2(-,?).
terminate_propagator_queue2(PQ,X) :-
  (PQ=solution(_,_,PQT) -> terminate_propagator_queue(X,PQT)
    ; add_error(terminate_propagator_queue2,'Illegal tail (should be solution/3): ',PQ)).

:- block add_to_propagator_queue(-,?,?,?,?).
add_to_propagator_queue(Result,ID,PropagatorQueue,NewPQ,Done) :-
  add_solution_to_queue(PropagatorQueue,Result,ID,NewPQ,Done).

add_solution_to_queue(PropagatorQueue,Result,ID,NewPQ,Done) :- %% print(add_prop(Result,ID,PropagatorQueue)),nl, %%
  (var(PropagatorQueue) -> PropagatorQueue = solution(Result,ID,NewPQ), Done=true
   ; PropagatorQueue = solution(_,_,Tail) -> add_solution_to_queue(Tail,Result,ID,NewPQ,Done)
   ; add_error_fail(add_solution_to_queue,'Illegal PropagatorQueue: ', PropagatorQueue)
  ). 
  */
% -------------------- END of experimental version for image


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_for_closure1_wf([(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],[int(2)],[int(1),int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_for_closure1_wf([(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],[],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_for_closure1_wf([(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],[int(3)],[int(3)],WF),WF)).
% first experimental version for computing closure1(Rel)[S]
:- block image_for_closure1_wf(-,?,?,?),image_for_closure1_wf(?,-,?,?).
image_for_closure1_wf(Rel,S,Res,_) :- (Rel==[] ; S==[]),!,empty_set(Res).
image_for_closure1_wf(Rel,Set,Res,WF) :- try_expand_and_convert_to_avl_unless_large(Set,ESet),
   image_for_closure1_wf_aux(Rel,ESet,Res,WF).
   
image_for_closure1_wf_aux(Rel,S,Res,WF) :- 
   ((nonvar(S),S=avl_set(_)) 
     -> closure1_for_explicit_set_from(Rel,S,Closure1Rel),!, % if S is known: start from S
         % print_term_summary(closure1_from(Rel,S,Closure1Rel,Res)),nl,
        range_wf(Closure1Rel,Res,WF)
     ; % print_term_summary(clo1(Rel)),nl,
       closure1_for_explicit_set(Rel,Closure1Rel),!, % we can compute it effiently; don't use code below
       % print_term_summary(clo1_res(Closure1Rel,S,Res)),nl,
       image_wf(Closure1Rel,S,Res,WF)
   ).
image_for_closure1_wf_aux(Rel,S,Res,WF) :-
    %when(nonvar(Res),(print(inst(Res,for(S))),nl,trace)),
    % print_term_summary(image_for_closure1_iterate(Rel,S,[],Res,WF)),nl,
    image_for_closure1_iterate(Rel,S,[],Res,WF).

image_for_closure1_iterate(Rel,S,Acc,Res,WF) :-
    image_wf0(Rel,S,Res1,WF),
    when(ground(Res1), image_for_closure1_check_fix(Rel,Acc,Res1,Res,WF)).
image_for_closure1_check_fix(Rel,Acc,Res1,Res,WF) :-
    difference_set(Res1,Acc,New),
    (not_empty_set_wf(New,WF),
     union(New,Acc,Acc1),
     image_for_closure1_iterate(Rel,New,Acc1,Res,WF)
    ;
    empty_set(New),equal_object(Acc,Res)).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:element_of_closure1_wf((int(1),int(3)),[(int(1),int(2)),(int(2),int(1)),(int(2),int(3))],WF),WF)).

% the code below could be further optimized: merge with code above and stop iterating when B appears in New
element_of_closure1_wf((A,B),Relation,WF) :- %print(in_closure1(A,B,Relation)),nl,
  image_for_closure1_wf(Relation,[A],Image,WF), %print(computed_image(Image)),nl,
  check_element_of_wf(B,Image,WF).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_element_of_closure1_wf((int(1),int(3)),[(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],WF),WF)).
not_element_of_closure1_wf((A,B),Relation,WF) :- %print(not_in_closure1(A,B,Relation)),nl,
  image_for_closure1_wf(Relation,[A],Image,WF), %print(computed_image(Image)),nl,
  not_element_of_wf(B,Image,WF).

% -----------------------------------

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:apply_to([(int(2),int(22))],int(2),int(22),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:apply_to([(int(1),int(22)),(int(3),int(33)),(int(4),int(44))],int(3),int(33),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:apply_to([(int(1),[int(22)]),(int(3),[int(32),int(33)]),(int(4),[int(44)])],int(3),[int(32),int(33)],WF),WF)).
:- assert_must_succeed(bsets_clp:apply_to([(int(1),int(2))],int(1),int(2),_WF)).
:- assert_must_succeed((bsets_clp:apply_to(F,int(3),int(2),_WF),F=[(int(3),int(2)),(int(2),int(1))])).
:- assert_must_succeed((bsets_clp:apply_to(F,X,int(1),_WF),F=[(int(3),int(2)),(int(2),int(1))],X=int(2))).
:- assert_must_succeed((bsets_clp:apply_to(F,int(3),_,_WF),F=[(int(3),[int(2),int(3)]),(int(2),[])])).

:- assert_must_fail(bsets_clp:apply_to([(int(1),int(2)),(int(1),int(3))],int(1),int(3),_WF)).
  /* input not a function */
apply_to(R,X,Y,WF) :- apply_to2(R,X,Y,R,unknown,WF).

apply_to(R,X,Y,Span,WF) :-  %prints(apply(X,Y,WF)),
  apply_to2(R,X,Y,R,Span,WF).


:- use_module(preferences,[preference/2]).
:- block apply_to2(-,-,?,?,?,?).
apply_to2(R,X,Y,InitalRel,Span,WF) :-  % prints(apply2(R,X,Y,InitalRel,WF)),
  (var(R) -> apply_to_var(R,X,Y,InitalRel,Span,WF) ;  apply_to_nonvar(R,X,Y,InitalRel,Span,WF)).

apply_to_var(R,X,Y,InitalRel,Span,WF) :-
    when((ground(X);nonvar(R)),  % only instantiate R when X sufficiently instantiated (TO DO: maybe use some for of equality_objects with existing relation R set up so far ??)
        (var(R) -> apply_to_var2(R,X,Y,InitalRel,Span,WF) ; apply_to_nonvar(R,X,Y,InitalRel,Span,WF))).
apply_to_var2(R,X,Y,_InitalRel,_,_WF) :- preferences:preference(find_abort_values,false),!,R=[(X,Y)|_].
apply_to_var2([(X,Y)|_],X,Y,_,_,_).
apply_to_var2(R,X,_Y,InitalRel,Span,WF) :- not_in_domain(X,R),
    when((ground(X),ground(R)),
            add_wd_error_span('function applied outside of domain (#1): ', '@'(X,InitalRel),Span,WF)).

apply_to_nonvar([],X,_Y,InitalRel,Span,WF) :- preferences:preference(find_abort_values,true),  
   %%print(apply_to_empty(X,_Y,InitalRel)),nl,
   when(ground(X),add_wd_error_span('function applied outside of domain (#2): ', '@'(X,InitalRel),Span,WF)).
apply_to_nonvar([(X2,Y2)|T],X,Y,InitalRel,Span,WF) :- 
   equality_objects(X2,X,EqRes),
   apply_to4(EqRes,Y2,T,X,Y,InitalRel,Span,WF).
%apply_to_nonvar(closure(Parameters,PT,Cond),X,Y,WF) :-  !,
%   check_element_of_wf((X,Y),closure(Parameters,PT,Cond),WF).
apply_to_nonvar(avl_set(A),X,Y,_InitalRel,Span,WF) :- 
           %%debug:watch(10,custom_explicit_sets:apply_to_explicit_set(avl_set(A),X,Y,WF)). %%
           apply_to_explicit_set(avl_set(A),X,Y,Span,WF).
apply_to_nonvar(closure(P,T,B),X,Y,_InitalRel,Span,WF) :- 
    %is_custom_explicit_set(Closure,apply),   % should also work for avl_set,...
    (is_recursive_closure(P,T,B)
     -> %print(apply_recursive_closure(X,P,T,B)),
        %profiler:add_profile_hit(rec_apply_closure_to_nonvar(X,Y,P,T,B,Span,WF)),
        when(ground(X),apply_closure_to_nonvar(X,Y,P,T,B,Span,WF))
     ;  %profiler:add_profile_hit(apply_closure_to_nonvar(X,Y,P,T,B,Span,WF)),
         apply_closure_to_nonvar(X,Y,P,T,B,Span,WF)).

% if we first check preferences:preference(find_abort_values,false) to avoid a choice
% point, we get a big slow-down on Alstom models; e.g., vesg_Mar12
apply_closure_to_nonvar(X,Y,P,T,B,_,WF) :- 
    %(preferences:preference(find_abort_values,true) -> true ; !),  % slow down ???!
    check_element_of_function_closure(X,Y,P,T,B,WF) .
apply_closure_to_nonvar(X,_,P,T,B,Span,WF) :- % removing this clause doubles runtime of COMPUTE_GRADIENT_CHANGE
    %profiler:add_profile_hit(clause2(X,P,T,B,Span,WF),2),
    preferences:preference(find_abort_values,true),
    not_in_domain(X,closure(P,T,B)),
    when((ground(X),ground(closure(P,T,B))),
           add_wd_error_span('function applied outside of domain (#3): ', '@'(X,closure(P,T,B)),Span,WF)).
   
:- block apply_to4(-,?,?,?,?,?,?,?).
apply_to4(pred_true,Y2,_T,_X,Y,_InitalRel,_,_WF) :- equal_object(Y2,Y).
apply_to4(pred_false,_Y2,T,X,Y,InitalRel,Span,WF) :- apply_to2(T,X,Y,InitalRel,Span,WF).
  % (nonvar(T) -> apply_to2(T,X,Y,InitalRel,WF)
  %  ; when((ground(X);nonvar(T)), apply_to2(T,X,Y,InitalRel,WF))).




/*******************************/
/*     sigma(Set,Exp,Res)      */
/*  Res = SIGMA(z).(Set | Exp) */
/*******************************/

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:sum_over_range([(int(1),int(2))],int(2)))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:sum_over_range([],int(0)))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:sum_over_range([(int(1),int(2)),(int(1),int(3))],int(5)))).

:- block sum_over_range(-,?).
sum_over_range([],R) :- !, R=int(0).
sum_over_range(CS,R) :- sum_of_range_custom_explicit_set(CS,Sum),!, R=Sum.
sum_over_range(Set,R) :- expand_custom_set_to_list(Set,EResult,_,sum_over_range),
  sum_over_range_aux(EResult,R).
  
%simple_sum_list([],A,A).
%simple_sum_list([(_,int(H))|T],Acc,R) :- NA is Acc+H, simple_sum_list(T,NA,R).

:- block sum_over_range_aux(-,?).
sum_over_range_aux(Set,int(Sum)) :-    % prints(sum_over_range(Set,Sum)),
  %print_term_summary(sum_over_range(Set)),debug:time(bsets_clp:sum_over_range2(Set,Sum)). % ,prints(res(Sum)).
  % Note : we cannot call range as it would collapse multiple occurences of a number
  %sum_over_range2(Set,0,int(Sum)).
  %debug:timer_call(bsets_clp:sum_range_int_project(Set,[],Sum)).
  sum_range_int_project(Set,[],Sum).

:- use_module(clpfd_interface,[clpfd_sum/3]).
:- block sum_range_int_project(-,?,?).
sum_range_int_project([],SumList,Sum) :- !, %print(posting_clpfd_sum(SumList,Sum)),nl,
  clpfd_sum(SumList,Sum,Posted),
  (Posted==true -> true ; sum_list(SumList,int(0),int(Sum))).
sum_range_int_project([(_,int(RanElement))|T],Acc,Sum) :- !, sum_range_int_project(T,[RanElement|Acc],Sum).
sum_range_int_project(Set,Acc,Sum) :-
   add_internal_error('Not a list as argument: ',sum_range_int_project(Set,Acc,Sum)),fail.

% non-CLPFD backup, when posting fails or CLPFD turned off
:- block sum_list(-,?,?).
sum_list([],Acc,Acc).
sum_list([H|T],Acc,Res) :- int_plus(int(H),Acc,NewAcc), sum_list(T,NewAcc,Res).


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:mul_over_range([(int(1),int(2))],int(2)))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:mul_over_range([],int(1)))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:mul_over_range([(int(1),int(2)),(int(1),int(3))],int(6)))).

:- block mul_over_range(-,?).
mul_over_range([],R) :- !, R=int(1).
mul_over_range(CS,R) :- mul_of_range_custom_explicit_set(CS,Mul),!, R=Mul.
mul_over_range(Set,R) :- expand_custom_set_to_list(Set,EResult,_,mul_over_range),
  mul_over_range_aux(EResult,R).


mul_over_range_aux(X,int(1)) :- empty_set(X).
mul_over_range_aux(X,Sum) :-
   remove_exact_first_element(First,X,Rest),
   second_of_pair(First,SumElement),
   mul_over_range(Rest,RSum),
   times(SumElement,RSum,Sum).


/********************************************/
/* surjection_relation(R,Domain,Range)      */
/*         R : Domain <->> Range            */
/********************************************/
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:surjection_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:surjection_relation_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(6),int(7)],WF),WF)).

surjection_relation_wf(R,Domain,Range,WF) :-
  is_surjective(R,Range,WF),
  % TODO: is not optimal since ran(R)<:Range is already implied by is_surjective and
  % checked a second time by relation_over_wf/4
  relation_over_wf(R,Domain,Range,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_surjection_relation_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(6),int(7)],WF),WF)).

not_surjection_relation_wf(R,Domain,Range,WF) :- 
   expand_custom_set_to_list(R,ER,_,not_surjection_relation_wf),
   not_tot_surj_rel(ER,[],Domain,Range,Range,WF).

/*********************************************/
/* total_surjection_relation(R,Domain,Range) */
/*         R : Domain <<->> Range            */
/*********************************************/
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_surjection_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:total_surjection_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).


:- assert_must_succeed((findall(R,bsets_clp:total_surjection_relation(R,[int(1)],[int(11),int(12)]),L), length(L,1))).
%:- assert_must_succeed((findall(R,bsets_clp:total_surjection_relation(R,[int(1),int(2)],[int(11),int(12)]),L), length(L,7))).
% the new domain predicate also instantiates from result; meaning that duplicate solutions are now generated
:- assert_must_succeed((findall(SR,(bsets_clp:total_surjection_relation(R,[int(1),int(2)],[int(11),int(12)]),sort(R,SR)),L), sort(L,SL),length(SL,7))).
:- assert_must_succeed((findall(R,bsets_clp:total_surjection_relation(R,[int(1),int(2)],[int(11)]),L), length(L,1))).

total_surjection_relation(R,Domain,Range) :- init_wait_flags(WF),
    total_surjection_relation_wf(R,Domain,Range,WF), ground_wait_flags(WF).

total_surjection_relation_wf(R,Domain,Range,WF) :-
  relation_over_wf(R,Domain,Range,WF),
  check_relation_is_total(R,Domain,WF), % calls domain which now instantiates R if Domain known
  check_relation_is_surjective(R,Range,WF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_relation_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_total_surjection_relation_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

not_total_surjection_relation_wf(R,Domain,Range,WF) :- 
   expand_custom_set_to_list(R,ER,_,not_total_surjection_relation_wf),
   not_tot_surj_rel(ER,Domain,Domain,Range,Range,WF).


/********************************************/
/*  partial_surjection(R,DomType,RangeType) */
/*         R : DomType +->> RangeType       */
/********************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6),int(2)],WF),WF)).  
:- assert_must_succeed((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_succeed((bsets_clp:partial_surjection(X,[int(1),int(2),int(3)],global_set('Name')),
                     kernel_objects:equal_object(X,[(int(2),fd(1,'Name')),(int(1),fd(2,'Name')),(int(3),fd(3,'Name'))]))).
:- assert_must_succeed((bsets_clp:partial_surjection_wf(X,[int(1),int(2),int(3)],global_set('Name'),_WF),
                     kernel_objects:equal_object(X,[(int(2),fd(1,'Name')),(int(1),fd(2,'Name')),(int(3),fd(3,'Name'))]))).
:- assert_must_succeed((bsets_clp:partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_succeed_multiple((bsets_clp:partial_surjection(X,[int(1),int(2),int(3),int(4)],[int(7),int(6)]),
                    % print(X),nl,
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6)),(int(3),int(6))]))).  /* mult. */
:- assert_must_succeed((X=[(int(2),int(7)),(int(1),int(6)),(int(3),int(6))],
                        bsets_clp:partial_surjection(X,[int(1),int(2),int(3),int(4)],[int(7),int(6)]))). 
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(3),int(8))])).
:- assert_must_succeed_multiple((bsets_clp:partial_surjection(_X,
                        [int(1),int(2),int(3),int(4),int(5),int(6),int(7)],[int(2),int(3),int(4)]) )).
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(2),int(6))])).
                     
partial_surjection(R,Domain,Range) :- init_wait_flags(WF),
  partial_surjection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

partial_surjection_wf(R,Domain,Range,WF) :-
  check_card_greater_equal(Domain,geq,Range),
  is_surjective(R,Range,WF),
  partial_function_wf(R,Domain,Range,WF).
  

% check_card_greater_equal(A,B) : quick check that card(A) >= card(B); also works with infinite cardinality
% TO DO: replace by a better constraint propagating predicate (also working for partially instantiated lists,...)
% compared with computing card and setting up < constraint: will only compute card if it can be done efficiently + deals with inf
% check_card_greater_equal(SetA,EQ,SetB) ;  EQ=eq or geq
:- block check_card_greater_equal(-,?,?).
check_card_greater_equal([],_,R) :- !, empty_set(R).
check_card_greater_equal(A,EQ,B) :- check_card_greater_equal2(A,EQ,B).

:- use_module(inf_arith,[block_inf_greater_equal/2]).
:- block check_card_greater_equal2(?,?,-).
check_card_greater_equal2(A,EQ,B) :-
   efficient_card_for_set(A,CardA,CodeA),
   efficient_card_for_set(B,CardB,CodeB),!,
   %print(check_geq(A,B,CardA,CodeA,CardB,CodeB)),nl,
   call(CodeA), call(CodeB),
   %% print(check_geq_val(CardA,CardB)),nl, %%
   (EQ=eq -> CardA=CardB ; inf_arith:block_inf_greater_equal(CardA,CardB)).
check_card_greater_equal2(_A,_,_B).


:- block is_surjective(-,-,?).
is_surjective(R,Range,WF) :- % print(is_surjective(R,Range,WF)),nl, %
  (var(R) -> setup_surj_range(Range,R)
           ; range_wf(R,Range,WF)).  
  
setup_surj_range(Range,R) :- 
   setup_range(Range,Res,DONE),  %print(setup_range(Range,Res,DONE)),nl,
   equal_when_done(Res,R,DONE).
:- block equal_when_done(?,?,-).
equal_when_done(Res,R,_DONE) :- equal_object(Res,R).
    
   
:- block setup_range(-,?,?).
setup_range(global_set(G),Res,DONE) :-
    expand_custom_set(global_set(G),ES), setup_range(ES,Res,DONE).
setup_range(avl_set(S),Res,DONE) :- expand_custom_set(avl_set(S),ES), setup_range(ES,Res,DONE).
setup_range(closure(P,T,B),Res,DONE) :-
    expand_custom_set(closure(P,T,B),ES), setup_range(ES,Res,DONE).
setup_range([],_,done).
setup_range([H|T],[(_,H)|ST],DONE) :- setup_range(T,ST,DONE).


   
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],
                                                  [int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],
                                                  [int(7),int(6),int(2)],WF),WF)).  
:- assert_must_fail((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_fail((bsets_clp:not_partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(3),int(8))])).

/* /: Domain +->> Range */
not_partial_surjection(R,Domain,Range) :- init_wait_flags(WF),
  not_partial_surjection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).
  
:- block not_partial_surjection_wf(-,?,?,?), not_partial_surjection_wf(?,-,?,?),
   not_partial_surjection_wf(?,?,-,?). 
not_partial_surjection_wf(R,DomType,RangeType,WF) :-   
    % TO DO: create a new partial_function_test kernel_equality like function instead
    % The fact that R is not a partial surjection is difficult to exploit !?
    when(ground(R), not_partial_surjection2(R,DomType,RangeType,WF)).
%:- block not_partial_surjection2(?,?,?,?,-).  
not_partial_surjection2(R,DomType,RangeType,WF) :- 
   not_partial_function(R,DomType,RangeType,WF).
not_partial_surjection2(R,DomType,RType,WF) :- partial_function_wf(R,DomType,RType,WF),
    invert_relation_wf(R,IR,WF),
    not_total_relation_wf(IR,RType,DomType,WF). 

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:total_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- assert_must_succeed((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_succeed((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7))])).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(8))])).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(3),int(6)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [])).


total_relation_wf(R,Domain,Range,WF) :- relation_over_wf(R,Domain,Range,WF),
   check_relation_is_total(R,Domain,WF).

% this predicates assume that the relation's range and domain have already been checked
check_relation_is_total(Relation,Domain,WF) :- domain_wf(Relation,Domain,WF).
check_relation_is_surjective(Relation,Range,WF) :-
    range_wf(Relation,Range,WF). % we could also call is_surjective (which does setup_surj_range) ?
  
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_total_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_fail((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(8))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(3),int(6)),(int(1),int(7))])).

:- block not_total_relation_wf(-,?,?,?).
not_total_relation_wf(FF,Domain,Range,_WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range), 
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    not_equal_object(FFDomain,Domain).
not_total_relation_wf(FF,Domain,Range,WF) :-  nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    kernel_equality:equality_objects(FFDomain,Domain,Result), % not yet implemented ! % TODO ! -> sub_set,equal,super_set
    when(nonvar(Result),(Result=pred_false -> true ; not_subset_of_wf(FFRange,Range,WF))). 
not_total_relation_wf(R,Domain,Range,WF) :- 
   expand_custom_set_to_list(R,ER,_,not_total_relation_wf),
   not_tot_surj_rel(ER,Domain,Domain,[],Range,WF).
 
 % can be used to check not total, not surj, not total surj relation
:- block not_tot_surj_rel(-,?,?,?,?,?).
not_tot_surj_rel([],DelDomain,_,DelRange,_,WF) :- 
    (not_empty_set_wf(DelDomain,WF) ; empty_set(DelDomain),not_empty_set_wf(DelRange,WF)).
not_tot_surj_rel([(X,Y)|T],DelDom,Dom,DelRan,Ran,WF) :- 
      membership_test(Dom,X,MemRes),
      not_tr2(MemRes,X,Y,T,DelDom,Dom,DelRan,Ran,WF).
  
:- block not_tr2(-,?,?,?,?,?,?,?,?).
 not_tr2(pred_false,_X,_Y,_T,_DelDom,_Dom,_DelRan,_Ran,_WF).
 not_tr2(pred_true,X,Y,T,DelDom,Dom,DelRan,Ran,WF) :-
       delete_element_wf(X,DelDom,DelDom2,WF),  % set DelDom initially to [] to avoid totality check
       membership_test(Ran,Y,MemRes),
       not_tr3(MemRes,Y,T,DelDom2,Dom,DelRan,Ran,WF).
 
:- block not_tr3(-,?,?,?,?,?,?,?).
 not_tr3(pred_false,_Y,_T,_DelDom2,_Dom,_DelRan,_Ran,_WF).
 not_tr3(pred_true,Y,T,DelDom2,Dom,DelRan,Ran,WF) :-
        delete_element_wf(Y,DelRan,DelRan2,WF),  % set DelRan initially to [] to avoid surjection check
        not_tot_surj_rel(T,DelDom2,Dom,DelRan2,Ran,WF).
 
/******************************************/
/*  total_surjection(R,DomType,RangeType) */
/*        R : DomType -->> RangeType      */
/******************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_surjection_wf([(int(2),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_succeed_check((bsets_clp:total_surjection_wf([(int(2),int(6)),(int(1),int(7)),(int(3),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),kernel_waitflags:ground_det_wait_flag(WF)))). %% TO DO: get rid of multiple solutions
:- assert_must_succeed((bsets_clp:total_surjection(X,[int(1)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:total_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_succeed((bsets_clp:total_surjection(X,[int(1),int(2)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_surjection([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_surjection([(int(7),int(7))],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_surjection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).
                     
                     
total_surjection(R,Domain,Range) :- init_wait_flags(WF),
  total_surjection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).
  
total_surjection_wf(R,DomType,RangeType,WF) :-
   check_card_greater_equal(DomType,geq,RangeType),
   total_function_wf(R,DomType,RangeType,WF),
  %  setup_surj_range(RangeType,R).
   check_relation_is_surjective(R,RangeType,WF).
   % invert_relation_wf(R,IR,WF), total_relation_wf(IR,RangeType,DomType,WF).
    
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7)),(int(3),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7)),(int(3),int(8))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- block not_total_surjection_wf(-,?,?,?), not_total_surjection_wf(?,-,?,?),
   not_total_surjection_wf(?,?,-,?). 
not_total_surjection_wf(R,DomType,RangeType,WF) :-  
    %get_middle_wait_flag(not_total_surjection,WF,LWF),
    % TO DO: create a new total_function_test kernel_equality like function instead; or write custom not_total_function code
    % The fact that R is not a total surjection is difficult to exploit !?
    when(ground(R), not_total_surjection2(R,DomType,RangeType,WF)).
%:- block not_total_surjection2(?,?,?,?,-).
not_total_surjection2(R,DomType,RangeType,WF) :-
    not_total_function(R,DomType,RangeType,WF).
not_total_surjection2(R,DomType,RangeType,WF) :-
    total_function_wf(R,DomType,RangeType,WF),
    not_partial_surjection_wf(R,DomType,RangeType,WF).

/*******************************************/
/*  partial_injection(R,DomType,RangeType) */
/*         R : DomType >+> RangeType       */
/*******************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_injection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_injection_wf([(int(1),int(6)),(int(4),int(7)),(int(2),int(8))],[int(1),int(2),int(3),int(4)],[int(7),int(6),int(8),int(9)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_injection_wf([(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed((bsets_clp:partial_injection(X,[int(1)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:partial_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_fail((bsets_clp:partial_injection(X,[int(1),int(2)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:partial_injection([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:partial_injection([(int(7),int(7))],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:partial_injection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)]))).
:- assert_must_fail((bsets_clp:partial_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).
                      

partial_injection(R,Domain,Range) :- init_wait_flags(WF),
  partial_injection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).
  
:- block partial_injection_wf(-,-,?,?).
partial_injection_wf(FF,Domain,Range,WF) :- nonvar(FF),
   custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),!,
   check_subset_of_wf(FFDomain,Domain,WF),
   check_subset_of_wf(FFRange,Range,WF).
partial_injection_wf(R,DomType,RangeType,WF) :-   
     try_expand_and_convert_to_avl_unless_large(R,ER), %try_expand_only_custom_closure_global(R,ER),
    partial_function_wf(ER,DomType,RangeType,WF),
    injective(ER,WF).
 %   invert_relation_wf(R,IR,WF),
 %    print(pf(IR)),nl,
  %  partial_function_wf(IR,RangeType,DomType,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective([(int(1),int(6)),(int(4),int(7)),(int(2),int(8))],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:injective([(int(1),int(6)),(int(4),int(7)),(int(2),int(7))],WF),WF)). 

:- block injective(-,?).
injective(FF,_WF) :- custom_explicit_sets:dom_range_for_specific_closure(FF,_FFDomain,_FFRange,function(bijection)),!.
injective(Rel,WF) :- expand_custom_set_to_list(Rel,ERel,_,injective),
   injective(ERel,[],WF).

:- block injective(-,?,?).
injective([],_,_).
injective([(_From,To)|T],SoFar,WF) :-  %print(inj_check(_From,To,SoFar,T)),nl,
    not_element_of_wf(To,SoFar,WF), /* check that it is injective */
   %print(not_el(To)),nl,
    add_new_element(To,SoFar,SoFar2),
    %print(added(To,SoFar2)),nl,
   injective(T,SoFar2,WF).
% no case for global_set: it cannot be a relation; two cases below not required because of expand_custom_set_to_list
%injective(avl_set(S),SoFar,WF) :- expand_custom_set(avl_set(S),ES), injective(ES,SoFar,WF).
%injective(closure(P,T,B),SoFar,WF) :- expand_custom_set(closure(P,T,B),ES), injective(ES,SoFar,WF).


   
/* /: Dom >+> R */

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(2),int(8))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(3),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- block not_partial_injection(-,?,?,?), not_partial_injection(?,-,?,?),
   not_partial_injection(?,?,-,?).
not_partial_injection(R,DomType,RangeType,WF) :- 
    %get_middle_wait_flag(not_partial_injection,WF,LWF),
    % TO DO: create a new partial_function_test kernel_equality like function instead
    when(ground(R), not_partial_injection2(R,DomType,RangeType,WF)).
%:- block not_partial_injection2(?,?,?,?,-). 
not_partial_injection2(R,DomType,RType,WF) :-
   not_partial_function(R,DomType,RType,WF).
not_partial_injection2(R,DomType,RType,WF) :- 
    partial_function_wf(R,DomType,RType,WF),
    invert_relation_wf(R,IR,WF),
%    print(not_partial_function(IR,RType,DomType)),nl,
    not_partial_function(IR,RType,DomType,WF). 
    
/*****************************************/
/*  total_injection(R,DomType,RangeType) */
/*       R : DomType >-> RangeType       */
/*****************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_injection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:total_injection_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).  
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_total_injection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_total_injection([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed((bsets_clp:total_injection(X,[int(1)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:total_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_fail((bsets_clp:total_injection(X,[int(1),int(2)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_injection([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_injection([(int(7),int(7))],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_injection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).
                      
                      
total_injection(R,Domain,Range) :- init_wait_flags(WF),
  total_injection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).
  
total_injection_wf(R,DomType,RangeType,WF) :-  % print_message(total_inj(R)),
    try_expand_and_convert_to_avl_unless_large(R,ER), %try_expand_only_custom_closure_global(R,ER), print(er(ER)),nl,
    total_function_wf(ER,DomType,RangeType,WF),  %  print_message(total_f(ER)),
    injective(ER,WF).
    
 
:- block not_total_injection(-,?,?,?), not_total_injection(?,-,?,?),
   not_total_injection(?,?,-,?). 
not_total_injection(R,DomType,RangeType,WF) :-  
    %get_middle_wait_flag(not_total_injection,WF,LWF),
    % TO DO: create a new total_function_test kernel_equality like function instead
    when((ground(R),ground(DomType)),
       not_total_injection2(R,DomType,RangeType,WF)).
%:- block not_total_injection2(?,?,?,?,-).
not_total_injection2(R,DomType,RangeType,WF) :- %print(not_tot_inj(R)),nl,
   not_total_function(R,DomType,RangeType,WF) . %, print(not_tot_fun(R)),nl.
not_total_injection2(R,DomType,RangeType,WF) :- % print(try2(R)),nl,
   total_function_wf(R,DomType,RangeType,WF), %print(tot_fun(R)),nl,
   invert_relation_wf(R,IR,WF), %print(inv(IR)),nl,
   not_partial_function(IR,RangeType,DomType,WF).

/***********************************/
/*  partial_bijection(R,DomType,RangeType) */
/*   R : DomType >+>> RangeType    */
/***********************************/


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_bijection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:partial_bijection_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).  
:- assert_must_succeed((partial_bijection(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(1),int(6)),(int(2),int(7))]))).
:- assert_must_succeed((partial_bijection(X,[int(1),int(2),int(3),int(4)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(3),int(6))])).
:- assert_must_fail((partial_bijection(X,[int(1),int(2)],[int(7),int(6),int(5)]),
                     X = [(int(2),int(7)),(int(1),int(6))])).
           
partial_bijection(R,Domain,Range) :- init_wait_flags(WF),
  partial_bijection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).
            
partial_bijection_wf(R,DomType,RangeType,WF) :-
    partial_injection_wf(R,DomType,RangeType,WF),
    partial_surjection_wf(R,DomType,RangeType,WF).

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_bijection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_partial_bijection([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_bijection([(int(2),int(7)),(int(1),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)). 

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_bijection([(int(2),int(7)),(int(3),int(6))],[int(1),int(2),int(3),int(4)],[int(7),int(6)],WF),WF)). 
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_partial_bijection([(int(2),int(7)),(int(1),int(6))],[int(1),int(2)],[int(7),int(6),int(5)],WF),WF)). 


:- block not_partial_bijection(-,?,?,?), not_partial_bijection(?,-,?,?),
   not_partial_bijection(?,?,-,?).
not_partial_bijection(R,DomType,RangeType,WF) :- 
    get_middle_wait_flag(not_partial_bijection,WF,LWF),
    not_partial_bijection2(R,DomType,RangeType,WF,LWF).
:- block not_partial_bijection2(?,?,?,?,-). 
not_partial_bijection2(R,DomType,RangeType,WF,_) :- 
    not_partial_injection(R,DomType,RangeType,WF).
not_partial_bijection2(R,DomType,RangeType,WF,_) :-
    partial_injection_wf(R,DomType,RangeType,WF),
    not_partial_surjection_wf(R,DomType,RangeType,WF).


/* ------------------ */

/* The transitive & reflexive closure of a relation */

:- assert_must_succeed((bsets_clp:relational_reflexive_closure([(int(1),int(4))],X),
                        kernel_objects:equal_object(X,[(int(1),int(1)),(int(1),int(4)),(int(4),int(4))]))).
:- assert_must_succeed((bsets_clp:relational_reflexive_closure([(int(1),int(4)),(int(4),int(2))],X),
                        kernel_objects:equal_object(X,[(int(1),int(1)),(int(2),int(2)),(int(4),int(4)),(int(1),int(4)),(int(4),int(2)),
                                             (int(1),int(2))]))).
:- assert_must_succeed((bsets_clp:relational_reflexive_closure([(int(1),int(4)),(int(4),int(2)),(int(2),int(3))],X),
kernel_objects:equal_object(X,[(int(1),int(1)),(int(2),int(2)),(int(3),int(3)),(int(4),int(4)),
                                      (int(1),int(4)),(int(4),int(2)),(int(2),int(3)),
                                             (int(4),int(3)),(int(1),int(2)),(int(1),int(3))]))).

:- block relational_reflexive_closure(-,?).
relational_reflexive_closure(Relation,Result) :-
    try_expand_and_convert_to_avl_with_check(Relation,ARelation,relational_reflexive_closure),
    relational_reflexive_closure2(ARelation,Result).
:- block relational_reflexive_closure2(-,?).
relational_reflexive_closure2(ARelation,Result) :- % print_term_summary(closure(ARelation,Result)),
    (closure_for_explicit_set(ARelation,Res)
      -> kernel_objects:equal_object(Res,Result)
      ; (expand_custom_set_to_list(ARelation,ERelation,_,relational_reflexive_closure2),
         is_full_relation(ERelation,WaitVar),
         compute_trans_closure(ERelation,Result1,WaitVar),
		% print(adding_id(ERelation,Result1)),nl,
		 rel_iterate0(ERelation,Result1,Result)
		%, print(done(Result)),nl
        )
    ).



/* The transitive (not reflexive) closure of a relation */

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:relational_trans_closure([(int(1),int(2)),(int(2),int(6))],[(int(1),int(2)),(int(1),int(6)),(int(2),int(6))]))). 
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:relational_trans_closure([(int(1),int(2)),(int(2),int(6)),(int(1),int(3))],[(int(1),int(2)),(int(1),int(3)),(int(1),int(6)),(int(2),int(6))]))). 
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:relational_trans_closure([(int(6),int(7)),(int(1),int(2)),(int(2),int(6)),(int(1),int(3))],[(int(1),int(2)),(int(1),int(3)),(int(1),int(6)),(int(2),int(6)),(int(1),int(7)),(int(2),int(7)),(int(6),int(7))]))). 
:- assert_must_succeed((bsets_clp:relational_trans_closure([(int(1),int(4))],X),
                        kernel_objects:equal_object(X,[(int(1),int(4))]))).
:- assert_must_succeed((bsets_clp:relational_trans_closure([(int(1),int(4)),(int(4),int(2))],X),
                        kernel_objects:equal_object(X,[(int(1),int(4)),(int(4),int(2)),
                                             (int(1),int(2))]))).
:- assert_must_succeed((bsets_clp:relational_trans_closure([(int(1),int(4)),(int(4),int(2)),(int(2),int(3))],X),
                        kernel_objects:equal_object(X,[(int(1),int(4)),(int(4),int(2)),(int(2),int(3)),
                                             (int(4),int(3)),(int(1),int(2)),(int(1),int(3))]))).

% transitive closure for relations (closure1)
:- block relational_trans_closure(-,?).
relational_trans_closure(Relation,Result) :- % print_term_summary(closure1(Relation,Result)),
    try_expand_and_convert_to_avl_with_check(Relation,ARelation,relational_trans_closure),
    relational_trans_closure2(ARelation,Result).
:- block relational_trans_closure2(-,?).
relational_trans_closure2(ARelation,Result) :- %print(closure1(ARelation)),nl,
    (closure1_for_explicit_set(ARelation,Res)
      -> % print_term_summary(closure1_explicit(ARelation,Res)),
         kernel_objects:equal_object(Res,Result)
      ;  expand_custom_set_to_list(ARelation,ERelation,_,relational_trans_closure2),
         is_full_relation(ERelation,WaitVar),
         compute_trans_closure(ERelation,Result,WaitVar)
    ).

:- block compute_trans_closure(?,?,-).
compute_trans_closure(Relation,Result,_) :- % print(computing_trans_closure(Relation)),nl,
  compute_trans_closure(Relation,Result). %,print(result(Result)),nl.
  
compute_trans_closure(Relation,Result) :- %print_message(iter(Relation)),
    one_closure_iteration(Relation,Relation,Relation,Result1),
    ( equal_object(Result1,Relation), equal_object(Result,Result1)
      ;
      not_equal_object(Result1,Relation), compute_trans_closure(Result1,Result)
    ).
     
one_closure_iteration([],_,Res,Res).
one_closure_iteration([(X,Y)|T],ExpandedPreviousRel,PreviousRel,OutRel) :- 
    add_tuples(ExpandedPreviousRel,X,Y,PreviousRel,IntRel),
    one_closure_iteration(T,ExpandedPreviousRel,IntRel,OutRel).
    
 
add_tuples([],_,_,OutRel,OutRel).
add_tuples([(X,Y)|T],OX,OY,InRel,OutRel) :-
   ( equal_object(Y,OX),add_element((X,OY),InRel,IntRel) ;
     not_equal_object(Y,OX), IntRel = InRel),
   add_tuples(T,OX,OY,IntRel,OutRel).


:- assert_must_succeed((is_full_relation(X,R),var(R),X=[],R==true)).
:- assert_must_succeed((is_full_relation(X,R),var(R),X=[(A,B)|T],var(R),A=int(1),var(R),B=A,var(R),T=[],R==true)).
:- block is_full_relation(-,?).
is_full_relation([],R) :- !,R=true.
is_full_relation([(X,Y)|T],W) :- !, when((nonvar(X),nonvar(Y)),is_full_relation(T,W)).
is_full_relation(X,R) :-
    add_internal_error('Illegal Set for is_full_relation: ',is_full_relation(X,R)),fail.

   
/* ------------------ */
                              

