% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(kernel_objects,[basic_type/2, 
                     enumerate_basic_type/2, all_objects_of_type/2,
                     max_cardinality/2,
                     enumerate_type/3, % last argument basic or tight
                     enumerate_tight_type/2, enumerate_tight_type/3,
                     enumerate_int/3, enumerate_natural/3,
                     enum_warning/3,
                     all_strings/1, is_string/2, is_not_string/1,
                     
                     top_level_dif/2, not_top_level_functor/2,
                     equal_object_optimized/2,
                     equal_object/2, not_equal_object/2,
                     equal_cons/3,  equal_cons_wf/4, equal_cons_lwf/4,
                     get_next_element/3,
                     
                     equality_objects/3,
                     membership_test/3,
                     
                     %is_a_set/1, 
                     empty_set/1, not_empty_set/1, not_empty_set_wf/2,
                     exact_element_of/2,
                     check_element_of/2, check_element_of_wf/3,
                     check_element_of_function_closure/6,
                     not_element_of/2, not_element_of_wf/3,
                     
                     add_element/3, add_element/4, add_new_element/3,
                     remove_element/3, remove_element_wf/4,remove_element_wf/5,remove_element_wf_if_not_infinite_or_closure/6,
                     remove_exact_element/3,
                     remove_exact_first_element/3,
                     delete_element/3, delete_element_wf/4,
                     
                     partition_wf/3, not_partition_wf/3,
                     %all_different/2,
                     disjoint_sets/3, not_disjoint_sets/3,
                     
                     union/3, union_wf/4, union_generalized/2, union_generalized_wf/3,
                     intersection/3, intersection_generalized_wf/4,
                     difference_set/3, difference_set_wf/4,
                     
                     strict_subset_of/2, strict_subset_of_wf/3,
                     check_subset_of/2, check_subset_of_wf/3, check_finite_subset_of_wf/3,
                     check_non_empty_subset_of_wf/3, check_finite_non_empty_subset_of_wf/3,
                     not_subset_of/2, not_subset_of_wf/3, not_both_subset_of/5,
                     not_finite_subset_of_wf/3,
                     not_strict_subset_of/2, not_strict_subset_of_wf/3,
                     not_non_empty_subset_of_wf/3, not_non_empty_finite_subset_of_wf/3,
                     both_global_sets/4,check_subset_of_global_sets/2,

	             first_of_pair/2, second_of_pair/2,
                     minimum_of_set_extension_list/4,
                     maximum_of_set_extension_list/4,
                     minimum_of_set/4, maximum_of_set/4,
                     is_finite_set_wf/2, is_infinite_set_wf/2, test_finite_set_wf/3,
                     finite_cardinality_as_int/3, cardinality_as_int_for_wf/2,
                     cardinality_as_int/2, cardinality/2, card_leq/2, card_geq/2,
                     cardval_greater_than/2,
                     cardinality_greater/5, cardinality_greater_equal/5,
                     cardinality_of_set_extension_list/2,
                     
                     cartesian_product/3, %  removed
                     is_cartesian_pair_wf/4, not_is_cartesian_pair/4, 
                     
                     power_set/2, non_empty_power_set/2,
                     
                     % is_boolean/1, %is_not_boolean/1,
                     is_integer/2, is_not_integer/1,
                     is_natural/2, is_natural1/2,
                     is_implementable_int/2,is_implementable_nat/2, is_implementable_nat1/2,
                     is_not_natural/1, is_not_natural1/1,
                     is_not_implementable_int/1,is_not_implementable_nat/1, is_not_implementable_nat1/1,
                     
                     less_than/2, less_than_equal/2, 
                     less_than_direct/2, less_than_equal_direct/2, 
                     safe_less_than_equal/2, safe_less_than_equal/3,
                     safe_pow2/2, safe_mul/3, safe_add/3, safe_pown/3,
                     greater_than/2, greater_than_equal/2,
                     int_plus/3,
                     division/5, modulo/5,
                     int_minus/3, unary_minus/2,
                     % nat_range/3, % removed
                     in_nat_range_wf/4, not_in_nat_range/3,
                     in_nat_range/3, % version without enumeration
                     times/3, square/3,
                     int_power/5,
                     % pred/2, succ/2, removed
                     integer_global_set/1,
                   
                     
                     element_of_global_set/2,element_of_global_set_wf/3,not_element_of_global_set/2,
                     
                     exhaustive_kernel_check/1, exhaustive_kernel_check_wf/2, exhaustive_kernel_check_wf/3,
                     exhaustive_kernel_check_wfdet/2,
                     exhaustive_kernel_succeed_check/1, exhaustive_kernel_fail_check/1,
                     exhaustive_kernel_fail_check_wf/2, exhaustive_kernel_fail_check_wfdet/2,
                     exhaustive_kernel_check/2, exhaustive_kernel_succeed_check/2, exhaustive_kernel_fail_check/2,

                     singleton_set_element/3
                     
                     ]).
                     
                     
%:- print(loading_kernel_objects),nl.

%portray_message(informational, _).
:- use_module(library(terms)).
:- use_module(self_check).

:- use_module(debug).
:- use_module(tools_printing,[print_term_summary/1]).
:- use_module(tools).   

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides operations for the basic datatypes of ProB (equal, not_equal, enumeration).').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(typechecker).
:- use_module(error_manager).

:- use_module(b_global_sets). %,[global_type/2, b_global_set_cardinality/2, b_empty_global_set/1]).

:- use_module(kernel_waitflags).
:- use_module(library(lists)).

%:- use_module(library(clpfd)).
%:- use_module(fd_utils).
:- use_module(fd_utils_clpfd).

:- use_module(kernel_freetypes).

:- use_module(custom_explicit_sets).


:- use_module(typechecker).

%:- use_module(clpfd_off_interface). %
% on a 32 bit system: use clpfd_off_interface; on 64 bit system clpfd_interface should be ok (integer overflows)
:- use_module(clpfd_interface). %


:- type atomic_type +--> (term(integer,[]) ; term(string,[]) ;constant ; constant(list(atomic)) ; abort ; boolean ; global(atomic)).
:- type atomic_any_type +--> (type(atomic_type) ; term(any,[]) ).
:- type basic_type_descriptor +--> (type(atomic_any_type) ; set(basic_type_descriptor) ;
                                    seq(basic_type_descriptor) ;
                                    couple(basic_type_descriptor,basic_type_descriptor) ;
                                    record(list(type(field_type))) ;
                                    freetype(atomic)).

:- type inferred_basic_type_descriptor +--> (var ; type(atomic_type) ; set(inferred_basic_type_descriptor) ;
                                    seq(inferred_basic_type_descriptor) ;
                                    couple(inferred_basic_type_descriptor,inferred_basic_type_descriptor)).

:- type fd_index +--> (integer ; var).
:- type fd_set +--> (atomic ; var).
:- type fd_term +--> fd(fd_index,fd_set).
:- type bsets_integer +--> int((integer ; var)).
:- type bsets_string +--> string((atomic ; var)).
:- type bsets_bool +--> (pred_false /* bool_false */ ; pred_true /* bool_true */).
:- type field_type +--> field(atomic,basic_type_descriptor).

%:- type bsets_sequence +--> (nil_seq ; cons(type(bsets_object),type(bsets_sequence))).
:- type bsets_set +--> vlist(type(bsets_object)).
:- type bsets_couple +--> term(',',[type(bsets_object),type(bsets_object)]).
:- type bsets_global +--> global_set((atomic ; var)).
:- type bsets_field +--> field(atomic,type(bsets_object)).
:- type bsets_record +--> rec((var ; list(bsets_field))).
:- type bsets_freetype +--> freeval(atomic,(atomic ; var),type(bsets_object)).
:- type bsets_avl +--> avl_set( ground ).
:- type bsets_closure +--> (closure(list(type(variable_id)),
                                   list(type(basic_type_descriptor)),type(boolean_expression))
                           ; closure_x(list(type(variable_id)),
                                   list(type(basic_type_descriptor)),type(boolean_expression),any)
                           ).

:- type bsets_object +--> (fd_term ; bsets_integer ; bsets_bool ; term(term,[any]) ; bsets_set ;
                           abort(any) ; bsets_couple ; bsets_string ; bsets_global ; var;
                           bsets_closure ; bsets_avl ; bsets_record ; bsets_freetype).


:- assert_must_succeed(kernel_waitflags:set_silent(true)). % disable waitflag store not init msgs


  
 
% a predicate to exhaustively check a kernel predicate with all possible modes


exhaustive_kernel_check(C) :- exhaustive_kernel_check4([],C,true,true).
exhaustive_kernel_check(Opts,C) :- exhaustive_kernel_check4(Opts,C,true,true).
exhaustive_kernel_check_wf(C,WF) :- exhaustive_kernel_check_wf([],C,WF).
exhaustive_kernel_check_wf(Opts,C,WF) :-
   exhaustive_kernel_check4(Opts,C,kernel_waitflags:init_wait_flags(WF),
                                 kernel_waitflags:ground_wait_flags(WF)).
exhaustive_kernel_check_wfdet(C,WF) :-
   exhaustive_kernel_check4([],C,kernel_waitflags:init_wait_flags(WF),
                                 kernel_waitflags:ground_det_wait_flag(WF)).

exhaustive_kernel_check4(Opts,Call,Pre,Post) :- enumerate_kernel_call(Call,Opts,ECall,Code),
   debug_println(9,exhaustive_kernel_check(ECall,Code)),
   flatten_call((Pre,ECall,Code,Post),FullCall),
   time_out_call(must_succeed_without_residue(FullCall)),debug_println(9,ok),
   fail.
exhaustive_kernel_check4(_,_,_,_).

flatten_call((A,B),Res) :- !,flatten_call(A,FA), flatten_call(B,FB), conjoin_call(FA,FB,Res).
flatten_call(Module:Call,Res) :- !, flatten_call(Call,F), Res=Module:F.
flatten_call(X,X).

conjoin_call(true,X,R) :- !,R=X.
conjoin_call(X,true,R) :- !, R=X.
conjoin_call(X,Y,(X,Y)).

exhaustive_kernel_succeed_check(C) :- exhaustive_kernel_succeed_check([],C).
exhaustive_kernel_succeed_check(Opts,Call) :- enumerate_kernel_call(Call,Opts,ECall,Code),
   debug_println(9,exhaustive_kernel_succeed_check(ECall,Code)),
   flatten_call((ECall,Code),FullCall),
   time_out_call(must_succeed(FullCall)),debug_println(9,ok),
   fail.
exhaustive_kernel_succeed_check(_,_).

exhaustive_kernel_fail_check(C) :- exhaustive_kernel_fail_check4([],C,true,true).
exhaustive_kernel_fail_check(Opts,C) :- exhaustive_kernel_fail_check4(Opts,C,true,true).
exhaustive_kernel_fail_check_wf(C,WF) :-
   exhaustive_kernel_fail_check4([],C,kernel_waitflags:init_wait_flags(WF),
                                      kernel_waitflags:ground_wait_flags(WF)).
exhaustive_kernel_fail_check_wfdet(C,WF) :-
   exhaustive_kernel_fail_check4([],C,kernel_waitflags:init_wait_flags(WF),
                                      kernel_waitflags:ground_det_wait_flag(WF)).
exhaustive_kernel_fail_check4(Opts,Call,Pre,Post) :- enumerate_kernel_call(Call,Opts,ECall,Code),
   debug_println(9,exhaustive_kernel_fail_check(ECall,Code)),
   flatten_call((Pre,ECall,Code,Post),FullCall),
   time_out_call(must_fail(FullCall)),debug_println(9,ok),
   fail.
exhaustive_kernel_fail_check4(_,_,_,_).

% enumerate_kernel_call(Call, OptionList, NewCall, CodeAfter)
enumerate_kernel_call((A,B),Opts,(EA,EB),(CA,CB)) :- !,
    enumerate_kernel_call(A,Opts,EA,CA), enumerate_kernel_call(B,Opts,EB,CB).
enumerate_kernel_call(Module:Call,Opts,Module:ECall,Code) :- !, enumerate_kernel_call(Call,Opts,ECall,Code).
enumerate_kernel_call(Call,Opts,ECall,Code) :- Call=..[KernelPred|CArgs],
   (member(commutative,Opts)
     -> (Args=CArgs ; CArgs=[A1,A2|T], Args=[A2,A1|T])
    ; Args=CArgs
   ),
   l_enumerate_kernel_args(Args,EArgs,Code), ECall=..[KernelPred|EArgs].
l_enumerate_kernel_args([],[],true).
l_enumerate_kernel_args([H|T],[EH|ET],Code) :-
   enumerate_kernel_args(H,EH,C1),
   l_enumerate_kernel_args(T,ET,C2),
   permute_code((C1,C2),Code).

permute_code((true,C),R) :- !,R=C.
permute_code((C,true),R) :- !, R=C.
permute_code((C1,C2),(C1,C2)).
permute_code((C1,C2),(C2,C1)).

enumerate_kernel_args(Var,Res,Code) :- var(Var),!, Res=Var, Code=true.
enumerate_kernel_args(X,Res,Code) :- do_not_delay(X),!, Res=X, Code=true.
enumerate_kernel_args(Arg,Arg,true).
enumerate_kernel_args(Arg,NewArg,Code) :- 
   (term_is_of_type(Arg,bsets_object,no)
    -> Code = equal_object(NewArg,Arg)
     ; Code = '='(NewArg,Arg)).
enumerate_kernel_args([H|T],[H|NewT],Code) :- Code = equal_object(NewT,T).
enumerate_kernel_args([H|T],Res,Code) :- try_expand_and_convert_to_avl([H|T],AVL),
    AVL \= [H|T], enumerate_kernel_args(AVL,Res,Code).
enumerate_kernel_args([H|T],Res,Code) :- ground([H|T]),generate_member_closure([H|T],Closure),
    enumerate_kernel_args(Closure,Res,Code).

do_not_delay(b(_,_,_)). % do not delay B predicates and expressions
generate_member_closure(ExplicitSet,closure(['_zzzz_unit_tests'],[Type],Pred)) :-
   infer_type(ExplicitSet,set(Type)),
   Pred = 
   b(member(b(identifier('_zzzz_unit_tests'),Type,[generated]),
    b(value(ExplicitSet),set(Type),[])),pred,[]).

infer_type(Value,Type) :- (infer_value_type(Value,Type)
  -> true %,print(inferred(Type,Value)),nl
  ;  print('### Could not infer type: '), print(Value),nl,fail).

infer_value_type([],set(any)).
infer_value_type([H|_],set(Type)) :- infer_value_type(H,Type).
infer_value_type(int(_),integer).
infer_value_type(string(_),string).
infer_value_type((A,B),couple(TA,TB)) :- infer_value_type(A,TA), infer_value_type(B,TB).
infer_value_type(fd(_,T),global(T)).
infer_value_type(pred_true /* bool_true */,boolean).
infer_value_type(pred_false /* bool_false */,boolean).
infer_value_type(rec(Fields),record(FieldTypes)) :- infer_field_types(Fields,FieldTypes).
infer_value_type(freeval(Id,_Case,_Val),freetype(Id)).

infer_field_types([],[]).
infer_field_types([field(Name1,Val)|T],[field(Name1,VT)|TT]) :-
     infer_value_type(Val,VT),
     infer_field_types(T,TT).

:- assert_pre(kernel_objects:basic_type(Obj,Type), (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:basic_type(Obj,_), type_check(Obj,bsets_object)).

%:- block basic_type(-,-).

basic_type(fd(X,T),global(T)) :- !, global_type(fd(X,T),T). % will set up CLP(FD) domain for X

basic_type(X,TY) :-
  nonvar(TY) ->  true %basic_type2(TY,X) %basic_symbreak(TY,X)
                %print(ignore_basic_type(X,Y)),nl %, basic_type2(TY,X) %%STILL REQUIRED ?????
                              ;  print(var_type_basic_type(X,TY)),nl.

/*
basic_type2(set(Type),Set) :- !,basic_type_set(Type,Set,inf).
basic_type2(_,_).

:- use_module(inf_arith).
basic_type_set(Type,Set,Bound) :-
  (kernel_objects:max_cardinality(Type,C)
    -> inf_arith:infmin(C,Bound,Card)
     ; Card=inf),
   %print(basic_type_set(Set,Card,Type)),nl,nl,
   basic_type_set(Set,[],Type,Card).

:- block basic_type_set(-,?,?,?).
%basic_type_set(S,Sofar,T) :- print_message(basic_type_set(S,Sofar,T)),fail.
basic_type_set(CustomSet,_,_Type,_) :- is_custom_explicit_set(CustomSet,basic_type_set),!.
basic_type_set([],_,_,_).
basic_type_set([H|T],ElementsSoFar,Type,Card) :-
   inf_arith:infdec_if_positive(Card,C1),
   basic_type(H,Type),
   not_element_of(H,ElementsSoFar), %consistent([H|ElementsSoFar]),
   (C1=0 -> empty_set(T) ; basic_type_set(T,[H|ElementsSoFar],Type,C1)).
   
:- use_module(kernel_sym_break_order).
basic_symbreak(set(_),Set) :- !,sym_break_set(Set).
basic_symbreak(seq(_),Set) :- !,sym_break_set(Set).
basic_symbreak(_,_). */

/*
basic_type2(global(T),fd(X,T)) :- global_type(fd(X,T),T). % will set up CLP(FD) domain for X
basic_type2(set(Type),Set) :- basic_type_set(Set,[],Type).
basic_type2(seq(X),Seq) :- basic_type(Seq,set(couple(integer,X))).
basic_type2(boolean,_).
%% basic_type2(boolean,pred_true). basic_type2(boolean,pred_false).
basic_type2(couple(XT,YT),(X,Y)) :- basic_type(X,XT), basic_type(Y,YT).
basic_type2(string,string(_)).
basic_type2(constant,term(_)).
basic_type2(integer,int(_X)). %  :- clpfd_interface:clpfd_inrange(_X,-999999,999999). %%%%
basic_type2(string,string(_)).
basic_type2(record(FieldTypes),rec(Fields)) :-
  basic_field_types(Fields,FieldTypes,rec(Fields),record(FieldTypes)).
basic_type2(freetype(Id),freeval(Id,Case,Val)) :- basic_type_freetype(Id,Case,Val).
basic_type2(abort,abort(_)).
basic_type2(any,_).

:- block basic_type_set(-,?,?).
%basic_type_set(S,Sofar,T) :- print_message(basic_type_set(S,Sofar,T)),fail.
basic_type_set(CustomSet,[],_Type) :- is_custom_explicit_set(CustomSet,basic_type_set),!.
basic_type_set([],_,_).
basic_type_set([H|T],ElementsSoFar,Type) :-
   basic_type(H,Type),
   not_element_of(H,ElementsSoFar), %consistent([H|ElementsSoFar]),
   basic_type_set(T,[H|ElementsSoFar],Type).
 


:- block basic_type_freetype(?,-,?).
basic_type_freetype(Id,Case,Val) :-
   registered_freetype(Id,Cases),
   member(case(Case,CT),Cases),!,
   basic_type(Val,CT).

:- block basic_field_types(-,-,?,?).
basic_field_types([],[],_,_).
basic_field_types([field(Name1,Val)|T],[field(Name2,VT)|TT],Orig1,Orig2) :-
     basic_field_types2(Name1,Val,T,Name2,VT,TT,Orig1,Orig2).

:- block basic_field_types2(?,?,?, -,?,?,?,?).
basic_field_types2(Name1,Val,T,Name2,VT,TT,Orig1,Orig2) :-
  (Name1=Name2
   -> (basic_type(Val,VT),
       basic_field_types(T,TT,Orig1,Orig2))
    ; (add_error(kernel_objects,'Record field names do not match: ',(Name1,Name2,Orig1,Orig2)),
       fail
      )
  ).


*/

/* ------------------------- */
/*   enumerate_basic_type/2  */
/* ------------------------- */
/* a version of basic_type that enumerates */

:- assert_must_succeed(enumerate_basic_type([],set(couple(integer,integer)) )).
:- assert_must_succeed(enumerate_basic_type([([],int(2)), ([int(3)],int(4))],
                                         set(couple(set(integer),integer)) )).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                             set(couple(integer,integer)) )).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                             seq(integer) )).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                             seq(integer) )).
:- assert_must_succeed((enumerate_basic_type(X,global('Name')),
                        equal_object(X,fd(1,'Name')) )).
:- assert_must_succeed((enumerate_basic_type(X,global('Name')),
                        equal_object(X,fd(2,'Name')) )).
:- assert_must_succeed((enumerate_basic_type(X,global('Name')),
                        X==fd(2,'Name')) ).
:- assert_must_succeed((enumerate_basic_type(X,record([field(a,global('Name'))])),
                        equal_object(X,rec([field(a,fd(1,'Name'))])) )).
:- assert_must_succeed((enumerate_basic_type(X,record([field(a,integer),field(b,global('Name'))])),
                        equal_object(X,rec([field(a,int(1)),field(b,fd(1,'Name'))])) )).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc1,[case(a,constant([a])),case(b,integer)]),
			kernel_freetypes:set_freetype_depth(2),
			enumerate_basic_type(X,freetype(selfc1)),equal_object(X,freeval(selfc1,a,term(a))),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc5,[case(a,constant([a])),case(b,integer)]),
			kernel_freetypes:set_freetype_depth(2),
			enumerate_basic_type(X,freetype(selfc5)),equal_object(X,freeval(selfc5,b,int(1))),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc7,[case(nil,constant([nil])),case(node,couple(freetype(selfc7),freetype(selfc7)))]),
			kernel_freetypes:set_freetype_depth(3),
			findall(X,enumerate_basic_type(X,freetype(selfc7)),Solutions),
			lists:length(Solutions,5),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc2,[case(a,constant([a])),case(b,freetype(selfc3))]),
			kernel_freetypes:add_freetype(selfc3,[case(c,constant([c])),case(d,freetype(selfc2))]),
			kernel_freetypes:set_freetype_depth(4),
			enumerate_basic_type(X,freetype(selfc2)),
			equal_object(X,freeval(selfc2,b,freeval(selfc3,d,freeval(selfc2,b,freeval(selfc3,c,term(c)))))),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(1,'Name'),fd(1,'Code'))])) ).
:- assert_must_succeed((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(2,'Name'),fd(1,'Code')), (fd(1,'Name'),fd(2,'Code'))])) ).
:- assert_must_succeed((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(1,'Name'),fd(2,'Code')), (fd(2,'Name'),fd(1,'Code'))])) ).
:- assert_must_succeed_any((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(1,'Name'),fd(2,'Code')), (fd(2,'Name'),fd(1,'Code'))])) ).
:- assert_must_succeed(enumerate_basic_type([(int(2),(int(1),int(2))),
                                                  (int(1),(int(3),int(4)))],
                                             set(couple(integer,couple(integer,integer))) )).
:- assert_must_succeed(enumerate_basic_type([(int(2),(int(1),int(2))),
                                                  (int(55),(int(3),int(4)))],
                                             set(couple(integer,couple(integer,integer))) )).
:- assert_must_succeed(enumerate_basic_type([abort('err')],set(abort))).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                               set(couple(integer,integer)))).
%:- assert_must_fail(enumerate_basic_type([weird(_)],_T)).
%:- assert_must_fail(enumerate_basic_type([(int(2),(term(a),term(2))),
%                                                  (int(3),(int(3),term(4)))],
%                                             set(couple(integer,couple(integer,constant))) )).
%:- assert_must_fail(enumerate_basic_type([(int(1),(int(1),term(2))),
%                                                  (int(2),(int(3),int(4)))],
%                                             set(couple(integer,couple(integer,constant))) )).

:- assert_must_succeed_multiple(enumerate_basic_type([(int(2),fd(_A,'Name')),(int(3),fd(_B,'Name')),
           (int(4),fd(_C,'Name')),(int(5),fd(_D,'Name')),(int(6),fd(_E,'Name')),(int(7),fd(_F,'Name')),
           (int(8),fd(_G,'Name')),(int(9),fd(_H,'Name')),(int(10),fd(_I,'Name')),
           (int(11),fd(_,'Name')),(int(12),fd(_,'Name')),(int(13),fd(_,'Name')),
           (int(14),fd(_,'Name'))],set(couple(integer,global('Name'))))).

:- assert_must_fail(( findall(XX,enumerate_basic_type(XX, set(set(global('Code')))) ,S), lists:member(X,S), remove(S,X,R), lists:member(X2,R), equal_object(X,X2)  )).   

:- assert_must_succeed((  enumerate_basic_type(global_set('Code'),
                   set(global('Code'))) )).

:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([(fd(1,'Name'),fd(2,'Code')), (fd(2,'Name'),fd(1,'Code'))],set(couple(global('Name'),global('Code')))))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([(fd(1,'Name'),pred_true), (fd(2,'Name'),pred_false), (fd(2,'Name'),pred_true)],set(couple(global('Name'),boolean))))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([pred_true,pred_false],set(boolean)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([[],[pred_true,pred_false]],set(set(boolean))))).

:- assert_pre(kernel_objects:enumerate_basic_type(Obj,Type),
             (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:enumerate_basic_type(Obj,_), (type_check(Obj,bsets_object),ground_check(Obj))).


:- block enumerate_basic_type(?,-).
enumerate_basic_type(Obj,Type) :- %print_message(call_enumerate_basic_type(Obj,Type)),
   %enumerate_basic_type2(Obj,Type).
   enumerate_basic_type3(Type,Obj,basic).
   %(ground(Obj) -> true ; enumerate_basic_type3(Type,Obj,basic)).


:- block enumerate_type(?,-,?). % last argument: basic or tight
enumerate_type(Obj,Type,Tight) :- %print_message(call_enumerate_basic_type(Obj,Type)),
   %enumerate_basic_type2(Obj,Type).
   enumerate_basic_type3(Type,Obj,Tight).
   
%enumerate_basic_type2(X,Type) :-
%  (ground(X) -> (basic_type(X,Type) -> true 
%                    ; add_internal_error('Type error: ',enumerate_basic_type2(X,Type)))
%             ; enumerate_basic_type3(Type,X)).

enumerate_basic_type3(global(T),R,_Tight) :- enum_global_type(R,T).
enumerate_basic_type3(set(X),Set,Tight) :-
     enumerate_basic_type_set(Set,X,Tight).
enumerate_basic_type3(seq(SeqRanType),Seq,Tight) :- 
     (Tight = tight -> enumerate_seq_type(Seq,SeqRanType)
       ; enumerate_basic_type3(set(couple(integer,SeqRanType)),Seq,basic)).
enumerate_basic_type3(couple(XT,YT),(X,Y),Tight) :-
     enumerate_type(X,XT,Tight),enumerate_type(Y,YT,Tight).
enumerate_basic_type3(boolean,pred_true /* bool_true */,_Tight).
enumerate_basic_type3(boolean,pred_false /* bool_false */,_Tight).
enumerate_basic_type3(string,string(S),_Tight) :- enumerate_string(S).
enumerate_basic_type3(constant,term(V),_Tight) :- (var(V) -> V=any  ; true).
enumerate_basic_type3(constant([V]),term(V),_Tight).
enumerate_basic_type3(record(FT),rec(F),Tight) :- enumerate_basic_field_types(F,FT,Tight).
enumerate_basic_type3(freetype(Id),freeval(Id,C,Value),Tight) :-
	 enumerate_freetype(Tight,freeval(Id,C,Value),freetype(Id)).
enumerate_basic_type3(freetype(Id,Depth),freeval(Id,C,Value),Tight) :-
	 enumerate_freetype(Tight,freeval(Id,C,Value),freetype(Id,Depth)).
enumerate_basic_type3(integer,int(N),Tight) :-
   (nonvar(N) -> (number(N) -> true ; add_internal_error('Illegal value: ',enumerate_basic_type3(integer,int(N),Tight)))
     ; enumerate_int(N)).
enumerate_basic_type3(abort,abort(V),_Tight) :- (var(V) -> V=any ; true).
enumerate_basic_type3(any,Obj,_Tight) :- enumerate_any(Obj).


max_cardinality_string(inf). % was 2
all_strings(AS) :- findall(string(S),enumerate_string(S),AS).
:- use_module(btypechecker,[machine_string/1]).
enumerate_string(S) :- atomic(S),!.
enumerate_string(S) :- %print('### WARNING, Enumerating STRING'),nl,
  % frozen(S,Goal), print(enum(S,Goal)),nl, 
  %  MAYBE TO DO: we could check if prolog:dif(S,'"STR1"') are in frozen Goal and then enumerate more?
  % if we do this we need to adapt dont_expand(global('STRING')) :- ... further below
  enum_warning('STRING',inf,'"STR1","STR2",...'),
  (S = '"STR1"' ; S = '"STR2"' ;
  btypechecker:machine_string(S)).

is_string(string(_),_WF).
is_not_string(X) :- top_level_dif(X,string).

:- block enumerate_any(-).
enumerate_any(fd(X,T)) :- !,when(ground(T),enum_global_type(fd(X,T),T)).
enumerate_any(int(N)) :- !,enumerate_basic_type3(integer,int(N),basic).
enumerate_any(term(X)) :- !, print_message(could_not_enumerate_term(X)).
enumerate_any(string(S)) :- !, enumerate_string(S).
enumerate_any(pred_true /* bool_true */) :- !.
enumerate_any(pred_false /* bool_false */) :- !.
enumerate_any([]) :- !.
enumerate_any([H|T]) :- !, enumerate_any(H), enumerate_any(T).
enumerate_any((H,T)) :- !, enumerate_any(H), enumerate_any(T).
enumerate_any(T) :- print_message(could_not_enumerate(T)).


:- use_module(preferences,[preference/2]).
:- use_module(library(clpfd),[labeling/2]). %, indomain/1]).
enumerate_int(N) :- clpfd_domain(N,FDLow,FDUp),  % print(enum(N,FDLow,FDUp)),nl,
   (finite_domain(FDLow,FDUp)
     ->  labeling([],[N]) % indomain(N) % enumerate_int1(N,FDLow,FDUp) % Note:  CLP(FD) labeling  does not necessarily try all values in range (disjunctive domains)
     ; make_domain_finite(FDLow,FDUp,Min,Max),
       enum_warning('INTEGER',FDLow:FDUp,Min:Max),
       enumerate_int1(N,Min,Max)
   ).
   
enum_warning(TYPE,RANGE,RESTRICTED_RANGE) :- %trace,
   (add_new_event_in_error_scope(enumeration_warning(enumerating,TYPE,RANGE,RESTRICTED_RANGE))
     -> print('### Warning: enumerating '), print(TYPE), print(' : '),
        print(RANGE), print(' ---> '),  print(RESTRICTED_RANGE),nl %% ,trace
      ;  true).
:- block finite_warning(-,?,?,?,?).
finite_warning(_,Par,Types,Body,Source) :-
   (add_new_event_in_error_scope(enumeration_warning(assuming_finite_closure,Par,Types,finite))
     -> print('### Warning: assuming set comprehension to be finite: '), translate:print_bvalue(closure(Par,Types,Body)),nl,
        print('###  Source: '), print(Source),nl
      ;  true).
     
:- block enumerate_natural(-,?,-).
enumerate_natural(N,From,_) :- nonvar(N) -> true ; enumerate_natural(N,From).
enumerate_natural(N,From) :- clpfd_domain(N,FDLow,FDUp),   % print(enumerate_nat(N,From,FDLow,FDUp)),nl,
   (finite_domain(FDLow,FDUp)
     ->  labeling([],[N]) % indomain(N) % enumerate_int1(N,FDLow,FDUp) % Note:  CLP(FD) labeling  does not necessarily try all values in range (disjunctive domains)
     ;    ((number(FDLow),FDLow>=From) -> FDLow1=FDLow ; FDLow1=From),
         make_domain_finite(FDLow1,FDUp,Min,Max),
         enum_warning('NATURAL(1)',FDLow1:FDUp,Min:Max),
         enumerate_int2(N,Min,Max)
   ).

% assumes one of FDLow and FDUp is not a number
make_domain_finite(FDLow,_FDUp,Min,Max) :- number(FDLow),!,Min=FDLow,
   preferences:preference(maxint,MaxInt),
   (MaxInt>=FDLow -> Max=MaxInt ; Max=FDLow). % ensure that we try at least one number
make_domain_finite(_FDLow,FDUp,Min,Max) :- number(FDUp),!,Max=FDUp,
   preferences:preference(minint,MinInt),
   (MinInt=<FDUp -> Min=MinInt ; Min=FDUp).
make_domain_finite(_FDLow,_FDUp,Min,Max) :-  
  ((preferences:preference(maxint,Max),
    preferences:get_preference(minint,Min))->true). % ensure that we try at least one number

enumerate_int1(N,Min,Max) :-
     (Min<0 /* enumerate positive numbers first; many specs only use NAT/NATURAL */
         ->  (enumerate_int2(N,0,Max) ; enumerate_int2(N,Min,-1))
          ;  enumerate_int2(N,Min,Max)
       ).
enumerate_int(X,Low,Up) :- get_int_domain(X,Low,Up,RL,RU),
   %% print(enumerate_int(X,Low,Up, RL,RU)),nl, %%
   enumerate_int2(X,RL,RU).

get_int_domain(X,Low,Up,RL,RU) :- clpfd_domain(X,FDLow,FDUp),
    fd_max(FDLow,Low,RL),fd_min(FDUp,Up,RU).

finite_domain(Low,Up) :- \+ infinite_domain(Low,Up).
infinite_domain(inf,_) :- !.
infinite_domain(_,sup).

fd_max(inf,L,R) :- !,R=L.
fd_max(FDX,Y,R) :- ((nonvar(FDX),nonvar(Y),FDX>Y) -> R=FDX ; R=Y).
fd_min(sup,L,R) :- !,R=L.
fd_min(FDX,Y,R) :- ((nonvar(FDX),nonvar(Y),FDX<Y) -> R=FDX ; R=Y).


enumerate_int2(N,X,Y) :- X=<Y,
  (N=X ; X1 is X+1, enumerate_int2(N,X1,Y)).
      
enumerate_basic_type_set(X,Type,Tight) :- var(X),!,
   max_cardinality_with_check(Type,Card),
   enumerate_basic_type_set2(X,[],Card,Type,none,Tight).
enumerate_basic_type_set([],_,_) :- !.
enumerate_basic_type_set(avl_set(_),_,_) :- !.
enumerate_basic_type_set(freetype(_),_,_) :- !.
enumerate_basic_type_set(global_set(GS),Type,_Tight) :- !,
     (Type = global(GT)
      -> (GS = GT -> true
          ; nonvar(GS), add_error_and_fail(enumerate_basic_type_set,'Type error in global set: ',GS:GT))
      ;  Type = integer,integer_global_set(GS)).
enumerate_basic_type_set(closure(_Parameters, _ParameterTypes, Body),_Type,_Tight) :- !,
      (ground(Body) -> true
        ;  print('### Enumerating non-ground closure-body: '), print(Body),nl,
           term_variables(Body,Vars), print('### Variables: '), print(Vars),nl,%trace,
           enumerate_values_inside_expression(Body)
      ).


enumerate_basic_type_set([H|T],Type,Tight) :- !,
   % collect bound elements; avoid enumerating initial elements with elements that already appear later
   collect_bound_elements([H|T], SoFar,Unbound,Closed),
   (Closed=false -> max_cardinality_with_check(Type,Card)
     ; Card = Closed),
     % print(enum(Card,Unbound,SoFar,[H|T],Closed)),nl,
    enumerate_basic_type_set2(Unbound,SoFar,Card,Type,none,Tight).
%enumerate_basic_type_set([H|T],Type,Tight) :- !,
%  (is_list_skeleton([H|T],Card) -> true
%     ;  max_cardinality_with_check(Type,Card)
%  ),
%   enumerate_basic_type_set2([H|T],[],Card,Type,none,Tight).
enumerate_basic_type_set(S,Type,Tight) :-
  add_internal_error('Illegal set: ',enumerate_basic_type_set(S,Type,Tight)).

enumerate_basic_type_set2(HT,ElementsSoFar,_Card,_Type,_Last,_Tight) :- nonvar(HT),
   is_custom_explicit_set(HT,enumerate_basic_type),!,
   disjoint_sets(HT,ElementsSoFar). % I am not sure this is necessary; probably other constraints already ensure this holds
enumerate_basic_type_set2([],_,_,_,_,_Tight).
enumerate_basic_type_set2(HT,ElementsSoFar,Card,Type,Last,Tight) :-  positive_card(Card),
   %debug:trace_point(enum(HT,ElementsSoFar,Card,Type,Last,Tight)),
   (var(HT) -> HT=[H|T], NewLast=NormH /* the enumerator has completely determined H */
    % Note: HT=[H|T] may wake up co-routines and then attach infos to H; but these should hold indpendently for all elements
    ; HT=[H|T],
      (unbound_value(H) -> NewLast=NormH /* the enumerator has completely determined H */
           ;  NewLast=Last)  /* H was not freely chosen by the enumerator */
   ),
   not_element_of(H,ElementsSoFar), % this is only needed for elements generated by the enumerator itself
   enumerate_type(H,Type,Tight),
   % TO DO: extract normal form from add_new_element
   val_greater_than(H,NormH,Last),
   %(val_greater_than(H,NormH,Last) -> print(ok(H)),nl ; print(pruned(H,NormH,Last)),nl,fail),
   %(ground(NormH) -> true ; print(not_ground(NormH)),nl),
   C1 is Card-1,
   add_new_element(NormH,ElementsSoFar,SoFar2), % we could use avl_store(NormH,A,true,A2) if ElementsSoFar=avl_set(A)
   %debug:trace_point(add_new_element(NormH,ElementsSoFar,SoFar2)),
   enumerate_basic_type_set2(T,SoFar2,C1,Type,NewLast,Tight).

:- assert_must_succeed((collect_bound_elements([int(1),int(2),int(4),X,int(5)|T],_,U,C),U==[X|T],C==false)).
:- assert_must_succeed((collect_bound_elements([int(1),int(2),int(4),X,int(5)],_,U,C),U==[X],C==1)).
:- assert_must_succeed(exhaustive_kernel_succeed_check(collect_bound_elements([int(1),int(2),int(4),int(5)],_,_,_))).

% collect the bound and unbound elements in a list; also return if the list is closed (then return length) or return false
collect_bound_elements(T, SoFar,Unbound,Closed) :- var(T),!, SoFar=[],Unbound=T,Closed=false.
collect_bound_elements([],[],[],0).
collect_bound_elements(avl_set(A),avl_set(A),[],0).
collect_bound_elements(global_set(GS),SoFar,Unbound,Closed) :- expand_custom_set(global_set(GS),ES),
   collect_bound_elements(ES,SoFar,Unbound,Closed).
collect_bound_elements(freetype(FS),SoFar,Unbound,Closed) :- expand_custom_set(freetype(FS),ES),
   collect_bound_elements(ES,SoFar,Unbound,Closed).
collect_bound_elements(closure(P,T,B),SoFar,Unbound,Closed) :- expand_custom_set(closure(P,T,B),ES),
   collect_bound_elements(ES,SoFar,Unbound,Closed).
collect_bound_elements([H|T],SoFar,Unbound,Closed) :-
   collect_bound_elements(T,TSoFar,TUnbound,TClosed),
   (ground(H) -> add_new_element(H,TSoFar,SoFar), Unbound=TUnbound, TClosed=Closed
     ;  SoFar = TSoFar, Unbound = [H|TUnbound],
       (TClosed=false -> Closed=false ; Closed is TClosed+1)
    ).

% use_module(kernel_objects),findall(A,enumerate_tight_type(A,set(set(boolean))),B),length(B,C).
% use_module(kernel_objects),findall(A,enumerate_tight_type(A,set(set(global('Name')))),B),length(B,C).
% use_module(kernel_objects),findall(A,enumerate_tight_type(A,set(global('Name'))),B),length(B,C).

% perform order checking on terms, normalising them first
% val_greater_than(A,NormA,NormB)
val_greater_than(A,NormA,NormB) :- !,
   (custom_explicit_sets:convert_to_avl_inside_set(A,NormA)
     -> (NormB==none -> true ; NormA @> NormB)
     ;  add_internal_error('Call failed: ',custom_explicit_sets:convert_to_avl_inside_set(A,NormA)),
        NormA = A).

positive_card(inf) :- !, print('$').
positive_card(C) :- (number(C) -> C>0
                               ; (add_error(positive_card,'Not number: ',C),fail)).

cardval_greater_than(_,inf) :- !,fail.
cardval_greater_than(inf,_) :- !.
cardval_greater_than(X,Y) :- X>Y.


:- block enumerate_basic_field_types(?,-,?).
enumerate_basic_field_types([],[],_Tight).
enumerate_basic_field_types(Fields,[field(Name,VT)|TT],Tight) :-
  enumerate_basic_field_types2(Fields,Name,VT,TT,Tight).

:- block enumerate_basic_field_types2(?,-,?,?,?).
enumerate_basic_field_types2([field(Name,V)|T], Name,VT,TT,Tight) :-
  enumerate_type(V,VT,Tight),
  enumerate_basic_field_types(T,TT,Tight).
  

:- block all_objects_of_type(-,?).
all_objects_of_type(Type,Res) :-
  findall(O,enumerate_basic_type(O,Type),Res).

is_list_skeleton(X,_) :- var(X),!,fail.
is_list_skeleton([],0).
is_list_skeleton([_|T],C1) :- is_list_skeleton(T,C), C1 is C+1.

list_length(X,L,Type) :- var(X),!,L=0, Type=open.
list_length([],0,closed).
list_length([_|T],C1,Type) :- list_length(T,C,Type), C1 is C+1.


:- assert_must_succeed((max_cardinality(set(couple(global('Name'),global('Code'))),64))).
:- assert_must_succeed((max_cardinality(set(set(set(couple(global('Name'),global('Code'))))),_))).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc4,[case(a,boolean),case(b,couple(boolean,boolean))]),
			max_cardinality(freetype(selfc4),6),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc6,[case(a,boolean),case(b,freetype(selfc6)),case(c,constant([c]))]),
			kernel_freetypes:set_freetype_depth(3),
			findall(X,enumerate_tight_type(X,freetype(selfc6)),Solutions),
			lists:length(Solutions,NumberOfSolutions),
			max_cardinality(freetype(selfc6),NumberOfSolutions),
			kernel_freetypes:reset_freetypes)).
 
max_cardinality_with_check(Set,CCard) :-
   (max_cardinality(Set,Card) ->
     (Card=inf
      -> debug_println(9,very_large_cardinality(Set,Card)),
         CCard = 20000000
      ;  CCard=Card,
         (Card>100 -> debug_println(9,large_cardinality(Set,Card)) ; true)  
     )
     ; print_error(failed(max_cardinality(Set,CCard))), CCard = 10
   ).
max_cardinality(global(T),Card) :- b_global_set_cardinality(T,Card).
max_cardinality(boolean,2).
max_cardinality(constant([_V]),1).
max_cardinality(any,inf) :- print_message(dont_know_card_of_any). /* what should we do here ? */
max_cardinality(constant,10) :- print_message(dont_know_card_of_constant). /* what should we do here ? */
max_cardinality(string,MC) :- max_cardinality_string(MC). % is inf now
max_cardinality(abort,1).
max_cardinality(integer,Card) :- Card=inf. %b_global_set_cardinality('INTEGER',Card).
max_cardinality(seq(X),Card) :- % Card=inf, unless a freetype can be of cardinality 0
   max_cardinality(set(couple(integer,X)),Card).
max_cardinality(couple(X,Y),Card) :-
   max_cardinality(X,CX), max_cardinality(Y,CY), safe_mul(CX,CY,Card).
max_cardinality(record([]),1).
max_cardinality(record([field(_,T1)|RF]),Card) :- 
   max_cardinality(record(RF),RC),
   max_cardinality(T1,C1),
   safe_mul(C1,RC,Card).
max_cardinality(set(X),Card) :- max_cardinality(X,CX),
  safe_pow2(CX,Card).
  % RealCard is 2**CX, (RealCard is inf -> Card is inf ; Card is integer(RealCard)).
max_cardinality(freetype(Id),Card) :- max_cardinality_freetype(freetype(Id),Card).
max_cardinality(freetype(Id,Depth),Card) :- max_cardinality_freetype(freetype(Id,Depth),Card).

:- assert_must_succeed((safe_pow2(3,R),R==8)).
:- assert_must_succeed((safe_pow2(inf,R),R==inf)).
:- assert_must_succeed((safe_pow2(3072,R),R==inf)).
:- assert_must_succeed((safe_pow2(18446744073709551616,R),R==inf)).
:- assert_must_succeed((safe_pow2(500,X), safe_pow2(501,X2), X2 is 2*X)).
% :- assert_must_succeed((kernel_objects:safe_pow2(1022,X), kernel_objects:safe_pow2(1023,X2), X2 is 2*X)).

safe_pow2(Exp,Res) :- (Exp==inf -> Res=inf
   ; Exp>1023 -> Res=inf  /* the limit where SICStus 4.2.1 reported inf; 4.2.3 goes further but uses a huge amount of memory */
   ; is_overflowcheck(Res, integer(2**Exp))).
% this seems to be either precise or give inf (at 1023 on 64 bit system)

:- assert_must_succeed((safe_pown(3,2,R),R==9)).
:- assert_must_succeed((safe_pown(2,3072,R),R==inf)).
:- assert_must_succeed((safe_pown(3,647,R),R==inf)).
:- assert_must_succeed((safe_pown(2,500,X), safe_pown(2,501,X2), X2 is 2*X)).
:- assert_must_succeed((safe_pown(2,500,X), safe_pow2(500,X))).
:- assert_must_succeed((kernel_objects:safe_pown(2,1022,X), kernel_objects:safe_pown(2,1023,X2), X2 is 2*X)).
:- assert_must_succeed((kernel_objects:safe_pown(3,500,X), kernel_objects:safe_pown(3,501,X3), X3 is 3*X)).
safe_pown(Base,Exp,Res) :-
  (Base=inf -> (Exp=0 -> Res=1 ; Res=inf)
    ; Exp=inf -> (Base=0 -> Res=0 ; Base=1 -> Res=1 ; Res=inf)
    ; infinite_pown(Base,Exp) -> Res=inf /* SICStus 4.2.1 reported inf; 4.2.3 goes further but uses a huge amount of memory */
    ; %is_overflowcheck(ResFloat, integer(Base**Exp)),
      overflow_float_pown(Base,Exp,ResFloat), % get result from float calculation converted to int
      (ResFloat=inf -> Res=inf
       ; Base = 2 -> Res = ResFloat % result is precise if not inf (check ??)
       ; (ResFloat>=0,ResFloat < 4503599627370496) -> Res = ResFloat % floating point has 52 bits precision (4503599627370496)
         % Note: this is true for both 32-bit and 64-bit SICStus
       ; safe_int_power(Base,Exp,Res)  % compute using Prolog arbitrary precision integers as result using float was probably imprecise
       % on SICStus 4.2.3 we could use Res is integer(Base**Exp) ??
      )
  ).
  
/* SICStus 4.2.1 reported inf; 4.2.3 goes further but uses a huge amount of memory */
infinite_pown(Base,Exp) :- Exp > 1023, !, Base >= 2. 
%infinite_pown(Base,Exp) :- Exp > 646, !, Base >= 3. % not really necessary any more when using overflow_float_pown
%infinite_pown(Base,Exp) :- Exp > 511, !, Base >= 4.
%infinite_pown(Base,Exp) :- Exp > 441, !, Base >= 5.

:- assert_must_succeed((safe_mul(3,2,R),R==6)).
:- assert_must_succeed((safe_mul(inf,2,R),R==inf)).
:- assert_must_succeed((safe_mul(2,inf,R),R==inf)).
:- assert_must_succeed((safe_mul(0,inf,R),R==0)).
:- assert_must_succeed((safe_mul(inf,0,R),R==0)).
% safe_multiplication for positive numbers
safe_mul(0,_,R) :- !, R=0. % true for cartesian product: card({}*INTEGER)=0
safe_mul(_,0,R) :- !, R=0. % ditto
safe_mul(inf,_,R) :- !, R=inf.
safe_mul(_,inf,R) :- !, R=inf.
safe_mul(X,Y,R) :- is_overflowcheck(R,X*Y).

safe_add(inf,_,R) :- !, R=inf.
safe_add(_,inf,R) :- !, R=inf.
safe_add(X,Y,R) :- is_overflowcheck(R,X+Y).

   
/* is with overflow check */
is_overflowcheck(Var,Expr) :-  %print(is_with_overflow(Var,Expr)),nl,
  (Expr=inf -> Var=inf
    ; on_exception(error(_,_), Var is Expr, Var=inf)).

overflow_float_pown(Base,Exp,Res) :-
  on_exception(error(_,_),
     (R1 is Base**Exp,  /* separate into two steps; SICStus 4.2.3 otherwise seems to use integer exponentation */
      Res is integer(R1)), Res=inf).

/*
% this code below would sometimes fail in spld generated code:
is_overflowcheck_old(Var,Expr) :- print(is_with_overflow(Var,Expr)),nl,
  (Expr=inf -> Var=inf
    ;  on_exception(error(type_error(evaluable,inf/0),_),
        on_exception(error(evaluation_error(float_overflow),_),
                     Var is Expr,
                     (print(float_overflow),nl,Var = inf)),
                     (print(evaluable_inf),nl,Var=inf))).
% catches any exception (e.g. also representation_error when converting inf to integer) 
% could be unsafe with timeout !! 
is_errc(Var,Expr) :-
  (Expr=inf -> Var=inf
            ;  safe_on_exception(_E, Var is Expr,  Var = inf)).
*/

/* ---------------------------- */


/* use a cleverer, better enumeration than enumerate_basic_type */
/* can only be used in certain circumstances: operation preconditions,
  properties,... but not for VARIABLES as there is no guarantee that
  something declared as a sequence will actually turn out to be a sequence */
  
:- assert_pre(kernel_objects:enumerate_tight_type(Obj,Type),
             (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:enumerate_tight_type(Obj,_), (type_check(Obj,bsets_object),ground_check(Obj))).
:- assert_pre(kernel_objects:enumerate_tight_type(Obj,Type,_),
             (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:enumerate_tight_type(Obj,_,_), (type_check(Obj,bsets_object),ground_check(Obj))).
  
:- assert_must_succeed(enumerate_tight_type([(int(1),int(2)),(int(2),int(4))],
                                             seq(integer) )).
:- assert_must_succeed(enumerate_tight_type([(int(1),int(2))],seq(integer) )).
:- assert_must_succeed(enumerate_tight_type([],seq(integer) )).
:- assert_must_succeed((enumerate_tight_type(X,record([field(a,integer),field(b,global('Name'))])),
                        equal_object(X,rec([field(a,int(1)),field(b,fd(1,'Name'))])) )).
:- assert_must_fail(enumerate_tight_type([(int(1),int(2)),(int(3),int(_))],
                                             seq(integer) )).
:- assert_must_fail(enumerate_tight_type([(int(3),int(_))],seq(integer) )).
:- assert_must_succeed((bsets_clp:is_sequence(X,global_set('Name')),
                        enumerate_tight_type(X,seq(global('Name')) ),
                        X = [(int(1),fd(2,'Name'))]     )).
:- assert_must_succeed(( enumerate_tight_type(XX, record([field(balance,integer),field(name,global('Name'))])) ,
  XX = rec([field(balance,int(1)),field(name,fd(3,'Name'))]) )).
:- assert_must_succeed(( enumerate_tight_type(XX, set(record([field(balance,global('Name')),field(name,global('Name'))]))) ,                /* STILL TAKES VERY LONG !! */
  XX = [rec([field(balance,fd(3,'Name')),field(name,fd(3,'Name'))])] )). 
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(balance,global('Name')),field(name,global('Name'))]))) ,S),
  length(S,Len), Len = 512 )).   
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(name,global('Code'))]))) ,S),
  length(S,Len), Len = 4 )).   
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(fname,global('Code')),field(name,global('Code'))]))) ,S),
  length(S,Len), Len = 16 )).  
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(fname,global('Code')),field(name,global('Name'))]))) ,S),
  length(S,Len), Len = 64 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(global('Name'))) ,S),
  length(S,Len), Len = 8 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(boolean))) ,S),
  length(S,Len), Len = 16 )).  
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(global('Name')))) ,S),
  length(S,Len), Len = 256 )).    
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(global('Code')))) ,S),
  length(S,Len), Len = 16 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(boolean))) ,S),
  length(S,Len), Len = 16 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(couple(global('Code'),global('Name')))) ,S),
  length(S,Len), Len = 64 )).  
%:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(couple(global('Code'),integer))) ,S),
%  length(S,Len), Len = 64 )).  
:- assert_must_succeed(( enumerate_tight_type(XX, set(record([field(balance,integer)]))) ,
  XX = [rec([field(balance,int(1))])] )).
:- assert_must_succeed(( enumerate_tight_type(global_set('Code'),set(global('Code'))) )).

:- block enumerate_tight_type(?,-,?).
enumerate_tight_type(Obj,Type,_VarID) :- %enumerate_tight_type2(Type,Obj).
  (ground(Obj) -> %print(value_determined(VarID,Obj,Type)),nl,
                  true
   ;  % print(enumerating(_VarID,Obj,Type)),nl, %%
    enumerate_basic_type3(Type,Obj,tight)
      %,print(try_enum(VarID,Obj)),nl
  ).
 
:- block enumerate_tight_type(?,-).
enumerate_tight_type(Obj,Type) :- %enumerate_tight_type2(Type,Obj).
  %%print_message(call_enumerate_tight_type(Obj,Type)),
  (ground(Obj) -> true ;  % print(enumerate_tight_type(Obj,Type)),nl,
    enumerate_basic_type3(Type,Obj,tight)
  ).


/* TO DO: provide tight enumerators for nat, functions, ... ?? */



:- assert_must_succeed((X=[(int(I1),pred_true /* bool_true */),Y], dif(I1,1),
                        kernel_objects:enumerate_seq_type(X,boolean),I1==2,Y=(int(1),pred_false /* bool_false */))).

enumerate_seq_type(X,Type) :-
  list_length(X,Card,ListType),
  (ListType=closed
     -> MaxIndex=Card
     ;  b_global_set_cardinality('NAT1',NatCard),
        %print('### WARNING: expanding seq up to length '), print(Card),nl
        (NatCard<Card -> MaxIndex=Card ; MaxIndex=NatCard),
        enum_warning('seq (length)',inf,MaxIndex)
  ),
   Card>=0,
   enumerate_seq(X,range(1,MaxIndex),MaxIndex,Type).
   
enumerate_seq([],_,_,_).
enumerate_seq(Seq,Indexes,Card,Type) :- 
   (unbound_variable_check(Seq) -> positive_card(Card),VarSeq=true ; VarSeq=false),
   Seq = [El|TSeq],
   (unbound_variable_check(El) -> VarEl=true ; VarEl=VarSeq),
   El = (int(Index),Element),
   (VarEl=true -> get_next_index(Indexes,Index,RemIndexes) 
     ; number(Index) -> remove_index_ground(Indexes,Index,RemIndexes)
      ; remove_index(Indexes,Index,RemIndexes)),
   enumerate_tight_type(Element,Type),
   C1 is Card-1,
   enumerate_seq(TSeq,RemIndexes,C1,Type).

get_next_index([Index1|RestIndexes],Index1,RestIndexes).
get_next_index(range(I1,I2),I1,Res) :- 
   I11 is I1+1, (I11>I2 -> Res=[] ; Res=range(I11,I2)).

remove_index_ground(Indexes,X,Res) :- get_next_index(Indexes,H,T),
   (X=H -> Res=T ; Res=[H|R2], remove_index_ground(T,X,R2)).
   
remove_index(Indexes,X,Res) :- get_next_index(Indexes,H,T),
   (X=H,Res=T ; X\==H, Res=[H|R2], remove_index(T,X,R2)).



/* a few more unit tests: */

:- assert_must_succeed(( findall(X,enumerate_type(X,set(couple(boolean,boolean)),tight) ,L), length(L,16) )).
:- assert_must_succeed(( findall(X,enumerate_type(X,set(couple(boolean,boolean)),basic) ,L), length(L,16) )).

:- assert_must_succeed(( enumerate_tight_type(
  [rec([field(balance,int(0)),field(name,fd(2,'Name'))])],[
   rec([field(balance,int(1)),field(name,fd(3,'Name'))]),
   rec([field(balance,int(1)),field(name,fd(2,'Name'))]),
   rec([field(balance,int(0)),field(name,fd(1,'Name'))]),
   rec([field(balance,int(-1)),field(name,fd(1,'Name'))])],
   set(record([field(balance,integer),field(name,global('Name'))]))) )).
:- assert_must_succeed(( enumerate_tight_type([
   rec([field(balance,int(1)),field(name,fd(2,'Name'))]),
   rec([field(balance,int(1)),field(name,fd(1,'Name'))]),
   rec([field(balance,int(0)),field(name,fd(1,'Name'))]),
   rec([field(balance,int(-1)),field(name,fd(1,'Name'))])|X],
   set(record([field(balance,integer),field(name,global('Name'))]))) , 
   X = [rec([field(balance,int(1)),field(name,fd(3,'Name'))])] )).

:- assert_must_succeed((not_element_of(X,[(pred_true /* bool_true */,pred_true /* bool_true */),
 (pred_true /* bool_true */,pred_false /* bool_false */),(pred_false /* bool_false */,pred_false /* bool_false */)]),
  enumerate_tight_type(X,couple(boolean,boolean)))).
  
:- assert_must_succeed(( not_equal_object(X,(pred_true /* bool_true */,pred_false /* bool_false */)),
                         not_equal_object(X,(pred_false /* bool_false */,pred_false /* bool_false */)),
                         not_equal_object(X,(pred_true /* bool_true */,pred_true /* bool_true */)),
                         enumerate_tight_type(X,couple(boolean,boolean)))).

:- assert_must_succeed(( X = [fd(3,'Name')|T],enumerate_tight_type(X,set(global('Name'))),
  T == [fd(1,'Name'),fd(2,'Name')] )).
  

   
unbound_value(V) :-
   (var(V) -> unbound_variable_check(V)
           ;  V = (V1,W1),unbound_value(V1), unbound_value(W1)).

enumerate_values_inside_expression(b(value(X),Type,_)) :- !, 
       (ground(Type) -> enumerate_basic_type(X,Type)
         ; add_internal_error('Value type not ground: ',type(Type):value(X))).
enumerate_values_inside_expression(b(Expr,_,_)) :- 
       nonvar(Expr), !, enumerate_expr(Expr).
enumerate_values_inside_expression(X) :-
   add_internal_error('Unexpected B expression: ',enumerate_values_inside_expression(X)).

:- block enumerate_expr(-).
enumerate_expr(identifier(_)) :- !.
enumerate_expr(integer(I)) :- !, enumerate_basic_type(int(I),integer).
enumerate_expr(comprehension_set(_V,B)) :- !, enumerate_values_inside_expression(B).
enumerate_expr(exists(_V,B)) :- !, enumerate_values_inside_expression(B).
enumerate_expr(Expr) :- Expr =..[_Op|Args], %print(enumerate_expr(_Op,Args)),nl,
       l_enumerate_values_inside_expression(Args).
    
:- block l_enumerate_values_inside_expression(-).
l_enumerate_values_inside_expression([]).
l_enumerate_values_inside_expression([H|T]) :-
  enumerate_values_inside_expression(H),
  l_enumerate_values_inside_expression(T).
 

/* --------------- */
/* top_level_dif/2 */
/* --------------- */
/* checks whether two terms have a different top-level functor */

:- assert_must_succeed(top_level_dif(a,b)).
:- assert_must_succeed(top_level_dif(f(_X),g(_Z))).
:- assert_must_fail(top_level_dif(f(a),f(_Z))).
:- assert_must_fail(top_level_dif(f(a),f(b))).

:- block top_level_dif(-,?),top_level_dif(?,-).
top_level_dif(X,Y) :-
    functor(X,FX,_),functor(Y,FY,_), FX\=FY.   /* check arities ? */

:- assert_must_succeed(not_top_level_functor(a,b)).
:- assert_must_succeed(not_top_level_functor(f(_X),g)).
:- assert_must_fail(not_top_level_functor(f(a),f)).
:- block not_top_level_functor(-,?).
not_top_level_functor(X,Functor) :-
       ((atomic(X);compound(X))
        -> functor(X,FX,_), dif(FX,Functor)
        ;  print(weird_term(X)),nl
       ).


/* ------------------------------------------------------------------- */
/*                             EQUAL OBJECT                            */
/* ------------------------------------------------------------------- */

sample_closure(C) :- 
   construct_closure([xx],[integer],Body,C),
   Body = b(conjunct(b(conjunct(
    b(member(b(identifier(xx),integer,[]),b(integer_set('NAT'),set(identifier(xx)),[])),pred,[]),
    b(greater(b(identifier(xx),integer,[]),b(integer(0),integer,[])),pred,[])),pred,[]),
    b(less(b(identifier(xx),integer,[]),b(integer(3),integer,[])),pred,[])),pred,[]).
  
:- assert_must_succeed(equal_object([int(3),int(1)],
     closure([zz],[integer],b(member(b(identifier(zz),integer,[]),b(value([int(1),int(3)]),set(integer),[])),pred,[])))).
:- assert_must_succeed(( equal_object( (fd(1,'Name'),fd(1,'Name')) , (fd(1,'Name'),fd(1,'Name')) ) )).
:- assert_must_succeed(( equal_object( (X,Y) , (fd(2,'Name'),fd(2,'Name')) ) , X = fd(2,'Name'), Y=fd(2,'Name') )).
:- assert_must_fail(equal_object(term(a),term(b))).
:- assert_must_fail(equal_object(int(1),int(2))).
:- assert_must_fail(equal_object([term(a),term(b)],[term(a),term(c)])).
:- assert_must_fail((equal_object([(int(1),[Y])],[(int(X),[Z])]),
                    Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[int(2)])).
:- assert_must_fail(equal_object(rec([field(a,int(1))]),rec([field(a,int(2))]))).
:- assert_must_fail(equal_object(rec([field(a,int(2)),field(b,int(3))]),
                       rec([field(a,int(2)),field(b,int(4))]))).
:- assert_must_succeed(equal_object(rec([field(a,int(2))]),rec([field(a,int(2))]))).
:- assert_must_succeed(equal_object(rec([field(a,int(2)),field(b,[int(3),int(2)])]),
                                        rec([field(a,int(2)),field(b,[int(2),int(3)])]) )).
:- assert_must_succeed(equal_object([(term(a),[])],[(term(a),[])])).
:- assert_must_succeed(equal_object(_X,[int(1),int(2)])).
:- assert_must_succeed(equal_object([int(1),int(2)],_X)).
:- assert_must_succeed((equal_object([(int(1),[Y])],[(int(X),[Z])]),
                    Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[])).
:- assert_must_succeed(equal_object([int(1),int(2)],[int(2),int(1)])).
:- assert_must_succeed(equal_object(global_set('Name'),[fd(2,'Name'),fd(3,'Name'),fd(1,'Name')])).
:- assert_must_succeed(equal_object(global_set('Name'),[fd(1,'Name'),fd(3,'Name'),fd(2,'Name')])).
:- assert_must_succeed((equal_object([fd(3,'Name'),fd(2,'Name'),fd(1,'Name')],global_set('Name')))).
%:- assert_must_succeed((equal_object([fd(3,'Name'),fd(2,'Name'),fd(1,'Name')],X),X=global_set('Name'))).
:- assert_must_succeed((equal_object(Y,X),X=global_set('Name'),equal_object(Y,[fd(3,'Name'),fd(2,'Name'),fd(1,'Name')]))).
:- assert_must_succeed((equal_object(X,X),X=global_set('Name'))).
:- assert_must_succeed((equal_object(_,X),X=global_set('Name'))).
:- assert_must_succeed((equal_object(X,global_set('Name')),X=global_set('Name'))).
:- assert_must_succeed((equal_object([_A,_B],[int(2),int(1)]))).
:- assert_must_fail((equal_object(X,global_set('Code')),X=global_set('Name'))).
:- assert_must_fail((equal_object(Y,global_set('Name')),Y=[fd(3,'Name'),fd(1,'Name')])).
:- assert_must_fail((equal_object(Y,global_set('Name')),Y=[_,_])).
:- assert_must_succeed((equal_object(X,closure([xx],[integer],true)),X==closure([xx],[integer],true))).
:- assert_must_succeed((sample_closure(C), equal_object([int(1),int(2)],C))).
:- assert_must_succeed((sample_closure(C), equal_object(C,[int(1),int(2)]))).
:- assert_must_fail((sample_closure(C), equal_object(C,[int(1),int(0)]))).
:- assert_must_fail((sample_closure(C), equal_object(C,global_set('NAT')))).
:- assert_must_succeed((equal_object(freeval(selfcx,a,int(5)),freeval(selfcx,a,int(5))))).
:- assert_must_fail((equal_object([int(1),int(2),int(3)],global_set('NATURAL1')))).
:- assert_must_fail((equal_object(X,global_set('NATURAL1')),equal_object(X,[int(1),int(2),int(3)]))).
:- assert_must_fail((equal_object(X,[int(1),int(2),int(3)]),equal_object(X,global_set('NATURAL1')))).
:- assert_must_fail((equal_object(X,global_set('NATURAL')),equal_object(X,global_set('NATURAL1')))).
:- assert_must_succeed((equal_object(X,global_set('NATURAL')),equal_object(X,global_set('NATURAL')))).
:- assert_must_fail((equal_object(freeval(selfcx,a,int(5)),freeval(selfcy,a,int(5))))).
:- assert_must_fail((equal_object(freeval(selfcx,b,int(5)),freeval(selfcx,a,int(5))))).
:- assert_must_fail((equal_object(freeval(selfcx,a,int(5)),freeval(selfcx,a,int(6))))).
:- assert_must_succeed((equal_object( 
 [[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
    [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(3,'Name'),fd(2,'Name')]]
,[[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
     [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(2,'Name'),fd(3,'Name')]])
     )). 
:- assert_must_succeed(exhaustive_kernel_check( (equal_object([int(3),int(2),int(1)],[int(2)|T]), 
equal_object(T,[int(1),int(3)])))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],equal_object([int(3),int(1)],[int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],equal_object([int(3),int(4),int(1)],[int(4),int(1),int(3)]))).   
     
%:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],equal_object([int(1),int(2),int(3)],global_set('NATURAL1')))).
:- assert_must_succeed((     equal_object([int(0),int(5)|T],avl_set(node(int(1),true,1,node(int(0),true,0,empty,empty),node(int(3),true,1,empty,node(int(5),true,0,empty,empty))))), nonvar(T),equal_object(T,[int(_A),int(_B)]) )).
% NOTE: had multiple solutions; after solving Ticket #227 it no longer has :-)
:- assert_must_succeed((     equal_object([int(0),int(5)|T],avl_set(node(int(1),true,1,node(int(0),true,0,empty,empty),node(int(3),true,1,empty,node(int(5),true,0,empty,empty))))), nonvar(T),equal_object(T,[_A,_B]) )).

:- assert_must_succeed((equal_object([_X,_Y],[int(1),int(2)]))).


:- use_module(bool_pred).
/* a version of equal_object which will convert lists to avl if possible */
equal_object_optimized(V1,V2) :- ( (var(V1),var(V2)) -> V1=V2
                                                      ; equal_object_opt2(V1,V2)).
equal_object_opt2(X,Y) :-  (nonvar(X) -> equal_object_opt3(X,Y)
                                       ; equal_object_opt3(Y,X)).

equal_object_opt3(X,S2) :-  unbound_variable_check(S2),should_be_converted_to_avl(X), !,
   construct_avl(X,S2).
%equal_object_opt3([H|T],S2) :-  var(S2),ground(H),ground(T), !, construct_avl([H|T],S2).
equal_object_opt3(X,Y) :- equal_object3(X,Y).

equal_object(V1,V2) :- ( (var(V1),var(V2)) -> V1=V2
                          ; nonvar(V1) -> equal_object3(V1,V2)
                          ;  equal_object3(V2,V1)).

%%equal_object3c(X,Y) :- if(equal_object3(X,Y),true,
%%   (print_message(equal_object3_failed(X,Y)),equal_object3(X,Y),fail)). %%
equal_object3(X,Y) :- nonvar(Y),type_error(X,Y),
    add_internal_error('Internal Typing Error (please report as bug !) : ',equal_object(X,Y)),fail.
equal_object3(closure(Par,ParTypes,Clo),Y) :- var(Y),!,
     ( closure_occurs_check(Y,Par,ParTypes,Clo)
       -> print(occurs_check(Y,Par)),nl,
          expand_custom_set(closure(Par,ParTypes,Clo),Expansion),
          equal_object_optimized(Y,Expansion)
       ;  Y = closure(Par,ParTypes,Clo)).
equal_object3(closure(Parameters,PT,Cond),Y) :- 
    equal_object_custom_explicit_set(closure(Parameters,PT,Cond),Y).
%equal_object3(Obj,Y) :- is_custom_explicit_set(Obj,equal_object3_Obj),
%   equal_object_custom_explicit_set(Obj,Y). % inlined below for performance
equal_object3(global_set(X),Y) :- equal_object_custom_explicit_set(global_set(X),Y).
equal_object3(freetype(X),Y) :- equal_object_custom_explicit_set(freetype(X),Y).
equal_object3(avl_set(X),Y) :- equal_object_custom_explicit_set(avl_set(X),Y).
equal_object3(pred_true /* bool_true */,pred_true /* bool_true */).
equal_object3(pred_false /* bool_false */,pred_false /* bool_false */).
equal_object3(term(X),term(X)).
equal_object3(string(X),string(X)).
equal_object3(rec(F1),rec(F2)) :- equal_fields(F1,F2).
equal_object3(freeval(Id,C,F1),freeval(Id,C,F2)) :- equal_object(F1,F2).
equal_object3(int(X),int(X)).
equal_object3(fd(X,Type),fd(Y,Type)) :- eq_fd(X,Y).
equal_object3((X,Y),(X2,Y2)) :-  equal_object(Y,Y2), equal_object(X,X2).
equal_object3([],X) :- empty_set(X).
equal_object3([H|T],S2) :- !, equal_cons(S2,H,T).


equal_object_custom_explicit_set(Obj,Y) :- %print(eq(Obj,Y)),nl,
    (var(Y) -> Y = Obj 
     ; (is_custom_explicit_set(Y,equal_object3_Y) -> equal_explicit_sets(Obj,Y)
         ; (Y=[] -> is_empty_explicit_set(Obj) 
             ; Y=[H|T] -> equal_cons(Obj,H,T) %was expand_custom_set(Obj,Expansion), equal_sets(Expansion,Y)
             ; add_internal_error('Illegal set: ',equal_object_custom_explicit_set(Obj,Y)),fail  
           )
     )).

:- block equal_fields(-,-).
equal_fields([],[]).
equal_fields([field(Name,V1)|T1],[field(Name,V2)|T2]) :- 
  equal_object(V1,V2),
  equal_fields(T1,T2).


% is just like equal_cons, but H and T are guaranteed by the caller to be free
% this just gives one next element of the set; can be used to iterate over sets.
get_next_element(R,H,T) :- var(R),!,R=[H|T].
get_next_element([H1|T1],H,T) :- !,(H1,T1)=(H,T).
get_next_element(R,H,T) :- equal_cons(R,H,T).

equal_cons_wf(R,H,T,WF) :-
   %get_cardinality_wait_flag(R,equal_cons_wf,WF,LWF), 
   %get_binary_choice_wait_flag(equal_cons_wf,WF,LWF), %old version
   LWF = lwf_card(R,equal_cons_wf,WF), % will be instantiated by instantiate_lwf
   equal_cons_lwf(R,H,T,LWF).

% a deterministic version; will never instantiate non-deterministically:
% probably better to use equal_cons_wf if possible
%equal_cons_det(R,H,T) :-  equal_cons_lwf(R,H,T,_).

equal_cons(R,H,T) :- equal_cons_lwf(R,H,T,2).

:- block blocking_equal_cons_lwf(-,?,?,?).
blocking_equal_cons_lwf(E,H,T,LWF) :- equal_cons_lwf(E,H,T,LWF).

equal_cons_lwf(R,H,T,_) :- var(R),!,add_new_el(T,H,R).
equal_cons_lwf([HR|TR],H,T,_) :- ground(H), delete_exact_member([HR|TR],H,Rest),!,
   equal_object(Rest,T).
equal_cons_lwf([HR|TR],H,T,LWF) :- equal_cons_cons(HR,TR,H,T,LWF).
equal_cons_lwf(avl_set(AVL),H,T,LWF) :-
 (is_one_element_custom_set(avl_set(AVL),El)
  -> T=[], equal_object(El,H)
  ;  T==[] -> fail % we have a one element set and AVL is not
  ;  element_can_be_added_or_removed_to_avl(H) ->
           remove_element_from_explicit_set(avl_set(AVL),H,AR), %print(removed(H)),nl,
           equal_object(AR,T)
  ;  nonvar(T),T=[H2|T2],element_can_be_added_or_removed_to_avl(H2) ->  %print(removed_next(H2)),nl,
           remove_element_from_explicit_set(avl_set(AVL),H2,AR),equal_object(AR,[H|T2])
           % TO DO: move all such H2 to the front ??
  ;  expand_custom_set(avl_set(AVL),ES),  %print(expanded(ES)),nl,
     equal_cons_lwf(ES,H,T,LWF) ).
equal_cons_lwf(closure(P,Ty,B),H,T,LWF) :- expand_custom_set(closure(P,Ty,B),ES), 
    blocking_equal_cons_lwf(ES,H,T,LWF).
equal_cons_lwf(global_set(G),H,T,LWF) :- equal_cons_global_set(G,H,T,LWF).

:- block equal_cons_global_set(-,?,?,?).
equal_cons_global_set(G,H,T,LWF) :- is_infinite_global_set(G),!,
    /* should normally fail, unless T is not a list but contains closure or global set */
    test_finite_set(T,Finite),
    when((nonvar(Finite);nonvar(LWF)),
           (Finite \== pred_true,
            /* expansion will generate warning */
                          expand_custom_set(global_set(G),ES), equal_cons_lwf(ES,H,T,LWF))).
equal_cons_global_set(G,H,T,LWF) :- 
    %(is_infinite_global_set(G) -> test_finite_set(T,Finite), Finite \== pred_true ; true),
    expand_custom_set(global_set(G),ES), equal_cons_lwf(ES,H,T,LWF).

add_new_el(T,H,R) :- var(T),!,R=[H|T].
add_new_el(T,H,R) :- nonvar(T), is_custom_explicit_set(T,add_new_el),
   add_element_to_explicit_set(T,H,Res),!,Res=R.
add_new_el([HT|TT],H,R) :- !,R=[H,HT|TT].
add_new_el([],H,R) :- !, R=[H].
add_new_el(Set,H,R) :- !, expand_custom_set_to_list(Set,ESet,_,add_new_el),
    add_new_el(ESet,H,R).

delete_exact_member(V,_,_) :- var(V),!,fail.
delete_exact_member([H|T],El,Res) :- 
 (H==El -> Res=T
   ; Res=[H|TR], delete_exact_member(T,El,TR)).

var_list(V) :- unbound_variable_check(V),!.
var_list([]).
var_list([H|T]) :- unbound_variable(H),var_list(T).

unbound_variable(V) :- free_var(V), frozen(V,true).
unbound_variable_check(V) :- free_var(V), 
  ((frozen(HR,Goal), Goal\=true)
    -> nl,print('### WARNING: goal attached to unbound variable expression'),nl,print(HR:Goal),nl,fail
    ; true).

equal_cons_cons(HR,TR,H,T,_LWF) :-   unbound_variable(HR),
   var_list(TR),!,   % print(eq_var_cons(HR,TR,H,T)),nl,
  % HR is variable: no constraints/co-routines attached to it; no other element in TR is constrained either
  (HR,TR)=(H,T).
equal_cons_cons(HR,TR,H,T,LWF) :-  
  (TR==[] -> T=[],equal_object(HR,H)
    ; T==[] -> TR=[], equal_object(HR,H)
    ; % here we use LWF for the first time
      instantiate_lwf(LWF,LWF2),
      % print(equality_objects_lwf(HR,H,EqRes,LWF2)),nl,
      equality_objects_lwf(HR,H,EqRes,LWF2),
      % print(equal_cons2(EqRes,HR,TR,H,T,LWF2)),nl,
      equal_cons2(EqRes,HR,TR,H,T,LWF2),
      propagate_card(TR,T,EqRes)).

% this will instantiate LWF if it has not yet been computed
% (Idea: get_cardinality_wait_flag can be expensive; only do it if we really need the wait_flag)
instantiate_lwf(LWF,R) :- var(LWF),!,R=LWF.
instantiate_lwf(lwf_card(Set,Info,WF),LWF) :- !, get_cardinality_wait_flag(Set,Info,WF,LWF).
instantiate_lwf(LWF,LWF).

:- block equal_cons2(-,?,?,?,?,?).
equal_cons2(pred_true,_HR,TR,_H,T,_) :- equal_object(TR,T).
equal_cons2(pred_false,HR,TR, H,T,LWF) :- % print(eq_cons2_neq(HR,TR,H,T)),nl,
    % print(equal_cons_lwf(T,HR,TR2,LWF)),nl,
    equal_cons_lwf(T,HR,TR2,LWF), % look for HR inside T 
    % print(res(T,HR,TR2,LWF)),nl,
    T2=TR2,
    equal_cons_lwf(TR,H,T2,LWF). %, was instead of T2=TR2: equal_object(TR2,T2).

:- block propagate_card(-,-,-).
propagate_card(X,Y,EqRes) :-
  (nonvar(EqRes) -> true % we no longer need to propagate; equal_cons will traverse
   ; nonvar(X) -> propagate_card2(X,Y,EqRes)
   ; propagate_card2(Y,X,EqRes)).
propagate_card2([],Y,_) :- !,empty_set(Y).
propagate_card2([_|TX],Y,EqRes) :- !, 
  (var(Y) -> Y= [_|TY], propagate_card(TX,TY,EqRes)
   ; Y=[] -> fail
   ; Y=[_|TY] -> propagate_card(TX,TY,EqRes)
   ; true
  ). % TO DO: add more propagation
propagate_card2(_,_,_).   

%same_card_and_expand(A,B,ExpA,ExpB) :- .... + reorder ??

:- assert_must_succeed(type_error([],int(1))).
:- assert_must_succeed(type_error((int(1),int(2)),[pred_true])).
:- assert_must_succeed(type_error(string('Name'),global_set('Name'))).
:- assert_must_fail((type_error([],[_]))).
type_error([],Y) :- no_set_type_error(Y).
type_error([_|_],Y) :- no_set_type_error(Y).
%type_error(X,Y) :- is_custom_explicit_set(X,type_error1), no_set_type_error(Y).
type_error(avl_set(_),Y) :- no_set_type_error(Y).
type_error(global_set(_),Y) :- no_set_type_error(Y).
type_error(freetype(_),Y) :- no_set_type_error(Y).
type_error(closure(P,_,B),Y) :- 
   (var(P) -> true ; var(B) -> true ; P=[] -> true ; P=[P1|_], var(P1) -> true ; no_set_type_error(Y)).
type_error((_,_),Y) :- Y \= (_,_).
type_error(fd(_,T1),Y) :- (Y= fd(_,T2) -> nonvar(T1),nonvar(T2),T1 \=T2 ; true).
type_error(int(_),Y) :- Y\= int(_).
type_error(term(_),Y) :- Y\= term(_).
type_error(rec(_),Y) :- Y \= rec(_).
type_error(freeval(_,_,_),Y) :- Y \= freeval(_,_,_).
type_error(string(_),Y) :- Y \= string(_).
% Should raise type error: kernel_objects:union([int(1)],[[]],R).

no_set_type_error(int(_)).
no_set_type_error(fd(_,_)).
no_set_type_error((_,_)).
no_set_type_error(rec(_)).
no_set_type_error(pred_true /* bool_true */).
no_set_type_error(pred_false /* bool_false */).
no_set_type_error(term(_)).
no_set_type_error(string(_)).
no_set_type_error(freeval(_,_,_)).



:- assert_must_succeed(not_equal_object(term(a),term(b))).
:- assert_must_succeed(not_equal_object(string('a'),string('b'))).
:- assert_must_succeed(not_equal_object(int(1),int(2))).
:- assert_must_succeed(not_equal_object(rec([field(a,int(1))]),rec([field(a,int(2))]))).
:- assert_must_succeed(not_equal_object(rec([field(a,int(1)),field(b,int(2))]),
              rec([field(a,int(1)),field(b,int(3))]))).
:- assert_must_fail(not_equal_object(rec([field(a,int(1))]),rec([field(a,int(1))]))).
:- assert_must_fail(not_equal_object(rec([field(a,int(1)),field(b,int(2))]),
              rec([field(a,int(1)),field(b,int(2))]))).
:- assert_must_fail(not_equal_object(abort(msg),int(2))).
:- assert_must_fail(not_equal_object(fd(1,a),abort(msg))).
:- assert_must_succeed(not_equal_object(global_set(a),global_set(b))).
:- assert_must_succeed(not_equal_object([term(a),term(b)],[term(a),term(c)])).
:- assert_must_succeed((not_equal_object([(int(1),[Y])],[(int(X),[Z])]),
                     Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[int(2)])).
:- assert_must_succeed(not_equal_object((int(1),int(2)),(int(3),int(4)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_equal_object((int(1),int(2)),(int(1),int(4))))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_equal_object((int(1),int(4)),(int(3),int(4))))).
:- assert_must_fail(not_equal_object((int(1),int(4)),(int(1),int(4)))).
:- assert_must_succeed(not_equal_object((int(1),string('a')),(int(1),string('b')))).
:- assert_must_fail(not_equal_object((int(1),string('b')),(int(1),string('b')))).
:- assert_must_fail(not_equal_object([(term(a),[])],[(term(a),[])])).
:- assert_must_fail((not_equal_object([(int(1),[Y])],[(int(X),[Z])]),
                     Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[])).
:- assert_must_fail(not_equal_object([int(1),int(2)],[int(2),int(1)])).
:- assert_must_fail(not_equal_object(abort(msg),abort(another_msg))).
:- assert_must_succeed(not_equal_object([int(1),int(2)],[int(0),int(4)])).
:- assert_must_fail((sample_closure(C),
    not_equal_object(C,[int(1),int(2)]))).
:- assert_must_succeed((sample_closure(C),
    not_equal_object(C,[int(1),int(0)]))).
:- assert_must_succeed((sample_closure(C),
    not_equal_object(C,global_set('NAT')))).
:- assert_must_fail((not_equal_object( 
 [[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
    [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(3,'Name'),fd(2,'Name')]]
,[[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
     [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(2,'Name'),fd(3,'Name')]])
     )).
:- assert_must_fail((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcx,a,int(2))))).
:- assert_must_succeed((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcy,a,int(2))))).
:- assert_must_succeed((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcx,b,int(2))))).
:- assert_must_succeed((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcx,a,int(3))))).

:- assert_must_succeed((not_equal_object(pred_true /* bool_true */,X), X==pred_false /* bool_false */)).
:- assert_must_succeed((not_equal_object([],X),X=[_|_])).
%:- assert_must_succeed((not_equal_object([],X), nonvar(X),X=[_|_])).
:- assert_must_succeed((not_equal_object(X,[]), X=[_|_])).
:- assert_must_succeed((not_equal_object(X,pred_false /* bool_false */), X==pred_true /* bool_true */)).

:- assert_must_succeed(not_equal_object([_X],[int(1),int(3)])). % Inefficiency example of setlog
:- assert_must_succeed_any(not_equal_object([_X],[int(1)])). % Inefficiency example of setlog
:- assert_must_succeed((not_equal_object([X],[pred_true /* bool_true */]),X==pred_false /* bool_false */)).
:- assert_must_succeed((not_equal_object([pred_true /* bool_true */],[X]),X==pred_false /* bool_false */)). 
:- assert_must_succeed((not_equal_object([[X]],[[pred_true /* bool_true */]]),X==pred_false /* bool_false */)).
:- assert_must_succeed((not_equal_object([[pred_true /* bool_true */]],[[X]]),X==pred_false /* bool_false */)).
:- assert_must_succeed((custom_explicit_sets:singleton_set(avl_set(N),pred_true /* bool_true */), kernel_objects:not_equal_object(avl_set(N),[X]), X==pred_false /* bool_false */)).
:- assert_must_succeed((custom_explicit_sets:singleton_set(avl_set(N),pred_true /* bool_true */), kernel_objects:not_equal_object([X],avl_set(N)), X==pred_false /* bool_false */)).
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([],[int(3333)]))).	
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([],[int(2),int(1),int(3)]))).	
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([int(3)],[int(2),int(1),int(3)]))).	
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([int(3),int(1),int(4)],[int(2),int(1),int(3)]))).	
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([int(2),int(1),int(3),int(5)],[int(2),int(1),int(3)]))).	
% X in 3..4, kernel_objects:not_equal_object([int(2),int(3)],[int(2),int(X)]), X==4. in clpfd Mode



:- block not_equal_object_wf(-,-,?).
 /* TO DO: implement a better _wf version ;  use bool_dif if possible */
not_equal_object_wf(X,Y,WF) :- (var(X) -> X\==Y, not_equal_object_wf1(Y,X,WF)
                                   ; not_equal_object_wf1(X,Y,WF)).

not_equal_object_wf1([],R,WF) :- !, not_empty_set_wf(R,WF). 
not_equal_object_wf1(X,Y,_) :- not_equal_object2(X,Y).

not_equal_object(X,Y) :-
 (nonvar(X) -> not_equal_object2(X,Y)
  ; nonvar(Y) -> not_equal_object2(Y,X)
  ;   X\==Y, when((?=(X,Y);nonvar(X);nonvar(Y)), not_equal_object0(X,Y))).

not_equal_object0(X,Y) :- (var(X) -> X\==Y, not_equal_object2(Y,X)
                                   ; not_equal_object2(X,Y)).

not_equal_object2(pred_true /* bool_true */,R) :- !, R=pred_false /* bool_false */.
not_equal_object2(pred_false /* bool_false */,R) :- !, R=pred_true /* bool_true */.
not_equal_object2(fd(X,Type),R) :- !, R=fd(Y,Type), neq_fd(X,Y,Type).
not_equal_object2(int(X),R) :- !, R=int(Y), integer_dif(X,Y).
not_equal_object2(string(X),R) :- !, R=string(Y), dif(X,Y).
not_equal_object2(term(X),R) :- !, R=term(Y), dif(X,Y).
not_equal_object2(rec(F1),R) :- !, R=rec(F2),
    not_equal_fields(F1,F2).
%not_equal_object2([],R) :- var(R),!, print(not_empty(R)),nl,R=[_|_].  % Dangerous for equal_object: we generate a non-enumerated variable !
not_equal_object2((X1,X2),R) :- !, R=(Y1,Y2),
    not_equal_couple(X1,Y1,X2,Y2).
not_equal_object2(X,Y) :- is_custom_explicit_set(X,not_equal_object2),!, 
  %  print(not_equal_object2(X,Y)),nl, % uncovered case; for efficiency better explicitly written above
    not_equal_explicit_set(X,Y).
not_equal_object2(X,Y) :- not_equal_object3(X,Y).

:- block not_equal_explicit_set(?,-).
not_equal_explicit_set(X,Y) :- 
    is_custom_explicit_set(Y,not_equal_explicit_set),!,
    % print(not_equal_explicit_sets(X,Y)),nl,
    not_equal_explicit_sets(X,Y).
not_equal_explicit_set(X,[]) :- !, %print(check_non_empty(X)),nl,
   is_non_empty_explicit_set(X).
not_equal_explicit_set(X,Y) :-  % print_term_summary(expanding(X)),
    expand_custom_set(X,EX), not_equal_object3(EX,Y).

:- block not_equal_object3(?,-).
not_equal_object3(X,Y) :- nonvar(X),type_error(X,Y),
    add_internal_error('Internal Typing Error (please report as bug !) : ',not_equal_object(X,Y)),fail.
%not_equal_object3(abort(_),Y) :- fail,  /* Reason: otherwise abort values can be assigned to 
%         local variables (ANY,exists,forall) and a problem could be hidden from the user */
%        not_top_level_functor(Y,abort).
%not_equal_object3(X,abort(_)) :- fail,  /* reason for fail: see above */
%		not_top_level_functor(X,abort).
%    /* check later: encode ints as finite domain variables? */
not_equal_object3(freeval(FT1,C1,F1),freeval(FT2,C2,F2)) :-
    when((ground(FT1),ground(FT2),ground(C1),ground(C2)),
	 (FT1==FT2 -> (C1==C2 -> not_equal_object(F1,F2)
	                       ; true)
	            ; true)).
not_equal_object3([],X) :- not_empty_set(X).
not_equal_object3([H|T],Set2) :- Set2=[] -> true
  ; cardinality([H|T],N1), cardinality(Set2,N2),
     when(?=(N1,N2),  % when we trigger code below, = can be decided:
        (N1=N2 -> neq_cons(Set2,H,T) ; true)).
%       (dif(N1,N2) ;  (N1=N2, neq_cons(Set2,H,T)))).  %not_equal_object_sets(Set1,Set2) )) ).

   
not_equal_object_sets([H|T],Set2) :- 
    ( Set2=[H2|_T2]
    -> not_equal_object_sets2(H,T,H2,Set2)
    ;  true
    ).
:- block not_equal_object_sets2(-,?,?,?), not_equal_object_sets2(?,?,-,?).
not_equal_object_sets2(H,_T,_H2,Set2) :-
    not_element_of(H,Set2).
not_equal_object_sets2(H,T,_H2,Set2) :-
    remove_element(H,Set2,Del2), %print_message(rem(H,Set2,Del2)),
    not_equal_object(T,Del2).
    

:- block neq_cons(-,?,?).
%neq_cons(X,Y,Z) :- print(neq_cons(X,Y,Z)),nl,fail.
neq_cons([],_,_).
neq_cons([H2|T2],H1,T1) :- 
   ((T2==[],T1==[])
     -> not_equal_object(H1,H2)
      ; check_and_remove([H2|T2],H1,NewSet2,RemoveSuccesful),
        % print(removed(H1,H2,T2,NewSet2,RemoveSuccesful)),
        neq_cons2(RemoveSuccesful,T1,NewSet2)
   ).
neq_cons(avl_set(A),H1,T1) :- element_can_be_added_or_removed_to_avl(H1),!,
    %print(removing(H1)),nl,
    (remove_element_from_explicit_set(avl_set(A),H1,RA)
      ->  not_equal_object(T1,RA)
      ;  true ).
neq_cons(ES,H1,T1) :- is_custom_explicit_set(ES,neq_cons),expand_custom_set(ES,ExpSet),
   neq_cons(ExpSet,H1,T1).

:- block neq_cons2(-,?,?).
neq_cons2(not_successful,_T1,_NewSet2).
neq_cons2(successful,T1,NewSet2) :-  not_equal_object_sets(T1,NewSet2).

% kernel_objects:not_equal_couple(int(1),int(Y),B,pred_true /* bool_true */).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(1),int(Y),B,pred_true /* bool_true */),Y=1, B==pred_false /* bool_false */)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),B,pred_true /* bool_true */),Y=1, B==pred_false /* bool_false */)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),B,pred_false /* bool_false */),Y=1, B==pred_true /* bool_true */)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),pred_false /* bool_false */,B),Y=1, B==pred_true /* bool_true */)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),B,pred_true /* bool_true */),Y=2, var(B))).
:- assert_must_succeed(( kernel_objects:not_equal_couple(B,pred_true /* bool_true */,int(Y),int(1)),Y=1, B==pred_false /* bool_false */)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(B,fd(C,'Code'),fd(Y,'Name'),F),F=fd(1,'Name'),Y=1,B=fd(1,'Code'),C=2 )).
:- assert_must_succeed(( kernel_objects:not_equal_couple(B,pred_true /* bool_true */,fd(Y,'Name'),F),F=fd(1,'Name'),Y=1, B==pred_false /* bool_false */)).
:- block not_equal_couple(-,?,-,?),not_equal_couple(?,-,?,-).
% (X1,X2) /= (Y1,Y2) 
not_equal_couple(X1,Y1,X2,Y2) :- %print(not_equal_couple(X1,Y1,X2,Y2)),nl,
  (atomic_value_pair(X1,Y1,X1V,Y1V,T1)
    -> (atomic_value_pair(X2,Y2,X2V,Y2V,T2)
         -> neq_atomic_pair(X1V,Y1V,T1,X2V,Y2V,T2)
         ;  when(?=(X1V,Y1V),((X1V=Y1V) -> not_equal_object(X2,Y2) ; true))
           % we could use kernel_equality for X2,Y2 ...; also true below
       )
     ; (atomic_value_pair(X2,Y2,X2V,Y2V,_T2)
         -> when(?=(X2V,Y2V),((X2V=Y2V) -> not_equal_object(X1,Y1) ; true))
         ;  not_equal_couple3(X1,Y1,X2,Y2)
       )
   ).


neq_atomic_pair(X1,Y1,T1,X2,Y2,T2) :-
  when((?=(X1,Y1);?=(X2,Y2)),
        (?=(X1,Y1) -> (X1 \= Y1 -> true ; dif(T2,X2,Y2))
                   ;  (X2 \= Y2 -> true ; dif(T1,X1,Y1))
        )).

 
atomic_value_pair(X1,Y1,X1V,Y1V,Type) :-
  (nonvar(X1)
   -> atomic_value(X1,X1V,Type), get_atomic_value(Type,Y1,Y1V)
   ;  nonvar(Y1), atomic_value(Y1,Y1V,Type), get_atomic_value(Type,X1,X1V)
   ).

get_atomic_value(integer,R,X) :- get_int(R,X). % warning: instantiating to int(X) can trigger co-routines; causing problems in the cut above ->
get_atomic_value(global(T),R,X) :- get_fd(R,T,X). % warning: ditto
get_atomic_value(boolean,X,X).

:- block get_int(-,?).
get_int(int(R),X) :- !, R=X.
get_int(R,X) :- add_internal_error('Type error: ',get_atomic_value(integer,R,X)),fail.
:- block get_fd(-,?,?).
get_fd(fd(R,T),T,X) :- !, R=X.
get_fd(R,T,X) :- add_internal_error('Type error: ',get_atomic_value(global(T),R,X)),fail.

atomic_value(fd(X,T),X,global(T)).
atomic_value(int(X),X,integer).
atomic_value(pred_true /* bool_true */,pred_true /* bool_true */,boolean).
atomic_value(pred_false /* bool_false */,pred_false /* bool_false */,boolean).
    
not_equal_couple3(X1,Y1,X2,Y2) :-  
  equality_objects(X1,Y1,EqRes),
  not_equal_couple4(EqRes,X2,Y2).

:- block not_equal_couple4(-,?,?).
not_equal_couple4(pred_true,X2,Y2) :- not_equal_object(X2,Y2).
not_equal_couple4(pred_false,_,_).

% we could use kernel_equality here ?!
   %when_sufficiently_instantiated(Y1,Y2,    mal: changed to ground on Oct 22 2006
%   when((ground(X2),ground(Y2)),
%    (not_equal_object(X2,Y2) ;     /* POSSIBLE BRANCHING HERE ! EG X1=int(_) X2=int(_) */
%     (equal_object(X2,Y2),not_equal_object(X1,Y1)))).
    
 /* To do: provide special support for things like
    couple of fd's [done], list of fd's, set of fd's */

:- block not_equal_fields(-,-).
not_equal_fields([field(ID1,V1)|T1],[field(ID2,V2)|T2]) :-
 % should we wait for ID1 or ID2 to become nonvar?
   check_field_name_compatibility(ID1,ID2),
   equality_objects(V1,V2,Res),
   not_equal_fields2(Res,T1,T2).
:- block not_equal_fields2(-,?,?).
not_equal_fields2(pred_false,_T1,_T2).
not_equal_fields2(pred_true,T1,T2) :-
    not_equal_fields(T1,T2).

check_field_name_compatibility(Name1,Name2) :-
   nonvar(Name1), nonvar(Name2), Name1 \= Name2, !,
   add_internal_error('Incompatible fields: ',check_field_name_compatibility(Name1,Name2)).
check_field_name_compatibility(_,_).


/* ------------------------------------------- */
/*      equality_objects/3  function     */
/* ------------------------------------------- */

%% :- ensure_loaded(kernel_equality).

% ----------------------------------------------------------
% ----------------------------------------------------------



:- use_module(kernel_equality).

% ----------------------------------------------------------
% ----------------------------------------------------------
 
/* ---------------> */
/* This should probably be more systematically applied before every kernel call
 + expanded for other symbolic representations !! */
 

 
/* underlying assumption: if G is a global set: we get back the
 global_set tag immediately: no need to use when to wait;
  better: ensure that b_compute_expression always returns a nonvar term */

integer_global_set('NAT').
integer_global_set('NATURAL').
integer_global_set('NAT1').
integer_global_set('NATURAL1').
integer_global_set('INT').
integer_global_set('INTEGER').


:- assert_must_succeed(( kernel_objects:element_of_global_set(int(0),'NATURAL'))).
:- assert_must_fail(( kernel_objects:element_of_global_set(int(0),'NATURAL1'))).
:- assert_must_fail(( kernel_objects:element_of_global_set(int(-1),'NATURAL'))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(int(-1),'INTEGER'))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(int(0),'NAT'))).
:- assert_must_fail(( kernel_objects:element_of_global_set(int(0),'NAT1'))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(X,'NAT'),X=int(1))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(X,'NATURAL'),X=int(1))).
                            
element_of_global_set(X,GS) :-
    init_wait_flags(WF),element_of_global_set_wf(X,GS,WF),ground_wait_flags(WF).

:- block element_of_global_set_wf(?,-,?).
element_of_global_set_wf(El,Set,WF) :- b_global_set(Set),!,
      global_type_wf(El,Set,WF).
element_of_global_set_wf(X,'STRING',_WF) :- !, X=string(_).
element_of_global_set_wf(int(X),GS,WF) :- 
      element_of_global_integer_set_wf(GS,X,WF).

/* what about BOOL ?? */
element_of_global_integer_set_wf('NAT',X,WF) :- 
   preferences:get_preference(maxint,MAXINT),
   in_nat_range_wf(int(X),int(0),int(MAXINT),WF).
element_of_global_integer_set_wf('NATURAL',X,WF) :- 
   (ground(X) -> X>=0
      ;  is_natural(int(X),WF),
         %get_last_wait_flag(element_of_global_set(int(X),'NATURAL'),WF,LWF),
         get_integer_enumeration_wait_flag(X,'NATURAL',WF,LWF),
         enumerate_natural(X,0,LWF)
   ).
element_of_global_integer_set_wf('NAT1',X,WF) :- 
   preferences:get_preference(maxint,MAXINT),
   in_nat_range_wf(int(X),int(1),int(MAXINT),WF).
element_of_global_integer_set_wf('NATURAL1',X,WF) :- 
   (ground(X) -> X>=1
      ;  is_natural1(int(X),WF),
         %get_last_wait_flag(element_of_global_set_wf(int(X),'NATURAL1'),WF,LWF),
         get_integer_enumeration_wait_flag(X,'NATURAL1',WF,LWF),
         enumerate_natural(X,1,LWF)
    ).
element_of_global_integer_set_wf('INT',X,WF) :- 
   preferences:get_preference(minint,MININT),
   preferences:get_preference(maxint,MAXINT),
   in_nat_range_wf(int(X),int(MININT),int(MAXINT),WF).
element_of_global_integer_set_wf('INTEGER',X,WF) :-
    (ground(X) -> true
     ;  %get_last_wait_flag(element_of_global_set_wf(int(X),'INTEGER'),WF,LWF),
        get_integer_enumeration_wait_flag(X,'INTEGER',WF,LWF),
        enumerate_int_wf(X,LWF) %when((nonvar(LWF);nonvar(X)),(ground(X)->true;enumerate_int(X))) 
    ). 


get_integer_enumeration_wait_flag(X,SET,WF,LWF) :-
   clpfd_domain(X,FDLow,FDUp), finite_domain(FDLow,FDUp),!,
   Size is 1+FDUp-FDLow,
   get_wait_flag(Size,element_of_global_set_wf(int(X),SET),WF,LWF).
get_integer_enumeration_wait_flag(X,SET,WF,LWF) :-
   get_last_wait_flag(element_of_global_set_wf(int(X),SET),WF,LWF).

:- assert_must_succeed((kernel_objects:enumerate_int_wf(X,4),X==2)).
:- block enumerate_int_wf(-,-).
enumerate_int_wf(X,_LWF) :- %print(enum_int(X)),nl,
   (ground(X) -> true ; enumerate_int(X)).

:- assert_must_succeed(not_element_of_global_set(int(-1),'NAT')).
:- assert_must_succeed(not_element_of_global_set(int(-1),'NATURAL')).
:- assert_must_succeed(not_element_of_global_set(int(0),'NAT1')).
:- assert_must_succeed(not_element_of_global_set(int(0),'NATURAL1')).
not_element_of_global_set(int(X),GS) :- 
   (var(GS) -> add_error(kernel_objects,var_not_element_of_global_set,(int(X),GS)) ; true),
   not_element_of_global_set2(GS,X).
not_element_of_global_set2('NAT',X) :- 
   preferences:get_preference(maxint,MAXINT),
   clpfd_not_in_non_empty_range(X,0,MAXINT). %when(nonvar(X), (X<0 ; X>MAXINT)).
not_element_of_global_set2('NATURAL',X) :- is_not_natural(int(X)).
not_element_of_global_set2('NAT1',X) :- 
   preferences:get_preference(maxint,MAXINT),
   clpfd_not_in_non_empty_range(X,1,MAXINT). %when(nonvar(X),(X<1 ; X>MAXINT)).
not_element_of_global_set2('NATURAL1',X) :- is_not_natural1(int(X)).
not_element_of_global_set2('INT',X) :- 
   preferences:get_preference(minint,MININT),
   preferences:get_preference(maxint,MAXINT),
   clpfd_not_in_non_empty_range(X,MININT,MAXINT). %when(nonvar(X), (X < MININT ; X > MAXINT)).
%not_element_of_global_set(string(_X),'STRING') :- fail.
%not_element_of_global_set(int(_X),'INTEGER') :- fail.
%not_element_of_global_set(_El,Set) :- b_global_set(Set), fail.  
 


/* ---- */
/* SETS */
/* ---- */

%:- block is_a_set(-).
%is_a_set(X) :- is_a_set2(X).
%is_a_set2([]) :- !.
%is_a_set2([_|_]) :- !.
%is_a_set2(X) :- is_custom_explicit_set(X,is_a_set2).




:- assert_must_succeed(exhaustive_kernel_fail_check(empty_set([int(4),int(3)]))).
:- assert_must_fail((empty_set([int(2),int(1)]))).
:- assert_must_fail((empty_set([int(1)]))).
:- assert_must_fail((empty_set([[]]))).
:- assert_must_fail((empty_set(global_set('Name')))).
:- assert_must_fail((empty_set(X),X=[int(1)])).
:- assert_must_succeed((empty_set([]))).
empty_set(X) :- (var(X) -> X=[]
                  ; X=[] -> true
                 % ; X=[_|_] -> fail
                 % ; X=avl_set(A) -> fail /* avl_sets should never be empty */
                  ; is_custom_explicit_set(X,empty_set),is_empty_explicit_set(X)).


:- assert_must_succeed(exhaustive_kernel_check(not_empty_set([int(4),int(3)]))).
:- assert_must_succeed((kernel_objects:not_empty_set([int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_empty_set([int(1)]))).
:- assert_must_succeed((kernel_objects:not_empty_set([[]]))).
:- assert_must_succeed((kernel_objects:not_empty_set(global_set('Name')))).
:- assert_must_succeed((kernel_objects:not_empty_set_lwf(X,1),nonvar(X),X=[_|_])).
:- assert_must_succeed((kernel_objects:not_empty_set_lwf([int(1)],_))).
:- assert_must_fail((kernel_objects:not_empty_set([]))).

not_empty_set_wf(S,_WF) :-
  ((var(S),preferences:preference(use_smt_mode,true))
    -> %print(setting_not_empty(S)),nl,
       S=[_|_]
    ; not_empty_set(S)
  %  get_wait_flag(3,not_empty_set_wf,WF,LWF),  not_empty_set_lwf(S,LWF)).
  ).

:- block not_empty_set_lwf(-,-).
% the instantiation with a list skeleton can easily cause multiple solutions for the same
% set to be found: hence we guard it by a wait flag
not_empty_set_lwf(S,_LWF) :- var(S),!, %print(setting_not_empty(S)),nl,
                             S=[_|_].
not_empty_set_lwf(S,_) :- not_empty_set(S).

:- block not_empty_set(-).
%not_empty_set(S) :- var(S),!, print(setting_not_empty(S)),nl,S=[_|_].
  % not_empty_set(X) :- not_equal_object([],X).
not_empty_set([_|_]).
not_empty_set(avl_set(_)).
not_empty_set(closure(P,T,B)) :- is_non_empty_explicit_set(closure(P,T,B)).
not_empty_set(global_set(Type)) :- b_non_empty_global_set(Type).




:- assert_must_succeed((exact_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed((exact_element_of(int(1),[int(2),int(3),int(4),int(1)]))).
:- assert_must_succeed((exact_element_of(int(4),[int(2),int(3),int(4),int(1)]))).
:- assert_must_succeed((exact_element_of(int(1),[int(2),int(3)|T]), T=[int(4),int(1)])).
:- assert_must_fail((exact_element_of(int(5),[int(2),int(3)|T]), T=[int(4),int(1)])).
:- assert_must_succeed((exact_element_of(fd(1,'Name'),global_set('Name')))).
:- assert_must_succeed((exact_element_of([int(2),int(1)],[[],[int(2),int(1)]]))).
:- assert_must_fail((exact_element_of([int(1),int(2)],[[],[int(2),int(1)]]))).
%:- assert_must_succeed((exact_element_of([(int(1),fd(2,'Name'))],
%      closure([zzzz],[set(couple(integer,global('Name')))], 'In'('ListExpression'(['Identifier'(zzzz)]),
%                        'Seq'(value([fd(1,'Name'),fd(2,'Name')]))))) )).
%:- assert_must_succeed((exact_element_of(XX,
%     closure([zzzz],[set(couple(integer,global('Name')))], 'In'('ListExpression'(['Identifier'(zzzz)]),
%                        'Seq'(value([fd(1,'Name'),fd(2,'Name')]))))),
%                        equal_object(XX,[(int(1),fd(1,'Name'))]) )).
%:- assert_must_succeed((
%exact_element_of(XX,closure([zzzz],[set(couple(integer,global('Name')))],
% 'In'('ListExpression'(['Identifier'(zzzz)]),
%   'Perm'(value([fd(1,'Name'),fd(2,'Name')]))))),
%   equal_object(XX,[(int(1),fd(2,'Name')),(int(2),fd(1,'Name'))]) )).

%:- assert_must_succeed(( exact_element_of(X,
% closure([zzzz],[set(record([field(balance,integer),field(name,global('Code'))]))],
% 'In'('ListExpression'(['Identifier'(zzzz)]),
% 'PowerSet'(value(closure([zzzz],
% [record([field(balance,integer),field(name,global('Code'))])],'In'('ListExpression'(['Identifier'(zzzz)]),
% 'SetOfRecords'(value(cons_expr(field(balance,global_set('NAT')),
% cons_expr(field(name,global_set('Code')),nil_expr))))))))))),
%  X=[rec([field(balance,int(0)),field(name,fd(2,'Code'))])] )).
%:- assert_must_fail(( exact_element_of(X,
%   closure([zzzz],[set(record([field(balance,integer),field(name,global('Code'))]))],
%  'In'('ListExpression'(['Identifier'(zzzz)]),
%  'PowerSet'(value(closure([zzzz],
%  [record([field(balance,integer),field(name,global('Code'))])],'In'('ListExpression'(['Identifier'(zzzz)]),
%  'SetOfRecords'(value(cons_expr(field(balance,global_set('NAT')),
%  cons_expr(field(name,global_set('Code')),nil_expr))))))))))), 
%  X=[rec([field(balance,int(-1)),field(name,fd(2,'Code'))])] )). 
  
                        
/* use this to compute elements */
exact_element_of(X,Set) :- %print_message(exact_element_of(X,Set)),
     dif(Set,[]),
     exact_element_of2(Set,X).
:- block exact_element_of2(-,?).
exact_element_of2([H|_],H).
exact_element_of2([_|T],E) :- exact_element_of3(T,E).
exact_element_of2(X,E) :- is_custom_explicit_set(X,exact_element_of2), check_element_of(E,X).
:- block exact_element_of3(-,?).
exact_element_of3([H|_],H).
exact_element_of3([_|T],E) :- exact_element_of3(T,E).


:- assert_must_succeed(exhaustive_kernel_check(check_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(check_element_of(int(3),[int(2),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(check_element_of(int(1),[]))).

/* uses equal_object instead of unification */
:- assert_must_succeed((check_element_of(X,
      [(int(1),(int(1),(int(1),int(1)))),(int(2),(int(1),(int(1),int(1)))),
       (int(1),(int(1),(int(1),int(2)))),(int(2),(int(1),(int(1),int(2))))]),
       equal_object(X, (int(2),(int(1),(int(1),int(2))))) )).
:- assert_must_succeed((check_element_of(X,
      [ (((int(1),int(1)),int(1)),int(1)), (((int(1),int(1)),int(1)),int(2)),
        (((int(1),int(1)),int(1)),int(3)), (((int(1),int(1)),int(1)),int(4)),
        (((int(1),int(1)),int(2)),int(1)), (((int(1),int(1)),int(2)),int(2))
        ]), equal_object(X,  (((int(1),int(1)),int(2)),int(1)))
      )).
:- assert_must_succeed((check_element_of(fd(1,'Name'),global_set('Name')))).
%:- assert_must_succeed_multiple(check_element_of(X,[[fd(1,'Name')],[]])).
:- assert_must_succeed((check_element_of((int(1),int(2)),[(int(1),int(2))]))).
:- assert_must_succeed((check_element_of((_X,_Y),[(fd(2,'Code'),fd(2,'Code'))]))).
:- assert_must_succeed((init_wait_flags(WF),
    check_element_of_wf((X,Y),[(fd(2,'Code'),fd(2,'Code'))],WF),
    ground_det_wait_flag(WF), X= fd(2,'Code'), Y= fd(2,'Code'),
   kernel_waitflags:ground_wait_flags(WF) )).
   :- assert_must_succeed((init_wait_flags(WF),
          check_element_of_wf((Y,X),[(fd(2,'Code'),fd(2,'Code'))],WF),
          ground_det_wait_flag(WF), X= fd(2,'Code'), Y= fd(2,'Code'),
   kernel_waitflags:ground_wait_flags(WF) )).
:- assert_must_succeed((check_element_of([int(1),int(2)],[[int(2),int(1)]]))).

:- assert_must_succeed((check_element_of([int(1),int(2)],[[],[int(2),int(1)]]))).
:- assert_must_succeed((check_element_of(X,[[],[int(2),int(1)]]), X==[] )).
:- assert_must_succeed((check_element_of_wf(X,[[],[int(2),int(1)]],_WF), 
    equal_object(X,[int(1),int(2)]) )).
:- assert_must_succeed((check_element_of_wf(XX,global_set('Name'),WF),kernel_waitflags:ground_wait_flags(WF), XX==fd(3,'Name') )).
:- assert_must_fail(check_element_of([fd(2,'Name')],[[fd(1,'Name')],[]])).
:- assert_must_fail((check_element_of([int(2)],[[],[int(2),int(1)]]))).
:- assert_must_succeed((check_element_of(int(1),_X))).
:- assert_must_succeed((check_element_of((int(2),_X),[(int(1),[(int(1),int(22))]),(int(2),[(int(1),int(55))])]))).

check_element_of(X,Set) :- init_wait_flags(WF),
  check_element_of_wf(X,Set,WF),
  ground_wait_flags(WF).
  
% new test: check_element_of(int(1),X).
% new test: check_element_of(int(1),[int(2)|X]).

% same as check_element_of_wf but does not wait on Y:
% should also work for relation ??
:- use_module(custom_explicit_sets,[element_of_closure/5]).
check_element_of_function_closure(X,Y, P,T,B, WF) :-
   element_of_closure((X,Y),P,T,B,WF).

check_element_of_wf(X,Set,WF) :- %%print_message(check_element_of_wf(X,Set)),
   dif(Set,[]),
   check_element_of1(X,Set,WF).

%check_element_of1(X,Set,WF) :- var(X),var(Set),unbound_variable_check(Set),!,
%   Set=[_|_], check_element_of2(Set,X,WF).
%:- block check_element_of1(-,-,?). %%
check_element_of1(X,Set,WF) :-   % print_term_summary(check_element_of1(X,Set,WF)),
  (unbound_variable_check(Set)
    ->  mark_as_non_free(X),
        Set=[X|_] % Note: X needs to be nonvar so that other code knows X is not free anymore
        % TO DO: normalise X ?
    ;  check_element_of2(Set,X,WF) 
  ).

% attach co-routine to mark a given term as not a real variable
mark_as_non_free(X) :- var(X) -> non_free(X) ; true.
:- block non_free(-).
non_free([H|T]) :- !, mark_as_non_free(H), mark_as_non_free(T).
non_free(_).

:- block check_element_of2(-,?,?).
check_element_of2(CS,El,WF) :-
              is_custom_explicit_set(CS,check_element_of2),!, element_of_custom_set_wf(El,CS,WF). 
check_element_of2([],_,_) :- !,fail.
check_element_of2([H|T],E,WF) :- !, % print(check_element_of4w(E,H,T,WF)),nl,
              %get_partial_set_priority([H|T],WF,LWF), %%
              %get_wait_flag(2,check_element_of2([H|T],E),WF,LWF), %%
              get_cardinality_wait_flag([H|T],check_element_of2,WF,LWF),
              check_element_of4w(E,H,T,WF,LWF).
check_element_of2(freetype(Id),E,WF) :- !, is_a_freetype_wf(E,Id,WF).
check_element_of2(Set,E,WF) :-
   add_internal_error('Illegal argument: ',check_element_of2(Set,E,WF)),fail.

% call if you already have an explicit waitflag (LWF) setup for the cardinality of the set
:- block check_element_of_wf_lwf(?,-,?,?).
check_element_of_wf_lwf(El,CS,WF,_LWF) :-
        is_custom_explicit_set(CS,check_element_of_wf_lwf),!, element_of_custom_set_wf(El,CS,WF). 
check_element_of_wf_lwf(E,[H|T],WF,LWF) :- check_element_of4w(E,H,T,WF,LWF).
check_element_of_wf_lwf(E,freetype(Id),WF,_) :- !, is_a_freetype_wf(E,Id,WF).

:- block check_element_of4w(-,?,-,?,-).
%check_element_of4w(E,H,T,_WF,_LWF) :- print(check_element_of4w(E,H,T,_WF,_LWF)),nl,fail.
check_element_of4w(E,H,T,_WF,_LWF) :- T==[],!,equal_object(E,H).
check_element_of4w(E,H,_T,_WF,_LWF) :- E==H ,!. %,print(eq(E,H)),nl.  % added by mal, 17.10 2007
check_element_of4w(E,H,T,WF,LWF) :- T\==[],
   equality_objects_lwf(E,H,Res,LWF),
   check_element_of4(Res,E,T,WF,LWF).

:- block check_element_of4(-,?,?,?,-).
check_element_of4(pred_true,_E,_,_WF,_LWF).
check_element_of4(pred_false,E,T,WF,LWF) :- 
   (var(T) -> T = [E|_] ; check_element_of5(E,T,WF,LWF)).

:- block check_element_of5(?,-,?,?).
check_element_of5(E,R,WF,LWF) :-  %print(equal_cons(R,H,T,for(E))),nl,
  get_next_element(R,H,T), 
  check_element_of4w(E,H,T,WF,LWF).

       

:- assert_must_succeed(exhaustive_kernel_check(not_element_of(int(3),[int(2),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_check(not_element_of(int(3),[int(2),int(1),int(4)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_element_of(int(3),[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_element_of(fd(1,'Name'),[]))).
:- assert_must_fail((kernel_objects:not_element_of(fd(1,'Name'),global_set('Name')))).
:- assert_must_succeed((kernel_objects:not_element_of(X,[fd(1,'Name')]),X = fd(2,'Name'))).
:- assert_must_fail((kernel_objects:not_element_of(X,[fd(1,'Name')]),X = fd(1,'Name'))).
:- assert_must_succeed(kernel_objects:not_element_of(term(a),[])).
:- assert_must_fail((kernel_objects:not_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_element_of([int(1),int(2)],
         [[int(1)],[int(0),int(4)],[int(0),int(3)],[int(0),int(1)],[int(0)],[]]))).
:- assert_must_fail((kernel_objects:not_element_of(abort(3),[int(2),int(1)]))).


not_element_of(X,Set) :- init_wait_flags(WF),
  not_element_of_wf(X,Set,WF),
  ground_wait_flags(WF).

:- block not_element_of_wf(-,-,?).
not_element_of_wf(_,Set,_) :- Set==[],!.
not_element_of_wf(El,Set,_) :- nonvar(El),El=fd(X,GS),b_fd_type(GS,N,N),!,
   % we have a global set with a single element; Set must be empty
   % print(not_el_of(El,Set)),nl,
   X=N,empty_set(Set).
not_element_of_wf(El,Set,WF) :- not_element_of_wf1(Set,El,WF).

:- block not_element_of_wf1(-,?,?).
%not_element_of_wf1(E,S,WF) :- print(not_el_of(E,S,WF)),nl,fail.
not_element_of_wf1(X,E,WF) :- is_custom_explicit_set(X,not_element_of_wf1),!,
   %print(check_not_el(E,X,WF)),nl,
   not_element_of_custom_set_wf(E,X,WF). % , print(ok(E,X,WF)),nl.
not_element_of_wf1([],_E,_WF).
not_element_of_wf1([H|T],E,WF) :- /* print(call(not_element_of_wf1([H|T],E))),nl, */
   not_equal_object_wf(E,H,WF),
   not_element_of_wf1(T,E,WF).

   
:- assert_must_succeed(exhaustive_kernel_check(add_element(int(3),[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(add_element(int(2),[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(add_element(int(4),[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed((kernel_objects:add_element(int(3),[int(2),int(1)],R),
                        kernel_objects:equal_object(R,[int(1),int(2),int(3)]))).
:- assert_must_succeed((kernel_objects:add_element([int(2)],[[int(2),int(1)],[]],R),
                        kernel_objects:equal_object(R,[[],[int(1),int(2)],[int(2)]]))).
:- assert_must_succeed((kernel_objects:add_element([int(1),int(2)],[[int(2),int(1)],[]],R),
                        kernel_objects:equal_object(R,[[],[int(1),int(2)]]))).
:- assert_must_succeed((kernel_objects:add_element(X,[int(2),int(1)],R),
                        kernel_objects:equal_object(R,[int(1),int(2)]), X = int(1))).
:- assert_must_succeed((kernel_objects:add_element([int(1),int(2)],
         [[int(1)],[int(0),int(4)],[int(0),int(3)],[int(0),int(1)],[int(0)],[]], _R))).

:- assert_must_succeed((kernel_objects:add_element(int(3),[int(X),int(1)],R,D),
        var(D), X=3, R==[int(3),int(1)], D==done)).
                        
:- assert_must_fail((kernel_objects:add_element(abort(msg),[int(2),int(1)],_R))).
:- assert_must_succeed((kernel_objects:add_element(int(3),[int(2),int(X)],R),
   nonvar(R), R =[H|T], H==int(2), nonvar(T),T=[_HH|TT],var(TT),
   X=4, T==[int(4),int(3)])).
:- assert_must_succeed((kernel_objects:add_element(int(3),[int(2),int(X)],R),
   nonvar(R), R =[H|T], H==int(2), nonvar(T),T=[_HH|TT],var(TT),
   X=3, T==[int(3)])).
:- assert_must_succeed((kernel_objects:add_element(int(3),X,[int(2),int(3)]),
   kernel_objects:equal_object(X,[int(2)]) )).
:- assert_must_succeed((kernel_objects:add_element(int(3),X,[int(3)]),
   kernel_objects:equal_object(X,[]) )).
:- assert_must_succeed((add_element(X,[int(1)],[int(1)]),X==int(1))).
:- assert_must_succeed((add_element(X,[],[int(1)]),X==int(1))).
% kernel_objects:add_element(E,[H],R,Done), H = int(X), E=int(Y), X in 1..10, Y in 11..20.


add_element(E,Set,NewSet) :- add_element(E,Set,NewSet,_).

:- block add_element(?,-,?,?).
add_element(Element,Set,NewSet,Done) :- Set==[],!, equal_object(NewSet,[Element]),Done=done.
add_element(E,Set,NewSet,Done) :- add_element1(E,Set,NewSet,Done).

:- block %add_element1(-,?,-,?), 
         add_element1(?,-,?,?).
add_element1(E,Set,NewSet,Done) :- var(E),!, add_element_var(Set,NewSet,E,Done).
add_element1(E,[H|T],NewSet,Done) :- E==H,!, % avoid running [H|T] through expand_custom_set_to_list, in case T is a variable this will create a pending co-routine
     equal_object(NewSet,[H|T]),Done=done.
add_element1(E,Set,NewSet,Done) :- 
     nonvar(Set), is_custom_explicit_set(Set,add_element), add_element_to_explicit_set(Set,E,R),!,
                            % print(add_to_explicit_set(E)),nl,
                             equal_object(R,NewSet),Done=done.
add_element1(E,Set,NewSet,Done) :-
    expand_custom_set_to_list(Set,ESet,_,add_element1), % we could avoid this expansion by treating avl_set,... below in add_element3
    add_element2(ESet,E,NewSet,Done).


add_element_var([],Res,Element,Done) :- !, % print(add_elvarempty(Element,[],Res)),nl,
   equal_cons(Res,Element,[]),Done=done.
add_element_var(Set,Res,Element,Done) :- Set \= [], Set \= closure(_,_,_),
  is_one_element_set(Res,ResEl), !, 
  % the result is a one element set; hence Element *must* be the element in that set
   %% print(add_elvar(Element,Set,Res,ResEl)),nl,
   equal_object(Element,ResEl), equal_object(Set,Res), Done=done.
add_element_var(Set,Res,Element,Done) :- %when(nonvar(Element), add_element(Element,Set,Res,Done)).
   expand_custom_set_to_list(Set,ESet,_,add_element_var),
   add_element2(ESet,Element,Res,Done).

is_one_element_set(S,_) :- var(S),!,fail.
is_one_element_set([H|T],H) :- T==[].
is_one_element_set(avl_set(S),El) :- is_one_element_custom_set(avl_set(S),El).

:- block add_element2(-,?,?,?).
% add_element2(E,S,R,_) :- print_message(add_el2(E,S,R)),fail.
add_element2([],E,Res,Done) :- var(Res),should_be_converted_to_avl(E),
         construct_avl([E],R),!,R=Res,Done=done.
add_element2(S,E,Res,Done) :- copy_list_skeleton(S,Res),add_element3(S,E,Res,Done).

:- block copy_list_skeleton(-,?).
copy_list_skeleton([],_) :- !.
copy_list_skeleton([H|T],R) :- !,equal_cons(R,H,RR), copy_list_skeleton(T,RR).
copy_list_skeleton(Set,R) :- !,is_custom_explicit_set(Set,copy_list_skeleton),
   expand_custom_set_to_list(Set,ESet,_,copy_list_skeleton), copy_list_skeleton(ESet,R).
copy_list_skeleton(Skel,R) :- add_internal_error('Argument not a set: ',copy_list_skeleton(Skel,R)).
 
:- block add_element3(-,?,?,?).
add_element3([],E,Res,Done) :- equal_cons(Res,E,[]),Done=done.
add_element3([H|T],E,Res,Done) :- equality_objects(H,E,EqRes), 
   equal_cons(Res,H,TailRes), % was: equal_object([H|TailRes],Res),  % use WF?
   (var(EqRes) -> has_not_to_be_added([H|T],Res,EqRes,0) ; true),
   add_element4(EqRes,T,E,TailRes,Done).

% check if an element has not to be added to arg1 to obtain arg2
:- block has_not_to_be_added(?,-,?,?),has_not_to_be_added(-,?,?,?).
%has_not_to_be_added(A,B,R,Sz) :- print(has_not_to_be_added(A,B,R,Sz)),nl,fail.
has_not_to_be_added([],[],R,Sz) :- !,(Sz<2 -> R=pred_true ; true).
has_not_to_be_added([],[_H|T],R,_Sz) :- !, %(var(R) -> print(add_f([],[_H|T],R,_Sz)),nl ; true),
   empty_set(T),R=pred_false.
has_not_to_be_added([_|_],[],_,_) :- !,fail. % we can either add or not; in both cases we do not obtain []
has_not_to_be_added([_|T1],[_|T2],R,Sz) :- !, S1 is Sz+1, has_not_to_be_added(T1,T2,R,S1).
has_not_to_be_added(_,_,_,_). % to do: support custom explicit sets

:- block add_element4(-,?,?,?,?).
add_element4(pred_true, T,_E,TRes,Done) :- equal_object(T,TRes), Done=done.
add_element4(pred_false,T, E,TRes,Done) :- add_element3(T,E,TRes,Done).


:- assert_must_succeed((kernel_objects:add_new_element(int(3),[int(2),int(1)],R),
                        kernel_objects:equal_object(R,[int(1),int(2),int(3)]))).
:- assert_must_succeed((kernel_objects:add_new_element([int(2)],[[int(2),int(1)],[]],R),
                        kernel_objects:equal_object(R,[[],[int(1),int(2)],[int(2)]]))).

% use when you are sure the element to add is not in the set
% to be used for adding elements to an accumulator
:- block add_new_element(?,-,?).
%%add_new_element(E,Set,NewSet) :- add_element(E,Set,NewSet). % TO DO : Improve
add_new_element(E,Set,NewSet) :- 
     is_custom_explicit_set(Set,add_element), add_element_to_explicit_set(Set,E,R),!,
                             %% print(add_new_to_explicit_set(E)),nl, %%
                             equal_object(R,NewSet).
add_new_element(E,Set,NewSet) :- expand_custom_set_to_list(Set,ESet,_,add_new_element),
         add_new_element2(ESet,E,NewSet).

:- block add_new_element2(-,?,?).
add_new_element2([],E,Res) :- var(Res),should_be_converted_to_avl(E),
         construct_avl([E],R),!,equal_object(R,Res).
add_new_element2(S,E,Res) :- equal_cons(Res,E,S).

:- assert_must_succeed(exhaustive_kernel_check(remove_element(int(3),[int(2),int(1),int(3)],[int(1),int(2)]))).
:- assert_must_succeed((kernel_objects:remove_element(fd(1,'Name'),X,[fd(2,'Name'),fd(3,'Name')]),
                        kernel_objects:equal_object(X,global_set('Name')))). 

:- assert_must_succeed((kernel_objects:remove_element(X,global_set('Name'),[fd(2,'Name'),fd(3,'Name')]),
                        X = fd(1,'Name'))).
:- assert_must_succeed((kernel_objects:remove_element(int(1),X,[int(2)]),
                        kernel_objects:equal_object(X,[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:remove_element([int(1),int(2)],X,[]),
                        kernel_objects:equal_object([[int(2),int(1)]],X))).
:- assert_must_fail((kernel_objects:remove_element(int(3),X,_),
                        (X = [int(2),int(1)] ; X=[], X = [int(2)]))).
:- assert_must_fail((kernel_objects:remove_element(int(1),X,Res),
                        X = [int(2),int(1)], (Res=[] ; Res = [_,_|_] ; Res = [int(1)]))).
                        
                        /* remove element is often used in enumerators !!! */

% remove element X from Set, yielding Res
remove_element(X,Set,Res) :- equal_cons(Set,X,Res).
     
     
     
:- assert_must_succeed(exhaustive_kernel_check(remove_element_wf(int(3),[int(3),int(1)],
 [int(1)],_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(remove_element_wf(int(1),[int(3),int(1)],
 [int(1)],_WF))).
:- assert_must_succeed((kernel_objects:remove_element_wf(fd(1,'Name'),X,[fd(2,'Name'),fd(3,'Name')],_WF),
                        kernel_objects:equal_object(X,global_set('Name')))). 
:- assert_must_succeed((kernel_objects:remove_element_wf(int(1),X,[int(2)],_WF),
                        kernel_objects:equal_object(X,[int(2),int(1)]))).
:- assert_must_succeed((     kernel_objects:remove_element_wf(int(1),[int(X),int(2)],R,WF), kernel_waitflags:ground_wait_flags(WF),X==1,R==[int(2)] )).
:- assert_must_succeed((     kernel_objects:remove_element_wf(X,[int(1),int(2)],R,WF), kernel_waitflags:ground_wait_flags(WF),X==int(2),R==[int(1)] )).
:- assert_must_succeed((     kernel_objects:remove_element_wf(X,[pred_true /* bool_true */,pred_false /* bool_false */],R,WF), kernel_waitflags:ground_wait_flags(WF),X==pred_false /* bool_false */,R==[pred_true /* bool_true */] )).

remove_element_wf(X,Set,Res,WF) :- remove_element_wf(X,Set,Res,WF,_DONE).

:- block remove_element_wf(?,-, -,?,?).
remove_element_wf(X,Set,Res,_WF,_DONE) :- Res==[],!, % we know that X must be the only element in Set
  equal_object(Set,[X]).
remove_element_wf(X,Set,Res,WF,DONE) :- 
  remove_element_wf1(X,Set,Res,WF,DONE).

:- block remove_element_wf1(?,-, ?,?,?).
remove_element_wf1(X,avl_set(A),Res,_WF,DONE) :- element_can_be_added_or_removed_to_avl(X),!,
  /* TO DO: try and move the check about whether X can be added to later; when either X is known
     or LWF is instantiated */
    remove_element_from_explicit_set(avl_set(A),X,AR),
    equal_object(AR,Res), DONE=done.
remove_element_wf1(X,Set,Res,WF,DONE) :- /* DONE is ground when element actually removed */
  expand_custom_set_to_list(Set,ESet,_,remove_element_wf1),
  %% nl,print(remove_element_wf1(X,Set,ESet,Res,WF,DONE)),nl,nl, %%
  remove_element_wf2(X,ESet,Res,LWF,DONE),
  %when(nonvar(DONE), print_bt_message(removed(X,ESet,Res,LWF))),
  (DONE==done -> true
   ;  get_cardinality_wait_flag(ESet,remove_element_wf(X,ESet),WF,LWF),
      quick_propagation_element_information(Set,X,WF) % use Set rather than ESet; better if still closure or AVL
   ).

:- block remove_element_wf2(?,-,?,?,?).
remove_element_wf2(H1,[H2|T],Res,LWF,DONE) :-  Res==[],!,
  equal_object(H1,H2),
  remove_element_wf3(pred_true,H1,H2,T,Res,LWF,DONE).
remove_element_wf2(H1,[H2|T],Res,LWF,DONE) :-
  prop_empty_set(T,EqRes),
  equality_objects_lwf(H1,H2,EqRes,LWF),
 %% print(rem(H1,H2,EqRes,T,Res)),nl,trace, %%
  %%((var(EqRes),var(LWF)) -> print(block_remove_element_wf3(EqRes,H1,H2,T,Res)),nl ; true),
  remove_element_wf3(EqRes,H1,H2,T,Res,LWF,DONE).
   /* important for total_bijection that this has higher priority than other expansions */

:- block prop_empty_set(-,?).
% force second argument to pred_true if first arg is empty set
prop_empty_set([],R) :- !,  R=pred_true.
prop_empty_set(_,_).

:- block remove_element_wf3(-,?,?,?,?,-,?).
% remove_element_wf3(EqRes,H1,H2,T,Res,LWF,DONE) :- print(remove_element_wf3(EqRes,H1,H2,T,Res,LWF,DONE)),nl,fail.
remove_element_wf3(pred_true,_H1,_H2,T,Res,_LWF,DONE) :-  equal_object(T,Res),DONE=done.
remove_element_wf3(pred_false,E,H,T,Res,LWF,DONE) :- 
     equal_object([H|RT],Res),
     remove_element_wf2(E,T,RT,LWF,DONE).
 
 /* the same as above: but do not remove if infinite or closure */

:- block remove_element_wf_if_not_infinite_or_closure(?,-,?,?,?,?).
 remove_element_wf_if_not_infinite_or_closure(X,Set, Res,WF,LWF,Done) :- 
        (dont_expand(Set)
         ->  check_element_of_wf(X,Set,WF), equal_object(Res,Set),
             Done=true % or should we wait until X known ?
             %(var(Res)->Res=Set ; equal_object(Res,Set))
          ;  expand_custom_set_to_list(Set,ESet,_,remove_element_wf_if_not_infinite_or_closure),
             remove_element_wf2(X,ESet,Res,LWF,Done)
        ).

:- use_module(bmachine_construction,[external_procedure_used/1]).
dont_expand(global_set('STRING')) :- !, 
    (external_procedure_used(_) -> true). % we could check if there is a STRING generating procedure involved
    % unless we use external functions, there is *no* way that new strings can be generated from a B machine !
    % Hence: we can expand STRING safely and thus avoid infinite enumeration of partial functions, ...
    % example:  procs : STRING +-> {"waiting"} &  card( dom(procs) ) = 6  thus fails quickly
dont_expand(Set) :- is_infinite_explicit_set(Set).
dont_expand(Set) :- is_non_expanded_closure(Set).

:- assert_must_succeed((kernel_objects:remove_exact_element([int(1),int(2)],X,[]),
                        X = [[int(1),int(2)]])).               
:- assert_must_fail((kernel_objects:remove_exact_element([int(1),int(2)],X,[]),
                        X = [[int(2),int(1)]])).
/* not used for the moment */                        
:- block remove_exact_element(?,-,?).
remove_exact_element(X,Set,Res) :- 
  expand_custom_set_to_list(Set,ESet,Done,remove_exact_element),
  remove_exact_element1(Done,X,ESet,Res).

:- block remove_exact_element1(-,?,?,?).
remove_exact_element1(_,X,S,R) :- remove_exact_element2(X,S,R).

:- block remove_exact_element2(?,-,?).
remove_exact_element2(H,[H|T],T).
remove_exact_element2(E,[H|T],[H|RT]) :- not_equal_object(E,H),
     remove_exact_element2(E,T,RT).
     
     
:- assert_must_succeed((kernel_objects:remove_exact_first_element([int(1),int(2)],X,[[]]),
                        X = [[int(1),int(2)],[]])).
:- assert_must_succeed((kernel_objects:remove_exact_first_element(X,global_set('Name'),T),
		         X==fd(1,'Name'),T==[fd(2,'Name'),fd(3,'Name')])).
:- assert_must_fail((kernel_objects:remove_exact_first_element([[]],X,_),
                        X = [[int(1),int(2)],[]])).  
                        
:- assert_must_succeed((kernel_objects:remove_exact_first_element(X,C,R),
      kernel_objects:gen_test_interval_closure(1,2,C),
                        X == int(1), R == [int(2)] )).  
   
gen_test_interval_closure(From,To,CL) :-
   CL=closure(['_zzzz_unary'],[integer],b(member( b(identifier('_zzzz_unary'),integer,[]),
 b(interval(b(value(int(From)),integer,[]),
            b(value(int(To)),integer,[])),set(integer),[])),pred,[])).
 
:- block remove_exact_first_element(?,-,?).
remove_exact_first_element(X,Set,Res) :- remove_exact_first_element1(Set,X,Res).

remove_exact_first_element1([],_,_) :- fail.
remove_exact_first_element1([H|T],H,T).
remove_exact_first_element1(avl_set(A),H,T) :- remove_minimum_element_custom_set(avl_set(A),H,T).
remove_exact_first_element1(global_set(GS),H,T) :- 
   remove_minimum_element_custom_set(global_set(GS),H,T).
remove_exact_first_element1(closure(P,Types,B),H,T) :- 
   remove_minimum_element_custom_set(closure(P,Types,B),H,T).




:- assert_must_succeed((kernel_objects:delete_element_wf(fd(1,'Name'),X,[fd(2,'Name'),fd(3,'Name')],_WF),
                        X = global_set('Name'))).
:- assert_must_succeed((kernel_objects:delete_element_wf(int(1),X,[int(2)],_WF),
                        X = [int(2),int(1)])).
:- assert_must_succeed((kernel_objects:delete_element_wf([int(1),int(2)],X,[],_WF),
                        X = [[int(2),int(1)]])).
:- assert_must_succeed((kernel_objects:delete_element_wf(int(3),X,[int(2),int(1)],_WF),
                        X = [int(2),int(1)])).
:- assert_must_succeed((kernel_objects:delete_element_wf(int(1),X,X,_WF),
                        X = [])).
:- assert_must_fail((kernel_objects:delete_element_wf(int(X),[int(1)],[int(1)],_WF),
                        X = 1)).
                        
                        /* WARNING: only use when R is not instantiated by something else;
                           (except for R=[]) */

:- block delete_element_wf(?,-,?,?).
delete_element_wf(X,Set,Res,_WF) :- delete_element(X,Set,Res). /* as X can be in or not in Set:
 no information can be gained by enumerating; WF not needed */


                           
:- assert_must_succeed((kernel_objects:delete_element(fd(1,'Name'),X,[fd(2,'Name'),fd(3,'Name')]),
                        X = global_set('Name'))).
:- assert_must_succeed((kernel_objects:delete_element(int(1),X,[int(2)]),
                        X = [int(2),int(1)])).
:- assert_must_succeed((kernel_objects:delete_element([int(1),int(2)],X,[]),
                        X = [[int(2),int(1)]])).
:- assert_must_succeed((kernel_objects:delete_element(int(3),X,[int(2),int(1)]),
                        X = [int(2),int(1)])).
     
:- block delete_element(?,-,?).
delete_element(X,Set,Res) :- 
  ground(X),
  try_expand_and_convert_to_avl_with_check(Set,ESet,delete_element),!,
  delete_element0(X,ESet,Res).
delete_element(X,Set,Res) :- delete_element1(X,Set,Res).
  
:- block delete_element0(?,-,?).
delete_element0(X,ESet,Res) :-
  ((is_custom_explicit_set(ESet,delete_element),
    delete_element_from_explicit_set(ESet,X,DS))
     -> equal_object(DS,Res)
     ; delete_element1(X,ESet,Res)
  ).
 
delete_element1(X,Set,Res) :- expand_custom_set_to_list(Set,ESet,_,delete_element1),
  %check_is_expanded_set(ESet,delete_element2),
  delete_element2(ESet,X,Res).

:- block delete_element2(-,?,?).
delete_element2([],_,[]). /* same as above, but allow element to be absent */
delete_element2([H2|T],E,R) :-
    equality_objects(H2,E,EqRes),
    delete_element3(EqRes,H2,T,E,R).
    %when_sufficiently_instantiated(E,H2,delete_element3(H1,[H2|T],R)). /* added by Michael Leuschel, 16/3/06 */

:- block delete_element3(-,?,?,?,?).
delete_element3(pred_true,_H2,T,_,R) :- equal_object(R,T).
delete_element3(pred_false,H2,T,E,Res) :- equal_cons(Res,H2,RT),delete_element2(T,E,RT).


:- assert_must_succeed(kernel_objects:check_is_expanded_set([int(1)],test)).

:- public check_is_expanded_set/2.
check_is_expanded_set(X,Source) :- 
  ((nonvar(X),(X=[] ; X= [_|_])) -> true
    ; add_internal_error('Is not expanded set: ',check_is_expanded_set(X,Source))
  ).
  

/* union/3 */

:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(3)],[int(2),int(1),int(3)],[int(1),int(3),int(2)]))).	
:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(1)],[int(1),int(2)],[int(1),int(2)]))).	
:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(3)],[int(2),int(1)],[int(1),int(3),int(2)]))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(3),int(2)],[int(2),int(1)],[int(1),int(3),int(2)]))). 
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],union([int(3),int(4)],[int(2),int(1)],[int(1),int(3),int(2)]))). 
:- assert_must_succeed((kernel_objects:union([int(1)],[int(2)],Res),kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union([],[int(2)],Res),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([int(2)],[],Res),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([int(2)],[int(2)],Res),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([int(1)],Res,[int(1),int(2)]),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([fd(1,'Name')],X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,X))).
:- assert_must_succeed((kernel_objects:union(X,global_set('Name'),Res),X=[fd(2,'Name'),fd(1,'Name')],
                         kernel_objects:equal_object(Res,global_set('Name')))).
:- assert_must_succeed((kernel_objects:union([fd(1,'Name')],[fd(3,'Name'),fd(2,'Name')],Res),
                         kernel_objects:equal_object(Res,global_set('Name')))).
%:- assert_must_succeed((kernel_objects:union([fd(1,'Name')],[fd(3,'Name'),fd(2,'Name')],Res),
%                         kernel_objects:equal_object(Res,X),X=global_set('Name'))).
:- assert_must_fail((kernel_objects:union([int(1)],[int(2)],Res),
            (kernel_objects:equal_object(Res,[_]);kernel_objects:equal_object(Res,[_,_,_|_])))).
:- assert_must_fail((kernel_objects:union([int(1)],[int(1)],Res),(Res=[];kernel_objects:equal_object(Res,[_,_|_])))).
:- assert_must_fail((kernel_objects:union([fd(1,'Name')],[fd(2,'Name')],Res),
                         kernel_objects:equal_object(Res,global_set('Name')))).
% kernel_objects:union([int(1),int(2)],X,[int(1),int(2),int(3)])

union(S1,S2,Res) :- init_wait_flags(WF), union_wf(S1,S2,Res,WF), ground_wait_flags(WF).

:- block union_wf(-,-,-,?).
union_wf(Set1,Set2,Res,_WF) :- Set1==[],!,equal_object(Set2,Res).
union_wf(Set1,Set2,Res,_WF) :- Set2==[],!,equal_object(Set1,Res).
union_wf(Set1,Set2,Res,_WF) :- Res==[],!,empty_set(Set1), empty_set(Set2).
union_wf(Set1,Set2,Res,WF) :- union0(Set1,Set2,Res,WF).

:- block union0(-,-,?,?), union0(-,?,-,?), union0(?,-,-,?). % require two arguments to be known
union0(Set1,Set2,Res,_WF) :- Set1==[],!,equal_object(Set2,Res).
union0(Set1,Set2,Res,_WF) :- Set2==[],!,equal_object(Set1,Res).
union0(Set1,Set2,Res,_WF) :- Res==[],!,empty_set(Set1), empty_set(Set2).
union0(Set1,Set2,Res,WF) :- (var(Set1) -> union1(Set2,Set1,Res,WF) ; union1(Set1,Set2,Res,WF)).

union1(Set1,Set2,Res,WF) :- 
   %print_term_summary(union1(Set1,Set2,Res,WF)),nl,
   try_expand_and_convert_to_avl_unless_large_or_closure(Set1,ESet1),
   try_expand_and_convert_to_avl_unless_large_or_closure(Set2,ESet2),
   union1e(ESet1,ESet2,Res,WF). %, print(union1e(ESet1,ESet2,Res,WF)),nl.

try_expand_and_convert_to_avl_unless_large_or_closure(Set,ESet) :- (var(Set);Set=closure(_,_,_)),!,ESet=Set.
try_expand_and_convert_to_avl_unless_large_or_closure(Set,ESet) :-
    try_expand_and_convert_to_avl_unless_large(Set,ESet).

union1e(Set1,Set2,Res,_WF) :-    % print_term_summary(union1e(Set1,Set2,Res)),
    is_custom_explicit_set(Set1,union1e),  %print(try_union(Set1,Set2)),nl,
    union_of_explicit_set(Set1,Set2,Union), !,
    %print_term_summary(explicit_set_union(Union)),
    equal_object(Union,Res).
union1e(Set2,Set1,Res,WF) :-  % Set2=avl_set(_),  nonvar(Set1), Set1 \= avl_set(_),
     nonvar(Set1), Set1=avl_set(_), Set2 \= avl_set(_), \+ ground(Set2),
     !, % avoid expanding Set2
     % print_term_summary(union_invert_arguments(Set2,Set1,Res)),
     expand_custom_set_to_list(Set1,ESet1,_,union1e_1), 
     union2(ESet1,Set2,Res,WF), lazy_check_subset_of(Set2,Res,WF).
union1e(Set1,Set2,Res,WF) :-    
     expand_custom_set_to_list(Set1,ESet1,_,union1e_2), % we could avoid this expansion by treating avl_set,... below in union2
     union2(ESet1,Set2,Res,WF),
     lazy_check_subset_of(Set2,Res,WF) % could perform additional constraint checking
   %%,nl,  print(union2_result(ESet1,Set2,Res)),nl,nl
    % ,try_prop_card_leq(ESet1,Res), try_prop_card_leq(Set2,Res).   %%% seems to slow down ProB: investigate
    .

lazy_check_element_of(H,Set,WF) :- nonvar(Set), 
   (Set=closure(_,_,_) ; Set=avl_set(_)),!,  % print(lazy_check(H,Set)),nl,
   check_element_of_wf(H,Set,WF).
lazy_check_element_of(_,_,_).

:- block lazy_check_subset_of(-,?,?), lazy_check_subset_of(?,-,?).
%lazy_check_subset_of(A,B) :- print(lazy_check_subset(A,B)),nl,fail.
lazy_check_subset_of(Set1,Set2,WF) :- nonvar(Set2), 
   (Set2=closure(_,_,_) ; Set2=avl_set(_)),!, lazy_check_subset_of2(Set1,Set2,WF).
lazy_check_subset_of(_,_,_). % ignore other set representations
:- block lazy_check_subset_of2(-,?,?).
lazy_check_subset_of2([],_,_WF) :- !.
lazy_check_subset_of2([H|T],Set,WF) :- !, check_element_of_wf(H,Set,WF), lazy_check_subset_of2(T,Set,WF).
lazy_check_subset_of2(_,_,_). % ignore other set representations

:- block union2(-,?,?,?).
%union2(A,B,R,_) :- print(union2(A,B,R)),nl,fail.
union2([],S,Res,_WF) :- equal_object(S,Res).
union2([H|T],Set2,Res,WF) :-  
   (T\==[],nonvar(Set2), Set2=[H2|T2], T2==[]  % minor optimisation for improved propagation; e.g., for x:S & S<:1..13 & S \/ {x} = S2 & x/: S2
   % the constraint is not yet detected straight away: x:S & S<:1..12 & S \/ {x} /= S
     -> union3(H2,T2,[H|T],Res,WF)
     ;  union3(H,T,Set2,Res,WF)
   ).
union3(H,T,Set2,Res,WF) :-  
      add_element(H,Set2,R,Done), %print(add_element(H,Set2,R,Done)),nl,
      lazy_check_element_of(H,Res,WF),  % TO DO: propagate constraint that H is in Res
      (T==[] -> union2(T,R,Res,WF) ; union4(Done,T,R,Res,WF)).
:- block union4(-,?,?,?,?).
union4(_Done,T,R,Res,WF) :- union2(T,R,Res,WF).

/* not used at the moment   :- block union_2(-,?,?).
union_2([],S,Res) :- equal_object(S,Res).
union_2([H|T],Set2,Res) :- delete_element(H,Set2,NSet2), 
   equal_cons(Res,H,TailRes), %equal_object([H|TailRes],Res), % use equal_cons_wf ?
   union_2(T,NSet2,TailRes). */

:- assert_must_succeed(exhaustive_kernel_check(union_generalized([[int(3)],[int(2),int(1),int(3)]],[int(1),int(3),int(2)]))).	
:- assert_must_succeed(exhaustive_kernel_check(union_generalized([[int(3),int(2)],[],[int(2),int(1),int(3)]],[int(1),int(3),int(2)]))).	
:- assert_must_succeed(exhaustive_kernel_fail_check(union_generalized([[int(3)],[int(3),int(4)],[int(2),int(1),int(3)]],[int(1),int(3),int(2)]))).	
:- assert_must_succeed((kernel_objects:union_generalized([[]],Res),Res=[])).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2)]],Res),
         kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2),int(1)]],Res),
         kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2),int(1)],[],[int(2)]],Res),
         kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2)],X],Res),
                        kernel_objects:equal_object(X,Res), X = [int(2),int(1),int(3)])).
:- assert_must_succeed((kernel_objects:union_generalized([global_set('Name'),X,X,X],Res),
                        kernel_objects:equal_object(global_set('Name'),Res), X = [fd(2,'Name'),fd(1,'Name')])).
:- assert_must_succeed((kernel_objects:union_generalized([X,global_set('Name')],Res),
                        kernel_objects:equal_object(global_set('Name'),Res), X = [fd(2,'Name'),fd(1,'Name')])).
:- assert_must_fail((kernel_objects:union_generalized([[int(1)],[int(2)]],Res),(Res=[_];
         kernel_objects:equal_object(Res,[_,_,_|_])))).
:- assert_must_fail((kernel_objects:union_generalized([[int(1)],[int(1)]],Res),(Res=[];
         kernel_objects:equal_object(Res,[_,_|_])))).

union_generalized(S,Res) :- init_wait_flags(WF), union_generalized_wf(S,Res,WF), ground_wait_flags(WF).

:- block union_generalized_wf(-,?,?).
union_generalized_wf(SetsOfSets,Res,WF) :-  % print(call_union_generalized(SetsOfSets,Res)),nl,
        expand_custom_set_to_list(SetsOfSets,ESetsOfSets,_,union_generalized_wf),  %print(exp(ESetsOfSets)),nl,
        union_generalized2(ESetsOfSets,[],Res,WF). %, print(res(Res)),nl.

:- block union_generalized2(-,?,?,?).
union_generalized2([],S,Res,_WF) :-  equal_object(S,Res).
union_generalized2([H|T],UnionSoFar,Res,WF) :- union_wf(H,UnionSoFar,UnionSoFar2,WF),
     %%  print_message(called_union(H,UnionSoFar,UnionSoFar2)), %%
     union_generalized2(T,UnionSoFar2,Res,WF).

      
:- assert_must_succeed(exhaustive_kernel_check([commutative],intersection([int(3)],[int(2),int(1),int(3)],[int(3)]))).	
:- assert_must_succeed(exhaustive_kernel_check([commutative],intersection([int(4),int(3),int(2)],[int(2),int(1),int(3)],[int(2),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],intersection([int(4),int(3),int(2)],[],[]))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],intersection([int(1),int(3)],[int(4),int(3),int(2)],[]))).
:- assert_must_succeed((kernel_objects:intersection(Y,X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,Y), Y =[fd(1,'Name')])).
:- assert_must_succeed((kernel_objects:intersection([int(1)],[int(2)],Res),Res=[])).
:- assert_must_succeed((kernel_objects:intersection([int(1)],[int(2)],[]))).
:- assert_must_fail((kernel_objects:intersection([int(1),int(4),int(3)],[int(2),int(3)],[]))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(1)],_))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(1)],[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(1)],[int(1),int(2)]))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(3)],Res),
         kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:intersection([int(2)],[int(2)],Res),
         kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:intersection([int(2),int(3)],[int(3),int(4),int(2)],Res),
         kernel_objects:equal_object(Res,[int(2),int(3)]))).
:- assert_must_fail((kernel_objects:intersection([int(1)],[int(2)],Res),(
         kernel_objects:equal_object(Res,[_|_])))).
:- assert_must_fail((kernel_objects:intersection([int(1)],[int(1)],Res),(Res=[];
         kernel_objects:equal_object(Res,[_,_|_])))).
:- assert_must_fail((kernel_objects:intersection([fd(1,'Name')],X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,X))).


intersection(S1,S2,Res) :- init_wait_flags(WF), intersection(S1,S2,Res,WF), ground_wait_flags(WF).

:- block intersection(-,-,?,?).
intersection(Set1,Set2,Res,_WF) :- (Set1==[] ; Set2==[]),!, empty_set(Res).
intersection(Set1,Set2,Res,WF) :-  % now we need to know at least a bit about both Set1 and Set2; at least given the current code below; TO DO: infer that {x} /\ s = {x}  =>  x:s
   % print(intersection0(Set1,Set2,Res,WF)),nl,
   intersection0(Set1,Set2,Res,WF).

:- block intersection0(-,?,?,?), intersection0(?,-,?,?).
intersection0(Set1,Set2,Res,_WF) :- (Set1==[] ; Set2==[]),!, empty_set(Res).
intersection0(Set1,Set2,Res,WF) :- 
   try_expand_and_convert_to_avl_unless_large(Set1,ESet1),
   try_expand_and_convert_to_avl_unless_large(Set2,ESet2),
   intersection1(ESet1,ESet2,Res,WF).

intersection1(Set1,Set2,Res,_WF) :- nonvar(Set1),is_custom_explicit_set(Set1,intersection),
     %% print_term_summary(try_inter(Set1,Set2,Res)),nl, %%
  intersection_of_explicit_set(Set1,Set2,Inter), !,
     %print_term_summary(explicit_set_inter(Set1,Set2,Inter)), %%
  equal_object(Inter,Res).
intersection1(Set1,Set2,Res,WF) :- 
     %%  print_term_summary(intersection(Set1,Set2,Res) ), %%
    (Res==[] -> % print_term_summary(disjoint(Set1,Set2)),
                disjoint_sets(Set1,Set2,WF)
     ;
       (swap_set(Set1,Set2)  ->  intersection2(Set2,Set1,Res,WF)
                              ;  intersection2(Set1,Set2,Res,WF))
    ).
   
swap_set(Set1,_Set2) :- var(Set1),!.
swap_set(_Set1,Set2) :- var(Set2),!,fail.
%swap_set(_Set1,Set2) :- is_infinite_explicit_set(Set2),!,fail.
swap_set(avl_set(_),Set2) :- Set2 \= avl_set(_), Set2 \= [], 
   Set2 \= closure(_,_,_), Set2 \= global_set(_). % if it was a small closure,  intersection_of_explicit_set should have triggered
swap_set(closure(_P,_T,_B),Set2) :- ok_to_swap(Set2). % TO DO: for two closures: we could try and use the smallest one as first argument to intersection2
swap_set(global_set(_GS),Set2) :- ok_to_swap(Set2).

ok_to_swap(global_set(GS)) :- !, \+ is_infinite_or_very_large_explicit_set(global_set(GS),1000000).
ok_to_swap(closure(P,T,B)) :- !,\+ is_infinite_or_very_large_explicit_set(closure(P,T,B),1000000).
ok_to_swap(_).
% maybe also use is_efficient_custom_set as below ??


intersection2(Set1,Set2,Res,WF) :- % print_term_summary(expand_inter1(Set1,Set2,Res,WF)),
                                expand_custom_set_to_list(Set1,ESet1,_,intersection2),
                                % print_term_summary(expanded_inter1(Set1)),
                                intersection3(ESet1,Set2,Res,WF).
:- block intersection3(-,?,?,?).
intersection3([],_,Res,_WF) :- empty_set(Res).
intersection3([H|T],Set,Res,WF) :- 
    (Res==[] -> %print(inter_res_empty(H,T,Set)),nl,
                not_element_of_wf(H,Set,WF),intersection3(T,Set,Res,WF) 
      ; membership_test(Set,H,MemRes),
        %print(mem_test(H,MemRes)),nl, %
        intersection4(MemRes,H,T,Set,Res,WF)
    ).

:- block intersection4(-,?,?, ?,?,?).
intersection4(pred_true,H,T,Set,Result,WF) :- %print(inter4_mem_obj(H)),nl,
                              %check_element_of(H,Set), 
                              equal_object([H|Res],Result),
                              intersection3(T,Set,Res,WF).
intersection4(pred_false,_H,T,Set,Res,WF) :- %not_element_of(H,Set), 
                              intersection3(T,Set,Res,WF).
 
 
:- assert_must_succeed(exhaustive_kernel_check_wfdet(disjoint_sets([int(5)],[int(2),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(disjoint_sets([int(5)],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(disjoint_sets([int(5),int(2)],[int(6),int(1),int(3)],WF),WF)).

distjoint_sets(S1,S2) :- 
   init_wait_flags(WF),disjoint_sets(S1,S2,WF),ground_wait_flags(WF).
:- block disjoint_sets(-,?,?), disjoint_sets(?,-,?).
disjoint_sets(S1,S2,WF) :- 
    % print_term_summary(disjoint_sets(S1,S2)),
    % TO DO: we could provide faster code for two avl sets or intervals; but probably caught in intersection code above?
        ((S1==[];S2==[]) -> true
         ; is_efficient_custom_set(S2) -> expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_1),
                                          disjoint_sets2(ESet1,S2,WF)
         ; is_efficient_custom_set(S1) -> expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_2),
                                          disjoint_sets2(ESet2,S1,WF)
         ; expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_3),
           %expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_4),
           disjoint_sets2(ESet1,S2,WF)
        ).
 
 % TO DO: we could infer some constraints on the possible max sizes of the sets
 % for finite types (sum of size must be <= size of type)
:- block disjoint_sets2(-,?,?).
disjoint_sets2([],_,_WF).
disjoint_sets2([H|T],S2,WF) :- not_element_of_wf(H,S2,WF), disjoint_sets2(T,S2,WF).

% NOT YET USED: not_disjoint_sets could be used for S /\ R /= {}
:- assert_must_succeed(exhaustive_kernel_check_wfdet(not_disjoint_sets([int(3)],[int(2),int(1),int(3)],WF),WF)).	
:- block not_disjoint_sets(-,?,?), not_disjoint_sets(?,-,?).
not_disjoint_sets(S1,S2,WF) :- 
        ((S1==[];S2==[]) -> fail
         ; is_efficient_custom_set(S2) -> expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_1),
                                          not_disjoint_sets2(ESet1,S2,WF)
         ; is_efficient_custom_set(S1) -> expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_2),
                                          not_disjoint_sets2(ESet2,S1,WF)
         ; expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_3),
           %expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_4),
           not_disjoint_sets2(ESet1,S2,WF)
        ).

:- block not_disjoint_sets2(-,?,?).
not_disjoint_sets2([],_,_WF).
not_disjoint_sets2([H|T],S2,WF) :- membership_test(S2,H,MemRes), not_disjoint3(MemRes,T,S2,WF).

:- block not_disjoint3(-,?,?,?).
not_disjoint3(pred_true,_,_,_).
not_disjoint3(pred_false,T,S2,WF) :- not_disjoint_sets2(T,S2,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(intersection_generalized_wf([[int(3)],[int(2),int(1),int(3)]],[int(3)],unknown,WF),WF)).	
:- assert_must_succeed(exhaustive_kernel_check_wfdet(intersection_generalized_wf([[int(3),int(2)],[int(2),int(1),int(3)],[int(4),int(3)]],[int(3)],unknown,WF),WF)).	
:- assert_must_succeed((kernel_objects:intersection_generalized_wf(avl_set(node(avl_set(node(fd(1,'Name'),true,1,empty,node(fd(2,'Name'),true,0,empty,empty))),
       true,1,empty,node(avl_set(node(fd(2,'Name'),true,1,empty,node(fd(3,'Name'),true,0,empty,empty))),true,0,empty,empty))),
       avl_set(node(fd(2,'Name'),true,0,empty,empty)),unknown,_WF))).      
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(1)],[int(2)]],Res,unknown,_WF),Res=[])).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(1)],[int(2),int(1)]],Res,unknown,_WF),
         kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(1)],X,[int(2),int(3),int(1)]],Res,unknown,_WF),
                       X = [int(2),int(1)], 
         kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([X,X,[int(2),int(3),int(1)]],Res,unknown,_WF),
                       X = [int(2),int(1)], kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(2),int(1),int(3)],X,[int(1),int(2)],X],Res,unknown,_WF),
                        kernel_objects:equal_object(X,Res), X = [int(2),int(1)])).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([global_set('Name'),X],Res,unknown,_WF),
                        kernel_objects:equal_object(X,Res), X = [fd(2,'Name'),fd(1,'Name')])).
:- assert_must_fail((kernel_objects:intersection_generalized_wf([[int(1)],[int(2)]],Res,unknown,_WF),(
         kernel_objects:equal_object(Res,[_|_])))).
:- assert_must_fail((kernel_objects:intersection_generalized_wf([[int(1)],[int(1)]],Res,unknown,_WF),(Res=[];
         kernel_objects:equal_object(Res,[_,_|_])))).
:- assert_must_abort_wf(kernel_objects:intersection_generalized_wf([],_R,unknown,WF),WF).

% code for general_intersection
:- block intersection_generalized_wf(-,?,?,?).
intersection_generalized_wf(SetsOfSets,Res,Span,WF) :-
  expand_custom_set_to_list(SetsOfSets,ESetsOfSets,_,intersection_generalized_wf),
  intersection_generalized2(ESetsOfSets,Res,Span,WF).
        
intersection_generalized2([],Res,Span,WF) :-  /* Atelier-B manual requires argument to inter to be non-empty */
   add_wd_error_set_result('inter applied to empty set','',Res,[],Span,WF).
intersection_generalized2([H|T],Res,_Span,WF) :- intersection_generalized3(T,H,Res,WF).       
:- block intersection_generalized3(-,?,?,?).
intersection_generalized3([],SoFar,Res,_WF) :- equal_object(SoFar,Res).
intersection_generalized3([H|T],InterSoFar,Res,WF) :-
      intersection(H,InterSoFar,InterSoFar2,WF), 
      intersection_generalized3(T,InterSoFar2,Res,WF).
      
:- assert_must_succeed(exhaustive_kernel_check(difference_set([int(3),int(2)],[int(2),int(1),int(3)],[]))).
:- assert_must_succeed(exhaustive_kernel_check(difference_set([int(3),int(2)],[int(2),int(1),int(4)],[int(3)]))).	
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[int(1),int(2)]],[]),
                        kernel_objects:equal_object(SSS,[[int(2),int(1)]]))).
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[int(1),int(2)]],R), kernel_objects:equal_object(R,[]),
                        kernel_objects:equal_object(SSS,[[int(2),int(1)]]))).
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[fd(1,'Name'),fd(2,'Name')]],R),
                        kernel_objects:equal_object(R,[]),
                        kernel_objects:equal_object(SSS,[[fd(2,'Name'),fd(1,'Name')]]))).
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[int(1),int(2)]],[]),
                        kernel_objects:equal_object(SSS,[[int(1),int(2)]]))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[int(1)],_))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[int(2)],_))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[int(2)],[int(1)]))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[],[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:difference_set([],[int(1),int(2)],[]))).
:- assert_must_succeed((kernel_objects:difference_set(Y,X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,[]), Y =[fd(1,'Name')])).
:- assert_must_succeed((kernel_objects:difference_set(X,Y,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,[fd(3,'Name'),fd(1,'Name')]), Y =[fd(2,'Name')])).
:- assert_must_fail((kernel_objects:difference_set(X,Y,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,[]), Y =[fd(1,'Name'),fd(2,'Name')])).
:- assert_must_fail((kernel_objects:difference_set(Y,X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,Y), Y =[fd(1,'Name')])).

difference_set(Set1,Set2,Res) :- init_wait_flags(WF),
  difference_set_wf(Set1,Set2,Res,WF),
  ground_wait_flags(WF).

:- block difference_set_wf(-,-,?,?).
difference_set_wf(Set1,_,Res,_WF) :- Set1==[],!,empty_set(Res). 
difference_set_wf(Set1,Set2,Res,_WF) :- Set2==[],!,equal_object(Set1,Res). 
difference_set_wf(Set1,Set2,Res,WF) :- difference_set1(Set1,Set2,Res,WF).


:- block difference_set1(?,-,-,?), difference_set1(-,?,-,?).
difference_set1(Set1,Set2,Res,_WF) :- nonvar(Set1),is_custom_explicit_set(Set1,difference_set),
   difference_of_explicit_set(Set1,Set2,Diff), !,
   %print_term_summary(explicit_set_diff(Set1,Set2,Diff)),%
   equal_object(Diff,Res).
difference_set1(Set1,Set2,Res,_WF) :- Set2==[],!,equal_object(Set1,Res). 
difference_set1(Set1,Set2,Res,WF) :- Res==[],!, check_subset_of_wf(Set1,Set2,WF). 
difference_set1(Set1,Set2,Res,WF) :- %print(diff(Set1,Set2,Res)),nl,%
  expand_custom_set_to_list(Set1,ESet1,_,difference_set1), %print_term_summary(expanded(Set1,ESet1)),
  compute_diff(ESet1,Set2,Res),
  %print_term_summary(computed_diff(ESet1,Set2,Res)),nl,%
  propagate_into2(Res,ESet1,Set2,WF).  %print(propagated(ESet1,Set2)),nl.
 
:- block compute_diff(-,?,?).
compute_diff([],_Set2,Res) :- empty_set(Res).
compute_diff([H|T],Set2,Res) :-
  membership_test(Set2,H,MemRes),compute_diff2(MemRes,H,T,Set2,Res).

:- block compute_diff2(-,?,?,?,?).
compute_diff2(pred_true,_H,T,Set2,Res) :- compute_diff(T,Set2,Res).
compute_diff2(pred_false,H,T,Set2,Res) :- equal_object([H|R2],Res),
  compute_diff(T,Set2,R2).
 
% propagate all elements from one set into another one; do not use for computation; may skip elements ...
/* this version not used at the moment: 
:- block propagate_into(-,?,?).
propagate_into(_,Set2,_WF) :- nonvar(Set2),
    is_custom_explicit_set(Set2,propagate_into),!. % second set already fully known
propagate_into([],_,_WF) :- !.
propagate_into([H|T],Set,WF) :- !,check_element_of_wf(H,Set,WF), propagate_into(T,Set,WF).
propagate_into(Set1,Set2,WF) :- is_custom_explicit_set(Set1,propagate_into),!,
   (is_infinite_explicit_set(Set1) -> true ;
    expand_custom_set_to_list(Set1,ESet1), propagate_into(ESet1,Set2,WF)). */

:- block propagate_into2(-,?,?,?).
propagate_into2(_,Set2,_NegSet,_WF) :- nonvar(Set2),
    is_custom_explicit_set(Set2,propagate_into),!. % second set already fully known
propagate_into2([],_,_,_WF) :- !.
propagate_into2([H|T],PosSet,NegSet,WF) :- !, 
  check_element_of_wf(H,PosSet,WF), 
  not_element_of_wf(H,NegSet,WF),propagate_into2(T,PosSet,NegSet,WF).
propagate_into2(Set1,PosSet,NegSet,WF) :- is_custom_explicit_set(Set1,propagate_into),!,
   (is_infinite_explicit_set(Set1) -> true ;
    expand_custom_set_to_list(Set1,ESet1,_,propagate_into2), propagate_into2(ESet1,PosSet,NegSet,WF)).
 

strict_subset_of(X,Y) :-
    init_wait_flags(WF),
    strict_subset_of_wf(X,Y,WF),
    ground_wait_flags(WF).

:- assert_must_succeed(exhaustive_kernel_check(strict_subset_of_wf([int(3),int(2)],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_check(strict_subset_of_wf([],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_check(strict_subset_of_wf([],[ [] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([int(3),int(2),int(1)],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([int(1),int(4)],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([[]],[],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([],[],_))).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:strict_subset_of(Y,X), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [[int(1),int(2)]], X=[[int(2)],[int(2),int(1)]])).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [fd(1,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [fd(3,'Name'),fd(2,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [fd(1,'Name'),fd(3,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(1),int(2)], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(2)], X=[int(2)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(2)], X=[int(1)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [], X=[int(1)])).

  
%:- block strict_subset_of_wf(-,-,?).
strict_subset_of_wf(Set1,Set2,_WF) :- Set1==[],!,not_empty_set(Set2).
%strict_subset_of_wf(Set1,Set2,WF) :- var(Set2),nonvar(Set1), print(subs(Set1,Set2)),nl,fail.
strict_subset_of_wf(Set1,Set2,WF) :- %print_quoted(check_subset_of(Set1,Set2)),nl,
   not_empty_set_wf(Set2,WF),
   get_cardinality_powset_wait_flag(Set2,WF,LWF),
   when(((ground(LWF),(nonvar(Set1);ground(Set2)))  ;  (nonvar(Set1),nonvar(Set2))  ),
     (both_global_sets(Set1,Set2,G1,G2)
       -> %(print(check_strict_subset_of_global_sets(G1,G2)),nl,
           check_strict_subset_of_global_sets(G1,G2)
        ;  ((var(Set1),ground(Set2))
               ->  expand_custom_set_to_list(Set2,ESet2,_,strict_subset_of_wf), %print(gen_strict_subsets(ESet2)),nl,
                   gen_strict_subsets(Set1,ESet2)
                ; strict_subset_of0(Set1,Set2,WF,LWF)
      ) ) ).
   

%:- block strict_subset_of0(-,?,?,?).  % required to wait: we know Set2 must be non-empty, but Set1 could be an avl-tree or closure
strict_subset_of0(Set1,Set2,WF,LWF) :- %not_empty_set(Set2),
      expand_custom_set_to_list(Set1,ESet1,_,strict_subset_of0),
      (ESet1==[] -> true %not_empty_set(Set2) /* special treatment, so that we know we must have explicit set for RemS below */
             ;  (try_expand_custom_set(Set2,ESet2),
                 %%try_prop_card_lt(ESet1,ESet2), try_prop_card_gt(ESet2,ESet1),
               %  print(strict_subset_of2(ESet1,[],ESet2,WF,LWF)),nl,
                 strict_subset_of2(ESet1,[],ESet2,WF,LWF))
       ).

:- block strict_subset_of2(-,?,?,?,-).
%strict_subset_of2(S,SoFar,Set2,WF) :- print(strict_subset_of2(S,SoFar,Set2,WF)),nl,fail.
strict_subset_of2([],_,RemS,WF,_LWF) :- not_empty_set_wf(RemS,WF). /* we know it must be explicit set */
strict_subset_of2([H|T],SoFar,Set2,WF,LWF) :- var(Set2),!,
    Set2 = [H|Set2R], 
    add_new_element(H,SoFar,SoFar2), %was SoFar2 = [H|SoFar],
    strict_subset_of2(T,SoFar2,Set2R,WF,LWF).
strict_subset_of2([H|T],SoFar,Set2,WF,LWF) :- 
 % when_sufficiently_for_member(H,Set2,WF,
     remove_element_wf(H,Set2,RS2,WF), not_empty_set_wf(RS2,WF),
     not_element_of_wf(H,SoFar,WF),  /* consistent((H,SoFar)), necessary? */
     when((nonvar(T) ; (ground(LWF),ground(RS2))),
           (add_new_element(H,SoFar,SoFar2), %SoFar2 = [H|SoFar],
            strict_subset_of2(T,SoFar2,RS2,WF,LWF) )).
     



:- assert_must_succeed(exhaustive_kernel_check(partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_check(partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_check(partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5)],[int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)], [int(3)] ],_))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ], _))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2)],[ [int(2)], [int(1)], [] ], _))).
:- assert_must_fail((kernel_objects:partition_wf([int(1),int(2)],[ [int(2)], [int(1),int(2)] ], _))).
:- assert_must_fail((kernel_objects:partition_wf([int(1),int(3)],[ [int(1)], [int(2)] ], _))).
:- assert_must_fail((kernel_objects:partition_wf([int(1),int(2),int(3)],[ [int(1)], [int(2)] ], _))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1)],[S1,S2],_WF), S1=[H|T], S2==[],T==[],H==int(1))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2)],[S1,S2],_WF), S1=[H|T], S2=[int(1)],(preferences:preference(use_clpfd_solver,true) -> T==[],H==int(2) ; T=[],H=int(2)))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2),int(3)],[S1,S2,S3],_WF), S1=[H2|T], S3=[int(3)],T=[H1|TT],H2=int(2),TT==[],S2==[],H1==int(1))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2),int(3)],[[int(1)],X,[int(2)]],_WF),
			X==[int(3)])).

:- use_module(bsets_clp,[disjoint_union_generalized_wf/3]).
:- block partition_wf(?,-,?).
partition_wf(Set,ListOfSets,WF) :- % print_term_summary(partition_wf(Set,ListOfSets,_WF)),nl, %
   (ground(Set),find_non_ground_set(ListOfSets,NGS,Rest) -> 
       % print(backwards_computation_of_partition(NGS)),nl, %
       bsets_clp:disjoint_union_generalized_wf(Rest,RestSet,WF),
       difference_set(Set,RestSet,NGS)
     ; bsets_clp:disjoint_union_generalized_wf(ListOfSets,Set,WF) 
   ),
   all_disjoint(ListOfSets,WF).  
 
 :- assert_must_succeed((kernel_objects:find_non_ground_set([x,y,A,z],B,C), B==A,C==[x,y,z])).
find_non_ground_set([H|T],NG,Rest) :-
  (ground(H) -> Rest=[H|TR], find_non_ground_set(T,NG,TR)
    ; ground(T),NG=H, Rest=T).
 
:- block all_disjoint(-,?).
all_disjoint(closure(P,T,B),WF) :-
   expand_custom_set(closure(P,T,B),ExpandedSet), all_disjoint(ExpandedSet,WF).
all_disjoint(avl_set(AVL),WF) :-
   expand_custom_set(avl_set(AVL),ExpandedSet), all_disjoint(ExpandedSet,WF).
% no case for global_set: cannot be a set of sets
all_disjoint([],_WF).
all_disjoint([H|T],WF) :- all_disjoint_with(T,H,WF),
   all_disjoint(T,WF).

:- block all_disjoint_with(-,?,?).
all_disjoint_with([],_,_WF).
all_disjoint_with([H|T],Set1,WF) :- disjoint_sets(Set1,H,WF), all_disjoint_with(T,Set1,WF).


/* not used anymore
% TO DO: use clpfd all_different for integers !?
:- block all_different(-,?).
%all_different(integer,Set) :- get_integer_list(Set,IntList), clpfd_alldifferent(IntList).
all_different(ListOfValues,_WF) :- debug_println(9,all_different(ListOfValues)),
   all_different2(ListOfValues,[]).
:- block all_different2(-,?).
all_different2([],_).
all_different2([H|T],SoFar) :- all_different3(SoFar,H), all_different2(T,[H|SoFar]).
all_different3([],_).
all_different3([H|T],X) :- not_equal_object(H,X), all_different3(T,X).
*/


:- assert_must_succeed(exhaustive_kernel_fail_check(not_partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5)],[int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_check(not_partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)], [int(3)] ],_))).
:- assert_must_fail((kernel_objects:not_partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ], _))).
:- assert_must_fail((kernel_objects:not_partition_wf([int(1),int(2)],[ [int(2)], [int(1)], [] ], _))).
:- assert_must_succeed((kernel_objects:not_partition_wf([int(1),int(2)],[ [int(2)], [int(1),int(2)] ], _))).
:- assert_must_succeed((kernel_objects:not_partition_wf([int(1),int(3)],[ [int(1)], [int(2)] ], _))).
:- assert_must_succeed((kernel_objects:not_partition_wf([int(1),int(2),int(3)],[ [int(1)], [int(2)] ], _))).

:- block not_partition_wf(?,-,?).
not_partition_wf(FullSet,ListOfSets,_WF) :- 
  expand_custom_set_to_list(ListOfSets,EListOfSets,_,not_partition_wf), % necessary ? ListOfSets is actually list 
  not_partition_wf2(EListOfSets,[],FullSet).

:- block not_partition_wf2(-,?,?).
%not_partition_wf2(Sets,SoFar,_) :- print(not_part2(Sets,SoFar)),nl,fail.
not_partition_wf2([],ElementsSoFar,FullSet) :- not_equal_object(ElementsSoFar,FullSet).
not_partition_wf2([Set1|Rest],ElementsSoFar,FullSet) :-
      expand_custom_set_to_list(Set1,ESet1,_,not_partition_wf2), not_partition_wf3(ESet1,ElementsSoFar,Rest,FullSet).

:- block not_partition_wf3(-,?,?,?).
not_partition_wf3([],S,OtherSets,FullSet) :-
   not_partition_wf2(OtherSets,S,FullSet). % finished treating this set
not_partition_wf3([H|T],ElementsSoFar,OtherSets,FullSet) :-
   membership_test(ElementsSoFar,H,MemRes),
   not_partition_wf4(MemRes,H,T,ElementsSoFar,OtherSets,FullSet).
   
:- block not_partition_wf4(-,?,?,?,?,?).
not_partition_wf4(pred_true,_,_,_,_,_). % Not disjoint
not_partition_wf4(pred_false,H,T,ElementsSoFar,OtherSets,FullSet) :-
   add_element(H,ElementsSoFar,ElementsSoFar2), % we could also already check whether H in FullSet or not
   not_partition_wf3(T,ElementsSoFar2,OtherSets,FullSet).

:- assert_must_succeed(exhaustive_kernel_succeed_check(check_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(check_subset_of([int(1),int(2),int(5)],[int(2),int(5),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(check_subset_of([int(1),int(3),int(5)],[int(2),int(5),int(1)]))).
:- assert_must_succeed((kernel_objects:power_set(global_set('Name'),PS),kernel_objects:check_subset_of(X,PS),
                        kernel_objects:equal_object(X,[[fd(2,'Name'),fd(1,'Name')]]))).
:- assert_must_succeed(findall(X,kernel_objects:check_subset_of(X,[[int(1),int(2)],[]]),[_1,_2,_3,_4])).
:- assert_must_succeed((kernel_objects:check_subset_of(X,[[int(1),int(2)],[]]),
                        nonvar(X),
                        kernel_objects:equal_object(X,[[int(2),int(1)]]))).
:- assert_must_succeed((kernel_objects:check_subset_of_wf(Y,X,_WF), Y = [fd(1,'Name')],
           nonvar(X),X=[H|T], var(T), H==fd(1,'Name'), X=Y)).
:- assert_must_succeed((kernel_objects:check_subset_of(Y,X), Y = [fd(1,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:check_subset_of(Y,X), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:check_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:sample_closure(C),kernel_objects:check_subset_of(C,global_set('NAT')))).
:- assert_must_succeed((kernel_objects:check_subset_of(global_set('NAT'),global_set('NAT')))).
:- assert_must_succeed((kernel_objects:check_subset_of(global_set('NAT'),global_set('NATURAL')))).
:- assert_must_fail((kernel_objects:check_subset_of(global_set('NAT'),global_set('NATURAL1')))).
:- assert_must_fail((kernel_objects:check_subset_of(global_set('NAT'),global_set('NAT1')))).
:- assert_must_fail((kernel_objects:check_subset_of(X,Y), Y = [fd(1,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
  /* TO DO: add special treatemt for closures and type checks !! */
  
check_subset_of(Set1,Set2) :- init_wait_flags(WF),
  check_subset_of_wf(Set1,Set2,WF),
  ground_wait_flags(WF).

check_finite_subset_of_wf(Set1,Set2,WF) :-
   check_subset_of_wf(Set1,Set2,WF),
   is_finite_set_wf(Set1,WF).

:- block check_subset_of_wf(-,-,?).
check_subset_of_wf(Set1,Set2,WF) :-    % print_term_summary(check_subset_of(Set1,Set2)),nl,
   (both_global_sets(Set1,Set2,G1,G2)
     -> ( %print(check_subset_of_global_sets(G1,G2)),nl,
         check_subset_of_global_sets(G1,G2))
      ;  check_subset_of0(Set1,Set2,WF)
   ).
     
both_global_sets(S1,S2,G1,G2) :- nonvar(S1),nonvar(S2),
  is_global_set(S1,G1), is_global_set(S2,G2).
  
% check if we have a global set or interval
% is_global_set([],R) :- !, R=interval(0,-1). % useful ???
is_global_set(global_set(G1),R) :- !,
   (convert_integerset_to_interval(G1,R) -> true ; R=G1).
is_global_set(Closure,R) :-
    custom_explicit_sets:is_interval_closure(Closure,Low,Up),!,
    R=interval(Low,Up).
convert_integerset_to_interval('INT',interval(MININT,MAXINT)) :-
   (preferences:preference(maxint,MAXINT),
    preferences:preference(minint,MININT)) -> true.
convert_integerset_to_interval('NAT',interval(0,MAXINT)) :-
   (preferences:preference(maxint,MAXINT)->true).
convert_integerset_to_interval('NAT1',interval(1,MAXINT)) :-
   (preferences:preference(maxint,MAXINT)->true).


:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(-200,1000),'INTEGER')).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets('NATURAL','NATURAL')).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets('NATURAL','INTEGER')).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets('NATURAL1','NATURAL')).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets('NATURAL1','INTEGER')).

check_subset_of_global_sets(X,Y) :- (var(X) ; var(Y)), 
   add_internal_error('Illegal call: ',check_subset_of_global_sets(X,Y)),fail.
check_subset_of_global_sets(interval(Low1,Up1),interval(Low2,Up2)) :- !,
   interval_subset(Low1,Up1,Low2,Up2).
check_subset_of_global_sets(X,X) :- !. % both args must be atomic and ground
check_subset_of_global_sets(X,Y) :- check_strict_subset_of_global_sets(X,Y).

:- block interval_subset(-,?,?,?), interval_subset(?,-,?,?).
interval_subset(Low1,Up1,_,_) :- safe_less_than(Up1,Low1). %Set 1 is empty.
interval_subset(Low1,Up1,Low2,Up2) :-
    safe_less_than_equal(Low1,Up1), % Set 1 is not empty
    less_than_equal_direct(Low2,Low1), less_than_equal_direct(Up1,Up2).

:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,2),interval(1,3))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,1),interval(1,2))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,1),interval(0,1))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(2,1),interval(33,34))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(3,1),interval(4,2))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(3,1),interval(2,1))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(1,2),interval(1,2))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(1,2),interval(2,3))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(2,3),interval(1,2))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(0,1000),'NATURAL')).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,1000),'NATURAL1')).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(-200,1000),'INTEGER')).
check_strict_subset_of_global_sets('NATURAL','INTEGER').
check_strict_subset_of_global_sets('NATURAL1','INTEGER').
check_strict_subset_of_global_sets('NATURAL1','NATURAL').
check_strict_subset_of_global_sets(interval(_,_),'INTEGER').
check_strict_subset_of_global_sets(interval(Low,Up),'NATURAL') :- 
    check_strict_subset_natural(Low,Up,0).
check_strict_subset_of_global_sets(interval(Low,Up),'NATURAL1') :- 
    check_strict_subset_natural(Low,Up,1).
check_strict_subset_of_global_sets(interval(Low1,Up1),interval(Low2,Up2)) :- 
	check_strict_subset_intervals(Low1,Up1,Low2,Up2).

:- block check_strict_subset_natural(-,?,?), check_strict_subset_natural(?,-,?).
check_strict_subset_natural(Low,Up,NatLowLim) :- 
    safe_less_than(Up,Low) % interval 1 empty
    ;
    safe_less_than_equal(Low,Up),
    safe_less_than_equal(NatLowLim,Low).

check_strict_subset_intervals(Low1,Up1,Low2,Up2) :- 
    less_than_equal_direct(Low2,Up2), % Low2..Up2 not empty
    check_strict_subset_intervals2(Low1,Up1,Low2,Up2).
:- block check_strict_subset_intervals2(-,?,?,?),check_strict_subset_intervals2(?,-,?,?),
   check_strict_subset_intervals2(?,?,-,?).
check_strict_subset_intervals2(Low1,Up1,_,_) :- safe_less_than(Up1,Low1). % interval 1 empty
check_strict_subset_intervals2(Low1,Up1,Low2,Up2) :-
    safe_less_than_equal(Low1,Up1), % interval 1 not empty
    ( safe_less_than(Low2,Low1), less_than_equal_direct(Up1,Up2)
      ;
      Low1=Low2,less_than_direct(Up1,Up2)
    ).

:- use_module(custom_explicit_sets,[is_definitely_maximal_set/1,singleton_set/2]).

check_subset_of0(Set1,_Set2,_WF) :- Set1==[],!.
check_subset_of0(Set1,Set2,_WF) :- Set2==[],
  %nonvar(Set2),Set2=[], %var(Set1),
  !,% print(checking_empty_set(Set1)),nl,
  empty_set(Set1). %  ,print(empty),nl.
check_subset_of0(_Set1,Set2,_WF) :- 
  nonvar(Set2),is_definitely_maximal_set(Set2),!. %  , print(subset_maximal(_Set1,Set2)),nl.
  %singleton
check_subset_of0(Set1,Set2,WF) :-  custom_explicit_sets:singleton_set(Set1,El),!,
   check_element_of_wf(El,Set2,WF).
check_subset_of0(Set1,Set2,_WF) :- subset_of_explicit_set(Set1,Set2,Code),!,
    % print(subset_explicit(Set1,Set2,Code)),nl,
   call(Code).
check_subset_of0(Set1,Set2,_) :- Set1==Set2,!.
check_subset_of0(Set1,Set2,WF) :- % TO DO: move into custom_explicit_sets
   %print(check_sub(Set1,Set2)),nl,
   nonvar(Set2),is_cartesian_product_closure(Set2,S21,S22),!,
   % TO DO: maybe don't do this if Set1 is avl_set ??
   debug_println(9,'Applying C <: S21*S22 <=> C : S21 <-> S22'),
   %print(sub_cart(Set1,S21,S22)),nl,
   bsets_clp:relation_over_wf(Set1,S21,S22,WF).
   %print('Applying C <: S21*S22 <=> dom(C) <:S21 & ran(C) <:S22'),nl,
   %bsets_clp:domain_wf(Set1,Dom,WF), % TO DO: do not compute domain for lists
   %check_subset_of_wf(Dom,S21,WF),
   %bsets_clp:range_wf(Set1,Ran,WF), % TO DO: do not compute range for lists
   %check_subset_of_wf(Ran,S22,WF).
check_subset_of0(Set1,Set2,WF) :- get_cardinality_powset_wait_flag(Set2,WF,LWF), 
    % print(check_subset_of0(Set1,Set2,WF)),nl,
   when((nonvar(Set1);(nonvar(LWF),ground(Set2))),check_subset_of0_lwf(Set1,Set2,WF,LWF)).

:- use_module(custom_explicit_sets,[is_infinite_or_very_large_explicit_set/2]).
check_subset_of0_lwf(Set1,Set2,_WF,_LWF) :- Set2==[],!,
   empty_set(Set1).
check_subset_of0_lwf(Set1,Set2,WF,LWF) :-
    (nonvar(Set1) ; nonvar(Set2),dont_expand_this_explicit_set(Set2)),!,
    expand_custom_set_to_list(Set1,ESet1,_,check_subset_of1),
    try_expand_and_convert_to_avl_unless_large(Set2,ESet2),
    % print(check_subset(ESet1,Set2,WF,LWF)),nl,
    % b_interpreter_components:observe_instantiation(ESet1,'ESet1',ESet1),
    check_subset_of2(ESet1,[],ESet2,WF,LWF,none).
check_subset_of0_lwf(Set1,Set2,_WF,_LWF) :- 
     expand_custom_set_to_list(Set2,ESet2,_,check_subset_of0_lwf), % Set2 is ground
     gen_subsets(Set1,ESet2).

:- block check_subset_of2(-,?,?,?,-, ?).
check_subset_of2([],_SoFar,_Set2,_WF,_LWF,_Last).
% :- print_tabs, print('SOLUTION '),  hashing:my_term_hash((_SoFar,_Set2),Hash), print(Hash),nl, (Hash = 30707821301826039 -> trace ; true).
check_subset_of2(HT,SoFar,Set2,WF,LWF,Last) :-  %print(chk(H,Set2,Last)),nl,
      ((var(HT),Set2 = avl_set(AVL))
        -> % the value is chosen by the enumerator
           %trace_point(check_subset_of2(HT,SoFar,Set2,WF,LWF,Last)),
           %print_tabs,print(' --> enum '), translate:print_bvalue(SoFar),nl,
           custom_explicit_sets:safe_avl_member(H,AVL),
           % this forces H to be ground; if Last /= none then it will be ground
           (Last==none -> true ; Last @< H), 
           % TO DO: we could write a safe_avl_member_greater_than(H,Last,AVL)
           %print_tabs,translate:print_bvalue(H),nl, %(H==avl_set(node([],true,0,empty,empty)) -> trace ; true),
           not_element_of_wf(H,SoFar,WF),
           NewLast=H,
           HT = [H|T]
        ;  % the value may have been chosen by somebody else or will not be enumerated in order below
           %trace_point(normal_checksubset(HT,SoFar,Set2,WF,LWF,Last)),
           HT = [H|T],
           not_element_of_wf(H,SoFar,WF),
           check_element_of_wf_lwf(H,Set2,WF,LWF),
           %check_element_of_wf(H,Set2,WF),
           
           NewLast = Last
      ),
      %print(check(H,Set2,WF,LWF,Last)),nl,
      check_subset_of3(H,T,SoFar,Set2,WF,LWF,NewLast).
      
% TO DO: write specific subsets code for avl_set(Set2) + try expand when becomes ground; merge with enumerate_tight_set ,...
% TO DO: ensure that it also works with global_set(T) instead of avl_set(_) or with interval closures


:- block check_subset_of3(?,-,-,?,?,-,?), check_subset_of3(?,-,?,-,?,-,?), check_subset_of3(?,-,-,-,?,?,?).
check_subset_of3(_,T,_,_Set2,_WF,_LWF,_) :- T==[],!.
check_subset_of3(H,T,SoFar,Set2,WF,LWF,Last) :- var(T),!,
    % Sofar, Set2 and LWF must be set
    when((nonvar(T);(ground(Set2),ground(H),ground(SoFar))),
                      (T==[] -> true
                       ; add_new_element(H,SoFar,SoFar2),  %SoFar2 = [H|SoFar],
                         check_subset_of2(T,SoFar2,Set2,WF,LWF,Last))).
check_subset_of3(H,T,SoFar,Set2,WF,LWF,Last) :-
    % T must be set and not equal to []
    T = [H2|T2],
	add_new_element(H,SoFar,SoFar2),  %SoFar2 = [H|SoFar],
    %check_subset_of2(T,SoFar2,Set2,WF,LWF))),
    check_element_of_wf(H2,Set2,WF),
    not_element_of_wf(H2,SoFar2,WF),  
    check_subset_of3(H2,T2,SoFar2,Set2,WF,LWF,Last).

/*
:- block check_subset_of2(-,?,?,?).
check_subset_of2([],_SoFar,_S,_WF).
check_subset_of2([H|T],SoFar,Set2,WF) :- 
      membership_test(Set2,H,MemRes),
      check_subset_of3(MemRes,H,T,SoFar,Set2,WF).
      
:- block check_subset_of3(-,?,?,?,?,?).
check_subset_of3(pred_true,H,T,SoFar,Set2,WF) :- check_subset_of2(T,[H|SoFar],Set2,WF).
*/
:- block gen_subsets(?,-).
gen_subsets([],_).
gen_subsets([H|T],Set) :- %print(gen_subsets(H,T,Set)),nl,
   ordered_delete(DH,Set,NewSet),
   equal_object(DH,H),
   gen_subsets(T,NewSet).

ordered_delete(H,[H|T],T).
ordered_delete(H,[_|T],R) :- ordered_delete(H,T,R).
    
gen_strict_subsets(T,[_H2|T2]) :- gen_subsets(T,T2).
gen_strict_subsets([H|T],[H2|T2]) :- equal_object(H,H2),
    gen_strict_subsets(T,T2).
    
:- assert_must_succeed(exhaustive_kernel_check_wf(check_finite_non_empty_subset_of_wf([int(1),int(5)], [int(2),int(5),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(check_finite_non_empty_subset_of_wf([int(1),int(5)], [int(5),int(1)],WF),WF)).
check_finite_non_empty_subset_of_wf(Set1,Set2,WF) :-
   check_non_empty_subset_of_wf(Set1,Set2,WF),
   is_finite_set_wf(Set1,WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(check_non_empty_subset_of_wf([int(1),int(5)], [int(2),int(5),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(check_non_empty_subset_of_wf([int(2)], [int(5),int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(check_non_empty_subset_of_wf([], [int(1)],WF),WF)).

check_non_empty_subset_of_wf(S1,S2,WF) :- not_empty_set_wf(S1,WF),
   check_subset_of_wf(S1,S2,WF).
         
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_subset_of([int(1),int(2),int(5)], [int(2),int(4),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1),int(3)]))).
:- assert_must_succeed((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name')], X=[fd(2,'Name')])).
:- assert_must_succeed((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name')], X=[fd(1,'Name'),fd(2,'Name')])).
:- assert_must_fail((kernel_objects:not_subset_of(Y,X), Y = [fd(1,'Name'),fd(3,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_subset_of(Y,X), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_subset_of(global_set('NAT'),global_set('NAT')))).
:- assert_must_succeed((kernel_objects:not_subset_of(global_set('NAT'),global_set('NAT1')))).


not_subset_of(Set1,Set2) :- init_wait_flags(WF),
  not_subset_of_wf(Set1,Set2,WF),
  ground_wait_flags(WF).

:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf([int(1),int(2),int(5)], [int(2),int(4),int(1),int(3)],_WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf(global_set('NATURAL'), global_set('INTEGER'),_WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf(global_set('INTEGER'), global_set('INTEGER'),_WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf([int(1)], [],_WF))).

:- block not_finite_subset_of_wf(-,?,?).
not_finite_subset_of_wf(Set1,Set2,WF) :- test_finite_set_wf(Set1,Finite,WF),
   not_finite_subset_of_wf_aux(Finite,Set1,Set2,WF).
:- block not_finite_subset_of_wf_aux(-,?,?,?).
not_finite_subset_of_wf_aux(pred_false,_Set1,_Set2,_WF).
not_finite_subset_of_wf_aux(pred_true,Set1,Set2,WF) :- not_subset_of_wf(Set1,Set2,WF).

:- block not_subset_of_wf(-,?,?).
not_subset_of_wf(Set1,Set2,WF) :-  % print_quoted(not_subset_of_wf(Set1,Set2)),nl,trace,
   (both_global_sets(Set1,Set2,G1,G2) % also catches intervals
     ->  not_subset_of_global_sets(G1,G2)
      ;  not_subset_of_wf1(Set1,Set2,WF)
   ).
not_subset_of_wf1(_Set1,Set2,_WF) :-
   nonvar(Set2), is_definitely_maximal_set(Set2),!,fail.
not_subset_of_wf1(Set1,Set2,WF) :- Set2==[],!, not_empty_set_wf(Set1,WF).
not_subset_of_wf1(Set1,Set2,WF) :-  custom_explicit_sets:singleton_set(Set1,El),!,
   not_element_of_wf(El,Set2,WF).
not_subset_of_wf1(Set1,Set2,WF) :- not_subset_of_explicit_set(Set1,Set2,Code,WF),!,
   call(Code).
not_subset_of_wf1(Set1,Set2,WF) :-
  expand_custom_set_to_list(Set1,ESet1,_,not_subset_of_wf1),
  % print_quoted(not_subset_of2(ESet1,Set2,WF)),nl,
  not_subset_of2(ESet1,Set2,WF).
 
 
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets(interval(0,2),interval(1,3))).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets(interval(1,2),interval(0,-1))).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets(interval(1,2),interval(4,3))).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets(interval(2,4),interval(1,3))).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets(interval(1,9000),interval(2,9999))).
:- assert_must_succeed((kernel_objects:not_subset_of_global_sets(interval(X2,X4),interval(1,3)),
 X2=2, X4=4)).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets(interval(2,4),interval(1,4))).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets(interval(2,4),interval(2,4))).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets(interval(2,4),interval(1,4))).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets(interval(2,4),interval(0,10))).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets(interval(0,2),'NATURAL1')).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets(interval(-1,2),'NATURAL')).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets(interval(1,2),'NATURAL1')).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets(interval(0,2),'NATURAL')).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets(interval(-1,2),'INTEGER')).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets('NATURAL','NATURAL1')).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets('INTEGER','NATURAL1')).
:- assert_must_succeed(kernel_objects:not_subset_of_global_sets('INTEGER','NATURAL')).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets('NATURAL1','INTEGER')).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets('NATURAL1','NATURAL1')).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets('NATURAL1','NATURAL')).
:- assert_must_fail(kernel_objects:not_subset_of_global_sets('NATURAL','NATURAL')).

:- block not_subset_of_global_sets(-,?), not_subset_of_global_sets(?,-).
not_subset_of_global_sets(interval(Low1,Up1),G2) :- !,
   safe_less_than_equal(Low1,Up1), % Set 1 is not empty; otherwise it will always be a subset
   not_subset_interval_gs_aux(G2,Low1,Up1).
not_subset_of_global_sets(G1,G2) :-
    \+ check_subset_of_global_sets(G1,G2).

:- use_module(library(clpfd)).
not_subset_interval_gs_aux(interval(Low2,Up2),Low1,Up1) :- !,
   post_constraint2((Low1 #<Low2 #\/ Up1 #> Up2 #\/ Up2 #< Low1),Posted),
   (Posted==true -> true ;  not_interval_subset(Low1,Up1,Low2,Up2)).
not_subset_interval_gs_aux(GS2,Low1,Up1) :-
   when((nonvar(Low1),nonvar(Up1)), \+ check_subset_of_global_sets(interval(Low1,Up1),GS2)).

:- block not_interval_subset(-,?,?,?), not_interval_subset(?,-,?,?),
         not_interval_subset(?,?,-,?), not_interval_subset(?,?,?,-).
not_interval_subset(Low1,Up1,Low2,Up2) :- \+ interval_subset(Low1,Up1,Low2,Up2).


:- block not_subset_of2(-,?,?).
%not_subset_of2([],_,_WF) :- fail.
not_subset_of2([H|T],Set2,WF) :- 
   membership_test(Set2,H,MemRes),
   not_subset_of3(MemRes,T,Set2,WF).

:- block not_subset_of3(-,?,?,?).
not_subset_of3(pred_false,_T,_Set2,_WF).
not_subset_of3(pred_true,T,Set2,WF)   :- not_subset_of2(T,Set2,WF).



:- assert_must_succeed(exhaustive_kernel_check_wf(not_both_subset_of([int(1),int(2),int(5)], []
,[int(2),int(4),int(1),int(3)],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(not_both_subset_of([int(1),int(2),int(5)], [int(3)],
    [int(2),int(5),int(1),int(3)],[int(1),int(4)],WF),WF)).

not_both_subset_of(Set1A,Set1B, Set2A,Set2B, WF) :- 
  kernel_equality:subset_test(Set1A,Set2A,Result,WF), % not yet implemented ! % TODO ! -> sub_set,equal,super_set
  not_both_subset_of_aux(Result,Set1B,Set2B,WF).

:- block not_both_subset_of_aux(-,?,?,?).
not_both_subset_of_aux(pred_false,_Set1B,_Set2B,_WF).
not_both_subset_of_aux(pred_true,Set1B,Set2B,WF) :- 
    not_subset_of_wf(Set1B,Set2B,WF).

/***********************************/
/* not_strict_subset_of(Set1,Set2) */
/*         Set1 /<<: Set2         */  
/**********************************/


:- assert_must_succeed(exhaustive_kernel_succeed_check(not_strict_subset_of([int(1),int(2),int(5)], [int(2),int(4),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_strict_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_strict_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1),int(3)]))).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [[int(1),int(2)]], X=[[int(2)],[int(2),int(1)]])).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [fd(1,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(1),int(2)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(2)], X=[int(2)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(2)], X=[int(1)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [], X=[int(1)])).                       
    
not_strict_subset_of(Set1,Set2) :- init_wait_flags(WF),
  not_strict_subset_of_wf(Set1,Set2,WF),
  ground_wait_flags(WF).
  
:- block not_strict_subset_of_wf(-,?,?).
not_strict_subset_of_wf(Set1,Set2,WF) :- %print_quoted(check_subset_of(Set1,Set2)),nl,
   (both_global_sets(Set1,Set2,G1,G2)
     -> (%print(check_not_strict_subset_of_global_sets(G1,G2)),nl,
         when((ground(G1),ground(G2)), \+check_strict_subset_of_global_sets(G1,G2)))
      ;  not_strict_subset_of_wf1(Set1,Set2,WF)
   ).
   
not_strict_subset_of_wf1(Set1,Set2,WF) :-
   % OLD VERSION: not_subset_of(Set1,Set2) ; check_equal_object(Set1,Set2).
    expand_custom_set_to_list(Set1,ESet1,_,not_strict_subset_of_wf1),
    not_strict_subset_of2(ESet1,Set2,WF).
  
:- block not_strict_subset_of2(-,?,?).
not_strict_subset_of2([],R,_WF) :- empty_set(R).
not_strict_subset_of2([H|T],Set2,WF) :- 
    membership_test(Set2,H,MemRes),
    not_strict_subset_of3(MemRes,H,T,Set2,WF).

:- block not_strict_subset_of3(-,?,?,?,?).
not_strict_subset_of3(pred_false,_H,_T,_Set2,_WF).
not_strict_subset_of3(pred_true,H,T,Set2,WF) :- 
      remove_element(H,Set2,RS2),not_strict_subset_of2(T,RS2,WF).
                      
/* Set1 /: FIN1(Set2) */     
:- assert_must_succeed((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X=[int(1)], Y=[int(1),int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[])).
:- assert_must_fail((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(1)])).   

:- block not_non_empty_finite_subset_of_wf(-,?,?).
not_non_empty_finite_subset_of_wf(Set1,Set2,WF) :- test_finite_set_wf(Set1,Finite,WF),
   not_non_empty_finite_subset_of_aux(Finite,Set1,Set2,WF).
:- block not_non_empty_finite_subset_of_aux(-,?,?,?).
not_non_empty_finite_subset_of_aux(pred_false,_Set1,_Set2,_WF).
not_non_empty_finite_subset_of_aux(pred_true,Set1,Set2,WF) :- not_non_empty_subset_of_wf(Set1,Set2,WF).
               
/* Set1 /: POW1(Set2) */
:- assert_must_succeed(exhaustive_kernel_check_wf(not_non_empty_subset_of_wf([int(1)], [int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(not_non_empty_subset_of_wf([int(2)], [int(2),int(3)],WF),WF)).
:- assert_must_succeed((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X=[int(1)], Y=[int(1),int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[])).
:- assert_must_fail((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(1)])).

:- block not_non_empty_subset_of_wf(-,-,?).
not_non_empty_subset_of_wf(Set1,Set2,WF) :- % Maybe introduce binary choice point ? 
   empty_set(Set1) ;
   not_subset_of_wf(Set1,Set2,WF).


/* min, max */

:- assert_must_succeed(exhaustive_kernel_check(minimum_of_set([int(1)],int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(minimum_of_set([int(2),int(3),int(1)],int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(minimum_of_set([int(2),int(3),int(1)],int(2),unknown,_WF))).
:- assert_must_succeed((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(1)])).
:- assert_must_succeed((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_fail((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(2), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_abort_wf(kernel_objects:minimum_of_set([],_R,unknown,WF),WF).
%:- must_succeed(kernel_waitflags:assert_must_abort2_wf(kernel_objects:minimum_of_set([],_R,WF),WF)).

:- block minimum_of_set_extension_list(-,?,?,?).
minimum_of_set_extension_list(ListOfValues,int(Min),Span,WF) :-
   minimum_of_set2(ListOfValues,Min,Span,WF).

:- block minimum_of_set(-,?,?,?).
minimum_of_set(Set1,Res,_Span,_WF) :- is_custom_explicit_set(Set1,minimum_of_set),
    min_of_explicit_set(Set1,Min), !,
 %%print_term_summary(explicit_set_min(Set1,Min)), %%
    equal_object(Min,Res).
minimum_of_set(Set1,int(Min),Span,WF) :- expand_custom_set_to_list(Set1,ESet1,_,minimum_of_set),
   minimum_of_set2(ESet1,Min,Span,WF).
:- block minimum_of_set2(-,?,?,?).
minimum_of_set2([],Res,Span,WF) :-
   add_wd_error_set_result('min applied to empty set','',Res,int(0),Span,WF).
minimum_of_set2([int(N)|T],Min,_,_) :- clpfd_geq2(N,Min),minimum_of_set3(T,N,Min,[N]).
   
:- if(current_prolog_flag(version_data,sicstus(4,2,1,_,_))).
:- block minimum_of_set3(-,?,?,?). % with CLPFD: makes sense to also unfold if Min Variable; hence no longer block on : minimum_of_set3(?,-,-).
minimum_of_set3([],MinSoFar,MinSoFar,ListOfValues) :-
   (var(MinSoFar) -> clpfd_minimum(MinSoFar,ListOfValues) ; true). % THIS CRASHES SICSTUS 4.2.0
:- else.
:- block minimum_of_set3(-,?,?,?).
minimum_of_set3([],MinSoFar,MinSoFar,_ListOfValues). 
:- endif.
minimum_of_set3([int(M)|T],MinSoFar,Min,ListOfValues) :- clpfd_geq2(M,Min), 
   minimum(M,MinSoFar,NewMinSoFar),  
   minimum_of_set3(T,NewMinSoFar,Min,[M|ListOfValues]).
   
   
:- block minimum(-,?,?), minimum(?,-,?).
minimum(M1,M2,Min) :- M1<M2 -> Min=M1 ; Min=M2. 

:- assert_must_succeed(exhaustive_kernel_check(maximum_of_set([int(1)],int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(maximum_of_set([int(2),int(3),int(1)],int(3),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(maximum_of_set([int(2),int(3),int(1)],int(2),unknown,_WF))).
:- assert_must_succeed((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(1)])).
:- assert_must_succeed((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(2), Y=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(3), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_fail((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(2), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_fail((preferences:preference(use_clpfd_solver,true),
      kernel_objects:maximum_of_set([int(X),int(_Y)],int(3),unknown,_WF), X = 4)). % in CLPFD modus
:- assert_must_fail((preferences:preference(use_clpfd_solver,true),
      kernel_objects:maximum_of_set([int(_),int(X)],int(3),unknown,_WF), X = 4)).% in CLPFD modus
:- assert_must_abort_wf(kernel_objects:maximum_of_set([],_R,unknown,WF),WF).

:- block maximum_of_set_extension_list(-,?,?,?).
maximum_of_set_extension_list(ListOfValues,int(Max),Span,WF) :-
   maximum_of_set2(ListOfValues,Max,Span,WF).

:- block maximum_of_set(-,?,?,?).
maximum_of_set(Set1,Res,_Span,_WF) :- %print(max(Set1,Res)),nl,
    is_custom_explicit_set(Set1,maximum_of_set),
    max_of_explicit_set(Set1,Max), !,
% print_term_summary(explicit_set_max(Set1,Max)), %%
    equal_object(Max,Res).
maximum_of_set(Set1,int(Max),Span,WF) :-
   expand_custom_set_to_list(Set1,ESet1,_,maximum_of_set),
   maximum_of_set2(ESet1,Max,Span,WF).
:- block maximum_of_set2(-,?,?,?).
maximum_of_set2([],Res,Span,WF) :-
  add_wd_error_set_result('max applied to empty set','',Res,int(0),Span,WF). %preferences:get_preference(maxint,R))). %R=abort(maximum_of_empty_set))).
maximum_of_set2([int(N)|T],Max,_Span,_) :- clpfd_geq2(Max,N), 
   maximum_of_set3(T,N,Max,[N]).

:- if(current_prolog_flag(version_data,sicstus(4,2,1,_,_))).
:- block maximum_of_set3(-,?,?,?). % with CLPFD: makes sense to also unfold if Max Variable; hence no longer block on : maximum_of_set3(?,-,-).
maximum_of_set3([],MaxSoFar,MaxSoFar,ListOfValues) :-
   (var(MaxSoFar) -> clpfd_maximum(MaxSoFar,ListOfValues) ; true). % THIS CRASHES SICSTUS 4.2.0
:- else.
:- block maximum_of_set3(-,?,?,?).
maximum_of_set3([],MaxSoFar,MaxSoFar,_ListOfValues). 
:- endif.
maximum_of_set3([int(M)|T],MaxSoFar,Max,ListOfValues) :- clpfd_geq2(Max,M), 
   maximum(M,MaxSoFar,NewMaxSoFar),  
   maximum_of_set3(T,NewMaxSoFar,Max,[M|ListOfValues]).
   
:- block maximum(-,?,?), maximum(?,-,?).
maximum(M1,M2,Max) :- M1>M2 -> Max=M1 ; Max=M2. 


:- assert_must_succeed((cardinality_of_set_extension_list([fd(1,'Name')],R), R = int(1))).
:- assert_must_succeed((cardinality_of_set_extension_list([int(X),int(Y)],int(1)), X=22, Y==22)).

cardinality_of_set_extension_list(List,int(Card)) :-
  length(List,MaxCard), less_than_equal_direct(Card,MaxCard),
  cardinality_of_set_extension_list2(List,[],0,MaxCard,Card).

:- block cardinality_of_set_extension_list2(-,?,?,?,?).
cardinality_of_set_extension_list2([],_,AccSz,_MaxCard,Res) :- Res=AccSz.
cardinality_of_set_extension_list2([H|T],Acc,AccSz,MaxCard,Res) :-
   % print(card_set_ext([H|T],Acc,AccSz,MaxCard,Res)),nl,
   membership_test(Acc,H,MemRes),
   (MaxCard==Res -> /* only solution is for H to be not in Acc */ MemRes=pred_false
    ; AccSz==Res -> /* only solution is for H to be in Acc */ MemRes=pred_true
    ; (var(Res),var(MemRes)) -> kernel_equality:equality_int(MaxCard,Res,EqMaxC),prop_if_pred_true(EqMaxC,MemRes,pred_false),
                                kernel_equality:equality_int(AccSz,Res,EqAccSz),prop_if_pred_true(EqAccSz,MemRes,pred_true)
    ; true),
   cardinality_of_set_extension_list3(MemRes,H,T,Acc,AccSz,MaxCard,Res).

:- block prop_if_pred_true(-,?,?).
prop_if_pred_true(pred_true,X,X).
prop_if_pred_true(pred_false,_,_).

:- block cardinality_of_set_extension_list3(-,?,?,?,?,?,?).
cardinality_of_set_extension_list3(pred_true,_,T,Acc,AccSz,MaxCard,Res) :-
   % H is a member of Acc, do not increase Acc nor AccSz; however MaxCard now decreases
   less_than_direct(Res,MaxCard), M1 is MaxCard-1, 
   cardinality_of_set_extension_list2(T,Acc,AccSz,M1,Res).
cardinality_of_set_extension_list3(pred_false,H,T,Acc,AccSz,MaxCard,Res) :-
   A1 is AccSz+1, less_than_equal_direct(A1,Res),
   cardinality_of_set_extension_list2(T,[H|Acc],A1,MaxCard,Res).

:- assert_must_succeed(exhaustive_kernel_check(is_finite_set_wf([fd(1,'Name'),fd(2,'Name')],_WF))).
:- assert_must_succeed((is_finite_set_wf(Y,_WF), Y = [])).
:- assert_must_succeed((is_finite_set_wf(Y,_WF), Y = [int(1),int(2)])).
:- use_module(typing_tools,[contains_infinite_type/1]).
:- use_module(custom_explicit_sets,[card_for_specific_custom_set/3]).
:- block is_finite_set_wf(-,?).
is_finite_set_wf([],_WF) :- !.
is_finite_set_wf([_|T],WF) :- !, is_finite_set_wf(T,WF). % what if Tail contains closure ??
is_finite_set_wf(avl_set(_),_WF) :- !.
is_finite_set_wf(closure(_P,T,_B),_WF) :- \+ contains_infinite_type(T),!.
% TO DO: add disjunct, closure1_value...
is_finite_set_wf(closure(P,T,B),WF) :- !, is_finite_set_closure(P,T,B,WF).
is_finite_set_wf(Set,_WF) :- explicit_set_cardinality(Set,Card), dif(Card,inf).

:- use_module(bsyntaxtree,[is_a_disjunct/3]).
is_finite_set_closure(P,T,B,WF) :- is_a_disjunct(B,D1,D2),!,
   is_finite_set_closure(P,T,D1,WF),
   is_finite_set_closure(P,T,D2,WF).
is_finite_set_closure(P,T,B,WF) :- !, 
   (card_for_specific_custom_set(closure(P,T,B),Card,Code) -> call(Code), dif(Card,inf)
     ;  get_enumeration_finished_wait_flag(WF,AWF), % only add warning if indeed we find a solution
        finite_warning(AWF,P,T,B,is_finite_set_closure(P))).

:- assert_must_succeed(exhaustive_kernel_fail_check(is_infinite_set_wf([fd(1,'Name'),fd(2,'Name')],_WF))).
:- assert_must_fail((is_infinite_set_wf(Y,_WF), Y = [int(1),int(2)])).

is_infinite_set_wf(Set,WF) :- test_finite_set_wf(Set,pred_false,WF).

test_finite_set(Set,PREDRES) :- test_finite_set_wf(Set,PREDRES,_WF). % _WF not used below.
:- block test_finite_set_wf(-,?,?).
test_finite_set_wf([],X,_WF) :- !, X=pred_true.
test_finite_set_wf([_|T],X,WF) :- !, test_finite_set_wf(T,X,WF).  % what if Tail contains closure ??
test_finite_set_wf(avl_set(_),X,_WF) :- !, X=pred_true.
test_finite_set_wf(closure(_P,T,_B),X,_WF) :- \+ contains_infinite_type(T), !, X=pred_true.
test_finite_set_wf(closure(P,T,B),X,_WF) :- !, test_finite_closure(P,T,B,X).
test_finite_set_wf(Set,X,_WF) :-  /* also deals with global_set(_) */
   explicit_set_cardinality(Set,Card), set_finite_result(Card,X).
   
:- use_module(bsyntaxtree,[is_a_disjunct/3]).
test_finite_closure(P,T,B,X) :- is_a_disjunct(B,D1,D2),!,
   test_finite_closure(P,T,D1,X1),
   (X1=pred_true -> test_finite_closure(P,T,D2,X)
     ; X = pred_false).
% TO DO: add is_closure1_value_closure 
test_finite_closure(P,T,B,X) :- !,
   (card_for_specific_custom_set(closure(P,T,B),Card,Code)
     -> call(Code), set_finite_result(Card,X)
     ;  finite_warning(now,P,T,B,test_finite_closure(P)),X=pred_true % we assume set to be finite, but print a warning
    ).

:- block set_finite_result(-,?).
set_finite_result(inf,X) :- !, X=pred_false.
set_finite_result(_,pred_true).


:- assert_must_succeed((finite_cardinality_as_int(Y,int(X),_WF), Y = [fd(1,'Name'),fd(2,'Name')],X==2)).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int([int(2),int(4),int(1)],int(3)))).
:- assert_must_succeed((cardinality_as_int(Y,int(2)), Y = [fd(1,'Name'),fd(2,'Name')])).
:- assert_must_succeed((cardinality_as_int(Y,int(2)), 
  nonvar(Y), Y = [_|YY], nonvar(YY), YY=[_])).
:- assert_must_succeed((cardinality_as_int([_|Y],int(3)), 
  nonvar(Y), Y = [_|YY], nonvar(YY), YY=[_])).
:- assert_must_succeed((cardinality_as_int(Y,int(1)), Y = [fd(1,'Name')])).
:- assert_must_succeed((cardinality_as_int(Y,int(0)), Y = [])).
:- assert_must_succeed((cardinality_as_int(X,int(3)), equal_object(X,global_set('Name')))).
:- assert_must_fail((cardinality_as_int(Y,int(X)), Y = [fd(1,'Name'),fd(2,'Name')],dif(X,2))).
:- assert_must_succeed_any((preferences:preference(use_clpfd_solver,false) ;
    cardinality_as_int(S,int(C)), clpfd_interface:post_constraint('#>='(C,2)), nonvar(S),S=[_|T],nonvar(T))).


finite_cardinality_as_int(Set,int(Card),WF) :-  %print(card(Set)),nl,
    cardinality_as_int1(Set,Card,CardValue), %print(card_value(CardValue)),nl,
    check_finite_card(CardValue,Card,Set,WF).

:- block check_finite_card(-,?,?,?).
% check before assigning that the result is not "inf" (otherwise we may trigger arithmetic co-routines before generating the error)
check_finite_card(CardValue,C,Set,WF) :- 
  (CardValue==inf -> add_wd_error('card applied to infinite (or very large) set: ',Set,WF) ; C=CardValue).
   
:- block cardinality_as_int(-,-).
cardinality_as_int(Set,int(Card)) :- cardinality_as_int1(Set,Card,Card).

% cardinality_as_int1(Set, CardValue, ComputedCardValue) : CardValue should be unified with ComputedCardValue afterwards
cardinality_as_int1(Set,Card,ResCard) :- 
   cardinality_as_int2(Set,0,Card,ResCard),
   force_non_empty(Set,0,Card,1). % we could consider using guarding this with a waitflag with priority 1.1

:- block cardinality_as_int2(-,?,-,?).
cardinality_as_int2(X,C,Res,ResultValue) :- % print(card2(X,C,Res,ResultValue)),nl,
   var(X),C==Res,!,empty_set(X),ResultValue=Res. % avoid choice point below
cardinality_as_int2(X,C,Res,ResultValue) :- nonvar(X),  X \= [], X\= [_|_],!, 
   (is_custom_explicit_set(X)
    -> explicit_set_cardinality(X,ESC), add_card(C,ESC,ResultValue)
     ;  add_error_fail(cardinality_as_int2,'First argument not set: ',cardinality_as_int2(X,C,Res))
   ).
cardinality_as_int2([],C,Res,ResultValue) :-  C=ResultValue, Res=ResultValue.
cardinality_as_int2([_H|T],C,Res,ResultValue) :- C1 is C+1,
  % print(card(_H,T,C,Res)),nl,
  (ground(Res) -> safe_less_than_equal(cardinality_as_int2,C1,Res)
                  /* check consistency so far if cardinality provided */
                ; clpfd_geq(Res,C1,_) %,(ground(Res)->print(forced(Res)),nl;true)
  ), 
  force_non_empty(T,C1,Res,1),
  cardinality_as_int2(T,C1,Res,ResultValue).

force_non_empty(Set,CSoFar,TotalCard,LWF) :- var(Set), var(TotalCard),!,
  % print(force_non_empty(CSoFar)),nl,
   clpfd_interface:post_constraint( (TotalCard#=CSoFar) #<=> EmptyR01),
   prop_non_empty(EmptyR01,Set,LWF).
force_non_empty(_,_,_,_).

:- block prop_non_empty(-,-,?).
prop_non_empty(_,X,_) :- nonvar(X),!. % do nothing; cardinality_as_int2 will be called anyway
prop_non_empty(0,X,LWF) :- /* X is var; first arg nonvar */ !, not_empty_set_lwf(X,LWF).
%prop_non_empty(1,X,_). % empty_set not really required: TotalCard is now instantiated; cardinality_as_int2 will get called
prop_non_empty(_,_,_).

:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int_for_wf(global_set('NATURAL'),inf))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int_for_wf([],0))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int_for_wf([int(2)],1))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int_for_wf([int(3),int(1),int(-1),int(100)],4))).
:- assert_must_succeed(exhaustive_kernel_fail_check(cardinality_as_int_for_wf([int(3),int(1),int(-1),int(100)],1000))).
:- assert_must_succeed(exhaustive_kernel_fail_check(cardinality_as_int_for_wf(global_set('NATURAL'),1000))).
% a simpler version without propagation to result; for waitflag priority computation or similar
cardinality_as_int_for_wf(Set,Card) :- cardinality_as_int_for_wf0(Set,0,Card).
:- block cardinality_as_int_for_wf0(-,?,-).
cardinality_as_int_for_wf0(X,C,Res) :-
  (nonvar(X) -> cardinality_as_int_for_wf1(X,C,Res)
    ; Res==inf -> cardinality_as_int_for_inf(X,C)
    ; cardinality_as_int_for_wf2(X,C,Res)).
           
:- block cardinality_as_int_for_inf(-,?).
cardinality_as_int_for_inf(X,C) :- cardinality_as_int_for_wf1(X,C,inf).

cardinality_as_int_for_wf1([],C,Res) :-  !,C=Res.
cardinality_as_int_for_wf1([_H|T],C,Res) :- !,C1 is C+1,
    cardinality_as_int_for_wf0(T,C1,Res).
cardinality_as_int_for_wf1(X,C,Res) :- is_custom_explicit_set(X),!,
    explicit_set_cardinality(X,ESC), add_card(C,ESC,Res).
cardinality_as_int_for_wf1(X,C,Res) :-
   add_internal_error('First arg not set: ',cardinality_as_int_for_wf1(X,C,Res)),fail.

% first argument was var, third argument not inf hence third arg must be set
cardinality_as_int_for_wf2([],C,C).
cardinality_as_int_for_wf2([_H|T],C,Res) :- C<Res, C1 is C+1,
    (var(T) -> cardinality_as_int_for_wf2(T,C1,Res) ; cardinality_as_int_for_wf1(T,C1,Res)).


:- block add_card(-,?,?), add_card(?,-,?).
add_card(inf,_,R) :- !,R=inf.
add_card(_,inf,R) :- !,R=inf.
add_card(X,Y,R) :- R is X+Y.

:- assert_must_succeed(exhaustive_kernel_check(cardinality([],0))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality([int(11)],s(0)))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality([int(11),int(22)],s(s(0))))).
% cardinality as peano number
:- block cardinality(-,-).
cardinality(Set,PCard) :-
 ((nonvar(Set),is_custom_explicit_set(Set,cardinality))
   ->  explicit_set_cardinality(Set,Card), % print(explicit_cardinality(Set,Card)),nl,
       card_convert_int_to_peano(Card,PCard)
   ;   cardinality3(Set,PCard)
 ).

:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(3,s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(2,S),S==s(s(0)))).
:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(X,s(s(s(0)))),X==3)).
:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(X,s(s(s(Y)))),X=4,Y==s(0))).
:- assert_must_fail((kernel_objects:card_convert_int_to_peano(X,s(s(s(_Y)))),X=2)).

:- block card_convert_int_to_peano(-,-).
card_convert_int_to_peano(X,S0) :- var(X), !,
   peel_s(S0,SX,RemS),  %% print(peel_s(S0,SX,RemS)),nl, %%
   (RemS==0 -> X=SX
    ; int_plus(int(X1),int(SX),int(X)),
      greater_than_equal(int(X1),int(0)),
      card_convert_int_to_peano(X1,RemS)).
card_convert_int_to_peano(inf,X) :- !, 
    add_message(cardinality,'*** WARNING: Large or infinite Cardinality; limiting to 100.'),nl,
    convert_int_to_peano(100,X).
card_convert_int_to_peano(X,P) :- convert_int_to_peano(X,P).

peel_s(0,0,0).
peel_s(s(X),Res,SX) :- (var(X) -> Res=1, SX=X ; peel_s(X,RX,SX), Res is RX+1).

:- block cardinality3(-,-).
cardinality3(Set,SC) :- var(Set),!,
  (SC=0 -> Set=[] ; SC=s(C),Set=[_|T],cardinality3(T,C)).
cardinality3([],0).
cardinality3([_|T],s(C)) :- cardinality3(T,C).
cardinality3(avl_set(AVL),Res) :- cardinality(avl_set(AVL),Res).
cardinality3(closure(P,T,B),Res) :- cardinality(closure(P,T,B),Res).




:- assert_must_succeed(exhaustive_kernel_check(card_leq([int(2),int(4),int(1)],s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_leq([int(1),int(2)],s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_leq(global_set('Name'),s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_leq([int(1),int(2)],s(s(0))))).
:- assert_must_fail((kernel_objects:card_leq([int(1),int(2)],s(0)))).
:- assert_must_fail((kernel_objects:card_leq(global_set('Name'),s(s(0))))).
:- block card_leq(-,?). 
card_leq(Set,Card) :- 
 ((nonvar(Set),is_custom_explicit_set(Set,card_leq))
   -> explicit_set_cardinality(Set,CCard), leq_int_peano(CCard,Card)
   ;  card_leq2(Set,Card) ).
:- block card_leq2(-,?).
card_leq2([],_C).
card_leq2([_|T],s(C)) :- card_leq2(T,C). %when(nonvar(T),card_leq2(T,C)).
leq_int_peano(0,_).
leq_int_peano(inf,_) :- !,fail.
leq_int_peano(X,s(C)) :- X>0, X1 is X-1, leq_int_peano(X1,C).



:- assert_must_succeed(exhaustive_kernel_check(card_geq([int(2),int(4),int(1)],s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_geq(global_set('Name'),s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_geq([int(1),int(2)],s(s(0))))).
:- assert_must_succeed((kernel_objects:card_geq([int(1),int(2)],s(0)))).
:- assert_must_fail((kernel_objects:card_geq(global_set('Name'),s(s(s(s(0))))))).
:- assert_must_fail((kernel_objects:card_geq([int(1),int(2)],s(s(s(0)))))).
:- block card_geq(-,-).
card_geq(Set,Card) :- 
 ((nonvar(Set),is_custom_explicit_set(Set,card_geq))
   -> explicit_set_cardinality(Set,CCard), geq_int_peano(CCard,Card)
   ;  card_geq2(Set,Card) ).
:- block card_geq2(?,-).
card_geq2(_,0).
card_geq2([_|T],s(C)) :- card_geq2(T,C).

:- block geq_int_peano(-,-).
geq_int_peano(_,0).
geq_int_peano(X,s(C)) :- geq_int_peano1(X,C).
:- block geq_int_peano1(-,?).
geq_int_peano1(inf,_) :- !.
geq_int_peano1(X,C) :- X>0, X1 is X-1, geq_int_peano(X1,C).

:- block convert_int_to_peano(-,?).
convert_int_to_peano(X,Y) :- convert_int_to_peano2(X,Y).
convert_int_to_peano2(inf,_).
convert_int_to_peano2(X,R) :- number(X),
 (X>100000 -> print('*** Warning: converting large integer to peano: '),print(X),nl ; true),
 convert_int_to_peano3(X,R).
convert_int_to_peano3(0,0).
convert_int_to_peano3(X,s(P)) :- 
  (X>0 -> X1 is X-1, convert_int_to_peano3(X1,P) 
   ; X<0 -> add_error_and_fail(convert_int_to_peano,'Negative nr cannot be converted to peano: ',X)
  ).

% not used:
%:- block convert_peano_to_int(-,?).
%convert_peano_to_int(0,0).
%convert_peano_to_int(s(P),X) :- convert_peano_to_int(P,X1), X is X1+1.

:- assert_must_succeed((kernel_objects:cardinality_greater_equal(Set,set(integer),int(X),integer,_WF), X=3,
 nonvar(Set),Set=[_|S2],nonvar(S2),S2=[_|S3],nonvar(S3),S3=[_|S4],var(S4) )).
:- assert_must_succeed((kernel_objects:cardinality_greater(Set,set(integer),int(X),integer,_WF), X=2,
 nonvar(Set),Set=[_|S2],nonvar(S2),S2=[_|S3],nonvar(S3),S3=[_|S4],var(S4) )).
   /* special predicates called for e.g. card(Set)>X */
cardinality_greater(Set,TypeSet,int(X),_,_WF) :- 
   kernel_objects:max_cardinality(TypeSet,MaxCard), 
   %print(check_less(X,MaxCard)),nl,
   (number(MaxCard) -> less_than(int(X),int(MaxCard)) ; true),
   card_greater2(Set,X).
:- block card_greater2(?,-).
card_greater2(Set,X) :-
    (X<0 -> true
      ; convert_int_to_peano(X,Peano),
        card_geq(Set,s(Peano))).
cardinality_greater_equal(Set,TypeSet,int(X),_,_WF) :- 
   kernel_objects:max_cardinality(TypeSet,MaxCard), 
   %print(check_less_equal(X,MaxCard)),nl,
   (number(MaxCard) -> less_than_equal(int(X),int(MaxCard)) ; true),
   card_greater_equal2(Set,X).
:- block card_greater_equal2(?,-).
card_greater_equal2(Set,X) :-
    (X<1 -> true
     ; convert_int_to_peano(X,Peano),
       card_geq(Set,Peano)).


 
%is_cartesian_pair_or_times(P,X,Y) :- is_cartesian_pair(P,X,Y).
%is_cartesian_pair_or_times(int(Z),int(X),int(Y)) :- times(int(X),int(Y),int(Z)).
    
is_cartesian_pair_wf((X,Y),XType,YType,WF) :- 
   check_element_of_wf(X,XType,WF), check_element_of_wf(Y,YType,WF).
   
:- assert_must_succeed((kernel_objects:not_is_cartesian_pair((int(1),int(1)),
           [int(1),int(2)],[int(2),int(3)],_WF))).
:- assert_must_succeed((kernel_objects:not_is_cartesian_pair((int(3),int(1)),
           [int(1),int(2)],[int(2),int(3)],_WF))).
:- assert_must_fail((kernel_objects:not_is_cartesian_pair((int(1),int(3)),
           [int(1),int(2)],[int(2),int(3)],_WF))).
:- assert_must_succeed((kernel_objects:not_is_cartesian_pair((X,int(3)),
           [int(1),int(2)],[int(2),int(3)],_WF),X=int(4))).


not_is_cartesian_pair((X,Y),XType,YType,WF) :- 
   not_is_cartesian_pair0(X,Y,XType,YType,WF).

:- block not_is_cartesian_pair0(-,-,?,?,?).
not_is_cartesian_pair0(X,Y,XType,YType,WF) :-
   (nonvar(X) -> not_is_cartesian_pair1(X,Y,XType,YType,WF)
    ; not_is_cartesian_pair1(Y,X,YType,XType,WF)).
 
not_is_cartesian_pair1(X,Y,XType,YType,WF) :-
  %print(member_ship_test(XType,X,MemResX)),nl,
  membership_test(XType,X,MemResX),
  (var(MemResX) -> membership_test(YType,Y,MemResY) ; true),
  not_is_cartesian_pair3(MemResX,X,XType,MemResY,Y,YType,WF).

:- block not_is_cartesian_pair3(-,?,?, -,?,?, ?).
not_is_cartesian_pair3(MemResX,X,XType, MemResY,Y,YType, WF) :-
  (MemResX==pred_false -> true
   ; MemResY==pred_false -> true
   ; MemResX==pred_true -> not_element_of_wf(Y,YType,WF)
   ; not_element_of_wf(X,XType,WF)
  ).



/***************************/
/*  power_set(Set,TypeSet) */
/*    Set : POW(TypeSet)   */
/***************************/
    
:- assert_must_succeed(exhaustive_kernel_check(power_set([int(2),int(4)],[[int(2)],
 [int(4)],[],[int(4),int(2)]]))).
:- assert_must_succeed(power_set([int(1)],[[int(1)],[]])).
:- assert_must_succeed((power_set([int(1),int(2)],R),
           equal_object(R,[[],[int(1)],[int(2)],[int(1),int(2)]]))).
:- assert_must_succeed(power_set([],[[]])).

:- block power_set(-,?).
power_set(S,Res) :-   %print_message(power_set(S,PowerS)),
    cardinality(S,Card),
    when(ground(Card),  /* when all elements are known */
          (try_expand_custom_set(S,SE),
           findall(Subset, generate_subsets(SE,Subset), PowerS), %print(powerS(PowerS)),nl,
           equal_object(PowerS,Res)
          )).


generate_subsets([],[]).
generate_subsets([H|T],R) :- (R=[H|GT]; R=GT), generate_subsets(T,GT).


:- assert_must_succeed(exhaustive_kernel_check(non_empty_power_set([int(2),int(4)],[[int(2)],
 [int(4)],[int(4),int(2)]]))).
:- assert_must_succeed(non_empty_power_set([int(1)],[[int(1)]])).
:- assert_must_succeed((non_empty_power_set([int(1),int(2)],R),
           equal_object(R,[[int(1)],[int(2)],[int(1),int(2)]]))).
:- assert_must_succeed(non_empty_power_set([],[])).

:- block non_empty_power_set(-,?).
non_empty_power_set(S,Res) :-
    cardinality(S,Card),
    when(ground(Card),  /* when all elements are known */
          (try_expand_custom_set(S,SE),findall(Subset, (generate_subsets(SE,Subset),not_empty_set(Subset)), PowerS),
           equal_object(PowerS,Res)) ).

            
            
/* ------- */
/* BOOLEAN */
/* ------- */

% following predicates are not used:
%is_boolean(pred_true /* bool_true */).
%is_boolean(pred_false /* bool_false */).
%is_not_boolean(X) :- dif(X,pred_true /* bool_true */), dif(X,pred_false /* bool_false */).

/* ------- */
/* NUMBERS */
/* ------- */


is_integer(int(X),_WF) :- when(ground(X),integer(X)).
:- block is_not_integer(-).
is_not_integer(X) :- X \= int(_), % will be called for x /: INTEGER; should always fail.
   add_internal_error('Wrong type argument: ',is_not_integer(X)),fail. 

is_natural(int(X),_WF) :- clpfd_geq2(X,0), number_geq(X,0).
is_natural1(int(X),_WF) :- clpfd_geq2(X,1), number_geq(X,1).

:- block number_geq(-,?).
number_geq(X,N) :- X>=N.
:- block number_leq(-,?).
number_leq(X,N) :- X=<N.

:- assert_must_succeed(is_implementable_int(int(0),_WF)).
:- assert_must_fail(is_not_implementable_int(int(0))).


is_implementable_int(int(X),WF) :-  element_of_global_integer_set_wf('INT',X,WF).
is_implementable_nat(int(X),WF) :- element_of_global_integer_set_wf('NAT',X,WF).
is_implementable_nat1(int(X),WF) :- element_of_global_integer_set_wf('NAT1',X,WF).
is_not_implementable_int(X) :- not_element_of_global_set(X,'INT').
is_not_implementable_nat(X) :- not_element_of_global_set(X,'NAT').
is_not_implementable_nat1(X) :- not_element_of_global_set(X,'NAT1').

is_not_natural(int(X)) :- clpfd_geq2(-1,X), number_leq(X,-1).
is_not_natural1(int(X)) :- clpfd_geq2(0,X), number_leq(X,0).

:- assert_must_succeed(exhaustive_kernel_check(less_than(int(2),int(3)))).
:- assert_must_succeed(( safe_less_than(A,B),A=3,B=5 )).
:- assert_must_succeed(( safe_less_than(A,B),B=5,A=3 )).
:- assert_must_fail(( safe_less_than(A,B),A=5,B=3 )).
:- assert_must_fail(( safe_less_than(A,B),B=3,A=5 )).
:- assert_must_fail(( safe_less_than(A,B),A=5,B=5 )).
:- assert_must_fail(( safe_less_than(A,B),B=5,A=5 )).

less_than(int(X),int(Y)) :- 
    ((number(X),number(Y)) -> X < Y
      ;  clpfd_lt(X,Y,Posted), 
        (Posted=true -> true  ; safe_less_than(X,Y))).
less_than_direct(X,Y) :- 
    ((number(X),number(Y)) -> X < Y
      ;  clpfd_lt(X,Y,Posted), 
        (Posted=true -> true  ; safe_less_than(X,Y))).
:- block safe_less_than(-,?), safe_less_than(?,-).
safe_less_than(X,Y) :- 
 ((number(X),number(Y)) -> X<Y
    ; add_internal_error('Arguments not numbers: ',safe_less_than(X,Y))).

:- assert_must_succeed(exhaustive_kernel_check(less_than_equal(int(33),int(33)))).
less_than_equal(int(X),int(Y)) :- 
    ((number(X),number(Y)) -> X =< Y
      ; clpfd_leq(X,Y,Posted), 
        (Posted=true -> true  ; safe_less_than_equal(less_than_equal,X,Y))).
less_than_equal_direct(X,Y) :- 
    ((number(X),number(Y)) -> X =< Y
      ; clpfd_leq(X,Y,Posted), 
        (Posted=true -> true  ; safe_less_than_equal(less_than_equal_direct,X,Y))).

safe_less_than_equal(X,Y) :-
   safe_less_than_equal(safe_less_than_equal,X,Y).
:- block safe_less_than_equal(?,-,?), safe_less_than_equal(?,?,-).
safe_less_than_equal(PP,X,Y) :- %print(slte(PP,X,Y)),nl,
 ((number(X),number(Y)) -> X=<Y
    ; add_internal_error('Arguments not numbers: ',safe_less_than_equal(PP,X,Y))).

:- assert_must_succeed(exhaustive_kernel_check(greater_than(int(2),int(1)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(greater_than(int(2),int(2)))).
greater_than(int(X),int(Y)) :- less_than_direct(Y,X).
:- assert_must_succeed(exhaustive_kernel_check(greater_than(int(2),int(1)))).
:- assert_must_succeed(exhaustive_kernel_check(greater_than_equal(int(2),int(2)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(greater_than_equal(int(1),int(2)))).
greater_than_equal(int(X),int(Y)) :- less_than_equal_direct(Y,X).





:- assert_must_succeed(exhaustive_kernel_check([commutative],int_plus(int(2),int(3),int(5)))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],int_plus(int(2),int(3),int(6)))).

:- assert_must_succeed(int_plus(int(1),int(2),int(3))).
:- assert_must_succeed(( int_plus2(A,B,C),A=3,B=2,C==5 )).
:- assert_must_succeed(( int_plus2(A,B,C),A=3,C=5,B==2 )).
:- assert_must_succeed(( int_plus2(A,B,C),B=2,A=3,C==5 )).
:- assert_must_succeed(( int_plus2(A,B,C),B=2,C=5,A==3 )).
:- assert_must_succeed(( int_plus2(A,B,C),C=5,A=3,B==2 )).
:- assert_must_succeed(( int_plus2(A,B,C),C=5,B=2,A==3 )).
:- assert_must_succeed(( int_plus2(A,B,C),A=0,B==C )).
:- assert_must_succeed(( int_plus2(A,B,C),B=0,A==C )).

int_plus(int(X),int(Y),int(Plus)) :- %print(int_plus(X,Y,Plus)),nl,
    (two_vars_or_more(X,Y,Plus)
      -> clpfd_eq(Plus,X+Y) % can have performance problems
       ; true % otherwise we can compute the value directly below; we could skip the block declaration
    ),
    int_plus2(X,Y,Plus).
two_vars_or_more(X,Y,Z) :- var(X),!, (var(Y) ; var(Z)).
two_vars_or_more(_X,Y,Z) :- var(Y) , var(Z).

:- block int_plus2(-,-,-).
int_plus2(X,Y,Plus) :-  %print(int_plus2(X,Y,Plus)),nl,
    ( ground(X) ->
	( X=0       -> Plus=Y
	; otherwise -> int_plus3(X,Y,Plus))
    ; ground(Y) ->
	( Y=0       -> Plus=X
	; otherwise -> int_plus3(Y,X,Plus))
    ; otherwise     -> int_minus3(Plus,X,Y)).
% int_plus3/3: the first argument must be ground when called
:- block int_plus3(?,-,-).
int_plus3(X,Y,Plus) :-
    ( var(Plus) -> Plus is X+Y
    ; otherwise -> Y is Plus-X).

:- assert_must_succeed(exhaustive_kernel_check(int_minus(int(2),int(3),int(-1)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(int_minus(int(2),int(3),int(1)))).
:- assert_must_succeed(int_minus(int(3),int(1),int(2))).
:- assert_must_succeed(( int_minus2(A,B,C),A=3,B=2,C==1 )).
:- assert_must_succeed(( int_minus2(A,B,C),A=3,C=1,B==2 )).
:- assert_must_succeed(( int_minus2(A,B,C),B=2,A=3,C==1 )).
:- assert_must_succeed(( int_minus2(A,B,C),B=2,C=1,A==3 )).
:- assert_must_succeed(( int_minus2(A,B,C),C=1,A=3,B==2 )).
:- assert_must_succeed(( int_minus2(A,B,C),C=1,B=2,A==3 )).
:- assert_must_succeed(( int_minus2(A,B,C),B=0,A==C )).
:- assert_must_succeed(( int_minus2(A,B,C),B=0,C=5,A==5 )).
:- assert_must_succeed(( int_minus2(A,B,5),B=0,A==5 )).

int_minus(int(X),int(Y),int(Minus)) :-  %print(int_minus(X,Y,Minus)),nl,
   int_minus2(X,Y,Minus),
   (two_vars_or_more(X,Y,Minus) -> clpfd_eq(Minus,X-Y) % can have performance problems.
     ; true). % we can compute the value directly anyway
:- block int_minus2(-,-,-).
int_minus2(X,Y,Minus) :- %print(int_minus2(X,Y,Minus)),nl,
    ( ground(Y) ->
	   ( Y=0       -> X=Minus
	    ; otherwise -> Y2 is -Y, int_plus3(Y2,X,Minus))
    ; ground(X) ->
	   int_minus3(X,Y,Minus)
    ; otherwise ->
	    int_plus3(Minus,Y,X)).
% int_minus3/3: the first argument must be ground when called
:- block int_minus3(?,-,-).
int_minus3(X,Y,Minus) :-
    ( var(Minus) -> Minus is X-Y
    ; otherwise  -> Y is X-Minus).

:- assert_must_succeed(exhaustive_kernel_check(division(int(2),int(3),int(0),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(7),int(2),int(3),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(8),int(2),int(4),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(9),int(2),int(4),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(2),int(-1),int(-2),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(9),int(-2),int(-4),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(-9),int(-3),int(3),unknown,_WF))).
:- assert_must_succeed((platform_is_64_bit -> exhaustive_kernel_check(division(int(4294967296),int(2),int(2147483648),unknown,_WF)) ; true)).
:- assert_must_succeed(exhaustive_kernel_check(division(int(4294967296),int(2147483648),int(2),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(division(int(2),int(3),int(1),unknown,_WF))).
:- assert_must_succeed(( division3(A,B,C,unknown,_),A=15,B=4,C==3 )).
:- assert_must_succeed(( division3(A,B,C,unknown,_),B=4,A=15,C==3 )).

division(int(X),int(Y),int(XDY),Span,WF) :- var(Y), (var(X) ; var(XDY)),
    preferences:preference(use_clpfd_solver,true),!,
    clpfd_eq_guarded_div(XDY,X,Y),
    check_nonzero(X,Y,XDY,Span,WF).
division(int(X),int(Y),int(XDY),Span,WF) :-
    %% clpfd_eq_expr(XDY,X/Y), % can have performance problems; could hide division by 0 !
    division3(X,Y,XDY,Span,WF).
    %((var(Y), preferences:preference(use_clpfd_solver,true))
    %  -> division_propagate_wo_divisor(X,Y,XDY)
    %   ; true).
% TO DO: add case when XDY and X are known ?? or Y=1 and XDY known ??
%:- block division2(-,?,?,?,?), division2(?,-,?,?,?).
%division2(X,Y,XDY,Span,WF) :-
%    ( Y\==0 -> XDY is X // Y
%    ; Y==0  -> add_wd_error_set_result('division by zero','/'(X,Y),XDY,0,Span,WF)).
% we could also add: post_constraint(Y#\=0 #=> X/Y#=XDY)
:- block check_nonzero(?,-,?,?,?).
check_nonzero(X,Y,XDY,Span,WF) :-
  (Y=0 -> add_wd_error_set_result('division by zero','/'(X,Y),XDY,0,Span,WF)
   ; true).
/* ---------
:- block division_propagate_wo_divisor(-,?,-).
division_propagate_wo_divisor(X,_Y,XDY) :-  % X / Y = XDY
% TO DO: we could extract fd_dom from Y; or already trigger the propagation earlier
% use clpfd_domain(Y,MinY,MaxY) ; we can also check if 0 in fd_dom(Y) and then call CLPFD version ?
   (nonvar(X)
     ->  (var(XDY) -> 
             MX is -X,
            (X>0 -> clpfd_inrange(XDY,MX,X) ; clpfd_inrange(XDY,X,MX))
            ; true)
       % we know XDY but not X
    ;   A is abs(XDY)-1, A>0 ->  % we know that X must be either >= abs(XDY) or <= -abs(XDY)
        MA is -A,
        clpfd_not_inrange(X,MA,A)
    ;   true % nothing to propagate, XDY is -1, 0 or 1
    ).
    */
:- block division3(?,-,?,?,?).
division3(X,Y,XDY,Span,WF) :-
    ( Y==0 -> add_wd_error_set_result('division by zero','/'(X,Y),XDY,0,Span,WF)
     ; nonvar(X) -> XDY is X // Y
     ; Y == 1 -> X=XDY
     ; (Y == -1,nonvar(XDY)) -> X is -XDY
     ; clpfd_eq_div(XDY,X,Y)). % we could setup constraint before Y is known; could hide division by 0 ?

:- assert_must_succeed(exhaustive_kernel_check(modulo(int(2),int(3),int(2),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(modulo(int(7),int(2),int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(modulo(int(8),int(2),int(0),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(modulo(int(9),int(2),int(1),unknown,_WF))).
:- assert_must_succeed((platform_is_64_bit -> exhaustive_kernel_check(modulo(int(4294967296),int(2147483648),int(0),unknown,_WF)) ; true)).
:- assert_must_succeed((platform_is_64_bit -> exhaustive_kernel_check(modulo(int(4294967299),int(2147483648),int(3),unknown,_WF)) ; true)).
:- assert_must_succeed(( modulo2(A,B,C,unknown,_),A=7,B=5,C==2 )).
:- assert_must_fail(( modulo2(A,B,C,unknown,_),A=7,B=5,C==3 )).

modulo(int(X),int(Y),int(Modulo),Span,WF) :- %print(modulo(X,Y,Modulo)),nl,
    %% clpfd_eq(Modulo,X mod Y), % can have performance problems; could hide division by 0 !
    %clpfd_modulo(X,Y,Modulo,WF), % maybe only call in non-CLPFD mode ?
    modulo2(X,Y,Modulo,Span,WF),
    % assert that Modulo<Y, Modulo>=0
    (nonvar(Modulo) -> true
      ; (number(Y),Y =< 0) -> true % in this case we will generate a well-definedness error; it would be more efficient from a constraint solving perspective to assume that there are no well-definedness errors and remove this case !!
      ; clpfd_modulo_prop(X,Y,Modulo,WF)
    ).
:- use_module(specfile,[animation_minor_mode/1]).
:- block modulo2(-,?,?,?,?), modulo2(?,-,?,?,?).
modulo2(X,Y,Modulo,Span,WF) :-  
    ( Y>0 -> (X<0 -> 
        (animation_minor_mode(tla) -> Modulo is X mod Y
         ; add_wd_error_set_result('mod not defined for negative numbers in B',mod(X,Y),Modulo,0,Span,WF))
                   ; Modulo is X mod Y)
    ; Y==0  -> add_wd_error_set_result('mod by zero',mod(X,Y),Modulo,0,Span,WF)
    ; Y<0 -> add_wd_error_set_result('mod not defined for negative numbers in B',mod(X,Y),Modulo,0,Span,WF)).
 
% propagate information aobut Moduloe result if part of the information known
clpfd_modulo_prop(_,Y,Modulo,_) :- %preferences:preference(use_clpfd_solver,true),!,
    % in CLP(FD) this is sufficient; for non-CLPFD mode it is better to call in_nat_range to restrict enumeration
    less_than_direct(Modulo,Y),
    less_than_equal_direct(0,Modulo).  % 0 <= Modulo < Y  -> by transitivity this forces Y>0  and we no longer detect wd-errors
    % we could reify: Y>0 => Modulo <Y ?  Is it worth it ?
    % what is still missing is that if Y < Modulo => X=Y  (CLP(FD) does this X in 0..100 , Y in 2..20 ,  X #= Y mod 30.)
/* clpfd_modulo_prop(X,Y,Modulo,WF) :- clpfd_modulo_noclp(X,Y,Modulo,WF).
:- block clpfd_modulo_noclp(-,-,-,?).
clpfd_modulo_noclp(X,Y,Modulo,WF) :-  print(mod(X,Y,Modulo,WF)),nl,
    var(X),var(Modulo),number(Y),!,
    Y1 is Y-1,
    in_nat_range_wf(int(Modulo),int(0),int(Y1),WF).  % problem: could enumerate lambda return variables !!
clpfd_modulo_noclp(_X,_Y,_Modulo,_WF).
*/

:- assert_must_succeed(exhaustive_kernel_check(unary_minus(int(2),int(-2)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(unary_minus(int(2),int(2)))).
:- assert_must_succeed(( unary_minus2(A,B),A=7,B== -7 )).
:- assert_must_succeed(( unary_minus2(A,B),A= -7,B==7 )).
:- assert_must_succeed(( unary_minus2(B,A),A=7,B== -7 )).
:- assert_must_succeed(( unary_minus2(B,A),A= -7,B==7 )).
:- assert_must_fail(( unary_minus2(B,A),A= -7,B=6 )).
:- assert_must_fail(( unary_minus2(A,B),A= -7,B=6 )).

unary_minus(int(X),int(MX)) :- 
    unary_minus2(X,MX),
    ((var(X),var(MX)) -> clpfd_eq(MX,0 - X) % can have performance problems
       ; true % we can compute the value without CLPFD
    ).
:- block unary_minus2(-,-).
unary_minus2(X,MX) :- 
    ( ground(X) -> MX is -X
    ; otherwise -> X is -MX).

:- assert_must_succeed(first_of_pair((int(1),int(2)),int(1))).
:- assert_must_succeed(second_of_pair((int(1),int(2)),int(2))).

first_of_pair((A,_B),R) :- equal_object(R,A).
second_of_pair((_A,B),R) :- equal_object(R,B).


:- assert_must_succeed(exhaustive_kernel_check(cartesian_product([int(2),int(4)],[int(3),int(1)],
 [(int(2),int(1)),(int(2),int(3)),(int(4),int(3)),(int(4),int(1))]))).
:- assert_must_succeed(exhaustive_kernel_check(cartesian_product([],[int(3),int(1)],[]))).
:- assert_must_succeed(exhaustive_kernel_check(cartesian_product([int(3)],[],[]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(cartesian_product([int(3)],[int(2)],[]))).
:- assert_must_succeed((cartesian_product(global_set('NAT'),[int(2)],_Res))).
:- assert_must_succeed((cartesian_product([int(1)],[int(2)],Res),
         equal_object(Res,[(int(1),int(2))]))).
:- assert_must_succeed((cartesian_product([int(1)],[int(2)],[(int(1),int(2))]))).
:- assert_must_succeed((cartesian_product([],[int(1),int(2)],Res),
         equal_object(Res,[]))).
:- assert_must_succeed((cartesian_product([int(1),int(2)],[],Res),
         equal_object(Res,[]))).
:- assert_must_succeed((cartesian_product([int(1),int(2)],[int(2),int(3)],Res),
 equal_object(Res,[(int(1),int(2)),(int(1),int(3)),(int(2),int(2)),(int(2),int(3))]))).
:- assert_must_succeed((cartesian_product([int(1)|T],[int(2)|T2],Res),
             T = [int(2)], T2 = [int(3)],
 equal_object(Res,[(int(1),int(2)),(int(1),int(3)),(int(2),int(2)),(int(2),int(3))]))).
:- assert_must_fail((cartesian_product([int(1)],[int(2),int(3)],Res),(Res=[_];
         equal_object(Res,[_,_,_|_])))).
    
  
:- block cartesian_product(-,?,?), cartesian_product(?,-,?).
cartesian_product(Set1,Set2,Res) :-
    expand_custom_set_to_list(Set1,ESet1,_,cartesian_product),
    try_expand_custom_set(Set2,ESet2),
    cartesian_product2(ESet1,ESet2,Res).

:- block cartesian_product2(-,?,?).                       
cartesian_product2([],_,Res) :- empty_set(Res).
cartesian_product2([H|T],Set2,Res) :-
    cartesian_el_product(Set2,H,Res,InnerRes),
    cartesian_product2(T,Set2,InnerRes).
    
:- block cartesian_el_product(-,?,?,?). 
cartesian_el_product([],_El,Res,InnerRes) :- equal_object(Res,InnerRes).
cartesian_el_product([H|T],El,ResSoFar,InnerRes) :-
    equal_object(ResSoFar,[(El,H)|NewResSoFar]),
    cartesian_el_product(T,El,NewResSoFar,InnerRes).
    


:- assert_must_succeed(exhaustive_kernel_check(in_nat_range(int(2),int(2),int(3)))).
:- assert_must_succeed(exhaustive_kernel_check(in_nat_range_wf(int(2),int(2),int(3),_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(in_nat_range_wf(int(2),int(3),int(2),_WF))).
:- assert_must_succeed((in_nat_range_wf(X,int(11),int(12),WF),
   kernel_waitflags:ground_wait_flags(WF), X==int(12) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(12),_WF), X=int(10) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(12),_WF),  X=int(13) )).
:- assert_must_succeed((in_nat_range_wf(X,int(11),int(12),_WF), X=int(11) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(10),_WF), X=int(11) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(10),_WF), X=int(10) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(10),_WF), X=int(12) )).

in_nat_range(int(X),int(Y),int(Z)) :- % does not enumerate, in contrast to in_nat_range_wf
  clpfd_inrange(X,Y,Z,Posted),  % better to call inrange rather than leq twice, avoids unecessary propagation
  (Posted==true -> true
   ;   safe_less_than_equal(in_nat_range,Y,X),
       safe_less_than_equal(in_nat_range,X,Z)
   ).
in_nat_range_wf(int(X),int(Y),int(Z),WF) :- %print(in_nat_range_wf(X,Y,Z)),nl,
  clpfd_inrange(X,Y,Z,Posted),  % better to call inrange rather than leq twice, avoids unecessary propagation
  (Posted==true -> 
     /* if the constraint was posted: we do not need to add safe_less_than_equal,...: if overflow happes whole computation will fail anyway */
       (ground(X) -> true ; add_fd_variable_for_labeling(X,WF))
   ;   safe_less_than_equal(in_nat_range_wf,Y,X),
       safe_less_than_equal(in_nat_range_wf,X,Z),
       (ground(X) -> true
       ; get_int_domain(X,Y,Z,RL,RU),get_nat_range_prio(X,RL,RU,WF,LWF),
         % print(register_enumerate_int(X,RL,RU,WF,LWF)),nl, %%
         % portray_waitflags(WF),nl,
         call_enumerate_int(X,RL,RU,LWF))
  ).
 % when((ground(X);nonvar(LWF)),(ground(X) -> true ; enumerate_int(X,RL,RU))).

:- block get_nat_range_prio(?,-,?,?,?), get_nat_range_prio(?,?,-,?,?).
get_nat_range_prio(Variable,Y,Z,WF,LWF) :- Size is Z+1-Y, %print(get_natrange_prio(Size,Y,Z,WF)),nl,
   (Size>1 -> 
     (preferences:preference(use_clpfd_solver,true) 
       -> add_fd_variable_for_labeling(Variable,Size,WF,LWF) % will use CLP(FD) labeling
       ;  get_wait_flag(Size,get_nat_range_prio(Y,Z),WF,LWF)
     )
    ; LWF=Size  /* Size=0 or 1 -> we can either fail or determine variable */).

:- assert_must_succeed((kernel_objects:call_enumerate_int(X,1,2,g), X==2)).
:- block call_enumerate_int(-,?,?,-).
call_enumerate_int(X,RL,RU,_LWF) :-  % print(enum(X,RL,RU,_LWF)),nl,%
   (ground(X) -> true
     ;   % print(enum_int(X,RL,RU)),nl, %%
       % get_int_domain(X,RL,RU,RLL,RUU) : if clp(fd) active then CLP(FD) labeling is used anyway
       enumerate_int(X,RL,RU)).




:- assert_must_succeed(exhaustive_kernel_check(not_in_nat_range(int(2),int(3),int(2)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_in_nat_range(int(2),int(2),int(3)))).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(12)), X=int(10) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(12)), X=int(13) )).
:- assert_must_fail((not_in_nat_range(X,int(11),int(12)), X=int(11) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(10)), X=int(11) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(10)), X=int(10) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(10)), X=int(12) )).
   
not_in_nat_range(int(X),int(Y),int(Z)) :-
  ((number(Y),number(Z))
    -> (Z>=Y -> clpfd_not_in_non_empty_range(X,Y,Z) ; true /* interval empty */)
    ;  clpfd_not_inrange(X,Y,Z)
  ).





:- assert_must_succeed(exhaustive_kernel_check_wf(square(int(3),int(9),WF),WF)).

square(int(X),int(Sqr),WF) :- % print(sqr(X,Sqr)),nl,
   int_square(X,Sqr,WF),
   (var(X) -> clpfd_eq(Sqr,X * X)
     ; true). % we can compute the value directly

:- block int_square(-,-,?).
int_square(X,Sqr,_) :- ground(X),!, Sqr is X*X.
int_square(X,Sqr,WF) :- get_binary_choice_wait_flag(implication,WF,WF2), int_square2(X,Sqr,WF2).
:- block int_square2(-,?,-).
int_square2(X,Sqr,_) :- ground(X),!, Sqr is X*X.
int_square2(X,Sqr,_WF2) :- integer_square_root(Sqr,X).

:- assert_must_succeed(( kernel_objects:integer_square_root(0,X),X==0 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(1,X),X==1 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(4,X),X==2 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(49,X),X==7 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(49,X),X==(-7) )).
:- assert_must_fail(( kernel_objects:integer_square_root(5,_) )).
:- assert_must_succeed(( X= 123456789, Y is X*X, kernel_objects:integer_square_root(Y,Z),Z==X)).
:- assert_must_fail(( X= 123456789, Y is 1+X*X, kernel_objects:integer_square_root(Y,_Z))).
:- assert_must_succeed(( X= 12345678900, Y is X*X, kernel_objects:integer_square_root(Y,Z),Z==X)).

integer_square_root(0,Root) :- !, Root = 0.
integer_square_root(Sqr,PMRoot) :- %print(sqrt(Sqr)),nl,
   Sqr>0, Root is truncate(sqrt(Sqr)), Sqr is Root*Root, %print(root(Root)),nl,
  (PMRoot = Root ; PMRoot is -(Root)).

times(int(X),int(Y),int(Times)) :- 
    int_times2(X,Y,Times),
    (two_vars_or_more(X,Y,Times) -> clpfd_eq(Times,X * Y) % can have performance problems.
      ; true). % we can compute the value directly

:- assert_must_succeed(exhaustive_kernel_check([commutative],times(int(2),int(3),int(6)))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],times(int(2),int(1),int(2)))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],times(int(2),int(0),int(0)))).
:- assert_must_succeed(exhaustive_kernel_check(times(int(0),int(1),int(0)))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],times(int(2),int(3),int(5)))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],times(int(1),int(3),int(2)))).
:- assert_must_succeed(( int_times2(A,B,C),A=3,B=2,C==6 )).
:- assert_must_succeed(( int_times2(A,B,C),A=3,C=6,B==2 )).
:- assert_must_succeed(( int_times2(A,B,C),B=2,A=3,C==6 )).
:- assert_must_succeed(( int_times2(A,B,C),B=2,C=6,A==3 )).
:- assert_must_succeed(( int_times2(A,B,C),C=6,A=3,B==2 )).
:- assert_must_succeed(( int_times2(A,B,C),C=6,B=2,A==3 )).
:- assert_must_succeed(( int_times2(A,_,C),A=0,C==0 )).
:- assert_must_succeed(( int_times2(_,B,C),B=0,C==0, nl,print(int_times2(B,C)),nl,nl )).
:- assert_must_succeed(( int_times2(A,B,C),A=1,B==C )).
:- assert_must_succeed(( int_times2(A,B,C),B=1,A==C )).
:- assert_must_succeed(( int_times2(A,1,C),A=2,C==2 )).
:- assert_must_succeed(( int_times2(_A,0,C),C==0 )).
:- assert_must_succeed(( int_times2(A,_,C),C=0,A=0 )).
:- assert_must_succeed(( int_times2(_,B,C),C=0,B=0 )).
:- assert_must_succeed(( int_times2(A,B,0),A=0,B=2 )).
:- assert_must_succeed(( int_times2(A,B,0),B=2,A=0 )).
:- assert_must_succeed(( int_times2(B,A,0),A=0,B=2 )).
:- assert_must_succeed(( int_times2(B,A,0),B=2,A=0 )).
:- assert_must_fail(( int_times2(A,_,C),A=3,C=7 )).
:- assert_must_fail(( int_times2(A,_,C),C=7,A=3 )).
:- assert_must_fail(( int_times2(_,B,C),B=2,C=7 )).
:- assert_must_fail(( int_times2(_,B,C),C=7,B=2 )).
:- assert_must_fail(( int_times2(A,_,C),C=7,A=0 )).
:- assert_must_fail(( int_times2(_,B,C),C=7,B=0 )).
:- assert_must_fail(( int_times2(B,A,0),B=2,A=1 )).

:- block int_times2(-,-,-).
int_times2(X,Y,Times) :-
    ( ground(X) ->
	( X==1      -> Y=Times
	; X==0      -> Times=0
	; otherwise -> int_times3(X,Y,Times))
    ; ground(Y) ->
	( Y==1      -> X=Times
	; Y==0      -> Times=0
	; otherwise -> int_times3(Y,X,Times))
    ; otherwise ->
	int_times4(X,Y,Times)).
% int_times3/3: First argument must be ground when called and non-zero
:- block int_times3(?,-,-).
int_times3(X,Y,Times) :-
    ( ground(Y) -> Times is X*Y
    ; otherwise -> Y is Times // X, Times is X*Y).
% int_times4/3: Third argument must be ground when called
:- block int_times4(-,-,?).
int_times4(X,Y,Times) :- %print(int_times4(X,Y,Times)),nl,
    ( Times==0 ->
	  ( ground(X) -> (X==0 -> true; Y=0 )
	  ; ground(Y) -> (Y==0 -> true; X=0 ))
    ; otherwise ->
	  ( ground(X) -> X\==0, Y is Times // X, Times is X*Y
	  ; ground(Y) -> Y\==0, X is Times // Y, Times is X*Y)).
    

:- assert_must_succeed(exhaustive_kernel_check(int_power(int(2),int(3),int(8),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(2),int(1),int(2),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(3),int(0),int(1),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(1),int(3),int(1),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(0),int(3),int(0),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(int_power(int(2),int(3),int(6),unknown,_))).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=2,B=5,C==32 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A= -2,B=5,C== -32 )).
%:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=1,B= -5,C==1 )). % now aborts !
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=1,C=1, B= -5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=1,C= 1,B = -5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=2,C=32,B==5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=10,C=1000,B==3 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A= -2,C= -32,B==5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A= -2,C= 16,B==4 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=2,C=1,B==0 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,B=2,C==0 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,C=0,B=2 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,B=0,C==1 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,C=1,B==0 )).
:- assert_must_succeed(( int_power2(17,13,C,unknown,_),C==9904578032905937 )).
:- assert_must_succeed(( int_power2(A,13,C,unknown,_),C=9904578032905937,A=17 )).
:- assert_must_fail(( int_power2(A,13,C,unknown,_),C=9904578032905936,A=17 )).
:- assert_must_fail(( int_power2(A,B,C,unknown,_),A=2,B=5,C=33 )).
:- assert_must_abort_wf(( int_power2(A,B,_,unknown,WF),A=2,B= -5 ),WF).
:- assert_must_fail(( int_power2(A,_,C,unknown,_),A= -2,C=32 )).
:- assert_must_fail(( int_power2(A,_,C,unknown,_),A= -2,C= -16 )).

% maybe X^Y with Y<0 should be undefined
% todo: calculate X from Y und Pow
int_power(int(X),int(Y),int(Pow),Span,WF) :- int_power2(X,Y,Pow,Span,WF).
:- block int_power2(-,?,?,?,?), int_power2(?,-,-,?,?).
int_power2(X,Y,Pow,Span,WF) :-
    ( ground(Y) ->
	  ( Y>=0      -> safe_int_power(X,Y,Pow) % has rounding errors: Pow is integer(exp(X,Y))
	    ; otherwise -> %X==1,Pow=1) % we could produce an abort_error here
	                 add_wd_error_set_result('power of negative number','**'(X,Y),Pow,1,Span,WF))
    ; otherwise ->  /* X & POW are ground */
		( X==1 -> Pow==1  /* 1**Y = 1 */
		; X==0, Pow==1 -> Y=0
		; X==0 -> Pow==0
		; X>0, Pow>0 ->
			checked_log(X,Y,Pow)
		; X<0, Pow<0 ->
			PosPow is -Pow,
			checked_log(-X,Y,PosPow),
			1 is Y mod 2
		; X<0, Pow>0 ->
			checked_log(-X,Y,Pow),
			0 is Y mod 2)).
checked_log(Base,Exp,Pow) :-
    Try is integer(log(Base,Pow)),
    % the calculation might have a rounding error
    ( safe_int_power(Base,Try,Pow) -> Exp=Try
    ; otherwise -> Exp is Try+1, safe_int_power(Base,Exp,Pow)).

% safe exponentiation using the squaring algorithm
% Note: in TLA mode 0^0 is undefined according to TLC
safe_int_power(Base,Exp,Result) :-
	Msb is msb(Exp),   % most significant bit
	ExpMask is 1<<Msb,
	safe_int_power2(ExpMask,Exp,Base,1,Result).
safe_int_power2(0,_,_,Prev,Result) :- !, Prev=Result.
safe_int_power2(Mask,Exp,Base,Prev,Result) :-
	P is Exp /\ Mask,   % P is Exp's highest bit
	Mask2 is Mask>>1,
	Quad is Prev*Prev,
	( P==0      -> Next = Quad
	; otherwise -> Next is Quad*Base),
	safe_int_power2(Mask2,Exp,Base,Next,Result).

 %% -------------------------------------------------------

:- assert_must_succeed(( singleton_set_element([int(1)],E,_WF), E==int(1) )).
:- assert_must_fail(singleton_set_element([int(1)],int(2),_WF) ).
:- assert_must_abort_wf(kernel_objects:singleton_set_element([int(1),int(2)],_E,WF),WF).
% This predicate computes the effect of the MU operator.
% Set should be a singleton set and Elem its only element.
% In case Set is empty  or has more than one element, an error
% message is generated.
% The waitflag store WFStore is needed to obtain a waitflag for the
% decision when to force whether Set is a singleton or not and
% to save the error if necessary.
singleton_set_element(Set,Elem,WFStore) :-
    cardinality_as_int(Set,Card),
    get_enumeration_finished_wait_flag(WFStore,LWF),
    equality_objects_lwf(Card,int(1),IsSingleton,LWF),
    singleton_set_element2(IsSingleton,Set,Elem,Card,WFStore).
:- block singleton_set_element2(-,?,?,?,?).
singleton_set_element2(pred_true,Set,Elem,_Card,_WFStore) :-
    exact_element_of(Elem,Set).
singleton_set_element2(pred_false,_Set,_Elem,Card,WFStore) :-
    ( Card=int(C) -> true ;  C = 'unknown'),
    add_wd_error('argument of MU expression must have cardinality 1, but has ', C, WFStore).


%:- print(finished_loading_kernel_objects),nl.