% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% each cli_testcase consists of: id, list of categories, arguments to probcli, comment

:- module(testcases, [cli_testcase/5, cli_testcase_diff_check_output/3, extra_testcase_file/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module defines the (integration) tests.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

cli_testcase(Id,Groups,TestConfiguration,Arguments,Description) :-
    cli_testcase(Id,Groups,ArgumentsTemp,Description),
    options_all_tests(AdditionalArguments),
    append(ArgumentsTemp,AdditionalArguments,Arguments),
    (additional_configuration(Id,Conf) -> TestConfiguration = Conf ; TestConfiguration = []).

% options that should be used in all testcases
options_all_tests(['-strict','-p','STRICT_RAISE_WARNINGS','TRUE']).

additional_configuration(50,  [conditional_skip(current_prolog_flag(os_data,os(_F,win32nt,_E)))]).
additional_configuration(161, [conditional_skip(current_prolog_flag(os_data,os(_F,win32nt,_E)))]).
additional_configuration(271, [conditional_skip(current_prolog_flag(os_data,os(_F,win32nt,_E)))]).
additional_configuration(324, [skip]). /* because only works if SMT mode is allowed to instantiate */
additional_configuration(539, [conditional_skip(current_prolog_flag(os_data,os(_F,win32nt,_E)))]).
additional_configuration(865, [conditional_skip(current_prolog_flag(os_data,os(_F,win32nt,_E)))]).
additional_configuration(967, [skip]). % the proof of the invariant relies on the theorem, which is proven but has wd-problem
%additional_configuration(968, [skip]). % the proof of the invariant relies on the theorem, which is proven but has wd-problem
additional_configuration(987, [skip]). % Skip test case for now because of low performance while running the test

cli_testcase(0, [unit], ['-selfcheck', '-strict', '-p', 'CLPFD', 'FALSE'], 'Unit tests').
cli_testcase(1, [unit], ['-selfcheck', '-strict', '-p', 'CLPFD', 'TRUE'], 'Unit tests (with CLPFD)').
cli_testcase(2, [unit], ['-selfcheck', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SYMBOLIC', 'TRUE'], 'Unit tests (with CLPDFD and SYMBOLIC)').
cli_testcase(3, [b_test], ['-t', '-strict', 'examples/B/FeatureChecks/DEFINITIONS/SETPREF.mch', '-p', 'MAXINT', '1024', '-card', 'ID', '3'], '').
cli_testcase(4, [b_test], ['-t', '-strict', 'examples/B/FeatureChecks/DEFINITIONS/DefinitionScopeOk.mch'], '').
cli_testcase(5, [b_test], ['-t', '-strict', 'examples/B/FeatureChecks/ForAll_VariousVersions.mch'], '').
cli_testcase(6, [b_test], ['examples/B/FeatureChecks/NotMemberCheck.mch', '-t', '-strict', '-assertions'], '').
cli_testcase(7, [bench], ['examples/B/Benchmarks/CarlaTravelAgencyErr.mch', '-t', '-strict'], '').
cli_testcase(8, [bench], ['examples/B/Benchmarks/RouteIsSeq.mch', '-t', '-strict'], '').
cli_testcase(9, [bench], ['examples/B/Benchmarks/Sets2.mch', '-t', '-strict'], '').
cli_testcase(10, [bench], ['examples/B/Benchmarks/Doors.mch', '-t', '-strict'], '').
cli_testcase(11, [bench], ['examples/B/Benchmarks/DSP0.mch', '-t', '-mc', '100', '-noinv', '-strict', '-nodead'], '').
cli_testcase(12, [bench], ['examples/B/Benchmarks/DSP0_complicated_initialisation.mch', '-t', '-mc', '100', '-noinv', '-strict', '-nodead'], '').
cli_testcase(13, [bench], ['examples/B/Benchmarks/CSM.mch', '-mc', '1000', '-strict', '-cc', '341', '1229'], '').
cli_testcase(14, [bench], ['examples/B/Benchmarks/scheduler.mch', '-t', '-mc', '100', '-strict', '-cc', '36', '121'], '').
cli_testcase(15, [bench], ['examples/B/Benchmarks/phonebook7.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(16, [bench], ['examples/B/Benchmarks/tictac.mch', '-mc', '100', '-noinv', '-nodead', '-strict'], '').
cli_testcase(17, [bench], ['examples/B/Benchmarks/Teletext_bench.mch', '-mc', '100', '-strict'], '').
cli_testcase(18, [bench], ['examples/B/Benchmarks/Cruise_finite1.mch', '-t', '-strict'], '').
cli_testcase(19, [bench], ['examples/B/BZTT/TestBZTT_Data.mch', '-mc', '100', '-cc', '26', '176', '-strict', '-l', 'log/TestBZTT_Data.log'], '').
cli_testcase(20, [performance_tests], ['examples/B/Daniel/Kodkod/Power2Nr.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE', '-aa', '4', '0', '0'], '').
cli_testcase(21, [performance_tests], ['examples/B/Daniel/Kodkod/Power2Nr.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE', '-ma', '4', '0', '0'], '').
cli_testcase(22, [performance_tests], ['examples/B/Daniel/Kodkod/StarTopologySingleSol.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(23, [performance_tests], ['examples/B/PerformanceTests/conc_seq_test.mch', '-strict', '-t'], '').
cli_testcase(24, [performance_tests], ['examples/B/PerformanceTests/LargeSets/PartialFunCheck.mch', '-strict', '-t'], '').
cli_testcase(25, [performance_tests], ['examples/B/PerformanceTests/ForAll/ForAll_NATURAL.mch', '-strict', '-t'], '').
cli_testcase(26, [performance_tests], ['examples/B/PerformanceTests/ForAll/ForAllExpansion.mch', '-strict', '-t'], '').
cli_testcase(27, [performance_tests], ['examples/B/PerformanceTests/ForAll/ForAllExpansian_Cartesian.mch', '-strict', '-t', '-assertions'], '').
cli_testcase(28, [performance_tests], ['examples/B/PerformanceTests/ForAll/ForAllExpansion_Cartesian2.mch', '-strict', '-t', '-assertions'], '').
cli_testcase(29, [performance_tests], ['examples/B/PerformanceTests/ForAll/ForAllExpansion_Cartesian2.mch', '-strict', '-t', '-assertions', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(30, [performance_tests], ['examples/B/PerformanceTests/ForAll/ForAllExpansion_Cartesian3.mch', '-strict', '-t', '-assertions'], '').
cli_testcase(31, [performance_tests], ['examples/B/PerformanceTests/LargeSets/SeqCheck.mch', '-strict', '-t'], '').
cli_testcase(32, [performance_tests], ['examples/B/PerformanceTests/SetLogInefficiencies.mch', '-t', '-strict'], '').
cli_testcase(33, [performance_tests], ['examples/B/PerformanceTests/SetLogInefficiencies2.mch', '-t', '-strict'], '').
cli_testcase(34, [performance_tests], ['examples/B/PerformanceTests/SetLogInefficiencies3.mch', '-t', '-strict'], '').
cli_testcase(35, [performance_tests], ['examples/B/PerformanceTests/InfiniteClosures.mch', '-strict', '-t'], '').
cli_testcase(36, [performance_tests], ['examples/B/PerformanceTests/InfiniteClosures2.mch', '-strict', '-t'], '').
cli_testcase(37, [performance_tests], ['examples/B/PerformanceTests/InfiniteAbsSqrt.mch', '-strict', '-t'], '').
cli_testcase(38, [performance_tests], ['examples/B/PerformanceTests/InfCardChecks.mch', '-strict', '-t'], '').
cli_testcase(39, [performance_tests], ['examples/B/PerformanceTests/ID_Checks.mch', '-strict', '-t', '-assertions'], '').
cli_testcase(40, [performance_tests], ['examples/B/PerformanceTests/DirectProductPerformance.mch', '-strict', '-init', '-p', 'TIME_OUT', '20000'], 'Test the performance of the direct product operator.').
cli_testcase(41, [performance_tests], ['examples/B/PerformanceTests/DirectProductPerformance.mch', '-strict', '-init', '-p', 'SYMBOLIC', 'TRUE', '-p', 'TIME_OUT', '15000'], '').
cli_testcase(42, [performance_tests], ['examples/B/Mathematical/Sieve.mch', '-mc', '1000', '-strict', '-expcterr', 'goal_found', '-l', 'log/Sieve.log'], '').
cli_testcase(43, [performance_tests], ['examples/B/Mathematical/BinomialCoefficients.mch', '-t', '-strict'], '').
cli_testcase(44, [performance_tests], ['examples/B/Mathematical/SieveAlternate.mch', '-mc', '1000', '-strict', '-expcterr', 'goal_found', '-l', 'log/SieveAlternate.log'], '').
cli_testcase(45, [performance_tests], ['examples/B/Mathematical/GraphIso/CheckGraphIsomorphism.mch', '-t', '-strict', '-l', 'log/CheckGraphIsomorphism.log'], '').
cli_testcase(46, [performance_tests], ['examples/B/Mathematical/GraphIso/CheckLargeGraphIsomorphismAsConstants.mch', '-t', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '5000'], '').
cli_testcase(47, [performance_tests], ['examples/B/Mathematical/Theories/FiniteSequences1.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(48, [performance_tests], ['examples/B/Mathematical/Theories/FiniteSequences2.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(49, [performance_tests], ['examples/B/PerformanceTests/SATLIB/flat200-90.mch', '-t', '-p', 'STATIC_ORDERING', 'TRUE', '-strict', '-p', 'TIME_OUT', '60000', '-l', 'log/SAT_flat200-90.log'], '').
cli_testcase(50, [performance_tests], ['examples/B/PerformanceTests/SATLIB/sudoku.mch', '-init', '-p', 'STATIC_ORDERING', 'TRUE', '-strict', '-p', 'TIME_OUT', '60000', '-p', 'LARGE_PARSER_JVM', 'TRUE'], 'Large machine to test the parser/JVM, skipped on Windows because the Windows build bot can not reserve enought memory.').
cli_testcase(51, [performance_tests], ['examples/B/PerformanceTests/ConstraintPropagation/MaxPropagation.mch', '-strict', '-t'], '').
cli_testcase(52, [performance_tests], ['examples/B/PerformanceTests/ConstraintPropagation/MaxPropagation2.mch', '-strict', '-t'], '').
cli_testcase(53, [performance_tests], ['examples/B/PerformanceTests/ConstraintPropagation/SetMembershipInference.mch', '-strict', '-t'], '').
cli_testcase(54, [performance_tests], ['examples/B/PerformanceTests/ConstraintPropagation/SetMembershipReification.mch', '-strict', '-t'], '').
cli_testcase(55, [performance_tests], ['examples/B/Demo/SortByPermutation_v2.mch', '-strict', '-t','-p', 'TIME_OUT', '7000'], '').
cli_testcase(56, [performance_tests], ['examples/B/IPD/colouring/GraphColouringNAT2.mch', '-t', '-strict'], '').
cli_testcase(57, [performance_tests], ['examples/EventBPrologPackages/Puzzles/Instance_small_60_400_9_ctx.eventb', '-init', '-strict', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(58, [performance_tests], ['examples/B/BZTT/PO_Scheduler.mch', '-mc', '100', '-strict', '-t', '-nodead'], '').
cli_testcase(59, [performance_tests], ['examples/B/PerformanceTests/SMTLIB/IDL_SimpleUnsat.mch', '-strict', '-init', '-expcterr', 'setup_constants_fails'], '').
cli_testcase(60, [performance_tests], ['examples/B/PerformanceTests/SMTLIB/int_completeness1.mch', '-strict', '-init', '-expcterr', 'setup_constants_fails'], '').
cli_testcase(61, [performance_tests], ['examples/B/PerformanceTests/SMTLIB/int_completeness1_event.mch', '-strict', '-t'], '').
cli_testcase(62, [performance_tests], ['examples/B/PerformanceTests/SMTLIB/SquareEquivalenceProblem.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE', '-mc', '10', '-cc', '2', '2'], '').
cli_testcase(63, [performance_tests], ['examples/B/PerformanceTests/clpfd_constraint_checks/IntegerComprSets.mch', '-strict', '-t'], '').
cli_testcase(64, [performance_tests], ['examples/B/PerformanceTests/SeqEnumWithLargeMAXINT.mch', '-strict', '-t'], '').
cli_testcase(65, [performance_tests], ['examples/B/PerformanceTests/ConstraintPropagation/SequenceSolving.mch', '-strict', '-t'], '').
cli_testcase(66, [performance_tests], ['examples/B/PerformanceTests/SAT_Tests/SubsetReificationTest.mch', '-strict', '-t'], '').
cli_testcase(67, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/monitor2.mch', '-mc', '100', '-cc', '15', '92', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(68, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/Symetric.mch', '-mc', '100', '-cc', '2', '6', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(69, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '11', '47', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(70, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '36', '156', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(71, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '11', '47', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], '').
cli_testcase(72, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '36', '73', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(73, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '150', '-cc', '145', '447', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(74, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '29', '94', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(75, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '29', '94', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], '').
cli_testcase(76, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '150', '-cc', '145', '210', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(77, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '51', '211', '-p', 'DEFAULT_SETSIZE', '4', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(78, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '51', '211', '-p', 'DEFAULT_SETSIZE', '4', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(79, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '81', '405', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(80, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '81', '405', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(81, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '120', '701', '-p', 'DEFAULT_SETSIZE', '6', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(82, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '120', '701', '-p', 'DEFAULT_SETSIZE', '6', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(83, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '169', '1127', '-p', 'DEFAULT_SETSIZE', '7', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(84, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '169', '1127', '-p', 'DEFAULT_SETSIZE', '7', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], '').
cli_testcase(85, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/PairRelation.mch', '-mc', '100', '-cc', '39', '173', '-p', 'DEFAULT_SETSIZE', '3', '-strict'], '').
cli_testcase(86, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/PairRelation.mch', '-mc', '100', '-cc', '11', '47', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(87, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/PairRelation.mch', '-mc', '100', '-cc', '39', '75', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(88, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/SetRelation.mch', '-mc', '100', '-cc', '38', '172', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(89, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/SetRelation.mch', '-mc', '100', '-cc', '10', '46', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(90, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/SetRelation.mch', '-mc', '100', '-cc', '38', '74', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(91, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '81', '177', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'off', '-nogoal', '-strict'], '').
cli_testcase(92, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '48', '105', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-nogoal', '-strict'], '').
cli_testcase(93, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '48', '105', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-nogoal', '-strict', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(94, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '81', '138', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'flood', '-nogoal', '-strict'], '').
cli_testcase(95, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/TicTacToe_SimplerSym.mch', '-mc', '1500', '-cc', '1450', '4547', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-nogoal', '-nodead', '-strict', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(96, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/TicTacToe_SimplerSym.mch', '-mc', '1500', '-cc', '1450', '4547', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-nogoal', '-nodead', '-strict'], '').
cli_testcase(97, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/TicTacToe_SimplerSym.mch', '-mc', '6000', '-cc', '5480', '8577', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'flood', '-nogoal', '-nodead', '-strict'], '').
cli_testcase(98, [symmtest], ['examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '355', '11220', '-strict'], '').
cli_testcase(99, [symmtest], ['examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '355', '11220', '-strict'], '').
cli_testcase(100, [symmtest], ['examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'flood', '-cc', '694', '11572', '-strict'], '').
cli_testcase(101, [symmtest], ['examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '10000', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '3013', '97730', '-strict'], '').
cli_testcase(102, [symmtest,proz], ['examples/Z/Derrick/video_club_abz08.fuzz', '-strict', '-card', 'TITLE', '2', '-card', 'PERSON', '2', '-mc', '100000', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '113', '596', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(103, [symmtest], ['examples/B/Puzzles/Hanoi6Sym.mch', '-mc', '10000', '-p', 'SYMMETRY_MODE', 'off', '-cc', '731', '2187', '-nogoal', '-strict'], '').
cli_testcase(104, [symmtest], ['examples/B/Puzzles/Hanoi6Sym.mch', '-mc', '10000', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '367', '1096', '-nogoal', '-strict'], '').
cli_testcase(105, [symmtest], ['examples/B/Puzzles/Hanoi6Sym.mch', '-mc', '10000', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '367', '1096', '-nogoal', '-strict'], '').
cli_testcase(106, [symmtest2], ['-p', 'MAXINT', '5', 'examples/B/Demo/DiningCryptographers.mch', '-mc', '100', '-cc', '65', '96', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(107, [symmtest2], ['-p', 'MAXINT', '5', 'examples/B/Demo/DiningCryptographers.mch', '-mc', '100', '-cc', '65', '96', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(108, [symmtest2], ['-p', 'MAXINT', '5', 'examples/B/Demo/DiningCryptographers.mch', '-mc', '100', '-cc', '65', '96', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], '').
cli_testcase(109, [symmtest2], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/Fraikin/Library_FullSym.mch', '-mc', '10000', '-cc', '1158', '12064', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'hash', '-strict', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(110, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '149', '248'], '').
cli_testcase(111, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '82', '137', '-p', 'SYMMETRY_MODE', 'nauty'], '').
cli_testcase(112, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '82', '137', '-p', 'SYMMETRY_MODE', 'hash'], '').
cli_testcase(113, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '149', '204', '-p', 'SYMMETRY_MODE', 'flood'], '').
cli_testcase(114, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '144', '332', '-p', 'SYMMETRY_MODE', 'nauty'], '').
cli_testcase(115, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '144', '332', '-p', 'SYMMETRY_MODE', 'hash', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(116, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '4', '-mc', '10000', '-strict', '-nodead', '-cc', '360', '1093', '-p', 'SYMMETRY_MODE', 'nauty'], '').
cli_testcase(117, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '4', '-mc', '10000', '-strict', '-nodead', '-cc', '360', '1093', '-p', 'SYMMETRY_MODE', 'hash' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(118, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '1339', '3787', '-p', 'SYMMETRY_MODE', 'hash' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(119, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '1339', '3787', '-p', 'SYMMETRY_MODE', 'nauty'], '').
cli_testcase(120, [symmtest2], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '3893', '10970', '-p', 'SYMMETRY_MODE', 'off'], '').
cli_testcase(121, [symmtest2], ['examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '8', '18', '-strict' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(122, [symmtest2], ['examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '8', '18', '-strict'], '').
cli_testcase(123, [symmtest2], ['examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '13', '47', '-strict'], '').
cli_testcase(124, [symmtest2], ['examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '13', '47', '-strict' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(125, [symmtest2], ['examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '53', '345', '-strict'], '').
cli_testcase(126, [symmtest2], ['examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '53', '345', '-strict' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(127, [eventb_test], ['examples/EventBPrologPackages/TestValidation_ctx.eventb', '-init', '-aa', '11', '0', '0', '-strict'], '').
cli_testcase(128, [eventb_test], ['examples/EventBPrologPackages/platoon/platoon_1_mch.eventb', '-t', '-strict'], '').
cli_testcase(129, [eventb_test], ['examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-t', '-strict'], '').
cli_testcase(130, [eventb_test], ['examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-t', '-p', 'CLPFD', 'TRUE', '-strict'], '').
cli_testcase(131, [eventb_test], ['examples/EventBPrologPackages/platoon/platoon_3_0_mch.eventb', '-t', '-strict'], '').
cli_testcase(132, [eventb_test], ['examples/EventBPrologPackages/platoon/platoon_4_0_mch.eventb', '-t', '-strict'], '').
cli_testcase(133, [eventb_test], ['examples/EventBPrologPackages/Records/SampleThreeRecordsTranslated.mch', '-t', '-strict', '-mc', '100', '-cc', '4', '6'], '').
cli_testcase(134, [eventb_test], ['examples/EventBPrologPackages/Records/INTEGER_Record.mch', '-t', '-strict', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE'], '').
cli_testcase(135, [eventb_test], ['examples/EventBPrologPackages/Test1.0/Testid_mch.eventb', '-t', '-strict', '-pref_group', integer, int32, '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(136, [eventb_test], ['examples/EventBPrologPackages/sieve.eventb', '-t', '-strict'], '').
cli_testcase(137, [eventb_test], ['examples/EventBPrologPackages/TrafficLight/tlm_0.eventb', '-t', '-strict'], '').
cli_testcase(138, [eventb_test], ['examples/EventBPrologPackages/TrafficLight/tlm_1.eventb', '-t', '-strict'], '').
cli_testcase(139, [eventb_test], ['examples/EventBPrologPackages/TrafficLight/tlm_3.eventb', '-t', '-strict'], '').
cli_testcase(140, [eventb_test], ['examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-t', '-strict'], '').
cli_testcase(141, [eventb_test], ['examples/EventBPrologPackages/SET_Game/SET_Game_mch.eventb', '-t', '-strict', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(142, [eventb_test], ['examples/EventBPrologPackages/Stefan_SecureBuilding/M21_mch.eventb', '-mc', '10000', '-t', '-strict'], '').
cli_testcase(143, [eventb_test], ['examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas.eventb', '-mc', '200', '-t', '-strict', '-noass'], '').
cli_testcase(144, [eventb_test], ['examples/EventBPrologPackages/Test1.0/Testid_mch.eventb', '-t', '-strict', '-p', 'MAXINT', '99999', '-p', 'MININT', '-99999', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(145, [eventb_test], ['examples/EventBPrologPackages/Test1.0/Testprj_mch.eventb', '-t', '-strict', '-p', 'MAXINT', '99999', '-p', 'MININT', '-99999'], '').
cli_testcase(146, [eventb_test], ['examples/EventBPrologPackages/Test1.0/TestPartition_mch.eventb', '-t', '-strict'], '').
cli_testcase(147, [eventb_test], ['examples/B/EventB_AtelierB/Lift.sys', '-t', '-strict'], '').
cli_testcase(148, [eventb_test], ['examples/EventBPrologPackages/MultiLevel/TestMC_m1_mch.eventb', '-mc', '1000', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:evt1:simulation_error'], '').
cli_testcase(149, [eventb_test], ['examples/EventBPrologPackages/Convergence/countdown_r_mch.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-cc', '7', '7'], '').
cli_testcase(150, [eventb_test], ['examples/EventBPrologPackages/Convergence/anticipated_grd_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:decr:variant_negative'], '').
cli_testcase(151, [eventb_test], ['examples/EventBPrologPackages/Convergence/anticipated_var_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:inc:invalid_variant'], '').
cli_testcase(152, [eventb_test], ['examples/EventBPrologPackages/Convergence/anticipated_svar_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:add:invalid_variant'], '').
cli_testcase(153, [eventb_test], ['examples/EventBPrologPackages/Convergence/convergent_grd_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:decr:variant_negative'], '').
cli_testcase(154, [eventb_test], ['examples/EventBPrologPackages/Convergence/convergent_var_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:decr:invalid_variant'], '').
cli_testcase(155, [eventb_test], ['examples/EventBPrologPackages/Convergence/convergent_svar_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:do_nothing:invalid_variant'], '').
cli_testcase(156, [eventb_test], ['examples/EventBPrologPackages/Convergence/variant_integer.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-cc', '8', '19'], '').
cli_testcase(157, [eventb_test], ['examples/EventBPrologPackages/Convergence/variant_set.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-cc', '17', '97'], '').
cli_testcase(158, [eventb_test], ['examples/EventBPrologPackages/Merging/merging_R2.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-cc', '12', '11'], '').
cli_testcase(159, [eventb_test], ['examples/EventBPrologPackages/MultiLevel/TheoremsInGuardsOK.eventb', '-mc', '10', '-nodead', '-strict'], '').
cli_testcase(160, [eventb_test], ['examples/EventBPrologPackages/MultiLevel/TheoremsInGuardsKO.eventb', '-mc', '10', '-nodead', '-expcterr', 'event_error:inc:invalid_theorem_in_guard', '-strict'], '').
cli_testcase(161, [csp_test], ['examples/CSP/mydemos/CompareWithFDR/GenPrime.csp', '-t', '-strict'], '').
cli_testcase(162, [csp_test], ['examples/CSP/mydemos/demo/buses.csp', '-t', '-strict'], '').
cli_testcase(163, [csp_test], ['examples/CSP/mydemos/demo/Peterson_v2.csp', '-t', '-strict', '-mc',1000, '-cc', 233, 465], '').
cli_testcase(164, [csp_test], ['-c', 'examples/CSP/mydemos/demo/mbuff.csp', '-t', '-strict'], '').
cli_testcase(165, [csp_test], ['examples/CSP/mydemos/demo/crossing.csp', '-t', '-strict', '-mc', '2000'], '').
cli_testcase(166, [csp_test], ['examples/CSP/mydemos/SetCompTests.csp', '-t', '-strict'], '').
cli_testcase(167, [csp_test], ['examples/CSP/mydemos/SimpleCompLinkedPar.csp', '-t', '-strict'], '').
cli_testcase(168, [csp_test], ['-c', 'examples/CSP/mydemos/SequenceComprTests2.csp', '-t', '-strict', '-v'], '').
cli_testcase(169, [csp_test], ['examples/CSP/mydemos/BigUnionInterTests.csp', '-t', '-strict'], '').
cli_testcase(170, [csp_test], ['examples/B/Tickets/299/faust_deadlock.csp', '-mc', '100000', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(171, [csp_test], ['examples/CSP/FDRFeatureTests/LargeStateSpace.csp', '-mc', '999999', '-cc', '101102', '101101', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(172, [csp_test], ['examples/CSP/FDRFeatureTests/RenameRecordPattern.csp', '-mc', '999999', '-t', '-strict'], '').
cli_testcase(173, [csp_test], ['examples/CSP/FDRFeatureTests/RenameRecordPattern2.csp', '-t', '-assertions', '-strict'], '').
cli_testcase(174, [csp_test], ['examples/CSP/FDRFeatureTests/SimpleRefTest.csp', '-assertions', '-strict'], '').
cli_testcase(175, [csp_test], ['examples/CSP/fontaine/RecDataType.csp', '-t', '-strict'], '').
cli_testcase(176, [csp_test], ['examples/CSP/fontaine/RecDataTypeFunctions.csp', '-t', '-strict'], '').
cli_testcase(177, [csp_test], ['examples/CSP/simple/InSituRefChecks.csp', '-cspref', 'Q', '[F=', 'MAIN', '-cspref', 'MAIN', '[F=', 'Q', '-cspref', 'Q', '[T=', 'R', '-cspref', 'R', '[F=', 'Q', '-strict'], '').
cli_testcase(178, [csp_test], ['examples/CSP/simple/InSituRefChecks.csp', '-cspref', 'Q', '[F=', 'MAIN', '-cspref', 'Q', '[F=', 'R', '-strict', '-expcterr', 'refinement_check_fails'], '').
cli_testcase(179, [csp_test], ['examples/CSP/UML_Compass/Microwave.csp', '-assertions', '-strict'], '').
cli_testcase(180, [csp_test], ['examples/CSP/mydemos/demo/roscoe_section2-2.csp', '-assertions', '-strict'], '').
cli_testcase(181, [csp_test,csp_bench], ['examples/CSP/ref_benchmarks/SimpleDeadlockCheck.csp', '-assertions', '-strict'], '').
cli_testcase(182, [csp_test,csp_bench], ['examples/CSP/simple/DivChecks.csp', '-assertions', '-strict'], '').
cli_testcase(183, [csp_test,csp_bench], ['examples/CSP/other/Ivo/Failure_Divergence_Tests.csp', '-assertions', '-strict'], '').
cli_testcase(184, [csp_test], ['examples/CSP/other/Ivo/Deadlock_Skip_Tests.csp', '-assertions', '-strict'], '').
cli_testcase(185, [csp_test], ['examples/CSP/other/Ivo/Failures_Tests.csp', '-assertions', '-strict'], '').
cli_testcase(186, [csp_test], ['examples/CSP/other/Ivo/AgentErrorsLetWithIn.csp', '-assertions', '-strict'], '').
cli_testcase(187, [csp_test], ['examples/CSP/other/Ivo/Deterministic1.csp', '-assertions', '-strict'], '').
cli_testcase(188, [csp_test], ['examples/CSP/other/Ivo/Deterministic_simple_v2.csp', '-assertions', '-strict'], '').
cli_testcase(189, [csp_test], ['examples/CSP/ref_benchmarks/abp_chapter5_roscoe_bench.csp', '-assertions', '-strict'], '').
cli_testcase(190, [csp_test], ['examples/CSP/ref_benchmarks/ucs1.csp', '-assertions', '-strict'], '').
cli_testcase(191, [csp_test], ['examples/CSP/simple/check_print.csp', '-t', '-strict'], '').
cli_testcase(192, [csp_test], ['examples/CSP/simple/check_set_compr_pat.csp', '-t', '-mc', '1000', '-nodead', '-cc', '6', '10', '-strict'], '').
cli_testcase(193, [csp_test], ['examples/CSP/ref_benchmarks/failsim.csp', '-assertions', '-strict'], '').
cli_testcase(194, [typechecker_test], ['examples/B/Tickets/305/Environment.mch', '-t', '-strict'], '').
cli_testcase(195, [typechecker_test], ['examples/B/Tickets/308/Basic_Controller.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(196, [typechecker_test,unit], ['-typecheckertests', '-strict', '-expcterr', bmachine_static_checks], '').
cli_testcase(197, [counterex], ['examples/B/Counterexamples/CarlaTravelAgencyErr.mch', '-mc', '1000', '-expcterr', 'invariant_violation', '-strict'], '').
cli_testcase(198, [counterex], ['examples/EventBPrologPackages/Stefan_SecureBuilding/M21_err3_mch.eventb', '-mc', '1000', '-expcterr', 'invariant_violation', '-strict'], '').
cli_testcase(199, [counterex], ['examples/B/PerformanceTests/dfcheck_houseset.mch', '-mc', '1000', '-df', '-expcterr', 'invariant_violation', '-expcterr', 'invariant_violation', '-strict'], '').
cli_testcase(200, [counterex], ['examples/B/PerformanceTests/BreadthFirstTest.mch', '-mc', '100000', '-expcterr', 'invariant_violation', '-strict'], '').
cli_testcase(201, [counterex], ['examples/B/PerformanceTests/DepthFirstTest.mch', '-mc', '100000', '-expcterr', 'invariant_violation', '-strict'], '').
cli_testcase(202, [counterex], ['examples/B/SymmetryReduction/Peterson_err.mch', '-mc', '10000', '-expcterr', 'invariant_violation', '-strict'], '').
cli_testcase(203, [counterex], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-mc', '10000', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(204, [counterex], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M3_mch.eventb', '-mc', '10000', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(205, [counterex], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_infinite_int.eventb', '-t', '-p', 'CLPFD', 'TRUE', '-strict'], '').
cli_testcase(206, [log_performance_tests], ['examples/EventBPrologPackages/sieve.eventb', '-mc', '1000', '-l', 'log/sieve.log', '-strict'], '').
cli_testcase(207, [puzzles_test], ['examples/B/Demo/RussianPostalPuzzle.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(208, [puzzles_test], ['examples/B/Demo/TrainTorchPuzzle.mch', '-mc', '10000', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(209, [puzzles_test], ['examples/B/Puzzles/BlocksWorldGeneric.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(210, [puzzles_test], ['examples/B/Puzzles/Farmer.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(211, [puzzles_test], ['examples/B/Puzzles/Hanoi.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(212, [puzzles_test], ['examples/B/Puzzles/Hanoi6.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(213, [puzzles_test], ['examples/B/Puzzles/Hanoi8.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(214, [puzzles_test], ['examples/B/Puzzles/Hanoi6Sym.mch', '-p', 'SYMMETRY_MODE', 'hash', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(215, [puzzles_test], ['examples/B/Puzzles/Hanoi6Sym.mch','-p','SYMMETRY_MODE','hash','-mc','10000','-card','Stakes','4','-expcterr','goal_found'], '').
cli_testcase(216, [puzzles_test], ['examples/B/Puzzles/Hanoi8Sym.mch', '-p', 'SYMMETRY_MODE', 'hash', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(217, [puzzles_test], ['-mc', '10000', 'examples/XTL/Hanoi.P', '-nodead', '-expcterr', 'xtl_error'], '').
cli_testcase(218, [puzzles_test], ['examples/B/Puzzles/water.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(219, [puzzles_test], ['examples/B/Puzzles/LatinSquares.mch', '-mc', '10000', '-expcterr', deadlock], '').
cli_testcase(220, [puzzles_test], ['examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-p', 'TIME_OUT', '5000'], '').
cli_testcase(221, [puzzles_test], ['examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-p', 'TIME_OUT', '8000', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(222, [b_test], ['examples/B/Benchmarks/Cruise_finite1.mch', '-mc', '1400', '-cc', '1361', '25696', '-strict'], '').
cli_testcase(223, [error_checks], ['examples/B/ErrorMachines/CardInfOverflow.mch', '-mc', '100', '-strict', '-expcterr', 'well_definedness_error', '-expcterr', 'invariant_violation'], '').
cli_testcase(224, [error_checks], ['examples/B/ErrorMachines/TwoOpsSameName.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(225, [error_checks], ['examples/B/ErrorMachines/PreCondViolated.mch', '-mc', '10', '-strict', '-expcterr', 'precondition_error'], '').
cli_testcase(226, [error_checks], ['examples/B/ErrorMachines/AssertCondViolated.mch', '-mc', '10', '-strict', '-expcterr', 'assert_error'], '').
cli_testcase(227, [error_checks,while], ['examples/B/ErrorMachines/WhileLoopOk.mch', '-t', '-mc', '20', '-strict'], '').
cli_testcase(228, [error_checks,while], ['examples/B/ErrorMachines/WhileLoopVariantError.mch', '-mc', '10', '-strict', '-expcterr', 'while_variant_error'], '').
cli_testcase(229, [error_checks,while], ['examples/B/ErrorMachines/WhileLoopVariantErr2.mch', '-mc', '10', '-strict', '-expcterr', 'while_variant_error'], '').
cli_testcase(230, [error_checks,while], ['examples/B/ErrorMachines/WhileLoopInvariantError.mch', '-mc', '10', '-strict', '-expcterr', 'while_invariant_violation'], '').
cli_testcase(231, [error_checks], ['examples/B/ErrorMachines/FunLaws_wderror.mch', '-mc', '1000', '-strict', '-expcterr', 'well_definedness_error', '-expcterr', 'invariant_violation'], '').
cli_testcase(232, [error_checks], ['examples/B/ErrorMachines/UndefinedArithmetic/DivByZero.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(233, [error_checks], ['examples/B/ErrorMachines/TestWDProblems_mch.eventb', '-init', '-animate', '1', '-strict', '-expcterr', 'well_definedness_error', '-expcterr', 'invariant_violation', '-expcterr', 'event_error:divby0:action_not_executable'], '').
cli_testcase(234, [error_checks], ['examples/B/ErrorMachines/UndefinedArithmetic/DivByZeroInsideNotExists.mch', '-strict', '-init', '-expcterr', 'well_definedness_error', '-expcterr', 'setup_constants_fails'], '').
cli_testcase(235, [error_checks], ['examples/B/ErrorMachines/UndefinedArithmetic/PowerOfNegativeNumber.mch', '-strict', '-t', '-expcterr', 'well_definedness_error'], '').
cli_testcase(236, [error_checks], ['examples/B/ErrorMachines/UndefinedInter.mch', '-init', '-animate', '1', '-strict', '-expcterr', 'well_definedness_error','-expcterr', 'animate'], '').
cli_testcase(237, [error_checks], ['examples/B/ErrorMachines/UndefinedBigINTER.mch', '-init', '-animate', '1', '-strict', '-expcterr', 'well_definedness_error','-expcterr', 'animate'], '').
cli_testcase(238, [error_checks], ['examples/B/ErrorMachines/UndefinedArithmetic/MaxOfEmptySet.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(239, [error_checks], ['examples/B/ErrorMachines/UndefinedArithmetic/MinOfEmptySet.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(240, [error_checks], ['examples/B/ErrorMachines/UndefinedSeqOperations/LastEmptySeq.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(241, [error_checks], ['examples/B/ErrorMachines/UndefinedSeqOperations/FrontEmptySeq.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(242, [error_checks], ['examples/B/ErrorMachines/OpCalls/Caller.mch', '-strict', '-expcterr', 'precondition_error', '-mc', '100'], '').
cli_testcase(243, [error_checks], ['examples/B/ErrorMachines/MultiAssignInit_bug.mch', '-t', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(244, [error_checks], ['examples/B/ErrorMachines/TypeError1.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(245, [error_checks], ['examples/B/ErrorMachines/TypeError2.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(246, [error_checks], ['examples/B/ErrorMachines/TypeError3.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(247, [error_checks], ['examples/B/ErrorMachines/TypeError4.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(248, [error_checks], ['examples/B/ErrorMachines/CstDefError1.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(249, [error_checks], ['examples/B/ErrorMachines/CstDefError2.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(250, [error_checks], ['examples/B/ErrorMachines/ParameterError1.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(251, [error_checks], ['examples/B/ErrorMachines/TicTacToe_ErrOp.mch','-strict','-expcterr','type_error','-expcterr','load_main_file','-expcterr',bmachine_static_checks], '').
cli_testcase(252, [error_checks], ['examples/B/ErrorMachines/AssignmentToParameterErr.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(253, [error_checks], ['examples/B/ErrorMachines/NoTypingForPara.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(254, [error_checks], ['examples/B/ErrorMachines/RecursiveDefinition.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(255, [tickets], ['examples/B/Tickets/184/ModeProtocolMachine_mch.eventb', '-mc', '20000', '-noass', '-card', 'MANAGERS', '1', '-p', 'MAX_INITIALISATIONS', '10', '-p', 'CLPFD', 'TRUE', '-strict' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(256, [tickets], ['examples/B/Tickets/184/ModeProtocolMachine_mch.eventb', '-mc', '200000', '-noass', '-p', 'MAX_INITIALISATIONS', '10', '-c', '--goal', 'card(incoming(MnAOC))>0 & card(outgoing(MnAOC))>0', '-strict', '-expcterr', 'goal_found'], '').
cli_testcase(257, [tickets], ['examples/B/Tickets/102/Bijections_mch.eventb', '-t', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(258, [tickets], ['examples/B/Tickets/102/PartialInjections_mch.eventb', '-t', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(259, [tickets], ['examples/B/Tickets/102/PartialFunctions_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(260, [tickets], ['examples/B/Tickets/102/Relations_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(261, [tickets], ['examples/B/Tickets/102/TotalFunctions_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(262, [tickets], ['examples/B/Tickets/102/TotalInjections_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(263, [tickets], ['examples/B/Tickets/102/TotalSurjections_mch.eventb', '-mc', '1000', '-t', '-nodead', '-strict'], '').
cli_testcase(264, [tickets], ['examples/B/Tickets/102/PartialSurjections_mch.eventb', '-mc', '1000', '-t', '-nodead', '-strict'], '').
cli_testcase(265, [tickets], ['examples/B/Tickets/102/TotalRelations_mch.eventb', '-mc', '1000', '-t', '-nodead', '-strict'], '').
cli_testcase(266, [tickets], ['examples/B/Tickets/Sven1/AbsIntern.mch', '-t', '-strict'], '').
cli_testcase(267, [tickets], ['examples/B/Tickets/Sven1/AbsInternCst.mch', '-t', '-strict'], '').
cli_testcase(268, [tickets], ['examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-properties'], '').
cli_testcase(269, [tickets], ['examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'TRUE'], '').
cli_testcase(270, [tickets], ['examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-p', 'MAXINT', '100', '-properties', '-p', 'SYMBOLIC', 'TRUE'], ''). % , '-p', 'EXPAND_CLOSURES_FOR_STATE', 'FALSE'
cli_testcase(271, [tickets], ['examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-p', 'MAXINT', '100', '-properties', '-p', 'SYMBOLIC', 'TRUE'], 'Valerio1 Ticket. Can not be run on windows due to memory restrictions.'). % , '-p', 'EXPAND_CLOSURES_FOR_STATE', 'TRUE'
cli_testcase(272, [tickets], ['examples/B/Tickets/Butler1/M2_mch.eventb', '-mc', '1000', '-strict', '-expcterr', 'invariant_violation', '-expcterr', 'event_error:Enter:simulation_error'], '').
cli_testcase(273, [tickets], ['examples/B/Tickets/Valerio3/BV16_DEFINITION.mch', '-t', '-strict', '-properties', '-p', 'SYMBOLIC', 'TRUE', '-p', 'CLPFD', 'TRUE'], ''). % , '-p', 'EXPAND_CLOSURES_FOR_STATE', 'FALSE'
cli_testcase(274, [tickets], ['examples/B/Tickets/Valerio3/POWER2.mch', '-init', '-strict', '-aa', '19', '0', '0'], '').
cli_testcase(275, [tickets], ['examples/B/Tickets/Valerio3/POWER2.mch', '-init', '-strict', '-ma', '18', '0', '0'], '').
cli_testcase(276, [tickets], ['examples/B/Tickets/Valerio4/FailingInit1.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expcterr', bmachine_static_checks], '').
cli_testcase(277, [tickets], ['examples/B/Tickets/Valerio4/FailingInit2.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expecterr', 'well_definedness_error', '-expcterr', bmachine_static_checks], '').
cli_testcase(278, [tickets], ['examples/B/Tickets/Valerio4/FailingInit3.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expecterr', 'well_definedness_error', '-expcterr', bmachine_static_checks], '').
cli_testcase(279, [tickets], ['examples/B/Tickets/Valerio4/FailingInit4.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expcterr', bmachine_static_checks], '').
cli_testcase(280, [tickets], ['examples/B/Tickets/Valerio4/FailingInit5.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expcterr', 'initialisation_fails', '-expcterr', bmachine_static_checks], '').
cli_testcase(281, [tickets], ['examples/B/Tickets/Valerio6/Z80_Simple.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(282, [tickets], ['examples/B/Tickets/Alstom1/TestBecomesElementOf.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(283, [tickets], ['examples/B/Tickets/Alstom1/TestPFBecomeSuch.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(284, [tickets], ['examples/B/Tickets/Alstom1/MAXINT_M1.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(285, [tickets], ['examples/B/Tickets/93/lift.eventb', '-strict', '-t', '-expcterr', 'event_error:ScheduleGoingUp:invalid_theorem_in_guard', '-expcterr', 'event_error:ScheduleGoingUp:invalid_modification'], '').
cli_testcase(286, [tickets], ['examples/B/Tester/TestExistsLetExtraction.mch', '-t', '-strict'], '').
cli_testcase(287, [tickets], ['examples/B/Tickets/177/loop_invariant.imp', '-strict', '-t'], '').
cli_testcase(288, [tickets], ['examples/B/Tickets/181/alstom_memory.mch', '-strict', '-t', '-mc', '100'], '').


cli_testcase(289, [laws,rel_fnc], ['examples/EventBPrologPackages/Laws/ExplicitComputationsEventB_ctx.eventb', '-init', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '7000'], 'Many explicit computations which check that all operators work as expected on various concrete data (for Event-B).').
cli_testcase(290, [laws], ['examples/EventBPrologPackages/Laws/LawsEventBSpecific.eventb', '-t', '-mc', '1000', '-nodead', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], '').
cli_testcase(291, [laws], ['examples/B/Laws/EqualityLaws.mch', '-t', '-mc', '100000', '-nodead', '-c', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], '').
cli_testcase(292, [laws,rel_fnc], ['examples/B/Laws/ExplicitComputations.mch', '-t', '-mc', '100000', '-nodead', '-c', '-strict', '-p', 'TIME_OUT', '7000'], 'Many explicit computations which check that all operators work as expected on various concrete data.').
cli_testcase(293, [laws,rel_fnc], ['examples/B/Laws/ExplicitComputations.mch', '-t', '-mc', '100000', '-nodead', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '9000'], 'Many explicit computations which check that all operators work as expected on various concrete data (CLPFD mode).').
cli_testcase(294, [laws], ['examples/B/Laws/BoolLaws.mch', '-mc', '100000', '-nodead', '-c', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], 'Various laws about BOOL datatype and boolean connectives.').
cli_testcase(295, [laws], ['examples/B/Laws/ArithmeticLaws.mch', '-mc', '100000', '-c', '-nodead', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], 'Various arithmetic laws.').
cli_testcase(296, [laws], ['examples/B/Laws/NatRangeLaws.mch', '-mc', '100000', '-c', '-nodead', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], '').
cli_testcase(297, [laws,rel_fnc], ['examples/B/Laws/RelLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], 'Use model checking to check a variety of laws on relations.').
cli_testcase(298, [laws], ['examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict'], '').
cli_testcase(299, [laws], ['examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'SYMBOLIC', 'TRUE'], '').
cli_testcase(300, [laws], ['examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(301, [laws], ['examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SYMBOLIC', 'TRUE'], '').
cli_testcase(302, [laws], ['examples/B/Laws/SetLawsNat.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict'], '').
cli_testcase(303, [laws], ['examples/B/Laws/SetLawsNat.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], '').
cli_testcase(304, [laws], ['examples/B/Laws/SetLawsNat.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'SYMBOLIC', 'TRUE'], '').
cli_testcase(305, [laws,repl], ['examples/B/Laws/REPL_UNIT_TESTS0.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '28000'], '').
cli_testcase(306, [laws,repl], ['examples/B/Laws/REPL_UNIT_TESTS1.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '28000'], '').
cli_testcase(307, [laws,repl], ['examples/B/Laws/REPL_UNIT_TESTS.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '45000'], '').
cli_testcase(308, [laws,rel_fnc], ['examples/B/Daniel/rel_fnc.mch', '-t', '-mc', '20000', '-strict', '-p', 'CLPFD', 'TRUE'], 'A small machine to check that rel/fnc work as expected.').
cli_testcase(309, [laws], ['examples/B/Laws/CardinalityLaws.mch', '-mc', '2500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(310, [cbc_deadlock], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(311, [cbc_deadlock], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(312, [cbc_deadlock], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M2_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(313, [cbc_deadlock], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M3_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(314, [cbc_deadlock], ['examples/B/Demo/Lift.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(315, [cbc_deadlock], ['examples/B/Benchmarks/Cruise_finite1.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(316, [cbc_deadlock], ['examples/B/Demo/RussianPostalPuzzle.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(317, [cbc_deadlock], ['examples/B/CBC/NoDeadlock.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(318, [cbc_deadlock], ['examples/B/CBC/SimpleDeadlock1.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(319, [cbc_deadlock], ['examples/B/CBC/DeadlockIntBools.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(320, [cbc_deadlock], ['examples/B/CBC/ProcessIDStarvation.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(321, [cbc_deadlock], ['examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(322, [cbc_deadlock], ['examples/B/CBC/MinSet_Deadlock.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(323, [cbc_deadlock], ['examples/EventBPrologPackages/Vizing_Breitner/m12_Deadlock_Freedom_mch.eventb', '-p','TIME_OUT','200','-cbc_deadlock', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-p','DEFAULT_SETSIZE',1], ''). /* with size 2 only works if SMT mode is allowed to instantiate */ 
cli_testcase(324, [cbc_deadlock], ['examples/EventBPrologPackages/Vizing_Breitner/m11_Conv_Stages_mch.eventb', '-cbc_deadlock', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict','-p', 'DEFAULT_SETSIZE',2], '').
cli_testcase(325, [cbc_deadlock], ['examples/EventBPrologPackages/Deadlock/BrakeSystem_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(326, [cbc_deadlock], ['examples/EventBPrologPackages/Deadlock/QueenEventsTF_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(327, [cbc_deadlock], ['examples/EventBPrologPackages/Deadlock/QueenEventsTF_mch.eventb', '-cbc_deadlock_pred', 'n=15', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(328, [cbc_deadlock], ['examples/EventBPrologPackages/Deadlock/QueenEventsTF_mch.eventb', '-cbc_deadlock_pred', 'n=32', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(329, [cbc_deadlock], ['examples/EventBPrologPackages/Alexei/ttrain_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(330, [cbc_deadlock], ['examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(331, [cbc_deadlock], ['examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(332, [cbc_deadlock], ['examples/EventBPrologPackages/platoon/platoon_1_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(333, [cbc_deadlock], ['examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(334, [cbc_deadlock], ['examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(335, [cbc_deadlock], ['examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m0_mch_deadlock.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(336, [cbc_deadlock], ['examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m0_mch_deadlock.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(337, [cbc_deadlock], ['examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m0_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(338, [cbc_deadlock], ['examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m1_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(339, [cbc_deadlock], ['examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m2_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(340, [cbc_deadlock], ['examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m3_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(341, [cbc_deadlock], ['examples/EventBPrologPackages/Abrial_Teaching/search/m_1b_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(342, [cbc_deadlock], ['examples/EventBPrologPackages/TrafficLight/tlm_1.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(343, [cbc_deadlock], ['examples/EventBPrologPackages/TrafficLight/tlm_2.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(344, [cbc_deadlock], ['examples/EventBPrologPackages/TrafficLight/tlm_3.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(345, [cbc_deadlock], ['examples/EventBPrologPackages/ProofDirected/benchmarks/siemens_mch_0.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(346, [cbc_deadlock], ['examples/EventBPrologPackages/ProofDirected/benchmarks/scheduler.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(347, [cbc_deadlock], ['examples/EventBPrologPackages/ProofDirected/benchmarks/earley_2.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(348, [cbc_deadlock], ['examples/EventBPrologPackages/ProofDirected/benchmarks/mondex_m2.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(349, [cbc_deadlock], ['examples/EventBPrologPackages/ProofDirected/benchmarks/cxcc0.eventb', '-cbc_deadlock', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-strict'], '').
cli_testcase(350, [cbc_deadlock], ['examples/EventBPrologPackages/ProofDirected/benchmarks/cxcc0.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'USE_RECORD_CONSTRUCTION', 'FALSE'], '').
cli_testcase(351, [cbc_deadlock], ['examples/EventBPrologPackages/ProofDirected/benchmarks/FMCH02.eventb', '-p', 'DEFAULT_SETSIZE', '1', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(352, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_mch_WithInit.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(353, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(354, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v2.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(355, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v3.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(356, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v4.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(357, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '1'], '').
cli_testcase(358, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '2', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(359, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '2', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(360, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '1', '-p', 'SMT', 'TRUE'], '').
cli_testcase(361, [cbc_deadlock], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SMT', 'TRUE'], '').
cli_testcase(362, [cbc_deadlock], ['examples/B/LTL/Turcanu/machineEco_mch4.eventb', '-cbc_deadlock', '-p', 'SMT', 'TRUE', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(363, [cbc_deadlock], ['examples/B/Mathematical/GraphIso/HardGraph.mch', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check_time_out', '-p', 'TIME_OUT', '80', '-p', 'MAX_INITIALISATIONS', '0', '-strict'], 'Test that time-out triggered within CBC Deadlock check').
cli_testcase(364, [cbc_deadlock], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(365, [puzzles_test], ['examples/B/Puzzles/StableMarriages.mch', '-t', '-strict', '-cc', '7', '15', '-mc', '100'], '').
cli_testcase(366, [puzzles_test], ['examples/B/Puzzles/StableMarriagesLarger.mch', '-t', '-strict', '-cc', '7', '46', '-mc', '100'], '').
cli_testcase(367, [puzzles_test], ['examples/B/Puzzles/GolombRuler.mch', '-t', '-p', 'CLPFD', 'TRUE', '-strict'], '').
cli_testcase(368, [puzzles_test], ['examples/B/Puzzles/Tischordnung.mch', '-t', '-strict', '-cc', '3', '7', '-mc', '100'], '').
cli_testcase(369, [puzzles_test], ['examples/B/Puzzles/Wegenetz.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(370, [puzzles_test], ['examples/B/Demo/RussianPostalPuzzle.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(371, [puzzles_test], ['examples/B/Demo/TrainTorchPuzzle.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(372, [puzzles_test], ['examples/B/Puzzles/BlocksWorldGeneric.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(373, [puzzles_test], ['examples/B/Puzzles/Farmer.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(374, [puzzles_test], ['examples/B/Puzzles/Hanoi.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(375, [puzzles_test], ['examples/B/Puzzles/water.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(376, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/EventBPrologPackages/Puzzles/CopyPastAsEvent_mch.eventb', '-t', '-strict'], '').
cli_testcase(377, [puzzles_test], ['examples/EventBPrologPackages/Puzzles/Sudoku_ctx.eventb', '-t', '-strict'], '').
cli_testcase(378, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/EventBPrologPackages/Puzzles/NQueensAsEvent_mch.eventb', '-t', '-strict'], '').
cli_testcase(379, [puzzles_test], ['examples/B/Puzzles/PexSolve.mch', '-t', '-strict'], '').
cli_testcase(380, [puzzles_test], ['examples/B/Puzzles/KnightsTour.mch', '-t', '-strict'], '').
cli_testcase(381, [puzzles_test], ['examples/B/Puzzles/KnightsTour_v2.mch', '-t', '-strict'], '').
cli_testcase(382, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/SudokuAsConstant.mch', '-t', '-strict'], '').
cli_testcase(383, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/SudokuHexAsConstant.mch', '-t', '-properties', '-sptxt', 'examples/B/Puzzles/SudokuHexAsConstant_sptxt.txt', '-strict'], '').
cli_testcase(384, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/SudokuAsConstantNoSolution.mch', '-init', '-strict', '-expcterr', 'setup_constants_fails'], '').
cli_testcase(385, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/LatinSquaresBig.mch', '-t', '-strict'], '').
cli_testcase(386, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/NQueens50.mch', '-t', '-properties', '-strict'], '').
cli_testcase(387, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/TwoQueensSevenKnights.mch', '-t', '-assertions', '-strict'], '').
cli_testcase(388, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/SendMoreMoney.mch', '-t', '-strict'], '').
cli_testcase(389, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/KissPassion.mch', '-t', '-strict'], '').
cli_testcase(390, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/CopySavePasteTools.mch', '-t', '-strict'], '').
cli_testcase(391, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/NoNeighboursProblem.mch', '-t', '-strict'], '').
cli_testcase(392, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/SubsetSum.mch', '-t', '-strict'], '').
cli_testcase(393, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/SubsetSum2.mch', '-t', '-strict'], '').
cli_testcase(394, [puzzles_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/WhoKilledAgatha.mch', '-t', '-mc', '100', '-p', 'MAX_INITIALISATIONS', '100', '-strict'], '').
cli_testcase(395, [puzzles_test], ['examples/EventBPrologPackages/SET_Game/SET_GAM_Sym_NoSet20_mch.eventb', '-mc', '1000', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'TIME_OUT', '7000', '-p', 'CLPFD', 'TRUE', '-df', '-goal', 'n=18', '-p', 'MAX_OPERATIONS', '20', '-strict', '-expcterr', 'goal_found'], '').
cli_testcase(396, [time_out_tests], ['examples/B/PerformanceTests/SMTLIB/DTP_k2_n35_c175_s1.mch', '-strict', '-init', '-expcterr', 'setup_constants_fails', '-expcterr', 'time_out'], '').
cli_testcase(397, [cbc], ['examples/B/Benchmarks/scheduler.mch', '-cbc', 'all', '-strict', '-p', 'SMT', 'TRUE'], '').
cli_testcase(398, [cbc], ['examples/B/CBC/NATURAL_Add.mch', '-cbc', 'all', '-strict', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(399, [cbc], ['examples/B/Demo/Lift.mch', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(400, [cbc], ['examples/B/Tutorial/DoubleCounter.mch', '-cbc', 'Double', '-strict', '-expcterr', 'cbc', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(401, [cbc], ['examples/B/Tutorial/DoubleCounter.mch', '-cbc', 'Double', '-strict', '-expcterr', 'cbc', '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', '150'], '').
cli_testcase(402, [cbc], ['examples/B/Special/ConstrainedBasedChecking/mesiSet.mch', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(403, [cbc], ['examples/B/Special/ConstrainedBasedChecking/mesiSet.mch', '-cbc', 'write', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(404, [cbc], ['examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(405, [cbc], ['examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'Delete', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(406, [cbc], ['examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'Run', '-strict'], '').
cli_testcase(407, [cbc], ['examples/EventBPrologPackages/Alexei/ttrain_mch.eventb', '-cbc', 'all', '-strict'], '').
cli_testcase(408, [cbc], ['examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc', 'bridge', '-strict'], '').
cli_testcase(409, [cbc], ['examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc', 'approach', '-strict', '-p', 'SMT', 'TRUE'], '').
cli_testcase(410, [cbc], ['examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc', 'exit', '-strict'], '').
cli_testcase(411, [cbc], ['examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc', 'bridge', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(412, [cbc], ['examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc', 'approach', '-strict', '-p', 'SMT', 'TRUE'], '').
cli_testcase(413, [cbc], ['examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc', 'exit', '-strict'], '').
cli_testcase(414, [cbc], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc', 'all', '-p', 'DEFAULT_SETSIZE', '4', '-p', 'SMT', 'TRUE', '-strict'], '').
cli_testcase(415, [cbc], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc', 'all', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SMT', 'TRUE', '-strict'], '').
cli_testcase(416, [cbc], ['examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc', 'all', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SMT', 'TRUE', '-strict'], '').
cli_testcase(417, [cbc], ['examples/EventBPrologPackages/Tests/Ticket167.eventb', '-cbc', 'inc', '-strict'], '').
cli_testcase(418, [cbc], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'redeal', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(419, [cbc], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'HasFlush', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(420, [cbc], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'hasFourOfAKind', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(421, [cbc], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'HasStraight', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(422, [cbc], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'AddACredit', '-cbc', 'deal', '-strict', '-cbc', 'selectme', '-cbc', 'invselectme'], '').
cli_testcase(423, [cbc], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-cbc', 'hasFullHouse', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(424, [cbc], ['examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(425, [cbc], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-cbc', 'all', '-strict'], '').
cli_testcase(426, [cbc], ['examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-cbc', 'all', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE'], '').
cli_testcase(427, [kodkod_test], ['-kodkod_comparision', '2', 'examples/B/Kodkod/expressions/functions.mch', '-strict'], '').
cli_testcase(428, [kodkod_test], ['-p', 'CLPFD', 'true', '-kodkod_comparision', '2', 'examples/B/Kodkod/expressions/triples.mch', '-strict'], '').
cli_testcase(429, [kodkod_test], ['-kodkod_comparision', '3', 'examples/B/Kodkod/expressions/relations.mch', '-strict'], '').
cli_testcase(430, [kodkod_test], ['-kodkod_comparision', '0', 'examples/B/Kodkod/expressions/sequences.mch', '-strict'], '').
cli_testcase(431, [kodkod_test], ['-kodkod_comparision', '1', 'examples/B/Kodkod/NoNeighboursProblem.mch', '-strict'], '').
cli_testcase(432, [kodkod_test], ['-p','MAX_INITIALISATIONS',50,'-kodkod_comparision', '0', 'examples/B/Kodkod/WhoKilledAgatha.mch', '-strict'], '').
cli_testcase(433, [analysis_test], ['-pacheck', 'examples/B/predicate_analysis/basic.mch', '-strict'], '').
cli_testcase(434, [analysis_test], ['-pacheck', 'examples/B/predicate_analysis/function_cards.mch', '-strict'], '').
cli_testcase(435, [analysis_test], ['-pacheck', 'examples/B/predicate_analysis/setext_dombug.mch', '-strict'], '').
cli_testcase(436, [csp_test], ['examples/CSP/ref_benchmarks/general4.csp', '-assertions', '-strict'], '').
cli_testcase(437, [b_test], ['examples/B/FeatureChecks/TestPrimedBecomeSuch.mch', '-t', '-strict'], 'Test for $0 usage in become such substitutions.').
cli_testcase(438, [csp_test], ['examples/CSP/ref_benchmarks/basin_olderog_bank.csp', '-assertions', '-strict'], '').
cli_testcase(439, [b_test], ['examples/B/FeatureChecks/EmptyString.mch', '-t', '-strict', '-assertions'], 'Check that we allow the empty string.').

cli_testcase(440, [b_test,animate], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-goal', 'c=10', '-check_goal', '-p', 'MAXINT', '100'], 'Check that animate executes exactly 10 operations after INITIALISATION.').
cli_testcase(441, [b_test,animate], ['examples/B/Simple/UpCounter.mch','-animate', '10', '-strict', '-goal', 'c=8', '-check_goal', '-p', 'MAXINT', '100'], 'Without -init we exacte SETUP_CONSTANTS, INITIALISATION and 8 operations.').
cli_testcase(442, [b_test,animate], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-goal', 'c=11', '-check_goal', '-p', 'MAXINT', '100', '-expcterr', 'check_goal'], 'Check that -check_goal fails if GOAL false.').
cli_testcase(443, [b_test,animate], ['examples/B/Simple/UpCounter.mch','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-det_check'], 'Check that all animation steps are deterministic.').
cli_testcase(444, [b_test,animate], ['examples/B/Simple/UpCounter_NonDet.mch','-animate', '4', '-strict', '-p', 'MAXINT', '100', '-det_check', '-expcterr', 'det_check'], 'Check that all -det_check detects non-deterministic steps.').
cli_testcase(445, [b_test,animate], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-goal', 'c=10 & c>1 & c<mx', '-check_goal', '-p', 'MAXINT', '100'], 'Check that -check_goal can also deal with conjuncts.').
cli_testcase(446, [b_test,animate], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '30', '-strict', '-goal', 'c=20', '-check_goal', '-p', 'MAXINT', '20', '-expcterr', 'animate'], 'Check that error raised if we cannot animate all steps -animate (but that we stay in the final state).').
cli_testcase(447, [b_test,animate], ['examples/B/Simple/UpCounter.mch', '-init','-animate_all', '-strict', '-goal', 'c=20', '-check_goal', '-p', 'MAXINT', '20'], 'Check that no error raised if we cannot animate all steps -animate_all (and that we stay in the final state).').

cli_testcase(448, [tickets], ['examples/B/Tickets/Hansen1/Bi_func.mch', '-mc', '1000', '-t', '-strict'], 'Check issue with existential quantifier inside set comprehension.').
cli_testcase(449, [tickets], ['examples/B/Tickets/Burdy1/TestOrApplyFun.mch', '-mc', '1000', '-t', '-strict'], 'Check issue with undefined function application in context of disjunct.').
cli_testcase(450, [tickets], ['examples/B/Tickets/Plagge1/KnightKnaves2.mch', '-init', '-strict'], 'Check that no time-out occurs in compiled mode').

cli_testcase(451, [b_test,animate,history], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', 'examples/B/Simple/UpCounter_his.txt'], 'Check that history properly written.').
cli_testcase(452, [b_test,animate,history], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', 'examples/B/Simple/UpCounter_his.txt','-his_option', 'show_states'], 'Check that history option -his_option show_states works.').
cli_testcase(453, [b_test,animate,history], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', 'examples/B/Simple/UpCounter_his.txt','-his_option', 'show_states','-his_option', 'show_init'], 'Check that history option -his_option show_states & show_init works.').
cli_testcase(454, [b_test,animate,history], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', 'examples/B/Simple/UpCounter_his.txt','-his_option', 'show_init'], 'Check that history option -his_option show_init works.').
cli_testcase(455, [b_test,animate,sptxt], ['examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-sptxt', 'examples/B/Simple/UpCounter_sptxt.txt'], 'Check that history option -sptxt works.').

cli_testcase(456, [puzzles_test], ['examples/B/Puzzles/Queens/QueensAllSolutions.mch', '-t', '-strict'], 'A version of N-Queens where all solutions are computed and we try to find N.').

cli_testcase(457, [laws,rel_fnc], ['examples/B/Laws/FunLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], 'Use model checking to check a variety of laws on functions.').
cli_testcase(458, [laws,rel_fnc], ['examples/B/Laws/FunLawsWithLambda.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], 'Use model checking to check a variety of laws on functions, including lambda abstractions.').
cli_testcase(459, [b_test,history], ['examples/B/Daniel/rel_fnc.mch', '-t',  '-strict', '-p', 'CLPFD', 'TRUE','-his', 'examples/B/xxdoesnotexist/rel_fnc_his.txt','-expcterr','open_file','-expcterr','history'], 'Check that we generate an error if hist file destination does not exist.').
cli_testcase(460, [b_test,sptxt], ['examples/B/Daniel/rel_fnc.mch', '-t',  '-strict', '-p', 'CLPFD', 'TRUE','-sptxt', 'examples/B/xxdoesnotexist/rel_fnc_his.txt','-expcterr','open_file','-expcterr','sptxt'], 'Check that we generate an error if sptxt file destination does not exist.').
cli_testcase(461, [b_test,history], ['examples/B/Daniel/rel_fnc.mch', '-t',  '-strict', '-p', 'CLPFD', 'TRUE','-his', 'examples/B/Daniel/rel_fnc_his.txt','-his_option', 'show_states','-his_option', 'show_init', '-p','SHOW_EVENTB_ANY_VALUES','TRUE'], 'A small machine to check that rel/fnc work as expected; here we check it in conjunction with -his option.').
cli_testcase(462, [b_test,history], ['examples/B/Simple/UpCounter.mch', '-mc', '100',  '-goal', 'c=10','-strict', '-p', 'MAXINT', '100', '-his', 'examples/B/Simple/UpCounter_his.txt','-his_option', 'show_states','-his_option', 'show_init','-expcterr','goal_found'], 'Check that history option works together with model checking.').
cli_testcase(463, [cbc_deadlock,history], ['examples/B/CBC/SimpleDeadlock1.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check','-his', 'examples/B/CBC/SimpleDeadlock1_his.txt','-his_option', 'show_states'], 'Check that -his works together with cbc_deadlock').
cli_testcase(464, [cbc,history], ['examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'Delete', '-strict', '-expcterr', 'cbc', '-his', 'examples/B/CBC/ProcessIDStarvation_InvErr_his.txt','-his_option', 'show_states'], 'Check that -his works with cbc invariant checking.').
cli_testcase(465, [b_test,animate,load_state], ['examples/B/Simple/UpCounter.mch', '-animate', '5', '-strict', '-p', 'MAXINT', '100', '-load_state', 'examples/B/Simple/UpCounter_saved10.P', '-goal', 'c=15',  '-check_goal'], 'Check that animate works after load_state.').
cli_testcase(466, [b_test,animate,while], ['examples/B/PerformanceTests/WhileLoopBench.mch', '-animate', '22', '-strict', '-p', 'TIME_OUT', 28000, '-his', 'examples/B/PerformanceTests/WhileLoopBench_his.txt', '-his_option' , 'show_states'], 'Check animate with operation with while loop.').
cli_testcase(467, [b_test,history], ['examples/B/FeatureChecks/IFTHENELSE.mch', '-t', '-strict','-his', 'examples/B/FeatureChecks/IFTHENELSE_his.txt','-his_option', 'show_states'], 'Check simple if-then-else works correctly, and check -his option writes operation arguments into file.').
cli_testcase(468, [b_test,history,sptxt], ['examples/B/Puzzles/Farmer.mch', '-mc', '1000', '-bf','-strict', '-nodead', '-expcterr', 'goal_found','-his', 'examples/B/Puzzles/Farmer_his.txt','-sptxt','examples/B/Puzzles/Farmer_sptxt.txt'], 'Check that -his and -sptxt work when used together').
cli_testcase(469, [b_test,sptxt], ['examples/B/PerformanceTests/LargeSets/LargeExplicitSet.mch', '-animate', '7', '-strict', '-nodead','-sptxt','examples/B/PerformanceTests/LargeSets/LargeExplicitSet_sptxt.txt'], 'Check that -sptxt expands large sets.').
cli_testcase(470, [b_test,sptxt,history], ['examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas.eventb', '-t', '-strict', '-noass','-his', 'examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_his.txt','-his_option','show_states','-sptxt','examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_sptxt.txt'], 'Check -his and sptxt on a larger Event-B model.').

cli_testcase(471, [b_test,sptxt,history], ['examples/B/Benchmarks/CarlaTravelAgencyErr.mch', '-t', '-strict', '-his', 'examples/B/Benchmarks/CarlaTravelAgencyErr_his.txt','-sptxt','examples/B/Benchmarks/CarlaTravelAgencyErr_sptxt.txt'], 'Check -his and -sptxt on a larger classical B model.').
cli_testcase(472, [b_test,sptxt,history], ['examples/B/SchneiderBook/Chapter12/TownsRRR.ref', '-t', '-strict', '-his', 'examples/B/SchneiderBook/Chapter12/TownsRRR_his.txt','-his_option','show_states','-sptxt','examples/B/SchneiderBook/Chapter12/TownsRRR_sptxt.txt'], 'Check -his and -sptxt on a classical B refinement.').
cli_testcase(473, [b_test,sptxt,history], ['examples/B/NewSyntax/SET_Game_Rec.mch', '-t', '-strict', '-his', 'examples/B/NewSyntax/SET_Game_Rec_his.txt','-his_option','show_states','-sptxt','examples/B/NewSyntax/SET_Game_Rec_sptxt.txt'], 'Check -his and -sptxt for classical B records.').
cli_testcase(474, [b_test,sptxt,history,dot], ['examples/B/ExpressionViewer/TotFunAnalyseTest.mch', '-ma', 1, 4, 0, '-dot_output', 'examples/B/ExpressionViewer/dot_output/', '-dot_all'], 'Check -dot_output, -dot_all works.').
cli_testcase(475, [laws,repl], ['examples/B/Laws/REPL_UNIT_TESTS2.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '60000'], 'Check that REPL unit tests work').
cli_testcase(476, [tickets,dot], ['examples/B/Tickets/Mejia1/TestInit_mch.eventb', '-t', '-mc', '1000', '-cc', 4, 13, '-strict', '-p', 'CLPFD', 'TRUE', '-spdot', 'examples/B/Tickets/Mejia1/TestInit_statespace.dot', '-p', 'DOT_PROP', 'TRUE', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Check that -spdot output works and that no warning generated for existential quantifier on disjunction in Event-B mode').
cli_testcase(477, [cbc], ['-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', 'examples/B/Puzzles/WhoKilledAgatha.mch', '-cbc_assertions'], 'Check CBC static assertion checking.').
cli_testcase(478, [cbc], ['-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', 'examples/B/Puzzles/WhoKilledAgatha_wrongassert.mch','-strict',  '-cbc_assertions','-expcterr',cbc_assertions], 'Check CBC static assertion checking with counter example.').
cli_testcase(479, [b_test,dot], ['-p', 'CLPFD', 'TRUE', '-mc', '100', '-p', 'MAX_INITIALISATIONS', 7, '-bf', 'examples/B/Puzzles/WhoKilledAgatha.mch','-spdot',  'examples/B/Puzzles/WhoKilledAgatha_statespace.dot','-strict', '-p', 'DOT_COLOR_ARC', steelblue, '-p', 'DOT_COLOR_NODE', olivedrab2, '-p', 'DOT_PEN_WIDTH', 2], 'Check Dot output with proper newline.').
cli_testcase(480, [b_test,dot], ['-p', 'CLPFD', 'TRUE', '-p', 'DOT_LOOPS', 'FALSE', 
'-p', 'DOT_ROOT', 'FALSE', '-p', 'DOT_LEAVES', 'FALSE', 
'examples/B/Benchmarks/Cruise_finite1.mch','-t' ,'-spdot',  'examples/B/Benchmarks/Cruise_finite1_statespace.dot','-strict', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Check Dot output with proper newline.').
cli_testcase(481, [b_test], ['examples/B/FeatureChecks/LET_multiple_variables.mch', '-t', '-strict'], 'Check LET with multiple variables.').
cli_testcase(482, [b_test], ['examples/B/FeatureChecks/LET_multiple_variables.mch', '-mc', 100, '-expcterr','general_error_occurred', '-expcterr','illegal_let', '-strict'], 'Check that LET with non-equation predicate does not work.').
cli_testcase(483, [csp_test,csp_bench1], ['examples/CSP/ref_benchmarks/abp_chapter5_roscoe_DLK_DIV_bench.csp', '-assertions', '-strict'], 'Check performance of Deadlock & DIV checks').
cli_testcase(484, [csp_test,csp_bench2], ['examples/CSP/NastyFDRDemos/LargeSpecRefineCheck.csp', '-assertions', '-strict'], 'Check performance with Large Spec and small Impl').
cli_testcase(485, [csp_test,csp_bench3], ['examples/CSP/simple/DeadlockTimeout.csp', '-assertions', '-strict'], 'Check that deadlock checker works with timeout generated tau transitions').
cli_testcase(486, [csp_test,csp_bench4], ['examples/CSP/simple/ParallelIntChoice.csp', '-assertions', '-strict'], 'Check that deadlock checker works with internal choice generated tau transitions').
cli_testcase(487, [csp_test], ['examples/CSP/simple/DivChecks3.csp', '-assertions', '-strict'], 'Check that ProB deals with Divergence correctly').
cli_testcase(488, [csp_test,csp_bench4], ['examples/CSP/simple/ParallelIntChoice.csp', '-assertions', '-strict', '-p',  'CSP_STRIP_SOURCE_LOC', 'TRUE'], 'Check that stripping of source location works.').
cli_testcase(489, [csp_test,csp_bench1], ['examples/CSP/ref_benchmarks/section7-1.csp', '-assertions', '-strict'], 'Check more CSP assertions from Roscoe book usc').
cli_testcase(490, [csp_test], ['examples/CSP/simple/exception_operator.csp', '-assertions', '-strict'], 'Check that new exception operator works').
cli_testcase(491, [csp_test,csp_bench], ['examples/CSP/ref_benchmarks/section7-2.csp', '-assertions', '-strict'], 'Check more CSP assertions from Roscoe book usc, here with exception operator').
cli_testcase(492, [csp_test],['examples/CSP/evans/traffic_spec.csp', '-assertions', '-strict'], 'Checking CSP assertions.').
cli_testcase(493, [tickets], ['examples/B/Tickets/Mejia2/ticket_simplified.mch', '-t', '-mc',100,'-assertions', '-strict'], 'Check cartesian product ticket').
cli_testcase(494, [laws], ['examples/B/Laws/SubstitutionLaws.mch', '-mc', '100', '-c', '-t', '-nodead', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE'], 'Check basic properties of B substitutions.').

cli_testcase(495, [error_checks], ['examples/B/ErrorMachines/SubstitutionSeqCompAssertErrCheck.mch', '-mc', '100', '-nodead', '-strict','-expcterr', 'assert_error'], 'Check that ProB finds error in Substitution assertion').
cli_testcase(496, [laws], ['examples/B/Laws/ExplicitComputationsCLPFD.mch', '-assertions', '-mc', '100000', '-nodead', '-c', '-strict', '-p', 'TIME_OUT', '7000'], 'Many explicit computations which check that all operators work as expected on various concrete data.').
cli_testcase(497, [b_test], ['examples/B/Benchmarks/Cruise_finite1.mch', '-pp', 'examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(498, [b_test], ['examples/B/Benchmarks/Cruise_finite1_prettyprinted.mch', '-pp', 'examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(499, [b_test], ['examples/B/Benchmarks/Cruise_finite1.mch', '-ppf', 'examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(500, [b_test], ['examples/B/Benchmarks/Cruise_finite1_prettyprintedf.mch', '-ppf', 'examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(501, [b_test], ['examples/B/Benchmarks/Cruise_finite1.mch', '-mc', '1400', '-cc', '1361', '25696', '-p', 'COMPRESSION', 'TRUE', '-strict'], 'Check that compression works').
cli_testcase(502, [laws], ['examples/B/Abrial_BBook/BBook_Page80.mch', '-init', '-assertions', '-strict'], 'Check example from B-Book').
cli_testcase(503, [laws], ['examples/B/Abrial_BBook/BBook_Page83.mch', '-init', '-assertions', '-strict'], 'Check example from B-Book').
cli_testcase(504, [laws], ['examples/B/Abrial_BBook/MembershipProperty4_Page96.mch', '-init',  '-expcterr', 'setup_constants_fails', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check rule from B-Book').
cli_testcase(505, [laws], ['examples/B/Abrial_BBook/MembershipProperty5_Page96.mch', '-init', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check that ProB finds counter example to erroneous rule from B-Book').
cli_testcase(506, [laws], ['examples/B/Abrial_BBook/MembershipProperty6_Page96.mch', '-init', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check that ProB finds counter example to erroneous rule from B-Book').
cli_testcase(507, [laws], ['examples/B/Laws/BoolLaws_SetCompr.mch', '-mc', '1000', '-nodead', '-c', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(508, [laws], ['examples/B/Laws/BoolLaws_SetComprCLPFD.mch', '-mc', '1000', '-nodead', '-c', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(509, [laws], ['examples/B/Laws/StringLaws.mch', '-mc', '500',  '-nogoal', '-nodead', '-strict'], 'Some Laws about the STRING datatype').
cli_testcase(510, [laws], ['examples/B/Laws/RelLaws.mch', '-init','-ppf', 'examples/B/Laws/RelLaws_internal.mch', '-strict'], 'Check that Pretty Printing works on a file with many operators').
cli_testcase(511, [csp_test,csp_deterministic1], ['examples/CSP/other/Ivo/Deterministic1.csp', '-assertions', '-strict'], 'Check more CSP deterministic assertions').
cli_testcase(512, [csp_test,csp_deterministic2], ['examples/CSP/other/Ivo/Deterministic2_v2.csp', '-assertions', '-strict'], 'Check more CSP deterministic assertions').
cli_testcase(513, [csp_test,csp_waterboiler], ['examples/CSP/other/Ivo/WaterBoiler.csp', '-assertions', '-strict'], 'Check CSP assertions from WaterBoiler.csp example.').
cli_testcase(514, [tickets,dot], ['examples/B/Mathematical/Voting/NoCondorcetWinnerKodkod.mch', '-t', '-strict', '-assertions','-dot_output', 'examples/B/Mathematical/Voting/dot_output/', '-dot_all'], 'Check that we have no problem with the existential quantifier; check dot output for assertion with perm.').
cli_testcase(515, [b_test,dot], ['-p', 'MAXINT', '5', 'examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '36', '156', '-p', 'DEFAULT_SETSIZE', '3', '-df', '-strict','-spmdot','examples/B/Demo/scheduler_spmdot.dot','-spddot','examples/B/Demo/scheduler_spddot.dot','-spdot','examples/B/Demo/scheduler_spdot.dot', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Test reduced state space dot output.').
cli_testcase(516, [symmtest], ['examples/B/SymmetryReduction/Contrived/PartiallyEnumeratedSet.mch', '-mc', '100', '-cc', '3', '16', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], 'Check that we detect unused enumerated set elements as symmetrical').
cli_testcase(517, [symmtest], ['examples/B/SymmetryReduction/Contrived/PartiallyEnumeratedSet.mch', '-mc', '100', '-cc', '3', '16', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], 'Check that we detect unused enumerated set elements as symmetrical').
cli_testcase(518, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/monitor2.mch', '-mc', '100', '-cc', '15', '92', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], 'Check that we detect unused enumerated set elements as symmetrical').
cli_testcase(519, [symmtest], ['-p', 'MAXINT', '5', 'examples/B/Demo/monitor2.mch', '-mc', '100', '-cc', '81', '529', '-p', 'SYMMETRY_MODE', 'off', '-strict'], 'Check original without symmetry').
cli_testcase(520, [tickets,csp_test], ['examples/CSP/Tickets/Shi/hanoi7.csp', '-mc', '100', '-t', '-strict'], 'Check that exception does not occur (Ticket)').
cli_testcase(521, [tickets,csp_test], ['examples/CSP/Tickets/Shi/hanoi4.csp', '-assertions', '-strict'], 'Check that exception does not occur (Ticket)').
cli_testcase(522, [tickets,csp_test], ['examples/CSP/Tickets/ErrManagerLoop/PrintRoom.csp', '-mc', 100, '-strict','-expcterr', 'general_error_occurred', '-expcterr', check_channel_output_value], 'Check that channel output error detected and no infinite loop in printing error').
cli_testcase(523, [b_test], ['examples/B/PerformanceTests/InfiniteClosures4.mch', '-strict','-assertions'], 'Check that infininite function represented by union is detected and can be applied; also test setting MAXINT using 2**31-1').
cli_testcase(524, [tickets], ['examples/B/Tickets/Beauger1/PredProblem.mch', '-t', '-strict'], 'Check that we can use relational composition ; with pred.').
cli_testcase(525, [b_test], ['examples/B/PerformanceTests/InfiniteParityFunction.mch', '-t', '-strict', '-assertions'], 'Some more tests with infinite functions.').
cli_testcase(526, [b_test], ['examples/EventBPrologPackages/Tests/FiniteInfiniteFunctions_ctx.eventb', '-assertions', '-strict'], 'Some more tests with infinite functions and the finite check in Event-B.').
cli_testcase(527, [b_test], ['examples/B/PerformanceTests/InfinitePartialSqrtFunction.mch', '-t', '-strict'], 'Check that we can deal with a partial infinite function.').
cli_testcase(528, [b_test,external], ['examples/B/ExternalFunctions/UsesLibraryMath.mch', '-t', '-strict'], 'Check that external functions work.').
cli_testcase(529, [b_test,external], ['examples/B/ExternalFunctions/PrologFun.mch', '-assertions', '-strict'], 'Check that external functions PROLOG_FUN work.').
cli_testcase(530, [b_test,external], ['examples/B/ExternalFunctions/PrintExternal.mch', '-t', '-strict'], 'Check that external printing functions work.').
cli_testcase(531, [b_test,external], ['examples/B/ExternalFunctions/LibraryStrings.mch', '-assertions', '-strict'], 'Check that external String functions work.').
cli_testcase(532, [symmtest], ['examples/B/SymmetryReduction/mondex_m1_c1_2.eventb', '-mc', 1000, '-p','SYMMETRY_MODE', hash, '-p','MAXINT',20, '-p','CLPFD', 'TRUE','-cc', 269, 4079, '-strict'], 'Check that symmetry reduction and detection of unused constants also works if we have too many disequality axioms.').
cli_testcase(533, [b_test], ['examples/B/PerformanceTests/InfiniteLanguageConcatenation.mch', '-mc', 1000, '-nodead','-assertions', '-strict'], 'Check that this function is detected as infinite and not expanded.').
cli_testcase(534, [tickets], ['examples/B/Tickets/Hansen3/TestingCircDef2.mch', '-mc', 1000, '-nodead','-assertions', '-strict'], 'Check that the erroneous circular definition problem is gone.').
cli_testcase(535, [tickets,bparser], ['examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-evalt', '1+1=3', '-expcterr', 'eval_string'], 'Test that Prolog parser for eval strings works and that erroneous result caught.').
cli_testcase(536, [tickets,bparser], ['examples/B/PerformanceTests/InfiniteParityFunction.mch', '-init', '-strict', '-evalt', 'not(finite({x|x>2}))','-evalt', 'not(finite(NATURAL))', '-evalf', 'finite({x|x>2})', '-evalt', 'finite(NAT1)', '-evalt', 'NATURAL /: FIN1(INTEGER)'], 'Test that Prolog parser for eval strings works.').
cli_testcase(537, [tickets,bparser], ['examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-p','MAXINT',10, '-evalt', 'NATURAL1/<:{1,2,3,7,9,10}', '-expcterr', 'eval_string_enum_warning'], 'Test that eval_string catches expansion errors.').
cli_testcase(538, [tickets,bparser], ['examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-p','MAXINT',10, '-evalt', '1..10 <: NATURAL1'], 'Test that these predicates do not cause enumeration warnings.').
cli_testcase(539, [b_test,kodkod_test], ['examples/B/Kodkod/CrewAllocationConstantsLarge.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-p', 'KODKOD', 'TRUE', '-properties', '-assertions'], 'Check that Kodkod translator works and that solution is ok (via -assertions). Skipped on Windows because of Memory restrictions.').
cli_testcase(540, [cbc], ['examples/B/TestCaseGeneration/SimpleCounter.mch', '-cbc_tests', 5 ,'count=MAX-1', 'examples/B/TestCaseGeneration/SimpleCounter_results.xml', '-cbc_cover', 'Increase', '-cbc_cover',  'Reset', '-cbc_cover',  'Halfway', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check that CBC test-case generation works.').
cli_testcase(541, [b_test,ltl], ['examples/CSP/Tickets/WilliamsLTL/SimpleLTL.csp', '-ltlformula', 'G([a] => F(e(a)))', '-strict'],'Check LTL formula in File.').

cli_testcase(542, [csp_test,ltl], ['examples/CSP/Tickets/WilliamsLTL/SimpleLTL.csp', '-ltlformula', 'not (GF e(a) => GF [a]) => F [b]', '-strict'],'Check LTL formula in File.').
cli_testcase(543, [csp_test,ltl], ['examples/CSP/other/Ivo/VendingMachine.csp', '-ltlformula', 'G([coin] => F (e(coffee) & e(tea)))', '-strict'],'Check LTL formula in File.').
cli_testcase(544, [csp_test,ltl], ['examples/CSP/other/Ivo/VendingMachine.csp', '-ltlformula', 'GF(e(coin))', '-strict'],'Check LTL formula in File.').
cli_testcase(545, [csp_test,ltl], ['examples/CSP/other/Ivo/dot_tuples_ltl_simple.csp', '-ltlformula', 'G([d] => F(e(ch.a)))', '-strict'],'Check LTL formula in File.').
cli_testcase(546, [csp_test,ltl], ['examples/CSP/other/Ivo/dot_tuples_ltl_simple.csp', '-ltlformula', 'GF(e(c.b.1.10))', '-strict'],'Check LTL formula in File.').

cli_testcase(547, [csp_test,ltl], ['examples/CSP/other/Ivo/Simple_LTL_Tupels.csp', '-ltlformula', 'GF([tau] => e(c.b.10.10))', '-strict'],'Check LTL formula in File.').
cli_testcase(548, [csp_test,ltl], ['examples/CSP/other/Ivo/Simple_LTL_Tupels.csp', '-ltlformula', 'GF([tau] => not e(c.a.10.10))', '-strict'],'Check LTL formula in File.').
cli_testcase(549, [csp_test,ltl], ['examples/CSP/other/Ivo/Simple_LTL_Tupels.csp', '-ltlformula', '[e] U [c.a.1.10]', '-strict'],'Check LTL formula in File.').


cli_testcase(550, [b_test,recursion,external], ['examples/B/ExternalFunctions/SortSet.mch', '-assertions', '-strict', '-p', 'CLPFD', 'TRUE'], 'Check that recursive sorting function works.').
cli_testcase(551, [csp_test], ['examples/CSP/other/Ivo/TestingCardinality.csp', '-assertions', '-strict'], '').
cli_testcase(552, [csp_test], ['examples/CSP/other/Ivo/DotTupleTestCase.csp', '-assertions', '-strict'], '').
cli_testcase(553, [b_test,recursion,external], ['examples/B/RecursiveFunctions/RecursiveSeqLenFunction.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that recursive functions work in non-symbolic mode.').
cli_testcase(554, [b_test,recursion,external], ['examples/B/RecursiveFunctions/Factorial.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that other recursive function works.').
cli_testcase(555, [b_test,recursion,external], ['examples/B/RecursiveFunctions/RecursiveSigmaCHOOSEv1.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE', '-assertions'], 'Test that union of explicit set and recursive lambda function works.').
cli_testcase(556, [tickets], ['examples/B/Tickets/DefinitionCallSwallow/PerserDefBug.mch', '-strict', '-mc', '1'], 'Regression test for Definitions Bug in Parser').
cli_testcase(557, [b_test,recursion,external], ['examples/B/RecursiveFunctions/Tests/NestedDef.mch', '-init', '-assertions','-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that we do not accidentally treat nested definion of variable with same name as abstract constant.').
cli_testcase(558, [b_test,recursion,external], ['examples/B/RecursiveFunctions/Tests/RecursiveIntegerEquation.mch', '-init', '-assertions','-strict', '-p', 'SYMBOLIC', 'FALSE', '-mc', 100, '-nodead'], 'Test that we do not accidentally treat equation over non-relation type.').
cli_testcase(559, [b_test,recursion,external], ['examples/B/RecursiveFunctions/Tests/RecursiveEnumEquation.mch', '-init', '-assertions','-strict', '-p', 'SYMBOLIC', 'FALSE', '-mc', 100, '-nodead'], 'Test that we do not accidentally treat equation over non-relation type.').
cli_testcase(560, [b_test,csp_test], ['examples/CSPB/Dining/Dining.mch', '-mc', '1000', '-strict'], 'Test that the dining philosophers without guidance by CSP have no deadlock').
cli_testcase(561, [b_test,csp_test], ['examples/CSPB/Dining/Dining.mch', '-csp-guide', 'examples/CSPB/Dining/Dining.csp', '-mc', '1000', '-strict', '-expecterr', 'deadlock'], 'Test that the dining philosophers with guidance by CSP have a deadlock').
cli_testcase(562, [b_test], ['examples/EventBPrologPackages/Pitesti/BankAccount_mch.eventb', '-test_description', 'examples/EventBPrologPackages/Pitesti/test_description.xml', '-strict'], 'Simple constraint based test case generation').
cli_testcase(563, [b_test], ['examples/B/Daniel/TestM.mch', '-mcm_tests', '20', '1000', 'state = End', 'examples/B/Daniel/TestM_testcases.xml', '-strict'],'State based test generation').
cli_testcase(564, [b_test,csp_test], ['examples/CSPB/Tickets/Treharne1/Test_corrected.mch', '-csp-guide', 'examples/CSPB/Tickets/Treharne1/Test-ctrl.csp', '-mc', '1000', '-strict', '-cc', 5, 6], 'Test that BOOL synchronises with CSP booleans').

cli_testcase(565, [b_test,proz],['examples/Z/Testcases/comsets.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(566, [b_test,proz],['examples/Z/Testcases/distributed_concatenation.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(567, [b_test,proz],['examples/Z/Testcases/compaction.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(568, [b_test,proz],['examples/Z/Testcases/zrm_refcard_ops.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(569, [b_test,ltl], ['examples/B/SymmetryReduction/Fraikin/Library.mch','-p','IGNORE_HASH_COLLISIONS',
'TRUE','-p','FORGET_STATE_SPACE','FALSE','-p','SYMMETRY_MODE','hash', '-expcterr', 'model_check_incomplete','-mc','9999',
'-ltlformula','G( e(Reserve(m1,b1)) => not({b1:dom(reservation) & m1 : ran(reservation(b1))}) )', '-strict'],'Check enabled e() with multiple parameters').
cli_testcase(571, [b_test], ['examples/B/Daniel/CommonSubLetBug.mch','-mc', '100', '-nodead', '-strict'], 'Regression test regarding let predicate and common subexpression elimination').
cli_testcase(572, [b_test,ltl],['examples/B/LTL/DerrickSmith/Abstract.mch', '-ltlassertions', '-strict'],'Check ltl assertions from DEFINITIONS').
cli_testcase(573, [b_test,ltl],['examples/B/LTL/Mikrowelle.mch', '-ltlassertions', '-strict'],'Check ltl assertions from DEFINITIONS').
cli_testcase(574, [b_test,ltl],['examples/CSPB/Dining/Dining.mch', '-ltlassertions', '-expcterr', ltl, '-strict'],'Check ltl assertions from DEFINITIONS in Dining.mch').
cli_testcase(575, [csp_test,ltl],['examples/CSPB/Dining/Dining.csp', '-ltlassertions', '-strict'],'Check ltl assertions from Pragmas in Dining.csp').
cli_testcase(576, [b_test,ltl],['examples/CSPB/Dining/Dining.mch', '-csp-guide', 'examples/CSPB/Dining/Dining.csp', '-ltlassertions', '-strict'],'Check ltl assertions from both Dining models in Csp and B.').
cli_testcase(577, [b_test,ltl],['examples/B/LTL/DerrickSmith/Abstract.mch', '-ltlformula', 'G([AOp2] => F(sink))', '-strict'],'Check ltl formula').
cli_testcase(578, [b_test,ltl],['examples/B/LTL/DerrickSmith/Abstract.mch', '-ltlformula', 'G([AOp2] => F(deadlock))', '-strict'],'Check ltl formula').
cli_testcase(579, [csp_test],['examples/CSP/mydemos/ReplicatedSharing.csp', '-mc', '100000', '-t', '-strict'],'Model check and trace replicated sharing processes.').
cli_testcase(580, [csp_test],['examples/CSP/mydemos/SimpleReplicatedLinkedParallel.csp', '-mc', '100000', '-t', '-strict'],'Model check and trace replicated linked parallel processes.').
cli_testcase(581, [csp_test],['examples/CSP/mydemos/SeqTests.csp', '-mc', '100000', '-nodead', '-t', '-strict'], 'Testing CSP sequences.').
cli_testcase(582, [csp_test],['examples/CSP/other/Ivo/Sequences.csp', '-mc', '100000', '-t', '-strict'], 'Testing CSP sequence structures.').
cli_testcase(583, [b_test,tla],['examples/TLA/GraphIso/GraphIso.tla', '-t', '-strict'], 'Testing TLA2B translation and loading.').
cli_testcase(584, [b_test,tla],['examples/TLA/ErrorTests/GraphIso.tla', '-init', '-strict', 
  '-expcterr', tla2b_parser, '-expcterr', load_main_file],'Check that TLA2B does not overwrite pre-existing machine file').
cli_testcase(585, [b_test,bparser],['-strict', 
  '-eval_file', 'examples/Eval/SendMoreMoney.eval', '-p', 'CLPFD', 'TRUE'],'Check that we can evaluate expr/pred from file without loading another machine').
cli_testcase(586, [b_test,tla],['examples/TLA/Login/Login_Deadlock.tla', '-init', '-strict', 
  '-mc', 100, '-p', 'SYMMETRY_MODE', hash, '-expcterr', deadlock],'Check that config file found and model checking TLA works').
cli_testcase(587, [csp_test],['examples/CSP/other/Ivo/SeqsAndSetsTest.csp', '-t', '-strict'], 'Testing the CSP-M builtin functions seq(-) and set(-).').
cli_testcase(588, [b_test],['examples/B/Pragmas/SymbolicPragmas.mch', '-init', '-assertions', '-strict'], 'Test that symbolic annotation pragma works.').
cli_testcase(589,[csp_test], ['examples/CSP/other/Ivo/InputOutputChannelValues.csp', '-mc', '10000', '-t', '-strict'], 'Testing complicated CSP-M datatypes.').
cli_testcase(590,[csp_test], ['examples/CSP/other/Ivo/CurryAgents.csp', '-mc', '10000', '-t', '-strict'], 'Testing Curry Agents in CSP-M models.').
cli_testcase(591,[csp_test], ['examples/CSP/other/Ivo/replicates.csp', '-mc', '10000', '-t', '-strict'], 'Testing replicate processes.').
cli_testcase(592,[csp_test], ['examples/CSP/other/Ivo/replicates2.csp', '-mc', '10000', '-t', '-strict'], 'Testing replicate processes.').
cli_testcase(593, [b_test], ['examples/B/Benchmarks/Cruise_finite1.mch', '-mc', '1400', '-cc', '1361', '25696', '-strict',  '-p', 'OPERATION_REUSE', 'TRUE'], 'Test Operation Reuse').
cli_testcase(594, [b_test,bparser], ['examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-evalt', '!x.(x:1..10=>x<11)', '-evalf', '!y.(y:1..10 => y<10)'], 'Test that universally quantified predicates work in eval_string.').
cli_testcase(595, [b_test,bparser],['-strict', 
  '-eval_rule_file', 'examples/Eval/KnightsKnaves.eval', '-p', 'CLPFD', 'TRUE'],'Check that we can evaluate expr/pred from file with universal quantification').
cli_testcase(596, [b_test,laws],['examples/B/Laws/TautologiesPL.mch', '-p', 'CLPFD', 'TRUE', '-mc',100, '-nodead', '-cc',9,8, '-strict'],'Check a list of tautologies').
cli_testcase(597,[csp_test], ['examples/CSP/other/Ivo/DotDatatypes.csp', '-mc', '10000', '-t', '-strict'], 'Testing complex datatypes.').
cli_testcase(598,[tickets], ['examples/B/Tickets/DefMchSameName/DefMchSameName.mch', '-ma', 7, 0 , 0, '-strict'], 'Testing that -main_assertions work properly.').
cli_testcase(599,[tickets], ['examples/B/Tickets/858/ubahn_mch4.eventb', '-init', '-expcterr', setup_constants_fails , '-strict'], 'Testing that no call_residue.').
cli_testcase(600,[tickets,symmetry], ['examples/B/Tickets/1043/SymTest.mch', '-mc', 300, '-cc', 3, 44 , '-p', 'SYMMETRY_MODE', hash, '-strict'], 'Testing that symmetry detected for enumerated set elements which are not used.').
cli_testcase(601,[tickets,symmetry], ['examples/B/Tickets/1043/SymTest.mch', '-mc', 300, '-cc', 3, 44 , '-p', 'SYMMETRY_MODE', nauty, '-strict'], 'Testing that symmetry detected for enumerated set elements which are not used.').
cli_testcase(602,[tickets,symmetry], ['examples/B/Tickets/1043/SymTest.mch', '-mc', 300, '-cc', 197, 238 , '-p', 'SYMMETRY_MODE', flood, '-strict'], 'Testing that symmetry detected for enumerated set elements which are not used.').
cli_testcase(603,[b_test], ['examples/B/Tester/DefinitionMultiInclude/Main.mch', '-init', '-aa', 3, 0,0,  '-strict'], 'Test same DEFINITION file include twice in different contexts.').
cli_testcase(604, [csp_test,ltl],['examples/CSP/mydemos/demo/Peterson_v2.csp', '-ltlformula', 'G(not [critical.2])', '-strict', '-expcterr' ,ltl ],'Check ltl formula is reported as false').
cli_testcase(605, [b_test,bparser], ['examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-eval', '(A <=> (waiting={})) & A', '-p', 'BOOL_AS_PREDICATE', 'FALSE',  '-expcterr' ,parse_machine_predicate_error], 'Test that BOOL as predicate can be disabled.').
cli_testcase(606, [b_test,bparser], ['examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-evalt', '(A <=> (waiting={})) & A', '-evalf', '(B <=> (waiting={})) & (B <=> not(waiting={}))', '-evalt', '!x.(x:1..10 => (A <=> x>z)) & z:0..11 & not(A)', '-p', 'BOOL_AS_PREDICATE', 'TRUE'], 'Test that BOOL as predicate works.').
cli_testcase(607, [b_test,bparser],['-strict', 
  '-eval_file', 'examples/Eval/SAT_uf20-02_pred.eval', '-p', 'CLPFD', 'TRUE', '-p', 'BOOL_AS_PREDICATE', 'TRUE'],'Test that BOOL as predicate works.').
cli_testcase(608,[tickets], ['examples/B/Tickets/ConcatDisjUnion/SimpleGrammar.mch', '-init', '-mc', 100 , '-nodead', '-cc', 4, 3, '-strict'], 'Testing that concat does not generate wrong solutions.').
cli_testcase(609,[tickets,b_test], ['examples/B/Compilers/SyracuseGrammar.mch', '-init', '-mc', 100 , '-expcterr', deadlock, '-strict', '-p', 'CLPFD', 'FALSE'], 'Testing that concat works efficiently enough.').
cli_testcase(610,[tickets,b_test], ['examples/B/Compilers/SyracuseGrammar.mch', '-init', '-mc', 100 , '-expcterr', deadlock, '-strict', '-p', 'CLPFD', 'TRUE'], 'Testing that concat works efficiently enough.').
cli_testcase(611,[b_test], ['examples/B/PerformanceTests/InfiniteRelationImage.mch', '-init', '-assertions', '-properties', '-strict', '-p', 'CLPFD', 'TRUE'], 'Testing that image for infinite relation works.').
cli_testcase(612, [b_test,recursion,external], ['examples/B/Compilers/NFA.mch', '-t', '-mc', 100, '-cc', 6, 12, '-assertions','-strict'], 'Test recursive function with recursive call inside UNION.').
cli_testcase(613, [b_test],['examples/B/Pragmas/AssertionPragmas.mch', '-init', '-pp', 'examples/B/Pragmas/AssertionPragmas_internal.mch', '-strict'], 'Test that labels recognised.').
cli_testcase(614, [b_test],['examples/B/Compilers/NFA_det.ref', '-init', '-refchk', 'examples/B/Compilers/NFA_nondet_refine_spec.P', '-cc', 10, 22, '-strict'], 'Test trace refinement checking and ABSTRACT_CONSTANT reuse in NFA_det.').
cli_testcase(615,[b_test,ltl], ['examples/B/LTL/Tests/LTL_1State.mch', 'examples/B/LTL/Tests/LTL_2States.mch', '-init', '-ltlassertions', '-strict'], 'Testing LTL formulas on simple statespaces.').
cli_testcase(616, [b_test],['examples/B/Tester/Partition.mch', '-init',  '-assertions','-pp','examples/B/Tester/Partition_internal.mch', '-strict'], 'Test that partition properly solved and printed.').
cli_testcase(617, [b_test,tickets],['examples/B/Tickets/RefLoadLoop/NFA_det.ref', '-expcterr', 'load_main_file', '-expcterr', invalid_machine_reference, '-expcterr', b_load_machine, '-strict'], 'Test that parser does not loop.').
cli_testcase(618,[b_test,ltl], ['examples/B/LTL/Tests/LTL_4States.mch', '-init', '-ltlassertions', '-strict'], 'Testing LTL formulas on simple statespaces.').
cli_testcase(619,[b_test,ltl], ['examples/B/LTL/Tests/CTL_SetMove_2States.mch', '-init', '-ctlformula', 'AG(EX{xx=bb})', '-ctlformula', 'AGEFEX[Move]true', '-ctlformula', 'AGEFEX[Set(aa)]{xx=aa}', '-ctlformula', 'not(EF EX[Set(aa)]{xx=bb})', '-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(620,[b_test,ltl], ['examples/B/LTL/Tests/CTL_SetMove_2States.mch', '-init', '-ctlformula', '(EF EX[Set(aa)]{xx=bb})',  '-expcterr', ctl,'-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(621,[b_test,ltl], ['examples/B/Benchmarks/scheduler.mch', '-init', '-ctlformula', 'AG ( EX[swap]true => {active /= {}})','-ctlformula', 'AG ( (EX[new]true)  or (EX[del]true) or (EX[swap]true) )', '-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(622,[b_test,ltl], ['examples/B/Benchmarks/scheduler.mch', '-init','-ctlformula', 'AG ( (EX[new]true)  or (EX[del]true) )', '-expcterr', ctl,'-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(623,[tickets], ['examples/B/PerformanceTests/Closure1_Computations.mch', '-init', '-assertions','-strict'], 'Check that closure1 works well for large or infinite sets.').
cli_testcase(624,[csp_test], ['examples/CSP/Tickets/3675/AIDCPRB_small.csp', '-mc', '10000', '-t', '-strict'], 'Testing complex datatypes.').
cli_testcase(625, [b_test,tla],['examples/TLA/Tests/ChooseTest.tla', '-mc', 100, '-strict'], 'Testing TLA2B translation and loading.').
cli_testcase(626, [b_test],['examples/B/PerformanceTests/TestFD_Domain_Narrowing.mch', '-t', '-p', 'CLPFD', 'TRUE', '-strict'], 'Test CLPFD on enumerated sets.').
cli_testcase(627, [cbc], ['examples/B/Benchmarks/scheduler.mch', '-cbc_sequence', 'new;new;del;ready', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE'], 'Check cbc_sequence').
cli_testcase(628, [cbc], ['examples/B/Benchmarks/scheduler.mch', '-cbc_sequence', 'new;new;del;del;del', '-strict', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE', '-expcterr', cbc_sequence_no_solution_found], 'Check cbc_sequence').
cli_testcase(629, [tickets], ['examples/B/Tickets/Burdy1/TestOrApplyFunErr.mch', '-mc', '1000', '-expcterr', well_definedness_error, '-strict'], 'Check issue with undefined function application in context of disjunct.').
cli_testcase(630, [tickets], ['examples/B/Tickets/PROBCORE-45/NotIntervalSol.mch', '-init', '-p', 'MAX_INITIALISATIONS', 20, '-mc', '1000', '-nodead', '-assertions', '-cbc_assertions', '-strict'], 'Check problem with not subset for intervals.').
cli_testcase(631, [tickets], ['examples/B/Tickets/PROBCORE-45/NotIntervalSol.mch', '-init', '-p', 'MAX_INITIALISATIONS', 20, '-p', 'CLPFD', 'TRUE', '-mc', '1000', '-nodead', '-assertions', '-cbc_assertions', '-strict'], 'Check problem with not subset for intervals.').
cli_testcase(632, [cbc], ['-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-p', 'DEFAULT_SETSIZE', 2, 'examples/B/CBC/data_structure1_ctx_theorem_error.eventb', '-cbc_assertions', '-expcterr',cbc_assertions, '-p', 'PROOF_INFO', 'TRUE'], 'Check CBC static assertion checking.').
cli_testcase(633, [tickets], ['examples/EventBPrologPackages/ProofDirected/TestR_label_reuse.eventb', '-mc', 1000, '-noass', '-strict', '-expcterr', 'invariant_violation', '-p', 'PROOF_INFO', 'TRUE'], 'Test that PROOF_INFO not confused by two invariants with same name (inv3).').
cli_testcase(634, [b_test,kodkod], ['examples/B/Kodkod/Loop.mch', '-strict', '-mc', 20, '-nodead', '-bf', '-p', 'MAX_INITIALISATIONS', 8, '-p', 'CLPFD', 'TRUE', '-cc', 15, 14, '-spdot', 'examples/B/Kodkod/Loop_statespace.dot'], 'Test performance on example where ProB used to be very slow').
cli_testcase(635, [b_test,kodkod], ['examples/B/Kodkod/Loop.mch', '-strict', '-mc', 20, '-nodead', '-bf', '-p', 'MAX_INITIALISATIONS', 8, '-p', 'KODKOD', 'TRUE', '-cc', 15, 14], 'Test performance on Loop example with Kodkod').
cli_testcase(636, [b_test], ['examples/EventBPrologPackages/ConfigChooser/PetChooser_mch.eventb', '-strict', '-mc', 300, '-p', 'CLPFD', 'TRUE', '-cc', 41, 236], 'Test performance on example where ProB used to be very slow').
cli_testcase(637, [b_test,kodkod], ['examples/B/Kodkod/LoopAsStrings.mch', '-strict', '-mc', 20, '-nodead', '-bf', '-p', 'MAX_INITIALISATIONS', 8, '-p', 'CLPFD', 'TRUE', '-cc', 15, 14], 'Test performance on example where ProB used to be very slow').
cli_testcase(638, [b_test,recursion,external], ['examples/B/Tickets/TransitiveClosure/ComputeTClos_mch.eventb', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE', '-assertions'], 'Test that recursive function works in Event-B.').
cli_testcase(639, [laws], ['examples/B/Laws/SetLawsPow.mch', '-mc', '500', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'FALSE'], 'Laws about sets of sets').
cli_testcase(640, [laws], ['examples/B/Laws/SetLawsPow.mch', '-mc', '500', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Laws about sets of sets').
cli_testcase(641, [laws], ['examples/B/Laws/SetLawsPow.mch', '-mc', '500', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SYMBOLIC', 'TRUE'], 'Laws about sets of sets').
cli_testcase(642, [laws], ['examples/B/Laws/SetLawsPowPow.mch', '-mc', '250', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Laws about sets of sets').
cli_testcase(643, [laws], ['examples/EventBPrologPackages/Deploy/s1_mch1_switches_mch.eventb', '-mc', '250', '-p', 'DEFAULT_SETSIZE', 2, '-strict', '-p', 'CLPFD', 'TRUE', '-cc', 10, 30], 'Check that deferred set is set to at least 3').
cli_testcase(644, [laws], ['examples/EventBPrologPackages/Laws/TestPartitionDeferredSetCardMch_mch.eventb', '-mc', '250', '-nodead', '-p', 'DEFAULT_SETSIZE', 2, '-strict', '-p', 'CLPFD', 'TRUE', '-cc', 5, 6], 'Check that deferred set is set to at least 6').
cli_testcase(645, [plugin_units], ['examples/B/PragmasUnits/Tests/WrongComparisonInInvariant.mch', '-plugin', 'units', '-expecterr', 'well_definedness_error', '-strict', '-mc', 5], '').
cli_testcase(646, [b_test,cache], ['examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-cache', 'cache_does_not_exist/', '-expecterr', cache], 'Check that -cache option works on simple example.').
cli_testcase(647, [b_test,cache], ['examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-cache', 'examples/cache/'], 'Check that -cache option works on simple example.').
cli_testcase(648, [b_test,cache], ['examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-cache', 'examples/cache/'], 'Check that -cache option works on simple example (2nd time).').
cli_testcase(649, [tickets,'PROB-83'], ['examples/B/Tickets/PROB-83/Time.mch', '-t', '-strict'], 'Check infinite function works.').
cli_testcase(650, [tickets,cbc], ['examples/B/Tickets/Berkani1/CartesianProductOverrideAssertion.mch', '-cbc_assertions', '-strict', '-expcterr',cbc_assertions], 'Check that counter example for Cartesian product found.').
cli_testcase(651, [tickets], ['examples/B/Tickets/Berkani1/CartesianProductOverride.mch', '-init', '-strict', '-p', 'CLPFD', 'TRUE', '-properties', '-assertions'], 'Check that counter example for Cartesian product found.').
cli_testcase(652, [tickets], ['examples/B/Compilers/TuringMachine_Goedelisation.mch', '-mc', 300, '-p', 'CLPFD', 'TRUE', '-expecterr', deadlock, '-strict', '-cc', 245, 244], 'Check machine with several complicated infinite & recursive functions.').
cli_testcase(653, [tickets], ['examples/B/Tickets/Mejia5/essai_fncs.mch', '-t', '-strict'], 'Check ticket (with NAT >+> NAT)').
cli_testcase(654, [laws], ['examples/B/Laws/SetRelLaws_NatBool.mch', '-mc', '750', '-nodead', '-c', '-strict','-p', 'NEGATED_INVARIANT_CHECKING', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Various laws about relations.').
cli_testcase(655, [typechecker_test,tickets], ['-mc','1000','examples/B/Tickets/PROB-46/Validate.mch','-expecterr',type_error ,'-expecterr', load_main_file, '-strict'], 'Check type-error: Operation calls another operation of the same machine.').
cli_testcase(656, [b_test], ['examples/B/Mathematical/PostCorrespondence_Turing_MC.mch', '-mc', '750', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE', '-expecterr', goal_found], 'A non-trivial machine.').
cli_testcase(657, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage_internal.mch'], '').
cli_testcase(658, [cbc,cbc_deadlock], ['examples/EventBPrologPackages/Deploy/s4_mch9_schedule_mch.eventb', '-cbc_deadlock', '-p', 'SMT', 'TRUE', '-strict', '-p', 'CLPFD', 'TRUE', '-p','DEFAULT_SETSIZE',1], 'Check no typing error.').
cli_testcase(659, [cbc,cbc_deadlock], ['examples/EventBPrologPackages/Deploy/s4_mch9_schedule_mch.eventb', '-cbc', 'all', '-p', 'SMT', 'TRUE', '-strict', '-p', 'CLPFD', 'TRUE', '-p','DEFAULT_SETSIZE',1], '').
cli_testcase(660, [b_test,external], ['examples/B/ExternalFunctions/TestLibraryIO.mch', '-t', '-strict', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(661, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion_internal.mch'], '').
cli_testcase(662, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent_internal.mch'], '').
cli_testcase(663, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/SimpleFunctionApplication.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/SimpleFunctionApplication_internal.mch'], '').
cli_testcase(664, [ltl], ['examples/B/TestCaseGeneration/SimpleCounter.mch', '-ltllimit', '50', '-ltlformula', 'G{1=1}', '-strict', '-expcterr', ltl_incomplete], 'check that no error reported for incomplete model check').
cli_testcase(665, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car_internal.mch'], '').
cli_testcase(666, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight_internal.mch'], '').
cli_testcase(667, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System_internal.mch'], '').
cli_testcase(668, [ltl,csp_test], ['examples/CSP/other/Ivo/BobAlice.csp', '-ltlassertions', '-strict'],'Check LTL formulas in File.').
cli_testcase(669, [plugin_units], ['examples/B/PragmasUnits/Tests/IncorrectUnitDefinition.mch', '-plugin', 'units', '-strict', '-expecterr', 'incorrect_unit_definition'], '').
cli_testcase(670, [plugin_units], ['examples/B/PragmasUnits/Tests/BoolAndString.mch', '-plugin', 'units', '-strict','-mc', 5], 'Bool and String are not annotated / used by the unit plugin. Just check that there is no error.').
cli_testcase(671, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/TypedGlobalSet.mch', '-plugin', 'units', '-strict', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/TypedGlobalSet_internal.mch'], '').
cli_testcase(672, [plugin_units], ['examples/B/PragmasUnits/Tests/WrongOperation.mch', '-plugin', 'units', '-strict', '-expecterr', 'Multiple Units inferred', '-mc', 5], '').
cli_testcase(673, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions_internal.mch'], '').
cli_testcase(674, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/Functions.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/Functions_internal.mch'], '').
cli_testcase(675, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash.mch','-plugin','units','-expecterr','Multiple Units inferred','-strict','-pp','examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash_internal.mch'], '').
cli_testcase(676, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias_internal.mch'], '').
cli_testcase(677, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_internal.mch'], '').
cli_testcase(678, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i.imp', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i_internal.mch'], '').
cli_testcase(679, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0_internal.mch'], '').
cli_testcase(680, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i.imp', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i_internal.mch'], '').
cli_testcase(681, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations.mch', '-plugin', 'units', '-strict', '-expecterr', 'well_definedness_error', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations_internal.mch','-p', 'plugin(units,fixpoint_halt_on_error)','false'], '').
cli_testcase(682, [ltl,csp_test], ['examples/CSP/slicing/peterson_modelcheck.csp', '-ltlassertions', '-strict'],'Check LTL formulas in File.').
cli_testcase(683, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion_internal.mch'], '').
cli_testcase(684, [modelcheck,csp_test], ['examples/CSP/Tickets/3675/AIDCPRB.csp', '-mc', '10000', '-nodead', '-strict'],'Model checking test for fixed bug number 3675.').
cli_testcase(685, [modelcheck,csp_test], ['examples/CSP/Tickets/3675/SafetyPRB.csp', '-mc', '150000', '-nodead', '-strict'],'Model checking test for fixed bug number 3675.').
cli_testcase(686, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations_internal.mch'], '').
cli_testcase(687, [b_test,csp_test], ['examples/CSPB/FeatureTests/StringMachine.mch', '-csp-guide', 'examples/CSPB/FeatureTests/StringMachine.csp', '-mc', '1000', '-strict', '-expecterr', 'deadlock'], 'Test that the machine with guidance by CSP have a deadlock').
cli_testcase(688, [plugin_units], ['examples/B/PragmasUnits/Tests/SomeUnitErrors.mch', '-plugin', 'units', '-strict', '-expecterr', 'well_definedness_error','-expecterr','assertion_violation', '-mc', 5], '').
cli_testcase(689, [cbc], ['examples/EventBPrologPackages/Advance/Safety_Metro_23_08_2012/GCDoor_M2_mch.eventb', '-cbc', all, '-strict'], 'Test that proof info works well and no invariant checked').
cli_testcase(690, [cbc], ['examples/EventBPrologPackages/ProofDirected/M1_err_mch.eventb', '-cbc', all, '-strict', '-expecterr', cbc, '-expecterr', 'event_refinement', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Test that error detected').
cli_testcase(691, [eventb_test], ['examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_mch.eventb', '-mc', 100, '-strict', '-expecterr', 'invariant_violation', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', 1], 'Test that error detected').
cli_testcase(692, [eventb_test], ['examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_mch.eventb', '-mc', 100, '-nodead', '-strict', '-expecterr', 'event_error:setX:action_not_executable', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Test that error detected').
cli_testcase(693, [eventb_test], ['examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_rename.eventb', '-mc', 100, '-strict', '-expecterr', 'invariant_violation', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', 1], 'Test that error detected').
cli_testcase(694, [eventb_test], ['examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_rename.eventb', '-mc', 100, '-nodead', '-strict', '-expecterr', 'event_error:setX3:action_not_executable', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Test that error detected').
cli_testcase(695, [plugin_units], ['examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits_internal.mch'], '').
cli_testcase(696, [plugin_units], ['examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat.mch', '-plugin', 'units', '-strict','-expecterr', 'Multiple Units inferred', '-pp', 'examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat_internal.mch'], '').
cli_testcase(697, [tickets,external], ['examples/B/Tickets/Burlando1/machine_test_PRINTF.mch', '-t', '-strict'], '').
cli_testcase(698, [b_test], ['-p', 'CLPFD', 'TRUE', 'examples/B/Puzzles/SudokuAsConstant.mch', '-t','-strict', '-p', 'TRACE_INFO', 'TRUE'], 'Check TRACE_INFO flag').
cli_testcase(699, [cbc,cbc_deadlock], ['examples/B/CBC/ExistsDeadlockTests.mch', '-cbc_deadlock', '-p', 'TRACE_INFO', 'TRUE','-expcterr', 'cbc_deadlock_check', '-strict'], 'Check that guard simplification works correctly').
cli_testcase(700, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0_internal.mch'], '').
cli_testcase(701, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1_internal.mch'], '').
cli_testcase(702, [b_test,tla],['examples/TLA/Tests/SetLaws.tla', '-mc', 1000, '-nodead', '-strict'], 'Testing TLA2B translation laws.').
cli_testcase(703, [b_test,tla],['examples/TLA/Tests/NatLaws.tla', '-mc', 1000, '-nodead', '-strict'], 'Testing TLA2B translation laws.').
cli_testcase(704, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0_internal.mch'], '').
cli_testcase(705, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1.mch','-expcterr',bmachine_static_checks, '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1_internal.mch'], '').
cli_testcase(706, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2.mch','-expcterr',bmachine_static_checks, '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2_internal.mch'], '').
cli_testcase(707, [plugin_units], ['examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3.mch','-expcterr',bmachine_static_checks, '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3_internal.mch'], '').
cli_testcase(708, [laws],['examples/B/Kodkod/LawsArithmeticCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'CLPFD', 'TRUE'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(709, [laws],['examples/B/Kodkod/LawsArithmeticCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'KODKOD', 'TRUE'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(710, [laws],['examples/B/Kodkod/LawsSetsCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'CLPFD', 'TRUE', '-strict'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(711, [laws],['examples/B/Kodkod/LawsRelCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'KODKOD', 'TRUE'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(712, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/AppendBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the built-in operator ^ raises expected error.').
cli_testcase(713, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/ConcatBuiltInFunctionTypeError1.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the concat(<...>) built-in function raises expected error.').
cli_testcase(714, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/ConcatBuiltInFunctionTypeError2.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the concat(<...>) built-in function raises expected error.').
cli_testcase(715, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/ExpandingInfiniteSequenceError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing of raising error by expanding infinite sequences.').
cli_testcase(716, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/HeadBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the head(-) built-in function raises expected error.').
cli_testcase(717, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/IsElemListTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'csp_error'], 'Testing if the elem(-,-) built-in function raises expected error.').
cli_testcase(718, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/LengthBuiltInFunctionInfiniteSeqError.csp', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock', '-strict'], 'Testing if the length(-) built-in function raises expected error by calling it with infinite sequence.').
cli_testcase(719, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/LengthBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the length(-) built-in function raises expected error.').
cli_testcase(720, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/NullBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the null(-) built-in function raises expected error.').
cli_testcase(721, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/SeqBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the seq(-) built-in function raises expected error.').
cli_testcase(722, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/SetBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the set(-) built-in function raises expected error.').
cli_testcase(723, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sequences/TailBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the tail(-) built-in function raises expected error.').
cli_testcase(724, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/ReplicatedInternalChoiceEmptySetError.csp', '-strict', '-mc', 10,'-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing of error raising by defining a replicated internal choice process over an empty set.').
cli_testcase(725, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/EmptySetTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the empty(-) built-in function raises expected error.').
cli_testcase(726, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/ReplicatedOperatorsOverEmptyPatternSets.csp', '-strict', '-mc', 10], 'Testing replicated operators over over empty pattern sets (||| x:{}@ P, ; x:<>@P, || x:{} @[a] P and [|a|] x:{} @P).').
cli_testcase(727, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/EmptySetInReplicatedLinkedParallelError.csp', '-strict', '-mc', 10,'-expcterr', 'haskell_csp', '-expcterr', 'csp_sequences', '-expcterr', 'general_error_occurred'], 'Testing of error raising by defining a replicated linked parallel process over an empty set.').
cli_testcase(728, [error_checks, csp_test], ['examples/CSP/ErrorModels/haskell_csp/InterleavingSpecialCase.csp', '-strict', '-mc', 10, '-expcterr', 'deadlock'], 'Testing a special case of the |||-operator (STOP ||| STOP).').
cli_testcase(729, [error_checks, csp_test], ['examples/CSP/ErrorModels/haskell_csp/UndefinedConstructError.csp', '-strict', '-mc', 10, '-expcterr', 'csp_interpreter', '-expcterr', 'general_error_occurred'], 'Testing for raising undefined constructor error.').
cli_testcase(730, [error_checks, csp_test], ['examples/CSP/ErrorModels/haskell_csp/TooManyOutputValuesOnChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_action', '-expcterr', 'general_error_occurred'], 'Testing for raising an error, when there are too many output values on channel.').
cli_testcase(731, [error_checks, csp_test], ['examples/CSP/ErrorModels/haskell_csp/NotABooleanValueErrorIfThenElse.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising an error, when there is no boolean value expression in the if-then-else-condition.').
cli_testcase(732, [laws],['examples/B/Kodkod/LawsSetsCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'KODKOD', 'TRUE'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(733, [b_test,csp_test], ['examples/CSPB/TestVariableProbing.mch', '-csp-guide', 'examples/CSPB/TestVariableProbing.csp', '-t', '-strict'], 'Test probing of B variables').
cli_testcase(734, [laws], ['examples/B/Laws/BoolLaws_SetCompr.mch', '-init', '-c', '-p', 'KODKOD','TRUE', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(735, [laws], ['examples/B/Laws/BoolLaws_SetComprCLPFD.mch', '-init', '-c', '-p', 'KODKOD','TRUE', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(736, [b_test,csp_test], ['examples/CSPB/EventB/m_partner_behaviour.eventb', '-csp-guide', 'examples/CSPB/EventB/m_partner_behaviour.csp', '-t', '-strict'], 'Test Event-B+CSP + discharged calc').
cli_testcase(737, [b_test,external], ['examples/B/ExternalFunctions/LibraryFiles.mch', '-init', '-assertions','-strict'], 'Check that external functions work.').
cli_testcase(738, [b_test,proz],['examples/Z/Testcases/freetypes.fuzz', '-mc', '1000', '-nodead', '-strict'],'ProZ test for freetypes').
cli_testcase(739, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/ChannelOutputValuesError.csp', '-strict', '-mc', 10, '-expcterr', 'evaluate_channel_outputs', '-expcterr', 'deadlock'], 'Testing raising an error when no channel tuple is given as prefix event.').
cli_testcase(740, [error_checks, csp_test], ['examples/CSP/ErrorModels/haskell_csp/TypeErrorInChannelOutput.csp', '-strict', '-mc', 10, '-expcterr', 'check_channel_output_value', '-expcterr', 'general_error_occurred'], 'Testing raising of channel output error.').
cli_testcase(741, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/ChannelNotDeclaredError.csp', '-strict', '-mc', 10, '-expcterr', 'evaluate_channel_outputs', '-expcterr', 'unify_spans1','-expcterr', 'general_error_occurred', '-expcterr', 'enumerate_action'], 'Testing raising of channel not declared errors.').
cli_testcase(742, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/ListProjectionError1.csp', '-strict', '-mc', 10, '-expcterr', 'list_projection','-expcterr', 'general_error_occurred'], 'Testing for list projection errors inside local definitions (let-within clauses).').
cli_testcase(743, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/ListProjectionError2.csp', '-strict', '-mc', 10, '-expcterr', 'list_projection','-expcterr', 'deadlock'], 'Testing for list projection errors inside local definitions (let-within clauses).').
cli_testcase(744, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/ListProjectionError3.csp', '-strict', '-mc', 10, '-expcterr', 'list_projection','-expcterr', 'deadlock'], 'Testing for list projection errors inside local definitions (let-within clauses).').
cli_testcase(745, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/TupleProjectionError1.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(746, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/TupleProjectionError2.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(747, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/NaTupleProjectionError1.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for not-associative tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(748, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_tupels/NaTupleProjectionError2.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for not-associative tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(749, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/CardinalityOfInfiniteSetError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the card(-) built-in function raises expected error by calling it with infinite set.').
cli_testcase(750, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/CardinalityTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the card(-) built-in function raises expected type error by calling it with expression, which is not a set.').
cli_testcase(751, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/InterEmptySetError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the Inter(-) built-in function raises expected error by calling it with an empty set.').
cli_testcase(752, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/ExpandClosureSetError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing for errors by expanding closure set elements (error occurs if the element is not a channel name e.g. just constructor).').
cli_testcase(753, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/ExpandingSymbolicSetError1.csp', '-strict', '-mc', 10,'-expcterr', 'expand_symbolic_set', '-expcterr', 'deadlock'], 'Testing for errors by expanding symbolic sets.').
cli_testcase(754, [csp_test], ['examples/CSP/mydemos/demo/Peterson_v2.csp', '-mc', 1000, '-strict', '-p', 'COMPRESSION', 'TRUE', '-cc', 233, 465, '-statistics'], 'Test compression with CSP').
cli_testcase(755, [csp_test], ['examples/CSPB/TestVariableProbing.csp', '-MAIN', 'OTHER', '-mc', 100, '-cc', 4, 4, '-strict'], 'Test setting main process').
cli_testcase(756, [promela], ['examples/Promela/atomic.pml', '-t', '-mc', 100, '-strict'], 'Test Promela mode').
cli_testcase(757, [error_checks,csp_repl], ['examples/CSP/ErrorModels/haskell_csp/TestingForParseAndLoadErrors.csp', '-strict', '-expcterr', 'haskell_csp'], 'Testing the error clauses of the pasrse and load predicates in the haskell_csp module.').
cli_testcase(758, [csp_repl], ['examples/CSP/other/Ivo/Farmer.csp', '-strict', '-get_csp_assertions_as_string'], 'Testing getting Csp assertions from file.').
cli_testcase(759, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/EnumerateInfiniteChannelInputError1.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_channel_input_value', '-expcterr', 'general_error_occurred'], 'Testing enumerating infinite channel input error.').
cli_testcase(760, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/EnumerateIllegalTypeOfChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_channel_input_value', '-expcterr', 'general_error_occurred'], 'Testing enumerating infinite channel input error.').
cli_testcase(761, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/EnumerateInfiniteChannelInputError2.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_channel_input_value', '-expcterr', 'general_error_occurred'], 'Testing for raising enumerating infinite channel type error for sequences.').
cli_testcase(762, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/ChannelInputSetEnumerationResultError.csp', '-strict', '-mc', 10, '-expcterr', 'set_enumeration_tuple_result', '-expcterr', 'deadlock'], 'Testing for raising type error on channel input.').
cli_testcase(763, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/ChannelTooFewValuesOnChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_action', '-expcterr', 'general_error_occurred'], 'Testing for raising type error in channel input.').
cli_testcase(764, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/InterleavedIfThenElseBlocks.csp', '-strict', '-mc', 1000], 'Testing interleaved if-then-else blocks.').
cli_testcase(765, [csp_repl], ['-strict', '-cspm_eval', '1+1 == 2'], 'Testing that evaluating an expression with the CSPM-tool works.').
cli_testcase(766, [csp_repl], ['examples/CSP/mydemos/OtherCoverageTests/TestingCspmEvalFlag.csp', '-strict', '-cspm_eval', 'reverse(<1,2,3>) == <3,2,1>'], 'Testing that evaluating an expression with the CSPM-tool depending on the implementation in the respective file works.').
cli_testcase(767, [private,data_validation], ['../probprivate/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld.mch','-strict', '-t', '-p',
 'CLPFD', 'TRUE', '-aa', 27, 0, 0, '-p', 'MAX_OPERATIONS', 1, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE', '-sptxt', '../probprivate/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld_sptxt.txt'], 'Siemens Data Validation Test.').
cli_testcase(777, [private,data_validation], ['../probprivate/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_ctx_aiguille.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 49, 2, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0, '-strict'], 'Siemens Data Validation Test.').
cli_testcase(778, [private,data_validation], ['../probprivate/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_cfg_ipart.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 144, 4, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(779, [private,data_validation], ['../probprivate/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_cfg_ipart.mch','-strict', '-p','CLPFD', 'TRUE', '-ma', 95, 2, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(780, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_easitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 127, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(781, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_easitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-ma', 22, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(782, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_etors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 142, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(783, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_stors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 147, 1, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(784, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_mes_as_env_inv_etors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 125, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(785, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_etors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 188, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(786, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_etors_dist_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 184, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(787, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_mesitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 182, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(788, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_pasitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 176, 1, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(789, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_se_dj_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 174, 2, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(790, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_ss_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 192, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(791, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_stors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 192, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(792, [private,data_validation], ['../probprivate/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_tcs_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 183, 8, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(793, [private,data_validation], ['../probprivate/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_etorf.mch','-strict', '-p','CLPFD', 'TRUE', '-init', '-expcterr', 'setup_constants_fails', '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(794, [private,data_validation], ['../probprivate/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_etorf.mch','-strict', '-p','CLPFD', 'TRUE', '-init', '-expcterr', 'setup_constants_fails', '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(795, [private,data_validation], ['../probprivate/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/cmnb_heure_cst_gen.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 7, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(796, [private,data_validation], ['../probprivate/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_ctx_trainstop.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 6, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(797, [private,data_validation], ['../probprivate/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_quai.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 51, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(798, [private,data_validation], ['../probprivate/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_zaum.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 40, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(799, [private,data_validation], ['../probprivate/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_cv_sui.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 62, 1, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 25000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(800, [private], ['../probprivate/examples/RodinModels/Bosch/CrCtl_0.5.0_v9/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10 & P_CrCtl_Mode=STANDBY', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(801, [private], ['../probprivate/examples/RodinModels/Bosch/CrCtl_0.5.0_v8/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(802, [private], ['../probprivate/examples/RodinModels/Bosch/CrCtl_0.5.0_v7/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(803, [private], ['../probprivate/examples/RodinModels/Bosch/CrCtl_0.5.0_v6/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(804, [private], ['../probprivate/examples/RodinModels/Bosch/CrCtl_0.5.0_v5/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(805, [private], ['../probprivate/examples/RodinModels/Bosch/CrCtl_0.5.0_v5/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(806, [private], ['../probprivate/examples/RodinModels/Bosch/CrCtl_0.5.0_v4/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(807, [tickets,csp_test,'PROB-43'], ['examples/CSP/Tickets/Shi/milnerSched5.csp', '-assertions', '-strict'], 'Check that exception does not occur (Ticket)').
cli_testcase(808, [b_test,dot], ['examples/B/Benchmarks/CarlaTravelAgencyErr.mch', '-t', '-strict', '-sgdot', 'examples/B/Benchmarks/CarlaTravelAgencyErr_state_graph.dot'], 'Check State Graph Generation').
cli_testcase(809, [private,data_validation], ['examples/B/Tickets/LargeMachTypeErr/LargeValueOk.mch', '-assertions', '-strict', '-p', 'LARGE_PARSER_JVM', 'TRUE'], 'Check loading large data').
cli_testcase(810, [private], ['examples/B/Tickets/LargeMachTypeErr/sequenceur_values.mch', '-expcterr', type_error, '-expcterr', load_main_file, '-strict', '-p', 'LARGE_PARSER_JVM', 'TRUE'], 'Check large number of errors').
cli_testcase(811, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_tupels/ChannelNameUsedAsValueError.csp', '-strict', '-mc', 10, '-expcterr', 'valid_value', '-expcterr', 'general_error_occurred','-expcterr', 'check_channel_output_value'], 'Testing for raising channel name used as value error.').
cli_testcase(812, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_tupels/IllegalDataValueError.csp', '-strict', '-mc', 10, '-expcterr', 'valid_value', '-expcterr', 'general_error_occurred','-expcterr', 'check_channel_output_value'], 'Testing for raising illegal data value error.').
cli_testcase(813, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_tupels/MismatchNumberOfArgumentsForChannelSynchrError.csp', '-strict', '-mc', 10, '-expcterr', 'unify_tuple_args', '-expcterr', 'deadlock'], 'Testing for raising mismatch number of arguments error.').
cli_testcase(814, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality1.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'csp_error'], 'Testing for type errors in equality.').
cli_testcase(815, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality2.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'csp_error'], 'Testing for type errors in equality.').
cli_testcase(816, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality3.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'csp_error'], 'Testing for type errors in equality.').
cli_testcase(817, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality4.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'csp_error'], 'Testing for type errors in equality.').
cli_testcase(818, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality5.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(819, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality6.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(820, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality7.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(821, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality8.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(822, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality9.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(823, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality10.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(824, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality11.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(825, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality12.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(826, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality1.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(827, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality2.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'csp_error'], 'Testing for type errors in equality.').
cli_testcase(828, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality3.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'csp_error'], 'Testing for type errors in equality.').
cli_testcase(829, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality4.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'csp_error'], 'Testing for type errors in equality.').
cli_testcase(830, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality5.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(831, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality6.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(832, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality7.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(833, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality8.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(834, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality9.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(835, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality10.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(836, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality11.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(837, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality12.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(838, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/TestingComprehensionSetGuards.csp', '-strict', '-mc', 10000], 'Testing set comprehension guards.').
cli_testcase(839, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-strict', '-mc', 10], 'Testing expanding symbolic sets.').
cli_testcase(840, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_sets/ChannelNameUsedForLocalVariableError.csp', '-strict', '-expcterr', 'csp_sets'], 'Testing for raising channel name used as local variable error.').
cli_testcase(841, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_sets/PatternListTooLongForChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing raising pattern list too long error.').
cli_testcase(842, [error_checks, csp_test], ['examples/CSP/ErrorModels/haskell_csp/TooManyValuesInRenamePatternError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing raising too many values in rename pattern error.').
cli_testcase(843, [error_checks, csp_test], ['examples/CSP/ErrorModels/haskell_csp/TypeErrorForChannelInputVariable.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_datatype_el', '-expcterr', 'deadlock'], 'Testing raising type error for channel input variable exception.').
cli_testcase(844, [csp_repl], ['examples/CSP/other/Ivo/Farmer.csp', '-strict', '-get_csp_enumerated_assertions_as_string'], 'Testing getting enumerated assertions from file.').
cli_testcase(845, [csp_repl], ['examples/CSP/other/Ivo/Farmer.csp', '-strict', '-prologOut', 'examples/CSP/other/Ivo/Farmer_temp.pl'], 'Testing translating CSP-M file to specific prolog file.').
cli_testcase(846, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/IsMemberSetCouldNotEvaluateError.csp', '-strict', '-mc', 10, '-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing for raising could not evaluate error for the member(-,-) built-in function.').
cli_testcase(847, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/NotASingletonSetError.csp', '-strict', '-mc', 10, '-expcterr', 'singleSetElement', '-expcterr', 'deadlock'], 'Testing for raising not a singleton set error.').
cli_testcase(848, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality13.csp', '-strict', '-mc', 10000], 'Testing for type errors in equality.').
cli_testcase(849, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality13.csp', '-strict', '-mc', 10000], 'Testing for type errors in disequality.').
cli_testcase(850, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/NotABooleanExpressionError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising not a boolean expression error.').
cli_testcase(851, [csp_test], ['examples/CSP/mydemos/subtype_ex.csp', '-strict', '-mc', 1000,'-expcterr', 'deadlock' ,'-df'], 'Testing complex subtype expressions.').
cli_testcase(852, [csp_test], ['examples/CSP/mydemos/subtype_nametype_ex.csp', '-strict', '-mc', 1000,'-expcterr', 'deadlock'], 'Testing complex subtype expressions.').
cli_testcase(853, [private], ['../probprivate/examples/CSP/Winter/miniAlvey.csp', '-strict', '-t', '-c', '-v'], 'Testing complex CSP spec.').
cli_testcase(854, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/NoMatchingCaseForFunctionCallError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing for no matching case for function call error.').
cli_testcase(855, [b_test,cache], ['-p', 'MAXINT', '5', 'examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '81', '177', '-p', 'DEFAULT_SETSIZE', '2', '-nogoal', '-strict', '-cache', 'examples/cache/'], 'Check that no error in reuse of cached operations').
cli_testcase(856, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality14.csp', '-strict', '-mc', 10000], 'Testing for type errors in equality.').
cli_testcase(857, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality14.csp', '-strict', '-mc', 10000], 'Testing for type errors in disequality.').
cli_testcase(858, [coverage,csp_test], ['examples/CSP/other/Ivo/Farmer.csp', '-strict', '-mc', 1000, '-expcterr', 'deadlock', '-save', 'examples/CSP/other/Ivo/Farmer_state.txt'], 'Testing saving XSB Specification state.').
cli_testcase(859, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/NotEnoughArgumentsOnChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'l_enumerate_channel_input_value', '-expcterr', 'deadlock'], 'Testing for raising not enough arguments on channel error.').
cli_testcase(860, [coverage, csp_test], ['examples/CSP/mydemos/OtherCoverageTests/TestingIsMemberBuiltInFunction.csp', '-mc', '100000', '-nodead', '-strict', '-assertions'], 'Testing member(-,-) builtin function.').
cli_testcase(861, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/UnknownExpandChannelPatternExpressionError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising unknown expand channel pattern expression error (possible when expanding sharing, synchronisation-sets/closures).').
cli_testcase(862, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/UnsupportedRefinementOperatorError.csp', '-strict', '-assertions', '-expcterr', 'unsupported_refinement_operator', '-expcterr', 'check_assertions', '-expcterr', 'checkAssertion'], 'Testing for raising unsupproted refinement operator error (e.g. [R=).').
cli_testcase(863, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/NonAtomicArgumentError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'check_channel_output_value', '-expcterr', 'general_error_occurred', '-bf'], 'Testing for raising Non-atomic argument error.').
cli_testcase(864, [coverage, csp_test], ['examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures.csp', '-strict', '-csp_add_unicode','examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode.csp'], 'Testing the unicode features of the CSPM tool.').
cli_testcase(865, [coverage, csp_test], ['examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode_saved.csp', '-strict', '-csp_remove_unicode', 'examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_no_unicode.csp'], 'Testing the unicode features of the CSPM tool.').
cli_testcase(866, [error_checks,csp_test], ['examples/CSP/ErrorModels/csp_sets/CouldNotExpandInfiniteDatatypeBody.csp', '-strict', '-mc', 10, '-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing for raising could not expand infinite datatype body error.').
cli_testcase(867, [coverage, csp_test], ['examples/CSP/mydemos/OtherCoverageTests/ExpandingDatatypeBodies.csp', '-strict', '-mc', 10], 'Testing expanding datatype bodies with complicated datat structures.').
cli_testcase(868, [tickets], ['examples/CSPB/Dining/Dining.mch', '-ltlformula', 'G {taken= {} ', '-strict', '-expcterr', 'ltl'], 'Testing that LTL parser does not loop. (PARSERLIB-11)').
cli_testcase(869, [tickets], ['examples/CSPB/Dining/Dining.mch', '-ctlformula', 'AG {taken= {} ', '-strict', '-expcterr', 'ctl'], 'Testing that CTL parser does not loop. (PARSERLIB-11)').
cli_testcase(870, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/Renaming.csp', '-strict', '-mc', 10], 'Testing for raising could not expand infinite datatype body error.').
cli_testcase(871, [tickets], ['examples/EventBPrologPackages/openETCS/ATB/Model_ctx.eventb', '-init', '-strict', '-properties'], 'Check that no time-out occurs; check C <: S21*S22 <=> C : S21 <-> S22 rule in check_subset_of0').
cli_testcase(872, [plugin_units], ['examples/B/PragmasUnits/Tests/WrongOperationTopParameter.mch', '-plugin', 'units', '-strict', '-expecterr', well_definedness_error, '-mc', 5], '').
cli_testcase(873, [plugin_units], ['examples/B/PragmasUnits/Tests/WrongConversionMultiplication.mch', '-plugin', 'units', '-strict', '-expecterr', well_definedness_error, '-mc', 5], '').
cli_testcase(874, [plugin_units], ['examples/B/PragmasUnits/Tests/WrongConversionAddition.mch', '-plugin', 'units', '-strict', '-expecterr', well_definedness_error, '-mc', 5], '').
cli_testcase(875, [tickets,error_checks,'PROB-213'], [], ['-eval', 'g = {x|->y | x=y+1 }', '-expecterr', type_expression_error], 'Check that a type error instead of an internal failure is reported when an expression is used where an identifier is expected (Ticket PROB-213).').
cli_testcase(876, [csp_test], ['examples/CSP/mydemos/demo/mbuff.csp', '-assertions', '-strict'], 'Checking CSP assertions.').
cli_testcase(877, [csp_test], ['examples/CSP/mydemos/demo/crossing.csp', '-assertions', '-strict'], 'Checking CSP assertions.').
cli_testcase(878, [csp_repl], ['-csp_eval', '(seq(set(<1,2,3>))==<1,2,3>) and (set(seq({1,2,3}))=={1,2,3})', '-strict'], 'Checking CSP expressions without loading a specific model first.').
cli_testcase(879, [csp_test], ['examples/CSP/mydemos/demo/crossing.csp', '-csp_eval', 'diff(NonTock,{|enter,leave|}) == {error,gate.down,gate.go_down,gate.go_up,gate.up,sensor.in,sensor.out}', '-strict'], 'Checking CSP expressions.').
cli_testcase(880, [csp_test], ['examples/CSP/mydemos/demo/crossing.csp', '-csp_eval', 'diff(NonTock,{|enter,leave|}', '-expcterr', 'parsing_csp_expression', '-strict'], 'Checking raising parse error while evaluating syntactically non correct CSP expressions.').
cli_testcase(881, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/Choice.mch', '-plugin', 'units', '-strict','-expecterr', 'Multiple Units inferred','-pp','examples/B/PragmasUnits/InternalRepresentationTests/Choice_internal.mch'], '').
cli_testcase(882,[private,data_validation],['../probprivate/examples/B/Alstom/Ticket_Oct11_2012/Rule_DB_SIGAREA_0024.mch','-strict', '-t', '-assertions'], 'Issue related to cut in b_try_check_boolean_expression_lwf.').
cli_testcase(883, [csp_test, assertions], ['examples/CSP/ucsexamples/chapter14/maxpro.csp', '-assertions', '-strict'], 'Checking CSP assertions from file').
cli_testcase(884, [b_test,proz], ['examples/Z/Woodcock/InnerMember.fuzz', '-t', '-mc', 100, '-strict'], 'Checking Z model').
cli_testcase(885, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/TryingToApplyUndefinedFunctionError.csp', '-strict', '-mc', 10, '-expcterr', 'evaluate_agent_call_parameters', '-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing for raising trying to apply undefined process/function error.').
cli_testcase(886, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/CouldNotUnifyLambdaArgsError.csp', '-strict', '-mc', 10, '-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing for raising of could not unify lambda args error.').
cli_testcase(887, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp/CallingNonAtomicFunctionArgumentError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising of trying to call non-atomic function error.').
cli_testcase(888, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp_analyzer/UnsupportedAppendPatternError.csp', '-expcterr', 'compile_head_para', '-strict'], 'Testing for raising of unsupported append pattern error.').
cli_testcase(889, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/TestingComplexChannelTypes.csp', '-assertions', '-strict'], 'Testing outputing channel values of complex types.').
cli_testcase(890, [error_checks,csp_test], ['examples/CSP/ErrorModels/haskell_csp_analyzer/IllegalASTCSPProcessInIllegalPositionError.csp', '-expcterr', 'haskell_csp_analyzer', '-strict'], 'Testing for raising of unsupported append pattern error.').
cli_testcase(891, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/PatternMatchingAppendPatternsAsChannelInputValues.csp', '-mc', 1000, '-nodead', '-assertions', '-strict'], 'Testing for pattern matching append patterns as input values.').
cli_testcase(892, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/CheckingPossibleInputVariablesOnChannel.csp', '-assertions', '-strict'], 'Testing for possible input patterns on channels.').
cli_testcase(893,[private,'PROB-215'],['../probprivate/examples/B/L4MicroKernel/L4API/API.mch','-strict', '-cbc_sequence', 'TimeoutIPC','-p', 'DEFAULT_SETSIZE', 3], 'Check that no errors produced by cbc_sequence (Ticket PROB-215).').
cli_testcase(894,['PROB-215'],['-enabling_analysis_csv','examples/B/Daniel/minienabling_analysis.csv','examples/B/Daniel/minienabling.mch'], 'Check CSV output of a simple enabling analysis (Ticket PROB-215).').
cli_testcase(895, [modelcheck,csp_test], ['examples/CSP/Tickets/3675/SafetyPRB.csp', '-assertions', '-strict'],'Assertions check of processes with complex events.').
cli_testcase(896, [kodkod_test,'PROB-218'], ['-p','KODKOD','TRUE','-eval','card(10..x) = 0 & x : 0..100 & (x<10 => x+x>5)', '-strict'],'The translation to Kodkod should not fail because a helper variable has no possible values (Ticket PROB-218)').
cli_testcase(897, [kodkod_test,'PROB-220'], ['-p','KODKOD','TRUE','-eval','S<:1..10 & S/={}', '-strict'],'Checking the information about the integer range of the elements of an empty set (Ticket PROB-220)').
cli_testcase(898, [b_test], ['-evalt', 'x \\/ 2..n = 0..(n+10) & x /\\ 2..n = {} & x <: 0..n+20 & n=10000', '-strict'],'Checking set partition inside eval').
cli_testcase(899, [kodkod_test,'PROB-217'], ['-p','KODKOD','TRUE','-mc',1000,'examples/B/Tickets/PROB-217/kodkod_overflow.mch', '-nodead', '-strict'],'The output of the predicate analysis alone does not prevent integer overflows -  (Ticket PROB-217)').
cli_testcase(900, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspref', 'Q(3)', '[F=', 'Q3', '-cspref', 'Q3', '[F=', 'Q(3)', '-cspref', 'Q(4)', '[T=', 'Q(1,3)', '-cspref', 'Q(1,3)', '[F=', 'Q(4)', '-strict'], 'Checking various refinement assertions of processes with arguments.').
cli_testcase(901, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspdeadlock', 'Q(10)',  'FD', '-expcterr', 'model_check_fails', '-strict'], 'Deadlock checking of a process called with argument.').
cli_testcase(902, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csplivelock', 'Q2(9)', '-strict'], 'Livelock checking of a process called with argument.').
cli_testcase(903, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspdeterminism', 'Q2(9)', 'FD', '-expcterr', 'model_check_fails','-strict'], 'Determinism checking of a process called with argument (must fail).').
cli_testcase(904, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspdeadlock', 'a.1 -> a.2 -> STOP',  'FD', '-expcterr', 'model_check_fails', '-strict'], 'Deadlock checking of a process called with argument.').
cli_testcase(905, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspref', 'a.1 -> a.2 -> STOP',  '[T=', 'a.1 -> STOP', '-strict'], '').
cli_testcase(906, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspref', 'a.1 -> a.2 -> STOP',  '[F=', 'a.1 -> STOP', '-expcterr', 'refinement_check_fails', '-strict'], '').
cli_testcase(907, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspref', 'a.1 -> a.2 -> STOP',  '[F=', 'a.1 -> b.1 -> STOP', '-expcterr', 'checkAssertion', '-expcterr', 'parse_single_csp_expression_file', '-strict'], '').
cli_testcase(908, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspref', 'a.1 -> a.2 -> STOP',  '[F=', 'a.1 -> a.1.1 -> STOP', '-expcterr', 'checkAssertion', '-expcterr', 'enumerate_action', '-strict'], '').
cli_testcase(909, [csp_test], ['examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-cspref', 'Q(4) [] Q(2)',  '[FD=', 'Q(2,2) [] Q(1,1)', '-strict'], '').
cli_testcase(910, [analysis_test], ['-pacheck','examples/B/predicate_analysis/sets.mch', '-strict'], 'Test cases for the predicate analysis: rules about sets').
cli_testcase(911, [analysis_test], ['-pacheck','examples/B/predicate_analysis/arithmetic.mch', '-strict'], 'Test cases for the predicate analysis: some arithmetic (mainly SIGMA)').
cli_testcase(912, [kodkod_test,cbc], ['examples/B/Kodkod/TestCBCAssertionCheck.mch', '-cbc_assertions', '-strict', '-p','KODKOD', 'TRUE', '-expcterr',cbc_assertions], 'Check that Kodkod CBC Checking works.').
cli_testcase(913, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval', 'D1 == {A,B.0,B.1,B.2,B.3,C.0.0,C.0.1,C.0.2,C.0.3,C.1.0,C.1.1,C.1.2,C.1.3,C.2.0,C.2.1,C.2.2,C.2.3,C.3.0,C.3.1,C.3.2,C.3.3}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(914, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval', '{|c.C|} =={c.(C.0.0),c.(C.0.1),c.(C.0.2),c.(C.0.3),c.(C.1.0),c.(C.1.1),c.(C.1.2),c.(C.1.3),c.(C.2.0),c.(C.2.1),c.(C.2.2),c.(C.2.3),c.(C.3.0),c.(C.3.1),c.(C.3.2),c.(C.3.3)}', '-strict'], 'Evaluating CSP-M expressions').
cli_testcase(915, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval', '{1..3} == {1,2,3}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(916, [csp_test], ['examples/CSP/simple/verysimple.csp', '-csp_eval', 'Events', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(917, [csp_test], ['examples/CSP/simple/verysimple.csp', '-csp_eval', '{|a|}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(918, [error_checks, csp_test], ['examples/CSP/simple/verysimple.csp', '-csp_eval', '{1..}','-strict','-expcterr',expand_symbolic_set], 'Testing raising of infinite set error.').
cli_testcase(919, [b_test, csp_test], ['examples/CSP/Tickets/WilliamsCSPB/Airports.mch', '-csp-guide', 'examples/CSP/Tickets/WilliamsCSPB/AirportsCtrl.csp', '-assertions', '-strict'], 'Checking CSP-M assertions for CSP-M || B models.').
cli_testcase(920, [csp_test], ['examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval', 
             '{x | x <- D1, (B.1) != x} == {A,B.0,B.2,B.3,C.0.0,C.0.1,C.0.2,C.0.3,C.1.0,C.1.1,C.1.2,C.1.3,C.2.0,C.2.1,C.2.2,C.2.3,C.3.0,C.3.1,C.3.2,C.3.3}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(921, [error_checks, csp_test], ['examples/CSP/ErrorModels/csp_sets/EmptySetNotAllowedForInterError.csp', '-mc', 10,'-expcterr', 'is_member_set2', '-expcterr', 'general_error_occurred', '-strict'], 'Testing raising of empty set not allowed for Inter(-) error.').
cli_testcase(922, [csp_test], ['-csp_eval', 'if member(1,{1..10}) then true else false', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(923, [error_checks,'PROB-226'], ['-mc', 1000, 'examples/B/Tickets/PROB-226/parallel_subst.mch', '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict'], 'Parallel assignments to the same variable should be forbidden.').
cli_testcase(924, [error_checks], ['-mc', 1000, 'examples/B/ErrorMachines/UnknownPromotedOp.mch', '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict'], 'Type-checker error when promoting an operation that does not exist.').
cli_testcase(925, [error_checks], ['-mc', 1000, 'examples/B/ErrorMachines/IncludeSeenOnlyOnce.mch', '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict'], 'Expecting error when including a machine twice that is used as a seen/used machine.').
cli_testcase(926, [csp_test,ltl], ['examples/CSP/simple/verysimple.csp', '-ltlformula', 'GF [a]', '-expcterr', 'ltl', '-strict'],'Checking for raising number of arguments not correct error.').
cli_testcase(927, [typechecker_test], ['-eval', 'a-b=c & a=0', '-strict'], 'Type-checking a minus when the type of the arguments is determinded later (an integer).').
cli_testcase(928, [typechecker_test], ['-eval', 'a-b=c & a={0}', '-strict'], 'Type-checking a minus when the type of the arguments is determinded later (a set).').
cli_testcase(929, [csp_test,ltl], ['examples/CSP/ErrorModels/ltl/ExtraArgumentsInCSPLTLPatternError.csp', '-ltlformula', 'GF [a.red.red.red]', '-expcterr', 'check_csp_arguments', '-expcterr', 'enumerate_action', '-expcterr', 'ltl', '-strict'],'Checking for raising number of arguments not correct error.').
cli_testcase(930, [csp_test,ltl], ['examples/CSP/ErrorModels/ltl/MissingElementsInCSPPatternError.csp', '-ltlformula', 'GF [a.<1>]', '-expcterr', 'check_csp_arguments','-expcterr', 'ltl', '-strict'],'Checking for raising number of arguments not correct error.').
cli_testcase(931, [csp_test], ['examples/CSP/ucsexamples/chapter20/phone.csp', '-csp_eval', 'Phones', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(932, [typechecker_test,error_checks], ['-eval', 'a-b=c & a=TRUE', '-expcterr', 'type_expression_error', '-strict'], 'Type-checking a minus when the type of the arguments is determinded later (type error).').
cli_testcase(933, [typechecker_test,error_checks], ['-eval', 'a*b=c & a=TRUE', '-expcterr', 'type_expression_error', '-strict'], 'Type-checking * when the type of the arguments is determinded later (type error).').
cli_testcase(934, [typechecker_test], ['-eval', '[a,c:=1,3](a<b&b<c)', '-strict'], 'Type-checking a substitution on a predicate').
cli_testcase(935, [typechecker_test,error_checks], ['-eval', '[a,c:=1,3,5](a<b&b<c)', '-expcterr', 'type_expression_error', '-strict'], 'Type-checking a substitution on a predicate (type error)').
cli_testcase(936, [b_test, csp_test,ltl], ['examples/CSP/Tickets/WilliamsCSPB/Airports.mch', '-csp-guide', 'examples/CSP/Tickets/WilliamsCSPB/AirportsCtrl.csp', '--csp_event', '-ltlformula', 'G not [link.AMS.AMS]', '-strict'], 'Checking ltl formulas for CSP-M || B models.').
cli_testcase(937, [b_test, csp_test,ltl], ['examples/CSP/Tickets/WilliamsCSPB/Airports.mch', '-csp-guide', 'examples/CSP/Tickets/WilliamsCSPB/AirportsCtrl.csp', '-ltlformula', 'G not [link(AMS,AMS)]', '-strict'], 'Checking ltl formulas for CSP-M || B models.').
cli_testcase(938, [error_checks], ['-mc', 1000, '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict', 'examples/B/ErrorMachines/MachineWithParametersErr1.ref'], 'Check for error: Wrong number of parameters in refinement.').
cli_testcase(939, [error_checks], ['-mc', 1000, '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict', 'examples/B/ErrorMachines/MachineWithParametersErr2.ref'], 'Check for error: Wrong parameters in refinement.').
cli_testcase(940, [cbc], ['-cbc', all, '-p','CLPFD', 'TRUE', '-strict', 'examples/B/Tickets/Treharne1/House_Tracker16.mch'], 'Check that we quickly see that there is no counterexample.').
cli_testcase(941, [cbc], ['-cbc', all, '-p','CLPFD', 'TRUE', '-strict', 'examples/B/CBC/ReservationLivre.mch', '-expcterr',cbc], 'Check that we find counterexample with larger scope.').
cli_testcase(942, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/Case.mch', '-plugin', 'units', '-strict', '-expcterr', 'well_definedness_error', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/Case_internal.mch','-p','plugin(units,fixpoint_halt_on_error)','false'], '').
cli_testcase(943, [typechecker_test], ['examples/B/Pragmas/UnrecognisedPragma.mch','-strict','-expcterr',pragma],'').
cli_testcase(944, [dot], ['-evalt','r:1..3-->>1..3 & r(1)/=1 & r(1)/=2 & r(2)/=2', '-evaldot', 'examples/Eval/evalout.dot','-strict'],'').
cli_testcase(945, [eventb_test], ['-t', '-strict', 'examples/EventBPrologPackages/Theory/TestBoolOps_mch.eventb'], 'Check BoolOps Theory.').
cli_testcase(946, [cbc], ['examples/B/Tickets/CBC_Refinement/M2_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', all, '-strict', '-expcterr', 'cbc', '-p', 'SMT', 'TRUE'], 'Check counter-example found in SMT mode').
cli_testcase(947, [cbc], ['examples/B/Tickets/CBC_Refinement/M2_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'RemAuth', '-strict', '-expcterr', 'cbc', '-p', 'SMT', 'TRUE'], 'Check counter-example found in SMT mode').
cli_testcase(948, [csp_test], ['examples/CSP/other/Ivo/WaterBoiler.csp', '-csp_eval', 'member(fill.3,{|fill|}) == true', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(949, [tickets], ['-assertions', '-strict', 'examples/B/Tickets/PrjMemCheck/PrjTestsOK.mch'], 'Check that we prj1/prj2 work correctly.').
cli_testcase(950, [tickets], ['-assertions', '-strict', 'examples/B/Tickets/PrjMemCheck/PrjTestsKO1.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(951, [tickets], ['-assertions', '-strict', 'examples/B/Tickets/PrjMemCheck/PrjTestsKO2.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(952, [tickets], ['-assertions', '-strict', 'examples/B/Tickets/PrjMemCheck/PrjTestsKO3.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(953, [tickets], ['-assertions', '-strict', 'examples/B/Tickets/PrjMemCheck/PrjTestsKO4.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(954, [eventb_test], ['-t','-strict','examples/EventBPrologPackages/Theory/DumbDataTheoryModel.eventb'],'Check model that uses user defined datatypes via the Theory Plugin.').
cli_testcase(955, [eventb_test], ['-mc',100,'-strict','examples/EventBPrologPackages/Theory/NonTrivialTypeParameters.eventb'],'Theory Plugin: Datatypes whose type parameters are not only type expressions').
cli_testcase(956, [eventb_test], ['-mc',1000,'-strict','examples/EventBPrologPackages/Puzzles/Hanoi_mch_3s_6d.eventb', '-noass', '-cc', 731, 2186, '-p', 'CLPFD', 'TRUE'],'Check that constraint solver not perplexed by use of card() for last').
cli_testcase(957, [plugin_units], ['examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations.mch', '-plugin', 'units', '-strict', '-pp', 'examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations_internal.mch'], '').
cli_testcase(958, [eventb_test], ['-t', '-p', 'PROOF_INFO', 'FALSE'  , '-p', 'TIME_OUT', '200', '-expcterr', 'time_out',
'-strict','examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant2_v2.eventb'],'Check that Time Out occurs without Proof Info usage').
cli_testcase(959, [eventb_test], ['-mc',100, '-t',  '-p', 'TIME_OUT', '200', '-strict','examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant2_v2.eventb'],'Check that Proof Info works').
cli_testcase(960, [eventb_test], ['-mc',0,  '-p', 'PROOF_INFO', 'FALSE'  , '-p', 'TIME_OUT', '200', '-expcterr', 'time_out', '-strict','examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant_INITCheck_mch.eventb'],'Check that Proof Info works also for INITIALISATION').
cli_testcase(961, [eventb_test], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-p', 'TIME_OUT', '200', '-strict','examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant_INITCheck_mch.eventb'],'Check that Proof Info works also for INITIALISATION').
cli_testcase(962, [eventb_test], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-p', 'TIME_OUT', '200', '-strict','examples/EventBPrologPackages/ProofDirected/ComplicatedProvenDynamicTheorem_mch.eventb'],'Check that Proof Info works also for dynamic theorems').

cli_testcase(963, [csp_test,ltl], ['examples/CSP/other/Ivo/WaterBoiler.csp', '-ltlformula', 'GF [fill._]', '-strict'],'Check LTL formula in File.').
cli_testcase(964, [eventb_test], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', well_definedness_error,  '-expcterr', invariant_violation, '-strict','examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Inv_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined invariants').
cli_testcase(965, [eventb_test], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', well_definedness_error,  '-expcterr', assertion_violation, '-strict','examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Thm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems').
cli_testcase(966, [eventb_test], ['-mc',100,  '-noass', '-nodead', '-p', 'PROOF_INFO', 'FALSE'  , '-expcterr', invariant_violation, '-strict','examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Thm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems and -noass checking').
cli_testcase(967, [eventb_test], ['-mc',100,  '-noass', '-nodead', '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', invariant_violation, '-strict','examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Thm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems and -noass checking'). 
cli_testcase(968, [eventb_test], ['-mc',100,  '-noass', '-nodead', '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', 'model_check_incomplete', '-strict','examples/EventBPrologPackages/ProofDirected/TestProofDirected_FalseThm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems and -noass checking'). 
cli_testcase(969, [b_test,recursion,external], ['-assertions', '-strict','examples/B/Tickets/Bracknell1/C0.mch'],'Check that no recursion detected').
cli_testcase(970, [b_test,recursion,external], ['-init', '-strict','examples/B/Tickets/Bracknell1/C0_ctx.eventb'],'Check that no recursion detected').  
cli_testcase(971, [b_test], ['examples/B/Tester/Prj1NATURAL.mch', '-t', '-mc', '1000', '-strict'], 'Check that prj1(NATURAL,...) not expanded').
cli_testcase(972, [cbc], ['examples/B/CBC/plavis_MemoryLoad_SP_14_TestComp.mch', '-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 3, '-p', 'MININT', -1, '-strict'], 'Check that partitioning into components works').
cli_testcase(973, [cbc], ['examples/B/CBC/ComponentPartitioningTest.mch', '-p', 'CLPFD', 'TRUE', '-init', '-strict'], 'Check that partitioning into components works').
cli_testcase(974, [cbc], ['examples/B/CBC/PartitionConstraintAssertionCheck.mch', '-p', 'CLPFD', 'TRUE', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict', '-v'], 'Check that partitioning into components works').
cli_testcase(975, [por], ['examples/B/Ivo/PruningActionsTest1.mch', '-mc', '1000000', '-POR', '-expcterr', invariant_violation, '-strict'], 'POR test.').
cli_testcase(976, [por], ['examples/B/Ivo/PruningActionsTest2.mch', '-mc', '1000000', '-POR', '-expcterr', deadlock, '-strict'], 'POR test..').
cli_testcase(977, [por], ['examples/B/Ivo/BridgeTransitions.mch', '-mc', '1000000', '-POR', '-get_min_max_coverage', 'examples/B/Ivo/BridgeTransitions_por.cov', '-strict'], 'POR test.').
cli_testcase(978, [por], ['examples/B/Ivo/CycleDetections.mch', '-mc', '1000000', '-POR', '-expcterr', invariant_violation, '-strict'], 'POR test.').
cli_testcase(979, [private,data_validation], ['../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.mch', '-animate_all', '-strict', '-cc', 113, 112, '-p', 'TIME_OUT', 15000, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.his','-expcterr', invariant_violation], 'Rule Test 4 of Alstom compilation project').
cli_testcase(980, [private,data_validation], ['../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.mch', '-animate_all', '-strict', '-cc', 40, 39, '-p', 'TIME_OUT', 15000, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.his'], 'Rule Test 5 of Alstom compilation project').
cli_testcase(981, [private,data_validation], ['../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.mch', '-animate_all', '-strict', '-cc', 97, 96, '-p', 'TIME_OUT', 15000, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.his'], 'Rule Test 1 of Alstom compilation project').
cli_testcase(982, [private,data_validation], ['../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori.mch', '-animate_all', '-strict', '-cc', 48, 47, '-p', 'TIME_OUT', 45000, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori.his','-expcterr', invariant_violation], 'Rule Test 3 of Alstom compilation project').
cli_testcase(983, [private,data_validation], ['../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_ESA_TRIGGER_0009_ori.mch', '-strict', '-cc', 112, 111, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-t'], 'Rule Test 2 of Alstom compilation project').
cli_testcase(984, [private,symmetry], ['../probprivate/examples/B/BossaAcadie/scheduler.mch', '-strict', '-cc', 145, 1936, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-mc', 9000, '-p', 'MAX_OPERATIONS', 20, '-p', 'SYMMETRY_MODE', hash], 'Bossa example').
cli_testcase(985, [cbc], ['examples/B/Mathematical/GraphIso/HardGraph_Nr5.mch', '-strict', '-cc', 3, 3, '-p', 'TIME_OUT', 5000,  '-mc', 20, '-p', 'MAX_OPERATIONS', 4, '-p', 'CLPFD', 'TRUE'], 'Hard graph; isomorphism detected means no deadlock').
cli_testcase(986, [eventb_test], ['-t','-mc',100,'-strict','examples/EventBPrologPackages/Puzzles/MRushHour_new.eventb','-expcterr', invariant_violation],'Check that new Proof Info works with wd in guards/actions accepted').
cli_testcase(987, [por], ['examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2.eventb', '-mc', '1000000', '-noass', '-nogoal', '-POR', '-get_min_max_coverage', 'examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2_por.cov', '-strict'], 'Testing if coverages of the constants and values of the generated reduced state space are matching with the coverages of the constants and values of the full state space.').
cli_testcase(988, [csp,dot], ['examples/CSP/simple/Copy.csp', '-mc', '1000', '-cc', '14', '24', '-strict','-spmdot','examples/CSP/simple/Copy_spmdot.dot','-spddot','examples/CSP/simple/Copy_spddot.dot', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Test reduced state space dot output for CSP.').
cli_testcase(989, [eventb_test], ['-assertions','-strict','examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx.eventb','-pp', 'examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx_internal.mch'],'Check that set comprehensions converted properly').
cli_testcase(990, [b_test], ['-evalt', 'ran(%v0.(v0: 0..30001 | v0)) = 0..30001', '-p', 'SMT', 'TRUE', '-strict'],'Checking non-regresson against issue reported by Valério Medeiros Jr on 28.1.2013').
cli_testcase(991, [tickets,csp], ['examples/CSP/Tickets/RecursiveInclude/Procs.csp', '-mc', 100000, '-t', '-strict'], 'Testing of recursive and local (by means of no absolute path files must be given in order to include a CSP model to another CSP model) including for CSP models.').
cli_testcase(992, [b_test], ['-evalt', '{x|  x = x mod 33 & x = x mod 44} = 0..32', '-p', 'CLPFD', 'TRUE', '-strict'],'Ensure mod constraint propagation works').
cli_testcase(993, [csp], ['../probprivate/examples/CSP/Kharmeh_PhD/4-isa-system-evaluation.csp', '-mc', 2000, '-cc', 292, 487, '-strict'],'Non trivial test case').
cli_testcase(994, [b_test], ['examples/B/FeatureChecks/CaseStmt.mch', '-t', '-mc', 200, '-cc', 4, 13, '-strict'],'Check case statement with else').
cli_testcase(995, [laws,rel_fnc], ['examples/B/Laws/FunLawsWithLambda.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'MEMO', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Laws this time with MEMO and CLPFD').
cli_testcase(996, [b_test], ['-evalt', '{x|x/2 : 20..30} = 40..61', '-evalt', '{x|x/20 : 22..23} = 440..479', '-evalt', 'ran(%v0.(v0: 0..20000 | [v0 mod 2/1,v0 mod 4/2,v0 mod 8/4,v0 mod 16/8,v0 mod 32/16,v0 mod 64/32,v0 mod 128/64])) = 1..7 --> 0..1', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict'],'Ensure division constraint propagation works').
cli_testcase(997, [csp_test,refusal_trace_refinement], ['examples/CSP/RefusalBasedRefinement/RefusalBasedRefinement_simple.csp', '-assertions', '-strict'],'Simple test case for refusal based refinement checking (CSP).').
cli_testcase(998, [csp_test,refusal_trace_refinement], ['examples/CSP/RefusalBasedRefinement/RefTestRoscoeBook.csp', '-assertions', '-strict'],'Test case for refusal based refinement checking (CSP) from Roscoe\'s book. Note: Difference in the results delivered by FDR and ProB of the commented out assertion "P2 [R= P1".').
cli_testcase(999, [eventb_test], ['examples/B/Tickets/197/robot_ctx02_ctx.eventb', '-init', '-properties', '-p', 'DEFAULT_SETSIZE', 2, '-strict'],'Check WD Axiom loading').
cli_testcase(1000, [b_test], ['-evalt', 'x \\/ y \\/ z <: 1011..1013 & card(x) > card(y) & card(y)>card(z)', '-evalt', '{x,y,z|x \\/ y \\/ z <: 21011..21012 & card(x) > card(y) & card(y)>card(z)} = {(({21011,21012}|->{21011})|->{}),(({21011,21012}|->{21012})|->{})}', '-p', 'CLPFD', 'TRUE', '-strict'], 'Ensure that domains propagated to union arguments.').
cli_testcase(1001, [b_test], ['-evalt', '{x|x <| [1,2,3] = [1,2] & card(x)=2} = {{1,2}}',
 '-evalt', 'x <| s = [1,2,3] \\/ {29|->29}  & x <: 1..1000 & s = %i.(i:1..300|i)', 
 '-evalt', 's |> x = [1,2,3] \\/ {29|->29}  & x <: 1..1000 & s = %i.(i:1..300|i)', 
 '-evalt', '{x| [1,4,3]|>x = [1,4] & card(x)=2}={{1,4}}',
 '-evalf', 's |> x = [1,2,3] \\/ {29|->29}  & x <: 2..3000 & s = %i.(i:1..900|i) ',
 '-evalt', '{x|#s.(x <| s = [1,2,3] \\/ {29|->29}  & x <: 1..30 & s = %i.(i:1..300|i))} = {{1,2,3,29}}',
 '-evalt', '{z,x|[1,2,z] |>> x = [1,2]  & z:12..13 & x<:11..12}={(12|->{11,12}),(12|->{12})}',
 '-evalt', 'x <<| [11,12,13,14,z] = [11,12,13,14]  & z:12..13 ',
 '-evalt', 'x <| [111,222,333,444,555]  = [111,222,333,444]',
 '-evalt', 'x <<| [111,222,333,444,555]  = [111,222,333,444]',
 '-evalt', '[111,222,333,444,555] |>x  = [111,222,333,444]',
 '-evalt', '[111,222,333,444,555] |>>x  = [111,222,333,444]',
 '-evalt', '{x|{x} <<| [111,222,333,444,555,666]  = [111,222,333,444,555]} = {6}',
 '-evalt', '{x| [111,222,333,444,555,666] |>> {x}  = [111,222,333,444,555]} = {666}',
 '-evalt', '{x| {x} <| [111,222,333,444,555,666]  = {5|->555}} = {5}',
 '-evalt', '{x| [111,222,333,444,555,666] |> {x}  = {5|->555}} = {555}',
 '-evalt', '{x| [111,222,333,444,555,555] |> {x}  = {5|->555,6|->555}} = {555}',
  '-p', 'CLPFD', 'TRUE', '-strict'], 'Ensure that domain/range restriction properly treated.').
cli_testcase(1002, [cbc], ['-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 127, '-p', 'MININT', -128,
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bank-Extraer_SP_29.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_22.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_11.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_12.prob',
  '-expcterr', 'eval_string_enum_warning'], 'SetLog tests now succeeding tests').
cli_testcase(1003, [cbc], ['-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 127, '-p', 'MININT', -128,
  %sc
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/sc-SCAddCat_SP_4.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/sc-SCAddCat_SP_22.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/sc-SCSetLevel_DNF_3.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/sc-SCSetSC_DNF_3.prob',
  %bank
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bank-CerrarCaja_SP_12.prob',
  %banco G
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_18.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_2.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_24.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_25.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_4.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_5.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_8.prob',
  %qsee
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_10.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_9.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpAux_SP_6.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpAux_SP_8.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_3.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_6.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_8.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitScientificData_SP_16.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitScientificData_SP_24.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/qsee-TransmitScientificData_SP_8.prob',
  %plavis ; two more missing
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_45.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_52.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-MemoryLoad_SP_55.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-RetrieveEData_SP_23.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_13.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_24.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_3.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_31.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_38.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/plavis-TransData_SP_39.prob',
  % iae
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_103.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_146.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_148.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_149.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_18.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_56.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_58.prob',
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_59.prob',
  % sensors
  '-evalt_file', '../probprivate/examples/Setlog/prob-ttf/sensors-KeepMaxReading_SP_5.prob',
  '-expcterr', 'eval_string_enum_warning'], 'SetLog tests some originally succeeding tests').
cli_testcase(1004, [cbc_refinement], ['examples/EventBPrologPackages/RefinementChecking/CCounterErr_mch.eventb', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1005, [cbc_refinement], ['examples/EventBPrologPackages/RefinementChecking/CCounter_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1006, [cbc_refinement], ['examples/EventBPrologPackages/RefinementChecking/CEnterLeave_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1007, [cbc_refinement], ['examples/EventBPrologPackages/RefinementChecking/CEnterLeaveErr1_mch.eventb', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1008, [cbc_refinement], ['examples/EventBPrologPackages/RefinementChecking/CEnterLeave_Rename_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1009, [cbc_refinement], ['examples/EventBPrologPackages/RefinementChecking/CMergeErr_mch.eventb', '-cbc_refinement', '-expcterr', cbc_refinement, '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(1010, [cbc_refinement], ['examples/EventBPrologPackages/RefinementChecking/CMerge_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1011, [cbc_refinement], ['examples/ForDistribution/Refinement/ServerR.ref', '-cbc_refinement'], '').
cli_testcase(1012, [cbc_refinement], ['examples/ForDistribution/Refinement/ServerR_wrong.ref', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1013, [cbc_refinement], ['examples/B/Implementations/ConcreteVars/MI.imp', '-cbc_refinement'], '').
cli_testcase(1014, [b_test], ['examples/B/Mathematical/FibMemo.mch', '-mc', 120, '-p', 'CLPFD', 'TRUE', '-strict', '-expcterr', 'deadlock', '-cc', 100, 99], 'Test that no CLPFD overflow occurs'). % no warning is generated for clpfd overflows at the moment; so the test is not really that useful
cli_testcase(1015, [cbc_refinement], ['examples/B/Implementations/Copy/ReturnEqualI.imp', '-cbc_refinement'], '').
cli_testcase(1016, [cbc_refinement], ['examples/B/Implementations/Division/DivisionImplOk.imp', '-cbc_refinement'], '').
cli_testcase(1017, [cbc_refinement], ['examples/B/Implementations/Division/DivisionImplWrongCondition.imp', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1018, [btest, refinement], ['examples/B/RefinementChecking/Schedular/Schedular1.mch', '-refchk', 'examples/B/RefinementChecking/Schedular/Schedular0_refine_spec.P', '-expcterr', 'refinement_check_fails', '-strict'], 'Checking for expected refinement error. Example taken from the paper: Automatic Refinement Checking for B.').
cli_testcase(1019, [cbc_refinement], ['examples/B/Implementations/FTP/FileTransfer1.ref', '-cbc_refinement'], '').
cli_testcase(1020, [cbc_refinement], ['examples/B/Implementations/FTP/FileTransfer2.ref', '-cbc_refinement'], '').
cli_testcase(1021, [cbc_refinement], ['examples/B/Implementations/Mul/MulImp.imp', '-cbc_refinement'], '').
cli_testcase(1022, [cbc_refinement], ['examples/B/Implementations/Mul/MulImpErr.imp', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1023, [cbc_refinement], ['examples/B/Implementations/Power/AExpBImpl.imp', '-cbc_refinement'], '').
cli_testcase(1024, [cbc_refinement], ['examples/B/Implementations/Simple10/Simple10Imp.imp', '-cbc_refinement'], '').
cli_testcase(1025, [cbc_refinement], ['examples/B/Implementations/Simple10/Simple10ImpErr.imp', '-cbc_refinement', '-expcterr', cbc_refinement, '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(1026, [cbc_refinement], ['examples/B/Implementations/Simple10/Simple10ImpErr2.imp', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1027, [cbc_refinement], ['examples/B/RefinementChecking/ServerR.ref', '-cbc_refinement'], '').
cli_testcase(1028, [cbc_refinement], ['examples/B/RefinementChecking/ServerR_wrong.ref', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1029, [cbc_refinement], ['examples/B/RefinementChecking/SwapR.ref', '-cbc_refinement'], '').
cli_testcase(1030, [b_test], ['examples/B/FunctionalProgramming/SequenceTheory.mch', '-t', '-assertions'], 'Check infinite functions and sets').
cli_testcase(1031, [b_test], ['examples/B/FeatureChecks/DeferredSetAndConstants.mch', '-init',
 '-evalt', 'ID3=3 & ID1:ID2 & ID0:ID2 & (xx=ID5 => not(xx=ID6)) & (xx=ID6 => not(xx=ID5))'],'Check that name clashes between deferred set ids and other ids handled gracefully').

% sicstus -l src/test_runner.pl --goal "id(690-694)."
% INSERT NEW TESTCASES ABOVE THIS LINE; cli_testcase is sorted in ascending order
% ----------------------------------------------------------------------
% IF A NEW TESTCASE REQUIRES CHECKING OUTPUT, ADD NEW LINE(S) BELOW
% cli_testcase_diff_check_output is sorted in descending order
cli_testcase_diff_check_output(989,'examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx_internal.mch','examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx_internal_saved.mch').
cli_testcase_diff_check_output(988,'examples/CSP/simple/Copy_spmdot.dot','examples/CSP/simple/Copy_spmdot_saved.dot').
cli_testcase_diff_check_output(988,'examples/CSP/simple/Copy_spddot.dot','examples/CSP/simple/Copy_spddot_saved.dot').
cli_testcase_diff_check_output(987,'examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2_por.cov','examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2.cov').
cli_testcase_diff_check_output(977,'examples/B/Ivo/BridgeTransitions_por.cov','examples/B/Ivo/BridgeTransitions.cov').
cli_testcase_diff_check_output(982,'../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori.his','../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori_saved.his').
cli_testcase_diff_check_output(981,'../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.his','../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori_saved.his').
cli_testcase_diff_check_output(980,'../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.his','../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori_saved.his').
cli_testcase_diff_check_output(979,'../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.his','../probprivate/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori_saved.his').
cli_testcase_diff_check_output(957,'examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations_internal_saved.mch').
cli_testcase_diff_check_output(944,'examples/Eval/evalout.dot','examples/Eval/evalout_saved.dot').
cli_testcase_diff_check_output(942,'examples/B/PragmasUnits/InternalRepresentationTests/Case_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/Case_internal_saved.mch').
cli_testcase_diff_check_output(894,'examples/B/Daniel/minienabling_analysis.csv',
                               'examples/B/Daniel/minienabling_analysis_saved.csv').
cli_testcase_diff_check_output(881,'examples/B/PragmasUnits/InternalRepresentationTests/Choice_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/Choice_internal_saved.mch').
cli_testcase_diff_check_output(865,'examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_no_unicode.csp', 
'examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_no_unicode_saved.csp').
cli_testcase_diff_check_output(864,'examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode.csp', 
'examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode_saved.csp').
cli_testcase_diff_check_output(808,'examples/B/Benchmarks/CarlaTravelAgencyErr_state_graph.dot', 
'examples/B/Benchmarks/CarlaTravelAgencyErr_state_graph_saved.dot').
cli_testcase_diff_check_output(767,'../probprivate/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld_sptxt.txt',
'../probprivate/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld_sptxt_saved.txt').
cli_testcase_diff_check_output(707,'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3_internal.mch',
'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3_internal_saved.mch').
cli_testcase_diff_check_output(706,'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2_internal.mch',
'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2_internal_saved.mch').
cli_testcase_diff_check_output(705,'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1_internal.mch',
'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1_internal_saved.mch').
cli_testcase_diff_check_output(704,'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0_internal.mch',
'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0_internal_saved.mch').
cli_testcase_diff_check_output(701,'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1_internal.mch',
'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1_internal_saved.mch').
cli_testcase_diff_check_output(700,'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0_internal.mch',
'examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0_internal_saved.mch').
cli_testcase_diff_check_output(696,'examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat_internal.mch',
'examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat_internal_saved.mch').
cli_testcase_diff_check_output(695,'examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits_internal.mch',
'examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits_internal_saved.mch').
cli_testcase_diff_check_output(686,'examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations_internal_saved.mch').
cli_testcase_diff_check_output(683,'examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion_internal_saved.mch').
cli_testcase_diff_check_output(681,'examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations_internal_saved.mch').
cli_testcase_diff_check_output(680,'examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i_internal.mch',
'examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i_internal_saved.mch').
cli_testcase_diff_check_output(679,'examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0_internal.mch',
'examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0_internal_saved.mch').
cli_testcase_diff_check_output(678,'examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i_internal.mch',
'examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i_internal_saved.mch').
cli_testcase_diff_check_output(677,'examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_internal.mch',
'examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_internal_saved.mch').
cli_testcase_diff_check_output(676,'examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias_internal_saved.mch').
cli_testcase_diff_check_output(675,'examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash_internal_saved.mch').
cli_testcase_diff_check_output(674,'examples/B/PragmasUnits/InternalRepresentationTests/Functions_internal.mch',
'examples/B/PragmasUnits/InternalRepresentationTests/Functions_internal_saved.mch').
cli_testcase_diff_check_output(673,'examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions_internal.mch',
 'examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions_internal_saved.mch').
cli_testcase_diff_check_output(671,'examples/B/PragmasUnits/InternalRepresentationTests/TypedGlobalSet_internal.mch',
 'examples/B/PragmasUnits/InternalRepresentationTests/TypedGlobalSet_internal_saved.mch').
cli_testcase_diff_check_output(667,'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System_internal.mch',
 'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System_internal_saved.mch').
cli_testcase_diff_check_output(666,'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight_internal.mch',
 'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight_internal_saved.mch').
cli_testcase_diff_check_output(665,'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car_internal.mch',
 'examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car_internal_saved.mch').
cli_testcase_diff_check_output(663,'examples/B/PragmasUnits/InternalRepresentationTests/SimpleFunctionApplication_internal.mch',
 'examples/B/PragmasUnits/InternalRepresentationTests/SimpleFunctionApplication_internal_saved.mch').
cli_testcase_diff_check_output(662,'examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent_internal.mch',
 'examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent_internal_saved.mch').
cli_testcase_diff_check_output(661,'examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion_internal.mch',
 'examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion_internal_saved.mch').
cli_testcase_diff_check_output(657,'examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage_internal.mch',
 'examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage_internal_saved.mch').
cli_testcase_diff_check_output(634,'examples/B/Kodkod/Loop_statespace.dot',
 'examples/B/Kodkod/Loop_statespace_saved.dot').
cli_testcase_diff_check_output(616,'examples/B/Tester/Partition_internal.mch',
 'examples/B/Tester/Partition_internal_saved.mch').
cli_testcase_diff_check_output(613,'examples/B/Pragmas/AssertionPragmas_internal.mch',
 'examples/B/Pragmas/AssertionPragmas_internal_saved.mch').
cli_testcase_diff_check_output(540,'examples/B/TestCaseGeneration/SimpleCounter_results.xml',
 'examples/B/TestCaseGeneration/SimpleCounter_results_saved.xml').

cli_testcase_diff_check_output(514,'examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A0_true.dot','examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A0_true_saved.dot').
cli_testcase_diff_check_output(514,'examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A1_true.dot','examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A1_true_saved.dot').
cli_testcase_diff_check_output(514,'examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A2_true.dot','examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A2_true_saved.dot').
cli_testcase_diff_check_output(514,'examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A3_true.dot','examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A3_true_saved.dot').

cli_testcase_diff_check_output(563,'examples/B/Daniel/TestM_testcases.xml', 'examples/B/Daniel/TestM_testcases_saved.xml').
cli_testcase_diff_check_output(562,'examples/EventBPrologPackages/Pitesti/bank_testfile.xml','examples/EventBPrologPackages/Pitesti/bank_testfile_saved.xml').
cli_testcase_diff_check_output(515,'examples/B/Demo/scheduler_spdot.dot','examples/B/Demo/scheduler_spdot_saved.dot').
cli_testcase_diff_check_output(515,'examples/B/Demo/scheduler_spmdot.dot','examples/B/Demo/scheduler_spmdot_saved.dot').
cli_testcase_diff_check_output(515,'examples/B/Demo/scheduler_spddot.dot','examples/B/Demo/scheduler_spddot_saved.dot').
cli_testcase_diff_check_output(510,'examples/B/Laws/RelLaws_internal.mch','examples/B/Laws/RelLaws_internal_saved.mch').
cli_testcase_diff_check_output(500,'examples/B/Benchmarks/Cruise_finite1_pp.mch','examples/B/Benchmarks/Cruise_finite1_prettyprintedf.mch').
cli_testcase_diff_check_output(499,'examples/B/Benchmarks/Cruise_finite1_pp.mch','examples/B/Benchmarks/Cruise_finite1_prettyprintedf.mch').
cli_testcase_diff_check_output(498,'examples/B/Benchmarks/Cruise_finite1_pp.mch','examples/B/Benchmarks/Cruise_finite1_prettyprinted.mch').
cli_testcase_diff_check_output(497,'examples/B/Benchmarks/Cruise_finite1_pp.mch','examples/B/Benchmarks/Cruise_finite1_prettyprinted.mch').
cli_testcase_diff_check_output(451,'examples/B/Simple/UpCounter_his.txt','examples/B/Simple/UpCounter_his_saved.txt').
cli_testcase_diff_check_output(452,'examples/B/Simple/UpCounter_his.txt','examples/B/Simple/UpCounter_his_saved2.txt').
cli_testcase_diff_check_output(453,'examples/B/Simple/UpCounter_his.txt','examples/B/Simple/UpCounter_his_saved3.txt').
cli_testcase_diff_check_output(454,'examples/B/Simple/UpCounter_his.txt','examples/B/Simple/UpCounter_his_saved4.txt').
cli_testcase_diff_check_output(455,'examples/B/Simple/UpCounter_sptxt.txt','examples/B/Simple/UpCounter_sptxt_saved.txt').
cli_testcase_diff_check_output(461,'examples/B/Daniel/rel_fnc_his.txt','examples/B/Daniel/rel_fnc_his_saved.txt').

cli_testcase_diff_check_output(462,'examples/B/Simple/UpCounter_his.txt','examples/B/Simple/UpCounter_his_saved3.txt').
cli_testcase_diff_check_output(463,'examples/B/CBC/SimpleDeadlock1_his.txt','examples/B/CBC/SimpleDeadlock1_his_saved.txt'). % FAILS with old write_history
cli_testcase_diff_check_output(464,'examples/B/CBC/ProcessIDStarvation_InvErr_his.txt','examples/B/CBC/ProcessIDStarvation_InvErr_his_saved.txt').
cli_testcase_diff_check_output(466,'examples/B/PerformanceTests/WhileLoopBench_his.txt','examples/B/PerformanceTests/WhileLoopBench_saved_his.txt').
cli_testcase_diff_check_output(467,'examples/B/FeatureChecks/IFTHENELSE_his.txt','examples/B/FeatureChecks/IFTHENELSE_saved_his.txt').
cli_testcase_diff_check_output(468,'examples/B/Puzzles/Farmer_sptxt.txt','examples/B/Puzzles/Farmer_saved_sptxt.txt').
cli_testcase_diff_check_output(468,'examples/B/Puzzles/Farmer_his.txt','examples/B/Puzzles/Farmer_saved_his.txt').
cli_testcase_diff_check_output(469,'examples/B/PerformanceTests/LargeSets/LargeExplicitSet_sptxt.txt','examples/B/PerformanceTests/LargeSets/LargeExplicitSet_saved_sptxt.txt').
cli_testcase_diff_check_output(470,'examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_his.txt','examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_saved_his.txt').
cli_testcase_diff_check_output(470,'examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_sptxt.txt','examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_saved_sptxt.txt').
cli_testcase_diff_check_output(471,'examples/B/Benchmarks/CarlaTravelAgencyErr_sptxt.txt','examples/B/Benchmarks/CarlaTravelAgencyErr_saved_sptxt.txt').
cli_testcase_diff_check_output(471,'examples/B/Benchmarks/CarlaTravelAgencyErr_his.txt','examples/B/Benchmarks/CarlaTravelAgencyErr_saved_his.txt').
cli_testcase_diff_check_output(472,'examples/B/SchneiderBook/Chapter12/TownsRRR_sptxt.txt','examples/B/SchneiderBook/Chapter12/TownsRRR_saved_sptxt.txt').
cli_testcase_diff_check_output(472,'examples/B/SchneiderBook/Chapter12/TownsRRR_his.txt','examples/B/SchneiderBook/Chapter12/TownsRRR_saved_his.txt').
cli_testcase_diff_check_output(473,'examples/B/NewSyntax/SET_Game_Rec_sptxt.txt','examples/B/NewSyntax/SET_Game_Rec_saved_sptxt.txt').
cli_testcase_diff_check_output(473,'examples/B/NewSyntax/SET_Game_Rec_his.txt','examples/B/NewSyntax/SET_Game_Rec_saved_his.txt').
cli_testcase_diff_check_output(474,'examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A0_true.dot','examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A0_true_saved.dot').
cli_testcase_diff_check_output(474,'examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A1_false.dot','examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A1_false_saved.dot').
cli_testcase_diff_check_output(474,'examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A2_false.dot','examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A2_false_saved.dot').
cli_testcase_diff_check_output(474,'examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A3_false.dot','examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A3_false_saved.dot').
cli_testcase_diff_check_output(474,'examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A4_false.dot','examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A4_false_saved.dot').
cli_testcase_diff_check_output(476,'examples/B/Tickets/Mejia1/TestInit_statespace.dot','examples/B/Tickets/Mejia1/TestInit_statespace_saved.dot').
cli_testcase_diff_check_output(479,'examples/B/Puzzles/WhoKilledAgatha_statespace.dot','examples/B/Puzzles/WhoKilledAgatha_statespace_saved.dot').
cli_testcase_diff_check_output(480,'examples/B/Benchmarks/Cruise_finite1_statespace.dot','examples/B/Benchmarks/Cruise_finite1_statespace_saved.dot').
cli_testcase_diff_check_output(383,'examples/B/Puzzles/SudokuHexAsConstant_sptxt.txt','examples/B/Puzzles/SudokuHexAsConstant_sptxt_saved.txt').


extra_testcase_file(305,'examples/B/Laws/REPL_UNIT_TESTS0.def').
extra_testcase_file(306,'examples/B/Laws/REPL_UNIT_TESTS1.def').
extra_testcase_file(307,'examples/B/Laws/REPL_UNIT_TESTS.def').
extra_testcase_file(465,'examples/B/Simple/UpCounter_saved10.P').
extra_testcase_file(472,'examples/B/SchneiderBook/Chapter12/TownsRR.ref').
extra_testcase_file(472,'examples/B/SchneiderBook/Chapter12/TownsR.ref').
extra_testcase_file(472,'examples/B/SchneiderBook/Chapter12/Towns.mch').
extra_testcase_file(475,'examples/B/Laws/REPL_UNIT_TESTS2.def').

% use
%    sicstus -l src/test_runner.pl --goal "run_last_test, halt ; halt."
% or
%    make runlast
% to just run last test case
% use
%    sicstus -l src/test_runner.pl --goal "id(ID), halt ; halt."
% to run a specific test
